/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package LuiCollection;

import static LuiCollection.LuiString.LTrim;
import static LuiCollection.LuiString.NVL;
import com.google.gson.Gson;
import java.util.LinkedHashMap;



public class LuiCollection extends LinkedHashMap<String, String> {
  private String DefaultValue = "";
  /**
   * Конструктор создаёт Map из строковой коллекции
   * @param StringCollection - коллекция как строка code:value;code:value;
   */
  public LuiCollection(String StringCollection) {
    super();
    if (StringCollection==null)StringCollection="";
    if (StringCollection.length() > 0) {
      String RepChar;
      int L = LuiString.InStr(StringCollection, ";", 0, -1);
      if (L < 0 || L+1 == StringCollection.length()) RepChar = "#$"; else {
        RepChar = StringCollection.substring(L+1);
        StringCollection = StringCollection.substring(0, L);
      }
      int State = 0;
      String Key = "";
      String Value = "";
      for (char c:StringCollection.toCharArray()) {
        switch (State) {
          case 0: // Парсим default value
            if (c == ';') {
              DefaultValue = Key.replace(RepChar, ":");
              Key = "";
              State = 1;
              continue;
            }
            if (c == ':') State = 1;
            else {
              Key += c;
              continue;
            }
          case 1: // Парсим код
            if (c == ':') {
              State = 2;
              Value = "";
              continue;
            }
            Key += c;
            continue;
          case 2: // Парсим значение
            if (c == ';') {
              Key=Key.replace(RepChar, ":");
              if (!Key.equals(" "))Key=LTrim(Key," ");
              Key=LTrim(Key,"\n");
              if (!Key.equals(" "))Key=LTrim(Key," ");
              put(Key.replace(RepChar, ":"), Value.replace(RepChar, ";"));
              State = 1;
              Key = "";
              continue;
            }
            Value += c;
        }
      }
      if (State == 0) DefaultValue = StringCollection.replace(RepChar, ":");
    }
  }
  /**
   * Добавление или изменение значения элемента по ключу
   * @param nKey ключ элемента
   * @param Value  значение элемента
   */
  public void Add_Element(String nKey, String Value) {
    put(nKey, Value);
  }
  /**
   * Удаление элемента
   * @param nKey ключ элемента
   */
  public void Erase_Element(String nKey) {
    remove(nKey);
  }
  /**
   * Определение наличия элемента
   * @param nKey
   * @return 
   */
  public boolean Exists_Element(String nKey) {
    return containsKey(nKey);
  }
  /**
   * Установка умолчания при ненахождении элемента коллекции
   * @param DefValue 
   */
  public void Set_Default(String DefValue) {
    DefaultValue = DefValue;
  }
  /**
   * Извлечение элемента коллекции
   * @param nKey
   * @return 
   */
  public String Get_Element(String nKey) {
    return getOrDefault(nKey, DefaultValue);
  }
  /**
   * Сляние двух коллекций. Т.е. в к существующей коллекции в объекте добаляется ещё одна
   * @param Add_Collection 
   */
  public void Add_Collection(LuiCollection Add_Collection) {
    for (String S : Add_Collection.keySet()) put(S, Add_Collection.get(S));
  }
  /**
   * Вычисление размера коллекции
   * @return 
   */
  public int Count_Element() {
    return size();
  }
  /**
   * Из Map в строкувую коллекцию
   * @param Mask
   * @return 
   */
  public String toString(String Mask) {
    String RetVal = DefaultValue.replace(";", "").replace(":", Mask.equals("")?"#$":Mask) + (DefaultValue.equals("")?"":";");
    for (String I:keySet())
      RetVal += I.replace(":", Mask.equals("")?"#$":Mask) + ":"+ NVL(Get_Element(I),"").replace(";", Mask.equals("")?"#$":Mask) + ";";
    return RetVal + Mask;
  }
  
  @Override
  public String toString() {
    return toString("");
  }
  public String toMultiLineText(){
    String RetVal = "";
    for (String key:keySet()){
        RetVal=RetVal+key+":"+Get_Element(key)+"\n";
    }
    return RetVal;
}
  public String ToJSON() {
    Gson g=new Gson();
    return g.toJson(this);
  }

}

