/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package LuiCollection;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class LuiString {
  public static final String TableNameDelimiter=" ,;.+*()\'\\/{}\n\r\f\t";
  public static final String[] SQLkeyWords  = new String[] {"select","insert","update","delete","from",
       "where","and","or","group","by","order","limit","offset","having","as","case",
        "when","then","else","end","type","left","right","join","on","outer","desc","asc","union","create","table","primary","key","if",
        "foreign","not","references","default","null","inner","cross","natural","database","drop","grant"};
  public static final String Delimiters = " ~`!@#$%^&*()+-=\\|<>?/:;\"\',{}[]№1234567890\r\n\f\t";
  public static final String Literal_Chars =
  "QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnmЙЦУКЕНГШЩЗХЪФЫВАПРОЛДЖЭЯЧСМИТЬБЮЁйцукенгшщзхъфывапролджэячсмитьбюё";
  public static final String Digit10_Chars = "0123456789";
  public static final String Digit16_Chars = "0123456789ABCDEF";
  
  public static String NVL(String pStr, String Def) {
      if (pStr==null)return Def;
      if (pStr.equals(""))return Def;
      return pStr;
    //return pStr==null?Def:pStr;
  }
  public static boolean is_not_null(String pStr){
    return !is_null(pStr);
  }
  public static boolean is_null(String pStr){
   if (pStr==null) return true;
   if (pStr.equals("")) return true;
   return false;
  }
  public static String NVL(String pStr) {
    return pStr==null?"":pStr;
  }
  
  /*public static boolean like(String str, String expr) {
      if (str==null) return false;
    return str.matches(NVL(expr, "").replaceAll("([\\[|\\]|\\(|\\)|\\{|\\}|\\.|\\*|\\+|\\?|\\$|\\^|\\||\\#|\\\\])", "\\\\$1").replace("_", ".").replace("%", ".*"));
  }*/
 public static boolean like(String str, String expr) {
    if (str==null) return false;
    return Pattern.compile(NVL(expr, "").replaceAll("([\\[|\\]|\\(|\\)|\\{|\\}|\\.|\\*|\\+|\\?|\\$|\\^|\\||\\#|\\\\])", "\\\\$1").replace("_", ".").replace("%", ".*"), Pattern.DOTALL).matcher(str).matches();
  }


  public static String Extract_Word(String Str, int N, String Delimiters) {
// Выделить N-ное по счёту слово от начала (или конца) строки Str (начиная с 1)
// Словом считается любой слитный набор символов, не входящий в строку разделителей Delimiters
    if (N==0) return "";
    ArrayList<String> S = Words(Str, Delimiters);
    if (java.lang.Math.abs(N)>S.size()) return "";
    if (N>0) return S.get(N-1); else return S.get(S.size()+N);
  }

  public static String Extract_Word(String Str, int N) {
    return Extract_Word(Str, N, " ");
  }

  
  public static ArrayList<String> Words(String Str, String Delimiters) {
// Возвращает все слова, найденные в строке
// Словом считается любой слитный набор символов, не входящий в строку разделителей Delimiters
// Для получения всех строк используйте конструкцию:
// for (String Iterator:Words(Str, Delimiters)) {Iterator...}
    ArrayList<String> S = new ArrayList<>();
    if (Str == null || Delimiters == null) return S;
    Normalize_String(Str, Delimiters, (str) -> {S.add(str);return str;});
    return S;
  }
  
  public static String Normalize_Words(String Str, String Delimiters, String NewDelimiter) {
// Возвращает строку, состоящую из слов строки Str, разделённые символами из строки Delimiters
// Слова в возвращаемой строке отделяются друг от друга строкой NewDelimiter
    String RetVal = "";
    for (String Iterator:Words(Str, Delimiters)) {
      RetVal = Add_If_Not_Null(RetVal, null, NewDelimiter)+Iterator;
    }
    return RetVal;
  }
  
  
  public static String Normalize_String(String Str, String Delimiters, LuiStringFunc NewWords, LuiStringFunc NewDelimiters) {
// Ищет в строке Str слова, разделённые символами из Delimiters.
// Со всеми словами производит обработку функцией NewWords
// Со всеми промежутками производит обработку функцией NewDelimiters
    String RetVal = "";
    if (Str == null) return "";
    int State = 0;
    int B = 0;
    char[] CharArray = Str.toCharArray();
    int i = 0;
    for (char c:CharArray) {
      if (State == 0) {
        if (Delimiters.indexOf(c) == -1) {
          if (i>B) RetVal += NewDelimiters.Func(Str.substring(B, i));
          B = i;
          State = 1;
        }
      } else {
        if (Delimiters.indexOf(c) > -1) {
          RetVal += NewWords.Func(Str.substring(B, i));
          B = i;
          State = 0;
        }
      }
      i++;
    }
    if (State == 0) {
      RetVal += NewDelimiters.Func(Str.substring(B));
    } else {
      RetVal += NewWords.Func(Str.substring(B));
    }
    return RetVal;
  }
  public static String RPad(String Str,char ch,int len){
      for(int i=Str.length();i<len;i++) Str=Str.concat(String.valueOf(ch));
      return Str;
  }
  public static String LPad(String Str,char ch,int len){
      for(int i=Str.length();i<len;i++) Str=String.valueOf(ch).concat(Str);
      return Str;
  }
  public static String Radd(String Str,char ch,int len){
      for(int i=0;i<len;i++) Str=Str.concat(String.valueOf(ch));
      return Str;
  }
  public static String Ladd(String Str,char ch,int len){
      for(int i=0;i<len;i++) Str=String.valueOf(ch).concat(Str);
      return Str;
  }
  public static String Normalize_String(String Str, String Delimiters, LuiStringFunc NewWords) {
    return Normalize_String(Str, Delimiters, NewWords, (S) -> {return S;});
  }
  
  public static String Add_If_Not_Null(String ToStr, String Pre, String Post) {
// Добавляет в начало и/или в конец строки текст, если строка не пустая
    if (ToStr == null) return "";
    if ("".equals(ToStr)) return "";
    return NVL(Pre) + ToStr + NVL(Post);
  }

  public static String Add_If_Not_Null(String ToStr, String Pre) {
    return Add_If_Not_Null(ToStr, Pre, null);
  }
  
  public static String Filter(String Str, String Valid_Symbols) {
// Оставляет в строке Str только те символы, которые содержатся в строке Valid_Symbols
    return Normalize_String(Str, Valid_Symbols, (S) -> {return "";}, (S) -> {return S;});
  }

  public static String FilterOff(String Str, String Invalid_Symbols) {
// Оставляет в строке Str только те символы, которые не содержатся в строке Valid_Symbols
    return Normalize_String(Str, Invalid_Symbols, (S) -> {return S;}, (S) -> {return "";});
  }
  
  public static String All_Words_UpCase(String Str) {
    return Normalize_String(Str, Literal_Chars, (S) -> {return S;}, (S) -> {return S.substring(0, 1).toUpperCase() + S.substring(1).toLowerCase();});
  }
  
  

  
  static String FreeFormat(String FormatStr, LuiCollection Elements, boolean IsExists) {
    int B;
    int E;
    String PStr;
    String TStr;
    if (NVL(FormatStr, "") != "") {
      for (;;) {
        B = -1;
        E = -1;
        PStr = FormatStr;
        for (;;) {
          TStr = PStr.replaceAll("\\[([^\\]\\[]*)\\]", "($1)");
          if (TStr.equals(PStr)) break;
          B = PStr.indexOf("[");
          E = PStr.indexOf("]", B);
          PStr = TStr;
        }
        if (B <0 || E < 0) break;
        PStr = FreeFormat(FormatStr.substring(B+1, E), Elements, false);
        if (PStr.length() > 0) IsExists = true;
        FormatStr = FormatStr.substring(0, B) + PStr + FormatStr.substring(E+1);
      }
      for (;;) {
        B = FormatStr.lastIndexOf("{");
        if (B == -1) break;
        E = FormatStr.indexOf("}", B);
        if (E == -1) break;
        PStr = Elements.Get_Element(FormatStr.substring(B+1, E));
        if (PStr.length() > 0) IsExists = true;
        FormatStr = FormatStr.substring(0, B) + PStr + FormatStr.substring(E+1);
      }
    }
    if (IsExists) return FormatStr; else return "";
  }
  
  public static String FreeFormat(String FormatStr, LuiCollection Elements) {
// Возвращает строку с данными из коллекции Elements по формату FormatStr
// Ссылки на элементы коллекции вставляются в форматную строку в виде {КодЭлемента}
// Форматная строка может содержать любое количество вложенных друг в друга
// подстрок в формате [текст]. Подстрока не будет присутсвовать в результате, если
// внутри неё не будет непустых значений элементов данных и вложенных подстрок.
// Пример:
// FormatStr: [{CityType}{City}, ][{StreetType}{Street}, ][д.{House}, [корп.{Corpus}, ][стр.{Building}, ]][кв.{Flat}[ ({Floor} эт.)]]
// Elements = City:Москва;CityType:гор.;Street:Вернадского;StreetType:пр-т.;House:15;Flat:50;Floor:2;
    return FreeFormat(FormatStr, Elements, true);
  }
  
  public static int InStr(String Str, String SubStr, int Offset, int N) {
// Возвращает смещение от начала строки Str N-го вхождения в неё подстроки (с начала или конца), если считать со смещения Offset
    if (N == 0) return -1;
    int RetVal;
    if (N>0) {
      RetVal = Offset-1;
      for (; N > 0 ; N--) {
        RetVal = Str.indexOf(SubStr, RetVal+1);
        if (RetVal < 0) return -1;
      }
    } else {
      RetVal = Str.length();
      for (; N < 0; N++) {
        RetVal = Str.substring(0, RetVal).lastIndexOf(SubStr);
        if (RetVal < Offset) return -1;
      }
    }
    return RetVal;
  }

  public static int Pair(String Pattern, String OpenStr, String CloseStr, int OffsetInPattern) {
// Поиск парной скобки (группы символов) в строке Pattern для группы символов OpenStr,
// расположенных после пропуска OffsetInPattern первых символов
// возвращается смещение парной ей группы символов CloseStr от начала Pattern
// Пример:
// Pair("<DIV class=1>Hello<DIV class=2>World<DIV>!</DIV></DIV></DIV>", "<DIV", "</DIV>", 4)
// Вернёт 48 -------------------------------------------/\
    if (Pattern == null || OpenStr == null || CloseStr == null || OffsetInPattern < 0) return -1;
    int PO = Pattern.indexOf(OpenStr, OffsetInPattern);
    int EndPos;
    int i = 1;
    for (;;) {
      EndPos = InStr(Pattern, CloseStr, PO, i);
      if (EndPos == -1 || InStr(Pattern + OpenStr, OpenStr, PO, ++i) >= EndPos) return EndPos;
    }
  }
  
  public static int Proportional(String Str) {
// Возвращает длину строки Str пропорциональным шрифтом в размерности символов моноширинного шрифта 
    return (int)((
    (Str.length() - (Str = FilterOff(Str, "#$+0123456789<=>?FLTZ_abcdeghknopqsuvxz~§«ёє»ГЗКТУХабвдеийлнопрсцчъэя")).length()) * 4 +
    (Str.length() - (Str = FilterOff(Str, "QERYUOPASGHKXVBNwЁЄАБВДЕИЙЛНОПРСФЦЧЬЭЯжмыю")).length()) * 5 +
    (Str.length() - (Str = FilterOff(Str, "!\"'(),.:;I[\\/]`fijlt|‚„‘’“”¦ ")).length()) * 2 +
    (Str.length() - (Str = FilterOff(Str, "*-J^ry{}­гзктухь")).length()) * 3 +
    (Str.length() - (Str = FilterOff(Str, "%CDMm©®МЪфшщ")).length()) * 6 +
    (Str.length() - (Str = FilterOff(Str, "@W…ЖШЩЮЫ№")).length()) * 7 +
     Str.length() * 4
    ) / 4.0 + 0.999999);
  }
  
  public static int proportional(String Str) {
    int RetVal = 0;
    for (char c:Str.toCharArray()) {
      if ("#$+0123456789<=>?FLTZ_abcdeghknopqsuvxz~§«ёє»ГЗКТУХабвдеийлнопрсцчъэя".indexOf(c) > -1) RetVal += 4; else
      if ("QERYUOPASGHKXVBNwЁЄАБВДЕИЙЛНОПРСФЦЧЬЭЯжмыю".indexOf(c) > -1) RetVal += 5; else 
      if ("!\"'(),.:;I[\\/]`fijlt|‚„‘’“”¦ ".indexOf(c) > -1) RetVal += 2; else
      if ("*-J^ry{}­гзктухь".indexOf(c) > -1) RetVal += 3; else
      if ("%CDMm©®МЪфшщ".indexOf(c) > -1) RetVal += 6; else
      if ("@W…ЖШЩЮЫ№".indexOf(c) > -1) RetVal += 7; else
      RetVal += 4;
    }
    return (int)(RetVal/4.0 + 0.999999999) ;
  }
  
  public static String RTrim(String Str, String Tail) {
// Откусывает с конца строки Str все символы из Tail
    if (Str == null) return "";
    if (Tail == null) return Str;
    for (int i = Str.length() - 1; i>=0; i--) if (!Tail.contains(Str.substring(i, i + 1))) return Str.substring(0, i + 1);
    return "";
  }

  public static String RTrim(String Str) {
// Откусывает с конца строки Str все пробелы
    return RTrim(Str, " ");
  }
  
  public static String LTrim(String Str, String Tail) {
// Удаляет из начала строки Str все символы из Tail
    if (Str == null) return "";
    if (Tail == null) return Str;
    for (int i = 0; i<Str.length(); i++) if (!Tail.contains(Str.substring(i, i + 1))) return Str.substring(i);
    return "";
  }

  public static String LTrim(String Str) {
// Удаляет из начала строки Str все пробелы
    return LTrim(Str, " ");
  }

  public static String Trim(String Str) {
// Удаляет из строки Str все пробелы в начале и конце
    return LTrim(RTrim(Str));
  }
  public static String withCapitalLetter(String Str){
// Переводит строку в lowercase, а первый её символ в Upper  
  if (is_null(Str))return"";
   return (Str.substring(0,1).toUpperCase()+Str.substring(1).toLowerCase());
  }
  public static boolean compareString(String pStr1,String pStr2){
    if (pStr1== null && pStr2==null) return true;  
    if (pStr1== null && pStr2!=null) return false;  
    if (pStr1!= null && pStr2==null) return false;  
    return pStr1.equals(pStr2);
  }
//
   static int ConvertingFlags(String modifier) {
    return  Integer.signum(modifier.indexOf("i")+1)*Pattern.CASE_INSENSITIVE |
            Integer.signum(modifier.indexOf("n")+1)*Pattern.DOTALL |
            Integer.signum(modifier.indexOf("m")+1)*Pattern.MULTILINE |
            Integer.signum(modifier.indexOf("x")+1)*Pattern.COMMENTS;
  }

  /**
   * Определяет соответствие строки sample шаблону pattern
   * @param sample строка
   * @param pattern шаблон регулярных выражений
   * @param modifier Модификаторы - строка с буквами-флагами, следующими в любом порядке:
   * i - при наличии поиск не чувствителен к регистру букв; n - любой набор символов, обозначаемый точкой, включает переводы строк;
   * m - Знаки ^ и $, обозначающие начало и конец строки, будут относиться только к началу и концу всего фрагмента;
   * x - Игнорировать пробелы.
   * @return true, если строка соответствует шаблону и false, если нет.
   */
  public static boolean RegExp_Like(String sample, String pattern, String modifier) {
    Matcher M = Pattern.compile(pattern, ConvertingFlags(modifier)).matcher(sample);
    return M.matches();
  }

  public static boolean RegExp_Like(String sample, String pattern) {
    return sample.matches(pattern);
  }

  /**
   * Ищет подстроку в строке sample согласно шаблону pattern. 
   * This function complies with the POSIX regular expression standard
   * @param sample Строка, в которой ищется подстрока
   * @param pattern Шаблон регулярных выражений
   * @param position Смещение в sample, с которого производится поиск соответствия шаблону. По умолчанию - 1
   * @param occurrence Номер найденного соответствия. По умолчанию - 1
   * @param modifier Модификаторы - строка с буквами-флагами, следующими в любом порядке:
   * i - при наличии поиск не чувствителен к регистру букв; n - любой набор символов, обозначаемый точкой, включает переводы строк;
   * m - Знаки ^ и $, обозначающие начало и конец строки, будут относиться только к началу и концу всего фрагмента;
   * x - Игнорировать пробелы.
   * @param subexpression Номер возвращаемого фрагмента, заданного в pattern круглыми скобками
   * @return Найденный фрагмент в виде строки или null, если поиск неуспешный
   */
  public static String RegExp_SubStr(String sample, String pattern, int position, int occurrence, String modifier, int subexpression) {
    Matcher M = Pattern.compile(pattern, ConvertingFlags(modifier)).matcher(sample);
    if (!M.find(position-1)) return null;
    for (;--occurrence>0;) if (!M.find()) return null;
    try {return M.group(subexpression);} catch(Exception ex) {return null;}
  }

  public static String RegExp_SubStr(String sample, String pattern, int position, int occurrence, String modifier) {
    return RegExp_SubStr(sample, pattern, position, occurrence, modifier, 0);
  }

  public static String RegExp_SubStr(String sample, String pattern, int position, int occurrence) {
    return RegExp_SubStr(sample, pattern, position, occurrence, "", 0);
 }

  public static String RegExp_SubStr(String sample, String pattern) {
    try {
     Matcher M = Pattern.compile(pattern).matcher(sample);
     if (!M.find()) return null;
     return M.group();
    } catch(Exception ex) {return null;}
  }
  
  /**
   * Заменяет в строке sampleStr фрагменты по шаблону pattern на replaceStr
   * @param sampleStr изначальная строка
   * @param pattern шаблон искомого фрагмента
   * @param replaceStr строка замны
   * @param position начиная с позиции в строке (по умолчанию - 1)
   * @param occurrence какой фразмент по счёту. 0 - все фрагменты (по умолчанию - 0)
   * @param modifier Модификаторы - строка с буквами-флагами, следующими в любом порядке:
   * i - при наличии поиск не чувствителен к регистру букв; n - любой набор символов, обозначаемый точкой, включает переводы строк;
   * m - Знаки ^ и $, обозначающие начало и конец строки, будут относиться только к началу и концу всего фрагмента;
   * x - Игнорировать пробелы.
   * @return строку с заменами
   */
  public static String RegExp_Replace(String sampleStr, String pattern, String replaceStr, int position, int occurrence, String modifier ) {
//    if (occurrence==0) return sampleStr.substring(0, position-1) + RegExp_Replace(sampleStr.substring(position-1), pattern, replaceStr);
    Pattern P = Pattern.compile(pattern, ConvertingFlags(modifier));
    if (occurrence>1) {
      Matcher M = P.matcher(sampleStr);
      if (!M.find(position-1)) return sampleStr;
      for (int i=occurrence;--i>0;) if (!M.find()) return sampleStr;
      position = M.end();
    }
    if (occurrence==0) return sampleStr.substring(0, position-1) + P.matcher(sampleStr.substring(position-1)).replaceAll(replaceStr);
    return sampleStr.substring(0, position-1) + P.matcher(sampleStr.substring(position-1)).replaceFirst(replaceStr);
  }

  public static String RegExp_Replace(String sampleStr, String pattern, String replaceStr) {
    return sampleStr.replaceAll(pattern, replaceStr);
  }
    /**
   * Заменяет символы, вредные для XML на их обозначения в XML
   * @param sample Строка
   * @return Строка после замены
   */
  public static String MaskXMLChars(String sample) {
   if (sample==null) return null;
     return sample.replace("&", "&amp;").replace("\"", "&quot;").replace("<", "&lt;").replace(">", "&gt;").replace("\t", "&#9;").replace("\n", "&#10;").replaceAll("[\\x00-\\x08\\x0B\\x0C\\x0E-\\x1F]", "\u3098");
  }

  /**
   * Заменяет служебные непечатные символы из начала диапазона на символ 'эквадратик'
   * @param sample Строка
   * @return Строка после замены
   */
  public static String MaskAncillaryChars(String sample) {
   if (sample==null) return null;
      return sample.replaceAll("[\\x00-\\x08\\x0B\\x0C\\x0E-\\x1F]", "\u3098");
  }
  /**
   * Преобразование строки к целому числу
     * @param pString -- исходная строка
     * @param def -- значение если преобразование не получилось
     
   */
  public static Integer String2Integer(String pString,Integer def){
     try{
     return (Integer.valueOf(pString));
     }
     catch (NumberFormatException ex){
     return def;        
             }
  }
 public static String RemoveIco(String pString) {
   int i=pString.indexOf("[ico:");
   if (i>=0){
     int j=pString.indexOf("]", i);
     if (j>0){
         String icohint=pString.substring(i, j+1);
         pString=pString.replace(icohint,"");
     }
   }
   return pString;  
 }
 public static String getIco(String pString) {
   int i=pString.indexOf("[ico:");
   if (i>=0){
     int j=pString.indexOf("]", i);
     if (j>0){
         return pString.substring(i+1, j);
         
     }
   }
   return "";  
 }
public static String GetIcoFromSquareBracket(String pString) {
   int i=pString.indexOf("[ico:");
   if (i>=0){
     int j=pString.indexOf("]", i);
     if (j>0){
         return pString.substring(i+5, j);
     }
   }
   return "";  
 }
public static String GetTableAndAliasFromCursor(String pString,int cursor){
   String retVal="";
   String alias="";
   char ch;
   int ri;
   if (is_null(pString)) return pString;
   int len=pString.length();
   if (cursor<0 || cursor>=len) return pString;
   // двигаемся в лево
   for(ri=cursor;ri>=0;ri--){
    ch=pString.charAt(ri);
    if (TableNameDelimiter.indexOf(ch)>-1 && ch!='.') break; // точку слева разрешим это может быть имя схемы
    retVal=String.valueOf(ch)+retVal;
   }
   // двигаемся в право
   for(ri=cursor+1;ri<len;ri++){
       ch=pString.charAt(ri);
    if (TableNameDelimiter.indexOf(ch)>-1) break;
    retVal=retVal+String.valueOf(ch);
   }
   //Проверим retVal на ключевое слово
   for (int k=0;k<SQLkeyWords.length;k++){
     if (SQLkeyWords[k].equals(retVal.toLowerCase())) {
         return "";
     }
   }
   // продолжаем двигаться вправо до непробела
   for(;ri<len;ri++){
       ch=pString.charAt(ri);
    if (ch!=' ' && ch!='\n' && ch!='\t' && ch!='\r' && ch!='\f') break;
   }
   // двигаемся вправо накапливая алиас
   for(;ri<len;ri++){
       ch=pString.charAt(ri);
       if (TableNameDelimiter.indexOf(ch)>-1) break;
       alias=alias+String.valueOf(ch);
   }
   if (is_not_null(alias)){
   for (int k=0;k<SQLkeyWords.length;k++){
     if (SQLkeyWords[k].equals(alias.toLowerCase())) {
         alias="";
         break;
     }
   }
   } 
   if (is_not_null(alias)){
   return retVal+" "+alias;
   }
   return retVal;
}
 }
