/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package WikiToRichText;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.RichTextString;
import org.apache.poi.ss.usermodel.ShapeTypes;
import org.apache.poi.ss.usermodel.VerticalAlignment;
import org.apache.poi.xssf.usermodel.TextAlign;
import org.apache.poi.xssf.usermodel.XSSFClientAnchor;
import org.apache.poi.xssf.usermodel.XSSFDrawing;

import org.apache.poi.xssf.usermodel.XSSFRichTextString;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFSimpleShape;
import org.apache.poi.xssf.usermodel.XSSFTextParagraph;





/*

+---------------------------------+

  +----------------------------+
  +--+   +-----+     +---------+
                

+-+--+---+-----+-----+---------+--+

*/


/**
 *
 * @author vsenchenko
 */
public class WikiToExcell
{
    Workbook wb;
    Sheet sheet;
    java.awt.Color fc;

    /*
    private enum FmtType {
        F_BOLD(1),
        F_ITALIC(2),
        F_UNDERLINE(4),
        F_SIZE_PLUS(8),
        F_SIZE_MINUS(16)        
    };
    */
    
    private class MarkItem
    {
        String type;
        int start;
        int end;
        short volume;
        Font font;
    };
    
    ArrayList <MarkItem> markList = new ArrayList<>();
    
    HashMap <Integer, MarkItem> markMap = new HashMap<>();
    
    public class Paragraph
    {
        XSSFRichTextString rts;
        //String align;
        TextAlign align;
        //boolean mono = false;
        
        /*
        Paragraph()
        {
            align = "LEFT";
        }
        */
    };

    
    ArrayList<Paragraph> ps = new ArrayList<>();

    private String matches(String s, String p)
    {
        Pattern pattern = Pattern.compile(p);

        Matcher matcher = pattern.matcher(s);

        if( !matcher.find() )
            return "";

        return matcher.group(1);
    }
    
    public WikiToExcell(Sheet sheet, String str, boolean paragraphsMode, java.awt.Color fc)
    {
        if(str.equals("")) return;

        this.sheet = sheet;
        this.wb = sheet.getWorkbook();
        this.fc = fc;
        
        String lines[] = str.split("\\r?\\n");
        
        //String lines[];
        
        /*
        if( paragraphsMode )
            lines = str.split("\\r?\\n");
        else
        {
            lines = new String[1];
            lines[0] = str;
        }
        */

        for(int i = 0; i < lines.length; i++)
        {
            if( lines[i].matches("^----$") )
                lines[i] = lines[i].replaceAll("----", "^^^^");
        }

        if( !paragraphsMode )
        {
            String a = "";
            for( String s : lines ) a = a.concat(s+"\n");
            lines = new String[1];
            lines[0] = a.substring(0, a.length()-1);
        }

        
        for(int i = 0; i < lines.length; i++)
        {
            boolean p_mono = false;
            int p_font_size = 0;
        //String  p_align = "LEFT";
            TextAlign p_align = TextAlign.LEFT;

            String s = lines[i];

            if( s.matches("^ .+$") )
            {
                s = s.substring(1);
                p_mono = true;
            }
            /*
            else
            if( s.matches("^----$") )
            {
                ;
            }
            */

            String v = matches(s, "^(--+).+$");
            
            if( !v.equals("") )
            {
                s = s.substring(v.length());
                p_font_size = - v.length() * 20;
            }           

            v = matches(s, "^(\\+\\++).+$");
            
            if( !v.equals("") )
            {
                s = s.substring(v.length());
                p_font_size = + v.length() * 20;
            }           

            
            
            //else
            if( s.matches("^=.+=$") )
            {
                s = s.substring(1, s.length()-1);
                //p_align = "CENTER";
                p_align = TextAlign.CENTER;
            }           
            else
            if( s.matches("^==.+$") )
            {
                s = s.substring(2);
                //p_align = "LEFT";
                p_align = TextAlign.LEFT;
            }
            else
            if( s.matches("^.+==$") )
            {
                s = s.substring(0, s.length()-2);
                //p_align = "RIGHT";
                p_align = TextAlign.RIGHT;
            }

            markList.clear();
            markMap.clear();

            //Paragraph p = new Paragraph();           
            //Paragraph p = make_paragraph(s);
            Paragraph p = make_paragraph(s, p_mono, p_font_size);
            
            p.align = p_align;
            
            //s = do_format_str(s);

            /*
            if(p_mono)
            {
                Font font = wb.createFont();
                font.setFontName("Courier New");
                p.rts.applyFont(font);
            }
            */
            
            ps.add(p);
        }
    }

    public void split_mark(MarkItem m1, int pos)
    {
        MarkItem m2 = new MarkItem();
        
        m2.type   = m1.type;
        m2.volume = m1.volume;
        m2.start  = pos;
        m2.end    = m1.end;
        
        markList.add(m2);
        //markList.add(0, m2);
        
        m1.end = pos;
    }
    
    public void printMarkList()
    {
//--        System.out.println("---------------");
        
        for(int i = 0; i < markList.size(); i++)
        {
            MarkItem m = markList.get(i);
//--            System.out.printf("markList(%d):  type=%s, [%d]  %d .. %d\n", i, m.type, m.volume, m.start, m.end);
        }
    }
    
    //public String do_format_str(String str)
    public Paragraph make_paragraph(String str, boolean mono, int font_size)
    {
        Paragraph p = new Paragraph();
        boolean flag = true;
        String s = str;
                
/*
        while(flag)
        {
            if( s.matches("/^ (.+)/") )
            {
                
            }
        
        }
*/

        /*
        Pattern pattern = Pattern.compile("'''|''|-+|\\++|~+");
       
        Matcher matcher = pattern.matcher(str);

        System.out.println("n = " + matcher.groupCount());

        
        while( matcher.find() )
        {
            System.out.println("----------");

            
            for( int i=0; i<=matcher.groupCount(); i++ )
            {
                //System.out.println(i + " => " + matcher.group(i));
                System.out.println("(" + i + ") " + matcher.group(i) + "   " + matcher.start(i) + "/" + matcher.end(i));
            }
            
            
            
        }
        */

        
        
        if( mono || font_size != 0 )
        {
            MarkItem item = new MarkItem();
            item.type  = "";
            item.start = 0;
            item.end   = -1;
            markList.add(item);        
        }
        
        
        
        
        Pattern pattern = Pattern.compile("(''')|('')|(-+)|(\\++)|(~~)");
        
        while(true)
        {
            Matcher matcher = pattern.matcher(s);
        
            if( matcher.find() )
            {
                //matcher.start(0) + "/" + matcher.end(0)


//--                System.out.println("==> " + matcher.group(0) + " --> " + matcher.start(0));

                //System.out.println("s.indexOf(matcher.group(0)));            
                
                //s = s.replaceFirst(matcher.group(0), "");
                
                /*
                String a = "";
                if(matcher.start(0) > 0)
                    a = s.substring(0, matcher.start(0));
                
                if(matcher.end(0) < s.length())
                        a = a + s.substring(matcher.end(0))

                s = a;
                */


                MarkItem item = new MarkItem();
                
                item.type  = matcher.group(0);
                item.start = matcher.start(0);
                item.end   = -1;
                
                
                //String delExpr = "";
                
                if( matcher.group(1) != null )
                {
                    // '''
                    s = s.replaceFirst("'''", "");
                }
                else if( matcher.group(2) != null )
                {
                    // ''
                    s = s.replaceFirst("''", "");
                }
                else if( matcher.group(3) != null )
                {
                    // -+
                    item.volume = (short)item.type.length();
                    item.type = "--";
                    s = s.replaceFirst("-+", "");
                }
                else if( matcher.group(4) != null )
                {
                    // \++
                    item.volume = (short)item.type.length();
                    item.type = "++";
                    s = s.replaceFirst("\\++", "");
                }
                else if( matcher.group(5) != null )
                {
                    // ~+
                    s = s.replaceFirst("~~", "");
                }

                markList.add(item);
                
                continue;
            }

            break;
        }
        
        
//--        System.out.println("str = " + s);


        
        printMarkList();

//--        System.out.println("do merge pairs");
        
        //while(true)
        do
        {
            full_loop:

            for(int i = 0; i < markList.size(); i++)
            {
                MarkItem m1 = markList.get(i);

                if(m1.end == -1)
                {
                    for(int j = 0; j < markList.size(); j++)
                    {
                        if(i==j) continue;
                            
                        MarkItem m2 = markList.get(j);

                        if(m2.end != -1) continue;

                        if( (m1.type.equals("++") || m1.type.equals("--")) && (m2.type.equals("++") || m2.type.equals("--")) )
                        {
                            //m1.end = m2.start - 1;    //??
                            m1.end = m2.start;
                            continue full_loop;                        
                        }
                        
                        if(m1.type.equals(m2.type))
                        {
                            m1.end = m2.start;
                            markList.remove(j);
                            continue full_loop;
                        }
                    }
                    
                    if(m1.end == -1)
                    {
                        m1.end = s.length();
                    }
                }
            }
            //break;
        }
        while(false);
        
        printMarkList();
        
//--        System.out.println("do split marks");
        
        do
        {
            full_loop:
            
            for(int i = 0; i < markList.size(); i++)
            {
                MarkItem m1 = markList.get(i);

                for(int j = 0; j < markList.size(); j++)
                {
                    if(i!=j)
                    {
                        MarkItem m2 = markList.get(j);

                        if(m1.start < m2.start && m2.start < m1.end)
                        {
                            split_mark(m1, m2.start);
                            continue full_loop;
                        }
                    
                        if(m1.start < m2.end && m2.end < m1.end)
                        {
                            split_mark(m1, m2.end);
                            continue full_loop;
                        }
                    }
                }
            }
        }
        while(false);
        

        printMarkList();


//--        System.out.println("do make marks map");
        
        for(int i = 0; i < markList.size(); i++)
        {
            MarkItem m = markList.get(i);

            /*
            if(markMap.containsKey(m.start))
            {
                MarkItem m1 = markMap.get(m.start);
                m1.type += "," + m.type;
                markMap.put(m1.start, m1);
            }
            else
                markMap.put(m.start, m);
            */
            /*
            if(markMap.containsKey(m.start))
            {
                MarkItem m1 = markMap.get(m.start);
                m1.type += "," + m.type;
                //markMap.put(m1.start, m1);
            }
            else
                markMap.put(m.start, m);
            */
            
            MarkItem m1;
            
            if(markMap.containsKey(m.start))
            {
                m1 = markMap.get(m.start);
                
                // for test
                m1.type += "," + m.type;
            }
            else
            {
                m1 = m;
                m1.font = wb.createFont();

                if(mono) m1.font.setFontName("Courier New");
                if(font_size != 0) m1.font.setFontHeight( (short)(m1.font.getFontHeight() + font_size));
            }

            
            switch(m.type)
            {
                case "''":
                    m1.font.setItalic(true);
                break;
            
                case "'''":
                    m1.font.setBold(true);
                break;
                
                case "~~":
                    m1.font.setUnderline(Font.U_SINGLE);
                break;

                case "++":
                    //System.out.println("FontHeight = " + m1.font.getFontHeight());
                    m1.font.setFontHeight( (short)(m1.font.getFontHeight() + 20 * m1.volume) );
                break;

                case "--":
                    //System.out.println("FontHeight = " + m1.font.getFontHeight());
                    m1.font.setFontHeight( (short)(m1.font.getFontHeight() - 20 * m1.volume) );
                break;
            }
            
            
            // будет срабатывать вхолостую для повторов, ну да ладно...
            markMap.put(m1.start, m1);
        }

        p.rts = new XSSFRichTextString(s.replaceAll("\\^\\^\\^\\^", "----"));
        
        for(Integer k: markMap.keySet())
        {
            MarkItem m = markMap.get(k);
            
//--            System.out.printf("%d .. %d: %s\n", m.start, m.end, m.type);

            p.rts.applyFont(m.start, m.end, m.font);
        }
        

       
        //return s;
        return p;
    }
    
    
    
    public void makeDrawing(int x, int y, int dx, int dy)
    {
        XSSFDrawing drawing = ((XSSFSheet)sheet).createDrawingPatriarch();
        
        XSSFClientAnchor anchor = drawing.createAnchor(0, 0, 0, 0, x, y, x+dx, y+dy);
        XSSFSimpleShape shape = drawing.createSimpleShape(anchor);
        
        shape.setShapeType(ShapeTypes.RECT);
        //java.awt.Color fc = ColorFactory.valueOf(head_color);
        
        //shape.setFillColor(211, 211, 211);
        shape.setFillColor(fc.getRed(), fc.getGreen(), fc.getBlue());

        //shape.getTextParagraphs().clear();
        shape.clearText();
        
        shape.setVerticalAlignment(VerticalAlignment.CENTER);
        
        
        
        //Source source = new Source(angular);
        
        for( int i = 0; i < ps.size(); i++ )
        {
            Paragraph p = ps.get(i);
        
            XSSFTextParagraph tp = shape.addNewTextParagraph(p.rts);
            
            tp.setTextAlign(p.align);
            
            /*
            switch(p.align)
            {
                case "LEFT":
                    tp.setTextAlign(TextAlign.LEFT);
                break;
                case "RIGHT":
                    tp.setTextAlign(TextAlign.RIGHT);
                break;
                case "CENTER":
                    tp.setTextAlign(TextAlign.CENTER);
                break;
            
            }
            */
            
        }
    }
    
    public RichTextString getSingle()
    {
        return ps.get(0).rts;
    }
    
}
