/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

    /* 
     *   (С) 2018  ООО ФОРС-Центр Разработки, ООО ФОРС-Телеком
     *   http://www.fors.ru
     *   http://www.forstelecom.ru
     *
     *   Класс доступа к данным в СУБД Postgres
     *   
     *   РЕВИЗИЯ:
     *  $Rev:: 1            $:  Ревизия файла
     *  $Author:: achibuk   $:  Автор последних изменений
     *  $Date:: 2018-06-26 1#$: Дата последних изменений
     *   ПРИМЕЧАНИЯ:
    ****************************************************************************** */
package LuiDb;

import LuiCollection.LuiCollection;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbCur;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.sql.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;



public abstract class LuiDbConnect implements LuiDbConn {
    PreparedStatement pst;
    FileInputStream fin;
    FileOutputStream fos;
    InputStream dbin;
    Reader rea;
    byte[] buf = new byte[4096];
    
    protected Connection conn;
    protected String dbURL;
    protected String username;
    protected String password;
    protected String schema;
    protected String dbVersion;
    protected boolean autoCommit;
    private boolean readOnly;
    protected long lastActivity;
    //protected Statement stmt;
    protected HashSet<Statement> mStatements = new HashSet<>();
    public String sessionInfo;
    protected String locale;
    protected String sid;
    protected final ArrayList<String> alString= new ArrayList<>();
    protected final HashMap<Integer,LuiDbCur> openedCursors = new HashMap<>();
    protected int rowCount;
    protected int cursorCount=0;

    /**
     * Конструктор
     */
    public LuiDbConnect(){
        conn=null;
        //stmt=null;
        lastActivity = System.currentTimeMillis();
}

    
    /**
     * Соединение с БД
     * @param pDataBaseURL URL для доступа к БД например "jdbc:oracle:thin:@fst1:1521:fst" или ""jdbc:postgresql://localhost:5432/test"
     * @param pUsername имя пользователя
     * @param pPassword пароль
     * @param pAutocommit режим автокоммита
     * @param pReadonly режим только чтения
     * @throws SQLException
     */
    @Override
    public void connect(String pDataBaseURL, String pUsername, String pPassword, String pSchema, boolean pAutocommit, boolean pReadonly) throws SQLException {
        if (conn!=null) {
            try {
                if (!conn.isClosed()) disconnect();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        autoCommit = pAutocommit;
        readOnly = pReadonly;
        dbURL = pDataBaseURL;
        password = pPassword;
        username = pUsername;
        schema = pSchema;
        lastActivity = System.currentTimeMillis();

        conn = DriverManager.getConnection(pDataBaseURL, pUsername, pPassword);
        conn.setAutoCommit(pAutocommit);
        conn.setReadOnly(pReadonly);
        
        //stmt = conn.createStatement();
        
        setSessionInfo("LUI");
        dbVersion = getDbVersion();
    }

    /**
     * Соединение с БД
     * @param pDataBaseURL URL для доступа к БД например "jdbc:oracle:thin:@fst1:1521:fst" или ""jdbc:postgresql://localhost:5432/test"
     * @param pUsername имя пользователя
     * @param pPassword пароль
     * @param pSchema схема
     * @throws SQLException
     */
    @Override
    public void connect(String pDataBaseURL, String pUsername, String pPassword, String pSchema) throws SQLException {
        connect(pDataBaseURL, pUsername, pPassword, pSchema, false, false);
    }

    
    /**
     * Отсоединение от БД
     * @throws SQLException
     */
    @Override
    public void disconnect() throws SQLException {
        
        for(Statement aSt: this.mStatements) {
            if(!aSt.isClosed()) { aSt.close(); }
        }
        mStatements.clear();
                
        //stmt //if (stmt!=null) {if (!stmt.isClosed()) stmt.close();}
        CloseAllCursors();        
        if (conn!=null) {if (!conn.isClosed()) conn.close();}
    }
    
    
    /**
     * Определение версии базы данных, к которой выполняется/выполнен соединение
     * @return версия БД
     * @throws SQLException
     */
    public abstract String getDbVersion() throws SQLException;

    /**
     * Выполнение DDL - операций
     * @param pStmt оператор SQL
     * @throws SQLException
     */
    @Override
    public void executeDDL(String pStmt) throws SQLException {
        lastActivity = System.currentTimeMillis();        
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            this.mStatements.add(aSt);
            aSt.executeUpdate(pStmt);
            //stmt //int i = stmt.executeUpdate(pStmt);
        } catch(SQLException e) {
            if (!autoCommit) conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                this.mStatements.remove(aSt);
                aSt.close();
            }
        }
    }

    /**
     * Выполнение Insert Update DElete
     * @param pStmt оператор SQL
     * @return количество изменённых/удалённых/вставленных строк
     * @throws SQLException
     */
    @Override
    public int executeDML(String pStmt) throws SQLException {
        lastActivity = System.currentTimeMillis();
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            this.mStatements.add(aSt);
            return aSt.executeUpdate(pStmt);
            //stmt //int i = stmt.executeUpdate(pStmt);
            //return i;
        } catch(SQLException e) {
            if (!autoCommit) conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
    }

    @Override
    public abstract ArrayList<String> executeIfNoLocks(String pStmt) throws SQLException;

    /**
     * Выполняет генерацию одного или более операторов для попытки 
     * блокировки данных извлекаемых оператором select. 
     * Обычно это делается добавлением фразы for update of ...
     * @param pStmt -- оператор select пример: select cl.id,cl.fullname from cl_t_client cl
     * @param uk -- список столбцов являющихся членами уникального ключа строк извлекаемых запросом
     * @return список строк - операторов
     * @throws SQLException 
    **/
    @Override
    public abstract ArrayList<String> select4update(String pStmt, ArrayList<ColInfo> uk) throws SQLException;

    @Override
    public abstract String getSelectResult(String pStmt, int pRows) throws SQLException;

    @Override
    public String getSelectResult(String pStmt) throws SQLException {
        return getSelectResult(pStmt, 1);
    }

    @Override
    public ArrayList<String> getSelectFirstRowValList(String pStmt) throws SQLException {
        ArrayList<String> res;
        lastActivity = System.currentTimeMillis();
        res = null;
        rowCount=0;
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            this.mStatements.add(aSt);
            try( ResultSet rs = aSt.executeQuery(pStmt))
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt))
            {
                while (rs.next()) {
                    rowCount++;
                    res = new ArrayList<String>();
                        for (int i=1; i<rs.getMetaData().getColumnCount()+1; i++){
                            res.add(rs.getString(i));
                        }
                    return res;
                }   
                return res;
            }
        } catch(SQLException e) {
            if (!autoCommit)conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
    }

    @Override
    public ArrayList<ArrayList<String>> getSelectAllRowsValList(String pStmt) throws SQLException {
        ArrayList<ArrayList<String>> res = null;
        ArrayList<String> row = null;
        rowCount = 0;
        lastActivity = System.currentTimeMillis();
        
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            this.mStatements.add(aSt);
        
            try( ResultSet rs = aSt.executeQuery(pStmt)) {
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt)) {
                while (rs.next()) {
                    rowCount++;
                    if (res==null) res = new ArrayList<>();
                    row = new ArrayList<>();
                    for (int i=1; i<rs.getMetaData().getColumnCount()+1; i++) {
                        row.add(rs.getString(i));
                    }
                    res.add(row);
                }
                return res;
            }
        } catch(SQLException e) {
            if (!autoCommit)conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
    }

    @Override
    public ArrayList<LuiCollection> getSelectAllRowsValMap(String pStmt) throws SQLException {
        ArrayList<LuiCollection> res = null;
        LuiCollection row = null;
        rowCount = 0;
        lastActivity = System.currentTimeMillis();
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            this.mStatements.add(aSt);
            try(ResultSet rs = aSt.executeQuery(pStmt)) {
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt)) {
                while(rs.next()) {
                    rowCount++;
                    if (res==null) res = new ArrayList<>();
                    row = new LuiCollection("");
                    for (int i=1; i<rs.getMetaData().getColumnCount()+1; i++) {
                        row.Add_Element(rs.getMetaData().getColumnLabel(i), rs.getString(i));
                    }
                    res.add(row);
                } 
                return res;
            }
        } catch(SQLException e) {
            if (!autoCommit) { conn.rollback(); }
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
        
    }

    @Override
    public LuiCollection getSelectFirstRowValMap(String pStmt) throws SQLException {
        LuiCollection res = null;
        rowCount = 0;
        lastActivity = System.currentTimeMillis();
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            mStatements.add(aSt);
        
            try( ResultSet rs = aSt.executeQuery(pStmt)) {
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt)) {
                while (rs.next()) {
                    rowCount++;  
                    res = new LuiCollection("");
                    for (int i=1; i<rs.getMetaData().getColumnCount()+1; i++) {
                        res.Add_Element(rs.getMetaData().getColumnLabel(i), rs.getString(i));
                    }
                    return res;     
                }   
                return res;
            }
        } catch(SQLException e) {
            if (!conn.getAutoCommit())
            if (!autoCommit)conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                mStatements.remove(aSt);
            }
        }
    }

    @Override
    public LuiCollection getSelectAllRowsTwoColsValMap(String pStmt) throws SQLException {
        LuiCollection res = null;
        rowCount = 0;
        lastActivity = System.currentTimeMillis();
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            mStatements.add(aSt);
            try( ResultSet rs = aSt.executeQuery(pStmt)) {
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt)) {
                while (rs.next()) {
                    rowCount++;
                    if (res==null) {res = new LuiCollection("");}
                    res.Add_Element(rs.getString(1), rs.getString(2));
                }
                return res;
            }
        } catch(SQLException e) {
            if (!autoCommit)conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
    }
    
    protected void selectAndIgnore(String pSQL) throws SQLException
    {
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            mStatements.add(aSt);
            try(ResultSet aRs = aSt.executeQuery(pSQL)) { }
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
    }

    protected int executeUpdate(String pStmt) throws SQLException
    {
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            mStatements.add(aSt);
            return aSt.executeUpdate(pStmt);
        }
        finally {
            if(aSt != null) {
                aSt.close();
                this.mStatements.remove(aSt);
            }
        }
    }
    
    /**
     * Сommit
     * @throws SQLException
    */
    @Override
    public void commit() throws SQLException {
        lastActivity = System.currentTimeMillis();
        conn.commit();
    }

    /**
     * Rollback
     * @throws SQLException
    */
    @Override
    public void rollback() throws SQLException {
        lastActivity = System.currentTimeMillis();
        if (!autoCommit)conn.rollback();
    }

    /**
     * Savepoint
     * @param pPointName
     * @throws SQLException
    */
    @Override
    public void savepoint(String pPointName) throws SQLException {
        lastActivity = System.currentTimeMillis();
        executeDDL("savepoint " + pPointName);
    }

    /**
     * Rollback to Savepoint
     * @param pPointName
     * @throws SQLException
    */
    @Override
    public abstract boolean rollbackToSavepoint(String pPointName) throws SQLException;

    @Override
    public abstract boolean isTransaction() throws SQLException;

    /**
     * Возвращает объект JDBC Connection
     * @return JDBC Connection
    */
    @Override
    public Connection getConnection() {
        return conn;
    }

    @Override
    public abstract int CursorOpen(String pStmt) throws SQLException;

    @Override
    public void CursorClose(int pCursID) throws SQLException {
        lastActivity = System.currentTimeMillis();
        try{
            openedCursors.get(pCursID).close();
            openedCursors.remove(pCursID);
        } catch(SQLException e) {
            if (!autoCommit)conn.rollback();
            throw e;
        }
    }

    @Override
    public boolean CursorNextRow(int pCursID) throws SQLException {
        lastActivity = System.currentTimeMillis();
        try{    
            return openedCursors.get(pCursID).nextRow();
        } catch(SQLException e) {
            if (!autoCommit)conn.rollback();
            throw e;
        }
    }

    @Override
    public boolean CursorisOpen(int pCursID) {
        lastActivity = System.currentTimeMillis();
        if (openedCursors.containsKey(pCursID)){
            return openedCursors.get(pCursID).isOpen();
        }
        return false;
    }

    @Override
    public int CursorGetCols(int pCursID) {
        lastActivity = System.currentTimeMillis();
        return openedCursors.get(pCursID).getCols();
    }

    @Override
    public ArrayList<ColInfo> CursorGetColsInfo(int pCursID) {
        lastActivity = System.currentTimeMillis();
        return openedCursors.get(pCursID).getColsInfo();
    }

    @Override
    public ArrayList<String> CursorGetRow(int pCursID) {
        lastActivity = System.currentTimeMillis();
        return openedCursors.get(pCursID).getRow();
    }
    
    @Override
    public int CursorOpenedCount() {
        return openedCursors.size();
    }

    /**
     * Закрытие всех открытых курсоров
     * @throws SQLException
     */
    @Override
    public void CloseAllCursors(boolean pDebug) throws SQLException {
        
        if(openedCursors.isEmpty()) { return; }
        
        if(pDebug) { 
            System.out.printf("\n===== UNCLOSED CURSORS: %d =====\n", openedCursors.size()); 
        }
        
       
        ArrayList<LuiDbCur> aCursors = new ArrayList<>(openedCursors.values());
        openedCursors.clear();
        for(LuiDbCur aCur: aCursors) {
            if(pDebug) { System.out.printf("\t[%s]\n", aCur.getSQLText()); }
            aCur.close();
        }
        
        /*
        for(Map.Entry<Integer, LuiDbCur> item : openedCursors.entrySet()){
            item.getValue().close();
        }
        openedCursors.clear();
        */
    }

    @Override
    public long getLastActivity() {
        return lastActivity;
    }

    @Override
    public void setSessionInfo(String pInfo) throws SQLException {
        sessionInfo = pInfo;
    }

    @Override
    public String getSessionInfo() {
        return sessionInfo;
    }


   // @Override
   // public abstract String StringToStringByFormat(String pString, String pFormat, String pLuiDataType) throws SQLException;

    @Override
    public int getRowCount() {
        return rowCount;
    }

    @Override
    public abstract String NextVal(String pSequence) throws SQLException;

    @Override
    public abstract String setOffsetClause(String query,String offsetColAlias,String hint,int rows_fetched);
    @Override
    public abstract String getOffsetColumn(String alias);

//    /**
//     * Передать в сервер БД переменную
//     * @param varCode - код переменной
//     * @param varValue - значение переменной
//     * @throws SQLException
//    */
//@Override
//    public void varToServer(String varCode, String varValue) throws SQLException {
//     getSelectResult("select pgv_set_text('UI_VARS','"+varCode+"','"+varValue+"')");
//    }

//    /**
//     * Забрать переменную с сервера БД
//     * @param varCode - код переменной
//     * @return
//     * @throws java.sql.SQLException
//     */
//@Override
//    public String varFromServer(String varCode) throws SQLException {
//     return (getSelectResult("select pgv_get_text('UI_VARS','"+varCode+"')"));
//    }

    
    
    
   /**
 * Установка для соединения NLS-настроек
 * @param pLocale 
 */
public void setLocale(String pLocale) {
    this.locale=pLocale;
}

    /**
     * Выделение основной фразы pClause (SELECT, FROM, WHERE, ORDER BY, GROUP BY, CONNECT BY) из текста sql-запроса
     * @param pQuery - оператор select
     * @param pClauseType - тип основной фразы в запросе
     * @return - фраза заданного типа из основного запроса, содержащегося в pStmt
     */
    public String getQueryMainClause(String pQuery, String pClauseType) {
        // Основная фраза pClause определяется как фраза pClause, которая не заключена в круглые скобки.
        // На самом деле, достаточно найти фразу pClause, слева от которой нет незакрытой скобки.
        String query = pQuery.toUpperCase();
        pClauseType=pClauseType.toUpperCase();
        query=query.replace("\n"+pClauseType, " "+pClauseType);
        query=query.replace(pClauseType+"\n", pClauseType+" ");
        query=query.replace(pClauseType+"\r", pClauseType+" ");
        query=query.replace(pClauseType+"(", pClauseType+" (");
        query=query.replace(")"+pClauseType,") "+pClauseType);
               
        String before;
        String clause = "";
        String from, where, connectby, groupby, orderby;
        from = where = connectby = groupby = orderby = "";
        int idx = query.length();
        int idx123, idx125; // 123 - открывающая скобка, 125 - закрывающая скобка
        
        
        
        while (true) {
            idx = query.lastIndexOf(pClauseType+" ", idx-1);
            if (idx < 0) break;
            if (idx == 0) {
                clause = pQuery;
                break;
            }
            before = query.substring(0, idx);
            while (true) {
                idx123 = before.lastIndexOf("(", before.length()-1);
                if (idx123 < 0) break;
                idx125 = before.indexOf(")", idx123+1);
                if (idx125 < 0) break;
                before = before.replace(before.substring(idx123, idx125+1), "");
            }
            if (idx123 < 0) {
                clause = pQuery.substring(idx).trim();
                break;
            }
        }
        // Здесь break нарочно не используется, чтобы, например, для варианта "SELECT" выполнились присвоения значений всем перечисленным переменным.
        switch (pClauseType) {
            case "SELECT":
                from = getQueryMainClause(pQuery, "FROM");
            case "FROM":
                where = getQueryMainClause(pQuery, "WHERE");
            case "WHERE":    
                connectby = getQueryMainClause(pQuery, "CONNECT BY");
            case "CONNECT BY":    
                groupby = getQueryMainClause(pQuery, "GROUP BY");
            case "GROUP BY":    
                orderby = getQueryMainClause(pQuery, "ORDER BY");
        }
        if (!from.isEmpty() && clause.contains(from)) clause = clause.substring(0, clause.indexOf(from));
        if (!where.isEmpty() && clause.contains(where)) clause = clause.substring(0, clause.indexOf(where));
        if (!connectby.isEmpty() && clause.contains(connectby)) clause = clause.substring(0, clause.indexOf(connectby));
        if (!groupby.isEmpty() && clause.contains(groupby)) clause = clause.substring(0, clause.indexOf(groupby));
        if (!orderby.isEmpty() && clause.contains(orderby)) clause = clause.substring(0, clause.indexOf(orderby));
        return clause;
    }
/**
 * 
 * @param pStmt -- оператор SQL (Insert или Update) с одной единственной bind-переменной заданной в виде знака вопроса
 * нпример:INSERT INTO table_images(data) VALUES(?)
 * @param pFile -- -- строка с путём и именем файла
 
 **/
    @Override
    public void saveFileToBlob(String pStmt,String pFile) throws SQLException,IOException{
    pst=null;
    fin=null;
    try {
    pst=conn.prepareStatement(pStmt);
    fin = new FileInputStream(pFile);
    
    pst.setBinaryStream(1,fin);
    pst.execute();
    //pst.executeQuery();
    } catch (SQLException | FileNotFoundException ex){
      throw ex;  
    }
    finally {
    if (fin!=null) fin.close();
    if (pst!=null) pst.close();
    }
}
    
 /** извлекает blob в файл
 * @param pStmt -- оператор Select извлекающий одну строку (рассмотрена будет только первая) и один столбец (blob(bytea)).
     * @param pFile -- строка с путём и именем файла
 
     * @throws java.sql.SQLException
     * @throws java.io.IOException
 *
 */
    @Override
    public void readBlobToFile(String pStmt,String pFile) throws SQLException,IOException {
    pst=null;
    fos=null;
    ResultSet rs=null;
    int size;
    try {
    pst=conn.prepareStatement(pStmt);
    fos = new FileOutputStream(pFile);
    rs=pst.executeQuery();
    if (rs.next()) {
    dbin=rs.getBinaryStream(1);
    if (dbin!=null){
    while (true){
      size=dbin.read(buf);
      if (size<0) break;
      fos.write(buf, 0, size);
    }
    }
    }
    } catch (SQLException | IOException ex){
      throw ex;  
    }
    finally {
    if (fos!=null) fos.close();
    if (rs!=null && !rs.isClosed())rs.close();
    if (pst!=null) pst.close();
    }
}   

/**
 * Сохраняет файл в clob
 * @param pStmt -- оператор SQL (Insert или Update) с одной единственной bind-переменной заданной в виде знака вопроса
 * нпример:INSERT INTO table_images(data) VALUES(?)
 * @param pFile -- -- строка с путём и именем файла
 
 **/
    @Override
    public void saveFileToClob(String pStmt,String pFile) throws SQLException,IOException{
    pst=null;
    rea=null;
    try {
    pst=conn.prepareStatement(pStmt);
    
    //rea = new FileReader(pFile);
    //pst.setCharacterStream(1, rea);
    
    
        fin = new FileInputStream(pFile);
        //pst.sStream(1, fin);
    pst.setBinaryStream(1, fin);
    //pst.executeQuery();
    pst.execute();
    } catch (FileNotFoundException ex){
      throw ex;  
    }catch (SQLException ex){
      throw ex;  
    }
    finally {
    //if (rea!=null) rea.close();
    if (fin!=null) fin.close();
    if (pst!=null) pst.close();
    }
}
    
 /** извлекает blob в файл
 * @param pStmt -- оператор Select извлекающий одну строку (рассмотрена будет только первая) и один столбец (blob(bytea)).
     * @param pFile -- строка с путём и именем файла
 
     * @throws java.sql.SQLException
     * @throws java.io.IOException
 *
 */
    @Override
    public void readClobToFile(String pStmt,String pFile) throws SQLException,IOException {
    pst=null;
    fos=null;
    int size;
    try {
    pst=conn.prepareStatement(pStmt);
    fos = new FileOutputStream(pFile);
    ResultSet rs=pst.executeQuery();
    if (rs.next()) {
    dbin=rs.getBinaryStream(1);
    while (true){
      size=dbin.read(buf);
      if (size<0) break;
      fos.write(buf, 0, size);
    }
    }
    } catch (SQLException | IOException ex){
      throw ex;  
    }
    finally {
    if (fos!=null) fos.close();
    if (pst!=null) pst.close();
    }

    }   

/* возвращает имя схемы */
public String getSchema(){
    return this.schema;
}
}
