/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

 package LuiDb;

import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.like;
import java.sql.SQLException;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;
import LuiDbMyS.LuiDbConnMyS;
import LuiDbPG.LuiDbConnPG;
import LuiDbOra.LuiDbConnOra;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;

/**
 *
 * @author aliubushkin
 */
public class LuiDbConnectPool implements LuiDbConnPool {
    private final LinkedList<LuiDbConn> freeList = new LinkedList<>();
    private final HashSet<LuiDbConn> busyList = new HashSet<>();
    private final String dbURL;
    private final String username;
    private final String password;
    private final String schema;
    private final boolean autoCommit;
    private final boolean readOnly;
    private final String poolInfo;
    private final String database;
    public int appCnt=0;
    private final AtomicBoolean mShutdown = new AtomicBoolean(false);
    
        
    public LuiDbConnectPool(String pDataBaseURL, String pUsername, String pPassword, String pSchema, boolean pAutocommit, boolean pReadonly, String pPoolInfo) {
        /*freeList = new ArrayList<>();
        busyList = new ArrayList<>();*/
        dbURL = pDataBaseURL;
        username = pUsername;
        password = pPassword;
        schema = pSchema;
        autoCommit = pAutocommit;
        readOnly = pReadonly;
        poolInfo = pPoolInfo;
        
        
        String aDatabase = "PG";
        if (like(dbURL.toLowerCase(),"%postgresql%")) {
            aDatabase = "PG"; 
        }
        if (like(dbURL.toLowerCase(),"%ora%")) {
            aDatabase = "ORA";
        }
        if (like(dbURL.toLowerCase(),"%mysql%")) {
            aDatabase = "MYS";
        }
        database = aDatabase;
        
        Maintainer.add(this);
    }
    
    @Override
    /**
     * Получить из пула свободное соединение с указанными параметрами
     * (если такого нет - создать свободное соединение)
     * соединение перестаёт быть свободным
     * @param pLocale -- региональные настройки
     * @param pLuiLang -- текущий язык приложения
     * @param pAppCode -- код приложения
     * @param pLuiSessionID -- id записи в БД о сеансе в LUI
     * @param pAppUser - имя пользователя приложения
     * @throws SQLException
     */
    public  LuiDbConn GetFree(String pLocale,String pLuiLang,String pAppCode, String pLuiSessionID,String pAppUser) throws SQLException{
    
        if(mShutdown.get()) {
            throw new SQLException("Connection pool in shutdown state");
        }
    
        LuiDbConn c;

        
        // Цикл перебора свободных соединений
        while(true) {
            synchronized(freeList) {
                if(freeList.isEmpty()) { break; }
                c = freeList.removeFirst();
            }
            
            try { // проверяем, хорошее ли соединение
                if(c.getConnection().isValid(3)) { 
                    // Если хорошее - его и возвращаем
                    synchronized(busyList) {
                        busyList.add(c); // Добавляем в набор занятых
                    }
                    
                    if(pLocale != null) {
                        c.setLocale(pLocale);
                    }
                    String info="";
                        /*
                        if (is_not_null(pAppCode)){
                        info=info+"LUI App:"+pAppCode+";";   
                        }
                        if (is_not_null(pLuiLang)){
                        info=info+"Lang:"+pLuiLang+";";   
                        }
                        if (is_not_null(pLuiSessionID)){
                        info=info+"SesID:"+pLuiSessionID+";";   
                        }
                        if (is_not_null(pAppUser)){
                        info=info+"AppUsr:"+pAppUser+";";   
                        }
                        info=info+"Pool:"+poolInfo+";";
                        */
//                              1        2                    3                      4                        5                    6
                    info="LUI;"+NVL(pAppCode,"")+";"+NVL(pLuiLang,"")+";"+NVL(pAppUser,"")+";"+NVL(pLuiSessionID,"")+";"+NVL(poolInfo,"");
                    c.setSessionInfo(info);
                    return c; 
                }
            } catch (SQLException ignore) { }
            
            // Соединение плохое - убиваем
            try {
                synchronized(busyList) {  // уберём соединение из списка занятых!
                    busyList.remove(c);
                }
                c.disconnect();
            }
            catch(Exception ignore) {}
        }
        
        
        
        
        switch (database) {
            case "PG": c = new LuiDbConnPG(); break;
            case "ORA": c = new LuiDbConnOra(); break;
            case "MYS": c = new LuiDbConnMyS(); break;
            default: return null;
        }
        c.connect(dbURL, username, password, schema, autoCommit, readOnly);
        String info="";
                        /*if (is_not_null(pAppCode)){
                        info=info+"LUI App:"+pAppCode+";";   
                        }
                        if (is_not_null(pLuiLang)){
                        info=info+"Lang:"+pLuiLang+";";   
                        }
                        if (is_not_null(pLuiSessionID)){
                        info=info+"SesID:"+pLuiSessionID+";";   
                        }
                        info=info+"Pool:"+poolInfo+";";*/
//                              1        2                    3                      4                        5                    6                        
                        info="LUI;"+NVL(pAppCode,"")+";"+NVL(pLuiLang,"")+";"+NVL(pAppUser,"")+";"+NVL(pLuiSessionID,"")+";"+NVL(poolInfo,"");
        c.setSessionInfo(info);
        if(pLocale != null) { 
            c.setLocale(pLocale);
        }
        synchronized(busyList) {
            busyList.add(c);
            return c;
        }
    }

    
    private static final AtomicLong stConnectionCleanerCounter = new AtomicLong(0);
    private static final Executor stConnectionCleanerExecutor = Executors.newCachedThreadPool((Runnable r) -> {
        Thread aT = new Thread(r);
        aT.setName("Connection Cleaner #"+ stConnectionCleanerCounter.incrementAndGet() );
        aT.setDaemon(true);
        return aT;
    });
   
        
    /**
    * Сделать соединение свободным
    * @param pDbConnect - соединение
    * @throws java.sql.SQLException
    */
    @Override
    public void SetFree(final LuiDbConn pDbConnect) throws SQLException {
        if (pDbConnect==null) { return; }
        boolean aOwn;
        
        synchronized( busyList) {
            aOwn = busyList.remove(pDbConnect);
        }
        
        if (aOwn) {
            
            stConnectionCleanerExecutor.execute(() -> 
                {
                    try {

                        pDbConnect.CloseAllCursors(true);
                        if (pDbConnect.getConnection().getAutoCommit() != true) {
                            pDbConnect.commit();
                        }

                        pDbConnect.setSessionInfo(poolInfo + " (free connection)");

                        if (mShutdown.get()) {
                            try {
                                pDbConnect.disconnect();
                            } catch (SQLException ignore) { }
                        } else {
                            synchronized (freeList) {
                                freeList.add(pDbConnect);
                            }
                        }
                    } catch (SQLException ignore) {}
                }
            );
        }
    }

    @Override
    public void DisconnectAll() {
        
        synchronized( freeList ) {
            freeList.forEach((LuiDbConn c) -> {
                try {
                    c.disconnect();
                } catch (SQLException ignore) {}
            });
            freeList.clear();
        }
        
        synchronized( busyList ) {
            busyList.forEach((LuiDbConn c) -> {
                try {
                    c.disconnect();
                } catch (SQLException ignore) {}
            });
            busyList.clear();
        }
        
        /*
        for (int i=0; i<freeList.size(); i++){
            freeList.get(i).disconnect();
        }
        for (int i=0; i<busyList.size(); i++){
            busyList.get(i).disconnect();
        }
        */
        freeList.clear();
        busyList.clear();
    }
 
 
    @Override
    public int GetFreeListSize() {
        synchronized(freeList) {
            return freeList.size();
        }
    }

    
@Override
    public String getURL() {
        return this.dbURL;
    }

    @Override
    public String getSchema() {
        return this.schema;
    }
@Override
    public String getUser() {
        return this.username;
    }
    /**
     * Процедура обслуживания пула.
     * Если свободных соединений больше 10, то закрываются все, кроме десяти.
     * Если свободных соединений больше трёх, то закрывается одно.
     */
    @Override
    public void maintenance() {
        
        // Цикл закрытия лишних соединений
       boolean killed=false; 
       int aSz;
       LuiDbConn aConn;
        while(true) {
            
            
            synchronized(freeList) {
                aSz = freeList.size();
                if(aSz < 3 ) { break; } //3
                aConn = freeList.removeFirst();
            }
            try {
                aConn.disconnect();
                killed=true;
            }
            catch(Throwable ignore) {}
            
            if(aSz < 10) { break; }//10
        }
        if (!killed && !poolInfo.equals("LUI MetaData Access")){ //если никого не убили и это не служебное соединение
            synchronized(freeList) {
                aSz = freeList.size();
                if (aSz>0) { // а свободные есть
                aConn=freeList.removeFirst();
                try {
                aConn.disconnect();
                killed=true;
            }
            catch(Throwable ignore) {}
                }
            }    
        }
        /*
        boolean killed=true;
        int freeSize;
        Long now = System.currentTimeMillis();
        while(killed){
            killed=false;
            synchronized(freeList) {
              freeSize=freeList.size();
              if (freeSize>0){
                if (killed == false) {   
                for (int i=0;i<freeSize;i++){
                    if ((now - freeList.get(i).getLastActivity())/ 1000 > 600){ //если минуту ни чем не заят
                        try {
                        freeList.get(i).disconnect();
                        }catch(Throwable ignore) {}
                     killed = true; 
                     freeList.remove(i);
                     break;
                    }
                }  
                    }    
              }
            }
        }*/
    }
    
    
    /** 
     * Переход пула в режим закрытия
     */
    @Override
    public void setShutdown() {
        mShutdown.set(true);
    }

    @Override
    public String getDBvendor() {
        return database;
    }
    
    public static void main(String[] pArgs)
    {
	System.out.println("Connect Poooool");
    }
//Экономия пулов ради
    @Override
    public boolean getAutoCommit() {
        return this.autoCommit;
    }

    @Override
    public String getPoolInfo() {
        return this.poolInfo;
    }

    @Override
    public void appCntIncrement() {
        synchronized(this){
        this.appCnt++;
        }
    }

    @Override
    public void appCntDEcrement() {
        synchronized(this){
        this.appCnt--;}
    }

    @Override
    public int getAppCnt() {
      return this.appCnt;
    }
}
