/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package LuiDbInterface;

import java.sql.SQLException;

/**
 *
 * @author aliubushkin
 */
public interface LuiDbConnPool extends NeedMaintenance {
    /**
     * В реализации этого интерфейса д.б. конструктор с параметрами:
     * @param pDataBaseURL
     * @param pUsername
     * @param pPassword
     * @param pSchema
     * @param pAutocommit
     * @param pReadonly
     * @param pPoolInfo
     * @throws SQLException
     Примерно так:  
    public LuiDbConnPool(String pDataBaseURL,String pUsername,String pPassword,String pSchema,boolean pAutocommit,boolean pReadonly,String pPoolInfo) throws SQLException;
    */
    /**
     * Получить из пула свободное соединение с указанными параметрами
     * (если такого нет - создать свободное соединение)
     * соединение перестаёт быть свободным
     * @param pLocale -- региональные настройки
     * @param pLuiLang -- текущий язык приложения
     * @param pAppCode -- код приложения
     * @param pLuiSessionID -- id записи в БД о сеансе в LUI
     * @param pAppUser -- имя пользователя приложения
     * @throws SQLException
     */
    public  LuiDbConn GetFree(String pLocale,String pLuiLang,String pAppCode, String pLuiSessionID, String pAppUser) throws SQLException;    
    
    public void SetFree(LuiDbConn c) throws SQLException;
    
    public void DisconnectAll() throws SQLException;
    
    //public LuiDbConn GetConnectionFromFreeList(int i);
    
    public int GetFreeListSize();
    
    //public void KillFreeConnection(int i);

    public String getSchema();    
    
    public String getUser();
    
    public String getURL();    
    
    public String getDBvendor();
    
    public boolean getAutoCommit();
    
    public String getPoolInfo();
    /**
     * Переход пула в режим гашения
     */
    public void setShutdown();
    
    public void appCntIncrement();
    
    public void appCntDEcrement();
    
    public int getAppCnt();
}
