/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package LuiDbInterface;

import java.sql.SQLException;
import java.util.ArrayList;


public interface LuiDbCur {
    
    public default String getSQLText() { return "unreleased"; }
    
    /**
    * открытие курсора
    * @throws java.sql.SQLException
    */
    public void open(String pStmt) throws SQLException;

    /**
    * закрытие курсора
    * @throws java.sql.SQLException
    */
    public void close() throws SQLException;

    /**
    * Извлечение очередной строки из курсора
    * @return true - строка извлечена false -- данных больше нет
    * @throws java.sql.SQLException
    */
    public boolean nextRow() throws SQLException;
  
    /**
    * Проверка открытости курсора
    * @return true - открыт, false - закрыт
    */
    public boolean isOpen();

    /**
    * Возвращает количество столбцов в строке курсора
    * @return количество столбцов в строке
    */
    public int getCols();
    
    /**
    * Получить описание столбцов запроса
    * @return список объектов ColInfo
    */
    public ArrayList<ColInfo> getColsInfo();

    /**
    * Получить текущую строку c данными курсора (последняя извлечённая строка запроса)
    * @return список объектов ColInfo
    */
    public ArrayList<String> getRow();
    
}
