/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

    /* 
     *   (С) 2018  ООО ФОРС-Центр Разработки, ООО ФОРС-Телеком
     *   http://www.fors.ru
     *   http://www.forstelecom.ru
     *
     *   Класс доступа к данным в СУБД Postgres
     *   
     *   РЕВИЗИЯ:
     *  $Rev:: 1            $:  Ревизия файла
     *  $Author:: achibuk   $:  Автор последних изменений
     *  $Date:: 2018-06-26 1#$: Дата последних изменений
     *   ПРИМЕЧАНИЯ:
    ****************************************************************************** */
package LuiDbOra;


import static LuiCollection.LuiString.*;
import LuiDb.LuiDbConnect;
//import LuiCollection.LuiCollection;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbCur;
import java.sql.*;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicLong;
import oracle.jdbc.OracleResultSet;
import oracle.jdbc.OracleResultSetMetaData;
import oracle.jdbc.OracleTypes;



public class LuiDbConnOra extends LuiDbConnect {
String offset="N";
    /**
     * Соединение с БД
     * @param pDataBaseURL URL для доступа к БД например "jdbc:oracle:thin:@fst1:1521:fst" или ""jdbc:postgresql://localhost:5432/test"
     * @param pUsername имя пользователя
     * @param pPassword пароль
     * @param pAutocommit режим автокоммита
     * @param pReadonly режим только чтения
     * @throws SQLException
     */
    @Override
    public void connect(String pDataBaseURL, String pUsername, String pPassword, String pSchema, boolean pAutocommit, boolean pReadonly) throws SQLException {
        super.connect(pDataBaseURL, pUsername, pPassword, pSchema, pAutocommit, pReadonly);
        sid = getSelectResult("select userenv('SID') from dual");
        if (schema!=null){
         if (!schema.equals("")){
            this.executeDDL("alter session set CURRENT_SCHEMA = "+schema);
            
         }
        }
        try{
        offset=getSelectResult("select 'Y' from dual offset 0 rows");
        //offset="N";
        } catch (SQLException ex){
            offset="N";
        }
    }

    /**
     * Определение версии базы данных, к которой выполняется/выполнен соединение
     * @return версия БД
     * @throws SQLException
     */
    public String getDbVersion() throws SQLException {
        //if (dbVersion == null) return getSelectResult("select version from v$instance");
        if (dbVersion == null) return getSelectResult("select banner||'; ' from v$version");
        return dbVersion;
    };
    
    
    @Override
    public int CursorOpen(String pStmt) throws SQLException {
        try{
            Cursor c = new Cursor();
            c.open(pStmt);
            cursorCount++;
            openedCursors.put(cursorCount,c);
            lastActivity = System.currentTimeMillis();
            return cursorCount;
        } catch(SQLException e) {
            if  (!conn.getAutoCommit()){
                conn.rollback();
            }
            throw e;
        }
    }
    
    /**
     * Выполняет генерацию одного или более операторов для попытки 
     * блокировки данных извлекаемых оператором select. 
     * Обычно это делается добавлением фразы for update of ...
     * @param pStmt -- оператор select пример: select cl.id,cl.fullname from cl_t_client cl
     * @param uk -- список столбцов являющихся членами уникального ключа строк извлекаемых запросом
     * @return список строк - операторов
     * @throws SQLException 
    **/
    @Override
    public ArrayList<String> select4update(String pStmt, ArrayList<ColInfo> uk) throws SQLException {
        // Для oracle так: select null from (pStmt) for update of col1, col2 wait 1
        // Для postgres так: select null from (pStmt) for update of table1, table2 nowait
        ArrayList<String> s4u = new ArrayList<>();
        String forUpd = "select null from (" + pStmt + ")";
        // фраза "where"
        for (int i=0; i<uk.size(); i++) {
            if (i == 0) {
                forUpd = forUpd + " where ";
            } else {
                forUpd = forUpd + " and ";
            }
            forUpd = forUpd + uk.get(i).name;
            forUpd = forUpd +uk.get(i).value;
            /*switch (NVL(uk.get(i).LuiDATATYPE,"CHAR")) {
                case "CHAR":
                    if (uk.get(i).value == null) {
                        forUpd = forUpd + " is null ";   
                    } else {
                        forUpd = forUpd + " = '" + uk.get(i).value + "'";
                    }
                    break;
                case "NUMBER":
                    if (uk.get(i).value == null) {
                        forUpd = forUpd + " is null ";   
                    } else {    
                        forUpd = forUpd + " = " + uk.get(i).value;
                    }
                    break;
                case "DATE":
                    if (is_not_null(uk.get(i).LuiFORMAT)) {
                        forUpd = forUpd + " = to_date('" + uk.get(i).value + "','" + uk.get(i).LuiFORMAT + "')";
                    } else {
                        forUpd = forUpd + " = to_date('" + uk.get(i).value + "')";
                    }    
                    break;
                case "DATETIME":
                    if (is_not_null(uk.get(i).LuiFORMAT)) {
                        forUpd = forUpd + " = to_timestamp('" + uk.get(i).value + "','" + uk.get(i).LuiFORMAT + "')";
                    } else {
                        forUpd = forUpd + " = to_timestamp('" + uk.get(i).value + "')";
                    }   
                    break;
            }*/
        }
        // фраза "for update of"
        forUpd = forUpd + " for update";
        for (int i=0; i<uk.size(); i++) {
            if (i == 0) {
                forUpd = forUpd + " of ";
            } else {
                forUpd = forUpd + ", ";
            }
            forUpd = forUpd + uk.get(i).name;
        }
        forUpd = forUpd + " wait 1";
        s4u.add(forUpd);
        return s4u;
    }
    
    /**
     * Выполнение оператора если он не вызовет зависание в блокировке
     * @param pStmt оператор SQL
     * @return список строк. В первой - либо "locked", либо "success". Во второй - либо имя блокирующего пользователя, либо количество изменённых строк
     * @throws SQLException
     */
    @Override
    public ArrayList<String> executeIfNoLocks(String pStmt) throws SQLException {
        // К сожалению, в oracle нет возможности установить время ожидания блокировки для чего-либо кроме select for update.
        // Поэтому оптимистичное выполнение pStmt в общем случае с установкой предельного времени ожидания блокировки, например, 0.5 сек, не получится.
        // Необходимо всегда на всякий случай сразу запускать тред поиска блокировок. И уже затем выполнять сам оператор.
        // Если будет обнаружна блокировка, то выполнение pStmt прервётся.
        lastActivity=System.currentTimeMillis();
        ThreadForFindLocks t4l;
        int cnt;
        
        try {
            t4l = new ThreadForFindLocks(); // запускаем тред поиска блокировок   
           
            try {
                if (like((LTrim(pStmt.toUpperCase())),"SELECT%")){
                    selectAndIgnore(pStmt);
                    //stmt //try( ResultSet aRs = stmt.executeQuery(pStmt) ) { } //++ Было странно
                    cnt=1;
                } else {
                    cnt= executeUpdate(pStmt); // stmt.executeUpdate(pStmt); // пытаемся выполнить оператор
                }
                // // выполнился таки! если тред жив ждём его завершения *//
                if (t4l.thrd.isAlive()) t4l.thrd.join(); 
                alString.clear(); // очищаем список возвращаемых строк, возможно испорченный тредом
                alString.add("success");
                alString.add(String.valueOf(cnt));
                return alString;
            } catch (SQLException e){ // попали на исключение
                if (e.getErrorCode()!=1013) { // если оно не наше (т.е. сообщение не о "query canceled")
                    throw e; // отдаём и уходим
                }
            } catch (InterruptedException e) {
                // ничего не делаем, если основной тред прервут                
            }
            // в эту точку попадаем, если оператор повисал в блокировке
            rollbackToSavepoint("UI4l"); // Откат испорченной транзакции
            return alString; // массив строк заполнен в треде поиска блокировок
        } catch(SQLException e) {
            if (!autoCommit){
                conn.rollback();
            }
            throw e;
        }
    }

    
    /**
     * Rollback to Savepoint
     * @param pPointName
     * @throws SQLException
    */
    @Override
    public boolean rollbackToSavepoint(String pPointName) throws SQLException {
        lastActivity = System.currentTimeMillis();
        try {    
            executeDDL("rollback to savepoint " + pPointName);
            return true;
        } catch (SQLException e) { // возникло исключение
            if (e.getErrorCode()!=1086){ // это не наше исключение?
                throw e; // не наше -- уходим отдавая исключение
            } else {
                return false;    
            }
        }
    }
    
    @Override
    public void setSessionInfo(String pInfo) throws SQLException {
        super.setSessionInfo(pInfo);
        executeDDL("begin dbms_application_info.set_module('"+pInfo+"',''); end;");
    }
    
    @Override
    public String NextVal(String pSequence) throws SQLException {
        return this.getSelectResult("select "+pSequence+".nextval from dual");
    }
@Override
    public String getOffsetColumn(String alias) {
      if ("Y".equals(offset))return "";  
      else {
          return "rownum "+alias;
      }
    }
    @Override
public String setOffsetClause(String query,String offsetColAlias,String hint,int rows_fetched){
    if ("Y".equals(offset)){
        return query+" offset " + String.valueOf(rows_fetched) + " rows";
    } else {
       if (is_not_null(hint)) 
        return "select /*+"+hint+"*/ * from ("+query+")q$offset where "+offsetColAlias+">"+String.valueOf(rows_fetched);
       else
        return "select * from ("+query+")q$offset where "+offsetColAlias+">"+String.valueOf(rows_fetched);    
    }
}

    /**
     * Преобразование данных из строки полученной в результате выполнения запроса в строку по указанному формату
     * Все данные в результате выполнения запроса преобразовываются  к строке по умолчанию
     * их необходимо преобразовать в нужный тип целевой БД, а потом опять в строку, но по заданному в параметре формату
     * @param pResultsetString -
     * @param pColInfo
     * @return
    */
/*    @Override
    public String fromRStoStringByFormat(String pStringFromResultSet, ColInfo pColInfo, String pFormat) throws SQLException {
        // Работаем с оракловыми типами date и timestamp одинаково, используя to_char и to_date (и "забывая", что есть функция to_timestamp)
        String dType = pColInfo.LuiDATATYPE.toLowerCase();
        if (!pColInfo.LuiDATATYPE.isEmpty()) {
            if (dType.equals("datetime")) dType = "date";
        } else {
            switch (pColInfo.columnType) {
                case OracleTypes.SMALLINT:
                case OracleTypes.INTEGER:
                case OracleTypes.BIGINT:
                case OracleTypes.DECIMAL:
                case OracleTypes.DOUBLE:
                case OracleTypes.FLOAT:
                case OracleTypes.REAL:
                case OracleTypes.NUMBER: dType = "number"; break;
                case OracleTypes.VARCHAR:
                case OracleTypes.CHAR:
                case OracleTypes.FIXED_CHAR:
                case OracleTypes.CLOB: dType = "char"; break;   
                case OracleTypes.DATE: dType = "date"; break;
                case OracleTypes.TIMESTAMP:
                case OracleTypes.TIMESTAMPTZ: dType = "date"; break;
//                case OracleTypes.TIMESTAMPTZ: dType = "timestamp"; break;
            }
        }
        return getSelectResult("select to_char(to_" + dType + "('" + pStringFromResultSet + "'),'" + pFormat + "') from dual");
    };

*/
    /**
     * Преобразование строки к числу/дате/времени... и обратно к строке по указанному формату
     * исходная строка может не соответствовать формату, но нужно попытаться её преобразовать
     * с использованием подмен , на .
     * @param pString - исходная строка
     * @param pFormat - формат
     * @param pLuiDataType - тип данных Lui:NUMBER DATE DATETIME
     * @return
     * @throws SQLException 
     */
/*    @Override
    public String StringToStringByFormat(String pString, String pFormat, String pLuiDataType) throws SQLException {
        // Работаем с оракловыми типами date и timestamp одинаково, используя to_char и to_date (и "забывая", что есть функция to_timestamp)
        switch (pLuiDataType) {   
            case "NUMBER":
                pString = pString.replace(" ", ""); // удаляем пробелы
                pString = pString.replace(".", ","); // заменяем точку на запятую
                if (is_not_null(pFormat)){
                    return getSelectResult("select to_char(to_number('" + pString + "'),'" + pFormat + "') from dual");
                } else {
                    return getSelectResult("select to_number('" + pString + "') from dual");    
                }
            case "DATE":
            case "DATETIME":
                if (is_not_null(pFormat)) {
                    return getSelectResult("select to_char(to_date('" + pString + "','" + pFormat + "'),'" + pFormat + "') from dual");
                } else {
                    return getSelectResult("select to_date('" + pString + "') from dual");        
                }
        }   
        return pString;     
    }
*/
    /**
     * Проверка наличия несохранённых изменений (транзакции)
     * @return boolean true - есть, false - нет
     * @throws SQLException
     */
    @Override
    public boolean isTransaction() throws SQLException {
        String r = getSelectResult("select 1 from v$transaction t, v$session s where t.addr = s.taddr and s.sid = " + sid + " and rownum = 1");
        if (r == null) return false;
        return (r.equals("1"));
    }

    @Override
    public String getSelectResult(String pStmt, int pRows) throws SQLException {
        StringBuilder res = null;
        String s = null;
        int i = 0;
        rowCount = 0;
        lastActivity = System.currentTimeMillis();
    
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            mStatements.add(aSt);
            
            try( ResultSet rs = aSt.executeQuery(pStmt)) {
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt)) {
                while (rs.next()) {
                    rowCount++;
                    if(res==null) res = new StringBuilder();
                    s = rs.getString(1);
                    if (!(s==null)){
                        res.append((String)s);
                    }
                    i++;
                    if (i>=pRows & pRows>0)break;
                }   
                
                if (res != null) return res.toString();
                return null;
            }
        } catch(SQLException e) {
            String cn=e.getClass().getCanonicalName();
            if (e.getErrorCode()==17128) return ""; // если это ошибка связанная с тем, что это не запрос а что-то иное, то просто вернём пустую строку
            if (e.getSQLState().equals("99999") && e.getErrorCode()==17003) return "";
            if (!autoCommit) conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                mStatements.remove(aSt);
            }
        }
    }

    
    /**
     * Формирование строки, содержащей обращение к данным в заданном столбце, для её использования в SQL-запросе
     * @param pColumn - код столбца
     * @param pLuiDataTypeSrc - исходный тип данных (LUI)
     * @param pLuiDataTypeTgt - целевой тип данных (LUI)
     * @param pLuiFormat - формат данных
     * @return 
     */
    @Override
    public String getColumnSQLClause (String pColumn, String pLuiDataTypeSrc, String pLuiDataTypeTgt, String pLuiFormat) {
        // Работаем с оракловыми типами date и timestamp одинаково, используя to_char и to_date (и "забывая", что есть функция to_timestamp)
        String s = null;
        switch (NVL(pLuiDataTypeSrc,"CHAR")) {
            case "CHAR":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        s = pColumn; 
                        break;
                    case "NUMBER": 
                        if (is_not_null(pLuiFormat)) s = "to_number(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = "to_number(" + pColumn + ")";
                        break;
                    case "DATE":
                    case "DATETIME":
                        if (is_not_null(pLuiFormat)) s = "to_date(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = "to_date(" + pColumn + ")";
                        break;
                }
                break;
            case "NUMBER":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        if (is_not_null(pLuiFormat)) s = "to_char(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = "to_char(" + pColumn + ")";
                        break;
                    case "NUMBER":
                    case "DATE":
                    case "DATETIME":
                        s = pColumn;
                        break;
                }
                break;
            case "DATE":
            case "DATETIME":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        if (is_not_null(pLuiFormat)) s = "to_char(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = "to_char(" + pColumn + ")";
                        break;
                    case "NUMBER": 
                    case "DATE":
                    case "DATETIME":
                        s = pColumn;
                        break;
                }
                break;
        }
        return s;
    }

    @Override
    public String getDBvendor() {
        return "ORA";
    }
    /**
     * возвращает select для извлечения данных о столбцах таблицы(view)
       в запросе д.б. столбцы:
       SCHEMA- имя схемы(владельца в Oracle)
       NAME- имя столбца
       DATATYPE- тип данных СУБД
       COMMENT- комментарий
       PK- признак вхождения в перв.ключь (Y/N)
       NULLABLE- признак НЕ обязательности данных (Y/N)
     * @param pTable - имя таблицы, возможно с именем схемы и даже псевдонимом 
     * пример: lui.lui_t_user usr
     * @return - запрос
     */

    @Override
public String getQuery4Column(String pTable){
String q="select tc.OWNER SCHEMA,\n" +
"tc.COLUMN_NAME NAME,\n" +
"tc.DATA_TYPE\n" +
",(select cmnt.COMMENTS from sys.all_col_comments cmnt where \n" +
" cmnt.OWNER=tc.OWNER and \n" +
" cmnt.TABLE_NAME=tc.TABLE_NAME and \n" +
" cmnt.COLUMN_NAME=tc.COLUMN_NAME) CMNT\n" +
",(select decode(count(*),0,'N','Y') from sys.all_cons_columns cc,\n" +
"              sys.all_constraints co\n" +
"              where cc.OWNER=co.OWNER and co.TABLE_NAME=cc.TABLE_NAME\n" +
"              and cc.CONSTRAINT_NAME=co.CONSTRAINT_NAME and co.CONSTRAINT_TYPE='P'\n" +
"              and cc.OWNER=tc.OWNER and cc.TABLE_NAME=tc.TABLE_NAME\n" +
"              and cc.COLUMN_NAME=tc.COLUMN_NAME\n" +
"            ) PK\n" +
",tc.NULLABLE from sys.all_tab_columns tc where tc.table_name = '<table>'\n" +
"and tc.OWNER='<owner>'\n" +
"order by 1,2";
String pOwner="";
if (pTable.indexOf(".")>-1){
 pOwner=Extract_Word(pTable,1,".");
 pTable=Extract_Word(pTable,2,".");
}
//if (is_null(pOwner)) { pOwner=this.schema;}
if (is_null(pOwner)) { pOwner="";q=q.replace("'<owner>'","sys_context('userenv','CURRENT_SCHEMA')");}
if (pTable.indexOf(" ")>-1){
    pTable=Extract_Word(pTable,1," ");
}
pTable=LTrim(pTable);
pTable=RTrim(pTable);
pTable=pTable.toUpperCase();
pOwner=LTrim(pOwner);
pOwner=RTrim(pOwner);
pOwner=pOwner.toUpperCase();
q=q.replace("<owner>", pOwner);
q=q.replace("<table>",pTable);
return q;
}
/**
     * возвращает select для извлечения данных о таблицах или view
       в запросе д.б. столбцы:
       SCHEMA- имя схемы(владельца в Oracle)
       TABLE_NAME- имя таблицы
       TABLE_TYPE- тип данных СУБД
       COMMENTS- комментарий
       
     */    @Override
    public String getQuery4Table() {
       return "select tc.OWNER SCHEMA,tc.TABLE_NAME , tc.TABLE_TYPE, tc.COMMENTS  from sys.all_tab_comments tc where tc.owner=sys_context('userenv','CURRENT_SCHEMA')";
    }

    /**
    * Вложенный класс - курсор
    */
    class Cursor implements LuiDbCur {
        
        private ArrayList<ColInfo> colList; //список столбцов
        private PreparedStatement prstmt; //Объект JDBC "Оператор"
        private OracleResultSet resSet; // Набор строк
        private OracleResultSetMetaData resSetMetaData; // JDBC описание столбцов
        private ArrayList<String> row;
        private int cols;
        private boolean is_executed;

        /**
        * Конструктор - инициализатор
        * 
        */
        Cursor() {
            colList=new ArrayList<ColInfo>();
            cols=0;
            row=new ArrayList<String>();
            prstmt=null;
            resSet=null;
            is_executed=false;
        }

        @Override
        public void open(String pStmt) throws SQLException {
            try {
                if (isOpen()){
                    close();
                }
            } catch (Exception ex){
            
            }
            colList.clear();
            row.clear();
            prstmt=conn.prepareStatement(pStmt);
            resSetMetaData=(OracleResultSetMetaData) prstmt.getMetaData();
            if (resSetMetaData==null){
                getSelectResult(pStmt,0);
            }
            if (resSetMetaData==null)return;
            cols=resSetMetaData.getColumnCount();
            colList=new ArrayList<>();
            for (int i = 1; i <= cols; i++) {
                ColInfo ci;
                ci = new ColInfo(resSetMetaData.getSchemaName(i), 
                                 resSetMetaData.getTableName(i), // oracle не даёт имя таблицы
                                 resSetMetaData.getTableName(i), // вот здесь собака oracle не даёт алиас имени таблицы и из-за этого невозможно правильно сочинить
                                                         // select ... from table1 t1,table2 t2... for update of t1,t2 
                                                         // И не откуда взять имена этих алиасов
                                                         // 
                                 resSetMetaData.getColumnName(i),
                                 resSetMetaData.getColumnLabel(i),
                                 resSetMetaData.getColumnClassName(i),
                                 resSetMetaData.getColumnTypeName(i),
                                 resSetMetaData.getColumnType(i));
                ci.LuiDATATYPE="CHAR";
                switch (ci.columnType) {
                    case OracleTypes.SMALLINT:
                    case OracleTypes.INTEGER:
                    case OracleTypes.BIGINT:
                    case OracleTypes.DECIMAL:
                    case OracleTypes.DOUBLE:
                    case OracleTypes.FLOAT:
                    case OracleTypes.REAL:
                    case OracleTypes.NUMBER: ci.LuiDATATYPE="NUMBER"; break;
                    case OracleTypes.VARCHAR:
                    case OracleTypes.CHAR:
                    case OracleTypes.FIXED_CHAR:
                    case OracleTypes.CLOB: ci.LuiDATATYPE="CHAR"; break;   
                    case OracleTypes.DATE: ci.LuiDATATYPE="DATETIME"; break;
                    case OracleTypes.TIMESTAMP:
                    case OracleTypes.TIMESTAMPTZ: ci.LuiDATATYPE="DATETIME"; break;
                }
                colList.add(ci);
            }
        }
    
        /**
        * закрытие курсора
        * @throws java.sql.SQLException
        */
        @Override
        public void close() throws SQLException {
            if (resSet!=null) {if (!resSet.isClosed()) resSet.close();}
            if (prstmt!=null) {if (!prstmt.isClosed()) prstmt.close();}
            is_executed=false;
        }

        /**
        * Извлечение очередной строки из курсора
        * @return true - строка извлечена false -- данных больше нет
        * @throws java.sql.SQLException
        */
        @Override
        public boolean nextRow() throws SQLException {
            if (!is_executed){
            resSet=(OracleResultSet) prstmt.executeQuery();
            is_executed=true;
        }
            if (resSet.next()){
                row.clear();
                for (int i=1;i<=cols;i++) {
                    row.add(resSet.getString(i));
                }
                return true;
            }
            return false;
        }
        
        /**
        * Проверка открытости курсора
        * @return true - открыт, false - закрыт
        */
        @Override
        public boolean isOpen() {
            return cols>0;
        }

        /**
        * Возвращает количество столбцов в строке курсора
        * @return количество столбцов в строке
        */
        @Override
        public int getCols() {
            return cols;
        }
        
        /**
        * Получить описание столбцов запроса
        * @return список объектов ColInfo
        */
        @Override
        public ArrayList<ColInfo> getColsInfo() {
            return colList;
        }

        /**
        * Получить текущую строку c данными курсора (последняя извлечённая строка запроса)
        * @return список объектов ColInfo
        */
        @Override
        public ArrayList<String> getRow() {
            return row;
        }
        
    }
    
    
    private static final AtomicLong stThreadNumber = new AtomicLong(0);
    
    /* Вложенный класс для обнюхивания другим тредом "блокировок текущего процесса другим сеансом" */    
    class ThreadForFindLocks implements Runnable{
        Thread thrd; // тред
        Connection conn4l; // соединение
        Statement stmt4l;
        ResultSet rs;
      
        /**
        * Конструктор объектов класса 
        * @throws java.sql.SQLException
        */
        public ThreadForFindLocks() throws SQLException{
            thrd = new Thread(this);
            thrd.setDaemon(true);
            thrd.setName(String.format("Oracle Lock Finder #%d", stThreadNumber.incrementAndGet()));
            this.thrd.start();
        }
    
        @Override
        public void run() {
            //  System.out.println("Thread running");
            try {
                Thread.sleep(500);
                try { 
                    conn4l = DriverManager.getConnection(dbURL, username, password);
                    stmt4l = conn4l.createStatement();
                    rs = stmt4l.executeQuery("select blocking_session from v$session where sid = " + sid + " and blocking_session_status = 'VALID'");
                    if (rs.next()) { // запрос вернул строку - блокировка найдена
                        String locking_sid = rs.getString(1); // извлекаем sid (session id) блокирующего процесса
                        if (locking_sid != null){ // если не пуст совсем
                            if (!locking_sid.isEmpty()) { // если не пуст
                                rs.close(); // закрываем запрос
                                rs = stmt4l.executeQuery("select username from v$session where sid = " + locking_sid); // ищем сеанс с таким sid
                                if (rs.next()) { // есть сеанс, извлекаем данные о пользователе
                                    String locking_usr = rs.getString(1);
                                    alString.clear();
                                    alString.add("locked");
                                    alString.add(locking_usr);
                                    
                                    for(Statement aSt: mStatements) {
                                        aSt.cancel();
                                    }
                                    mStatements.clear();
                                    //stmt.cancel(); // убиваем оператор основного треда, выполняющийся и заблокировавшийся
                                }
                            }
                        } 
                    }
                    stmt4l.close();
                    conn4l.close(); // 
                } catch (SQLException e) {
                    System.err.println(e);
                }
            } catch (InterruptedException e) {}
        }
    }

    
}


