/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

    /* 
     *   (С) 2016  ООО ФОРС-Центр Разработки, ООО ФОРС-Телеком
     *   http://www.fors.ru
     *   http://www.forstelecom.ru
     *
     *   Класс доступа к данным в СУБД Postgres
     *   
     *   РЕВИЗИЯ:
     *  $Rev:: 1            $:  Ревизия файла
     *  $Author:: aliubushkin   $:  Автор последних изменений
     *  $Date:: 2016-10-14 1#$:  Дата последних изменений
     *   ПРИМЕЧАНИЯ:
    ****************************************************************************** */
package LuiDbPG;

import static LuiCollection.LuiString.*;
import LuiDb.LuiDbConnect;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbCur;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
//import java.sql.*;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicLong;
import org.postgresql.jdbc.PgResultSet;
import org.postgresql.jdbc.PgResultSetMetaData;
//import org.postgresql.jdbc.PgConnection;
//import org.postgresql.largeobject.LargeObject;
//import org.postgresql.largeobject.LargeObjectManager;

public class LuiDbConnPG extends LuiDbConnect {
    
    // Блокиратор одновременного соединения с базой данных
    private static final Object PGCONNECTLOCK = new Object();
    
    /**
     * Соединение с БД
     * @param pDataBaseURL URL для доступа к БД например "jdbc:oracle:thin:@fst1:1521:fst" или ""jdbc:postgresql://localhost:5432/test"
     * @param pUsername имя пользователя
     * @param pPassword пароль
     * @param pAutocommit режим автокоммита
     * @param pReadonly режим только чтения
     * @throws SQLException
     */
    @Override
    public void connect(String pDataBaseURL, String pUsername, String pPassword,String pSchema,boolean pAutocommit, boolean pReadonly) throws SQLException {
//PgConnection pgc=(PgConnection) super.conn;        
//LargeObject lo;  -- это мысли про LargeObject
//LargeObjectManager lom;
//lom.
        // Есть существенное подозрение, что PgJDBC не может корректно устанавливать несколько соединений одновременно
        // Сериализуем, на всякий случай.
        synchronized(PGCONNECTLOCK) {
            super.connect(pDataBaseURL, pUsername, pPassword, pSchema, pAutocommit, pReadonly);
        }
        
        sid = getSelectResult("select pg_backend_pid()");
        if (schema!=null){
         if (!schema.equals("")){
            this.executeDDL("set session search_path="+schema+",\"$user\",public"); // !!! если не добавить public, то преобразование к String данных некоторых типов будет давать null если запрос выполняется к чужой таблице
                                                                                        // !!! а если не добавлять "$user" то к собственным даннм придётся обращаться с префиксом юзера
            if (!pAutocommit) //lal: если не автокоммит, то надо коммит а то после rollbak search_path вернётся в "$usr",public
            commit();
         }
        }
    }
    
    /**
     * Определение версии базы данных, к которой выполняется/выполнен соединение
     * @return версия БД
     * @throws SQLException
     */
    public String getDbVersion() throws SQLException {
        if (dbVersion == null) return getSelectResult("select replace(version(),'PostgreSQL','') vers");
        return dbVersion;
    };
    
    
    @Override
    public int CursorOpen(String pStmt) throws SQLException {
          try{
            Cursor c = new Cursor();
            c.open(pStmt);
            cursorCount++;
            openedCursors.put(cursorCount,c);
            lastActivity = System.currentTimeMillis();
            return cursorCount;
        } catch(SQLException e) {
            if  (!conn.getAutoCommit()){
                conn.rollback();
            }
            throw e;
        }
    }
    
    /**
     * Выполняет генерацию одного или более операторов для попытки 
     * блокировки данных извлекаемых оператором select. 
     * Обычно это делается добавлением фразы for update of ...
     * @param pStmt -- оператор select пример: select cl.id,cl.fullname from cl_t_client cl
     * @param uk -- список столбцов являющихся членами уникального ключа строк извлекаемых запросом
     * @return список строк - операторов
     * @throws SQLException 
    **/
    @Override
    public ArrayList<String> select4update(String pStmt, ArrayList<ColInfo> uk) throws SQLException {
        // Для oracle так: select null from (pStmt) for update of col1, col2 wait 1
        // Для postgres так: select null from (pStmt) for update of table1, table2 nowait
        ArrayList<String> s4u = new ArrayList<>();
        String forUpd = "";
        String from = "";
        String where = getQueryMainClause(pStmt, "WHERE");
        if (where.isEmpty()) from = getQueryMainClause(pStmt, "FROM");
        // фраза "where"
        for (int i=0; i<uk.size(); i++) {
            if (where.isEmpty() && i == 0) {
                forUpd = forUpd + " where ";
            } else {
                forUpd = forUpd + " and ";
            }
            if (uk.get(i).tableAlias.isEmpty()) forUpd = forUpd + NVL(uk.get(i).name,uk.get(i).alias);
            else forUpd = forUpd + uk.get(i).tableAlias + "." + NVL(uk.get(i).name,uk.get(i).alias);
            forUpd = forUpd +uk.get(i).value;
          /*  switch (NVL(uk.get(i).LuiDATATYPE,"CHAR")) {
                case "CHAR":
                    if (uk.get(i).value == null) {
                        forUpd = forUpd + " is null ";   
                    } else {
                        forUpd = forUpd + " = '" + uk.get(i).value + "'";
                    }
                    break;
                case "NUMBER":
                    if (uk.get(i).value == null) {
                        forUpd = forUpd + " is null ";   
                    } else {    
                        forUpd = forUpd + " = " + uk.get(i).value;
                    }
                    break;
                case "DATE":
                    if (is_not_null(uk.get(i).LuiFORMAT)) {
                        forUpd = forUpd + " = to_date('" + uk.get(i).value + "','" + uk.get(i).LuiFORMAT + "')";
                    } else {
                        forUpd = forUpd + " = date '" + uk.get(i).value + "'";
                    }    
                    break;
                case "DATETIME":
                    if (is_not_null(uk.get(i).LuiFORMAT)) {
                        forUpd = forUpd + " = to_timestamp('" + uk.get(i).value + "','" + uk.get(i).LuiFORMAT + "')";
                    } else {
                        forUpd = forUpd + " = timestamp '" + uk.get(i).value + "'";
                    }   
                    break;
            }*/
        }
        if (where.isEmpty()) {
            // добавляем после фразы FROM условия по значениям полей, входящих в уникальный ключ
            if (from.length()>1)
            forUpd = pStmt.replace(from, from + forUpd); 
        } else {
            // обрамляем содержимое фразы WHERE основного запроса круглыми скобками и добавляем после него условия по значениям полей, входящих в уникальный ключ
            forUpd = pStmt.replace(where, where.substring(0,6) + "(" + where.substring(6) + ")" + forUpd);
        }
        // фраза "for update of"
        ArrayList<String> ukTables = new ArrayList<>();
        boolean addTable;
        forUpd = forUpd + " for update";
        int j=0;
        for (int i=0; i<uk.size(); i++) {
            addTable = true;
            for (String s : ukTables) {
                if (s.equals(uk.get(i).tableAlias)) {
                    addTable = false;
                    break;
                }
            }
            if (!addTable || uk.get(i).tableAlias.isEmpty()) continue;
            if (j == 0) {
                forUpd = forUpd + " of ";
                j++;
            } else {
                forUpd = forUpd + ", ";
            }
            forUpd = forUpd + uk.get(i).tableAlias;
            ukTables.add(uk.get(i).tableAlias);
        }
//        forUpd = forUpd + " nowait";
if (ukTables.size()==1 && where.length()<2 && from.length()<2) {
    forUpd="select * from "+ukTables.get(0)+" "+forUpd;
}
        s4u.add(forUpd);
        return s4u;
    }
    
/**
     * Выполняет генерацию одного или более операторов для попытки 
     * блокировки данных извлекаемых оператором select. 
     * Обычно это делается добавлением фразы for update of ...
     * @param pStmt -- оператор select пример: select cl.id,cl.fullname from cl_t_client cl
     * @param uk -- список столбцов являющихся членами уникального ключа строк извлекаемых запросом
     * @return список строк - операторов
     * @throws SQLException 
**/
/*    
@Override
    public ArrayList<String> select4update(String pStmt, ArrayList<ColInfo> uk) throws SQLException {
        boolean isCol=false;
        boolean isNeedTab=true;
        class tableEntry {
          String tableName;
          ArrayList<ColInfo> keyCols;
        }
        ArrayList<tableEntry> tabs;
        tabs=new ArrayList<>();
        ArrayList<String> s4u=new ArrayList<>();
        for (int i=0;i<uk.size();i++) {
            isNeedTab=true;
            // такая таблица в списке есть?
           for (int j=0;j<tabs.size();j++) {
               if (tabs.get(j).tableName.equals(uk.get(i).tableName)) { // есть
                   // проверим есть ли в ней такой столбец
                   isCol=false;
                   for (int k=0;k<tabs.get(j).keyCols.size();k++){
                       if (uk.get(i).name.equals(tabs.get(j).keyCols.get(k).name)) {
                           //есть!
                           isCol=true;
                           break;
                       }                    
                   }
                   if (isCol==false) { // нет в ней такого столбца - значит добавляем и новой таблицы вводить не надо
                       tabs.get(j).keyCols.add(uk.get(i));
                       isNeedTab=false;
                       break;
                   }
               }
           }
           if (isNeedTab){
            tableEntry te=new tableEntry();   
            te.tableName=uk.get(i).tableName;
            te.keyCols=new ArrayList<>();
            te.keyCols.add(uk.get(i));
            tabs.add(te);
           }
        }
        String forUpd;
        for (int i=0;i<tabs.size();i++){
           forUpd="select null from "+tabs.get(i).tableName+" where ";
          for (int j=0;j<tabs.get(i).keyCols.size();j++) {
              if (j>0) {
                  forUpd=forUpd+" and ";
              }
              forUpd=forUpd+tabs.get(i).keyCols.get(j).name;
              switch (NVL(tabs.get(i).keyCols.get(j).LuiDATATYPE,"CHAR")){
                     case "CHAR":
                     if (tabs.get(i).keyCols.get(j).value==null){
                     forUpd=forUpd+" is null ";   
                     } else {
                     forUpd=forUpd+"='"+tabs.get(i).keyCols.get(j).value+"'";
                     }
                     break;
                     case "NUMBER":
                     if (tabs.get(i).keyCols.get(j).value==null){
                     forUpd=forUpd+" is null ";   
                     } else {    
                     forUpd=forUpd+"="+tabs.get(i).keyCols.get(j).value;
                     }
                     break;
                     case "DATE":
                       if (is_not_null(tabs.get(i).keyCols.get(j).LuiFORMAT)) {
                         forUpd=forUpd+"=to_date('"+tabs.get(i).keyCols.get(j).value+"','"+tabs.get(i).keyCols.get(j).LuiFORMAT+"')";
                       } else
                       {
                        forUpd=forUpd+"=date '"+tabs.get(i).keyCols.get(j).value+"'";
                       }    
                     break;
                     case "DATETIME":
                       if (is_not_null(tabs.get(i).keyCols.get(j).LuiFORMAT)) {
                         forUpd=forUpd+"=to_timestamp('"+tabs.get(i).keyCols.get(j).value+"','"+tabs.get(i).keyCols.get(j).LuiFORMAT+"')";
                       } else
                       {
                        forUpd=forUpd+"=timestamp '"+tabs.get(i).keyCols.get(j).value+"'";
                       }   
                     break;
              }
          }
          forUpd=forUpd+" for update";
          s4u.add(forUpd);
        }
      
      return s4u;
    }
*/    

    /**
     * Выполнение оператора если он не вызовет зависание в блокировке
     * @param pStmt оператор SQL
     * @return список строк. В первой - либо "locked", либо "success". Во второй - либо имя блокирующего пользователя, либо количество изменённых строк
     * @throws SQLException
     */
    @Override
    public ArrayList<String> executeIfNoLocks(String pStmt) throws SQLException {
        //во- первых попробуем оптимистично выполнить оператор установив предельное время ожидания блокировки 0.5 сек
        // читаем умолчание
        lastActivity=System.currentTimeMillis();
        ThreadForFindLocks t4l;
        String dflt;
        int cnt;
        
        try {
            boolean isSP=false;
            dflt = getSelectResult("select ''''||setting||unit||'''' from pg_settings where name='lock_timeout'");
            alString.clear(); // очищаем массив строк
            executeDDL("set local lock_timeout = 1000");
            try {
              try {
                  savepoint("UI4l");
                  isSP=true;
              }
              catch (SQLException spex){
                 if ("25P01".equals(spex.getSQLState())) {
                   // это значит что SP не установился из-за отсутствия транзакции  
                 }
                 else throw spex;
              }
                
                if (like((LTrim(pStmt.toUpperCase())),"SELECT%")) {
                    selectAndIgnore(pStmt);
                    //stmt //try( ResultSet aRs = stmt.executeQuery(pStmt) ) { } //++ Было странно
                    cnt=1;    
                } else {
                    cnt = executeUpdate(pStmt); //stmt.executeUpdate(pStmt);
                }
                //Ура - всё получилось
                alString.add("success"); // пишем успех
                alString.add(String.valueOf(cnt));// и количество изменённых строк
                executeDDL("set local lock_timeout="+dflt); // восстановим умолчание
                return alString;
            } catch  (SQLException e) { // возникло исключение
                if (!e.getSQLState().equals("55P03")) { // это не наше исключение?  (т.е. сообщение не о "lock_not_available")
                    if (isSP){
                    rollbackToSavepoint("UI4l"); // Откат испорченной транзакции
                    isSP=false;
                    }
                    executeDDL("set local lock_timeout="+dflt); // восстановим умолчание
                    throw e; // не наше -- уходим, отдавая исключение
                }
            }
            // В эту точку попадаем если оптимистичный сценарий не удался
            if (isSP){
            rollbackToSavepoint("UI4l"); // Откат испорченной транзакции
            isSP=false;
            }
            try {
                  savepoint("UI4l");// Новая точка возврата
                  isSP=true;
              }
              catch (SQLException spex){
                 if ("25P01".equals(spex.getSQLState())) {
                   // это значит что SP не установился из-за отсутствия транзакции  
                 }
                 else throw spex;
              }
            
            executeDDL("set local lock_timeout="+dflt); // восстановление умолчания
            t4l = new ThreadForFindLocks(); // запускаем тред поиска блокировок   
           
            try {
                if (like((LTrim(pStmt.toUpperCase())),"SELECT%")){
                    selectAndIgnore(pStmt);
                    //stmt //try( ResultSet aRs = stmt.executeQuery(pStmt) ) { } //++ Было странно
                    cnt=1;
                } else {
                    cnt= executeUpdate(pStmt); //stmt.executeUpdate(pStmt); // пытаемся выполнить оператор
                }
                // // выполнился таки! если тред жив ждём его завершения *//
                if (t4l.thrd.isAlive()) t4l.thrd.join(); 
                alString.clear(); // очищаем список возвращаемых строк, возможно испорченный тредом
                alString.add("success");
                alString.add(String.valueOf(cnt));
                return alString;
            } catch (SQLException e){ // попали на исключение
                if (!e.getSQLState().equals("57014")){ // если оно не наше (т.е. сообщение не о "query canceled")
                    throw e; // отдаём и уходим
                }
            } catch (InterruptedException e) {
                // ничего не делаем, если основной тред прервут                
            }
            // в эту точку попадаем, если оператор повисал в блокировке
            if (isSP) {
            rollbackToSavepoint("UI4l"); // Откат испорченной транзакции
            isSP=false;}
            return alString; // массив строк заполнен в треде поиска блокировок
        } catch(SQLException e) {
            if (!autoCommit){
                conn.rollback();
            }
            throw e;
        }
    }


    /**
     * Rollback to Savepoint
     * @param pPointName
     * @throws SQLException
    */
    @Override
    public boolean rollbackToSavepoint(String pPointName) throws SQLException {
        lastActivity = System.currentTimeMillis();
        try {    
            executeDDL("rollback to savepoint " + pPointName);
            return true;
        } catch (SQLException e) { // возникло исключение
            if (!e.getSQLState().equals("3B001")){ // это не наше исключение?
                throw e; // не наше -- уходим отдавая исключение
            } else {
                return false;    
            }
        }
    }
    
    @Override
    public void setSessionInfo(String pInfo) throws SQLException {
        super.setSessionInfo(pInfo);
        executeDDL("set session application_name='"+pInfo+"'");
        if (!this.autoCommit){
            this.commit();
        }
    }

    @Override
    public String NextVal(String pSequence) throws SQLException {
        return this.getSelectResult("select nextval('"+pSequence+"')");
    }
@Override
    public String getOffsetColumn(String alias) {
      return "";  
    }
     @Override
    public String setOffsetClause(String query,String offsetColAlias,String hint,int rows_fetched){ 
        return query+" offset " + String.valueOf(rows_fetched);
    }

    /**
     * Преобразование данных из строки полученной в результате выполнения запроса в строку по указанному формату
     * Все данные в результате выполнения запроса преобразовываются  к строке по умолчанию
     * их необходимо преобразовать в нужный тип целевой БД, а потом опять в строку, но по заданному в параметре формату
     * @param pResultsetString -
     * @param pColInfo
     * @return
    */
/*    @Override
    public String fromRStoStringByFormat(String pStringFromResultSet, ColInfo pColInfo, String pFormat) throws SQLException {
        return getSelectResult("select to_char((" + pColInfo.columnTypeName + " '" + pStringFromResultSet + "'),'" + pFormat + "')");
    };
  */  
    
    /**
     * Преобразование строки к числу/дате/времени... и обратно к строке по указанному формату
     * исходная строка может не соответствовать формату, но нужно попытаться её преобразовать
     * с использованием подмен , на .
     * @param pString - исходная строка
     * @param pFormat - формат
     * @param pLuiDataType - тип данных Lui:NUMBER DATE DATETIME
     * @return
     * @throws SQLException 
     */
/*    @Override
    public String StringToStringByFormat(String pString, String pFormat, String pLuiDataType) throws SQLException {
        switch (pLuiDataType) {   
            case "NUMBER":
                pString = pString.replace(" ", ""); // удаляем пробелы
                pString = pString.replace(",", "."); // заменяем запятую на точку
                if (is_not_null(pFormat)){
                    return getSelectResult("select to_char(numeric '" + pString + "','" + pFormat + "')");
                } else {
                    return getSelectResult("select numeric '" + pString + "'");    
                }
            case "DATE":
                if (is_not_null(pFormat)) {
                    return getSelectResult("select to_char(to_date('" + pString + "','" + pFormat + "'),'" + pFormat + "')");
                } else {
                    return getSelectResult("select date '" + pString + "'");        
                }
            case "DATETIME":
                if (is_not_null(pFormat)) {
                    return getSelectResult("select to_char(to_timestamp('" + pString + "','" + pFormat + "'),'" + pFormat + "')");
                } else {
                    return getSelectResult("select timestamp '" + pString + "'");           
                }
        }   
        return pString;     
    }
  */  
    /**
     * Проверка наличия несохранённых изменений (транзакции)
     * @return boolean true - есть, false - нет
     * @throws SQLException
     */
    @Override
    public boolean isTransaction() throws SQLException {
        //String r = getSelectResult("select 1 from pg_catalog.pg_locks where transactionid is not null and database is not null and pid = " + sid);
        // эта верхняя строка закоментирована LAL, т.к. в случае сложного PLPGSQL постгрес не всегда может определить database и relation
        // и они остаются пустыми, а транзакция есть!
        String r = getSelectResult("select 1 from pg_catalog.pg_locks where transactionid is not null and pid = " + sid+ "limit 1");
        if (r == null) return false;
        return (r.equals("1"));
    }
    

    @Override
    public String getSelectResult(String pStmt, int pRows) throws SQLException {
        StringBuilder res = null;
        String s = null;
        int i = 0;
        rowCount = 0;
        lastActivity = System.currentTimeMillis();
    
        Statement aSt = null;
        try {
            aSt = conn.createStatement();
            mStatements.add(aSt);
        
            try( ResultSet rs = aSt.executeQuery(pStmt)) {
            //stmt //try (ResultSet rs = stmt.executeQuery(pStmt)) {
                while (rs.next()) {
                    rowCount++;
                    if(res==null) res = new StringBuilder();
                    s = rs.getString(1);
                    if (!(s==null)){
                        res.append((String)s);
                    }
                    i++;
                    if (i>=pRows & pRows>0)break;
                }   
                if (res != null) return res.toString();
                    return null;
            }
        } catch(SQLException e) {
            if (e.getSQLState().equals("02000")) return ""; // если это ошибка связанная с тем, что это не запрос а что-то иное, то просто вернём пустую строку
            if (!autoCommit) conn.rollback();
            throw e;
        }
        finally {
            if(aSt != null) {
                aSt.close();
                mStatements.remove(aSt);
            }
        }
    }

    
    /**
     * Формирование строки, содержащей обращение к данным в заданном столбце, для её использования в SQL-запросе
     * @param pColumn - код столбца
     * @param pLuiDataTypeSrc - исходный тип данных (LUI)
     * @param pLuiDataTypeTgt - целевой тип данных (LUI)
     * @param pLuiFormat - формат данных
     * @return 
     */
    @Override
    public String getColumnSQLClause (String pColumn, String pLuiDataTypeSrc, String pLuiDataTypeTgt, String pLuiFormat) {
        String s = null;
        switch (NVL(pLuiDataTypeSrc,"CHAR")) {
            case "CHAR":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        s = pColumn; 
                        break;
                    case "NUMBER": 
                        if (is_not_null(pLuiFormat)) s = "to_number(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = pColumn + "::numeric";
                        break;
                    case "DATE":
                        if (is_not_null(pLuiFormat)) s = "to_date(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = pColumn + "::date";
                        break;
                    case "DATETIME":
                        if (is_not_null(pLuiFormat)) s = "to_timestamp(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = pColumn + "::timestamp";
                        break;
                }
                break;
            case "NUMBER":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        if (is_not_null(pLuiFormat)) s = "to_char(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = pColumn + "::text";
                        break;
                    case "NUMBER":
                    case "DATE":
                    case "DATETIME":
                        s = pColumn;
                        break;
                }
                break;
            case "DATE":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        if (is_not_null(pLuiFormat)) s = "to_char(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = pColumn + "::text";
                        break;
                    case "NUMBER": 
                    case "DATE": 
                        s = pColumn;
                        break;
                    case "DATETIME":
                        s = pColumn + "::timestamp";
                        break;
                }
                break;
            case "DATETIME":
                switch (NVL(pLuiDataTypeTgt,"CHAR")) {
                    case "CHAR": 
                        if (is_not_null(pLuiFormat)) s = "to_char(" + pColumn + ", '" + pLuiFormat + "')";
                        else s = pColumn + "::text";
                        break;
                    case "NUMBER": 
                    case "DATETIME":
                        s = pColumn;
                        break;
                    case "DATE":
                        s = pColumn + "::date";
                        break;
                }
                break;
        }
        return s;
    }

    @Override
    public String getDBvendor() {
     return "PG";
    }

   
    
    /**
    * Вложенный класс - курсор
    */
    class Cursor implements LuiDbCur {
        
        private ArrayList<ColInfo> colList; //список столбцов
        //private Statement stmt; //Объект JDBC "Оператор"
        private PreparedStatement prstmt;
        private PgResultSet resSet; // Набор строк
        private PgResultSetMetaData resSetMetaData; // JDBC описание столбцов
        private ArrayList<String> row;
        private int cols;
        private boolean is_executed;
        
        
        private String mSQLText = null;
        @Override
        public String getSQLText() { return mSQLText; }

        /**
        * Конструктор - инициализатор
        * 
        */
        Cursor() {
            colList=new ArrayList<>();
            cols=0;
            row=new ArrayList<>();
            prstmt=null;
            resSet=null;
            is_executed=false;
            
        }

        @Override
        public void open(String pStmt) throws SQLException {
            try {
                if (isOpen()) close();
            } catch (Exception ex){}
            colList.clear();
            row.clear();
            prstmt=conn.prepareStatement(mSQLText = pStmt);
            resSetMetaData=(PgResultSetMetaData) prstmt.getMetaData();
            cols=resSetMetaData.getColumnCount();
            colList=new ArrayList<>();
            for (int i = 1; i <= cols; i++) {
                ColInfo ci;
                //String pSchemaName,String pTableName,String pTableAlias,String pName,String pAlias,String pColumnClassName,String pColumnTypeName,int pColumnType
                ci = new ColInfo(resSetMetaData.getBaseSchemaName(i),
                                 resSetMetaData.getBaseTableName(i),
                                 "",
//                                 resSetMetaData.getTableName(i), // вот здесь собака postgres не даёт алиас имени таблицы и из-за этого невозможно правильно сочинить
                                                         // select ... from table1 t1,table2 t2... for update of t1,t2 
                                                         // И не откуда взять имена этих алиасов
                                                         // 
                                 resSetMetaData.getBaseColumnName(i),
                                //resSetMetaData.getColumnName(i),
                                 resSetMetaData.getColumnLabel(i),
                                 resSetMetaData.getColumnClassName(i),
                                 resSetMetaData.getColumnTypeName(i),
                                 resSetMetaData.getColumnType(i));
                ci.LuiDATATYPE="CHAR";
                switch (ci.columnTypeName){
                    case "smallint":
                    case "integer":
                    case "bigint":
                    case "decimal":
                    case "numeric":
                    case "real":
                    case "double precision":
                    case "smallserial":
                    case "serial":
                    case "bigserial":
                    case "money":
                    case "int4":
                    case "int8":
                    case "int":
                    case "interval": ci.LuiDATATYPE="NUMBER"; break;
                    case "varchar":
                    case "character varying":
                    case "character":
                    case "text": ci.LuiDATATYPE="CHAR"; break;    
                    case "date": ci.LuiDATATYPE="DATETIME"; break;
                    case "timestamp":
                    case "timestamp with time zone":
                    case "timestamp without time zone":
                    case "timestamptz": ci.LuiDATATYPE="DATETIME"; break;
                }
                colList.add(ci);
            }
        }
    
        /**
        * закрытие курсора
        * @throws java.sql.SQLException
        */
        @Override
        public void close() throws SQLException {
            if (resSet!=null) {if (!resSet.isClosed()) resSet.close();}
            if (prstmt!=null) {if (!prstmt.isClosed()) prstmt.close();}
            is_executed=false;
        }

        /**
        * Извлечение очередной строки из курсора
        * @return true - строка извлечена false -- данных больше нет
        * @throws java.sql.SQLException
        */
        @Override
        public boolean nextRow() throws SQLException {
        if (!is_executed){
           prstmt.setFetchSize(200);
            resSet=(PgResultSet) prstmt.executeQuery();
      
            is_executed=true;
        }
            if (resSet.next()){
                row.clear();
                for (int i=1;i<=cols;i++) {
                    row.add(resSet.getString(i));
                }
                return true;
            }
            return false;
        }
        
        /**
        * Проверка открытости курсора
        * @return true - открыт, false - закрыт
        */
        @Override
        public boolean isOpen() {
            return cols>0;
        }

        /**
        * Возвращает количество столбцов в строке курсора
        * @return количество столбцов в строке
        */
        @Override
        public int getCols() {
            return cols;
        }
        
        /**
        * Получить описание столбцов запроса
        * @return список объектов ColInfo
        */
        @Override
        public ArrayList<ColInfo> getColsInfo() {
            return colList;
        }

        /**
        * Получить текущую строку c данными курсора (последняя извлечённая строка запроса)
        * @return список объектов ColInfo
        */
        @Override
        public ArrayList<String> getRow() {
            return row;
        }
        
    }

    
    private static final AtomicLong stThreadNumber = new AtomicLong(0);

    /* Вложенный класс для обнюхивания другим тредом "блокировок текущего процесса другим сеансом" */    
    class ThreadForFindLocks implements Runnable{
        Thread thrd; // тред
        Connection conn4l; // соединение
        Statement stmt4l;
        ResultSet rs;
      
        /**
        * Конструктор объектов класса 
        * @throws java.sql.SQLException
        */
        public ThreadForFindLocks() throws SQLException{
            thrd = new Thread(this);
            thrd.setDaemon(true);
            thrd.setName(String.format("PG Lock Finder #%d", stThreadNumber.incrementAndGet()));
            this.thrd.start();
        }
    
        @Override
        public void run() {
            //  System.out.println("Thread running");
            try {
                Thread.sleep(500);
                try { 
                    conn4l = DriverManager.getConnection(dbURL, username, password);
                    stmt4l = conn4l.createStatement();
                    rs = stmt4l.executeQuery("select replace(replace(pg_blocking_pids(" + sid + ")::text,'{',''),'}','')");
                    if (rs.next()) { // запрос вернул строку - блокировка найдена
                        String locking_pid = rs.getString(1); // извлекаем pid (session id) блокирующего процесса
                        if (locking_pid != null){ // если не пуст совсем
                            if (!locking_pid.isEmpty()) { // если не пуст
                                rs.close(); // закрываем запрос
                                rs = stmt4l.executeQuery("select usename from pg_stat_activity where pid = " + locking_pid); // ищем сеанс с таким pid
                                if (rs.next()) { // есть сеанс, извлекаем данные о пользователе
                                    String locking_usr = rs.getString(1);
                                    alString.clear();
                                    alString.add("locked");
                                    alString.add(locking_usr);
                                    
                                    for(Statement aSt : mStatements) {
                                        aSt.cancel();
                                    }
                                    mStatements.clear();
                                    //stmt.cancel(); // убиваем оператор основного треда, выполняющийся и заблокировавшийся
                                }
                            }
                        } 
                    }
                    stmt4l.close();
                    conn4l.close(); // 
                } catch (SQLException e) {
                    System.err.println(e);
                }
            } catch (InterruptedException e) {}
        }
    }

 public String getQuery4Column(String pTable){
String q="select\n" +
"    cols.table_schema \"SCHEMA\",\n" +
"    cols.column_name \"NAME\",\n" +
"    cols.data_type \"DATA_TYPE\",\n" +
"    (\n" +
"        SELECT\n" +
"            pg_catalog.col_description(c.oid, cols.ordinal_position::int)\n" +
"        FROM\n" +
"            pg_catalog.pg_class c\n" +
"        WHERE\n" +
"            c.oid = (SELECT ('\"' || cols.table_name || '\"')::regclass::oid)\n" +
"            AND c.relname = cols.table_name\n" +
"    ) AS \"CMNT\",\n" +
"    (\n" +
/*"select case count(*) when 0 then 'N' else 'Y' end from\n" +
"    information_schema.table_constraints co,\n" +
"    information_schema.constraint_column_usage cc\n" +
"    where \n" +
"    co.table_catalog=CURRENT_CATALOG AND\n" +
"    co.table_name=cols.table_name and\n" +        
"    co.table_schema=cols.table_schema \n" +        
"    and  co.constraint_type='PRIMARY KEY'\n" +
"    and cc.constraint_catalog=co.constraint_catalog\n" +
"    and cc.constraint_name=co.constraint_name\n" +
"    and cc.table_catalog=co.table_catalog\n" +
"    and cc.table_name=co.table_name\n" +
"    and cc.table_schema=co.table_schema\n" +
"    and cc.column_name=cols.column_name    \n" +*/
"select case count(*) when 0 then 'N' else 'Y' end \n" +
"from pg_constraint c,\n" +
"     pg_namespace ns,\n" +
"     pg_class rel,\n" +
"     pg_attribute a\n" +
" where c.contype='p' and rel.oid=c.conrelid AND\n" +
" rel.oid = a.attrelid and a.attnum = any (c.conkey)\n" +
" and ns.oid = rel.relnamespace\n" +
" and rel.relname=cols.table_name\n" +
" and ns.nspname=cols.table_schema\n" +
" and a.attname=cols.column_name"+        
"    ) as \"PK\",\n" +
"     substring(cols.is_nullable,1,1) \"NULLABLE\"\n" +
"FROM\n" +
"    information_schema.columns cols\n" +
"WHERE\n" +
"    cols.table_catalog    = CURRENT_CATALOG\n" +
"    AND cols.table_name   = '<table>'\n" +
"    AND cols.table_schema = '<owner>'\n" +
"order by 1,2";
String pOwner="";
if (pTable.indexOf(".")>-1){
 pOwner=Extract_Word(pTable,1,".");
 pTable=Extract_Word(pTable,2,".");
}
if (is_null(pOwner)) { 
    //if (is_null(pOwner)) { pOwner="";q=q.replace("'<owner>'","CURRENT_SCHEMA");}
    if (is_null(pOwner)) { 
    pOwner="";
    q=q.replace("'<owner>'"," any (current_schemas(true))");
    }
}
if (pTable.indexOf(" ")>-1){
    pTable=Extract_Word(pTable,1," ");
}
pTable=LTrim(pTable);
pTable=RTrim(pTable);

pOwner=LTrim(pOwner);
pOwner=RTrim(pOwner);

q=q.replace("<owner>", pOwner);
q=q.replace("<table>",pTable);
return q;
}
 /**
     * возвращает select для извлечения данных о таблицах или view
       в запросе д.б. столбцы:
       SCHEMA- имя схемы(владельца в Oracle)
       TABLE_NAME- имя таблицы
       TABLE_TYPE- тип данных СУБД
       COMMENTS- комментарий
       
     */    @Override
    public String getQuery4Table() {
       return "select table_schema \"SCHEMA\",table_name \"TABLE_NAME\",\n" +
"case table_type when 'BASE TABLE' then 'TABLE' else table_type end \"TABLE_TYPE\",\n" +
"(select pg_catalog.obj_description(oid)\n" +
"from pg_catalog.pg_class cla where cla.oid=((tb.table_name)::regclass::oid)) \"COMMENTS\"\n" +
"from information_schema.tables tb \n" +
"where tb.table_schema=CURRENT_SCHEMA and\n" +
"tb.table_catalog=CURRENT_CATALOG";
        
    }
  
}