/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс для хранения массива накопленных сообщений об ошибках и не об ошибках
 */
package LuiExceptionMsg;
import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;
import LuiNLS.LuiNLS;
import java.io.PrintStream;
import java.sql.SQLException;
import java.util.ArrayList;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiClCommand;
import luiserver.LuiClCommandGroup;
import luiserver.LuiClCommandWriter;
/**
 *
 * @author aliubushkin
 */
public class LuiMsgWriter {
public ArrayList<LuiMsg>  msgArray;    //массив накапливаемых сообщений    
private LuiDbConnPool connPool; //пул соединений для доступа в БД к таблице сообщений используется если приложение работает без выделения индивидуальных сессий
private LuiDbConn conn; //соединение для доступа к таблице сообщений, либо оно постоянное переданное в конструкторе, либо пусто
private LuiNLS nls; // текущий NLS
private String AppCode; //код прикладной системы
private static final String STMT="select message_text,help_text,severety from lui_t_message where app_code='<app_code>' and code='<code>'";
private static final int SD=2; //глубина поиска стека

private LuiMsg m;
private PrintStream errStream;
private PrintStream outStream;
private LuiApplication app;
//private LuiMsg lastm=null;
/**
 * Конструктор с пулом. Используется если приложение работает без выделения индивидуальных сессий
 * @param pConnPool -пул соединений для доступа в БД к таблице сообщений
 * @param pNLS - ссылка на объект NLS
 * @param pAppCode - код прикладной системы
 */
public LuiMsgWriter(LuiDbConnPool pConnPool,LuiNLS pNLS,String pAppCode){
    connPool=pConnPool;
    conn=null;
    msgArray=new ArrayList<>();
    nls=pNLS;
    AppCode=pAppCode;
    
}
/**
 * Конструктор с выделенным коннектом. Используется если приложение работает с выделением индивидуальных сессий
 * @param pConn - соединение для доступа в БД к таблице сообщений
 * @param pNLS - языковые настройки
 * @param pAppCode - код приложения
 */
public LuiMsgWriter(LuiDbConn pConn,LuiNLS pNLS,String pAppCode){
    conn=pConn;
    connPool=null;
    msgArray=new ArrayList<>();
    nls=pNLS;
    AppCode=pAppCode;
    

}
/**
 * Конструктор с выделенным коннектом и записью только в PrinstStream. 
 * Используется при старте сервера
 * @param pConn - соединение для доступа в БД к таблице сообщений
 * @param pNLS языковые настройки
 * @param pAppCode код приложения
 * @param pSerr поток вывода ошибок
 * @param pSout поток вывода не ошибок
 */
public LuiMsgWriter(LuiDbConn pConn,LuiNLS pNLS,String pAppCode,PrintStream pSerr,PrintStream pSout){
    conn=pConn;
    connPool=null;
    msgArray=new ArrayList<>();
    nls=pNLS;
    AppCode=pAppCode;

    errStream=pSerr;
    outStream=pSout;
}
public void setApplication(LuiApplication pApp){
    app=pApp;
    
}
public void sqlError(LuiDbConn plc,SQLException ex, LuiAppProp prop, String stmt) throws LuiException, SQLException{
    
String msg=ex.getLocalizedMessage();
if (app!=null) {
  if (app.luiSession!=null) {  
  //if (app.getProperty("DEBUG_MODE").equals("N")){     
   
         
     // перехват
     if (app.interceptor4sql_error!=null) {
     if (app.interceptor4sql_error.size()>0){        
     for (String key:app.interceptor4sql_error.keySet()){
       if (like(msg,app.interceptor4sql_error.Get_Element(key))){
           this.write(key,msg);
           return;
           
       }  
     }
     }
     }
     // конец перехвата      
   // }
     if ("PG".equals(plc.getDBvendor())) { // переделка SQL-ошибки PG в LUI-exception
      if (ex.getSQLState().equals("22000")){
        int i=msg.indexOf("LUI_MESSAGE_CODE");
        if (i>=0) {
            int j=msg.indexOf("‡");
            LuiCollection msgCollection=new LuiCollection(msg.substring(i,j));
            String msgCode=msgCollection.Get_Element("LUI_MESSAGE_CODE");
            if (is_not_null(msgCode)) {
               String p1=NVL(msgCollection.Get_Element("p1"));
               String p2=NVL(msgCollection.Get_Element("p2"));
               String p3=NVL(msgCollection.Get_Element("p3"));
               String p4=NVL(msgCollection.Get_Element("p4"));
               String p5=NVL(msgCollection.Get_Element("p5"));
               String p6=NVL(msgCollection.Get_Element("p6"));
               String p7=NVL(msgCollection.Get_Element("p7"));
               String p8=NVL(msgCollection.Get_Element("p8"));
               String p9=NVL(msgCollection.Get_Element("p9"));
               String p10=NVL(msgCollection.Get_Element("p10"));
               write(msgCode,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10 );
               return;
                }
        }
      }   
     }
//      write("LUI-00034","\n"+ex.getMessage(),"");
if (prop==null){
write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+stmt+") ","");    
} else
write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+stmt+") ",prop.getGenesis4msg());

      return;
  
} 
}
if (prop==null){
write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+stmt+") ","");    
} else
write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+stmt+") ",prop.getGenesis4msg());
}
/**
 * Запись сообщения
 * @param pCode - код
 * @param pParams - параметры
 * @throws LuiException проблемы
 */
public void write(String pCode,String ... pParams)throws LuiException {
ArrayList<String> errTabRow=null;    
int i=0;
LuiDbConn con=null;
String schema=null;
if (connPool!=null) {schema=connPool.getSchema();}
if (is_not_null(schema))schema=schema+".";
// выясним точку возникновения    
String aCalling="thread:\""+Thread.currentThread().getName()+"\" method:"+Thread.currentThread().getStackTrace()[SD].getClassName()+"."
        + Thread.currentThread().getStackTrace()[SD].getMethodName()+" line:"
        +Thread.currentThread().getStackTrace()[SD].getLineNumber();
// если надо получаем соединение
if (conn==null && connPool!=null) {
    try {
    String sesid="";
    String appusr="";
    if (app!=null){
      if (app.luiSession!=null){
         sesid=app.luiSession.id;
         appusr=app.getAppUser();
      }
    }
    con=connPool.GetFree(nls.currentLang.locale,nls.currentLang.code,AppCode,sesid,appusr);
    } catch (SQLException ex) {
    errTabRow=new ArrayList<String>();
    pCode=AppCode+"-00000";
    errTabRow.add("Can't connect for find message "+pCode+" in \"lui_t_messages\" table:"+ex.getLocalizedMessage());
    errTabRow.add("");
    errTabRow.add("E");
    }
} else 
{con=conn;}
if (con!=null){
    try {
        // ищем мессагу в таблице
        errTabRow=con.getSelectFirstRowValList(STMT.replace("<app_code>", AppCode).replace("<code>", pCode));
        if (errTabRow==null) { 
            if (!AppCode.equals("LUI")){
                errTabRow=con.getSelectFirstRowValList(STMT.replace("<app_code>","LUI").replace("<code>", pCode));
            }
            if (errTabRow==null) { 
            errTabRow=new ArrayList<String>();
            pCode=AppCode+"-00000";
            errTabRow.add("Can't find message "+pCode+" in \"lui_t_messages\" table.");
            errTabRow.add("");
            errTabRow.add("E");
            }
        }
    } catch (SQLException ex) {
    errTabRow=new ArrayList<String>();
    pCode=AppCode+"-00000";
    errTabRow.add("Can't find message "+pCode+" in \"lui_t_messages\" table:"+ex.getLocalizedMessage());
    errTabRow.add("");
    errTabRow.add("E");
    }
}    
// подготовим тело сообщения
String mt="";
if (errTabRow!=null && errTabRow.size()>0)
        mt=NVL(nls.NLS(errTabRow.get(0)),"");
        // подготовим help сообщения
String mh="";
if (errTabRow!=null && errTabRow.size()>1)
       mh=NVL(nls.NLS(errTabRow.get(1)),"");
        
        // пропарсим параметры
        if (pParams!=null){ 
        for(i=0;i<pParams.length;i++){
        if (pParams[i]!=null){
        mt=mt.replace("<p"+(i+1)+">", pParams[i].toString());
        mh=mh.replace("<p"+(i+1)+">", pParams[i].toString());
        }
         }  
        }
        // если что-то осталось не пропарсено
        for(;i<100;i++){
           mt=mt.replace("<p"+(i+1)+">", "");
           mh=mh.replace("<p"+(i+1)+">", "");
        }
        // теперь создаём объект "сообщение"
        m=new LuiMsg(pCode, //Код
                mt, //Сообщение
                mh, // Help
                errTabRow.get(2), // Severety
                aCalling); //Стек
        
if (errStream!=null) {
    if (m.severity.equals("E")){
// errStream.println("("+m.severity+")"+m.code+":"+m.messageText);
    }else {
 
 outStream.println("("+m.severity+")"+m.code+":"+m.messageText);         
    }
        
} else {
// заносим его в список сообщений
/*
if (lastm!=null) {
  if (!(lastm.toString().equals(m.toString()))){
msgArray.add(m);      
  } else {
      m=lastm;
  }
} else {
msgArray.add(m);
}
lastm=m;
}*/
msgArray.add(m);
}
// если надо вернём соединение в пул
if (connPool!=null & con!=null){
    try {
        connPool.SetFree(con);
    } catch (SQLException ex) {
        
    }
}
if (app!=null){
     if (app.luiSession!=null){
       if (app.log!=null){
          switch (m.severity){
           case "E":
           app.log.addRecord("ERROR",m.code+"("+m.severity+"):"+m.messageText, aCalling,"");
           break;
           case "W":
           app.log.addRecord("WARNING",m.code+"("+m.severity+"):"+m.messageText, aCalling,"");
           break;
           case "I":
           app.log.addRecord("INFO",m.code+"("+m.severity+"):"+m.messageText, aCalling,"");
           break;
          }
       }
     }   
    }
if (m.severity.equals("E")){
    //throw new LuiException(m.code,m.messageText,m.severity,msgArray,aCalling);
    String mtxt=m.messageText;
    
    if (is_not_null(m.helpText)){
            mtxt=mtxt+"\n"+this.nls.NLS("RUS:Совет: ;ENG:Advise: ;")+m.helpText;
    }       
    throw new LuiException(m.code,m.messageText,m.helpText,m.severity,msgArray.size()-1,aCalling,app);
}
}

public void Remove(int pI){
    msgArray.remove(pI);
}
public void ClearMsgArray(){
    if (msgArray!=null){
        msgArray.clear();
    }
}
public void msgs2commands(LuiClCommandWriter commandWriter){
if (app!=null){
  if (app.properties.containsKey("AsyncTask")) {
     ClearMsgArray();
     return;
  }  
}    
for (int i=0;i<msgArray.size();i++){
  LuiMsg m=msgArray.get(i);
  if (app!=null){
   String errfrmCode=null;
   try{   
   errfrmCode=app.getProperty("ERROR_MESSAGE_FORM");   
   
              String FormID=NVL(app.getVal("FormID"),"");
              String FormCode="";
              if (app.forms.containsKey(FormID)){
                  FormCode=NVL(app.forms.get(FormID).getProperty("MetadataCode"),"");
              }
    if (is_not_null(errfrmCode)) {
     if (!errfrmCode.equals(FormCode)){      
     LuiCollection parcol=new LuiCollection("");
     parcol.Add_Element("CONFIG","DEFAULT");
     parcol.Add_Element("MSG_SEVERETY",m.severity);
     parcol.Add_Element("MSG_CODE",m.code);
     parcol.Add_Element("MSG_TEXT",m.messageText);
     parcol.Add_Element("MSG_HELPTEXT",m.helpText);
     parcol.Add_Element("MSG_CALLING",m.calling);
     app.executeForm(errfrmCode, parcol.toString(), app.getVal("FormID"),null/*,"N"*/);
        break;
    }else errfrmCode="";
    }
     } catch(LuiException|SQLException ex){}
  }
  LuiClCommandGroup XMLcg=commandWriter.addGroup("MSG");
  LuiClCommand XMLcmd=XMLcg.addCommand("Message");
  String title="";
  if (app!=null){
       switch (m.severity) {
          case "E":
           title=app.NLS.NLS("RUS:Ошибка ;ENG:Error ;");
          break;
          case "W":
          case "M":
           title=app.NLS.NLS("RUS:Сообщение ;ENG:Message ;");
          break;
          case "I":
           title=app.NLS.NLS("RUS:Информация ;ENG:Information ;");   
          break;
      }
  } else {
      switch (m.severity) {
          case "E":
           title="Error ";
          break;
          case "M":
          case "W":
           title="Message ";
          break;
          case "I":
           title="Information ";   
          break;
      }
  }
  title=title+m.code;
  if (m.severity.equals("E")) {
      XMLcmd.addParameter("Type", "Fail");
      XMLcmd.addParameter("Title", title);
  }
  if (m.severity.equals("M") || m.severity.equals("W"))  {
      XMLcmd.addParameter("Type", "Error");
      XMLcmd.addParameter("Title", title);
  }
  if (m.severity.equals("I"))  {
      XMLcmd.addParameter("Type", "Info");
      XMLcmd.addParameter("Title", title);
  }
      
                         XMLcmd.addParameter("Text", m.messageText);
                         XMLcmd.addParameter("Notes",m.toString());
}
ClearMsgArray();
}
public void setNLS(LuiNLS pNLS){
 nls=pNLS;   
}
public void setAppCode(String pAppCode){
  AppCode=pAppCode;
}
}