/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс, обеспечивающий работу LUI на нескольких языках RUS & ENG
 */
package LuiNLS;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_null;
import LuiDbInterface.LuiDbConn;
import java.sql.SQLException;
import java.util.HashMap;

/**
 *
 * @author aliubushkin
 */
public class LuiNLS {

    public HashMap<String, LuiLang> langList;
    public LuiLang currentLang;
    public String langs;

    /**
     * Конструктор класса
     *
     * @param pConn - соединение с БД для доступа к lui_t_language
     * @param pAppSysCode - код прикладной системы
     * @throws SQLException ошибка при выполнении SQL-запроса
     */
    public LuiNLS(LuiDbConn pConn, String pAppSysCode) throws SQLException {
        Integer cursor_id = null;
        langList = new HashMap<String, LuiLang>();
        langs = "";
        String cl = "";

        try {
            cursor_id = pConn.CursorOpen("select code,name,language_code,is_def,locale from lui_t_language where enable='Y' and app_code='" + pAppSysCode + "'");
            while (pConn.CursorNextRow(cursor_id)) {
                langList.put(pConn.CursorGetRow(cursor_id).get(0), // ключ мапа
                        new LuiLang(pConn.CursorGetRow(cursor_id).get(0), pConn.CursorGetRow(cursor_id).get(1), pConn.CursorGetRow(cursor_id).get(2), pConn.CursorGetRow(cursor_id).get(4)) // язык
                );
                if (pConn.CursorGetRow(cursor_id).get(3).equals("Y")) {
                    cl = pConn.CursorGetRow(cursor_id).get(0);
                }
            }
        } finally {
            if (cursor_id != null) {
                if (pConn.CursorisOpen(cursor_id)) {
                    pConn.CursorClose(cursor_id);
                }
            }
        }
        if ("".equals(cl)) {
            for (String k : langList.keySet()) {
                cl = k;
                break;
            }
        }
        setCurrentLang(cl);

        //langs = "";
        StringBuilder aLangsBuilder = new StringBuilder();
        for (String lk : langList.keySet()) {
            aLangsBuilder.append(lk).append(':')
                    .append(NLS(langList.get(lk).name, NVL(currentLang.code, lk)))
                    .append(';');
            //langs = langs + lk + ":" + NLS(langList.get(lk).name, NVL(currentLang.code, lk)) + ";";
        }
        langs = aLangsBuilder.toString();
    }

    /**
     * Устанавливает текущий язык
     *
     * @param pLangCode код языка
     */
    public void setCurrentLang(String pLangCode) {
        currentLang = langList.get(pLangCode);
        if (currentLang.code.equals("RUS")) {
           // Locale l = new Locale("ru");
           // Locale.setDefault(l);
        } else {
            //Locale.setDefault(Locale.ENGLISH);
        }
        langs = "";
        for (String lk : langList.keySet()) {
            langs = langs + lk + ":" + NLS(langList.get(lk).name, NVL(currentLang.code, lk)) + ";";
        }
    }

    public String NLS(String pMultiLangText, String pLangPref) {
        if (is_null(pMultiLangText)) return "";
        
        String l;
        l = currentLang.code;
        if (pLangPref != null) {
            if (!pLangPref.equals("")) {
                l = pLangPref;
            }
        }
        return new LuiCollection(pMultiLangText).Get_Element(l);
    }

    public String NLS(String pMultiLangText) {
        String res = null;
        String fr=null;
        String lc = currentLang.code;
        LuiCollection c = new LuiCollection(pMultiLangText);
        if (c.size()==0) return pMultiLangText;
        for (int i = 0; i < langList.size(); i++) {
            res = c.get(lc);
            if (res != null) {
                if (i==0) {
                        fr=""; 
                    }
                if (!res.equals("")) {
                    return res;
                }
            }
            String pl=langList.get(lc).parentLangCode;
            if (pl==null) {
                //return pMultiLangText;
                return res;
            }
            if (pl.equals("")) {
                //return pMultiLangText;
                return res;
            }
            lc = langList.get(lc).parentLangCode;
        }
        if (fr!=null) return fr;
        return pMultiLangText;
       
    }
}
