/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * вычисление значение свойства приложения
 * Ситаксис
ApplicationProperty:<код свойства>
 * установка значения свойства приложения
 * Ситаксис
ApplicationProperty:<код свойства>=<значение>
 */
package Standard;

import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;
import static LuiCollection.LuiString.LTrim;
import static LuiCollection.LuiString.NVL;
import LuiDbInterface.LuiDbConn;

/**
 *
 * @author aliubushkin
 */
public class ApplicationProperty implements LuiPlugin{
    LuiApplication app;

    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
          return comp(pString,pProp,false);
    }
private String comp(String pString, LuiAppProp pProp, boolean pFindOnly) throws LuiException, SQLException {
    
    if (app.properties.containsKey(pString)) { // если св-во есть
        app.properties.get(pString).setDepend(pProp); // то говорим, что оно определяет свойство pProp
          if (pFindOnly) return null;
          return NVL(app.properties.get(pString).getComputedProperty(),"");
    } else { // если такого нет
     // поищем его в справочнике и узнаем должно ли быть оно в памяти
     LuiDbConn lc=null;   
     try{
     lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
     String in_memory=lc.getSelectResult("select in_memory from lui_t_application_property where app_code='"+app.code+
             "' and code='"+pString+"'");
     if (in_memory==null){ // нет такого в БД
        in_memory="Y"; // значит д.б. в памяти
     }
     // читаем значение из БД
     String dbv=
     lc.getSelectResult("select lui_f_get_application_property('"+app.code+"','"+pString+"','"+app.NLS.currentLang.code+"')");
     // добавляем его к пропертям
     LuiAppProp p=app.addProperty(pString,NVL(dbv,""));
     // вычисляем из памяти
     dbv=app.getProperty(pString);
     if (in_memory.equals("Y")){ // если оно должно остаться в памяти
     p.setDepend(pProp); // установим зависимость
     }
     else { // если не должно оставаться в памяти - удаляем
      app.properties.remove(pString);
     }
     return NVL(dbv,"");
     } finally {
         if (lc!=null){
             app.metaConnPool.SetFree(lc);
         }
     }   
    }   
}
    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        // выясняем код свойства
        pString=LTrim(pString);
        int eq=pString.indexOf("=");
        String pc=pString.substring(0, eq);
        // выясняем значение
        String v=pString.substring(eq+1);
        
        app.addProperty(pc, v);
        if ("DEBUG_MODE".equals(pc)){
            if (app.luiSession!=null){
                app.luiSession.setSessionProp(pc, v);
            }
        }
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
     this.comp(pString, pProp, true);
    }
    
}
