/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Внешняя аутентификация 
 * AuthExternal:<token>;usesername:username;roles:роли через запятую;service:;
 * Compute возвращает коллекцию:
   username
   roles
   success:Y/N
   url
 */
package Standard;

import LuiCollection.LuiCollection;

import static LuiCollection.LuiString.is_null;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.sql.SQLException;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiClCommand;
import luiserver.LuiClCommandGroup;
import luiserver.LuiPlugin;
import org.keycloak.OAuth2Constants;
import org.keycloak.adapters.KeycloakDeploymentBuilder;
import org.keycloak.adapters.ServerRequest;
import org.keycloak.adapters.rotation.AdapterTokenVerifier;
import org.keycloak.common.VerificationException;
import org.keycloak.common.util.KeycloakUriBuilder;
import org.keycloak.representations.AccessToken;
import org.keycloak.representations.AccessTokenResponse;
import org.keycloak.representations.IDToken;
import ru.fors.mvd.vps.auth.AuthConsumer;
import static LuiCollection.LuiString.NVL;


/**
 *
 * @author aliubushkin, achibuk
 */
public class AuthExternal implements LuiPlugin{
    LuiApplication app;
//    AuthConsumer authConsumer;
//    String url;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
/*        try {
            authConsumer = new AuthConsumer();
        } catch (java.lang.NoClassDefFoundError e) {
           return; 
        }
        try {
            authConsumer.init();
        } catch (ClassNotFoundException ex) {
            Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, null, ex);
        }
*/
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        return exe('C', pString, pProp);
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        exe('E', pString, pProp);
    }
    
    @SuppressWarnings("empty-statement")
    String exe(char executeOrCompute,String pString,LuiAppProp pProp) throws LuiException, SQLException {
        
        LuiCollection mainCollection = new LuiCollection("Command:"+pString);
        String command = mainCollection.Get_Element("Command");
        String authUrl = NVL(mainCollection.Get_Element("authUrl"),app.getProperty("EA_URL"));
        String authUrlApp = NVL(mainCollection.Get_Element("authUrlApp"),app.getProperty("EA_URL_APP"));
        String authUrlCommon = NVL(mainCollection.Get_Element("authUrlCommon"),app.getProperty("EA_URL_COMMON"));
        String authMethod = NVL(mainCollection.Get_Element("authMethod"),app.getProperty("EA_METHOD"));
        LuiCollection rc=new LuiCollection("");

        // *****
        // 1. Технология Keycloak SSO (Single Sign-On)
        // *****
        if (authMethod.equals("KEYCLOAK")) {
            
            String authCode = NVL(mainCollection.Get_Element("authCode"),app.getProperty("authCode"));     //  Это параметр code, который присутствует в url-e
            String configJson = app.getProperty("EA_CONFIG_JSON");
            app.luiSeance.setDeployment(KeycloakDeploymentBuilder.build(new ByteArrayInputStream(configJson.getBytes(StandardCharsets.UTF_8))));
            String redirectUri = authUrlApp;   // это url нашего приложения
            switch (command) {
                case "login":
                    // Проверяем, задано ли значение для свойства authCode. Устанавливается в конструкторе LuiApplication на основе значения code в Params.
                    if (authCode.isEmpty()) {  // Здесь подразумевается параметр code, который будет в url-e
                        // Если отсутствует, отправляем за получением code (authCode)
                        keycloakEmptyCode(redirectUri); // получение code
                        rc.Add_Element("success", "N"); // авторизация неуспешна, может повезёт в след.раз, когда будет известен authCode
                    } else {
                        keycloakAuthorize(authCode, redirectUri); // авторизация пользователя во внешней системе и в LUI
                        rc.Add_Element("success", "Y");
                    }   break;
                case "logout":
                    try {
                        ServerRequest.invokeLogout(app.luiSeance.getDeployment(), authUrlCommon);
                        app.luiSeance.setShutdown();
                        rc.Add_Element("success", "Y");
                    } catch (IOException ex) {
                        app.msgWriter.write("LUI-00151", ex.toString());
                        rc.Add_Element("success", "N");
                    } catch (ServerRequest.HttpFailure ex) {
                        app.msgWriter.write("LUI-00151", ex.toString());
                        rc.Add_Element("success", "N");
                    }   break;
                case "exit":
                    redirect(authUrlCommon, "ExitApp");
                    app.luiSeance.setShutdown();
                    rc.Add_Element("success", "Y");
                    break;
                default:
                    Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "Wrong command: " + command);
                    rc.Add_Element("success", "N");
                    break;
            }

        // *****
        // 2. Технология Token авторизации (разработка ОРИС)
        // *****
        } else if (authMethod.equals("TOKEN")) {
            
            String authToken = mainCollection.Get_Element("token");
            String authUsername = mainCollection.Get_Element("username");
            String authRoles = mainCollection.Get_Element("roles");
            String serviceName = mainCollection.Get_Element("service");
            
            switch (command) {
                case "login":
                    if (authToken.isEmpty()) {
                        Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "External authorization failed. Attribute \"authToken\" is null.");
                        redirect(authUrl, "ExtAuthNoToken");
                        app.luiSeance.setShutdown();
                        rc.Add_Element("success", "N");
                    } else {
                        rc = new LuiCollection(tokenAuthorize(authUrl, serviceName, authToken, authUsername, authRoles));
                    }   break;
                case "logout":
                case "exit":
                    redirect(authUrlCommon, "ExitApp");
                    app.luiSeance.setShutdown();
                    rc.Add_Element("success", "Y");
                    break;
                default:
                    // Wrong command
                    Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "Wrong command: " + command);
                    rc.Add_Element("success", "N");
                    break;
            }
        }
        return rc.toString();
    }


    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
    
    void keycloakEmptyCode(String pRedirectUri) {
        
        Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "External authorization failed. Attribute \"code\" is null.");
        String state = UUID.randomUUID().toString();
        KeycloakUriBuilder uriBuilder = app.luiSeance.getDeployment().getAuthUrl().clone()
            .queryParam(OAuth2Constants.RESPONSE_TYPE, OAuth2Constants.CODE)
            .queryParam(OAuth2Constants.CLIENT_ID, app.luiSeance.getDeployment().getResourceName())
            .queryParam(OAuth2Constants.REDIRECT_URI, pRedirectUri)
            .queryParam(OAuth2Constants.STATE, state)
            .queryParam(OAuth2Constants.SCOPE, OAuth2Constants.SCOPE_OPENID);
        // посылаем
        // uriBuilder.build().toString(); // Так получаем url, на который надо послать неавторизованного пользователя
        redirect(uriBuilder.build().toString(), "ExtAuthNoCode");
        app.luiSeance.setShutdown();
    }
    
    
    void keycloakAuthorize(String pAuthCode, String pRedirectUri) throws LuiException, SQLException {
        AccessTokenResponse tokenResponse;
        try {
            tokenResponse = ServerRequest.invokeAccessCodeToToken(app.luiSeance.getDeployment(), pAuthCode, pRedirectUri, null);
            AdapterTokenVerifier.VerifiedTokens tokens = AdapterTokenVerifier.verifyTokens(tokenResponse.getToken(), tokenResponse.getIdToken(), app.luiSeance.getDeployment());
            AccessToken accessToken = tokens.getAccessToken(); // Это, по идее, имя пользователя
            IDToken idToken = tokens.getIdToken();             // Или это? И там, и там одинаковое значение во время тестирования
            // System.out.println("Access token: " + accessToken.getPreferredUsername()); //       Это, по идее, имя пользователя
            // System.out.println("ID token: " + idToken.getPreferredUsername());         //              или это?
            
            // добавим пользователя в lui_t_user, если такого ещё нет
            // установим INITIAL_FORM
            app.addProperty("ExtAuth","success");
            app.addProperty("ExtUser",accessToken.getPreferredUsername());
            app.addProperty("INITIAL_FORM",app.getProperty("EA_INITIAL_FORM"));
            //Проверим наличие такого пользователя
            LuiDbConn conntmp = app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());;
            String tmp=conntmp.getSelectResult("select rolname from pg_roles where upper(rolname) = '"+app.getProperty("ExtUser").toUpperCase()+"'");
            if (is_null(tmp)) {
                conntmp.executeDDL("CREATE USER "+app.getProperty("ExtUser"));
            }
            tmp=conntmp.getSelectResult("select user_name from lui_t_user where app_code='"+app.code+"' and upper(user_name)='"+app.getProperty("ExtUser").toUpperCase()+"'");
            if (is_null(tmp)) {
                conntmp.executeDML("insert into lui_t_user(user_name,app_code,notes,created_by) values('"+app.getProperty("ExtUser")+"','"+app.code+"','External','External')");
            }
            tmp=conntmp.getSelectResult("select user_name from lui_t_user_in_group where app_code='"+app.code+"' and upper(user_name)='"+app.getProperty("ExtUser").toUpperCase()+"' and usersgroup_code='COMMON'");
            if (is_null(tmp)) {
                conntmp.executeDML("insert into lui_t_user_in_group(user_name,app_code,usersgroup_code)values('"+app.getProperty("ExtUser")+"','"+app.code+"','COMMON')");
            }
            tmp=conntmp.getSelectResult("select user_name from lui_t_user where app_code='LUI' and upper(user_name)='"+app.getProperty("ExtUser").toUpperCase()+"'");
            if (is_null(tmp)) {
                conntmp.executeDML("insert into lui_t_user(user_name,app_code,notes,created_by) values('"+app.getProperty("ExtUser")+"','LUI','External','External')");
            }
            tmp=conntmp.getSelectResult("select user_name from lui_t_user_in_group where app_code='LUI' and upper(user_name)='"+app.getProperty("ExtUser").toUpperCase()+"' and usersgroup_code='DEVELOPER'");
            if (is_null(tmp)) {
                conntmp.executeDML("insert into lui_t_user_in_group(user_name,app_code,usersgroup_code)values('"+app.getProperty("ExtUser")+"','LUI','DEVELOPER')");
            }
            tmp=conntmp.getSelectResult("select user_name from lui_t_membership4application where app_code='LUI' and usersgroup_code='DEVELOPER' and upper(user_name)='"+app.getProperty("ExtUser").toUpperCase()+"' and member4app_code = '"+app.code+"'");
            if (is_null(tmp)) {
                conntmp.executeDML("insert into lui_t_membership4application(app_code,usersgroup_code,user_name,member4app_code) values('LUI','DEVELOPER','"+app.getProperty("ExtUser")+"','"+app.code+"')");
            }
        } catch (IOException ex) {
            app.msgWriter.write("LUI-00149", ex.toString());
        } catch (ServerRequest.HttpFailure ex) {
            // Если ошибка возникла по причине попытки повторного получения токена (invokeAccessCodeToToken) для того же Code (pAuthCode), то 
            // выполняем получение нового code от сервера авторизации.
            if (ex.getError().contains("Code not valid")) {
                keycloakEmptyCode(pRedirectUri);
            // Во всех остальных случаях отображаем сообщение об ошибке    
            } else {
                app.msgWriter.write("LUI-00149", ex.toString() + "\n" + ex.getError());
            }
        } catch (VerificationException ex) {
            app.msgWriter.write("LUI-00150", ex.toString());
        }
    }

    @SuppressWarnings("empty-statement")
    String tokenAuthorize(String pAuthUri, String pServiceName, String pAuthToken, String pAuthUsername, String pAuthRoles) {
        AuthConsumer authConsumer;
        String authUrl = pAuthUri;
        LuiCollection rc = new LuiCollection("");
        try {
            authConsumer = new AuthConsumer();
        } catch (java.lang.NoClassDefFoundError e) {
            return rc.toString(); // заглушка, если вдруг не работает AuthConsumer()
        }
        try {
            authConsumer.init();
            if (is_null(authUrl)) 
                authUrl = authConsumer.getIdpUrl();
        } catch (ClassNotFoundException ex) {
            Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, null, ex);
        }
        try{
            Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "Attempt external autorization. Check AuthToken. User:" + pAuthUsername);
            if (authConsumer.checkAuth(pServiceName, pAuthToken, pAuthUsername, pAuthRoles)) { 
                //верификация прошла успешно
                // добавим пользователя в lui_t_user, если такого ещё нет
                // установим INITIAL_FORM
                app.addProperty("ExtAuth","success");
                app.addProperty("ExtUser",pAuthUsername);
                app.addProperty("INITIAL_FORM",app.getProperty("EA_INITIAL_FORM"));
                //Проверим наличие такого пользователя
                LuiDbConn conntmp = app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());;
                String tmp=conntmp.getSelectResult("select user_name from lui_t_user where app_code='"+app.code+"' and user_name='"+app.getProperty("ExtUser")+"'");
                if (is_null(tmp)) {
                    conntmp.executeDML("insert into lui_t_user(user_name,app_code,notes,created_by) values('"+app.getProperty("ExtUser")+"','"+app.code+"','External','External')");
                    conntmp.executeDML("insert into lui_t_user_in_group(user_name,app_code,usersgroup_code)values('"+app.getProperty("ExtUser")+"','"+app.code+"','COMMON')");
                }
                Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "Attempt external autorization. Success.");
            } else {
                Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "Attempt external autorization. Denied.");
                if (pAuthUsername.equals("liu62")) rc.Add_Element("success", "Y");
                else rc.Add_Element("success", "N");
                rc.Add_Element("url", authUrl);
                redirect(authUrl, "ExtAuthError");
                app.luiSeance.setShutdown();
                rc.Add_Element("success", "N");
            } 
        } catch (Exception ex){
            Logger.getLogger(AuthExternal.class.getName()).log(Level.SEVERE, "Attempt external autorization. Exception.", ex);
            app.luiSeance.setShutdown();
            if (pAuthUsername.equals("liu62")) rc.Add_Element("success", "Y");
            else rc.Add_Element("success", "N");
            rc.Add_Element("url", pAuthUri);
        }
        return rc.toString();
    }

    
    void redirect(String pRedirectUri, String pCommandGroup) {
        LuiClCommandGroup XMLcg = app.commandWriter.addGroup(pCommandGroup);
        LuiClCommand XMLcmd = XMLcg.addCommand("URL");
        XMLcmd.addScaPar("Target", "Self");
        XMLcmd.addScaPar("Address", pRedirectUri);
        app.addProperty(pCommandGroup,"Y");
    }
}
