/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Выполнение подключения к БД или вычисление свойств имеющнгося соединения
 * Cинтаксис для execute:
 * DBConnect:Код_соединения;NAME:Название;MODE:COMMON/PERSONAL;USERNAME:имя_юзера;PASSWORD:пароль;URL:url;SCHEMA:схема;SQL:<оператор SQL выполняемый сразу после успешго входа>;
 * Вместо PASSWORD м.б. использовано слово PASSWORDC - это криптованный пароль (применяется в DTDBC)
* Синтаксис для compute:
DBConnect:Код_соединения
возвращяет строку вида NAME:Название;MODE:COMMON/PERSONAL;USERNAME:имя_юзера;URL:url;SCHEMA:схема;DBMS_TYPE:субд(PG/ORA);
или null если соединение не найдено
 */
package Standard;

import LuiCollection.LuiCollection;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import luiserver.LuiPlugin;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.RTrim;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import LuiDbInterface.LuiDbConn;
import luiserver.LuiAppConn;
import luiserver.LuiAppForm;

/**
 *
 * @author aliubushkin
 */
public class DBConnect implements LuiPlugin{
LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) {
        pString=RTrim(pString," ");
        pString=RTrim(pString,"\n");
        pString=RTrim(pString," ");
        pString=RTrim(pString,";");
        pString=RTrim(pString," ");
        
     if (app.userDbConnections.containsKey(pString)) {
        LuiAppConn ac= app.userDbConnections.get(pString);
        LuiCollection luiCol=new LuiCollection("");
        luiCol.Add_Element("CODE",ac.code);
        luiCol.Add_Element("NAME",ac.name);
        luiCol.Add_Element("MODE",ac.mode);
        luiCol.Add_Element("USERNAME",ac.username);
        luiCol.Add_Element("PASSWORDC", ac.getPWDC());
        luiCol.Add_Element("URL",ac.dbURL);
        luiCol.Add_Element("schema",ac.schema);
        luiCol.Add_Element("DBMS_TYPE",ac.connPool.getDBvendor());
        return luiCol.toString();
     } else {
         return null;
     }
     
    }

    @Override
    public void execute(String pString, LuiAppProp pProp)throws LuiException, SQLException {
  
        LuiCollection c=new LuiCollection("DBConnect:"+pString);
        String connCode=c.Get_Element("DBConnect");
        String pwd="";
        String mode=NVL(c.Get_Element("MODE"),"PERSONAL");
        if (c.containsKey("PASSWORD")){
        pwd=c.Get_Element("PASSWORD");
        } else {
         if (c.containsKey("PASSWORDC")){
        pwd=c.Get_Element("PASSWORDC");
        pwd=app.crypter.decrypt(pwd);
         }
        }
     LuiAppConn ac;   
    if (mode.equals("COMMON")){
        ac=app.addUserDbConnection(connCode,
                c.Get_Element("NAME"),
                "PERSONAL",
                c.Get_Element("USERNAME"),
                pwd,
                c.Get_Element("URL"),
                c.Get_Element("SCHEMA")
                );
        ac.disconnect();
        app.userDbConnections.remove(connCode);
    }
          ac=app.addUserDbConnection(connCode,
                c.Get_Element("NAME"),
                mode,
                c.Get_Element("USERNAME"),
                pwd,
                c.Get_Element("URL"),
                c.Get_Element("SCHEMA")
                );
    LuiAppForm frm=pProp.ownerObj.getForm();
    if (frm==null) {
    String curFrmID=app.getVal("FormID");        
    if (app.forms.containsKey(curFrmID)){
       frm=app.forms.get(curFrmID);
    }
    }
     if (app.properties.containsKey("Connection")){
       if (is_null(app.getProperty("Connection"))) {
         app.setProperty("Connection",connCode);  
       }  
     } else // если ваще такой проперти нет
         app.addProperty("Connection",connCode);
     if (frm!=null) frm.setConnectionProperty("");
        if (c.containsKey("SQL")){
            String stmt=c.Get_Element("SQL");
            if (is_not_null(stmt)) {
            LuiDbConn lc=null;
                if (ac.getSessionMode()=='C'){
                  lc=ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                } else lc=ac.currentConnection;
            if (lc==null){ // если соединение так и не нашлось, то
                app.msgWriter.write("LUI-00020",". Lui:Null pointer detected!");
              }
            try {
                lc.getSelectResult(stmt);
                if (!lc.getConnection().getAutoCommit()){
                lc.commit();}
            } finally {
                if (ac.getSessionMode()=='C') {
                    //if (lc!=null) {
                        ac.connPool.SetFree(lc);
                    //}
                }
            }
            }
        }
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {

    }
 @Override
    public String quoteMask(String prefixString, String Value) {
        return Value.replace(";","#$");
    }        
}
