/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Вычисление и присваивание свойства элемента формы
*  Вычисление
* ElementProperty:<код элемента>.<код свойства>[|C]
[|C] означает что надо взять currentValue
* Пример: ElementProperty:CODE.NAME
*  Присваивание
* ElementProperty:<код элемента>.<код свойства>=<значение>
 */
package Standard;

import LuiCollection.LuiString;
import static LuiCollection.LuiString.LTrim;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.like;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import luiserver.LuiAppElement;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;

/**
 *
 * @author aliubushkin
 */
public class ElementProperty implements LuiPlugin{
LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    
    public String comp(String pString, LuiAppProp pProp,boolean pFindOnly) throws LuiException, SQLException {
    LuiAppForm frm=pProp.ownerObj.getForm();
    if (frm==null) {
    String curFrmID=app.getVal("FormID");        
    if (app.forms.containsKey(curFrmID)){
       frm=app.forms.get(curFrmID);
    } else return "";
    }
    
    if (like(pString,"%.%.%")){
     String FrmID=pString.split("\\.",2)[0];
     if (app.forms.containsKey(FrmID)){
        frm=app.forms.get(FrmID);
        pString=pString.split("\\.",2)[1];
     }
    }
    LuiAppElement el=null;
    String elCode=null;
    String propCode="";
    elCode=LuiString.Extract_Word(pString, 1,".");
    propCode=LuiString.Extract_Word(pString, 2,".");
    boolean cv=false;
    if (propCode.indexOf("|C")>0){
       propCode=propCode.split("\\|")[0]; 
       cv=true;
    }
    if (frm.elements.containsKey(elCode)){ // если есть такой элемент
        el=frm.elements.get(elCode);
    if (el.properties.containsKey(propCode)) { // если у указанного элемента есть указанное св-во
        if (!cv || pFindOnly)
        el.properties.get(propCode).setDepend(pProp); // то говорим, что оно определяет свойство pProp
        if (pFindOnly) return null;
        if (cv)return NVL(el.properties.get(propCode).currentValue,"");
          return NVL(el.properties.get(propCode).getComputedProperty(),"");
    } else { // если у указанного элемента нет указанного св-ва - создаём его с пустым значением
        el.addProperty(propCode, "").setDepend(pProp);
        if (pFindOnly) return null;
        return "";
        
    }
    } return "";
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
    LuiAppForm frm=pProp.ownerObj.getForm();
    if (frm==null) {
    String curFrmID=app.getVal("FormID");        
    if (app.forms.containsKey(curFrmID)){
       frm=app.forms.get(curFrmID);
    } else return;
    }
    LuiAppElement el=null;
    String elCode=null;
    String propCode="";
    String propVal="";
    // выясняем код свойства
        pString=LTrim(pString);
        int eq=pString.indexOf("=");
        String pc=pString.substring(0, eq);
        // выясняем значение
        propVal=pString.substring(eq+1);
        
    
    String w[]=pc.split("\\.");
    if (w.length==2) {
        elCode=w[0];
        propCode=w[1];
    } 
    if (w.length==3){
      if (!app.forms.containsKey(w[0])) return; 
        frm=app.forms.get(w[0]);
        elCode=w[1];
        propCode=w[2];
    }
   
    
    if (frm.elements.containsKey(elCode)){ // если есть такой элемент
        el=frm.elements.get(elCode);
        el.addProperty(propCode, propVal);
    }
    }
    
    

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        comp(pString,pProp,true);
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        return comp(pString,pProp,false);
    }
    
}
