/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 */
package Standard;

import LuiCollection.LuiCollection;
import LuiCollection.LuiString;
import static LuiCollection.LuiString.Add_If_Not_Null;
import static LuiCollection.LuiString.GetIcoFromSquareBracket;
import static LuiCollection.LuiString.MaskAncillaryChars;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.RTrim;
import static LuiCollection.LuiString.Trim;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import luiserver.LuiAppElement;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;

import luiserver.LuiApplication;
import luiserver.LuiClCommand;
import luiserver.LuiClCommandGroup;
import luiserver.LuiParser;
import luiserver.LuiPlugin;

/**
 *
 * @author aliubushkin
 */
public class Help implements LuiPlugin{
    LuiApplication app;
StringBuilder content;
LuiAppForm frm;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
    frm=pProp.ownerObj.getForm();
    if (frm==null) {
    String curFrmID=app.getVal("FormID");        
    if (app.forms.containsKey(curFrmID)){
       frm=app.forms.get(curFrmID);
    } else return "";
    }        
    content=new StringBuilder();        
    line("<html>");
    line("<head>");
    line("<title>");line(pString);line("</title>");
    //line("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=windows-1251\"/>");
    line("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"/>");
    // читаем css
    String css="";
    LuiDbConn lc=null;
    try{
    lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());  
    css=lc.getSelectResult("select lui_f_get_application_property('"+app.code+"','HELP_CSS','"+app.NLS.currentLang.code+"')");
    } finally {
        if (lc!=null) {
        app.metaConnPool.SetFree(lc);
        lc=null;
        }
    }
    //int odoa=css.indexOf("\r\n");
    content.append(css);
    line("<link rel=\"stylesheet\" type=\"text/css\" href=\""+app.getProperty("ApplicationURL")+"resources/desktop/images/icons.css\">");
    line("<link rel=\"stylesheet\" type=\"text/css\" href=\""+app.getProperty("ApplicationURL")+"custom/default-desktop.css\">");
    line("<link rel=\"stylesheet\" type=\"text/css\" href=\""+app.getProperty("ApplicationURL")+"custom/"+app.getProperty("STYLE_CURRENT")+"-desktop.css\">");
    
    line("</head>");        
    line("<body>");
    String ftype=frm.getType();
    switch (ftype){
        case "BLANK":
            ftype=app.NLS.NLS("RUS:Бланк;ENG:Blank;");
        break;
        case "LIST":
            ftype=app.NLS.NLS("RUS:Список;ENG:List;");
        break;
        case "MENU":
            ftype=app.NLS.NLS("RUS:Меню;ENG:Menu;");
        break;        
        case "CHART":
            ftype=app.NLS.NLS("RUS:Диаграмма;ENG:Chart;");
        break;        
        case "TREE":
            ftype=app.NLS.NLS("RUS:Дерево;ENG:Tree;");
          if (!(frm.getVal("RootFormID").equals(frm.getFormID()))) {
            ftype=app.NLS.NLS("RUS:Ветка;ENG:Tree branch;");  
          }
        break;
    }
    String formNotes="";
/*    lc=null;
    try {
    lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID());
    formNotes=lc.getSelectResult("SELECT value FROM lui_t_value v  where v.property_code='FORM_NOTES' and v.app_code='"+app.code+"' and v.form_code='"+frm.code+"'");
    if (formNotes==null) formNotes="RUS:;ENG:;";
    } finally {
        if (lc!=null)
        app.metaConnPool.SetFree(lc);
    }
    frm.addProperty("FORM_NOTES", formNotes);
    formNotes=frm.getProperty("FORM_NOTES");
    frm.properties.remove("FORM_NOTES");*/
    formNotes=GetVal("","FORM_NOTES");
    line(ftype);
    line("<div class=\"Help_Title\">"+pString+"</div>"+app.NLS.NLS(formNotes));
    line("<div id=\"InParam-"+frm.getFormID()+"\" class=\"Container Hide\">");
    line("<div class=\"TitleShow\" onclick=\"Hide($('InParam-"+frm.getFormID()+"'))\">"+app.NLS.NLS("RUS:Техническая информация;ENG:Technical information;")+"</div>");
    line("<div class=\"TitleHidden\" onclick=\"Show($('InParam-"+frm.getFormID()+"'))\">"+app.NLS.NLS("RUS:Отобразить техническую информацию;ENG:Show technical information;")+"</div>");
    line("<div class=\"Hided\">");
    String trpar=frm.getVal("TransitParam");
    if (is_not_null(trpar)){ //Входные параметры
        line("<div class=\"Title\">"+app.NLS.NLS("RUS:Входные параметры;ENG:Input parameters;")+"</div>");  
     line("(Явно указанные при вызове параметры)");   
     line("<table class=\"TabInfo\">");
     line("<tr class=\"Title\">"+
             "<th>"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Значение;ENG:Value;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Пояснения;ENG:Description;")+"</th>"+"</tr>");
    LuiCollection trparcol=new LuiCollection(trpar);
    for (String pc:trparcol.keySet()){
                    line("<tr>");
                    line("<td>"+LuiParser.reMask(pc)+"</td>");
                    line("<td>"+LuiParser.reMask(trparcol.get(pc))+"</td>");
                    line("<td>"+LuiParser.reMask(NVL(frm.elements.get(pc).getProperty("Name"),frm.elements.get(pc).getProperty("DESCRIPTION")))+"</td>");
                    line("</tr>");
    }
    line("</table>");
    }//Входные параметры
line("<div class=\"Title\">"+app.NLS.NLS("RUS:Техническая информация;ENG:Technical information;")+"</div>"); // Техническая информация
line("<table class=\"TabInfo\">");
line("<tr><th>"+app.NLS.NLS("RUS:Идентификатор экземпляра формы в памяти сервера;ENG:The ID of the form instance in memory on the server;")+"</th><td>"+frm.getFormID()+"</td></tr>");
line("<tr><th>"+app.NLS.NLS("RUS:Код формы (для прав);ENG:Form code for access control;")+"</th><td>"+frm.getCode()+"</td></tr>");
 if (is_not_null(frm.getProperty("MetadataCode"))){
     line("<tr><th>"+app.NLS.NLS("RUS:Код метаданных;ENG:Form code in database;")+"</th><td>"+LuiParser.reMask(frm.getProperty("MetadataCode"))+"</td></tr>");
     line("<tr><th>"+app.NLS.NLS("RUS:Код конфигурации;ENG:Cofiguration code;")+"</th><td>"+LuiParser.reMask(frm.getProperty("ConfigCode"))+"</td></tr>");
     line("<tr><th>"+app.NLS.NLS("RUS:Название конфигурации;ENG:Cofiguration name;")+"</th><td>"+LuiParser.reMask(frm.getProperty("ConfigName"))+"</td></tr>");
     String usrconf=frm.getVal("UserConfigurationName");
     if (is_not_null(usrconf)){
         line("<tr><th>"+app.NLS.NLS("RUS:Название конфигурации;ENG:Cofiguration name;")+"</th><td>"+LuiParser.reMask(usrconf)+"</td></tr>");
     }
 }
line("</table>");
 if (frm.getType().equals("LIST")||frm.getType().equals("TREE")) {// Отмеченные строки
     line("<div class=\"Title\">");
     if (frm.getVal("Inversion").equals("Y"))
         line(app.NLS.NLS("RUS:Разотмеченные ;ENG:Unmarket ;"));
         else     
         line(app.NLS.NLS("RUS:Отмеченные ;ENG:Market ;"));
     switch (frm.getType()){
         case "LIST":
             line(app.NLS.NLS("RUS:строки;ENG:rows;"));break;
         case "TREE":
             line(app.NLS.NLS("RUS:узлы;ENG:nodes;"));break;
         default:
             line(app.NLS.NLS("RUS:элементы;ENG:elements;"));break;    
     }
     line("</div>");
     line("<table class=\"TabInfo\">");
     line("<tr class=\"Title\">");
     line("<th>"+app.NLS.NLS("RUS:Значения ключей;ENG:Key values;")+"</th>");
     line("</tr>");
     for (LuiCollection kv:frm.checkRows.keySet()){
     line("<tr><td>"+kv.toString()+"</td></tr>");    
     }
     
  line("</table>");
 }
line("</div>");
line("</div>");
if ("CHART".equals(frm.getType())){ // если диаграмма
 // действия уровня формы
for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
     if (el.getType().equals("ACTION") 
           && !(like(el.getCode(),"SYSMENU$%"))){
         
     }
 
} 
}
if ("LIST".equals(frm.getType())){ // если список
 line("<div id=\"Cols-"+frm.getFormID()+"\" class=\"Container\">");
 line("<div class=\"Title\">"+app.NLS.NLS("RUS:Столбцы списка;ENG:List columns;")+"</div>");
 line("<table class=\"TabInfo\">");
 line("<tr class=\"Title\">"+
      "<th class=\"Switchable Code\">"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
      "<th>"+app.NLS.NLS("RUS:Заголовок;ENG:Header;")+"</th>"+
      "<th>"+app.NLS.NLS("RUS:Описание;ENG:Description;")+"</th>"+
      "</tr>");    
 int cnt=help4groupOfList("Form");
       line("</table>");
 if (cnt>0) {
       addSwitch("Cols-"+frm.getFormID(),"HC",app.NLS.NLS("RUS:Отобразить невидимые столбцы Списка;ENG:Show invisible columns;"),
               app.NLS.NLS("RUS:Скрыть невидимые столбцы;ENG:Hide invisible columns"));
 }
       addSwitch("Cols-"+frm.getFormID(),"Code",app.NLS.NLS("RUS:Отобразить коды столбцов;ENG:Show column codes;"),
               app.NLS.NLS("RUS:Скрыть коды столбцов;ENG:Hide column codes;"));
 line("</div>");
 line("<div id=\"Act-"+frm.getFormID()+"\" class=\"Container\">");
 line("<div class=\"Title\">Действия списка</div>");
 line("<table class=\"TabInfo\">");
line("<tr class=\"Title\">"+
             "<th class=\"Switchable Code\">"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Название;ENG:Name;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Область применения;ENG:Usage area;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Способ активации;ENG:How to execute;")+"</th>"+     
             "<th>"+app.NLS.NLS("RUS:Описание;ENG:Description;")+"</th>"+     
             "</tr>");
 cnt=0;
for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
   if (el.getType().equals("ACTION") 
           && !(like(el.getCode(),"SYSMENU$%"))
           && !(like(el.getCode(),"COLMENU$%"))
           && !(like(el.getCode(),"QBE$ITEM%"))
           ){
    if (!(GetVal(el.getCode(),"ACTIVE","START","").equals("N"))){
      String TStr="";
      if (!el.isActionActive()){
          cnt++;
          TStr=" class=\"Switchable HC\"";
      }
            line("<tr"+TStr+"><th class=\"Switchable Code NoWrap\">"+MaskAncillaryChars(el.getCode())+"</th>");// Код
            line("<td class=\"NoWrap\">"+MaskAncillaryChars(GetVal(el.getCode(), "PROMPT"))+"</td>"); // Название      
      if (el.getParent().equals("Form")) {
        TStr="";
        switch(GetVal(el.getCode(),"ACTIVE")){
            case "Y":
                line("<td>"+app.NLS.NLS("RUS:Весь список;ENG:List;")+"</td>");
                break;
            case "EXIST":
                line("<td>"+app.NLS.NLS("RUS:Строка списка;ENG:Row;")+"</td>");
                break;
            default:
                line("<td>"+app.NLS.NLS("RUS:Недоступно;ENG:Unavailable;")+"</td>");
                break;
        }
      }else {
              TStr=NVL(Add_If_Not_Null(GetVal(el.getParent(), "CAPTION"),"&laquo;","&raquo;"),el.getParent());
              line("<td>"+app.NLS.NLS("RUS:Ячейка столбца ;ENG:Cell ;")+TStr+"</td>");
              TStr=app.NLS.NLS("RUS:в ячейке столбца ;ENG:at cell of ;")+TStr;          
      }
      line("<td class=\"NoWrap\">"); // Способ активации
      String atext="";
      switch(GetVal(el.getCode(),"DISPLAY")){
          case "Y":
          case "EXIST":  //Локальное меню
           if (is_not_null(GetVal(el.getCode(),"PROMPT"))) {
             //не понял селект на 687 
             //Select Count(*) into N from Table(UI_P_VIRTUAL.Loop_WorkID) where Parent=WorkID and UI_P_VIRTUAL.GetVal(WorkID,'CallAction')=A.Code;
             atext=app.NLS.NLS("RUS:Пункт локального меню;ENG:Local menu item;")+"<br>";
           }   
          break;
          case "":
          case "N": 
          break;
          default: //размножение???
            atext=app.NLS.NLS("RUS:Пункты локального меню;ENG:Local menu items;")+"<br>";  
          break;
      }
      if (is_not_null(GetVal(el.getCode(),"ICON"))){
          //atext=app.NLS.NLS("RUS:Иконка;ENG:Icon;")+" &laquo;"+GetVal(el.getCode(),"ICON")+"&raquo; ";
          atext=app.NLS.NLS("RUS:Иконка;ENG:Icon;")+" "+ico(GetVal(el.getCode(),"ICON"))+" ";
        switch (el.getParent()) {
            case "Form":
            atext=atext+app.NLS.NLS("RUS:в верхней панели инструментов;ENG:in the top toolbar;");
            break;
            case "Item":
            default:
            atext=atext+app.NLS.NLS("RUS:в ячейке;ENG:in the cell;");    
            break;
        }
        atext=atext+"<br>";
      }
      if (is_not_null(GetVal(el.getCode(),"BUTTON"))){
       atext=atext+app.NLS.NLS("RUS:Кнопка;ENG:Button;")+" &laquo;"+GetVal(el.getCode(),"BUTTON")+"&raquo;<br>";
      }
      if (like(GetVal(el.getCode(),"ROLE"),"%DBLCLICK%")){
       atext=atext+app.NLS.NLS("RUS:Клавиша;ENG:Key;")+" &laquo;Enter&raquo;<br>"; 
      }
      if (like(GetVal(el.getCode(),"ROLE"),"%ADD%")){
       atext=atext+app.NLS.NLS("RUS:Клавиша;ENG:Key;")+" &laquo;F6&raquo;<br>"; 
      }
      if (like(GetVal(el.getCode(),"ROLE"),"%DELETE%")){
       atext=atext+app.NLS.NLS("RUS:Клавиша;ENG:Key;")+" &laquo;Shift+F6&raquo;<br>"; 
      }
      if (like(GetVal(el.getCode(),"ROLE"),"%CANCEL%")){
       atext=atext+app.NLS.NLS("RUS:Клавиша;ENG:Key;")+" &laquo;Esc&raquo;<br>"; 
      }
      line(atext+"</td>");
      atext="";
      switch(GetVal(el.getCode(),"APPLY_TO")){
          case "PROGRESS":
              atext=app.NLS.NLS("RUS:Действие может применяться к нескольким [[выделенным]] записям;ENG:The action can be applied to multiple [[marked]] rows;");
          break;
          default:
          break;
      }
      atext=MergeText(MergeText(ifExist(GetVal(el.getCode(),"PROMPT"),GetVal(el.getCode(),"EXPLANATION")),GetVal(el.getCode(),"ACTION_NOTES")),atext);
      String role=GetVal(el.getCode(),"ROLE");
      if (like(role,"%DBLCLICK%")){
          atext=MergeText(atext,app.NLS.NLS("RUS:Для быстрого выполнения действия можно сделать [[двойной щелчок мышью]] "+
          NVL(TStr,"или нажать клавишу [[Enter]]")+        
                  ";ENG:To perform an action quickly, you can do [[doubleclick]] "+
          NVL(TStr,"or press [[Enter]] key")+                  
                  ";"));
      }
      if (like(role,"%ADD%")){
          atext=MergeText(atext,app.NLS.NLS("RUS:Для быстрого выполнения действия можно нажать клавишу [[F6]];ENG:To perform an action quickly, you can press [[F6]] key;"));
      }
      if (like(role,"%DELETE%")){
          atext=MergeText(atext,app.NLS.NLS("RUS:Для быстрого выполнения действия можно нажать клавишу [[Shift+F6]];ENG:To perform an action quickly, you can press [[Shift+F6]] key;"));
      }
      if (like(role,"%CLOSE%")){
          atext=MergeText(atext,app.NLS.NLS("RUS:Служебное действие при закрытии списка;ENG:Service action when list closing;"));          
      }
      if (like(role,"%AUTO%")){
          atext=MergeText(atext,app.NLS.NLS("RUS:Служебное действие для отображения детальных форм при навигации по строкам списка;ENG:Service action to display detailed forms when navigating through the list rows;"));
      }
      if (like(role,"%POSTEDIT%")){
          atext=MergeText(atext,app.NLS.NLS("RUS:Служебное действие для обработки изменения пользователем данных в ячейке списка;ENG:Service action to handle user changes to data in a list cell;"));          
      }
      
          
      if (is_not_null(GetVal(el.getCode(),"ICON")) && el.getParent().equals("Form")) {
      atext=MergeText(atext,app.NLS.NLS("RUS:Действие может активизироваться [[кнопкой]] на панели инструментов;ENG:The action can be activated [[button]] on the toolbar;"));
      }
      atext=MergeText(atext,".");
      atext=RTrim(atext);
      atext=LuiString.MaskAncillaryChars(atext);
      atext="<td>"+atext+"</td>";
      line(atext);
    }
   }
} //конец цикоа по действиям
      line("</table>");
if (cnt>0){
  addSwitch("Act-"+frm.getFormID(),"HC",
          app.NLS.NLS("RUS:Отобразить недоступные действия списка;ENG:Show unavailable actions;"),
          app.NLS.NLS("RUS:Скрыть недоступные действия;ENG:Hide unavailable actions;"));
}
addSwitch("Act-"+frm.getFormID(),"Code",
        app.NLS.NLS("RUS:Отобразить коды действий;ENG:Show action codes;"),
        app.NLS.NLS("RUS:Скрыть коды действий;ENG:Hide action codes;")
        );
    line("</div>");

} // если список
if ("MENU".equals(frm.getType())){ // если меню
  line("<div id=\"Fields-"+frm.getFormID()+"\" class=\"Container\">");
  line("<div class=\"Title\">"+app.NLS.NLS("RUS:Пункты меню;ENG:Menu items;")+"</div>");
  line("<table class=\"TabInfo\">");    
}line("<table class=\"TabInfo\">");
    line("<tr class=\"Title\">"+
               "<th class=\"Switchable Code\">"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Иконка;ENG:Icon;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Название;ENG:Name;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Способ выполнения;ENG:How execute;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Описание;ENG:Description;")+"</th>"+
               "</tr>");
 help4groupOfMenu("Form","");
 line("</table>");
     addSwitch("Fields-"+frm.getFormID(),"Code",app.NLS.NLS("RUS:Отобразить коды;ENG:Show codes;"),app.NLS.NLS("RUS:Скрыть коды полей;ENG:Hide field codes;"));
      line("</div>");
    line("</div>");
if ("BLANK".equals(frm.getType())){ // если бланц
    line("<div id=\"Fields-"+frm.getFormID()+"\" class=\"Container\">");
    line("<div class=\"Title\">"+app.NLS.NLS("RUS:Поля бланка;ENG:Form fields;")+"</div>");
    line("<table class=\"TabInfo\">");
    line("<tr class=\"Title\">"+
               "<th class=\"Switchable Code\">"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Группа;ENG:Group;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Подпись;ENG:Prompt;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Способ редактирования;ENG:Edit mode;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Описание;ENG:Description;")+"</th>"+
               "</tr>");
 help4groupOfBlank("Form");
 line("</table>");
     addSwitch("Fields-"+frm.getFormID(),"Code",app.NLS.NLS("RUS:Отобразить коды полей;ENG:Show field codes;"),app.NLS.NLS("RUS:Скрыть коды полей;ENG:Hide field codes;"));
      line("</div>");
    line("</div>");
    line("<div id=\"Actions-"+frm.getFormID()+"\" class=\"Container\">");
    line("<div class=\"Title\">"+app.NLS.NLS("RUS:Действия бланка;ENG:Form actions;")+"</div>");
   line("<table class=\"TabInfo\">");
          line("<tr class=\"Title\">"+
               "<th class=\"Switchable Code\">"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Название;ENG:Name;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Способ активации;ENG:How to execute;")+"</th>"+     
             "<th>"+app.NLS.NLS("RUS:Описание;ENG:Description;")+"</th>"+     
             "</tr>");
String atext="";          
for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
   if (el.getType().equals("ACTION") 
           && !(like(el.getCode(),"SYSMENU$%"))
           ){
     if (!(GetVal(el.getCode(),"DISPLAY").equals("Y"))) continue; 
     if (GetVal(el.getParent(),"EType").equals("FIELD")){
      if (is_not_null(GetVal(el.getParent(),"ForNLSSupport")))continue;   
     }
     line("<tr>");
     line("<td class=\"Switchable Code NoWrap\">"+MaskAncillaryChars(el.getCode())+"</td>"); // Код
     line("<td class=\"NoWrap\">"+MaskAncillaryChars(NVL(GetVal(el.getCode(),"PROMPT"),GetVal(el.getCode(),"Name")))+"</td>"); // Названи
     line("<td>"); // Способ активации
     atext=app.NLS.NLS("RUS:Кнопка ;ENG:Button ;")+"&laquo;"+GetVal(el.getCode(),"PROMPT")+"&raquo;#<br>";
     if (GetVal(el.getParent(),"EType").equals("FIELD")){
       if (is_not_null(GetVal(el.getCode(),"PROMPT"))){
       // не понял select на 823
       // Select Count(*) into N from Table(UI_P_VIRTUAL.Loop_WorkID) where Parent=WorkID and UI_P_VIRTUAL.GetVal(WorkID,'CallAction')=A.Code;
       atext=app.NLS.NLS("RUS:Локальное меню в поле ;ENG:Local menu for field ;")+"&laquo;"+
               NVL(GetVal(el.getParent(),"PROMPT"),GetVal(el.getParent(),"Name"))+"&raquo;#<br>";       
       }
     }
     if (like(GetVal(el.getCode(),"ROLE"),"%DBLCLICK%")){// Enter
       atext=atext+app.NLS.NLS("RUS:Клавиша;ENG:Key;")+ " Enter<br>";
     }
     String gname="";
    if (!("Form".equals(el.getParent()))){
        gname=NVL(GetVal(el.getParent(),"TITLE"),GetVal(el.getParent(),"Name"));
        gname=app.NLS.NLS("RUS: в группе ;ENG: in the group ;")+gname;
    }
    line(atext.replace("#",gname)+"</td>");
    String role=GetVal(el.getCode(),"ROLE");
    atext="";
    if (like(role,"%DBLCLICK%")){
      atext=app.NLS.NLS("RUS:Действие выполняется при нажатии [[Enter]];ENG:The action is performed by pressing [[Enter]] key;");
    }
    if (like(role,"%CLOSE%")){
      atext=app.NLS.NLS("RUS:Технологическое действие, выполняющееся при закрытии бланка;ENG:Service action is performed when form closing;");
    }
    if (like(role,"%MATCH%")){
      atext=app.NLS.NLS("RUS:Действие выполняется в случае соответствия значения в поле шаблону;ENG:The action is performed if the value in the field matches the template;")+
              " [["+GetVal(el.getCode(),"PATTERN")+"]]";
    }
    atext=MergeText(
            MergeText(ifExist(GetVal(el.getCode(),"PROMPT"),
                    GetVal(el.getCode(),"EXPLANATION")),
                    GetVal(el.getCode(),"ACTION_NOTES"))
            ,atext);
    atext=LuiString.MaskAncillaryChars(atext);
    line("<td>"+atext+"</td>");
    line("</tr>");
   }
} //конец цикла по действиям бланка  
 line("</table>");
addSwitch("Actions-"+frm.getFormID(),"Code",app.NLS.NLS("RUS:Отобразить коды действий;ENG:Show action codes;"),app.NLS.NLS("RUS:Скрыть коды действий;ENG:Hide action codes;"));
line("</div>");
line("</div>");
 
} //если бланц
if ("TREE".equals(frm.getType())){ // если дуб
    line("<div id=\"Actions-"+frm.getFormID()+"\" class=\"Container\">");
    line("<div class=\"Title\">"+app.NLS.NLS("RUS:Действия с узлами;ENG:Node actions;")+"</div>");
        line("<table class=\"TabInfo\">");
          line("<tr class=\"Title\">"+
               "<th class=\"Switchable Code\">"+app.NLS.NLS("RUS:Код;ENG:Code;")+"</th>"+
               "<th>"+app.NLS.NLS("RUS:Название;ENG:Name;")+"</th>"+
             "<th>"+app.NLS.NLS("RUS:Способ активации;ENG:How to execute;")+"</th>"+     
             "<th>"+app.NLS.NLS("RUS:Описание;ENG:Description;")+"</th>"+     
             "</tr>");
for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
   if (el.getType().equals("ACTION") 
           && !(like(el.getCode(),"SYSMENU$%"))
           && !(like(el.getCode(),"COLMENU$%"))
           ){
 if (el.getParent().equals("Form")){
 if (!(GetVal(el.getCode(),"ACTIVE","START","").equals("N"))){
line("<tr>");     
line("<td class=\"Switchable Code NoWrap\">"+MaskAncillaryChars(el.getCode())+"</td>"); // Код
line("<td class=\"NoWrap\">"+MaskAncillaryChars(NVL(GetVal(el.getCode(),"PROMPT"),GetVal(el.getCode(),"Name")))+"</td>"); // Названи
line("<td>"); // Способ активации
String atext="";
if (like(GetVal(el.getCode(),"ROLE"),"%AUTO")) {
   atext=app.NLS.NLS("RUS:При навигации по узлам;ENG:When you navigate through the nodes;")+"<br>";
}else if (!("N".equals(GetVal(el.getCode(),"DISPLAY","START","")))){
    if (is_not_null(GetVal(el.getCode(),"PROMPT"))){
     atext=app.NLS.NLS("RUS:Локальное меню;ENG:Local menu;")+"<br>";   
    }
}
if (like(GetVal(el.getCode(),"ROLE"),"%DBLCLICK%")){
   atext=app.NLS.NLS("RUS:Клавиша Enter или двойной щелчок левой кнопкой мыши на узле;ENG:Enter key or double-click the left mouse button on the node;")+"<br>";    
}
line(atext+"</td>");
line("<td>"+LuiString.MaskAncillaryChars(
                  MergeText(ifExist(GetVal(el.getCode(),"PROMPT"),GetVal(el.getCode(),"EXPLANATION")),GetVal(el.getCode(),"ACTION_NOTES"))
                )+"</td>");
line("</tr>");     
 }
 }
   }
} //конец действий
line("</table>");
addSwitch("Actions-"+frm.getFormID(),"Code",app.NLS.NLS("RUS:Отобразить коды действий;ENG:Show action codes;"),app.NLS.NLS("RUS:Скрыть коды действий;ENG:Hide action codes;"));
    line("</div>");
    line("</div>");
} // если баобаб
//общие действия и функции

  line("<div id=\"Common-"+frm.getFormID()+"\" class=\"Container Hide\">");
  line("<div class=\"TitleShow\" onclick=\"Hide($('Common-"+frm.getFormID()+"'))\">"+app.NLS.NLS("RUS:Общая функциональность;ENG:Common functions;")+"</div>");
  line("<div class=\"TitleHidden\" onclick=\"Show($('Common-"+frm.getFormID()+"'))\">"+app.NLS.NLS("RUS:Отобразить общую функциональность;ENG:Show common functions;")+"</div>");
  line("<div class=\"Hided\">");
  String frmtype="";
  switch (frm.getType()){
      case "LIST" :
          frmtype=app.NLS.NLS("RUS:Действия, возможные в каждом cписке;ENG:Functions available in any list;");
          break;
      case "BLANK" :
          frmtype=app.NLS.NLS("RUS:Действия, возможные в каждом бланке;ENG:Functions available in any blank;");
          break;          
      case "TREE":
          frmtype=app.NLS.NLS("RUS:Действия, возможные в каждом дереве;ENG:Functions available in any tree;");
          break;          
  }
  line(frmtype);
  line("<table class=\"TabInfo\">");
  if ("LIST".equals(frm.getType())){ // если cписок
       line(app.NLS.NLS("RUS:<tr><th>Щелчок левой кнопкой мыши на ячейке</th><td>Сделать строку и ячейку в позиции курсора текущей (контекстной)</td></tr>;ENG:<tr><th>Click with the left mouse button in the cell</th><td>Make the row and the cell at the cursor position of the current (context)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Стрелка вверх]]</th><td>Сделать текущей предыдущую строку (выше на одну той, что текущая сейчас)</td></tr>;ENG:<tr><th>Key [[UpArrow]]</th><td>Make the current previous row (one higher than the current one)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Стрелка вниз]]</th><td>Сделать текущей следующую строку (ниже на одну той, что текущая сейчас)</td></tr>;ENG:<tr><th>Key [[DownArrow]]</th><td>Make the current one the next line (below the one that the current now)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Tab]]</th><td>Сделать текущей следующую ячейку (правее той, что текущая сейчас)</td></tr>;ENG:<tr><th>Key [[Tab]]</th><td>Make the current next cell (to the right of the current one)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Shift-Tab]]</th><td>Сделать текущей предыдущую ячейку (левее той, что текущая сейчас)</td></tr>;ENG:<tr><th>Key [[Shift-Tab]]</th><td>make the current previous cell (to the left of the current one)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[PageUp]]</th><td>Сделать текущей строку на несколько строк выше той, что текущая сейчас (листание страницы вверх)</td></tr>;ENG:<tr><th>Key [[Page Up]]</th><td>Make the current row a few rows higher than the current one (page up)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[PageDown]]</th><td>Сделать текущей строку на несколько строк ниже той, что текущая сейчас (листание страницы вниз)</td></tr>;ENG:<tr><th>Key [[PageDown]]</th><td>Make the current row a few rows below the current one (page down)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Enter]] или Двойной щелчок мышью</th><td>Выполнение первого по порядку, возможного в настоящий момент типового действия для текущей строки (и текущего поля)</td></tr>;ENG:<tr><th>Key [[Enter]] or Double clicking with the mouse</th><td>implementation of the first order, is possible in the present model the actions for the current row (and current field)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Insert]]</th><td>Выполнение первого по порядку, возможного в настоящий момент действия [[Добавления]]</td></tr>;ENG:<tr><th>Press [[Insert]]</th><td>to Execute the first in order that is currently possible [Add]]</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Delete]]</th><td>Выполнение первого по порядку, возможного в настоящий момент действия [[Удаления]]</td></tr>;ENG:<tr><th>Key [[Delete]]</th><td>Implementation of the first order, it is possible to currently [[Remove]]</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Щелчок правой кнопкой мыши или [[Alt-X]]в ячейке Списка</th><td>Отображение списка действий, применимых в настоящий момент ко всему списку, к текущей строке и текущей ячейке в этой строке (локальное меню)</td></tr>;ENG:<tr><th>Right-click in List cell or key [[Alt-X]]</th><td>Display a list of actions that are currently applicable to the entire list, the current row, and the current cell in that row (local menu)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Щелчок левой кнопкой мыши на заголовке столбца</th><td>Вызов меню для действий со столбцами (изменение порядка сортировки и перемещения столбца из группы в группу)</td></tr>;ENG:<tr><th>Left-click on the column header</th><td>Open the menu for column actions (changing the sort order and moving a column from group to group)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Щелчок правой кнопкой мыши на заголовке столбца</th><td>Вызов меню для изменения перечня видимых столбцов и перемещения их между группами</td></tr>;ENG:<tr><th>Right-click on the column header</th><td>Open the menu to change the list of visible columns and move them between groups</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Перемещение мышью границы столбцов</th><td>Изменение ширины столбца слева</td></tr>;ENG:<tr><th>Move the column border </th><td>To change the column width on the left</td></tr>;"));
       if (!(frm.getProperty("SelectionMode").equals("None"))){
       line(app.NLS.NLS("RUS:<tr><th>Щелчок левой кнопкой мыши на иконке слева от заголовка первого столбца</th><td>Вкл/Откл режим множественного выделения строк для действий над несколькими выделенными строками</td></tr>;ENG:<tr><th>Left-click on the icon to the left of the first column header</th><td>On/Off multiple selection mode to operate on multiple selected rows</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Щелчок правой кнопкой мыши на иконке слева от заголовка первого столбца в режиме множественного выделения строк</th><td>Вызов локального меню для массовых действий с отмеченными строками (Выделить всё, Инверсия, Снять выделения)</td></tr>;ENG:<tr><th>Right-Click on the icon to the left of the title of the first column in the multiple selection mode of rows</th><td>Call the local menu for mass mark actions (Mark all, Invert, Unmark all)</td></tr>;"));
       }
       line(app.NLS.NLS("RUS:<tr><th>Иконка "+ico("QBE")+" или клавиша [[F7]]</th><td>Переход списка в режим ввода условий отбора по образцам (QBE)</td></tr>;ENG:<tr><th>Icon "+ico("QBE")+" or [[F7]] key</th><td>Switch on mode of input criteria to select rows (QueryByExample)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Иконка "+ico("refresh")+" или клавиша [[F8]]</th><td>Обновить список отображаемых строк из БД. Выполнить запрос по условиям отбора, указанным в режиме QBE</td></tr>;ENG:<tr><th>Icon "+ico("refresh")+" or [[F8]] key</th><td>Refresh the list of displayed rows from the database. Execute the query according to the criteria specified in the QBE mode</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[F9]] в режиме QBE</th><td>Отображение наиболее типичных значений текущего поля (если таковые есть) с возможностью вставки в поле выбранного значения</td></tr>;ENG:<tr><th>Key [[F9]]] in the QBE mode </th><td>displays the most typical values of the current field (if any) with the possibility to insert them into the field of the selected value</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Enter]] или двойной клик в ячейке в режиме QBE</th><td>Ввод критерия поиска в специальном бланке</td></tr>;ENG:<tr><th>Key [[Enter]]] or double click in the QBE mode</th><td>Call special blank to input criteria</td></tr>;"));
       
       
  }
  if ("BLANK".equals(frm.getType())){ // если бланц
       line(app.NLS.NLS("RUS:<tr><th>Щелчок правой кнопкой мыши на поле ввода</th><td>Отображение локального меню действий для текущего поля</td></tr>;ENG:<tr><th>Right-click on the input field</th><td>Displays the local action menu for the current field</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Иконка "+ico("LOV")+" справа от поля или клавиша [[F9]]</th><td>Вызов списка с возможными или наиболее типичными значениями ([[LOV]] - List of values)</td></tr>;ENG:<tr><th>Icon "+ico("LOV")+" on the right of field or [[F9]] key</th><td>List of possible or common values ([[LOV]] - List of values)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Иконка "+ico("Calendar")+" справа от поля или клавиша [[F9]]</th><td>Вызов календаря</td></tr>;ENG:<tr><th>Icon "+ico("Calendar")+" on the right of field or [[F9]] key</th><td>Calendar for date(time) input</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Иконка "+ico("DropDown")+" справа от поля или клавиша [[F9]]</th><td>Выпадающий список возможных значений</td></tr>;ENG:<tr><th>Icon "+ico("DropDown")+" on the right of field or [[F9]] key</th><td>Drop-down list of possible values</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[F4]]</th><td>Заполнить нетронутые поля значениями из предыдущего бланка (\"копирование\" для ускорения ввода множества похожих данных)</td></tr>;ENG:<tr><th>[[F4]] key</th><td>Fill empty fields with values from the previous form (\"copy\" to speed up the input of many similar data)</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Tab]]</th><td>Перемещение курсора на следующее поле или следующую кнопку</td></tr>;ENG:<tr><th>[[Tab]] key</th><td>Move cursor to next field or next button</td></tr>;"));
       line(app.NLS.NLS("RUS:<tr><th>Клавиша [[Shift-Tab]]</th><td>Перемещение курсора на предыдущее поле или предыдущую кнопку</td></tr>;ENG:<tr><th>[[Shift-Tab]] key</th><td>Move cursor to previous field or button</td></tr>;"));
       
       
  }
  if ("TREE".equals(frm.getType()) ||"MENU".equals(frm.getType())){ // если дуб
      line(app.NLS.NLS("RUS:<tr><th>Щелчок левой кнопкой мыши на узле</th><td>Сделать узел текущим (контекстным)</td></tr>;ENG:<tr><th>Click with the left mouse button on the node</th><td>Make the current node (context)</td></tr>;"));
      line(app.NLS.NLS("RUS:<tr><th>Щелчок правой кнопкой мыши на узле или [[Alt-X]]</th><td>Отображение списка действий, применимых в настоящий момент к текущему узлу (локальное меню)</td></tr>;ENG:<tr><th>Right-click a node or [[Alt-X]]</th><td>Displays a list of actions currently applicable to the current node (local menu)</td></tr>;"));
      line(app.NLS.NLS("RUS:<tr><th>Щелчок левой кнопкой мыши на узле при нажатой клавише [[Ctrl]]</th><td>Переход уровня в режим многоузлового выделения</td></tr>;ENG:<tr><th>Left-click on the node while pressing [[Ctrl]]</th><td>Switch the tree level to multi-node selection mode</td></tr>;"));
      line(app.NLS.NLS("RUS:<tr><th>Щелчок правой кнопкой мыши на выделенном узле</th><td>Отображение списка действий, применимых в настоящий момент ко всем выделенным узлам</td></tr>;ENG:<tr><th>Right-click the selected node</th><td>Display a list of actions that are currently applicable to all selected nodes</td></tr>;"));
  }
  line(app.NLS.NLS("RUS:<tr><th>Перетаскивание мышью границы окна</th><td>Изменение размера окна с привязкой значений к текущей конфигурации и текущему пользователю</td></tr>;ENG:<tr><th>Drag the window border</th><td>Change the window size and save size values for current configuration and the current user</td></tr>;"));
  if (!frm.getVal("SysMenu").equals("N")) {
  line(app.NLS.NLS("RUS:<tr><th>Иконка "+ico("winmenu-hover")+" в левой части заголовка формы или клавиша [[Alt-Z]]</th>;ENG:<tr><th>"+ico("winmenu-hover")+" Icon on the left side of the window title or [[Alt-Z]]</th>;"));
  line("<td id=\"Sys-"+frm.getFormID()+"\" class=\"Container\">"+app.NLS.NLS("RUS:Отображение списка действий с общей функциональностью (системное меню);ENG:Display a list of actions with common functionality (system menu);"));
  line("<table class=\"Switchable SysMenu TabInfo\">");
  for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
   if (el.getType().equals("ACTION") 
           && (like(el.getCode(),"SYSMENU$%"))
           && !(el.getProperty("ACTIVE").equals("N"))){
      line("<tr><th>"+ico(GetIcoFromSquareBracket(el.getProperty("PROMPT")))+GetVal(el.getCode(),"PROMPT")+"</th><td>"+GetVal(el.getCode(),"EXPLANATION")+"</td></tr>");
   }
  }
  
  addSwitch("Sys-"+frm.getFormID(),"SysMenu",app.NLS.NLS("RUS:Показать пункты меню;ENG:Show menu items;"),app.NLS.NLS("RUS:Скрыть;ENG:Hide;"));
  }
    return content.toString();   
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
    frm=pProp.ownerObj.getForm();
    if (frm==null) {
    String curFrmID=app.getVal("FormID");        
    if (app.forms.containsKey(curFrmID)){
       frm=app.forms.get(curFrmID);
    } else return;
    }
    if (is_not_null(app.getProperty("EHELP_URL"))){
        LuiDbConn mlc=null;
        String ehelp="";
        try {
        mlc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());  
        ehelp=mlc.getSelectResult("select lui_f_is_form_in_group('"+frm.getProperty("MetadataCode")+
                "',lui_f_get_application_property('"+app.code+"','GROUP4GENFORM'),'"+app.code+"')");
        if (is_not_null(ehelp))ehelp=frm.getType();
        if (frm.elements.containsKey(frm.getType()+"_TYPE")) {
            ehelp=ehelp+"_"+frm.elements.get(frm.getType()+"_TYPE").getProperty("Value");
        }
                } finally{
            if (mlc!=null) app.metaConnPool.SetFree(mlc);
        }
        if (is_not_null(ehelp)) 
        app.parser.plugins.get("URL").execute("(Target:T)"+app.getProperty("EHELP_URL")+ehelp+"-"+frm.getProperty("ConfigCode"), pProp);
        else
        app.parser.plugins.get("URL").execute("(Target:T)"+app.getProperty("EHELP_URL")+frm.getProperty("MetadataCode")+"-"+frm.getProperty("ConfigCode"), pProp);
        return;
    }
           LuiClCommandGroup XMLcg=app.commandWriter.addGroup("HELP");
        LuiClCommand XMLcmd=XMLcg.addCommand("Info");
        XMLcmd.addScaPar("Code",app.code);
        XMLcmd.addScaPar("Title",NVL(pString,"Help"));
        XMLcmd.addScaPar("Target","Tab");
        String cont=compute(pString,pProp);
        XMLcmd.addScaPar("Content",cont);
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       
    }
    
private void line(String pString){
  content.append(pString.replace("\n","<BR>").replace("[[[","<span class=Dynamic>").replace("]]]","</span>").replace("[[","<span>").replace("]]","</span>"));
  content.append("\n");
}
private void help4groupOfBlank(String pParent) throws LuiException, SQLException{
    String gname="";
    String tstr="";
    if ("Form".equals(pParent)){
        gname=app.NLS.NLS("RUS:Форма;ENG:Form;");
    } else {
        //gname=NVL(frm.elements.get(pParent).getProperty("TITLE"),frm.elements.get(pParent).getProperty("Name"));
        gname=NVL(GetVal(pParent,"TITLE"),GetVal(pParent,"Name"));
    }
        
    for (int i=0;i<frm.elements.size();i++){
    LuiAppElement el=frm.elSort(i);
   if (el.getParent().equals(pParent)){
   if (el.getType().equals("GROUP")){
   help4groupOfBlank(el.getCode());
   }    
   if (el.getType().equals("FIELD")){
     if (is_not_null(el.getProperty("ForNLSSupport")))continue;
     line("<tr>");  
     line("<td class=\"Switchable Code NoWrap\">"+LuiString.MaskAncillaryChars(el.getCode())+"</td>"); // Код
     line("<td>"+MaskAncillaryChars(gname)+"</td>"); // Группа
     line("<td class=\"NoWrap\">"+MaskAncillaryChars(GetVal(el.getCode(),"CAPTION"))+"</td>"); // Подпись
     switch (GetVal(el.getCode(),"CHANGING")) {
         case "B":
             tstr=app.NLS.NLS("RUS:Переключатель значений;ENG;Value switch;");
         break; 
         case "R":
             tstr=app.NLS.NLS("RUS:Радиогруппа;ENG;Radiogroup;");
         break; 
         case "C":
             tstr=app.NLS.NLS("RUS:Выбор значения из списка возможных значений;ENG;Select a value from the list of possible values;");
         break;             
         case "H":
             tstr=app.NLS.NLS("RUS:Ввод значения без отображения символов;ENG;Enter a value without displaying characters;");
         break;             
         case "Y":
         case "U":
          tstr=app.dataConverter.LuiDataTypes.get(GetVal(el.getCode(),"DATA_TYPE")).name;
             String format=GetVal(el.getCode(),"FORMAT");
             if (is_not_null(format)){
                   tstr=tstr+app.NLS.NLS("RUS: в формате ;ENG: by formst ;")+format;
             }
             if (GetVal(el.getCode(),"CHANGING").equals("U")){
                 tstr=tstr.replace("#", app.NLS.NLS("RUS: в верхнем регистре;ENG: in uppercase;"));
             } else {
                 tstr=tstr.replace("#","");
             }
         Integer lines=LuiString.String2Integer(GetVal(el.getCode(),"VISIBLE"), 0);
         if (lines>1){
           tstr=MergeText(tstr,app.NLS.NLS("RUS:Mногострочное поле.;ENG:Multiline field.;"));
         }
         break;
         default:
             tstr=app.NLS.NLS("RUS:Неизменяемое поле. Только для отображения;ENG;Display only;");
         break;
     }
     if (GetVal(el.getCode(),"OPTIONAL").equals("N")){
       switch(GetVal(el.getCode(),"CHANGING")){
           case "Y":case "C":case "U":case "H":
           tstr=MergeText(tstr,app.NLS.NLS("RUS:Обязательно для заполнения;ENG:Mandatory to fill;"));    
           break;
       }
     }
     line("<td>"+MaskAncillaryChars(tstr)+"</td>");
     line("<td>"+MaskAncillaryChars(RTrim(MergeText(ifExist(GetVal(el.getCode(),"CAPTION"),GetVal(el.getCode(),"HINT")),GetVal(el.getCode(),"ITEM_NOTES")),"."))+"</td>");  // Описание
     line("</tr>");
   }
    }
     }
    
}
private int help4groupOfMenu(String pParent,String nbsp) throws LuiException, SQLException{
int cnt=0;
String atext;
String role="";
for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
   if (el.getParent().equals(pParent)){
   if (el.getType().equals("ACTION")){
    if (like(el.getCode(),"SYSMENU$%"))continue;
    if (!GetVal(el.getCode(),"DISPLAY","CALC","").equals("Y"))continue;
    
    line("<tr>");
     line("<td class=\"Switchable Code NoWrap\">"+MaskAncillaryChars(el.getCode())+"</td>"); // Код
     line("<td>"+nbsp+ico(el.getProperty("ICON"))+"</td>"); // иконка
     line("<td class=\"NoWrap\">"+nbsp+MaskAncillaryChars(NVL(GetVal(el.getCode(),"PROMPT"),GetVal(el.getCode(),"Name")))+"</td>"); // Название    
     //готовим способ выполнения
     atext="";
     if ((is_not_null(GetVal(el.getCode(),"COMMAND","START",""))||
         is_not_null(GetVal(el.getCode(),"PROC_FINALLY","START",""))||
     is_not_null(GetVal(el.getCode(),"PROC_INITIALLY","START",""))
                      ) && !GetVal(el.getCode(),"ACTIVE","START","").equals("N")){
      role=el.getProperty("ROLE");
      if (like(NVL(role,"DBLCLICK"),"%DBLCLICK%")){
      atext=app.NLS.NLS("RUS:Двойной клик или Enter;ENG:Double click or Enter key;")+"<br>";
      }    
      if (like(role,"%RMB%")){
      atext=app.NLS.NLS("RUS:Меню всплываюшее по правой кноаке мыши или по кнопке Alt+x;ENG:pop-up Menu on the right mouse button or Alt+x;")+"<br>";
      }
      if (like(role,"%CLOSE%")){
     }   
     }
     atext="<td>"+atext+"</td>";
     line(atext);
     //готовим описание
     atext="";
     atext=MergeText(MergeText(ifExist(GetVal(el.getCode(),"PROMPT"),GetVal(el.getCode(),"EXPLANATION")),GetVal(el.getCode(),"ACTION_NOTES")),atext);
     atext="<td>"+atext+"</td>";
     line(atext);
     line("</tr>");
     if (!like(role,"%CLOSE%")){
     help4groupOfMenu(el.getCode(),nbsp+"&nbsp;&nbsp;");
     }
   }
   }    
}
   


return cnt;
}
private int help4groupOfList(String pParent) throws LuiException, SQLException{
String GT="";
int cnt=0;
if (!"Form".equals(pParent)){
  GT=frm.elements.get(pParent).getProperty("GROUP_TYPE");
}
for (int i=0;i<frm.elements.size();i++){
   LuiAppElement el=frm.elSort(i);
   if (el.getParent().equals(pParent)){
   if (el.getType().equals("GROUP")){
   cnt=cnt+help4groupOfList(el.getCode());
   }
   if (el.getType().equals("ITEM")){
   if (el.isVisible(true)){
     String TStr=GT;
     if (GetVal(el.getCode(),"Width").equals("0")){
       TStr=TStr+" Switchable HC";
       cnt=cnt++;   
     }
     if (is_not_null(TStr)){
           TStr=" class=\""+Trim(TStr)+"\"";
     }
     line("<tr"+TStr+"><th class=\"Switchable Code NoWrap\">"+MaskAncillaryChars(el.getCode())+"</th>");
     line("<td class=\"NoWrap\">"+MaskAncillaryChars(GetVal(el.getCode(),"CAPTION"))+"</td>");
     line("<td>"+MaskAncillaryChars(MergeText(ifExist(GetVal(el.getCode(),"CAPTION"),GetVal(el.getCode(),"HINT")),GetVal(el.getCode(),"ITEM_NOTES")))+"</td></tr>");
   }
   }
   }
 } 
return cnt;
}
private String GetVal(String pElCode,String pPropertyCode,String pMapMode,String pIfFail) throws LuiException, SQLException{
  String RetVal="";
  String eCode=NVL(pElCode,"Form");
  
  
  if ("CALC".equals(pMapMode)){
      if ("Form".equals(eCode)) {
       if (frm.properties.containsKey(pPropertyCode)) RetVal=frm.getProperty(pPropertyCode);
       else RetVal=pIfFail;
      } else {
       if (frm.elements.containsKey(pElCode)){   
        if (frm.elements.get(eCode).properties.containsKey(pPropertyCode)) RetVal=frm.elements.get(eCode).getProperty(pPropertyCode);
        else RetVal=pIfFail;
       } else RetVal=pIfFail;
              } 
  } else {
      if ("Form".equals(eCode)) {
        if (frm.properties.containsKey(pPropertyCode)) RetVal=frm.properties.get(pPropertyCode).getNonCompValue();
        else RetVal=pIfFail;
      } else {
      if (frm.elements.containsKey(pElCode)){    
       if (frm.elements.get(eCode).properties.containsKey(pPropertyCode)) RetVal=frm.elements.get(eCode).properties.get(pPropertyCode).getNonCompValue();
       else RetVal=pIfFail;
      } else RetVal=pIfFail;        
      } 
  }
    if (is_not_null(pMapMode)){
        
    }  
  return LuiString.RemoveIco(RetVal);
}
private String GetVal(String pElCode,String pPropertyCode) throws LuiException, SQLException{
    return GetVal(pElCode,pPropertyCode,"CALC","");
}
private String MergeText(String existStr,String newStr){
    String W="";
    int i;
    if (is_null(existStr))return newStr;
    if (is_null(newStr))return existStr;
    //if (newStr.length()>existStr.length()) return newStr;
    i=1;
    while(true){
     W=LuiString.Extract_Word(existStr, i," ,.;!?-()/\\:'\"<>").toUpperCase();
     if (is_null(W)) break;
     if (newStr.toUpperCase().indexOf(W)<0)break;
     i++;
    }
    if (is_null(W)){
        return newStr;
    }
    i=1;
    while(true){
     W=LuiString.Extract_Word(newStr, i," ,.;!?-()/\\:'\"<>").toUpperCase();
     if (is_null(W)) break;
     if (existStr.toUpperCase().indexOf(W)<0)break;
     i++;
    }
    if (is_null(W)){
        return existStr;
    }
    i=RTrim(existStr).length();
    char lastCh=RTrim(existStr).charAt(i-1);
    if (!(lastCh=='.' || lastCh=='!' || lastCh==';' || lastCh=='?')) {
        return RTrim(existStr)+". "+newStr;
    }
    return existStr+" "+newStr;
}
private String ifExist(String pattern,String newStr){
String W="";
int i=1;
if (is_null(pattern))return newStr;
i=1;
while(true){
    W=LuiString.Extract_Word(newStr, i," ,.;!?-()/\\:'\"<>").toUpperCase();
    if (is_null(W))break;
    if (pattern.toUpperCase().indexOf(W)<0)break;
    i++;
}
if (is_null(W))return "";
return newStr;
}
private void addSwitch(String elemId,String code,String titleOn,String titleOff){
    line("<div class=\"Variables UnSwitch"+code+"\" onclick=\"Set($('"+elemId+"'),'+"+code+"')\">"+titleOn+"</div>");
    line("<div class=\"Variables Switchable "+code+"\" onclick=\"Set($('"+elemId+"'),'-"+code+"')\">"+titleOff+"</div>");
}
private String ico(String pICON){
 return "<div class=\"help-ico ico-"+pICON+"\"></div>";
}
}
