/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Интеграция с LDAP
Compute возвращает Y/N и может вызвать Exception при ошибках импорта членства в группах:
{LADP:URL;DOMAIN:;USERNAME:;PASSWORD:;IMPORT_MEMBERSHIP:N/Y/E;}
Execute вызывает LUI-00009 если аутентификация не прошла и может вызвать Exception при ошибках импорта членства в группах
URL- url для коннекта к LDAP-серверу пример:"ldap://ldap.fors.ru:389"
DOMAIN - домен, в котором разыскивается пользователь. Пример:fors.ru
USERNAME - имя пользователя
PASSWORD - пароль
IMPORT_GROUP_MEMBERSHIP - следует ли импортировать принадлежность к группам 
Если N (по умолчанию), то ничего не происходит - просто аутентификация и всё
Если Y, то список групп из LDAP-сервера заносится в текстовый JSON массив и сохраняется в переменной сеанса в БД (LUI$) LUI.LDAP_MEMBERSHIP
Если E, то список групп обрабатывается следующим образом:
      Для данного приложения выполняется просмотр групп пользователей в LUI (lui_t_usergroup) с непустым полем LDAP
      Для группы пользователей LUI выполняется вычисление выражения хранящегося в поле LDAP. Вычисление выполняется последовательно 
      для групп считанных из LDAP-сервера, как только выражение венёт 'Y' группа LUI будет помещена в JSON массив.
      Накопленный JSON массив будет содержать перечень групп прикладной системы, в которых состоит пользователь по данным из LDAP-сервера
      Этот массив будет использоваться при определении прав доступа пользователя к элементам форм и, возможно к другим объектам приложения, но не к объектам БД.
      Объекты БД могут оставаться закрытыми на уровне СУБД.
      
 */
package Standard;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.Hashtable;
import javax.naming.AuthenticationException;
import javax.naming.AuthenticationNotSupportedException;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;

/**
 *
 * @author aliubushkin
 */
public class LDAP implements LuiPlugin{
LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
         return exe('C',pString,pProp);
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       exe('E',pString,pProp);  
    }
     private String exe(char executeOrCompute,String pString,LuiAppProp pProp) throws  LuiException,SQLException{
       
       LuiCollection col= new LuiCollection("LDAP:"+pString);
       String url=col.Get_Element("LDAP");
       String domain=col.Get_Element("DOMAIN");
       String username=col.Get_Element("USERNAME");
       String password=col.Get_Element("PASSWORD");
       String domainDC="DC="+domain.replace(".",",DC=");
       String imp=NVL(col.Get_Element("IMPORT_MEMBERSHIP"),"Y");
       Hashtable<String, String> environment = new Hashtable<String, String>();

        environment.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
        //environment.put(Context.INITIAL_CONTEXT_FACTORY, "vasa");
        environment.put(Context.PROVIDER_URL, url);
        environment.put(Context.SECURITY_AUTHENTICATION, "simple");
        environment.put(Context.SECURITY_PRINCIPAL, username+"@"+domain);
        environment.put(Context.SECURITY_CREDENTIALS, password);
        DirContext context =null;
        LuiDbConn lc=null;
        String memberShip="";
        try 
        {
            context = new InitialDirContext(environment);
           if (!"N".equals(imp)){ 
            SearchControls ctrls = new SearchControls();
            ctrls.setSearchScope(SearchControls.SUBTREE_SCOPE);
            NamingEnumeration results = context.search(domainDC,"(& (userPrincipalName="+username+"@"+domain+")(objectClass=user))", ctrls);
            if(!results.hasMore()) 
                throw new AuthenticationException("Principal name not found");
            SearchResult result = (SearchResult) results.next();
            Attribute memberOf = result.getAttributes().get("memberOf");
            if ("Y".equals(imp)){
            for(int i=0;i<memberOf.size();i++){
               if (lc==null){
                   lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                   memberShip=lc.getSelectResult("select set_config('LUI.LDAP_MEMBERSHIP',(to_jsonb('"+memberOf.get(i).toString().replace("'","''")+"'::text))::text,false)");
               } else {
             memberShip=lc.getSelectResult("select set_config('LUI.LDAP_MEMBERSHIP',(current_setting('LUI.LDAP_MEMBERSHIP',true)::jsonb||to_jsonb('"+memberOf.get(i).toString().replace("'","''")+"'::text))::text,false)");
            }
            }
           } //"Y"
           if ("E".equals(imp)){ 
                 // выясняем код формы
                LuiAppForm frm=pProp.ownerObj.getForm();
                if (frm==null) {
                String curFrmID=app.getVal("FormID");        
                if (app.forms.containsKey(curFrmID)){
                   frm=app.forms.get(curFrmID);
                }
                }
                if (frm==null) return null;
               lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
               int luiMemCnt=0;
               LuiCollection LuiAllGrp=lc.getSelectAllRowsTwoColsValMap("select code,ldap from lui_t_usergroup where app_code='"+app.code+"' and ldap is not null");
               if (LuiAllGrp!=null){
                
                 for (String luiGrp:LuiAllGrp.keySet()){
                   for(int i=0;i<memberOf.size();i++){
                    String ldapGrp=memberOf.get(i).toString();
                    frm.addProperty("LuiGroup",luiGrp);   
                    frm.addProperty("LdapGroup", ldapGrp);
                    frm.addProperty("LuiGroupExpression",LuiAllGrp.Get_Element(luiGrp));
                    if (frm.getProperty("LuiGroupExpression").equals("Y")){
                      if (luiMemCnt>0){
                          memberShip=memberShip+",";
                      } 
                         memberShip=memberShip+"\""+luiGrp+"\"";
                        luiMemCnt++;
                        break;
                    }
                    frm.properties.remove("LuiGroup");
                    frm.properties.remove("LdapGroup");
                    frm.properties.remove("LuiGroupExpression");
                   }  
                 }
                 if (luiMemCnt>0){
                     memberShip="{"+memberShip+"}";
                     memberShip=lc.getSelectResult("select set_config('LUI.LUI_MEMBERSHIP','"+memberShip+"'::text,false)");
                 }
               }
           }
           }
        } catch (AuthenticationNotSupportedException ex) 
        {
            app.msgWriter.write("LUI-00135",url,ex.getMessage());
        }

        catch (AuthenticationException ex)
        {
            if ("E".equals(executeOrCompute))
            app.msgWriter.write("LUI-00009");
            else 
            return "N";
        }

        catch (NamingException ex)
        {
            app.msgWriter.write("LUI-00135",url,ex.getMessage());
        }
        finally {
            if (context!=null){
                try {
                    context.close();
                } catch (NamingException ex) {
                   
                }
            }
            if (lc!=null) {
                app.metaConnPool.SetFree(lc);
            }
        }
       return "Y";
     }
    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        
    }
    
}
