/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Формирует HTML-документ из данных списка во временном каталоге сеанса
 Compute возвращает url на этот файл
 Execute шлёт команды клиенту для открытия файла
 синтаксис List2XSSF:nnnn;TOTALS:Y;
 */
package Standard;

import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.Writer;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import LuiCollection.LuiCollection;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;

import luiserver.LuiAppConn;
import luiserver.LuiAppElement;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiClCommand;
import luiserver.LuiClCommandGroup;
import luiserver.LuiClCommandWriter;
import luiserver.LuiClEvent;
import luiserver.LuiClEventReader;
import luiserver.LuiPlugin;


/**
 *
 * @author 
 */
public class List2HTML implements LuiPlugin
{
    String s160 = " ";
    LuiApplication app;
    LuiClCommandWriter cw;
    LuiClEventReader er;

    String bookFileName;
    Writer filew=null;
    LuiAppForm frmInd = null;
    LuiAppElement elInd = null;
    LuiAppElement elIndOK = null;
    LuiClCommandGroup XMLcg = null;
    LuiClCommand XMLcmd = null;
    HashMap <String, String> storedRow = new HashMap<>();
    HashMap <String, String> mapCellStyles = new HashMap<>();
    HashMap <String, Integer> mapCellStylesIdx = new HashMap<>();
    HashMap <String, String> mapTextStyle = new HashMap<>();
    HashMap <String, Integer> mapTextStyleIdx = new HashMap<>();
    Integer styleIdx = 0;
    LuiCollection cs = null;
    HashMap <String, String> mapCharts = new HashMap<>();
    LuiCollection colored_cols = null;
    boolean p_stream_mode;
    boolean p_coloring;
    String head_decor = "BACKGROUND(#e8eef4)";
    

    ArrayList<String> lines = new ArrayList<>();
    String htmlHead, htmlFloor;
    
    
    public class TabHeadColumn
    {
        public String head0;
        public String head1;
        public int width;
        
        public TabHeadColumn(String h0, String h1)
        {
            this.head0 = h0;
            this.head1 = h1;
            this.width = 0;
        }
    };
    
    
    @Override
    public void setContext(LuiApplication pContext)
    {
        app=pContext;
        er=app.eventReader;
        cw=app.commandWriter;        
    }

    // SVA ---------------------------------------------------------------------

    private String getParameter(String pParamCode)  throws LuiException, SQLException
    {
        String value = "";
        LuiDbConn lc = null;
        
        try
        {
            lc = app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
            value = lc.getSelectResult("select lui_f_get_application_property('"+app.code+"','"+pParamCode+"','"+app.NLS.currentLang.code+"')");
        }
        finally
        {
            if( lc != null )
            {
                app.metaConnPool.SetFree(lc);
                lc = null;
            }
        }
        
        return value;
    }
    
    // SVA ---------------------------------------------------------------------

    private void line(String pString) throws IOException
    {
        filew.write(pString.replace("\n","<BR>").replace("[[[","<span class=Dynamic>").replace("]]]","</span>").replace("[[","<span>").replace("]]","</span>") + "\r\n");
    }

    // SVA ---------------------------------------------------------------------

    private void aline(String pString)
    {
        lines.add(pString);
    }
    
    // SVA ---------------------------------------------------------------------
    
    private void pause(int secs)
    {
        try
        {
            Thread.sleep(secs * 1000);
        }
        catch( InterruptedException ex)
        {
            System.out.println("IOException: " + ex.getLocalizedMessage());
        }
    }

    // SVA ---------------------------------------------------------------------

    private void clearMemory()
    {
        filew = null;
        htmlHead = "";
        htmlFloor = "";
        lines.clear();
        mapCellStyles.clear();
        mapCellStylesIdx.clear();
        styleIdx = 0;
        mapTextStyle.clear();
        mapTextStyleIdx.clear();
        if(colored_cols != null) colored_cols.clear();
        System.gc();
    }

    // SVA ---------------------------------------------------------------------

    
    public void saveImage(String formId, String imageBase64) throws LuiException, SQLException
    {
        mapCharts.put(formId, imageBase64);
    }   
    
    // SVA ---------------------------------------------------------------------


    private void makeChartSheet(String fc)
    {
        aline("<img style=\"cursor:pointer\" onclick=\"img_resize(this)\" src=\"" + mapCharts.get(fc) + "\" />");
    }
    
    // SVA ---------------------------------------------------------------------
    
    private String getElementVisualProps(LuiAppElement el) throws LuiException, SQLException
    {
        // Извлекаем свойства ячейки
        String visualProps = el.properties.get("VISUALIZATION").getComputedProperty().trim();
        String alignmProps = el.getProperty("ALIGNMENT").trim();
        String props = (visualProps + " " + alignmProps).trim();
        
        return props;
    }

    // SVA ---------------------------------------------------------------------

    private String makeCellStyle(String props)
    {
        if( is_null(props) )
        {
            props = "<<EMPTY>>";
            // Вообще-то наверно пусто не бывает. Всегда есть ALIGNMENT?
        }

        if( mapCellStyles.containsKey(props) )
        {
            return "st" + mapCellStylesIdx.get(props);
        }

        String style = "";

        // default colors
        String fnColor = "#000000";
        String bgColor = "#FFFFFF";
        
        String textAlign = "";
        String vertAlign = "";
                
        String[] prop_list = props.split("\\s+");

        for(int k = 0; k < prop_list.length; k++)
        {                                   
            String prop = prop_list[k];
            String arg = "";
            String template = "^(\\S+?)\\((\\S+)\\)$";

            if( prop.matches(template) )
            {
                Pattern pattern = Pattern.compile(template);
                Matcher matcher = pattern.matcher(prop);
                if(matcher.matches())
                {
                    prop = matcher.group(1);
                    arg  = matcher.group(2);
                }
            }

            switch (prop)
            {                
                // -------------------------------------------------------------

                case "RIGHT":   textAlign = "right";  break;
                case "CENTER":  textAlign = "center"; break;
                case "LEFT":    textAlign = "left";   break;
 
                // -------------------------------------------------------------

                case "TEXTCOLOR":   fnColor = arg;  break;
                case "BACKGROUND":  bgColor = arg;  break;                                            

                // -------------------------------------------------------------

                case "ALL_SELECTED ":   //?? {background-color:#DDF; font-weight:600}
                    fnColor = "#DDF";
                    style = style.concat("font-weight:600; ");
                break;

                case "ATTENTION":       //?? {color:#970; font-weight:600}
                    fnColor = "#970";
                    style = style.concat("font-weight:600; ");
                break;

                case "BACKGROUND_SELECTED":
                    bgColor = "#DDF";
                break;

                case "BAD":      //?? {text-decoration:line-through; text-decoration-color:red; color:#888}
                    //fnColor = "#888";
                    //font.setStrikeout(true);
                    style = style.concat("text-decoration:line-through; text-decoration-color:red; ");
                break;

                case "BLUE":          //?? {color:royalblue}
                    fnColor = "royalblue";
                break;

                case "BOLD":            //? BOLD {font-weight:bolder}
                    style = style.concat("font-weight:bolder; ");
                break;

		/*
                // BOUNDARY {border-bottom:2px solid darkred}

                case "BROWN":          //?? {color:brown}
                    fnColor = "brown";
                break;

                case "CROSSEDOUT":      //?? {text-decoration:line-through; text-decoration-color:red}
                    font.setStrikeout(true);
                    // text-decoration-color:red ???
                break;

                case "CYAN":          //?? {color:cyan; text-shadow:1px 1px 1px darkcyan}
                    fnColor = "cyan";
                    // text-shadow:1px 1px 1px darkcyan ???
                break;

                case "DANGER":      //?? {color:#A00; font-weight:600}
                    fnColor = "#A00";
                    font.setBold(true);                                    
                break;

                case "FONT_SELECTED":   //?? {font-weight:600}
                    font.setBold(true);                                    
                break;

                case "GRAY":      //?? {color:#AAA}
                    fnColor = "#AAA";
                break;

                case "GREEN":          //?? {color:forestgreen}
                    fnColor = "forestgreen";
                break;

                case "INHERIT":
                    bgColor = "#DDD";
                    fnColor = "#444";
                break;

                case "INTERIM":
                    bgColor = "#FF7";
                break;

                case "ITALIC":
                    font.setItalic(true);
                break;

                // LARGE {font-size:larger}

                case "LOWERED":
                    fnColor = "#BBB";
                break;

                case "MAGENTA":          //?? {color:magenta}
                    fnColor = "magenta";
                break;

                case "MONOSPACE":   //?? {font-family:monospace}
                    font.setFontName("Courier New");
                break;

                case "NEGATIVE":    //?? {color:white; background-color:gray; font-weight:600}
                    fnColor = "white";
                    bgColor = "gray";
                    font.setBold(true);                                    
                break;

                //NORMAL {color:black; font-weight:normal; background-color:transparent; text-shadow:none; position:static; text-decoration:none; border:0}

                case "NOTES":
                    font.setItalic(true);
                break;

                case "OK":          //?? {color:#4A0; font-weight:600}
                    fnColor = "#4A0";
                    font.setBold(true);                                    
                break;

                case "ORANGE":          //?? {color:orange}
                    fnColor = "orange";
                break;

                case "SECTION":          //?? {background-color:#888; font-weight:600; color:#FFF}
                    bgColor = "#888";
                    fnColor = "#FFF";
                    font.setBold(true);                                    
                break;

                case "SHADOW":          //?? {text-shadow:1px 0 3px grey}
                    fnColor = "gray";
                    font.setBold(true);                                    
                break;

                case "START":          //?? {background-color:#8F8}
                    bgColor = "#8F8";
                break;
		*/

                case "STOP":          //?? {background-color:#F88}
                    bgColor = "#F88";
                break;

		/*
                case "TILT":
                    font.setItalic(true);
                break;

                case "UNDERLINE":
                    font.setUnderline(Font.U_SINGLE);
                break;
		*/                
            }
        }

        // Установим параметры ячейки
        
        style = style.concat("background-color:" + bgColor + "; ");
        style = style.concat("color:" + fnColor + "; ");

        if( !textAlign.equals("") ) style = style.concat("text-align:" + textAlign + "; ");
        if( !vertAlign.equals("") ) style = style.concat("vertical-align:" + vertAlign + "; ");
                
        mapCellStyles.put(props, style);
        mapCellStylesIdx.put(props, styleIdx);

	//System.out.println("makeCellStyle: props = " + props);
	//System.out.println("makeCellStyle: style = " + style);
	//System.out.println("makeCellStyle: return = " + "st" + styleIdx.toString());

        String ret = "st" + styleIdx.toString();

        styleIdx++;
        
        return ret;
    }
    
    // SVA ---------------------------------------------------------------------
    
    private String makeFontForRichText(String getup)
    {
        if( mapTextStyle.containsKey(getup) )
        {
            return "st" + mapTextStyleIdx.get(getup);
        }
        
        String style = "";

        boolean fnBold = false;
        boolean fnStrikeout = false;
        boolean fnItalic = false;

        String fnColor = "#000000";
        String prop = getup;
        String arg = "";
        String template = "^(\\S+)\\((\\S+)\\)$";

        if( prop.matches(template) )
        {
            Pattern pattern = Pattern.compile(template);
            Matcher matcher = pattern.matcher(prop);
            if(matcher.matches())
            {
                prop = matcher.group(1);
                arg  = matcher.group(2);
            }
        }

        switch( prop )
        {
            case "TEXTCOLOR":                                               
                fnColor = arg;
            break;

            case "UNDERLINE":
                style = style.concat("text-decoration:underline; text-decoration-color:"+arg+"; ");
            break;

            // -----------------------------------------------------------------
            /*
            case "ALL_SELECTED ":   //?? {background-color:#DDF; font-weight:600}
                fnColor = "#aeF8";
                fnBold = true;
            break;

            case "ATTENTION":       //?? {color:#970; font-weight:600}
                fnColor = "#e2b602";
                fnBold = true;
            break;

            case "BACKGROUND_SELECTED":
                fnColor = "#aef8";
            break;

            case "BAD":      //?? {text-decoration:line-through; text-decoration-color:red; color:#888}
                fnColor = "#888";
                fnStrikeout = true;
                style = style.concat("text-decoration-color:red; ");
            break;

            case "BLUE":          //?? {color:royalblue}
                fnColor = "royalblue";
            break;

            case "BOLD":
                fnBold = true;
            break;

            //case "BOUNDARY":          //?? {border-bottom:2px solid darkred}
            //    style = style.concat("border-bottom:2px solid darkred; ");
            //break;
            
            case "BROWN":          //?? {color:brown}
                fnColor = "brown";
            break;

            case "CROSSEDOUT":      //?? {text-decoration:line-through; text-decoration-color:red}
                fnStrikeout = true;
                style = style.concat("text-decoration-color:red; ");
            break;

            case "CYAN":          //?? {color:cyan; text-shadow:1px 1px 1px darkcyan}
                fnColor = "cyan";
                style = style.concat("text-shadow:1px 1px 1px darkcyan; ");
            break;

            case "DANGER":      //?? {color:#A00; font-weight:600}
                fnColor = "#A00";
                fnBold = true;
            break;

            case "FONT_SELECTED":   //?? {font-weight:600}
                fnBold = true;
            break;

            case "GRAY":      //?? {color:#AAA}
                fnColor = "#AAA";
            break;

            case "GREEN":          //?? {color:forestgreen}
                fnColor = "forestgreen";
            break;

            case "ITALIC":
                fnItalic = true;
            break;

            case "LARGE":           //?? {font-size:larger}
                style = style.concat("font-size:larger; ");
            break;

            case "LOWERED":
                fnColor = "#BBB";
            break;

            case "MAGENTA":          //?? {color:magenta}
                fnColor = "magenta";
            break;

            case "MONOSPACE":   //?? {font-family:monospace}
                style = style.concat("font-family:monospace; ");
            break;

            //NORMAL {color:black; font-weight:normal; background-color:transparent; text-shadow:none; position:static; text-decoration:none; border:0}

            case "NOTES":
                fnItalic = true;
            break;

            case "OK":          //?? {color:#4A0; font-weight:600}
                fnColor = "#4A0";
                fnBold = true;
            break;

            case "ORANGE":          //?? {color:orange}
                fnColor = "orange";
            break;

            case "SHADOW":          //?? {text-shadow:1px 0 3px grey}
                fnColor = "gray";
                style = style.concat("text-shadow:1px 0 3px grey; ");
            break;

            case "TILT":
                fnItalic = true;
            break;
            */
            
            default:
                return prop;
            //break;
        }

        style = "color:" + fnColor + "; " + style;

        if(fnBold)      style = style.concat("font-weight:600; ");
        if(fnStrikeout) style = style.concat("text-decoration:line-through; ");
        if(fnItalic)    style = style.concat("font-style:italic; ");
        

        mapTextStyle.put(prop, style);
        mapTextStyleIdx.put(prop, styleIdx);

        String ret = "st" + styleIdx.toString();

        styleIdx++;
        
        return ret;
    }

    // SVA ---------------------------------------------------------------------
      
    private String makeRichText(String makeup, String getup, String value)
    {
        String[] prop_list = getup.split("\\s+");
        String flags_pattern = "^\\/(.*)\\/([igmsuy]+)$";
        String regexp = makeup;
        String flags = "";

        if( regexp.matches(flags_pattern) )
        {
            Pattern pattern = Pattern.compile(flags_pattern);
            Matcher matcher = pattern.matcher(regexp);
            if(matcher.matches())
            {
                regexp = matcher.group(1);
                flags  = matcher.group(2);
            }
        }
        
        // decode regexp flags
        
        int jflags = 0;
        
        for( int i=0; i<flags.length(); i++ )
        {
            switch( flags.substring(i, 1) )
            {
                case "i":  jflags |= Pattern.CASE_INSENSITIVE;  break;
                //case "g":  jflags |= ; break;
                case "m":  jflags |= Pattern.MULTILINE;         break;
                case "s":  jflags |= Pattern.DOTALL;            break;
                //case "u":  jflags |= ; break;
                //case "y":  jflags |= ; break;
            }
        }

        // make RichText

        StringBuilder sv = new StringBuilder("");

        Pattern pattern = Pattern.compile(regexp, jflags);
        Matcher matcher = pattern.matcher(value);

        int lastPos = 0;


        while( matcher.find() )
        {
            for( int i=1; i<=matcher.groupCount(); i++ )
            {
                if( is_not_null(matcher.group(i)) )
                {
                    
                    if( lastPos <  matcher.start(i))
                        sv.append(value.substring(lastPos, matcher.start(i)));
                    
                    String textStyle = makeFontForRichText(prop_list[i]);
                    
                    String val = value.substring(matcher.start(i), matcher.end(i));
                                       
                    val = "<span class=\""+textStyle+"\">" + val + "</span>";

                    sv.append(val);
                    
                    lastPos = matcher.end(i);
                }
            }
        }        

        
        if( lastPos < value.length()-1 )
                sv.append(value.substring(lastPos));
        
        return sv.toString();
    }    

    // SVA ---------------------------------------------------------------------

    private void printStyles() throws IOException
    {
        line("<style>");
        
        for( String key : mapCellStyles.keySet() )
        {

            line("td.st" + mapCellStylesIdx.get(key) + " { " + mapCellStyles.get(key) + "}");
        }

        
        for( String key : mapTextStyle.keySet() )
        {
            line("span.st" + mapTextStyleIdx.get(key) + " { " + mapTextStyle.get(key) + "}");
        }
        
        
        line("</style>");
    }

    // SVA ---------------------------------------------------------------------

    private void storeCurrentPosition(LuiAppForm frm, ArrayList<ColInfo> colList) throws LuiException, SQLException
    {
        storedRow.clear();

        storedRow.put("<<<CurrentRow>>>", frm.getVal("CurrentRow"));
      
        for( int i=0; i<colList.size(); i++ )
        {
            String colCode = colList.get(i).alias;

            if( !frm.elements.containsKey(colCode) )
                continue;

            LuiAppElement el = frm.elements.get(colCode);

            storedRow.put(colCode, el.getValue());
        }
    }

    // SVA ---------------------------------------------------------------------

    private void restoreCurrentPosition(LuiAppForm frm, ArrayList<ColInfo> colList) throws LuiException, SQLException
    {
        if( colList != null && storedRow.size() > 0 )
        {
            frm.setVal("CurrentRow", storedRow.get("<<<CurrentRow>>>"));

            for( int i=0; i<colList.size(); i++ )
            {
                String colCode = colList.get(i).alias;

                if( !frm.elements.containsKey(colCode) )
                    continue;

                LuiAppElement el = frm.elements.get(colCode);

                el.setValue(storedRow.get(colCode), false, "", false);
            }
        }

        storedRow.clear();
    }

    // SVA ---------------------------------------------------------------------
    
    private void makeListSheet(String fc) throws LuiException, SQLException //, IOException
    {
        boolean needTot = false;
        HashMap<String,String> tot = null;
        int realCol = 1;
        int totalRows = 0;
        int tot4Ind = 0;

        String maxRowsS = cs.Get_Element("Print");
        String needTotS = cs.Get_Element("TOTALS");
        int maxRows = Integer.valueOf(NVL(maxRowsS, "1000"));
       
        totalRows = maxRows;
        tot4Ind = maxRows;

	//System.out.println("makeListSheet: fc = " + fc);

        LuiAppForm frm = app.forms.get(fc);

        LuiDbConn lc = null;

        frmInd = null;
        elInd = null;
        elIndOK = null;

        fc = app.getVal("FormID");

	//System.out.println("makeListSheet: fc = " + fc);
    
        if( app.forms.containsKey(fc) )
        {
           frmInd = app.forms.get(fc);

           if( frmInd.elements.containsKey("PROGRESS$INDICATOR") )
           {
               elInd=frmInd.elements.get("PROGRESS$INDICATOR");

               if( frmInd.elements.containsKey("OK") )
               {
                   elIndOK = frmInd.elements.get("OK");
               }
           }
           else
           {
               frmInd = null;
           }
        }

        
        aline("<div class=\"tableFixHead\">");
        aline("<table class=\"data-table\">");


        String queryText = frm.getVal("MainQuery");
            
        if( is_not_null(queryText) )
        {
            String acc = frm.getProperty("Connection");
            if( is_not_null(acc) )
            {
                if (app.userDbConnections.containsKey(acc))
                {
                    LuiAppConn ac = app.userDbConnections.get(acc);
                    lc = null;
                    Integer cursor = null;

                    if( ac.getSessionMode() == 'C' )
                        lc = ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                    else
                        lc = ac.currentConnection;
                                       
                    // формируем жирную ячейку с заголовком списка

                    ArrayList<ColInfo> colList = null;
                    int r = 0;
                    
                    try
                    {
                        cursor = lc.CursorOpen(queryText);
                        int cols = lc.CursorGetCols(cursor); // курсор
                        colList = lc.CursorGetColsInfo(cursor);
                        ArrayList<Boolean> colVisible = new ArrayList<Boolean>();
                        int FixedColumns = 0;

                        storeCurrentPosition(frm, colList);
                        
			ArrayList<String> colNames0 = new ArrayList<>();
			ArrayList<String> colNames1 = new ArrayList<>();

			ArrayList<TabHeadColumn> tabHeader = new ArrayList<>();

                        // формирование заголовков столбцов

                        for( int i=0; i<colList.size(); i++ )
                        {
                            String colCode = "";
                            colCode = colList.get(i).alias;
                            if( !frm.elements.containsKey(colCode) ) continue;
                            LuiAppElement el = frm.elements.get(colCode);

                            String colWidth = frm.getVisualPref(colCode, "ColWidth", "!");

                            if( colWidth.equals("!") )
                            {
                                colWidth = el.getProperty("WIDTH");
                            }

                            colVisible.add(i, !colWidth.equals("0") && el.getProperty("VISIBLE").equals("Y") && el.getProperty("Read").equals("Y"));


                            if( colVisible.get(i) )
                            {
                                String colSuperName = null;
                                String colNam = el.getName4show();

				colNames1.add(el.getName4show());
				//System.out.println("=1=> " + el.getName4show());

                                if( !el.getParent().equals("Form") )
                                {
				    colNames0.add(NVL(frm.elements.get(el.getParent()).getProperty("TITLE"), ""));
				    //System.out.println("=0=> " + frm.elements.get(el.getParent()).getProperty("TITLE"));

				    colSuperName = frm.elements.get(el.getParent()).getProperty("TITLE");

                                    //SVA если группа типа FIXED то посчитать для Freeze
                                    if( frm.elements.get(el.getParent()).getProperty("GROUP_TYPE").equals("FIXED") )
                                        FixedColumns++;
                                }
                                else
                                {
				    colNames0.add("пусто");
                                }

				tabHeader.add(new TabHeadColumn(colSuperName, colNam));
                                
                                realCol++;
                                if( is_not_null(el.getProperty("TOTAL")) ) needTot = true;
                            }
                        }

                        /*
                        System.out.printf("colNames0.size = %d\n", colNames0.size());
                        System.out.printf("colNames1.size = %d\n", colNames1.size());

                        for( int x=0; x<colNames1.size(); x++ )
                        {
                            System.out.printf("(%d) [%s]  %s\n", x, colNames0.get(x), colNames1.get(x));
                        }

                        System.out.println("---------------------");
                        System.out.printf("tabHeader.size = %d\n", tabHeader.size());

                        for( int x=0; x<tabHeader.size(); x++ )
                        {
                            System.out.printf("(%d) [%s]  %s\n", x, tabHeader.get(x).head0, tabHeader.get(x).head1);
                        }
                        */


                        // SVA Объединение заголовков
                        {
                            int col_count = realCol - 1;
                            int x = 0;

                            while( x < col_count )
                            {
                                if( is_null(tabHeader.get(x).head0)  )
                                {
                                    tabHeader.get(x).head0 = tabHeader.get(x).head1;
                                    tabHeader.get(x).head1 = "";
                                }
                                else
                                {
                                    int last_idx = x;

                                    String s0 = tabHeader.get(x).head0;

                                    while( last_idx < col_count-1 && s0.equals( tabHeader.get(last_idx+1).head0 ))
                                    {
                                        tabHeader.get(last_idx+1).head0 = "";
                                        last_idx++;
                                    }

                                    if( last_idx > x )
                                    {
                                        tabHeader.get(x).width = last_idx - x + 1;
                                    }

                                    x = last_idx;
                                }

                                x++;            
                            }
                        }


                        //System.out.println("---------------------");
                        //System.out.printf("tabHeader.size = %d\n", tabHeader.size());

                        //for( int x=0; x<tabHeader.size(); x++ )
                        //{
                        //    System.out.printf("(%d) [%s]  %s\n", x, tabHeader.get(x).head0, tabHeader.get(x).head1);
                        //}

                        aline("<thead>");
                        aline("<tr>");
                        
                        for( int x=0; x<tabHeader.size(); x++ )
                        {
                            if( is_not_null(tabHeader.get(x).head0) )
                            {
                                String classFirstCol = (x==0?" firstcol":"");
                                String cellSpan = "";

                                if( tabHeader.get(x).width > 0 )
                                    cellSpan = " colspan=\"" + tabHeader.get(x).width + "\"";
                                else if( is_null(tabHeader.get(x).head1) )
                                    cellSpan = " rowspan=\"2\"";

                                aline("<th class=\"h1" + classFirstCol + "\"" + cellSpan + ">");        

                                aline("<div class=\"table-header\">");
                                aline("<div class=\"header-title\">" + tabHeader.get(x).head0 + "</div>");
                                aline("</div>");
                                aline("<div class=\"column-resizer\" onmousedown=\"resizerMouseDown(event, this)\"></div>");
                                aline("</th>");
                            }
                        }
                        aline("</tr>");

                        aline("<tr>");
                        for( int x=0; x<tabHeader.size(); x++ )
                        {
                            if( is_not_null(tabHeader.get(x).head1) )
                            {
                                aline("<th class=\"h2\">");
                                aline("<div class=\"table-header\">");
                                aline("<div class=\"header-title\">" + tabHeader.get(x).head1 + "</div>");
                                aline("</div>");
                                aline("<div class=\"column-resizer\" onmousedown=\"resizerMouseDown(event, this)\"></div>");
                                aline("</th>");
                            }
                        }
                        aline("</tr>");


                        aline("</thead>");


                        // -----------------------------------------------------
                        
                        ArrayList<String> srow;

                        // расчёт итогов
                        if( needTotS.equals("Y") )
                        {
                            if( app.NLS.currentLang.code.equals("RUS") )
                            {
                                //if(!progress("Подсчёт количества...", "")) return "";
                                //SVA ??? проверить
                                if(!progress("Подсчёт количества...", "")) return;
                            }
                            else
                            {
                                //if(!progress("Calculating totals...", "")) return ""; 
                                //SVA ??? проверить
                                if(!progress("Calculating totals...", "")) return; 
                            }

                            queryText = "count(*) \"row$cnt\"";

                            for( int i=0; i<cols; i++ )
                            {
                                if( colVisible.get(i) )
                                {
                                    String colCode = "";
                                    colCode = colList.get(i).alias;
                                    if( !frm.elements.containsKey(colCode) ) continue;
                                    LuiAppElement el = frm.elements.get(colCode);
                                    String totalExpr = "";
                                    totalExpr = el.getProperty("TOTAL");

                                    if( is_not_null(totalExpr) )
                                    {
                                        String format = el.getProperty("FORMAT");
                                        if( is_not_null(format) )
                                        {
                                            totalExpr = "to_char(" + totalExpr + ",'" + format + "')";
                                        }

                                        queryText = queryText + ", " + totalExpr + " \"" + el.getCode() + "\" \n";
                                    }
                                }
                            }

                            if( is_not_null(queryText) )
                            {
                                queryText = "select /*Total Row*/\n"+queryText+" from (\n"+frm.makeMainQuery()+"\n)q$total ";

                                String qbewhere = frm.addWhere(lc, frm.getProperty("inner_qbe"),false);

                                if( is_not_null(qbewhere) )
                                {
                                    if( frm.getProperty("inner_qbe").equals("Y") )
                                    {
                                        if( queryText.indexOf("/*ADD_LUI$_INNER_QBE*/") > 0 )
                                                queryText = queryText.replace("/*ADD_LUI$_INNER_QBE*/"," and ("+qbewhere.substring(7)+")");
                                        else
                                        {
                                            if( queryText.indexOf("/*LUI$_INNER_QBE*/") > 0 )
                                                queryText = queryText.replace("/*LUI$_INNER_QBE*/", qbewhere);
                                        }
                                    } 
                                    else
                                        queryText = queryText + qbewhere + "\n";
                                }
                            }

                            tot = lc.getSelectFirstRowValMap(queryText);
                            totalRows = Integer.valueOf(tot.get("row$cnt"));
                            tot4Ind = totalRows;
                            if( tot4Ind > maxRows ) tot4Ind = maxRows;
                        }


                        aline("<tbody>");
                        
                        //основной цикл
                        // расчитаем сколько раз показывать progress-bar

                        int delta=1;
                        if( tot4Ind < 200 ) { delta = 20; } //5раз
                        else if( tot4Ind < 1000 ) { delta = 10; } //10раз
                        else if( tot4Ind < 20000 ) { delta = 5; } //20раз
                        else if( tot4Ind < 50000 ) { delta = 2; } //50раз

                        int oldPct = 0;
                        int newPct = 0;

                        while( lc.CursorNextRow(cursor) )
                        {
                            r++;
                            newPct = (r-1)*100/tot4Ind;

                            if( newPct-oldPct >= delta )
                            {
                                oldPct = newPct;
                                if( !progress(String.valueOf(r) + "/" + String.valueOf(tot4Ind), String.valueOf(newPct)) )
                                    //return "";
                                    //SVA ??? проверить
                                    return;
                            }

                            aline("<tr>");



                            if( r > maxRows )
                            {
                                String s;
                                if( app.NLS.currentLang.code.equals("RUS") )
                                    s = "...Превышено разрешённое количество выводимых строк!";
                                else
                                    s = "...Exceeded the allowed number of output lines!";
                                
                                aline("<tr><td colspan=\""+cols+"\" class=\""+makeCellStyle("STOP")+"\">"+s+"</td></tr>");
                                break;
                            }

                            srow = lc.CursorGetRow(cursor); 

                            // укажем форме текущую строку
                            frm.setVal("FetchedRows", String.valueOf(r));
                            frm.setVal("CurrentRow",  String.valueOf(r));

                            //устанавливаем значения элементов
                            for( int i=0; i<cols; i++ )
                            {
                                String colCode = "";
                                colCode = colList.get(i).alias;

                                if( !frm.elements.containsKey(colCode) )
                                    continue;

                                LuiAppElement el = frm.elements.get(colCode);

                                String ldt = el.getProperty("DATA_TYPE");
                                String valBuffer = srow.get(i);

                                if( valBuffer == null )
                                {
                                    if( "EMPTY_STRING".equals(app.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                                    && is_not_null(app.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol) )
                                    {
                                        //если есть особый режим работы с NULL
                                        valBuffer = app.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
                                    }
                                }

                                if( "".equals(valBuffer) )
                                {
                                    if( "NULL".equals(app.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                                    && is_not_null(app.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol) )
                                    {
                                        //если есть особый режим работы с NULL
                                        valBuffer = app.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
                                    }
                                }      

                                // и запихиваем в значение столбца
                                el.setValue(valBuffer, false, "", false);

                            }

                            // значения столбцов установлены       

                            int icn = 0;
                            String dv = "";

                            for( int i=0; i<cols; i++ )
                            {
                                String colCode = "";
                                colCode = colList.get(i).alias;

                                if( !frm.elements.containsKey(colCode) )
                                    continue;

                                if( colVisible.get(i) )
                                {
                                    if( srow.get(i) == null )
                                        dv = "";
                                    else
                                    {
                                        if( srow.get(i).length() > 32767 )
                                            dv = srow.get(i).substring(0, 32000);
                                        else
                                            dv = srow.get(i);
                                    }

                                    icn++;

                                    LuiAppElement el = frm.elements.get(colCode);
                                    

                                    if( !is_null(dv) )
                                    {
                                        if( p_coloring )
                                        {
                                            String makeup = el.getProperty("MAKEUP");
                                            String getup  = el.getProperty("GETUP");

                                            if( is_not_null(makeup) && is_not_null(getup) )
                                                if( colored_cols.Get_Element(colCode).equals("Y") )
                                                    dv = makeRichText(makeup, getup, dv);

                                        }
                                    }

                                    aline("<td class=\""+makeCellStyle(getElementVisualProps(el))+"\">"+dv+"</td>");
                                }
                            }
                        }

                        aline("</tr>");

                        //SVA
                        //printStyles();


                        //if( !progress("", "100") ) return "";
                        //SVA ??? проверить
                        if( !progress("", "100") ) return;


                        //итоги                     

                        if( needTotS.equals("Y") )
                        {
                            r++;

                            String a = (app.NLS.currentLang.code.equals("RUS") ? "Всего строк: ":"Total rows: ") + totalRows;
                            aline("<tr><td colspan=\"" + cols + "\" class=\""+makeCellStyle(head_decor)+"\">" + a + "</td></tr>");

                            r++;
                            aline("<tr>");

                            int icn = 0; 

                            for( int i=0; i<cols; i++ )
                            {
                                if( colVisible.get(i) )
                                {
                                    String colCode = colList.get(i).alias;
                                    if( !frm.elements.containsKey(colCode) ) continue;
                                    LuiAppElement el = frm.elements.get(colCode);
                                    icn++;
                                    String totalExpr = "";

                                    totalExpr = el.getProperty("TOTAL");

                                    String dv = tot.get(colCode);

                                    if( is_not_null(totalExpr) )
                                    {
                                        if(tot==null || !tot.containsKey(colCode) ||is_null(tot.get(colCode)) )
                                        {
                                            dv="";
                                        }
                                        else
                                            dv=tot.get(colCode);
                                    }

                                    if( dv == null ) dv = "";

                                    aline("<td class=\""+makeCellStyle(head_decor + " " + el.getProperty("ALIGNMENT"))+"\">"+dv+"</td>");
                                }
                            }

                            aline("</tr>");
                        }

                        restoreCurrentPosition(frm, colList);                        
                    }
                    finally
                    {
                            if( cursor!= null && lc.CursorisOpen(cursor) ) lc.CursorClose(cursor);
                            if( ac.getSessionMode() == 'C' )
                            {
                                ac.connPool.SetFree(lc);
                            }

                        restoreCurrentPosition(frm, colList);
                    }
                }   
            }  
        }
        
        aline("</tbody>");
        aline("</table>");
        aline("</div>");

    }

    // SVA ---------------------------------------------------------------------
    
    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        ArrayList<String> formList = new ArrayList<>();
        LuiClEvent ev;
        String imageBase64;
        byte[] imageData;
        BufferedImage bufferedImage;
        LuiAppForm frm;
        String curFrmID = null;       
        LuiCollection cs = new LuiCollection("Print:" + pString);
        String maxRowsS = cs.Get_Element("Print");
        String needTotS = cs.Get_Element("TOTALS");
        int maxRows = Integer.valueOf(NVL(maxRowsS, "1000"));
        boolean singleForm = cs.Get_Element("SCOPE").equals("F");
        String mainFormID = cs.Get_Element("FormID");

        if( is_null(mainFormID) ) return "";
        if( !app.forms.containsKey(mainFormID) ) return "";
        frm = app.forms.get(mainFormID);
       
        if( frm == null )
        {
            curFrmID = app.getVal("FormID");
            
            if( app.forms.containsKey(curFrmID) )
            {
               frm = app.forms.get(curFrmID);
            }
            else
                return "";
        }
        
        curFrmID = frm.getFormID();

        // поскольку вероятно придётся давать команды браузеру для сохранения форм,
        // то сначала мы отправим серверу визуализации все имеющиеся в буферах команды
        app.changes();
        app.getSeance().remZeroCmd(); // удалить пустышки из очереди команд (пустые группы)
        app.getSeance().remFrmWinApp(); // удаление форм и окон...

        //SVA  cw.flush();
        //SVA сейчас нельзя. 
        
        // кроме того удалим из очереди событий клиента всё что там есть сейчас
        er.eventList.clear();
        
        // и проверим не исчезла ли наша форма и её окно
        if( !app.forms.containsKey(curFrmID) )
            return "";

        String winID = frm.getVal("WindowID"); // узнаём код окна формы

        if( !app.windows.containsKey(winID) )
            return"";
       
        // SVA
        LuiClCommandGroup XMLcg = cw.addGroup("GetChart");
        Boolean doGetChart = false;       
        
        for( String fc : app.forms.keySet() ) // крутим все формы
        {
            if( singleForm && (!fc.equals(mainFormID)) )
                continue;

            // расположенные в том же окне и являющиеся диаграммами
           
            if (!app.forms.get(fc).getVal("WindowID").equals(winID))
                continue;

            if (app.forms.get(fc).getType().equals("LIST"))
            {
                //nTabs++;
                formList.add(fc);
                continue;
            }
            
            if (app.forms.get(fc).getType().equals("CHART"))
            {
                //nTabs++;
                formList.add(fc);

                LuiAppForm cfrm = app.forms.get(fc); // берём форму-диаграмму
                
                // выполняем подготовку команды серверу визуализации для того чтобы браузер сгенерировал PNG
                LuiClCommand XMLcmd = XMLcg.addCommand("GetChart"); //вот это совершенно новая команда
                XMLcmd.addScaPar("FormID", fc);

                doGetChart = true;

                // теперь надо отправить команду Жене
                //SVA cw.flush();
                //SVA сейчас нельзя отправлять. Отправим один раз после того как все команды быдут добавлены.
            }
        }

        if( doGetChart )
        {
            // отправляем команды GetChart

            cw.flush();

            //SVA ВОЗМОЖНО ЭТО ПОТЕНЦИАЛЬНАЯ ПРОБЛЕМА
            // Если картинок нет, то ничего не отправляем. И тогда не уйдут команды,
            // которые возможно былы в буферах. 
            
           
            // в ответ мы ждём событе Image о готовности диаграммы в base64

            er.readEvents();

            // цикл обработки событий
            while( (ev = er.getEvent()) != null  )
            {
                if( ev.name.equals("Image") )
                {
                    saveImage(ev.getParValue("FormID"), ev.getParValue("Data"));
                }
            }
        }

        
        aline("<div class=\"html-report\">");
        aline("<div class=\"tab\">");

        for(int i = 0; i < formList.size(); i++)
        {
            LuiAppForm ifrm = app.forms.get(formList.get(i));

            String title;
            
            if( ifrm.getType().equals("CHART") )
            {
                title = ifrm.DgV2.ChartTitle;
                
                if( is_null(title) ) title = ifrm.getFormID();
            }
            else
            {
                title = ifrm.getProperty("HEADER");
            }
                       
            aline("<div class=\"tablinks"+(i==0?" active":"")+"\" onclick=\"showTab(event, 'tab"+i+"')\" title=\""+title+"\">"+title+"</div>");                      
        }

        aline("</div>");           
        
        int i = 0;
        
        // снова крутим формы
        for( String fc : app.forms.keySet() ) // крутим все формы
        {
            if( singleForm && (!fc.equals(mainFormID)) )
                continue;

            if( !app.forms.get(fc).getVal("WindowID").equals(winID) )
                continue; // расположенные в том же окне 

            if( app.forms.get(fc).getType().equals("CHART") )
            {
                aline("<div id=\"tab"+(i)+"\" class=\"tabcontent"+(i==0?" active":"")+"\">");

                makeChartSheet(fc);
                aline("</div>");
                i++;
            }

            if( app.forms.get(fc).getType().equals("LIST") )
            {
                //aline("<div id=\"tab"+(i++)+"\""+(i==0?" class=\"tabcontent active\"":"")+">");
                aline("<div id=\"tab"+(i)+"\" class=\"tabcontent"+(i==0?" active":"")+"\">");
                aline("<div class=\"table-wrap\">");
                makeListSheet(fc);
                aline("</div>");
                aline("</div>");
                i++;
            }
        }
              
        aline("</div>");
        
        /*
        }
        catch (IOException ex)
        {
            if (filew!=null){
                              try{
                                  filew.flush();
                                  filew.close();
                                  filew=null;
                              } catch(Exception ef) {}
                            }
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());

        }
        */
        
        //SVA DEBUG
        if( System.getProperty("lui_debug", "N").equals("Y") ) return null;
        
        return bookFileName;
    }

    // SVA ---------------------------------------------------------------------

    protected boolean createWorkbook(String formTitle) throws LuiException, SQLException//, FileNotFoundException
    {
        clearMemory();

        bookFileName = app.getTmpDir()+"..\\output.html";
       
        String test = getParameter("TEST");

        htmlHead  = getParameter("LIST2HTML_PAGE");
        htmlFloor = getParameter("LIST2HTML_PAGE2");
               
        try
        {
            File file = new File(bookFileName);
            file.createNewFile(); // если файл существует - команда игнорируется
            FileOutputStream fileOutputStream = new FileOutputStream(file, false);
            filew = new OutputStreamWriter(fileOutputStream, StandardCharsets.UTF_8);

            //System.out.println("bookFileName = " + bookFileName);

            line("<!DOCTYPE html>");
            line("<html>");
            line("<head>");
            line("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"/>");
            line("<title>"+formTitle+"</title>");
        }
        catch( FileNotFoundException ex )
        {
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
            return false;
        }
        catch( IOException ex )
        {
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
        }

        
        return true;
    }
    
    // SVA ---------------------------------------------------------------------

    protected void closeWorkbook() throws LuiException
    {
        try
        {
            filew.write(htmlHead);
            line("");   

            printStyles();        

            line("</head>");
            line("<body onload=\"doc_resize()\">");
            
            for(int i = 0; i < lines.size(); i++)
            {
                line(lines.get(i));
            }

            filew.write(htmlFloor);

            line("</body>");
            line("</html>");
           
            filew.flush();
            filew.close();
        }
        catch( FileNotFoundException ex )
        {
            // SVA зачем это все???
            /*
            if( filew != null )
            {           
                try
                {
                    filew.flush();
                    filew.close();
                    filew = null;
                } catch(Exception ef) {}
            }
            */
            
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
        }
        
        catch( IOException ex )
        {
            /*
            if( filew != null )
            {
                try
                {
                    filew.flush();
                    filew.close();
                    filew = null;
                }
                catch(Exception ef) {}
            }
            */
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
        }

        clearMemory();
    }

    // SVA ---------------------------------------------------------------------
    
    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        String adr = "";
        String fc = app.getVal("FormID");
        if( is_null(fc) ) return;
        if( !app.forms.containsKey(fc) ) return;

        cs = new LuiCollection("Print:" + pString);
       
        LuiAppForm frm=app.forms.get(fc);
        String formTitle = frm.getProperty("HEADER");
        
        if( createWorkbook(formTitle) )
        {
            p_stream_mode = cs.Get_Element("STREAM_MODE").equals("Y");
            p_coloring    = cs.Get_Element("COLORING").equals("Y");
	    p_coloring    = true;

            String columns = cs.Get_Element("COLUMNS").replace("=", ":").replace("#", ";");

            if( p_stream_mode ) p_coloring = false;

            if( p_coloring ) colored_cols = new LuiCollection(columns);
            
            adr = compute(pString, pProp);
            
            closeWorkbook();
        }

        if( is_not_null(adr) )
        {
            LuiClCommandGroup XMLcg = app.commandWriter.addGroup("PRINT");
            LuiClCommand XMLcmd = XMLcg.addCommand("URL");
            XMLcmd.addScaPar("Address", adr);
            XMLcmd.addScaPar("Target", "Tab");
        }
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
    
    private boolean progress(String pText,String pPct) throws LuiException, SQLException {
        boolean retVal=true;

        if( elInd != null )
        {
            elInd.setProperty("VISIBLE","1");
            
            if( elIndOK != null )
            {
                elIndOK.setProperty("ACTIVE","N");
            }
            
            if( is_not_null(pText) )
            {
                elInd.setProperty("HELPER_TEXT", "   " + pText);
            }
            else
            {
                elInd.setProperty("HELPER_TEXT", "");
            }

            if( is_not_null(pPct) )
            {
                elInd.setProperty("VISUALIZATION", "PROGRESS PROC" + pPct);
                elInd.setProperty("Value", "   " + pPct+"%");
            }
            else
            {
                elInd.setProperty("VISUALIZATION", "");
                elInd.setProperty("Value", "");
            }

            elInd.properties.get("Value").setChangeFlag();
            elInd.properties.get("HELPER_TEXT").setChangeFlag();
            elInd.properties.get("VISUALIZATION").setChangeFlag();
            elInd.properties.get("Show").setChangeFlag();
            frmInd.changes();
            XMLcg = app.commandWriter.addGroup("CallBack");
            XMLcmd = XMLcg.addCommand("Event");
            XMLcmd.addScaPar("EventCode", "ProgressContinue");
            XMLcmd.addScaPar("MS","0");
            app.commandWriter.flush();
            app.eventReader.readEvents();

            while (true)
            {
                LuiClEvent ev = app.eventReader.getEvent();

                if( ev == null ) break;

                if( ev.name.equals("ProgressContinue") || ev.name.equals("GotoField") )
                    retVal = true;
                else
                    retVal = false;        
            }
        }

        return retVal;
    }
}
