/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Формирует XSLX-файл output.xslx из данных списка во временном каталоге сеанса
 Compute возвращает url на этот файл
 Execute шлёт команды клиенту для открытия файла
 синтаксис List2XSSF:nnnn;TOTALS:Y;
 */
package Standard;

import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import java.util.*;

import LuiCollection.LuiCollection;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;

import luiserver.LuiAppConn;
import luiserver.LuiAppElement;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiClCommand;
import luiserver.LuiClCommandGroup;
import luiserver.LuiClCommandWriter;
import luiserver.LuiClEvent;
import luiserver.LuiClEventReader;
import luiserver.LuiPlugin;

import ColorFactory.ColorFactory;
import WikiToRichText.WikiToExcell;

import org.openxmlformats.schemas.drawingml.x2006.main.CTPositiveSize2D;
import org.openxmlformats.schemas.spreadsheetml.x2006.main.CTCol;

import org.apache.poi.ss.usermodel.VerticalAlignment;
import org.apache.poi.ss.usermodel.CreationHelper;
import org.apache.poi.ss.usermodel.Drawing;
import org.apache.poi.ss.usermodel.ClientAnchor;
import org.apache.poi.ss.usermodel.Picture;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.PrintOrientation;
import org.apache.poi.ss.usermodel.PrintSetup;
import org.apache.poi.ss.usermodel.BorderStyle;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.CellType;
import org.apache.poi.ss.usermodel.FillPatternType;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.HorizontalAlignment;
import org.apache.poi.ss.usermodel.IndexedColors;
import org.apache.poi.ss.usermodel.ShapeTypes;

import org.apache.poi.ss.util.WorkbookUtil;
import org.apache.poi.ss.util.CellRangeAddress;

import org.apache.poi.util.POILogFactory;
import org.apache.poi.util.POILogger;
import org.apache.poi.util.Units;

import org.apache.poi.xssf.streaming.SXSSFPicture;
import org.apache.poi.xssf.usermodel.XSSFClientAnchor;
import org.apache.poi.xssf.usermodel.XSSFPicture;
import org.apache.poi.xssf.usermodel.XSSFPictureData;
import org.apache.poi.xssf.usermodel.XSSFPrintSetup;
import org.apache.poi.xssf.usermodel.XSSFRichTextString;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.apache.poi.xssf.streaming.SXSSFSheet;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.apache.poi.xssf.usermodel.DefaultIndexedColorMap;
import org.apache.poi.xssf.usermodel.TextAlign;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFFont;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFDrawing;
import org.apache.poi.xssf.usermodel.XSSFSimpleShape;
import org.apache.poi.xssf.usermodel.XSSFTextParagraph;

import net.htmlparser.jericho.Config;
import net.htmlparser.jericho.Element;
import net.htmlparser.jericho.Source;
import net.htmlparser.jericho.Attribute;
import net.htmlparser.jericho.Attributes;
import HtmlToRichText.HTMLtoExcel;
import java.time.LocalTime;
import org.apache.poi.ss.usermodel.DataFormat;
import org.apache.poi.ss.usermodel.IgnoredErrorType;
import org.apache.poi.ss.util.CellReference;
import org.apache.poi.xssf.usermodel.TextAutofit;

/**
 *
 * @author vsenchenko
 */

public class List2XSSF2 implements LuiPlugin
{
    boolean debug_mode = false;
    String s160 = " ";
    LuiApplication app;
    LuiClCommandWriter cw;
    LuiClEventReader er;
    String bookFileName;
    FileOutputStream filew = null;
    Workbook workbook = null;
    LuiAppForm frmInd = null;
    LuiAppElement elInd = null;
    LuiAppElement elIndOK = null;
    LuiClCommandGroup XMLcg = null;
    LuiClCommand XMLcmd = null;
    HashMap <String, String> storedRow = new HashMap<>();
    HashMap <String, CellStyle> mapCellStyles = new HashMap<>();
    HashMap <String, Font> mapFonts = new HashMap<>();
    LuiCollection cs = null;
    HashMap <String, byte[]> mapCharts = new HashMap<>();
    LuiCollection colored_cols = null;
    boolean p_stream_mode;
    boolean p_coloring;
    String head_color = "#E8E8E8";
    String head_decor = "BACKGROUND(" + head_color + ")";

    XSSFColor borderColor = new XSSFColor(ColorFactory.valueOf("#404040"));
    
    private static final POILogger logger = POILogFactory.getLogger(XSSFPicture.class);
    private static float DEFAULT_COLUMN_WIDTH = 9.140625f;

    // --- For GRID
    String default_separator;
    String default_fractional;
    LuiPlugin cellPropertyPlugin;

    public class valClass
    {
        int value;
        
        valClass(int val)
        {
            value = val;
        }
    };
       
    // SVA ---------------------------------------------------------------------

    public class Decor
    {
        HashMap <String, String> props = new HashMap<>();

        Decor(String ps)
        {
            setProps(ps);
        }
        
        public void setProps(String ps)
        {           
            Map<String, String> prop_list = splitParams(ps);
            
            for( String prop : prop_list.keySet() )
            {
                String arg = prop_list.get(prop);
                addProp(prop, arg.equals("")?"<noval>":arg);
            }
            
        }

        public void addProp(String prop, String val, String defaultVal)
        {
            props.put(prop, val.equals("")?defaultVal:val.replaceAll(" ", ""));
        }

        public void addProp(String prop, String val)
        {
            if(!val.equals(""))
                addProp(prop, val, "");
        }

        public void addProp(String prop)
        {
            addProp(prop, "<noval>");
        }
        
        public String getProps()
        {
            StringBuffer s = new StringBuffer();
            
            for( String prop : props.keySet() )
            {
                String val = props.get(prop);
                s.append(" " + prop);
                if( !val.equals("<noval>") )
                    s.append("(" + val + ")");
            }

            return s.substring(1).toString();
        }
    };

    // SVA ---------------------------------------------------------------------
    
    @Override
    public void setContext(LuiApplication pContext)
    {
        app=pContext;
        er=app.eventReader;
        cw=app.commandWriter; 

        debug_mode = System.getProperty("lui_debug", "N").equals("Y");
    }

   
    // SVA ---------------------------------------------------------------------

    private void pause(int secs)
    {
        try
        {
            Thread.sleep(secs * 1000);
        }
        catch( InterruptedException ex)
        {
            DEBUG("IOException: " + ex.getLocalizedMessage());
        }
    }

    // SVA ---------------------------------------------------------------------

    private void DEBUG(String str)
    {
       if(debug_mode) System.out.println(str);
    }

    private void DEBUGP(String str)
    {
       if(debug_mode) System.out.print(str);
    }

    private void DEBUG(String format, Object... arguments)
    {
        if(debug_mode) System.out.printf(format+"\n", arguments);
    }

    // SVA ---------------------------------------------------------------------
    
    private void DEBUG_FORM_INFO(LuiAppForm frm) throws LuiException, SQLException
    {
        if(!debug_mode) return;

        // все параметры формы (для отладки)
        DEBUG("------ Form Properties ------");
        for(String k : frm.properties.keySet())
        {
            DEBUG("frm.prop %s = %s", k, frm.getProperty(k));            
        }

        DEBUG("~~~~~~ Form Properties ~~~~~~");

        // все элементы формы
        DEBUG("------ Form Elements ------");
        for( String ek : frm.elements.keySet() )
        {
            DEBUG("--->>> %s (%s)", frm.elements.get(ek).getType(), ek);                       
            LuiAppElement el = frm.elements.get(ek);            
            DEBUG("el.Code(%s) el.Type=%s", ek, el.getType());

            for( String pk : el.properties.keySet() )
            {
                LuiAppProp p = el.properties.get(pk);               
                DEBUG("    [%s] %s", pk, p.getCurrentValue());
            }
        }
        DEBUG("~~~~~~ Form Elements ~~~~~~");

        if(frm.MatrixMap == null) return;
        
        DEBUG("------ MatrixMap ------");

        for(String k : frm.MatrixMap.keySet())
        {
            String s = "";
            ArrayList<String> cellList = frm.MatrixMap.get(k);
            for(int i = 0; i < cellList.size(); i++)
            {
                s = s + cellList.get(i) + "<"+frm.elements.get(cellList.get(i)).getProperty("CELL_SHOW")+">, ";
            }

            DEBUG("-- [%s] => %s", k, s);
        }

        DEBUG("~~~~~~ MatrixMap ~~~~~~");
    }
    
    // SVA ---------------------------------------------------------------------

    private static Map<String, String> splitParams(String str)
    {
        HashMap<String, String> params = new HashMap<>();

        Pattern pattern = Pattern.compile("(\\S+)\\((\\S+\\([\\ \\S]+?\\))\\)|(\\S+)\\(([\\ \\S]+?)\\)|(\\S+)");
       
        Matcher matcher = pattern.matcher(str);
        
        while( matcher.find() )
        {
            int i=0;
            
            if( matcher.group(5) != null )
                i = 5;
            else if( matcher.group(3) != null )
                i = 3;
            else if( matcher.group(1) != null )
                i = 1;

            if(i==0)
                throw new ArithmeticException("Impossible error!");

            
            String arg = i<5?matcher.group(i+1):"";
            //System.out.println(matcher.group(p) + " = " + arg);

            params.put(matcher.group(i), arg);
        }

        return params;
    }
    
    // SVA ---------------------------------------------------------------------

    private static String cellNum(int x, int y)
    {
        return (new CellReference(y, x)).formatAsString();
    }

    // SVA ---------------------------------------------------------------------
    
    
    private void clearMemory()
    {
        filew = null;
        workbook = null;
        mapCellStyles.clear();
        mapFonts.clear();
        mapCharts.clear();
        if(colored_cols != null) colored_cols.clear();
        System.gc();
    }

    // SVA ---------------------------------------------------------------------

    private XSSFColor cfColor(String color)
    {
        java.awt.Color c = ColorFactory.valueOf(color);
        
        return new XSSFColor(c);
    }

    // SVA ---------------------------------------------------------------------
    
    private XSSFColor cfColor(String color, String bgColor)
    {
        java.awt.Color c = ColorFactory.valueOf(color);

        //DEBUG("cfColor: c = " +  c.toString() + " a = " + c.getAlpha());

        int alpha = c.getAlpha();
                
        if( alpha < 255 )
        {
            java.awt.Color b = ColorFactory.valueOf(bgColor);

            //DEBUG("cfColor: b = " +  b.toString() + " a = " + b.getAlpha());
            
            java.awt.Color q = new java.awt.Color(
                c.getRed()   * alpha / 255 +  b.getRed()   * ( 255 - alpha ) / 255,
                c.getGreen() * alpha / 255 +  b.getGreen() * ( 255 - alpha ) / 255,
                c.getBlue()  * alpha / 255 +  b.getBlue()  * ( 255 - alpha ) / 255
            );

            //DEBUG("cfColor: q = " +  q.toString() + " a = " + q.getAlpha());
            
            return new XSSFColor(q);
        }
        
        return new XSSFColor(c);
    }
   
    // SVA ---------------------------------------------------------------------
    
    public void saveImage(String formId, String imageBase64) throws LuiException, SQLException
    {
        byte[] imageData = java.util.Base64.getDecoder().decode(imageBase64.substring(imageBase64.indexOf(",") + 1));

        mapCharts.put(formId, imageData);
    }   
    
    // SVA ---------------------------------------------------------------------

    // SVA taken from SXSSFPicture.java for correct resize XSSFPicture to original size
    private float getColumnWidthInPixels(XSSFPicture pict, int columnIndex){
        XSSFSheet sheet = pict.getSheet();

        CTCol col = sheet.getColumnHelper().getColumn(columnIndex, false);
        double numChars = col == null || !col.isSetWidth() ? DEFAULT_COLUMN_WIDTH : col.getWidth();

        return (float)numChars*Units.DEFAULT_CHARACTER_WIDTH;
    }
    
    // SVA taken from SXSSFPicture.java for correct resize XSSFPicture to original size
    private static float getRowHeightInPixels(XSSFPicture pict, int rowIndex) {
        // THE FOLLOWING THREE LINES ARE THE MAIN CHANGE compared to the non-streaming version: use the SXSSF sheet,
		// not the XSSF sheet (which never contais rows when using SXSSF)
        //SVA XSSFSheet xssfSheet = pict.getSheet();
        //SVA SXSSFSheet sheet = _wb.getSXSSFSheet(xssfSheet);

        XSSFSheet sheet = pict.getSheet();

        Row row = sheet.getRow(rowIndex);

        float height = row != null ?  row.getHeightInPoints() : sheet.getDefaultRowHeightInPoints();
        return height * Units.PIXEL_DPI / Units.POINT_DPI;
    }
    
    // SVA taken from SXSSFPicture.java for correct resize XSSFPicture to original size
    public XSSFClientAnchor getPreferredSize(XSSFPicture pict, double scale){
        XSSFClientAnchor anchor = pict.getClientAnchor();
        if (anchor == null) {
            logger.log(POILogger.WARN, "picture is not anchored via client anchor - ignoring resize call");
            return null;
        }

        XSSFPictureData data = pict.getPictureData();

        //SVA (protected)Dimension size = pict.getImageDimension(data.getPackagePart(), data.getPictureType());
        Dimension size = pict.getImageDimension();

        double scaledWidth = size.getWidth() * scale;
        double scaledHeight = size.getHeight() * scale;

        float w = 0;
        int col2 = anchor.getCol1()-1;

        while (w <= scaledWidth) {
            w += getColumnWidthInPixels(pict, ++col2);
        }

        assert (w > scaledWidth);
        double cw = getColumnWidthInPixels(pict, col2);
        double deltaW = w - scaledWidth;
        int dx2 = (int)(Units.EMU_PER_PIXEL * (cw - deltaW));

        anchor.setCol2(col2);
        anchor.setDx2(dx2);

        double h = 0;
        int row2 = anchor.getRow1()-1;

        while (h <= scaledHeight) {
            h += getRowHeightInPixels(pict, ++row2);
        }

        assert (h > scaledHeight);
        double ch = getRowHeightInPixels(pict, row2);
        double deltaH = h - scaledHeight;
        int dy2 = (int)(Units.EMU_PER_PIXEL * (ch - deltaH));
        anchor.setRow2(row2);
        anchor.setDy2(dy2);

        CTPositiveSize2D size2d =  pict.getCTPicture().getSpPr().getXfrm().getExt();
        size2d.setCx((long)(scaledWidth * Units.EMU_PER_PIXEL));
        size2d.setCy((long)(scaledHeight * Units.EMU_PER_PIXEL));

        return anchor;
    }
    
    // SVA taken from SXSSFPicture.java for correct resize XSSFPicture to original size
    public void pict_resize(XSSFPicture pict, double scale){
        XSSFClientAnchor anchor = pict.getClientAnchor();
        XSSFClientAnchor pref = getPreferredSize(pict, scale);
        if (anchor == null || pref == null) {
            DEBUG("picture is not anchored via client anchor - ignoring resize call");
            return;
        }

        int row2 = anchor.getRow1() + (pref.getRow2() - pref.getRow1());
        int col2 = anchor.getCol1() + (pref.getCol2() - pref.getCol1());

        anchor.setCol2(col2);
        anchor.setDx1(0);
        anchor.setDx2(pref.getDx2());

        anchor.setRow2(row2);
        anchor.setDy1(0);
        anchor.setDy2(pref.getDy2());
    }

    // SVA for correct resize XSSFPicture to original size
    private void resize_pict_to_original_size(Picture pict)
    {
        if( pict instanceof XSSFPicture )
        {
            pict_resize(((XSSFPicture)pict), 1.0);
        }

        if( pict instanceof SXSSFPicture )
        {
            pict.resize(1.0);
        }
    }

    // SVA ---------------------------------------------------------------------

    private void buildChartSheet(String fc) throws LuiException, SQLException
    {
        LuiAppForm frm = app.forms.get(fc);
        
        Sheet sheet = workbook.createSheet(
            WorkbookUtil.createSafeSheetName(fc)
            //WorkbookUtil.createSafeSheetName(frm.getProperty("HEADER"))
            //WorkbookUtil.createSafeSheetName(frm.DgV2.ChartTitle)
        );

        CreationHelper helper = workbook.getCreationHelper();

        int pictureIdx = workbook.addPicture(mapCharts.get(fc), Workbook.PICTURE_TYPE_PNG);

        Drawing drawing = sheet.createDrawingPatriarch();

        ClientAnchor anchor = new XSSFClientAnchor();

        anchor.setAnchorType(ClientAnchor.AnchorType.DONT_MOVE_AND_RESIZE);

        Picture pict = drawing.createPicture(anchor, pictureIdx);
       
        resize_pict_to_original_size(pict);
        
        Dimension dim = pict.getImageDimension();        

        if( dim.getWidth() > dim.getHeight() )
        {
            PrintSetup ps = sheet.getPrintSetup();
            ((XSSFPrintSetup)ps).setOrientation(PrintOrientation.LANDSCAPE);
        }
    }
    
    // SVA ---------------------------------------------------------------------
    
    private String getElementVisualProps(LuiAppElement el) throws LuiException, SQLException
    {
        // Извлекаем свойства ячейки
        String visualProps = el.properties.get("VISUALIZATION").getComputedProperty().trim();
        String alignmProps = el.getProperty("ALIGNMENT").trim();
        String props = (visualProps + " " + alignmProps).trim();
        
        return props;
    }

    // SVA ---------------------------------------------------------------------
            
    private CellStyle makeCellStyle(String props)
    {
        if( is_null(props) )
        {
            props = "<<EMPTY>>";
            // Вообще-то наверно пусто не бывает. Всегда есть ALIGNMENT?
        }

        if( mapCellStyles.containsKey(props) )
        {
            return mapCellStyles.get(props);
        }
            
        CellStyle style = workbook.createCellStyle();

        
        // ВРЕМЕННО: рамки для списков?
        style.setBorderBottom(BorderStyle.THIN);
        style.setBottomBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setBottomBorderColor(borderColor);
        style.setBorderLeft(BorderStyle.THIN);
        style.setLeftBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setLeftBorderColor(borderColor);
        style.setBorderRight(BorderStyle.THIN);
        style.setRightBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setLeftBorderColor(borderColor);
        style.setBorderTop(BorderStyle.THIN);
        style.setTopBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setTopBorderColor(borderColor);
        
        
        Font font = workbook.createFont();

        // default colors
        String fnColor = "#000000";
        String bgColor = "#FFFFFF";
        //String bgColor = "#FAFAFA";


        Map<String, String> prop_list = splitParams(props);
        
        for( String prop : prop_list.keySet() )
        {
            String arg = prop_list.get(prop);           
            
            switch (prop)
            {                
                // Мои
                
                case "NUMFORMAT":
                    DataFormat format = workbook.createDataFormat();
                    style.setDataFormat(format.getFormat(arg));
                break;

                
                // Старый вариант
                //case "VATOP":    style.setVerticalAlignment(VerticalAlignment.TOP);    break;
                //case "VACENTER": style.setVerticalAlignment(VerticalAlignment.CENTER); break;
                //case "VABOTTOM": style.setVerticalAlignment(VerticalAlignment.BOTTOM); break;

                // Новый вариант
                case "VALIGN":
                    switch (arg)
                    {
                        case "TOP":    style.setVerticalAlignment(VerticalAlignment.TOP);    break;
                        case "CENTER": style.setVerticalAlignment(VerticalAlignment.CENTER); break;
                        case "BOTTOM": style.setVerticalAlignment(VerticalAlignment.BOTTOM); break;
                    }
                break;

                
                case "WRAPTEXT":  style.setWrapText(true); break;
                               
                case "BORDER":
                    //headerCellStyle.setBorderBottom(BorderStyle.THIN);
                    style.setBorderBottom(BorderStyle.DOUBLE);
                    style.setBottomBorderColor(IndexedColors.GREY_50_PERCENT.getIndex());
                    style.setBorderLeft(BorderStyle.THIN);
                    style.setLeftBorderColor(IndexedColors.GREY_50_PERCENT.getIndex());
                    style.setBorderRight(BorderStyle.THIN);
                    style.setRightBorderColor(IndexedColors.GREY_50_PERCENT.getIndex());
                    style.setBorderTop(BorderStyle.THIN);
                    style.setTopBorderColor(IndexedColors.GREY_50_PERCENT.getIndex());
                break;

                case "BBORDER":
                    style.setBorderBottom(BorderStyle.THIN);
                    style.setBottomBorderColor(IndexedColors.BLACK.getIndex());

                    style.setBorderLeft(BorderStyle.THIN);
                    style.setLeftBorderColor(IndexedColors.BLACK.getIndex());

                    style.setBorderRight(BorderStyle.THIN);
                    style.setRightBorderColor(IndexedColors.BLACK.getIndex());

                    style.setBorderTop(BorderStyle.THIN);
                    style.setTopBorderColor(IndexedColors.BLACK.getIndex());
                break;
                
                
                //case "FONTNAME":  font.setFont(font); break;
                case "FONTHEIGHT": font.setFontHeightInPoints( Short.parseShort(arg) ); break;
                
                // -------------------------------------------------------------

                // Старый вариант
                case "RIGHT":   style.setAlignment(HorizontalAlignment.RIGHT);  break;
                case "CENTER":  style.setAlignment(HorizontalAlignment.CENTER); break;
                case "LEFT":    style.setAlignment(HorizontalAlignment.LEFT);   break;

                // Новый вариант
                case "HALIGN":
                    switch (arg)
                    {
                        case "RIGHT":   style.setAlignment(HorizontalAlignment.RIGHT);  break;
                        case "CENTER":  style.setAlignment(HorizontalAlignment.CENTER); break;
                        case "LEFT":    style.setAlignment(HorizontalAlignment.LEFT);   break;                    
                    }
                break;
                
                // -------------------------------------------------------------

                case "TEXTCOLOR":   fnColor = arg;  break;
                case "BACKGROUND":  bgColor = arg;  break;                                            

                // -------------------------------------------------------------

                case "BOTTOM2TOP":   style.setRotation((short)90);  break;
                case "TOP2BOTTOM":   style.setRotation((short)180);  break;

                // -------------------------------------------------------------
                
                
                case "ALL_SELECTED ":   //?? {background-color:#DDF; font-weight:600}
                    fnColor = "#DDF";
                    font.setBold(true);
                break;

                case "ATTENTION":       //?? {color:#970; font-weight:600}
                    fnColor = "#970";
                    font.setBold(true);
                break;

                case "BACKGROUND_SELECTED":
                    bgColor = "#DDF";
                break;

                case "BAD":      //?? {text-decoration:line-through; text-decoration-color:red; color:#888}
                    fnColor = "#888";
                    font.setStrikeout(true);
                break;

                case "BLUE":          //?? {color:royalblue}
                    fnColor = "royalblue";
                break;

                case "BOLD":
                    font.setBold(true);                                    
                break;

                // BOUNDARY {border-bottom:2px solid darkred}

                case "BROWN":          //?? {color:brown}
                    fnColor = "brown";
                break;

                case "CROSSEDOUT":      //?? {text-decoration:line-through; text-decoration-color:red}
                    font.setStrikeout(true);
                    // text-decoration-color:red ???
                break;

                case "CYAN":          //?? {color:cyan; text-shadow:1px 1px 1px darkcyan}
                    fnColor = "cyan";
                    // text-shadow:1px 1px 1px darkcyan ???
                break;

                case "DANGER":      //?? {color:#A00; font-weight:600}
                    fnColor = "#A00";
                    font.setBold(true);                                    
                break;

                case "FONT_SELECTED":   //?? {font-weight:600}
                    font.setBold(true);                                    
                break;

                case "GRAY":      //?? {color:#AAA}
                    fnColor = "#AAA";
                break;

                case "GREEN":          //?? {color:forestgreen}
                    fnColor = "forestgreen";
                break;

                case "INHERIT":
                    bgColor = "#DDD";
                    fnColor = "#444";
                break;

                case "INTERIM":
                    bgColor = "#FF7";
                break;

                case "ITALIC":
                    font.setItalic(true);
                break;

                // LARGE {font-size:larger}

                case "LOWERED":
                    fnColor = "#BBB";
                break;

                case "MAGENTA":          //?? {color:magenta}
                    fnColor = "magenta";
                break;

                case "MONOSPACE":   //?? {font-family:monospace}
                    font.setFontName("Courier New");
                break;

                case "NEGATIVE":    //?? {color:white; background-color:gray; font-weight:600}
                    fnColor = "white";
                    bgColor = "gray";
                    font.setBold(true);                                    
                break;

                //NORMAL {color:black; font-weight:normal; background-color:transparent; text-shadow:none; position:static; text-decoration:none; border:0}

                case "NOTES":
                    font.setItalic(true);
                break;

                case "OK":          //?? {color:#4A0; font-weight:600}
                    fnColor = "#4A0";
                    font.setBold(true);                                    
                break;

                case "ORANGE":          //?? {color:orange}
                    fnColor = "orange";
                break;

                case "SECTION":          //?? {background-color:#888; font-weight:600; color:#FFF}
                    bgColor = "#888";
                    fnColor = "#FFF";
                    font.setBold(true);                                    
                break;

                case "SHADOW":          //?? {text-shadow:1px 0 3px grey}
                    fnColor = "gray";
                    font.setBold(true);                                    
                break;

                case "START":          //?? {background-color:#8F8}
                    bgColor = "#8F8";
                break;

                case "STOP":          //?? {background-color:#F88}
                    bgColor = "#F88";
                break;

                case "TILT":
                    font.setItalic(true);
                break;

                case "UNDERLINE":
                    font.setUnderline(Font.U_SINGLE);
                break;
            }
        }

        // Установим параметры ячейки
       
        // 18.12.2020 ((XSSFFont)font).setColor(new XSSFColor(ColorFactory.valueOf(fnColor),  new DefaultIndexedColorMap()));
        //((XSSFFont)font).setColor(cfColor(fnColor));
        ((XSSFFont)font).setColor(cfColor(fnColor, bgColor));
        
        style.setFont(font);

        style.setFillPattern(FillPatternType.SOLID_FOREGROUND);

        //SVA 18.12.2020 ((XSSFCellStyle)style).setFillForegroundColor(new XSSFColor(ColorFactory.valueOf(bgColor), new DefaultIndexedColorMap()));
        //((XSSFCellStyle)style).setFillForegroundColor(new XSSFColor(ColorFactory.valueOf(bgColor)));
        ((XSSFCellStyle)style).setFillForegroundColor(cfColor(bgColor));

        
        /*
        SVA 17.02.2021  ПРОВЕРИТЬ, ДЛЯ ЧЕГО ЭТО БЫЛО НУЖНО???

        style.setBorderBottom(BorderStyle.THIN);
        style.setBottomBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setBottomBorderColor(borderColor);
        style.setBorderLeft(BorderStyle.THIN);
        style.setLeftBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setLeftBorderColor(borderColor);
        style.setBorderRight(BorderStyle.THIN);
        style.setRightBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setLeftBorderColor(borderColor);
        style.setBorderTop(BorderStyle.THIN);
        style.setTopBorderColor(IndexedColors.GREY_25_PERCENT.getIndex());
        //((XSSFCellStyle)style).setTopBorderColor(borderColor);
        */
        
        
        mapCellStyles.put(props, style);
        
        return style;
    }
    
    // SVA ---------------------------------------------------------------------
    
    private Font makeFontForRichText(String getup)
    {
        if( mapFonts.containsKey(getup) )
        {
            return mapFonts.get(getup);
        }
        
        Font font = workbook.createFont();

        String fnColor = "#000000";
        String prop = getup;
        String arg = "";
        String template = "^(\\S+)\\((\\S+)\\)$";

        if( prop.matches(template) )
        {
            Pattern pattern = Pattern.compile(template);
            Matcher matcher = pattern.matcher(prop);
            if(matcher.matches())
            {
                prop = matcher.group(1);
                arg  = matcher.group(2);
            }
        }

        switch (prop)
        {
            case "TEXTCOLOR":                                               
                fnColor = arg;
            break;

            // -----------------------------------------------------------------

            case "ALL_SELECTED ":   //?? {background-color:#DDF; font-weight:600}
                fnColor = "#DDF";
                font.setBold(true);
            break;

            case "ATTENTION":       //?? {color:#970; font-weight:600}
                fnColor = "#970";
                font.setBold(true);
            break;

            case "BAD":      //?? {text-decoration:line-through; text-decoration-color:red; color:#888}
                fnColor = "#888";
                font.setStrikeout(true);
            break;

            case "BLUE":          //?? {color:royalblue}
                fnColor = "royalblue";
            break;

            case "BOLD":
                font.setBold(true);                                    
            break;

            // BOUNDARY {border-bottom:2px solid darkred}

            case "BROWN":          //?? {color:brown}
                fnColor = "brown";
            break;

            case "CROSSEDOUT":      //?? {text-decoration:line-through; text-decoration-color:red}
                font.setStrikeout(true);
            break;

            case "CYAN":          //?? {color:cyan; text-shadow:1px 1px 1px darkcyan}
                fnColor = "cyan";
                // text-shadow:1px 1px 1px darkcyan ???
            break;

            case "DANGER":      //?? {color:#A00; font-weight:600}
                fnColor = "#A00";
                font.setBold(true);                                    
            break;

            case "FONT_SELECTED":   //?? {font-weight:600}
                font.setBold(true);                                    
            break;

            case "GRAY":      //?? {color:#AAA}
                fnColor = "#AAA";
            break;

            case "GREEN":          //?? {color:forestgreen}
                fnColor = "forestgreen";
            break;

            case "ITALIC":
                font.setItalic(true);
            break;

            // LARGE {font-size:larger}

            case "LOWERED":
                fnColor = "#BBB";
            break;

            case "MAGENTA":          //?? {color:magenta}
                fnColor = "magenta";
            break;

            case "MONOSPACE":   //?? {font-family:monospace}
                font.setFontName("Courier New");
            break;

            //NORMAL {color:black; font-weight:normal; background-color:transparent; text-shadow:none; position:static; text-decoration:none; border:0}

            case "NOTES":
                font.setItalic(true);
            break;

            case "OK":          //?? {color:#4A0; font-weight:600}
                fnColor = "#4A0";
                font.setBold(true);                                    
            break;

            case "ORANGE":          //?? {color:orange}
                fnColor = "orange";
            break;

            case "SHADOW":          //?? {text-shadow:1px 0 3px grey}
                fnColor = "gray";
                font.setBold(true);                                    
            break;

            case "TILT":
                font.setItalic(true);
            break;

            case "UNDERLINE":
                font.setUnderline(Font.U_SINGLE);
            break;
        }

        ((XSSFFont)font).setColor(new XSSFColor(ColorFactory.valueOf(fnColor),  new DefaultIndexedColorMap()));

        mapFonts.put(prop, font);
        
        return font;
    }

    // SVA ---------------------------------------------------------------------
      
    private XSSFRichTextString makeRichText(String makeup, String getup, String value)
    {
        String[] prop_list = getup.split("\\s+");
        String flags_pattern = "^\\/(.*)\\/([igmsuy]+)?$";
        String regexp = makeup;
        String flags = "";
        //DEBUG("prop_list.length = " + prop_list.length);
        
        if( regexp.matches(flags_pattern) )
        {
            Pattern pattern = Pattern.compile(flags_pattern);
            Matcher matcher = pattern.matcher(regexp);
            if(matcher.matches())
            {
                regexp = matcher.group(1);
                flags  = matcher.group(2);
            }
        }
        
        // decode regexp flags
        
        int jflags = 0;
        
        if( flags != null )
        {
            for( int i = 0; i < flags.length(); i++ )
            {
                switch( flags.substring(i, 1) )
                {
                    case "i":  jflags |= Pattern.CASE_INSENSITIVE;  break;
                    //case "g":  jflags |= ; break;
                    case "m":  jflags |= Pattern.MULTILINE;         break;
                    case "s":  jflags |= Pattern.DOTALL;            break;
                    //case "u":  jflags |= ; break;
                    //case "y":  jflags |= ; break;
                }
            }
        }
        // make RichText
        
        XSSFRichTextString rt = new XSSFRichTextString(value);
        Pattern pattern = Pattern.compile(regexp, jflags);
        Matcher matcher = pattern.matcher(value);

        //DEBUG("regexp = " + regexp);
        //DEBUG("value = " + value);

        while( matcher.find() )
        {
            //DEBUG("matcher.find");
            for( int i=0; i<=matcher.groupCount(); i++ )
            {
                //DEBUG("matcher.group(%d) = %s", i, matcher.group(i));
                if( is_not_null(matcher.group(i)) )
                {
                    //DEBUG("set font to " + prop_list[i]);
                    Font font = makeFontForRichText(prop_list[i]);
                    rt.applyFont(matcher.start(i), matcher.end(i), font);
                }
            }
        }        
        
        return rt;
    }    

    // SVA ---------------------------------------------------------------------

    private void printStyles() throws LuiException, SQLException
    {
        PrintWriter sf;

        try
        {
             sf = new PrintWriter(app.getTmpDir()+"styles.txt");
        }
        catch( FileNotFoundException ex )
        {
            DEBUG("ERROR: printStyles: FileNotFoundException");
            return;
        }
                

        sf.println("cell stypes:");
        sf.println("");
        
        for( String key : mapCellStyles.keySet() )
        {
            sf.println(key);
        }

        sf.println("");
        sf.println("fonts:");
        sf.println("");
        
        for( String key : mapFonts.keySet() )
        {
            sf.println(key);
        }
        
        sf.close();
    }

    // SVA ---------------------------------------------------------------------

    private void storeCurrentPosition(LuiAppForm frm, ArrayList<ColInfo> colList) throws LuiException, SQLException
    {
        storedRow.clear();

        //DEBUG("CurrentRow = " + frm.getVal("CurrentRow"));

        storedRow.put("<<<CurrentRow>>>", frm.getVal("CurrentRow"));
      
        for( int i=0; i<colList.size(); i++ )
        {
            String colCode = colList.get(i).alias;

            if( !frm.elements.containsKey(colCode) )
                continue;

            LuiAppElement el = frm.elements.get(colCode);

            storedRow.put(colCode, el.getValue());
        }
    }

    // SVA ---------------------------------------------------------------------

    private void restoreCurrentPosition(LuiAppForm frm, ArrayList<ColInfo> colList) throws LuiException, SQLException
    {
        if( colList != null && storedRow.size() > 0 )
        {
            frm.setVal("CurrentRow", storedRow.get("<<<CurrentRow>>>"));

            for( int i=0; i<colList.size(); i++ )
            {
                String colCode = colList.get(i).alias;

                if( !frm.elements.containsKey(colCode) )
                    continue;

                LuiAppElement el = frm.elements.get(colCode);

                el.setValue(storedRow.get(colCode), false, "", false);
            }
        }

        storedRow.clear();
    }

    // SVA ---------------------------------------------------------------------
    
    private void buildListSheet(String fc) throws LuiException, SQLException
    {
        boolean needTot = false;
        HashMap<String,String> tot = null;
        int realCol = 1;
        int totalRows = 0;
        int tot4Ind = 0;

        String maxRowsS = cs.Get_Element("Print");
        String needTotS = cs.Get_Element("TOTALS");
        int maxRows = Integer.valueOf(NVL(maxRowsS, "1000"));
       
        totalRows = maxRows;
        tot4Ind = maxRows;

        LuiAppForm frm = app.forms.get(fc);

        if(debug_mode) DEBUG_FORM_INFO(frm);                      
        
        LuiDbConn lc = null;

        frmInd = null;
        elInd = null;
        elIndOK = null;

        fc = app.getVal("FormID");
    
        if( app.forms.containsKey(fc) )
        {
           frmInd = app.forms.get(fc);

           if( frmInd.elements.containsKey("PROGRESS$INDICATOR") )
           {
               elInd=frmInd.elements.get("PROGRESS$INDICATOR");

               if( frmInd.elements.containsKey("OK") )
               {
                   elIndOK = frmInd.elements.get("OK");
               }
           }
           else
           {
               frmInd = null;
           }
        }


        Sheet sheet = workbook.createSheet(
            WorkbookUtil.createSafeSheetName(frm.getProperty("HEADER"))
        );

        if( sheet instanceof SXSSFSheet )
            ((SXSSFSheet)sheet).trackAllColumnsForAutoSizing();

        Row row0;
        Row row;
        Cell cell0;
        Cell cell;

        String queryText = frm.getVal("MainQuery");

        queryText = queryText.replaceAll("\\x01", "{").replaceAll("\\x02", "}");
        
        //DEBUG("queryText = " + queryText);                    
        
        
        if( is_not_null(queryText) )
        {
            String acc = frm.getProperty("Connection");
            if( is_not_null(acc) )
            {
                if (app.userDbConnections.containsKey(acc))
                {
                    LuiAppConn ac = app.userDbConnections.get(acc);
                    lc = null;
                    Integer cursor = null;

                    if( ac.getSessionMode() == 'C' )
                        lc = ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                    else
                        lc = ac.currentConnection;
                                       
                    // формируем жирную ячейку с заголовком списка

                    ArrayList<ColInfo> colList = null;
                    int r = 0;

                    try
                    {
                        cursor = lc.CursorOpen(queryText);
                        int cols = lc.CursorGetCols(cursor); // курсор
                        colList = lc.CursorGetColsInfo(cursor);


                        // Любушкин 2021.05.16
                        // SVA для чего это?
                        int qcc=0;
                        for(int iq=0;iq<frm.elements.size();iq++){
                            LuiAppElement el=frm.elSort(iq);
                            if (!el.getType().equals("ITEM"))continue;
                             if (el.colinfo==null) continue;
                             colList.get(qcc).alias=el.Code;
                             qcc++;
                        }


                        ArrayList<Boolean> colVisible = new ArrayList<Boolean>();
                        int FixedColumns = 0;

                        storeCurrentPosition(frm, colList);
                        
                        row0 = sheet.createRow(r);
                        row  = sheet.createRow(++r);

                        // формирование заголовков столбцов

                        CellStyle headerCellStyle = makeCellStyle(head_decor + " CENTER VALIGN(CENTER) WRAPTEXT BORDER FONTHEIGHT(12)"); // FONTHEIGHT(12)

                        for( int i=0; i<colList.size(); i++ )
                        {
                            String colCode = "";
                            colCode = colList.get(i).alias;
                            if( !frm.elements.containsKey(colCode) ) continue;
                            LuiAppElement el = frm.elements.get(colCode);

                            String colWidth = frm.getVisualPref(colCode, "ColWidth", "!");

                            if( colWidth.equals("!") )
                            {
                                colWidth = el.getProperty("WIDTH");
                            }

                            colVisible.add(i, !colWidth.equals("0") && el.getProperty("VISIBLE").equals("Y") && el.getProperty("Read").equals("Y"));


                            if( colVisible.get(i) )
                            {
                                cell0 = row0.createCell(realCol-1);
                                cell = row.createCell(realCol-1);                                           
                                String colNam = el.getName4show();

                                if( !el.getParent().equals("Form") )
                                {
                                    cell0.setCellValue(frm.elements.get(el.getParent()).getProperty("TITLE"));

                                    //SVA если группа типа FIXED то посчитать для Freeze
                                    if( frm.elements.get(el.getParent()).getProperty("GROUP_TYPE").equals("FIXED") )
                                        FixedColumns++;
                                }

                                cell.setCellValue(colNam);
                                cell.setCellStyle(headerCellStyle);
                                cell0.setCellStyle(headerCellStyle);

                                realCol++;
                                if( is_not_null(el.getProperty("TOTAL")) ) needTot = true;
                            }
                        }

                        sheet.createFreezePane(FixedColumns, 2);

                        // SVA Объединение заголовков

                        int col_count = realCol - 1;
                        int x = 0;

                        while( x < col_count )
                        {
                            if( row0.getCell(x).getStringCellValue().equals("")  )
                            {
                                row0.getCell(x).setCellValue( row.getCell(x).getStringCellValue() );
                                sheet.addMergedRegion(new CellRangeAddress(0, 1, x, x));
                            }
                            else
                            {
                                int last_idx = x;

                                String s0 = row0.getCell(x).getStringCellValue();

                                while( last_idx < col_count-1 && s0.equals( row0.getCell(last_idx+1).getStringCellValue() ))
                                {
                                    last_idx++;
                                }

                                if( last_idx > x )
                                {
                                    sheet.addMergedRegion(new CellRangeAddress(0, 0, x, last_idx));
                                }

                                x = last_idx;
                            }

                            x++;            
                        }


                        ArrayList<String> srow;

                        // расчёт итогов
                        if( needTotS.equals("Y") )
                        {
                            if( app.NLS.currentLang.code.equals("RUS") )
                            {
                                //if(!progress("Подсчёт количества...", "")) return "";
                                //SVA ??? проверить
                                if(!progress("Подсчёт количества...", "")) return;
                            }
                            else
                            {
                                //if(!progress("Calculating totals...", "")) return ""; 
                                //SVA ??? проверить
                                if(!progress("Calculating totals...", "")) return; 
                            }

                            queryText = "count(*) \"row$cnt\"";

                            for( int i=0; i<cols; i++ )
                            {
                                if( colVisible.get(i) )
                                {
                                    String colCode = "";
                                    colCode = colList.get(i).alias;
                                    if( !frm.elements.containsKey(colCode) ) continue;
                                    LuiAppElement el = frm.elements.get(colCode);
                                    String totalExpr = "";
                                    totalExpr = el.getProperty("TOTAL");

                                    if( is_not_null(totalExpr) )
                                    {
                                        String format = el.getProperty("FORMAT");
                                        if( is_not_null(format) )
                                        {
                                            totalExpr = "to_char(" + totalExpr + ",'" + format + "')";
                                        }

                                        queryText = queryText + ", " + totalExpr + " \"" + el.getCode() + "\" \n";
                                    }
                                }
                            }

                            if( is_not_null(queryText) )
                            {
                                queryText = "select /*Total Row*/\n"+queryText+" from (\n"+frm.makeMainQuery()+"\n)q$total ";

                                String qbewhere = frm.addWhere(lc, frm.getProperty("inner_qbe"),false);

                                if( is_not_null(qbewhere) )
                                {
                                    if( frm.getProperty("inner_qbe").equals("Y") )
                                    {
                                        if( queryText.indexOf("/*ADD_LUI$_INNER_QBE*/") > 0 )
                                                queryText = queryText.replace("/*ADD_LUI$_INNER_QBE*/"," and ("+qbewhere.substring(7)+")");
                                        else
                                        {
                                            if( queryText.indexOf("/*LUI$_INNER_QBE*/") > 0 )
                                                queryText = queryText.replace("/*LUI$_INNER_QBE*/", qbewhere);
                                        }
                                    } 
                                    else
                                        queryText = queryText + qbewhere + "\n";
                                }
                            }

                            tot = lc.getSelectFirstRowValMap(queryText);
                            totalRows = Integer.valueOf(tot.get("row$cnt"));
                            tot4Ind = totalRows;
                            if( tot4Ind > maxRows ) tot4Ind = maxRows;
                        }

                        
                        if( sheet instanceof XSSFSheet )
                        {
                            ((XSSFSheet) sheet).addIgnoredErrors(new CellRangeAddress(2,  2 + tot4Ind, 0, cols), IgnoredErrorType.NUMBER_STORED_AS_TEXT);
                        }

                        
                        //основной цикл
                        // расчитаем сколько раз показывать progress-bar

                        int delta=1;
                        if( tot4Ind < 200 ) { delta = 20; } //5раз
                        else if( tot4Ind < 1000 ) { delta = 10; } //10раз
                        else if( tot4Ind < 20000 ) { delta = 5; } //20раз
                        else if( tot4Ind < 50000 ) { delta = 2; } //50раз

                        int oldPct = 0;
                        int newPct = 0;

                        while( lc.CursorNextRow(cursor) )
                        {
                            r++;
                            
                            //SVA
                            /*
                            if( r==98 )
                            {
                                //SVA??? sheet.trackAllColumnsForAutoSizing();
                                if( sheet instanceof SXSSFSheet )
                                    ((SXSSFSheet)sheet).trackAllColumnsForAutoSizing();
                            }
                            */
                            
                            newPct = (r-1)*100/tot4Ind;

                            if( newPct-oldPct >= delta )
                            {
                                oldPct = newPct;
                                if( !progress(String.valueOf(r) + "/" + String.valueOf(tot4Ind), String.valueOf(newPct)) )
                                    //return "";
                                    //SVA ??? проверить
                                    return;
                            }

                            row = sheet.createRow(r);

                            if( r > maxRows )
                            {
                                cell = row.createCell(0);
                                cell.setCellStyle(makeCellStyle("STOP"));
                                if( app.NLS.currentLang.code.equals("RUS") )
                                    cell.setCellValue("...Превышено разрешённое количество выводимых строк!");
                                else
                                    cell.setCellValue("...Exceeded the allowed number of output lines!");        

                                sheet.addMergedRegion(new CellRangeAddress(r, r, 0, col_count-1));

                                break;
                            }

                            srow = lc.CursorGetRow(cursor); 

                            // укажем форме текущую строку
                            frm.setVal("FetchedRows", String.valueOf(r));
                            frm.setVal("CurrentRow",  String.valueOf(r));

                            //устанавливаем значения элементов
                            for( int i=0; i<cols; i++ )
                            {
                                String colCode = "";
                                colCode = colList.get(i).alias;

                                if( !frm.elements.containsKey(colCode) )
                                    continue;

                                LuiAppElement el = frm.elements.get(colCode);

                                String ldt = el.getProperty("DATA_TYPE");
                                String valBuffer = srow.get(i);

                                if( valBuffer == null )
                                {
                                    if( "EMPTY_STRING".equals(app.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                                    && is_not_null(app.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol) )
                                    {
                                        //если есть особый режим работы с NULL
                                        valBuffer = app.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
                                    }
                                }

                                if( "".equals(valBuffer) )
                                {
                                    if( "NULL".equals(app.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                                    && is_not_null(app.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol) )
                                    {
                                        //если есть особый режим работы с NULL
                                        valBuffer = app.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
                                    }
                                }      

                                // и запихиваем в значение столбца
                                el.setValue(valBuffer, false, "", false);

                            }

                            // значения столбцов установлены       

                            int icn = 0;
                            String dv = "";

                            for( int i=0; i<cols; i++ )
                            {
                                String colCode = "";
                                colCode = colList.get(i).alias;

                                if( !frm.elements.containsKey(colCode) )
                                    continue;

                                if( colVisible.get(i) )
                                {
                                    if( srow.get(i) == null )
                                        dv = "";
                                    else
                                    {
                                        if( srow.get(i).length() > 32767 )
                                            dv = srow.get(i).substring(0, 32000);
                                        else
                                            dv = srow.get(i);
                                    }

                                    cell = row.createCell(icn);
                                    icn++;

                                    LuiAppElement el = frm.elements.get(colCode);

                                    if( is_null(dv) )
                                    {
                                        cell.setCellValue("");
                                    }
                                    else
                                    {
                                        String dt = el.getProperty("DATA_TYPE");

                                        if( !el.getCode().equals("GG")||!frm.getCode().equals("LUI_L_DATA_GROUPING_SHOW") )
                                        {
                                            try
                                            {
                                                String dvn = dv.replaceAll("\\s+","");
                                                dvn = dvn.replaceAll(s160,"");
                                                /*double d =*/ Double.valueOf(dvn);
                                                dt = "NUMBER";
                                            }
                                            catch (Exception e){}                                
                                        }

                                        switch(dt)
                                        {
                                            case "DATETIME":
                                                cell.setCellValue(dv);
                                            break;

                                            case "NUMBER":
                                                dv = dv.replaceAll("\\s+","");
                                                dv = dv.replaceAll(s160,"");
                                                //cell.setCellType(CellType.NUMERIC);
                                                //SVA 2020.12.16                                
                                                //cell.setCellValue(Double.valueOf(dv));
                                                cell.setCellValue(Double.valueOf(dv.replace(',', '.')));
                                            break;

                                            default:
                                                cell.setCellValue(dv);
                                        }
                                    }

                                    cell.setCellStyle(makeCellStyle(getElementVisualProps(el)));

                                    if( p_coloring && cell instanceof XSSFCell )
                                    {
                                        String makeup = el.getProperty("MAKEUP");
                                        String getup  = el.getProperty("GETUP");
        
                                        if( is_not_null(makeup) && is_not_null(getup) )
                                            // 23.09.2021 Ранее в форме LUI_B_EXPLISTDATA отключено формирование списка цветных колонок
                                            // поэтому будем временно красить все, у которых задана раскраска
                                            //if( colored_cols.Get_Element(colCode).equals("Y") )
                                                ((XSSFCell)cell).setCellValue(makeRichText(makeup, getup, dv));                                        
                                    }
                                }
                            }
                        }

                        //SVA
                        //printStyles();


                        //if( !progress("", "100") ) return "";
                        //SVA ??? проверить
                        if( !progress("", "100") ) return;


                        //итоги                     

                        if( needTotS.equals("Y") )
                        {
                            r++;
                            row = sheet.createRow(r); 
                            cell = row.createCell(0);
                            cell.setCellStyle(makeCellStyle(head_decor + " LEFT"));

                            String a = (app.NLS.currentLang.code.equals("RUS") ? "Всего строк: ":"Total rows: ") + totalRows;
                            cell.setCellValue(a);

                            sheet.addMergedRegion(new CellRangeAddress(r, r, 0, col_count-1));

                            r++;
                            row = sheet.createRow(r);

                            int icn = 0; 

                            for( int i=0; i<cols; i++ )
                            {
                                if( colVisible.get(i) )
                                {
                                    String colCode = colList.get(i).alias;
                                    if( !frm.elements.containsKey(colCode) ) continue;
                                    LuiAppElement el = frm.elements.get(colCode);
                                    cell = row.createCell(icn);icn++;
                                    String totalExpr = "";

                                    totalExpr = el.getProperty("TOTAL");

                                    String dv = tot.get(colCode);

                                    if( is_not_null(totalExpr) )
                                    {
                                        cell.setCellStyle(makeCellStyle(head_decor + " " + el.getProperty("ALIGNMENT")));

                                        if( is_null(tot.get(colCode)) )
                                        {
                                            cell.setCellValue("");
                                        }
                                        else
                                        {
                                            String dt=el.getProperty("DATA_TYPE");

                                            try
                                            {
                                                String dvn=dv.replaceAll("\\s+","");
                                                dvn = dvn.replaceAll(s160,"");
                                                /*double d = */Double.valueOf(dvn);
                                                dt = "NUMBER";
                                            }
                                            catch (Exception e){}

                                            switch(dt)
                                            {
                                                case "DATETIME":
                                                    cell.setCellValue(dv);
                                                break;
                                                case "NUMBER":
                                                    dv = dv.replaceAll("\\s+","");
                                                    dv = dv.replaceAll(s160,"");
                                                    cell.setCellValue(Double.valueOf(dv));    
                                                break;
                                                default:
                                                    cell.setCellValue(tot.get(colCode));       
                                            }
                                        }
                                    }
                                    else
                                    {
                                        cell.setCellStyle(makeCellStyle(head_decor));
                                        cell.setCellValue("");
                                    }
                                }
                            }
                        }

                        // выравнивание
                        for( int i=0; i<col_count; i++ )
                        {
                            sheet.autoSizeColumn(i, true);
                        }
                        
                        restoreCurrentPosition(frm, colList);
                        
                    }

                    // временно
                    //catch( Exception ex)
                    //{
                    //    DEBUG("Exception: " + ex.getLocalizedMessage());
                    //}

                    finally
                    {
                        if( lc != null )
                        {
                            if( lc.CursorisOpen(cursor) ) lc.CursorClose(cursor);
                            if( ac.getSessionMode() == 'C' )
                            {
                                ac.connPool.SetFree(lc);
                            }
                        }

                        restoreCurrentPosition(frm, colList);
                    }
                }   
            }  
        }
    }
   
    // SVA ---------------------------------------------------------------------

    public class GridInfo
    {
        public class CellAddr
        {
            int x;
            int y;
        }

        Sheet testSheet;
        Cell testCell;
        Cell rtsCell;
        CellStyle defaultCellStyle;

        LuiAppForm frm;
        Sheet sheet;

        String gridTitle;
        
        int zeroCol = 0;
        int zeroRow = 0;
    
        int colsHeadWidth;
        int colsHeadDepth;

        int rowsHeadWidth;
        int rowsHeadDepth;

        int colsHeadDepthFull;
        int rowsHeadDepthFull;
               
        int colInfo = 0;
        int rowInfo = 0;

        String colInfoType = "N";
        String rowInfoType = "N";
        
        CellAddr rowsHeadStart = new CellAddr();
        CellAddr colsHeadStart = new CellAddr();
        
        CellAddr rowsInfoStart = new CellAddr();
        CellAddr colsInfoStart = new CellAddr();
        
        ArrayList<XSSFRichTextString> angular = new ArrayList<XSSFRichTextString>();
        
        public GridInfo(LuiAppForm frm, Sheet sheet) throws LuiException, SQLException 
        {
            testSheet = workbook.createSheet("<test>");
            testCell = testSheet.createRow(0).createCell(0);
            rtsCell = testCell.getRow().createCell(1);
            defaultCellStyle = testCell.getCellStyle();
            this.frm = frm;
            this.sheet = sheet;

            // -----------------------------------------------------------------

            valClass _colsHeadDepth = new valClass(0);
            valClass _rowsHeadDepth = new valClass(0);

            int hx;
            hx = findHeadElems(this, 0, 0,   _colsHeadDepth, "COL", "COLS");
            hx = findHeadElems(this, 0, hx+1, _colsHeadDepth, "COL", "Form");
            //hx++;

            DEBUG("hx = " + hx + " depth = " + _colsHeadDepth.value);

            /*gi.*/colsHeadWidth = hx;
            /*gi.*/colsHeadDepth = _colsHeadDepth.value;
            ///*gi.*/colsHeadDepthFull = /*gi.*/colsHeadDepth + /*gi.*/colInfo;

            //progress(stage, stages);

            int hy;
            hy = findHeadElems(this, 0, 0, _rowsHeadDepth, "ROW", "ROWS");
            hy = findHeadElems(this, 0, hy+1, _rowsHeadDepth, "ROW", "Form");
            //hy++;

            DEBUG("hy = " + hy + " depth = " + _rowsHeadDepth.value);

            /*gi.*/rowsHeadWidth = hy;
            /*gi.*/rowsHeadDepth = _rowsHeadDepth.value;           
            ///*gi.*/rowsHeadDepthFull = /*gi.*/rowsHeadDepth + /*gi.*/rowInfo;

            // -----------------------------------------------------------------
            
            String gridTitle = frm.getProperty("GRID_TITLE");

            if( !gridTitle.equals("") )
            {
                zeroRow = 1;
                
                // GRID_TITLE_BORDER
            }

            
            String s = frm.getProperty("COL_INFO");
            if( !s.equals("N") )
            {
               colInfo = 1;
               colInfoType = s.substring(1, 2);
               if(colInfoType.equals("O"))
                   colsHeadDepth = 0;
            }
            
            s = frm.getProperty("ROW_INFO");
            if( !s.equals("N") )
            {
               rowInfo = 1;
               rowInfoType = s.substring(1, 2);
               if(rowInfoType.equals("O"))
                   rowsHeadDepth = 0;
            }
            
            colsHeadDepthFull = colsHeadDepth + colInfo;
            rowsHeadDepthFull = rowsHeadDepth + rowInfo;
            
            switch (colInfoType)
            {
                case "T":
                    colsHeadStart.x = zeroCol + rowInfo + rowsHeadDepth;
                    colsHeadStart.y = zeroRow + colInfo;
                    colsInfoStart.x = zeroCol + rowInfo + rowsHeadDepth;
                    colsInfoStart.y = zeroRow;
                    break;
                case "B":
                    colsHeadStart.x = zeroCol + rowInfo + rowsHeadDepth;
                    colsHeadStart.y = zeroRow;
                    colsInfoStart.x = zeroCol + rowInfo + rowsHeadDepth;
                    colsInfoStart.y = zeroRow + colsHeadDepth;
                    break;
                case "O":
                    colsInfoStart.x = zeroCol + rowInfo + rowsHeadDepth;
                    colsInfoStart.y = zeroRow;
                    break;                    
                default:
                    colsHeadStart.x = zeroCol + rowInfo + rowsHeadDepth;
                    colsHeadStart.y = zeroRow;                    
            }

            switch (rowInfoType)
            {
                case "L":
                    rowsHeadStart.x = zeroCol + rowInfo;
                    rowsHeadStart.y = zeroRow + colInfo + colsHeadDepth;
                    rowsInfoStart.x = zeroCol;
                    rowsInfoStart.y = zeroRow + colInfo + colsHeadDepth;
                    break;                    
                case "R":
                    rowsHeadStart.x = zeroCol;
                    rowsHeadStart.y = zeroRow + colInfo + colsHeadDepth;
                    rowsInfoStart.x = zeroCol + rowsHeadDepth;
                    rowsInfoStart.y = zeroRow + colInfo + colsHeadDepth;
                    break;                    
                case "O":
                    rowsInfoStart.x = zeroCol;
                    rowsInfoStart.y = zeroRow + colInfo + colsHeadDepth;
                    break;                    
                default:
                    rowsHeadStart.x = zeroCol;
                    rowsHeadStart.y = zeroRow + colInfo + colsHeadDepth;                    
            }           
            
        }
       
        public void clear()
        {           
            for( int i = 0; i < workbook.getNumberOfSheets(); i++ )
            {
                Sheet sheet = workbook.getSheetAt(i);
                if( sheet.getSheetName().equals("<test>"))
                {
                    workbook.removeSheetAt(i);
                }
            }       
        }

        
        public int calcRichTextWidth(XSSFRichTextString rts)
        {
            CellStyle style = defaultCellStyle;

            //DEBUG("style = " + style);           
            //DEBUG("RTS formattingRuns = " + rts.numFormattingRuns());

            if( rts.hasFormatting() )
            {
                //DEBUG("RTS font = " + rts.getFontAtIndex(0));
                //DEBUG("RTS fontName = " + rts.getFontAtIndex(0).getFontName());
                //DEBUG("RTS fontHeight = " + rts.getFontAtIndex(0).getFontHeight());

                style = workbook.createCellStyle();
                //style.setFont(rts.getFontAtIndex(0));
                // ничего не понимаю! так должно работать, но не работает...
                
                Font rtsFont = rts.getFontAtIndex(0);
                Font font = workbook.createFont();
                font.setFontHeight(rtsFont.getFontHeight());
                font.setFontName(rtsFont.getFontName());
                font.setBold(rtsFont.getBold());
                font.setItalic(rtsFont.getItalic());                
                style.setFont(font);
            }
            
            testCell.setCellStyle(style);

            //DEBUG("style = " + style);
            //DEBUG("testCell font = " + ((XSSFCellStyle)testCell.getCellStyle()).getFont());
            //DEBUG("testCell fontHeight = " + ((XSSFCellStyle)testCell.getCellStyle()).getFont().getFontHeight());
            
            //testCell.setCellStyle(defaultCellStyle);
            testCell.setCellValue(rts);
            testSheet.autoSizeColumn(0, true);
            return testSheet.getColumnWidth(0) - 449;
        }

        /*
        public int calcRichTextMinWidth(XSSFSimpleShape shape)
        {
            for(int i = 0; i < shape.getTextParagraphs().size(); i++)
            {
                XSSFTextParagraph p = shape.getTextParagraphs().get(i);
                //p.get
                //rtsCell.setCellValue(rts);
            }
            
            return 0;
        }
        */

        public int calcRichTextMinWidth()
        {
            int width = 0;
            
            for(int i = 0; i < angular.size(); i++)
            {
                CellStyle style = defaultCellStyle;

                XSSFRichTextString rts = angular.get(i);

                if( rts.hasFormatting() )
                {
                    style = workbook.createCellStyle();
                    //style.setFont(rts.getFontAtIndex(0));
                    //DEBUG("rts  font h = " + rts.getFontAtIndex(0).getFontHeight());
                    //DEBUG("sty font h = " + ((XSSFCellStyle)style).getFont().getFontHeight());
                    Font rtsFont = rts.getFontAtIndex(0);
                    Font font = workbook.createFont();
                    font.setFontHeight(rtsFont.getFontHeight());
                    font.setFontName(rtsFont.getFontName());
                    font.setBold(rtsFont.getBold());
                    font.setItalic(rtsFont.getItalic());                
                    style.setFont(font);
                }

                rtsCell.setCellStyle(style);
                
                rtsCell.setCellValue(rts);

                int w = calcColWidth(rtsCell, rts.getString());
                //DEBUG("calcRichTextMinWidth: (" + i + ") w = " + w);
                if(w > width) width = w;
            }
            
            return width;
        }
        

        // ----------------------------------------------
        public int calcWordWidh(String str)
        {
            testCell.setCellValue(str);
            testSheet.autoSizeColumn(0, true);
            return testSheet.getColumnWidth(0) - 449;
        }
              

        public int calcColWidth(String str)
        {
            Sheet sheet = testCell.getRow().getSheet();
            //String[] list = str.split("\\s+");
            // TEST
            //String[] list = str.split("[\\s\\-]+");
            String[] list = str.replaceAll("-", "- ").split("\\s+");
            int width = 0;

            for(int i = 0; i < list.length; i++)
            {                                   

                testCell.setCellValue(list[i]);
                sheet.autoSizeColumn(0, true);
                int w = sheet.getColumnWidth(0);
                if(w > width) width = w;
                //DEBUG("calcColWidth: " + list[i] + " " + w);
            }

            return width;
        }   

        public int calcColWidth(Cell cell, String str)
        {
            testCell.setCellStyle(cell.getCellStyle());
            
            return calcColWidth(str);
        }

        // ------------------------------
        
        public int calcCellLines(Cell cell, int width, String str)
        {
            String line = "";
            int lines = 0;

            //DEBUG("$$$ calcCellLines -> width = %d", width );


            // if str empty then return 0;

            //        CellStyle style = cell.getCellStyle();
            //        CellStyle testStyle = workbook.createCellStyle();
            //        testStyle.setFont( workbook.getFontAt(style.getFontIndexAsInt()) );
            //        gi.testCell.setCellStyle(testStyle);
            /*gi.*/testCell.setCellStyle(cell.getCellStyle());

            //DEBUG("$$$ calcCellLines -> cell0 font height = " + /*gi.*/testCell.getSheet().getWorkbook().getFontAt( cell.getCellStyle().getFontIndexAsInt() ).getFontHeight() );
            //DEBUG("$$$ calcCellLines -> cell  font height = " + cell.getSheet().getWorkbook().getFontAt( cell.getCellStyle().getFontIndexAsInt() ).getFontHeight() );


            String[] list = str.split("\\s+");

            for(int i = 0; i < list.length; i++)
            {                                   
                int w = /*gi.*/calcWordWidh(line + (line.length()==0?"":" ") + list[i]);

                if(w <= width)
                {
                    line += (line.length()==0?"":" ") + list[i];
                }
                else
                {
                    lines++;
                    line = list[i];
                }
                //DEBUG("$$$ calcCellLines -> w=%d, line = '%s'", w, line );
            }

            if( line.length() > 0 ) lines++;

            //DEBUG("$$$ calcCellLines -> lines=%d", lines );

            return lines;
        }
    }
    
    // SVA ---------------------------------------------------------------------

    public CellRangeAddress getCellRange(Sheet sheet, int x, int y)
    {
        for(int i = 0; i < sheet.getNumMergedRegions(); i++)
        {
            CellRangeAddress cr = sheet.getMergedRegion(i);

            if( cr.getFirstRow() == y && cr.getFirstColumn() == x )
            {
                return cr;
            }
        }

        CellRangeAddress cellRange = new CellRangeAddress(y, y, x, x);
        
        return cellRange;
    }
            
    // SVA ---------------------------------------------------------------------   
    
    // SVA 2021.03.04 Как это работает? Не учитывается стиль ячейки, для которой считается.
    // Надо переделать что бы учитывался стиль?
    // Перенести в класс gridInfo?  //28.03.2021 воткнул. проверить, а здесь удалить.
    
    public /*int*/ void calcColWidth(GridInfo gi, LuiAppElement el) throws LuiException, SQLException
    {
        String str = el.getProperty("CAPTION");

        ArrayList<Integer> words = new ArrayList<>();
        Sheet sheet = gi.testCell.getRow().getSheet();
        String[] list = str.split("\\s+");
        int width = 0;
        
        for(int i = 0; i < list.length; i++)
        {                                   
            gi.testCell.setCellValue(list[i]);
            sheet.autoSizeColumn(0, true);
            int w = sheet.getColumnWidth(0);
            if(w > width) width = w;
            words.add(w);
        }

        el.setProperty("_WIDTH", Integer.toString(width));        
    }   

    // SVA ---------------------------------------------------------------------

    // SVA 2021.03.04 А вот здесь стиль учитывается!
    // Перенести в класс gridInfo?
    
    public int calcCellLines(GridInfo gi, Cell cell, int width, String str)
    {
        String line = "";
        int lines = 0;

        //DEBUG("$$$ calcCellLines -> width = %d", width );


        // if str empty then return 0;
        
        //CellStyle style = cell.getCellStyle();
        //CellStyle testStyle = workbook.createCellStyle();
        //testStyle.setFont( workbook.getFontAt(style.getFontIndexAsInt()) );
        //gi.testCell.setCellStyle(testStyle);

        gi.testCell.setCellStyle(cell.getCellStyle());

        //DEBUG("$$$ calcCellLines -> cell0 font height = " + gi.testCell.getSheet().getWorkbook().getFontAt( cell.getCellStyle().getFontIndexAsInt() ).getFontHeight() );
        //DEBUG("$$$ calcCellLines -> cell  font height = " + cell.getSheet().getWorkbook().getFontAt( cell.getCellStyle().getFontIndexAsInt() ).getFontHeight() );

        
        String[] list = str.split("\\s+");
        
        for(int i = 0; i < list.length; i++)
        {                                   
            //gi.testCell.setCellValue(line + " " + list[i]);
            //gi.testSheet.autoSizeColumn(0, true);
            //int w = sheet.getColumnWidth(0);
            //int w = gi.calcWordWidh(line + " " + list[i]);
            int w = gi.calcWordWidh(line + (line.length()==0?"":" ") + list[i]);
            
            if(w <= width)
            {
                line += (line.length()==0?"":" ") + list[i];
            }
            else
            {
                lines++;
                line = list[i];
            }
            //DEBUG("$$$ calcCellLines -> w=%d, line = '%s'", w, line );
        }

        if( line.length() > 0 ) lines++;

        //DEBUG("$$$ calcCellLines -> lines=%d", lines );

        return lines;
    }
    
    // SVA ---------------------------------------------------------------------

    private int resizeHeadElems(GridInfo gi, String elType, String parentEelement) throws LuiException, SQLException
    {
        int levelWidth = 0;
        
        for( String ek : gi.frm.elements.keySet() )
	{
            LuiAppElement el = gi.frm.elements.get(ek);
            
            if( el.getType().equals(elType) && el.getProperty("Parent").equals(parentEelement) )
            {
                int subElWidth = resizeHeadElems(gi, elType, el.Code);

                int calculatedWidth = Integer.parseInt(el.getProperty("_WIDTH"));

                int elementWidth = subElWidth;
                
                if( calculatedWidth > subElWidth)
                {
                    int px = gi.zeroCol + gi.rowsHeadDepth + Integer.parseInt(el.getProperty("_IDX"));
                    int dx = Integer.parseInt(el.getProperty("_SPAN"));
                    int py = gi.zeroRow + Integer.parseInt(el.getProperty("_LEV"));
                    int dy = el.getProperty("_LAST").equals("Y") ? gi.colsHeadDepth - py : 1;

                    int deltaWidth = 0;
                    if (dx!=0)deltaWidth=(calculatedWidth - subElWidth) / dx;
                    
                    for( int x = px; x < px+dx; x++ )
                    {
                        gi.sheet.setColumnWidth(py, gi.sheet.getColumnWidth(py) + deltaWidth);
                        elementWidth += deltaWidth;
                    }
                    
                    Cell cell = gi.sheet.getRow(py).getCell(px);

                    String cellValue = el.getProperty("CAPTION");

                    int lines = calcCellLines(gi, cell, calculatedWidth, cellValue);

                    int fontHeight = workbook.getFontAt( cell.getCellStyle().getFontIndexAsInt() ).getFontHeight();
                    int cellHeight = (fontHeight + fontHeight/2) * lines;
                    int cellHeightSum = 0;

                    for( int y = 0; y < dy; y++ )
                        cellHeightSum += gi.sheet.getRow(y).getHeight();

                    if( cellHeight > cellHeightSum )
                    {
                        int deltaHeight = (cellHeight - cellHeightSum) / dy;

                        for( int y = 0; y < dy; y++ )
                        {
                            //gi.sheet.getRow(y).setHeight( gi.sheet.getRow(y).getHeight() + deltaHeight );
                            
                            short d = (short) (gi.sheet.getRow(y).getHeight() + deltaHeight);
                            gi.sheet.getRow(y).setHeight( d );
                        }
                    }
                }
                
                levelWidth += elementWidth;
            }
        }
    
        return levelWidth;
    }
    
    // SVA ---------------------------------------------------------------------
    
    private int findHeadElems(GridInfo gi, int lev, int idx, valClass depth, String elType, String parentEelement) throws LuiException, SQLException
    {
        //DEBUG("$$$ findHeadElems -> " + parentEelement);

        ArrayList<LuiAppElement> items = new ArrayList<>();
        
        for( String ek : gi.frm.elements.keySet() )
	{
            LuiAppElement el = gi.frm.elements.get(ek);

            if( el.getType().equals(elType) && el.getProperty("Parent").equals(parentEelement) )
            {
                //DEBUG("==> %s [%s] seq=%s %s", elType, ek, el.getProperty("Seq"), el.getProperty("CAPTION"));
                //DEBUG("Seq =  " + el.getProperty("Seq"));
                //items.add(Integer.parseInt(el.getProperty("Seq")), ek);
                double seq = Double.parseDouble(el.getProperty("Seq"));

                Boolean flag = true;

                for( int i=0; i<items.size(); i++ )
                {
                    if( Double.parseDouble(items.get(i).getProperty("Seq")) > seq )
                    {
                        items.add(i, el);
                        flag = false;
                        break;
                    }
                }
                
                if(flag)
                {
                    items.add(el);
                }
            }
        }

        //DEBUG("items =  " + items.size() + "  parentEelement = " + parentEelement);

        if( items.size() == 0 )
        {
            if( !parentEelement.equals("Form") ) // временно! Разобраться.
            gi.frm.elements.get(parentEelement).setProperty("_LAST", "Y");

            return idx;
        }

        if( lev+1 > depth.value)
            depth.value = lev+1;
                       
       
        for( int i = 0; i < items.size(); i++ )
        {
            if(i>0) idx++;
                
            LuiAppElement el = items.get(i);

            el.setProperty("_IDX", Integer.toString(idx));
            el.setProperty("_LEV", Integer.toString(lev));
          
            //DEBUG("%s [%s] <%d> seq=%s %s", elType, el.Code, idx, el.getProperty("Seq"), el.getProperty("CAPTION"));

            int ix = findHeadElems(gi, lev+1, idx, depth, elType, el.Code);

            int span = ix - idx + 1;

            el.setProperty("_SPAN", Integer.toString(span));
            
            //el.setProperty("_WIDTH", Integer.toString(calcColWidth(el.getProperty("CAPTION"))));
            calcColWidth(gi, el);

            DEBUG("%s [%s] <%d> span=%d lev=%d seq=%s %s", elType, el.Code, idx, span, lev, el.getProperty("Seq"), el.getProperty("CAPTION"));
            
            idx = ix;
        }

        return idx;
    }

    // SVA ---------------------------------------------------------------------
    
    public String getCellDecor(LuiAppElement el) throws LuiException, SQLException
    {
        //DEBUG("getCellDecor -> " + el.getProperty("EType") + " " + el.getCode());
        String decor = el.getProperty("VISUALIZATION");
        //DEBUG("cell decor = " + decor);
        
        if( decor.equals("") )
        {           
            decor = el.getForm().elements.get(el.getProperty("COL")).getProperty("VISUALIZATION");
            
            if( decor.equals("") )
            {
                decor = el.getForm().elements.get(el.getProperty("ROW")).getProperty("VISUALIZATION");
            }
        }
        
        return decor;
    }
    
    // SVA ---------------------------------------------------------------------

    public Boolean isShaded(LuiAppElement el) throws LuiException, SQLException
    {
        /*
        Элемент скрывается элементом, заголовки которого совпадают
        или являются предками заголовков нашего элемента.
        
        Очевидно, не должно быть элементов, с совпадающими обоими заголовками - 
        они скроют друг друга. То же самое будет с перекрывающимися элементами 
        (или даже ошибка выполнения).
        */
        
        String curCode = el.getCode();

        //DEBUG("isShaded: (" + curCode + ")");


        //HashMap<String, LuiAppElement> elements = el.getForm().elements;
        
        ArrayList<String> colHList = new ArrayList<>();
        ArrayList<String> rowHList = new ArrayList<>();

        String colHCode = el.getProperty("COL");
        String rowHCode = el.getProperty("ROW");

        //DEBUG("isShaded: colHCode = " + colHCode);
        //DEBUG("isShaded: rowHCode = " + rowHCode);
        //DEBUGP("isShaded: COLS:");
        
        while(true)
        {
            //LuiAppElement ell = elements.get(colHCode);
            LuiAppElement ell = el.getForm().elements.get(colHCode);
            if( !ell.getType().equals("COL") ) break;
            //DEBUGP(" " + colHCode);
            colHList.add(colHCode);
            colHCode = ell.getProperty("Parent");
            if( colHCode.equals("Form") ) break;
        }

        //DEBUG(";");
        
        //DEBUGP("isShaded: ROWS:");

        while(true)
        {
            //LuiAppElement ell = elements.get(rowHCode);
            LuiAppElement ell = el.getForm().elements.get(rowHCode);
            if( !ell.getType().equals("ROW") ) break;
            //DEBUGP(" " + rowHCode);
            rowHList.add(rowHCode);
            rowHCode = ell.getProperty("Parent");
            if( rowHCode.equals("Form") ) break;
        }

        //DEBUG(";");
        
        for( String ek : el.getForm().elements.keySet() )
	{
            if( ek != curCode )
            {
                LuiAppElement el2 = el.getForm().elements.get(ek);
                
                if( el2.getType().equals("CELL") )
                {
                    String el2ColHCode = el2.getProperty("COL");
                    
                    int colLev = 0;
                
                    for( int i = 0; i < colHList.size(); i++ )
                    {
                        if( el2ColHCode.equals( colHList.get(i) ) )
                        {
                            colLev = i + 1;
                            //DEBUG("isShaded: --> (%s) %s == %s colLev = %d", el2.getProperty("Code"), el2ColHCode, colHList.get(i), colLev);
                            break;
                        }
                    }
                
                    String el2RowHCode = el2.getProperty("ROW");

                    if( colLev == 0 )
                            continue;
                    
                    int rowLev = 0;
                
                    for( int i = 0; i < rowHList.size(); i++ )
                    {
                        if( el2RowHCode.equals( rowHList.get(i) ) )
                        {
                            rowLev = i + 1;
                            //DEBUG("isShaded: --> (%s) %s == %s  rowLev = %d", el2.getProperty("Code"), el2RowHCode, rowHList.get(i), rowLev);
                            break;
                        }
                    }

                    if( /*colLev > 0 &&*/ rowLev > 0 /* && ( colLev > 1 || rowLev > 1 ) */ )
                    {
                        return true;
                    }
                }                
            }
        }
        
        return false;        
    }
     
    // SVA ---------------------------------------------------------------------

    private int rightSubColumns(Cell rcell, CellRangeAddress cellRange)
    {
        int w = 0;
        
        //if( cellRange.getFirstColumn() < cellRange.getLastColumn() )
        for( int i = cellRange.getFirstColumn(); i < cellRange.getLastColumn(); i++ )
        {
            w += rcell.getSheet().getColumnWidth(i+1);
        }
        
        return  w;
    }

    // SVA ---------------------------------------------------------------------
    
    private void makeAngular(Sheet sheet, String angular, int dx, int dy, GridInfo gi)
    {
        int x = 0, y = 0;
        
        DEBUG("Angular = " + angular);

        if( angular.equals("") ) return;
        
        // НОВЫЙ ВАРИАНТ

        java.awt.Color fc = ColorFactory.valueOf(head_color);

        WikiToExcell wk = new WikiToExcell(sheet, angular, true, fc);

        wk.makeDrawing(gi.zeroCol, gi.zeroRow, dx, dy);

        if(true) return;

        sheet.getRow(gi.zeroRow).getCell(gi.zeroCol).setCellValue(angular);       
        
        if(true) return;

        // СТАРЫЙ ВАРИАНТ
        
        XSSFDrawing drawing = ((XSSFSheet)sheet).createDrawingPatriarch();
        
        //XSSFClientAnchor anchor = drawing.createAnchor(0, 0, 0, 0, 1, 10, 5, 16);
        //XSSFClientAnchor anchor = drawing.createAnchor(0, 0, 0, 0, x, x+xd-1, y, y+yd-1);
        XSSFClientAnchor anchor = drawing.createAnchor(0, 0, 0, 0, x, y, x+dx, y+dy);
        XSSFSimpleShape shape = drawing.createSimpleShape(anchor);
        
        shape.setShapeType(ShapeTypes.RECT);
        //java.awt.Color fc = ColorFactory.valueOf(head_color);
        
        //shape.setFillColor(211, 211, 211);
        shape.setFillColor(fc.getRed(), fc.getGreen(), fc.getBlue());

        //shape.getTextParagraphs().clear();
        shape.clearText();
        
        shape.setVerticalAlignment(VerticalAlignment.CENTER);
        
        
        //XSSFRichTextString rt = new XSSFRichTextString(angular);
        //XSSFTextParagraph tp;
        //tp = shape.addNewTextParagraph(rt);
        //if(true) return;
        
        
        Source source = new Source(angular);
        
        for(Element el : source.getAllElements("p"))
        {
            DEBUG("el == " + el.toString());

            HTMLtoExcel htmlToExcel = new HTMLtoExcel();           

            XSSFRichTextString rts = (XSSFRichTextString)htmlToExcel.fromHtmlToCellValue(el.toString(), workbook);
            //XSSFRichTextString rts = new XSSFRichTextString(el.toString());;

            gi.angular.add(rts);

            DEBUG(" RTS = <" + rts.getString() + ">");
            DEBUG(" RTS width = " + gi.calcRichTextWidth(rts));

            XSSFTextParagraph tp = shape.addNewTextParagraph(rts);
            
            for( Attribute a : el.getStartTag().getAttributes() )
            {
                DEBUG("  a => " + a.getName() + " = " + a.getValue());

                if( a.getName().equalsIgnoreCase("style") )
                {
                    for( String style : a.getValue().toLowerCase().split(";") )
                    {
                        String[] avp = style.split(":");

                        if( avp != null && avp.length > 1 )
                        {
                            if( avp[0].trim().toLowerCase().equals("text-align") )
                            {
                                switch(avp[1].trim().toLowerCase())
                                {
                                    case "left":
                                        tp.setTextAlign(TextAlign.LEFT);
                                    break;
                                    case "right":
                                        tp.setTextAlign(TextAlign.RIGHT);
                                    break;
                                    case "center":
                                        tp.setTextAlign(TextAlign.CENTER);
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            
            //DEBUG(" tp.getTextRuns().size() = " + tp.getTextRuns().size());
            //DEBUG(" tp run0 font size  = " + tp.getTextRuns().get(0).getFontSize());
            //DEBUG(" tp run0 font name  = " + tp.getTextRuns().get(0).getFontFamily());
            //DEBUG(" tp run0 text = " + tp.getTextRuns().get(0).getText());
	}
        
        DEBUG(" Paragraphs().size = " + shape.getTextParagraphs().size());
        
        //shape.setTextAutofit(TextAutofit.SHAPE);

        DEBUG(" calcRichTextMinWidth = " + gi.calcRichTextMinWidth());

        /*
        tp = shape.addNewTextParagraph("Текст справа");
        tp.setTextAlign(TextAlign.RIGHT);
        */
    }
   
    // SVA ---------------------------------------------------------------------

    private String getHeaderFormat(String p, LuiAppElement el) throws LuiException, SQLException
    {
        String format = "";
        //DEBUG("  -- getHeaderFormat: el = " + el);
        //DEBUG("  -- getHeaderFormat: type = " + el.getType());

        if( el.getType() == "COL" || el.getType() == "ROW" )
        {
            format = getFormatVal(p, el.getProperty("ITEM_SEMANTICS"));

            if( format.equals("") )
                format = getHeaderFormat(p, el.getForm().elements.get(el.getProperty("Parent")));
        }
        
        return format;
    }

            
    private String getFormatVal(String p, String v)
    {       
        String ps = "\"" + p + "\":\"";
                
        int v1 = v.indexOf(ps);

        if( v1 > -1 )
        {
            //p = p.substring(p1+10);
            //p = p.substring(0, p.indexOf("\""));
            v = v.substring(v1 + ps.length(), v.indexOf('"', v1 + ps.length()));
            
            if( !v.equals("") )
                return v;

                //String a = ".0000000000";
                //int b = Integer.parseInt( v.substring(v.length()-1) );
                //return (v.length()==3?"#,##":"") + "0" + (b>0?a.substring(0, b+1):"");
        }

        //SVA: переделать на JSON? Сравнить производительность.        
        return "";
    }
    
    private String getCellFormatVal(String p, LuiAppElement el) throws LuiException, SQLException
    {
        DEBUG("  -- getCellFormat: type = " + el.getType() + " p = " + p);

        String format = getFormatVal(p, el.getProperty("CELL_SEMANTICS"));
    
        if( format.equals(""))
        {
            format = getHeaderFormat(p, el.getForm().elements.get(el.getProperty("COL")));
            
            if( format.equals(""))
            {
                format = getHeaderFormat(p, el.getForm().elements.get(el.getProperty("ROW")));                
            }
        }

        DEBUG("  -- getCellFormat: format = " + format);

        return format;
    }

    private String getCellFormat_MY(LuiAppElement el) throws LuiException, SQLException
    {
        //decimal - десятичный разделитель
        //separator - разделитель разрядов
        //fractional - длина дробной части
        
        String separator = getCellFormatVal("separator", el);

        if( separator.equals("") ) separator = default_separator;
        
        String fractional = getCellFormatVal("fractional", el);
        
        if( fractional.equals("") ) fractional = default_fractional;
        
        String a = ".0000000000";

        int b = 0;

        if( !fractional.equals("") ) b = Integer.parseInt( fractional );

        return (separator.length()>0?"#,##":"") + "0" + (b>0?a.substring(0, b+1):"");
    }

    // попробуем сделать Сашин вариант
    
    private String getCellFormat(LuiAppElement el) throws LuiException, SQLException
    {
        LuiPlugin plugin = app.parser.plugins.get("MatrixSemantics");

        String s = el.getForm().getFormID() + "." + el.Code + ".";
        
        //compute(frm.getFormID()+"."+((LuiAppElement)p.ownerObj).Code+"."+sem, pProp);
                
        //decimal - десятичный разделитель
        //separator - разделитель разрядов
        //fractional - длина дробной части
        
        String separator = plugin.compute(s + "separator", null); //getCellFormatVal("separator", el);

        if( separator.equals("") ) separator = default_separator;
        
        String fractional = plugin.compute(s + "fractional", null); //getCellFormatVal("fractional", el);
        
        if( fractional.equals("") ) fractional = default_fractional;
        
        String a = ".0000000000";

        int b = 0;

        if( !fractional.equals("") ) b = Integer.parseInt( fractional );

        return (separator.length()>0?"#,##":"") + "0" + (b>0?a.substring(0, b+1):"");
    }
   
    // SVA ---------------------------------------------------------------------   
    
    private String getHeadProperty(LuiAppElement el, String elType, String prop) throws LuiException, SQLException
    {
        // здесь нужна рекурсия?
        //DEBUG("getHeadProperty: el %s prop %s", el.getType(), prop);
        //return el.getProperty(prop);

        String val = "";
                
        if( el.getType() ==  elType )
        {
            val = el.getProperty(prop);
            
            if(val.equals(""))
                val = getHeadProperty(el.getForm().elements.get(el.getProperty("Parent")), elType, prop);
        }

        return val;
    }
    
    private String getCellProperty(LuiAppElement el, String prop, String headProp) throws LuiException, SQLException
    {
        //DEBUG("getCellProperty:  el %s prop %s headProp %s", el.getType(), prop, headProp);
        String val = el.getProperty(prop);
        
        if(val.equals(""))
            val = getHeadProperty(el.getForm().elements.get(el.getProperty("COL")), "COL", headProp);
        
        if(val.equals(""))
            val = getHeadProperty(el.getForm().elements.get(el.getProperty("ROW")), "ROW", headProp);

        return val;
    }
    
    // SVA ---------------------------------------------------------------------
    // ВАРИАНТ ЧЕРЕЗ ПЛАГИН
    
    private String getCellProperty(String cellAddr, String prop) throws LuiException, SQLException
    {
        //return cellPropertyPlugin.compute(cellAddr + "." + prop, null);
        String val = cellPropertyPlugin.compute(cellAddr + "." + prop, null);
        DEBUG(cellAddr + "." + prop + " = " + val);
        return val;
    }
    
    // SVA ---------------------------------------------------------------------

    private int setValueOfCell(Cell cell, String val)
    {           
        DEBUG("val = " + val);

        try
        {
            //val.replace(",", ".")

            cell.setCellValue(Double.parseDouble(val.replace(" ", "").replace(",", ".")));
            //cell.setCellValue(Double.parseDouble(val.indexOf(".") > -1 ? "$" : val));

            DEBUG("-- is number --");

            /*
            if(elType == "CELL")
            {
                String cell_format = getCellFormat(el);
                DEBUG("cell_format = " + cell_format);
                //style.setDataFormat(format.getFormat("#,##0.0000"));

//--                    if( !cell_format.equals("") )
//--                        decor += " NUMFORMAT(" + cell_format + ")";
                decor.addProp("NUMFORMAT", cell_format);
            }
            */
            
            return 1;
        }
        catch(NumberFormatException e)
        {
            DEBUG("-- NumberFormatException --");
            cell.setCellValue(val);
            return 0;
        }  
    }

    // SVA ---------------------------------------------------------------------

    //private int setValueOfHeadCell(Cell cell, String val)
    private void setValueOfHeadCell(Cell cell, String val)
    {           
        // SVA 26.09.2021 Появились случаи, когда значение ячейки заголовка is null
        if( val.equals("") )
        {
            cell.setCellValue("");
            return;
        }
        
        try
        {
            cell.setCellValue(Double.parseDouble(val));
            //return 1;
        }
        catch(NumberFormatException e)
        {
            //cell.setCellValue(val);
            cell.setCellValue( (new WikiToExcell(cell.getSheet(), val, false, java.awt.Color.WHITE)).getSingle() );
            //return 0;
        }  
    }

    // SVA ---------------------------------------------------------------------

    /*
    private void setColInfoCell(GridInfo gi, LuiAppElement el)
    {
        String elType = el.getType();
        
        if( elType.equals("COL") && gi.colInfo == 1 )
        {
            String ci =  el.getProperty( gi.colInfoType == "Y" ? "TUPLE_INFO" : "ltn" );
            //sheet.getRow(py+1).getCell(px).setCellValue(ci);
            //setValueOfCell(sheet.getRow(py+rowSpan).getCell(px), ci);

            Cell cell = sheet.getRow(py+rowSpan).getCell(px);
            setValueOfCell(cell, ci);
            cell.setCellStyle(makeCellStyle(head_decor + " VALIGN(CENTER) HALIGN(CENTER) BBORDER"));
        }
        else if( elType.equals("ROW") && gi.rowInfo == 1 )
        {
        
        }
    }
    */    
    // SVA ---------------------------------------------------------------------

    
    private void buildGridSheet(String fc) throws LuiException, SQLException
    {       
        LuiAppForm frm = app.forms.get(fc);

        cellPropertyPlugin = app.parser.plugins.get("CellProperty");
        String formId = frm.getFormID();
        
        if(debug_mode) DEBUG_FORM_INFO(frm);                      
                
        progres_initialize(fc);

        Sheet sheet = workbook.createSheet(
            //WorkbookUtil.createSafeSheetName(fc)
            WorkbookUtil.createSafeSheetName(frm.getProperty("HEADER"))
            //WorkbookUtil.createSafeSheetName(frm.DgV2.ChartTitle)
        );

        int stage = 0, stages = 9;
        progress(stage, stages);
        
        GridInfo gi = new GridInfo(frm, sheet);


        DEBUG("---");
        DEBUG("colsHeadDepth = " + gi.colsHeadDepth);
        DEBUG("rowsHeadDepth = " + gi.rowsHeadDepth);
        DEBUG("---");
        DEBUG("colsInfoStart.x = " + gi.colsInfoStart.x);
        DEBUG("colsInfoStart.y = " + gi.colsInfoStart.y);
        DEBUG("colsHeadStart.x = " + gi.colsHeadStart.x);
        DEBUG("colsHeadStart.y = " + gi.colsHeadStart.y);
        DEBUG("---");
        DEBUG("rowsInfoStart.x = " + gi.rowsInfoStart.x);
        DEBUG("rowsInfoStart.y = " + gi.rowsInfoStart.y);
        DEBUG("rowsHeadStart.x = " + gi.rowsHeadStart.x);
        DEBUG("rowsHeadStart.y = " + gi.rowsHeadStart.y);
        DEBUG("---");
        
        //if(true) return;
        
        LuiAppElement elRows = frm.elements.get("ROWS");
        LuiAppElement elCols = frm.elements.get("COLS");
        
        //LuiAppElement elCell = frm.elements.get("CELL");

        DEBUG("elRows = " + elRows);
        DEBUG("elCols = " + elCols);
        //DEBUG("elCell = " + elCell);
       
        //sheet.createFreezePane(gi.rowsHeadDepth, colsHeadDepth.value);
        sheet.createFreezePane(gi.zeroCol + gi.rowsHeadDepthFull, gi.zeroRow + gi.colsHeadDepthFull);
        
        progress(stage, stages);

        // Create rows

        progress(stage, stages);
        
        
        //sheet.getRow(0).removeCell(testCell);
        //testCell = null;
        
        // сначала создаем все ячейки таблицы с пустыми значениями
        for( int i = 0; i < gi.zeroRow + gi.colsHeadDepthFull + gi.rowsHeadWidth; i++)
        {
            Row row = sheet.createRow(i);

            for( int j = 0; j < gi.zeroCol + gi.rowsHeadDepthFull + gi.colsHeadWidth; j++)
            {
                Cell cell = row.createCell(j);
                cell.setCellValue("");
            }
        }

        /*
        SVA 24.05.2021 IgnoredErrorType.NUMBER_STORED_AS_TEXT
        Раскомментировать, если потребуется.
        if( sheet instanceof XSSFSheet )
        {
            ((XSSFSheet) sheet).addIgnoredErrors(new CellRangeAddress(gi.colsHeadDepthFull, gi.colsHeadDepthFull + gi.rowsHeadWidth, gi.rowsHeadDepthFull,  gi.rowsHeadDepthFull + gi.colsHeadWidth), IgnoredErrorType.NUMBER_STORED_AS_TEXT);
        }
        */
        
        if( gi.colsHeadDepthFull > 1 || gi.rowsHeadDepthFull > 1 )
            //sheet.addMergedRegion(new CellRangeAddress(gi.zeroCol, gi.zeroCol+gi.colsHeadDepthFull-1, gi.zeroRow, gi.zeroRow+gi.rowsHeadDepthFull-1));
            sheet.addMergedRegion(new CellRangeAddress(gi.zeroRow, gi.zeroRow+gi.colsHeadDepthFull-1, gi.zeroCol, gi.zeroCol+gi.rowsHeadDepthFull-1));

        CellStyle angularCellStyle = makeCellStyle(head_decor + " CENTER VALIGN(CENTER) BBORDER"); // FONTHEIGHT(12) 
        for(int y = gi.zeroRow; y < gi.zeroRow + gi.colsHeadDepthFull; y++)
            for(int x = gi.zeroCol; x < gi.zeroCol + gi.rowsHeadDepthFull; x++)
                sheet.getRow(y).getCell(x).setCellStyle(angularCellStyle);

                
        progress(stage, stages);
        
        // Create cells

        default_separator  = "";
        default_fractional = "";
        if(frm.elements.get("SEPARATOR") != null)
        default_separator  = frm.elements.get("SEPARATOR").getProperty("Value");
        if(frm.elements.get("FRACTIONAL") != null)
        default_fractional = frm.elements.get("FRACTIONAL").getProperty("Value");
        if(default_fractional.equals("")) default_fractional = "2";
        
        
        
        for( String ek : frm.elements.keySet() )
	{
            DEBUG("-------------------------------");
            LuiAppElement el = frm.elements.get(ek);
            String elType = el.getType();
            String val;
            int px, py, colSpan, rowSpan;
            boolean lastLeaf;
            int elementIndex;
            
            Decor decor = new Decor(/* CENTER */ "VALIGN(CENTER) BBORDER WRAPTEXT" /*FONTHEIGHT(12) WRAPTEXT*/);
           
            //DEBUG("###>>> %s (%s)", elType, el.Code);

            switch( elType )
            {
                case "COL":
                    frm.addProperty("CurrentTuple", el.Code);                    
                    elementIndex = Integer.parseInt(el.getProperty("_IDX"));
                    lastLeaf = el.getProperty("_LAST").equals("Y");
                    
                    if( lastLeaf && gi.colInfo == 1 )
                    {
                        Cell cell = sheet.getRow(gi.colsInfoStart.y).getCell(gi.colsInfoStart.x + elementIndex);
                        String ci =  el.getProperty( gi.colInfoType == "Y" ? "TUPLE_INFO" : "ltn" );
                        setValueOfCell(cell, ci);
                        cell.setCellStyle(makeCellStyle(head_decor + " VALIGN(CENTER) HALIGN(CENTER) BBORDER"));
                    }

                    if(gi.colsHeadDepth == 0)
                            continue;
                    
                    int colLevel = Integer.parseInt(el.getProperty("_LEV"));
                    
                    if( lastLeaf )
                        rowSpan = gi.colsHeadDepth - colLevel;
                    else
                        rowSpan = 1;
                    
                    px = gi.colsHeadStart.x + elementIndex;
                    py = gi.colsHeadStart.y + colLevel;
                    colSpan = Integer.parseInt(el.getProperty("_SPAN"));
                    val = el.getProperty("CAPTION");
                    
                    decor.setProps(head_decor);

                    decor.addProp("HALIGN", "CENTER");
                    decor.setProps(getHeadProperty(el, elType, "TUPLE_VISUAL"));
                break;

                case "ROW":
                    frm.addProperty("CurrentTuple", el.Code);                   
                    elementIndex = Integer.parseInt(el.getProperty("_IDX"));
                    lastLeaf = el.getProperty("_LAST").equals("Y");
                    
                    if( lastLeaf && gi.rowInfo == 1 )
                    {
                        Cell cell = sheet.getRow(gi.rowsInfoStart.y + elementIndex).getCell(gi.rowsInfoStart.x);
                        String ci =  el.getProperty( gi.rowInfoType == "Y" ? "TUPLE_INFO" : "ltn" );
                        setValueOfCell(cell, ci);
                        cell.setCellStyle(makeCellStyle(head_decor + " VALIGN(CENTER) HALIGN(CENTER) BBORDER"));
                    }                        

                    if(gi.rowsHeadDepth == 0)
                            continue;
                    
                    int rowLevel = Integer.parseInt(el.getProperty("_LEV"));

                    if( lastLeaf )
                        colSpan = gi.rowsHeadDepth - rowLevel;
                    else
                        colSpan = 1;
                        
                    py = gi.rowsHeadStart.y + Integer.parseInt(el.getProperty("_IDX"));
                    px = gi.rowsHeadStart.x + rowLevel;
                    rowSpan = Integer.parseInt(el.getProperty("_SPAN"));
                    val = el.getProperty("CAPTION");
                    
                    decor.setProps(head_decor);
                    
                    decor.addProp("HALIGN", el.getProperty("ALIGNMENT"), "LEFT");
                    decor.setProps(getHeadProperty(el, elType, "TUPLE_VISUAL"));
                    
                break;

                //case "CELL":
                //    continue;                    
                //break;

                default:
                    continue;
            }

            decor.addProp("DIRECTION");
            
            
            DEBUG("==> %s [%s] rows %d..%d, cols %d..%d  (%s)", elType, el.Code, py, py+rowSpan-1, px, px+colSpan-1, val);

            Row row = sheet.getRow(py);

            Cell cell = row.getCell(px);
            setValueOfHeadCell(cell, val);
                       
            if( p_coloring && cell instanceof XSSFCell && elType == "CELL" )
            {
                String makeup =     getCellProperty(el, "CELL_GETUP", "GETUP");
                String getup  = getCellProperty(el, "CELL_MAKEUP", "MAKEUP");

                if( is_not_null(makeup) && is_not_null(getup) )
                    //if( colored_cols.Get_Element(colCode).equals("Y") )
                        ((XSSFCell)cell).setCellValue(makeRichText(makeup, getup, val));
            }           

            if( rowSpan > 1 || colSpan > 1 )
            {
                sheet.addMergedRegion(new CellRangeAddress(py, py+rowSpan-1, px, px+colSpan-1));
            }

            DEBUG("    decor = <" + decor.getProps() + ">");
            
            for(int y = py; y < py + rowSpan; y++)
                for(int x = px; x < px + colSpan; x++)
                {
                    sheet.getRow(y).getCell(x).setCellStyle(makeCellStyle(decor.getProps()));
                }
        }
        // конец выгрузки элементов COL, ROW
        

        
        /*
        // Выгружаем ячейки
        for(String k : frm.MatrixMap.keySet())
        {
            DEBUG("-------------------------------");

            ArrayList<String> cellList = frm.MatrixMap.get(k);

            LuiAppElement el = frm.elements.get(cellList.get(cellList.size()-1));

            String val;
            int px, py, colSpan, rowSpan;

            Decor decor = new Decor(/ * CENTER * / "VALIGN(CENTER) BBORDER WRAPTEXT" / *FONTHEIGHT(12) WRAPTEXT* /);
            
            //DEBUG("###>>> %s (%s)", elType, el.Code);
            
            LuiAppElement colHeadEl = frm.elements.get( el.getProperty("COL") );
            LuiAppElement rowHeadEl = frm.elements.get( el.getProperty("ROW") );
            
            px = gi.zeroCol + gi.rowsHeadDepthFull + Integer.parseInt(colHeadEl.getProperty("_IDX"));
            py = gi.zeroRow + gi.colsHeadDepthFull + Integer.parseInt(rowHeadEl.getProperty("_IDX"));
            colSpan = Integer.parseInt(colHeadEl.getProperty("_SPAN"));
            rowSpan = Integer.parseInt(rowHeadEl.getProperty("_SPAN"));
            val = el.getProperty("CELL_TEXT");
            decor.setProps(getCellDecor(el));

            if( isShaded(el) )
                continue;

            decor.setProps(getCellProperty(el, "CELL_VISUAL", "VISUALIZATION"));
            decor.addProp("HALIGN", getCellProperty(el, "CELL_HALIGN", "ALIGNMENT"), "CENTER");
            decor.addProp("VALIGN", getCellProperty(el, "CELL_VALIGN", "VALIGN"), "CENTER");

            decor.addProp("DIRECTION");

            //DEBUG("==> %s [%s] rows %d..%d, cols %d..%d  (%s)", elType, el.Code, py, py+rowSpan-1, px, px+colSpan-1, val);

            Row row = sheet.getRow(py);

            Cell cell = row.getCell(px);

            if( setValueOfCell(cell, val) == 1 )
            {
                String cell_format = getCellFormat(el);
                DEBUG("cell_format = " + cell_format);
                decor.addProp("NUMFORMAT", cell_format);
            }

            if( p_coloring && cell instanceof XSSFCell / *&& elType == "CELL"* / )
            {
                String makeup =     getCellProperty(el, "CELL_GETUP", "GETUP");
                String getup  = getCellProperty(el, "CELL_MAKEUP", "MAKEUP");

                if( is_not_null(makeup) && is_not_null(getup) )
                        ((XSSFCell)cell).setCellValue(makeRichText(makeup, getup, val));
            }

            if( rowSpan > 1 || colSpan > 1 )
            {
                sheet.addMergedRegion(new CellRangeAddress(py, py+rowSpan-1, px, px+colSpan-1));
            }

            DEBUG("    decor = <" + decor.getProps() + ">");
            
            for(int y = py; y < py + rowSpan; y++)
                for(int x = px; x < px + colSpan; x++)
                {
                    sheet.getRow(y).getCell(x).setCellStyle(makeCellStyle(decor.getProps()));
                }
        }
        */


        // Выгружаем ячейки
        // НОВЫЙ ВАРИАНТ
        for(int y = 0; y < gi.rowsHeadWidth; y++)
        for(int x = 0; x < gi.colsHeadWidth; x++)
        {
            DEBUG("~cell~~~~~~~~~~~~~~~~~~~~~~~~~~");

            Decor decor = new Decor(/* CENTER */ "VALIGN(CENTER) BBORDER WRAPTEXT" /*FONTHEIGHT(12) WRAPTEXT*/);

            int px = gi.zeroCol + gi.rowsHeadDepthFull + x;
            int py = gi.zeroRow + gi.colsHeadDepthFull + y;

            Cell cell = sheet.getRow(py).getCell(px);
            
            String idx = (y + 1) + ":" + (x + 1);
            DEBUG("idx = " + idx);
            
            ArrayList<String> cellList = frm.MatrixMap.get(idx);

            if(cellList != null)
            {    
                LuiAppElement el = frm.elements.get(cellList.get(cellList.size()-1));           

                String cellAddr = formId + "." + el.getProperty("ROW") + ":" + el.getProperty("COL");
                String val = getCellProperty(cellAddr, "CELL_SHOW");

                decor.setProps(getCellDecor(el));

                decor.setProps(getCellProperty(cellAddr, "CELL_VISUAL"/*, "VISUALIZATION"*/));
                decor.addProp("HALIGN", getCellProperty(cellAddr, "CELL_HALIGN"/*, "ALIGNMENT"*/), "CENTER");
                decor.addProp("VALIGN", getCellProperty(cellAddr, "CELL_VALIGN"/*, "VALIGN"*/), "CENTER");

                decor.addProp("DIRECTION");

                if( setValueOfCell(cell, val) == 1 )
                {
                    String cell_format = getCellFormat(el);
                    DEBUG("cell_format = " + cell_format);
                    decor.addProp("NUMFORMAT", cell_format);
                }

                if( p_coloring && cell instanceof XSSFCell /*&& elType == "CELL"*/ )
                {
                    String makeup =     getCellProperty(cellAddr, "CELL_GETUP"/*, "GETUP"*/);
                    String getup  = getCellProperty(cellAddr, "CELL_MAKEUP"/*, "MAKEUP"*/);

                    if( is_not_null(makeup) && is_not_null(getup) )
                            ((XSSFCell)cell).setCellValue(makeRichText(makeup, getup, val));
                }

                DEBUG("    decor = <" + decor.getProps() + ">");            
            }

            cell.setCellStyle(makeCellStyle(decor.getProps()));
        }

        progress(stage, stages);
        
        // попробуем ресайзить заголовки прямо в эксель таблице

        DEBUG("gi.colsHeadWidth = " + gi.colsHeadWidth);
        DEBUG("gi.colsHeadDepth = " + gi.colsHeadDepth);
        DEBUG("gi.rowsHeadWidth = " + gi.rowsHeadWidth);
        DEBUG("gi.rowsHeadWidth = " + gi.rowsHeadDepth);
        

        // Пробуем сделать масимальный ресайз всей таблицы

        int tabWith = 0;
        
        for(int x = gi.zeroCol; x < gi.zeroCol + gi.rowsHeadDepthFull + gi.colsHeadWidth; x++)
        {
            sheet.autoSizeColumn(x, true);
            tabWith += sheet.getColumnWidth(x);
        }
        
        DEBUG("tabWith = " + tabWith);

        // если таблица получилась узкая, то и хорошо
        if(tabWith <= 39000) // for 1280
        {
            // Увеличим высоту строк на треть
            for(int y = gi.zeroRow; y < gi.zeroRow + gi.colsHeadDepthFull + gi.rowsHeadWidth; y++)
            {
                //sheet.getRow(y).setHeight( (short) Math.round(sheet.getRow(y).getHeight() * 1.3f) );
                sheet.getRow(y).setHeight( (short) (sheet.getRow(y).getHeight() * 2) );
            }
        }
        else // иначе, если таблица получилась широкая, то будем ресайзить
        {
            // сделать столбцы узкими
            for(int x = gi.zeroCol; x < gi.zeroCol + gi.rowsHeadDepthFull + gi.colsHeadWidth; x++)
            {
                sheet.setColumnWidth(x, 10);
            }

            // делаем за два прохода
            // сначала подбираем ширину столбцов,
            // а вторым проходом подбираем высоту
            for(int k=0; k<2; k++)
            {
                // Верхний заголовок
                for(int y=gi.zeroRow+gi.colsHeadDepthFull-1; y>=gi.zeroRow; y--)
                {
                    for(int x = gi.zeroCol + gi.rowsHeadDepthFull; x < gi.zeroCol + gi.rowsHeadDepthFull + gi.colsHeadWidth; x++)
                    {
                        Row row = sheet.getRow(y);
                        Cell rcell = row.getCell(x);
                        String val = (rcell.getCellType() == CellType.NUMERIC) ? String.valueOf(rcell.getNumericCellValue()) : rcell.getStringCellValue();


                        if( !val.equals("") )
                        {
                            CellRangeAddress cellRange = getCellRange(sheet, x, y);

                            int calculatedColWidth = gi.calcColWidth(val);

                            //calculatedColWidth += 500;

                            int colWidth = 0;
                            for( int rx = x; rx <= cellRange.getLastColumn(); rx++ )
                                colWidth += sheet.getColumnWidth(rx);

                            if( calculatedColWidth > colWidth)
                            {
                                int deltaWidth = calculatedColWidth - colWidth / (cellRange.getLastColumn() - x + 1);                        
                                colWidth = 0;

                                for( int rx = x; rx <= cellRange.getLastColumn(); rx++ )
                                {    
                                    int curWidth = sheet.getColumnWidth(x);
                                    sheet.setColumnWidth(x, curWidth + deltaWidth);
                                    colWidth += curWidth + deltaWidth;
                                }
                            }

                            if(k>0) // второй проход - подбираем высоту верхних заголовков
                            {
                                int lines = 1 + calcCellLines(gi, rcell, colWidth, val);

                                float fontHeight = workbook.getFontAt( rcell.getCellStyle().getFontIndexAsInt() ).getFontHeight();
                                int cellHeight = Math.round((fontHeight*1.3f) * lines);
                                int cellHeightSum = 0;

                                for( int ry = y; ry <= cellRange.getLastRow(); ry++ )
                                    cellHeightSum += sheet.getRow(ry).getHeight();

                                if( cellHeight > cellHeightSum )
                                {
                                    int deltaHeight = (cellHeight - cellHeightSum) / (cellRange.getLastRow() - y + 1);

                                    for( int ry = y; ry <= cellRange.getLastRow(); ry++ )
                                    {
                                        short d = (short) (sheet.getRow(ry).getHeight() + deltaHeight);
                                        sheet.getRow(ry).setHeight( d );
                                    }
                                }
                            }
                        }
                    }
                }

                // Левый заголовок
                for(int x = gi.zeroCol + gi.rowsHeadDepthFull-1; x>=gi.zeroCol; x--)
                {
                    for(int y = gi.zeroRow + gi.colsHeadDepthFull; y < gi.zeroRow + gi.colsHeadDepthFull + gi.rowsHeadWidth; y++)
                    {
                        Row row = sheet.getRow(y);
                        Cell rcell = row.getCell(x);
                        String val = (rcell.getCellType() == CellType.NUMERIC) ? String.valueOf(rcell.getNumericCellValue()) : rcell.getStringCellValue();

                        DEBUG("===> (%s) x=%d, y=%d val=%s", cellNum(x,y), x, y, val);

                        if( !val.equals("") )
                        {
                            CellRangeAddress cellRange = getCellRange(sheet, x, y);

                            int calculatedColWidth = gi.calcColWidth(val);

                            // 26.02.2021 Если поле объединеное, то из расчетной ширины 
                            // надо вычесть суммарную ширину колонок правее.

                            calculatedColWidth -= rightSubColumns(rcell, cellRange);

                            int colWidth = 0;
                            for( int rx = x; rx <= cellRange.getLastColumn(); rx++ )
                                colWidth += sheet.getColumnWidth(rx);

                            if( calculatedColWidth > colWidth)
                            {
                                int deltaWidth = calculatedColWidth - colWidth / (cellRange.getLastColumn() - x + 1);                        
                                colWidth = 0;

                                for( int rx = x; rx <= cellRange.getLastColumn(); rx++ )
                                {    
                                    int curWidth = sheet.getColumnWidth(x);
                                    sheet.setColumnWidth(x, curWidth + deltaWidth);
                                    colWidth += curWidth + deltaWidth;
                                }
                            }

                            if(k>0) // второй проход - подбираем высоту левых заголовков
                            {
                                int lines = 1 + calcCellLines(gi, rcell, colWidth, val);

                                float fontHeight = workbook.getFontAt( rcell.getCellStyle().getFontIndexAsInt() ).getFontHeight();
                                DEBUG("font height = " + fontHeight );
                                //int cellHeight = (fontHeight + fontHeight/2) * lines;
                                int cellHeight = Math.round((fontHeight*1.3f) * lines);
                                int cellHeightSum = 0;

                                for( int ry = y; ry <= cellRange.getLastRow(); ry++ )
                                    cellHeightSum += sheet.getRow(ry).getHeight();

                                if( cellHeight > cellHeightSum )
                                {
                                    int deltaHeight = (cellHeight - cellHeightSum) / (cellRange.getLastRow() - y + 1);

                                    for( int ry = y; ry <= cellRange.getLastRow(); ry++ )
                                    {
                                        short d = (short) (sheet.getRow(ry).getHeight() + deltaHeight);
                                        sheet.getRow(ry).setHeight( d );
                                    }
                                }
                            }                    
                        }
                    }
                }
            }
        }

        /*
        {
            int w1 = sheet.getColumnWidth(1);
            int w2 = sheet.getColumnWidth(2);
            DEBUG("*** w1=%d, w2=%d, w1+w2=%d", w1, w2, w1+w2);
        }
        */

        progress(stage, stages);

        // раздвинуть ячейки по ширине данных
        
        for( int x = gi.zeroCol + gi.rowsHeadDepthFull; x < gi.zeroCol + gi.rowsHeadDepthFull + gi.colsHeadWidth; x++ )
        {
            for( int y = gi.zeroRow + gi.colsHeadDepthFull; y < gi.zeroRow + gi.colsHeadDepthFull + gi.rowsHeadWidth; y++ )
            {
                Row row = sheet.getRow(y);
                Cell rcell = row.getCell(x);

                /* не работает. раздвигает по длинным заголовкам
                // ПРОБА: Если число, то просто делаем автосайз
                if( rcell.getCellType() == CellType.NUMERIC )
                {
                    sheet.autoSizeColumn(x, true);
                    continue;
                }
                */
                
                //String val = rcell.getStringCellValue();
                String val = (rcell.getCellType() == CellType.NUMERIC) ? String.valueOf(rcell.getNumericCellValue()) : rcell.getStringCellValue();
                //DEBUG("& (%s) %s", cellNum(x,y), val);

                if( !val.equals("") )
                {
                    CellRangeAddress cellRange = getCellRange(sheet, x, y);
                    
                    //int calculatedColWidth = gi.calcColWidth(val);
                    int calculatedColWidth = gi.calcColWidth(rcell, val);
			calculatedColWidth -= rightSubColumns(rcell, cellRange);
                    
                    int colWidth = 0;
                    for( int rx = x; rx <= cellRange.getLastColumn(); rx++ )
                        colWidth += sheet.getColumnWidth(rx);

                    //DEBUG("& colWidth = %d, calculatedColWidth = %d", colWidth, calculatedColWidth);

                    if( calculatedColWidth > colWidth)
                    {
                        int deltaWidth = calculatedColWidth - colWidth / (cellRange.getLastColumn() - x + 1);
                        colWidth = 0;
                        
                        for( int rx = x; rx <= cellRange.getLastColumn(); rx++ )
                        {    
                            int curWidth = sheet.getColumnWidth(x);
                            sheet.setColumnWidth(x, curWidth + deltaWidth);
                            colWidth += curWidth + deltaWidth;
                        }
                    }
		}
            }
        }

        progress(stage, stages);

        makeAngular(sheet, frm.getProperty("ANGULAR"), gi.rowsHeadDepthFull, gi.colsHeadDepthFull, gi);
        //makeAngular(sheet, frm.getProperty("ANGULAR"), gi.rowsHeadDepth, gi.colsHeadDepth, gi);
        //makeAngular(sheet, "ANGULAR", gi.rowsHeadDepth, gi.colsHeadDepth);



        /*
        заголовок над таблицей

        grid_title
        grid_title_border        

        col_info надо ли выводить информацию о столбце	N/Y/O
        row_info 

        frm.prop COL_INFO = O
        frm.prop ROW_INFO = Y        

        Y - из свойства TUPLE_INFO
        O (order) - порядковый номер столбца/строки из свойства ltn (leaf tuple number)
        
        */

        String gridTitle = frm.getProperty("GRID_TITLE");

        DEBUG("GRID_TITLE = " + gridTitle);
        
        if( !gridTitle.equals("") )
        {
            Cell cell = sheet.getRow(0).getCell(0);

            cell.setCellValue(gridTitle);

            sheet.addMergedRegion(new CellRangeAddress(0, 0, 0, gi.zeroCol + gi.rowsHeadDepthFull + gi.colsHeadWidth - 1));
            
            if(frm.getProperty("GRID_TITLE_BORDER") == "Y")
            {
                
            }
        }
        
        gi.clear();
        
        progress(stage, stages);
        
        DEBUG("-END-");
    }
    
    // SVA ---------------------------------------------------------------------
    
    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        LuiClEvent ev;
        String imageBase64;
        byte[] imageData;
        BufferedImage bufferedImage;
        LuiAppForm frm;
        String curFrmID = null;       
        LuiCollection cs = new LuiCollection("Print:" + pString);
        String maxRowsS = cs.Get_Element("Print");
        String needTotS = cs.Get_Element("TOTALS");
        int maxRows = Integer.valueOf(NVL(maxRowsS, "1000"));
        boolean singleForm = cs.Get_Element("SCOPE").equals("F");
        String mainFormID = cs.Get_Element("FormID");
             
        if( is_null(mainFormID) ) return "";
        if( !app.forms.containsKey(mainFormID) ) return "";

        // SVA 2021.06.01
        String statFormID = cs.Get_Element("StatFormID");
        DEBUG("StatFormID = " + statFormID);
        if( statFormID.equals("") )
        {
            statFormID = "<<NULL>>";
            
        }
        
        frm = app.forms.get(mainFormID);
        
        if( frm == null )
        {
            curFrmID = app.getVal("FormID");
            
            if( app.forms.containsKey(curFrmID) )
            {
               frm = app.forms.get(curFrmID);
            }
            else
                return "";
        }
        
        curFrmID = frm.getFormID();

        // поскольку вероятно придётся давать команды браузеру для сохранения форм,
        // то сначала мы отправим серверу визуализации все имеющиеся в буферах команды
        app.changes();
        app.getSeance().remZeroCmd(); // удалить пустышки из очереди команд (пустые группы)
        app.getSeance().remFrmWinApp(); // удаление форм и окон...

        //SVA  cw.flush();
        //SVA сейчас нельзя. 
        
        // кроме того удалим из очереди событий клиента всё что там есть сейчас
        er.eventList.clear();
        
        // и проверим не исчезла ли наша форма и её окно
        if( !app.forms.containsKey(curFrmID) )
            return "";

        String winID = frm.getVal("WindowID"); // узнаём код окна формы

        if( !app.windows.containsKey(winID) )
            return"";
       
        // SVA
        LuiClCommandGroup XMLcg = cw.addGroup("GetChart");
        Boolean doGetChart = false;

        for( String fc : app.forms.keySet() ) // крутим все формы
        {
            if( singleForm && (!fc.equals(mainFormID)) )
                continue;

            // расположенные в том же окне и являющиеся диаграммами
            if (!app.forms.get(fc).getVal("WindowID").equals(winID) || !app.forms.get(fc).getType().equals("CHART"))
                continue;

            LuiAppForm cfrm = app.forms.get(fc); // берём форму-диаграмму

            // выполняем подготовку команды серверу визуализации для того чтобы браузер сгенерировал PNG
            LuiClCommand XMLcmd = XMLcg.addCommand("GetChart"); //вот это совершенно новая команда
            XMLcmd.addScaPar("FormID", fc);

            doGetChart = true;
            
            // теперь надо отправить команду Жене
            //SVA cw.flush();
            //SVA сейчас нельзя отправлять. Отправим один раз после того как все команды быдут добавлены.
        }

        if( doGetChart )
        {
            // отправляем команды GetChart

            cw.flush();

            //SVA ВОЗМОЖНО ЭТО ПОТЕНЦИАЛЬНАЯ ПРОБЛЕМА
            // Если картинок нет, то ничего не отправляем. И тогда не уйдут команды,
            // которые возможно былы в буферах. 
            
           
            // в ответ мы ждём событе Image о готовности диаграммы в base64

            er.readEvents();

            // цикл обработки событий
            while( (ev = er.getEvent()) != null  )
            {
                if( ev.name.equals("Image") )
                {
                    saveImage(ev.getParValue("FormID"), ev.getParValue("Data"));
                }
            }
        }
       
        // снова крутим формы
        for( String fc : app.forms.keySet() ) // крутим все формы
        {
            //DEBUG("[1]--form--> " + fc);
            if( singleForm && (!fc.equals(mainFormID)) )
                continue;

            //DEBUG("[2]--form--> " + fc);
            if( !app.forms.get(fc).getVal("WindowID").equals(winID) && !fc.equals(statFormID) )
                continue;
            
            // расположенные в том же окне или форма, для которой посчитана статистика

            //DEBUG("[3]--form--> " + fc);
            String formType = app.forms.get(fc).getType();
            
            //DEBUG("FORM TYPE = " + formType);

            if( formType.equals("CHART") )
            {
                buildChartSheet(fc);
            }

            if( formType.equals("LIST") )
            {
                buildListSheet(fc);
            }

            if( formType.equals("GRID") )
            {
                buildGridSheet(fc);
            }
        }

        //SVA DEBUG
        if( debug_mode ) return null;
        
        return app.getTmpDirURL() + "output.xlsx";
    }

    // SVA ---------------------------------------------------------------------

    protected boolean createWorkbook() throws LuiException, SQLException//, FileNotFoundException
    {
        clearMemory();

        bookFileName = app.getTmpDir() + "output.xlsx";

        //SVA DEBUG
        //if( debug_mode ) bookFileName = "/xls/output.xlsx";
        //bookFileName = app.getTmpDir() + "output" + (cs.Get_Element("STREAM_MODE").equals("Y")?"_SXSSF_":"_XSSF_") + ".xlsx";

        try
        {
            filew = new FileOutputStream(new File(bookFileName));
        }
        catch (FileNotFoundException ex)
        {
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
            return false;
        }

        if( cs.Get_Element("STREAM_MODE").equals("Y") )
            workbook = new SXSSFWorkbook(100);
        else
            workbook = new XSSFWorkbook();
        
        return true;
    }
    
    // SVA ---------------------------------------------------------------------

    protected void closeWorkbook() throws LuiException
    {
        if (workbook==null) return;
        try
        {
            workbook.write(filew);
            
            workbook.close();

            if( workbook instanceof SXSSFWorkbook )
                ((SXSSFWorkbook)workbook).dispose();
            
            filew.flush();
            filew.close();
        }
        catch( FileNotFoundException ex )
        {
            // SVA зачем это все???
            /*
            if( filew != null )
            {
                if( workbook != null )
                {
                    try
                    { 
                        workbook.close();
                        if( workbook instanceof SXSSFWorkbook )
                            ((SXSSFWorkbook)workbook).dispose();
                    }
                    catch(Exception wex){}
                }
            
                try
                {
                    filew.flush();
                    filew.close();
                    filew = null;
                } catch(Exception ef) {}
            }
            */
            
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
        }
        
        catch( IOException ex )
        {
            /*
            if (workbook != null)
            {
                try
                { 
                    workbook.close();
                    if( workbook instanceof SXSSFWorkbook )
                        ((SXSSFWorkbook)workbook).dispose();
                }
                catch(Exception wex){}
            }
            
            if( filew != null )
            {
                try
                {
                    filew.flush();
                    filew.close();
                    filew = null;
                }
                catch(Exception ef) {}
            }
            */
            app.msgWriter.write("LUI-00097", bookFileName, ex.getMessage());
        }

        clearMemory();
    }

    // SVA ---------------------------------------------------------------------
    
    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        DEBUG("START execute at " + LocalTime.now());
        DEBUG("pString = " + pString);
       
        String adr = "";
        String fc = app.getVal("FormID");
        if( is_null(fc) ) return;
        if( !app.forms.containsKey(fc) ) return;
        
        DEBUG("TmpDir    = " + app.getTmpDir());
        DEBUG("TmpDirURL = " + app.getTmpDirURL());
               
        cs = new LuiCollection("Print:" + pString);

        if( createWorkbook() )
        {
            p_stream_mode = cs.Get_Element("STREAM_MODE").equals("Y");
            p_coloring    = cs.Get_Element("COLORING").equals("Y");

            //DEBUG("p_stream_mode = " + p_stream_mode);
            //DEBUG("p_coloring    = " + p_coloring);
            //DEBUG("COLUMNS = " + cs.Get_Element("COLUMNS"));

            String columns = cs.Get_Element("COLUMNS").replace("=", ":").replace("&", ";");

            //DEBUG("columns = " + columns);
            if( p_stream_mode ) p_coloring = false;

            if( p_coloring ) colored_cols = new LuiCollection(columns);
            
            adr = compute(pString, pProp);
            
            closeWorkbook();
        }

        if( is_not_null(adr) )
        {
            LuiAppForm frm = app.forms.get(fc);
            LuiClCommandGroup XMLcg = app.commandWriter.addGroup("PRINT");
            LuiClCommand XMLcmd = XMLcg.addCommand("URL");
            XMLcmd.addScaPar("Address", adr);
            XMLcmd.addScaPar("Target", "Tab");
        }

        //DEBUG("END execute at " + LocalTime.now());
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private void progres_initialize(String fc)
    {
        frmInd = null;
        elInd = null;
        elIndOK = null;

        fc = app.getVal("FormID");
    
        if( app.forms.containsKey(fc) )
        {
           frmInd = app.forms.get(fc);

           if( frmInd.elements.containsKey("PROGRESS$INDICATOR") )
           {
               elInd=frmInd.elements.get("PROGRESS$INDICATOR");

               if( frmInd.elements.containsKey("OK") )
               {
                   elIndOK = frmInd.elements.get("OK");
               }
           }
           else
           {
               frmInd = null;
           }
        }
    }

    
    private boolean progress(String pText,String pPct) throws LuiException, SQLException
    {
        boolean retVal=true;

        if( elInd != null )
        {
            elInd.setProperty("VISIBLE","1");
            
            if( elIndOK != null )
            {
                elIndOK.setProperty("ACTIVE","N");
            }
            
            if( is_not_null(pText) )
            {
                elInd.setProperty("HELPER_TEXT", "   " + pText);
            }
            else
            {
                elInd.setProperty("HELPER_TEXT", "");
            }

            if( is_not_null(pPct) )
            {
                elInd.setProperty("VISUALIZATION", "PROGRESS PROC" + pPct);
                elInd.setProperty("Value", "   " + pPct+"%");
            }
            else
            {
                elInd.setProperty("VISUALIZATION", "");
                elInd.setProperty("Value", "");
            }

            elInd.properties.get("Value").setChangeFlag();
            elInd.properties.get("HELPER_TEXT").setChangeFlag();
            elInd.properties.get("VISUALIZATION").setChangeFlag();
            elInd.properties.get("Show").setChangeFlag();
            frmInd.changes();
            XMLcg = app.commandWriter.addGroup("CallBack");
            XMLcmd = XMLcg.addCommand("Event");
            XMLcmd.addScaPar("EventCode", "ProgressContinue");
            XMLcmd.addScaPar("MS","0");
            app.commandWriter.flush();
            app.eventReader.readEvents();

            while (true)
            {
                LuiClEvent ev = app.eventReader.getEvent();

                if( ev == null ) break;

                if( ev.name.equals("ProgressContinue") || ev.name.equals("GotoField") )
                    retVal = true;
                else
                    retVal = false;        
            }
        }

        return retVal;
    }

    // SVA ---------------------------------------------------------------------

    private void progress(int stage, int stages) throws LuiException, SQLException
    {
        ++stage;
        progress(String.valueOf(stage) + "/" + String.valueOf(stages), String.valueOf(stage*100/stages));
    }
}
