/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * синтаксис Compute:
LuiFile:
 */
package Standard;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import LuiExceptionMsg.LuiException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.sql.SQLException;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.Writer;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;



/**
 *
 * @author aliubushkin
 */
public class LuiFile implements LuiPlugin{

    LuiApplication app;
    HashMap<String,LuiFileDescriptor>  fileDescrMap;
   // HashMap<FileDescriptor,File>  fileStreamMap = new HashMap<>();

    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
        
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       return exe('C',pString,pProp);
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       if (pString.equals("CloseAll"))CloseAll();
       else
       exe('E',pString,pProp);
    }
    
    String exe(char executeOrCompute,String pString,LuiAppProp pProp) throws LuiException, SQLException {
        String res="";
        LuiCollection mainCollection = new LuiCollection("Command:"+pString);
        String command = mainCollection.Get_Element("Command");
        String fileName = mainCollection.Get_Element("FileName");
        String srcFileName = mainCollection.Get_Element("SrcFileName");
        String dstFileName = mainCollection.Get_Element("DstFileName");
        String dirName = mainCollection.Get_Element("DirName");
        String srcDirName = mainCollection.Get_Element("SrcDirName");
        String dstDirName = mainCollection.Get_Element("DstDirName");
        String option = mainCollection.Get_Element("Option");
        String output = mainCollection.Get_Element("Output");
        String openMode = mainCollection.Get_Element("OpenMode");
        String textLine = mainCollection.Get_Element("TextLine");
        String fileDescrStr = mainCollection.Get_Element("FileDescr");
        if (fileName.length() > 0) if (!fileName.contains(System.getProperty("file.separator"))) fileName = app.getTmpDir() + fileName;
        if (srcFileName.length() > 0) if (!srcFileName.contains(System.getProperty("file.separator"))) srcFileName = app.getTmpDir() + srcFileName;
        if (dstFileName.length() > 0) if (!dstFileName.contains(System.getProperty("file.separator"))) dstFileName = app.getTmpDir() + dstFileName;
        if (dirName.length() > 0) if (!dirName.contains(System.getProperty("file.separator"))) dirName = app.getTmpDir() + dirName;
        if (srcDirName.length() > 0) if (!dirName.contains(System.getProperty("file.separator"))) srcDirName = app.getTmpDir() + srcDirName;
        if (dstDirName.length() > 0) if (!dirName.contains(System.getProperty("file.separator"))) dstDirName = app.getTmpDir() + dstDirName;
        File file = null;
        File dir = null;
        File srcFile = null;
        File dstFile = null;
                
        switch (command) {
            case "CheckFile":
                option = NVL(option,"Y");
                output = NVL(output,"N");
                if (fileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя файла (FileName)");
                file = new File(fileName);
                // файл существует
                if (file.exists()) {
                    if (output.equals("N")) return file.getAbsolutePath(); // возврат полного имени файла
                    else if (output.equals("L")) return String.valueOf(file.length()); // возврат длины файла в байтах
                    else {
                        long timestamp = file.lastModified();
                        return (new Date(timestamp)).toString(); // возврат даты модификации файла
                    }
                // файл не существует (не найден)
                } else {
                    if (option.equals("Y")) return ""; // возврат пустого значения
                    else app.msgWriter.write("LUI-00117", fileName); // генерация сообщения об ошибке
                }
                break;
            case "CreateFile": 
                option = NVL(option,"Y");
                if (fileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя файла (FileName)");
                file = new File(fileName);
                try {
                    // файл создан
                    if (file.createNewFile()) return file.getAbsolutePath(); // возврат полного имени файла 
                    // файл не создан (уже существует)
                    else {
                        switch (option) {
                            case "C": backupFile(file); // создание бэкапной копии существующего файла
                            case "Y": clearFile(file); // очищаем файл
                                      return file.getAbsolutePath(); // возврат полного имени файла   
                        }
                        // сюда придём только в случае, когда следует сообщить, что файл не создан
                        app.msgWriter.write("LUI-00118", fileName); // генерация сообщения об ошибке
                    }
                // ошибка ввода/вывода при создании файла
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00118", fileName, e.toString()); // генерация сообщения об ошибке. Ошибка создания файла.
                }
                break;
            case "CopyFile":
                if (srcFileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя копируемого файла (SrcFileName)");
                if (dstFileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя для копии файла (DstFileName)");
                srcFile = new File(srcFileName);
                dstFile = new File(dstFileName);
                try {
//                    Files.copy(srcFile.toPath(), dstFile.toPath(), StandardCopyOption.REPLACE_EXISTING); // с заменой существующего файла dstFileName
                    Files.copy(srcFile.toPath(), dstFile.toPath()); // если файл dstFileName уже существует, то будет ошибка копирования
                    return dstFile.getAbsolutePath(); // возврат полного имени нового файла
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00131", srcFileName, e.toString()); // ошибка копирования файла
                }
                break;
            case "RenameFile": 
                if (srcFileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано исходное имя файла (SrcFileName)");
                if (dstFileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано новое имя файла (DstFileName)");
                srcFile = new File(srcFileName);
                dstFile = new File(dstFileName);
                if (srcFile.renameTo(dstFile)) return dstFile.getAbsolutePath(); // возврат полного нового имени файла 
                else app.msgWriter.write("LUI-00122", srcFileName); // не удалось переименовать файл
                break;        
            case "DeleteFile":
                option = NVL(option,"Y");
                if (fileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя файла (FileName)");
                file = new File(fileName);
                if (file.delete() || option.equals("Y")) return file.getAbsolutePath(); // возврат полного имени файла
                else app.msgWriter.write("LUI-00123", fileName); // генерация сообщения об ошибке. Ошибка удаления файла.
                break;
            case "OpenFile": 
                if (fileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя файла (FileName)");
                if (openMode.length() == 0) app.msgWriter.write("LUI-00128", "Не указан режим (OpenMode), в котором необходимо открыть файл");
                openMode = openMode.toLowerCase();
                try {
                    file = new File(fileName);
                    if (!file.exists()) app.msgWriter.write("LUI-00117", fileName); // генерация сообщения об ошибке. Файл не найден
                    FileDescriptor fileDescr = null;
                    LuiFileDescriptor luiFD = null;
                    boolean appendMode = false;
                    switch (openMode) {
                        case "r": // чтение из файла
                            FileInputStream fileInputStream = new FileInputStream(file);
                            fileDescr = fileInputStream.getFD();
                            BufferedReader reader = new BufferedReader(new InputStreamReader(fileInputStream, StandardCharsets.UTF_8));
//                            BufferedReader reader = new BufferedReader(new FileReader(fileDescr));
                            luiFD = new LuiFileDescriptor(fileDescr, file, reader, null);
                            break;
                        case "a": // запись в конец файла
                            appendMode = true;
                        case "w": // запись в начало файла
                            FileOutputStream fileOutputStream = new FileOutputStream(file, appendMode);
                            fileDescr = fileOutputStream.getFD();
                            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fileOutputStream, StandardCharsets.UTF_8));
//                            BufferedWriter writer = new BufferedWriter(new FileWriter(fileDescr));
                            luiFD = new LuiFileDescriptor(fileDescr, file, null, writer);
                            break;
                    }
                    fileDescrMap.put(fileDescr.toString(), luiFD);
                    return fileDescr.toString();
                } catch (FileNotFoundException e) {
                    app.msgWriter.write("LUI-00117", fileName, e.toString()); // генерация сообщения об ошибке. Файл не найден
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00121", fileName, openMode, e.toString()); // генерация сообщения об ошибке. Ошибка при открытии файла
                }
                break;
            case "ReadData": // протестировано чтение содержимого только текстовых файлов
                option = NVL(option,"Y");
                if (fileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя файла (FileName)");
                String s;
                StringBuilder filedata = new StringBuilder("");
                try {
                    file = new File(fileName);
                    // файл существует
                    if (file.exists()) {
                        FileInputStream fileInputStream = new FileInputStream(file);
                        BufferedReader reader = new BufferedReader(new InputStreamReader(fileInputStream, StandardCharsets.UTF_8));
                        do {
                            s = reader.readLine();
                            if (filedata.length() == 0) filedata.append(s); 
                            else {
                                if (s != null) {
                                    filedata.append("\n");
                                    filedata.append(s);
                                }
                            }
                        } while (s != null);
                        reader.close();
                        return filedata.toString(); // вернём всё содержимое файла
                    // файл не существует (не найден)
                    } else {
                        if (option.equals("Y")) return ""; // возврат пустого значения
                        else app.msgWriter.write("LUI-00117", fileName); // генерация сообщения об ошибке
                    }
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00128", "Чтение из файла. Что-то пошло не так.", e.toString()); // генерация сообщения об ошибке 
                }    
                break;
            case "SaveData": 
                option = NVL(option,"Y");
                if (fileName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя файла (FileName)");
                break;
            case "ReadString":
                if (fileDescrStr.length() == 0) app.msgWriter.write("LUI-00128", "Не указан дескриптор файла (FileDescr)"); 
                try {
                    LuiFileDescriptor luiFD = fileDescrMap.get(fileDescrStr);
                    FileDescriptor fileDescr = luiFD.getFD();
                    if (!fileDescr.valid()) app.msgWriter.write("LUI-00119", fileName); // генерация сообщения об ошибке. Дескриптор испорчен
//                    return NVL(luiFD.getReader().readLine(),"");
                    return NVL(luiFD.getReader().readLine(),"~EOF~");
//                    return luiFD.getReader().readLine();
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00128", "Чтение из файла. Что-то пошло не так.", e.toString()); // генерация сообщения об ошибке 
                } catch (NullPointerException e) {
                    app.msgWriter.write("LUI-00128", "Чтение из файла. Указанный дескриптор файла не существует.", e.toString()); // генерация сообщения об ошибке 
                }
                break;
            case "SaveString":
                if (fileDescrStr.length() == 0) app.msgWriter.write("LUI-00128", "Не указан дескриптор файла (FileDescr)"); 
                try {
                    LuiFileDescriptor luiFD = fileDescrMap.get(fileDescrStr);
                    FileDescriptor fileDescr = luiFD.getFD();
                    if (!fileDescr.valid()) app.msgWriter.write("LUI-00120", fileName); // генерация сообщения об ошибке. Дескриптор испорчен
                    BufferedWriter writer = luiFD.getWriter();
                    writer.write(textLine);
                    writer.append('\n');
                    writer.flush();
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00128", "Запись в файл. Что-то пошло не так.", " (" + e.toString() + ")"); // генерация сообщения об ошибке 
                } catch (NullPointerException e) {
                    app.msgWriter.write("LUI-00128", "Запись в файл. Указанный дескриптор файла не существует.", e.toString()); // генерация сообщения об ошибке 
                }
                break;
            case "CloseFile": 
                try {
                    if (fileDescrStr.length() == 0) app.msgWriter.write("LUI-00128", "Не указан дескриптор файла (FileDescr)"); 
                    LuiFileDescriptor luiFD = fileDescrMap.get(fileDescrStr);
                    FileDescriptor fileDescr = luiFD.getFD();
                    if (!fileDescr.valid()) app.msgWriter.write("LUI-00128", "Ошибка закрытия файла " + fileName); // генерация сообщения об ошибке. Дескриптор испорчен
                    luiFD.closeIOStream();
                    fileDescrMap.remove(fileDescrStr);
                    return luiFD.getFile().getAbsolutePath(); // возврат полного имени закрытого файла
                } catch (IOException e) {
                    app.msgWriter.write("LUI-00128", "Ошибка закрытия файла " + fileName, e.toString()); // генерация сообщения об ошибке. Не удалось закрыть поток работы с файлом
                } catch (NullPointerException e) {
                    app.msgWriter.write("LUI-00128", "Ошибка закрытия файла. Указанный дескриптор файла не существует.", e.toString()); // генерация сообщения об ошибке 
                }
                break;
            case "CheckDir":
                option = NVL(option,"Y");
                if (dirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя каталога (DirName)");
                file = new File(dirName);
                // каталог существует
                if (file.exists()) return file.getAbsolutePath(); // возврат полного имени каталога
                // каталог не существует (не найден)
                else {
                    if (option.equals("Y")) return ""; // возврат пустого значения
                    else app.msgWriter.write("LUI-00124", dirName); // генерация сообщения об ошибке
                }
                break;
            case "CreateDir": 
                option = NVL(option,"Y");
                if (dirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя каталога (DirName)");
                file = new File(dirName);
                try {
                    // каталог создан
                    if (file.mkdir()) return file.getAbsolutePath(); // возврат полного имени каталога
                    // каталог не создан (уже существует)
                    else {
                        if (option.equals("Y")) return file.getAbsolutePath(); // возврат полного имени каталога
                        // сюда придём только в случае, когда следует сообщить, что каталог не создан
                        app.msgWriter.write("LUI-00125", dirName); // генерация сообщения об ошибке
                    }
                // ошибка ввода/вывода при создании каталога
                } catch (SecurityException e) {
                    app.msgWriter.write("LUI-00125", dirName, e.toString()); // генерация сообщения об ошибке. Ошибка создания каталога.
                }
                break;
            case "RenameDir": 
                if (srcDirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано исходное имя каталога (SrcDirName)");
                if (dstDirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано новое имя каталога (DstDirName)");
                File srcDir = new File(srcDirName);
                File dstDir = new File(dstDirName);
                if (srcDir.renameTo(dstDir)) return dstDir.getAbsolutePath(); // возврат полного нового имени каталога
                else app.msgWriter.write("LUI-00126", srcDirName); // не удалось переименовать каталог
                break;        
            case "DeleteDir": 
                option = NVL(option,"Y");
                if (dirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя каталога (DirName)");
                file = new File(dirName);
                if (file.delete() || option.equals("Y")) return file.getAbsolutePath(); // возврат полного имени файла
                else app.msgWriter.write("LUI-00127", dirName); // генерация сообщения об ошибке. Ошибка удаления файла.
                break;
            case "GetFiles": 
                option = NVL(option,"Y");
                if (dirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя каталога (DirName)");
                dir = new File(dirName);
                if (!dir.exists() && option.equals("N")) app.msgWriter.write("LUI-00124", dirName); // генерация сообщения об ошибке. Ошибка получения списка файлов в каталоге.
                if (dir.listFiles()!=null)
                for (File dfile : dir.listFiles()) {
                    if (dfile!=null)
                    res = res + dfile.getName() + ":~$";
                }
                break;
            case "DelFiles": 
                option = NVL(option,"Y");
                if (dirName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано имя каталога (DirName)");
                dir = new File(dirName);
                if (!dir.exists() && option.equals("N")) app.msgWriter.write("LUI-00124", dirName); // генерация сообщения об ошибке. Ошибка получения списка файлов в каталоге.
                if (dir.listFiles()!=null)
                for (File dfile : dir.listFiles()) {
                    if (dfile!=null)
                    if (!dfile.delete() && option.equals("N")) app.msgWriter.write("LUI-00123", fileName); // генерация сообщения об ошибке. Ошибка удаления файла.
                }
                res = dirName;
                break;
        }
        return res;
    }
    
    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
    
    public void clearFile (File file) throws IOException {
        Writer fstream = new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8); // конструктор с одним параметром - для перезаписи
//        FileWriter fstream = new FileWriter(file); // конструктор с одним параметром - для перезаписи
        try (BufferedWriter out = new BufferedWriter(fstream)) { //  создаём буферезированный поток
            out.write(""); // очищаем, перезаписав поверх пустую строку
            out.flush(); 
        } 
    }
    
    public void backupFile (File file) throws IOException {
        Date currentDate = new Date();
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddhhmmss");
        File target = new File(file.getPath() + "." + dateFormat.format(currentDate) + ".bak");
        Files.copy(file.toPath(), target.toPath());
    }

    public void CloseAll(){
        if (fileDescrMap!=null){
        for (String fk:fileDescrMap.keySet()){
          try{  fileDescrMap.get(fk).closeIOStream();} catch(Exception ex){}
          
        }
    }
    }
    
 // Класс-обертка для FileDescriptor для плагина LuiFile, хранящий в себе ссылку на файл, дескриптор файла и поток, из которого этот дескриптор получен.
    public class LuiFileDescriptor {
        FileDescriptor fileDescr = null;
        File file = null;
        BufferedReader reader = null; 
        BufferedWriter writer = null; 
        
        public LuiFileDescriptor (FileDescriptor pFileDescr, File pFile, BufferedReader pReader, BufferedWriter pWriter) {
          fileDescr = pFileDescr;
          file = pFile;
          reader = pReader;
          writer = pWriter;
        }
        
        public FileDescriptor getFD() {
            return fileDescr;
        }

        public File getFile() {
            return file;
        }

        public BufferedReader getReader() {
            return reader;
        }

        public BufferedWriter getWriter() {
            return writer;
        }
        
        public void closeIOStream() throws IOException {
            if (reader != null) reader.close();
            if (writer != null) writer.close();
        }
       
    }}
