/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package Standard;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import LuiExceptionMsg.LuiException;
import com.google.gson.JsonElement;
import java.sql.SQLException;
import luiserver.LuiAppElement;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;
//import com.google.gson.JsonElement;
import java.util.ArrayList;
/**
 * Вычисление семантики для элемента матрицы 
 * @author aliubushkin
 * MatrixSemantics:FormID.Element.Semanitic;[Descr:N];[Inherit:Y/L/A];
 * FormID может отсутствовать тогда
 * MatrixSemantic:Element.Semanitic  будет браться форма где вычисляется данное свойство
 * Элемент может тоже отсутствовать но только если форма тоже не указана. Тогда будет браться элемент, где это свойство вычисляется
 * Возвращает значение семантики
 * Если Descr:Y то возвращает коллекцию 
 * Value:Значение;Descr:Описание происхождения семантики;
 * Inherit:Y - не ищем в родном
 * Inherit:A - ищем в родном и старших
 * Inherit:L - ищем только в родном
 */
public class MatrixSemantics implements LuiPlugin {
 LuiApplication app;
 Integer shift=null;
    @Override
    public void setContext(LuiApplication pContext) {
      app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        shift=0;
        ArrayList<String> acell=null;
        LuiAppForm frm=null;
        LuiAppElement el=null;
        String fc=null;
        String ec=null;
        String sem=null;
        
        if (!like(pString,"%;"))pString=pString+";";
        int commai=pString.indexOf(";");
        String ms=pString.substring(0,commai);
        LuiCollection Col=new LuiCollection(pString.substring(commai));
        
        String Descr="N";
        String InhO="A";
        if (Col.Get_Element("Descr").equals("Y"))Descr="Y";
        if (Col.Get_Element("Inherit").equals("Y"))InhO="Y";
        if (Col.Get_Element("Inherit").equals("L"))InhO="L";
        String words[];
        words = ms.split("\\.",3);
        switch (words.length){
            
            case 3:
                fc=words[0];
                ec=words[1];
                sem=words[2];
                break;
            case 2:
                frm=pProp.ownerObj.getForm();
                if (frm!=null)fc=frm.code;
                else fc=app.getVal("FormID");
                ec=words[0];
                sem=words[1];
                break;
            case 1:
                frm=pProp.ownerObj.getForm();
                if (frm!=null)fc=frm.code;
                else fc=app.getVal("FormID");
                try{
                 el=(LuiAppElement)pProp.ownerObj;
                 ec=el.Code;
                }catch(Exception ex){};
                sem=words[0];
                break;
        }
        if (fc!=null && app.forms.containsKey(fc))frm=app.forms.get(fc);
        if (frm==null)return"";
/*                if (ec.equals("C254") && sem.equals("fractional")){
            String here="here";
        }*/
    if (sem==null)return"";
    if (sem.equals("god"))shift=0;
    /***статистика***/
    String stat="";
    try{
    stat=frm.getProperty("StatMatrixSemantics");
    }catch(Exception ex){}
    int StatCount=0;
    long StatTime=0;
    if (is_not_null(stat)){
        StatCount=Integer.valueOf(stat.split(" ")[0]);
        StatTime=Integer.valueOf(stat.split(" ")[1]);
    }
    StatCount++;
    long StatStart=System.currentTimeMillis();
    /***статистика***/
    try{
        
        if (ec!=null){
            if (frm.elements.containsKey(ec))el=frm.elements.get(ec);
            else {
                if (like(ec,"%_:_%")){
                   if (frm.MatrixMap.containsKey(ec)){
                      ec=frm.MatrixMap.get(ec).get(0);
                      if (frm.elements.containsKey(ec))el=frm.elements.get(ec);
                   } else {
                       String r=ec.split(":",2)[0];
                       if (frm.elements.containsKey(r)){
                           r=frm.elements.get(r).getProperty("ltn");
                       }
                       String c=ec.split(":",2)[1];
                       if (frm.elements.containsKey(c)){
                           c=frm.elements.get(c).getProperty("ltn");
                       }
                       ec=r+":"+c;
                       if (frm.MatrixMap.containsKey(ec)){
                       ec=frm.MatrixMap.get(ec).get(0);
                      if (frm.elements.containsKey(ec))el=frm.elements.get(ec);
                      }
                   }
                }
            }
        }
        String Description="";
        String Semantica=null;
        String Semantics="";
        if (el!=null) {
        String ps;
       if (el.getType().equals("CELL"))ps="CELL_SEMANTICS";
       else ps="ITEM_SEMANTICS"; 
       if (ps.equals("CELL_SEMANTICS") ) { //если это ячейка
              String r=el.getProperty("RowNo"); //номер строки
              String c=el.getProperty("ColNo"); // номер столбца
              if (frm.MatrixMap!=null && frm.MatrixMap.containsKey(r+":"+c))
              acell=frm.MatrixMap.get(r+":"+c);
              if (acell!=null && !acell.contains(el.Code))acell=null;
           }    
           
       if (!InhO.equals("Y")) {
          
           // поиск в родном элементе
           if (acell!=null) { //если это стопка ищем в ней
              Semantica=null;
              LuiAppElement  sel=null;
             for(int si=acell.size()-1;si>=0;si--){
               if (!frm.elements.containsKey(acell.get(si)))continue;
               sel=frm.elements.get(acell.get(si));
               Semantics=NVL(sel.getProperty(ps),"");
               //if(is_not_null(Semantics)){
               if (Semantics.indexOf("\""+sem+"\"")>=0){
               JsonElement JSONRoot= app.JSONparser.parse("{"+Semantics+"}");
               JsonElement JSONEl=null;
                try{
                JSONEl=JSONRoot.getAsJsonObject().get(sem);
                }catch (Exception ex){} 
                if (JSONEl!=null && !JSONEl.isJsonNull()) {
                  if (!JSONEl.getAsString().equals("NaN"))
                    Semantica=JSONEl.getAsString();
                }
                if (Semantica!=null) break;
               }
             }
             Semantica=checkShift(Semantica);
            if (Semantica!=null && sel!=null) { // нашли в стопке
                if (Descr.equals("Y")){
              LuiCollection col=new LuiCollection("");
              col.Add_Element("Value",Semantica);
              col.Add_Element("Descr", NVL(sel.getProperty("CAPTION"),sel.getProperty("Name")));
              return col.toString();
              }
              else return Semantica; 
            } // нашли в стопке
           } //стопка
           else {// не стопка
            Semantica=null;
            Semantics=NVL(el.getProperty(ps),"");
            //if(is_not_null(Semantics)){
            if (Semantics.indexOf("\""+sem+"\"")>=0){
            JsonElement JSONRoot= app.JSONparser.parse("{"+Semantics+"}");
            JsonElement JSONEl=null;
                try{
                JSONEl=JSONRoot.getAsJsonObject().get(sem);
                }catch (Exception ex){} 
                if (JSONEl!=null  && !JSONEl.isJsonNull()) {
                    if (!JSONEl.getAsString().equals("NaN"))
                    Semantica=JSONEl.getAsString();
                }
            // проверим на + и -
            Semantica=checkShift(Semantica);
            }
            if (Semantica!=null) {
                if (Descr.equals("Y")){
              LuiCollection col=new LuiCollection("");
              col.Add_Element("Value",Semantica);
              col.Add_Element("Descr", NVL(el.getProperty("CAPTION"),el.getProperty("Name")));
              return col.toString();
              }
              else return Semantica; 
            } 
       } //не cтопка
       
       }
            // в родном элементе не нашли
            
            // если сказано Inherit:L -- уходим
            if (InhO.equals("L"))return "";
            if (el.getType().equals("CELL")){
         /*    if (sem.equals("god")){
                  String here= "here";
              }*/
              String p=el.getProperty("COL");
              if (frm.elements.containsKey(p)){
              
              Semantica=findInParent(frm.elements.get(p),Descr,sem);
              if (Semantica!=null){
                return Semantica;  
              }
                  else{
                  p=el.getProperty("ROW");
                  if (frm.elements.containsKey(p)){
                     Semantica=findInParent(frm.elements.get(p),Descr,sem);
                     if (is_not_null(Semantica)) return Semantica;
                  }
              }
              }
            } else {
                String p=el.getParent();
                if (frm.elements.containsKey(p)){
                     Semantica=findInParent(frm.elements.get(p),Descr,sem);
                     if (is_not_null(Semantica)) return Semantica;
                  }
            }
    }         
            if (is_null(Semantica)){ // ищем в параметрах
              for (String pk:frm.elements.keySet()){
                  if (!frm.elements.get(pk).Code.equals(sem.toUpperCase()))continue;
                  if (!frm.elements.get(pk).getType().equals("PARAM"))continue;
                  if (!frm.elements.get(pk).getParent().equals("Form"))continue;
                  Semantica=frm.elements.get(pk).getProperty("Value");
                  if (is_not_null(Semantica) ){
                    if (shift!=0){
                        try{
                         shift=shift+Integer.valueOf(Semantica);
                         Semantica=shift.toString();
                        } catch(Exception ex){}
                    }
                      if (Descr.equals("Y")){
                        LuiCollection col=new LuiCollection("");
              col.Add_Element("Value",Semantica);
              col.Add_Element("Descr", NVL(frm.elements.get(pk).getProperty("CAPTION"),frm.elements.get(pk).getProperty("Name")));  
              return col.toString();
                      } else return Semantica;
                  }
              }  
            }
      if (Descr.equals("Y"))return "Value:;Descr:;";
      if (shift!=0)return String.valueOf(shift); //если сдвиг !=0 его и возвращаем
      return "";  
    }finally{
        frm.setProperty("StatMatrixSemantics",String.valueOf(StatCount)+" "+String.valueOf(StatTime+System.currentTimeMillis()-StatStart));   
    }
    }
    public String checkShift(String sem) {
        if (shift!=null){
            if (sem==null)return null;
            String s=sem.replace(" ","");
            if (is_null(s))return null;
            Integer inc=null;
            if (s.charAt(0)=='-' || s.charAt(0)=='+'){
            try{
                inc=Integer.valueOf(s);
                shift=shift+inc;
                return null;
            } catch (Exception ex){}
            } else {
                if (shift!=0){
                    try{
                    inc=Integer.valueOf(s);
                    inc=inc+shift;
                    return inc.toString();
                    }
                  catch (Exception ex){}
                }
            }
        }
        return sem;
    }
    public String findInParent(LuiAppElement el,String Descr,String sem) throws LuiException, SQLException{
       LuiAppElement tel=el;
       String ps;
       String Semantics="";
       if (el.getType().equals("CELL"))ps="CELL_SEMANTICS";
       else ps="ITEM_SEMANTICS";
       while (tel!=null){
            String Semantica=null;     
            Semantics=NVL(tel.getProperty(ps),"");
            //if (is_not_null(Semantics)){
            if (Semantics.indexOf("\""+sem+"\"")>=0){
            JsonElement JSONRoot= app.JSONparser.parse("{"+Semantics+"}");
            JsonElement JSONEl=null;
                try{
                JSONEl=JSONRoot.getAsJsonObject().get(sem);
                }catch (Exception ex){} 
                if (JSONEl!=null && !JSONEl.isJsonNull()) {
                    if (!JSONEl.getAsString().equals("NaN"))
                    Semantica=JSONEl.getAsString();
                }
                
            Semantica=checkShift(Semantica);
            }
            if (Semantica!=null){
              if (Descr.equals("Y")){
              LuiCollection col=new LuiCollection("");
              col.Add_Element("Value",Semantica);
              String pr="";
              if (tel.getForm().application.code.equals("MVD_VPS_IAM")){
                  if (tel.getType().equals("COL")) pr="Графа ";
                  if (tel.getType().equals("ROW")) pr="Строка ";
              }
              col.Add_Element("Descr",pr+NVL(tel.getProperty("CAPTION"),tel.getProperty("Name")));
              return col.toString();
              }
              else return Semantica;
                
            } else {
                String p=tel.getParent();
                if (tel.getForm().elements.containsKey(p)) {
                   tel= tel.getForm().elements.get(p);
                } else tel=null;
            }
       }
       return null;
    }
    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {

    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {

    }
    
}
