/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 */
package Standard;

import LuiCollection.LuiCollection;
import LuiDbInterface.LuiDbConn;

import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.ArrayList;
import luiserver.LuiAppConn;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;

/**
 *
 * @author achibuk
 */


public class Ora2PgExternalPackages implements LuiPlugin{
    LuiApplication app;
//    String ls;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
//        ls=System.getProperty("line.separator");
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       return null;
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        LuiAppConn oraAppConn=null, pgAppConn=null;
        LuiDbConn oraDbConn=null, pgDbConn=null;
        Integer oraDTypeCursor=null, oraDTypeUsageCursor=null, oraProcCursor=null, oraArgCursor=null;
//        Integer pgObjCursor=null;
        
        LuiCollection mainCollection = new LuiCollection("Project:"+pString);
        String projectCode = mainCollection.Get_Element("Project");         // код проекта миграции
        String oraConnCode = mainCollection.Get_Element("OraConnCode");     // код используемого ораклового соединения
        String pgConnCode = mainCollection.Get_Element("PgConnCode");       // код используемого постресового соединения
        int acSessionID = Integer.parseInt(mainCollection.Get_Element("ACSessionID"));          // ID сеанса миграции
        int projectObjectID = Integer.parseInt(mainCollection.Get_Element("ProjectObjectID"));  // ID объекта миграции в проекте
        String pkgName = mainCollection.Get_Element("PkgName");             // наименование внешнего ораклового пакета
        
        if (projectCode.length() == 0) app.msgWriter.write("LUI-00128", "Не указан код проекта миграции");
        if (oraConnCode.length() == 0) app.msgWriter.write("LUI-00128", "Не указан код соединения Oracle");
        if (pgConnCode.length() == 0) app.msgWriter.write("LUI-00128", "Не указан код соединения Postgres");
        if (Integer.toString(acSessionID).length() == 0) app.msgWriter.write("LUI-00128", "Не указан ID сеанса миграции");
        if (Integer.toString(projectObjectID).length() == 0) app.msgWriter.write("LUI-00128", "Не указан ID объекта миграции в проекте");
        if (pkgName.length() == 0) app.msgWriter.write("LUI-00128", "Не указано наименование внешнего ораклового пакета");

        ArrayList<PgObjectScript> pgObjectScripts = new ArrayList<>();
        try {
            // Проверим, вдруг данные об объектах, соответствующих содержимому пакета pkgName, уже добавлены для сессии acSessionID
            pgAppConn = app.userDbConnections.get(pgConnCode);
            pgDbConn = getDbConn(pgAppConn);
            
            String objCnt = pgDbConn.getSelectResult(
                "select count(*) from o2p_t_object " + 
                 "where acsession4o_id = 0" + acSessionID + " and postgres_schema = upper('" + pkgName + "')");
            if (objCnt.compareTo("0") != 0) {
                // Если да, то ссылки на эти объекты проставим для указанного projectObjectID
                pgDbConn.executeDML(
                    "do $$\n" +    
                    "declare obj record;\n" +
                    "begin\n" + 
                      "for obj in (select id from o2p_t_object o " +
                                   "where o.acsession4o_id = 0" + acSessionID + " and postgres_schema = upper('" + pkgName + "') " +
                                     "and not exists (select 1 from o2p_t_object4projectobject where po4opo_id = 0" +  projectObjectID + " and o4opo_id = o.id) " +       
                                   "order by id)\n" +
                      "loop\n" +
                        "insert into o2p_t_object4projectobject (id, po4opo_id, o4opo_id)\n" +
                          "values (nextval('o2p_q_object4projectobject'), 0" +  projectObjectID + ", obj.id);\n" +      
                      "end loop;\n" +
                    "end;\n" +
                    "$$");
                    if (pgAppConn.getSessionMode() == 'P') pgDbConn.commit();
            } 
            else 
            {
                // Если не добавлены, то подготовим их и затем уже добавим
                try {
                    oraAppConn = app.userDbConnections.get(oraConnCode);
                    oraDbConn = getDbConn(oraAppConn);
                    String objectScript;
            
                    // 1. Скрипт для создания схемы БД в PosgreSQL, наименование которой совпадает с наименованием пакета pkgName
                    pgObjectScripts.add(pgObjectScripts.size(), new PgObjectScript("schema", pkgName, pkgName, 
                        "drop schema if exists " + pkgName + " cascade;\n" +
                        "create schema " + pkgName + ";\n"));
            
                    // 2. Скрипты для создания типов данных, определённых в пакете.
                    // (Пока не рассматриваются типы данных, определённые как TABLE или VARRAY)
                    oraDTypeCursor = oraDbConn.CursorOpen(
                        "select distinct type_owner, type_name, type_subname " +
                          "from all_arguments " +
                         "where package_name = upper('" + pkgName + "') and data_level = 0 and type_subname is not null " +
                           "and data_type not like '%TABLE%' " +
                           "and data_type not like '%VARRAY%'");
                    ArrayList<String> oraDTypeRow;
                    while (oraDbConn.CursorNextRow(oraDTypeCursor)) {
                        oraDTypeRow = oraDbConn.CursorGetRow(oraDTypeCursor);
                        if (oraDTypeRow != null) {
                            oraDTypeUsageCursor = oraDbConn.CursorOpen(
                                "select type_subname, data_level, sequence, argument_name, pls_type from all_arguments " +
                                 "where package_name = upper('" + pkgName + "') " +
                                   "and subprogram_id in " +
                                     "(select min(subprogram_id) from all_arguments " +
                                       "where package_name = upper('" + pkgName + "') " +
                                         "and type_owner = '" + oraDTypeRow.get(0) + "' " +
                                         "and type_name = '" + oraDTypeRow.get(1) + "' " +
                                         "and type_subname = '" + oraDTypeRow.get(2) + "') " +
                                 "order by sequence");
                            ArrayList<String> oraDTypeUsageRow;
                            int oraDTypeUsageSeq = 0;
                            objectScript = "";
                            while (oraDbConn.CursorNextRow(oraDTypeUsageCursor)) {
                                oraDTypeUsageRow = oraDbConn.CursorGetRow(oraDTypeUsageCursor);
                                if (oraDTypeUsageRow != null) {
                                    if (oraDTypeUsageRow.get(0) != null && oraDTypeRow.get(2) != null &&
                                        oraDTypeUsageRow.get(0).compareTo(oraDTypeRow.get(2)) == 0 && oraDTypeUsageRow.get(1).compareTo("0") == 0) 
                                    {
                                        oraDTypeUsageSeq = Integer.valueOf(oraDTypeUsageRow.get(2));
                                        continue;
                                    }
                                    if (oraDTypeUsageSeq > 0) {
                                        if (oraDTypeUsageRow.get(1).compareTo("1") == 0) {
                                            if (objectScript.length() > 0) objectScript = objectScript + ", ";
                                            objectScript = objectScript + oraDTypeUsageRow.get(3) + " " + getPgType4OraType(oraDTypeUsageRow.get(4));
                                        } else {
                                            break;
                                        }
                                    }
                                }
                            }
                            if (objectScript.length() > 0) {
                                objectScript = "create type " + pkgName + "." + oraDTypeRow.get(2) + " as (" + objectScript + ");\n";
                                pgObjectScripts.add(pgObjectScripts.size(), new PgObjectScript("type", pkgName, oraDTypeRow.get(2), objectScript));
                            }
                        }
                    }

                    // 3. Скрипты для создания функций, определённых в пакете
                    // (Пока не рассматриваются не функции/процедуры, в числе атрибутов которых есть атрибуты, определённые как TABLE или VARRAY, 
                    // а также атрибуты, тип которых отличается от стандартного и не определён внутри пакета pkgName)
                    String dType = "";
                    String dTypeReturned = "";
                    oraProcCursor = oraDbConn.CursorOpen(
                        "select owner, object_name, procedure_name, subprogram_id " + 
                          "from all_procedures p " + 
                         "where object_name = upper('" + pkgName + "') and object_type = 'PACKAGE' and procedure_name is not null " + 
                           "and not exists (select 1 from all_arguments " + 
                                            "where owner = p.owner " + 
                                              "and package_name = p.object_name " +
                                              "and object_name = p.procedure_name " +
                                              "and (data_type in ('TABLE','VARRAY') or " + 
                                                   "data_type = 'OBJECT' and (argument_name is null or type_name <> '" + pkgName + "' )))" +
                         "order by subprogram_id");
                    ArrayList<String> oraProcRow;
                    while (oraDbConn.CursorNextRow(oraProcCursor)) {
                        oraProcRow = oraDbConn.CursorGetRow(oraProcCursor);
                        if (oraProcRow != null) {
                            oraArgCursor = oraDbConn.CursorOpen(
                                "select position, argument_name, replace(in_out,'/') in_out, " +
                                        "case when data_type like '%XMLTYPE%' then 'XMLTYPE' else pls_type end pls_type, " +
                                        "type_subname, defaulted " +
                                  "from all_arguments " + 
                                 "where owner = '" + oraProcRow.get(0) + "' " + 
                                   "and package_name = '" + oraProcRow.get(1) + "' " +
                                   "and object_name = '" + oraProcRow.get(2) + "' " +
                                   "and data_level = 0 " + 
                                   "and subprogram_id = '" + oraProcRow.get(3) + "' " +
                                "order by sequence");
                            ArrayList<String> oraArgRow;
                            boolean haveOutParam = false;
                            objectScript = "(";
                            dTypeReturned = "";
                            while (oraDbConn.CursorNextRow(oraArgCursor)) {
                                oraArgRow = oraDbConn.CursorGetRow(oraArgCursor);
                                if (oraArgRow != null) {
                                    if (oraArgRow.get(0).compareTo("0") == 0) {
                                        dTypeReturned = (oraArgRow.get(3)!=null && oraArgRow.get(3).length()>0)?getPgType4OraType(oraArgRow.get(3)):pkgName+"."+oraArgRow.get(4);
                                        continue;
                                    }
                                    if (oraArgRow.get(2).contains("OUT")) haveOutParam = true; // признак того, что у функции есть хотя бы один out-параметр
                                    if (oraArgRow.get(0).compareTo("1") == 0); // Теперь всегда есть открывающая скобка для параметров функции, даже если никаких параметров у функции нет // objectScript = objectScript + "(";
                                    else objectScript = objectScript + ", ";
                                    // Если задан тип данных в поле pls_type (oraArgRow.get(3)), то используем его, если нет, то используем тип данных, указанный в поле type_subname (oraArgRow.get(4))
                                    dType = (oraArgRow.get(3)!=null && oraArgRow.get(3).length()>0)?getPgType4OraType(oraArgRow.get(3)):pkgName+"."+oraArgRow.get(4);
                                    objectScript = objectScript + oraArgRow.get(1) + " " + oraArgRow.get(2) + " " + dType;
                                    // Если для параметра функции должно быть указано значение по умолчанию (all_arguments.defaulted = 'Y')...
                                    if (oraArgRow.get(5).compareTo("Y") == 0) {
                                        // Баг оракла с 1996 года. Значение поля all_arguments.default_value = всегда null. 
                                        // В данном случае это не имеет значения, т.к. всё равно создаётся функция пустышка, которую при желании можно отредактировать.
                                        objectScript = objectScript + " = null::" + dType;
                                    }
                                }
                            }
                            if (objectScript.length() > 0) objectScript = objectScript + ")";
                            
                            if (haveOutParam) ; // если есть out-параметр, то ничего возвращать нельзя (функция в этом случае сама вернёт record на основании out-параметров)
                            else if (dTypeReturned.length() > 0) objectScript = objectScript + " returns " + dTypeReturned;
                            else objectScript = objectScript + " returns void";
                            
                            objectScript = objectScript + " as $body$\n";
                            objectScript = objectScript + "begin\n";
                            if (dTypeReturned.length() > 0) objectScript = objectScript + " return null::" + dTypeReturned + ";\n";
                            else objectScript = objectScript + " return;\n";
                            objectScript = objectScript + "end;\n$body$\nlanguage plpgsql;\n";
                            objectScript = "create or replace function " + pkgName + "." + oraProcRow.get(2) + " " + objectScript;
                            pgObjectScripts.add(pgObjectScripts.size(), new PgObjectScript("function", pkgName, oraProcRow.get(2), objectScript));
                        }
                    }
                } finally {
                    if (oraDTypeCursor!=null) oraDbConn.CursorClose(oraDTypeCursor);
                    if (oraDTypeUsageCursor!=null) oraDbConn.CursorClose(oraDTypeUsageCursor);
                    if (oraProcCursor!=null) oraDbConn.CursorClose(oraProcCursor);
                    if (oraArgCursor!=null) oraDbConn.CursorClose(oraArgCursor);
                    if (oraAppConn != null && oraDbConn != null && oraAppConn.getSessionMode() == 'C') { // если соединение коммунальное, то вернём соединение в пул свободных
                        oraAppConn.connPool.SetFree(oraDbConn);
                    }
                }

                // 4. Добавление в соответствующие таблицы (o2p) подготовленных скриптов, сохранённых в pgObjectScripts
                for (int i = 0; i < pgObjectScripts.size(); i++) {
                    pgDbConn.executeDML(
                        "do $$\n" +    
                        "declare objid o2p_d_id;\n" +
                        "begin\n" +
                          "objid := nextval('o2p_q_object');\n" +      
                          "insert into o2p_t_object (id, num, name, postgres_schema, postgres_script, ostatus4o_code, acsession4o_id, pgtype4o_code)\n" +
                            "values (objid, objid, upper('" + pgObjectScripts.get(i).objectName + 
                              "'), upper('" + pgObjectScripts.get(i).objectSchema + "'), '" + pgObjectScripts.get(i).objectScript + 
                              "', 'AUTO', 0" + acSessionID + 
                              ", upper('" + pgObjectScripts.get(i).objectType + "'));\n" +
                          "insert into o2p_t_object4projectobject (id, po4opo_id, o4opo_id)\n" +
                            "values (nextval('o2p_q_object4projectobject'), 0" +  projectObjectID + ", objid);\n" +      
                        "end;\n" +
                        "$$");
                    if (pgAppConn.getSessionMode() == 'P') pgDbConn.commit();
                }
            }   
        } finally {
//            if (pgObjCursor!=null) pgDbConn.CursorClose(pgObjCursor);
            if (pgAppConn != null && pgDbConn != null && pgAppConn.getSessionMode() == 'C') { // если соединение коммунальное, то вернём соединение в пул свободных
                pgAppConn.connPool.SetFree(pgDbConn);
            }
        } 
    }
    
    public LuiDbConn getDbConn(LuiAppConn luiAppConn) throws SQLException, LuiException {
        LuiDbConn luiDbConn;
        if (luiAppConn.getSessionMode() == 'C') { // если соединение коммунальное, то получим из пула свободное
            luiDbConn = luiAppConn.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
        } else {//в противном случае возьмём текущее 
            luiDbConn = luiAppConn.currentConnection;
        }
        if (luiDbConn == null){ // если соединение так и не нашлось, то
            app.msgWriter.write("LUI-00020",". Lui:Null pointer detected!");
        }
        return luiDbConn;
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        
    }
    
    public String getPgType4OraType(String oraType) {
        if (oraType == null || oraType.length() == 0) return oraType;
        switch (oraType.toUpperCase()) {
            case "CHAR": 
            case "BOOLEAN": return oraType;
            case "VARCHAR2": return "VARCHAR";
            case "TIMESTAMP":
            case "NUMBER": return "NUMERIC";
            case "INTEGER":
            case "BINARY_INTEGER": return "INTEGER";
            case "CLOB": return "TEXT";
            case "BLOB":
            case "LONG RAW":
            case "RAW": return "BYTEA";
            case "XMLTYPE": return "XML";
            case "UROWID": return "OID";
        }
        return "TEXT";  
    }
    
    // Класс для хранения данных о типе объекта, схеме БД, наименовании объекта и скрипте для создания объекта в БД PostgreSQL.
    public class PgObjectScript {
        public String objectType, objectSchema, objectName, objectScript;
        public PgObjectScript(String objectType, String objectSchema, String objectName, String objectScript) {
            this.objectType = objectType;
            this.objectSchema = objectSchema;
            this.objectName = objectName;
            this.objectScript = objectScript;
        }        
    }
    
}
