/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package Standard;

import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.PreparedStatement;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import luiserver.LuiAppConn;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiParser;
import luiserver.LuiPlugin;

/**
 * Парсинг SQL без выполнения
 * Execute ни чего не делает
 * Compute возвращает ошибку если она есть
 * Синтаксис: {ParseSQL:(код соединения)SQL-оператор}
 * @author aliubushkin
 */
public class ParseSQL implements LuiPlugin{
 LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
    String retval="";    
    LuiAppConn ac=null;
    LuiDbConn lc=null;
    String stmt;
    PreparedStatement ps=null;
    ResultSetMetaData resSetMetaData=null;
    // выясняем код формы
    LuiAppForm frm=pProp.ownerObj.getForm();
    if (frm==null) {
    String curFrmID=app.getVal("FormID");        
    if (app.forms.containsKey(curFrmID)){
       frm=app.forms.get(curFrmID);
    }
    }
    if (frm==null) return "";
      ArrayList<String> acs=app.getAppConnAndStmt(pString,"",frm.getFormID());
        
        ac=app.userDbConnections.get(acs.get(0));
        stmt=LuiParser.reMask(acs.get(1));
        if (ac.getSessionMode()=='C') { // если соединение коммунальное, то получим из пула свободное
            lc=ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
        }
        else {//в противном случае возьмём текущее 
            lc=ac.currentConnection;
        }
        if (lc==null){ // если соединение так и не нашлось, то
        app.msgWriter.write("LUI-00020",". Lui:Null pointer detected!");
        return "";
        }
        if (lc.getDBvendor().equals("PG"))stmt="explain "+stmt;
    try {
        ps=lc.getConnection().prepareStatement(stmt);
        if (lc.getDBvendor().equals("PG")) {
        ps.execute();
        } else {
        resSetMetaData= ps.getMetaData();
        }
        retval="";
    } catch (SQLException ex){
      retval=ex.getLocalizedMessage();
      if (ac.getSessionMode()!='C') lc.rollback();
    } finally {
        if (ps!=null) ps.close();
        //if (lc!=null){
            if (ac.getSessionMode()=='C') { // если соединение коммунальное, то вернём соединение в пул свободных
                ac.connPool.SetFree(lc);
            }
       // }
    }
    return retval;
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        compute(pString, pProp);
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       
    }
    
}
