/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * вычисление значение свойства элемента или его старшего элемента
 * а если не найдено, то формы или приложения
 * Ситаксис
Property:<код свойства>
 */
package Standard;

import static LuiCollection.LuiString.Extract_Word;
import static LuiCollection.LuiString.LTrim;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.RTrim;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiAppWFE;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;

/**
 *
 * @author aliubushkin
 */
public class Property implements LuiPlugin {
LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    /**
     * вычисление значение свойства элемента или его старшего элемента
     * а если не найдено, то формы или приложения
     * @param pString программа
     * @param pProp свойство
     * @return значение
     * @throws LuiExceptionMsg.LuiException логическая ошибка
     * @throws java.sql.SQLException ошибка выполнения SQL-запроса
     */
    @Override
    public  String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
      
        return comp(pString,pProp,false);
    }
    /**
     * 
     * @param pString
     * @param pProp
     * @param pFindOnly
     * @return
     * @throws LuiException
     * @throws SQLException 
     */
    private String comp(String pString, LuiAppProp pProp, boolean pFindOnly) throws LuiException, SQLException {
      //System.out.print(pString+"  "+pProp.getGenesis()+" "+pFindOnly+""+app.currentSCN);
      // вычислим форму
              LuiAppForm frm=pProp.ownerObj.getForm();
              if (frm==null) {
              String curFrmID=app.getVal("FormID");        
              if (app.forms.containsKey(curFrmID)){
              frm=app.forms.get(curFrmID);
              } else {
        //          System.out.println("Форма не нашлась!");
                  return null;
              }
              }
      LuiAppWFE ownerObj;
      boolean B=true;
      ownerObj=pProp.ownerObj; // берём хозяина свойства которое вычисляем
      while (B){
      B=false;
      if (ownerObj.properties.containsKey(pString)) {//если есть такое свойство
          ownerObj.properties.get(pString).setDepend(pProp); // то говорим, что оно определяет свойство pProp
          if (pFindOnly) return null;
          String retVal=NVL(ownerObj.properties.get(pString).getComputedProperty(),"");
        //  System.out.println(" return:"+retVal);
          return retVal;
      }
      
      // в этой точке обнаружили, что среди свойств хозяина нет свойств с кодом pString
      // тогда определим имеется ли свойство с кодом Parent
      if (ownerObj.properties.containsKey("Parent")) {
          String parent=ownerObj.getProperty("Parent");
          if (parent.equals("Application")){
            ownerObj=ownerObj.ownerObj;
            B=true;
          } else{
          if (parent.equals("Form")){ // если родитель это форма, то
              ownerObj=ownerObj.ownerObj;
              B=true;
          } else { // если родитель это не форма, а другой элемент формы и теперь надо поковыряться в свойствах уже этого элемента
              
              //String curFrmID=app.getVal("FormID");
              if (frm.elements.containsKey(parent)){ //если есть такой элемент
                  ownerObj=frm.elements.get(parent);
                  B=true;
              }
          }}
      }
      }
      //System.out.println(" return:NULL");
      return null; // так и не нашли
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        // выясняем код свойства
        pString=LTrim(pString);
        int eq=pString.indexOf("=");
        String pc=pString.substring(0, eq);
        // выясняем значение
        String v=pString.substring(eq+1);
        
    LuiAppWFE ownerObj;
      boolean B=true;
      ownerObj=pProp.ownerObj; // берём хозяина свойства которое вычисляем
      while (B){
      B=false;
      if (ownerObj.properties.containsKey(pc)) {//если есть такое свойство
          ownerObj.addProperty(pc,NVL(v,"")); // записываем в него значение
          return;
      }
      
      // в этой точке обнаружили, что среди свойств хозяина нет свойств с кодом pString
      // тогда определим имеется ли свойство с кодом Parent
      if (ownerObj.properties.containsKey("Parent")) {
          String parent=ownerObj.getProperty("Parent");
          if (parent.equals("Application")){
            ownerObj=ownerObj.ownerObj;
            B=true;
          } else{
          if (parent.equals("Form")){ // если родитель это форма, то
              ownerObj=ownerObj.ownerObj;
              B=true;
          } else { // если родитель это не форма, а другой элемент формы и теперь надо поковыряться в свойствах уже этого элемента
              String curFrmID=app.getVal("FormID");
              if (app.forms.get(curFrmID).elements.containsKey(parent)){ //если есть такой элемент
                  ownerObj=app.forms.get(curFrmID).elements.get(parent);
                  B=true;
              }
          }}
      }
      }
    }
     
    

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
         comp(pString,pProp,true);
    }
    
}
