/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Запись в файл содержимого свойства Элемента или формы
* Compute отсутствует
* Синтаксис Execute:
Property2File:Имя свойства;Filename:Имя файла;Mode:[W]/A;LineSeparator:null/OS/CRLF/CR/LF;
CR=/r
LF=/n
Имя свойства задаётся в формате:
  <ID формы>.<Код элемента или "Form">.<Код свойства>
Пример:
Property2File:{Property:FormID}.Field1.Value;File:my_file.txt;
Property2File:{Property:FormID}.Field2.Value;File:my_file.txt;Mode:A;
 */
package Standard;

import LuiCollection.LuiCollection;
import LuiCollection.LuiString;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_null;

import LuiExceptionMsg.LuiException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.sql.SQLException;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiParser;
import luiserver.LuiPlugin;

/**
 *
 * @author aliubushkin
 */
public class Property2File implements LuiPlugin{
LuiApplication app;
String ls;
public static final String RC="‡";
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
        ls=System.getProperty("line.separator");
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       return null;
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
       LuiCollection pcol=new LuiCollection("Property2File:"+pString);
       FileOutputStream fileOutputStream=null;
       Writer filew=null;
       String propPath=pcol.Get_Element("Property2File");
       String FrmID=LuiString.Extract_Word(propPath,1,".");
       String v2f="";
       LuiAppProp p=app.findProperty(propPath);
       
       if (p!=null) v2f=p.getComputedProperty();
       v2f=LuiParser.reMask(v2f);
       String filename=pcol.get("Filename");
       if (is_null(filename)) return; 
       if (filename.indexOf("/")!=-1 || filename.indexOf("\\")!=-1){
           // имя файла указано с путём
       } else {
       String tmpdir=app.getTmpDir();
       filename=tmpdir+filename;
       }
    try {
        boolean append=false;
        String mode=pcol.get("Mode");
        if (NVL(mode,"W").equals("A")) append=true;
        File file = new File(filename);
        file.createNewFile(); // если файл существует - команда игнорируется
        fileOutputStream = new FileOutputStream(file, append);
        filew = new OutputStreamWriter(fileOutputStream, StandardCharsets.UTF_8);
//        FileWriter filew=new FileWriter(filename,append);
        String LineSep=NVL(pcol.Get_Element("LineSeparator"),"OS");
        
        switch (LineSep){
            case "null":
            break;
            case "OS":
            v2f=v2f.replace("\n",ls);
            break;
            case "LF":
            v2f=v2f.replace("\r\n","\n");
            v2f=v2f.replace("\r","\n");
            break;
            case "CR":
            v2f=v2f.replace("\r\n","\r");
            v2f=v2f.replace("\n","\r");                
            break;
            case "CRLF":
            v2f=v2f.replace("\r\n",RC);
            v2f=v2f.replace("\r",RC);
            v2f=v2f.replace("\n",RC);
            v2f=v2f.replace("RC","\r\n");
            break;
        }
        filew.write(v2f);
        
    } catch (IOException ex) {
        app.msgWriter.write("LUI-00097",filename,ex.getMessage());
    } finally{
        if (filew!=null)try{filew.close();}catch(Exception ex){}
        if (fileOutputStream!=null)try{fileOutputStream.close();}catch(Exception ex){}
        }
    }
     
       
    

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        
    }
    
}
