/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Сохранение пользовательской конфигурации
 * SaveConfig:FormID;NAME:....;DEFAULT:Y/N;MENU:Y/N;
 */
package Standard;

import LuiCollection.LuiCollection;

import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.HashMap;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;
import static LuiCollection.LuiString.NVL;

/**
 *
 * @author aliubushkin
 */
public class SaveConfig implements LuiPlugin {
private static final String STMT4INSERTCONF="INSERT INTO lui_t_user_config(id,parent_app,parent_form,parent_config,name,order_by,menu_appr,user_name)\n" +
"VALUES (<id>,'<parent_app>','<parent_form>','<parent_config>','<name>','<order_by>','<menu_appr>','<user_name>');";
private static final String STMT4SELECTCONF="select max(ID) from lui_t_user_config c where c.parent_app='<parent_app>' and c.parent_form='<parent_form>' \n" +
"and c.parent_config='<parent_config>' and c.name='<name>' and c.user_name='<user_name>'";
private static final String STMT4UPDATECONF="update lui_t_user_config c SET\n" +
"menu_appr='<menu_appr>',\n" +
"name='<name>',\n" +
"order_by='<order_by>'\n" +
"where c.id=<id>";
private static final String STMT4DELCOND="delete from lui_t_user_condition c where c.usrcfg_id=<id>";
private static final String STMT4INSERTCOND="INSERT INTO lui_t_user_condition(row_number,column_code,condition,usrcfg_id)\n" +
"VALUES (<row_number>,'<column_code>','<condition>',<id>)";
private static final String STMT4DELPREF1="delete from lui_t_user_pref p\n" +
"where p.app_code='<app_code>' and\n" +
"p.form_code='<form_code>' and\n" +
"p.config_code='<config_code>' and\n" +
"p.user_name='<user_name>' and\n" +
"p.usrcfg_id=<id>";
private static final String STMT4INSPREF1="INSERT INTO lui_t_user_pref\n" +
"(\n" +
"  value,\n" +
"  user_name,\n" +
"  element_code,\n" +
"  form_code,\n" +
"  property_code,\n" +
"  config_code,\n" +
"  app_code,\n" +
"  work_place,\n" +
"  usrcfg_id\n" +
")\n" +
"select \n" +
"  value,\n" +
"  user_name,\n" +
"  element_code,\n" +
"  form_code,\n" +
"  property_code,\n" +
"  config_code,\n" +
"  app_code,\n" +
"  work_place,\n" +
"  <id>\n" +
"  from lui_t_user_pref p where \n" +
"  p.app_code='<app_code>' and\n" +
"  p.form_code='<form_code>' and\n" +
"  p.config_code='<config_code>' and\n" +
"  p.user_name='<user_name>' and\n" +        
"  coalesce(p.usrcfg_id,0)=<old_id>";
private static final String STMT4DELPREF2="delete from lui_t_user_pref p\n" +
"where p.app_code='<app_code>' and\n" +
"p.form_code='<form_code>' and\n" +
"p.config_code='<config_code>' and\n" +
"p.user_name='<user_name>' and\n" +
"p.element_code is null and\n" +
"p.property_code='DEF_USER_CFG'";
private static final String STMT4INSPREF2="INSERT INTO \n" +
" lui_t_user_pref\n" +
"(\n" +
"  value,\n" +
"  user_name,\n" +
"  form_code,\n" +
"  property_code,\n" +
"  config_code,\n" +
"  app_code\n" +
")\n" +
"VALUES (\n" +
"  '<id>',\n" +
"  '<user_name>',\n" +
"  '<form_code>',\n" +
"  'DEF_USER_CFG',\n" +
"  '<config_code>',\n" +
"  '<app_code>'\n" +
")";
private static final String STMT4DELPREF3="delete from lui_t_user_pref p\n" +
"where p.app_code='<app_code>' and\n" +
"p.form_code='<form_code>' and\n" +
"p.config_code='<config_code>' and\n" +
"p.user_name='<user_name>' and\n" +
"p.element_code is null and\n" +
"p.value='<id>' and\n" +
"p.property_code='DEF_USER_CFG'";

private static final String STMT4DELPREF4="delete from lui_t_user_pref p\n" +
"where p.app_code='<app_code>' and\n" +
"p.form_code='<form_code>' and\n" +
"p.config_code='<config_code>' and\n" +
"p.user_name='<user_name>' and\n" +
"p.element_code is null and\n" +
"p.property_code='<prop_code>' and\n" +
"p.usrcfg_id=<id>";

private static final String STMT4INSPREF4="INSERT INTO \n" +
" lui_t_user_pref\n" +
"(\n" +
"  value,\n" +
"  user_name,\n" +
"  form_code,\n" +
"  property_code,\n" +
"  config_code,\n" +
"  app_code,\n" +
"  usrcfg_id\n"+        
")\n" +
"VALUES (\n" +
"  '<value>',\n" +
"  '<user_name>',\n" +
"  '<form_code>',\n" +
"  '<prop_code>',\n" +
"  '<config_code>',\n" +
"  '<app_code>',\n" +
"  '<id>'\n" +        
")";
    LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        return null;
    }

    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        String id=null;
        LuiCollection coll=new LuiCollection("SAVE_CONFIG:"+pString);
        String FormID=coll.Get_Element("SAVE_CONFIG");
        String user_name=app.getAppUser();
        if (is_null(user_name)) return;
        if (!app.forms.containsKey(FormID)) return;  
        String menu_appr=coll.Get_Element("MENU");
        String is_default=coll.Get_Element("DEFAULT");
        String name=coll.Get_Element("NAME");
        if (is_null(name)) return;
        LuiAppForm frm=app.forms.get(FormID);
        String parent_form=frm.getProperty("FEATURE");
        String parent_app=app.code;
        //String parent_app=app.getProperty("AppCode");
        String parent_config=frm.getConfigCode();
        LuiCollection order_by= new LuiCollection("");
        
        frm.setVal("UserConfigurationShow","Y");
        for (int i=0;i<frm.el4order.size();i++){
          order_by.Add_Element(frm.el4order.get(i).getCode(),frm.el4order.get(i).getProperty("SORT"));
        }
        LuiDbConn lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
        String stmt;
        try {
            stmt=STMT4SELECTCONF.replace("<parent_app>",parent_app);
            stmt=stmt.replace("<parent_form>",parent_form);
            stmt=stmt.replace("<parent_config>",parent_config);
            stmt=stmt.replace("<user_name>",user_name);
            stmt=stmt.replace("<name>",name);
            id=lc.getSelectResult(stmt);
            if (is_null(id)){
            id=lc.NextVal("lui_q_user_config");
            stmt=STMT4INSERTCONF.replace("<parent_app>",parent_app);
            stmt=stmt.replace("<parent_form>",parent_form);
            stmt=stmt.replace("<parent_config>",parent_config);
            stmt=stmt.replace("<user_name>",user_name);
            stmt=stmt.replace("<id>", id);
            stmt=stmt.replace("<name>",name);
            stmt=stmt.replace("<order_by>", order_by.toString());
            stmt=stmt.replace("<menu_appr>",menu_appr);
            lc.executeDML(stmt);
            } else {
            stmt=STMT4UPDATECONF.replace("<id>",id);
            stmt=stmt.replace("<name>",name);
            stmt=stmt.replace("<order_by>", order_by.toString());
            stmt=stmt.replace("<menu_appr>",menu_appr);
            lc.executeDML(stmt);
            stmt=STMT4DELCOND.replace("<id>", id);
            lc.executeDML(stmt);
            }
         HashMap<String,HashMap<String,String>>listQBE=frm.listQBE;
         for (String row_number:listQBE.keySet()){
             HashMap<String,String> rowQBE=listQBE.get(row_number);
            for (String column_code:rowQBE.keySet()){
                String condition=rowQBE.get(column_code);
                if (is_not_null(condition)){
                 stmt=STMT4INSERTCOND.replace("<row_number>",row_number);
                 stmt=stmt.replace("<column_code>", column_code);
                 stmt=stmt.replace("<condition>",condition);
                 stmt=stmt.replace("<id>", id);
                 lc.executeDML(stmt);
                }
            } 
         }
         String old_id=NVL(frm.getVal("UserConfiguration"),"0");
         if (!old_id.equals(id)){
         stmt=STMT4DELPREF1.replace("<app_code>", parent_app);
         stmt=stmt.replace("<form_code>",parent_form);
         stmt=stmt.replace("<config_code>",parent_config);
         stmt=stmt.replace("<id>",id);
         stmt=stmt.replace("<user_name>",user_name);
         lc.executeDML(stmt);
         stmt=STMT4INSPREF1.replace("<app_code>",parent_app);
         stmt=stmt.replace("<form_code>",parent_form);
         stmt=stmt.replace("<config_code>",parent_config);
         stmt=stmt.replace("<id>",id);
         stmt=stmt.replace("<user_name>",user_name);
         stmt=stmt.replace("<old_id>",old_id);
         lc.executeDML(stmt);
         }
         if (is_default.equals("Y")){
         stmt=STMT4DELPREF2.replace("<app_code>",parent_app);
         stmt=stmt.replace("<form_code>",parent_form);
         stmt=stmt.replace("<config_code>",parent_config);
         stmt=stmt.replace("<user_name>",user_name);
         lc.executeDML(stmt);
         stmt=STMT4INSPREF2.replace("<app_code>",parent_app);
         stmt=stmt.replace("<form_code>",parent_form);
         stmt=stmt.replace("<config_code>",parent_config);
         stmt=stmt.replace("<user_name>",user_name);
         stmt=stmt.replace("<id>",id);
         lc.executeDML(stmt);
         }
         if (!is_default.equals("Y")){
         stmt=STMT4DELPREF3.replace("<app_code>",parent_app);
         stmt=stmt.replace("<form_code>",parent_form);
         stmt=stmt.replace("<config_code>",parent_config);
         stmt=stmt.replace("<user_name>",user_name);
         stmt=stmt.replace("<id>",id);
         lc.executeDML(stmt);
         }
         stmt=STMT4DELPREF4.replace("<app_code>",parent_app);
         stmt=stmt.replace("<form_code>",parent_form);
         stmt=stmt.replace("<config_code>",parent_config);
         stmt=stmt.replace("<user_name>",user_name);
         stmt=stmt.replace("<id>",id);
         stmt=stmt.replace("<prop_code>","QBE_NOT");
         lc.executeDML(stmt);
         
         if (frm.getProperty("QBE_NOT").equals("Y")) {
           stmt=STMT4INSPREF4.replace("<app_code>",parent_app);
           stmt=stmt.replace("<form_code>",parent_form);
           stmt=stmt.replace("<config_code>",parent_config);
           stmt=stmt.replace("<user_name>",user_name);
           stmt=stmt.replace("<id>",id);
           stmt=stmt.replace("<prop_code>","QBE_NOT");  
           stmt=stmt.replace("<value>","Y");
            lc.executeDML(stmt);
         }
         if (frm.getProperty("QBE_COL").equals("Y")) {
           stmt=STMT4INSPREF4.replace("<app_code>",parent_app);
           stmt=stmt.replace("<form_code>",parent_form);
           stmt=stmt.replace("<config_code>",parent_config);
           stmt=stmt.replace("<user_name>",user_name);
           stmt=stmt.replace("<id>",id);
           stmt=stmt.replace("<prop_code>","QBE_COL");  
           stmt=stmt.replace("<value>","Y");
            lc.executeDML(stmt);
         }
        } finally {
            app.metaConnPool.SetFree(lc);
            frm.setUsrConfigListProperty();
            if (id!=null){
                frm.loadUserConfig(id);
            }
        }
         }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        
    }
    
}
