/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Генерация скриншота окна браузера.
 */

package Standard;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.sql.SQLException;

import java.util.*;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import luiserver.LuiAppForm;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiClCommand;
import luiserver.LuiClCommandGroup;
import luiserver.LuiClCommandWriter;
import luiserver.LuiClEvent;
import luiserver.LuiClEventReader;
import luiserver.LuiPlugin;

/**
 *
 * @author vsenchenko
 */
public class Screenshot implements LuiPlugin
{
    boolean print_debug = true;
    LuiApplication app;
    LuiCollection cs = null;
    LuiClCommandWriter cw;
    LuiClEventReader er;

    // SVA ---------------------------------------------------------------------

    private void DEBUG(String str)
    {
       if(print_debug) System.out.println(str);
    }

    private void DEBUGP(String str)
    {
       if(print_debug) System.out.print(str);
    }

    private void DEBUG(String format, Object... arguments)
    {
        if(print_debug) System.out.printf(format+"\n", arguments);
    }

    // SVA ---------------------------------------------------------------------

    @Override
    public void setContext(LuiApplication pContext)
    {
        app=pContext;
        er=app.eventReader;
        cw=app.commandWriter;

        /*
        System.out.println("Screenshot.setContext");

        try
        {
            DEBUG("app.getTmpDir() = " + app.getTmpDir());
        }
        catch (LuiException ex)
        {
            DEBUG("ERR: " + ex.getMessage());
        }
        catch (SQLException ex)
        {
            DEBUG("ERR: " + ex.getMessage());
        }
        */
    }

    // SVA ---------------------------------------------------------------------

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        throw new UnsupportedOperationException("Not supported yet.");
        //To change body of generated methods, choose Tools | Templates.
    }

    // SVA ---------------------------------------------------------------------
    
    private String getParameter(String pParamCode)  throws LuiException, SQLException
    {
        String value = "";
        LuiDbConn lc = null;
        
        try
        {
            lc = app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
            value = lc.getSelectResult("select lui_f_get_application_property('"+app.code+"','"+pParamCode+"','"+app.NLS.currentLang.code+"')");
        }
        finally
        {
            if( lc != null )
            {
                app.metaConnPool.SetFree(lc);
                lc = null;
            }
        }
              
        return value;
    }

    // SVA ---------------------------------------------------------------------

    public void saveFile(String fname, String data) //throws IOException
    {       
        try
        {
            FileUtils.writeStringToFile(new File(fname), data, "UTF-8");
        }
        catch( IOException ex )
        {
            DEBUG("IOException: " + ex.getLocalizedMessage());
        }
    }
    
    // SVA ---------------------------------------------------------------------
    
    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        String fName = "";
        String url = "";       
        LuiClEvent ev;

        DEBUG("compute");
        
        String formId = app.getVal("FormID");

        DEBUG("formId = " + formId);

        LuiAppForm frm = app.forms.get(formId);

        DEBUG("frm.code = " + frm.code);
        DEBUG("form type = " + frm.getType());

        // поскольку вероятно придётся давать команды браузеру для сохранения форм,
        // то сначала мы отправим серверу визуализации все имеющиеся в буферах команды
        app.changes();
        app.getSeance().remZeroCmd(); // удалить пустышки из очереди команд (пустые группы)
        app.getSeance().remFrmWinApp(); // удаление форм и окон...

        // кроме того удалим из очереди событий клиента всё что там есть сейчас
        er.eventList.clear();
        
        // и проверим не исчезла ли наша форма и её окно
        if( !app.forms.containsKey(formId) )
            return "";

        String winID = frm.getVal("WindowID"); // узнаём код окна формы

        if( !app.windows.containsKey(winID) )
            return"";

        LuiClCommandGroup XMLcg = cw.addGroup("Capture");
       
        LuiClCommand XMLcmd = XMLcg.addCommand("Capture");
        XMLcmd.addScaPar("FormID", formId);

        cw.flush();

        er.readEvents();

        // цикл обработки событий
        while( (ev = er.getEvent()) != null  )
        {
            DEBUG("ev.name = " + ev.name);

            if( ev.name.equals("Capture") )
            {
                String data = ev.getParValue("Data");
                String w = ev.getParValue("Width");
                String h = ev.getParValue("Height");

                DEBUG("w = " + w);
                DEBUG("h = " + h);

                DEBUG("app.getTmpDir() = " + app.getTmpDir());
                
                //String 
                        fName = app.getTmpDir() + "screenshot_" + formId;
                
                //fName = fName.replace("o:\\\\", "o:\\");

                saveFile(fName.concat(".html"), data);
               
                String cmd = getParameter("MAKE_SCREENSHOT");

                DEBUG("cmd = " + cmd);
                
                cmd = cmd.replace("<w>", w).replace("<h>", h)
                        .replace("<img_fname>", fName.concat(".png"))
                        .replace("<html_fname>", fName.concat(".html"));
                
                DEBUG("cmd = " + cmd);
                
                Runtime runtime = Runtime.getRuntime();

                try
                {
                    Process process = runtime.exec(cmd);
                    process.waitFor();
                    int exitVal = process.exitValue();
                    DEBUG("exitVal = " + exitVal);
                    
                    if( exitVal == 0 )
                    {
                        url = app.getTmpDirURL().concat("screenshot_").concat(formId).concat(".png");
                    }
                }
                catch (InterruptedException e)
                {
                    DEBUG("InterruptedException: ");
                    e.printStackTrace();                
                }
                catch (IOException e)
                {
                    DEBUG("IOException: ");
                    e.printStackTrace();
                }                
            }
        }


        if( is_not_null(url) )
        {
            byte[] imageData = null;

            try( InputStream is = new FileInputStream(fName.concat(".png")) )
            {
                imageData = IOUtils.toByteArray(is);
                is.close();        
            }
            catch( FileNotFoundException ex )
            {
                //System.out.println("FileNotFoundException: " + ex.getMessage());
                app.msgWriter.write("LUI-00097", fName.concat(".png"), ex.getMessage());
            }
            catch( IOException ex )
            {
                //System.out.println("IOException: " + ex.getMessage());
                app.msgWriter.write("LUI-00097", fName.concat(".png"), ex.getMessage());
            }
                    
            String imageBase64 = "";
            if (imageData!=null)imageBase64=java.util.Base64.getEncoder().encodeToString(imageData);                   

            //LuiClCommandGroup 
                    XMLcg = app.commandWriter.addGroup("SCREENSHOT");
            //LuiClCommand 
                    XMLcmd = XMLcg.addCommand("DownloadFile");
            XMLcmd.addScaPar("FormID", formId);
            XMLcmd.addScaPar("FileName", "screenshot_".concat(formId).concat(".png"));
            XMLcmd.addScaPar("Base64", imageBase64);
        }
        
        return url;
    }
    
    // SVA ---------------------------------------------------------------------
    
    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException
    {
        String fc = app.getVal("FormID");
        if( is_null(fc) ) return;
        if( !app.forms.containsKey(fc) ) return;
        
        String url = compute(pString, pProp);

        /*
        DEBUG("url = " + url);

        
        LuiClCommandGroup XMLcg=application.commandWriter.addGroup("SaveChart");
        LuiClCommand XMLcmd=null;
        XMLcmd=XMLcg.addCommand("SaveChart");
        XMLcmd.addScaPar("FormID",this.getFormID());
        XMLcmd.addScaPar("FileName",NVL(NVL(Jsoup.parse(getProperty("HEADER")).text(),getProperty("ConfigName")),this.code));

        
        if( is_not_null(url) )
        {
            LuiAppForm frm = app.forms.get(fc);
            //LuiClCommandGroup XMLcg = app.commandWriter.addGroup("PRINT");
            LuiClCommandGroup XMLcg = app.commandWriter.addGroup("SCREENSHOT");

//            LuiClCommand XMLcmd = XMLcg.addCommand("URL");
//            XMLcmd.addScaPar("Address", url);
//            XMLcmd.addScaPar("Target", "Tab");

            LuiClCommand XMLcmd = XMLcg.addCommand("DownloadFile");
            XMLcmd.addScaPar("FormID", fc);
            XMLcmd.addScaPar("FileName", url);
            XMLcmd.addScaParameter("Path",actionCommand);
            
            
            DEBUG("url = " + url);
        }
        */
    }    
}
