/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Отправка e-mail
 * Compute отсутствует
 * Execute
 * SendMail:<Тема>;Content:<Текст>;ContentType:<html/plain>;Charset:<>;Host:<ip_дрес или доменное имя>;Port:<порт>;SSL:Y/N;TLS:Y/N;User:<>;Password:<>;Sender:<>;Recipients:<>,<>;AttFiles:<>,<>;AttBLOBs<>,<>;
 */
package Standard;

import LuiCollection.LuiCollection;

import static LuiCollection.LuiString.NVL;

import LuiExceptionMsg.LuiException;
import java.io.File;
import java.sql.SQLException;
import luiserver.LuiAppProp;
import luiserver.LuiApplication;
import luiserver.LuiPlugin;
import ru.fors.utils.lightsmtp.MimeAddress;
import ru.fors.utils.lightsmtp.MimeAttachment;
import ru.fors.utils.lightsmtp.MimeMessage;
import ru.fors.utils.lightsmtp.MimeMessage.RecipientType;
import ru.fors.utils.lightsmtp.MimeMultipart;
import ru.fors.utils.lightsmtp.MimeTextBody;
import ru.fors.utils.lightsmtp.SMTPClient;



/**
 *
 * @author aliubushkin
 */
public class SendMail implements LuiPlugin{
LuiApplication app;
    @Override
    public void setContext(LuiApplication pContext) {
        app=pContext;
    }

    @Override
    public String compute(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        return "";
    }
    
    
    @Override
    public void execute(String pString, LuiAppProp pProp) throws LuiException, SQLException {

        try {
            LuiCollection aParams = new LuiCollection("Subject:" + pString);
            
            String aTmp = aParams.Get_Element("Debug");
            boolean aDebug = aTmp != null && "Y".equalsIgnoreCase(aTmp);

            MimeMessage aMessage = new MimeMessage();
            
            
            // от кого
            aTmp = aParams.Get_Element("Sender");
            if (aTmp == null || aTmp.isEmpty()) {
                throw new RuntimeException("Sender address required");
            }
            aMessage.setSender(new MimeAddress(aTmp));

            // кому
            aTmp = aParams.Get_Element("Recipients");
            if (aTmp == null || aTmp.isEmpty()) {
                throw new RuntimeException("Recipient address required");
            }
            for(String aRcpt: aTmp.split(",")) {
                aMessage.addRecipient(new MimeAddress(aRcpt));
            }
            
            
            // копии
            aTmp = aParams.Get_Element("CopyTo");
            if (aTmp != null && !aTmp.isEmpty()) {
                for(String aRcpt: aTmp.split(",")) {
                    aMessage.addRecipient(RecipientType.CC, new MimeAddress(aRcpt));
                }
            }

            // скрытые копии
            aTmp = aParams.Get_Element("BlindCopyTo");
            if (aTmp != null && !aTmp.isEmpty()) {
                for(String aRcpt: aTmp.split(",")) {
                    aMessage.addRecipient(RecipientType.BCC, new MimeAddress(aRcpt));
                }
            }


            // тема
            aTmp = aParams.Get_Element("Subject");
            if (aTmp != null && !aTmp.isEmpty()) {
                aMessage.setHeader("Subject", aTmp);
            }

            MimeMultipart aContent = new MimeMultipart("multipart/mixed");
            MimeTextBody aMessageText = new MimeTextBody();;
            
            aMessageText.setMimeType(NVL(aParams.Get_Element("ContentType"), "text/plain"));
            aMessageText.append(NVL(aParams.Get_Element("Content"), ""));
            
            aContent.add(aMessageText);
            aMessage.setBody(aContent);

            // Вложения
            aTmp = aParams.Get_Element("AttFiles");
            if (aTmp != null && !aTmp.isEmpty()) {
                for (String aFileName : aTmp.trim().split(",")) {
                    File aFile = new File(aFileName);
                    if (!aFile.exists()) {
                        throw new RuntimeException("File '" + aFileName + "' not exists");
                    }
                    if (!aFile.isFile()) {
                        throw new RuntimeException("File '" + aFileName + "' is not a plain file");
                    }
                    if (!aFile.canRead()) {
                        throw new RuntimeException("Can not read '" + aFileName + "'");
                    }

                    MimeAttachment aAttach = new MimeAttachment(aFile);
                    aContent.add(aAttach);
                }
            }

            

            // ==== отправка=========
            String aHost = aParams.Get_Element("Host");
            String aPort = aParams.Get_Element("Port");
            int aPortN = 0;
            String aTLS = aParams.Get_Element("TLS");
            String aSSL = aParams.Get_Element("SSL");
            String aUser = NVL(aParams.Get_Element("User"), aParams.Get_Element("Username"));
            String aPasswd = NVL(aParams.Get_Element("Password"), aParams.Get_Element("Passwd"));
            
            if(aPort != null && !aPort.isEmpty()) {
                try {
                    aPortN = Integer.parseInt(aPort);
                } catch(Exception ignore) {}
                if(aPortN < 1 || aPortN > 65535) {
                    throw new RuntimeException("SMTP Port must be a number between 1 and 65536");
                }
            }
            
            SMTPClient aSMTP = null;
            
            if(aHost == null || aHost.isEmpty()) { throw new RuntimeException("SMTP Host address required"); }
            if(aSSL != null && aSSL.equalsIgnoreCase("Y")) {
                if(aPortN == 0) { aPortN = 465; }
                aSMTP = SMTPClient.connectSSL(aHost, aPortN);
            }
            else if(aTLS != null && aTLS.equalsIgnoreCase("Y")) {
                aSMTP = SMTPClient.connectTLS(aHost, aPortN, true);
            }
            else {
                aSMTP = SMTPClient.connectTLS(aHost, aPortN, false);
            }
            
            try {
                if(aUser != null && (!aUser.isEmpty()) && aPasswd != null) {
                    if(!aSMTP.login(aUser, aPasswd))  {
                        throw new RuntimeException("SMTP Auth failed");
                    }
                }
                
                aSMTP.sendMessage(aMessage);
            }
            finally {
                aSMTP.close();
            }
            
            

        } catch (Throwable ex){
            app.msgWriter.write("LUI-00111",ex.getMessage());  
        }
    }

    @Override
    public void findDependency(String pString, LuiAppProp pProp) throws LuiException, SQLException {
        
    }
     @Override
    public String quoteMask(String prefixString, String Value) {
        return Value.replace(";","#$");
    }
}
