/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс поддержки соединения Приложения с прикладной БД
 
 */
package luiserver;

import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;

/**
 *
 * @author aliubushkin
 */
public class LuiAppConn implements LuiAppBreakable{
public    String name;
public    String mode; // PERSONAL/COMMON
public    String username;
    String password;
public    String dbURL;
public    String schema;
public    LuiDbConn currentConnection;
public    LuiDbConnPool connPool=null;
public String code;
    LuiApplication app;

    
public LuiAppConn(LuiApplication pApp,String pCode,String pName,String pMode,String pUsername,String pPassword,String pDbURL,String pSchema){
    app=pApp;
    name=pName;
    mode=pMode;
    username=pUsername;
    password=pPassword;
    dbURL=pDbURL;
    schema=pSchema;
    code=pCode;
}
public char getSessionMode(){
    return mode.charAt(0);
}
/**
 * Установление соединения для пользовательского приложения
 * если соединение коммунальное, то оно после проверки соединимости тутже отдаётся в пул свободных
 * @throws SQLException ошибка при выполнении SQL-запроса
 * @throws LuiException ещё какая-нибудь ошибка
 */
public void connect() throws SQLException, LuiException {
            if (getSessionMode()=='C') {// и если не требуется инд. постоянная сессия - создаём пул c автокоммитом
              
                
                //for(int i=0;i<app.poolArray)  
              connPool=app.poolArray.addPool(dbURL,username,password,schema,true,false,name);
              }
            else{ //требуется постоянная сессия - создаём пул без автокоммита
              connPool=app.poolArray.addPool(dbURL,username,password,schema,false,false,name);
                 }   
if (connPool==null) {
  app.msgWriter.write("LUI-000036", dbURL);
}
             //Выполняем попытку соединения
             try{
                     currentConnection=connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser()); 
                     } catch (SQLException ex){
                         app.poolArray.removePool(connPool); // при неудачной попытке соединения удаляем пул совсем
                             throw ex;
                             }     
                     // если постоянного соединения не требуется, то освободим его
                     if (getSessionMode()=='C') {
                         connPool.SetFree(currentConnection);
                     }
                     // занесём его для break
                       app.backList.add(this);
             }
    
public void disconnect(){
    //Экономия пулов ради
    try{
    synchronized(connPool) {   
    if (connPool.getAutoCommit()) {
        connPool.appCntDEcrement();
        if (connPool.getAppCnt()>0) { //если какие-то ещё подписанты остались
            return;
        }
    }
     connPool.DisconnectAll();
    
    app.poolArray.removePool(connPool);
    }
    } catch (SQLException ex){
        
    }
    
}    

    @Override
    public void back() {
      if (app.userDbConnections.containsKey(code)){
        app.userDbConnections.remove(code);
       } 
      disconnect(); 
    }
public String getPWDC(){
    return app.crypter.encrypt(this.password);
}
}

