/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;


import LuiCollection.LuiCollection;
import LuiCollection.LuiString;
import static LuiCollection.LuiString.LTrim;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.RemoveIco;
import static LuiCollection.LuiString.String2Integer;
import static LuiCollection.LuiString.compareString;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.script.ScriptException;





/**
 *
 * @author aliubushkin
 */
public class LuiAppElement extends LuiAppWFE{
     public      double sourceSeq; //порядковый номер элемента
     public      double mutSeq; //мутированный номер элемента (для изменения порядка столбцов)
           public ColInfo colinfo; //информация о столбце запроса из JDBC
           LuiAppForm frm;
           LuiApplication app;
           String CURSORSIGN="¦";
           public String Code;
           
 /**
 * Конструктор элемента 
 * @param pObj - хозяин
 * @param pType - тип элемента
 * @param pCode - код элемента
 * @param pSeq  - порядковый номер
 */
public LuiAppElement(LuiAppWFE pObj,String pType,String pCode,double pSeq){
    sourceSeq=pSeq;
    ownerObj=pObj;
    mutSeq=sourceSeq;
    Code=pCode;
    frm=(LuiAppForm)ownerObj;
    app=frm.application;
//properties=super.properties;
    setProperty("Code",pCode);
    setProperty("EType",pType);
    addProperty("Metadata","Y");
    setChangeFlag();
}
/**
 * Возвращает код элемента
 * @return код элемента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getCode() throws LuiException, SQLException{
    return this.getNonCompValue("Code");
}
/**
 * Возвращает тип элемента
 * @return тип элемента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getType() throws LuiException, SQLException{
        
        return this.getNonCompValue("EType");
}
/**
 * Устанавливает значение поля (property Value)
 *
 * @param pValue -- новое значение
 * @param pHT -- руками или роботом true-руками false-роботом
 * @param pIsCorrectByFormat -- следует ли причёсывать под формат
 * @param pCheckChanging -- проверять возможность ввода руками
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setValue(String pValue,boolean pHT,String pIsCorrectByFormat,boolean pCheckChanging) throws LuiException, SQLException{
    String newVal=LuiParser.mask(pValue);//-- Value всегда замаскированный
    
    String cc=null;
    LuiAppConn ac=null;
    LuiDbConn lc=null;
    try{
    if (properties.containsKey("Value")){ // если у элемента уже есть такое 
    // блок сдесь написанный может быть пропущен т.к. у параметров ещё может не быть value
    String oldVal=LuiParser.mask(properties.get("Value").currentValue);//-- СТАРЫЙ ТОЖЕ  всегда замаскированный
    if (pIsCorrectByFormat.equals("Y")|| !compareString(oldVal,newVal)){ // если надо новое значение причесать под формат, или оно не равно старому
     if (is_not_null(newVal) && getType().equals("FIELD")) { // и если новое значение не пусто, а тип элемента = поле бланка
      if (pHT && pCheckChanging){ // если руками и надо проверять возможность ввода руками
         if (!(getProperty("CHANGING").equals("Y") ||   
          getProperty("CHANGING").equals("C") ||    
          getProperty("CHANGING").equals("U") ||
          getProperty("CHANGING").equals("B") ||
          getProperty("CHANGING").equals("R") ||
          getProperty("CHANGING").equals("H"))) {// и если поле не предназначено для ввода - уходим
          return; 
      }
      }
      //if (pIsCorrectByFormat.equals("Y") ) {// если указано, что надо причёсывать данные под формат 
      
      String dataFormat=getProperty("FORMAT");
      //if (is_not_null(dataFormat)) {// и если задан формат данных
      try {
      String fieldDatatype=getProperty("DATA_TYPE");    
      String AorF=app.dataConverter.LuiDataTypes.get(fieldDatatype).AlphabetOrFormat;
      String tmpVal=newVal;
      if (AorF.equals("A") && pIsCorrectByFormat.equals("Y")) { // проверяем алфавит если надо
                newVal=checkAlpha(dataFormat, newVal, getName4show());
      }
      if (is_not_null(newVal) && pIsCorrectByFormat.equals("Y") && !AorF.equals("A")) { // если не пусто и надо проверять формат и поле не алфавитное
      // причёсываем данные под формат и выполняем попытку конвертирования
                 cc=ownerObj.getProperty("Connection"); // выясняем код соединения
                 if (is_null(cc)) {// если у бланка нет соединения поищем у приложения
                   for(String ccc:getApplication().userDbConnections.keySet()){
                       cc=ccc;
                       break;
                   }
                 }
                 if (is_not_null(cc)) { //если  не пуст
                    if (getApplication().userDbConnections.containsKey(cc)) { // и такое соединение есть
                      ac=getApplication().userDbConnections.get(cc);
                     if (ac.getSessionMode()=='C') { // если соединение коммунальное, то получим из пула свободное
                       lc=ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                     }
                    else {//в противном случае возьмём текущее 
                       lc=ac.currentConnection;
                        }  
                    } 
                 }
                 if (lc==null){ // если соединение так и не нашлось - возьмём из пула соединений к мета данным
                     lc=getApplication().metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                 }
      newVal=app.dataConverter.convAndCheck(newVal, this, lc);                  
      // конец  причёсывания данные под формат и попытки конвертирования
      
      } // если не пусто
      
       setProperty("Correct","Y");    
       if (!pIsCorrectByFormat.equals("Y") ) {
           newVal=tmpVal;
       }
     // }
      } 
      catch (LuiException lex){
            Integer mi=lex.getIndex();
            if (mi!=null) {
                app.msgWriter.msgArray.remove((int)mi);
            }
        }
      catch (SQLException sex){
       setProperty("Correct","N");   
      }
      catch(Exception ex){
       setProperty("Correct","N");
      }
      finally {
        if (ac!=null) {
            if (ac.getSessionMode()=='C') {
                ac.connPool.SetFree(lc);
            }
        } else {
          if (lc!=null){
             getApplication().metaConnPool.SetFree(lc);
          }
        }
      }
     }   
     if (!compareString(oldVal,newVal)) { // ЕСЛИ СТАРОЕ != НОВОМУ 
         if (getType().equals("ITEM")){
             addProperty("Value",newVal);
         }
         properties.get("Value").clearComputedValue(false); // очищаем вычисленное значение
                                                       // И ВСЕ КТО ОТ ЭТОГО ПОЛЯ ЗАВИСЯТ ТОЖЕ ОЧИТЯТСЯ
        if (pHT){
          setProperty("UserChanged","Y"); // метим поле введённым
          ownerObj.setVal("Changed", "Y"); // метим форму изменённой руками
        }
     properties.get("Value").setChangeFlag();
     }
    }
    }
    else {
    addProperty("Value",newVal);
    }
    properties.get("Value").currentValue=newVal;
    properties.get("Value").setChangeFlag();
    // если есть серверный верификатор
      if (properties.containsKey("VERIFIER_SRV_PRG")){
        if (pIsCorrectByFormat.equals("Y")){ // и проверка нужна  
        if   ("Y".equals(getProperty("Correct"))){ // и на предыдущих этапах проверка дала положительный результат
         LuiAppProp p=properties.get("VERIFIER_SRV_PRG");
        try { 
         p.recalculate=true;
         if (is_not_null(p.getComputedProperty())) setProperty("Correct","N");  
        } catch (LuiException lex){
            Integer mi=lex.getIndex();
            if (mi!=null) {
                app.msgWriter.msgArray.remove((int)mi);
            }
        }
          catch (Exception ex){
          setProperty("Correct","N");  
        } 
      }
      }
      }
    if (properties.get("Value").recalculate!=null)
  {
      // у параметров Value - динамика!
      properties.get("Value").recalculate=false; // говорим что вычислен чтобы пользователський ввод не затёрся динамикой
  } else {  }
    } finally {
        if (ac!=null) {
            if (ac.getSessionMode()=='C') {
                ac.connPool.SetFree(lc);
            }
        } else {
          if (lc!=null){
             getApplication().metaConnPool.SetFree(lc);
          }
        }
            
    }  
    
}
@Override
public LuiAppProp setProperty(String pPropertyCode, String pPropertyValue) {
    if (pPropertyCode.equals("UserChanged") && "Y".equals(pPropertyValue)){
    ownerObj.setVal("Changed", "Y"); // метим форму изменённой руками
    }
    return super.setProperty(pPropertyCode, pPropertyValue);
}
    

    /**
     * Возвращает значение свойства "Value" элемента
     *
     * @return значение свойства "Value"
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса     
     */
    public String getValue() throws LuiException, SQLException {
        return getProperty("Value");
    }

/**
 * возвращает код родительского элемента
 * @return код родительского элемента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getParent()throws LuiException, SQLException{
    if (properties.containsKey("Parent")){
    return properties.get("Parent").getComputedProperty();
    }
     return "Form";
}

/**
 * Вычисляет имя элемента для показа в ошибках и логах
 * @return имя элемента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getName4show() throws LuiException, SQLException{
  String retVal="";
  
    switch (getType()){
        case "ITEM":
        case "FIELD":
        retVal= getProperty("CAPTION");
        break;
        case "ACTION":
        retVal= getProperty("PROMPT");
        break;
        case "GROUP":
        retVal= NVL(LuiString.LTrim(getProperty("TITLE")),getProperty("Name"));
        break;
    }
  
  if (is_null(retVal)) retVal=getProperty("Name");
  return retVal;
}

/**
 * Выполнение проверки значения на соответствие алфавиту
 * @param pAlphaCode код алфавита
 * @param pValue проверяемое значение
 * @param pName имя значения
 * @return не знаб, что, пусть автор напишет
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String checkAlpha(String pAlphaCode,String pValue,String pName) throws SQLException, LuiException{
  if (is_null(pAlphaCode)) return pValue;
  
  LuiDbConn lc=null;
  try {
    lc  = app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());   
    String aaCode=app.code;
    LuiCollection cf=null;
    String Stmt="select code,final_check from lui_t_alphabet where app_code='"+aaCode+"' and code='"+pAlphaCode+"'";
    cf=lc.getSelectFirstRowValMap(Stmt);
    if (cf==null && !app.code.equals("LUI")) {
      aaCode="LUI";
      Stmt="select code,final_check from lui_t_alphabet where app_code='LUI' and code='"+pAlphaCode+"'";
      cf=lc.getSelectFirstRowValMap(Stmt);
      if (cf==null) return pValue;
   }

    Stmt="select seq,pattern,correction,error_code from lui_t_alphabet_rule where alphabet_code='"+pAlphaCode+"' and app_code='"+aaCode+"' order by seq";
    ArrayList<LuiCollection> rules= lc.getSelectAllRowsValMap(Stmt);
    for (int i=0; i<rules.size(); i++) {
    String pattern = rules.get(i).get("pattern");
    String Correction = rules.get(i).get("correction");
    String ErrorCode = rules.get(i).get("error_code");
    boolean macho=false;
    try {
      macho=pValue.matches(pattern);
    } catch (Exception ex){
      app.msgWriter.write("LUI-00088", pAlphaCode,rules.get(i).get("seq"),ex.getMessage());
    }
    if (is_not_null(ErrorCode) && !macho) {
        app.msgWriter.write(ErrorCode,pName);
    }
    if (is_not_null(Correction) || is_null(ErrorCode)) {
        Pattern pmacho = null;
        try{
        pmacho=Pattern.compile(pattern);
        } catch(Exception ex){
          app.msgWriter.write("LUI-00088", pAlphaCode,rules.get(i).get("seq"),ex.getMessage());
        }
        Matcher mmacho=null;
        try{
        mmacho=pmacho.matcher(pValue);
        } catch (Exception ex){
          app.msgWriter.write("LUI-00088", pAlphaCode,rules.get(i).get("seq"),ex.getMessage());
        }
        try {
        pValue=mmacho.replaceAll(NVL(Correction,""));
        } catch (Exception ex){
          app.msgWriter.write("LUI-00089", pAlphaCode,rules.get(i).get("seq"),ex.getMessage());
        }
    }
    }
   if (cf!=null && is_not_null(cf.Get_Element("final_check"))) {
       LuiAppProp p=addProperty("FinalCheck",cf.Get_Element("final_check"));
       if (properties.containsKey("Value")) {
       properties.get("Value").currentValue=pValue;
       }
       pValue=p.getComputedProperty();
       for (LuiAppProp pt:p.dependsFrom.keySet()){
         pt.depends.remove(p);
       }
       properties.remove("FinalCheck");
   }
  } finally {
    if (lc!=null)  app.metaConnPool.SetFree(lc);
  }
  
  return pValue;
}
/**
 * Поле может быть изменено каким-то образом пользователем?
 * @return признак возможности изменения поля пользователем
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean isFieldChangeable() throws LuiException, SQLException{
  if (!getType().equals("FIELD"))  return false;
  switch (getProperty("CHANGING")){
      case "Y":
      case "C":
      case "U":
      case "P":
      case "B":
      case "H":
      case "R":
      return true;
  }
  return false;
}
/**
 * Элемент видим?
     * @param withparent -- с учётом видимости родителей или без учёта
 * @return признак видимости элемента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean isVisible(boolean withparent)throws LuiException, SQLException{
  switch (getType()){
      case "FIELD":
      case "ITEM":
      case "ACTION":
      String visible="";
          if (getType().equals("ACTION")) {
      visible=getProperty("DISPLAY");              
          } 
          else {
      visible=getProperty("VISIBLE");
          }
      if (visible.equals("N")) return false;
      if (like(visible,"0/%")) return false;
      if (like(visible,"%/%")) visible="Y";
      if (visible.equals("Y")) visible="1";
      try{
      Integer v=String2Integer(visible,null);
      if (v==null) return false;
      if (v<=0) return false;
        if (v>0) {
            if (getParent().equals("Form")||!withparent) return true;
            return (frm.elements.get(getParent()).isVisible(withparent));
        }
      } catch (Exception ex){
          return false;
      }
      
      break;
      case "GROUP":
       String gvi=NVL(getProperty("SEEABLE"),"Y");
       if (like(gvi,"0/%")) gvi="N";
        if (like(gvi,"%/%")) gvi="Y";
       if (is_not_null(gvi)){
       try{
       Integer ivg=Integer.valueOf(gvi);
       if (ivg!=null && ivg>0) gvi="Y";
       } catch (Exception e){
           
       }
       }
       if (!NVL(gvi,"N").equals("Y")) return false;
       if (getParent().equals("Form")||!withparent) return true;
            return (frm.elements.get(getParent()).isVisible(withparent));
      
  }
  return true;  
}
public String getFieldShow() throws LuiException, SQLException{
    String oldShow="";
    if (properties.containsKey("Show")) {
        oldShow=LuiParser.reMask(properties.get("Show").currentValue);
    } else setProperty("Show","");
    String Show=getFieldSh();
    if (!oldShow.equals(Show)){
        setProperty("Show",Show);properties.get("Show").changeFlag=true;
    } else properties.get("Show").changeFlag=false;
    return Show;
}
/**
 * Вычисление значения SHOW для поля бланка
 * @return значение SOW для поля бланка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getFieldSh() throws LuiException, SQLException{
    String upd=NVL(getProperty("CHANGING"),"N");
    if (!(upd.equals("Y") || upd.equals("C") || upd.equals("B") || upd.equals("U") || upd.equals("H") ||upd.equals("R"))) upd="N";
    String curVal=getProperty("Value");
    if (is_null(getNonCompValue("SHOW"))){ // если SHOW не определено
      if (upd.equals("C")||upd.equals("N")||upd.equals("B")){// Выбор только из списка?
         String lov=getProperty("VARIANTS"); // берём LOV
         lov=LuiString.RemoveIco(lov);
         if (is_not_null(lov)) { // если он есть, то надо понять что это
                                 // LOVQ:(CONN)Select code,name from .... -- запрос к указанному соединению
                                 // LOVQ:Select -- запрос к соедлинению указанному в свойстве Connection бланка или к соединию какому- нибудь
                                 // select -- запрос к соедлинению указанному в свойстве Connection бланка или к соединию какому- нибудь
                                 // LOVD:Y:да;N:нет;
                                 // Y:да;N:нет;
          lov=LuiString.LTrim(lov); // отсекаем пробелы слева
          if (like(lov,"LOVC:%") || like(lov,"LOVD:%")) { // если это коллекция
             LuiCollection coll=new LuiCollection(lov.substring(5)); 
             if (coll.Exists_Element(curVal)){ // если в коллекции есть элемент с ключём= вычисленное значение, то
                 
                 return coll.Get_Element(curVal); // возвращаем расшифровку значения
             } else { //// если в коллекции НЕТ элемента с ключём= вычисленное значение, то
                 
                 return NVL(curVal,""); // возвращаем вычисленное значение
             }
                 
          }
          //if (like(lov,"LOVQ:%")) {// если это запрос, то определяем текст запроса и код соединения
          if (lov.indexOf("LOVQ:",0)!=-1) {// если это запрос, то определяем текст запроса и код соединения
              ArrayList<String> ac=app.getAppConnAndStmt(lov.substring(5), ownerObj.getProperty("Connection"),this.frm.getFormID());
              LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
              LuiDbConn lc=null;
              Integer cursor=null;
              String valFromLOV=null;
              String query=ac.get(1);
              try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
              } else lc=appConn.currentConnection;
              // открываем курсор по запросу LOV'а
              
              cursor=lc.CursorOpen(query);
              ArrayList<ColInfo> colsInfo=lc.CursorGetColsInfo(cursor); // получаем описание столбцов
              ColInfo cf;
              String alias1="\""+colsInfo.get(0).alias+"\"";
              String alias2=alias1;
              cf=colsInfo.get(0);
              if (colsInfo.size()>1) {
              alias2="\""+colsInfo.get(1).alias+"\"";
              cf=colsInfo.get(1);
              }  
              
              query="select "+alias2+" from ("+ac.get(1)+")q$1 where ";
              cf.LuiDATATYPE=this.getProperty("DATA_TYPE");
              cf.LuiFORMAT="";
              String cla=app.dataConverter.getConvString(cf, lc, true,true);
              if (is_null(cla)){
                  query=query+alias1+"='"+NVL(curVal,"").replace("'","''")+"'";
              } else {
                  addProperty("v", alias1);
                  addProperty("f", "");
                  LuiAppProp p=addProperty("ConvertExpression",cla);
                  cla=p.getComputedProperty();
                  query=query+cla+"='"+NVL(curVal,"").replace("'","''")+"'";   
              }
             
              valFromLOV=lc.getSelectResult(query);
              if (lc.getRowCount()==0 && is_null(curVal)){ // если ни чего не извлеклось и вычисленное значение пусто
              query="select "+alias2+" from ("+ac.get(1)+")q$1 where "+alias1+" is null";
              valFromLOV=lc.getSelectResult(query);    
              }
              if (lc.getRowCount()==0 && upd.equals("N")){// если ничего не извлеклось и поле не редактируемое
                 
                  return NVL(curVal,""); // вернём значение поля
              }
              
              return NVL(valFromLOV,""); // вернём извлечённое значение, но не null!
              } 
              catch (SQLException ex){
            app.msgWriter.write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+
                    " ("+query+") ",
                    this.properties.get("VARIANTS").getGenesis() );
            
            return "";
        }
              finally {
                  if (lc!=null) {
                      if (cursor!=null) lc.CursorClose(cursor);
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }
              
          } else { // это не LOVD и не LOVQ
            
              return NVL(curVal,"");
          }
         } else{ // если нет
            
            return NVL(curVal,""); // возвращаем вычисленное значение 
         }
          
      } else { // если допустим ввод руками, то
          
           return NVL(curVal,""); // возвращаем вычисленное значение 
      }
    } else { // show указан
        if (upd.equals("C")||upd.equals("N")||upd.equals("B")){// Выбор только из списка?
          
            return getProperty("SHOW"); // вычисляем и возвращаем show
        } else { // если допустим ввод руками, то
            if ((ownerObj.getVal("CurrentField")).equals(getCode())) {// если это поле текущее
          
                return NVL(curVal,""); // возвращаем вычисленное значение 
            } else { //если не текущее
          
            return NVL(getProperty("SHOW"),""); // вычисляем и возвращаем show   
            }
        }
    }
//return "";
}
/**
 * является ли ячейка грида редактируемой?
 * @return признак редактируемости ячейки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean isEditCell() throws LuiException, SQLException{
    for (String elcode:frm.elements.keySet()){
      LuiAppElement el=frm.elements.get(elcode);  
       if (el.getParent().equals(getCode())){
          if (like(el.getProperty("ROLE"),"%POSTEDIT%")){
            if (el.getProperty("Run").equals("Y")){  
            switch (el.getProperty("ACTIVE")){
                case "Y" : case "EXIST": 
                return true;
            }
            }
          } 
       }
    }
  return false;  
}
/** 
 * Умеет определять активность действия ActionCode в форме FormID, исходя из
 * ACTION, SelectionMode, APPLY_TO  и количества отмеченных строк
 */


/**
 * Вычисляет активность действия
 * @return признак активности действия
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean isActionActive() throws LuiException, SQLException{
  
  String aia=getProperty("ACTIVE");
  boolean multiRowmode=false;
  if (frm.getProperty("SelectionMode").equals("Multi")) multiRowmode=true;
  if (aia.equals("Y") || aia.equals("EXIST")) {
    if (aia.equals("EXIST")) {
        if (frm.isNoSelectedRows()) return false;
        if (!NVL(frm.getVal("CurrentNodeType"),"ROW").equals("ROW")) return false;
    }
  //  if (getProperty("Parent").equals("Form")){
      String aim=getProperty("APPLY_TO");  
      if (aim.equals("N") && multiRowmode){
          if (aia.equals("EXIST")) return false;
      }
      if (aim.equals("Y") && frm.isNoSelectedRows()) return false;
  //  }
    return true;
} else return false;
}

/**
 * вычисляет максимальный seq среди элементов для которых данный элемент - отец
 * @return значение максимального seq
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public double getMaxChildSeq() throws LuiException, SQLException{
  String parent=getCode();
  double rc=-32767.0;
  for (String elc:frm.elements.keySet()){
      if (frm.elements.get(elc).getParent().equals(parent)){
         if (rc<frm.elements.get(elc).sourceSeq) rc=frm.elements.get(elc).sourceSeq;
      }
  }
  if (rc<0)return 0;
  return rc;  
}
public void cellButton(String pButton) throws LuiException, SQLException{
    if (is_null(pButton)) return;
    app.changeSCN("CellButton");
    String icon;
    for(int i=0;i<frm.elements.size();i++){
        LuiAppElement el=frm.elSort(i);
        if (el.getType().equals("ACTION")){
          if (el.getParent().equals(this.getCode())) {
            icon=el.getProperty("ICON");
            if (is_not_null(icon)){
              if (icon.equals(pButton)){
                String AV=el.getProperty("DISPLAY");
                if (AV.indexOf("SQL:", 0)>-1) { // в DISPLAY размножитель SELECT
          ArrayList<String> ac=app.getAppConnAndStmt(AV.substring(4), frm.getProperty("Connection"),this.frm.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          if (propsArray!=null){
          if (propsArray.size()>0) {
          LuiClCommandGroup XMLcg=app.commandWriter.addGroup("CellToolBarButton"); 
          XMLcg.addParameter("FormID",frm.getFormID());
          XMLcg.addParameter("Type","Action");
          LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
              // крутим извлечённые строки
              for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                el.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
            
                if (!compareString(el.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");  
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          }
           } finally {
                  if (lc!=null) {
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }         
                } else if (AV.indexOf("JS:", 0)>-1) { // в DISPLAY размножитель JS
                         try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.app.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          LuiClCommandGroup XMLcg=app.commandWriter.addGroup("CellToolBarButton"); 
          XMLcg.addParameter("FormID",frm.getFormID());
          XMLcg.addParameter("Type","Action");
          LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");

          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного массива
           el.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
            if (!compareString(el.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);    
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          } 
                    
                } else {// нет размножителя
                    if (el.isActionActive()){
                      frm.runAction(el.getCode());
                      app.checkDetails("");
                      break;
                    }
                }
              }
            }
          }
        }
    }
}
/**
 * Обработка нажатия ПКМ в поле бланка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void fieldRMB() throws LuiException, SQLException{
   boolean is_edtfld =false;
   
   //frm.makeElList("Form");
   app.changeSCN("FieldRMB");
   
   LuiClCommandGroup XMLcg=app.commandWriter.addGroup(getCode()+" LocalMenu");
   LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
   XMLcmd.addScaPar("FormID",frm.getFormID());
   XMLcmd.addScaPar("Type", "Action");
   for (int i=0;i<frm.elements.size();i++){
    LuiAppElement ael=frm.elSort(i);
    if (!like(ael.getCode(),"SYSMENU$%") && !like(ael.getCode(),"SYSACTION$%")){ // исключаем действия предусмотренные в шаблонах уровня формы
    if (compareString(ael.getType(),"ACTION")&& compareString(ael.getParent(),this.getCode())){
      if (ael.getProperty("Run").equals("Y")){
          if (like(ael.getProperty("ROLE"),"%FIELD_RMB%")){
            ael.properties.get("ACTIVE").clearComputedValue(false);  
            if (ael.isActionActive()) {
              app.commandWriter.removeGroup(getCode()+" LocalMenu");
              frm.runAction(ael.getCode());
              return;
            }
          }
          String AV=ael.getProperty("DISPLAY");
          if (like(ael.getCode(),"EDITOR$%")) is_edtfld=true;
          if (AV.equals("Y")){ // если нет ни JS, ни SQL
            String prompt=ael.getProperty("PROMPT");
            if (is_not_null(prompt)){
                XMLcmd=XMLcg.addCommand("AddMenuItem");
                XMLcmd.addScaPar("Code",ael.getCode());
                XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
                if (!compareString(ael.getProperty("ACTIVE"),"Y")){
                    XMLcmd.addScaPar("Enable","N");
                    prompt=RemoveIco(prompt);
                }
                XMLcmd.addScaPar("Label",prompt);
            }
          } else if (like(AV,"SQL:%")) { // в DISPLAY размножитель SELECT
          ArrayList<String> ac=app.getAppConnAndStmt(AV.substring(4), frm.getProperty("Connection"),this.frm.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          if (propsArray!=null){
          // крутим извлечённые строки
          for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                ael.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=ael.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
            if (!compareString(ael.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }  
          } else if (like(AV,"JS:%"))  { // в DISPLAY размножитель JS
            try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.app.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного массива
           ael.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          String Prompt=ael.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
            if (!compareString(ael.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          }  
          } // JS конец
      }
    }
   }
   }
   //Пункты поддержки многоязычности
   //???????
   //Чисто клиентские
  //Их генерирует сервер отображения
 //дадим ему это делать если в уже сгенерированных меню нет EDITOR$%
 if (!is_edtfld && NVL(app.getProperty("FIELD_RMB_DEFAULT_MENU"),"Y").equals("Y")){
   XMLcmd=XMLcg.addCommand("AddLocalItems");
   XMLcmd.addScaPar("Group", "Last");
 }
}
/** 
 * Обработка нажатия ПКМ в заголовке группы столбцов списка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса 
 */

public void groupRMB() throws LuiException, SQLException{
    if (compareString(frm.getVal("Is_QBE_Mode"),"Y")) return; // если щас QBE - уходим 
    app.changeSCN("GroupRMB"); // меняем SCN
    frm.setVal("GroupMenu", getCode());
    LuiClCommandGroup XMLcg=app.commandWriter.addGroup("GroupRMB");
    XMLcg.addParameter("FormID",frm.getFormID());
    LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
    // крутим действия уровня формы или столбца с кодом "GROUPMENU$%"
    for (int pcfi=1;pcfi<3;pcfi++) {
    String parent="Form";
    if (pcfi==1)parent=this.Code;
    for (int i=0;i<frm.elements.size();i++){
     LuiAppElement el=frm.elSort(i);
     if (!like(el.getCode(),"GRPMENU$%"))continue;
     if (!el.getParent().equals(parent))continue;
     String AV=el.getProperty("DISPLAY");    
     if (AV.equals("N"))continue;
     if (AV.equals("Y")) {
         XMLcmd=XMLcg.addCommand("AddMenuItem");
         XMLcmd.addScaPar("Type", "ACTION");
         XMLcmd.addScaPar("Code", el.getCode());  
         String Prompt=el.getProperty("PROMPT");
         //XMLcmd.addScaPar("Group",el.getParent()/*+String.valueOf(Double.valueOf(el.getProperty("Seq"))/10)*/); //String.valueOf(el.sourceSeq/10));
         if (!el.isActionActive()){
          XMLcmd.addScaPar("Enable","N");   
          Prompt=RemoveIco(Prompt);
       }
         XMLcmd.addScaPar("Label",Prompt);
         } 
    }
    }
}
/** 
 * Обработка нажатия ПКМ в заголовке столбца списка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса 
 */
public void columnRMB() throws LuiException, SQLException{
  frm.setVal("ColumnMenu", getCode()); // установим Var ColumnMenu для вычисления Parent действий COLMENU%
  
  String tstr=frm.getVisualPref(getCode(),"ColWidth", ""); // читаем сохранённую пользовательскоую ширину столбца
  if (is_not_null(tstr)){
    setProperty("UserColumnWidth",tstr);  // и установим св-во UserColumnWidth этого элемента для вычисления активности действий COLMENU%
  } else {
    setProperty("UserColumnWidth",""); // если пользовательская ширина не установлена
  }
  LuiCollection gc=new LuiCollection(""); // создаём для элемента коллекцию возможных групп
  int ovfGroupCount=0;
  int tableGroupCount=0;
  int fixGroupCount=0;
  for (int i=0;i<frm.elements.size();i++){
    LuiAppElement el=frm.elSort(i);
   if (el.getType().equals("GROUP")) {
     if (el.getProperty("GROUP_TYPE").equals("OVERFLOW")) ovfGroupCount++;   
     if (el.getProperty("GROUP_TYPE").equals("FIXED")) fixGroupCount++;   
     if (NVL(el.getProperty("GROUP_TYPE"),"Form").equals("Form")) tableGroupCount++; 
     if (el.getCode().equals(this.getParent())) continue; // если наш столбец уже в этой группе пропускаем
    gc.Add_Element(el.getCode(), NVL(el.getProperty("TITLE"),el.getCode())); // добавляем группу в коллекцию групп
   } 
  }
  if (ovfGroupCount==0 && !getParent().equals("OVERFLOW")){
      gc.Add_Element("OVERFLOW", app.NLS.NLS("RUS:\"Переполнение\";ENG:Overflow zone;")); // добавляем OVERFLOW группу в коллекцию групп
  }
  if (fixGroupCount==0 && !getParent().equals("FIXED")){
      gc.Add_Element("FIXED", app.NLS.NLS("RUS:\"Нет горизонтальной прокрутки\";ENG:Not scrollable columns;")); // добавляем OVERFLOW группу в коллекцию групп
  }
  if (tableGroupCount==0 && !getParent().equals("Form")){
      gc.Add_Element("Form", app.NLS.NLS("RUS:\"Табличная часть\";ENG:Grid zone;")); // добавляем Табличную часть в коллекцию групп
  }
  // устанавливаем свойство 
  setProperty("GroupsToMove",gc.toString());
  
  app.changeSCN("ColumnRMB"); // меняем SCN
  //if (compareString(frm.getVal("Is_QBE_Mode"),"Y")) return; // если щас QBE - уходим 
    LuiClCommandGroup XMLcg=app.commandWriter.addGroup("ColumnRMB");
    XMLcg.addParameter("FormID",frm.getFormID());
    
    LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
    // крутим действия уровня формы или столбца с кодом "COLMENU$%"
    for (int pcfi=1;pcfi<3;pcfi++) {
    String parent="Form";
    if (pcfi==1)parent=this.Code;
    for (int i=0;i<frm.elements.size();i++){
     LuiAppElement el=frm.elSort(i);
     if (el.getType().equals("ACTION")){
         if (like(el.getCode(),"COLMENU$%")) {
         String ep=el.getParent();
         if (el.getParent().equals(parent)) {
         String AV=el.getProperty("DISPLAY");    
         if (AV.equals("Y")) {
         XMLcmd=XMLcg.addCommand("AddMenuItem");
         XMLcmd.addScaPar("Type", "ACTION");
         XMLcmd.addScaPar("Code", el.getCode());  
         String Prompt=el.getProperty("PROMPT");
         XMLcmd.addScaPar("Group",el.getParent()/*+String.valueOf(Double.valueOf(el.getProperty("Seq"))/10)*/); //String.valueOf(el.sourceSeq/10));
         if (!el.isActionActive()){
          XMLcmd.addScaPar("Enable","N");   
          Prompt=RemoveIco(Prompt);
       }
         XMLcmd.addScaPar("Label",Prompt);
         } 
         if (like(AV,"SQL:%")) { // в DISPLAY лежит SELECT
             ArrayList<String> ac=app.getAppConnAndStmt(AV.substring(4), ownerObj.getProperty("Connection"),this.frm.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          // крутим извлечённые строки
          if (propsArray!=null){
              for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                el.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=el.getProperty("PROMPT");
            
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
            if (!el.isActionActive()){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }  
         }// в DISPLAY лежит SELECT
         
         
        }   
       }
     }
    }
}
    /*XMLcmd=XMLcg.addCommand("AddMenuItem");
    XMLcmd.addScaPar("Type", "Sort");
    XMLcmd.addScaPar("Code", getCode());*/
    
    
    
}
    /**
     * Обработка нажатия ПКМ в ячейке сипска 
     *
     * @param pRow - номер строки
     * @param pContext -контекст коллекция
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
/*public void cellRMB(String pRow,String pContext) throws LuiException, SQLException{
  
  app.changeSCN("CellRMB");
 // if (!compareString(frm.getVal("Is_QBE_Mode"),"Y")){ // если щас не QBE
  LuiCollection Context=new LuiCollection(pContext);
  LuiCollection CurrentContext = null;
  Integer Row=null;
  int RN=Integer.valueOf(NVL(frm.getVal("CurrentRow"),"0"));
    LuiClCommandGroup XMLcg=app.commandWriter.addGroup("CellRMB");
    XMLcg.addParameter("FormID",frm.getFormID());
    XMLcg.addParameter("Type", "Action");
    if (is_not_null(pRow)){
    Row=Integer.valueOf(NVL(pRow,"0"));
    }
    if (Row==0){//клик в области грида, где нет строк
        CurrentContext=frm.getFieldsValues();
        frm.setVal("CurrentRow","");
        frm.clearElementsValue();
        frm.clearChecking();
    } else {
        frm.setElementsValues(Context);
    }
    //Готовим список действий 
    ArrayList<LuiAppElement> actionListCol=new ArrayList<>();
    ArrayList<LuiAppElement> actionListRow=new ArrayList<>();
    LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
    String AV;
    String parent;
    LuiAppElement ael;
    for (int i=0;i<frm.elements.size();i++) {
        ael=frm.elSort(i);
        //if (compareString(frm.getVal("Is_QBE_Mode"),"Y") && !like(ael.getCode(),"CELLMENU$%")) continue;
        if (compareString(frm.getVal("Is_QBE_Mode"),"Y") && !like(ael.getCode(),"QBE$ITEM%")) continue;
        if (ael.getType().equals("ACTION") && !like(ael.getCode(),"SYSMENU$%") && !like(ael.getCode(),"COLMENU$%")){  
          parent=ael.getParent();  
         if ((parent.equals("Form") && (Row>0 ||Row==null))
                 ||parent.equals(getCode())){ 
         if (ael.getProperty("Run").equals("Y")) {
          AV=ael.getProperty("DISPLAY");
          if (AV.equals("Y") || (AV.equals("EXIST") && Row>0) || like(AV,"SQL:%") || like (AV,"JS:%")){
            if (is_not_null(ael.getProperty("PROMPT"))){  
             if (parent.equals("Form")){    
               actionListRow.add(ael);
             } else {
               actionListCol.add(ael);  
             } 
            }
          }  
         }
        }
    }
    }
    // слияние
    for (int i=0;i<actionListRow.size();i++) actionListCol.add(actionListRow.get(i));
    // изготовление команд
    for (int i=0;i<actionListCol.size();i++){
       ael=actionListCol.get(i);
       AV=ael.getProperty("DISPLAY");
       if (like(AV,"SQL:%")) { // в DISPLAY лежит SELECT
          ArrayList<String> ac=app.getAppConnAndStmt(AV.substring(4), ownerObj.getProperty("Connection"),this.frm.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          // крутим извлечённые строки
          if (propsArray!=null){
          for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                ael.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=ael.getProperty("PROMPT");
            
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
            if (!ael.isActionActive()){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }
       } else if (like(AV,"JS:%")) { // JSON-размножитель
          try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.app.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного массива
           ael.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          String Prompt=ael.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
                if (!ael.isActionActive()){
            XMLcmd.addScaPar("Enable","N");   
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          }
        } // JSON
         else{ // не Select и не JS
       String Prompt=ael.getProperty("PROMPT");           
       if (is_not_null(Prompt)) {
       XMLcmd=XMLcg.addCommand("AddMenuItem");
       XMLcmd.addScaPar("Code",ael.getCode());
       XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
       if (!ael.isActionActive()){
       XMLcmd.addScaPar("Enable","N");   
       Prompt=RemoveIco(Prompt);
       }
       XMLcmd.addScaPar("Label",Prompt);
       }
       }
    }
    //Чисто клиентские пункты
      XMLcmd=XMLcg.addCommand("AddLocalItems");
      XMLcmd.addScaPar("Group","Local");
      frm.setVal("CurrentRow",String.valueOf(RN));
      if (Row==0){
          frm.setElementsValues(CurrentContext);
        if (RN>0){
            frm.checkRow("Y",frm.getUK());
        }  
      }
  //} 
}
*/
/**
 * Помещение коллекции значений столбца списка в ListQBE
 * коллекция должна представлять собой набор элементов вида номер:значение
 * если передана пустая строка - выполняется попытка взять коллекцию из свойства формы LOV_MULTI
*/
public void collection2QBE(String pCol) throws LuiException, SQLException{
    if (is_null(pCol)){
        if (is_not_null(frm.getProperty("MULTI_LOV"))){
            pCol=frm.getProperty("MULTI_LOV");
        }
    }
    if (is_null(pCol))return;
    if (!frm.getVal("Is_QBE_Mode").equals("Y"))return;
    String currentRow=frm.getVal("CurrentRow");
    int cr=Integer.valueOf(NVL(currentRow,"1"))-1;
    LuiCollection col=new LuiCollection(pCol);
    for (String rn:col.keySet()){
        currentRow=String.valueOf(cr+Integer.valueOf(rn));
        HashMap<String,String> rowQBE; 
        if (!frm.listQBE.containsKey(currentRow)){
            rowQBE=new HashMap<>();
            frm.listQBE.put(currentRow, rowQBE);
              } else {
                 rowQBE=frm.listQBE.get(currentRow);
              }
        if (rowQBE.containsKey(getCode())) {
                  rowQBE.replace(getCode(),col.Get_Element(rn));
              } else {
                  rowQBE.put(getCode(), col.Get_Element(rn));
              }
    }
}
/** обработка изменений введённых в поле руками пользователя
 * 
 * @param pValue -- новый текст в поле null если текст в поле не меняется а перемещается курсор
 * @param pStartCursor - позиция курсора выделения
 * @param pEndCursor - конец позиции курсора выделения
 * @param pIsCheckFormat - следует ли проверять формат
 * @param pCheckChanging - true - проверять возможность ввода руками
 * @param pFromClient - true - прилетело от клиента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void newValue(String pValue,String pStartCursor,String pEndCursor,String pIsCheckFormat,boolean pCheckChanging,boolean pFromClient) throws LuiException, SQLException{
    if (pValue==null && "Y".equals(pIsCheckFormat)) pValue=getValue();
    if (pValue==null && is_null(pStartCursor) && is_null(pEndCursor) ) return; // это если ничего не изменилось и не надо проверять формат
    if (pValue!=null && pFromClient && this.properties.containsKey("Var OldValueFrom")){ 
        // если значение есть и оно от браузера пришло и есть свойство Var OldValueFrom (слайдерное)
        // запишем его в этот Old чтобы не слать зря обратно
        setVal("OldValueFrom",pValue);
    }
    if (getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR")){ //если это цвет
        if (pValue!=null && pValue.replace(" ","").equals("rgba(255,255,255,0)") && getProperty("OPTIONAL").equals("Y")) // и передан белый прозрачный и поле допускает пустоту
        pValue=""; // то вернём пустоту
    }
    if (pIsCheckFormat.equals("Y")){
        app.changeSCN("NewValue");
    }else{
        app.changeSCN("EditValue");
    }
    if (pValue==null &&(is_not_null(pStartCursor) || is_not_null(pEndCursor))) {//это если поменялась только позиция курсора
    if (is_not_null(pStartCursor)) {setProperty("StartCursor",pStartCursor);if (pFromClient) properties.get("StartCursor").changeFlag=false;}
    if (is_not_null(pEndCursor))  {setProperty("EndCursor",pEndCursor);if (pFromClient) properties.get("EndCursor").changeFlag=false;}
     frm.checkFrmDetails();
     return;
    }
    
    if (is_not_null(pStartCursor)) {setProperty("StartCursor",pStartCursor);if (pFromClient) properties.get("StartCursor").changeFlag=false;}
    if (is_not_null(pEndCursor))  {setProperty("EndCursor",pEndCursor);if (pFromClient) properties.get("EndCursor").changeFlag=false;}
    if (!pFromClient && pValue!=null){
        int ic=pValue.indexOf(CURSORSIGN);
        if (ic>-1){
           setProperty("StartCursor",String.valueOf(ic));
           setProperty("EndCursor",String.valueOf(ic));
           pValue=pValue.replace(CURSORSIGN,"");
                  }
    }
    if (pFromClient && pValue!=null) {
           setProperty("Show",pValue);properties.get("Show").changeFlag=false;
                            }
    String FType=frm.getType();
    switch (FType){
        case "LIST":
        case "TREE":
            if (frm.getVal("Is_QBE_Mode").equals("Y")){
              HashMap<String,String> rowQBE;  
              String currentRow=frm.getVal("CurrentRow");
              if (!frm.listQBE.containsKey(currentRow)){
                 rowQBE=new HashMap<>();
                 frm.listQBE.put(currentRow, rowQBE);
              } else {
                 rowQBE=frm.listQBE.get(currentRow);
              }
              if (rowQBE.containsKey(getCode())) {
                  rowQBE.replace(getCode(), pValue);
              } else {
                  rowQBE.put(getCode(), pValue);
              }
            } else
            {
              this.addProperty("OldValue",this.properties.get("Value").currentValue);
              setValue(pValue,true, pIsCheckFormat,pCheckChanging);
              String actionCode=frm.checkHotKey("POSTEDIT", getCode());
              if (is_not_null(actionCode)){ // если есть действие c hjkm. POSTEDIT
                  //Ui_p_Service.Run_Action(FormID, ActionCode);
                  frm.runAction(actionCode);
              }
            }
        break;
        case "BLANK":
        case "CHART":
        String curVal=getValue();
        if (getProperty("CHANGING").equals("B")){ // CheckBox
        if (pFromClient){                                          // обработка револьверного значения если это пришло от клиента
         String variants=getProperty("VARIANTS");
         if (variants.indexOf("LOVQ:")>-1){// варианты извлекаются запросом
             
         } else 
         if (variants.indexOf("LOVD:")>-1){// варианты извлекаются из простой коллекции
           LuiCollection vc=new LuiCollection(variants.substring(5)); // создаём коллекцию
           String nv=null; // новое значение неизвестно
           Boolean b=false; // следует ли брать ключ на очердной итерации
           for (String k:vc.keySet()){// крутим элементы
             if (b) { // если надо брать 
                nv=k; // берём и уходим
                break;
             }  
             if (k.equals(curVal)) b=true; // попали на текущее значение поля - установми признак взятия на следующей итерации
           }
           if (nv==null) { // на первом проходе не найдено следующее значение
           for (String k:vc.keySet()){// опять крутим элементы    
                nv=k; // берём первый и уходим
                break;    
           }
           }
           setValue(nv,true, pIsCheckFormat,pCheckChanging);// устанавливаем значение следующее за текущим
         }
        } else { 
           setValue(pValue,true, pIsCheckFormat,pCheckChanging);        // если CHECKBOX но не от клиента 
        }
        } else { // если не checkbox
           setValue(pValue,true, pIsCheckFormat,pCheckChanging);           
        }
        if (!compareString(curVal,pValue)) { // если старое значение не равно новому
            String actionCode=frm.checkHotKey("POSTEDIT", getCode());
            if (is_not_null(actionCode)){ // если есть действие c hjkm. POSTEDIT
              String oldCurrentField=frm.getVal("CurrentField"); // запомним какое поле текущее
              //Ui_p_Service.Run_Action(FormID,ActionCode);
              frm.runAction(actionCode);
              String newCurrentField=frm.getVal("CurrentField"); // какое поле текущее теперь?
              if (!compareString(newCurrentField,oldCurrentField)){ // если текущее стало другое
                  LuiClCommandGroup XMLcg=getApplication().commandWriter.addGroup("RestoreCurrentField");
                  LuiClCommand XMLcmd=XMLcg.addCommand("CurrentField");
                  XMLcmd.addScaParameter("FormID", frm.getFormID(),"");
                  XMLcmd.addScaParameter("Field",oldCurrentField ,"");
              }
            }
        }
        if (is_null(pIsCheckFormat) || !pIsCheckFormat.equals("Y")){ // если нет проверки формата
           if (getProperty("CHANGING").equals("Y")) { // и если разрешён ручной ввод
             String LS=getProperty("LIVESEARCH"); // читаем св-во LIVESEARCH
             if (is_not_null(LS)) { // и  если оно не пусто
                 // там может быть : SQL(CONN):select value from ....
                 ArrayList<String> acstmt=app.getAppConnAndStmt(LS, null,this.frm.getFormID());
                 String acc=acstmt.get(0);
                 String stmt=acstmt.get(1);
                 LuiAppConn ac=app.userDbConnections.get(acc);
                 LuiDbConn lc=null;
                 Integer cursor=null;
                 LuiCollection values = new LuiCollection("");
                 String v;
                 try {
                    if (ac.getSessionMode()=='C'){
                        lc=ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                    }
                        else lc=ac.currentConnection;
 
                    cursor=lc.CursorOpen(stmt);
                    while (lc.CursorNextRow(cursor)){
                        v=lc.CursorGetRow(cursor).get(0);
                        values.Add_Element(v,v);
                     }
                   //if (values.size()>0) {
                     LuiClCommandGroup XMLcg=getApplication().commandWriter.addGroup("LiveSearch");  
                     LuiClCommand XMLcmd=XMLcg.addCommand("DropDown");
                     XMLcmd.addScaPar("FormID",frm.getFormID());
                     XMLcmd.addScaPar("Field",getCode());
                     XMLcmd.addScaPar("Contents",values.toString());
                   // }
                     
                 } catch (SQLException ex ){
                     app.msgWriter.sqlError(lc, ex, properties.get("LIVESEARCH"), stmt);
                 }
                 finally {
                     if (ac.getSessionMode()=='C'){
                        if (lc!=null) {
                            if (cursor!=null){
                                if (lc.CursorisOpen(cursor)){
                                    lc.CursorClose(cursor);
                                }
                            }
                           if (ac.getSessionMode()=='C'){
                              ac.connPool.SetFree(lc);
                           }
                          } 
                     }
                 }
             }
           }
            
        }
        break;
    }
    if (pIsCheckFormat.equals("Y") || is_not_null(pStartCursor) || is_not_null(pEndCursor)){
         frm.checkFrmDetails();
    }
}
/** 
 * вычисляет количество строк для поля, которое надо отправить на сервер визуализации
 * @return Количество строк. По необъяснимиым причинам - в виде строки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getLines4view() throws LuiException, SQLException{
  //return LuiString.String2Integer(getProperty("VISIBLE").replaceFirst("^Y$", "1").split("\\d+(?:\\/\\d)*")[0], 0).toString();  
  String v=getProperty("VISIBLE");
  if (like(v,"%/%")) return v;
  if (v.equals("N")) return "0";
  if (v.equals("Y")) return "1";
  if (like(v,"Y%")) return "0";
  return NVL(getProperty("VISIBLE"),"0").replaceFirst("^Y$", "1");  
}
/**
 * Нажатие стрелки вниз в поле с LOVD
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса 
 */
public void downArrayBar() throws LuiException, SQLException{
    if (like(getProperty("VARIANTS"),"LOVD%")){
       String action=frm.checkHotKey("F9", getCode());
       if (action!=null){
           frm.runAction(action);
       }
    }
}
/**
 * сбрасывает флаг изменённости
 */
public void resetChangeFlag(){
    if (this.changeFlag){
     for (String pkey:properties.keySet()){
       properties.get(pkey).changeFlag=false;
     }   
     this.changeFlag=false;    
    }
}
@Override
public String getGenesis4dmp(){
    try {    
    //return frm.getGenesis4dmp()+" "+getType()+":"+getCode()+";";
    return getType()+":"+getCode();
            } catch (LuiException ex) {
 return "";
               } catch (SQLException ex) {
 return "";                  
               }
}

@Override
public String getGenesis(){
               try {    
String parent =this.getParent();                  
 if ("Form".equals(parent)) {
               return ownerObj.getGenesis()+" "+getType()+":"+getCode()+";";
 } 
 if (!frm.elements.containsKey(parent)){
               return ownerObj.getGenesis()+" "+getType()+":"+getCode()+";";
 }
 parent=frm.elements.get(parent).getGenesis();
 return parent+" "+getType()+":"+getCode()+";";
               } catch (LuiException ex) {
 return "";
               } catch (SQLException ex) {
 return "";                  
               }
}
@Override
    public LuiAppElement getParentElement(String pElType) throws LuiException, SQLException{
     if (getType().equals(pElType)){
         return this;
     } else {
        String p=getParent(); 
        if (!"Form".equals(p)) {
            if (frm.elements.containsKey(p)){
                return frm.elements.get(p).getParentElement(pElType);
            }
        }
     }
     return null;
  }

@Override
public LuiAppForm getForm(){
 return frm;
}
/**
 * Установка значения или добавление св-ва элемента
 * Переопределено с супер класса для обеспечения обработки особых св-в
 * @param pCode -- код св-ва
 * @param pValue-- зн-е св-ва
 * @return объёкт "свойство"
 */
@Override
public LuiAppProp addProperty(String pCode,String pValue){
 if (pCode.equals("SORT") && is_not_null(pValue) && !frm.getVal("Status").equals("Loading")) {
  try {    
 Integer newsort=Integer.valueOf(pValue);
 if (newsort!=null){
    for (String ec:frm.elements.keySet()) {
     LuiAppElement el=frm.elements.get(ec);
     if (!getCode().equals(ec)) {
     if (el.getType().equals("ITEM")) {
       Integer elsort;
       //включим подавление ошибки
          /* String safe=null;
           if (frm.properties.containsKey("SuppressErrors")){
             safe=frm.getProperty("SuppressErrors");
           }
           if (safe==null || !safe.equals("Y")) {
           frm.setProperty("SuppressErrors","Y");
           }
       */
       try{
           elsort=Integer.valueOf(el.getProperty("SORT"));
       } catch (Exception iex){
           elsort=null;
           app.msgWriter.ClearMsgArray();
       } finally {
          // if (safe!=null) frm.setProperty("SuppressErrors",safe);
       }
       
       if (elsort!=null){
         if (Math.abs(newsort)<=Math.abs(elsort)) {
           if (elsort<0) {
               elsort--;
           }  else {
               elsort++;
           }
           //el.properties.get("SORT").currentValue=String.valueOf(elsort);
           el.addProperty("SORT",String.valueOf(elsort));
          break;
         }  
       }
     }
   }
   }
 return super.addProperty(pCode,pValue);
 
 
 } else
  return super.addProperty(pCode, pValue);
  
 } catch (Exception ex){
     return super.addProperty(pCode, pValue);
 }
}
if (pCode.equals("UserColumnWidth")) {
     try {
         frm.changeVisualParam(getCode(), "ColWidth", pValue);
         LuiClCommandGroup XMLcg=frm.application.commandWriter.addGroup("ColWidth");
         LuiClCommand XMLcmd=XMLcg.addCommand("Column");
         XMLcmd.addScaPar("FormID",frm.getFormID());
         XMLcmd.addScaPar("Code",getCode());
         XMLcmd.addScaPar("Width",pValue);
     } catch (Exception ex) {
         return super.addProperty(pCode, pValue);
     }
  return super.addProperty(pCode, pValue);
} 
try{
if (pCode.equals("Parent") && frm.getType().equals("LIST") && getType().equals("ITEM")){
 if (properties.containsKey(pCode)) { // такое св-во есть
   String pv=getProperty(pCode);
   if (is_null(pv)){ // если оно пусто - уходим т.к. наверно это создание элемента идёт
     return super.addProperty(pCode, pValue);  
   }
   if (pv.equals(pValue)) { // если оно совпадает с требуемым - тоже уходим
     return super.addProperty(pCode, pValue);    
   }
 } else { // если нет и столбец пихают  просто в форму
   if ("Form".equals(pValue)){
     return super.addProperty(pCode, pValue);      
   }  
 }
 LuiAppElement newgrp;
  if (!frm.elements.containsKey(pValue) && !pValue.equals("Form")) { // если гуппы нет - добавим
      newgrp=frm.addElement(pValue,"GROUP", null, pValue, "Form");
      newgrp.addProperty("GROUP_TYPE", pValue);
      newgrp.addProperty("POSITION", "UP");
      newgrp.addProperty("GROUP_NOTES", "");
      //frm.makeElList("Form");
      if (is_not_null(frm.getVal("CurrentAction"))) { // если добавление группы идёт по действию пользователя то добавим команду создания группы на экране
      LuiClCommandGroup XMLcg=frm.application.commandWriter.addGroup("Add grpou by end user");
      LuiClCommand XMLcmd=XMLcg.addCommand("Group");
      XMLcmd.addScaPar("FormID", frm.getFormID());
      XMLcmd.addScaPar("Code", pValue);
      XMLcmd.addScaPar("Type", LuiString.withCapitalLetter(pValue));
      XMLcmd.addScaPar("State", "Visible");
      XMLcmd.addScaPar("FieldsPromptPosition", "Up");
      }
  }
  if (is_not_null(frm.getVal("CurrentAction"))) { // если перемещение столбца идёт по действию пользователя
  LuiClCommandGroup XMLcg=frm.application.commandWriter.addGroup("Col2Grp");
  LuiClCommand XMLcmd=XMLcg.addCommand("Column");
  XMLcmd.addScaPar("FormID", frm.getFormID());
  XMLcmd.addScaPar("Code", getCode());
  if (NVL(pValue,"Form").equals("Form")) {
  XMLcmd.addScaPar("Group", "");    
  } else {
  XMLcmd.addScaPar("Group", pValue);
  }
  frm.changeVisualParam(getCode(), pCode, pValue);
  }
  return super.addProperty(pCode, pValue);    
}
}
catch (Exception ex){
  return super.addProperty(pCode, pValue);  
}
if ("UserChanged".equals(pCode)&& "Y".equals(pValue)){
ownerObj.setVal("Changed", "Y"); // метим форму изменённой руками
}
  return super.addProperty(pCode, pValue);
}
/**
 * Установка признаков необходимости вычислений для всех свойств элемента
 */
public void clearComputedValues(){
    for (String key:properties.keySet()){
       if  (properties.get(key).recalculate!=null) {
       if (properties.get(key).recalculate==false) {
           properties.get(key).clearComputedProp();
       }    
    }
    }
   app.dependsCount++; 
}
    /**
     * При отсутствии свойтсва basePropertyCode у действия создаёт зависимости
     * от него у всех других динамических свойств этого действия и подчинённых
     * выходных параметров. Безусловно устанавливает значение newValue в
     * свойство BasePropertyCode
     *
     * @param basePropertyCode код свойства
     * @param newValue значение свойства
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void makeRunSCN(String basePropertyCode,String newValue) throws LuiException, SQLException{
  if (!properties.containsKey(basePropertyCode)) {
    LuiAppProp bp=setProperty(basePropertyCode,newValue);
    // для св-в самого действия
    for (String pc:properties.keySet()){
     LuiAppProp p=properties.get(pc);
     if (p.recalculate!=null) {
       if (p.recalculate){
           bp.setDepend(p);
       }
     }
    }
    // для выходных параметров
    for (String apc:frm.elements.keySet()){
      if (!frm.elements.get(apc).getType().equals("PARAM")) continue;
      if (!frm.elements.get(apc).getParent().equals(getCode())) continue;
      LuiAppElement par=frm.elements.get(apc);
      for (String pc:par.properties.keySet()){
     LuiAppProp p=par.properties.get(pc);
     if (p.recalculate!=null) {
       if (p.recalculate){
           bp.setDepend(p);
       }
     }
    }
    }
  }
  else {
   setProperty(basePropertyCode,newValue);   
  }  
}
    /**
     *
     * @param lc -- соединение с БД
     * @param colVal -- критерий на языке QBE
     * @return какое-то условие
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public String getSQLCondition(LuiDbConn lc,String colVal,String pInnerQBE) throws LuiException, SQLException {
  LuiDbConn llc=null;
  LuiAppConn      ac=null;
    if (lc==null){
      String cc=ownerObj.getProperty("Connection"); // выясняем код соединения
                 if (is_null(cc)) {// если у бланка нет соединения поищем у приложения
                   for(String ccc:getApplication().userDbConnections.keySet()){
                       cc=ccc;
                       break;
                   }
                 }
                 if (is_not_null(cc)) { //если  не пуст
                    if (getApplication().userDbConnections.containsKey(cc)) { // и такое соединение есть
                      ac=getApplication().userDbConnections.get(cc);
                     if (ac.getSessionMode()=='C') { // если соединение коммунальное, то получим из пула свободное
                       llc=ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                     }
                    else {//в противном случае возьмём текущее 
                       llc=ac.currentConnection;
                        }  
                    } 
                 }
                 if (llc==null){ // если соединение так и не нашлось - возьмём из пула соединений к мета данным
                     llc=getApplication().metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                 }
  lc=llc;
    }
try{  
  if (colVal==null){
      //colVal=getProperty("Show");
              HashMap<String,String> rowQBE;  
              String currentRow=frm.getVal("CurrentRow");
              if (frm.listQBE.containsKey(currentRow)){
                 rowQBE=frm.listQBE.get(currentRow);
                 if (rowQBE.containsKey(getCode())) {
                  colVal=rowQBE.get(getCode());
                 }
              }
    if (colVal==null)  colVal=getProperty("Show");            
  }
  if (this.isWinFunc() && is_not_null(getProperty("FORMAT"))){ // если на оконную ф-ю наложен формат
     String wfc="";
     if (pInnerQBE.equals("Y") && is_not_null(getProperty("INNER_COLUMN"))) wfc=getProperty("INNER_COLUMN");
     else {
     String wfC="\""+this.Code+"\"";    
     wfc= app.dataConverter.getFromLuiConversion(wfC, this, lc);
     wfc=wfc.replace("'"+wfC+"'", wfC);
     }
     
     addProperty("QBEcolumn",wfc);
  } else{
  //if (!this.isWinFunc() && (pInnerQBE.equals("Y") || getProperty("Sintetic").equals("Y"))) addProperty("QBEcolumn","("+NVL(getProperty("INNER_COLUMN"),getProperty("DEFINITION"))+")");
  if ((!this.isWinFunc() || is_not_null(getProperty("INNER_DEFINITION")) ) && (pInnerQBE.equals("Y") || getProperty("Sintetic").equals("Y"))) addProperty("QBEcolumn","("+NVL(getProperty("INNER_DEFINITION"),getProperty("DEFINITION"))+")");
  else addProperty("QBEcolumn","\""+getCode()+"\"");
}
  addProperty("QBEcolumn2str",app.dataConverter.getToLuiConversion(this,lc,pInnerQBE,true));
  if ("Y".equals(app.dataConverter.LuiDataTypes.get(getProperty("DATA_TYPE")).tl_where_clause)){ // если столбец всегда надо конвертировать в строку для создания условия
  addProperty("QBEcolumn",getProperty("QBEcolumn2str"));
  }
  String V=colVal;
  if ((V+" ").substring(0,1).equals("!") && V.length()>1) {
      addProperty("QBEnot","not "); 
      addProperty("QBEnotSign","!"); 
      V=colVal.substring(1);
      addProperty("v",V);
  }
  else {
      addProperty("QBEnot","");
      addProperty("QBEnotSign",""); 
      addProperty("v",V);
  }
  addProperty("QBEess","");
  addProperty("QBEnss","");
  LuiDataType ldt=app.dataConverter.LuiDataTypes.get(getProperty("DATA_TYPE"));
  if (ldt.emptyStringFromLuiMode.equals("NULL")) 
  {addProperty("QBEess",ldt.emptyStringSpecialSymbol);}
  else
  {addProperty("QBEnss",ldt.nullSpecialSymbol);    }
  addProperty("QBEsign","");
  addProperty("QBEoperator","");
  addProperty("QBEcriteria","");
  addProperty("QBEcriteria2 ","");
  addProperty("QBEcriteriaRaw","");
  addProperty("QBEcriteriaRaw2","");
  addProperty("QBESQLcondition","");
  if (is_null(colVal)) return "";
  Integer op_index=null;
  LuiCollection compOpCol=null;
  for (int i=0;i<ldt.compOp.size();i++){ // крутим операторы по порядку
   compOpCol=ldt.compOp.get(i);
   if (!compOpCol.Get_Element("dbms_type").equals("any") && !compOpCol.Get_Element("dbms_type").equals(lc.getDBvendor())) continue;
   if (is_null(compOpCol.Get_Element("markov_exp"))) continue;
   addProperty("QBEsign",compOpCol.Get_Element("sign"));
   addProperty("QBEmarkov_exp",compOpCol.Get_Element("markov_exp"));
   properties.get("QBEmarkov_exp").clearCompVal();
   if (!"Y".equals(getProperty("QBEmarkov_exp"))) continue;
   op_index=i;break;
  }
  if (op_index==null) {
        int s;
        Integer si=null;
   for (int i=0;i<ldt.compOp.size();i++){ // крутим операторы по порядку
       compOpCol=ldt.compOp.get(i);
   if (!compOpCol.Get_Element("dbms_type").equals("any") && !compOpCol.Get_Element("dbms_type").equals(lc.getDBvendor())) continue;
       s=Integer.valueOf(ldt.compOp.get(i).Get_Element("seq"));
     if (si==null) {op_index=i;si=s; continue;}
     if (si>s) {op_index=i;si=s; continue;}
   }   
  
  if (op_index==null) { // если так и не нашли
      addProperty("QBEciteriaRaw",V);
      V="'"+V.replace("'","''")+"'";
      addProperty("QBEciteria",V);
      addProperty("QBEoperator","=");
      addProperty("QBEsign","=");
      addProperty("QBESQLcondition",getProperty("QBEcolumn2str")+"="+V);
      return getProperty("QBESQLcondition");
  } else { // нашли умолчание, теперь возможно надо в критерий добавить знак операции
      compOpCol=ldt.compOp.get(op_index);
      String qbe_constructor=compOpCol.Get_Element("qbe_constructor"); // берём QBE-конструктор
      if (like(qbe_constructor,"{P:QBEsign}%") || //если конструктор qbe-условия начинается со знака
      like(qbe_constructor,compOpCol.Get_Element("sign")+"%") ||   
      like(qbe_constructor,"{P:QBEnotSign}{P:QBEsign}%") || // или с отрицание+знак
      like(qbe_constructor,"{P:QBEnotSign}"+compOpCol.Get_Element("sign")+"%") 
          ) {
          if (!like(V,compOpCol.Get_Element("sign")+"%")) { // и V не начинается с этого, то
              V=compOpCol.Get_Element("sign")+V;
          }
            }
  }
  }
  compOpCol=ldt.compOp.get(op_index);
  addProperty("QBEsign",compOpCol.Get_Element("sign"));
  
  addProperty("QBEoperator",compOpCol.Get_Element("operator_code"));
  if (like(compOpCol.Get_Element("predicate_constructor"),"%{P:QBEcriteria2}%")||
      like(compOpCol.Get_Element("predicate_constructor"),"%{P:QBEcriteriaRaw2}%")) {
  String V2=LuiString.Extract_Word(V, 2, getProperty("QBEsign"));
  V=LuiString.Extract_Word(V, 1, getProperty("QBEsign"));
  addProperty("QBEcriteriaRaw",V);
  addProperty("QBEcriteriaRaw2",V2);
  try{// пробуем преобразовать V и V2 к целевому типу 
   app.dataConverter.convAndCheck(V, this, lc);
   V=app.dataConverter.getFromLuiConversion(V, this, lc);
   app.dataConverter.convAndCheck(V2, this, lc);
   V2=app.dataConverter.getFromLuiConversion(V2, this, lc);
  } catch (Exception ex){ // неудачное преобразование к целевому типу
   addProperty("QBEcolumn",getProperty("QBEcolumn2str"));   
   V="'"+V.replace("'","''");
   V2="'"+V2.replace("'","''");
  }
  addProperty("QBEcriteria",V);
  addProperty("QBEcriteria2",V2);
  } else {
  // теперь надо очистить V от знака операции
  String qbe_constructor=compOpCol.Get_Element("qbe_constructor"); // берём QBE-конструктор
  if (compOpCol.Get_Element("operator_code").toLowerCase().equals("is null")) {
      V="";
  } else{
  if (like(qbe_constructor,"{P:QBEsign}%") || //если конструктор qbe-условия начинается со знака
      like(qbe_constructor,"\\"+getProperty("QBEsign")+"%") ||   
      like(qbe_constructor,"{P:QBEnotSign}{P:QBEsign}%") || // или с отрицание+знак
      like(qbe_constructor,"{P:QBEnotSign}"+"\\"+getProperty("QBEsign")+"%") 
          ) {
      V=LTrim(V,getProperty("QBEsign")); // то осечём их слева
  }  else {
      if (like(qbe_constructor,"%_{P:QBEsign}%")) // если упоминание знака операции находится в середине,то 
              {
      V=V.replace(getProperty("QBEsign"),""); // заменим на пустую строку
      } else {
          // иначе ни чего не делаем
      } 
  } 
  }
  addProperty("QBEcriteriaRaw",V);
  try{// пробуем преобразовать V к целевому типу 
   V=app.dataConverter.convAndCheck(V, this, lc);
   V=app.dataConverter.getFromLuiConversion(V, this, lc);
  } catch (Exception ex){ // неудачное преобразование к целевому типу
   addProperty("QBEcolumn",getProperty("QBEcolumn2str"));   
   V="'"+V.replace("'","''")+"'";
  }
   addProperty("QBEcriteria",V);
  }   
    addProperty("QBESQLcondition",compOpCol.Get_Element("predicate_constructor"));
   properties.get("QBESQLcondition").clearCompVal();
   return getProperty("QBESQLcondition");
 
}      finally {
        if (ac!=null) {
            if (ac.getSessionMode()=='C') {
                ac.connPool.SetFree(llc);
            }
        } else {
          if (llc!=null){
             getApplication().metaConnPool.SetFree(llc);
          }
        }
      }

}
/**
 * Установка или удаление признака доступности столбца для построения RT-curve
 * Если столбец пригоден, то и весь список станет пригодным
 */
public void setRTproperty() throws LuiException, SQLException{
    boolean b=false;
    if (frm.getType().equals("LIST")){
      if ("ITEM".equals(getType())){  
       if (isVisible(true)){ 
           if ("NUMBER".equals(getProperty("DATA_TYPE"))){
             if (!"Y".equals(getProperty("FOR_UNIQUE_KEY"))){  
               b=true;
             }
           }
       }
    }
    }
if (b){
  setVal("rtEnable","Y");
  frm.setVal("rtEnable","Y");  
} else {
    properties.remove("Var rtEnable");
}
}
@Override
           public LuiAppElement getElement(){
      return this;
  }
/**
 * опредеоение того, имеется ли в столбце оконная функция
 */

public boolean isWinFunc() throws LuiException, SQLException{
  if (getProperty("IS_WIN_FUNC").equals("Y"))return true;
  if (like(getProperty("INNER_DEFINITION").toLowerCase(),"%over%(%)%")) return true;
  if (like(getProperty("DEFINITION").toLowerCase(),"%over%(%)%")) return true;
  return false;  
}
}