/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс для хранения свойства формы или элемента
 */
package luiserver;


import static LuiCollection.LuiString.compareString;
import LuiExceptionMsg.LuiException;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.util.HashMap;

import java.util.regex.Pattern;



/**
 *
 * @author aliubushkin
 */
public class LuiAppProp {
  //public Boolean isDyn=null;
  //public static final String DYNAMICCONDITION=".*\\{.+\\}.*";
  public Pattern DYNAMICCONDITION = Pattern.compile(".*\\{.+?\\}.*",  Pattern.MULTILINE | Pattern.DOTALL);
  public static final String OPENINGBRASKET="{";
  public static final String CLOSINGBRASKET="}";
  //public static final String MASKOPENINGBRASKET=new String(new byte[]{1},Charset.forName("windows-1251"));
  public static final String MASKOPENINGBRASKET=new String(new byte[]{1},Charset.forName("UTF8"));
  //public static final String MASKCLOSINGBRASKET=new String(new byte[]{2},Charset.forName("windows-1251"));
  public static final String MASKCLOSINGBRASKET=new String(new byte[]{2},Charset.forName("UTF8"));
  public Boolean recalculate; // признак необходимости вычисления null-статично,true-надо перевычислять, false-не надо акутально уже
                                   
      public int statComp = 0;          // Количество вычислений динамического значения
      public int statGet = 0;           // Количество обращений за значением
      public int statTime = 0;          // Время всех вычислений в секундах
      boolean changeFlag;        // Признак изменённого значения по сравнению с предыдущей отправкой Жене, сбрасывается после отправки всех накопленных команд
      public String startedValue;       // Значение - шаблон для динамических свойств это то что написано в метаданных
      public String currentValue;       // Значение при последнем обращении (динамика) или статическое значение
      
      public HashMap<LuiAppProp,Integer> depends; // Коллекция свойств элементов, которые зависят от данного свойства
      public HashMap<LuiAppProp,Integer> dependsFrom; // Коллекция свойств элементов, которые определяют данное свойство
      public boolean is_depends=false; //Признак того что это свойство зависит от каких-либо других свойств //06.11.2018
      public LuiAppWFE ownerObj; //ссылка на объект, к которому относится свойство
      public String code; //код свойства
      public char genesis='A';
      //public boolean is_meta=false; //признак "считано из метаданных"
      public boolean gen4menu=false; //признак того, что добавлено в ходе формирования динамических действий формы типа MENU
      public boolean gen4html=false; //признак того, что добавлено для обработки HTML
      LuiApplication app;
/**
 * Конструктор
 * @param pObj объект для которого создаётся свойство
 * @param pCode код свойства
 * @param pValue значение свойства
 */
public LuiAppProp(LuiAppWFE pObj,String pCode,String pValue) { 
        depends = new HashMap<>();
        dependsFrom=new HashMap<>();
        ownerObj=pObj;
        code=pCode;
        setStartedValue(pValue);
        app=ownerObj.getApplication();
      }
public void setCurrentValue(String pValue){
    currentValue=pValue;
}
/** 
 * Изменение начального значения параметра и вычисленного знгачения если начальное значение является статическим
 * @param pValue  значение
 */
public void setStartedValue(String pValue){
    startedValue = pValue;
    //isDyn=null;
    if (isDynamic()){ 
          recalculate=true; // указываем что ещё не вычислялось и при первом обращении надо вычислить
          currentValue="";
        } else {
          recalculate=null;
          currentValue = pValue;
        }
        setChangeFlag();
}

/** 
 * Извлечение текущего значения без перевычисления динамики
 * @return текущее значение
 */
public String getCurrentValue(){
    return currentValue;
}
public String getNonCompValue(){
    if (this.recalculate==null) return currentValue;
    return startedValue;
}
/**
     * Установка признака, что свойство pProp зависит от этого свойства
     * @param pProp зависимое свойство
     */
public void setDepend(LuiAppProp pProp) {
      depends.put(pProp,0);
      pProp.is_depends=true; //06.11.2018
      pProp.dependsFrom.put(this,0);
    }

/**
     * Сброс вычисленного или насильно изменённого значения свойства т.е. делание его "невычисленным" для динамики 
     * (???или просто приведение в исходное состояние для статики???)
     */
    public void clearCompVal() {

       if (this.recalculate!= null) { // Если динамика
      this.recalculate=true;      
      setChangeFlag();
      } 
       
    }      

    /**
     * Сброс вычисленных и насильно изменённых значений свойства,
     * а также всех зависимых свойств.
     * @param notChangeDepCnt - true не увиличивать счётчик зависимостей, false - увеличивать
     */
    public void clearComputedValue(boolean notChangeDepCnt) {
        clearComputedProp();
        if (app!=null){
        if (!notChangeDepCnt)
            app.dependsCount++;
        }
    }
    /** внутренная процедура рекурсивной последовательности очисток
     * 
     */
   public void clearComputedProp() {
        clearCompVal();
        if (depends!=null){
        for (LuiAppProp dp:depends.keySet()) { // цикл по свойствам которые зависят от данного свойства
          if (depends.get(dp)!=app.dependsCount)  { //если зависимость = счётчику
          depends.replace(dp, app.dependsCount); // запачкаем зависимость счётчиком
          dp.clearComputedProp(); // и вызовем сброс значения для найденного свойства
          
        }
        }
    }
    }
    /**
     * Принудительно устанавливает текущее значение свойства и делает его статическим
     * @param pValue значение
     */
    public void setProperty(String pValue) {
        
      if (!compareString(currentValue,pValue)) // если текущее значение не равно требуемому
          clearComputedValue(false); // то выполним сброс значения в исходное состояние
      
      currentValue = LuiParser.mask(pValue); // если не равно, то в текущее значение записывается, то что дали с заменёнными скобками
      recalculate=null;// и устанавливаем признак статики
      setChangeFlag();
    }    
    

 
/**
 * Установка флага изменённости т.е. признака необходимости отправки на клиента
*/
public void setChangeFlag(){
    changeFlag=true;
    if (ownerObj!=null) ownerObj.setChangeFlag();
}

/**
 * Снятие флага изменённости т.е. признака необходимости отправки на клиента
*/
public void resetChangeFlag(){
    changeFlag=false;
    //if (ownerObj!=null) ownerObj.setChangeFlag();
}

  /** добывает текущий SCN из приложения
   * 
   */
  private Integer getCurrentSCN(){
   /*  if (ownerObj.getClass().getName().equals("LuiApplication")){
         return ((LuiApplication)ownerObj).currentSCN;
     }
     if (ownerObj.getClass().getName().equals("LuiAppWindow")){
         return ((LuiAppWindow)ownerObj).application.currentSCN;
     }
     if (ownerObj.getClass().getName().equals("LuiAppForm")){
         return ((LuiAppForm)ownerObj).application.currentSCN;
     }
     if (ownerObj.getClass().getName().equals("LuiAppElement")){
          if (((LuiAppElement)ownerObj).ownerObj.getClass().getName().equals("LuiAppForm")) return ((LuiAppForm)((LuiAppElement)ownerObj).ownerObj).application.currentSCN;
          if (((LuiAppElement)ownerObj).ownerObj.getClass().getName().equals("LuiAppWindow")) return ((LuiAppWindow)((LuiAppElement)ownerObj).ownerObj).application.currentSCN;
     }*/
   return ownerObj.getApplication().getCurrentSCN();
   //return ownerObj.getCurrentSCN(); 14.11.2018
  }
 /** добывает текущий ClearGroup из приложения
   * 
   */
  private Integer getClearGroup(){
        return ownerObj.getClearGroup();
  }
 
 /**
  * Устанавливает что value изменён не пользователем
  */
  private void setValueNonUserChanged(){
      ownerObj.setProperty("UserChanged","");
  }
  /**
   * Возвращает вычисленное значение свойства и при неоходимости предварительно вычисляет
   * @return  вычисленное значение
   * @throws LuiException логическая ошибка
   * @throws SQLException ошибка выполнения SQL-запроса
   */
public String getComputedProperty() throws LuiException, SQLException {
      Boolean recalc;
      recalc=recalculate;
      if (recalc!=null) recalculate=false;
      statGet++;
      if (recalc==null || !recalc){
        // Если уже не требуется вычисления
        //resetChangeFlag();
        return currentValue; // вернём запомненный результат вычисления
      } 
     
      String RetVal="";
      if (startedValue!=null) RetVal= startedValue;
      if (isDynamic()){//Есть ли динамическа?
/*        ownerObj.getApplication().luiSeance.getServer().perfStat.start(
                ownerObj.getNeedPerfStat(), 
                app.getSeanceIdAndSessionID(), 
                app.code,
                app.getUserName(), 
                ownerObj.getFormCode(),
                ownerObj.getElementCode(),this.code);
  */              
        statComp++;
        long StartTime = System.currentTimeMillis();
        //System.out.println("+P!");
        RetVal = getApplication().parser.compute(RetVal, this);
        statTime += System.currentTimeMillis() - StartTime;
        //если автоматическое вычисление поля изменило значение поля, то считаем, что оно изменено не человеком
        if (code.equals("Value") && !compareString(RetVal,currentValue)) setValueNonUserChanged();
    /*    ownerObj.getApplication().luiSeance.getServer().perfStat.end(
                //ownerObj.getNeedPerfStat(), 
                app.getSeanceIdAndSessionID(), 
                app.code,
                app.getUserName(), 
                ownerObj.getFormCode(),
                ownerObj.getElementCode(),this.code);*/
      } else recalculate=null;
/* рудимент???      
if ("Value".equals(this.code)) { // для Value храним только замаскированное содержимое 
    RetVal=LuiParser.mask(RetVal); // так надо. Иначе если пользователь введёт в поле фигурку и есть JS-код который опирается на Value этого поля,
    // то вычисление JS-выражений происходит с ошибками иногда Exception - иногда просто неверный результат
} 
*/    
   if (!compareString(RetVal,currentValue)) setChangeFlag(); // это добавили 27.02.2020 потому что потеряно при переписывании с PL/SQL
   //else resetChangeFlag();// это закоментили 17.03.2020 потому, что нельзя сбрасывать признак тут Он дожен только в конце обработки всех событий для visual Server
   currentValue = RetVal;
   return RetVal;
    }
public String getGenesis4dmp(){
   //return ownerObj.getGenesis4dmp()+"PropertyCode:"+code+";";
   return ownerObj.getGenesis4dmp()+"."+code;
}

public String getGenesis(){
   return ownerObj.getGenesis()+"PropertyCode:"+code+";";
}
public String getGenesis4msg(){
try {
    if ("Y".equals(ownerObj.getApplication().getProperty("DEBUG_MODE"))){
   return "("+ownerObj.getGenesis()+"PropertyCode:"+code+";)";
  } else return "";
} catch (Exception ex) {
   return ""; 
}
}

public LuiAppElement getParentElement(String pElType) throws LuiException, SQLException{
     return ownerObj.getParentElement(pElType);
  }
private LuiApplication getApplication(){
   return app;
}
/**
 * Делает запрос об изменённости свойства элемента формы с момента изменения последнего SCN Флаги изменённости не сбрасывает. Детерминирована в рамказ одного SCN
 * Было ли изменено или нет с момента последней отправки клиенту?
 * @return  признак изменённости свойства
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public  boolean changed() throws LuiException, SQLException{
    getComputedProperty();
    return changeFlag;
}
public boolean isDynamic(){
  //if (isDyn!=null)return isDyn;
  if (startedValue==null)return false;
  if (startedValue.equals("")) return false;
  if (DYNAMICCONDITION.matcher(startedValue).matches()){
     if (app==null || app.parser==null || app.parser.plugins==null)return true;
     if (startedValue.indexOf("{F:")>-1)return true;
     if (startedValue.indexOf("{F2SQL:")>-1)return true;
     if (startedValue.indexOf("{SQL:")>-1)return true;
     if (startedValue.indexOf("{SQL:")>-1)return true;
     if (startedValue.indexOf("{Property:")>-1)return true;
     if (startedValue.indexOf("{ApplicationProperty:")>-1)return true;
     if (startedValue.indexOf("{FormProperty:")>-1)return true;
     if (startedValue.indexOf("{ElementProperty:")>-1)return true;
 for (String luilang:app.parser.plugins.keySet()){
   if (startedValue.indexOf("{"+luilang+":")>-1) return true;
}      
  }
return false;
}
}