/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Образец класса для Приложения Окна Формы и Элемента
 */
package luiserver;

import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.compareString;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.TreeMap;


/**
 *
 * @author aliubushkin
 */
public class LuiAppWFE {
public static final String VALPREFIX="Var ";
public HashMap<String,LuiAppProp> properties;
boolean changeFlag;
public    LuiAppWFE ownerObj;
public String needPerfStat;// F/E/P/N
public LuiAppWFE(){
properties=new HashMap<>();
}

/**
 * Добавление свойства для хозяйского объекта. Выполняется анализ на динамику. Устанавливается startValue
 * @param pPropertyCode код свойства
 * @param pPropertyValue значение свойства
 * @return LuiAppProp объект "свойство"
 */
public LuiAppProp addProperty(String pPropertyCode, String pPropertyValue) {
    LuiAppProp prop;
      if (!properties.containsKey(pPropertyCode)) { //если свойства  ещё нет
        prop = new LuiAppProp(this,pPropertyCode,NVL(pPropertyValue,""));
        properties.put(pPropertyCode, prop);
        setChangeFlag();
        return prop;
      }
      prop = properties.get(pPropertyCode); //если свойство уже есть или только что создали
      if (!compareString(pPropertyValue,prop.getNonCompValue())) { // сравним начальные значения и если они не совпадают
           prop.clearComputedValue(false); // уводим в исходное значение и чистим зависимости свойства от других свойств
           prop.setStartedValue(pPropertyValue);  //  установим начальное и конечное значение одинаковыми т.е. исходное состояние
      }
      setChangeFlag();
      return prop;
}

/**
 * Установка свойства элемента или формы или окна
 * @param pPropertyCode код свойства
 * @param pPropertyValue значение свойства
 * @return объект "свойство"
 */
public LuiAppProp setProperty(String pPropertyCode, String pPropertyValue) {
      LuiAppProp prop;
      if (!properties.containsKey(pPropertyCode)) { //если свойства  ещё нет
        prop= addProperty(pPropertyCode,pPropertyValue);
      } else {
          prop= properties.get(pPropertyCode);
      }
      prop.setProperty(pPropertyValue);
       //.setStartedValue(pPropertyValue); -- так было до 14.11.18 но выяснилось что какие-то проперти не становятся инвалидами
      
      return prop;
    }    
    
/**
 * Чтение свойства элемента или формы или окна
 * @param pPropertyCode код свойства
 * @return значение свойства
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getProperty(String pPropertyCode) throws LuiException, SQLException{
    if (properties.containsKey(pPropertyCode)) return LuiParser.reMask(properties.get(pPropertyCode).getComputedProperty());
        return "";
}
public String getPropertyIfNotExists(String pPropertyCode, String valueIfNotExist) throws LuiException, SQLException{
    if (properties.containsKey(pPropertyCode)) return LuiParser.reMask(properties.get(pPropertyCode).getComputedProperty());
    return valueIfNotExist;
}
/**
 * Создание или изменение вечно статического свойства
 * @param pValCode код свойства
 * @param pValValue значение свойства
 * @return  объект "свойство"
 */
public LuiAppProp setVal(String pValCode,String pValValue){
    LuiAppProp prop;
 if (!properties.containsKey(VALPREFIX+pValCode)) {//если переменной ещё нет
     prop= new LuiAppProp(this,VALPREFIX+pValCode,pValValue);
     prop.recalculate=null;
     prop.startedValue=null;
     prop.currentValue=pValValue;
     properties.put(VALPREFIX+pValCode, prop);
     setChangeFlag();
     //return prop;
 }
 prop = properties.get(VALPREFIX+pValCode); //если свойство уже есть
 prop.setProperty(pValValue);
 return prop;
}
/**
 * Чтение вечно статического свойства
 * @param pValCode код свойства
 * @return значение свойства
 */
public String getVal(String pValCode){
 if (properties.containsKey(VALPREFIX+pValCode)) return properties.get("Var "+pValCode).getCurrentValue();
 return "";
}
/** 
 * Чтение свойства без его вычисления
 * @param pPropertyCode код свойства
 * @return значение свойства
 */
public String getNonCompValue(String pPropertyCode){
    if (properties.containsKey(pPropertyCode)) return properties.get(pPropertyCode).getNonCompValue();
    return "";
}
/**
 * Установка флага изменённости
*/
public void setChangeFlag(){
   changeFlag=true;
   if (ownerObj!=null){ownerObj.setChangeFlag();}      
}
public TreeMap<String,LuiAppProp> getSortedProps(){
 TreeMap<String,LuiAppProp> sp=new TreeMap<>();   
 for (String pc:properties.keySet()){
     sp.put(pc, properties.get(pc));
 }
 return sp;
}
public int getCurrentSCN(){
    if (ownerObj==null) {throw new RuntimeException("No owner for "+this.getClass().getSimpleName()+" ["+this.getGenesis()+"]");};
    return ownerObj.getCurrentSCN();
}
 public Integer getClearGroup(){
      return ownerObj.getClearGroup();
 }
 public String getGenesis(){
     return "";
 }
 public String getGenesis4dmp(){
     return "";
 }
  public LuiApplication getApplication(){
     return ownerObj.getApplication();
 }
  public LuiAppForm getForm() {
     return null;  
  }
  public String getFormCode(){
    LuiAppForm frm=getForm();
    if (frm==null)return null;
    return frm.code;
  }
  public LuiAppElement getElement(){
      return null;
  }
  public String getElementCode(){
    LuiAppElement el=getElement();
    if (el==null) return null;
    return el.Code;
  }
  public LuiAppElement getParentElement(String pElType) throws LuiException, SQLException{
     return null;
  }
  public String getNeedPerfStat(){
      if (needPerfStat!=null)return needPerfStat;
      return ownerObj.getNeedPerfStat();
  }
}
