/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;

import static LuiCollection.LuiString.compareString;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import LuiExceptionMsg.LuiException;

import java.sql.SQLException;




/**
 *
 * @author aliubushkin
 */
public class LuiAppWindow extends LuiAppWFE{

//HashMap<String,LuiAppElement> elements;
LuiApplication application;

/**
 * Конструктор
     * @param pOwnerObj -- хозяин
     * @param pWindowID -- код окна
     * @param pParentWindowID -- код родительского окна
 */
public LuiAppWindow(LuiAppWFE pOwnerObj,String pWindowID,String pParentWindowID){
    ownerObj=pOwnerObj;
    application=(LuiApplication)ownerObj;

  //  elements=new HashMap<>();
    properties=super.properties;
    setProperty("WindowID",pWindowID);
    setProperty("ParentID",pParentWindowID);
    setProperty("Position","Cascade");
    setProperty("Title","");
}
public String getWinID() throws LuiException, SQLException{
    return getProperty("WindowID");
}
public String getParentID() throws LuiException, SQLException{
    return getProperty("ParentID");
}

public void show(){
    setProperty("ShowStatus","Show");
}
public void hide(){
    setProperty("ShowStatus","Hide");
}
public boolean isActivable() throws LuiException, SQLException{
    boolean rc=false;
    // крутим формы этого окна
    for (String fc:application.forms.keySet()){
        if (application.forms.get(fc).getVal("WindowID").equals(getWinID())){
            if (application.forms.get(fc).elements.containsKey("WINACTIVEABLE$")){ // если в форме есть параметр WINACTIVEABLE$
               LuiAppElement el= application.forms.get(fc).elements.get("WINACTIVEABLE$");
               if (el.getParent().equals("Form")&& el.getType().equals("PARAM")){
                   if (el.getProperty("Value").equals("Y")){ // и он = Y
                    rc=true;
                    break;
                   }
               }
            }
        }
    }
    return rc;
}
public void activate() throws LuiException, SQLException{
    if (activateRT()) return; // если что-то сработает по бегущим графикам - и слава богу
    if (this.getProperty("AbleStatus").equals("Enable"))return;
    // крутим формы этого окна
    if (isActivable()){
        while(!application.commandWriter.isCommandToEnableWindow(getWinID())) { // пока окно не станет активно
                         for (String wc:application.windows.keySet()){ // крутим другие в поисках активного
                            if (application.windows.get(wc).getWinID().equals(this.getWinID())) continue;
                            if (application.windows.get(wc).getProperty("AbleStatus").equals("Enable")){
                              application.windows.get(wc).Close(false);
                              break;  
                            }
                         }
                     }
    }
    /*for (String fc:application.forms.keySet()){
        if (application.forms.get(fc).getVal("WindowID").equals(getWinID())){
            if (application.forms.get(fc).elements.containsKey("WINACTIVEABLE$")){ // если в форме есть параметр WINACTIVEABLE$
               LuiAppElement el= application.forms.get(fc).elements.get("WINACTIVEABLE$");
               if (el.getParent().equals("Form")&& el.getType().equals("PARAM")){
                   if (el.getProperty("Value").equals("Y")){ // и он = Y
                    
                     // то такое оно можно активировать
                     while(!application.commandWriter.isCommandToEnableWindow(getWinID())) { // пока окно не станет активно
                         for (String wc:application.windows.keySet()){ // крутим другие в поисках активного
                            if (application.windows.get(wc).getWinID().equals(this.getWinID())) continue;
                            if (application.windows.get(wc).getProperty("AbleStatus").equals("Enable")){
                              application.windows.get(wc).close();
                              break;  
                            }
                         }
                     }
                   }
               }
            }
        }
    }
    */
}
public boolean activateRT() throws LuiException, SQLException{
  if (is_null(application.getVal("rtChart"))) return false; // уйдём если нет графиков реального времени
  if (!application.forms.containsKey(application.getVal("rtChart")))return false; // или нет формы с графиками уйдём
  LuiAppForm chfrm=application.forms.get(application.getVal("rtChart")); // берём форму с графиками
  if (chfrm.RTdashboard==null) return false; // если в ней нет грфического движка тоже уйдём
  if (chfrm.getVal("WindowID").equals(this.getWinID())){ // если ткнули в окно формы с графиками
   // ищем активную форму
   LuiAppForm pfrm=null;
   String cFrmID=application.getVal("FormID");
   if (!chfrm.getFormID().equals(cFrmID) &&  // если это не форма с графиками
        application.forms.containsKey(cFrmID) // и она есть
       ) {
       pfrm=application.forms.get(cFrmID);
         } 
   else { // если вдруг текущей формой оказалась форма с графиками то будем искать активное окно
      for (String aw:application.windows.keySet()){
        if (application.windows.get(aw).getProperty("AbleStatus").equals("Enable")){
          for (String fc:application.forms.keySet()){
            if (application.forms.get(fc).getVal("WindowID").equals(aw)){
              if (!fc.equals(chfrm.getFormID())){  
                pfrm=application.forms.get(fc);break;
              }
            }
          }
          if (pfrm!=null)break;
        }
      }
   }
   if (pfrm!=null){
       if (!pfrm.getFormID().equals(chfrm.getFormID())){ // если pfrm это не тоже самое что chfrm
           if (pfrm.elements.containsKey("SYSMENU$RT-CHARTS")){ // и у этой формы есть действие для вызова бегущих графиков
               pfrm.setVal("CurrentAction","");
               pfrm.runAction("SYSMENU$RT-CHARTS");
               return true;
           }
       }
   }
  }
  // ткнули не в окно с графиками
  if (!application.forms.containsKey(chfrm.getVal("rtParentFormID")))return false; // уйдём если возврат из формы сграыиками ведёт в никуда
  LuiAppForm pfrm=application.forms.get(chfrm.getVal("rtParentFormID")); // берём форму в которую надо возвращаться из графиков
  if (pfrm.getVal("WindowID").equals(this.getWinID())){ // если ткнули именно в её окно
    if (chfrm.elements.containsKey("BACK")){  
        pfrm.setVal("CurrentAction","");
        chfrm.runAction("BACK"); // прыгаем туда из графиков
        return true;
    }
    if (chfrm.elements.containsKey("SYSMENU$BACK")){  
        pfrm.setVal("CurrentAction","");
        chfrm.runAction("SYSMENU$BACK"); // прыгаем туда из графиков
        return true;
    }
  }
return false;
}
public void disCover(LuiClCommandGroup pXMLcg,String pNotCoverFrmID) throws LuiException, SQLException{
    addProperty("Discover","Y");
    if (pXMLcg==null ){
        pXMLcg=application.commandWriter.addGroup("Disable Window "+this.getWinID());
    }
    if (pXMLcg==null)return;
         LuiClCommand XMLcmd=pXMLcg.addCommand("DisableWindow");
          XMLcmd.addScaPar("WindowID",getWinID());
    if (isActivable()){
          XMLcmd.addScaPar("Cover","rgba(255,255,255,0)");
    }
    if (is_not_null(pNotCoverFrmID)){ // если надо покрыть ввуалью все формы кроме заданной, то
        // крутим формы этого окна
    for (String fc:application.forms.keySet()){
        if (application.forms.get(fc).getVal("WindowID").equals(getWinID())){
            if (!fc.equals(pNotCoverFrmID)){
                XMLcmd=pXMLcg.addCommand("Form");
                XMLcmd.addScaPar("WindowID", getWinID());
                XMLcmd.addScaPar("FormID", fc);
                XMLcmd.addScaPar("Cover", "rgba(0,0,0,0.25)");
            }
        }
    }
    }
}
public void disable(){
    setProperty("AbleStatus","Disable");
}
public void enable() {
    //for(String wid:application.windows.keySet()){
      //  if (wid.equals(getWinID())) {
            setProperty("AbleStatus","Enable");
            setProperty("Discover","N");
        //} else{
          //  application.windows.get(wid).disable();
        //}
    //}
}
public void close() throws LuiException, SQLException{
    Close(true);
}
/**
 * Обработка события закрытия окна c параметром, который определяет необходимость выполнения финальных действий в формах окна
 * @throws LuiException логическая ошибка
  * @throws SQLException ошибка выполнения SQL-запроса
 */
    public void Close(boolean pDoCloseAction) throws LuiException, SQLException{
    application.changeSCN("CloseWindow");
    if (application.forms!=null){
        if (application.forms.size()<1)return;
        if (pDoCloseAction) {
        for (String frmc:application.forms.keySet()){
        
        LuiAppForm frm=application.forms.get(frmc);
        if (compareString(frm.getVal("WindowID"),this.getWinID())){
          if (!compareString(frm.getVal("Status"),"Closing")){ 
              if (frm.getProperty("NeverClosing").equals("Y"))return;
          }
        }
        }
        }
    boolean b=true;
    while (b){
    b=false;
    for (String frmc:application.forms.keySet()){
        LuiAppForm frm=application.forms.get(frmc);
        if (compareString(frm.getVal("WindowID"),this.getWinID())){
          if (!compareString(frm.getVal("Status"),"Closing")){  
            if (compareString(frm.getVal("MasterForm"),"Y")){
                if (compareString(frm.getVal("ReadOnly"),"Y")){
                    frm.close("","CLOSE" );
                } else {
                   String ActionCode=frm.checkHotKey("CLOSE","Form");
                   if (is_not_null(ActionCode) && pDoCloseAction){
                       frm.addProperty("ActionBreak","");
                       frm.runAction(ActionCode);
                       if (frm.getProperty("ActionBreak").equals(ActionCode)){
                       frm.addProperty("ActionBreak","");    
                       } else {
                       if (!compareString(frm.getVal("Status"),"Closing")){ 
                       frm.close("","");    
                       } 
                       b=true;
                       break;
                       }
                   } else{
                       frm.close("","");
                       b=true;
                       break;
                   }
                }
            }
          }
        }
    }
    }
}
   
    }
@Override
public String getGenesis(){
    try {    
        return ownerObj.getGenesis()+"Window:"+getWinID()+";";
    } catch (LuiException ex) {
        return "";
    } catch (SQLException ex) {
        return "";
    }
}   
@Override
public LuiApplication getApplication(){
    return application;
}
}
