/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс для хранения данных о прикладной системе
 */
package luiserver;

import LuiCollection.LuiCollection;
import LuiCollection.LuiString;
import static LuiCollection.LuiString.Extract_Word;
import static LuiCollection.LuiString.GetTableAndAliasFromCursor;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.RTrim;
import static LuiCollection.LuiString.compareString;

import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;

import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import LuiNLS.LuiNLS;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonParser;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileOutputStream;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import javax.script.ScriptEngine;

import org.apache.commons.io.IOUtils;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.Scriptable;
/**
 *
 * @author aliubushkin
 */
public class LuiApplication extends LuiAppWFE{
 private static final String STMT4UPDATEUSRPREF="update lui_t_user_pref set value <value> where app_code='<app_code>' and form_code <form_code> and user_name='<user_name>' and property_code='<property_code>' and element_code <element_code> and config_code <config_code> and usrcfg_id <usrcfg_id>";
 private static final String STMS4INSERTUSRPREF="insert into lui_t_user_pref(value,user_name,element_code,form_code,usrcfg_id,property_code,config_code,app_code) values (<value>,'<user_name>',<element_code>,<form_code>,<usrcfg_id>,'<property_code>',<config_code>,'<app_code>')";
 private static final String STMT4LOADFORMTYPE="select formtype_code FROM lui_t_form where app_code='<app_code>' and code='<form_code>'";    
 private static final String STMT4LOADAPP="SELECT nls(name,'<LANG>'),owner FROM lui_t_application where code='";
 private static final String STMT4LOADPROP="select p.code \"property_code\",COALESCE(p.value,p.def_value) \"value\",'Y'  from lui_t_application_property p where p.in_memory='Y' and p.app_code='<app_code>'";
 public boolean decr=true;
 String AppFilePath;               
 String TmpFilePath;
 public Integer dependsCount=0; // хранит счётчик для обеспечения зависимостей между свойствами элементов
 public String code;
 public LuiGmap gMap;
 public String LuiSrvName;
 public String name;
 public String owner;
 public LuiDbConnPool metaConnPool; //пул для доступа к метаданным
 public LuiDbConnPoolArray poolArray; //массив пулов для регистрации в нём прикладных пулов соединений
 public LuiNLS NLS;
 private int sesUniqueID=0;
 public HashMap<String,LuiAppWindow> windows;
 public LinkedHashMap<String,LuiAppForm> forms;
 public LuiMsgWriter msgWriter ;
 public HashMap<String,LuiAppForm> verifiers;
 public HashMap<String,LuiAppForm> formTemplates; // шаблонные свойства форм и их элементов ключ этого массива соответствует коду типа формы
 //public HashMap<String,LuiAppForm> osCommands; //команды ОС причём ключём этого map является <КодФормы>.<Конфигурация>
 public LuiClCommandWriter commandWriter;// писатель комманд
 public LuiClEventReader eventReader; //читатель событий нужен LOVам
 public LuiParser parser; // разборщик выражений и выполнятор программ
 public HashMap<String,LuiAppConn> userDbConnections; // набор прикладных соединений с БД
 public LuiAppLog log;//журнал
 public ArrayList<LuiAppBreakable> backList; // список объектов отката в случае аномального завершения действия
 public LuiProcTimer TotalTimer;
 public HashMap<String,Class> stdPluginClasses;
 public HashMap<String, String> stdPluginIsInter; // признаки интерактивности плагинов
 public LuiCollection commonForms;
 public ScriptEngine JSengine; //движок JavaScript
 public jdk.nashorn.api.scripting.ScriptObjectMirror JSContents; // класс для передачи данных между Java и JS
 public Context RhinoContext;
 public Scriptable     RhinoScope;
 public LuiSession luiSession=null;
 public LuiProgressBar progressBar=null;
 public LuiElementComparator elComparator=null;
 public LuiCrip crypter=null; 
 public LuiCollection interceptor4sql_error=null;
 public LuiCollection lastFieldValues=null;
 public LuiSeance luiSeance;
 private EventChannel sChannel; // служебный канал (secondary)
 private LuiClCommandWriter commandWriterSecondary;

 private Gson gson;
 private GsonBuilder gsonBuilder;
 public JsonParser JSONparser=new JsonParser();
 public LuiDataConverter dataConverter;
 public HashMap<String,LuiAppForm> frmCache=new HashMap<>();
 public ArrayList<String> frmCacheList=new ArrayList<>();
 public Long frmCacheCrTime=null;
 public long frmCacheLifeTime=0;
 //public LuiCollection LastLOVchoice=new LuiCollection("");

 
 /**
  * Конструктор объекта. Выполняет поиск и загрузку данных о приложении
  * @param pLuiSeance сеанс LUI
  * @param pCode  -- код прикладной системы
  * @param pLangCode код языка
  * @param pMetaConnPool -- пул соединений дgetля доступа к мета
  * @param pMsg   -- писатель сообщений
  * @param pPoolArray -- массив пулов для создания нового пола
  * @param pNLS НЛС
  * @param pFormTemplates коллекция шаблонов форм
  * @param pCommandWriter писатель команд
  * @param pStdPluginClasses коллекция плагинов
  * @param PstdPluginIsInter коллекция признаков интерактивности плагинов
  * @param pEventReader читатель событий
  * @param pTimer таймер
  * @param pCommonForms коллекция общих форм
  * @param pHostName имя хоста
  * @param pCrypter шифровальник (?)
  * @param pClientID Идентификатор клиента
  * @param pMsgPort порт для сообшений
  * @param pLuiSrvName имя LUI-сервера
  * @param pAppFilePath путь к каталогу с файлами приложения
  * @param pLuiDatConv конвертер дат
  * @param Params параметры (а до этого что было?)
  * @throws LuiException логическая ошибка
  * @throws SQLException ошибка выполнения SQL-запроса
  */
public LuiApplication(LuiSeance pLuiSeance, String pCode,String pLangCode,LuiDbConnPool pMetaConnPool,LuiMsgWriter pMsg,LuiDbConnPoolArray pPoolArray,LuiNLS pNLS,HashMap<String,LuiAppForm> pFormTemplates,LuiClCommandWriter pCommandWriter,HashMap<String,Class> pStdPluginClasses,HashMap<String, String> PstdPluginIsInter,LuiClEventReader pEventReader,LuiProcTimer pTimer,LuiCollection pCommonForms,String pHostName,LuiCrip pCrypter,String pClientID, String pMsgPort,String pLuiSrvName,String pAppFilePath,LuiDataConverter pLuiDatConv,String Params) throws LuiException, SQLException{
    
    dataConverter=pLuiDatConv;
    LuiSrvName=pLuiSrvName;
    crypter=pCrypter; 
    elComparator= new LuiElementComparator();
    log=new LuiAppLog(this);
    eventReader=pEventReader;
    if (eventReader!=null)
    eventReader.app=this;
    poolArray=pPoolArray;
    commandWriter=pCommandWriter;
    TotalTimer=pTimer;
    code=pCode;
    NLS=pNLS;
    msgWriter=pMsg;
    properties=super.properties;
    luiSeance=pLuiSeance;
    needPerfStat="P";
if (pFormTemplates!=null){
    formTemplates= new HashMap<>(pFormTemplates);
}
    AppFilePath=pAppFilePath;
     metaConnPool=pMetaConnPool;
     windows=new HashMap<>();
     forms=new LinkedHashMap<>();
     stdPluginClasses=pStdPluginClasses;
     stdPluginIsInter=PstdPluginIsInter;
     commonForms=pCommonForms;
     parser=new LuiParser(this,stdPluginClasses,stdPluginIsInter);
     if (commandWriter!=null)     commandWriter.setJSengine(JSengine);
     userDbConnections=new HashMap<>();
     LuiDbConn conntmp = null;
     luiSession=null;
     progressBar=null;
     
     try {
         conntmp  = metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
         LuiNLS newNLS=new LuiNLS(conntmp,code); // ставим новый NLS
         
         NLS=newNLS;
         msgWriter.setAppCode(code); msgWriter.setApplication(this); msgWriter.setNLS(NLS); // обновляем писателя сообщений
         if (is_not_null(pLangCode)){
             if (NLS.langList.containsKey(pLangCode)){
                 NLS.setCurrentLang(pLangCode);
             }
         }
         Integer nl=NLS.langList.size();
         setVal("NLSnumberOfLanguages",nl.toString());
         setVal("NLSlanguages",NLS.langs);
         setVal("NLScurrentLanguage",NLS.currentLang.code);
         setProperty("CURRENT_LANG",NLS.currentLang.code);
         setVal("NLSJSlanguages",new LuiCollection(NLS.langs).ToJSON());
         setProperty("DEBUG_MODE","N");
     backList=new ArrayList<>();
     addProperty("AppCode",code);
     addProperty("OS Name",System.getProperty("os.name"));
     addProperty("OS Version",System.getProperty("os.version"));
     addProperty("QBE_LOV_ROLE","F9");
     addProperty("QBE_MULTI_LOV_ROLE","");
   // нуждается ли это приложение в собственном объекте конвертер данных?
    if (!code.equals("LUI")){
      if (conntmp.getSelectResult("select 1 from lui_t_datatype where app_code='"+code+"' limit 1")!=null){
     dataConverter=new  LuiDataConverter(code,conntmp);    
     }
    }
     addProperty("ClientHostName",pHostName);
   if (luiSeance!=null){
     addProperty("INITIAL_FORM", code); //Код начальной формы по умолчанию совпадает с кодом приложения
     gson=luiSeance.gson;
     gsonBuilder=luiSeance.gsonBuilder;
     
   } else {
     gsonBuilder = new GsonBuilder();
     gson =gsonBuilder.create();   
     
   }
     addProperty("PathSeparator",System.getProperty("file.separator"));    
         
         ArrayList<String> naow= conntmp.getSelectFirstRowValList(STMT4LOADAPP.replace("<LANG>",NLS.currentLang.code)+code+"'"); //название приложения и вледелец
         name=naow.get(0);
         owner=naow.get(1);
         if (is_null(name)) { msgWriter.write("LUI-00007",code); }
         addProperty("AppName",name);
         // читаем все свойства приложения
         ArrayList<ArrayList<String>>props=conntmp.getSelectAllRowsValList(STMT4LOADPROP.replace("<app_code>",code));
         for(int i=0;i<props.size();i++){
             ArrayList<String> row=props.get(i);
             if (NVL(row.get(2),"N").equals("Y")){
             addProperty(row.get(0),NLS.NLS(row.get(1)));    
             } else {
             addProperty(row.get(0),row.get(1));
             }
         }
        if (is_not_null(Params)){
            LuiCollection parcol = new LuiCollection(Params);
            for(String pc:parcol.keySet()){
                 if (like(pc.toUpperCase(),"LANG") || like(pc.toUpperCase(),"APPCODE")) continue;
                String prop_code=pc;
                if (pc.toUpperCase().equals("DUMPKEYDISABLE"))prop_code="DUMP_KEY_DISABLE";
                if (pc.toUpperCase().equals("APPEARANCE")) prop_code="APPEARANCE";
                if (pc.toUpperCase().equals("STYLE")){
                    addProperty("STYLE_AFTER_LOGON",parcol.get(pc));
                    addProperty("STYLE_BEFORE_LOGON",parcol.get(pc));
                    continue;
                }
                if (pc.toUpperCase().equals("CODE")) {
                    if (getProperty("EA_METHOD").equals("KEYCLOAK")) prop_code="AuthCode";
                }
                if (pc.toUpperCase().equals("EXTAUTH")) prop_code="EXTERNAL_AUTHORIZATION";

                addProperty(prop_code,parcol.get(pc));
                
                /*if (pc.toUpperCase().equals("DUMPKEYDISABLE")){
                   if (parcol.get(pc).equals("Y")) {
                       addProperty("DUMP_KEY_DISABLE","Y");
                   }
                }
                if (!like(pc.toUpperCase(),"LANG") && !like(pc.toUpperCase(),"APPCODE")){
                  if (pc.toUpperCase().equals("APPEARANCE")) {
                      addProperty("APPEARANCE",parcol.get(pc));
                  } else
                    addProperty(pc,parcol.get(pc));
                }*/
                
            }
        }
 addProperty("APP_FILES_PATH",AppFilePath); 
 

//*** ВНЕШНЯЯ АВТОРИЗАЦИЯ (начало)
if (getProperty("EXTERNAL_AUTHORIZATION").toUpperCase().equals("Y")) {
    // заполнение коллекции для плагина AuthExternal
    LuiCollection acol = new LuiCollection("");
    String authMethod = getProperty("EA_METHOD");
    acol.Add_Element("AuthExternal","login");
    acol.Add_Element("authUrl",getProperty("EA_URL"));
    acol.Add_Element("authUrlApp",getProperty("EA_URL_APP"));
    acol.Add_Element("authUrlCommon",getProperty("EA_URL_COMMON"));
    acol.Add_Element("authMethod",getProperty("EA_METHOD"));
    if (authMethod.equals("KEYCLOAK")) {
        acol.Add_Element("authCode",getProperty("AuthCode"));
        properties.remove("authCode");
    } else if (authMethod.equals("TOKEN")) {
        acol.Add_Element("token",getProperty("authToken"));
        acol.Add_Element("username",getProperty("sudisUser"));
        acol.Add_Element("roles",getProperty("sudisRoles"));
        acol.Add_Element("service",getProperty("serviceName"));
        properties.remove("authToken");
    }
    addProperty("ExtAuth","{"+acol.toString()+"}");
    acol = new LuiCollection(properties.get("ExtAuth").getComputedProperty()); // место обращения к плагину AuthExternal
}
//*** ВНЕШНЯЯ АВТОРИЗАЦИЯ (конец)


if (luiSeance!=null){ // если приложение создаётся внутри сеанса
 addProperty("AppDirURL",luiSeance.url+"appfiles"+"/"+code+"/"); 
 try {
 frmCacheLifeTime=1000*Long.valueOf(getProperty("FORMS_CACHE_LIFETIME"));
 } catch(Exception ecf){}
 gMap=new LuiGmap(this);
         // Создаётся ещё 1 сокет/канал (для служебных сообщений), устанавливается его связь с текущим сеансом (заполняется sChannel для сеанса)
         luiSeance.createSecondaryChannel(pMsgPort, pClientID);
         
         //тип стартовой формы
         String Ftype=conntmp.getSelectResult(STMT4LOADFORMTYPE.replace("<app_code>",code).replace("<form_code>",getProperty("INITIAL_FORM")));
         if (is_null(Ftype)){
          msgWriter.write("LUI-00015", getProperty("INITIAL_FORM"));
         }
         addProperty("INITIAL_FORMTYPE",Ftype);
         LuiClCommandGroup XMLcg=this.commandWriter.addGroup("Init");
         LuiClCommand XMLcmd=XMLcg.addCommand("Init");
         
         // передача appearance
         String SS=getProperty("APPEARANCE");
         XMLcmd.addScaPar("Appearance",NVL(SS,"Windows"));
         
         // передача стилей визуализации
         SS=getProperty("STYLE_BEFORE_LOGON");
         setProperty("STYLE_CURRENT",NVL(SS,"default"));
         XMLcmd.addScaPar("Skin",NVL(SS,"default"));
         //локаль
         String locale=NLS.currentLang.locale;
         if (locale.indexOf("_")>0){
             locale=locale.substring(0,locale.indexOf("_"));
         }
         XMLcmd.addScaPar("Locale",locale.toLowerCase());
         
         
         if (!this.code.equals("LUI")){
         // передача Application
         SS=NVL(getProperty("BROWSER_WINDOW_TITLE"),this.code);
         XMLcmd.addScaPar("Application",SS);
         // иконка в заголовке окна браузера
         SS=getProperty("BROWSER_WINDOW_ICON");
         if (is_not_null(SS)){
          if (!like(SS,"http%")){   
          SS=luiSeance.url+"custom/"+getProperty("STYLE_BEFORE_LOGON")+"-desktop/"+SS;
         }
          XMLcmd.addScaPar("Favicon",SS);   
         }
         }
         // передача словаря
         LuiCollection dict=conntmp.getSelectAllRowsTwoColsValMap("select code,NLS(message_text,'"+
                 NLS.currentLang.code
                 +"','LUI') from lui_t_message where severety='V' and app_code='LUI'");
         XMLcg=this.commandWriter.addGroup("Dictionary");
         for (String wc:dict.keySet()) { // крутим элементы словаря
          String ownVal="";   
          if (!"LUI".equals(this.code)) {  // если это система не LUI пытаемся найти собственный элемент словаря
          ownVal=conntmp.getSelectResult("select NLS(message_text,'"+NLS.currentLang.code+"','"+this.code+"') from lui_t_message where severety='V' and app_code='"+this.code+"' and code='"+wc+"'");
          }
          XMLcmd=XMLcg.addCommand("Dict");
          XMLcmd.addScaPar("Key",wc);
          XMLcmd.addScaPar("Value",NVL(ownVal,dict.get(wc)));
         }
        
}
         setProperty("dbURL",metaConnPool.getURL());
         setProperty("dbSCHEMA",metaConnPool.getSchema());
         pMetaConnPool.SetFree(conntmp);
     } 
     catch (SQLException ex) {
         msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());       
     } catch (NullPointerException ex) {
        StackTraceElement[] ste = ex.getStackTrace();
        String m = NVL(ex.getLocalizedMessage(), " null pointer ");
        m = m + " in ";
        for (int i = 0; i < ste.length; i++) {
            StackTraceElement st = ste[i];
            m = m + "\n";
            m = m + st.getClassName() + "." + st.getMethodName() + ":line" + st.getLineNumber();
        }
        msgWriter.write("LUI-00006",code,"\n"+ m);
     }
     catch (Exception ex) {
         msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());       
     } finally{
         try {
           metaConnPool.SetFree(conntmp);
         } catch (SQLException ex1) {
         }
     }
  getAppDir();   //каталогприложения
  addProperty("TmpDir","{JS:App.getTmpDir();}");  // каталог временных файлов
  addProperty("TmpDirURL","{JS:App.getTmpDirURL();}");  // URL временных файлов
  if (luiSeance!=null){
  loadTemplates();  // загрузка шаблонов 
  verifiers=new HashMap<>();
  if (loadVerifiers()){ // загрузка верификаторов и генерацияфайла verify.js
                        // tru если файл имеется
    LuiClCommandGroup XMLcg=this.commandWriter.addGroup("Verifier script");
         LuiClCommand XMLcmd=XMLcg.addCommand("Script");
         XMLcmd.addScaPar("Name", getProperty("AppDirURL")+"verify.js");
  }
  }
  // служебное соединение добавим в массив прикладных
  
  LuiAppConn srvc = new LuiAppConn(this,"LUI$","LUI service connection","COMMON",metaConnPool.getUser(),"***","","lui");
  srvc.connPool=this.metaConnPool;
  srvc.dbURL=this.metaConnPool.getURL();
  userDbConnections.put("LUI$", srvc);
  
  
  
  //конец конструктора
  
}
/**
 * Загрузка верификаторов из БД и выгрузка из в файл verify.js в каталог приложения
 * @return какой-то признак
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean loadVerifiers() throws SQLException, LuiException{
    String conf=this.getProperty("VERIFIER_CONFIGURATION");
    String filename=AppFilePath+"verify.js";  
   if (is_null(conf))conf="DEFAULT";
    LuiDbConn lc=null;
    boolean fg=false;
    LuiAppForm verifier;
    ArrayList<LuiCollection> vrf;
    Writer filew =null;
    OutputStreamWriter OSW = null;
    FileOutputStream FOS=null;
    try{
    if (!code.equals("LUI")){//если прикладная система не LUI
        for(String commonVrf:commonForms.keySet()){
            if (commonForms.Get_Element(commonVrf).equals("VERIFIER")){
            verifier=new LuiAppForm(this,commonVrf,"VERIFIER");
            verifier.setProperty("ConfigCode",conf);
            verifier.load("LUI",true);
            verifiers.put(commonVrf, verifier);
            }
        }
    }
    
    //String RU4L4MVD=null;
    lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
    vrf=lc.getSelectAllRowsValMap("select code from lui_t_form where formtype_code='VERIFIER' and app_code='"+code+"' and version_number is null");
    if (vrf!=null){
        for (int i=0;i<vrf.size();i++){
            
            if (verifiers.containsKey(vrf.get(i).Get_Element("code"))){
                verifiers.remove(vrf.get(i).Get_Element("code"));
            }
            
            verifier=new LuiAppForm(this,vrf.get(i).Get_Element("code"),"VERIFIER");            
            verifier.setProperty("ConfigCode","DEFAULT");
            verifier.load(code,true);
            verifiers.put(vrf.get(i).Get_Element("code"), verifier);
        }
    }
    String verifierJS="";
        for(String VrfCode:verifiers.keySet()){
     //     if (like(VrfCode,"RU4L4MVD%")){
     //         RU4L4MVD=VrfCode;continue;
     //     }
            
           if (verifiers.get(VrfCode).properties.containsKey("VERIFIER_PRG")){
            verifierJS=verifierJS+verifiers.get(VrfCode).getProperty("VERIFIER_PRG")+"\n";
//            verifiers.get(VrfCode).properties.remove("VERIFIER_PRG"); // чтобы не хранить
           }
        }
      
if (is_not_null(verifierJS)){
     FOS=new FileOutputStream(filename);
     OSW=new OutputStreamWriter(FOS, "UTF-8");
     filew=new BufferedWriter(OSW);
     filew.write("var fieldVerify={};\n"+verifierJS);
 /*     if (RU4L4MVD!=null){
      filew.write("\n");
      filew.write("fieldVerify."+RU4L4MVD+" = function (param) { \n" +
"   var ret = // returned object\n" +
"   {   exitCode:0, // returned code: =0 - success\n" +
"                   // !=0 - error\n" +
"       exitMessage:\"OK\", // additional message\n" +
"       value:param.value, // output value\n" +
"       cursor:param.cursor // new cursor position\n" +
"   };   \n" +
"echarts.registerMap('RU4L4MVD',\n");
      filew.write(lc.getSelectResult("select convert_from(valueb,'UTF8') \n" +
"from lui_t_value \n" +
"where app_code='LUI' and form_code='RU4L4MVD' and config_code='DEFAULT' and element_code is null and property_code='GEOJSON'"));
      filew.write(");\n" +
"console.log('Map registred!');"+
              "return ret;\n" +
"}; ");
      }   */
        //filew.close();
        fg=true;
}
    } 
    catch (IOException ioex){
        this.msgWriter.write("LUI-00114",filename,ioex.getMessage());
    }
    
    finally {
        if (filew!=null)try{filew.close();}catch(Exception ex){}
        if (OSW!=null)try{OSW.close();}catch(Exception ex){}
        if (FOS!=null)try{FOS.close();}catch(Exception ex){}
        if (lc!=null) metaConnPool.SetFree(lc);
    }
return fg;
}
/**
 * Загрузка шаблонных форм
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void loadTemplates() throws SQLException, LuiException{
   String conf=this.getProperty("TEMPLATE_CONFIGURATION");
   if (is_null(conf))conf="DEFAULT";
   LuiAppForm tmpl;
   // загружаем нестандартные упомянутые шаблоны первыми чтобы они не запвчкались стандартными
LuiDbConn mlc=null;
ArrayList<LuiCollection> local_other_templates=null;
try{
mlc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
local_other_templates=mlc.getSelectAllRowsValMap("select distinct t.code,t.formtype_code from lui_t_form f,lui_t_form t where t.app_code='"+this.code+"' "
        + "and t.app_code=f.app_code and t.version_number is null and f.template=t.code "
        + "and f.version_number is null");
} finally{
    if (mlc!=null) this.metaConnPool.SetFree(mlc);
}
if (local_other_templates!=null){
    for (int i=0;i<local_other_templates.size();i++){
        String lft=local_other_templates.get(i).Get_Element("formtype_code");
        String lfc=local_other_templates.get(i).Get_Element("code");
        try{
        tmpl=new LuiAppForm(this,lfc,lft);
        tmpl.setProperty("ConfigCode",conf);
        tmpl.load(code,true);
        this.formTemplates.put(lfc, tmpl);
        }catch(LuiException lex) {
            String here="here";
        }
    }
}
// загружаем стандартные шаблоны
//formTemplates.clear();
   tmpl=new LuiAppForm(this,"BLANK_TEMPLATE","BLANK");
   tmpl.setProperty("ConfigCode",conf);
   try{
         tmpl.load(null,true);
         //tmpl.makeElList("Form");
         formTemplates.put("BLANK_TEMPLATE", tmpl);
         } catch(LuiException lex){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
           //      tmpl.makeElList("Form");
                 this.formTemplates.put("BLANK_TEMPLATE", tmpl);
                 } catch(Exception ex){
                     msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
         }
   tmpl=new LuiAppForm(this,"LIST_TEMPLATE","LIST");
   tmpl.setProperty("ConfigCode",conf);
         try{
         tmpl.load(null,true);
//         tmpl.makeElList("Form");
         this.formTemplates.put("LIST_TEMPLATE", tmpl);
         } catch(LuiException lex2){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
 //                tmpl.makeElList("Form");
                 this.formTemplates.put("LIST_TEMPLATE", tmpl);
                 } catch(Exception ex){
                 msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
             }
tmpl=new LuiAppForm(this,"TREE_TEMPLATE","TREE");
   tmpl.setProperty("ConfigCode",conf);
         try{
         tmpl.load(null,true);
 //        tmpl.makeElList("Form");
         this.formTemplates.put("TREE_TEMPLATE", tmpl);
         } catch(LuiException lex2){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
//                 tmpl.makeElList("Form");
                 this.formTemplates.put("TREE_TEMPLATE", tmpl);
                 } catch(Exception ex){
                 msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
             }         
tmpl=new LuiAppForm(this,"MENU_TEMPLATE","MENU");
   tmpl.setProperty("ConfigCode",conf);
         try{
         tmpl.load(null,true);
 //        tmpl.makeElList("Form");
         this.formTemplates.put("MENU_TEMPLATE", tmpl);
         } catch(LuiException lex2){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
//                 tmpl.makeElList("Form");
                 this.formTemplates.put("MENU_TEMPLATE", tmpl);
                 } catch(Exception ex){
                 msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
             }         
tmpl=new LuiAppForm(this,"CHART_TEMPLATE","CHART");
   tmpl.setProperty("ConfigCode",conf);
         try{
         tmpl.load(null,true);
 //        tmpl.makeElList("Form");
         this.formTemplates.put("CHART_TEMPLATE", tmpl);
         } catch(LuiException lex2){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
//                 tmpl.makeElList("Form");
                 this.formTemplates.put("CHART_TEMPLATE", tmpl);
                 } catch(Exception ex){
                 msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
             }         
   tmpl=new LuiAppForm(this,"HTML_TEMPLATE","HTML");
   tmpl.setProperty("ConfigCode",conf);
         try{
         tmpl.load(null,true);
 //        tmpl.makeElList("Form");
         this.formTemplates.put("HTML_TEMPLATE", tmpl);
         } catch(LuiException lex2){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
//                 tmpl.makeElList("Form");
                 this.formTemplates.put("HTML_TEMPLATE", tmpl);
                 } catch(Exception ex){
                 msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
             }         
tmpl=new LuiAppForm(this,"GRID_TEMPLATE","GRID");
   tmpl.setProperty("ConfigCode",conf);
         try{
         tmpl.load(null,true);
 //        tmpl.makeElList("Form");
         this.formTemplates.put("GRID_TEMPLATE", tmpl);
         } catch(LuiException lex2){
             this.msgWriter.ClearMsgArray();
                 try{tmpl.load("LUI",true);
//                 tmpl.makeElList("Form");
                 this.formTemplates.put("GRID_TEMPLATE", tmpl);
                 } catch(Exception ex){
                 msgWriter.write("LUI-00006",code,"\n"+ex.getLocalizedMessage());
                 }
             }          

}
/**
 * Загрузка команд OC
 */
/*public void loadOScommands() throws SQLException{
  LuiDbConn lc=null;
  String Stmt="select form_code oscode,c.code config_code from lui_t_config c,lui_t_form f where f.code=c.form_code and f.formtype_code='HOST' and f.version_number is null and f.app_code='<app_code>' order by 1,2";
  try {
    lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID());
    // в начале грузим команды OS приложения
     ArrayList<LuiCollection> oscList=lc.getSelectAllRowsValMap(Stmt);
   for (int i=0;i<oscList.size();i++){
     oscList.get(i).Get_Element("oscode")
   }
  } finally {
    metaConnPool.SetFree(lc);
  }  
}*/


/**
 * Создание новой формы и занесение её в массив форм
     * @param pFormID - ID формы, если пусто вычислит сама
     * @param pCode - код метаданных; может состоять из &lt;КодПриложения&gt;.&lt;Код формы&gt;
     * @param pType - тип формы BLANK/LIST не может быть пусто22222
 * @return форма
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiAppForm addForm(String pFormID,String pCode,String pType/*,String formStat*/) throws LuiException, SQLException{
    String mdApp=null;
    LuiDbConn lc=null;
    
    int point=pCode.indexOf(".");
    if (point>1) {
      mdApp=pCode.substring(0, point);
     try {
       lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
       String tmp=null;
       tmp=lc.getSelectResult("select code from lui_t_application where code='"+mdApp+"'");
       if (is_null(tmp)) {
         mdApp=code;  
       } else {
           pCode=pCode.substring(point+1);
       }
    } finally {
        if (lc!=null){
          metaConnPool.SetFree(lc);
        }
    } 
    } else mdApp=getProperty("AppCode");
    LuiAppForm frm= new LuiAppForm(this,pCode,pType);
    frm.setProperty("MetadataCode", pCode);
    String mdType=null;
    lc=null;
    try {
       lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
       mdType=lc.getSelectResult("select formtype_code from lui_t_form where app_code='"+mdApp+"' and code='"+pCode+"'");
       if (is_null(mdType)) {
       mdType=lc.getSelectResult("select formtype_code from lui_t_form where app_code='LUI' and formgroup_code='LUI_COMMON' and code='"+pCode+"'");    
        if (is_null(mdType)) {
        this.msgWriter.write("LUI-00015", pCode);
        return null;
        } else 
       frm.setProperty("MetadataApp","LUI");
       } else
       frm.setProperty("MetadataApp",mdApp);
       /*if (getProperty("FORM_STAT").equals("Y") && formStat.equals("Y")){
          lc.getSelectResult("select lui_r_register_form_event('"+frm.getProperty("MetadataApp")+"','"+pCode+"','RUN')");
       }*/
    } finally {
        if (lc!=null){
          metaConnPool.SetFree(lc);
        }
    }
    
    if (is_null(pFormID)) {
      pFormID="FRM"+getNextUniqueID();
    }
    forms.put(pFormID, frm);//заносим в массив
    frm.addProperty("FormID",pFormID); // FormID как св-во формы чтобы работали {Property:FormID}
    if (!(mdType.equals("HOST"))) {//если это не команда oc 
    setVal("FormID",pFormID); // текущая форма
    }
    // устанавливаем свойства формы по умолчанию
    if (is_null(pType)) {
      pType=mdType;  
    } else {
      if (mdType.equals("LIST")&& "TREE".equals(pType)) { // если список и его надо открыть деревом
      frm.setVal("RootFormID",pFormID);
        }
      if (mdType.equals("LIST") && "BLANK".equals(pType)){ // если сказано бланк а на самом деле это список
      pType=mdType;    
      }
      if (mdType.equals("BLANK") && "LIST".equals(pType)){ // если сказано спиисок а на самом деле это бланк
      pType=mdType;    
      }
      if (mdType.equals("CHART") && "LIST".equals(pType)){ // если сказано спиисок а на самом деле это CHART
      pType=mdType;    
      }
      if (mdType.equals("CHART") && "BLANK".equals(pType)){ // если сказано бланк а на самом деле это CHART
      pType=mdType;    
      }
      if (mdType.equals("HTML") && "LIST".equals(pType)){ // если сказано спиисок а на самом деле это HTML
      pType=mdType;    
      }
      if (mdType.equals("HTML") && "BLANK".equals(pType)){ // если сказано бланк а на самом деле это HTML
      pType=mdType;    
      }
      if (mdType.equals("GRID") && "LIST".equals(pType)){ // если сказано спиисок а на самом деле это CHART
      pType=mdType;    
      }
      if (mdType.equals("GRID") && "BLANK".equals(pType)){ // если сказано бланк а на самом деле это CHART
      pType=mdType;    
      }

    }
    frm.setProperty("FType", pType);
    if (pType.equals("TREE"))pType="LIST";
    if (formTemplates!=null){
    if (formTemplates.containsKey(pType)){
    for(String propcode:formTemplates.get(pType).properties.keySet()){
//     if (formTemplates.get(pType).properties.get(propcode).is_meta)       
     if (formTemplates.get(pType).properties.get(propcode).genesis!='A')       
        frm.addProperty(propcode,formTemplates.get(pType).getNonCompValue(propcode));
    }
    }
    }
    return frm;
}

public void actionBreak() throws LuiException, SQLException{
//   GL_P_SQL.Stop_Percent; -- Закроем прогресс-бар  
if (progressBar!=null){
progressBar.stopProgress();
}
   for (String frmid:forms.keySet()){
     if (NVL(forms.get(frmid).getVal("Status"),"Loading").equals("Loading")){//Если форма ещё не запустилась...
     forms.get(frmid).close(null,"CANCEL"); //то благополучно удаляем её (на клиенте о ней ничего ещё не знают)
    //this.removeForm(frmid);
     } else if (forms.get(frmid).getVal("Status").equals("Closing")) { // Если форма собиралась закрыться...
         forms.get(frmid).setVal("Status","Run"); //то оставим её в живых и рабогтающих (клиент ещё не получил команд на удаление формы)
     }
   }
   // второй цикл по формам
   boolean is_need4break=true;
   for (String frmid:forms.keySet()){
    LuiAppForm frm=forms.get(frmid);
    if (is_not_null(frm.getVal("CurrentAction"))) {//Рассматриваем только те формы, в которых выполняется некоторое действие 
      is_need4break=true;  
     //теперь надо найти форму которая открылась по этому действию
      for (String childfrmid:forms.keySet()){
          LuiAppForm cfrm=forms.get(childfrmid);
          if (cfrm.getVal("ParentFormID").equals(frmid)) {
              if (cfrm.getVal("CallAction").equals(frm.getVal("CurrentAction"))){
                  is_need4break=false; // нашлась такая, то её не прерываем
                  // т.к. она вызвала другую и ждёт её завершения
              } 
          }
      }
    if (is_need4break) {  
    boolean b=frm.rollback(); // Откат к началу действия 
    // в этой точке надо освободить(закрыть) соединения открытые по данному действию
    for (int i=backList.size()-1;i>=0;i--){
        backList.get(i).back();
    }
    backList.clear();
    
     frm.setVal("CurrentAction",null); 
     //19.03.2020 frm.checkRunTimer(); //-- восстанавливаем таймер, если нужно 
    }
    }
   }
  pingRunTimer();//19.03.2020 
}

/**
 * Создание нового соединения с БД для приложения
 * @param pCode - код
 * @param pName - название
 * @param pSTC  - Session P-personal C-common; 
 * @param pUsername - username
 * @param pPassword - password
 * @param pDbURL - URL для связи с БД
 * @param pSchema - схема внутри БД
 
 * @return новое соединение с базой данных для приложения
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiAppConn addUserDbConnection(String pCode,String pName,String pSTC,String pUsername,String pPassword,String pDbURL,String pSchema) throws LuiException, SQLException{
    if ("LUI$".equals(pCode)) return null; // ни чего не делаем при попытке подмены служебного соединения
    if (userDbConnections.containsKey(pCode)){
        userDbConnections.get(pCode).disconnect();
        userDbConnections.remove(pCode);
    }
    LuiAppConn c=new LuiAppConn(this,pCode,pName,pSTC,pUsername,pPassword,pDbURL,pSchema);
    try {
    c.connect();
    } catch (SQLException ex){
        String sqlstate=ex.getSQLState();
        if (sqlstate.equals("28P01"))
        msgWriter.write("LUI-00009");    
         else
        msgWriter.write("LUI-00008",ex.getLocalizedMessage());
    }
    userDbConnections.put(pCode,c);
    if (is_null(getProperty("dbUser"))){ // если свойство "Пользователь БД" ещё не установлено
        setProperty("dbUser",c.username); // в принципе надо дать возможность менять это свойство программно т.е. из JS или другого языка
        setProperty("dbURL",c.dbURL);
        setProperty("dbSCHEMA",c.schema);
        setProperty("dbPassword",this.crypter.encrypt(c.password));
     
    }
    // если исчезло служебное соединение
          if (!userDbConnections.containsKey("LUI$")){
  LuiAppConn srvc = new LuiAppConn(this,"LUI$","LUI service connection","COMMON",metaConnPool.getUser(),"***","","lui");
  srvc.connPool=metaConnPool;
  userDbConnections.put("LUI$", srvc);
        }
    return c;
}
/**
 * Выделяет в строке код типа соединения и оператор, подлежащий выполнению
 * @param pString -- строка вида (CONN_CODE)select .... from ....
 * @param pDefConn -- соединение по умочанию
 * @param pForFormID -- для какой формы ищем соедение с БД
 * @return список 1-й элемент это код соединения 2-й элемент это оператор
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public ArrayList<String> getAppConnAndStmt(String pString,String pDefConn,String pForFormID) throws LuiException, SQLException{
    ArrayList<String> ac=null;
    if (is_not_null(pString)) {
    pString=RTrim(pString);
    if (pString.charAt(0)=='(') { // если начинается со скобки
            // ищем где она закрывается
            int cb=pString.indexOf(")"); 
            if (cb>0){// и если она закрывается
                String connCode=pString.substring(1,cb); // будем подозревать, что там между скобками спрятан код соединения
                if (userDbConnections.containsKey(connCode)) { // и если такое соединение найдено
                 ac=new ArrayList<>();
                 ac.add(connCode);
                 ac.add(pString.substring(cb+1));
                 return ac;
                 
                } else { 
                    msgWriter.write("LUI-00019", connCode,pString); // если не найдено
                }
                    
            }
        }
    }
    //(...) отсутствует берём, то что указано в pDefConn
    if (is_not_null(pDefConn)){
        if (userDbConnections.containsKey(pDefConn)){
          ac=new ArrayList<>();
          ac.add(pDefConn);
          ac.add(pString);
          return ac;
        }
    }
    //если соединение не указано в pDefConn, то пытаемся взять из пропертей формы
    String FrmID="";
    if (is_not_null(pForFormID))FrmID=pForFormID;
    else FrmID=getVal("FormID");
    if (is_not_null(FrmID) /*&& forms.containsKey(FrmID)*/){
     String fc=forms.get(FrmID).getProperty("Connection");
     if (is_not_null(fc)){
       if (this.userDbConnections.containsKey(fc)){  
          ac=new ArrayList<>();
          ac.add(fc);
          ac.add(pString);
          return ac; 
       }
     }
    }
     //если и у фомы нет коннекта берём любой кроме служебного
            for(String cc:userDbConnections.keySet()){
               if (!("LUI$".equals(cc))) {
                 ac=new ArrayList<>();
                 ac.add(cc);
                 ac.add(pString);
                 return ac;}
            }
    msgWriter.write("LUI-00020",pString);
    return ac;
}
public String getNextUniqueID(){
  sesUniqueID++;
  return String.valueOf(sesUniqueID);
}
public void run() throws LuiException, SQLException{
  
  LuiCollection params=new LuiCollection("POSITION:WorkCenter;");
  LuiDbConn lc=null;
  try{
      
  lc=this.metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
  params=lc.getSelectAllRowsTwoColsValMap("select code,v.value  from lui_t_element e , lui_t_value v \n"+
"where e.app_code='"+code+"'\n"+
"and e.form_code='"+getProperty("INITIAL_FORM")+"'\n"+
"and e.etype='PARAM'\n"+
"and e.element_code is null\n"+
"and v.app_code=e.app_code\n"+
"and v.form_code=e.form_code\n"+
"and v.config_Code='DEFAULT'\n"+
"and v.element_code=e.code\n"+
"and v.property_code ='VALUE'");
  params=null;
 if (params==null){
    params=new LuiCollection("POSITION$:WorkCenter;"); 
 }
 
  } finally {
    if (lc!=null) this.metaConnPool.SetFree(lc);
  }
    this.executeForm(getProperty("INITIAL_FORM"),params.toString(), "", getProperty("INITIAL_FORMTYPE")/*,"Y"*/);
}

/**
 * Запуск формы 
 * @param pMetaDataCode - код формы в метаданных; "" - если программная
 * @param pParams -- строка -коллекция парамтров
 * @param pParentFormID -- вызывающая форма
 * @param pFormType -- тип формы используется только если список желаем сделать деревом в остальных случаях можно null
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String executeForm(String pMetaDataCode, String pParams,String pParentFormID,String pFormType/*,String formStat*/) throws LuiException, SQLException{
 String parentWindowID="";
 String parentAction="";
 String parentActionRole="";
 String parentActionCommand="";
 String newFrmID="";
 String windowID="WIN"+getNextUniqueID();
 LuiClCommandGroup XMLcg=null;
 LuiClCommand XMLcmd=null;
 String formID=null;
 LuiAppForm frm=null;
 //LuiDbConn c=null;
 LuiAppForm pfrm=null;
 LuiAppElement pel=null;
 boolean this_is_detail=false;
 boolean doNotRemove=false;
this.luiSeance.licAtt(11,20);
 XMLcg=commandWriter.addGroup("");
     if (is_not_null(pParentFormID)){ //Вызов посредсвом выполнения действия в родительской форме
        for(String fid:forms.keySet()){ // просматриваем все формы с темже отцом
           if (forms.get(fid).getVal("ParentFormID").equals(pParentFormID)){
               if (forms.get(fid).getVal("Status").equals("Loading")){
                   //Закрываем формы от предыдущих попыток
                   //removeForm(forms.get(fid).getFormID());
                   forms.get(fid).close(null,null);
               }
           }
        }
      parentWindowID=forms.get(pParentFormID).getVal("WindowID");
      parentAction=forms.get(pParentFormID).getVal("CurrentAction");
      
      if (is_not_null(parentAction)){
        if (forms.get(pParentFormID).elements.containsKey(parentAction)){
          parentActionRole=forms.get(pParentFormID).elements.get(parentAction).getProperty("ROLE");
          parentActionCommand=forms.get(pParentFormID).elements.get(parentAction).getProperty("COMMAND");
          if (like(parentActionRole,"%AUTO%")/*&& !like(parentActionRole,"%WIZARD%")03 06 2021 комент ради действий двойного назначения*/ 
                  && !like(parentActionRole,"%CLOSE%") /* 11 06 2021 ради того чтобы окно закрылось с логоном после внешней авторизации */
                  ){
              
              windowID="";
              //formID=pParentFormID+"+"+parentAction;
              formID=Extract_Word(pParentFormID,1,"+")+"+"+
                       String.valueOf(Integer.valueOf(NVL(Extract_Word(pParentFormID,2,"+"),"0"))+
                       Math.abs(parentActionCommand.hashCode())+ 
                       Math.abs(parentAction.hashCode()));
                      //String.valueOf(Double.valueOf(parentActionCommand.hashCode()))+
              //formID=pParentFormID+"+"+String.valueOf(parentActionCommand.hashCode())+"+"+parentAction;
              pfrm=forms.get(pParentFormID);
              pfrm.setVal("CurrentAction","");//Детали не взводят флаг выполнения действия
              pel=pfrm.elements.get(parentAction);
              this_is_detail=true;
              if (forms.containsKey(formID) && forms.get(formID).getType().equals("GRID")){
                  commandWriter.removeGroup(formID);
                  XMLcmd=XMLcg.addCommand("ClearGrid");
                  XMLcmd.addScaPar("FormID",formID);
                  //for (String ek:)
                  doNotRemove=true;
              }
          }
        }
      }
    }
    LuiCollection paramCollection=new LuiCollection(pParams);
    if (is_not_null(pMetaDataCode)){// если форма метаданковая
        if (is_null(formID)){
            formID="FRM"+getNextUniqueID();
        }
    
        // узнаем конфигурацию формы
    String config=NVL(paramCollection.Get_Element("CONFIG"),"DEFAULT");

    // создаём форму и заносим в массив форм
    if (!forms.containsKey(formID)) { // если формы с этим formID ещё нет
    frm=addForm(formID,pMetaDataCode,pFormType/*,formStat*/); // создаём новую
    
    newFrmID=formID;
    } else { // если есть такая - берём её
    frm=forms.get(formID);
    
    if (this_is_detail && 
            !frm.getType().equals("GRID")) {// если это деталь но не GRID
      if (pMetaDataCode.equals(frm.getProperty("MetadataCode"))) { // если совпадают коды
      if (config.equals(frm.getProperty("ConfigCode"))) { // и конфигурации
       if (frm.getVal("Status").equals("Run"))  { 
       //поищем в детальной форме действие с ролью DETAIL
         for (int di=0;di<frm.elements.size();di++){
           LuiAppElement el=frm.elSort(di);
           if (el.getType().equals("ACTION") && el.getParent().equals("Form") &&
                   like(el.getProperty("ROLE"),"%DETAIL%") ) {
               if (el.isActionActive()){
               // да есть такое действие , тогда 
               // начинаем вычислять новую сигнатуру параметров вызывающего действия
        String currentParamsHash=String.valueOf(pel.getProperty("COMMAND").hashCode()); ;
        for(int pi=0;pi<pfrm.elements.size();pi++){ // циклим все элементы формы
            LuiAppElement par=pfrm.elSort(pi);
            if (par.getParent().equals(pel.getCode()) && par.getType().equals("PARAM")){// если это параметр действия el
                //String v=par.getProperty("VALUE");
                currentParamsHash=String.valueOf(Integer.valueOf(currentParamsHash)+(par.getCode()+":"+par.getProperty("VALUE")+";").hashCode());
            }
        }
        pel.setProperty("ParamsHash",currentParamsHash);
               //и устанавливаем входные параметры этой детальной формы
               for (int i=0;i<frm.elements.size();i++){
                   LuiAppElement ppel=frm.elSort(i);
                   if (ppel.getType().equals("PARAM")&& ppel.getParent().equals("Form")){
                     ppel.properties.get("Value").clearCompVal();
                     if (paramCollection.containsKey(ppel.getCode())){
                       ppel.addProperty("Value", paramCollection.get(ppel.getCode()));
                     }
                   }
               }
               frm.runAction(el.getCode());//выполняем это действие в вызываемой форме и всё
               return "";
           }
         }
        }
      }
      }
      }   
    }
    }
    
    setVal("FormID",formID); // текущая форма
    frm.setVal("Status","Loading");//Флаг стадии загрузки. Если что-то случится на этой стадии, то удалим загруженное
    frm.setVal("CommitMode",getVal("COMMIT")); //Все действия будут выполняться в этом режиме Commit;
    frm.setProperty("ConfigCode", config);
    }
    else { // форма программная
        // значит к этому моменту addForm уже выполнен и ID формы лежит в getVal("FormID"
        formID=getVal("FormID");
        frm=forms.get(formID);

    }
    if (is_not_null(pParentFormID)){ //Вызов посредсвом выполнения действия в родительской форме
      if (forms.containsKey(pParentFormID)){
        pfrm=forms.get(pParentFormID);
        frm.setConnectionProperty(pParams);
        frm.setVal("ParentFormID",pParentFormID);
        frm.setVal("ParentFormType",pfrm.getType());
        
      if (is_not_null(parentAction)) {
          frm.setVal("CallAction", parentAction);
           pfrm.elements.get(parentAction).setProperty("CalledFormID",formID);
          String chco=pfrm.elements.get(parentAction).getProperty("CHID_COMMIT");
          if ("Y".equals(chco)) {
              frm.addProperty("COMMIT","commit");
          } else if ("N".equals(chco)) {
              frm.addProperty("COMMIT","Insert into LUI_T_DUMMY_COMMIT values (null)");
          } else {
              frm.addProperty("COMMIT",pfrm.getProperty("COMMIT"));
          }
          
      }
      if (pfrm.getVal("IsNeedLock").equals("Y")) {//требовалась блокировка, и не получилось заблокировать
          frm.setVal("ReadOnly", "Y");
      } else {
          frm.setVal("ReadOnly", "N");
      }
    }   
    }
  if (is_not_null(pMetaDataCode)){// если форма метаданковая, то загружаем её
      if (like(pMetaDataCode,"O2P_L_OBJECT%")){
          System.out.println("before load:"+System.currentTimeMillis()); // perfomance

      }
      loadFormFromMetadata(formID, pParams, "REQUERY");
      if (frm.getType().equals("GRID")){
          frm.prepareMatrixTemplEl("PARAM");
          if (this_is_detail) { // если матрица - деталь
        //и устанавливаем входные параметры этой детальной формы
        for (String pcode:paramCollection.keySet()){
            if (!frm.elements.containsKey(pcode))continue;
                   LuiAppElement ppel=frm.elements.get(pcode);
                   if (ppel.getType().equals("PARAM")&& ppel.getParent().equals("Form")){
                       ppel.properties.get("Value").clearCompVal();
                       ppel.addProperty("Value",paramCollection.get(pcode));
                   }
        }
               
    }
      }
     if (like(pMetaDataCode,"O2P_L_OBJECT%")){
          System.out.println("after load:"+System.currentTimeMillis()); // perfomance

      }
  }
       
  
  
  if (is_not_null(windowID)){  // если надо открыть новое окно - открываем
     
      LuiAppWindow w=new LuiAppWindow(this,windowID,parentWindowID);
      windows.put(windowID, w);
      
      if (is_not_null(parentWindowID)) { // если есть родительское окно - гасим его
        
        if (like(parentActionRole,"%NOTCOVER%"))
        windows.get(parentWindowID).disCover(XMLcg,pParentFormID);
        else
        windows.get(parentWindowID).disCover(XMLcg,null);
      }
      XMLcmd=XMLcg.addCommand("Window"); // команда создания окна
      XMLcmd.addScaParameter("WindowID", windowID,"");
      
      frm.setVal("MasterForm","Y");
      
     /* if (paramCollection.Exists_Element("WINDOW_TYPE$")){
          frm.setProperty("WindowType", paramCollection.Get_Element("WINDOW_TYPE$"));
      } */
        if (frm.elements.containsKey("WINDOW_TYPE$")){
            if (frm.elements.get("WINDOW_TYPE$").getType().equals("PARAM") && frm.elements.get("WINDOW_TYPE$").getParent().equals("Form")) {
             frm.setProperty("WindowType", frm.elements.get("WINDOW_TYPE$").getProperty("Value"));
            }
        }  
      
      XMLcmd.addScaPar("WindowType",NVL(frm.getProperty("WindowType"),"Common"));
      String winMax=frm.getVisualPref("","WinMax", "");
      /*if (is_null(winMax)){
          winMax=paramCollection.Get_Element("WINMAX$");
      }*/
      if (is_null(winMax)){
        if (frm.elements.containsKey("WINMAX$")){
          if (frm.elements.get("WINMAX$").getType().equals("PARAM")){
              if (frm.elements.get("WINMAX$").getParent().equals("Form")){
                  winMax=frm.elements.get("WINMAX$").getValue();
              }
          }  
        }
      }
      String winTop=frm.getVisualPref("","WinTop", "");
      /*if (is_null(winTop)){
          winTop=paramCollection.Get_Element("WINTOP$");
      }*/
      if (is_null(winTop)){
        if (frm.elements.containsKey("WINTOP$")){
          if (frm.elements.get("WINTOP$").getType().equals("PARAM")){
              if (frm.elements.get("WINTOP$").getParent().equals("Form")){
                  winTop=frm.elements.get("WINTOP$").getValue();
              }
          }  
        }
      }
      if (is_not_null(winTop)){
          XMLcmd.addScaPar("Top",winTop);
      }
      String winLeft=frm.getVisualPref("","WinLeft", "");
     /* if (is_null(winLeft)){
          winLeft=paramCollection.Get_Element("WINLEFT$");
      }*/
      if (is_null(winLeft)){
        if (frm.elements.containsKey("WINLEFT$")){
          if (frm.elements.get("WINLEFT$").getType().equals("PARAM")){
              if (frm.elements.get("WINLEFT$").getParent().equals("Form")){
                  winLeft=frm.elements.get("WINLEFT$").getValue();
              }
          }  
        }
      }
      if (is_not_null(winLeft)){
          XMLcmd.addScaPar("Left",winLeft);
      }
      String winWidth=frm.getVisualPref("","WinWidth", "");
      /*if (is_null(winWidth)){
          winWidth=paramCollection.Get_Element("WINWIDTH$");
      }*/
      if (is_null(winWidth)){
        if (frm.elements.containsKey("WINWIDTH$")){
          if (frm.elements.get("WINWIDTH$").getType().equals("PARAM")){
              if (frm.elements.get("WINWIDTH$").getParent().equals("Form")){
                  winWidth=frm.elements.get("WINWIDTH$").getValue();
              }
          }  
        }
      }
      String WinHeight=frm.getVisualPref("","WinHeight", "");
     /* if (is_null(WinHeight)){
          WinHeight=paramCollection.Get_Element("WINHEIGHT$");
      }*/
      if (is_null(WinHeight)){
        if (frm.elements.containsKey("WINHEIGHT$")){
          if (frm.elements.get("WINHEIGHT$").getType().equals("PARAM")){
              if (frm.elements.get("WINHEIGHT$").getParent().equals("Form")){
                  WinHeight=frm.elements.get("WINHEIGHT$").getValue();
              }
          }  
        }
      }
      XMLcmd.addScaPar("Width",winWidth);
      XMLcmd.addScaPar("Height",WinHeight);
      XMLcmd.addScaPar("Max", winMax);
      String maxButton="";
      
     /* if (paramCollection.Exists_Element("WINMAXBUTTON$")){
          maxButton=paramCollection.Get_Element("WINMAXBUTTON$");
      } else { */
         if (frm.elements.containsKey("WINMAXBUTTON$")){
          if (frm.elements.get("WINMAXBUTTON$").getType().equals("PARAM")){
              if (frm.elements.get("WINMAXBUTTON$").getParent().equals("Form")){
                  maxButton=frm.elements.get("WINMAXBUTTON$").getValue();
              }
          }  
        } 
      //}
      if (frm.getProperty("WindowType").equals("Logon") && maxButton.equals(""))maxButton="N";
      if (is_not_null(maxButton)){
          XMLcmd.addScaPar("MaxButton", maxButton);
      }
      String closeAction=frm.checkHotKey("CLOSE", "Form");
      if (is_not_null(closeAction)){
          if (frm.elements.get(closeAction).getNonCompValue("ACTIVE").equals("Y")){ //если действие всгда активно
           if (is_null(frm.elements.get(closeAction).getNonCompValue("PROC_INITIALLY"))){ // и нет программы "перед"
               if (is_null(frm.elements.get(closeAction).getNonCompValue("COMMAND"))){ // и нет команды
                   if (is_null(frm.elements.get(closeAction).getNonCompValue("PROC_FINALLY"))){ // и нет программы "после"
                       if (frm.elements.get(closeAction).getNonCompValue("AFTER").equals("NONE")){ //и нет действий "после"
      XMLcmd.addScaParameter("CloseButton","N",""); // то кнопку закрытия не рисуем                    
                       }
                   }
               }
           }
          }
      }
      if (paramCollection.Exists_Element("POSITION$")){
          frm.setVal("Position", paramCollection.Get_Element("POSITION$"));
      } else {
        if (frm.elements.containsKey("POSITION$")){
            if (frm.elements.get("POSITION$").getType().equals("PARAM") && frm.elements.get("POSITION$").getParent().equals("Form")) {
             frm.setVal("Position", frm.elements.get("POSITION$").getProperty("Value"));
            }
        }  
      }
          
      String pos=frm.getVal("Position");
      if (is_not_null(pos)) {
          if (!"TopLeft".equals(pos)) {
          XMLcmd.addScaParameter("Position",frm.getVal("Position"),"Cascade");
          XMLcmd.addScaParameter("ParentID", parentWindowID,"");
          }
      } else {
          XMLcmd.addScaParameter("ParentID", parentWindowID,"");
      }
  } else { // не надо нового окна
     if (!doNotRemove)
     commandWriter.removeGroup(formID);// удалим группу комманд, которая носит имя создаваемой формы 
  }
 XMLcmd=XMLcg.addCommand("Form");
 XMLcmd.addScaParameter("FormID",formID,"");
 String xmlFtype=LuiString.withCapitalLetter(frm.getType());
 if (xmlFtype.equals("Blank"))xmlFtype="Blanc";
 if (xmlFtype.equals("Menu"))xmlFtype="Tree";
 XMLcmd.addScaParameter("Type",xmlFtype,"");
 String ParentID=null;
 if (is_not_null(pParentFormID)) {
     if (forms.containsKey(pParentFormID)){
      if (is_not_null(forms.get(pParentFormID).getVal("RootFormID"))){
         ParentID=forms.get(pParentFormID).getVal("RootFormID");
      } else ParentID=pParentFormID;
     }
 }
 if (ParentID!=null) XMLcmd.addScaParameter("ParentID",ParentID,"");
 if (is_null(windowID)){ //Используем окно родительской формы
     windowID=parentWindowID;
     String pos="Bottom";
     if (like(parentActionRole,"%AUTOTABBOTTOM%"))
      pos="TabBottom";
     else
     if (like(parentActionRole,"%AUTOTABRIGHT%"))
      pos="TabRight";
     else
     if (like(parentActionRole,"%AUTOBOTTOM%"))
      pos="Bottom";
     else
     if (like(parentActionRole,"%AUTORIGHT%"))
      pos="Right";
      else {
     if (!NVL(pFormType,frm.getType()).equals("LIST")) {
         pos="Right";      
     }
     if (is_not_null(pParentFormID)) {
        if (forms.get(pParentFormID).getType().equals("MENU")){
           pos="Right";      
        } 
      }
     }
     
         frm.setVal("Position",pos);
     
     XMLcmd.addScaPar("Position",pos);
     XMLcmd.addScaParameter("Width",frm.getVisualPref("","ChildWidth", "") , "");
     XMLcmd.addScaParameter("Height",frm.getVisualPref("","ChildHeight", "") , "");
     
 }
 frm.setVal("WindowID",windowID);
 XMLcmd.addScaParameter("WindowID",windowID,"");
 if (is_not_null(windowID)){
     commandWriter.removeCommadFromAllGroups("Focus");
     XMLcmd=XMLcg.addCommand("Focus");
     XMLcmd.addScaParameter("FormID",NVL(frm.getVal("RootFormID"),formID) ,"");
 }
 //Загрузка пользовательской конфигурации
 if (is_not_null(paramCollection.Get_Element("DefCfg"))) {
     frm.loadUserConfig(paramCollection.Get_Element("DefCfg"));
 }
     if (like(pMetaDataCode,"O2P_L_OBJECT%")){
          System.out.println("before reload:"+System.currentTimeMillis()); // perfomance

      }
 // если список или матрица на закладке установим параметр ожидания активизации закладки (для оптимизации)
 if (like(frm.getVal("Position"),"Tab%") && (frm.getType().equals("LIST")||frm.getType().equals("GRID"))){
     frm.setVal("WaitingForFormActivate","Y");
 }
 frm.reloadForm(); //выполняем певичное отображение элементов формы
 
 
 
 if (like(pMetaDataCode,"O2P_L_OBJECT%")){
          System.out.println("after reload:"+System.currentTimeMillis()); // perfomance

      }
 // проверим нужность системного меню
 XMLcmd=XMLcg.addCommand("Window");
 XMLcmd.addScaParameter("WindowID", windowID, "");
 String visibleSysMenu="N";
 for(String fc:forms.keySet()){
    if (windowID.equals(forms.get(fc).getVal("WindowID"))) {
      if (NVL(forms.get(fc).getVal("SysMenu"),"Y").equals("Y")){
          visibleSysMenu="Y";break;
      } 
    }
 }
 XMLcmd.addScaPar("SysMenu",visibleSysMenu); // иконку системного меню
 
 //  DBMS_APPLICATION_INFO.set_action(GetWindowTitle(WindowID));
 //frm.setVal("Status","Run");-- Это уже сделано в reloadForm см.выше
 return newFrmID;
}


    /**
     * Загрузка формы из мета данных
     *
     * @param pFormID - id формы
     * @param pParams - строка параметров
     * @param pIsInteractive - в результате чего загружается
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void loadFormFromMetadata( String pFormID, String pParams, String pIsInteractive) throws SQLException, LuiException{
   LuiAppForm frm = forms.get(pFormID);
   String metadataApp=frm.getProperty("MetadataApp");
   if ("REFRESH".equals(pIsInteractive)){
     frm.checkRows.clear();
     frm.el4order.clear();
//     frm.elList.clear();
     frm.listQBE.clear();
     frm.elements.clear();
//     frm.properties.clear();
     frm.setVal("UserConfiguration", null);
    } else if ("REQUERY".equals(pIsInteractive)) {
        // ни чего не делаем
    } else {
      String Status=getVal("Status");
       removeForm(pFormID);
       setVal("Status",Status);
    }
    frm.setVal("Changed","N");
    frm.setVal("TransitParam",pParams);
    String key=metadataApp+"."+frm.code+"."+frm.getConfigCode();
    if (frmCacheLifeTime==0) {
       frm.load(metadataApp,false);
       frm.loadParams(pParams);
       frm.resetRecalc();
       //frm.resetChangeFlag();
    } else {
          if (frmCacheCrTime!=null && 
                (System.currentTimeMillis()-frmCacheCrTime)>frmCacheLifeTime){
            frmCache.clear();frmCacheList.clear();frmCacheCrTime=null;
        } 
       
    if (!this.frmCache.containsKey(key)){
        if (frmCacheList.size()>20){
             frmCache.remove(frmCacheList.get(0));
             frmCacheList.remove(0);
        }
        LuiAppForm cfrm=new LuiAppForm(this,frm.code,frm.getType());
        cfrm.addProperty("ConfigCode", frm.getConfigCode());
        cfrm.load(metadataApp,false);
        frmCache.put(key, cfrm);
        frmCacheList.add(key);
        if (frmCacheCrTime==null)frmCacheCrTime=System.currentTimeMillis();
    } 
    LuiAppForm cfrm=frmCache.get(key);
       for (String cfp:cfrm.properties.keySet()){
          LuiAppProp cp=cfrm.properties.get(cfp);
          LuiAppProp p=frm.addProperty(cfp,cfrm.properties.get(cfp).currentValue); 
          p.changeFlag=cp.changeFlag;
          p.currentValue=cp.currentValue;
          p.gen4menu=cp.gen4menu;
          //p.is_meta=true;
          p.gen4html=cp.gen4html;
          p.genesis=cp.genesis;
          p.recalculate=cp.recalculate;
          p.startedValue=cp.startedValue;
          if (p.recalculate!= null) { // Если динамика
      p.recalculate=true;      
      p.changeFlag=true;
      }
         
       }
       for (String cfe:cfrm.elements.keySet()){
           LuiAppElement cel=cfrm.elements.get(cfe);
           LuiAppElement el=frm.addElement(cel.Code,cel.getType(),cel.sourceSeq,cel.getProperty("Name"),cel.getProperty("Parent"));
         for (String cfep:cel.properties.keySet()){
             LuiAppProp cp=cel.properties.get(cfep);
          LuiAppProp p=
             el.addProperty(cfep,cel.properties.get(cfep).currentValue);
          p.changeFlag=cp.changeFlag;
          p.currentValue=cp.currentValue;
          p.gen4menu=cp.gen4menu;
          //p.is_meta=true;
          p.gen4html=cp.gen4html;
          p.genesis=cp.genesis;
          p.recalculate=cp.recalculate;
          p.startedValue=cp.startedValue;
          if (p.recalculate!= null) { // Если динамика
      p.recalculate=true;      
      p.changeFlag=true;
      }
         }  
       }
    frm.loadParams(pParams);
    }
    /*
    for (String fpc:frm.properties.keySet()) frm.properties.get(fpc).clearCompVal();
    for (String fec:frm.elements.keySet()){
        for (String epc:frm.elements.get(fec).properties.keySet()) 
            frm.elements.get(fec).properties.get(epc).clearCompVal();
    }*/
}
/**
 * Закрывает форму и удаляет её
 * @param pFormID идентификатор формы
 */
public void removeForm(String pFormID){
   //  UI_P_LIST.Clear_Cheking(WorkID, null);
   if (forms.containsKey(pFormID)){
       try{
      LuiAppForm frm=forms.get(pFormID);
       frm.clearChecking();
      String cc=frm.getProperty("Connection");
      if (is_not_null(cc)){
        if (userDbConnections.containsKey(cc)){
            LuiAppConn ac=userDbConnections.get(cc);
            /*if (ac.getSessionMode()!='C'){
              ac.currentConnection.CloseAllCursors();
            }*/
            if (is_not_null(frm.getVal("MainCursor")) && ac.getSessionMode()!='C'){
                ac.currentConnection.CursorClose(Integer.valueOf(frm.getVal("MainCursor")));
            }
        }
       
               }           
      if (!frm.getVal("Status").equals("Closing") &&
          !frm.getVal("Status").equals("Loading")
              ){
        frm.close(null, null);  
      }
// чистим зависимости
      for (String elcode:frm.elements.keySet()){//крутим элементы
        LuiAppElement el=frm.elements.get(elcode);
        for (String epcode:el.properties.keySet()){// и их св-ва
             LuiAppProp ep=el.properties.get(epcode); // берём св-во
             
           for (LuiAppProp pfrom:ep.dependsFrom.keySet()){ // от чего оно зависит?
               if (pfrom.depends.containsKey(ep)){ // есть на него ссылка?
                 pfrom.depends.remove(ep);
               }
           }
             
        }
      }
       forms.remove(pFormID);
       } catch(Exception e){}
    
   }
}
/**
 * Отображение системного меню
 * @param pWindowID идентификатор окна
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void systemMenu(String pWindowID) throws LuiException, SQLException{
    this.changeSCN("SystemMenu");
    LinkedHashMap <String,LuiCollection> sysMenu=new LinkedHashMap<>();
    //int item_seq;
    String item_code;
    String item_name;
    String group_code;
    LuiAppElement ael=null;
    int items4Show=0;
    
    int listCnt=0; //количество списков
    int treeCnt=0; //количество деревьев
    int blankCnt=0; //количество бланков
    int chartCnt=0; //количество диаграмм
    int gridCnt=0; //количество матриц
    LuiClCommandGroup XMLcg=commandWriter.addGroup("SystemMenu");
    LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
    //каково количество списков в окне?
    for (String frmcode:forms.keySet()){ //крутим формы окна для которых есть системное меню
        if (forms.get(frmcode).getVal("WindowID").equals(pWindowID) && 
                NVL(forms.get(frmcode).getVal("SysMenu"),"Y").equals("Y")){
      switch (forms.get(frmcode).getType()){
          case "LIST": listCnt++;break;
          case "TREE": 
          case "MENU":
              if ("Y".equals(forms.get(frmcode).getVal("MasterForm"))){
              treeCnt++;
              }
              break;
          case "BLANK": blankCnt++;break;
          case "CHART": chartCnt++;break;
          case "GRID":gridCnt++;break;
      }  }
    }
    for (String frmcode:forms.keySet()){ //крутим формы
      LuiAppForm frm=forms.get(frmcode);
      
      if (frm.getVal("WindowID").equals(pWindowID)&& 
                NVL(forms.get(frmcode).getVal("SysMenu"),"Y").equals("Y")){ // которые находятся в указанном окне
         if (!frm.getType().equals("GRID"))frm.elSort(0);
         for (int i=0;i<frm.elSortedList.size();i++) { // крутим элементы
           LuiAppElement el=frm.elSortedList.get(i);
           if (el.getType().equals("ACTION")){
            if (like(el.getCode(),"SYSMENU$%")){
              //if (el.getProperty("ACTIVE").equals("Y")){
              if (el.getPropertyIfNotExists("DISPLAY",el.getProperty("ACTIVE")).equals("Y")){
                  items4Show++;
                  ael=el;
                  item_code=el.getCode()/*.replace("SYSMENU_","")*/;
                  item_name=el.getProperty("PROMPT");
                  group_code="Usr";
                  switch (item_code){
                      case "SYSMENU$ENTERQUERY":
                      case "SYSMENU$EXECUTEQUERY":
                      case "SYSMENU$SORTS":
                      case "SYSMENU$ROWCOUNT":
                      case "SYSMENU$PRINT":
                      case "SYSMENU$SAVECONFIG":
                      case "SYSMENU$CHART":
                      case "SYSMENU$DATA_GROUPING":
                      //case "SYSMENU$RT-CHARTS":
                  if (listCnt>1){
                    item_name=item_name+"|"+NVL(frm.getProperty("HEADER"),NVL(frm.getProperty("ConfigName"),frm.getCode()));
                  } 
                  group_code="List";
                  break;
                      case "SYSMENU$QBENOT":
                  group_code="List";
                  break;
                      case "SYSMENU$CLEARVISUAL":
                      case "SYSMENU$FULLPATH":
                      case "SYSMENU$CALCULATOR":
                  group_code="AddOns";
                  break;
                      case "SYSMENU$CHOOSELANGUAGE":
                  group_code="Lang";
                 /* if (Integer.valueOf(getVal("NLSnumberOfLanguages"))>1){
                   LuiCollection langs=new LuiCollection(getVal("NLSlanguages"));
                   for (String lk:langs.keySet()){
                  if(!sysMenu.containsKey(item_name+"|"+langs.get(lk))){
                  XMLcmd=XMLcg.addCommand("AddMenuItem");
                  XMLcmd.addScaParameter("Type", "System",null);
                  XMLcmd.addScaParameter("Code",item_code,null);
                  XMLcmd.addScaParameter("Group",group_code,null);
                  XMLcmd.addScaParameter("Label",item_name+"|"+langs.get(lk),null);
                  sysMenu.put(item_name+"|"+langs.get(lk),true);
                  XMLcmd.addScaParameter("Params",lk,null);
                  if (lk.equals(getProperty("CURRENT_LANG"))){
                   XMLcmd.addScaParameter("Style","Checked",null);   
                  }
                   }
                  }}
                  continue;*/
                      case "SYSMENU$DEBUGON":
                      case "SYSMENU$DEBUGOFF":
                      case "SYSMENU$GETSTAT":
                      case "SYSMENU$GETTRACEINFO":
                  group_code="Supp";
                  break;          
                      case "SYSMENU$HELP":
                  group_code="Help"; 
                  if (listCnt+blankCnt+treeCnt+chartCnt>1){
                    item_name=item_name+"|"+NVL(frm.getProperty("HEADER"),NVL(frm.getProperty("ConfigName"),frm.getCode()));
                  }
                  break;
                      case "SYSMENU$ABOUT":
                  group_code="Help";
                  if (listCnt+blankCnt+treeCnt+chartCnt>1){
                    item_name=item_name+"|"+NVL(frm.getProperty("HEADER"),NVL(frm.getProperty("ConfigName"),frm.getCode()));
                  }
                  break;
                  }
                  // сочиняем коллекцию параметров пункта меню
                  LuiCollection menuItemCo=new LuiCollection("");
                  menuItemCo.Add_Element("Code", item_code);
                  menuItemCo.Add_Element("Group", group_code);
                  menuItemCo.Add_Element("FormID",frm.getFormID());
                  if (!el.isActionActive()) {
                  menuItemCo.Add_Element("Enable","N");
                  menuItemCo.Add_Element("Label",LuiString.RemoveIco(item_name));    
                  }
                  else {
                  menuItemCo.Add_Element("Enable","Y");    
                  menuItemCo.Add_Element("Label",item_name);    
                  }
                  if(!sysMenu.containsKey(item_name)){ //если такого элемента ещё нет
                      sysMenu.put(item_name, menuItemCo);
                  } else { //если уже есть
                      sysMenu.replace(item_name, menuItemCo);
                  }
                  
              } 
            }
           }
         }
      }
         
      
    }
    for (String iname:sysMenu.keySet()){
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Type", "System");
            XMLcmd.addScaPar("Code", sysMenu.get(iname).Get_Element("Code"));
            XMLcmd.addScaPar("Label", sysMenu.get(iname).Get_Element("Label"));
            XMLcmd.addScaPar("Group", sysMenu.get(iname).Get_Element("Group"));
            XMLcmd.addScaPar("FormID", sysMenu.get(iname).Get_Element("FormID"));
            if (sysMenu.get(iname).Get_Element("Enable").equals("N")) XMLcmd.addScaPar("Enable","N");
            XMLcmd.addScaPar("Params","WindowID:"+pWindowID+";");
        }
    if (items4Show==1 && ael!=null && !ael.isActionActive()){
        commandWriter.removeGroup("SystemMenu");
        ael.getForm().runAction(ael.getCode());
    } 
}

/**
 * Задаёт пользователю вопрос и получает от него ответ
 * @param pQuestion -- текст вопроса
 * @param pVariants -- варианты ответов в виде строковой коллекции
 * @param pProp -- свойство
 * @return ответ
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */   
public String ask(String pQuestion,String pVariants,LuiAppProp pProp) throws LuiException, SQLException{
this.luiSeance.licAtt(21,99999999);
    if (is_null(pQuestion))return "";
    String retVal=null;
    boolean NoAsActionBreak=false;
    if (!like(pQuestion,"Question:%")){
      if (is_null(pVariants)){
          if (like(RTrim(pQuestion," \n"),"%?")){
              NoAsActionBreak=true;
              pVariants="Y:"+this.NLS.NLS("RUS:Да;ENG:Yes;")+";N:"+this.NLS.NLS("RUS:Нет;ENG:No;")+";";
          } else {
              pVariants="Y:"+this.NLS.NLS("RUS:Продолжить;ENG:Continue;")+";N:"+this.NLS.NLS("RUS:Отменить;ENG:Cancel;")+";";
          }
      }  
    LuiCollection col=new LuiCollection("");
    col.Add_Element("Question", pQuestion);
    retVal=col.toString().substring(9)+pVariants;
    retVal=parser.plugins.get("Question").compute(retVal, pProp);
    } else {
    //retVal=parser.compute(pQuestion, pProp);
    retVal=parser.plugins.get("Question").compute(pQuestion.replace("Question:",""), pProp);
    }
    retVal=LuiParser.reMask(retVal);
    if (NoAsActionBreak){
    if (NVL(retVal,"N").equals("N")){
        //actionBreak();
          //String FormID=getVal("FormID");
            //  if (is_not_null("FormID")){
              //    if (forms.containsKey(FormID)){
                //      if (is_not_null(forms.get(FormID).getVal("CurrentAction"))){
                              throw new LuiException("break","","","",null,"",this);
                  //    }
                 // }
             // }
    }}
    
return retVal;
}
void makeCheckDetList(String pParent, ArrayList<LuiAppForm> pCheckList) throws LuiException, SQLException{
  String contextForkID;  
  for (String fc:forms.keySet()){
//      String wID=forms.get(fc).getVal("WindowID");
//      if (!windows.containsKey(wID))continue;
 //     if (pWin!=null && !pWin.equals(wID))continue;
  //    if (!windows.get(wID).getProperty("AbleStatus").equals("Enable")) continue;
      
        if (forms.get(fc).getVal("ParentFormID").equals(pParent) ||
                (is_null(forms.get(fc).getVal("ParentFormID")) && is_null(pParent))){
          if (is_null(forms.get(fc).getVal("ParentNode"))){//Рассматриваем только бланки, списки и корни деревьев
            contextForkID=fc;
            if (is_not_null(forms.get(fc).getVal("ContextForkID"))){
            contextForkID=forms.get(fc).getVal("ContextForkID");//-- Контекстная рабочая область. Актуально для дерева.
            }
            if (forms.containsKey(contextForkID)){
            if (forms.get(contextForkID).getVal("Status").equals("Run")){// Только работающие (активные) формы
             if (is_null(forms.get(contextForkID).getVal("CurrentAction"))){//Только формы, где сейчас не выполняется ни одного действия
               pCheckList.add(forms.get(contextForkID));
              
               // makeCheckDetList(contextForkID,pCheckList);
               
             }
             
             makeCheckDetList(contextForkID,pCheckList);
             
            }}
          }
        }
    }
}
public void checkDetails(String pParent) throws LuiException, SQLException{
 ArrayList<LuiAppForm> checkList=new ArrayList<LuiAppForm>();   
 makeCheckDetList(pParent,checkList);
 for (int i=0;i<checkList.size();i++)
 {
     checkList.get(i).checkFrmDetails();
 }
 /*String contextForkID;
 try {
    for (String fc:forms.keySet()){
        if (forms.get(fc).getVal("ParentFormID").equals(pParent) ||
                (is_null(forms.get(fc).getVal("ParentFormID")) && is_null(pParent))){
          if (is_null(forms.get(fc).getVal("ParentNode"))){//Рассматриваем только бланки, списки и корни деревьев
            contextForkID=fc;
            if (is_not_null(forms.get(fc).getVal("ContextForkID"))){
            contextForkID=forms.get(fc).getVal("ContextForkID");//-- Контекстная рабочая область. Актуально для дерева.
            }
            if (forms.get(contextForkID).getVal("Status").equals("Run")){// Только работающие (активные) формы
             if (is_null(forms.get(contextForkID).getVal("CurrentAction"))){//Только формы, где сейчас не выполняется ни одного действия
             forms.get(contextForkID).checkFrmDetails();
             }
             checkDetails(contextForkID);
            }
          }
        }
    }
        } catch (Exception ex){}
 */
 }

/**
 * Обработка выбора пункта меню
 * @param pType - тип иеню
 * @param pCode - код элемента типа ACTION
 * @param pFormID - ID формы
 * @param pParams - коллекция доп. параметров
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void choiceMenuItem(String pType, String pCode, String pFormID, String pParams) throws LuiException, SQLException{
  this.changeSCN("ChoiceMenuItem");
 LuiAppForm frm;
 LuiAppElement el=null;
 if (forms.containsKey(pFormID)) {
   frm=forms.get(pFormID);
   el=frm.findAndMute(pCode);
   if (el!=null){
    frm.setVal("CurrentNodeType","");
    frm.runAction(el.getCode());
   }
 }
            
    
}
/**
 * Установка контекста (значений элементов списка) в текущей ноде дерева - в соответствии с её состоянием
 * @param pWorkID идентификатор рабочей области
 * @param pContextStr строка контекста (?)
 * @param pOpenWorkID  идентификатор неизвестно чего (?)
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setContext4Node(String pWorkID,String pContextStr,String pOpenWorkID) throws LuiException, SQLException{
  
 if (is_not_null(pWorkID)){
     setVal("FormID",pWorkID);
   if (forms.containsKey(pWorkID)){
     LuiAppForm frm=forms.get(pWorkID);
     if (!"MENU".equals(frm.getType())){
     LuiCollection c =new LuiCollection(pContextStr);
     frm.setElementsValues(c);}
   }  
 }
 if (is_not_null(pOpenWorkID)){
     if (forms.containsKey(pOpenWorkID)){
     LuiAppForm frm=forms.get(pOpenWorkID);
     if (!"MENU".equals(frm.getType())){
     frm.setVal("CurrentRow","0");
     LuiCollection EmptyContext=new LuiCollection("");
     for(String elcode:frm.elements.keySet()){
         LuiAppElement el=frm.elements.get(elcode);
         if (el.getType().equals("ITEM")){
             EmptyContext.Add_Element(el.getCode(),"");
         }
     }
     setVal("FormID",pOpenWorkID);
     frm.setElementsValues(EmptyContext);
 }}
 }
}
/**
 * Сворачивание иерархии открытых форм узла дерева
 * @param pFormID - ID родительской формы
 * @param pNode - узел дерева
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void collapseNodeForms(String pFormID,String pNode) throws LuiException, SQLException{
   for (String frmc:forms.keySet()){
       LuiAppForm frm=forms.get(frmc);
      if (frm.getParentID().equals(pFormID)){
       String pn=NVL(frm.getVal("ParentNode"),"");
       if ( pn.equals(pNode) || NVL(LuiString.Extract_Word(pn,1,"#"),"").equals(pNode)){
       collapseNodeForms(frmc,pNode);
       frm.close("","");}
   }
   }
}

    /**
     * Ещё одно Сворачивание иерархии открытых форм узла дерева используется в
     * requery
     *
     * @param pFormID - ID родительской формы
     * @param notCheckParentNode - режим без проверки родительского узла
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void collapseNodeForms2(String pFormID, boolean notCheckParentNode) throws LuiException, SQLException{
   for (String frmc:forms.keySet()){
       LuiAppForm frm=forms.get(frmc);
      if (frm.getParentID().equals(pFormID)){
       if (notCheckParentNode || is_not_null(frm.getVal("ParentNode"))){
       collapseNodeForms2(frmc,true);
       frm.close("","");
       }
   }
   }
}
/**
 * Сохранение значения Value пользовательского свойства Property в контексте WindowID, FormID и Element
 * @param pWindowID идентификатор окна
 * @param pFormID идентификатор формы
 * @param pElement идентификатор элемента
 * @param pProperty идентификатор свойства
 * @param pValue значение
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setUserProperty(String pWindowID,String pFormID,String pElement,String pProperty,String pValue) throws LuiException, SQLException{
this.changeSCN("SetUserProperty");
    String frmID=pFormID;
    
 if (is_null(frmID) && is_not_null(pWindowID)){
     for (String key:forms.keySet()){
         if (forms.get(key).getVal("WindowID").equals(pWindowID)){
            if  (forms.get(key).getVal("MasterForm").equals("Y")){
              frmID=key;
            }
         }
     }
 }
if (is_not_null(frmID) && is_not_null(pProperty)){
    forms.get(frmID).changeVisualParam(pElement, pProperty, pValue);
}
}
/**
 * Сохраняет предпочтение пользователя
 * @param pFormCode код формы
 * @param pConfigCode код конфигурации
 * @param pElementCode код элемента
 * @param pPropertyCode код свойства
 * @param pUserCfgID идентификатор пользовательской конфигурации
 * @param pValue значение
 * @param pIns какой-то флаг
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setVisualParams(String pFormCode,String pConfigCode,String pElementCode,String pPropertyCode,String pUserCfgID,String pValue,boolean pIns) throws SQLException, LuiException{
if (pUserCfgID.equals("0"))pUserCfgID="";
    if (is_null(getAppUser()))return;
    LuiDbConn lc=null;
    String stmt=STMT4UPDATEUSRPREF;
    stmt=stmt.replace("<app_code>",code);
    stmt=stmt.replace("<user_name>",getAppUser());
    if (is_null(pFormCode)){
    stmt=stmt.replace("<form_code>","is null");
    }
    else{
    stmt=stmt.replace("<form_code>","='"+pFormCode+"'");    
    }
    if (is_null(pConfigCode)){
    stmt=stmt.replace("<config_code>","is null");    
    }else{
    stmt=stmt.replace("<config_code>","='"+pConfigCode+"'");        
    }
    if (is_null(pElementCode)){
    stmt=stmt.replace("<element_code>","is null");    
    }else{
    stmt=stmt.replace("<element_code>","='"+pElementCode+"'");        
    }
    stmt=stmt.replace("<property_code>", pPropertyCode);
    if (is_null(pUserCfgID)){
    stmt=stmt.replace("<usrcfg_id>","is null");
    } else {
    stmt=stmt.replace("<usrcfg_id>","='"+pUserCfgID+"'");
    }
/*    if (is_null(pValue)){
    stmt=stmt.replace("<value>","=null");
    } else {
    stmt=stmt.replace("<value>","='"+pValue+"'");    
    } */
    stmt=stmt.replace("<value>","='"+NVL(pValue,"")+"'");    
    try{
    lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
    if (lc.executeDML(stmt)==0 && pIns){
        stmt=STMS4INSERTUSRPREF;
        stmt=stmt.replace("<user_name>",getAppUser());
        stmt=stmt.replace("<app_code>",code);
        if (is_null(pFormCode)){
        stmt=stmt.replace("<form_code>","null");
        }
        else{
        stmt=stmt.replace("<form_code>","'"+pFormCode+"'");    
        }
        if (is_null(pConfigCode)){
        stmt=stmt.replace("<config_code>","null");    
        }else{
        stmt=stmt.replace("<config_code>","'"+pConfigCode+"'");        
        }
        if (is_null(pElementCode)){
        stmt=stmt.replace("<element_code>","null");    
        }else{
        stmt=stmt.replace("<element_code>","'"+pElementCode+"'");        
        }
        stmt=stmt.replace("<property_code>", pPropertyCode);
        if (is_null(pUserCfgID)){
        stmt=stmt.replace("<usrcfg_id>","null");
        } else {
        stmt=stmt.replace("<usrcfg_id>","'"+pUserCfgID+"'");
        }
 /*       if (is_null(pValue)){
        stmt=stmt.replace("<value>","null");
        } else {
        stmt=stmt.replace("<value>","'"+pValue+"'");    
        }   */
        stmt=stmt.replace("<value>","'"+NVL(pValue,"")+"'");    
        lc.executeDML(stmt);
    }
    }
    finally{
    metaConnPool.SetFree(lc);
    }
}

/**
 * Установка текущего языка
 * @param pLangCode -- код языка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setCurrentLANG(String pLangCode) throws LuiException, SQLException{
  if (NLS.langList.containsKey(pLangCode)){
      if (getProperty("CURRENT_LANG").equals(pLangCode))return;
      NLS.setCurrentLang(pLangCode);
      setVal("NLScurrentLanguage",NLS.currentLang.code);
      setProperty("CURRENT_LANG",NLS.currentLang.code);
      loadTemplates();
  } else{
     msgWriter.write("LUI-00031",pLangCode); 
  }
    
}
/**
 * Посылка изменений форм на клинета
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void changes() throws LuiException, SQLException{
/* Это неудачная попытка воспроизведения порядка столцов на сервере по
    результатам их перемещания на клиенте
    String frm2Sort=getVal("NeedSort");
if (forms.containsKey(frm2Sort)){
   forms.get(frm2Sort).elSortedList.clear();
   forms.get(frm2Sort).elSort(0);
   properties.remove("Var NeedSort");
} */
 // оптимизация для rt-chart
if (!eventReader.is_non_interval && eventReader.is_interval){ // если тоько интервал
 if (is_not_null(getVal("rtChart"))){ // и есть форма real-time графиков
   if (forms.containsKey(getVal("rtChart"))){
     LuiAppForm rtFrm=forms.get(getVal("rtChart"));  // берём её
     if (forms.get(getVal("rtChart")).RTdashboard!=null){ // если у неё есть машина графики
         String win=rtFrm.getVal("WindowID"); // возмём окно формы
          if (windows.containsKey(win)){
              if ("Enable".equals(windows.get(win).getProperty("AbleStatus"))) { // если оно активно
                  rtFrm.changes(); // берём только изменения графиков
                  return; // и уходим
              }
                      }            
     }  
   }  
 }   
}
// конец оптимизации для rt-chart
if (forms!=null){
ArrayList<String> formIDs;
boolean b=true;
while (b){
b=false;
// поскольку состав и количество форм могут измениться в ходе changes мы создадим перечень их ID
formIDs=new ArrayList<>();
for(String fc:forms.keySet()){
    formIDs.add(fc);
}        
for(int i=0;i<formIDs.size();i++){
      if (!forms.containsKey(formIDs.get(i))){b=true;break;}
      LuiAppForm frm=forms.get(formIDs.get(i));
      if (frm.changeFlag){
         if (frm.getVal("Status").equals("Run")){
           if (!eventReader.is_non_interval && eventReader.is_interval && compareString(frm.getVal("Is_QBE_Mode"),"Y")) continue;// если тоько интервал и форма в режиме запроса
           frm.changes(); 
           
           
           if (is_null(getVal("rtChart")) 
                   && !frm.getType().equals("TREE") // а это сюда добавлено 05.06.2021 в связи с тем, что если скакать по дереву то не закрывался детельный бланк вызванный в списке 
                   // уже покинутого узла дерева т.е. бланк уже был помечен Clising а это действие его реанимировало и получались активными два детальных бланка
                   // открытых из разных узлов дерева
                   )
           frm.checkFrmDetails(); // вот это добавлено для случая необходимости обновления деталей (это когда нет димаческих графиков)
                                  // а ситуацич такая Master-List Detail-Blanc
                                  // в мастере выхываем другой список и в нём что-то сделалось что данные бланка д.б. другия
                                  // этот другой список закрываем крестом
                                  // обновления детального бланка не выполняется и всё! Хот тресни. Поэтому добавил сюда.14092020
           frm.changeFlag=false;
           if  (formIDs.size()!=forms.size()){
               b=true;break;
           }
           
         } 
      }
    }
 
}
// и ещё раз
b=true;
while (b){
b=false;
// поскольку состав и количество форм могут измениться в ходе changes мы создадим перечень их ID
formIDs=new ArrayList<>();
for(String fc:forms.keySet()){
    formIDs.add(fc);
}        
for(int i=0;i<formIDs.size();i++){
      if (!forms.containsKey(formIDs.get(i))){b=true;break;}
      LuiAppForm frm=forms.get(formIDs.get(i));
      if (frm.changeFlag){
         if (frm.getVal("Status").equals("Run")){
           if (!eventReader.is_non_interval && eventReader.is_interval && compareString(frm.getVal("Is_QBE_Mode"),"Y")) continue;// если тоько интервал и форма в режиме запроса
           frm.changes(); 
           
           if (is_null(getVal("rtChart")) 
                   && !frm.getType().equals("TREE") // а это сюда добавлено 05.06.2021 в связи с тем, что если скакать по дереву то не закрывался детельный бланк вызванный в списке 
                   // уже покинутого узла дерева т.е. бланк уже был помечен Clising а это действие его реанимировало и получались активными два детальных бланка
                   // открытых из разных узлов дерева
                   )
           frm.checkFrmDetails(); // вот это добавлено для случая необходимости обновления деталей (это когда нет димаческих графиков)
                                  // а ситуацич такая Master-List Detail-Blanc
                                  // в мастере выхываем другой список и в нём что-то сделалось что данные бланка д.б. другия
                                  // этот другой список закрываем крестом
                                  // обновления детального бланка не выполняется и всё! Хот тресни. Поэтому добавил сюда.14092020
           frm.changeFlag=false;
           if  (formIDs.size()!=forms.size()){
               b=true;break;
           }
           
         } 
      }
    }
 
}
// конец ещё раза
}
}
/**
 * Обработка события Interval
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */

public void interval() throws LuiException, SQLException {
   if (is_null(getProperty("LUI Timer"))) return; 
   this.changeSCN("Interval");
   //boolean b=false;
   String Timer="";
   for(String fc:forms.keySet()){
      LuiAppForm frm=forms.get(fc);
      
      if (!frm.getVal("Status").equals("Run")) continue;
      if (is_null(frm.getVal("Timer"))) continue;
      if (frm.getVal("Is_QBE_Mode").equals("Y")) continue;
      if (is_not_null(frm.getVal("CurrentAction")) && frm.RTdashboard==null) continue; // и нет выполняемого действия и это не чарт
      String win=frm.getVal("WindowID");
          if (!windows.containsKey(win)) continue;
              if ("Enable".equals(windows.get(win).getProperty("AbleStatus")) 
                      || frm.RTdashboard!=null){// и если форма в активном окне или RT-чарт
               for (int i=0;i<frm.elements.size();i++){
                   LuiAppElement el=frm.elSort(i);
                   if (el.getType().equals("ACTION")) { // крутим действия
                     if (like(el.properties.get("ROLE").getNonCompValue(),"%TIME%")){ // если в исходном выражении св-ва ROLE замечен timer
                       if (like(el.getProperty("ROLE"),"%TIME%")){// и в свежевычесленном тоже
                           if (el.isActionActive()) { // и действие активно
                              frm.runAction(el.getCode());
                              //this.checkDetails("");
                           if(eventReader.is_non_interval || !frm.getFormID().equals(getVal("rtChart"))) {
                              frm.checkFrmDetails(); // только если не интервальное событие или не realTime - chart форма
                           }
                           }
                       }  
                     if (is_null(Timer)){
                         Timer="1000";  // было что-то похожее на timer  
                     }   
                       if (like(el.properties.get("ROLE").getNonCompValue(),"%TIME500%"))Timer="500";
                       
                     
                     }
                   }
               
              }
             }
        
   }      
       
 if (is_not_null(Timer)) {
 if (forms.containsKey(getVal("rtChart"))){
  if (forms.get(getVal("rtChart")).RTdashboard!=null){
      
      Timer=forms.get(getVal("rtChart")).RTdashboard.getDelayTime().toString();
      
  }   
 }
 LuiClCommandGroup XMLcg=this.commandWriter.addGroup("Timer");
 LuiClCommand XMLcmd=XMLcg.addCommand("Event"); 
 XMLcmd.addScaPar("EventCode", "Interval");
 XMLcmd.addScaPar("MS", Timer);
  }  else {
     setProperty("LUI Timer","");
 }
}
/**
 * Проверка необходимости взведения таймера на клиенте для обработки возможных действий формы WorkID с ACTION_ROLE = TIME
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void pingRunTimer() throws LuiException, SQLException{
    if (is_not_null(getProperty("LUI Timer"))) return;
    for (String fc:forms.keySet()){
        LuiAppForm frm=forms.get(fc);
        if (is_null(frm.getVal("Timer"))) continue;
        setProperty("LUI Timer","Y");
        LuiClCommandGroup XMLcg=this.commandWriter.addGroup("Timer");
        LuiClCommand XMLcmd=XMLcg.addCommand("Event"); 
        XMLcmd.addScaPar("EventCode", "Interval");
        XMLcmd.addScaPar("MS","0");
        return;
 
    }
}

/**
 * сбрасывает флаг изменённости
 */
public void resetChangeFlag(){
    if (this.changeFlag){
     for (String pkey:properties.keySet()){
       properties.get(pkey).changeFlag=false;
     }   
     for (String frmc:forms.keySet()){
       forms.get(frmc).resetChangeFlag();
     }
     this.changeFlag=false;    
    }
}
/**
 * Закрывает приложение
 */
public void close(){
 while (forms.size()>0){
 for (String frmc:forms.keySet()){
     removeForm(frmc);
     break;
 }}
if (this.luiSession!=null){
    luiSession.endSession(true);
    luiSession=null;
}
for(String c:userDbConnections.keySet()){
    if (!("LUI$".equals(c))){ // закрываем соединения кроме служебного
       userDbConnections.get(c).disconnect();
    }
   }
   userDbConnections.clear();// чистим массив прикладных соединений
if (windows!=null) windows.clear();
try{
Context.exit();
}catch(Exception ex){}
if (this.parser!=null){
    if (parser.plugins.containsKey("LuiFile")){
        try{
        parser.plugins.get("LuiFile").execute("CloseAll", null);
        } catch(Exception ex){}
    }
}
}

@Override
 /**
 * Возвращает текущий SCN
 */ 

 public int getCurrentSCN(){
  if (properties.containsKey("SCN")){
      try {
       return Integer.valueOf(getProperty("SCN"));
      } catch (Exception ex){
          return 0;
      }
  }
  else {
       return 0;    
  }  
}
 /**
  * Увеличивает текущий SCN
  * @param pEvent событие
  * @throws LuiException логическая ошибка
  * @throws SQLException ошибка выполнения SQL-запроса
  */
 public void changeSCN(String pEvent) throws LuiException, SQLException{
  Integer SCN;   
  setVal("SCN name",pEvent);  
  if (properties.containsKey("SCN"))
       SCN=Integer.valueOf(getProperty("SCN"))+1;
  else {
       SCN=1;    
  }
  setProperty("SCN",SCN.toString());
  
  
 }

@Override
public String getGenesis(){
 return "Application:"+code+";";    
}
@Override
public LuiApplication getApplication(){
    return this;
}
@Override
public LuiAppForm getForm(){
   if (is_not_null(getVal("FormID"))) return forms.get(getVal("FormID"));
    return null;
}

public String getDbUser() throws LuiException, SQLException{
    return getProperty("dbUser");
}
public String getAppUser() throws LuiException, SQLException{
    if (properties.containsKey("AppUser")) return getProperty("AppUser");
    return getDbUser();
}
public String getWorkPlace() throws LuiException, SQLException{
    return getProperty("WorkPlace");
}
/**
 * Вычисляет наличие права
 * @param pObjType - Тип объекта
 * @param pObj - Идентификатор объекта
 * @param pRgh - Право
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 * @return эффективные права
*/
public String getEffectiveRight(String pObjType,String pObj,String pRgh) throws LuiException, SQLException{ 
    if ("N".equals(getProperty("ACCESS_RIGHT"))) return "Y";
    String pUser=getAppUser();
    String stmt="";
    if (is_null(pUser)) return "Y";
    if (pUser.equalsIgnoreCase("LUI")) {
            return "Y";
        }
    if (pUser.equals(owner)){
            return "Y";
    }
  //  if (pObj.equals("SD_B_MSG+LUI_L_LINKED_FILE.FILETYPE_NAME")){
  //      String str="here";
  //  }
    String retVal="N";
    LuiDbConn c=null;
    try{
    c=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
    stmt="select lui_f_get_effective_right('"+pObjType+"','"+pObj+"','"+pRgh+"','"+pUser+"','"+code+"')";
    retVal=c.getSelectResult(stmt);
    } 
    catch (SQLException ex){
      this.msgWriter.sqlError(c, ex, null, stmt);
    }
    finally{
        if (c!=null){
         metaConnPool.SetFree(c);
        }
    }
    return retVal;
}
/**
 * Возвращает ID сеанса в LUI или null если сеанса ещё нет
 * @return ID сеанса в LUI
 */
public String getLuiSessionID(){
  if (this.luiSession!=null)return luiSession.id;
  return null;
}

/**
 * Возвращает URL для временного каталога сеанса
 * @return URL временного каталога сеанса
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getTmpDirURL() throws LuiException, SQLException{
    if (properties.containsKey("TmpDirURL")){
        String v=properties.get("TmpDirURL").getCurrentValue();
        if (is_null(v)) getTmpDir();
        return properties.get("TmpDirURL").getCurrentValue();
    } else {
        getTmpDir();
        if (properties.containsKey("TmpDirURL")){
            return properties.get("TmpDirURL").getCurrentValue();
        } else return "";
    }
        
}
/**
 * Возвращает имякаталога для хранения файлов приложения общих для всех сеансов этого приложения
 * @return имякатаорга для хранения файлов приложения общих для всех сеансов этого приложения
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getAppDir() throws LuiException, SQLException{
    String delimiter=System.getProperty("file.separator");
 if (is_null(AppFilePath)){ // если путь к файлам приложения ещё не известен
         if (!properties.containsKey("APP_FILES_PATH")){ //  и нет св-ва APP_FILES_PATH то возьмём из LUI
          LuiDbConn lc=null;
          try {
            lc  = metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());   
           AppFilePath=lc.getSelectResult("select lui_f_get_application_property('LUI','APP_FILES_PATH')");
          } 
            finally {
              metaConnPool.SetFree(lc);
          }
         } else { // если есть APP_FILES_PATH
           AppFilePath=getProperty("APP_FILES_PATH");
         }
  if (is_null(AppFilePath)){ // если так и не нашли
      msgWriter.write("LUI-00056"); // сообщим об этом
  }
 }
     File dir = new File(AppFilePath);
     if (AppFilePath.charAt(AppFilePath.length()-1)!=delimiter.charAt(0)){
          AppFilePath=AppFilePath+delimiter;
      }
     dir.mkdir();
  if (!like(AppFilePath,"%"+delimiter+code+delimiter)) { //если не содержит имя приложения
      // добавляем его через разделитель (если его нет)
      if (AppFilePath.charAt(AppFilePath.length()-1)!=delimiter.charAt(0)){
          AppFilePath=AppFilePath+delimiter;
      }
      AppFilePath=AppFilePath+code+delimiter;
      dir = new File(AppFilePath);
      dir.mkdir();
  }
 addProperty("AppFilePath",AppFilePath);
 
      
    return AppFilePath;
}
/**
 * Возвращает имя каталога для генерации временных файлов в течении сеанса
 * если каталога нет, то он будет создан
 * если создать не удаётся - будет ошибка
 * @return имя каталога для генерации временных файлов
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getTmpDir() throws LuiException, SQLException{
//    String tmpdir=System.getProperty("java.io.tmpdir");
  if (this.luiSession==null || is_null(getProperty("AppUser"))){ // попытка создания до начала сеанса
     msgWriter.write("LUI-00057","TmpDir"); // сообщим об этом
 } 

 if (is_null(TmpFilePath)) {
     String delimiter=System.getProperty("file.separator");
     TmpFilePath=AppFilePath;
     TmpFilePath=TmpFilePath+getProperty("AppUser")+delimiter;
     File dir = new File(TmpFilePath);
     dir.mkdir();
     TmpFilePath=TmpFilePath+luiSession.id+delimiter;
     dir = new File(TmpFilePath);
     dir.mkdir();
     if (dir.exists()==false){ // если не удалось создать
      msgWriter.write("LUI-00058","TmpFilePath"); // сообщим об этом   
     }
 String Tmp=getProperty("ApplicationURL")+"appfiles"+"/"+code+"/"+getProperty("AppUser")+"/"+luiSession.id+"/";
 addProperty("TmpDirURL",Tmp);
 }
 
     return TmpFilePath;
 
}

    /**
     * Разбор командной строки на отдельные параметры. 
     * Параметр - последовательность непробельных символов
     * Пробельные символы - пробел, табуляция, перевод строки
     * Экранировние:
     *    - символ \ экранирует (отменяет специальное значение) следующего символа
     *    - символы ' и " отменяют специальное значение символов до следующего такого же символа
     * 
     * @param pCmdLine строка для разбора
     * @return список из параметров
     */
    public  LinkedList<String> parseCommandLine( String pCmdLine ) throws LuiException {
        
        boolean aSpaces = true;  // режим "в пробелах"
        boolean aEscape = false; // режим "в экране"
        char aWaitFor = 0;        // режим - "в ожидании символа"
        
        StringBuilder aSB = null;  // накопитель значения параметра, создаётся при каждом смене режитма "в пробелах"
        LinkedList<String> aParts = new LinkedList<>(); // накопитель найденных значений параметров
        
        for(char c: pCmdLine.toCharArray()) { // посимвольный разбор строки
            
            if(aSpaces) {       // если "в проблах" - ничего не делаем, пока не найдём непробельный символ
                switch(c) {
                    case ' ':   case '\t': case '\n': continue;
                    default:    aSpaces = false;
                                aSB = new StringBuilder();
                                break;
                }
            }
            
            if(aEscape) {      // если предыдущий символ - "экран", то просто добавляем текущий символ к текущему значению
                switch(c) {
                    case ' ': case '\t': case '\n': case '\\': case '"': case '\'':  aSB.append(c); break; 
                    default:  aSB.append('\\').append(c); break;
                }
                aEscape = false; 
                continue; 
            }
            
            if(aWaitFor != 0) {  // если внутри "строки" - работаем с одиночными жкранами, и ждём закрывающего символа
                if(c== '\\') { aEscape = true; continue; }
                if(c==aWaitFor) {
                    aWaitFor = 0;
                    continue;
                }
                aSB.append(c); 
                continue;
            }
            
            
            switch(c) {
                case ' ': case '\t': case '\n':   // переход в режим "пробелов"
                    aSpaces = true;
                    aParts.add(aSB.toString());
                    aSB = null;
                    break;
                case '\\': aEscape = true;  break;     // переход в режим экрана
                case '\'': case '\"': aWaitFor = c; break;  // переход в режим строки
                default: aSB.append(c); break; // добавление символа к значению
            }
        }
        
        // строка закончилась, проверяем режимы...
        if( aWaitFor != 0) { 
            this.msgWriter.write("LUI-00129","("+pCmdLine+")");
            //throw new RuntimeException(String.format("Unbalanced [%c] character", aWaitFor)); 
                                      // Обнаружены непарные кавычки (aWaitFor)
                           }
        if( aEscape ) { 
            this.msgWriter.write("LUI-00130","("+pCmdLine+")");
                                      // Обнаружена незакрытая escape-последовательность "\" в конце строки
        }
        if( !aSpaces ) { aParts.add(aSB.toString()); }
        
        return aParts;
    }


    /**
     * Выполнение команд OC
     *
     * @param commandCode - код комады OC
     * @param pParams - строка параметров (коллекция)
     * @return результат выполнения (?)
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
    public String HOST(String commandCode, String pParams) throws LuiException, SQLException {
        LuiCollection colpar = new LuiCollection(pParams);
        LuiAppForm oscfrm;
        String parentFormID = getVal("FormID");
        String config = NVL(colpar.Get_Element("CONFIG"), "DEFAULT");
        int procRowNum = 0;
        int totRowNum = 0;
        // ищем такую команду в массиве форм
        if (forms.containsKey(commandCode + "." + config)) {
            oscfrm = forms.get(commandCode + "." + config);
            oscfrm.reloadForm();
        } else { // такой команды ещё нет
            oscfrm = this.addForm(commandCode + "." + config, commandCode, "HOST"/*,"N"*/);
            /*oscfrm=new LuiAppForm(this,commandCode,"HOST");
        for(String propcode:formTemplates.get("HOST").properties.keySet()){
     if (formTemplates.get("HOST").properties.get(propcode).is_meta)       
        oscfrm.addProperty(propcode,formTemplates.get("HOST").getNonCompValue(propcode));
      }*/

            oscfrm.setProperty("ConfigCode", config);
            oscfrm.setVal("ParentFormID", parentFormID);
            oscfrm.setProperty("FormID", commandCode + "." + config);
            oscfrm.load(oscfrm.getProperty("MetadataApp"),false);
            oscfrm.loadParams(pParams);
            oscfrm.setProperty("HEADER", oscfrm.getProperty("ConfigName"));
            oscfrm.elSort(0);
            //forms.put(commandCode+"."+config, oscfrm);

        }
        // готовим Property:Variables
        if (oscfrm.properties.containsKey("Variables")) {
            oscfrm.properties.remove("Variables");
        }
        String Variables = "";
        String V = "";
        for (String parCode : colpar.keySet()) { // для параметров, 
            if (oscfrm.elements.containsKey(parCode)) { // если такой параметр в HOST - форме есть 
                LuiAppElement vel = oscfrm.elements.get(parCode);
                if (vel.getParent().equals("Form") && vel.getType().equals("PARAM")) {
                    V = colpar.Get_Element(parCode);
                    vel.properties.get("Value").startedValue = V;
                    //if (like(V, "%{%}%")) {//19/05/2020
                    if (vel.properties.get("Value").isDynamic()){
                        vel.properties.get("Value").recalculate = true;
                    } else {
                        vel.properties.get("Value").currentValue = V;
                        vel.properties.get("Value").recalculate = null;
                    }

                    if (vel.getProperty("isVariable").equals("Y")) { // являющихся переменными (т.е. не описанными в команде OS
                        vel.addProperty("FormattedValue", oscfrm.getProperty("COMMAND_VARIABLE_FORMAT"));
                        Variables = Variables + " " + vel.getProperty("FormattedValue");
                    }
                }
            }
        }
        oscfrm.addProperty("Variables", Variables);

        String cmd = oscfrm.getProperty("COMMAND_WRKDIR"); // рабочий каталог
        File wrkdir = cmd==null || cmd.isEmpty() ? null : new File(cmd);
        
        oscfrm.properties.get("COMMAND_TEXT").clearCompVal();
        if (pParams.indexOf("LUI$LOG:N;") < 0 && "Y".equals(properties.get("DEBUG_MODE").currentValue)) { // если надо логировать
            this.decr = false;
            cmd = oscfrm.getProperty("COMMAND_TEXT"); // текст команды    
            this.decr = true;
            oscfrm.properties.get("COMMAND_TEXT").clearCompVal();
            LuiCollection allpar = new LuiCollection("");
            for (int i = 0; i < oscfrm.elements.size(); i++) {
                LuiAppElement el = oscfrm.elSort(i);
                allpar.Add_Element(el.getCode(), el.getValue());
            }
            if (is_not_null(oscfrm.getVal("ParentFormID"))) {
                if (forms.containsKey(oscfrm.getVal("ParentFormID"))) {
                    procRowNum = Integer.valueOf(NVL(forms.get(oscfrm.getVal("ParentFormID")).getProperty("ProcessingRowNumber"), "0"));
                    totRowNum = Integer.valueOf(NVL(forms.get(oscfrm.getVal("ParentFormID")).getProperty("TotalRowNumber"), "0"));
                }
            }
            if (procRowNum < 10 || procRowNum == totRowNum) {
                log.addRecord("ACTION",
                        NLS.NLS("RUS:Выполнение команды ОС;ENG:Execute OS command;") + " " + commandCode,
                        NLS.NLS("RUS:Входные параметры;ENG:Input paremeters;") + ":\n" + allpar.toString().replace(";", "\n")
                        + NLS.NLS("RUS:Полный текст команды;ENG:Input paremeters;") + ":\n" + cmd, oscfrm.getFormID());
            }
        }
        cmd = oscfrm.getProperty("COMMAND_TEXT"); // текст команды    
        
        try {
            String[] aCmd = parseCommandLine(cmd).toArray(new String[]{});
            
/* achibuk (21.04.2020) Закомментирована прежняя версия выполнения команды ОС при помощи Runtime.getRuntime().exec, как устаревшая.            
            Process p = Runtime.getRuntime().exec( 
                aCmd
              , null
              , wrkdir);
            int rc = p.waitFor();
*/          

/* achibuk (21.04.2020) Вместо прежней версии добавлена новая с выполнением команды ОС при помощи класса ProcessBuilder, что позволяет перенаправлять 
потоки вывода и ошибок в файлы (можно в один и тот же файл) */
   // if (getProperty("DEBUG_MODE").equals("Y")){
   //         System.out.println("HOST:");
   //         for (int wi=0;wi<aCmd.length;wi++){
   //             System.out.println(" "+wi);
   //         }
   // }        
            ProcessBuilder pb = new ProcessBuilder(aCmd);
            pb.redirectErrorStream(true);
            pb.directory(wrkdir);
            String outFileName = oscfrm.getProperty("COMMAND_STDOUT_2_FILE"); // имя файла, в который необходимо перенаправить поток вывода
            String errFileName = oscfrm.getProperty("COMMAND_STDERR_2_FILE"); // имя файла, в который необходимо перенаправить поток ошибок
            File outFile = null;
            if (outFileName.length() > 0) {
                outFile = new File(outFileName);
                pb.redirectOutput(outFile);
            }
            if (errFileName.length() > 0) {
                if (errFileName.equals(outFileName)) pb.redirectError(outFile); // если имена файлов для перенаправления в них потоков вывода и ошибок совпадают, то используем ту же ссылку на файл
                else pb.redirectError(new File(errFileName));
            }
            Process p = pb.start();

            StringWriter writer = new StringWriter();
            new Thread(new Runnable() {
                public void run() {
                    try {
                        IOUtils.copy(p.getInputStream(), writer);
                    } catch (IOException e) {}
                }
            }).start();

            int rc = p.waitFor();
            String pOutput = writer.toString();
           // if (getProperty("DEBUG_MODE").equals("Y")){
           // System.out.println("return code:"+String.valueOf(rc));
           // }
// завершение добавления от 21.04.2020            

            
            oscfrm.addProperty("RESULT_CODE", String.valueOf(rc));
            setProperty("LastHostResultCode", String.valueOf(rc));
            cmd = oscfrm.getProperty("COMMAND_RESULT");
            setProperty("LastHostResult", cmd);
            return cmd;
        } catch (IOException ex) {
            msgWriter.write("LUI-00058", commandCode, cmd, ex.getMessage());
        } catch (InterruptedException ex) {
            msgWriter.write("LUI-00059", commandCode, cmd, ex.getMessage());
        }
        return "";
    }

//---------------------------//

    public LuiSeance getSeance() {
        return luiSeance;
    }

    public LuiSession getSession() {
        return luiSession;
    }

    /**
     * Получение ссылки на приложение по известному id сессии пользователя
     * @param pSessionID - id сессии
     * @return Приложение LUI
    */
    public LuiApplication getAppBySessionID (String pSessionID) {
        // Получим ссылку на LuiServer для данного LuiApplication
        LuiSrv luiServer = getSeance().getServer();
        LuiApplication lapp;
        LuiSession lses;
        // Просмотрим все LuiApplication'ы этого сервера через список сеансов
        for (LuiSeance seance : luiServer.getSeances()) {
            lapp = seance.getApplication();
            if (lapp!=null) {
                lses = lapp.getSession();
                if (lses!=null) {
                    if (lses.id.equals(pSessionID)) return lapp;
                }
            }
        }
        return null;
    }
    
    /**
     * Отправка текстового сообщения пользователю, создавшему указанную сессию
     * @param pSessionID - id сессии
     * @param pMsgType - тип сообщения (информационное, предупреждающее)
     * @param pMsgText - текст сообщения
    */
    public void sendMsg2Session(String pSessionID, String pMsgType, String pMsgText) {
        if (!luiSeance.getMessaging()) return;
        LuiApplication app = getAppBySessionID(pSessionID);
        if (app != null) app.show_message(this, pMsgType, pMsgText);
    }


    /**
     * Отображение полученного сообщение в приложении
     * @param pAppFrom - из кого приложения получено сообщение
     * @param pMsgType - тип сообщения (информационное, предупреждающее)
     * @param pMsgText - текст сообщения
    */
    public void show_message(LuiApplication pAppFrom, String pMsgType, String pMsgText) {
        try {
            if (!luiSeance.getMessaging()) pAppFrom.msgWriter.write("LUI-00100", getSession().id);
            switch (pMsgType) {
                case "I": luiSeance.msgWriterSecondary.write("LUI-00101", pMsgText); break;
                case "W": luiSeance.msgWriterSecondary.write("LUI-00102", pMsgText); break;
                case "E": luiSeance.msgWriterSecondary.write("LUI-00103", pMsgText); break;
                default: luiSeance.msgWriterSecondary.write("LUI-00101", pMsgText);
            }
            luiSeance.msgWriterSecondary.msgs2commands(luiSeance.commandWriterSecondary); //Выгребание из очереди сообщений и отображение их клиенту
            luiSeance.commandWriterSecondary.flush();
        } catch (Exception e) {
        }
    }
/**
 * Поиск свойства по пути
 * Имя свойства задаётся в формате:   &lt;ID формы&gt;.&lt;Код элемента или "Form"&gt;.&lt;Код свойства&gt;
 * @param pPath путь к свойству (?)
 * @return Свойство
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiAppProp findProperty (String pPath) throws LuiException, SQLException{
    String frmID;
    String elcode;
    String pcode;
    
    if (pPath.indexOf(".")<0) { // если точек нет
        frmID=getProperty("FormID");
        elcode="Form";
        pcode=pPath;
    } else {
    frmID=LuiString.Extract_Word(pPath,1,".");
    elcode=LuiString.Extract_Word(pPath, 2,".");
    pcode=LuiString.Extract_Word(pPath, 3,".");
     if (is_null(pcode)) {
         pcode=elcode;
         elcode=frmID;
         frmID=getProperty("FormID");
     }
    }
    if (is_null(frmID) || is_null(elcode) || is_null(pcode)) return null;
    if (!forms.containsKey(frmID)) return null;
    LuiAppForm frm=forms.get(frmID);
    if ("Form".equals(elcode)){
        if (!frm.properties.containsKey(pcode)) return null;
        return frm.properties.get(pcode);
    } else {
        if (!frm.elements.containsKey(elcode)) return null;
    LuiAppElement el=frm.elements.get(elcode);
    if (!el.properties.containsKey(pcode))return null;
    return el.properties.get(pcode);
    }
}

    /**
     * Возвращает имя таблицы (возможно со схемой и псевдонимом) из строки в которой курсор
     * @param pQuery - запрос
     * @param pCursor - позиция курсора
     * @return имя таблицы
     */
    public String getTableAndAliasFromQueryEditor(String pQuery,String pCursor){
    int i;
    try{
       i=Integer.valueOf(pCursor); 
    } catch (Exception ex){i=0;}   
    return GetTableAndAliasFromCursor(pQuery, i);

    }

    public String getColumnsFromDTDBC2JSON4Field(String pFormID,String pElCode) throws LuiException, SQLException{
       if (forms.containsKey(pFormID)){
         LuiAppForm frm=forms.get(pFormID);
         if (frm.elements.containsKey(pElCode)){
             LuiAppElement el=frm.elements.get(pElCode);
             if (el.properties.containsKey("Value")&&el.properties.containsKey("StartCursor")){
                 String table=getTableAndAliasFromQueryEditor(el.getProperty("Value"),el.getProperty("StartCursor"));
                 if (is_not_null(table)){
                   return getColumnsFromDTDBC2JSON(table);
                 }
             }
         }
       } 
       String retval="[";
                retval=retval+"{";
                 retval=retval+"\"CONN\":\""+"\",";
                 retval=retval+"\"SCHEMA\":\""+"\",";
                 retval=retval+"\"NAME\":\""+"\",";
                 retval=retval+"\"DATA_TYPE\":\""+"\",";
                 retval=retval+"\"CMNT\":\""+"\",";
                 retval=retval+"\"PK\":\""+"\",";
                 retval=retval+"\"NULLABLE\":\""+"\"";
                retval=retval+"}";
                retval=retval+"]";    
        return retval;
    }
    /**
     * Возвращает JSON массив описаний таблиц Каждый элемент массчива состоит из полей
     *  CONN - код соединения
     *  SCHEMA - схема
     *  TABLE_NAME - имя таблицы
     *  TABLE_TYPE - тип TABLE или VIEW
     *  CMNT - коментарий
     */
public String  getTablesFromDTDBC2JSON() throws LuiException{
 String retval="";
 LuiCollection dtdbcCol=null;
    LuiAppConn ac=null;
    LuiDbConn mlc=null;
    LuiDbConn lc=null;
    String stmt="";
    LuiCollection tabDescr=null;
    ArrayList<LuiCollection> tabsDescr=null;
    try{
     mlc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
    dtdbcCol=mlc.getSelectAllRowsTwoColsValMap("select code,name from lui_t_dtdbc where app_code='"+getProperty("EditAppCode")+"'"); //получаем список соединений
    if (dtdbcCol==null) {
        dtdbcCol=new LuiCollection("LUI$:;");
    }
    for (String dtdbc:dtdbcCol.keySet()){ // крутим соединения
        ac=null;
        lc=null;
        tabsDescr=null;
        if (!userDbConnections.containsKey(dtdbc)) continue; // если его реально нет - уход на следующее
        ac=userDbConnections.get(dtdbc);
        if (ac==null) continue;
        try{
        if (ac.getSessionMode()=='C') { // если соединение коммунальное, то получим из пула свободное
            lc=ac.connPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
        }
        else {//в противном случае возьмём текущее 
            lc=ac.currentConnection;
        }  
        if (lc==null) continue;
        stmt=lc.getQuery4Table();
        if (is_null(stmt)) continue;
        tabsDescr=lc.getSelectAllRowsValMap(stmt);
        if (tabsDescr==null) continue;
        for (int i=0;i<tabsDescr.size();i++){
            tabDescr=tabsDescr.get(i);
            tabDescr.Add_Element("CONN",dtdbc);
            if (is_not_null(retval))retval=retval+",";
                retval=retval+tabDescr.ToJSON();
        }
        }catch (Exception ex){}
      finally {
        //  if (ac!=null) {
              if (ac.getSessionMode()=='C') { // если соединение коммунальное, то вернём соединение в пул свободных
                try{  
                ac.connPool.SetFree(lc);
                } catch (Exception ex){}
            }
          //}
      }
    }      
    }catch (Exception ex){}
      finally {if (mlc!=null) {
          try {metaConnPool.SetFree(mlc);
      }catch (Exception ex2){}
              
      }
    }
        if (is_null(retval)){
        retval="[";
                retval=retval+"{";
                 retval=retval+"\"CONN\":\""+"\",";
                 retval=retval+"\"SCHEMA\":\""+"\",";
                 retval=retval+"\"TABLE_NAME\":\""+"\",";
                 retval=retval+"\"TABLE_TYPE\":\""+"\",";
                 retval=retval+"\"COMMENTS\":\""+"\"";
                retval=retval+"}";
                retval=retval+"]";    
    } else retval="["+retval+"]";

 return retval;
}
    /**
     * Возвращает JSON массив описаний столбцов таблицы Каждый элемент массчива состоит из полей
     *  CONN - код соединения
     *  SCHEMA - схема
     *  NAME - имя столбца
     *  DATA_TYPE - тип данных СУБД
     *  LUI_DATA_TYPE - тип данных LUI
     *  CMNT - коментарий
     *  PK - входит в ПК? (Y/N)
     *  NULLABLE - null можно? (Y/N)
     * @param pTable -имя таблицы (возможно со схемой и псевдонимом)
     * @return - JSON массив описаний столбцов
     * @throws LuiException логическая ошибка
     */
    public String getColumnsFromDTDBC2JSON(String pTable) throws LuiException{
    String STMT4TYPE="select dc.lui_datatype from lui_t_data_convert dc where dc.dbms_type='<dbms_type>' and dc.app_code='<app_code>' and dc.dbms_datatype='<dbms_datatype>' order by dc.is_default4dbms2lui desc";
    String retval="";
    String dt="";
    String ldt="";
    LuiCollection dtdbcCol=null;
    LuiAppConn ac=null;
    LuiDbConn mlc=null;
    LuiDbConn lc=null;
    LuiCollection colDescr=null;
    ArrayList<LuiCollection> colsDescr=null;
    String stmt="";
    try{
    mlc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
    dtdbcCol=mlc.getSelectAllRowsTwoColsValMap("select code,name from lui_t_dtdbc where app_code='"+getProperty("EditAppCode")+"'"); //получаем список соединений
    
    
    if (dtdbcCol==null) {
        dtdbcCol=new LuiCollection("LUI$:;");
    }
    
    for (String dtdbc:dtdbcCol.keySet()){ // крутим соединения
        ac=null;
        lc=null;
        colsDescr=null;
        if (!userDbConnections.containsKey(dtdbc)) continue; // если его реально нет - уход на следующее
        ac=userDbConnections.get(dtdbc);
        if (ac==null) continue;
      try{
        if (ac.getSessionMode()=='C') { // если соединение коммунальное, то получим из пула свободное
            lc=ac.connPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,code,getLuiSessionID(),getAppUser());
        }
        else {//в противном случае возьмём текущее 
            lc=ac.currentConnection;
        }  
        if (lc==null) continue;
        stmt=lc.getQuery4Column(pTable);
        if (is_null(stmt)) continue;
        colsDescr=lc.getSelectAllRowsValMap(stmt);
        if (colsDescr!=null){
            for (int i=0;i<colsDescr.size();i++){
                colDescr=colsDescr.get(i);
             dt=colDescr.Get_Element("DATA_TYPE");
             ldt=null;
             ldt=mlc.getSelectResult(STMT4TYPE.replace("<app_code>",getProperty("EditAppCode")).replace("<dbms_type>",lc.getDBvendor()).replace("<dbms_datatype>",dt),1);
             if (is_null(ldt)){
            ldt=mlc.getSelectResult(STMT4TYPE.replace("<app_code>","LUI").replace("<dbms_type>",lc.getDBvendor()).replace("<dbms_datatype>",dt),1);    
            }
            if (is_null(ldt)){
                ldt="CHAR";
            }
                colDescr.Add_Element("CONN",dtdbc);
                colDescr.Add_Element("LUI_DATA_TYPE",ldt);
                if (is_not_null(retval))retval=retval+",";
                retval=retval+colDescr.ToJSON();
            }
        }
      } catch (Exception ex){}
      finally {
          //if (ac!=null) {
              if (ac.getSessionMode()=='C') { // если соединение коммунальное, то вернём соединение в пул свободных
                try{  
                ac.connPool.SetFree(lc);
                } catch (Exception ex){}
            }
        // }
      }
    }// конец кручения соединений
    } catch (Exception ex){}
      finally {if (mlc!=null) {
          try {metaConnPool.SetFree(mlc);
      }catch (Exception ex2){}
              
      }
    }
    if (is_null(retval)){
        retval="[";
                retval=retval+"{";
                 retval=retval+"\"CONN\":\""+"\",";
                 retval=retval+"\"SCHEMA\":\""+"\",";
                 retval=retval+"\"NAME\":\""+"\",";
                 retval=retval+"\"DATA_TYPE\":\""+"\",";
                 retval=retval+"\"CMNT\":\""+"\",";
                 retval=retval+"\"PK\":\""+"\",";
                 retval=retval+"\"NULLABLE\":\""+"\"";
                retval=retval+"}";
                retval=retval+"]";    
    } else retval="["+retval+"]";
    return retval;
    }
public String str2JSON(String s){
    return gson.toJson(s);
}
/*public int countCloseForm(){ //here is debug
 int rc=0;
 for(int g=0;g<commandWriter.commandGroupList.size();g++){
  for (int c=0;c<commandWriter.commandGroupList.get(g).commandList.size();c++){
    if (commandWriter.commandGroupList.get(g).commandList.get(c).name.equals("CloseForm"))rc++;  
  }  
 }
    return rc;        
}*/
public String getConnCollection() {
  LuiCollection con=new LuiCollection("");
  for (String conCode:this.userDbConnections.keySet()){
    con.Add_Element(conCode, conCode+"-"+userDbConnections.get(conCode).name);
  }
  
  return con.toString();
}
public String toJSON(Object o){
  if (luiSeance==null) return gson.toJson(o);
  return luiSeance.toJSON(o);
          
}
public String rtGetCommand4chartControl(){
  if (forms.containsKey(getVal("rtChart"))){
    if (forms.get(getVal("rtChart")).RTdashboard!=null){
       return  forms.get(getVal("rtChart")).RTdashboard.getCommand4ChartControl();
    }  
  }  
  return "setOption";
}
public String getSeanceIdAndSessionID(){
  String seanceId="";
  if (this.luiSeance!=null){
      seanceId=String.valueOf(luiSeance.mSeanceId);
  }
  String sesId="";
  if (this.luiSession!=null){
      sesId=String.valueOf(luiSession.id);
  }
  return seanceId+":"+sesId;
}
public String getUserName(){
    if (this.luiSession!=null){
        return luiSession.userName;
    }
    return "";
}

/** Повторное выполнение действия вызвавшего форму pFrmID
     * @param pFrmID - ID формы
**/
public void reRun(String pFrmID) throws LuiException, SQLException{
    String parentFormID=forms.get(pFrmID).getVal("ParentFormID");
    String parentAction=forms.get(pFrmID).getVal("CallAction");
    LuiAppElement ael=forms.get(parentFormID).elements.get(parentAction);
    for (String pk:ael.properties.keySet()){
        ael.properties.get(pk).clearCompVal(); // перевычисление свойств действия
    }
    for (String ec:forms.get(parentFormID).elements.keySet()){
        if (forms.get(parentFormID).elements.get(ec).getParent().equals(parentAction)){
            ael=forms.get(parentFormID).elements.get(ec);
            for (String pk:ael.properties.keySet()){
            ael.properties.get(pk).clearCompVal(); // перевычисление свойств параметра
    }
        }
    }
    forms.get(pFrmID).close(null,null);
    forms.get(parentFormID).runAction(parentAction);
}
public String clearWiki(String pstring){
    pstring=pstring.replace("-----","");
    pstring=pstring.replace("----","");
    pstring=pstring.replace("---","");
    pstring=pstring.replace("--","");
    pstring=pstring.replace("+++++","");
    pstring=pstring.replace("++++","");
    pstring=pstring.replace("+++","");
    pstring=pstring.replace("++","");
    pstring=pstring.replace("'''","");
    pstring=pstring.replace("''","");
    pstring=pstring.replace("==","");    
    return pstring;
}
  
 
}
