/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/**
 * Класс - писатель групп команд от сервера клиенту
 * его задача вести список групп команд
 * предоставляя методы для добавления и удаления групп
 * а также для добавления и удаления команд указанной группы
 * а также для удаления команд заданного имени из всех групп
 * Метод flush выполняет отправку на клиент всех подготовленных групп команд
 */
 
package luiserver;


import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.StringWriter;

import java.util.ArrayList;
import java.util.Set;
import javax.script.ScriptEngine;
import javax.script.ScriptException;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

/**
 *
 * @author aliubushkin
 */

public class LuiClCommandWriter {
private LuiMsgWriter messages;    
public ArrayList<LuiClCommandGroup> commandGroupList;    // список групп команд
public ArrayList<LuiClCommandGroup> AddcommandGroupList;    // список групп команд
private BufferedWriter writer; //писатель в tcp-клиент
private DocumentBuilderFactory DOC_FACTORY;
private TransformerFactory TRANS_FACTORY;
private Transformer aTR;
private DocumentBuilder aDB;
//private LuiClEventReader EventReader;
//private boolean really;
private LuiProcTimer timer;
public ScriptEngine JSengine;
public boolean really=true;


//public String realTime="";
/**
 * Конструктор писателя
 * @param pWriter - писатель в tcp-клиент
 * @param pTimer - таймер
  */
public LuiClCommandWriter(BufferedWriter pWriter,LuiProcTimer pTimer){

writer=pWriter;
timer=pTimer;
commandGroupList=new ArrayList<LuiClCommandGroup>();
DOC_FACTORY = DocumentBuilderFactory.newInstance();
TRANS_FACTORY = TransformerFactory.newInstance();
//boolean really=true;

    try { 
        aTR = TRANS_FACTORY.newTransformer();
        
        aDB = DOC_FACTORY.newDocumentBuilder();
    } catch (TransformerConfigurationException | ParserConfigurationException ex) {
        
    }
                
}
/**
 * Добавление группы команд
 * @param pName - имя группы
 * @return -- ccылка на созданную группу
 */
public LuiClCommandGroup addGroup(String pName){
 synchronized(commandGroupList){
 LuiClCommandGroup g=new LuiClCommandGroup(pName);
 if (!really)g.really=false;
 commandGroupList.add(g);
 return g;
 }
}
/**
 * Добавление доп группы команд
 * @param pName - имя группы
 * @return -- ccылка на созданную группу
 */
public LuiClCommandGroup addAddGroup(String pName){
 if (this.AddcommandGroupList==null)AddcommandGroupList=new ArrayList<LuiClCommandGroup>();
    synchronized(AddcommandGroupList){
 LuiClCommandGroup g=new LuiClCommandGroup(pName);
 if (!really)g.really=false;
 AddcommandGroupList.add(g);
 return g;
 }
}
/**
 * возвращает последнюю группу а если их нет - создаёт с именем=""
 * @return -- ссылка на последнюю созданную группу
 */
/*public LuiClCommandGroup getLastGroup(String pNewGrp){
    int s=commandGroupList.size();
    if (s>0){
        return commandGroupList.get(s-1);
    } 
    return addGroup(NVL(pNewGrp,""));
}*/
/**
 * Удаление группы команд
 * @param pName - имя группы
  */
public void removeGroup(String pName){
 synchronized(commandGroupList){
 boolean removed=true;
 while (removed){
     removed=false;
 for (int i=0;i<commandGroupList.size();i++){
     if (commandGroupList.get(i).name.equals(pName)){
         commandGroupList.remove(i);
         removed=true;
         break;
     }
 }
 }
}
}
/**
 * Удаление команд c указанным именем из всех групп
 * @param pName - имя команды
  */
public void removeCommadFromAllGroups(String pName){
synchronized(commandGroupList){
    for (int i=0;i<commandGroupList.size();i++){
        commandGroupList.get(i).removeCommands(pName);
    }
}
}
/**
 * Получение группы по ёё имени
 * извлекается первая группа с указанным именем
 * @param pName - имя группв
 * @return -- ссылка на группу
 */
public LuiClCommandGroup getGroup(String pName){
synchronized(commandGroupList){    
    for (int i=0;i<commandGroupList.size();i++){
     if (commandGroupList.get(i).name.equals(pName)){
     return commandGroupList.get(i);
     }
   }
     return null; 
}
}
/**
 * Отправка всех групп команд клиенту
 * @throws LuiException логическая ошибка
 */
public void flush() throws LuiException {
    /*if (EventReader!=null){
    if (EventReader.getEventListSize()>0) return; // если есть считанные, но не обработанные события
    }*/
    
   synchronized(timer){
   synchronized(aTR){
    synchronized(aDB){
    synchronized(writer){
    synchronized(commandGroupList){
        if (AddcommandGroupList!=null ){
       //  try{
       //      Thread.sleep(5000);
       //  } catch(Exception ex){}
       for(int i=0;i<AddcommandGroupList.size();i++){
         commandGroupList.add(AddcommandGroupList.get(i));
       }
       AddcommandGroupList.clear();
       AddcommandGroupList=null;
   } 
        Document aDoc = aDB.newDocument();
        Element aRoot = aDoc.createElement("Command");
        
        
        aRoot.setAttribute("Time",String.valueOf(timer.stopAndGetValue()));
        
        
        aDoc.appendChild(aRoot);
        for (int cg=0;cg<commandGroupList.size();cg++){ // цикл по группам команд
            
           Element aGroup=null;
      //    if (commandGroupList.get(cg).params.size()>0){ 
            aGroup=aDoc.createElement("W");
            aGroup.setAttribute("N", commandGroupList.get(cg).name);
            Set<String> parKeySet=commandGroupList.get(cg).params.keySet(); //набор ключей параметров круппы комманд
            for (String key:parKeySet){ //цикл по ключам параметров группы команд
                Element aGroupPar=aDoc.createElement("P"); // создаём элемент документа "параметр группы"
                aGroupPar.setAttribute("N",key); // добавляем атрибут "N" - имя параметра
                if ((commandGroupList.get(cg).params.get(key).type)=='S') {// если параметр просто сторока
                aGroupPar.appendChild(aDoc.createTextNode(commandGroupList.get(cg).params.get(key).get_value()));
                } 
                else {
                 if ((commandGroupList.get(cg).params.get(key).type)=='J') {// если параметр JSON
                     String v=commandGroupList.get(cg).params.get(key).get_value();
                     try {
                    Object bl = JSengine.eval("JSON.stringify({"+v+"})");//запускаем ввычислятор
                    v=String.valueOf(bl); //  и берём то что он вернул
                     aGroupPar.appendChild(aDoc.createTextNode(v));
                     }
                     catch (ScriptException jsex){
                      String here="here";   
                     }
                     //stringify???
                 } else {
                 // коллекция 
                 aGroupPar.setAttribute("T","Collection"); // добавляем атрибут "T" - Тип параметра
                 Set<String> parValKeySet=commandGroupList.get(cg).params.get(key).value.keySet();
                 for (String valKey:parValKeySet){ //цикл по ключам значений параметра группы команд
                 Element aGroupParKey=aDoc.createElement("X");
                 aGroupParKey.appendChild(aDoc.createTextNode(valKey));
                 aGroupPar.appendChild(aGroupParKey);
                 Element aGroupParKeyVal=aDoc.createElement("V");
                 aGroupParKeyVal.appendChild(aDoc.createTextNode(commandGroupList.get(cg).params.get(key).get_value(valKey)));
                 aGroupPar.appendChild(aGroupParKeyVal);
                }    
                 }    
            }
                
            aGroup.appendChild(aGroupPar); // добавляем в группу команд сформированный параметр группы
            }
            aRoot.appendChild(aGroup);
       //   }

            for (int c=0;c<commandGroupList.get(cg).commandList.size();c++){ //цикл по списку команд группы
                if (!commandGroupList.get(cg).commandList.get(c).really) continue;
                Element aCommand=aDoc.createElement("C");
                aCommand.setAttribute("N",commandGroupList.get(cg).commandList.get(c).name); // устанавливаем имя команды
                Set<String> parComKeySet=commandGroupList.get(cg).commandList.get(c).params.keySet(); //набор ключей параметров круппы комманд
                for (String key:parComKeySet){ //цикл по ключам параметров команды
                    Element aCommandPar=aDoc.createElement("P"); // создаём элемент документа "параметр команды"
                    aCommandPar.setAttribute("N",key); // добавляем атрибут "N" - имя параметра
                    if (commandGroupList.get(cg).commandList.get(c).params.get(key).type=='S') {// если параметр просто строка
                       aCommandPar.appendChild(aDoc.createTextNode(commandGroupList.get(cg).commandList.get(c).params.get(key).get_value()));
                       //aCommandPar.setNodeValue();
                    } 
                    else {
                     if (commandGroupList.get(cg).commandList.get(c).params.get(key).type=='J') {// если параметр JSON
                     String v=commandGroupList.get(cg).commandList.get(c).params.get(key).get_value();
                     if (v.length()>7){
                     if (v.charAt(0)!='{' || v.charAt(1)!='\"'){
                     if (v.charAt(0)=='r' && v.charAt(1)=='e' && v.charAt(2)=='a' && v.charAt(3)=='l'
                      && v.charAt(4)=='T' && v.charAt(5)=='i' && v.charAt(6)=='m' && v.charAt(7)=='e'     
                             ){
                       //synchronized(realTime){  
                       //v=realTime;  
                       //}
                     }
                     else    
                     {
                     try{
                     Object bl;    
                     if (v.charAt(0)=='A' && v.charAt(1)=='p' && v.charAt(2)=='p' && v.charAt(3)=='.') {
                         synchronized(JSengine){
                         bl = JSengine.eval(v);//запускаем ввычислятор    
                      }
                     } else {   
                      bl = JSengine.eval("JSON.stringify({"+v+"})");//запускаем ввычислятор
                     }
                     v=String.valueOf(bl); //  и берём то что он вернул
                     } catch (ScriptException jsex){
                         String here="here";
                     }
                     }
                     }
                     }
                     aCommandPar.appendChild(aDoc.createTextNode(v));
                     //stringify???
                 } else {   
                        // коллекция
                      aCommandPar.setAttribute("T","Collection"); // добавляем атрибут "T" - Тип параметра  
                      Set<String> parComValKeySet=commandGroupList.get(cg).commandList.get(c).params.get(key).value.keySet(); // набор ключей значений параметра
                      for (String valKey:parComValKeySet){ //цикл по ключам значений параметра команды
                      Element aCommandParKey=aDoc.createElement("X");   
                      aCommandParKey.appendChild(aDoc.createTextNode(valKey));
                      aCommandPar.appendChild(aCommandParKey);
                      Element aCommandParKeyVal=aDoc.createElement("V");
                      aCommandParKeyVal.appendChild(aDoc.createTextNode(commandGroupList.get(cg).commandList.get(c).params.get(key).get_value(valKey)));
                      aCommandPar.appendChild(aCommandParKeyVal);
                      }
                    }
                }
                aCommand.appendChild(aCommandPar); // добавляем в команду сформированный параметр
                }
               //if (aGroup!=null) {
                aGroup.appendChild(aCommand); // добавляем в группу команд сформированную команду
               //} else {
               // aRoot.appendChild(aCommand); // если нет группы добавляем в корень
               //}
            }
        }
        
        StringWriter strWriter = new StringWriter();
        StreamResult aSR = new StreamResult(strWriter);
        aTR.setOutputProperty(OutputKeys.INDENT, "yes");
        
        //XMLOutputter rawOutputter;
        //rawOutputter = new XMLOutputter();
        //rawOutputter.setFormat(Format.getRawFormat().setLineSeparator("").setTextMode(TextMode.PRESERVE));

        //aTR.setOutputProperty(OutputKeys.ENCODING, "utf8");
        //aTR.setOutputProperty(OutputKeys.VERSION, "1.0");
        //aTR.setOutputProperty(OutputKeys.METHOD, "xml");
    try {
    
        aTR.transform(new DOMSource(aDoc), aSR);
     
        String s=strWriter.getBuffer().toString()+"\n";        
        writer.write(s);
        writer.flush();}
     
     catch (TransformerException ex) {
        messages.write("LUI-00003", ex.getLocalizedMessage());
    } catch (IOException ex) {
       try { 
        messages.write("LUI-00004", ex.getMessage());
       } catch (Exception ex2){
           String aCalling="thread:\""+Thread.currentThread().getName()+"\" method:"+Thread.currentThread().getStackTrace()[2].getClassName()+"."
        + Thread.currentThread().getStackTrace()[2].getMethodName()+" line:"
        +Thread.currentThread().getStackTrace()[2].getLineNumber();

           throw new LuiException("IOEcception",ex.getMessage(),"","E",null,aCalling,null);
       }
    } 
commandGroupList.clear();
    }
  
        }    
   }
}
}
   
}   
public BufferedWriter getWriter(){
    return writer;
}
public void setJSengine(ScriptEngine pJSengine){
    JSengine=pJSengine;
}
public boolean isCommandToEnableWindow(String pWID){
   for (int i=0;i<commandGroupList.size();i++){
     for (int j=0;j<commandGroupList.get(i).commandList.size();j++){
         if (commandGroupList.get(i).commandList.get(j).name.equals("EnableWindow")){
             if (commandGroupList.get(i).commandList.get(j).getParValue("WindowID").equals(pWID)){
                 return true;
             }
         }
     }  
   } 
   return false; 
}
}

