/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс для поддержки классов команд клиенту и событий от клиента
 */
package luiserver;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.MaskAncillaryChars;

import java.util.HashMap;
import java.util.LinkedHashMap;

/**
 *
 * @author aliubushkin
 */
public class LuiClEvCo {
public String name;
public LinkedHashMap<String,LuiClParameter> params;
public static final String RC="‡"; /*ascii 135 заменитель перевода строк в значениях параметров*/
    /**
     * Конструктор 
     * @param pName - имя события
     */
    public LuiClEvCo(String pName){
      name=pName;  
      params=new LinkedHashMap<String,LuiClParameter>();
    }
    /**
     * Добавление параметра 
     * @param pName - имя параметра
     * @param pType - тип параметра 'C' - коллекция, 'S' - просто строка
     */
    private LuiClParameter addPar(String pName,char pType){
     LuiClParameter p=new LuiClParameter(pType);
     synchronized (params){
     params.put(pName, p);
     }
     return p;
    }
    
    /**
     * Добавление ключ-значения параметра 
     * @param pName - имя параметра
     * @param pKey - ключ значения
     * @param pValue - значение ключа
     */
    public void addParameter(String pName,String pKey,String pValue){
        synchronized (params){
        LuiClParameter p=params.get(pName);
        if (p==null){
         p=addPar(pName,'C');
        }
        p.add_value(pKey,MaskAncillaryChars(pValue.replaceAll("\r\n",RC).replaceAll("\n", RC)));
        //p.add_value("V", MaskAncillaryChars(pValue).replaceAll("\n", "&#10;").replaceAll("\r", "&#13;"));  
        //p.add_value("V", MaskAncillaryChars(pValue.replaceAll("\r\n","\n").replaceAll("\n", "&#10;")));
       // p.add_value(pKey, pValue.replace("\n", RC));
        }
    }
    /**
     * Добавление ключ-значения параметра без крестов
     * @param pName - имя параметра
     * @param pKey - ключ значения
     * @param pValue - значение ключа
     */
    public void addParameterRaw(String pName,String pKey,String pValue){
        synchronized (params){
        LuiClParameter p=params.get(pName);
        if (p==null){
         p=addPar(pName,'C');
        }
        p.add_value(pKey,MaskAncillaryChars(pValue));
        //p.add_value("V", MaskAncillaryChars(pValue).replaceAll("\n", "&#10;").replaceAll("\r", "&#13;"));  
        //p.add_value("V", MaskAncillaryChars(pValue.replaceAll("\r\n","\n").replaceAll("\n", "&#10;")));
       // p.add_value(pKey, pValue.replace("\n", RC));
        
    }}
    /**
     * Добавление скалярного параметра без преобразований
     * @param pName - имя параметра
     * @param pValue - значение ключа
     */
    public void addRawPar(String pName,String pValue){
      synchronized (params){
        LuiClParameter p=params.get(pName);
        if (p==null){
         p=addPar(pName,'S');
        }
        p.add_value("V", pValue);
    }}
    /**
     * Добавление скалярного параметра 
     * @param pName - имя параметра
     * @param pValue - значение ключа
     */
    public void addParameter(String pName,String pValue){
     synchronized (params){
     LuiClParameter p=params.get(pName);
        if (p==null){
         p=addPar(pName,'S');
        }
        if (pValue==null){
        p.add_value("V","");    
        } else
        p.add_value("V", MaskAncillaryChars(pValue.replaceAll("\r\n",RC).replaceAll("\n", RC)));
       //  p.add_value("V", MaskAncillaryChars(pValue).replaceAll("\n", "&#10;").replaceAll("\r", "&#13;"));  
       //     p.add_value("V", MaskAncillaryChars(pValue.replaceAll("\r\n","\n").replaceAll("\n", "&#10;")));
        //p.add_value("V", pValue.replace("\n", RC));*/
    }
    }
    /**
     * Добавление скалярного параметра 
     * @param pName - имя параметра
     * @param pValue - значение ключа
     */
    public void addParameterRaw(String pName,String pValue){
     synchronized (params){
        LuiClParameter p=params.get(pName);
        if (p==null){
         p=addPar(pName,'S');
        }
        if (pValue==null){
        p.add_value("V","");    
        } else
        p.add_value("V", MaskAncillaryChars(pValue));
       //  p.add_value("V", MaskAncillaryChars(pValue).replaceAll("\n", "&#10;").replaceAll("\r", "&#13;"));  
       //     p.add_value("V", MaskAncillaryChars(pValue.replaceAll("\r\n","\n").replaceAll("\n", "&#10;")));
        //p.add_value("V", pValue.replace("\n", RC));*/
    }}
    /**
     * Извлечение значения параметра 
     * @param pName - имя параметра
     * @return значение параметра
     */
    public String getParValue(String pName){
      synchronized (params){  
        if (params.containsKey(pName)){
        switch (params.get(pName).type){
            case 'S':
            return params.get(pName).get_value().replace(RC,"\n");
            case 'C':
                   LuiClParameter p=params.get(pName);
                   LuiCollection c=new LuiCollection("");
                   HashMap<String,String> hm=p.value;
                   for (String key:hm.keySet()){
                       c.Add_Element(key, hm.get(key).toString());
                   }
                   return c.toString();
        }
        
        return params.get(pName).toString().replace(RC,"\n");
        }
        return "";
    }}
    /**
     * Извлечение одного из значений параметра по ключу значения
     * @param pName - имя параметра
     * @param pKey - ключ
     * @return значение параметра
     */
    public String getParKeyValue(String pName,String pKey){
        synchronized (params){
        return params.get(pName).get_value(pKey);
    }}
    public void addJparameter(String pName,String pValue){
      synchronized (params){  
      LuiClParameter p=params.get(pName);
        if (p==null){
         p=addPar(pName,'J');
        }
        p.add_value("V", pValue);
//p.add_value("V", MaskAncillaryChars(pValue.replaceAll("\r\n",RC).replaceAll("\n", RC)));
    }}
}
