/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс - читатель событий от клиента
 * его задача высасывать из клиента всё, пока не придёт пустая строка
 * Полученные данные из XML-сруктруры заносятся в список событий
 * Выдача событий из этого списка производится по одному
 * Если список событий изсяк, то производится новое чтение данных от клиента.
 * Это чтение может зависнуть на долгом ожидании, если на клиенте ничего не происходит.
 */
package luiserver;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import javax.script.ScriptException;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;

/**
 *
 * @author aliubushkin
 */
public class LuiClEventReader {
  private LuiMsgWriter msgWriter;
  private Document XMLEvents;
  private SAXBuilder xmlBuilder;    
  private String eventsString;  
  public  ArrayList<LuiClEvent> eventList;
  private LuiClEvent currentEvent;
  private BufferedReader reader; //читатель из tcp-клиента
  private LuiProcTimer timer;
  public  LuiApplication app;
  public  boolean is_interval=false;
  public  boolean is_non_interval=false;
  Element e;
  /**
   * Конструктор читателя
   * он создаёт объект под массив событий
   * @param pReader -- буфферизированный читатель из сокета
   * @param pMsg -- писатель сообщений об ошибках и не только об ошибках
   * @param pTimer таймер
   */
  public LuiClEventReader(BufferedReader pReader,LuiMsgWriter pMsg,LuiProcTimer pTimer){
    msgWriter=pMsg;
    timer=pTimer;
    eventList=new ArrayList<LuiClEvent>();
    reader=pReader;
    xmlBuilder=new SAXBuilder();
  
  }
  
 /* Зта программа ранее использовалась в LOVQ,LOVC,LOVD,Question,Calendar,UploadFile,
  для ограничения реакции на тики
  public void readEventsII() throws LuiException{
  while(true){    
    readEvents();
    // вычищаем Interval
    boolean b=true;
    while (b){
    b=false;
    for (int i=0;i<this.eventList.size();i++){
        if (this.eventList.get(i).name.equals("Interval")){
            this.eventList.remove(i);
            b=true;
            if (app!=null)
            app.setProperty("LUI Timer","");
            break;
        }
    }
    }
    // осталось ли что-нибудь
    if (eventList.size()>0) break;
    if (app!=null) {
        app.commandWriter.flush();
    }
  }
  }*/
  public void readEvents() throws LuiException{
    is_interval=false;
    is_non_interval=false;
    if (app!=null){ // ловля зависших в сервере
        if (app.luiSession!=null){
            LuiDbConn lc=null;
            try{
                lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                lc.executeDML("update lui_t_session set event_processing_stop=clock_timestamp() where id="+app.luiSession.id);
            } catch(SQLException ex){
            String here="here"; 
            }
                finally {
                try{
                if (lc!=null)
                app.metaConnPool.SetFree(lc);
                } catch(SQLException ex){}
            }
        }
    }
    try {
        eventsString=reader.readLine();

    } catch ( IOException ex) {
            msgWriter.write("LUI-00001",ex.getLocalizedMessage());
        }
      
    if (app!=null){ // ловля зависших в сервере
        if (app.luiSession!=null){
            LuiDbConn lc=null;
            try{
                lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
                lc.executeDML("update lui_t_session set event_processing_stop=null,event_processing_start=clock_timestamp() where id="+app.luiSession.id);
            } catch(SQLException ex){
            String here="here"; 
            }
                finally {
                try{
                if (lc!=null)
                app.metaConnPool.SetFree(lc);
                } catch(SQLException ex){}
            }
        }
    }
    timer.reStart();
    if (eventsString!=null) {
    string2eventArray();
    } else {
       // msgWriter.write("LUI-00001","Error reading data from client. Null value detected.");
    }
  }
          
  /** Получение очердного события
   * 
   * @return - событие
   */
  public LuiClEvent getEvent() {
    if (eventList.size()==0) return null;
    currentEvent=eventList.get(0);
    eventList.remove(0);
    return currentEvent;
  }
  public int getEventListSize(){
      return eventList.size();
  }
  public BufferedReader getReader(){
      return reader;
  }
  public void string2eventArray() throws LuiException{
  List<Element> evList = null;
  List<Element> evParList;
  List<Element> evParValList;
  String evName;
  String parName;
  String parKey;
  String parKeyVal;
  //String evString;
  String parType;
  LuiClEvent LuiClEv = null;

      try {
          XMLEvents=(Document) xmlBuilder.build(new StringReader(eventsString));
      } catch (JDOMException ex) {
          msgWriter.write("LUI-00002",ex.getLocalizedMessage());
      } catch (IOException ex) {
          msgWriter.write("LUI-00001",ex.getLocalizedMessage());
      }
      if (LuiClEv!=null){
      eventList.add(LuiClEv);
      }
      else {
      e=XMLEvents.getRootElement();
      if (e.getName().equalsIgnoreCase("Events")){
          evList=e.getChildren();
      }
      else if (e.getName().equalsIgnoreCase("E")){
          evList=new ArrayList<Element>();
          evList.add(e);
      }
      else {// В этой точке обнаружено что корневой элемент не Events и не E
          
      }
      if (evList==null)return;
      // Цикл по массиву событий
      for (int eCounter=0;eCounter<evList.size();eCounter++){
        if (evList.get(eCounter).getName().equalsIgnoreCase("E")){ // Если элемент массива носит имя E
          // выясняем имя события
          evName=null;
          evName=evList.get(eCounter).getAttributeValue("N");
          if (evName!=null){// если имя события не пустое
             LuiClEv=new LuiClEvent(evName);
             evParList=evList.get(eCounter).getChildren(); // список подчинённых элементов события
             // открываем цикл по подчинённым элементам события
              for (int parCounter=0;parCounter<evParList.size();parCounter++){
                  if (evParList.get(parCounter).getName().equalsIgnoreCase("P")){//Если подчинённый элемент носит имя P
                      // выясняем имя параметра
                      parName=null;
                      parName=evParList.get(parCounter).getAttributeValue("N");
                      if (parName!=null){ // если имя параметра не пустое
                          //выясняем тип параметра
                        parType=null;
                        parType=evParList.get(parCounter).getAttributeValue("T");
                        if (parType==null){parType="S";}
                        if (parType.equals("Collection")){ //если коллекция
                            
                             evParValList=evParList.get(parCounter).getChildren(); // получаем список элеменов подчинённых параметру
                             //цикл по значениям параметра
                             parKey="V";
                             for (int evParValCounter=0;evParValCounter<evParValList.size();evParValCounter++){
                                 if (evParValList.get(evParValCounter).getName().equals("X")) { //если имя элемента X, то его значение это код элемента коллекции
                                    parKey=evParValList.get(evParValCounter).getValue();
                                 }
                                 if (evParValList.get(evParValCounter).getName().equals("V")) { //если имя элемента V, то его значение это значение элемента коллекции
                                    parKeyVal=evParValList.get(evParValCounter).getValue(); 
                                    LuiClEv.addParameterRaw(parName,parKey,parKeyVal);
                                    
                                 }
                             }
                        }
                        else{//если строка 
                            parKeyVal=evParList.get(parCounter).getValue();
                            LuiClEv.addParameterRaw(parName,parKeyVal);
                        }
                      }
                  }
              }
           
           eventList.add(LuiClEv); //добавляем событие в список событий со всеми его параметрами
           if (LuiClEv.name.equals("Interval")){
             is_interval=true;
           } else {
             is_non_interval=true;  
           }
          
                  }          
          
      }
  }
      
  }
   
  }
}
