/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;

import java.io.IOException;
import org.apache.commons.codec.binary.Base64;
import java.io.UnsupportedEncodingException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
 
/**
 * Класс для шифрования и дешифрования строк
 * Использует библиотеку Apache Codec http://commons.apache.org/codec/

 */
public class LuiCrip {
 
    /**
     * Упрощенный конструктор. Создает  с ключом DESSecretKey со значением по умолчанию (не рекомендуется)
     */
    public LuiCrip() {
        this(new byte[]{1, 2, 3, 4, 5, 6, 7, 8});
    }
     
    /**
     * Упрощенный конструктор. Создает  с ключом 
     * DESSecretKey (алгоритм шифрования DES) со значением key. 
     * Ключ key должен иметь длину 8 байт
     * @param key ключ
     */
    public LuiCrip(byte[] key) {
        try {
            updateSecretKey(new DESSecretKey(key));
        } catch (InvalidKeyException | NoSuchAlgorithmException | NoSuchPaddingException ex) {
            
        }
    }
 
    public LuiCrip(SecretKey key) throws NoSuchPaddingException,
            NoSuchAlgorithmException,
            InvalidKeyException {
        updateSecretKey(key);
    }
 
    private void updateSecretKey(SecretKey key) throws NoSuchPaddingException,
            NoSuchAlgorithmException,
            InvalidKeyException {
        ecipher = Cipher.getInstance(key.getAlgorithm());
        dcipher = Cipher.getInstance(key.getAlgorithm());
        ecipher.init(Cipher.ENCRYPT_MODE, key);
        dcipher.init(Cipher.DECRYPT_MODE, key);
    }
 
    public static class DESSecretKey implements SecretKey {
 
        private final byte[] key;
 
        /**
         * ключ должен иметь длину 8 байт
         * @param key ключ
         */
        public DESSecretKey(byte[] key) {
            this.key = key;
        }
 
        @Override
        public String getAlgorithm() {
            return "DES";
        }
 
        @Override
        public String getFormat() {
            return "RAW";
        }
 
        @Override
        public byte[] getEncoded() {
            return key;
        }
    }
 
    private Cipher ecipher;
    private Cipher dcipher;
 
    /**
     * Функция шифрования
     *
     * @param str строка открытого текста
     * @return зашифрованная строка в формате Base64
     */
    public String encrypt(String str) {
        try {
            byte[] utf8 = str.getBytes("UTF8");
            byte[] enc = ecipher.doFinal(utf8);
            return Base64.encodeBase64String(enc);
        } catch (IllegalBlockSizeException | BadPaddingException | UnsupportedEncodingException ex) {
            
        }
        return null;
    }
 
    /**
     * Функция дешифрования
     *
     * @param str зашифрованная строка в формате Base64
     * @return расшифрованная строка
     */
    public String decrypt(String str) {
        try {
            byte[] dec = Base64.decodeBase64(str);
            byte[] utf8 = dcipher.doFinal(dec);
            return new String(utf8, "UTF8");
        } catch (IllegalBlockSizeException | BadPaddingException | IOException ex) {
         
        }
        return null;
    }
}