/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Конвертер данных
 */
package luiserver;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;

import static LuiCollection.LuiString.is_null;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 *
 * @author aliubushkin
 */
public class LuiDataConverter {
    public HashMap<String,LuiDataConvElement> convElems; 
    public HashMap<String,LuiDBMStype> dbmsTypes;
    public HashMap<String,LuiDataType> LuiDataTypes;
//    public LuiApplication app;
    public String appCode;
public LuiDataConverter(String pAppCode,LuiDbConn plc) throws LuiException, SQLException{
        convElems = new HashMap<>();
        dbmsTypes= new HashMap<>();
        LuiDataTypes= new HashMap<>();
    appCode=pAppCode;
    // Типы данных LUI
    ArrayList<LuiCollection> DT=plc.getSelectAllRowsValMap("select app_code,lui_datatype,name,alphabet_or_format,tl_select_clause,"
            + "tl_where_clause,tl_order_by_clause,"
            + "empty_string_from_lui_mode,empty_string_special_symbol,null_special_symbol"
            + " from lui_f_get_data_types4app('"+appCode+"')");
    for(int i=0;i<DT.size();i++){
        LuiCollection DTcol=DT.get(i);
        LuiDataType ldt=new LuiDataType(DTcol.Get_Element("app_code"),
                                        DTcol.Get_Element("name"),
                                        DTcol.Get_Element("alphabet_or_format"),
                                        DTcol.Get_Element("tl_select_clause"),
                                        DTcol.Get_Element("tl_where_clause"),
                                        DTcol.Get_Element("tl_order_by_clause"),
                DTcol.Get_Element("empty_string_from_lui_mode"),
                DTcol.Get_Element("empty_string_special_symbol"),
                DTcol.Get_Element("null_string_special_symbol")
        );
        LuiDataTypes.put(DTcol.Get_Element("lui_datatype"), ldt);
    }
    // Типы СУБД
    ArrayList<LuiCollection> DBT=plc.getSelectAllRowsValMap("select code,name,dummy_query,empty_string_is_null,datatype_ident,for_update_of from lui_t_dbms_type");
     for(int i=0;i<DBT.size();i++){
         LuiCollection DBTcol=DBT.get(i);
         LuiDBMStype ldbt=new LuiDBMStype(DBTcol.Get_Element("name"),DBTcol.Get_Element("dummy_query"),DBTcol.Get_Element("empty_string_is_null"),DBTcol.Get_Element("datatype_ident"),DBTcol.Get_Element("for_update_of"));
         dbmsTypes.put(DBTcol.Get_Element("code"), ldbt);
     }
    // Типы данных в СУБД 
    ArrayList<LuiCollection> DBDT=plc.getSelectAllRowsValMap("select dbms_type,dbms_datatype,name,numeric_code from lui_t_dbms_datatype");
    for (int i=0;i<DBDT.size();i++){
      LuiCollection DBDTcol=DBDT.get(i);
      LuiDBMSdatatype dbdt=new LuiDBMSdatatype();
      dbdt.name=DBDTcol.Get_Element("name");
      dbdt.numeric_code=Integer.valueOf(DBDTcol.Get_Element("numeric_code"));
      dbmsTypes.get(DBDTcol.Get_Element("dbms_type")).dataTypes.put(DBDTcol.Get_Element("dbms_datatype"), dbdt);
    }
    
    // Элементы конвертера
    // Крутим типы данных LUI
    for (String ldtkey:LuiDataTypes.keySet()){
       // Извлекаем операторы сравнения для данного типа данных
       LuiDataTypes.get(ldtkey).compOp=
               plc.getSelectAllRowsValMap("select operator_code,sign,name,icon,dbms_type,seq,markov_exp,predicate_constructor,qbe_constructor from lui_t_datatype_comp_op where app_code='"+
               LuiDataTypes.get(ldtkey).appCode+
               "' and lui_datatype='"+ldtkey+"' order by markov_seq");
       if (LuiDataTypes.get(ldtkey).compOp==null){
         LuiDataTypes.get(ldtkey).compOp=new ArrayList<>();
         LuiDataTypes.get(ldtkey).compOp.add(new LuiCollection("operator_code:=;sign:=;icon:co_equal;name:=;dbms_type:any;"));
         LuiDataTypes.get(ldtkey).compOp.add(new LuiCollection("operator_code:%;sign:%;icon:co_percent;name:like;dbms_type:any;"));
       }
       // Крутим СУБД
       for (String dbmsKey:dbmsTypes.keySet()){
        //Крутим типы данных Субд
         for (String dbdtKey:dbmsTypes.get(dbmsKey).dataTypes.keySet()) {
             ArrayList<LuiCollection> CE=plc.getSelectAllRowsValMap("select is_default4dbms2lui,dbms_type,dbms_datatype,to_lui,from_lui,to_lui_by_format,from_lui_by_format,is_default from lui_f_get_data_convert4app('"+
                   appCode+"','"+dbmsKey+"','"+dbdtKey+"','"+ldtkey+"')");
             if (CE!=null) {
           for(int i=0;i<CE.size();i++){
           LuiCollection CEcol=CE.get(i);
           if (CEcol.Get_Element("is_default4dbms2lui").equals("Y")){
               dbmsTypes.get(dbmsKey).dataTypes.get(dbdtKey).LuiDataType=ldtkey;
           }
           LuiDataConvElement ldce=new LuiDataConvElement(CEcol.Get_Element("to_lui"),CEcol.Get_Element("from_lui"),CEcol.Get_Element("to_lui_by_format"),CEcol.Get_Element("from_lui_by_format"));
           if (CEcol.Get_Element("is_default").equals("Y")){
               LuiDataTypes.get(ldtkey).defConvEls.put(CEcol.Get_Element("dbms_type"),ldce);
           }
           String mapKey=ldtkey+"."+CEcol.Get_Element("dbms_type")+".";
           if (dbmsTypes.get(CEcol.Get_Element("dbms_type")).identDataTypeByName) {
               mapKey=mapKey+CEcol.Get_Element("dbms_datatype");
           } else {
               mapKey=mapKey+String.valueOf(dbmsTypes.get(CEcol.Get_Element("dbms_type")).dataTypes.get(CEcol.Get_Element("dbms_datatype")).numeric_code);
           }
           convElems.put(mapKey, ldce);
                 } 
             }
             
         }
            }
    }
     
     
    
}
/**
 * Возвращает строку для конвертирования данных из Lui и в Lui
 * @param pColInfo информация о столбце
 * @param lc соединение с базой данныз
 * @param pIsToLui направление конвертации
 * @return строка для конвертирования
 */
public String getConvString(ColInfo pColInfo,LuiDbConn lc,boolean pIsToLui,boolean format){
    String dbType=lc.getDBvendor();
    String mapKey=pColInfo.LuiDATATYPE+"."+dbType;
    if (dbmsTypes.get(dbType).identDataTypeByName)mapKey=mapKey+"."+pColInfo.columnTypeName;
    else mapKey=mapKey+"."+String.valueOf(pColInfo.columnType);
    LuiDataConvElement el=null;
    if (convElems.containsKey(mapKey)){
       el=convElems.get(mapKey);
    } else {
        if (LuiDataTypes.containsKey(pColInfo.LuiDATATYPE)){
        if (LuiDataTypes.get(pColInfo.LuiDATATYPE).defConvEls.containsKey(dbType)){
        el=LuiDataTypes.get(pColInfo.LuiDATATYPE).defConvEls.get(dbType);
        }
        } else {
//            String here="here";
        }
    }
    if (el!=null){
      if (!format ||is_null(pColInfo.LuiFORMAT) || !this.LuiDataTypes.get(pColInfo.LuiDATATYPE).AlphabetOrFormat.equals("F")) {
          if (pIsToLui) {
              return el.toLui;
          } else {
              return el.fromLui;
          }
      } else {
          if (pIsToLui) {
              return el.toLuiByFormat;
          } else {
              return el.fromLuiByFormat;
          }
      } 
   
    } else
    return null;
}
public String convAndCheck(String pValue,LuiAppElement pel,LuiDbConn lc) throws LuiException, SQLException{
    ColInfo ci=pel.colinfo;
    if (ci==null){ // если нет данных о свойствах столбца запроса, то создадим их
        //ColInfo(String pSchemaName,String pTableName,String pTableAlias,String pName,String pAlias,String pColumnClassName,String pColumnTypeName,int pColumnType)
        ci=new ColInfo();
        ci.LuiDATATYPE=NVL(pel.getProperty("DATA_TYPE"),"CHAR"); 
        ci.LuiFORMAT=NVL(pel.getProperty("FORMAT"),""); 
        ci.alias=pel.getCode();
        ci.columnType=0;
        ci.columnTypeName="?";
    }
   
    String fromLui=this.getFromLuiConversion(pValue, pel, lc); // получаем выражение для преобразования pValue к типу данных СУБД
//this.getConvString(ci, lc, false);
    //if (is_null(fromLui)) return pValue;
    //String toLui=this.getToLuiConversion(pel, lc);
    String toLui=this.getConvString(ci, lc, true,true); // получаем выражение для преобразования из СУБД в LUI
    //if (is_null(toLui)) return pValue;
    pel.addProperty("v", "'"+pValue.replace("'","''")+"'");  // у поля создаём св-во v-value взяв в кавычки pvalue и задвоив их
    pel.properties.get("v").recalculate=false;               // и говорим что его не надо перевычислять
    pel.properties.get("v").currentValue=pel.properties.get("v").startedValue; // и говорим что текущее равно начальному
    pel.addProperty("f", "'"+pel.getProperty("FORMAT")+"'"); // у поля создаём св-во f-FORMAT
    LuiAppProp p=pel.addProperty("ConvertExpression",LuiParser.mask(fromLui)); // у поля создаём св-во ConvertExpression
    fromLui=p.getComputedProperty();
    fromLui=LuiParser.reMask(fromLui);
    pel.addProperty("v", fromLui);
    pel.properties.get("v").recalculate=false;  
    pel.properties.get("v").currentValue=pel.properties.get("v").startedValue;
    p=pel.addProperty("ConvertExpression",toLui);
    try{
    toLui=p.getComputedProperty();
    String stmt=this.dbmsTypes.get(lc.getDBvendor()).dummyQuery;
    stmt=stmt.replace("?", toLui);
    pValue=lc.getSelectResult(stmt);
    } catch (SQLException sex){
     if (pel.getProperty("DATA_TYPE").equals("NUMBER") &&
         is_not_null(ci.LuiFORMAT)){ // если это NUMBER и есть формат
        // то надо попробовать неформатное преобразование к числу и если это успешно 
        // переобразовать по формату к строке
        // это нужно для того, чтобы срабатывал ввод пользователем 123456 при формате 999G999G999
        // ведь далеко не каждый юзер будет вводить пробелы: 123 456
        // поэтому временно лишаем поле форматный свойств
        String cif=pel.colinfo.LuiFORMAT;
        pel.colinfo.LuiFORMAT="";
        try{
        pValue=this.convAndCheck(pValue, pel, lc);
        pel.addProperty("v", pValue);
        pel.colinfo.LuiFORMAT=cif; // восстанавливаем формат
        toLui=this.getConvString(ci, lc, true,true); // вновь получаем выражение для преобразования из СУБД в LUI
        pel.properties.get("v").currentValue=pel.properties.get("v").startedValue;
        p=pel.addProperty("ConvertExpression",toLui);
        toLui=p.getComputedProperty();
    String stmt=this.dbmsTypes.get(lc.getDBvendor()).dummyQuery;
    stmt=stmt.replace("?", toLui);
    pValue=lc.getSelectResult(stmt);
        } catch(SQLException iex){
            pel.colinfo.LuiFORMAT=cif; // восстанавливаем формат
            throw sex;
            
        }
        
        
     } else throw sex;   
    }
    return pValue;
 }


    /**
     * возвращает выражение для конвертирования значения pValue из Lui в СУБД
     *
     * @param pValue значение
     * @param pel элемент
     * @param lc соединение с БД
     * @return выражение для конвертирования
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */

public String getFromLuiConversion(String pValue,LuiAppElement pel,LuiDbConn lc) throws LuiException, SQLException{
    String v;
    
    if (pValue==null) v="null";
    else v="'"+pValue.replace("'","''")+"'";
    if (pel.colinfo==null) {
      String ldt=pel.getProperty("DATA_TYPE");
      if (is_null(ldt)) ldt="CHAR";
      pel.colinfo=new ColInfo();
      pel.colinfo.LuiDATATYPE=ldt;
      pel.colinfo.LuiFORMAT=pel.getProperty("FORMAT");
      
    }
    if (pel.colinfo.LuiDATATYPE==null)pel.colinfo.LuiDATATYPE="CHAR";
    
    switch (LuiDataTypes.get(pel.colinfo.LuiDATATYPE).emptyStringFromLuiMode){ 
        case "NULL": // если для данного LUI-типа сказано, что делать из пустой строки NULL
        if ("".equals(pValue)){// и если значение=пустая строка
           //return "null"; 
           v="null";
         } else // если не пустая строка
            if (is_not_null(LuiDataTypes.get(pel.colinfo.LuiDATATYPE).emptyStringSpecialSymbol)){ // и задан символ пустой строки
             if (pValue.equals(LuiDataTypes.get(pel.colinfo.LuiDATATYPE).emptyStringSpecialSymbol)){ // и значение равно этому спец символу
                 v="''";
             }
            
        }
        
        break;
        case "EMPTY_STRING": // если для данного LUI-типа сказано, что делать из пустой строки пустую строку
        if (pValue.equals("")){// и если значение=пустая строка
           //return "''"; 
           v="''";
         } else  // если не пустая строка  
            if (is_not_null(LuiDataTypes.get(pel.colinfo.LuiDATATYPE).nullSpecialSymbol)){ // и задан символ для NULL
             if (pValue.equals(LuiDataTypes.get(pel.colinfo.LuiDATATYPE).nullSpecialSymbol)){ // и значение равно этому спец символу
                 v="null";
             }
              }
        break;
    }
    String fromLui=this.getConvString(pel.colinfo, lc, false,true);
    if (is_null(fromLui)) return v;
    pel.addProperty("v",v);
    pel.properties.get("v").recalculate=false;
    pel.properties.get("v").currentValue=pel.properties.get("v").startedValue;
    pel.addProperty("f", "'"+pel.getProperty("FORMAT")+"'");
    LuiAppProp p=pel.addProperty("ConvertExpression",fromLui);
    fromLui=p.getComputedProperty();
    return fromLui;
}

public String getToLuiConversion(LuiAppElement el,LuiDbConn lc,String pInnerQBE,boolean format)throws LuiException, SQLException{
    String v="\""+el.colinfo.alias+"\"";
    //if (pInnerQBE.equals("Y")||el.getProperty("Sintetic").equals("Y"))v=NVL(el.getProperty("INNER_COLUMN"),el.getProperty("DEFINITION"));
    if (pInnerQBE.equals("Y")||el.getProperty("Sintetic").equals("Y"))v=NVL(el.getProperty("INNER_DEFINITION"),el.getProperty("DEFINITION"));
    if (el.getProperty("Sintetic").equals("Y") && is_null(el.colinfo.LuiFORMAT))v="("+v+")";      
    
    if (!this.LuiDataTypes.containsKey(el.colinfo.LuiDATATYPE)) return v;
    String allow=this.LuiDataTypes.get(el.colinfo.LuiDATATYPE).tl_select_clause;
    if (allow.equals("N")) return v;
    if (allow.equals("B") && (!el.getForm().getType().equals("BLANK"))) return v;
    if (allow.equals("L") && !el.getForm().getType().equals("LIST") && !el.getForm().getType().equals("TREE")) return v;
    String f="'"+el.getProperty("FORMAT")+"'";
    if (!format)f="''";
    el.addProperty("v", v);el.properties.get("v").recalculate=false;
    el.properties.get("v").currentValue=el.properties.get("v").startedValue;
    el.addProperty("f", f);
    String conv=getConvString(el.colinfo,lc,true,format);
    if (conv==null){
        if (format)
        return lc.getColumnSQLClause(v,el.colinfo.LuiDATATYPE,"CHAR",el.colinfo.LuiFORMAT);
        else return lc.getColumnSQLClause(v,el.colinfo.LuiDATATYPE,"CHAR","");
    } else {
    if (!conv.equals("")) {
    LuiAppProp p=el.addProperty("ColumnSQLClause",conv);
    conv=p.getComputedProperty();
    } else conv=v;
    
    return conv;
          }
}
public String getColumnSelectClause(LuiAppElement el,LuiDbConn lc,boolean format) throws LuiException, SQLException{
    //String v="\""+el.colinfo.alias+"\"";  //13.05.2021 Вынос order by наружу
    if (!format) return "\""+el.colinfo.alias+"\"";
    String conv=getToLuiConversion(el,lc,"N",format);
    //if (!conv.equals(v)) conv= conv+" "+v; //13.05.2021 Вынос order by наружу
    return conv;
  }


}