/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;

import LuiDbInterface.LuiDbConnPool;
import LuiDb.LuiDbConnectPool;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 *
 * @author aliubushkin
 */
 class LuiDbConnPoolArray {

    //public ArrayList<LuiDbConnPool> ConnPoolArray;
    final ArrayList<LuiDbConnPool> ConnPoolArray;
    //private ThreadForShrink t4s;

    public LuiDbConnPoolArray() {
        ConnPoolArray = new ArrayList<>();
        //t4s = new ThreadForShrink();
    }

    
    public synchronized void addPool( LuiDbConnectPool pPool )
    {
        ConnPoolArray.add(pPool);
        
    }
    
    public synchronized void setShutdown( ) {
        ConnPoolArray.forEach((p)->p.setShutdown());
    }
    
    /**
     * Создание пула соединений
     *
     * @param pDataBaseURL -- URL к БД
     * @param pUsername -- имя пользователя
     * @param pPassword -- пароль
     * @param pSchema -- схема
     * @param pAutocommit -- автокоммит
     * @param pReadonly -- только чтение
     * @param pPoolInfo -- информация о пуле сеансов
     * @return соединение
     */
    public LuiDbConnPool addPool(String pDataBaseURL, String pUsername, String pPassword, String pSchema, boolean pAutocommit, boolean pReadonly, String pPoolInfo) {
        LuiDbConnPool p = null;
        // экономия пулов
        if (pAutocommit){
        synchronized(this) {
          for (int i=0;i<ConnPoolArray.size();i++){
           p=ConnPoolArray.get(i);
           if (p.getAutoCommit() &&
                   p.getSchema().equals(pSchema) &&
                   p.getURL().equals(pDataBaseURL) &&
                   p.getUser().equals(pUsername) &&
                   
                   p.getPoolInfo().equals(pPoolInfo)){
           p.appCntIncrement(); // увеличим счётчик подписанотов
           return p;    
           } 
          } 
        }
        }
        // конец экономии пулов
        p = new LuiDbConnectPool(pDataBaseURL, pUsername, pPassword, pSchema, pAutocommit, pReadonly, pPoolInfo);
        p.appCntIncrement(); // тоже увеличим счётчик подписанотов
        synchronized(this) {
            ConnPoolArray.add(p);
        }
        return p;
    }

    public void removeAllPools() {
        synchronized(this) {
            new ArrayList<LuiDbConnPool>(ConnPoolArray).forEach(this::removePool);
            ConnPoolArray.clear();
        }
    }

    public void removePool(LuiDbConnPool p) {
        try {
            p.DisconnectAll();
        } catch (SQLException ex) {

        }
        ConnPoolArray.remove(p);
    }

    /*
    class ThreadForShrink implements Runnable {

        Thread thrd; // тред
        public volatile boolean stop;

        /**
         * Конструктор объектов класса
         *
         * @throws java.sql.SQLException
         * /
        public ThreadForShrink() {
            thrd = new Thread(this);
            this.thrd.start();
            stop = false;
        }

        @Override
        public void run() {
            long now;
            boolean killed;
            Thread.currentThread().setName("LUI DB session pool optimizer");
            try {
                while (stop != true) {
                    Thread.sleep(60000/*0* /); // спим 10 минут
                    now = System.currentTimeMillis(); // который час?
                    for (int i = 0; i < ConnPoolArray.size(); i++) { // осмотрим массив пулов
                        LuiDbConnPool pool = ConnPoolArray.get(i);
                        if (pool == null) {
                            continue;
                        }
                        killed = true; // предполагаем что убитые будут
                        while (killed) { // цикл пока есть убитый
                            killed = false; // а может не убитый
                            if (pool.GetFreeListSize() > 0) { // если свободных соединений много
                                if (killed == false) { // и в течении этого осмотра ещё никто не умирал
                                    for (int j = 0; j < pool.GetFreeListSize(); j++) { // берём одно из них
                                        if ((now - pool.GetConnectionFromFreeList(j).getLastActivity()) / 1000 > 60/*0* /) {// если нет активности более 600 сек
                                            System.out.println("Kill unused session");
                                            pool.KillFreeConnection(j); // убиваем его
                                            killed = true; // и утверждаем что убитые есть!
                                            break; // после убийства надо начать новый осмотр
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            } catch (InterruptedException ex) {

            }
        }
    }
    */
}
