/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Сравниватель элементов для выполнения сортировки 
 */
package luiserver;

import static LuiCollection.LuiString.like;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.Comparator;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * -1 если первый меньше 
 * 0 - равны 
 * 1 если второй меньше
 * @author aliubushkin
 */
public class LuiElementComparator implements Comparator<LuiAppElement> {

    @Override
    public int compare(LuiAppElement el1, LuiAppElement el2) {
  //   String s1;
  //   String s2;
     String p1;
     String p2;
     try {   
   //  s1=el1.getParent()+"."             +el1.getCode()+"("             +el1.getType()+")"             +el1.mutSeq;
   //  s2=el2.getParent()+"."             +el2.getCode()+"("             +el2.getType()+")"             +el2.mutSeq;
   //if (like(el1.getCode(),"NAME-RUS%") || like(el1.getCode(),"NAME-ENG%")) {
   //    String s3="here";
   //}
     LuiAppForm frm=el1.frm;
     if (frm!=el2.frm){
             return frm.getCode().compareTo(el2.frm.getCode());
     }
     p1=el1.getParent();
     if (!("Form".equals(p1))) {
         if (!frm.elements.containsKey(p1)) {
             p1="Form";
         }
     }
     p2=el2.getParent();
     if (!("Form".equals(p2))) {
         if (!frm.elements.containsKey(p2)) {
             p2="Form";
         }
     }
     if (!p1.equals(p2)) { // Разные родители
    if (el1.getCode().equals(p2)) { //если первый элемент является родителем для второго
        return -1;
    }     
    if (el2.getCode().equals(p1)) {// если второй элемент является родителем для первого
        return 1;
    }
    if ("Form".equals(p1)) {// если родитель первого = форма
        return compare(el1, frm.elements.get(p2)); // сравниваем первый с родителем второго
    }   
    if ("Form".equals(p2)) {// если родитель второго = форма 
        return compare(frm.elements.get(p1), el2); // сравниваем родителя первого со вторым
    }
    return compare(frm.elements.get(p1), frm.elements.get(p2)); // и у первого и у второго родители не формы. Тогда сравниваем родителей
    }
    if ("PARAM".equals(el1.getType()) && //  первый - параметр
      !("PARAM".equals(el2.getType()))) { // а второй - не параметр
          return -1;
      } 
    if ("PARAM".equals(el2.getType()) && //  второй - параметр
      !("PARAM".equals(el1.getType()))) { // а первый - не параметр
          return 1;
      } 
    if ("PARAM".equals(el2.getType()) && //  второй - параметр
       ("PARAM".equals(el1.getType()))) { // первый - параметр
          if (el1.mutSeq<el2.mutSeq) return -1;
          if (el1.mutSeq==el2.mutSeq) return 0;
          return 1;
      } 
    int S1=2, S2=2;
    switch (frm.getType()){
        case "TREE":
        case "LIST":
            if ("ACTION".equals(el1.getType())) S1=4; // Действия всегда в конце
            if ("ACTION".equals(el2.getType())) S2=4; // Действия всегда в конце
            if ("GROUP".equals(el1.getType())) { 
              if ("OVERFLOW".equals(el1.getProperty("GROUP_TYPE"))) S1=3; // Группа OVERFLOW в конце грида
              if ("FIXED".equals(el1.getProperty("GROUP_TYPE"))) S1=1; // Группа FIXED в начале
            }
            if ("GROUP".equals(el2.getType())) {
              if ("OVERFLOW".equals(el2.getProperty("GROUP_TYPE"))) S2=3; // Группа OVERFLOW в конце грида
              if ("FIXED".equals(el2.getProperty("GROUP_TYPE"))) S2=1; // Группа FIXED в начале
            }
            if (S1<S2) { /*если уровень1<уровня2*/  return -1; }
            if (S1>S2) { /*если уровень1>уровня2*/  return 1; }
            if (S1==S2) { /*если уровень1=уровня2*/
                if (el1.mutSeq<el2.mutSeq) return -1;
                if (el1.mutSeq==el2.mutSeq) return 0;
                return 1;
            }
        break;
        case "BLANK":
           if ("ACTION".equals(el1.getType()) && // если действие
              !"Form".equals(p1) &&  // не уровня формы
              !"SIDE_BY_SIDE".equals(frm.elements.get(p1).getProperty("GROUP_TYPE") // и у родителя св-во "тип группы"!=cайдбайсайд
              )) S1=3; // Действия в группе не - в конце
           if ("ACTION".equals(el2.getType()) && 
              !"Form".equals(p2) && 
              !"SIDE_BY_SIDE".equals(frm.elements.get(p2).getProperty("GROUP_TYPE")
              )) S2=3; // Действия в группе не - в конце
            if (S1<S2) { /*если уровень1<уровня2*/  return -1; }
            if (S1>S2) { /*если уровень1>уровня2*/  return 1; }
            if (S1==S2) { /*если уровень1=уровня2*/
                if (el1.mutSeq<el2.mutSeq) return -1;
                if (el1.mutSeq==el2.mutSeq) return 0;
                return 1;
            }
       break;
        default:
            if (el1.mutSeq<el2.mutSeq) return -1;
            if (el1.mutSeq==el2.mutSeq) return 0;
            return 1;
    }
     } catch (LuiException ex) {
             return 0;
         } catch (SQLException ex) {
             return 0;
         }
     return 0;   
    }
    
}
