/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Предназначен для выполнения загрузки свойств форм и элементов из справочников

 */
package luiserver;

import static LuiCollection.LuiString.NVL;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import LuiNLS.LuiNLS;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author aliubushkin
 */
public class LuiLoadFormTemplate {
 private static final String STMT4LOADELEMPROP="select p.code ,coalesce(pp.def_value,''),p.seq from lui_t_property p, lui_t_property_formtype pp where pp.property_code=p.code and pp.formtype_code='<formtype_code>' and p.etype='<etype>'";
 private static final String STMT4LOADFORMPROP="select p.code,coalesce(pp.def_value,'') from lui_t_property p,lui_t_property_formtype pp where pp.property_code=p.code and pp.formtype_code='<formtype_code>' and p.etype='FORM' and p.code!='FORM_DEV_NOTES' order by p.seq";
 private static final String STMT4LOADETYPE="select distinct etype,lower(etype) FROM lui_t_property p,lui_t_property_formtype pf where p.code=pf.property_code and pf.formtype_code='<formtype_code>' and pf.formtype_code!='FORM'";
 private static final String STMT4FORMTYPE="select code,name from lui_t_formtype";
 
 static void Load(LuiDbConn conn,LuiNLS NLS,LuiMsgWriter pMsg,HashMap<String,LuiAppForm> formTemplates) throws LuiException{
     try {    
         
         HashMap<String,String> formtypes=conn.getSelectAllRowsTwoColsValMap(STMT4FORMTYPE);
         if (formtypes==null){
             pMsg.write("LUI-00011");
         }
         for(String formtype:formtypes.keySet()){//цикл по типам форм
             LuiAppForm frm=new LuiAppForm(null,formtype+"_TEMPLATE",formtype); // создаём шаблон формы
             
        
         // Собственные свойства формы
         HashMap<String,String> props=conn.getSelectAllRowsTwoColsValMap(STMT4LOADFORMPROP.replace("<formtype_code>", formtype));
         if (props==null){
             //pMsg.write("LUI-00012",formtype);
             continue;
         }
         for(String Key:props.keySet()){
            LuiAppProp p=frm.addProperty(Key,props.get(Key));
            //p.is_meta=true;
            p.genesis='D';
         }
         //Типовые элементы формы
          HashMap<String,String> elems=conn.getSelectAllRowsTwoColsValMap(STMT4LOADETYPE.replace("<formtype_code>", formtype));
          if (elems==null){
              pMsg.write("LUI-00013",formtype);
          }
          double i=0;
          for(String elc:elems.keySet()){ // цикл по элементам
              LuiAppElement el;
              if (formtype.equals("BLANK") && elc.equals("ITEM")) { //добавляем поле в бланк
              el=new LuiAppElement(frm,"FIELD", "FIELD"+"_TEMPLATE",++i);    
              frm.elements.put("FIELD", el);
              }
              else {
              el=new LuiAppElement(frm,elc, elc+"_TEMPLATE",++i); //добавляем элемент в список
              frm.elements.put(elc, el);
                   }
              
          ArrayList<ArrayList <String>>elprops=conn.getSelectAllRowsValList(STMT4LOADELEMPROP.replace("<formtype_code>", formtype).replace("<etype>", elc));
          for(int ii=0;ii<elprops.size();ii++){//цикл по свойствам элемента
            LuiAppProp p=  el.addProperty(elprops.get(ii).get(0),NVL(elprops.get(ii).get(1),""));
            //p.is_meta=true;
            p.genesis='D';
          }
          }
          formTemplates.put(formtype, frm); // кладём шаблон в массив

         //
         }
         
         
     } catch (SQLException ex) {
         pMsg.write("LUI-00010",ex.getLocalizedMessage());
                
     
}
}
}
