/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Интерпретатор выражений и выполнятор действий
 */
package luiserver;

import LuiCollection.LuiString;
import static LuiCollection.LuiString.Extract_Word;
import static LuiCollection.LuiString.RTrim;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 *
 * @author aliubushkin
 */
public class LuiParser {
  public static final String OPENINGBRASKET="{";
  public static final String CLOSINGBRASKET="}";
  //public static final String MASKOPENINGBRASKET=new String(new byte[]{1},Charset.forName("windows-1251"));
  public static final String MASKOPENINGBRASKET=new String(new byte[]{1},Charset.forName("UTF8"));
  //public static final String MASKCLOSINGBRASKET=new String(new byte[]{2},Charset.forName("windows-1251"));
  public static final String MASKCLOSINGBRASKET=new String(new byte[]{2},Charset.forName("UTF8"));
  //public static final String HARDMASKOPENINGBRASKET=new String(new byte[]{3},Charset.forName("windows-1251"));
  public static final String HARDMASKOPENINGBRASKET=new String(new byte[]{3},Charset.forName("UTF8"));
  //public static final String HARDMASKCLOSINGBRASKET=new String(new byte[]{4},Charset.forName("windows-1251"));
  public static final String HARDMASKCLOSINGBRASKET=new String(new byte[]{4},Charset.forName("UTF8"));
  public final HashMap<String, LuiPlugin> plugins; // множество плагинов
  
  private LuiApplication application;

public LuiParser(LuiApplication pApp,HashMap<String,Class> pStdPluginClasses,HashMap<String, String> pStdPluginIsInter) throws LuiException{
    plugins=new HashMap<>();
    application=pApp;
    
    if (!application.code.equals("LUI")){
    LuiDbConn conntmp = null;
    try {
         conntmp  = application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
         String stmt="select code,filename,classname from lui_t_plugin where app_code='"+application.code+"'";
         if (application.luiSeance==null) {
             stmt=stmt+" and is_interactive='N'";
         }
         ArrayList<ArrayList<String>> plist=conntmp.getSelectAllRowsValList(stmt);
         if (plist!=null){
         for(int i=0;i<plist.size();i++){
         addPlugin(plist.get(i).get(0),plist.get(i).get(1),plist.get(i).get(2));
                 }
         }
         application.metaConnPool.SetFree(conntmp);
    } catch (Exception ex) {
         application.msgWriter.write("LUI-00018",application.code,ex.getLocalizedMessage());
         
     } finally{
         try {
           application.metaConnPool.SetFree(conntmp);
         } catch (SQLException ex1) {
         }
     } 
    }
    for (String plugCode:pStdPluginClasses.keySet()){
      LuiPlugin plugin;
    if (application.luiSeance==null &&  pStdPluginIsInter.get(plugCode).equals("Y")) continue; 
     if (!plugins.containsKey(plugCode)){
       try {
            plugin=(LuiPlugin)pStdPluginClasses.get(plugCode).newInstance();
            plugin.setContext(pApp);
            plugins.put(plugCode, plugin);
        } catch (InstantiationException ex) {
            application.msgWriter.write("LUI-00017", plugCode,ex.getLocalizedMessage());
        } catch (IllegalAccessException ex) {
System.out.println("LuiParser: plugCode = " + plugCode + " msg = " + ex.getLocalizedMessage());
//            application.msgWriter.write("LUI-00017", plugCode, ex.getLocalizedMessage());
            application.msgWriter.write("LUI-00017", plugCode, ex.getMessage());
        } catch (Exception ex){
            application.msgWriter.write("LUI-00017", plugCode,ex.getLocalizedMessage());
        }
     }
    }
}    
public void addPlugin(String pPlugCode,String pFileName,String pClassName) throws LuiException {
    File f;
    URL url;
    LuiPlugin plugin;
    try {
    f=new File(pFileName);
    url=f.toURI().toURL();
    URLClassLoader classLoader;
    classLoader=new URLClassLoader(new URL[]{url});
    plugin=(LuiPlugin) classLoader.loadClass(pClassName).newInstance();
    plugins.put(pPlugCode, plugin);
    } catch (MalformedURLException ex) {
         application.msgWriter.write("LUI-00017", pPlugCode,ex.getLocalizedMessage());
      } catch (ClassNotFoundException ex) {
         application.msgWriter.write("LUI-00017", pPlugCode,ex.getLocalizedMessage());
      } catch (InstantiationException ex) {
         application.msgWriter.write("LUI-00017", pPlugCode,ex.getLocalizedMessage());
      } catch (IllegalAccessException ex) {
         
      }
}

/**
   * Маскирут скобки
   * @param Str входная строка
   * @return строку после маскировки скобок
   */
  public static String mask(String Str) { // 
    if (Str==null) return null;
    if (Str.equals("")) return Str;
    //if (isValue)return Str.replace(OPENINGBRASKET, HARDMASKOPENINGBRASKET).replace(CLOSINGBRASKET, HARDMASKCLOSINGBRASKET);
    return Str.replace(OPENINGBRASKET, MASKOPENINGBRASKET).replace(CLOSINGBRASKET, MASKCLOSINGBRASKET);
  }

  /**
   * Восстанавливает скобки
   * @param Str входная строка
   * @return строку после восстановления скобок
   */
  public static String reMask(String Str) { // Восстанавливает замаскированные фигурные скобки
   // if (isValue)return  Str.replace(HARDMASKOPENINGBRASKET, OPENINGBRASKET).replace(HARDMASKCLOSINGBRASKET, CLOSINGBRASKET); 
    if (Str==null) return null;
    return Str.replace(MASKOPENINGBRASKET, OPENINGBRASKET).replace(MASKCLOSINGBRASKET, CLOSINGBRASKET);
  }
    /**
     * Выполняет программный код пользователя записанный в значении свойства
     * pProp
     *
     * @param RawString строка
     * @param pProp свойство
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
  /*
  public void execute(String RawString,LuiAppProp pProp) throws LuiException, SQLException {
    while (is_not_null(RawString)){
     RawString=LuiString.LTrim(RawString);
     RawString=LuiString.RTrim(RawString);
     RawString=LuiString.LTrim(RawString,"\n");
     RawString=LuiString.RTrim(RawString,"\n");
     String command=RawString; 
     if (RawString.charAt(0)=='['){ // если начинается с квадратной скобки
       // если начинается со скобки
            // ищем где она закрывается
            int cb=RawString.indexOf("]"); 
            if (cb>0){// и если она закрывается
                command=RawString.substring(1,cb);
                RawString=RawString.substring(cb+1);
            }else {RawString="";}
     } else {RawString="";}
      int ddPos=command.indexOf(":");
      if (ddPos<=0) {
        application.msgWriter.write("LUI-00021", pProp.getGenesis());
      }
      String langCode=command.substring(0,ddPos);
      // ищеем интерпретатор языка
      if (plugins.containsKey(langCode)) {
          plugins.get(langCode).execute(command.substring(ddPos+1), pProp);
      } else {
          application.msgWriter.write("LUI-00022",langCode,pProp.getGenesis());
      }
  }  
  }
  */
    public void execute(String RawString, LuiAppProp pProp) throws LuiException, SQLException {
        if (RawString==null) return; // Неизвестно что не парсим
        if (RawString.equals("")) return; // Пустую строку бессмысленно парсить
        int MaxddPos=-1; // Позиция последней программы
        String langCode=null; // Язык последней программы
        for (String plugKey:plugins.keySet()) { // Цикл по всем языкам программы

            int ddPos=("\n\n"+RawString).lastIndexOf("\n\n"+plugKey+":"); // Поиск начала программы на текущем языке
            if (ddPos>MaxddPos) { // Найдено последнее использование этой программы внутри текста, и она пока последняя среди всех других программ
                MaxddPos=ddPos; // запоминаем её начало
                langCode=plugKey; // запоминаем код её языка
            }
            
            
/* Код ниже работает неправильно, если перечень программ-плагинов такой:
------------------------
    Программа_1:
            
    Программа_2:
            
    Программа_1:
            
    Программа_2:
            
------------------------
 Если в цикле поиска среди plugins.keySet() сначала Программа_1, а затем Программа_2, то...
 в таком случае последнее использование Программы_2 он не видит (не ищет). Результатом поиска станет второй вызов Программы_1.
*/            
            
/*
            int ddPos=("\n\n"+RawString).indexOf("\n\n"+plugKey+":"); // Поиск начала программы на текущем языке
            if (ddPos>MaxddPos) { // Программа найдена и пока последняя
                MaxddPos=ddPos; // запоминаем её начало
                langCode=plugKey; // запоминаем код её языка
                // проверим наличие такой же, но ниже по тексту
                while(true){
                    int ddPos2=RawString.indexOf("\n\n"+plugKey+":", MaxddPos+1);
                    if (ddPos2>MaxddPos) {
                        MaxddPos=ddPos2+2;  
                    } else break;
                }
            }
*/            
        }
        if (langCode==null) // Ни одного языка в непустой программе не найдено!
            application.msgWriter.write("LUI-00021", pProp.getGenesis()); // Злобная ругань
        else { // Найден
            execute(RawString.substring(0, MaxddPos), pProp); // Парсим и выполняем всё, что до
            plugins.get(langCode).execute(RTrim(RawString.substring(MaxddPos+langCode.length()+1),"\n"), pProp); // Выполняем найденную программу (она последняя)
        }
    }

/**
 * 
 * @param pRawString строка
 * @param pProp свойство
 * @return вычисленное значение
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
  public String compute(String pRawString,LuiAppProp pProp) throws LuiException, SQLException {
      return comp(pRawString,pProp,false);
  }
  
    /**
     * Парсит динамику в строке. Динамика должна иметь вид: {Label:Код}, где
     * Label - имя интерпретатора, Код - код на языке интерпретатора
     *
     * @param pRawString - исходная строка
     * @param pProp - какую проперть вычисляем это нужно для того, чтобы ссылку
     * на неё внести в другие проперти если она от них зависит
     * @param pFindOnly - если true, то не надо выполнять реальных вычислений,
     * надо только пропарсить зависимости
     * @return пропарсенную строку
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
  public String comp(String pRawString,LuiAppProp pProp,boolean pFindOnly) throws LuiException, SQLException {
  
    String RetVal = pRawString;
    
    /*{ApplicationProperty:SCN}*/
    /*if (like(pRawString,"%META_Search%"))     
    {
       //     String here="here";
    }*/
    String Dynamic;
    String plugKey;
    String PStr;
    int B;
    int E;
    int L;
    int N = 1;
    //System.out.println(pRawString);
    for (;;) { // Цикл парсинга
      if ("".equals(RetVal)) break; // Пустая строка останавливает парсинг
      if ((B = LuiString.InStr(RetVal, OPENINGBRASKET, 0, -N)) == -1) break; // Ищем N-ю справа открывающую скобку. Останов, если нет.
      if ((E = LuiString.Pair(RetVal, OPENINGBRASKET, CLOSINGBRASKET, B)) == -1) {
          RetVal=RetVal.substring(0, B)+mask(OPENINGBRASKET)+RetVal.substring(B+1);
          continue;
          //break;
      } // Ищем парную ей закрывающую скобку. Останов, если нет.
      Dynamic = RetVal.substring(B + OPENINGBRASKET.length(), E); // Внутренность скобок (динамика)
      // детектор языка RC
      if (Dynamic.length()>0 && pProp.ownerObj.getForm()!=null){
          if (pProp.ownerObj.getForm().getType().equals("GRID")){
        boolean isrc=false;
        switch (Dynamic.charAt(0)){
            case ' ':case '-':case '+':case ',':case '0':case '1':case '2':case '3':case '4':case '5':case '6':case '7':case '8':case '9':isrc=true;break;
        }
        if (isrc){
          int zap=0;
          int dvo=0;
          for (int ir=0;ir<Dynamic.length();ir++){
            isrc=false;  
            switch (Dynamic.charAt(ir)){
            case ' ':case '-':case ':':case '+':case ',':case '0':case '1':case '2':case '3':case '4':case '5':case '6':case '7':case '8':case '9':isrc=true;break;
            }
            if (!isrc)break;  
            if (Dynamic.charAt(ir)==',')zap++;
            if (Dynamic.charAt(ir)==':')dvo++;
            if (dvo>1)isrc=false;
            if (zap>2)isrc=false;
            if (!isrc)break;  
          }
          if (isrc){
          if (zap==0)isrc=false;
          if (dvo<1 && zap>1)isrc=false;
          if (dvo==1 && zap<2)isrc=false;
          }
        }
        if (isrc){
          Dynamic="RC:"+Dynamic;  
        }
      }}
      L = Dynamic.indexOf(":"); // Смещение двоеточия
      if (L>0) { // Двоеточие есть
        plugKey = Dynamic.substring(0, L); // Код языка
        if (plugins.containsKey(plugKey)) { // Нашли язык
          if (pFindOnly) { // Только зависимости
            plugins.get(plugKey).findDependency(Dynamic.substring(plugKey.length()+1),pProp);
            N++;
          } else { // Парсинг
            PStr = mask(plugins.get(plugKey).compute(Dynamic.substring(plugKey.length()+1),pProp)); // Пропарсенное значение
            if (PStr == null) { // Язык не смог пропарсить
              PStr = mask(OPENINGBRASKET + Dynamic + CLOSINGBRASKET); 
            }
            if (plugins.get(plugKey).needRunImplant()){
            RetVal = Inplant(RetVal.substring(0, B) + "{ * }" + RetVal.substring(E + CLOSINGBRASKET.length()), " * ", PStr, plugins);
            } else {
            RetVal = RetVal.substring(0, B) + PStr + RetVal.substring(E + CLOSINGBRASKET.length());
            }
            N=1;
          }
        } else { // плагин не найден 
          N++;
        }
      } else { // плагин не найден 
        N++;
      }
    }
    return reMask(RetVal);
  }

  /*
  public String comp(String pRawString,LuiAppProp pProp,boolean pFindOnly) throws LuiException, SQLException {
    String CompareString;
   String RetVal = pRawString;
    String Dynamic;
    String plugKey;
    String PStr;
    int B;
    int E;
    int L;
    
    L=0;
    for (;;) {
      if ("".equals(RetVal)) break;
      if ((B = LuiString.InStr(RetVal, OPENINGBRASKET, 0, -1)) == -1) break;
      if ((E = LuiString.Pair(RetVal, OPENINGBRASKET, CLOSINGBRASKET, B)) == -1) {
        RetVal = RetVal.substring(0, B) + mask(RetVal.substring(B),false);
        continue;
      }
      CompareString = RetVal;
      Dynamic = RetVal.substring(B + OPENINGBRASKET.length(), E);
      PStr = mask(OPENINGBRASKET + Dynamic + CLOSINGBRASKET,false);
      L = Dynamic.indexOf(":");
      if (L>0) {
        plugKey = Dynamic.substring(0, L);
        if (plugins.containsKey(plugKey)) {
          if (pFindOnly) {
              plugins.get(plugKey).findDependency(reMask(Dynamic.substring(plugKey.length()+1),false),pProp);
          } else {
              // в этом месте если строка Dynamic.substring(plugKey.length()+1) содержит  *F:P_APPCODE*, где "*" это маски фигурок
              // программа remask восстанавливает скобки
              // и передаст в плагин ссылку на возможно существующий элемент P_APPCODE 
              // и если такой элемент в бланке есть то пипец - ввод в поле бланка руками строки вида {F:P_APPCODE} приведёт к сохранению в БД
              // пропарсенных данных
          PStr = plugins.get(plugKey).compute(reMask(Dynamic.substring(plugKey.length()+1),false),pProp);
          if (PStr == null) PStr = mask(OPENINGBRASKET + Dynamic + CLOSINGBRASKET,false);
          }
        } else { // плагин не найден - маскируем скобочки фигурные
            RetVal = RetVal.substring(0, B) + PStr + RetVal.substring(E + CLOSINGBRASKET.length());
            continue;
        }
      }
      
      RetVal = Inplant(RetVal.substring(0, B) + "{ * }" + RetVal.substring(E + CLOSINGBRASKET.length()), " * ", PStr, plugins);
      
    }
    return reMask(RetVal,true);
  }
*/  
  
    /**
     * Выявить зависимости прочих свойств от свойства переданного в параметре
     *
     * @param pRawString - исходная строка
     * @param pProp - какую проперть вычисляем это нужно для того, чтобы ссылку
     * на неё внести в другие проперти если она от них зависит
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void findDependency(String pRawString,LuiAppProp pProp) throws LuiException, SQLException{
    pProp.is_depends=false;
    if (is_not_null(pRawString))
    comp(pRawString,pProp,true);
    //06/06/2020 если это Value - сдалаем его не зависимым от всех прочих свойств
    if (pProp.code.equals("Value")){
      LuiAppWFE el=pProp.ownerObj;
      if (el.properties.containsKey("EType") && el.getProperty("Etype").equals("FIELD")){
     for (LuiAppProp rp:pProp.dependsFrom.keySet()){
       rp.depends.remove(pProp);
     }
     pProp.dependsFrom.clear();
      }
    }
    //06/06/2020 Конец
    //pProp.SCN=0; //06.11.2018
    if (pProp.code.equals("RENOVATION") && pProp.is_depends){ //06.11.2018
        LuiAppElement el=(LuiAppElement) pProp.ownerObj;
        if (el.properties.containsKey("Value")){ 
            LuiAppProp p=el.properties.get("Value");
            pProp.setDepend(p);
            p.startedValue="{Property:RENOVATION}";
            p.recalculate=false;
        }
    }//06.11.2018 
    else {
       //05.05.2021 
    LuiAppElement el=(LuiAppElement) pProp.ownerObj;
        if (el.properties.containsKey("Value")){ 
            LuiAppProp p=el.properties.get("Value");
            p.startedValue=p.getComputedProperty();
            p.is_depends=false;
        }
      //05.05.2021  
    }
}
public static String Inplant(String Str, String ParCode, String ParValue, HashMap<String, LuiPlugin> plugins) {
// Ищет в строке Str структуры типа {ParCode} и заментяет их на ParValue. I
    if (Str == null) return "";
    int P = Str.indexOf("{" + ParCode + "}");
    int M = 0;
    int j = 0;
    int BegStr = 0;
    String PName="";
    if (P > -1) {
      for (int i = P-1; i >=0; i--) {
        switch (Str.substring(i, i+1)) {
          case "{": M++; break;
          case "}": M--; break;
        }
        if (M > j) {
          PName = Extract_Word(Str.substring(i+1), 1, "{:");
          if (plugins.containsKey(PName)) {
            BegStr = i+1;
            break;
          } else {
            j++;
          }
        }
      }
     if (M==j) {
        PName = Extract_Word(Str, 1, ":");
      }

      // Имплантация вычисленного значения с учётом разных плагинов
      if (plugins.containsKey(PName)) {
       return Str.replace("{" + ParCode + "}", plugins.get(PName).quoteMask(Str.substring(BegStr, P), ParValue));
      }
      return Str.replace("{" + ParCode + "}", ParValue);
    }
    return Str;
  }        
}

