/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;


import LuiExceptionMsg.LuiException;
import javax.script.ScriptException;


/**
 *
 * @author aliubushkin
 */
public class LuiProgressBar {
int total=0;
int current=0;
int percent=0;
long showFreq=500;
String processId;
String title="";
String cancelable="N";
String percentText="";
String workText="";
String status="closed";
long lastShowTime;
LuiApplication app;
LuiClCommandGroup XMLcg;
LuiClCommand XMLcmd;
LuiClCommandWriter cw;
LuiClEventReader er;

// конструктор его вызов желательно делать после появления сеанса LUI
public LuiProgressBar(LuiApplication pApp){
    app=pApp;
    er=app.eventReader;//new LuiClEventReader(app.eventReader.getReader(),app.msgWriter,app.TotalTimer);
    cw=app.commandWriter;//new LuiClCommandWriter(app.commandWriter.getWriter(),app.TotalTimer); 
    if (app.luiSession!=null)
    processId=app.luiSession.id;
    else
    processId="1";
}

/**
 * Начало отображения progressbar
 * @param pTitle название процесса-прогресса
 * @param pCancelable принзак прерываемости
 * @param pPercentText текст "выполнено X из Y"
 * @param pWorkText ещё какой-то текст
 * @param pTotal всего
 * @param pCurrent сейчас
 * @return какой-то признак
 * @throws LuiException логическая ошибка
 */
public boolean startProgress(String pTitle,String pCancelable,String pPercentText,String pWorkText,int pTotal,int pCurrent) throws LuiException{
    boolean retVal=true;
    total=pTotal;
    current=pCurrent;
    lastShowTime=0;

    if (total>0)
    percent=100*pCurrent/total;
    else
    percent=0;    
    title=pTitle;
    cancelable=pCancelable;
    percentText=pPercentText;
    workText=pWorkText;
     XMLcg=cw.addGroup("Progress bar");
     XMLcmd=XMLcg.addCommand("Progress");
     XMLcmd.addScaPar("ProcessID",processId);
     XMLcmd.addScaPar("Title",title);
     XMLcmd.addScaPar("Cancelable",cancelable);
     XMLcmd.addScaPar("Percent",String.valueOf(percent));
     XMLcmd.addScaPar("PercentText",percentText);
     XMLcmd.addScaPar("WorkText",workText);
     XMLcmd=XMLcg.addCommand("Event");
     XMLcmd.addScaPar("EventCode", "ProgressContinue");
     XMLcmd.addScaPar("MS","0");
     XMLcmd.addScaPar("Params","ProcessID:"+processId+";");
     status="started";
     cw.flush();
     er.readEvents();
     while (true){
        LuiClEvent ev=er.getEvent();
        if (ev==null) break;
        switch (ev.name){
            case "ProgressContinue":   
            retVal=true;    
            break;
            case "ProgressCancel":
            retVal=false;    
            break;
            default:
            retVal=true;
        }
     }   
return retVal;
}
public boolean continueProgress(String pTitle,String pCancelable,String pPercentText,String pWorkText,int pTotal,int pCurrent) throws LuiException{
 boolean retVal=true;
 if (lastShowTime>0 && pCurrent<pTotal){ //если continueProgress не в первый раз и это не финал
     if (showFreq>(System.currentTimeMillis()-lastShowTime)) // и время лчередного показа ещё не пришло
         return true;
         }
  lastShowTime=System.currentTimeMillis();
  XMLcg=cw.addGroup("Progress bar");
  XMLcmd=XMLcg.addCommand("Progress");
  XMLcmd.addScaPar("ProcessID",processId);
  
  if (!pTitle.equals(title)){
     title=pTitle;
   XMLcmd.addScaPar("Title",title);  
  }
  if (current!=pCurrent || total!=pTotal){
   current=pCurrent;
   total=pTotal;
   if (total>0)
   percent=100*pCurrent/total;   
   else
   percent=0; 
   XMLcmd.addScaPar("Percent",String.valueOf(percent));
   
  }
  if (!pPercentText.equals(percentText)){
      percentText=pPercentText;
      XMLcmd.addScaPar("PercentText",percentText);
  }
  if (!pWorkText.equals(workText)){
      workText=pWorkText;
      XMLcmd.addScaPar("WorkText",workText);
  }
  XMLcmd=XMLcg.addCommand("Event");
     XMLcmd.addScaPar("EventCode", "ProgressContinue");
     XMLcmd.addScaPar("MS","0");
     XMLcmd.addScaPar("Params","ProcessID:"+processId+";");
     status="run";
     cw.flush();
     
     er.readEvents();
     while (true){
        LuiClEvent ev=er.getEvent();
        if (ev==null) break;
        switch (ev.name){
            case "ProgressContinue":   
            retVal=true;    
            break;
            case "ProgressCancel":
            retVal=false;    
            break;
            default:
            retVal=true;
        }
        
     }
     
 return retVal;
}
public void stopProgress() throws LuiException{
if (!"closed".equals(status)){
  XMLcg=cw.addGroup("Progress bar");
  XMLcmd=XMLcg.addCommand("ProgressClose");
      XMLcmd.addScaPar("ProcessID",processId);
total=0;
current=0;
percent=0;    
status="closed";
}
  //cw.flush();
  //app.commandWriter.flush();
}
}
