/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Cоздание условного выражения для WHERE
 */
package luiserver;


import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 *
 * @author aliubushkin
 */
public class LuiSQLCondition {
public    String isNot = ""; // Если в QBE отрицание, то будет содержать "not"
public    String relation = null; // Отношение в выражении (=, !=, <, >, >=, <=, <>, like, between)
public    String relSign =null;
public    boolean relEQ = false; // Признак выдуманного оператора сравнения '='
private   String DT; // Тип данных LUI
private   String DF; // Формат данных
private   LuiAppElement el;
public    String colValue=null; // правая часть выражения преобразованная к жеванию в SQL
public    String colValueRaw=null;// правая часть выражения as is
public    String colAlias=null; // левая часть выражения
public    String colValBetweenLeft; // левая часть between  преобразованная к жеванию в SQL
public    String colValBetweenRight; // правая часть between  преобразованная к жеванию в SQL
public    String colValBetweenLeftRaw; // левая часть between  as is
public    String colValBetweenRightRaw; // правая часть between as is
public    String condition=""; // полный SQL текст условия
private   LuiDbConn con;
private   boolean colAlias_converted=false; // признак сконвертированности столбца запроса в строку
/** Конструктор для обеспечения умолчаний в бланке LUI_B_QBE
 * 
 */
public LuiSQLCondition(){
    isNot="";
    relation="";
    colValue="";
    colValueRaw="";
    colAlias="";
    colValBetweenLeft="";
    colValBetweenRight="";
    colValBetweenLeftRaw="";
    colValBetweenRightRaw="";
    condition="";
    relEQ=false;
}

    /**
     * Конструктор Преобразует выражения для QBE в условия SQL для PostGreSQL и
     * др. СУБД
     *
     * @param lc - соединение с БД
     * @param ColVal - Выражение для QBE
     * @param pEl - столбец
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
  public LuiSQLCondition(LuiDbConn lc,LuiAppElement pEl, String ColVal) throws LuiException, SQLException {
  colAlias = "\""+pEl.getCode()+"\"";
  colValue = ColVal;
  DT = NVL(pEl.getProperty("DATA_TYPE"), "CHAR").toUpperCase();
  DF = pEl.getProperty("FORMAT");
  LuiDataType ldt=pEl.app.dataConverter.LuiDataTypes.get(DT);
  con=lc;
  el=pEl;
  if (pEl.app.dataConverter.LuiDataTypes.containsKey(DT)){ // если есть такой тип данных
    if ("Y".equals(pEl.app.dataConverter.LuiDataTypes.get(DT).tl_where_clause)){ // и столбец его надо конвертировать в строку для создания условия
        String conv=pEl.app.dataConverter.getConvString(pEl.colinfo, lc, true,true); // берём строку конвертера в LUI
        if (is_not_null(conv)){ // если она не пуста
        pEl.addProperty("v", colAlias);
        pEl.addProperty("f", DF);
        LuiAppProp p=pEl.addProperty("ConvertExpression",conv);
        conv=p.getComputedProperty(); // вычисляем конвертер получится например nls(name)
        if (is_not_null(conv)){ // если получилось
        //colAlias_converted=true; // запомним что столбец преобразован
        colAlias=conv; // и поместим выражение в левую часть условия
        }
        }
    }
  }
  
  if (colValue==null){
      relation="is null";
      colValue="";
    colValueRaw="";
    colValBetweenLeft="";
    colValBetweenRight="";
    colValBetweenLeftRaw="";
    colValBetweenRightRaw="";
      condition= colAlias+" is null";
      relEQ=false;
      return;
  } 
  if (!like(colValue,"=%") && like(colValue,"%#%")){ // Если выражение содержит # и не начинается со знака =
    int reshotka=colValue.indexOf("#");
    Integer apostrof_cnt=0;
    for(int i=0;i<reshotka;i++){
        if (colValue.charAt(i)=='\'')apostrof_cnt++;
    }
    if (Math.floorMod(apostrof_cnt,2)==0) {
     condition=colValue.replaceFirst("#", colAlias);
        return;
  }
      
  }
  
  if (colValue.matches("!.*")) { // Если выражение начинается на !, то это отрицание
    isNot = "not ";
    colValue = colValue.substring(1);
  }
  relation=null;
  String likeSign="%";
  String betweenSign="<>";
  String sign=null;
  //выявление оператора отношения - равно
  // для заданного типа данныхкрутим операторы сравнения кроме like и,between
  for (int i=0;i<ldt.compOp.size();i++){
  if (!ldt.compOp.get(i).Get_Element("dbms_type").equals("any") && 
      !ldt.compOp.get(i).Get_Element("dbms_type").equals(con.getDBvendor())) continue; 
   sign=ldt.compOp.get(i).Get_Element("sign");
   if (is_null(sign))continue;
   if (ldt.compOp.get(i).Get_Element("operator_code").equals("like")) {
       likeSign=sign; //запоминаем like
   } else if (ldt.compOp.get(i).Get_Element("operator_code").equals("between")) {
       betweenSign=sign;//запоминаем between
   } else {
     if (relation==null) {  //если оператор ещё не вычислен
      int l=sign.length(); // берём длину обозначения
      if (colValue.length()>=l){ // если длина вбитого юзером условия меньше или равна длине знака
      if  (colValue.substring(0,l).equals(sign)){ // и значение в ячейке QBE начинается на знак рассматриваемого оператора
      relation=ldt.compOp.get(i).Get_Element("operator_code");  // запомним его в relation
      relSign=sign;
      }
      }
    }    
   }
        }    
  if (relation==null) { // если оператор не нашёлся
    if (colValue.contains(likeSign)) { //а в строке есть "%"
        relation=" like ";
    } else {
      if (colValue.contains(betweenSign)) { //а в строке есть "><"
        relation=" between ";
      } else  {relation="="; // иначе считаем что отношение это '=' и запоминаем что '=' -- выдумано
               relEQ=true;
              }
      
    }
  } else { // оператор найден в начале заданной строки!
      relEQ=false;
      // вычисляем строку с которой будем строить предикат
      if (colValue.length()>relSign.length()) 
      {
          colValue = colValue.substring(relSign.length());//считаем, что за оператором идёт значение
      }   
      else {
      colValue="";
      }
  }
  // начало основных модификаций под настройки null и ''
  if (pEl.app.dataConverter.LuiDataTypes.get(DT).emptyStringFromLuiMode.equals("NULL")) { // если сказано что пустая строка это NULL
   if (relation.equals("=")) {   
    if (colValue.equals("")) relation=" is null";
  } }
  
  if (!"".equals(colValue)) { // если значение есть
      if (relation.equals(" between ")) { // если между
         colValBetweenLeftRaw=colValue.split(betweenSign)[0];
         colValBetweenLeft=makeValue(colValBetweenLeftRaw);
         colValBetweenRightRaw=colValue.split(betweenSign)[1];
         colValBetweenRight=makeValue(colValBetweenRightRaw);
         colValue=colValBetweenLeft+" and "+colValBetweenRight;
     } else {colValueRaw=colValue;
         colValue=makeValue(colValue);}
  }
  // конец основных модификаций под настройки null и ''
  // далее закоментированный фрагмент до модификаций под настройки null и ''
  /*
  if ("".equals(colValue)) { // если выражение пустая строка (а отношения сейчас уже выявлены)
   if (relation.equals("=")) relation=" is null";  
  } else { //значение есть!
     if (relation.equals(" between ")) { // если между
         colValBetweenLeftRaw=colValue.split(betweenSign)[0];
         colValBetweenLeft=makeValue(colValBetweenLeftRaw);
         colValBetweenRightRaw=colValue.split(betweenSign)[1];
         colValBetweenRight=makeValue(colValBetweenRightRaw);
         colValue=colValBetweenLeft+" and "+colValBetweenRight;
     } else {colValueRaw=colValue;
         colValue=makeValue(colValue);}
  }
  */
  // конец закоментированного фрагмента до модификаций под настройки null и ''
  relation=relation.trim();
  
  condition=isNot+colAlias+" "+NVL(relation)+" "+colValue;
  isNot=isNot.trim();
  if (relation.equals("is null")) colValueRaw="";
  if (!relation.equals("between")){
    colValBetweenLeft="";
    colValBetweenRight="";
    colValBetweenLeftRaw="";
    colValBetweenRightRaw="";
  }
}
private String makeValue(String V) throws LuiException, SQLException { // Технологическая функция
  
 if (" like ".equals(relation)) { //если like , то преобразовываем и сам столбец к строке
      if (!colAlias_converted){ // конечно, если он ещё не преобразован
      colAlias=el.app.dataConverter.getToLuiConversion(el,con,"N",true);
      colAlias_converted=true;
              }
   V= "'" + V.replace("'", "''")+"'";  
 } else { // если не like
  try{// пробуем преобразовать V к целевому типу 
   el.app.dataConverter.convAndCheck(V, el, con);
   V=el.app.dataConverter.getFromLuiConversion(V, el, con);
  } catch (Exception ex){ //если не получилось, то преобразовываем и сам столбец к строке
      if (!colAlias_converted){
      colAlias=el.app.dataConverter.getToLuiConversion(el,con,"N",true);
      colAlias_converted=true;
              }
    V= "'" + V.replace("'", "''")+"'";  
  }
 }
      
      ArrayList<LuiCollection> ops=el.app.dataConverter.LuiDataTypes.get(DT).compOp;
      for (int i=0;i<ops.size();i++) {
        if (!el.app.dataConverter.LuiDataTypes.get(DT).compOp.get(i).Get_Element("dbms_type").equals("any") && 
            !el.app.dataConverter.LuiDataTypes.get(DT).compOp.get(i).Get_Element("dbms_type").equals(con.getDBvendor())) continue;   
          if (ops.get(i).Get_Element("operator_code").equals(relation)) {
           String func4col=ops.get(i).Get_Element("function4column");
           if (is_not_null(func4col)) {
               el.addProperty("QBEColumAlias", colAlias);
               el.addProperty("QBEfunction4column", func4col);
               colAlias=el.getProperty("QBEfunction4column");
           }
           String func4cri=ops.get(i).Get_Element("function4criteria");
           if (is_not_null(func4col)) {
               el.addProperty("QBEColumCriteria", V);
               el.addProperty("QBEfunction4criteria", func4cri);
               V=el.getProperty("QBEfunction4criteria");
           }
          break;    
          }
      }
      return V;
 }
}
