/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 *  Класс сеанс в LUI
 */
package luiserver;

/**
 *
 * @author aliubushkin
 */
import LuiCollection.LuiCollection;

import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;
import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import LuiNLS.LuiNLS;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.like;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import java.io.IOException;
import java.net.Socket;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import javax.script.ScriptException;
import org.apache.commons.collections4.list.TreeList;
import org.keycloak.adapters.KeycloakDeployment;


public class LuiSeance implements Runnable {
    public  String evFormID=null;
    public  String url;
    private int licDaysOvf=0;
    private String LuiSrvName;
    private String AppFilePath;
    private  LuiClEvent ev;
    private boolean openMode = false;
    private boolean doNotWorkChanges=false;    
    private final EventChannel mChannel;
    private EventChannel sChannel; // служебный канал (secondary)
    private final AtomicBoolean mShutdown = new AtomicBoolean(false);
    //private BufferedWriter writer; //писатель tcp-клиенту
    private final LuiMsgWriter msgWriter; //хранитель и плеватель сообщений
    //private BufferedReader reader; //читатель из tcp-клиента
    private final LuiClEventReader eventReader;
    private final LuiClCommandWriter commandWriter;
    protected LuiMsgWriter msgWriterSecondary; //хранитель и плеватель служебных сообщений
    protected LuiClCommandWriter commandWriterSecondary;
    private final LuiNLS NLS;
    private LuiApplication appSys;
    private final LuiDbConnPoolArray connPoolArray;
    private final LuiDbConnPool metaConnPool;
    private final HashMap<String, LuiAppForm> formTemplates;
    private final HashMap<String, Class> stdPluginClasses; // множество стандартных плагинов
    private final HashMap<String, String> stdPluginIsInter; // признаки интерактивности плагинов
    private final LuiCollection commonForms;
    private final LuiDataConverter LuiDatConv;
    public LuiProcTimer procTimer;
    public Thread thread; // ссылка на работающий 
    
    private static final AtomicLong stSeanceCounter = new AtomicLong(0L);
    public final long mSeanceId;
    private final LuiSrv mOwnerServer;
    private LuiException lastLuiException;    
    private LuiCrip crypter;
    public  long LuiSessionID;
    private boolean messaging; // поддержка обмена сообщений с другими сеансами пользователей
    public GsonBuilder gsonBuilder = new GsonBuilder();
    public Gson gson = gsonBuilder.create(); 
  private Boolean JSONrunning=false; 
  private int  MONLIM;
    private KeycloakDeployment deployment; // хранит данные о размещении сеанса пользователя во внешней системе при выполнении авторизации при помощи SSO keycloak
//private String line; // 
//private StringBuilder eventString;
//private LuiHandler handler;

    public KeycloakDeployment getDeployment() {
        return deployment;
    }

    public void setDeployment(KeycloakDeployment deployment) {
        this.deployment = deployment;
    }
    

    /**
     * Конструктор класса "Сеанс в LUI"
     *
     * @param pServer -- это сам сервер
     * @param pChannel -- это канал связи с клиентом
     * @param p_ConnPoolArray -- массив пулов соединений к БД
     * @param p_metaConnPool -- спец пул для метаданных
     * @param pNLS -- Языковое окружение по умолчанию
     * @param pFormTemplates -- шаблоны форм
     * @param pStdPluginClasses -- стандартные плагины
     * @param p_commonForms -- перечень (код:имя;код:имя....) общих форм
     * вызываемых из метаданных прикладной системы LUI и работающих в служебном
     * соединении
     * @param PstdPluginIsInter -- какая-то карта интерактивности (?)
     * @param p_crypter -- шифровальник
     * @param pLuiSrvName -- имя LUI-сервера
     * @param pAppFilePath -- путь к каталогу файлов приложения
     * @param pLuiDatConv -- конвертер данных
     */
    public LuiSeance(LuiSrv pServer, EventChannel pChannel, LuiDbConnPoolArray p_ConnPoolArray, LuiDbConnPool p_metaConnPool, LuiNLS pNLS, HashMap<String, LuiAppForm> pFormTemplates, HashMap<String, Class> pStdPluginClasses,HashMap<String, String> PstdPluginIsInter, LuiCollection p_commonForms,LuiCrip p_crypter,String pLuiSrvName,String pAppFilePath,LuiDataConverter pLuiDatConv,int pMONLIM) {
    MONLIM=pMONLIM;
        LuiDatConv=pLuiDatConv;
        LuiSrvName=pLuiSrvName;
        mOwnerServer = pServer;
        mSeanceId = stSeanceCounter.incrementAndGet();
        mChannel = pChannel;
        connPoolArray = p_ConnPoolArray;
        metaConnPool = p_metaConnPool;
        NLS = pNLS;
        formTemplates = pFormTemplates;
        stdPluginClasses = pStdPluginClasses;
        stdPluginIsInter=PstdPluginIsInter;
        commonForms = p_commonForms;
        procTimer = new LuiProcTimer();
        messaging = false;
        crypter=p_crypter;
        AppFilePath=pAppFilePath;
        // MetadataDBconnPool=p_ConnPool;

        /*
        try {
            writer = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), "Windows-1251")); // буфер отдачи данных клиенту UTF!! 
            reader = new BufferedReader(new InputStreamReader(socket.getInputStream(), "Windows-1251")); // буфер чтения данных от клиента UTF!
        } catch (Exception ex) {
            System.out.println("Error creating reader/writer");
        }
        */

        msgWriter = new LuiMsgWriter(metaConnPool, NLS, "LUI");// создаём плевателя сообщений
        commandWriter = new LuiClCommandWriter(mChannel.writer(), procTimer);//создаём писателя команд
        eventReader = new LuiClEventReader(mChannel.reader(), msgWriter, procTimer); //создаём читателя событий
        

    }

    /**
     * Получить ID сеанса
     * @return ID сеанса
     */
    public long getSeanceId() { return mSeanceId; }
    
    @Override
    public void run() { //****
        //Thread.currentThread().setName("LUI Seance");
      int oldMsgSize = 0;  
  /*    ArrayList<Double> tl=new ArrayList<>();
      tl.add(11.0);
      tl.add(1.4);
      tl.add(1.01);
      Collections.sort(tl);
      for (int i=0;i<tl.size();i++){
          Double d=tl.get(i);
          d=d;
      }*/
      try {  
        while( ! mShutdown.get() ) {  // см выше :)
            
            if (eventReader.eventList.size() == 0) {
                try {
                    eventReader.readEvents(); // прочитали группу событий                    
                } catch (LuiException ex) {
                   if (ex.getCode().equals("LUI-00001")) {
                       setShutdown();
                   } else
                    LuiExceptionHandle(ex);
                }
            }
           if (mShutdown.get()) {
               break;
           } // если шутдаун тутже рвём итерацию цикла
            if (appSys != null) { // если приложение есть
                appSys.setVal("Close Detail Mode", "");

            }

            // начинаем обрабатывать считанные события
            try {
                openMode = false;
                while (true) {
                    ev = eventReader.getEvent(); // извлекаем событие
                    if (ev == null) {
                        break;
                    }
                    clEventHandler();          // обрабатываем событие
                    if (ev.name.equals("Open")) {
                        openMode = true;
                        break;
                    }

                } // конец цикла обработки событий
                if (appSys != null && !openMode) {
                    if (doNotWorkChanges){
                     doNotWorkChanges=false;   
                    } else {
                    appSys.changes();
                    }
                }
            } catch (LuiException exr) {
                // обработка Lui- ошибки
                    if (exr.getCode().equals("break")) {
                    try {
                        appSys.actionBreak();
                    } catch (Exception exb) {
                    }
                } else {
                    if (exr.getCode().equals("LUI-00001")) {
                       setShutdown();
                   } else
                    LuiExceptionHandle(exr);
                }
            } catch (SQLException exr) {
                // обработка SQL ошибки, которая возникла в ходе обработки событий
                SQLExceptionHandle(exr);
            } catch (Exception exr) {
                // обработка невateедомой ошибки, которая возникла в ходе обработки событий
                
               if (is_not_null(appSys.getVal("rtChart"))){
                 if (appSys.forms.containsKey(appSys.getVal("rtChart"))){
                   LuiAppForm cfrm=appSys.forms.get(appSys.getVal("rtChart"));
                   try{
                   cfrm.close("","");
                   } catch(Exception cex){
                      String here="here" ;
                   }
                   appSys.forms.remove(appSys.getVal("rtChart"));
                   appSys.setVal("rtChart","");
                 }  
               } 
                ExceptionHandle(exr);
            }

            //Фатальных ошибок не случилось
            this.remZeroCmd();
            msgWriter.msgs2commands(commandWriter); //Выгребание из очереди сообщений и отображение их клиенту
            if (appSys == null) { // если приложение не создалось
              if (commandWriter.commandGroupList.size()>0) {// а что-то есть в командах
                  try{
                  commandWriter.flush();
                     } catch (Throwable ignore) { }
              }
            }
            else {
               if (!openMode) {
                   this.remFrmWinApp();
               } else { // openMode=true
                   for (int i = 0; i < commandWriter.commandGroupList.size(); i++) { //крутим группы команд
                         if (appSys==null)continue;   
                        for (int j = 0; j < commandWriter.commandGroupList.get(i).commandList.size(); j++) {
                         if (appSys==null)continue;
                            LuiClCommand XMLcmd = commandWriter.commandGroupList.get(i).commandList.get(j);
                            String par;
                            switch (XMLcmd.name) {
                              case "EnableWindow":
                                case "Window":
                                    par = XMLcmd.getParValue("WindowID");
                                    
                                     if (appSys.windows!=null){
                                    if (appSys.windows.containsKey(par)) {
                                        appSys.windows.get(par).enable();
                                    }
                                    }
                                    
                                    break;
                                case "DisableWindow":
                                    par = XMLcmd.getParValue("WindowID");
                                    if (appSys.windows.containsKey(par)) {
                                        appSys.windows.get(par).disable();
                                    }
                                    break;  
                            }
                        }
                   }
               } // openMode=true
                if (eventReader.eventList.size() == 0) {
                    try {
                        lastLuiException=null;
                        synchronized(commandWriter){
                        // ПОСКОЛЬКО В ПРОЦЕСС FLUSH МОГУТ ВОЗНИКАТЬ lui ОШИБКИ ИЗМЕРИМ ПРЕДВАРИТЕЛЬНО
                        oldMsgSize=0;
                        if  (msgWriter.msgArray!=null)   
                        oldMsgSize=msgWriter.msgArray.size();
                        commandWriter.flush();
                        }
                    }  catch (ConcurrentModificationException cex){
                        cex.printStackTrace();
                    }  catch (LuiException lex){
                       msgWriter.msgs2commands(commandWriter); //Выгребание из очереди сообщений и отображение их клиенту 
                       try{
                       commandWriter.flush();
                       } catch (Exception ex2){try { mChannel.close(); } catch(Throwable ignore) {} 
                        setShutdown();}
                    }
                        
                        catch (Exception ex) {
                        if (msgWriter.msgArray!=null && oldMsgSize!=msgWriter.msgArray.size()){
                        //if (ext.indexOf("LUI-")>0){
                       commandWriter.commandGroupList.clear();
                       msgWriter.msgs2commands(commandWriter); //Выгребание из очереди сообщений и отображение их клиенту      
                       try{
                       commandWriter.flush();
                       } catch (Exception ex2){try { mChannel.close(); } catch(Throwable ignore) {} 
                        setShutdown();}
                        } else{
                        //????? в этом месте надо понимать, что связь с клиентом вероятно потеряна
                        // и не имеет смысла в него чё-то слать
                        // и если надо (debug) то скинуть мессагу в лог-файл
                        // а приложение надо бы удалить
                        // ё сам сеанс - утечка памяти!
                        System.out.println(ex.toString());
                        ex.printStackTrace();
                        //socket = null;
                        try { mChannel.close(); } catch(Throwable ignore) {} 
                        setShutdown();
                        }
                    }
                  if (mShutdown.get())  {
                      try {
                                            mChannel.close();
                                        } catch (Throwable ignore) { }
                  }
                }
            }

        } // пока сокет не закрыт
        if (appSys != null) {
            appSys.close();
        }
        System.out.println("Socket thread " + Thread.currentThread().getName() + " finished");
      }
      finally {
          mOwnerServer.removeSeance(this);
      }

    }

    /**
     * обработка события
     *
     */
    private void clEventHandler() throws LuiException, SQLException, ScriptException {
        
        
        if (appSys == null) { // если приложения ещё нет
            if (!ev.name.equals("Open")) { // а событие не open
                msgWriter.write("LUI-00005", ev.name);
            }
        }
        if (ev.name.equals("Open")) {
            if (appSys != null) {
                commandWriter.commandGroupList.clear();
                appSys.close();
            } // если уже есть приложение закроем его
            // создаём новое приложение
            String Params = ev.getParValue("Params");
            if (is_null(Params)) {
                Params = ev.getParValue("params");
            }
            String appcode;
            String HostName;
            String lang="";
            String browserlang="";
            
            if (is_not_null(Params)) {
                LuiCollection parcol = new LuiCollection(Params.toUpperCase());
                appcode = NVL(parcol.Get_Element("APPCODE"),mOwnerServer.defApp);
                lang = parcol.Get_Element("LANG");
               
            } else {
                appcode = NVL(ev.getParValue("AppCode"),mOwnerServer.defApp);
                lang = ev.getParValue("Lang");
                browserlang=ev.getParValue("UserLang");
                if (like(browserlang.toUpperCase(),"%RU%")){
                    browserlang="RUS";
                } else browserlang="ENG";
            }
            lang=NVL(lang,browserlang);
            HostName = ev.getParValue("HostName");
            url=ev.getParValue("URL");
            // получаем значение ClientID и MsgPort для нового (служебного) сокета
            String clientID=ev.getParValue("ClientID");
            String msgPort=ev.getParValue("MsgPort"); 
            // проверяем наличие прикладной системы
            LuiDbConn lc=null;
            
            try {
              lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code,appcode,"","");  
              if (is_null(lc.getSelectResult("select code from lui_t_application where code='"+appcode+"'"))) {
                msgWriter.write("LUI-00006",appcode, NLS.NLS("RUS:Приложение не найдено;ENG:Application not found;"));
              }
              String days=lc.getSelectResult("select now()::date-a.datetime::date\n" +
"from lui_t_audit a order by a.id limit 1 ");
              if (Integer.valueOf(NVL(days,"0"))>MONLIM*30){
                 licDaysOvf=Integer.valueOf(days)-MONLIM*30; 
              } else {
                  licDaysOvf=0;
              }
            } finally {
                if (lc!=null)metaConnPool.SetFree(lc);
            }
            
            appSys = new LuiApplication(this, appcode, lang, metaConnPool, msgWriter, connPoolArray, NLS,
                    formTemplates, commandWriter, stdPluginClasses,stdPluginIsInter, eventReader, procTimer, commonForms, HostName,crypter, clientID, msgPort,LuiSrvName,AppFilePath
            ,LuiDatConv,Params);
            appSys.addProperty("ApplicationURL", url);
            appSys.changeSCN("Open"); // меняем SCN
            if (appSys.getProperty("ExtAuthError").equals("Y")) {
              System.out.println("LuiSeance:Close application when external autorization denied.");
              appSys.close();
              appSys=null;
              eventReader.eventList.clear();
              this.setShutdown();
              return;  
            } 
            appSys.run();
            return;
        }
        checkLuiSession("EventHandler"); //проверим сессию
        // всё кроме open
        //appSys.changeSCN(ev.name); // меняем SCN - перенесено в каждую из программм
        // переопределим FormID
        evFormID = ev.getParValue("FormID");
        if (is_not_null(evFormID)) {
            if (!appSys.forms.containsKey(evFormID)) return; // уходим если формы уже нет
            appSys.setVal("FormID", evFormID);
        } else {
            evFormID = appSys.getVal("FormID");
        }
        if (!appSys.forms.containsKey(evFormID)){
            evFormID=appSys.getVal("FormID");
        }
        
        switch (ev.name) {
     case "Action":
     appSys.forms.get(evFormID).htmlAction(ev.getParValue("Code"),ev.getParValue("Parameters"));
     break;
     case "Activate":
     appSys.windows.get(ev.getParValue("WindowID")).activate();
     break;
     case "ButtonPressed":
     appSys.forms.get(evFormID).buttonPressed(ev.getParValue("Code"));
     break;    
     case "CellButton":
     appSys.forms.get(evFormID).elements.get(ev.getParValue("Column")).cellButton(ev.getParValue("Name"));
     break;    
     case "CellRMB":
     /*if (!ev.params.containsKey("Column") || !ev.params.containsKey("Row")){
     appSys.ask(NLS.NLS("RUS:Для получения локального меню установите указатель мыши в ячейку списка.;ENG:To get the local menu, place the mouse pointer in the list cell.;"),
             "OK:OK;", null);
     }  else  {*/
     appSys.changeSCN("CellRMB");
     appSys.forms.get(evFormID).cellRMB(ev.getParValue("Column"),ev.getParValue("Row"),ev.getParValue("Context"));
     //}
     break;
     case "ChangeCell":
    appSys.forms.get(evFormID).MatrixChangeCell(ev.getParValue("C-code"),ev.getParValue("R-code"));
     break;
     case "ChangeConfig":
     appSys.forms.get(evFormID).changeConfig(ev.getParValue("ConfigID"));
     break;
     case "ChangeMode":
     appSys.forms.get(evFormID).changeMode();
     break;
     case "ChangeNode":
     appSys.forms.get(evFormID).changeNode(ev.getParValue("Node"), ev.getParValue("State"), ev.getParValue("NType"),ev.getParValue("Context"),ev.getParValue("ContextID"),ev.getParValue("ChildsContextID"));
     break;    
     case "ChangeState":
     appSys.forms.get(evFormID).changeState(ev.getParValue("Node"),ev.getParValue("State"),ev.getParValue("NType"),ev.getParValue("Context"),ev.getParValue("ContextID"),ev.getParValue("ChildsContextID"));
     break;
     case "ChartClick":
     appSys.forms.get(evFormID).chartClick("Left",ev.getParValue("Parameters"),null);
     break;
     case "ChartOver":
    if (like(ev.getParValue("Parameters"),"{\"componentType\":\"title\"%")) //пока только для заголовков   
    appSys.forms.get(evFormID).chartClick("Over",ev.getParValue("Parameters"),null);    
     break;
     case "ChartOut":
     if (like(ev.getParValue("Parameters"),"{\"componentType\":\"title\"%")) //пока только для заголовков    
     appSys.forms.get(evFormID).chartClick("Out",ev.getParValue("Parameters"),null);    
     break;
     case "ChartRMB":
     appSys.forms.get(evFormID).chartClick("Right",ev.getParValue("Parameters"),null);
     break;
     case "ColormapDC":
     case "PiecewiseDC":
     appSys.forms.get(evFormID).chartClick("LeftDouble",ev.getParValue("Parameters"),"colormap");
     break;
     case "ChartDC":
     appSys.forms.get(evFormID).chartClick("LeftDouble",ev.getParValue("Parameters"),null);
     break;
     case "ChartLegendClick":
     appSys.forms.get(evFormID).chartLegendClick(ev.getParValue("Name"),"CLICK");   
     break;
     case "ChartLegendRMB":
     appSys.forms.get(evFormID).chartLegendClick(ev.getParValue("Name"),"RMB");   
     break;
     case "ChartLegendDC":
     appSys.forms.get(evFormID).chartLegendClick(ev.getParValue("Name"),"DC");
     break;
     case "ChartResize":
     appSys.forms.get(evFormID).chartResize(ev.getParValue("Width"),ev.getParValue("Height"));
     break;
     case "ChartZoom":
     appSys.forms.get(evFormID).chartZoom(ev.getParValue("Scale"));    
     break;
   /*  case "ChartMapPosition":
     appSys.forms.get(evFormID).chartMapPosition(ev.getParValue("Parameters"));
     break;*/
     case "CheckAll":
     appSys.forms.get(evFormID).checkAll();
     break;
     case "CheckNode":
     appSys.forms.get(evFormID).checkNode(ev.getParValue("Node"),ev.getParValue("State"),ev.getParValue("Context"),ev.getParValue("ContextID"),ev.getParValue("isCheck"));    
     break;
     case "CheckRow":
     appSys.forms.get(evFormID).checkRow(ev.getParValue("IsCheck"),appSys.forms.get(evFormID).getUK() );
     break;
     case "ChoiceMenuItem":
     appSys.choiceMenuItem(ev.getParValue("Type"), ev.getParValue("Code"), evFormID, ev.getParValue("Params"));
     break;
     case "CloseWindow":
     for (String fc:appSys.forms.keySet()){
         if (appSys.forms.get(fc).getVal("WindowID").equals(ev.getParValue("WindowID")) && appSys.forms.get(fc).code.equals(appSys.getProperty("ERROR_MESSAGE_FORM"))) {
          doNotWorkChanges=true;
          eventReader.eventList.clear();
          break;
         }
     }
     if (doNotWorkChanges)appSys.windows.get(ev.getParValue("WindowID")).Close(false);     
     else appSys.windows.get(ev.getParValue("WindowID")).close();     
     break;
     case "ColumnDC":
     if (!ev.params.containsKey("Column")){
     }  else  {
     appSys.changeSCN("ColumnDC_"+ev.getParValue("Column"));
     appSys.forms.get(evFormID).setVal("ColumnClick", ev.getParValue("Column"));
     String ColClickAction=appSys.forms.get(evFormID).checkHotKey("COLUMN_DBLCLICK", ev.getParValue("Column"));
     if (is_not_null(ColClickAction))appSys.forms.get(evFormID).runAction(ColClickAction);
     }    
     break;
     case "ColumnLMB":
     if (!ev.params.containsKey("Column")){
     }  else  {
     appSys.changeSCN("ColumnLMB_"+ev.getParValue("Column"));
     appSys.forms.get(evFormID).setVal("ColumnClick", ev.getParValue("Column"));
     String ColClickAction=appSys.forms.get(evFormID).checkHotKey("COLUMN_CLICK", ev.getParValue("Column"));
     if (is_not_null(ColClickAction))appSys.forms.get(evFormID).runAction(ColClickAction);
     } 
     break; 
     case "ColumnRMB": // отображение меню для заголовка столбца т.е. это действия COLMENU% уровня формы
     if (is_not_null(ev.getParValue("Column"))){    
     appSys.forms.get(evFormID).elements.get(ev.getParValue("Column")).columnRMB();
     }
     break;
     case "DefaultAction":
     appSys.forms.get(evFormID).defaultAction();
     break;
     case "DeleteBar":
     appSys.forms.get(evFormID).deleteBar();
     break;
     case "DeleteRow":
     appSys.forms.get(evFormID).deleteRow();
     break;
     case "DownArrayBar":
     appSys.forms.get(evFormID).elements.get(ev.getParValue("Element")).downArrayBar();
     break;
     case "DropDown": // это обрабатывается в плагине LOVD тоже
     appSys.forms.get(evFormID).elements.get(ev.getParValue("Field")).newValue(ev.getParValue("Value")+"¦", "", "", "Y",true,false);
     break;
     case "Dublicate":
     appSys.forms.get(evFormID).duplicate();    
     case "EnterBar":
/* проверим необходимость подавления 
if (appSys.forms.get(evFormID).getType().equals("BLANK")) {
    LuiAppForm bfrm=appSys.forms.get(evFormID);
    String curfld=bfrm.getVal("CurrentField");
    if (is_not_null(curfld)){
       if (bfrm.elements.containsKey(curfld)) {
         if (bfrm.elements.get(curfld).getProperty("CHANGING").equals("Y")) {
           String vsbl=bfrm.elements.get(curfld).getProperty("VISIBLE");  
           if (!vsbl.equals("1") &&!vsbl.equals("Y")) {  
            break;
           }
         }
       }
    }
}
        */
     appSys.forms.get(evFormID).defaultAction();
     break;
     case "EscapeBar":
     if (appSys.forms.containsKey(evFormID)){  
         
     String win=appSys.forms.get(evFormID).getVal("WindowID");/// текущее окно
     if (appSys.windows.containsKey(win)){    
     if (!appSys.forms.get(evFormID).backFromRTchart()){
        for (String fc:appSys.forms.keySet()){
         if (appSys.forms.get(fc).getVal("WindowID").equals(win) && appSys.forms.get(fc).code.equals(appSys.getProperty("ERROR_MESSAGE_FORM"))) {
          doNotWorkChanges=true;
          eventReader.eventList.clear();
          break;
         }
     } 
         appSys.windows.get(win).close();
     }
     
        }}
     break;
     case "F1Bar":
     case "F5Bar":
     case "F4Bar": // дублирование значений в бланк из предыдущего бланка
     case "F6Bar":
     case "F7Bar":
     case "F8Bar":
     case "F9Bar":
     case "QBE":
     case "InsertBar":
     case "ShiftF6Bar":
     appSys.forms.get(evFormID).fBar(ev);
     break;  
     case "F10Bar":
     if (appSys.forms.get(evFormID).getType().equals("BLANK")){
         appSys.forms.get(evFormID).defaultAction();
     }
     break;
     case "FetchRows":
     appSys.forms.get(evFormID).fetchRows();
     break;    
     case "FieldButton":
     appSys.forms.get(evFormID).fieldButton(ev.getParValue("Name"));     
     break;
     case "FieldRMB":
     appSys.forms.get(evFormID).elements.get(ev.getParValue("Field")).fieldRMB();
     break;
     case "FormActivate":
     //очищаем очередь от лишних FormActivate
     //вычислим отца формы evFormID
     
     String PevFormID=appSys.forms.get(evFormID).getParentID();
     LuiClEvent lev=ev;     
     boolean bi=true;
     while (bi){
     bi=false;
     for(int i=0;i<eventReader.eventList.size();i++){
         if (eventReader.eventList.get(i).name.equals("FormActivate")) {
             String ePevFormID = eventReader.eventList.get(i).getParValue("FormID"); //форма в событии
             if (appSys.forms.get(evFormID).getVal("Position").equals(appSys.forms.get(ePevFormID).getVal("Position"))){ // позиции совпадают TabLeft=TabLeft
             ePevFormID=appSys.forms.get(ePevFormID).getParentID();
              if (PevFormID.equals(ePevFormID)) { // отцы совпадают
              lev=eventReader.eventList.get(i); // запоминаем удаляемое
              eventReader.eventList.remove(i);bi=true; //   удаляем лишнее событие
              break;
              }
             }
         }
     }
     }
     ev=lev;
     evFormID = ev.getParValue("FormID");
     if (appSys.forms.get(evFormID).getVal("WaitingForFormActivate").equals("Y")){ // если форма ждёт активации
         appSys.forms.get(evFormID).setVal("WaitingForFormActivate","N");
         if (appSys.forms.get(evFormID).getType().equals("LIST")) // если список
         appSys.forms.get(evFormID).requery(); // выполняем запрос
         if (appSys.forms.get(evFormID).getType().equals("GRID")) // если матрица
         appSys.forms.get(evFormID).reloadForm();// выполняем запрос
         LuiClCommandGroup pXMLcg=this.commandWriter.addGroup("Set Focus to TabForm");
       LuiClCommand pXMLcmd=pXMLcg.addCommand("Focus");
      pXMLcmd.addScaPar("FormID",evFormID);
     }
     break;
     case "GetTotal":
         if (appSys.forms.containsKey(evFormID)){
     appSys.forms.get(evFormID).getTotal();    
         }
     break;
     case "InvertChecking":
     appSys.forms.get(evFormID).invertChecking();
     break;
     case "Interval":
     appSys.interval(); 
     break;
     case "GotoCell":
     appSys.forms.get(evFormID).gotoCell(ev.getParValue("RowNum"),ev.getParValue("Column"),new LuiCollection(ev.getParValue("Context")),false);
     break;        
     case "GotoField":
     appSys.forms.get(evFormID).setField(ev.getParValue("Field")/*,""ev.getParValue("Cursor")*/);
     break;    
     case "GroupDC":
     if (!ev.params.containsKey("Group")){
     }  else  {
     appSys.forms.get(evFormID).setVal("GroupClick", ev.getParValue("Group"));
     String GrpClickAction=appSys.forms.get(evFormID).checkHotKey("GROUP_DBLCLICK", ev.getParValue("Group"));
     if (is_not_null(GrpClickAction))appSys.forms.get(evFormID).runAction(GrpClickAction);
     }
     break;
     case "GroupLMB":
     if (!ev.params.containsKey("Group")){
     }  else  {
     appSys.forms.get(evFormID).setVal("GroupClick", ev.getParValue("Group"));
     String GrpClickAction=appSys.forms.get(evFormID).checkHotKey("GROUP_CLICK", ev.getParValue("Group"));
     if (is_not_null(GrpClickAction))appSys.forms.get(evFormID).runAction(GrpClickAction);
     }
     break;
     case "GroupRMB":
     if (appSys.forms.get(evFormID).elements.containsKey(ev.getParValue("Group"))){
     appSys.forms.get(evFormID).elements.get(ev.getParValue("Group")).groupRMB();
     }
     break;
     case  "RMB":
     case  "LMB":
     case  "DBLCLK":
     if (appSys.forms.get(evFormID).getType().equals("HTML")) {
         appSys.forms.get(evFormID).htmlMouseEvent(ev.name,ev.getParValue("Parameters"),ev.getParValue("Code"));
     }
     break;
     case "MatrixSelect":
     appSys.forms.get(evFormID).MatrixSelect(ev.getParValue("Columns"),ev.getParValue("Rows"),ev.getParValue("Cells"));
     break;
     case "MatrixRMB":
     appSys.forms.get(evFormID).MatrixRMB(ev.getParValue("C-code"),ev.getParValue("R-code"));
     break;
     case "MatrixClick":
     case "MatrixDC":
    String cCode=ev.getParValue("C-code");
    String rCode=ev.getParValue("R-code");
    String r="";
    String c="";
     if (is_not_null(cCode))   { appSys.forms.get(evFormID).addProperty("CurrentCOLN",appSys.forms.get(evFormID).elements.get(cCode).getProperty("ltn"));c=appSys.forms.get(evFormID).elements.get(cCode).getProperty("ltn");}
     if (is_not_null(rCode))   { appSys.forms.get(evFormID).addProperty("CurrentROWN",appSys.forms.get(evFormID).elements.get(rCode).getProperty("ltn"));r=appSys.forms.get(evFormID).elements.get(rCode).getProperty("ltn");}
    appSys.forms.get(evFormID).addProperty("CurrentCOL",cCode);
    appSys.forms.get(evFormID).addProperty("CurrentROW",rCode);
    appSys.forms.get(evFormID).addProperty("CurrentTuple","");
    if (is_not_null(appSys.forms.get(evFormID).getProperty("CurrentCOL")) && is_null(appSys.forms.get(evFormID).getProperty("CurrentROW"))){
    appSys.forms.get(evFormID).addProperty("CurrentTuple",appSys.forms.get(evFormID).getProperty("CurrentCOL"));    
    }
    if (is_null(appSys.forms.get(evFormID).getProperty("CurrentCOL")) && is_not_null(appSys.forms.get(evFormID).getProperty("CurrentROW"))){
    appSys.forms.get(evFormID).addProperty("CurrentTuple",appSys.forms.get(evFormID).getProperty("CurrentROW"));    
    }
    appSys.forms.get(evFormID).addProperty("CurrentCELL","");
    /*if (is_not_null(cCode) && is_not_null(rCode)){
    for (String ek:appSys.forms.get(evFormID).elements.keySet()){
        if (!appSys.forms.get(evFormID).elements.get(ek).getType().equals("CELL"))continue;
        if (appSys.forms.get(evFormID).elements.get(ek).getProperty("ROW").equals(rCode) && appSys.forms.get(evFormID).elements.get(ek).getProperty("COL").equals(cCode))
        {appSys.forms.get(evFormID).addProperty("CurrentCELL",ek);
        break;}
    }
    } */
    if (is_not_null(r)&& is_not_null(c)) {
        if (appSys.forms.get(evFormID).MatrixMap.containsKey(r+":"+c)){
            ArrayList<String> acell=appSys.forms.get(evFormID).MatrixMap.get(r+":"+c);
            if (acell!=null && acell.size()>0) appSys.forms.get(evFormID).addProperty("CurrentCELL",acell.get(acell.size()-1));
        }
    } 
     if (ev.name.equals("MatrixDC"))
     appSys.forms.get(evFormID).defaultAction();
     else {
     if (!appSys.forms.get(evFormID).getProperty("SELECTION_MODE").equals("SIMPLE")){
     String elClick="Form";
     String role="ANGULAR_CLICK";
     if (is_null(ev.getParValue("R")) && is_null(ev.getParValue("C"))) role="OUTSIDE_CLICK";
     if (is_not_null(appSys.forms.get(evFormID).getProperty("CurrentTuple")))elClick=appSys.forms.get(evFormID).getProperty("CurrentTuple");
     if (is_not_null(appSys.forms.get(evFormID).getProperty("CurrentROW")))role="ROW_CLICK";
     if (is_not_null(appSys.forms.get(evFormID).getProperty("CurrentCOL")))role="COLUMN_CLICK";
     String ClickAction=appSys.forms.get(evFormID).checkHotKey(role, elClick);
     if (is_not_null(ClickAction))appSys.forms.get(evFormID).runAction(ClickAction);
     }}
     break;
     case "NewValue":
     if (is_not_null(ev.getParValue("Field"))){    
     if (ev.params.containsKey("Value"))
     {   
         appSys.forms.get(evFormID).elements.get(ev.getParValue("Field")).newValue(ev.getParValue("Value"),ev.getParValue("StartCursor"),ev.getParValue("EndCursor"),ev.getParValue("IsCheckFormat"),true,true);
     }      
     else { 
         appSys.forms.get(evFormID).elements.get(ev.getParValue("Field")).newValue(null,ev.getParValue("StartCursor"),ev.getParValue("EndCursor"),ev.getParValue("IsCheckFormat"),true,true);
     }
             }
     break;
     case "NextPartLOV": //обрабатывается в плагине LOVQ
     break;  
     case "NodeDC":
         appSys.forms.get(evFormID).nodeDC(ev.getParValue("Node"),ev.getParValue("State"),ev.getParValue("NType"),ev.getParValue("Context"),ev.getParValue("ContextID"),ev.getParValue("ChildsContextID"));
     break;
     case "NodeRMB":
         appSys.forms.get(evFormID).nodeRMB(ev.getParValue("Node"),ev.getParValue("State"),ev.getParValue("NType"),ev.getParValue("Context"), ev.getParValue("ContextID"));
     break;
     case "PatternMatch": // соответствие текста набранного в поле некому регулярному выражанию 
     appSys.forms.get(evFormID).patternMatch();
     break;
     case "ProgressCancel": // соответствие текста набранного в поле некому регулярному выражанию 
     if (appSys.progressBar!=null) {   
      LuiClCommandGroup pXMLcg;
      LuiClCommand pXMLcmd;

      pXMLcg=this.commandWriter.addGroup("Progress bar");
      pXMLcmd=pXMLcg.addCommand("ProgressClose");
      pXMLcmd.addScaPar("ProcessID",ev.getParValue("ProcessID"));
     }
     break;

     case "Requery":
     appSys.forms.get(evFormID).requery();
     break;
     case "Redo":
     String redo=appSys.forms.get(evFormID).checkHotKey("REDO", "Form");
     if (is_not_null(redo))appSys.forms.get(evFormID).runAction(redo);
     break;
     case "SetUserProperty":
     appSys.setUserProperty(ev.getParValue("WindowID"),ev.getParValue("FormID"), ev.getParValue("Element"),ev.getParValue("Property"),ev.getParValue("Value"));
     break;
     case "SystemMenu":
     appSys.systemMenu(ev.getParValue("WindowID"));
     break;
     case "ToolBarButton":
     if (appSys.forms.get(evFormID).getType().equals("TREE")){
       String root=appSys.forms.get(evFormID).getVal("RootFormID");
       if (appSys.forms.containsKey(root)){
       String context=appSys.forms.get(root).getVal("ContextForkID");
       if (is_null(context))context=root;
       if (appSys.forms.containsKey(context)){
         appSys.setVal("FormID", context);
         appSys.forms.get(context).toolBarButton(ev.getParValue("Name"));
         
       }
               }
     } else
     appSys.forms.get(evFormID).toolBarButton(ev.getParValue("Name"));
     break;
     case "UnCheckAll":
     appSys.forms.get(evFormID).unCheckAll();
     break;
     case "Undo":
     String undo=appSys.forms.get(evFormID).checkHotKey("UNDO", "Form");
     if (is_not_null(undo))appSys.forms.get(evFormID).runAction(undo);
     break;
     
        }
        /*if (eventReader.getEventListSize()==0) { // если событий в массиве больше не накоплено
msgWriter.msgs2commands(commandWriter);
commandWriter.flush();
}*/
//commandWriter.commandGroupList.clear();
    }

    
    /**
     * Предварительное прерывание сеанса
     */
    public void preStop() {
setShutdown();
        if(thread != null && thread.isAlive()) {
            thread.interrupt();
        }
    }
    
    /**
     * Прерывание LUI-сеанса
     */
    public void stop() {
        try {
            if (appSys != null) {
                appSys.close();
            }
setShutdown();
            mChannel.close();
            
           /* if (socket.isClosed() != true) {
                socket.close();
                socket = null;
            }*/
        } catch (Exception e) {

        }
    }

    /**
     * Вычисляет наличие права
     *
     * @param pConn - соединение с БД
     * @param pAppCode - Прикладная система
     * @param pObjType - Тип объекта
     * @param pObj - Идентификатор объекта
     * @param pRgh - Право
     * @param pUser - Пользователь
     * @return признак наличия права
     * @throws SQLException ошибка выполнения SQL-запроса
     */
    public String getEffectiveRight(LuiDbConn pConn, String pAppCode, String pObjType, String pObj, String pRgh, String pUser) throws SQLException {
        if (pUser.equals("")) {
            return "Y";
        }
        if (pUser.equalsIgnoreCase("LUI")) {
            return "Y";
        }
        String lang="RUS";
        if (appSys!=null){
            lang=appSys.NLS.currentLang.code;
        if (pAppCode.equals(appSys.code)){
            if (pUser.equals(appSys.owner)) {
                return "Y";
            }
        }
            }
        String retVal;
        //retVal=pConn.getSelectResult("select lui_f_get_effective_right('"+pAppCode+"','"+pObjType+"','"+pObj+"','"+pRgh+"','"+pUser+"','"+lang+"')");
        retVal=pConn.getSelectResult("select lui_f_get_effective_right('"+pObjType+"','"+pObj+"','"+pRgh+"','"+pUser+"','"+pAppCode+"')");
        return retVal;
        /*
        String sqltext = null;
        
        try {
            retVal = pConn.getSelectResult("Select Min(RO.STATUS)"
                    + " from lui_t_object_right RO"
                    + "     ,lui_t_user_in_group UG"
                    + " where RO.app_code='" + pAppCode
                    + "' and ug.app_code='" + pAppCode
                    + "' and RO.usersgroup_code=UG.usersgroup_code"
                    + "  and UG.user_name='" + pUser
                    + "' and RO.object_type='" + pObjType
                    + "' and RO.object='" + pObj
                    + "' and RO.action_type='" + pRgh
                    + "' and RO.status in ('GRANT','REVOKE')");
            if (is_null(retVal)) {
                retVal = pConn.getSelectResult("Select Min(RO.STATUS)"
                        + " from lui_t_object_right RO"
                        + "     ,lui_t_user_in_group UG"
                        + " where RO.app_code='" + pAppCode
                        + "' and ug.app_code='" + pAppCode
                        + "' and RO.usersgroup_code=UG.usersgroup_code"
                        + "  and UG.user_name='" + pUser
                        + "' and RO.object_type='" + pObjType
                        + "' and RO.object like '" + pObj
                        + "' and RO.action_type='" + pRgh
                        + "' and RO.status in ('GRANT','REVOKE')");
            }
            if (is_null(retVal)) {
                ArrayList<String> als = pConn.getSelectFirstRowValList(
                        "select set_select,set_code,set_name,set_from,set_parent "
                        + "from lui_t_object_type_right where app_code='" + pAppCode
                        + "' and code='" + pObjType + "'"
                );
                if (als == null) {
                    return "N";
                }
                sqltext = "select " + als.get(0) + " " + als.get(1) + " code, " + als.get(2) + " name, " + als.get(4) + " Parent_Code from " + als.get(3);
                if (is_not_null(als.get(4))) {
                    if (!als.get(4).equalsIgnoreCase("null")) {
                        String newSelect = "Select Max(Parent_Code) from ("
                                + sqltext
                                + ") where code='" + pObj + "'";
                        String newObj = pConn.getSelectResult(newSelect);
                        if (is_not_null(newObj)) {
                            if (getEffectiveRight(pConn, pAppCode, pObjType, newObj, pRgh, pUser, pCascade).equals("Y")) {
                                retVal = "GRANT";
                            } else {
                                retVal = "REVOKE";
                            }
                        }
                    }
                }
            }
            if (is_null(retVal)) {
                retVal = pConn.getSelectResult("select def_status from lui_t_action_type4object_type where app_code='" + pAppCode + "' and object_type='" + pObjType + "' and action_type='" + pRgh + "'");
                if ("Y".equals(retVal)) {
                    retVal = "GRANT";
                } else {
                    retVal = "REVOKE";
                }
            }
            if (pCascade && NVL(retVal, "REVOKE").equals("REVOKE")) {
                if (is_null(sqltext)) {
                    ArrayList<String> als = pConn.getSelectFirstRowValList(
                            "select set_select,set_code,set_name,set_from,set_parent "
                            + "from lui_t_object_type_right where app_code='" + pAppCode
                            + "' and code='" + pObjType + "'"
                    );
                    if (als == null) {
                        return "N";
                    }
                    sqltext = "select " + als.get(0) + " " + als.get(1) + " code, " + als.get(2) + " name, " + als.get(4) + " Parent_Code from " + als.get(3);
                }
                String stmt = "select code from (" + sqltext + ") where Parent_Code";
                if (pObj == null) {
                    stmt = stmt + " is null";
                } else {
                    stmt = stmt + "='" + pObj + "'";
                }
                ArrayList<ArrayList<String>> newObjs = pConn.getSelectAllRowsValList(stmt);
                for (int i = 0; i < newObjs.size(); i++) {
                    if (getEffectiveRight(pConn, pAppCode, pObjType, newObjs.get(i).get(0), pRgh, pUser, pCascade).equals("Y")) {
                        retVal = "GRANT";
                        break;
                    } else {
                        retVal = "REVOKE";
                    }
                }
            }
            if (is_null(retVal)) {
                return "N";
            }
            if (retVal.equals("GRANT")) {
                return "Y";
            }
            return "N";
        } catch (Exception ex) {
        }
        return "N";
*/
    }

    /**
     * Обработчик Lui-exception
     *
     * @param ex исключение
     */
    public void LuiExceptionHandle(LuiException ex) {
        //if (socket != null) {
        //if(! mShutdown.get()) { // зачем раньше тут была проверка на shutdown?
            //eventReader.eventList.clear(); // чистим остатки жененых посылок
           commandWriter.commandGroupList.clear(); // чистип приготовленные команды
            if (appSys != null) {
                try {
                    appSys.actionBreak();
                } catch (LuiException ex1) {

                } catch (SQLException ex1) {

                }
            }
         boolean show=true;
         if (lastLuiException!=null){
           if (lastLuiException.toString().equals(ex.toString())) show=false; 
         }
         if (show){
          String errfrmCode=null;
           /*формирование сообщения для конечного пользователя*/
            if (appSys!=null) {
            try{ 
              errfrmCode=appSys.getProperty("ERROR_MESSAGE_FORM");  
              String FormID=NVL(appSys.getVal("FormID"),"");
              String FormCode="";
              if (appSys.forms.containsKey(FormID)){
                  FormCode=NVL(appSys.forms.get(FormID).getProperty("MetadataCode"),"");
              }
              if (is_not_null(errfrmCode)) {
               if (!errfrmCode.equals(FormCode)){   
                  LuiCollection parcol=new LuiCollection("");
                    parcol.Add_Element("CONFIG","DEFAULT");
                    parcol.Add_Element("MSG_SEVERETY",ex.severity);
                    parcol.Add_Element("MSG_CODE",ex.message_code);
                    parcol.Add_Element("MSG_TEXT",ex.message_text);
                    parcol.Add_Element("MSG_HELPTEXT",ex.message_helptext);
                    parcol.Add_Element("MSG_CALLING",ex.calling);
                    appSys.msgWriter.ClearMsgArray();
                    eventReader.eventList.clear();
                    doNotWorkChanges=true;
                    if (is_not_null(evFormID) && appSys.forms.containsKey(evFormID))
                    appSys.executeForm(errfrmCode, parcol.toString(), evFormID,null/*,"N"*/);
                    else
                    appSys.executeForm(errfrmCode, parcol.toString(), appSys.getVal("FormID"),null/*,"N"*/);
              } else errfrmCode="";
              }
            }
               catch (LuiException|SQLException lex){
                  errfrmCode=null;
               }
                    }
            /*Конец формирования сообщения для конечного пользователя*/
            if (is_null(errfrmCode)){
            LuiClCommandGroup cg = commandWriter.addGroup("Error");
            LuiClCommand c = cg.addCommand("Message");
            c.addParameter("Type", "Fail");
            String err="Error ";
            
            if (appSys!=null) {
            err=appSys.NLS.NLS("RUS:Ошибка ;ENG:Error ;");
                    }
            
            c.addParameter("Title",err + ex.getCode());
            c.addParameter("Text", ex.getMessage());
            c.addParameter("Notes", ex.toString());
            
                
         }
         }  
            this.lastLuiException=ex;
            Integer msgind = ex.getIndex();
            if (msgind != null && msgWriter.msgArray.size()>msgind) {
                 msgWriter.msgArray.remove((int) msgind);
            }
            /*try {
                         commandWriter.flush();
                         } catch (LuiException ex1) {
                             System.out.println("Ошибка отображения ошибки:"+ex1.toString());
                         }*/
      //  }
    }

    public void ExceptionHandle(Exception ex) {
        // if (socket != null) {
        if(! mShutdown.get() ) {
            //eventReader.eventList.clear(); // чистим остатки жененых посылок
            commandWriter.commandGroupList.clear();// чистип приготовленные команды
            if (appSys != null) {
                try {
                    appSys.actionBreak();
                } catch (LuiException ex1) {

                } catch (SQLException ex1) {

                }
            }
            LuiClCommandGroup cg = commandWriter.addGroup("Error");
            LuiClCommand c = cg.addCommand("Message");
            c.addScaPar("Type", "Fail");
            c.addScaPar("Title", "Java Error");
            StackTraceElement[] ste = ex.getStackTrace();
            String m = NVL(ex.getMessage(), " null pointer ");
            m = m + " in ";

            for (int i = 0; i < ste.length; i++) {
                StackTraceElement st = ste[i];
                //if (!like(st.getClassName(),"java.lang%")) {
                m = m + "\n";
                m = m + st.getClassName() + "." + st.getMethodName() + ":line" + st.getLineNumber();
                //}
            }

            //m = LuiString.MaskAncillaryChars(m);
            c.addScaPar("Text", m);
            //c.addParameter("Notes", ex.getStackTrace().toString());
        }
    }

    public void SQLExceptionHandle(Exception ex) {
        //if (socket != null) {
        if(!mShutdown.get()) {
            //eventReader.eventList.clear(); // чистим остатки жененых посылок
            commandWriter.commandGroupList.clear();// чистип приготовленные команды
            if (appSys != null) {
                try {
                    appSys.actionBreak();
                } catch (LuiException ex1) {

                } catch (SQLException ex1) {

                }
            }
             
           String errfrmCode=null;
           /*формирование сообщения для конечного пользователя*/
            if (appSys!=null) {
            try{ 
              errfrmCode=appSys.getProperty("ERROR_MESSAGE_FORM");  
              String FormID=NVL(appSys.getVal("FormID"),"");
              String FormCode="";
              if (appSys.forms.containsKey(FormID)){
                  FormCode=NVL(appSys.forms.get(FormID).getProperty("MetadataCode"),"");
              }
              if (is_not_null(errfrmCode)) {
                 if (!errfrmCode.equals(FormCode)){   
                  LuiCollection parcol=new LuiCollection("");
                    parcol.Add_Element("CONFIG","DEFAULT");
                    parcol.Add_Element("MSG_SEVERETY","E");
                    parcol.Add_Element("MSG_CODE","LUI-XXXXX");
                    parcol.Add_Element("MSG_TEXT",ex.getMessage());
                    parcol.Add_Element("MSG_HELPTEXT","");
                    parcol.Add_Element("MSG_CALLING","");
                    appSys.msgWriter.ClearMsgArray();
                    eventReader.eventList.clear();
                    doNotWorkChanges=true;
                    appSys.executeForm(errfrmCode, parcol.toString(), appSys.getVal("FormID"),null/*,"N"*/);
              }else errfrmCode="";
              }
            }
               catch (LuiException|SQLException lex){
                  errfrmCode=null;
               }
                    }
            /*Конец формирования сообщения для конечного пользователя*/
            if (is_null(errfrmCode)){
            LuiClCommandGroup cg = commandWriter.addGroup("Error");
            LuiClCommand c = cg.addCommand("Message");
            c.addParameter("Type", "Fail");
            c.addParameter("Title", "SQL Error");
            c.addParameter("Text", ex.getMessage());
            c.addParameter("Notes", "error note"/* ex.toString()*/);
            }
        }

    }
/**
 * Проверка наличия сеанса
 * @param pFrom  откуда (?)
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */    
    public void checkLuiSession(String pFrom) throws SQLException, LuiException{
        if (!mShutdown.get()) {
            
            if (appSys != null) {
                
                if (appSys.luiSession != null) {
                    
                    if (appSys.luiSession.id != null) {
                    
                        LuiDbConn lc = null;
                        try {
                          try{  
                            lc = appSys.metaConnPool.GetFree(appSys.NLS.currentLang.locale, appSys.NLS.currentLang.code, appSys.code, appSys.luiSession.id, appSys.getAppUser());
                          } catch (SQLException ex){
                            //Если сейчас не удалось получить сеанс
                            // то может удасться позже
                             return;
                          }                            
                            lc.getSelectResult("select 1 from lui_t_session where end_datetime is null and id=" + appSys.luiSession.id);
                            if (lc.getRowCount() < 1) {
                                //setShutdown();
                                
                                appSys.msgWriter.write("LUI-00042");
                                
                                LuiClCommandGroup XMLcgl= appSys.commandWriter.addGroup("LogOff");
                                XMLcgl.addCommand("LogOff");
                                appSys.commandWriter.flush();
                                setShutdown();
                                appSys.close();
                                appSys=null;
                            }
                        } catch (Exception ex){
                            //ошиька чтения сеанса из БД - гасим приложение 
                             
                               
                                appSys.msgWriter.write("LUI-00042");
                                
                               LuiClCommandGroup XMLcgl= appSys.commandWriter.addGroup("LogOff");
                               XMLcgl.addCommand("LogOff");
                                appSys.commandWriter.flush();
                                setShutdown();
                                appSys.close();
                                appSys=null;
                        }
                        finally {
                            if (lc != null) {
                                if (appSys != null) {
                                    if (appSys.metaConnPool != null) {
                                        appSys.metaConnPool.SetFree(lc);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
// ---------------------------------------- //

    public LuiSrv getServer() {
        return this.mOwnerServer;
    }
    
    public LuiApplication getApplication() {
        return this.appSys;
    }

    // Создание сокета/канала для служебных сообщений.
    public void createSecondaryChannel(String pMsgPort, String pClientID) throws LuiException {
        // После создания сокета - через out-поток этого сокета передаётся значение pClientID.
        // Перед созданием сокета надо вычислить ip-адрес из данных об имеющемся основном сокете/канале.
      if (sChannel!=null) return;  
      if( !(mChannel instanceof SocketChannel)) return;
        
        try {
            String s=((SocketChannel)mChannel).getSocket().getInetAddress().toString();
            Socket secondarySocket = new Socket(((SocketChannel)mChannel).getSocket().getInetAddress(), Integer.parseInt(pMsgPort));
            //sChannel = new SocketChannel(secondarySocket, Charset.forName("Windows-1251"));
            sChannel = new SocketChannel(secondarySocket, Charset.forName("UTF8"));
            msgWriterSecondary = new LuiMsgWriter(metaConnPool, NLS, "LUI"); // создаём плевателя служебных сообщений
            commandWriterSecondary = new LuiClCommandWriter(sChannel.writer(), procTimer); // создаём писателя служебных команд
            commandWriterSecondary.getWriter().write(pClientID + "\n"); // отправляем ID клиента для установки связи этого служебного сокета с конкретным браузером (браузер идентифицируется по ID клиента)
            commandWriterSecondary.getWriter().write("\n");
            commandWriterSecondary.getWriter().flush();
            messaging = true;
        } catch (IOException e) {
            msgWriter.write("LUI-00104", e.getMessage());
        } catch (Exception e) {
            msgWriter.write("LUI-00104", e.getMessage());
        }  
    }; 
    
    
    public boolean getMessaging() {
        return messaging;
    }
public void remFrmWinApp(){
                            for (int i = 0; i < commandWriter.commandGroupList.size(); i++) { //крутим группы команд
                         if (appSys==null)continue;   
                        for (int j = 0; j < commandWriter.commandGroupList.get(i).commandList.size(); j++) {
                         if (appSys==null)continue;
                            LuiClCommand XMLcmd = commandWriter.commandGroupList.get(i).commandList.get(j);
                            String par;
                            switch (XMLcmd.name) {
                                case "CloseForm":
                                    par = XMLcmd.getParValue("FormID");
                                    appSys.removeForm(par);
                                    break;
                                case "CloseWindow":
                                    par = XMLcmd.getParValue("WindowID");
                                    
                                    if (appSys.windows.containsKey(par)) 
                                        appSys.windows.remove(par);
                                    
                                    break;
                                case "HideWindow":
                                    par = XMLcmd.getParValue("WindowID");
                                    if (appSys.windows.containsKey(par)) {
                                        appSys.windows.get(par).hide();
                                    }
                                    break;
                                case "EnableWindow":
                                case "Window":
                                    par = XMLcmd.getParValue("WindowID");
                                    
                                     if (appSys.windows!=null){
                                    if (appSys.windows.containsKey(par)) {
                                        appSys.windows.get(par).enable();
                                    }
                                    }
                                    
                                    break;
                                case "DisableWindow":
                                    par = XMLcmd.getParValue("WindowID");
                                    if (appSys.windows.containsKey(par)) {
                                        appSys.windows.get(par).disable();
                                    }
                                    break;
                                case "LogOff":
                                    appSys.close();
                                    appSys = null;
                                    setShutdown(); // указываем что надо бы прекратить сеанс
                                break;
                            }
                        }
                    }
                    if (appSys != null) {
                        ArrayList<String> forErase=new ArrayList<>();
                        for (String frmc : appSys.forms.keySet()) {// крутим формы
                            if (appSys.forms.get(frmc).getVal("Status").equals("Loading")) {//если форма не дозагружена
                                if (is_not_null(appSys.forms.get(frmc).getVal("WindowID"))) {// и у неё есть окно
                                    forErase.add(frmc); // заносим форму в список удаляемых
                                    //appSys.removeForm(frmc);//удалим форму - это выкидыш
                                }
                            }
                        }
                        for(int i=0;i<forErase.size();i++){
                            appSys.removeForm(forErase.get(i));
                        }
                        appSys.resetChangeFlag(); /// сброс флажков изменённости
                    }

}    
public void remZeroCmd(){
boolean a = true;
            while (a) {
                a = false;
                for (int i = 0; i < commandWriter.commandGroupList.size(); i++) {
                if (!commandWriter.commandGroupList.get(i).really){a=true; commandWriter.commandGroupList.remove(i);break;}
                    boolean b = true;
                    while (b) {
                        b = false;
                        for (int j = 0; j < commandWriter.commandGroupList.get(i).commandList.size(); j++) {
                            if (!commandWriter.commandGroupList.get(i).commandList.get(j).really) {
                                b = true;
                                commandWriter.commandGroupList.get(i).commandList.remove(j);
                                break;
                            }
                        }
                    }
                    if (commandWriter.commandGroupList.get(i).commandList.size() == 0) {
                        commandWriter.commandGroupList.remove(i);
                        a = true;
                        break;
                    }
                }
            }    
}
public void setShutdown(){
mShutdown.set(true);    
}
public String toJSON(Object o){
  synchronized(this.JSONrunning)  {
    int i=0;  
    while (JSONrunning){
     if (i>9) break;
        try {
            Thread.sleep(100);
        } catch (InterruptedException ex) {
           
        }
     i++;
    }
    JSONrunning=true;
    try {
     return gson.toJson(o);
    } catch (ConcurrentModificationException cex){
        return "{}";
    } finally {
        JSONrunning=false;
    }
    
    
  }        
}
public void licAtt(int pDl,int pDr){
  if (licDaysOvf<pDl)return;
  if (licDaysOvf>pDr)return;
  /*
  String text="Внимание! Достигнут предел периода бесплатного использования LUI. Оставьте заявку на приобретение на сайте ";
  String texte="Attention! The limit of the free LUI usage period has been reached. Leave a purchase request on the website";
  String ctext=appSys.crypter.encrypt(text);
  String ctexte=appSys.crypter.encrypt(texte);
  String url="lui.fors.ru";
  String urle="lui.fors.com";
  String curl=appSys.crypter.encrypt(url);
  String curle=appSys.crypter.encrypt(urle);
  String apologies="Приносим свои извинения.";
  String apologiese="Please accept our apologies.";
  String capologies=appSys.crypter.encrypt(apologies);
  String capologiese=appSys.crypter.encrypt(apologiese);*/
  
  String text=this.crypter.decrypt("pAmKEvHLN6kp77vQ8WaIw4hHw6SNavLCF/k3eTXv+U29QT8WByNtmLqIBGvOLBztGXcjhMxOrm8WQAofeXjxaiNeig6KGSoMOco0O24Z4m2UVHlP8o+iyN58xq1c3kKLFJF1QR5flOMqBv98rWr37LC57RobVyG7mICM+bd++5sr5/hRT7/VcEPOrXS7JOJPD1Y5rfsMDe0OJ9aJSvu2ThHzpbj6vW7ZQol8TQo52939qYr8/+HCzc3e73LUVsMHVWH8IZfoX04=")+" ";
  String url=this.crypter.decrypt("24PrypEfa5mlXHG57iL1zg==");
  String apologies=this.crypter.decrypt("BWRD44+WF2tFbn0vUEdIl8KZs8wt0eyKPxIM01d8TDIi2eGhozy4D/w9yxep1zw1");
  if (!"RUS".equals(appSys.NLS.currentLang.code)){
  text=this.crypter.decrypt("EPxFSei16Y6Hy9yGjBlLKrjMR4HdDHsoOJ/SNW7JiOIu0eQb8SUN43hvJT5LUMqitMey2cvFsFh27fMw7t59wAVTwFh52ISAlj4s+YjoKxIbTm/uvkgTZAhfMxj11thWj7wsG8T26DHdUr2K+fsccg==");    
  url=this.crypter.decrypt("24PrypEfa5l/Tsp1XGMn4A==");
  apologies=this.crypter.decrypt("RUqSDl6R7FjsyxxVzqO2Q0Rm7zayB2W2zCsYiYeOiN0=");
  }
  
  try {
  appSys.parser.plugins.get("HTML").execute("(Title:;Target:M;)"+
                 "<!DOCTYPE html>\n" +
"<HTML><HEAD>\n" +
"<meta http-equiv=\"X-UA-Compatible\" content=\"IE=Edge\">\n" +
"<META http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n" +
"</HEAD>\n" +
"<body style=\"padding:10pt;margin:0;background-color: #172548;color:white;text-align:center;font: normal 180% Arial, Verdana, Tahoma;\">\n" +
"<h3 style=\"font-size:180%;color: #FFB;text-shadow: 1px 2px 3px red;\">"+text+"<a href=\"http://"+url+"/#distribution\" target=\"_blank\" style=\"color:#BFF\">"+url+"</a></h3>\n" +
"<p>"+apologies+"</p>\n" +
"</body>\n" +
"</html>"
                 ,appSys.properties.get("AppCode"));
  }catch(Exception ex){}
}

}
