/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 Создание и изменение данных о LUI-сеансе в БД
 */
package luiserver;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.sql.SQLException;

/**
 *
 * @author aliubushkin
 */
public class LuiSession {
 public String id;
 public String start_datetime;
 public String end_datetime;
 public String session_property;
 public LuiApplication app;
 public String userName;
 
 // конструктор
 public LuiSession(LuiApplication pApp,String pUser) throws LuiException, SQLException{
 app=pApp;
 userName=pUser;
 if (app.interceptor4sql_error!=null) {
     app.interceptor4sql_error.clear();
 }
 String stmt="";
 LuiDbConn lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,"","");
 if (lc!=null) {
 try {
     String async_task="null";
     if (app.properties.containsKey("AsyncTask")){
         async_task="'"+app.getProperty("AsyncTask")+"'";
     }
     stmt="select code,interceptor4sql_error from lui_t_message where app_code='"+app.code+"' and interceptor4sql_error>' ' order by code";
     app.interceptor4sql_error=lc.getSelectAllRowsTwoColsValMap(stmt);
     stmt="select nextval('lui_q_session')";
     id=lc.getSelectResult(stmt);
     if (app.luiSeance!=null){
     app.luiSeance.LuiSessionID=Long.valueOf(id);
     }
     stmt="select clock_timestamp()";
     start_datetime=lc.getSelectResult(stmt);
     String sp="DEBUG_MODE:"+app.getProperty("DEBUG_MODE")+";";
     stmt="INSERT INTO lui_t_session(id, app_code, start_datetime, end_datetime, user_name, session_property,client_hostname,lui_server,task_code)VALUES ("+
             id+",'"+app.code+"','"+start_datetime+"'::lui_d_datetime,null,'"+pUser+"','"+sp+"','"+app.getProperty("ClientHostName")+"','"+this.app.LuiSrvName+"',"+async_task+");";
     lc.executeDML(stmt);
     if (!async_task.equals("null")){
       lc.executeDML("update lui_v_task set last_session_id="+String.valueOf(id)+" where app_code='"+app.code+"' and code="+async_task);
     }
     app.addProperty("SessionID",id);
     app.addProperty("AppUser", pUser);
     
     app.luiSession=this;
     if (app.log!=null) app.log.setSID(this.id);
     app.progressBar=new LuiProgressBar(app); 
     
 }  
    catch (SQLException ex){
     app.msgWriter.write("LUI-00053",ex.getMessage(),stmt);
 } finally {
     app.metaConnPool.SetFree(lc);
 }
 }
 for (String cc:app.userDbConnections.keySet()){ // бежим по соединениям
     LuiAppConn ac=app.userDbConnections.get(cc);
     if (ac.getSessionMode()=='P'){ // если персональная
       String info="";
       //                              1        2                    3                      4                        5          6                        
                        info="LUI;"+NVL(app.code,"")+";"+NVL(app.NLS.currentLang.code,"")+";"+NVL(app.getAppUser(),"")+";"+NVL(id,"")+";"+NVL(ac.name,"");

                        /*info=info+"LUI App:"+app.code+";";   
                        info=info+"Lang:"+app.NLS.currentLang.code+";";   
                        info=info+"SesID:"+id+";";   
                        info=info+"AppUsr:"+app.getAppUser()+";";  
                        info=info+"Pool:"+ac.name+";";*/
       ac.currentConnection.setSessionInfo(info);
     }
 }
 if (!app.userDbConnections.containsKey("LUI$")){ // если нет LUI$-соединения то его надо добавить
     LuiAppConn srvc = new LuiAppConn(app,"LUI$","LUI service connection","COMMON",app.metaConnPool.getUser(),"***","","lui");
  srvc.connPool=app.metaConnPool;
  app.userDbConnections.put("LUI$", srvc);
 }
 }
// Завершение сеанса
public void endSession(boolean pCloseAppDBConnect) {
  if (app.luiSession==null) return;
if (is_not_null(app.getVal("rtChart"))){ //если есть графики реального времени - гасим их и их окно
   String chartFrmID=app.getVal("rtChart");
   if (app.forms.containsKey(chartFrmID)){
      try { 
       String win=app.forms.get(chartFrmID).getVal("WindowID");
       app.forms.get(chartFrmID).close("","");
       //app.forms.remove(chartFrmID);
       //app.windows.remove(win);
      } catch (Exception ex){
          
      }
   }
}  
  String stmt="";
  LuiDbConn lc=null;
   
 try {
 lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());  
 } catch(Exception ex) {
     return;
 }
 if (lc!=null) {
   try {
    if ("Y".equals(app.properties.get("DEBUG_MODE").currentValue) ||
            app.properties.containsKey("AsyncTask")){   
       stmt="select clock_timestamp()";
       end_datetime=lc.getSelectResult(stmt);
       stmt="update lui_t_session set end_datetime='"+end_datetime+"'::lui_d_datetime,event_processing_stop='"+end_datetime+"'::lui_d_datetime where id="+id;
       lc.executeDML(stmt); }
    else {
       stmt="delete from lui_t_session where id="+id;
       lc.executeDML(stmt);
    }
    lc.executeDML("delete from lui_t_data_grouping where is_stored='N' and session_id="+id);
    if (app.properties.containsKey("SessionID"))
    app.properties.remove("SessionID");
   }  catch (SQLException ex){
     //app.msgWriter.write("LUI-",ex.getMessage(),stmt);
 } finally {
   if (pCloseAppDBConnect) {
   for(String c:app.userDbConnections.keySet()){
    if (!("LUI$".equals(c))){ // закрываем соединения кроме служебного
       app.userDbConnections.get(c).disconnect();
    }
   }
   app.userDbConnections.clear();// чистим массив прикладных соединений  
   }
        if (lc!=null) {
     try{      
     app.metaConnPool.SetFree(lc);
     } catch (Exception ex) {
         
     }
       }
 }
 }  
 app.luiSession=null;
}
// Изменение имени пользователя
public void setUserNam(String pUserName){
 if (app.luiSession==null) return;
  String stmt="";
  LuiDbConn lc=null;
  
  try {
 lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());  
 } catch(Exception ex) {
     return;
 }
 if (lc!=null) {
   try {
       stmt="update lui_t_session set user_name='"+pUserName+"' where id="+id;
       lc.executeDML(stmt); 
   }  catch (SQLException ex){
     //app.msgWriter.write("LUI-",ex.getMessage(),stmt);
 } finally {
       if (lc!=null) {
     try{      
     app.metaConnPool.SetFree(lc);
     } catch (Exception ex) {
         
     }
       }
 }
 }
}
// Изменение свойств сеанса
public void setSessionProp(String pProp,String pVal) {
  if (app.luiSession==null) return; 
  String stmt="";
  LuiDbConn lc=null;
  try {
 lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());  
 } catch(Exception ex) {
     return;
 }
  if (lc!=null) {
   try {
       String curprops=lc.getSelectResult("select session_property from lui_t_session where id="+id);
       LuiCollection curpropscol=new LuiCollection(NVL(curprops,""));
       curpropscol.Add_Element(pProp, pVal);
       stmt="update lui_t_session set session_property='"+curpropscol.toString()+"' where id="+id;
       lc.executeDML(stmt); 
   }  catch (SQLException ex){
     //app.msgWriter.write("LUI-",ex.getMessage(),stmt);
 } finally {
       if (lc!=null) {
     try{      
     app.metaConnPool.SetFree(lc);
     } catch (Exception ex) {
         
     }
       }
 }
 }
}
}
