/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;

import LuiCollection.LuiCollection;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;
import LuiExceptionMsg.LuiException;
import LuiNLS.LuiNLS;
import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;

import java.util.Set;
import org.apache.commons.io.FileUtils;

/**
 *
 * @author aliubushkin
 */
public class LuiSessionDispatcher extends Thread {

    private final Set<LuiSeance> Seances; // = new HashSet<>();
    private final LuiSrv mSrv;
    private final LuiDbConnPool metaConnPool;
    private LuiNLS NLS;
    private long lastdelsestime=0;
    private String appFilePath;
    int sessDispatcherTimeOut;
    int daysToKeepOldSessions;


    public LuiSessionDispatcher(Set<LuiSeance> pSeances, LuiSrv pSrv,LuiDbConnPool PmetaConnPool,LuiNLS pNLS,int pSessDispatcherTimeOut,int pDaysToKeepOldSessions,String pAppFilePath) {
        Seances = pSeances;
        mSrv = pSrv;
        metaConnPool=PmetaConnPool;
        NLS=pNLS;
        sessDispatcherTimeOut=pSessDispatcherTimeOut;
        daysToKeepOldSessions=pDaysToKeepOldSessions;
        appFilePath=pAppFilePath;
        setDaemon(true);
        setName("LUI Seances Dispatcher");
    }

    @Override
    public void run() {

        while (!mSrv.isShutdown()) {

            try {
                Thread.sleep(sessDispatcherTimeOut);
            } catch (InterruptedException ignore) {
            }

            // Avoiding Concurrent Modification
            ArrayList<LuiSeance> aSeances;
            synchronized (Seances) {
                aSeances = new ArrayList<>(Seances);
            }

            aSeances.stream().forEach((s) -> {
                try {
                    s.checkLuiSession("SessionDispatcher");
                } catch (SQLException ex) {
                } catch (LuiException ex) {
                }
            });
            aSeances.clear();
            if (daysToKeepOldSessions==0)continue; // уходим на ожидание  если не надо удалять устаревшие сеансы
            if (System.currentTimeMillis()-lastdelsestime<1000*60*60*24) continue; // уходим на ожидание проверка устаревших делалась менее 24часов назад
            lastdelsestime=System.currentTimeMillis();
            // чистка каталогов
            File rootDir= new File(appFilePath);
            if(rootDir.isDirectory()) {
                 if (rootDir.listFiles()!=null)
                 for(File AppItem : rootDir.listFiles()){ // крутим каталоги прикладных систем
                   if (AppItem!=null && AppItem.isDirectory()) { // если это каталог
                      if (AppItem.listFiles()!=null)
                      for (File UsrItem:AppItem.listFiles()){ // крутим каталоги внутри прикладной системы - т.е. юзера
                          if (UsrItem!=null && UsrItem.isDirectory()) { // если это каталог
                             for (File SesItem:UsrItem.listFiles()) {// крутим каталоги внутри юзера прикладной системы - т.е. сеансы
                                 if (SesItem!=null && SesItem.isDirectory()) {
                                   boolean isNum=false;
                                   try {
                                   String n=SesItem.getName();
                                   Integer.valueOf(n);
                                   isNum=true;
                                           } catch (Exception ex){}
                                   if (isNum) {
                                             long dirAge=System.currentTimeMillis()-SesItem.lastModified();
                                     if (dirAge>daysToKeepOldSessions*1000*60*60*24){
                                       try {  
                                         FileUtils.deleteDirectory(SesItem);
                                       } catch (IOException ioe){
                                       }
                                     }
                                   }
                                 }
                             }
                             // делаем попытку удаление каталога юзера он не удалиться если не пуст
                            try{
                            UsrItem.delete();
                            } catch(Exception ued){}
                          }
                      } 
                            boolean AppDirEmpty=true;
                            //проверим наличие в каталоге прикладной системы файлов отличных от verify.js
                            /* удаление папок приложения оказалось вредно
                            for (File UsrItem:AppItem.listFiles()){
                              if (!"verify.js".equals(UsrItem.getName())){
                                  AppDirEmpty=false;
                                  break;
                              }  
                            }
                            if (AppDirEmpty){
                            // делаем попытку удаление каталога приложения
                            try{
                            FileUtils.deleteDirectory(AppItem);
                            } catch(Exception aed){} 
                            
                            } */
                   }  
                 }
            }
            synchronized (Seances) {
                aSeances = new ArrayList<>(Seances);
            }
            LuiDbConn lc=null;
            try {
                lc=metaConnPool.GetFree(NLS.currentLang.locale,NLS.currentLang.code," Async tasks dispatcher", null, null);
                ArrayList<LuiCollection> oldSes=lc.getSelectAllRowsValMap("select id,app_code,user_name,end_datetime,start_datetime from lui_t_session ss where ss.start_datetime<(current_timestamp-'"+String.valueOf(this.daysToKeepOldSessions)+" day'::interval)\n" +
                "/*and ss.lui_server='"+mSrv.LuiSrvName+"'*/");
                    if (oldSes==null) continue;
                for (int i=0;i<oldSes.size();i++){ // для каждого устаревшего сеанса проверяем его наличие в нашем JAVA-сервере
                   boolean alive=false; 
                   long lsid=Long.valueOf(oldSes.get(i).Get_Element("id"));
                   for(int si=0;si<aSeances.size();si++)    {
                       if (aSeances.get(si).LuiSessionID==lsid){
                           alive=true;
                           break;
                       }
                   }
                   if (!alive){ // и если у нас его нет
                       /* удаляем его каталог -- это выше сделано поэтому коммент
                       String tmpDir=mSrv.AppFilePath //+System.getProperty("file.separator")
                               +oldSes.get(i).Get_Element("app_code")
                               +System.getProperty("file.separator")
                               +oldSes.get(i).Get_Element("user_name")
                               +System.getProperty("file.separator")
                               +oldSes.get(i).Get_Element("id")
                               +System.getProperty("file.separator");
                       File dir=new File(tmpDir);
                       if (dir.isDirectory()) {
                       try {
                       FileUtils.deleteDirectory(dir);
                       System.out.println("Outdated user temporary directiory "+tmpDir+" deleted.");
                       } catch (IOException ioe) {
                        
                       }
                       }*/
                      // удаляем его самого 
                      lc.executeDML("delete from lui_t_session where id="+oldSes.get(i).Get_Element("id"));
                      // и его протухшие группировки данных
                      lc.executeDML("delete from lui_t_data_grouping where is_stored='N' and session_id="+oldSes.get(i).Get_Element("id"));
                      System.out.println("Outdated user session "+oldSes.get(i).Get_Element("app_code")+"."+oldSes.get(i).Get_Element("id")+
                              "("+oldSes.get(i).Get_Element("start_datetime")+"-"+oldSes.get(i).Get_Element("end_datetime")
                              +") deleted.");
                      
                      
                   }        
                }
            } catch (SQLException ex) {
               
            } finally {if (lc!=null) {
               try { 
                metaConnPool.SetFree(lc);
               } catch (Exception ex){}
            }
                }
        aSeances.clear();
        }
    }
}
