/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/* 
 *   (С) 2016  ООО ФОРС-Центр Разработки, ООО ФОРС-Телеком
 *   http://www.fors.ru
 *   http://www.forstelecom.ru
 *
 *   Lui TCP - сервер
 *   
 *   РЕВИЗИЯ:
 *  $Rev:: 1            $:  Ревизия файла
 *  $Author:: aliubushkin   $:  Автор последних изменений
 *  $Date:: 2016-10-14 1#$:  Дата последних изменений
 *   ПРИМЕЧАНИЯ:
****************************************************************************** */
package luiserver;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import LuiDb.LuiDbConnectPool;
import LuiDbInterface.LuiDbConn;
import LuiNLS.LuiNLS;
import LuiDbInterface.LuiDbConnPool;
import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.URLClassLoader;

import java.nio.charset.Charset;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 *
 * @author aliubushkin
 */
public class LuiSrv {
    int MONLIM=5;

    public static boolean CHARTLICENSE=true;
    private LuiDbConnPool metaConnPool;    // пул соединений для доступа к мета
    private LuiDbConnPoolArray connPoolArray;   // общий массив пулов
    //public String listenAddr; // ip-адреса которые можно слушать
    //public int listenPort; // порт прослушивателя
    //public ServerSocket serverSocket; // сокет сервера
    //public ArrayList<LuiSeance> LuiSeanceArray;
    public final Set<LuiSeance> mSeances = new HashSet<>();
    //public volatile boolean shutdown;
    private final AtomicBoolean mShutdown = new AtomicBoolean(false);
    private final Set<ServerThread> mListeningThreads = new HashSet<>();
    
    
    /*
    String metaDBurl;
    String metaDBusername;
    String metaDBpassword;
    String metaDBschema;
    */
    // public ServerThread serverThread; // мой вложенный объект "серверный тред"
    // public Thread srvThread; //Java thred серверного треда
    public final LuiNLS NLS;
    public final HashMap<String, LuiAppForm> formTemplates;
    public final LuiCollection commonForms;
    public final HashMap<String, Class> stdPluginClasses; // множество стандартных плагинов
    public final HashMap<String, String> stdPluginIsInter; // признаки интерактивности плагинов
    public final String LuiSrvName;
    public final String defApp;
    public LuiTaskDispatcher taskDisp;
    public File HomeDir;
    //public LuiSessionDispatcher sesDisp;
    public LuiDataConverter LuiDatConv;
    LuiCrip crypter;    
    int taskDispatcherTimeOut=0;
    int sessDispatcherTimeOut=60000;
    int daysToKeepOldSessions=5;
    public String AppFilePath="";
    public LuiPerformance perfStat=new LuiPerformance(); 
    
        /**
     * Конструктор сервера LUI
     * @param pPool пул соединений с базой матеданных
     * @param pPluginsLocation каталог, в котором лежат плагины
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     * @throws MalformedURLException неправильный URL
     * @throws ClassNotFoundException  не найден класс
     */
    public LuiSrv(LuiDbConnectPool pPool, String pPluginsLocation) throws SQLException, LuiException, MalformedURLException, ClassNotFoundException {
        this(pPool,null, pPluginsLocation, null,null,null,null,null,null,null,null);
    }
   
    
    public boolean isShutdown() {
        return mShutdown.get();
    }
    
    /**
     * Конструктор сервера LUI
     * @param pPool пул соединений с базой матеданных
     * @param pDefApp приложение по умолчанию
     * @param pPluginsLocation каталог, в котором лежат плагины
     * @param pLUIParams устанавливаемые параметры приложения - lui.lui_t_application_property, app_code='LUI'
     * @param pTaskDispatcherTimeOut таймаут диспетчера задач
     * @param pSessDispatcherTimeOut таймаут диспетчера сеансов
     * @param pDaysToKeepOldSessions сколько дней хранить устаревшие сеансы
     * @param pLuiSrvName имя сервера LUI
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     * @throws MalformedURLException неправильный URL
     * @throws ClassNotFoundException  не найден класс
     */
    @SuppressWarnings("empty-statement")
    public LuiSrv(LuiDbConnectPool pPool,String pDefApp, String pPluginsLocation, Map<String, String> pLUIParams,String pTaskDispatcherTimeOut,String pSessDispatcherTimeOut,String pDaysToKeepOldSessions,String pLuiSrvName,String pLicMon, String pLicChart,File pHomeDir) throws SQLException, LuiException, MalformedURLException, ClassNotFoundException {
    HomeDir=pHomeDir;
        try{
        MONLIM=Integer.valueOf(NVL(pLicMon,"120"));
        }catch (Exception ex){
        MONLIM=120;    
        }
        if (pLicChart!=null && pLicChart.equals("N"))
        CHARTLICENSE=false;
        LuiSrvName=pLuiSrvName;
        // Пулы ненужных соединений
        connPoolArray = new LuiDbConnPoolArray();
        connPoolArray.addPool(pPool);
        metaConnPool = pPool;
        defApp=NVL(pDefApp,"LUI");
        crypter=new LuiCrip(new byte[]{1,9,6,2,0,2,2,5});         
        try{
        taskDispatcherTimeOut=Integer.valueOf(pTaskDispatcherTimeOut);
        } catch (Exception ex){
        taskDispatcherTimeOut=0;    
        }
        try{
        sessDispatcherTimeOut=Integer.valueOf(pSessDispatcherTimeOut);
        } catch (Exception ex){
        sessDispatcherTimeOut=60000;    
        }
        try{
        daysToKeepOldSessions=Integer.valueOf(pDaysToKeepOldSessions);
        } catch (Exception ex){
        daysToKeepOldSessions=5;    
        }
        LuiDbConn aConn = pPool.GetFree(null,null,null,null,null);
        try {
            
            NLS = new LuiNLS(aConn, "LUI");
            aConn.setLocale(NLS.currentLang.locale);
            LuiMsgWriter aMsgWriter = new LuiMsgWriter(aConn, NLS, "LUI", System.err, System.out);
            if (HomeDir!=null)
            System.out.println("LUI Home directory is "+HomeDir.toString());
            // Загрузка типов данных
            System.out.print("Loading LUI datatypes...");
            try{
                LuiDatConv=new LuiDataConverter("LUI",aConn);
            } catch (Exception ex){
                System.out.println(" Error:"+ex.toString());
                throw ex;
            }
            System.out.println(" OK");
            
            // Загрузка списка общих форм
            commonForms = aConn.getSelectAllRowsTwoColsValMap("select code,formtype_code from lui_t_form where formgroup_code='LUI_COMMON' and version_number is null");
            formTemplates = new HashMap<>();
            LuiLoadFormTemplate.Load(aConn, NLS, aMsgWriter, formTemplates);

            // Загрузка плагинов
            stdPluginClasses = new HashMap<>();
            stdPluginIsInter = new HashMap<>();
            ArrayList<ArrayList<String>> plist =new ArrayList<>();
            if (pLUIParams!=null)
            plist = aConn.getSelectAllRowsValList( "select code,name,notes,filename,classname,is_interactive from lui_t_plugin where app_code='LUI' and ( plug = 'Y' or code in (" + pLUIParams.get("includePlugin") + ") ) and code not in (" + pLUIParams.get("excludePlugin") + ")" );
            else
            plist = aConn.getSelectAllRowsValList( "select code,name,notes,filename,classname,is_interactive from lui_t_plugin where app_code='LUI' and  plug = 'Y' " );    
            // соберём все URL'ы
            HashSet<URL> aURLSet = new HashSet<>();
            for (ArrayList<String> aRow : plist) {
                String aPath = aRow.get(3);
                if(pPluginsLocation != null) { 
                    aPath = pPluginsLocation + "/" + aPath;
                    
                }
                aURLSet.add(new File(aPath).toURI().toURL());
            }
            
            aURLSet.forEach((aURL) -> {
                System.out.printf("\tPlugin library: %s\n", aURL.toString());
            });
            
            URL[] aURLList = new URL[aURLSet.size()];
            URLClassLoader aPluginsLoader = new URLClassLoader(aURLSet.toArray(aURLList), this.getClass().getClassLoader());
            
            for (ArrayList<String> aRow : plist) {
                String aPluginCode = aRow.get(0);
                String aClassName = aRow.get(4);
                System.out.printf("\tPlugin: %s: %s ", aPluginCode, aClassName);
                Class<?> aPlugin = aPluginsLoader.loadClass(aClassName);
                System.out.println("OK");
                stdPluginClasses.put(aPluginCode, aPlugin);
                stdPluginIsInter.put(aPluginCode,aRow.get(5));
            }
            // параметры LUI-системы
            if(pLUIParams != null) {
                Connection aJConn = aConn.getConnection();
                try( CallableStatement aStmt = aJConn.prepareCall(
                        "UPDATE  lui_t_application_property\n" +
                                "   SET  VALUE=?\n"+
                                " WHERE  APP_CODE='LUI'\n"+
                                "        AND CODE=?\n"
                )) {
                    for(Map.Entry<String, String> aPair: pLUIParams.entrySet()) {
                        aStmt.setString(1, aPair.getValue() );
                        aStmt.setString(2, aPair.getKey());
                        aStmt.execute();
                    }
                }
                if(!aJConn.getAutoCommit()) {
                    aJConn.commit();
                }
            if (pLUIParams.containsKey("APP_FILES_PATH")) {
                AppFilePath=pLUIParams.get("APP_FILES_PATH");
            }
            }
            // диспетчер задач
       if (taskDispatcherTimeOut>0){     
        taskDisp=new LuiTaskDispatcher(this.metaConnPool,NLS,this.crypter,connPoolArray,this.stdPluginClasses,this.stdPluginIsInter,this.LuiSrvName,this.taskDispatcherTimeOut,this.commonForms,this.formTemplates,this.AppFilePath,this.LuiDatConv);
        taskDisp.thread = new Thread(taskDisp);
        taskDisp.thread.setDaemon(true);
        taskDisp.thread.setName("LUI async task dispatcher");
        taskDisp.thread.start();
       }
          // диспетчер сеансов для периодической проверки сеанса
          // это на тос случай если в БД запись о Lui-сеансе удалили а пользователь ни куда не тыкает
          // и тогда экземпляр его приложения занимает ресурсы в JAVA-машине
         if (sessDispatcherTimeOut>0){     
         new LuiSessionDispatcher(this.mSeances, this,metaConnPool,NLS,sessDispatcherTimeOut,daysToKeepOldSessions,AppFilePath).start();
         }
        } 
        
        
        catch(Throwable t) {
            connPoolArray.removeAllPools();
            throw t;
        }
        finally {
            pPool.SetFree(aConn);
        }
    }
    
    
    
    /**
     * Останов сервера
     */
    public void shutdown() {
    //public void Stop() {
        //прекращаем создавать новые соединения
        //shutdown = true;
        mShutdown.set(true);
        connPoolArray.setShutdown();
        /* try {
            serverSocket.close();
        } catch (IOException ex) {
        }
        */
        
        synchronized( mListeningThreads ) {
            mListeningThreads.stream().forEach( (t) -> t.shutdown() );
        }
        
        // гасим все треды сеансов
        // Прадварительное гашение - расстановка флагов shitdown, посылка сигналов interrupt
        synchronized(mSeances) {
            mSeances.stream().forEach( (s) -> s.preStop());
        }
        
        try {
            Thread.sleep(500);
        }
        catch(InterruptedException ignore) {}
        
        // жесткое гашение сеансов
        synchronized(mSeances) {
            mSeances.stream().forEach( (s) -> s.stop());
        }
        // гасим диспетчера задач
        //taskDisp.thread.interrupt();
        // гасим коннекты БД
        connPoolArray.removeAllPools();
    }

    /*
    public void Start() {
        serverThread = new ServerThread(); // Создание треда слушателя для новых подключений
        srvThread = new Thread(serverThread); // Запуск этого треда
        srvThread.start(); // врезультате этого выполняется run() помеченный ***
    }
    */
    
    /**
     * Запуск слушателя входящих TCP-соединений
     * @param pAddress адрес, на котором слушать входящие соединения, может быть null
     * @param pPort номер порта TCP для входящих соединений
     * @throws IOException ошибка ввода-вывода
     */
    public void startTCPListener( String pAddress, int pPort ) throws IOException {
        
        System.out.printf("Listening on %s:%s\n", pAddress, pPort);
        ServerThread aListener = new ServerThread(pAddress, pPort);
        synchronized(mListeningThreads) {
            aListener.start();
            mListeningThreads.add(aListener);
        }
    }
    
    
    public LuiSeance createSeance( EventChannel pChannel )
    {
        LuiSeance mSeance = new LuiSeance( this, pChannel , connPoolArray, metaConnPool, NLS, formTemplates, stdPluginClasses,stdPluginIsInter, commonForms,crypter,LuiSrvName,AppFilePath,LuiDatConv,MONLIM);  // создаём новый тред для обслуживания сокета кнового клиента
        
        mSeance.thread = new Thread(mSeance);
        
        mSeance.thread.setDaemon(true);
        mSeance.thread.setName(String.format("LUI Seance %d", mSeance.getSeanceId()));
        mSeance.thread.start();
        synchronized(mSeances) {
            mSeances.add(mSeance);
        }
        return mSeance;
    }
    

    /**
     * Удаление сеанса из списка активных.
     * @param pSeance удаляемый сеанса
     */
    protected void removeSeance( LuiSeance pSeance )
    {
        synchronized(mSeances) {
            mSeances.remove(pSeance);
        }
    }
    
    
    /**
     * Вложенный класс для прослушивания новых подключений
     */
    public class ServerThread extends Thread {
        
        private final String mListenAddr;
        private final int    mListenPort;
        private final ServerSocket mServerSocket;
        private final AtomicBoolean mShutdown = new AtomicBoolean(false);
        
        /**
         * Конструктор серверного слушателя. Создаёт серверный сокет.
         * @param pListenAddr адрес, на котором слушать (может быть null)
         * @param pListenPort номер порта TCP
         * @throws IOException ошибка ввода-вывода
         */
        public ServerThread( String pListenAddr, int pListenPort) throws IOException {
            mListenAddr = pListenAddr;
            mListenPort = pListenPort;
            
            ServerSocket aServer;
            
            if(pListenAddr == null || pListenAddr.isEmpty()) {
                aServer = new ServerSocket(pListenPort, 10);
            }
            else {
                aServer = new ServerSocket(pListenPort, 10, InetAddress.getByName(pListenAddr));
            }
            
            mServerSocket = aServer;
            mServerSocket.setSoTimeout(10000);
                    
            setName(String.format("LUI Listener %s:%d", mListenAddr, mListenPort ));
            setDaemon(true);
        }
        
        /**
         * Завершить приём соединений
         */
        public void shutdown() {
            mShutdown.set(true);
            
            try {
                mServerSocket.close();
            }
            catch(Throwable ignore) {}
        }

        @Override
        public void run() { //***
          try {  
            //Thread.currentThread().setName(String.format("LUI Listener %s:%d", mListenAddr, mListenPort ));
            
            /*
            try {
                serverSocket = new ServerSocket(); // Создание серверно сокета для новых подключений
                serverSocket.bind(new InetSocketAddress(listenAddr, listenPort)); // Указываем ему что надо слушать
                System.out.println("LUI: - \"Now I'm listening to " + serverSocket.getLocalSocketAddress() + "\"");
                System.out.println("LUI: - \"Press <Ctrl+C> or enter \"end\" to break  ....");
            } catch (Exception ex) {
                System.out.println("\nServer socket open error:");
                ex.printStackTrace();
                return;
            }
            */ 
            //while (!shutdown) { // цикл ожидания подключения нового клиента
            while(!mShutdown.get()) {
                // очистим дохлые сеансы
                /*
                boolean b = true;
                while (b) {
                    b = false;
                    
                    for (int i = 0; i < LuiSeanceArray.size(); i++) {
                        LuiSeance s = LuiSeanceArray.get(i);
                        if (s.thread != null) {
                            if (!s.thread.isAlive()) {
                                LuiSeanceArray.remove(i);
                                b = true;
                            }
                        } else {
                            LuiSeanceArray.remove(i);
                            b = true;
                        }
                    }
                    
                } // конец чистки*/
                synchronized( mSeances ) {
                        mSeances.removeIf( (s) ->  s.thread == null || !s.thread.isAlive());
                }
                
                try {
                    //clientSocket = serverSocket.accept(); // создаёт под нового клиента новый сокет
                    Socket clientSocket = mServerSocket.accept();
                    clientSocket.setKeepAlive(true);
                    clientSocket.setSoTimeout(0);
                    clientSocket.setSoLinger(false, MONLIM);
                    //createSeance( new SocketChannel(clientSocket, Charset.forName("Windows-1251")));
                    createSeance( new SocketChannel(clientSocket, Charset.forName("UTF8")));
                    
                    /*
                    LuiSeance seance = new LuiSeance( new SocketChannel(clientSocket, Charset.forName("Windows-1251")), connPoolArray, metaConnPool, NLS, formTemplates, stdPluginClasses, commonForms);  // создаём новый тред для обслуживания сокета кнового клиента
                    Thread t = new Thread(seance);
                    t.start(); // запуск треда для нового LUI-сеанса
                    seance.thread = t;
                    //LuiSeanceArray.add(seance); // записываем сеанс в массив сеансов
                    synchronized(mSeances) {
                        mSeances.add(seance);
                    }*/
                } 
                catch( SocketTimeoutException ignore) {}
                catch (SocketException ex) {
                    if (!ex.getMessage().equals("socket closed")) {
                        ex.printStackTrace();
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();

                }
            }
            System.out.println("Stop LUI-Server...");
          }
          finally {
            synchronized( mListeningThreads ) {
                mListeningThreads.remove(this);
            }
          }
            
        }

    }

// ---------------------------------- //
    
    public Set<LuiSeance> getSeances() {
        return mSeances;
    }
    public boolean getChartLicense(){
     return CHARTLICENSE;   
    }
}
