/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Выполнение задач
 */
package luiserver;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.sql.SQLException;
import java.util.ArrayList;


/**
 *
 * @author aliubushkin
 */
public class LuiTask{
    LuiApplication app;
    LuiAppForm frm;
    LuiAppProp pB;
    LuiAppProp pQ;
    LuiAppProp pR;
    LuiAppProp pRex;
    LuiAppProp pA;
    LuiAppProp pPBtit;
    LuiAppProp pPBpct;
    LuiAppProp pPBwrk;
    long startTime;
    long showFreq=500;
    long lastShowTime=0;
    String action;
    int total=0;
    int curcnt=0;
    int logsize=100;
    boolean pbResponce;
    String cancelable;
public LuiTask(LuiAppForm pFrm) {
    frm=pFrm;
    app=frm.application;
    startTime=System.currentTimeMillis();
}
private void do_action(LuiAppProp p) throws LuiException, SQLException{
    app.parser.execute(action, p);
}
/**
 * Это показывалка прогресса
 * @param whatToDo
 * @return
 * @throws LuiException
 * @throws SQLException 
 */
private boolean pb(String whatToDo) throws LuiException, SQLException{
    String title=null;
    String pct=null;
    String wrk=null;
    long timeElpSec=(System.currentTimeMillis()- startTime)/1000;
    long min=timeElpSec/60;
    long ss=timeElpSec-min*60;
    long hh=min/60;
    min=min-hh*60;
    int sesupd=0; 
    frm.setProperty("ProcessingElapsedTime",String.format("%02d",hh)+":"+String.format("%02d",min)+":"+String.format("%02d",ss));
    if (pPBtit!=null) {
      title=pPBtit.getComputedProperty();
    }
    title=NVL(title,"");
    switch (whatToDo){
        case "ActionBefore":
            //pct="0/0";
            pct=pPBpct.getComputedProperty();
            wrk=app.NLS.NLS("RUS:Действе в начале;ENG:Action before;");
            break;
        case "CountQweryRows":
            //pct="0/0";
            pct=pPBpct.getComputedProperty();
            wrk=app.NLS.NLS("RUS:Подсчёт строк;ENG:Count rows;");
            break;
        case "Processing":
     if (pPBpct!=null){
      pct=pPBpct.getComputedProperty();
      } else pct=null;
     pct=NVL(pct,String.valueOf(curcnt)+"/"+String.valueOf(total));
     if (pPBwrk!=null){
      wrk=pPBwrk.getComputedProperty();
    } else wrk=null;
      wrk=NVL(wrk,app.NLS.NLS("RUS:Обработка;ENG:Processing;"));
     break;
        case "ActionAfter":
            //pct=String.valueOf(total)+"/"+String.valueOf(total);
            pct=pPBpct.getComputedProperty();
            wrk=app.NLS.NLS("RUS:Действе в конце;ENG:Action after;");
     break;
        case "Complete":
            pct=pPBpct.getComputedProperty();
            wrk=app.NLS.NLS("RUS:Завершение процесса;ENG:Process complete;");
     break;
    }
  if (app.luiSeance!=null){  
    if ("closed".equals(app.progressBar.status)){
    pbResponce=app.progressBar.startProgress(title, cancelable, pct, wrk,total, curcnt);
    } else {
     if   (whatToDo.equals("Complete")) {
        app.progressBar.stopProgress();
     } else
    pbResponce=app.progressBar.continueProgress(title, cancelable, pct, wrk, total, curcnt);        
    }
  if (!pbResponce){
      app.progressBar.stopProgress();
      app.actionBreak();
  }
  return pbResponce;
  } else { // если задача асинхронная, то вместо прогресс-бара будем обновлять данные в сессии задачи если ничего не обновилось - значит сессию убили
           // кроме того узнать о судьбе задачи можно исследуя её текущее состояние
  pbResponce=true;
  LuiDbConn lc=null;
  try {
   lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
   /* слишком частое обновление вредно
     if (lastShowTime>0 && curcnt<total){ //если continueProgress не в первый раз и это не финал
     if (showFreq>(System.currentTimeMillis()-lastShowTime)) // и время лчередного показа ещё не пришло
         return true;
         }
  lastShowTime=System.currentTimeMillis();
  */
   // готовим коллекцию атрибутов
   LuiCollection bpCol=new LuiCollection("");
   bpCol.Add_Element("TASK_EXE_TIME", String.valueOf(timeElpSec));
   bpCol.Add_Element("TASK_ROWS_TOTAL", String.valueOf(total));
   bpCol.Add_Element("TASK_ROWS_CURRENT", String.valueOf(curcnt));
   bpCol.Add_Element("TASK_TITLE_TEXT", title);
   bpCol.Add_Element("TASK_PCT_TEXT", pct);
   bpCol.Add_Element("TASK_WRK_TEXT", wrk);
   String termMode=lc.getSelectResult("select terminate_mode from lui_v_task where app_code='"+app.code+"' and code='"+frm.code+"'");
   if (termMode==null) {
       pbResponce=false; // задачи нет уже - удалена
   } 
   if (is_not_null(termMode)) { // Задачу хотят остановить 
       if (termMode.equals("BREAK") || termMode.equals("ABORT")) { // причём немедленно
           pbResponce=false;
        }        
   }
      if (whatToDo.equals("Complete")){
         bpCol.Add_Element("TASK_COMPLETE_CAUSE","NORMAL"); 
      } 
   
   sesupd=lc.executeDML("update lui_t_session set session_property='"+bpCol.toString()+"' where end_datetime is null and id="+app.luiSession.id.toString());
   if (sesupd<1) {
     // если ни одной записи не обновлено
     // то вероятно, что сеанс уже исчез или имеет не пустое время завершения
     // это бывает при shutown сервера или...
     // такую задачу будем считать аварийно завершённой и она должна перейти в wait
     bpCol.Add_Element("TASK_COMPLETE_CAUSE","BREAK");
     sesupd=lc.executeDML("update lui_t_session set session_property='"+bpCol.toString()+"' where id="+app.luiSession.id.toString());
     pbResponce=false;
   }
  } finally {
      if (lc!=null) app.metaConnPool.SetFree(lc);
  }
  return pbResponce;
  }
}
/**
 * выполнение обработки строк запроса
 * Имя свойства задаётся в формате:  &lt;ID формы&gt;.&lt;Код элемента или "Form"&gt;.&lt;Код свойства&gt;
 * @param pBefore -- действие перед выполнением запроса
 * @param pQuery  -- запрос
 * @param pRow -- действие для строки
 * @param pRowEx -- обработка ошибок строки
 * @param pAfter -- действие после
 * @param pCancelable -- Y- возможно прерывание процесса
 * @param pPBtitle -- заголовок прогресс-бара
 * @param pPercentText -- текст на прогресс-баре
 * @param pWorkText -- ещё какой-то текст
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 * 
 */
public void procQueryRows(String pBefore,String pQuery, String pRow,String pRowEx, String pAfter, String pCancelable, String pPBtitle, String pPercentText, String pWorkText) throws LuiException, SQLException{
    pB=app.findProperty(pBefore);
    pQ=app.findProperty(pQuery);
    pR=app.findProperty(pRow);
    pRex=app.findProperty(pRowEx);
    pA=app.findProperty(pAfter);
    pPBtit=app.findProperty(pPBtitle);
    pPBpct=app.findProperty(pPercentText);
    pPBwrk=app.findProperty(pWorkText);
    cancelable=NVL(pCancelable,"Y");
    String charset="UTF-8";
    String filename="";
    total=0;
    curcnt=0;
    boolean fileMode=false;
    File file = null;
    int rowActionLogged=0;
    if (app.luiSeance==null){
    frm.setProperty("TaskMode","Async");
    } else {
    frm.setProperty("TaskMode","Sync");    
    }
    frm.setProperty("ProcessingElapsedTime","00:00:00");
    frm.setProperty("TotalRowNumber","0");
    frm.setProperty("ProcessingRowNumber","0");
    frm.setProperty("ProcessingPercent", "0");
    try{
      String lsz=frm.elements.get("P_LOGSIZE").getProperty("Value");
      int sz=Integer.valueOf(lsz);
      if (sz>0) logsize=sz;
    } catch (Exception ex){}
    // логирование параметров
    LuiCollection pCol=new LuiCollection("");
     for (int i=0;i<frm.elements.size();i++){
         LuiAppElement e=frm.elSort(i);
         if (e.getType().equals("PARAM")){
             if (e.getCode().equals("P_LOGSIZE")){
                 String lsz=e.getProperty("Value");
                 int sz=Integer.valueOf(lsz);
                 if (sz>0) logsize=sz;
             }
         pCol.Add_Element(e.getCode(),e.getProperty("Value"));
         }
     }
             app.log.addRecord("TASK_STARTING",
             app.NLS.NLS("RUS:Запуск задачи;ENG:Starting task;"),
             pCol.toString(),frm.code);
             
    // конец логирования параметров
    if (pB!=null){ // действие перед запросом
     action=pB.getComputedProperty();
     if (is_not_null(action)) {
    if (!pb("ActionBefore")) return;
    app.log.addRecord("TASK_ACTION_BEFORE",
    app.NLS.NLS("RUS:Выполнение программы в начале задачи;ENG:Execute programm before task;"), 
    action, 
    frm.code);
    do_action(pB);
     }
    }
    if (pQ!=null){ // если задано свойство для запроса
     String stmt=pQ.getComputedProperty();
     if (is_not_null(stmt)){
     fileMode=false;
         if (like(stmt,"FILE:%")){
         if (stmt.charAt(stmt.length()-1)!=';'){
            stmt=stmt+";";
        }
     LuiCollection fileCol=new LuiCollection(stmt);
     filename=fileCol.Get_Element("FILE");
     charset=NVL(fileCol.Get_Element("CHARSET"),"UTF-8");
     file=new File(filename);
     fileMode=file.exists();
     }
     if (fileMode) { // file
     try {    
          if (!pb("CountQweryRows")) return;
          app.log.addRecord("TASK_COUNT_ROWS",
    app.NLS.NLS("RUS:Подсчёт количества строк в файлк;ENG:Count file rows;"), 
    stmt, 
    frm.code);
          total=0;
          FileInputStream inF=new FileInputStream(file);
          InputStreamReader iSr=new InputStreamReader(inF,charset);
          BufferedReader reader = new BufferedReader(iSr);
          String fileRow;
          while (true){
              fileRow=reader.readLine();
              if (fileRow==null) {reader.close();
              //fr.close();
              iSr.close();
              inF.close();
              break;}
              total++;
          }
          frm.setProperty("TotalRowNumber",String.valueOf(total));
          // читаем файл
          //fr=new FileReader(file);
          inF=new FileInputStream(file);
          iSr=new InputStreamReader(inF,charset);
          reader = new BufferedReader(iSr);
          while (true){
              fileRow=reader.readLine();
              if (fileRow==null) {reader.close();
              //fr.close();
              iSr.close();
              inF.close();
              break;}
               curcnt++;
       frm.setProperty("ProcessingRowNumber",String.valueOf(curcnt));
        frm.setProperty("ProcessingPercent",String.valueOf(100*curcnt/total));
        if (!pb("Processing")) return;
              frm.setProperty("FileRecord",fileRow);
               if (pR!=null){
       //pR.clearComputedValue(false);
       action=pR.getComputedProperty();
       if (is_not_null(action) && (rowActionLogged<=logsize || curcnt==total)) {
           app.log.addRecord("TASK_ACTION_ROW",
           app.NLS.NLS("RUS:Выполнение программы для строки запроса "+String.valueOf(curcnt)+";ENG:Execute programm for query row "+String.valueOf(curcnt)+";"), 
           action, 
           frm.code);
           rowActionLogged++;
                                               }
       try {
        frm.setProperty("TASK_ERROR_TEXT","");   
        frm.setProperty("TASK_ERROR_CODE","");   
        do_action(pR);       //выполнение действия
       } catch (LuiException ex){
         if (pRex==null || is_null(pRex.getComputedProperty())) throw ex;
         else {
             frm.setProperty("TASK_ERROR_TEXT",ex.getMessage().replace("'", "''"));
             frm.setProperty("TASK_ERROR_CODE",ex.getCode());
             app.parser.execute(pRex.getComputedProperty(), pRex);
         }
       } 
       catch (SQLException ex){
         if (pRex==null || is_null(pRex.getComputedProperty())) throw ex;
         else {
             frm.setProperty("TASK_ERROR_TEXT",ex.getMessage().replace("'", "''"));
             frm.setProperty("TASK_ERROR_CODE",ex.getSQLState());
             app.parser.execute(pRex.getComputedProperty(), pRex);
         }
       }
       catch (Exception ex){
         if (pRex==null || is_null(pRex.getComputedProperty())) throw ex;
         else {
             frm.setProperty("TASK_ERROR_TEXT",ex.getMessage().replace("'", "''"));
             frm.setProperty("TASK_ERROR_CODE",ex.getClass().getCanonicalName());
             app.parser.execute(pRex.getComputedProperty(), pRex);
         }
       }
       
                            }
          }
          // конец файла
     }   catch (IOException ex) {  
         app.msgWriter.write("LUI-00109",filename,frm.code,ex.getMessage());
         }  
     } else { // select
     ArrayList<String> acstmt=app.getAppConnAndStmt(stmt,"",frm.getFormID());         
     LuiAppConn ac=app.userDbConnections.get(acstmt.get(0));
     LuiDbConn lc=null;
     Integer cursor=null;
     try {
        if (ac.getSessionMode()=='C'){
         lc=ac.connPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
         }
        else lc=ac.currentConnection;
        if (lc==null)app.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
        //подсчитываем записи
        if (!pb("CountQweryRows")) return;
        String qstmt="select count(*) from ("+stmt+")q$count";
    app.log.addRecord("TASK_COUNT_ROWS",
    app.NLS.NLS("RUS:Подсчёт количества строк запроса;ENG:Count query rows;"), 
    qstmt, 
    frm.code);
        total=Integer.valueOf(lc.getSelectResult(qstmt));
        frm.setProperty("TotalRowNumber",String.valueOf(total));
        //открываем курсор
        cursor=lc.CursorOpen(stmt);
        int colcnt=lc.CursorGetCols(cursor);
        ArrayList<ColInfo> ColsInfo=lc.CursorGetColsInfo(cursor);
        while(lc.CursorNextRow(cursor)){ // крутим строки
        curcnt++;
      /* это для замедленного выполнения  
            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) {
                Logger.getLogger(LuiTask.class.getName()).log(Level.SEVERE, null, ex);
            }
        */
        frm.setProperty("ProcessingRowNumber",String.valueOf(curcnt));
        frm.setProperty("ProcessingPercent",String.valueOf(100*curcnt/total));
        if (!pb("Processing")) return;
        ArrayList<String> rowfields=lc.CursorGetRow(cursor);
        for (int i=0;i<colcnt;i++){
            frm.setProperty("QC "+ColsInfo.get(i).alias, rowfields.get(i));
        }
      if (pR!=null){
       //pR.clearComputedValue(false);
       action=pR.getComputedProperty();
       if (is_not_null(action) && (rowActionLogged<=logsize || curcnt==total)) {
           app.log.addRecord("TASK_ACTION_ROW",
           app.NLS.NLS("RUS:Выполнение программы для строки запроса "+String.valueOf(curcnt)+";ENG:Execute programm for query row "+String.valueOf(curcnt)+";"), 
           action, 
           frm.code);
           rowActionLogged++;
       }
        try {
        frm.setProperty("TASK_ERROR_TEXT","");   
        frm.setProperty("TASK_ERROR_CODE","");   
            do_action(pR);       //выполнение действия
       } catch (LuiException ex){
         if (pRex == null || is_null(pRex.getComputedProperty())) throw ex;
         else {
             frm.setProperty("TASK_ERROR_TEXT",ex.getMessage().replace("'", "''"));
             frm.setProperty("TASK_ERROR_CODE",ex.getCode());
             app.parser.execute(pRex.getComputedProperty(), pRex);
         }
       } 
       catch (SQLException ex){
         if (pRex == null || is_null(pRex.getComputedProperty())) throw ex;
         else {
             frm.setProperty("TASK_ERROR_TEXT",ex.getMessage().replace("'", "''"));
             frm.setProperty("TASK_ERROR_CODE",ex.getSQLState());
             app.parser.execute(pRex.getComputedProperty(), pRex);
         }
       }
       catch (Exception ex){
         if (pRex == null || is_null(pRex.getComputedProperty())) throw ex;
         else {
             frm.setProperty("TASK_ERROR_TEXT",ex.getMessage().replace("'", "''"));
             frm.setProperty("TASK_ERROR_CODE",ex.getClass().getCanonicalName());
             app.parser.execute(pRex.getComputedProperty(), pRex);
         }
       }

      }
        } // конец строк курсора
     } finally {
        if (lc!=null) {
          if (cursor!=null) lc.CursorClose(cursor);
          if (ac.getSessionMode()=='C'){
              ac.connPool.SetFree(lc);
          }
        } 
     }
     }// select 
    //} // запрос не пуст
    } // проперть с запросом не пусто 
if (pA!=null){ // действие после запроса
     action=pA.getComputedProperty();
     if (is_not_null(action)) {
     pb("ActionAfter");
     app.log.addRecord("TASK_ACTION_AFTER",
    app.NLS.NLS("RUS:Выполнение программы в конце задачи;ENG:Execute programm after task;"), 
    action, 
    frm.code);
     do_action(pA);
     }
    }

pb("Complete");
    
}

}
}

