/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Асинхронное выполнение задач
 */
package luiserver;

import LuiCollection.LuiCollection;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.is_not_null;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiDbInterface.LuiDbConnPool;
import LuiExceptionMsg.LuiException;
import LuiExceptionMsg.LuiMsgWriter;
import LuiNLS.LuiNLS;
import java.sql.SQLException;
import java.util.HashMap;

/**
 *
 * @author aliubushkin
 */
public class LuiTaskAsync implements Runnable{
public LuiDataConverter LuiDatConv;
public Thread thread; // ссылка на работающий 
LuiCollection taskAttr;
public LuiApplication app;
String LuiSrvName;
public LuiAppForm frm;
LuiNLS NLS;
LuiDbConnPool MetaConnPool;
LuiMsgWriter MsgWriter;
LuiCrip crip;
LuiDbConnPoolArray connPoolArray;
HashMap<String, Class> stdPluginClasses; // множество стандартных плагинов
HashMap<String, String> stdPluginIsInter; // признаки интерактивности плагинов
LuiProcTimer procTimer;
HashMap<String, LuiAppForm> formTemplates;
LuiCollection commonForms;
String AppFilePath;

public LuiTaskAsync(LuiCollection pTaskAttr,LuiNLS pNLS,LuiDbConnPool pMetaConnPool,
        LuiCrip pCrip,LuiDbConnPoolArray pConnPoolArray
,HashMap<String, Class>PstdPluginClasses,HashMap<String, String> PstdPluginIsInter,String pLuiSrvName,LuiCollection pCommonForms,HashMap<String, LuiAppForm> pFormTemplates,String pAppFilePath,LuiDataConverter pLuiDatConv){
    taskAttr=pTaskAttr;
    LuiDatConv=pLuiDatConv;
    NLS=pNLS;
    LuiSrvName=pLuiSrvName;
     MetaConnPool=pMetaConnPool;
     MsgWriter = new LuiMsgWriter(MetaConnPool, NLS, "LUI");// создаём плевателя сообщений
     crip=pCrip;
     connPoolArray=pConnPoolArray;
     stdPluginClasses=PstdPluginClasses;
     stdPluginIsInter=PstdPluginIsInter;
     procTimer = new LuiProcTimer();
     commonForms=pCommonForms;
     formTemplates=pFormTemplates;
     AppFilePath=pAppFilePath;
}
    @Override
    public void run() {
    try {
    
        //commandWriter = new LuiClCommandWriter(mChannel.writer(), procTimer);//создаём писателя команд
        //eventReader = new LuiClEventReader(mChannel.reader(), msgWriter, procTimer); //создаём читателя событий    
    //создаём приложение
    app=new LuiApplication(null,taskAttr.Get_Element("app_code"),NLS.currentLang.code,MetaConnPool,
        MsgWriter
        ,connPoolArray
        ,NLS
        ,formTemplates
        ,null //LuiClCommandWriter pCommandWriter,
        ,this.stdPluginClasses
        ,this.stdPluginIsInter
        ,null //LuiClEventReader pEventReader,
        ,this.procTimer
        ,commonForms
        ,null //String pHostName
        ,crip //LuiCrip pCrypter
        ,null //String pClientID,
        ,null //String pMsgPort
        ,LuiSrvName
        ,AppFilePath,LuiDatConv,null
);
// устанавливаем свойство, указывающее что приложение создано для асинхронной задачи
app.setProperty("AsyncTask", taskAttr.Get_Element("code"));
app.setProperty("DEBUG_MODE", "Y");//это для логирования
// создаём в нём форму
frm=new LuiAppForm(app,this.taskAttr.Get_Element("code"),"TASK");

app.forms.put(frm.code,frm);   //записываем форму в массив форм
app.setVal("FormID",frm.code); // указываем текущую форму
// св-ва
frm.addProperty("FormID",frm.code);
frm.addProperty("TASK_ERROR_TEXT","");
frm.addProperty("TASK_PRG_BEGIN",taskAttr.Get_Element("action_before"));
frm.addProperty("TASK_QUERY",taskAttr.Get_Element("main_query"));
frm.addProperty("TASK_PRG_END",taskAttr.Get_Element("action_after"));
frm.addProperty("TASK_PRG_ROW",taskAttr.Get_Element("action_for_row"));
frm.addProperty("TASK_PRG_ROW_EX",taskAttr.Get_Element("action_for_row_exception"));
frm.addProperty("HEADER",taskAttr.Get_Element("progress_bar_title"));
frm.addProperty("TASK_BAR_PCT_TEXT",taskAttr.Get_Element("progress_bar_percent_text"));
frm.addProperty("TASK_BAR_WRK_TEXT",taskAttr.Get_Element("progress_bar_work_text"));
frm.addProperty("TASK_PRG_EXCEPTION",taskAttr.Get_Element("action_for_exception"));
frm.addProperty("ConfigCode",taskAttr.Get_Element("config_code"));
// параметры
LuiCollection parCol=new LuiCollection(taskAttr.Get_Element("params"));
double i=0;
LuiDbConn lc=null;
try {
lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());
for (String parCode:parCol.keySet()){
   LuiAppElement pel=frm.addElement(parCode, "PARAM",++i, parCode,null);
   pel.addProperty("Value", parCol.get(parCode));
   String dt=lc.getSelectResult("select lui_f_get_value('"+app.code+"','"+
           taskAttr.Get_Element("form_code")+"','"+
           taskAttr.Get_Element("config_code")+"','DATA_TYPE','"+
           parCode+"');"
           );
if (is_not_null(dt)){
    pel.addProperty("DATA_TYPE",dt);
String frmt=lc.getSelectResult("select lui_f_get_value('"+app.code+"','"+
           taskAttr.Get_Element("form_code")+"','"+
           taskAttr.Get_Element("config_code")+"','FORMAT','"+
           parCode+"');"
           );
frmt=NVL(frmt,"");
pel.addProperty("FORMAT",frmt);
ColInfo ci=new ColInfo();
ci.LuiDATATYPE=dt;
ci.LuiFORMAT=frmt;
pel.colinfo=ci;
}        
}
}finally {
    if (lc!=null) app.metaConnPool.SetFree(lc);
}
// создаём сеанс в БД если надо
if (is_not_null(taskAttr.Get_Element("db_url"))){
 
 LuiCollection connCol=new LuiCollection("");
 connCol.Add_Element("DBConnect", frm.code);
 connCol.Add_Element("NAME",frm.code+"(AsyncTask)");
 connCol.Add_Element("MODE",taskAttr.Get_Element("db_mode"));
 connCol.Add_Element("USERNAME",taskAttr.Get_Element("db_username"));
 connCol.Add_Element("PASSWORD","{DECR:"+taskAttr.Get_Element("db_password")+"}");
 connCol.Add_Element("URL",taskAttr.Get_Element("db_url"));
 connCol.Add_Element("SCHEMA",taskAttr.Get_Element("db_schema"));
 LuiAppProp p=app.addProperty("Connection",connCol.toString());
 app.parser.execute(app.getProperty("Connection"), p);
}
String fstatus="'WAIT'";//будущий статус
//и время выследующего выполнения
String fnexttime="lui_f_get_next_task_start(frequency,run_day,run_time,last_start)" ;
if (taskAttr.Get_Element("frequency").equals("PAUSE")) {
    fnexttime="lui_f_get_next_task_start(frequency,run_day,run_time,clock_timestamp()::lui_d_datetime)" ;
} 
if (taskAttr.Get_Element("frequency").equals("ONCE")) {
    fstatus="'DISABLE'";
    fnexttime="null";
}

String fprop="";//и проперти
String termMode="";
String termBy="";
// создаём сеанс в LUI
LuiSession luiSes=new LuiSession(app,taskAttr.Get_Element("started_by"));
// зовём выполнитель задач
LuiTask t=new LuiTask(frm);
// посмотрим а не надоли задаиться чуток в предбаннике
lc=null;
int prepcnt=1;

try {
lc=app.metaConnPool.GetFree(app.NLS.currentLang.locale,app.NLS.currentLang.code,app.code,app.getLuiSessionID(),app.getAppUser());    
    lc.getSelectResult("select pg_sleep_until(to_timestamp('"+taskAttr.Get_Element("next_start")+"','yyyy-mm-dd hh24:mi:ss'))");
   prepcnt=lc.executeDML("update lui_t_task set status='RUN',last_start=current_timestamp where app_code='"+app.code+"' and code='"+frm.code+"' and status='PREPARE'");
} catch (SQLException sex){
   // String here=sex.getLocalizedMessage();
}
finally {
    if (lc!=null) app.metaConnPool.SetFree(lc);
}
//<ID формы>.<Код элемента или "Form">.<Код свойства>
String errorMsg="";
if (prepcnt>0) { // если задача осталась в припарке (т.е. её не убили в предбаннике)
try {
t.procQueryRows(frm.code+".Form.TASK_PRG_BEGIN",
        frm.code+".Form.TASK_QUERY",
        frm.code+".Form.TASK_PRG_ROW",
        frm.code+".Form.TASK_PRG_ROW_EX",
        frm.code+".Form.TASK_PRG_END",
        "Y",
        frm.code+".Form.HEADER",
        frm.code+".Form.TASK_BAR_PCT_TEXT",
        frm.code+".Form.TASK_BAR_WRK_TEXT");
} 
catch         
(LuiException  lex){
    fstatus="'DISABLE'";
    fnexttime="null";
    errorMsg=lex.getMessage().replace("'", "''");
    frm.setProperty("TASK_ERROR_TEXT",errorMsg);
    frm.setProperty("TASK_ERROR_CODE",lex.getCode());
    fprop="lui_f_collection(lui_f_collection(session_property,'TASK_COMPLETE_CAUSE','ERROR'),'TASK_ERROR_TEXT','"+errorMsg+"')";
}
catch         
( SQLException lex){
    fstatus="'DISABLE'";
    fnexttime="null";
    errorMsg=lex.getMessage().replace("'", "''");
    frm.setProperty("TASK_ERROR_TEXT",errorMsg);
    frm.setProperty("TASK_ERROR_CODE",lex.getSQLState());
    fprop="lui_f_collection(lui_f_collection(session_property,'TASK_COMPLETE_CAUSE','ERROR'),'TASK_ERROR_TEXT','"+errorMsg+"')";
}
    }
// обновляем задачу ожиданием или неожиданием
lc=null;
try{    
lc=this.MetaConnPool.GetFree(NLS.currentLang.locale, NLS.currentLang.code, app.code, luiSes.id,taskAttr.Get_Element("started_by"));
// а не установлен ли статус завершения?
LuiCollection termCol=lc.getSelectFirstRowValMap("select terminate_mode,terminated_by from lui_v_task where app_code='"+taskAttr.Get_Element("app_code")+"' and code='"+taskAttr.Get_Element("code")+"'");
if (termCol!=null){
termMode=termCol.Get_Element("terminate_mode");
termBy=termCol.Get_Element("terminated_by");
}
if (is_not_null(termMode)){
fprop="lui_f_collection(session_property,'TASK_COMPLETE_CAUSE','"+termMode+"')";
    switch (termMode){
    case "BREAK":
        fstatus="'WAIT'";
        break;
    case "ABORT":
    case "STOP":
        fstatus="'DISABLE'";
        fnexttime="null";
        break;
}
}
lc.executeDML("update lui_v_task set status="+fstatus+",next_start="+fnexttime+" where app_code='"+taskAttr.Get_Element("app_code")+"' and code='"+taskAttr.Get_Element("code")+"'");
if (is_not_null(fprop)){
lc.executeDML("update lui_t_session set session_property="+fprop+" where id="+luiSes.id.toString());
}
if (is_not_null(termBy)){
    lc.executeDML("update lui_t_session set terminated_by='"+termBy+"' where id="+luiSes.id.toString());
}
if (is_not_null(errorMsg)){
    String action=frm.getProperty("TASK_PRG_EXCEPTION");
    if (is_not_null(action))
    try{    
      app.log.addRecord("TASK_ACTION_EXCEPTION",
    app.NLS.NLS("RUS:Выполнение действия при завершении по ошибке;ENG:Action on completion by error;"), 
    action, 
    frm.code);
      app.parser.execute(action,frm.properties.get("TASK_PRG_EXCEPTION"));
    } catch (Exception ex){
            }
}
} finally {
    if (lc!=null) MetaConnPool.SetFree(lc);
}
//закрываем сеанс
luiSes.endSession(true);
//удаляем форму
app.forms.remove(frm.code);
app.forms.clear();
//гасим приложение
app.close();

    } catch (LuiException ex) {
if (app!=null){
    if (frm!=null){
        app.forms.remove(frm.code);
        app.forms.clear();
    }
//гасим приложение
app.close();
}
    } catch (SQLException ex) {
if (app!=null){
    if (frm!=null){
        app.forms.remove(frm.code);
        app.forms.clear();
    }
//гасим приложение
app.close();
}

    }
    }
    
}
