/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.PipedReader;
import java.io.PipedWriter;

/**
 *
 * @author geliseev
 */
public class PipeChannel implements EventChannel {

    private final PipedReader mPipedReader;
    private final PipedWriter mPipedWriter;
    private final BufferedReader mReader;
    private final BufferedWriter mWriter;
    
    
    public PipeChannel() {
        mReader = new BufferedReader( mPipedReader = new PipedReader() );
        mWriter = new BufferedWriter( mPipedWriter = new PipedWriter() );
    }
    
    public PipeChannel(PipeChannel pChan) throws IOException {
        mReader = new BufferedReader( mPipedReader = new PipedReader( pChan.mPipedWriter ) );
        mWriter = new BufferedWriter( mPipedWriter = new PipedWriter( pChan.mPipedReader ) );
    }
    
    
    @Override
    public BufferedReader reader() {
        return mReader;
    }

    @Override
    public BufferedWriter writer() {
        return mWriter;
    }

    @Override
    public void close() throws Exception {
        mReader.close();
        mWriter.close();
        mPipedReader.close();
        mPipedWriter.close();
    }
    
}
