/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package luiserver;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.nio.charset.Charset;

/**
 *
 * @author geliseev
 */
public class SocketChannel implements EventChannel {

    private final Socket mSocket;
    private final BufferedReader mReader;
    private final BufferedWriter mWriter;
    
    
    public SocketChannel( Socket pSocket ) throws IOException {
        mSocket = pSocket;
        mReader = new BufferedReader(new InputStreamReader(pSocket.getInputStream(), Charset.forName("utf8")));
        mWriter = new BufferedWriter(new OutputStreamWriter(pSocket.getOutputStream(), Charset.forName("utf8")));
    }
    
    public SocketChannel( Socket pSocket, Charset pCharset) throws IOException {
        mSocket = pSocket;
        mReader = new BufferedReader(new InputStreamReader(pSocket.getInputStream(), pCharset));
        mWriter = new BufferedWriter(new OutputStreamWriter(pSocket.getOutputStream(), pCharset));
    }
    
    
    @Override
    public BufferedReader reader() {
        return mReader;
    }

    @Override
    public BufferedWriter writer() {
        return mWriter;
    }

    @Override
    public void close() throws Exception {
        mReader.close();
        mWriter.close();
        mSocket.close();
    }

    public Socket getSocket() {
        return mSocket;
    }
    
}
