/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.tlc.lui;

import static LuiCollection.LuiString.like;
import LuiDb.LuiDbConnectPool;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Properties;
import luiserver.LuiSrv;
import luiserver.LuiTaskAsync;

/**
 * Стартер демона в Linux
 * @author geliseev
 */
public class Starter {
    
    
    // Имя Lui-servera    
    private static String LuiSrvName = null;
    
    // Конфигурация    
    private static Properties stConfig = null;
    // Домашний каталог (где лежит LuiServer.jar_
    private static File stHomeDir = null;
    // Путь кконфигурационному файлу
    private static File stConfigFile = null;
    // Признак отладочного режима
    private static boolean stDebug = false;
    // Путь к каталогу аплоада файлов
    private static File  stAppFiles = null;
    // URL базы данных в формате JDBC
    private static String stDbURL = null;
    // URL базы данных в формате JDBC
    private static String taskDispatcherTimeOut="3000";
    private static String sessDispatcherTimeOut="60000";
    private static String daysToKeepOldSessions="5";
    private static String licenseMonths="120";
    
    private static String licenseChart="B";
    private static LuiSrv aLuiSrv=null;
    
    
    /**
     * Главная процедура запуска
     * @param pArgs аргументы командной строки
     */
    public static void main(String[] pArgs) {
    
	       try {
            if (Class.forName("org.postgresql.Driver") == null) {
                System.err.println("Can not load postgresql jdbc driver");
                System.exit(1);
            }
        } catch (ClassNotFoundException ex) {
            System.err.printf("Can no load postgresql driver: %s %s\n", ex.getClass().getCanonicalName(), ex.getMessage());
            System.exit(1);
        }

        try {

            try {
                parseParams(pArgs);
            } catch (URISyntaxException ex) {
                printError(100, "Can not acceess config file. Does LuiServer started not from a filesystem?\nError: %s\n", ex.getMessage());
            }

            checkDatabase();
            LuiDbConnectPool aConnectPool = new LuiDbConnectPool(
                    stDbURL,
                     stConfig.getProperty("dbuser"),
                     stConfig.getProperty("dbpasswd"),
                     stConfig.getProperty("schema"),
                     true,
                     false,
                     "LUI MetaData Access");

            //System.out.println("plugins = " + stConfig.getProperty("plugins"));
            //System.out.println("appfiles = " + stConfig.getProperty("appfiles"));
            HashMap<String, String> aLUIParams = new HashMap<>();
            aLUIParams.put("APP_FILES_PATH", new File(stConfig.getProperty("appfiles")).getAbsolutePath() + File.separator);
            aLUIParams.put("LUI_SCHEMA", stConfig.getProperty("schema"));
            aLUIParams.put("LUI_OWNER", stConfig.getProperty("owner"));
            aLUIParams.put("includePlugin", stConfig.getProperty("includePlugin"));
            aLUIParams.put("excludePlugin", stConfig.getProperty("excludePlugin"));
            aLuiSrv = new LuiSrv(aConnectPool,stConfig.getProperty("defApp"), stConfig.getProperty("plugins"), aLUIParams, 
                    taskDispatcherTimeOut,sessDispatcherTimeOut,daysToKeepOldSessions, 
                    LuiSrvName,licenseMonths,licenseChart,stHomeDir);

            aLuiSrv.startTCPListener(null, Integer.parseInt(stConfig.getProperty("listen")));

        } catch (Exception ex) {
            System.err.printf("Can not start LUI Server:\nException: %s\nError:%s\n", ex.getClass().getCanonicalName(), ex.getMessage());
            if (stDebug) {
                ex.printStackTrace(System.err);
            }
            System.exit(101);
        }
// Крюк гашения сервера        
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {

                if (aLuiSrv.taskDisp != null) {
                    //aLuiSrv.taskDisp.thread.interrupt(); -- так не получилось
                    aLuiSrv.taskDisp.clearNotRunTask(); // чистим мёртвые задачи
                    if (aLuiSrv.taskDisp.asyncTasks.size() > 0) { // если массив задач не пуст
                        System.out.println("Breaking working async tasks....");
                        System.out.flush();
                        for (String tk : aLuiSrv.taskDisp.asyncTasks.keySet()) { // крутим оставшиеся
                            LuiTaskAsync tas = aLuiSrv.taskDisp.asyncTasks.get(tk);
                            if (!tas.thread.isAlive()) {
                                continue; // если сдох смотрим следующий
                            }
                            String stmt = "select lui_r_stop_async_task('" + tas.app.code + "','" + tas.frm.code + "','BREAK','STOP')";
                            System.out.print(tas.frm.code + "...");
                            System.out.flush();
                            try {
                                aLuiSrv.taskDisp.lc.getSelectResult(stmt);
                            } catch (SQLException ex) {
                                System.out.println("SQLError:" + ex.getLocalizedMessage());
                            } catch (Exception ex) {
                                System.out.println("Error:" + ex.getLocalizedMessage());
                            }
                            System.out.println("OK");
                            System.out.flush();
                        }
                        while (true) {
                            aLuiSrv.taskDisp.clearNotRunTask();
                            if (aLuiSrv.taskDisp.asyncTasks.size() == 0) {
                                break;
                            }
                        }
                    }
                }

            }
        });

        // Бесконечный цикл выжидания
        while (true) {
            try {
                Thread.sleep(600000);
            } catch (InterruptedException ignore) {
            }

            // Раз в 10 минут - самообслуживание
            System.gc();
        }

    }
    
    
    /**
     * Проверка качества базы данных, выполнение функции refresh_roles
     */
    private static final void checkDatabase() {
        stDbURL = String.format("jdbc:postgresql://%s:%s/%s",  stConfig.getProperty("dbhost"), stConfig.getProperty("dbport"), stConfig.getProperty("database") );
        if(stDebug) {
            System.out.printf("Trying to connect database \"%s\"\n", stDbURL);
        }
        
        int aWaitDb = 0;
        try {
            aWaitDb = Integer.parseUnsignedInt(stConfig.getProperty("waitdb"));
        }
        catch(Exception ignore) {};
        
        long aStopWait = System.currentTimeMillis() + 1000*aWaitDb;
        
        Connection aConn = null;
        
        while(true) {
            try {
                aConn = DriverManager.getConnection(stDbURL, stConfig.getProperty("dbuser"), stConfig.getProperty("dbpasswd"));
                break;
            } catch (SQLException ex) {
                if(System.currentTimeMillis() > aStopWait) {
                    printError(20, "Connection to database \"%s\" failed with: %s\n", stDbURL, ex.getMessage());
                    return;
                }
                try { Thread.sleep(1000); } catch(InterruptedException ignore) {}
            }
        }
        
        // Проверка базы данных
        String aStatement;
        try( PreparedStatement aST = aConn.prepareStatement(aStatement = "SELECT COUNT(1) FROM pg_catalog.pg_tables WHERE schemaname=? and tablename in ( 'lui_t_value', 'lui_t_config', 'lui_t_language', 'lui_t_plugin', 'lui_t_message' )"))
        {
            aST.setString(1, stConfig.getProperty("schema"));
            try(ResultSet aRS = aST.executeQuery()) {
               aRS.next();
               if(aRS.getInt(1) != 5) {
                   printError(22, "Schema \"%s\" does not exists or corrupted in database \"%s\"", stConfig.getProperty("schema"), stDbURL);
               }
            }
        } catch (SQLException ex) {
            printError(21, "Can not read data from pg_catalog: %s", ex.getMessage());
            return;
        }
                  
        // Выполнение скрипта refresh_roles
        try
        {
             CallableStatement aST = aConn.prepareCall(aStatement = "SET SESSION SEARCH_PATH = \""+stConfig.getProperty("schema")+"\",\"$user\", public");
             aST.execute();
             aST.close();
             //aST = aConn.prepareCall(aStatement = "SELECT lui_f_get_lui_owner()");
             aST=aConn.prepareCall(aStatement = "SELECT lui_r_refresh_roles()");
             aST.execute();
             aST.close();
             
        } catch (SQLException ex) {
            printError(23, "Failed to run SQL statement '%s': %s", aStatement, ex.getMessage());
        }
        
        try {
            aConn.close();
        } catch(SQLException ex) {
            printError(24, "Unexpected error while disconnectiong from database: %s", ex.getMessage());
        }
    }
    
    
    
    /**
     * Печать ошибки, выход с кодом завершения
     * @param pCode код завершения
     * @param pFormat форматная строка ошибки
     * @param pArgs аргументы для подстановки в форматную строку
     */
    private static void printError( int pCode, String pFormat, Object ... pArgs ) {
        System.err.printf(pFormat, pArgs);
        if(!pFormat.endsWith("\n")){ System.err.println(); }
        System.err.println("Try \"--help\" option for more information");
        java.lang.System.exit(pCode);
    }
    
    
    
    private static void parseParams(String[] pArgs) throws URISyntaxException {
        LinkedList<String> aArgs = new LinkedList<>(Arrays.asList(pArgs));
    
        stHomeDir = new File(Starter.class.getProtectionDomain().getCodeSource().getLocation().toURI().getPath());
        if(!stHomeDir.isDirectory()) { stHomeDir = stHomeDir.getParentFile(); }
        
        // Ищем в параметрах путь к конфигурационному файлу
        while(!aArgs.isEmpty()) {
            String aParam = aArgs.pollFirst();
            if(aParam == null) { break; }
            if("--config".contentEquals(aParam)) {
                aParam = aArgs.pollFirst();
                if(aParam == null) { 
                    printError(10, "Parameter \"--config\" present in command line, but configuration file location not specified");
                }
                stConfigFile = new File(aParam);
            }
            if("--debug".contentEquals(aParam)) {
                stDebug = true;
            }
            if("--help".contentEquals(aParam)) {
                printHelp();
                return;
            }
        }
        aArgs = new LinkedList<>(Arrays.asList(pArgs));
        
        // Ищем конфигурационный файл
        if(stConfigFile == null) { stConfigFile = new File(stHomeDir, "LuiServer.cfg" ); }
        if(stDebug) { System.out.printf("Using configuration file \"%s\"\n", stConfigFile.getAbsolutePath()); }
        if( ! stConfigFile.exists() ) { printError(11, "Configuration file \"%s\" not exists", stConfigFile.getAbsolutePath()); }
        if( ! stConfigFile.isFile() ) { printError(11, "Configuration file \"%s\" is not a file", stConfigFile.getAbsolutePath()); }
        if( ! stConfigFile.canRead() ) { printError(11, "Configuration file \"%s\" is not readable", stConfigFile.getAbsolutePath()); }
        
        
        //  Параметры по-умолчанию
        Properties defConfig = new Properties();
        defConfig.setProperty("includePlugin","''");
        defConfig.setProperty("excludePlugin","''");
        defConfig.setProperty("dbuser","lui_common");
        defConfig.setProperty("dbpasswd","");
        defConfig.setProperty("dbhost","localhost");
        defConfig.setProperty("dbport","5432");
        defConfig.setProperty("database","postgres");
        defConfig.setProperty("schema","lui");
        defConfig.setProperty("owner","lui");
        defConfig.setProperty("listen","7777");
        defConfig.setProperty("appfiles",new File(stHomeDir,"appfiles").getAbsolutePath());
        defConfig.setProperty("plugins", new File(stHomeDir,"plugins").getAbsolutePath());
        defConfig.setProperty("taskDispatcherTimeOut","3000");
        defConfig.setProperty("sessDispatcherTimeOut","60000");
        defConfig.setProperty("daysToKeepOldSessions","5");
        defConfig.setProperty("licenseMonths","120");
        defConfig.setProperty("licenseChart","B");
        defConfig.setProperty("defApp","LUI");
        defConfig.setProperty("waitdb", "0");
        
        stConfig = new Properties(defConfig);
        
        // Читаем конфигурационный файл
        FileReader fr=null;
        try {
            fr=new FileReader(stConfigFile);
            stConfig.load(fr);
        } catch (FileNotFoundException ex) {
            printError(12, "Configuration file \"%s\" not found: %s", stConfigFile.getAbsolutePath(), ex.getMessage());
        } catch (IOException ex) {
            printError(12, "Failed to load configuration file \"%s\": %s", stConfigFile.getAbsolutePath(), ex.getMessage());
        }
        finally{
            if (fr!=null)try{fr.close();}catch(Exception ex){}
        }
        
        // проверяем наличие левых параметров в конфигурационном файле
        for( Object aParam: stConfig.keySet()) {
            if( !defConfig.containsKey(aParam) ) {
                printError(13,"Unexpected parameter \"%s\" in configuration file \"%s\"", aParam, stConfigFile );
            }
        }
        
        // Повторное чтение параметров командной строки
        // Результаты кладутся в stConfig
        String aParam;
        String aVal;
        int aPid;
        
        while (!aArgs.isEmpty()) {
            aParam = aArgs.pollFirst();
            if (aParam == null) {
                break;
            }
            switch (aParam) {
                case "--config":
                    aArgs.pollFirst();
                    break;
                case "--debug": break;
                case "--taskDispatcherTimeOut":
                case "--sessDispatcherTimeOut":
                case "--daysToKeepOldSessions":    
                case "--dbuser":
                case "--dbpasswd":
                case "--dbhost":
                case "--dbport":
                case "--database":
                case "--schema":
                case "--listen":
                case "--plugins":
                case "--defApp":
                case "--waitdb":
                case "--appfiles": aVal = aArgs.pollFirst();
                                   if(aVal == null || aVal.isEmpty()) {
                                       printError(10, "No value specified for parameter \"%s\"", aParam);
                                   }
                                   aParam = aParam.substring(2);
                                   stConfig.setProperty(aParam, aVal);
                                   break;
                case "--pidfile":
                                aVal = aArgs.pollFirst();
                                if(aVal == null || aVal.isEmpty()) {
                                    printError(10, "No value specified for parameter \"%s\"", aParam);
                                } 
                                
                                try {
                                    aPid = Integer.parseInt(new File("/proc/self").getCanonicalFile().getName());
                                }
                                catch(IOException | NumberFormatException ex) {
                                    System.err.printf("Warning: Can not stat /proc/self pseudofile: %s %s\n", ex.getClass().getCanonicalName(), ex.getMessage());
                                    break;
                                }
                                try (PrintWriter aPidWriter = new PrintWriter( aVal )){
                                    aPidWriter.printf("%d\n", aPid);
                                    }
                                catch(Exception ex) {
                                   System.err.printf("Warning: Can not write pid to '%s': %s %s\n", aVal, ex.getClass().getCanonicalName(), ex.getMessage());
                                }
                                break;
                               
                default: 
                  
                    printError(9, "Unexpected parameter \"%s\"", aParam); 
            }
        }
        
        // Отладка - вывод всех конфигурационных параметров
        if(stDebug) {
            System.out.println("Configuration:");
            stConfig.entrySet().forEach((aProp) -> {
                System.out.printf(" \"%s\":\"%s\"\n", aProp.getKey(), aProp.getValue());
            });
        }
        
        // Проверка параметров на "правильность"
        int aPort = -1;
        try { aPort = Integer.parseInt(stConfig.getProperty("dbport"));} catch (Exception ignore) {}
        if( aPort < 1 || aPort > 65535) {
            printError(14, "%s value must be an integer between 1 and 65535, not %s", "dbport", stConfig.getProperty("dbport") );
        }
        
        aPort = -1;
        try { aPort = Integer.parseInt(stConfig.getProperty("listen"));} catch (Exception ignore) {}
        if( aPort < 1 || aPort > 65535) {
            printError(14, "%s value must be an integer between 1 and 65535, not %s", "listen", stConfig.getProperty("listen") );
        }
        
        
        LuiSrvName=generateServerName();
        taskDispatcherTimeOut=stConfig.getProperty("taskDispatcherTimeOut");
        sessDispatcherTimeOut=stConfig.getProperty("sessDispatcherTimeOut");
        daysToKeepOldSessions=stConfig.getProperty("daysToKeepOldSessions");
        licenseMonths=stConfig.getProperty("licenseMonths");
        licenseChart=stConfig.getProperty("licenseChart");
        stAppFiles = new File(stConfig.getProperty("appfiles"));
        if(!stAppFiles.exists()) { printError(15, "appfiles directory \"%s\" does not exists", stConfig.getProperty("appfiles") ); }
        if(!stAppFiles.isDirectory()) { printError(15, "appfiles directory \"%s\" is not a directory", stConfig.getProperty("appfiles") ); }
        if(!stAppFiles.canWrite()) { printError(15, "appfiles directory \"%s\" is not writeable", stConfig.getProperty("appfiles") ); }
        
    }

    
    public static String generateServerName(){
        InetAddress addr; 
        String myLANIP="?";
        String myLANNAME="?";
        try {
            addr = InetAddress.getLocalHost();
            myLANIP = addr.getHostAddress();
            myLANNAME = addr.getCanonicalHostName();
        } catch (UnknownHostException ex) {
          myLANIP="?";
          myLANNAME="?";
        }
        
        return myLANNAME+"/"+myLANIP+":"+(stConfig==null?"?":stConfig.getProperty("listen"));
    }
    
    /**
     * Выаод справки по ключам запуска
     */
    private static void printHelp() {
        System.out.println(
                "Usage\n "
            +   "--config <path>        - use specified configuration file\n"
            +   "--debug                - run in debug mode\n"
            +   "--dbhost <ip|dns>      - PostgreSQL database address\n"
            +   "--dbport <number>      - PostgreSQL database TCP port (default 5432)\n"
            +   "--database <name>      - PostgreSQL database name\n"
            +   "--schema <name>        - LUI Metadata schema name in database\n"
            +   "--dbuser <name>        - LUI \"common\" user name\n"
            +   "--dbpasswd <password>  - LUI \"common\" user password\n"
            +   "--taskDispatcherTimeOut <msec>  - Async tasks dispatcher time out in milliseconds. If \"0\" then dispatcher not starting\n"                        
            +   "--appfiles <path>      - LUI file upload directory path (must be same to \"appfiles\" in LUI Web Application config\n"
            +   "--listen <number>      - TCP port to listen incoming connections from LUI Web Application=n"
            +   "--plugins <path>       - path to plugins directory\n"
            +   "--pidfile <path>       - file to write process id (On linux only. Not a reiable method. Uses /proc/self pseudo file))\n"
            +   "--waitdb <N>           - wait for initial database connections for N seconds. 0 (default) - do not wait, exit immediately if database os not available.\n"
            +   "--help                 - show this help\n"
        );
    }
    
}
