/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.telecom.webui.common;

import java.util.HashMap;
import java.util.regex.Pattern;

/**
 *
 * @author john
 */
public class Defaults {

    public static enum Instructions { 
        s, // Start 
        f, // Finish
        err, // Error
        le, // Last Event
        lc, // Last Command
        code, // code
        c, // Create
        position$,
        win, // WINdow
        swin, // Service window
        h, // Height 
        ha, // Height auto
        h$, // height calculate
        w, // Width
        wa, // Width auto
        w$, // Width calculate
        mnu, // MeNU on
        _mnu, // MeNU off
        cls, // CLoSe on 
        _cls, // CLoSe off
        ttl, // TiTLe 
        _ttl, // Hide TiTLe
        prnt, // Parent
        pos, // Position
        posc, // Position cascade
        post, // Position center
        posp, // Position parent
        posz, // Position zero
        pspr, // Position parent right
        pspb, // position parent bottom
        pslc, // position last click
        b, // Begin
        e, // End
        form, // FOrm
        tree, // TRee
        nd, // node 
        lbl, // LaBeL 
        hnt, // HiNT 
        _hnt, // remove hint
        ico, // ICOn 
        _ico, // ReMove ICon
        type, // type
        stat, // STATe
        sty, // STYle
        _sty, // remove STYle
        head, // HEADer
        txt, // TeXT
        _txt, // Empty text
        tbar, // ToolBAR 
        bbar, // ButtonBAR 
        btn, // BuTtoN 
        _btn, // Button end
        sep, // SEParator
        cfg, // ConFiGs 
        _cfg, // Configs end
        elm, // ELeMent
        _elm, // ELeMent end
        k, // Key 
        v, // Value
        vl, // Value
        itm, // ITeM 
        _itm, // ITeM end
        curr, // CURRent
        curf, // Current With False
        clr, // CLeaR
        bsmt, // BaSeMenT
        cont, // Content
        econ, // East cOntainer 
        scon, // South cOntainer
        evt, // EVenT 
        evtc, // Event Code 
        evtd, // Event Delay 
        par, // PARameter
        parv, // Parameter Value
        east$size, south$size,
        en, // ENable
        dis, //DISable
        show, // Show
        hide, // Hide
        menu, // Menu
        auto$class,
        grid, // Grid 
        col, // COLumn  
        row, // Row  
        cell, // Cell 
        srow, // Set row 
        scol, // Set column 
        hdr, // Header on 
        _hdr, // Header off
        ftr, // Footer on 
        _ftr, // Footer off
        foot, // Footer value
        alig, // Align 
        srt, // SoRT 
        _srt, // Remove sort 
        rall, // All Rows
        ftch, // Fetch Rows
        sel, // SElect
        _sel, // DeSelect
        rm, // ReMove
        clos, // Close
        gcol, // Group of columns 
        gfix, // Group fixed
        govr, // Group overflow 
        gfld, // Group of fields
        gtab, // Tab group
        ghrz, // Horizontal group
        gbch, // Bunch group
        tabs, // Group of tabs
        qbe, // QBE on,
        _qbe, // QBE off
        cqbe, // Clear QBE
        edit, // Edit
        _edt, // Nonedit
        ini, // INIt
        _ini, // 
        blnk, // Blank 
        fld, // Field 
        _prm, // No prompt
        prm, // Prompt 
        ro, // Readonly 
        rw, // Read-write 
        prml, // Prompt Left
        prmt, // Prompt Top
        prmw, // Prompt Width
        cb, // Cell buttons
        qstn, // Question
        msg, // Message
        note, //Note
        lov, // LOV 
        mndt, // Mandatory 
        ord, // Ordinary
        drs, // Dropdown select
        drp, // Dropdown start
        _drp, // Dropdown end
        dit, // Dropdown item start
        _dit, // Dropdown item end
        dkey, // Dropdown item key
        dval, // Dropdown item value
        lvl, // Level
        ctxt, // Context
        cal, // Calendar
        cyr, // Calendar year
        cmon, // Calendar month
        cday, // Calendar day
        crfr, // Calendar Refresh
        cdd, // Calendar Date Detail
        csd, // Calendar special date start
        _csd, // Calendar special date end
        chr, // Calendar hour
        cmin, // Calendar minutes
        csec, // Calendar seconds
        ctd, // Calendar time detail
        dl, // Download
        upl, // Upload
        _mlt, // Not multiple
        inft, // Info title
        infc, // Info content
        infa, // Info address
        infp, // Info params
        inf, // Info
        lck, // Lock
        _lck, // Unlock
        fcs, // Focus
        cfcs, // Change focus
        root, // Set root element
        init, // Init workarea
        acel, // Activate cell
        nav, // Navigable
        _nav, // Not navigable
        sc, // South container
        sc$, // South container with calculate
        ec, // East container
        ec$, // East container with calculate
        ow, // Overflow width
        ow$, // Overflow width with calculate
        poll, // Server polling
        prog, // Progress process
        prtl, // Progress title
        prpt, // Progress process text
        prwt, // Progress work text
        prct, // Progress percent
        prca, // Progress cancelable
        prcl, // Progress close
        prw, // Progress window
        rtf, // RTF editor
        snd, // Sound
        ok, // OK
        trb, // Trace begin
        tre, // Trace end
        trc, // Tracer
        trtl, // Trace title
        trti, // Trace time
        trty, // Trace type
        trms, // Trace message
        upd, // Update
        dk, // Dictionary key
        dv, // Dictionary value
        skin, // Styles skin
        pat, // Pattern
        nxt, // Next
        app, //Application
        prot, // Protocol
        mm, // Main menu
        run, // Run application
        lgn, // Logon window
        maxv, // Maximum value
        car, // Caret position
        mode, // Mode
        swtc, // Switch
        _gtf, // Go to field unlock
        fh, // Factor height
        loc, // Locale
        js, // JavaScript
        vrp, // Verifier parameter start
        _vrp, // Verifier parameter end
        vrpn, // Verifier parameter name
        vrpv, // Verifier parameter value
        vr, // Verifier
        rdp, // radio parameter start
        _rdp, // radio parametr end
        rdpd, // radio item disabled
        rdpn, // radio parameter name
        rdpv, // radio parameter value
        rd, // radio field start
        _rd, // radio field end
        fhlp, // Field Helper
        mkup, // Make up
        _mkp, // Remove make up
        mkps, // makeup style
        chrt, // Chart form
        html, // HTML form
        data, // Data
        gen, // Generate
        pie, // Pie chart
        cmd, // Command
        exec, // Execute
        fav, // Favicon
        save, // Save
        cur, // Cursor
        _cur, // Undo cursor
        lm, // Legend mode
        ee, // Emulated event
        em, // Empty
        _em, // Not Empty
        fsz, // fix size
        mapc, // geo map code
        map, // geo map
        max, // maximize
        rest, // restore
        mxm, // maximize button
        _mxm, // hide maximize button
        bg, // background
        _bg, // no background
        cpk, // color picker
        cpa, // color palette
        _cpa, // default color palette
        lft, // left
        lft$, // left with calculation
        top, // top
        top$, // top with calculation
        rmin, // range min
        rmax, // range max
        rstp, // range step
        rfrm, // range from
        cvr, // cover
        mtrx, // matrix
        ornt, // orientation
        ornt_, // orientation value only
        _orn, // disable orientation
        halg, // horiz align
        halg_, // horiz align value only
        valg, // vert align
        valg_, // vert align value only
        ovrf, // overflow
        ovrf_, // overflow value only
        clss, // classes
        clss_, // classes
        mucs, // muclasses
        ins, // insert
        css, // css
        scls, // self class
        scls_, // self class
        chld, // child
        get, // Get
        appr, // Appearance
        xlbl, // extended label
        brd, // border
        _brd, // no border
        cnf, // column info
        _cnf, // no column info
        rnf, //row info
        _rnf, // no row info
        minf, // matrix info
        ecod, // elem code
        prop, // properties
        mnf, // menu form
        mtr, // menu tree
        mnd, // menu node
        aw, // angular width
        sm, // Selection mode
        _sm, // Turn off selection mode
        rfr, // Refresh
        capt, // Capture
        tabf, // Tab form
        more, // more
        id, // ID
        ln, // link
        ids, // id's string
        rids, // Row id's string
        cids, // Column id's string
        wipe, // Wipe
        pln, // Plain
        undo, // Undo
        redo, // Redo
        ewin, // Embedded window
        nop, // Null operation 
        
    }
    
    public static enum Events {
        
    }
    
    public static enum WindowPosition {
        Cascade,
        WorkCenter,
        Parent,
        ParentRight,
        ParentBottom,
        LastClick,
        Zero
    }
    
    public static enum FormPosition {
        Main,
        Right,
        Bottom,
        TabRight,
        TabBottom
    }
    
    public static enum FormType {
        List,
        Blanc,
        Tree,
        Menu,
        Chart,
        Html,
        Grid
    }
    
    public static enum NodeState {
        Collapsed, 
        Expanded, 
        Leaf, 
        Empty, 
        Continue, 
        AutoExpand,
        Hidden
    };

    public static enum NodeType {
        ROW, 
        ACTION, 
        CONTINUE
    };
    
    public static enum ButtonState {
        Active, 
        Disable, 
        Hidden,
        More
    };
    
    public static enum MenuItemType {
        Action, 
        Functional, 
        Column, 
        Event, 
        Client, 
        Sort, 
        Lang, 
        System, 
        None, 
        Local$
    };
    
    public static enum MenuItemRole {
        DBLCLICK, 
        ADD, 
        DELETE, 
        CLOSE, 
        Copy, 
        View, 
        None
    };
    
    public static enum ColumnAlign {
        Left, 
        Center, 
        Right
    };
    
    public static enum Action {
        Fetch  
    };
    
    public static enum SelectionMode {
        Single,
        Multi,
        None
    }
    
    public static enum GroupType {
        Tab,
        Virtual,
        Bunch,
        SideBySide,
        Fixed,
        Overflow
    }
    
    public static enum GroupState {
        Hidden,
        Visible
    }
    
    public static enum FieldType {
        text,
        upper,
        passwd,
        checkbox,
        richtext,
        hidden,
        area,
        code, 
        color,
        range,
        radio
    }
    
    public static enum FieldHelper {
        none,
        ro
    }
    
    public static enum DateDetail {
        D,
        M,
        Y
    }
    
    public static enum TimeDetail {
        H,
        M,
        S
    }
    
    public static enum LOVRowType {
        Leaf,
        Text,
        Tree,
        Combo
    }
    
    public static enum SortDirection {
        Undefined,
        Asc,
        Desc
    }
    
    public static enum InfoTarget {
        Float,
        Modal,
        Tab,
        Window
    }
    
    public static enum URLTarget {
        Float,
        Self,
        Tab,
        Window
    };
    
    public static enum Sounds {
        OK,
        INFO,
        ATTENTION,
        DANGER
    };
    
    public static enum CaptionOrientation {
        H,
        V,
        BT,
        TB
    };
    
    public static enum HorizontalAlign {
        Left,
        Center,
        Right,
        Undefined
    };
    
    public static enum VerticalAlign {
        Top,
        Center,
        Bottom,
        Undefined
    };
    
    public static enum Overflow {
        Hidden,
        Expand,
        Wrap,
        Undefined
    };
    
    public static enum MatrixRowInfo {
        None,
        Left,
        Right,
        Only,
        Y
    };
    
    public static enum MatrixColumnInfo {
        None,
        Top,
        Bottom,
        Only,
        Y
    };
    
    public static enum FieldMode {
        Editable,
        Readonly,
        Wiki
    }
    
    public static enum MatrixSelectionMode {
        Standard,
        Simple
    }
    
    public static enum InfoCellPosition {
        Top,
        Bottom,
        Left,
        Right
    }
    
    public static enum ChartOrientation {
        Portrait,
        Landscape
    }

    public static int HTTP_OK = 200;
    public static int HTTP_INTERNAL_SERVER_ERROR = 500;
    public static String CONFIG_FILE = "/usr/local/etc/lui2config.xml";
    //public static String CONFIG_FILE = "/usr/local/etc/fstconfig.xml";
    //public static String CONFIG_FILE = "/usr/local/etc/adminconfig.xml";
    //public static String CONFIG_FILE = "/usr/local/etc/userconfig.xml";
    public static String CONFIG_FILE_UNIX = CONFIG_FILE;
    public static String CONFIG_FILE_WINDOWS = "c:/lui" + CONFIG_FILE;
    public static String SLOW_POLL_INTERVAL = "15000";
    public static String NORMAL_POLL_INTERVAL = "3000";
    public static String FAST_POLL_INTERVAL = "500";
    public static String COLLECTION_MASK = "#$";
    public static String LF_REPLACE = "‡";
    public static String CARET = "¦";
    public static String XML_1_0_VALID = "[^\u0009\r\n\u0020-\uD7FF\uE000-\uFFFD\ud800\udc00-\udbff\udfff]";
    public static Pattern XML_1_0_PATTERN = Pattern.compile(XML_1_0_VALID);
    public static String XML_1_1_VALID = "[^\u0001-\uD7FF\uE000-\uFFFD\ud800\udc00-\udbff\udfff]";
    public static Pattern XML_1_1_PATTERN = Pattern.compile(XML_1_1_VALID);
    public static String[] MATRIX_PROPERTIES = {"halg", "valg", "ovrf", "mkup", "mucl", "scls", "clss"};
    public static String[] MATRIX_CELLSET_PROPERTIES = {"scls"};
    public static String[] MATRIX_CELL_PROPERTIES = {"halg", "valg", "ovrf", "mkup", "mucl", "clss"};
    public static String[] MATRIX_CELL_HTML_PROPERTIES = {"halg", "valg", "ovrf", "clss"};
    public static String[] MATRIX_JOIN_PROPERTIES = {"clss", "scls"};
    public static String MATRIX_RESET_CLASS = "NORMAL";
    public static String COLON_REPLACE = "ƒ";
    public static HashMap<String, String> CLASS_TEMPLATES = new HashMap<String, String>() {
        {
            put("FONT_SELECTED", "font-weight:600!important;");
            put("ALL_SELECTED", "font-weight:600!important;background-color:#EEF!important;");
            put("DANGER", "font-weight:600!important;color:#A00!important;");
            put("ATTENTION", "font-weight:600!important;color:#970!important;");
            put("OK", "font-weight:600!important;color:#4A0!important;");
            put("SECTION", "font-weight:600!important;color:white!important;background-color:#888!important;");
            put("BACKGROUND_SELECTED", "background-color:#EEF!important;");
            put("LOWERED", "color:#BBB!important;");
            put("INHERIT", "color:#444!important;background-color:#DDD!important;");
            put("MONOSPACE", "font-family:monospace!important;");
            put("BOLD", "font-weight:bolder;");
            put("TILT", "font-style:italic;");
            put("LARGE", "font-size:larger;");
            put("SHADOW", "text-shadow: 1px 0 3px grey;");
            put("START", "text-shadow:-1px -1px 0 #000;background-color:#0b0;color:white;border-right:1px solid #040;border-bottom:1px solid #040;font-weight:600;");
            put("INTERIM", "text-shadow:-1px -1px 0 #fff;background-color:#ece600;color: #404;border:1px solid #440;font-weight:600;");
            put("STOP", "text-shadow:-1px -1px 0 #000;background-color:#d00;color:white;border-right:1px solid #400;border-bottom:1px solid #400;font-weight:600;");
            put("NOTES", "font-style:italic;");
            put("BAD", "text-decoration:line-through;text-decoration-color:red;color:#888;");
            put("GRAY", "color:#AAA;");
            put("RED", "color:red;");
            put("GREEN", "color:forestgreen;");
            put("BLUE", "color:royalblue;");
            put("BROWN", "color:brown;");
            put("CYAN", "color:cyan;text-shadow:1px 1px 1px darkcyan;");
            put("ORANGE", "color:orange;");
            put("MAGENTA", "color:magenta;");
            put("ITALIC", "font-style:italic;");
            put("UNDERLINE", "text-decoration:underline;");
            put("CROSSEDOUT", "text-decoration:line-through;text-decoration-color:red;");
            put("NEGATIVE", "color:white;background-color:gray;font-weight:600;");
            put("NORMAL", "color:black;font-weight:normal;background-color:transparent;text-shadow:none;position:static;text-decoration:none;border:none;");
            put("ENTIRE", "z-index:1;");
            put("HIDDEN", "display:none;");
        };    
    };
}
