/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.telecom.webui.common;

import com.google.gson.JsonObject;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import org.apache.commons.lang3.StringUtils;
import static org.apache.commons.lang3.StringUtils.indexOf;
import static org.apache.commons.lang3.StringUtils.splitByWholeSeparator;
import static org.apache.commons.lang3.StringUtils.substringAfter;
import static org.apache.commons.lang3.StringUtils.substringBefore;
import org.jdom2.Element;

/**
 *
 * @author john
 */
public class Utils {
    
    private static final String IPV4_PATTERN = "(([01]?\\d\\d?|2[0-4]\\d|25[0-5])\\.){3}([01]?\\d\\d?|2[0-4]\\d|25[0-5])";
    private static final String IPV6_PATTERN = "([0-9a-f]{1,4}:){7}([0-9a-f]){1,4}";
    private static final char[] BASE52_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".toCharArray();
    private static final char[] BASE10_CHARS = "0123456789".toCharArray();
    
    public static String generateId() {
        Random random = new Random();
	StringBuilder sb = new StringBuilder();
        for (int i = 0; i < 6; i++) {
            sb.append(BASE52_CHARS[random.nextInt(52)]);
        }
        return sb.toString();
    }
    
    public static String numberSeq() {
        Random random = new Random();
	StringBuilder sb = new StringBuilder();
        for (int i = 0; i < 8; i++) {
            sb.append(BASE10_CHARS[random.nextInt(10)]);
        }
        return sb.toString();
    }
    
    public static boolean testEmptyString(String s) {
        return (s == null) || (s.equals(""));
    }

    public static String string2String(String s) {
        if (testEmptyString(s)) return "";
        return s;
    }

    public static boolean hasIcons(String s) {
        return indexOf(s, "[ico:") > -1;
    }
    
    public static boolean isIpAddress(String ipAddress) {
        Pattern VALID_IPV4_PATTERN=null;
	Pattern VALID_IPV6_PATTERN=null;
        try {
            VALID_IPV4_PATTERN = Pattern.compile(IPV4_PATTERN, Pattern.CASE_INSENSITIVE);
            VALID_IPV6_PATTERN = Pattern.compile(IPV6_PATTERN, Pattern.CASE_INSENSITIVE);
        } catch (PatternSyntaxException e) {}
        Matcher ip4 = VALID_IPV4_PATTERN.matcher(ipAddress);
        if (ip4.matches()) {
            return true;
        }
        Matcher ip6 = VALID_IPV6_PATTERN.matcher(ipAddress);
        return ip6.matches();
    }
    
    public static String getExceptionStack(Exception e) {
        Writer result = new StringWriter();
        PrintWriter printWriter = new PrintWriter(result);
        e.printStackTrace(printWriter);
        return result.toString();
    }
    
    public static boolean testYString(String s) {
        return (s != null) && (("Y").equals(s.toUpperCase()));
    }
    
    public static JsonObject instruction(Defaults.Instructions instruction) {
        JsonObject ret = new JsonObject();
        ret.addProperty(instruction.toString(), "");
        return ret;
    }

    public static JsonObject instruction(Defaults.Instructions instruction, String parameter) {
        JsonObject ret = new JsonObject();
        ret.addProperty(instruction.toString(), parameter);
        return ret;
    }
    
    public static ArrayList<JsonObject> nop() {
        ArrayList<JsonObject> ret = new ArrayList<>();
        ret.add(instruction(Defaults.Instructions.s));
        ret.add(instruction(Defaults.Instructions.f));
        return ret;
    }
    
    public static void reorderXML(Element e) {
        if (e.getName().equalsIgnoreCase("W") || e.getName().equalsIgnoreCase("C")) {
            e.sortChildren(new XMLComparator());
        }
        List<Element> children = e.getChildren();
        Iterator<Element> iter = children.iterator();
        while (iter.hasNext()) {
            Element child = iter.next();
            reorderXML(child);
        }
    }
    
    public static String encodeBrackets(String s) {
        return s.replace("<", "&lt;").replace(">", "&gt;");
    }
    
    public static StringBuilder patternSafe(String s) {
        StringBuilder ret = new StringBuilder();
        ret.append("\\Q");
        ret.append(s);
        ret.append("\\E");
        return ret;
    }
    
    public static ArrayList<JsonObject> parseCaption(String s) {
        ArrayList<JsonObject> ret = new ArrayList<>();
        JsonObject o;
        if (StringUtils.indexOf(s, "[ico:") < 0) {
            o = new JsonObject();
            if (testEmptyString(s)) {
                o.addProperty("type", "split");
            } else {
                o.addProperty("type", "icon");
                o.addProperty("val", s);
            }    
            ret.add(o);
        } else {
            String text = substringBefore(s, "[ico:");
            if (!testEmptyString(text)) {
                o = new JsonObject();
                o.addProperty("type", "text");
                o.addProperty("val", text);
                ret.add(o);
            }
            String[] a = splitByWholeSeparator(StringUtils.removeStart(s, text), "[ico:");
            for (int i = 0; i < a.length; i++) {
                if (!testEmptyString(a[i])) {
                    String icon = substringBefore(a[i], "]");
                    text = substringAfter(a[i], "]");
                    if (testEmptyString(icon)) {
                        o = new JsonObject();
                        o.addProperty("type", "split");
                        ret.add(o);
                    } else {    
                        o = new JsonObject();
                        o.addProperty("type", "icon");
                        o.addProperty("val", icon);
                        ret.add(o);
                    }
                    if (!testEmptyString(text)) {
                        o = new JsonObject();
                        o.addProperty("type", "text");
                        o.addProperty("val", text);
                        ret.add(o);
                    }
                }
            }
        }
        return ret;
    }
}
