/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.telecom.webui.core;

import java.io.IOException;
import java.io.StringReader;
import java.sql.Array;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;
import java.sql.CallableStatement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import ru.fors.telecom.webui.common.Defaults;
import ru.fors.telecom.webui.common.SharedCommand;
import ru.fors.telecom.webui.common.Utils;
import java.util.concurrent.TimeUnit;
import org.apache.commons.dbutils.DbUtils;
import ru.fors.telecom.webui.common.SharedCommands;


/**
 *
 * @author john
 */
public class SharedArea {
    
    private Connection connection = null;
    private CallableStatement statement = null;
    private final String descriptor;
    private final Properties properties;
    private final SAXBuilder xmlBuilder;
    private final ArrayList<HashMap<String, String>> globalParams;
    private final HashMap<String, String> localParams;
    private final HashMap<String, String> effectiveParams;
    private String commandTag;
    private int commandLevel;
    private String commandToProcess;
    private String param;
    private String param2;
    private String tnsAdmin = "";
    private final HashMap<String, SharedCommands> commands;
    private final SharedCommands defaultCommands;

    public SharedArea (String desc, String tns, String user, String pwd) {
        properties = new Properties();
        properties.setProperty("user", user);
        properties.setProperty("password", pwd);
        properties.put("v$session.program", "WEBUI");
        descriptor = desc;
        tnsAdmin = tns;
        xmlBuilder = new SAXBuilder();
        globalParams = new ArrayList<>();
        localParams = new HashMap<>();
        effectiveParams = new HashMap<>();
        commandLevel = -1;
        commandToProcess = null;
        commands = new HashMap<>();
        defaultCommands = new SharedCommands();
        defaultCommands.setRetry(Defaults.NORMAL_POLL_INTERVAL);
    }
    
    public void startQuery () {
        Query query = new Query();
        Thread thread = new Thread(query);
        thread.start();
    }

    public class Query implements Runnable {

        @Override
        public void run() {
            while (true) {
                try {
                    if (connection == null) {
                        if (Utils.testEmptyString(tnsAdmin)) {
                            connection = DriverManager.getConnection("jdbc:oracle:thin:@" + descriptor, properties);
                        } else {
                            connection = DriverManager.getConnection("jdbc:oracle:thin:@" + descriptor + "?TNS_ADMIN=" + tnsAdmin, properties);
                        }
                        //connection = DriverManager.getConnection("jdbc:oracle:thin:@" + descriptor, properties);
                        connection.setAutoCommit(false);
                    } else {
                        statement = connection.prepareCall("begin ? := ui_p_common.poll(''); end;");
                        statement.registerOutParameter(1, Types.ARRAY, "XMLVARRAY");
                        statement.execute();
                        Array a = statement.getArray(1);
                        String[] s = (String[]) a.getArray();
                        StringBuilder res = new StringBuilder();
                        for (int i = 0; i < s.length; i++) {
                            res.append(s[i]);
                        }
                        processCommands(res.toString());
                        statement.close();
                        statement = null;
                    }
                    TimeUnit.MILLISECONDS.sleep(500);
                } catch (Exception ex) {
                    commands.clear();
                    if (statement != null) {
                        try {
                            statement.close();
                            DbUtils.close(statement);
                        } catch (SQLException ex1) {
                        } finally {
                            statement = null;
                        }
                    }
                    if (connection != null) {
                        try {
                            connection.close();
                            DbUtils.close(connection);
                        } catch (SQLException ex1) {
                        } finally {
                            connection = null;
                        }
                    }
                    try {
                        TimeUnit.SECONDS.sleep(90);
                    } catch (Exception ex1) {}
                }
            }
        }
 
    }
    
    public synchronized void processCommands(String rawCommands) throws JDOMException, IOException {
        Document XMLCommands = xmlBuilder.build(new StringReader(rawCommands));
        Utils.reorderXML(XMLCommands.getRootElement());
        List<Element> commandsList = XMLCommands.getRootElement().getChildren("W");
        commands.clear();
        if (commandsList.isEmpty()) {
            buildInstructions(XMLCommands.getRootElement());
        } else {
            Iterator<Element> iter = commandsList.iterator();
            while (iter.hasNext()) {
                buildInstructions(iter.next());
            }
        }
    }
    
    private void buildInstructions(Element elem) {
        globalParams.clear();
        localParams.clear();
        commandTag = "";
        processXML(elem, 0);
        processCommand();
    }
    
    private void processXML(Element e, int level) {
        if (e.getName().equalsIgnoreCase("Command") || (e.getName().equalsIgnoreCase("W"))) {
            commandTag = "W";
            processCommand();
        } else if (e.getName().equalsIgnoreCase("C")) {
            processCommand();
            commandTag = "C";
            commandToProcess = e.getAttributeValue("N");
            commandLevel = level;
        }
        if (e.getName().equalsIgnoreCase("P")) {
            if (commandTag.equalsIgnoreCase("W")) {
                if (globalParams.size() < level) {
                    for (int i = globalParams.size() + 1; i <= level; i++) {
                        globalParams.add(new HashMap<String, String>());
                    }
                }
                globalParams.get(level - 1).put(e.getAttributeValue("N"), e.getText());
            } else if (commandTag.equalsIgnoreCase("C")) {
                localParams.put(e.getAttributeValue("N"), e.getText());
            }
        }
        List<Element> children = e.getChildren();
        Iterator<Element> iter = children.iterator();
        while (iter.hasNext()) {
            Element child = iter.next();
            processXML(child, level + 1);
        }
    }
    
    private String checkParameter(String param) {
        if (effectiveParams.containsKey(param)) {
            return effectiveParams.get(param);
        }
        return null;
    }
    
    private void processCommand() {
        if (commandToProcess != null) {
            effectiveParams.clear();
            try {
                int i;
                for (i = 0; i < commandLevel; i++) {
                    if ((!globalParams.isEmpty()) && (!globalParams.get(i).isEmpty())) {
                        effectiveParams.putAll(globalParams.get(i));
                    }
                }
            } catch (Exception ex) {}
            effectiveParams.putAll(localParams);
            
            switch (commandToProcess) {
                
                case "Progress": 
                    param = checkParameter("UserName");
                    if (!Utils.testEmptyString(param)) {
                        param = param.toUpperCase();
                        param2 = checkParameter("ProcessID");
                        if (!Utils.testEmptyString(param2)) {
                            SharedCommands cmds;
                            if (commands.containsKey(param)) {
                                cmds = commands.get(param);
                            } else {
                                cmds = new SharedCommands();
                                commands.put(param, cmds);
                            }
                            cmds.setRetry(Defaults.FAST_POLL_INTERVAL);
                            SharedCommand cmd = new SharedCommand();
                            cmds.addCommand(cmd);
                            cmd.setName("progress");
                            cmd.addParam("id", param2);
                            param = checkParameter("Title");
                            if (!Utils.testEmptyString(param)) {
                                cmd.addParam("title", param);
                            }
                            param = checkParameter("PercentText");
                            if (!Utils.testEmptyString(param)) {
                                cmd.addParam("ptext", param);
                            }
                            param = checkParameter("WorkText");
                            if (!Utils.testEmptyString(param)) {
                                cmd.addParam("wtext", param);
                            }
                            param = checkParameter("Percent");
                            if (!Utils.testEmptyString(param)) {
                                cmd.addParam("percent", param);
                            }
                        }    
                    }
                    break;
                    
            }
         
            localParams.clear();
            commandToProcess = null; 
        }
        
    }
    
    public SharedCommands getDefaultCommands () {
        return defaultCommands;
    }
    
    public SharedCommands getCommands (String user) {
        if (!commands.isEmpty()) {
            if (commands.containsKey(user)) {
                return commands.get(user);
            }    
        }
        return getDefaultCommands();
    }

}
