/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.telecom.webui.core;

import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonParser;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Locale;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import ru.fors.telecom.webui.error.UiException;
import ru.fors.telecom.webui.common.Defaults;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import javax.servlet.ServletContext;
import javax.servlet.annotation.MultipartConfig;
import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import ru.fors.telecom.webui.common.ClientConfigEntry;
import ru.fors.telecom.webui.common.ServerConfigEntry;
import ru.fors.telecom.webui.common.SharedConfig;
import ru.fors.telecom.webui.common.Utils;
import ru.fors.telecom.webui.error.UiError;


/**
 *
 * @author john
 */
@MultipartConfig
public class WebUI extends HttpServlet {

    private final Dispatcher dispatcher;
    private final ConcurrentHashMap<String, WorkArea> workAreas;
    private final ArrayList<SharedArea> sharedAreas;
    private final ArrayList<ServerConfigEntry> serverConfig;
    private final ArrayList<ClientConfigEntry> clientConfig;
    private final SharedConfig sharedConfig;
    private final HashMap<String, Object> clientParams;
    private ArrayList<TestLauncher> launchers;
    private Integer protocol = 1;
    private String sockAddress = "";
    private String tnsAdmin = "";
    private String sockPort = "";
    private String msgPort = "";
    private boolean isFastcom = false;
    private ServerSocket msgServer;
    private BufferedReader msgReader;
    private Socket msgSocket;
    private boolean uiRunning = true;
    private Driver oracleDriver;
    private java.util.function.Function<? extends Object, ? extends Object> localFabric = null; //++BUNDLE--
    private final static String WA_NOT_FOUND = "Application data not found!\nTry to restart the application.";
    
    public WebUI() {
        this.dispatcher = new Dispatcher();
        this.workAreas = new ConcurrentHashMap<>();
        this.sharedAreas = new ArrayList<>();
        this.sharedConfig = new SharedConfig();
        this.clientConfig = new ArrayList<>();
        this.serverConfig = new ArrayList<>();
        this.clientParams = new HashMap<>();
    }

    @Override
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        Locale.setDefault(new Locale("ru", "RU"));
        // ++BUNDLE
        /*Object aLUIInternal = config.getServletContext().getAttribute("LUIInternal");
        if(aLUIInternal != null) {
            if(aLUIInternal instanceof Function) {
                try {
                    localFabric =  (Function<? extends Object, ? extends Object>) aLUIInternal;
                    protocol = 2;
                    sockAddress = null;
                    sockPort = null;
                    clientParams.put("protocol", protocol);
                    clientParams.put("name", config.getServletContext().getServletContextName());
                    sharedConfig.setTempDir(System.getProperty("java.io.tmpdir"));
                    return;
                }
                catch(Exception ignore) {}
            }
        }*/
        // --BUNDLE
        //++ Битва за нормальный конфиг.
        File configFile = null;
        String aConfigPath = config.getInitParameter("lui-config");
        if(aConfigPath!=null) {
            configFile = new File( config.getServletContext().getRealPath(aConfigPath) );
            if(!configFile.exists()) {
                configFile = new File( aConfigPath);
            }
        }
        if(configFile == null || !configFile.exists()) {
            /*File  //--Битва за нормальный конфиг*/ configFile = new File(Defaults.CONFIG_FILE_UNIX);
        }
        //File configFile = new File(Defaults.CONFIG_FILE_UNIX);
        if (!configFile.exists()) {
            configFile = new File(Defaults.CONFIG_FILE_WINDOWS);
            if (!configFile.exists()) {
                configFile = new File(System.getProperty("user.home") + Defaults.CONFIG_FILE);
                if (!configFile.exists()) {
                    System.out.println("Application configuration file not found");
                }
            }    
        }
        SAXBuilder builder = new SAXBuilder();
        Document configDoc = null;
        try {
            configDoc = builder.build(configFile);
        } catch (JDOMException | IOException ex) {
            throw new ServletException(
                    String.format(
                            "Failed to load configuration file \'%s\': %s %s"
                          , configFile
                          , ex.getClass().getCanonicalName()
                          , ex.getMessage()));
        }
        
        if(configDoc==null) {
            throw new ServletException(String.format("Configuration file \'%s\' not loaded", configFile));
        }

        Element rootNode =  configDoc.getRootElement();
        String connect = Utils.string2String(rootNode.getChildText("connect"));
        if (connect.equals("socket")) {
            protocol = 2;
            Element socket = rootNode.getChild("socket");
            sockAddress = socket.getChildText("address");
            sockPort = socket.getChildText("port");
            msgPort = socket.getChildText("msgport");
        } else {
            tnsAdmin = Utils.string2String(rootNode.getChildText("tnsadmin"));
            //if (Utils.testEmptyString(tnsAdmin)) {
            //    tnsAdmin = System.getenv("TNS_ADMIN");
            //}
        }
        clientParams.put("protocol", protocol);
        clientParams.put("name", config.getServletContext().getServletContextName());
        if (protocol == 1) {
            try {
                oracleDriver = new oracle.jdbc.OracleDriver();
                DriverManager.registerDriver(oracleDriver);
                System.out.println("Oracle driver registered");
            } catch (SQLException ex) {
                System.out.println("Error registering Oracle JDBC Driver");
            }
            isFastcom = Utils.testYString(Utils.string2String(rootNode.getChildText("fastcom")));
            String mobileServer = Utils.string2String(rootNode.getChildText("mobile-server"));
            String autoLogon = Utils.string2String(rootNode.getChildText("autologon"));
            clientParams.put("autologon", autoLogon);
            List<Element> serversList = rootNode.getChildren("server");
            for (int i = 0; i < serversList.size(); i++) {
                Element node = serversList.get(i);
                ServerConfigEntry serverConfigEntry = new ServerConfigEntry();
                ClientConfigEntry clientConfigEntry = new ClientConfigEntry();
                String alias = Utils.string2String(node.getChildText("alias"));
                clientConfigEntry.setAlias(alias);
                clientConfigEntry.setUserName(Utils.string2String(node.getChildText("username")));
                clientConfigEntry.setPassword(Utils.string2String(node.getChildText("password")));
                clientConfigEntry.setDefault(!Utils.testEmptyString(node.getChildText("default")));
                String description = Utils.string2String(node.getChildText("description")).replaceAll(" ", "").replaceAll("\n", "").replaceAll("\r", "").replaceAll("\t", "");
                serverConfigEntry.setDescription(description);
                if (alias.equals(mobileServer)) {
                    sharedConfig.setMobileDescriptor(description);
                    sharedConfig.setMobileIndex(i);
                }
                Element userlistRoot = node.getChild("userlist");
                if (userlistRoot != null) {
                    String group = "";
                    try {
                        group = userlistRoot.getChildText("group");
                    } catch (Exception ex) {
                    }
                    Integer valid = 0;
                    try {
                        valid = Integer.parseInt(userlistRoot.getChildText("valid"));
                    } catch (NumberFormatException ex) {
                    }
                    serverConfigEntry.setLoadUserList(true);
                    serverConfigEntry.setUserListGroup(group);
                    serverConfigEntry.setUserListValid(valid);
                }
                serverConfig.add(serverConfigEntry);
                clientConfig.add(clientConfigEntry);
                Element commonRoot = node.getChild("common");
                if (commonRoot == null) {
                    sharedAreas.add(new SharedArea(null, null, null, null));
                } else {
                    SharedArea sharedArea = new SharedArea(description, tnsAdmin, Utils.string2String(commonRoot.getChildText("username")), Utils.string2String(commonRoot.getChildText("password")));
                    sharedAreas.add(sharedArea);
                    sharedArea.startQuery();
                }
            }
            clientParams.put("config", clientConfig);
        } else if (protocol == 2) {
            startMessageServer();
        }
        sharedConfig.setTempDir(Utils.string2String(rootNode.getChildText("tempdir")));
    }

    @Override
    public void destroy() {
        super.destroy();
        if (protocol == 1) {
            try {
                DriverManager.deregisterDriver(oracleDriver);
                System.out.println("Oracle driver deregistered");
            } catch (SQLException ex) {
                System.out.println("Error deregistering Oracle JDBC driver");
            }
        } else {
            try {
                msgServer.close();
            } catch (IOException ex) {}
            uiRunning = false;
        }
    }
    
    /**
     * Processes requests for both HTTP <code>GET</code> and <code>POST</code>
     * methods.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    protected void processRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {

    }

    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        try {
            // Alive packet
            String pollId = request.getParameter("poll");
            if (pollId != null) {
                WorkArea wa = workAreas.get(pollId);
                if (wa == null) {
                    UiError.raise(WA_NOT_FOUND);
                } else {
                    Gson gson = new Gson();
                    sendResponse(response, Defaults.HTTP_OK, gson.toJson(wa.procPolling()));
                }
            }
        } catch (UiException e) {
            sendResponse(response, Defaults.HTTP_INTERNAL_SERVER_ERROR, e.getMessage());
        }
    }
    
    /**
     * Handles the HTTP <code>POST</code> method.
     *
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {

        final HttpServletRequest req = request;
        final HttpServletResponse resp = response;
        Gson gson = new Gson();
        String sessionId;
        WorkArea workArea = null;
        
        /*String authUrl = "https://mvd-vps-dev-app.fors.ru/auth"; // todo: поменять при необходимости
        String clientId = "atis"; // todo: поменять при необходимости
        String configJson =
            "{\n" +
            "  \"realm\": \"vps\",\n" +
            "  \"auth-server-url\": \"" + authUrl + "\",\n" +
            "  \"ssl-required\": \"external\",\n" +
            "  \"resource\": \"" + clientId + "\",\n" +
            "  \"public-client\": true,\n" +
            "  \"confidential-port\": 0\n" +
            "}";
        KeycloakDeployment deployment = KeycloakDeploymentBuilder.build(new ByteArrayInputStream(configJson.getBytes(StandardCharsets.UTF_8)));
        String code = "code";
        String redirectUri = "http://localhost:8082/sso";
        AccessTokenResponse tokenResponse;
        try {
            tokenResponse = ServerRequest.invokeAccessCodeToToken(deployment, code, redirectUri, null);
            AdapterTokenVerifier.VerifiedTokens tokens = AdapterTokenVerifier.verifyTokens(tokenResponse.getToken(), tokenResponse.getIdToken(), deployment);
            AccessToken accessToken = tokens.getAccessToken();
            IDToken idToken = tokens.getIdToken();
            System.out.println("Access token: " + accessToken.getPreferredUsername());
            System.out.println("ID token: " + idToken.getPreferredUsername());
        } catch (ServerRequest.HttpFailure ex) {
            Logger.getLogger(WebUI.class.getName()).log(Level.SEVERE, null, ex);
        } catch (VerificationException ex) {
            Logger.getLogger(WebUI.class.getName()).log(Level.SEVERE, null, ex);
        }
        String state = UUID.randomUUID().toString();
        KeycloakUriBuilder builder = deployment.getAuthUrl().clone()
                .queryParam(OAuth2Constants.RESPONSE_TYPE, OAuth2Constants.CODE)
                .queryParam(OAuth2Constants.CLIENT_ID, deployment.getResourceName())
                .queryParam(OAuth2Constants.REDIRECT_URI, redirectUri)
                .queryParam(OAuth2Constants.STATE, state)
                .queryParam(OAuth2Constants.SCOPE, OAuth2Constants.SCOPE_OPENID);
        resp.sendRedirect(builder.build().toString());*/
        
        try {
            if (ServletFileUpload.isMultipartContent(req)) {

                // Если загружаем или выгружаем файлы
                File repository = new File(sharedConfig.getTempDir());
                DiskFileItemFactory factory = new DiskFileItemFactory(DiskFileItemFactory.DEFAULT_SIZE_THRESHOLD, repository);
                ServletFileUpload servletUpload = new ServletFileUpload(factory);
                String reqAction = "download";
                String reqToken;
                
                req.setCharacterEncoding("UTF-8");
                try {
                    List<FileItem> fileItemList = servletUpload.parseRequest(req);
                    for (int i = 0; i < fileItemList.size(); i++) {
                        FileItem fileItem = fileItemList.get(i);
                        if (fileItem.isFormField()) {
                            switch (fileItem.getFieldName()) {
                                case "req-action":
                                    reqAction = fileItem.getString();
                                    break;
                                case "req-token":
                                    reqToken = fileItem.getString();
                                    if (!Utils.testEmptyString(reqToken)) {
                                        workArea = workAreas.get(reqToken);
                                        if (workArea == null) {
                                            UiError.raise(WA_NOT_FOUND);
                                        }
                                    }
                                    break;
                                default:
                                    break;
                            }
                        } else {
                            if (reqAction.equals("upload")) {
                                String file = fileItem.getName();
                                try (InputStream stream = fileItem.getInputStream()) {
                                    if ((file != null) && (workArea != null)) {
                                        workArea.uploadFile(file, stream);
                                    }
                                }
                            }
                        }
                    }
                    if (workArea != null) {
                        if (reqAction.equals("download")) {
                            InputStream inStream;
                            inStream = workArea.getDownloadStream();
                            if (inStream != null) {
                                String fn;
                                fn = workArea.getDownloadFilename();
                                if (!Utils.testEmptyString(fn)) {
                                    resp.setContentType("application/octet-stream");
                                    //URI uri;
                                    //uri = new URI(null, null, fn, null);
                                    //resp.setHeader("Content-Disposition", "attachment; filename*=UTF-8''" + uri.toASCIIString());
                                    String fnEncoded = URLEncoder.encode(fn, "UTF-8");
                                    resp.setHeader("Content-Disposition", "attachment; filename*=UTF-8''" + fnEncoded);
                                    
                                    try (OutputStream outStream = resp.getOutputStream()) {
                                        byte[] buffer = new byte[4096];
                                        int bytesRead;
                                        while ((bytesRead = inStream.read(buffer)) != -1) {
                                            outStream.write(buffer, 0, bytesRead);
                                        }
                                    }
                                }
                                inStream.close();
                            }
                            workArea.downloadFinish();
                        } else if (reqAction.equals("upload")) {
                            workArea.uploadFinish();
                            sendResponse(resp, Defaults.HTTP_OK, gson.toJson(workArea.getInstructions()));
                        }
                    }

                } catch (FileUploadException ex) {
                    UiError.raise("An error occurred while uploading the file.\nError code (FileUploadException)\n" + ex.getMessage());
                } catch (SQLException ex) {
                    UiError.raise("An error occurred while uploading the file.\nCan not open stream for reading.\nError code (SQLException)\n" + ex.getMessage());
                } //catch (URISyntaxException ex) {
                    //UiError.raise("An error occurred while uploading the file.\nError code (URISyntaxException)\n" + ex.getMessage());
                //}

            } else {
                StringBuilder builder = new StringBuilder();
                String line;
                BufferedReader reader = req.getReader();
                while ((line = reader.readLine()) != null) {
                    builder.append(line);
                }
                String reqData = builder.toString();
                JsonElement msgElement = new JsonParser().parse(reqData);
                JsonElement msgTypeElement = msgElement.getAsJsonObject().get("mt");
                JsonElement paramElement = null;
                String messageType = msgTypeElement.getAsString();

                if (null != messageType) {
                    switch (messageType) {
                        case "C":
                            WorkArea wa = null;
                            paramElement = msgElement.getAsJsonObject().get("par");
                            JsonElement mobileElement = paramElement.getAsJsonObject().get("mobile");
                            boolean isMobile = mobileElement.getAsBoolean();
                            JsonElement userElement = null;
                            JsonElement passwordElement = null;
                            JsonElement serverElement = null;
                            if (protocol == 1) {
                                userElement = paramElement.getAsJsonObject().get("user");
                                passwordElement = paramElement.getAsJsonObject().get("password");
                                serverElement = null;
                                if (!isMobile) {
                                    serverElement = paramElement.getAsJsonObject().get("server");
                                }
                            }
                            String remoteHost;
                            String remoteAddr = Utils.string2String(req.getRemoteAddr());
                            remoteHost = req.getHeader("X-FORWARDED-FOR");
                            if (Utils.testEmptyString(remoteHost)) {
                                remoteHost = req.getHeader("X_FORWARDED_FOR");
                                if (Utils.testEmptyString(remoteHost)) {
                                    remoteHost = req.getRemoteHost();
                                    if (Utils.testEmptyString(remoteHost)) {
                                        remoteHost = req.getRemoteAddr();
                                        if (Utils.testEmptyString(remoteHost)) {
                                            remoteHost = "unknown.host";
                                        }
                                    }
                                }
                            }
                            if (remoteHost.length() > 30) {
                                remoteHost = remoteHost.substring(0, 29);
                            }
                            if (protocol == 1) {
                                java.util.Properties connectionProperties = new java.util.Properties();
                                String userName = userElement.getAsString();
                                connectionProperties.setProperty("user", userName);
                                connectionProperties.setProperty("password", passwordElement.getAsString());
                                connectionProperties.put("v$session.terminal", remoteHost);
                                connectionProperties.put("v$session.program", "WEBUI");
                                String connectionDescriptor;
                                Connection connection;
                                int serverIndex = 0;
                                if (isMobile) {
                                    serverIndex = sharedConfig.getMobileIndex();
                                    connectionDescriptor = sharedConfig.getMobileDescriptor();
                                } else {
                                    serverIndex = serverElement.getAsInt();
                                    connectionDescriptor = serverConfig.get(serverIndex).getDescription();
                                }
                                try {
                                    if (Utils.testEmptyString(tnsAdmin)) {
                                        connection = DriverManager.getConnection("jdbc:oracle:thin:@" + connectionDescriptor, connectionProperties);
                                    } else {
                                        connection = DriverManager.getConnection("jdbc:oracle:thin:@" + connectionDescriptor + "?TNS_ADMIN=" + tnsAdmin, connectionProperties);
                                    }
                                    connection.setAutoCommit(false);
                                    wa = new WorkArea(userName, connection, dispatcher, sharedAreas.get(serverIndex), remoteAddr, isMobile, isFastcom);
                                } catch (SQLException ex) {
                                    ex.printStackTrace();
                                    UiError.raise(ex.getMessage());
                                }
                            } else if (protocol == 2) {
                                InetAddress address;
                                //++BUNDLE
                                if (sockAddress != null) {
                                    //--BUNDLE
                                    try {
                                        address = InetAddress.getByName(sockAddress);
                                        Integer port;
                                        port = Integer.parseInt(sockPort);
                                        Socket socket = new Socket(address, port);
                                        socket.setKeepAlive(true);
                                        wa = new WorkArea(socket, dispatcher, msgPort, remoteHost, isMobile);
                                        wa.setTempDir(sharedConfig.getTempDir());
                                    } catch (Exception ex) {
                                        UiError.raise("Error opening socket to server " + sockAddress + " on port " + sockPort);
                                    }
                                    //++BUNDLE
                                } else if (localFabric != null) {
                                    try {
                                        wa = new WorkArea(localFabric, dispatcher, remoteHost, isMobile);
                                    } catch (Exception ex) {
                                        UiError.raise("Error opening local connection: " + ex.getMessage());
                                    }
                                } else {
                                    UiError.raise("Bad Servlet configuration");
                                }
                                //--BUNDLE
                            }
                            if (wa == null) {
                                UiError.raise("Error creating workarea");
                            } else {
                                workAreas.put(wa.getToken(), wa);
                                sendResponse(resp, Defaults.HTTP_OK, gson.toJson(wa.getInstructions()));
                            }
                            
                            break;
                        case "G":
                            sendResponse(resp, Defaults.HTTP_OK, gson.toJson(clientParams));
                            break; 
                        case "T":
                            paramElement = msgElement.getAsJsonObject().get("par");
                            JsonElement appcodeElement = paramElement.getAsJsonObject().get("appcode");
                            JsonElement threadsElement = paramElement.getAsJsonObject().get("threads");
                            try {
                                String appCode = appcodeElement.getAsString();
                                Integer threads;
                                threads = Integer.parseInt(threadsElement.getAsString());
                                startTest(appCode, threads);
                            } catch (Exception ex) {}
                            sendResponse(resp, Defaults.HTTP_OK, gson.toJson(Utils.nop()));
                            break;   
                        case "_T":
                            for (int i = 0; i < launchers.size(); i++) {
                                System.out.println("Stopping thread #" + launchers.get(i).getID());
                                try {
                                    launchers.get(i).stop();
                                } catch (Exception ex) {};
                                //try {
                                //    Thread.currentThread().sleep(200);
                                //} catch (InterruptedException ex) {}
                            }
                            sendResponse(resp, Defaults.HTTP_OK, gson.toJson(Utils.nop()));
                            break;    
                        default:
                            JsonElement sessionIdElement = msgElement.getAsJsonObject().get("sid");
                            sessionId = sessionIdElement.getAsString();
                            workArea = workAreas.get(sessionId);
                            if (workArea == null) {
                                UiError.raise(WA_NOT_FOUND);
                            } else {
                                switch (messageType) {
                                    case "E": // Обычное событие
                                        //Date startDate = new Date();
                                        workArea.processEvent(msgElement);
                                        //Date finishDate = new Date();
                                        //System.out.println("Event processing> "+(finishDate.getTime()-startDate.getTime()));
                                        //startDate = new Date();
                                        gson.toJson(workArea.getInstructions());
                                        //finishDate = new Date();
                                        //System.out.println("Convert to json> "+(finishDate.getTime()-startDate.getTime()));
                                        sendResponse(resp, Defaults.HTTP_OK, gson.toJson(workArea.getInstructions()));
                                        
                                        break;

                                                                                                           
                                    case "Z": // Пользователь завершил работу
                                        workArea.close();
                                        sendResponse(resp, Defaults.HTTP_OK, "");

                                        break;

                                    default: {
                                        UiError.raise("Unknown request type: " + messageType);
                                    }
                                }
                                
                            }                        
                            break;
                    }
                } else {
                    UiError.raise("Request type is not defined");
                }

            }
            
        } catch (UiException e) {
            sendResponse(resp, Defaults.HTTP_INTERNAL_SERVER_ERROR, e.getMessage());
        }
    }
    
    private void sendResponse(HttpServletResponse response, int status, String message) throws IOException {
        response.setContentType("application/json;charset=UTF-8");
        response.setStatus(status);
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        PrintWriter out = new PrintWriter(new OutputStreamWriter(bout, "UTF-8"));
        out.print(message);
        out.flush();
        response.setContentLength(bout.size());
        bout.writeTo(response.getOutputStream());
        bout.flush();
    }
    
    public class Dispatcher {
        
        public void rm(String token) {
            Timer rmTimer = new Timer();
            RmTask rmTask = new RmTask(token);
            rmTimer.schedule(rmTask, 20000);
        }
        
        public class RmTask extends TimerTask {

            String token;
            
            public RmTask(String token) {
                this.token = token;
            }
            
            @Override
            public void run() {
                workAreas.remove(token);
                //System.out.println("Workarea " + token + " closed");
            }
            
        }
        
    }
    
    public void startMessageServer () {
        MessageServer ms = new MessageServer();
        Thread thread = new Thread(ms);
        thread.setDaemon(true);
        thread.start();
    }
    
    public class MessageServer implements Runnable {

        @Override
        public void run() {
            try {
                Integer i = Integer.parseInt(msgPort);
                if (i > 0) {
                    msgServer = new ServerSocket(i);
                    while (uiRunning) {
                        try {
                            msgSocket = msgServer.accept();
                            msgSocket.setKeepAlive(true);
                            msgSocket.setSoTimeout(500);
                            //msgReader = new BufferedReader(new InputStreamReader(msgSocket.getInputStream(), "Windows-1251"));
                            msgReader = new BufferedReader(new InputStreamReader(msgSocket.getInputStream(), "UTF8"));
                            String line;
                            String sessionId = null;
                            while ((line = msgReader.readLine()) != null) {
                                if (line.equals("")) {
                                    break;
                                } else {
                                    sessionId = line;
                                }
                            }
                            WorkArea wa = workAreas.get(sessionId);
                            if (wa == null) {
                                UiError.raise(WA_NOT_FOUND);
                            } else {
                                msgSocket.setSoTimeout(0);
                                MessageHandler msgHandler = new MessageHandler(msgSocket, msgReader);
                                Thread thread = new Thread(msgHandler);
                                thread.setDaemon(true);
                                thread.start();
                                wa.setMessageHandler(msgHandler);
                            }
                        } catch (Exception ex) {
                            try {
                                msgReader.close();
                                msgSocket.close();
                            } catch (Exception e) {
                            }
                        }
                    }
                }
            } catch (IOException ex) {
                System.out.println("Error open port " + msgPort + " for listening");
            }
            System.out.println("Messages server shutdown");
        }
        
    }
    
    public void startTest(String app, int threads) {
        launchers = new ArrayList<>();
        for (int i = 0; i < threads; i++) {
            TestLauncher launcher = new TestLauncher(i, app);
            launchers.add(launcher);
            Thread thread = new Thread(launcher);
            thread.setDaemon(true);
            thread.start();
            //try {
            //    Thread.currentThread().sleep(100);
            //} catch (InterruptedException ex) {}
        }
       System.out.println("------ Test started (" + launchers.size() + " threads) ------");
    }
    
    public class TestLauncher implements Runnable {
        
        private final int threadID;
        private final String appCode;
        private WorkArea wa;

        public TestLauncher(int id, String app) {
            threadID = id;
            appCode = app;
        }
        
        public void stop() {
            try {
                wa.stopTest();
            } catch (Exception ex) {}    
        }
        
        public int getID() {
            return threadID;
        }
        
        @Override
        public void run() {
            try {
                InetAddress address = InetAddress.getByName(sockAddress);
                Integer port;
                port = Integer.parseInt(sockPort);
                Socket socket = new Socket(address, port);
                socket.setKeepAlive(true);
                wa = new WorkArea(socket, dispatcher, msgPort, "stresstest#" + threadID, false);
                wa.setTempDir(sharedConfig.getTempDir());
                System.out.println("Start thread #" + threadID + "   workarea > " + wa.getToken());
                wa.startTest(appCode);
            } catch (Exception ex) {
                System.out.println("Thread #" + threadID + " run failed");
            } finally {
                System.out.println("Finish thread #" + threadID);
            }    
        }
        
    }

}
