/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.telecom.webui.menu;

import java.util.ArrayList;
import org.apache.commons.lang3.StringUtils;
import ru.fors.telecom.webui.common.Utils;

/**
 *
 * @author john
 */
public class UiMenuObject {

    private String id = null;
    protected String uid = null;
    protected String type = "";
    protected String role = "";
    protected String formId = "";
    protected String code = "";
    protected String params = "";
    protected boolean enabled = true;
    protected String label = "";
    protected String style = "";
    protected String parent = " parent='_root_'";
    private String group = null;
    private String firstIcon = "";
    protected ArrayList<UiMenuItemContent> content = new ArrayList<>();

    public UiMenuObject() {
        uid = Utils.generateId();
    }

    public UiMenuObject(String id) {
        setId(id);
        uid = Utils.generateId();
    }

    public UiMenuObject(String id, String label) {
        setId(id);
        uid = Utils.generateId();
        setLabel(label);
    }
    
    public String getUid() {
        return uid;
    }

    public final void setId(String id) {
        this.id = id;
    }

    public String getId() {
        return id;
    }

    public void setType(String type) {
        this.type = Utils.string2String(type);
    }
    
    public String getType() {
        return type;
    }

    public void setRole(String role) {
        this.role = Utils.string2String(role);
    }

    public void setCode(String code) {
        this.code = Utils.string2String(code);
    }
    
    public String getCode() {
        return code;
    }

    public void setParams(String params) {
        this.params = Utils.string2String(params);
    }
    
    public String getParams() {
        return params;
    }
    
    public void setParent(String parent) {
        this.parent = " parent='" + parent + "'";
    }

    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
    }

    public void setFormId(String formId) {
        this.formId = Utils.string2String(formId);
    }

    public String getFormId() {
        return formId;
    }
    
    public String getFirstIcon() {
        if (Utils.testEmptyString(firstIcon)) {
            return "";
        } else {
            return "ico-" + firstIcon;
        }
    }
    
    public final void setLabel(String label) {
        this.label = Utils.string2String(label);
        String tmp = label;
        if (StringUtils.startsWith(tmp, "[ico:")) {
            firstIcon = StringUtils.substringAfter(StringUtils.substringBefore(tmp, "]"), "[ico:");
            tmp = StringUtils.substringAfter(tmp, "]");
        }
        while (StringUtils.contains(tmp, "[ico:")) {
            String txt = StringUtils.substringBefore(tmp, "[ico:");
            content.add(new UiMenuItemText(txt));
            tmp = StringUtils.removeStart(tmp, txt);
            content.add(new UiMenuItemIcon(StringUtils.substringAfter(StringUtils.substringBefore(tmp, "]"), "[ico:")));
            tmp = StringUtils.substringAfter(tmp, "]");
        }
        if (!Utils.testEmptyString(tmp)) {
            content.add(new UiMenuItemText(tmp));
        }
    }

    public void setStyle(String style) {
        this.style = " " + Utils.string2String(style);
    }

    public void setGroup(String group) {
        this.group = group;
    }

    public String getGroup() {
        return group;
    }

}
