/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

/*
 * Класс для хранения данных о форме
 */
package luiserver;


import LuiCollection.LuiCollection;
import LuiCollection.LuiString;
import static LuiCollection.LuiString.Extract_Word;
import static LuiCollection.LuiString.Filter;

import static LuiCollection.LuiString.LTrim;
import static LuiCollection.LuiString.NVL;
import static LuiCollection.LuiString.RPad;
import static LuiCollection.LuiString.RTrim;
import static LuiCollection.LuiString.RemoveIco;
import static LuiCollection.LuiString.String2Integer;
import static LuiCollection.LuiString.compareString;
import static LuiCollection.LuiString.is_not_null;
import static LuiCollection.LuiString.is_null;
import static LuiCollection.LuiString.like;
import static LuiCollection.LuiString.withCapitalLetter;
import LuiDbInterface.ColInfo;
import LuiDbInterface.LuiDbConn;
import LuiExceptionMsg.LuiException;
import LuiNLS.LuiLang;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import java.nio.charset.Charset;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Matcher;

import java.util.regex.Pattern;
import javax.script.ScriptException;
import org.jsoup.Jsoup;


/**
 *
 * @author aliubushkin
 */
public class LuiAppForm extends LuiAppWFE{
public static final String CURSORSIGN="¦";
public static final String OFF$SET="off$set";
private static final String STMT4LOADFORMCONF="select name FROM lui_t_config  where app_code='<app_code>' and form_code='<form_code>' and code='<config_code>'";    
//private static final String STMT4LOADFORMPROP="select p.code,p.def_value from lui_t_property p,lui_t_property_formtype pp where pp.property_code=p.code and pp.formtype_code='<formtype_code>' and p.etype='FORM' order by p.seq";
//                                                      0    1   2     3     4    
private static final String STMT4LOADFORMELEM="select code,name,etype,seq,element_code from lui_t_element where app_code='<app_code>' and form_code='<form_code>' order by seq";
// 0=code 1-name 2-etype 3-seq 4-parent

//private HashMap<String,LuiAppElement> cellActions=null; // массив действий ячеек по time-обновлению ячеек 
private String defIfSecond="";
public LuiApplication  application;
public HashMap<String, LuiAppElement> elements; // элементы формы. Свойства самой формы хранятся в элементе с кодом "Form"
public ArrayList<LuiAppElement> elSortedList;// упорядоченный список элементов
public ArrayList<LuiAppElement> el4order; // упорядоченный перечень кодов эелементов для построения Order by
//public HashMap<String,String> rowQBE; //строка QBE
public HashMap<String,HashMap<String,String>> listQBE; // список строк QBE
private String lastCurrentField; //пока не ясно зачем хранить код последнего текущего поля
public LinkedHashMap<LuiCollection,Boolean> checkRows; //отмеченные строки
private boolean changeTotal=false;
public String code="";
public   jdk.nashorn.api.scripting.ScriptObjectMirror JSElements; // Массив элементов возможно изменяемых в JavaScript
public jdk.nashorn.api.scripting.ScriptObjectMirror JSFormProps; //Массив свойств формы возможно изменяемых в JavaScript
public LuiSQLCondition SQLCondition; // публичный класс для использования в бланке LUI_B_QBE
public LuiRTdashboard RTdashboard;
public Thread RTrunThread;
public LuiDgV2 DgV2;
public ArrayList<LuiCollection> collapsedDGLevels=new ArrayList<>();; // список свёрнутых уровней в группировке данных LUI_L_DATA_GROUPING_SHOW
public HashMap<String,ArrayList<String>> MatrixMap;//Массив кодов ячеек матрицы адресуемых по Кодстроки:Кодстолбца т.е. в каждой клетке стопка
                                                    // а в стопке верхним является элемент с последним номером
    /**
     * Конструктор формы
     *
     * @param pOwnerObj - хозяин (applicatio
     * @param pCode - Код ("&lt;придумай&gt;" - если форма программная,как в
     * метаданных если форма метаданковая)
     * @param pType - Тип (BLANK или LIST или TREE)
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public LuiAppForm(LuiAppWFE pOwnerObj,String pCode,String pType) throws LuiException, SQLException{
    ownerObj=pOwnerObj;
    code=pCode;
    application=(LuiApplication)ownerObj;
    elements=new HashMap<>();
    elSortedList=new ArrayList<>();
    el4order=new ArrayList<>();

    listQBE=new HashMap<>();
     checkRows=new LinkedHashMap<>();
    
    setProperty("FType",pType);
    setProperty("Parent","Application"); 
    addProperty("Code",pCode);
    addProperty("FEATURE",pCode);
    setVal("UserConfiguration","");
    setVal("UserConfigurationName","");
    setVal("UsrConfList","");
    setChangeFlag();
    
}

public String getCode() throws LuiException, SQLException{
    //return getProperty(`  );
    return this.code;
}
public void setUsrConfigListProperty() throws SQLException, LuiException{
    LuiCollection confColl = new LuiCollection("");
    if (is_not_null(application.getAppUser())){
    LuiDbConn lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
    try{
        confColl=lc.getSelectAllRowsTwoColsValMap("select id,name from lui_t_user_config where parent_app='"+
                application.code+"' and parent_form='"+
                //getCode()+"' and parent_config='"+
                getProperty("FEATURE")+"' and parent_config='"+
                getConfigCode()+"' and user_name='"+
                application.getAppUser()+"' order by name");
        
    } finally{
        application.metaConnPool.SetFree(lc);
    }
    }
    if (confColl==null){
        setVal("UsrConfList","");
        setVal("UsrConfNames","");
    }
    else{    setVal("UsrConfList",confColl.toString());
     LuiCollection names=new LuiCollection("");
     for (String id:confColl.keySet()){
         names.Add_Element(confColl.Get_Element(id),confColl.Get_Element(id));
     }
     setVal("UsrConfNames",names.toString());
    }
}
public String getConfigCode() throws LuiException, SQLException{
    return getNonCompValue("ConfigCode");
}
public String getType() throws LuiException, SQLException{
    return this.getNonCompValue("FType");
}
public String getFormID() throws LuiException, SQLException{
    return getNonCompValue("FormID");
}
public String getParentID() throws LuiException, SQLException{
    return getVal("ParentFormID");
}

public String getTypeLB() throws LuiException, SQLException{
    String t=getType();
    if (t.equals("TREE")) return "LIST";
    return t;
}

    /**
     * Создание элемента формы и добавление его в массив элементов
     *
     * @param pCode -код элемента
     * @param pType -тип элемента (!!!если в бланк будут добавляться элементы
     * типа ITEM, то они превратятся в FIELD)
     * @param pSeq - порядковый номер
     * @param pName - название
     * @param pParent - код старшего элемента если null или "", то будет "Form"
     * @return - элемент (LuiAppElement)
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public LuiAppElement addElement(String pCode,String pType,Double pSeq,String pName,String pParent) throws LuiException, SQLException{
  LuiAppElement el;
  if (this.getType().equals("BLANK") && pType.equals("ITEM")) {
     pType="FIELD";}
    if (!elements.containsKey(pCode)){
     if (pSeq==null) pSeq=getNewSeq(pType);
     el=new LuiAppElement(this,pType,pCode,pSeq);
     if (pType.equals("FIELD")) {
     el.addProperty("Value","");
      }
      elements.put(pCode, el);
     //добавляем стандартных свойств
     // устанавливаем свойства по умолчанию
   if (application.formTemplates!=null){
   if (application.formTemplates.containsKey(this.getTypeLB())){
     LuiAppForm ft=application.formTemplates.get(this.getTypeLB());
     if (application.formTemplates.get(getTypeLB()).elements.containsKey(pType)) {
    for(String propcode:application.formTemplates.get(getTypeLB()).elements.get(pType).properties.keySet()){
      //  if(application.formTemplates.get(getTypeLB()).elements.get(pType).properties.get(propcode).is_meta){
      if(application.formTemplates.get(getTypeLB()).elements.get(pType).properties.get(propcode).genesis!='A'){
        String pv=application.formTemplates.get(getTypeLB()).elements.get(pType).getNonCompValue(propcode);
        LuiAppProp p=el.addProperty(propcode,pv);
        p.genesis='D';
        
        }
    }
   }
   }
    }
    el.addProperty("Code",pCode);
    el.addProperty("EType",pType);
    el.addProperty("Name",pName);
    el.addProperty("Parent", NVL(pParent,"Form"));
    el.addProperty("Seq",String.valueOf(pSeq));
     setChangeFlag();
  } else {
        el=elements.get(pCode);
        if (pSeq==null)pSeq=el.sourceSeq;
        if (application.formTemplates!=null){
   if (application.formTemplates.containsKey(this.getTypeLB())){
     LuiAppForm ft=application.formTemplates.get(this.getTypeLB());
     if (application.formTemplates.get(getTypeLB()).elements.containsKey(pType)) {
    for(String propcode:application.formTemplates.get(getTypeLB()).elements.get(pType).properties.keySet()){
        //if(application.formTemplates.get(getTypeLB()).elements.get(pType).properties.get(propcode).is_meta){
        if(application.formTemplates.get(getTypeLB()).elements.get(pType).properties.get(propcode).genesis!='A'){
        String pv=application.formTemplates.get(getTypeLB()).elements.get(pType).getNonCompValue(propcode);
        LuiAppProp p=el.addProperty(propcode,pv);
        p.genesis='D';
        }
    }
   }
   }
    }
    }
  
  
  if (pSeq!=el.sourceSeq) {
  el.sourceSeq=pSeq; 
  el.mutSeq=pSeq;
  el.addProperty("Seq",String.valueOf(pSeq));
  setChangeFlag();
  }
  return el;  
}
/*public void loadFromCache(String app) throws LuiException, SQLException{
    String key=NVL(app,application.getProperty("AppCode"))+"."+getCode()+"."+this.getConfigCode();
    if (application.frmCashe.containsKey(key)){
       LuiAppForm cfrm=application.frmCashe.get(key);
       for (String cfp:cfrm.properties.keySet()){
         // LuiAppProp p=new LuiAppProp(cfrm,);
          addProperty(cfp,cfrm.properties.get(cfp).currentValue); 
       }
       for (String cfe:cfrm.elements.keySet()){
           LuiAppElement cel=cfrm.elements.get(cfe);
           LuiAppElement el=addElement(cel.Code,cel.getType(),cel.sourceSeq,cel.getProperty("Name"),cel.getProperty("Parent"));
         for (String cfep:cel.properties.keySet()){
             el.addProperty(cfep,cel.properties.get(cfep).currentValue);
         }  
       }
    
    }
    
}
public void moveToCache(String app) throws LuiException, SQLException{
    if (getType().equals("VERIFIER")) return;
    if (like(getCode(),"%_TEMPLATE")) return;
    String key=NVL(app,application.getProperty("AppCode"))+"."+getCode()+"."+this.getConfigCode();
    if (!application.frmCashe.containsKey(key)){
       LuiAppForm cfrm=new LuiAppForm(application,this.code,this.getType());
       for (String fp:properties.keySet()){
          cfrm.addProperty(fp,properties.get(fp).currentValue); 
       }
       for (String fe:elements.keySet()){
           LuiAppElement el=elements.get(fe);
           LuiAppElement cel=cfrm.addElement(el.Code,el.getType(),el.sourceSeq,el.getProperty("Name"),el.getProperty("Parent"));
         for (String fep:el.properties.keySet()){
             cel.addProperty(fep,el.properties.get(fep).currentValue);
         }  
       }
    application.frmCashe.put(key, cfrm);
    }
    
}
*/
/**
     * Загрузка из метаданных или кэша
     *
     * @param app код прикладной ситемы откуда брать (null-берём из
     * application.code
     * @param isTemplate - признак того, что грузится шаблон
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void load(String app,boolean isTemplate) throws LuiException, SQLException{
  LuiAppProp p;
        LuiDbConn conn=null;
       String formTemplate="";
    // пытаемся получить соединение
    try{
    String key=NVL(app,application.getProperty("AppCode"))+"."+getCode()+"."+this.getConfigCode();
   
    conn=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
    String confname=conn.getSelectResult(STMT4LOADFORMCONF.replace("<app_code>",NVL(app,application.getProperty("AppCode"))).replace("<form_code>", getCode()).replace("<config_code>",getConfigCode()));
    if (is_null(confname)){ // конфигурация отсутствует
        application.metaConnPool.SetFree(conn);
        application.msgWriter.write("LUI-00016", getCode(),getConfigCode());
    }
    if (!isTemplate){
        formTemplate=conn.getSelectResult("select template from lui_t_form where app_code='"+NVL(app,application.code)+"' and code='"+getCode()+"'");
        if (is_null(formTemplate) || getType().equals("TREE"))formTemplate=getType()+"_TEMPLATE";
    setVal("FormTemplate",formTemplate);
    }
    addProperty("ConfigName",application.NLS.NLS(confname));
    if (getType().equals("LIST") || getType().equals("TREE")){
    addProperty("CountSelected","{JS:App.forms.get(\"{Property:FormID}\").getCountMarkedRows();}");
    addProperty("RowsSelectedJSON","{JS:App.forms.get(\"{Property:FormID}\").getSelectedRowsJSON();}");
    }
    if (getType().equals("MENU")){
    addProperty("ItemsSelectedJSON","{JS:App.forms.get(\"{Property:FormID}\").getSelectedMenuItemsJSON();}");
    }
    setVal("Logging","Y");
    
    if (application.formTemplates!=null && application.formTemplates.containsKey(getTypeLB())){
        // циклим стандартные пропети
        for (String prcode:application.formTemplates.get(getTypeLB()).properties.keySet()){
     //if(application.formTemplates.get(getTypeLB()).properties.get(prcode).is_meta){   
     if(application.formTemplates.get(getTypeLB()).properties.get(prcode).genesis!='A'){   
     String prval=LuiMetaDataValue.get(app,application, conn,getCode(),getTypeLB(), getConfigCode(), null,null,null, prcode,formTemplate);
     if (is_not_null(prval))  {// если есть инд.значение, то установим его
         p=addProperty(prcode,prval);
         p.recalculate=true;
         p.genesis='M';
       if (prcode.equals("FEATURE")){
            addProperty("Code",prval);
        }
     } 
     }
      
    }
    }
    //теперь ищем элементы формы
    ArrayList<ArrayList<String>> elems = conn.getSelectAllRowsValList(STMT4LOADFORMELEM.replace("<app_code>",
            NVL(app,application.code)).replace("<form_code>",getCode()));
    if (elems!=null){
    for(int i=0;i<elems.size();i++){
    // 0=code 1-name 2-etype 3-seq 4-parent
    String ec=elems.get(i).get(0);
    
        LuiAppElement el=addElement(elems.get(i).get(0),elems.get(i).get(2),Double.valueOf(elems.get(i).get(3)),application.NLS.NLS(elems.get(i).get(1)),NVL(elems.get(i).get(4),"Form"));
      
if (application.formTemplates!=null){
    switch (el.getType()) {
        case "ITEM":
        case "PARAM":
        case "FIELD":
      String elprval=LuiMetaDataValue.get(app,application, conn, getCode(),getTypeLB(), getConfigCode(),el.getCode(),el.getType(),null,"DATA_TYPE",formTemplate);
      if (is_not_null(elprval)) {
      p=el.addProperty("DATA_TYPE", elprval);
      p.genesis='M';}
      elprval=LuiMetaDataValue.get(app,application, conn, getCode(),getTypeLB(), getConfigCode(),el.getCode(),el.getType(),null,"ITEM_TEMPLATE",formTemplate);
      if (is_not_null(elprval)) {
      p=el.addProperty("ITEM_TEMPLATE", elprval);
      p.genesis='M';
      }
    }
// циклим стандартые свойства элемента
  if (application.formTemplates.get(getTypeLB()).elements.containsKey(el.getType())){
        for(String elpropcode:application.formTemplates.get(getTypeLB()).elements.get(el.getType()).properties.keySet()){
        
        if (elpropcode.equals("DATA_TYPE"))continue;
       // if(application.formTemplates.get(getTypeLB()).elements.get(el.getType()).properties.get(elpropcode).is_meta){
        if(application.formTemplates.get(getTypeLB()).elements.get(el.getType()).properties.get(elpropcode).genesis!='A'){
       String elprval=LuiMetaDataValue.get(app,application, conn, getCode(),getTypeLB(), getConfigCode(),el.getCode(),el.getType(),el.getProperty("ITEM_TEMPLATE"), elpropcode,formTemplate);
       
        if (el.getType().equals("PARAM") && elpropcode.equals("VALUE") && el.getParent().equals("Form")){
             el.properties.remove("VALUE");    
             p=el.addProperty("Value", NVL(elprval,""));
             //if (elprval!=null) el.properties.get("Value").is_meta=true;
             //if (is_not_null(elprval))p.genesis='M';
             p.genesis=LuiMetaDataValue.genesis;
              
/* закоментил т.к. это успешно в addProperty делается. Да и like чёто не работает
             if (like(elprval,"%{%}%")){
         el.properties.get("Value").SCN=LuiAppProp.STARTSCN;
        } else {
            el.properties.get("Value").currentValue=NVL(elprval,"");
            el.properties.get("Value").SCN=0;
        }*/
        }
        else {
        //if (is_not_null(elprval)) {
          if (LuiMetaDataValue.genesis!='D'){
            p=el.addProperty(elpropcode, NVL(elprval,""));// если есть инд.значение, то установим его
            //el.properties.get(elpropcode).is_meta=true;
            p.genesis=LuiMetaDataValue.genesis;
        }
        }
        }
        }
}
    } 

    }// конец цикла по элементам
    
    }
    // верификаторы
if (getType().equals("BLANK")||getType().equals("LIST")){
if (application.verifiers!=null){
 if (application.verifiers.size()>0){   
 for (String elcode:elements.keySet()){
     LuiAppElement el=elements.get(elcode);
     if (el.getType().equals("FIELD")||el.getType().equals("ITEM")){
       if (el.properties.containsKey("VERIFIER")){
           String vrfCode=el.properties.get("VERIFIER").startedValue;
           if (is_not_null(vrfCode)){ // если есть верификатор
           if (application.verifiers!=null){ // и они есть в природе
           if (application.verifiers.containsKey(vrfCode)){ // и в них есть такой
             LuiAppForm vrfFrm=application.verifiers.get(vrfCode);  // берём форму верификатора
             if (vrfFrm.properties.containsKey("VERIFIER_SRV_PRG")) { // если есть серверный верификатор
              String srvVRF=NVL(vrfFrm.properties.get("VERIFIER_SRV_PRG").startedValue,vrfFrm.properties.get("VERIFIER_SRV_PRG").currentValue);
              if (is_not_null(srvVRF)){ // и он не пустой
                  el.addProperty("VERIFIER_SRV_PRG", srvVRF);
              }  
             }
             for(int pind=0;pind<vrfFrm.elements.size();pind++){ // крутим его элементы (там бывают только параметры
                 LuiAppElement pvrf=vrfFrm.elSort(pind); 
                 if (pvrf.getType().equals("PARAM")){ 
                    if (pvrf.properties.containsKey("Value")){
                      el.addProperty("VERIFIER_PARAM."+pvrf.getCode(),pvrf.properties.get("Value").startedValue); // под каждый параметр создаём проперть подя
                    }
                 }
             }
           }}
           }
       }
       }
     }
     }
 }
 }
    
    } catch (Exception exc){
        try{application.metaConnPool.SetFree(conn);}
        catch(Exception ex2){}
        application.msgWriter.write("LUI-00014",getType(),getCode(),getConfigCode(),exc.getLocalizedMessage());
        }
      finally{
        try{application.metaConnPool.SetFree(conn);}
        catch(Exception ex2){}
    }
if (is_not_null(getProperty("FEATURE"))){    
//code=getProperty("FEATURE"); /* Не знаю как это лечить почему FEATURE????
}
if (like(this.code,"O2P_L_OBJECT%")){
          System.out.println("after loadparam:"+System.currentTimeMillis()); // perfomance

      }
}
/**
 * Загрузка формы параметрами 
 * @param pParams
 * @throws LuiException
 * @throws SQLException 
 */
public void loadParams(String pParams) throws LuiException, SQLException{
    try{
    // разберёмся с входными параметрами
    if (is_not_null(pParams)){
       LuiCollection pc=new LuiCollection(pParams);
       String V=null;
       for (String K:pc.keySet()){ // крутим коллекцию параметров
           V=pc.Get_Element(K); //берём значение
         if ("LUI$LOG".equals(K)){ /// если отмена дурналирования
             setVal("Logging",V);
         }
         for (String elcode:elements.keySet()){ // крутим
           LuiAppElement el=elements.get(elcode);
           if (el.getType().equals("PARAM") && el.getParent().equals("Form")) { // параметры уровня формы
             if (K.equals(el.getProperty("NAME"))){ // если код выходного параметрв совпал с 
              K=elcode; // то запишем код элемента как код выходного параметра
              break;
             } 
           }
         }
      if (!elements.containsKey(K)) { // нет такого параметра "K" -- создаём
        Double seqAlfel=this.getNewSeq("PARAM");
        LuiAppElement Alfel=this.addElement(K,"PARAM", seqAlfel,K, "Form");
        Alfel.addProperty("isVariable","Y");
      } /*else {// нет то что здесь задумано - бред
          // если есть такой элемент
          if (!elements.get(K).getType().equals("PARAM") ||!elements.get(K).getParent().equals("Form")) {
              // но это не параметр или не на уровне формы, то мы мутируем код создаваемого параметра
              // постфиксом $IN проверив предварительно что такого эоемента нет
              if (!elements.containsKey(K+"$IN")){
                  K=K+"$IN";
                  
              }
          } 
      }*/
        LuiAppElement Alfel=elements.get(K);// извлекаем элемент - входной параметр
        String Alf=Alfel.getProperty("CHECK");
        if (is_not_null(Alf)){
            //?????????????????????  не дописано
                //V:=GL_P_SQL.Get_Execute_Result('begin :A:=GL_F_Check_Alpha(:A,'''||Alf||''','''||Replace(K,'''','''''')||''');end;',V);  
        }
        String parent;
        if (Alfel.properties.containsKey("Parent")){
        parent=Alfel.getParent();
        } else {
            Alfel.addProperty("Parent","Form");
            parent="Form";
        }
        Alfel.setProperty("Parent", parent);
        if (Alfel.properties.containsKey("VALUE")){
            Alfel.properties.remove("VALUE");
            Alfel.addProperty("Value", V);
        }
        if (!Alfel.properties.containsKey("Value")){
            Alfel.addProperty("Value", V);
        }
        Alfel.properties.get("Value").startedValue=V;
        /*if (like(V,"%{%}%")){   Это так было в Fastcom ПОТОМУ ЧТО ИНОГДА НАДО БЫЛО ПЕРЕДАТЬ ЧЕРЕЗ ПАРАМЕТР ПРОГРААМУ ВЫЧИСЛЕНИЯ ЧЕГО-НИБУДЬ 
        В LUI ЭТО НЕ НАДО Т,К. ЕСТЬ {Parse:
         Alfel.properties.get("Value").recalculate=true;
        } else {
            Alfel.properties.get("Value").currentValue=V;
            Alfel.properties.get("Value").recalculate=null;
        }*/
        if (Alfel.properties.get("Value").isDynamic()) {
            Alfel.properties.get("Value").recalculate=true;
        } else {
            Alfel.properties.get("Value").currentValue=V;
            Alfel.properties.get("Value").recalculate=null;
        }
        if (is_null(Alfel.getProperty("EType"))){
            Alfel.setProperty("EType","PARAM");
        }
        
        
        
        //???
       }
    }
    
    if (elements.containsKey("WINDOW_TYPE$")){
      if (elements.get("WINDOW_TYPE$").getParent().equals("Form")) {
        if (elements.get("WINDOW_TYPE$").getType().equals("PARAM")) {
            addProperty("WindowType",elements.get("WINDOW_TYPE$").getValue());
        }
      }
      }
    if (elements.containsKey("CONNECTION$")){
      if (elements.get("CONNECTION$").getParent().equals("Form")) {
        if (elements.get("CONNECTION$").getType().equals("PARAM")) {
         if (application.userDbConnections!=null){   
          if (application.userDbConnections.containsKey(elements.get("CONNECTION$").getValue())) {  
            addProperty("Connection",elements.get("CONNECTION$").getValue());
          }
        }}
      }
      }
    } catch (Exception exc){
        application.msgWriter.write("LUI-00014",getType(),getCode(),getConfigCode(),exc.getLocalizedMessage());
    }
}
/**
 * Генерация упорядоченного списка элементов для сортировки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void makeEl4order() throws LuiException, SQLException{
    int min=-32767;
    LuiAppElement el;
    el4order.clear();
    while (true){
      el=getNext4order(el4order,min,null);
      if (el==null)break;
      Integer m=String2Integer(el.getProperty("SORT"),null);
      if (m!=null) {       
      min=m;
      if (min<0) min=0-min;
      el4order.add(el);
      }
    }
        LuiCollection order_by= new LuiCollection("");
        for (int i=0;i<el4order.size();i++){
          order_by.Add_Element(el4order.get(i).getCode(),el4order.get(i).getProperty("SORT"));
        }  
        this.setVal("OrderBy",order_by.toString());
}

/**
 * Вычисление строки - характерисики предков элементта
 * @param pEl элемент приложения
 * @return строка - характерисика предков элементта
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getParentsNo(LuiAppElement pEl) throws LuiException, SQLException{
  String p =pEl.getParent();
  if (!elements.containsKey(p)){
      return "0";
  } else {
      return getParentsNo(elements.get(p))+"."+String.format("%05d",elements.get(p).sourceSeq)+elements.get(p).getCode();
  }
  
}
/**
 * Генерация упорядоченного списка элементов
 * @param pNum неизвестно для чего нужное число. Автор поленился описать вызов
 * @return элемент LUI
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiAppElement elSort(int pNum) throws LuiException, SQLException{
  //  if (getType().equals("GRID")){
    //    String here="here";
    //}
    return elSortC(pNum,false);
}
/**
 * Генерация упорядоченного списка элементов c учётом надо или нет сортироваять ячейки матрицы
 * @param pNum неизвестно для чего нужное число. Автор поленился описать вызов
 * @SortMatrixCell
 * @return элемент LUI
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiAppElement elSortC(int pNum,boolean SortMatrixCell) throws LuiException, SQLException{
  LuiAppElement el;  
  double MINELSEQ=-1000000;
  //if (pNum==0) {
  if (elements.size()!=elSortedList.size()){
  elSortedList.clear();
  
  if (!getType().equals("BLANK")) { // если не бланк мутируем seq левыми соседями
           for (String ec:elements.keySet()) {
               el=elements.get(ec);
               if (getType().equals("GRID") && (el.getType().equals("ROW")||el.getType().equals("COL")||el.getType().equals("CELL")))continue;
               if (el.getType().equals("ITEM")/*||el.getType().equals("GROUP")*/){
               el.mutSeq=el.sourceSeq;
               Double depth=null;
               
               LuiAppElement ep=el;
               int climit=0;
               while (true){
                climit++;
                if (climit>10000) break;
                if (ep.properties.containsKey("LeftNeighbor")) { // если у элемента есть такое св-во как левый сосед
                if (depth==null) {
                    depth=((double)1)/1000;
                }
                else  {
                    depth =  depth+((double)1)/1000;
                }
                
                String LN=NVL(ep.getProperty("LeftNeighbor"),"");
                if (!elements.containsKey(LN)) {
                 el.mutSeq=MINELSEQ+depth;
                    break;
                } else {
                 el.mutSeq=elements.get(LN).mutSeq+depth;   
                ep=elements.get(LN);
                }
                } else {
                  break;  
                }
               }
               
           } }}
 //for (String ec:elements.keySet()){ // заносим как попало элементы в список 
   //  el=elements.get(ec);
    // elSortedList.add(el);
 //}
 //elSortedList.addAll(elements.values());
 
       //   Collections.sort(elSortedList,application.elComparator); // делаем сортировку компаратором
       //    elSortedList.sort(application.elComparator);
    /*новая сортировка
   ArrayList<Double> dl=new ArrayList<>();
   for (String ek:elements.keySet()){
       dl.add(elements.get(ek).mutSeq);
   }
   Collections.sort(dl);
   
   for (int i=0;i<dl.size();i++){
      for (String ek:elements.keySet()){
        if (elSortedList.contains(elements.get(ek)))continue;
        if (elements.get(ek).mutSeq==dl.get(i)){
         elSortedList.add(elements.get(ek));
        }  
      }
   }
   elSortedList.clear();*/
   /* старая сортировка*/
  // boolean b=false;
   while (elements.size()>elSortedList.size()){
     // b=true;
      for (String ec:elements.keySet()){
      
       el=elements.get(ec);   
       if (!elSortedList.contains(el)){
         if (!el.getType().equals("CELL") || SortMatrixCell)  el=getLeastThen(el);
         if (!elSortedList.contains(el)){
            elSortedList.add(el);
         }  
       }
      } 
   }
 /* if (b && getType().equals("LIST")){
       for (int i=0;i<elSortedList.size();i++){
           elSortedList.get(i).mutSeq=i;
       }
   }*/
   /* */ 
 /*     String es="";
      for (int i=0;i<elements.size();i++) {
          LuiAppElement e=elSortedList.get(i);
          es=es+e.getParent()+"."+e.getCode()+"("+e.getType()+")"+e.sourceSeq+"\n";
      }
      es=es+"\n";
  */
  }
  return elSortedList.get(pNum);
      
}

public LuiAppElement getLeastThen(LuiAppElement pEl) throws LuiException, SQLException{
      for (String ec:elements.keySet()) {
       if (!pEl.getCode().equals(ec)){   
       LuiAppElement el=elements.get(ec);
       if (!elSortedList.contains(el)){
         int cr=application.elComparator.compare(el,pEl);  
         if (cr<0){
             return getLeastThen(el);
         } 
         
       }
       }
      }
   return pEl;   
}
/**
 * Перемещение элемента списка или бланка вверх внутри его родителя
 */
/*public void item2Left(String elCode)throws LuiException, SQLException{
    if (!elements.containsKey(elCode))return;
    LuiAppElement el=elements.get(elCode);
    double ss=el.sourceSeq;
    double ms=el.mutSeq;
    String seq=el.getProperty("Seq");
    LuiAppElement oel=null;
    
    for (int i=0;i<elements.size();i++){
       if (!elSort(i).getParent().equals(el.getParent())) continue;
       if (!elSort(i).getType().equals(el.getType())) continue;
       if (elSort(i).Code.equals(elCode)) break;
    oel=elSort(i);
           }
    el.sourceSeq=oel.sourceSeq;
    el.mutSeq=oel.mutSeq;
    el.addProperty("Seq",oel.getProperty("Seq"));
    oel.sourceSeq=ss;
    oel.mutSeq=ms;
    oel.addProperty("Seq",seq);
    this.elSortedList.clear();
    this.elSort(0);
} */
/**
 * выборка следующего для сортировки в el4order
 */
private LuiAppElement getNext4order(ArrayList<LuiAppElement> el4o, Integer pSeq,String pParent) throws LuiException, SQLException{
    LuiAppElement el=null;
    LuiAppElement elm=null;
    for (String key:elements.keySet()){
      el=elements.get(key);
      if (is_not_null(pParent)){
          if (!el.getParent().equals(pParent)) {
          el=null; 
          continue;}
      } 
      if (is_not_null(el.getProperty("SORT"))){
      if (!el4o.contains(el)){
      if (el.getType().equals("ITEM") || el.getType().equals("FIELD")){
     
     Integer order=String2Integer(el.getProperty("SORT"),null);
     if (order!=null){
     if (order<0) order = 0-order;
     if (order>=pSeq){
        elm=null;
        elm=getLeast4order(order);
        if (elm!=null) return elm;
        return el; 
     } else el=null;
    } else el=null;
    } else el=null;
    } else el=null;
    } else el=null;
}
return el;
}
/**
 * выборка предидущего для сортировки в el4order
 */
private LuiAppElement getLeast4order(int pSeq) throws LuiException, SQLException{
    LuiAppElement el=null;
    LuiAppElement elm=null;
 for (String key:elements.keySet()){
      el=elements.get(key);
      if (is_not_null(el.getProperty("SORT"))){
      if (!el4order.contains(el)){
      if (el.getType().equals("ITEM") || el.getType().equals("FIELD")){
              Integer order=String2Integer(el.getProperty("SORT"),null);
              if (order!=null){
              if (order<0) order = 0-order;
        if (order<pSeq){
            elm=null;
            elm=getLeast4order(order);
            if (elm!=null) return elm;
            return el;
      }else el=null;
      }else el=null;
 }else el=null;
}else el=null;
}else el=null;      
 }
return el; 
}

private String def(String defIfFirst){
    return NVL(defIfSecond,defIfFirst);
}
private String def(){
    return def("");
}
/**
 * (ре)инициализация формы. Выполняется посл Загрузки формы UI,
 *  на Последней стадии выполнения команды, если ACTION_REQUERY = REFRESH
 *  ,при изменении конфигурации
 *  ,при сбросе параметров отображения
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
 public void reloadForm() throws LuiException, SQLException{
 LuiClCommandGroup XMLcg=null;
 LuiClCommand XMLcmd=null;
 LuiClCommand XMLcmdG=null;
 addProperty("Elements","{ElementValues2JSONArray:"+getFormID()+";Short:Y;SCN:{ApplicationProperty:SCN};}");
 
 setVal("Changed","N"); // Установим, что ничего не менялось
 //сброс вычисленных значений входных параметров (после действия REFRESH) при этом если в параметре записана динамика,
 // то должен восстановится признак необходимости вычислений
 
ArrayList<String> el4del=new ArrayList<>(); // список удаления виртуальных элементов
 for (String elKey:elements.keySet()){
     LuiAppElement el=elements.get(elKey); 
  //if (!el.metadata) 
  if (el.getProperty("Metadata").equals("N"))
  {
     el4del.add(elKey);// если элемент не метаданковый - заносим его в список удаления
  }
  else {
     if (el.getType().equals("PARAM") && el.getProperty("Parent").equals("Form")){
      /*if (el.properties.containsKey("VALUE") && !el.properties.containsKey("Value")){
          LuiAppProp p=el.properties.get("VALUE");
          el.properties.put("Value", p);
          el.properties.remove("VALUE");
      }*/
         el.properties.get("Value").clearComputedValue(true);
}
     }
 }
 for (int edi=0;edi<el4del.size();edi++){//-- удаляем виртуальные элементы
    if (!getType().equals("GRID") || 
            (elements.get(el4del.get(edi)).getType().equals("ROW") && elements.get(el4del.get(edi)).getParent().equals("ROWS"))
            ||
            (elements.get(el4del.get(edi)).getType().equals("COL") && elements.get(el4del.get(edi)).getParent().equals("COLS"))
            ){ 
     if (XMLcg==null) {
         XMLcg=application.commandWriter.addGroup("Kill Elements");
         XMLcg.addParameter("FormID", getFormID());
     }
     XMLcmd=XMLcg.addCommand("Kill");
     XMLcmd.addScaPar("Code", el4del.get(edi));
    }
     elements.remove(el4del.get(edi));
 } 
if (el4del.size()>0){
elSortedList.clear();    
// elSort(0);
 el4del.clear();
el4del=null;
XMLcg=null;
XMLcmd=null;
}
 
if (getType().equals("GRID")){
    addProperty("CurrentCell","");
if (is_null(getVal("Second"))) { // Первый проход
         prepareMatrixTemplEl("CELL");
         prepareMatrixTemplEl("PARAM");
       }
}
 //вычисляем значения па/раметров и снимаем признак необходимости вычислений, т.е. делаем параметры статическими
 for (String elKey:elements.keySet()){
     LuiAppElement el=elements.get(elKey);
     if (el.getType().equals("PARAM") && el.getProperty("Parent").equals("Form")){
       if (this.code.equals("LUI_B_PROPERTY_COLOR") && elKey.equals("P_TASK_QUERY")) {
           el.properties.get("Value").startedValue=el.properties.get("Value").startedValue.replace("{SQL:", "{");
       }
         el.properties.get("Value").getComputedProperty();
     }
 }
   if (getType().equals("HOST")){ // если команда ОС сбросим все св-ва
     for (String prkey:properties.keySet()){
         properties.get(prkey).clearComputedValue(false);
     }
    return; 
   }
  if (getVal("Second").equals("Y")){
     //defIfSecond=new String(new byte[]{0},Charset.forName("windows-1251"));
     defIfSecond=new String(new byte[]{0},Charset.forName("UTF8"));
 }
 //loadUserPrefs(); //перенесено 20.03.20 вниз
       if (getType().equals("LIST")){ // если список
       setVal("CurrentRow",""); // текущая строка - неопределена
       if (!properties.containsKey("Var CurrentColumn")){
       setVal("CurrentColumn","");//и столбец если ещё нет такого св-ва
       }
 }
   if (getType().equals("BLANK")){//если бланк
       if (!properties.containsKey("Var CurrentField")) { // если нет такого св-ва
       setVal("CurrentField","");//текущее поле не определено
       }
   }
 //приготовление системного меню
 prepareSysMenu();  
 prepareAddons();
 if (getType().equals("CHART")){
     for(String pc:properties.keySet()){
         if (like(pc,"Click%")){
             properties.get(pc).currentValue="";
             properties.get(pc).startedValue="";
         }
     }
 }
 //Выполняем действие "при загрузке"
 if (properties.containsKey("ON_LOAD")){
   LuiAppProp prop=properties.get("ON_LOAD");
   String prg=prop.getComputedProperty();
   if (is_not_null(prg)) {
application.log.addRecord("ON_LOAD", application.NLS.NLS("RUS:Выполнение процедуры при загрузке формы ;ENG:Programm on load ;"), prg,this.getFormID());       
       application.parser.execute(prg, prop);
   
   }
 }
 if (this.DgV2!=null && getType().equals("LIST")) { // Если это список выпестованный из диаграммы группировки данных
     // то применим QBE сохранённые там и установим уомлания QBE-ешных свойств
     LuiDbConn mlc=null;
     try {
     mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
     String DEFQBE_NOT=mlc.getSelectResult("select lui_f_dg_get_chart_pref_val("+DgV2.DgId+","+DgV2.Level+",'QBE_NOT_LIST','')");
     addProperty("QBE_NOT",DEFQBE_NOT);addProperty("DEFQBE_NOT",DEFQBE_NOT);
     String DEFQBE_COL=mlc.getSelectResult("select lui_f_dg_get_chart_pref_val("+DgV2.DgId+","+DgV2.Level+",'QBE_COL_LIST','')");
     addProperty("QBE_COL",DEFQBE_COL);addProperty("DEFQBE_COL",DEFQBE_COL);
     String DEFQBE=mlc.getSelectResult("select lui_f_dg_get_chart_pref_val("+DgV2.DgId+","+DgV2.Level+",'QBE_LIST','')");
     addProperty("DEFQBE",DEFQBE);
     if (!DEFQBE.equals("{}")) {
     JsonElement JSONRoot = application.JSONparser.parse(DEFQBE);
     if (JSONRoot.isJsonObject()){
      JsonObject JSONOBJROOT=JSONRoot.getAsJsonObject();
       listQBE.clear(); // очищаем QBE
       for (int i=0;i<JSONOBJROOT.size();i++){ // крутим массив строк QBE
         HashMap<String,String> hmc;  
         if (listQBE.containsKey(String.valueOf(i+1))){
             hmc=listQBE.get(String.valueOf(i+1));
         } else hmc=new HashMap<>();  
           JsonElement JSQBEROWel=JSONOBJROOT.get(String.valueOf(i+1));
           if (JSQBEROWel!=null && JSQBEROWel.isJsonObject()){
               JsonObject JSQBEROW=JSQBEROWel.getAsJsonObject();
               Iterator<Entry<String,JsonElement>> iter=JSQBEROWel.getAsJsonObject().entrySet().iterator();
               while (iter.hasNext()){
                   Entry<String,JsonElement> keyVal=iter.next();
                   String key=keyVal.getKey();
                   String val=keyVal.getValue().getAsString();
                   hmc.put(key, val);
               }
               listQBE.put(String.valueOf(i+1), hmc);
           }
       }
     }
     }
             } finally{
         if (mlc!=null) application.metaConnPool.SetFree(mlc);
     }
 }
 loadUserPrefs(); //перенесено 20.03.20 сверху
 if (getVal("Second").equals("Y"))  {
     resetRecalc();
 }
 //loadUserPrefs(); 
 if (getType().equals("BLANK")){ // если это бланк
     // заполнение дефолтных полей данными из запроса
     queryToBlank();
     prepareBlank();
 } 
 if (getType().equals("LIST")) { // если это список
     prepareList();
 }
 if (!getType().equals("TREE") && !getType().equals("MENU") ){ // если это не дерево и не меню 
    
     for (String elCode:elements.keySet()){ // цикл по элементам и применение к ним прав и определение свойства LeftNeighbor
         switch (elements.get(elCode).getType()) {
             case "ITEM":
//              elements.get(elCode).setProperty("Read",application.getEffectiveRight("UI_ELEMENT",getCode()+"."+elCode ,"READ"));
              if (!getType().equals("GRID")) {
                elements.get(elCode).setProperty("Read",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"READ"));
              String tstr=getVisualPref(elCode,"LeftNeighbor", null);
              if (tstr!=null){
              elements.get(elCode).setProperty("LeftNeighbor", tstr);
              }              }
             break;
             case "FIELD":
              elements.get(elCode).setProperty("UserChanged","");
          //    if ((getVal("ReadOnly").equals("Y") || application.getEffectiveRight("UI_ELEMENT",getCode()+"."+elCode ,"UPD").equals("N"))){
                  if ((getVal("ReadOnly").equals("Y") || application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"UPD").equals("N"))){
              elements.get(elCode).setProperty("CHANGING","N");    
              }
          //    elements.get(elCode).setProperty("Read",application.getEffectiveRight("UI_ELEMENT",getCode()+"."+elCode ,"READ"));
                elements.get(elCode).setProperty("Read",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"READ"));  
             break;
             case "ACTION":
             if (like(elCode,"SYSMENU$%") || like(elCode,"COLMENU$%")  || like(elCode,"COLMENU$%")|| like(elCode,"CELLMENU$%")){ // если это действие системного меню или меню столбца или QBE меню ячейки
               elements.get(elCode).setProperty("Run",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"RUN"));     
             } else {
             if ((getVal("ReadOnly").equals("Y"))) {
              elements.get(elCode).setProperty("Run","N");
             } else{
                  elements.get(elCode).setProperty("Run",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"RUN"));   
             } }
             break;
             case "GROUP":
             //elements.get(elCode).setProperty("Read",application.getEffectiveRight("UI_ELEMENT",getCode()+"."+elCode ,"READ"));                 
             elements.get(elCode).setProperty("Read",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"READ"));                 
             break;
             case "PARAM":
             break;
         }
         }
  //06/06/2020:
  if (getType().equals("BLANK")){
  for (String elCode:elements.keySet()) {
    if (elements.get(elCode).getType().equals("FIELD")){ // просматриваем только поля бланка
        //elements.get(elCode).addProperty("Value", elements.get(elCode).getNonCompValue("DEFINITION")); 01 07 2021
        elements.get(elCode).addProperty("Value","{Property:DEFINITION}"); //01 07 2021
    }
  }
      }   
  //
     }
 else { // если дерево или меню
      for (String elCode:elements.keySet()){
        if (elements.get(elCode).getType().equals("ACTION")){ // просматриваем только действия
        //-- установка Run
        if (like(elCode,"SYSMENU$%") || like(elCode,"COLMENU$%")  || like(elCode,"COLMENU$%")|| like(elCode,"CELLMENU$%")){ // если это действие системного меню или меню столбца или QBE меню ячейки
               elements.get(elCode).setProperty("Run",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"RUN"));     
             } else {
             if ((getVal("ReadOnly").equals("Y"))) {
              elements.get(elCode).setProperty("Run","N");
             } else{
                  elements.get(elCode).setProperty("Run",application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+elCode ,"RUN"));   
             } }
          
           if (!elements.get(elCode).getProperty("Run").equals("Y") // если действие нельзя выполнять
           || getVal("ReadOnly").equals("Y") ) { // если прав на действие нет или режим тока чтение
            if (!getType().equals("MENU")){   
             if  (like(elements.get(elCode).properties.get("DISPLAY").getNonCompValue(),"%{Property:ACTIVE}%")) { // и если в DISPLAY ccылка на ACTIVE
               String nd=elements.get(elCode).properties.get("DISPLAY").getNonCompValue();
               nd=nd.replace("{Property:ACTIVE}",elements.get(elCode).properties.get("ACTIVE").getNonCompValue());
                 elements.get(elCode).setProperty("DISPLAY",nd);
           } 
            }
            elements.get(elCode).setProperty("ACTIVE","N");
           }
        }       
     }
   if (getType().equals("MENU")){ //если это меню
       elSort(0);
       String ucfgs="0";
     for (int i=0;i<this.elSortedList.size();i++){ // обходим последовательно элементы
            LuiAppElement mel=this.elSortedList.get(i);
      if (mel.getType().equals("ACTION")){ // просматриваем только действия не системного меню и не CLOSE и не RMB
          if (!like(mel.getCode(),"SYSMENU$%") && !like(mel.getProperty("ROLE"),"%CLOSE%") && !like(mel.getProperty("ROLE"),"%RMB%")){
          String pcmd=mel.getProperty("COMMAND"); // берём код вызываемой формы
         if (like(pcmd,"%(%)")){ //За командой может идти коллекция параметров в круглых скобках
         int li=pcmd.indexOf("(");
         pcmd=pcmd.substring(0, li);
         }
         if (is_not_null(pcmd)){
         String params="";
         params=getParamString(mel.getCode(),params);
         LuiCollection paramCol=new LuiCollection(params);
         String config=NVL(paramCol.Get_Element("CONFIG"),"DEFAULT"); // конфигурация
         if (is_not_null(application.getAppUser())){
         LuiDbConn lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
          try {
             ArrayList<LuiCollection> ucfgArray=lc.getSelectAllRowsValMap("select id,name from lui_t_user_config where id not in ("+ucfgs+") and menu_appr='Y' and parent_form='"+pcmd+"' and parent_config='"+config+"' and parent_app='"+application.code+"' and user_name='"+application.getAppUser()+"' order by name");
             if (ucfgArray!=null) {
               for (int ci=0;ci<ucfgArray.size();ci++){
                LuiCollection ucfgR=ucfgArray.get(ci);
                String ucfgId=ucfgR.Get_Element("id");
                ucfgs=ucfgs+","+ucfgId;
                String ucfgName=ucfgR.Get_Element("name");
                        LuiAppElement ucfgel=this.addAction(mel.getCode()+"-"+ucfgId, ci, 
                         ucfgName, 
                         mel.getCode(), //pParent
                         ucfgName, //pPrompt
                         application.NLS.NLS("RUS:Вызов пользовательской конфигурации;ENG:Call user saved configuration;"),//hint
                         mel.properties.get("ACTIVE").getNonCompValue(), // ACTIVE
                         mel.properties.get("DISPLAY").getNonCompValue(), // VISIBLE
                         mel.properties.get("PROC_INITIALLY").getNonCompValue(), 
                         mel.properties.get("COMMAND_TYPE").getNonCompValue(), 
                         mel.properties.get("COMMAND").getNonCompValue(), 
                         "N", 
                         mel.properties.get("QUESTION").getNonCompValue(), 
                         null, // pNotes
                         mel.properties.get("ROLE").getNonCompValue(), 
                         mel.properties.get("AFTER").getNonCompValue(), 
                         mel.properties.get("CHILD_COMMIT").getNonCompValue(), 
                         mel.properties.get("PROC_FINALLY").getNonCompValue(), 
                         null, //Icon
                         null, 
                         "N"
                         );
                   //теперь перенесём параметры
                   double pi=0;
                   for (int ip=0;ip<this.elSortedList.size();ip++){
                       LuiAppElement mpel=this.elSortedList.get(ip);
                       if (mpel.getType().equals("PARAM")){
                          if (mpel.getParent().equals(mel.getCode())){ 
                             pi++;
                          LuiAppElement newmpel=this.addElement(mpel.getCode()+"-"+ucfgId, "PARAM",pi,mpel.properties.get("Name").getNonCompValue(),ucfgel.getCode());
                          // теперь все св-ва параметров кроме Parent,Code,Name,Seq
                          newmpel.addProperty("NAME", mpel.properties.get("NAME").getNonCompValue());
                          newmpel.addProperty("VALUE",mpel.properties.get("VALUE").getNonCompValue());  
                          }
                       }
                   }
                   
                   //теперь добавим параметр DefCfg
                   LuiAppElement newmpel=this.addElement("DefCfg-"+ucfgId, "PARAM",pi,"DefCfg",ucfgel.getCode());
                   newmpel.addProperty("NAME", "DefCfg");
                   newmpel.addProperty("VALUE",ucfgId);
               }
             }
          } finally {
              application.metaConnPool.SetFree(lc);
         }
          }
      }
     }
     
   }   
  }
   this.elSortedList.clear(); // подчистим сортированный список для последующих полных сортировок
   } // если меню
   }// если дерево или меню
 setAllChangeFlag(true);
 setVal("CurrentAction", "");
 XMLcg=application.commandWriter.addGroup("ReloadForm");
 XMLcg.addParameter("FormID", getFormID());
 XMLcmd=XMLcg.addCommand("Form");
 if (!getProperty("NoTileOnReload").equals("Y"))
 XMLcmd.addScaParameter("Title",this.getFormTtle()+getProperty("Var QBE_Title"),def(""));
 XMLcmd.addScaParameter("SplitterRight",getVisualPref("","SplitterRight", ""),def());
 XMLcmd.addScaParameter("SplitterBottom",getVisualPref("","SplitterBottom", ""),def());
 LuiCollection toolBarButtons;
 int toolBarDecade=0;
 switch(getType()){
     // новый вариант матрицы 05.02.21
     case "GRID"://если матрица
   /*    if (is_null(getVal("Second"))) { // Первый проход
         prepareMatrixTemplEl("CELL");
         prepareMatrixTemplEl("PARAM");
       }*/
    addProperty("SelectedColumns","[]");
    addProperty("SelectedRows","[]");
    addProperty("SelectedCells","[]");
    //Обработка матрицы начало
if (!getVal("WaitingForFormActivate").equals("Y")) {    
           String TupleType;
           Integer cursor=null;
           String muSQLText; // Текст запроса размножителя
           String ItemText; //-- Значение свойства
           String PropName; //-- Код свойства
           String sizeNameUsr;
           String sizeNameVS;
           int L; //-- Количество столбцов в запросе
           String ItemCode;// -- Код элемента
           String RowCode; //-- Код строки
           String ColCode; //-- Код колонки
           String RowNum; //-- Код строки
           String ColNum; //-- Код колонки
           ArrayList<LuiCollection> LeafColumns=new ArrayList<>();
           ArrayList<LuiCollection> LeafRows=new ArrayList<>();
           int CellNum =1; //-- Номер кода ячейки  -- 25.03.2021
           String newCellCode=null; //переменная в которой бует код создаваемой ячейки -- 25.03.2021
           LuiDbConn lc=null;
           HashMap<String,LuiDbConn> lcComConn=new HashMap<>(); // массив соединений с короткими транзакциями для обеспечения работы размножителей в рамках одной сессии
           LuiAppConn ac=null;
           String firstRowsGroup=null;
           String firstColsGroup=null;
           
           ArrayList<ColInfo> cols;
           ArrayList<String> cursRow;
           this.elSortedList.clear();
           elSortC(0,true); // сортировка кандидатов строк и столбцов и ячеек
           this.prepareSysMenu();
           ArrayList<LuiAppElement> clist=new ArrayList<>(); // кандидаты в ячейки
           for (int i=0;i<elements.size();i++){ // 25.03.2021 теперь порядок важен
              LuiAppElement el=elSort(i);
              if (!el.getType().equals("CELL"))continue; 
              clist.add(el);
           }
             ArrayList<LuiAppElement> glist=new ArrayList<>();
           
             for (int i=0;i<elements.size();i++){
                 LuiAppElement el=elSort(i);
                 if (!el.getType().equals("GROUP")) continue;
                 glist.add(el);
             }
             for (int gi=0;gi<glist.size();gi++){// -- Обходим элементы типа «Размерность». Всё равно в каком порядке
              LuiAppElement Axisel=glist.get(gi);
              String gk=Axisel.getCode();
              TupleType=Axisel.getProperty("GROUP_TYPE");
              if (TupleType.equals("ROWS")){
                  TupleType="ROW";
                  if ( firstRowsGroup==null) firstRowsGroup=gk;
              }
              else {
                 if (TupleType.equals("COLS")){
                     TupleType="COL";
                     if ( firstColsGroup==null) firstColsGroup=gk;
                 } 
                 else TupleType="";
              }
              if (is_null(TupleType)) continue; //-- Тип кортежей определён
              ArrayList<LuiCollection> ItemIerar =new ArrayList<>();
              //ItemIerar=this.MatrixMakeItemIerarSimple(gk, "ITEM");
              MatrixMakeItemIerar(ItemIerar,gk,"ITEM");
              LinkedHashMap<String,ArrayList<String>> CortageTree =new LinkedHashMap<>();
              for (int ii=0;ii<ItemIerar.size();ii++){//-- Обходим по иерархии все подчинённые элементы размерности. Берём элементы типа ITEM. Обходим в поредке Seq на каждом уровне!
                  LuiCollection tuplecol=ItemIerar.get(ii);
                  String TupleCode=tuplecol.Get_Element("CODE");
                  LuiAppElement tupleEl=elements.get(TupleCode);
                  muSQLText=tupleEl.getProperty("ITEM_MULTIPLIER");//-- Берём свойство кортежа - размножитель кортежа
                 if (like(muSQLText,"JS:%")){//-- JS-Размножитель есть. Значит, это кортеж-репликатор. Пусть отработает, а потом мы его удалим.
                       tupleEl.properties.remove("ITEM_MULTIPLIER"); //-- Удаление свойства ITEM_MULTIPLIER (в репликациях оно не нужно и вредно)
                       try { 
                       application.JSContents.clear();
                       jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.application.JSengine.eval(muSQLText.substring(3));
                       jdk.nashorn.api.scripting.ScriptObjectMirror props;
                       for (String Ind:Arr.keySet()) { // крутим элементы верхего массива -- Каждsq - новый кортеж
                        props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
                        ItemCode=null;
                        for (String propkey:props.keySet()){ // крутим элементы вложенного массива   -- Первым делом нужно выяснить код нового кортежа
                          if (propkey.equals("CODE")) {//-- Ищем столбец с кодом Code 
                             ItemCode=props.get(propkey).toString();//-- Нашли. Записем в ItemCode
                             break; 
                          } 
                        }
                        if (ItemCode!=null && !ItemCode.equals(TupleCode)){ //-- если ItemCode не null и не равен Tuple.Code
                             MatrixCopyElement(TupleCode,ItemCode,TupleType);//--- Создаём копию элемента но другого типа
                           elements.get(ItemCode).addProperty("Metadata","N");
                           elements.get(ItemCode).addProperty("MetadataElement",TupleCode);  
                           for (String propkey:props.keySet()){ // крутим элементы вложенного массива каждое выражение - некое свойство кортежа
                              PropName=propkey; 
                              if (PropName.toUpperCase().equals("CODE")) continue; //-- код уже учли
                              ItemText=props.get(propkey).toString();//-- значение свойства
                              if (PropName.toUpperCase().equals("PARENT")) {
                                if (ItemText==null) continue;
                                if (ItemText.equals(""))ItemText="Form";          
                                PropName=LuiString.withCapitalLetter(PropName);
                            }
                              elements.get(ItemCode).addProperty(PropName,ItemText);//-- заменяем свойство в новом элементе типа TupleType  
                           }
                           // теперь берём отца
                           String cortParent=elements.get(ItemCode).getProperty("Parent");
                           if (!CortageTree.containsKey(cortParent)) { // иесли его ещё нет в дереве кортежей
                               ArrayList<String> childs=new ArrayList<>();
                               childs.add(ItemCode);
                               CortageTree.put(cortParent, childs);
                           } else {
                               CortageTree.get(cortParent).add(ItemCode);
                           }
                           // листовых кортежей даже не будет в CortageTree
                        }//-- если ItemCode не null и не равен Tuple.Code
                       }//// крутим элементы верхего массива -- Каждsq - новый кортеж
                       } catch (ScriptException sex) {
              
                       } catch (Exception ex){
              
                       }
                     elements.remove(TupleCode);//-- Удаление кортежа-репликатора  
                 } // JS конец
                 if (like(muSQLText,"SQL:%") && is_not_null(muSQLText.substring(4))){//-- SQL-Размножитель есть. Значит, это кортеж-репликатор. Пусть отработает, а потом мы его удалим.
                       tupleEl.properties.remove("ITEM_MULTIPLIER"); //-- Удаление свойства ITEM_MULTIPLIER (в репликациях оно не нужно и вредно)
                   try{
                   
                   //берём соединение
                   ArrayList<String> acstmt=application.getAppConnAndStmt(muSQLText.substring(4),getProperty("Connection"),getFormID());
                   String acc=acstmt.get(0);
                   String stmt=acstmt.get(1);
                   if (lcComConn.containsKey(acc)){
                   lc=lcComConn.get(acc);
                   } else {
                   ac=application.userDbConnections.get(acc);
                  if (ac.getSessionMode()=='C') { // если соединение коммунальное
                   lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
                   lcComConn.put(acc, lc);
                   } else{                  //в противном случае возьмём текущее cоединение
                       lc=ac.currentConnection;
                         }
                    if (lc==null) // если соединение так и не нашлось, то
                     application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
                    // соединение есть
                   }
                    cursor=lc.CursorOpen(stmt); // открываем курсор
                    cols=lc.CursorGetColsInfo(cursor);//получаем список столбцов
                    L=cols.size();
                    while(lc.CursorNextRow(cursor)){ //крутим строки  -- Каждая строка - новый кортеж
                      cursRow=lc.CursorGetRow(cursor);
                      ItemCode=null;
                      
                      for (int F=0;F<L;F++){//-- Первым делом нужно выяснить код нового кортежа
                        if (cols.get(F).alias.toUpperCase().equals("CODE")) {//-- Ищем столбец с кодом Code
                         ItemCode=cursRow.get(F);//-- Нашли. Записем в ItemCode
                         break;
                        }
                      
                      }
                      if (ItemCode!=null && !ItemCode.equals(TupleCode)){ //-- если ItemCode не null и не равен Tuple.Code
                              MatrixCopyElement(TupleCode,ItemCode,TupleType);//--- Создаём копию элемента но другого типа
                           elements.get(ItemCode).addProperty("Metadata","N");
                           elements.get(ItemCode).addProperty("MetadataElement",TupleCode);
                        for (int F=0;F<L;F++){//-- Смотрим все выражения запроса
                            PropName=cols.get(F).alias/*.toUpperCase()*//*.substring(0,31)*/; //-- каждое выражение - некое свойство кортежа
                            ItemText=cursRow.get(F);//-- значение свойства
                            if (PropName.toUpperCase().equals("CODE")) continue; //-- код уже учли
                            if (PropName.toUpperCase().equals("PARENT")) {
                                if (ItemText==null) continue;
                                if (ItemText.equals(""))ItemText="Form";          
                                PropName=LuiString.withCapitalLetter(PropName);
                            }
                            if (ItemText!=null)
                            elements.get(ItemCode).addProperty(PropName,ItemText);//-- заменяем свойство в новом элементе типа TupleType  
                        }
                        // теперь берём отца
                           String cortParent=elements.get(ItemCode).getProperty("Parent");
                           if (!CortageTree.containsKey(cortParent)) { // иесли его ещё нет в дереве кортежей
                               ArrayList<String> childs=new ArrayList<>();
                               childs.add(ItemCode);
                               CortageTree.put(cortParent, childs);
                           } else {
                               CortageTree.get(cortParent).add(ItemCode);
                           }
                           // листовых кортежей даже не будет в CortageTree
                      }
                    }
                    
                   } finally {
                    if (cursor!=null)
                    lc.CursorClose(cursor);
                    cursor=null;
                    /*if (lc!=null && ac!=null) {
                       if (ac.getSessionMode()=='C') { // если соединение коммунальное
                        ac.connPool.SetFree(lc); // освободим его
                      }
                       ac=null;
                       lc=null;
                     } */
                   }
                    elements.remove(TupleCode);//-- Удаление кортежа-репликатора
                    
                  }
                  if (!like(muSQLText,"SQL:%") && !like(muSQLText,"JS%"))
                  {// -- Это не шаблон корьежа, а статический кортеж
                     elements.get(TupleCode).addProperty("EType",TupleType);//-- заменяем тип кортежа на ROW или COL
                     elements.get(TupleCode).addProperty("MetadataElement",TupleCode);
                     elements.get(TupleCode).addProperty("Metadata","Y");
                     // теперь берём отца
                           String cortParent=elements.get(TupleCode).getProperty("Parent");
                           if (!CortageTree.containsKey(cortParent)) { // иесли его ещё нет в дереве кортежей
                               ArrayList<String> childs=new ArrayList<>();
                               childs.add(TupleCode);
                               CortageTree.put(cortParent, childs);
                           } else {
                               CortageTree.get(cortParent).add(TupleCode);
                           }
                           // листовых кортежей даже не будет в CortageTree
                  }
              } // конец кручения иерархии item ов
              // передача корежей клиенту
              //-- Обходим все кортежи текущей оси - по иерархии и в порядке Seq на каждом уровне
              //this.elSortedList.clear();
              ArrayList<LuiCollection> CortIerar=new ArrayList<>();
              //MatrixMakeItemIerar(CortIerar,gk,TupleType);
              MatrixMakeItemIerarSimple2(gk,CortageTree,CortIerar);
              //CortIerar=MatrixMakeItemIerarSimple(gk,TupleType);
                boolean leaf;
                int ltn=0; // нумерация листовых кортежей
              for (int ii=0;ii<CortIerar.size();ii++){
                  LuiCollection tuplecol=CortIerar.get(ii);
                  String TupleCode=tuplecol.Get_Element("CODE");
                  addProperty("CurrentTuple",TupleCode);
                  String TupleParent=tuplecol.Get_Element("PARENT");
                  LuiAppElement TupleEl = elements.get(TupleCode);
                  leaf=true;
                  for (String ce:elements.keySet()){
                 if (!elements.get(ce).getType().equals(TupleType))continue;
                 if (elements.get(ce).getParent().equals(TupleCode)){leaf=false;break;} // как отец где-то упомянут значит не листовой
               }
                   if (leaf){ltn++;
                     TupleEl.addProperty("ltn",String.valueOf(ltn));
                     if (TupleEl.getType().equals("ROW")){
                         LuiCollection lrc=new LuiCollection("");
                         lrc.Add_Element("rc",TupleEl.Code);
                         lrc.Add_Element("rn",String.valueOf(ltn));
                         String its=TupleEl.getProperty("ITEM_SEMANTICS");
                         if (is_not_null(its)){
                         String[] itemSemArray=its.split(",");
                         for (int isa=0;isa<itemSemArray.length;isa++){
                           String[] SemEl= itemSemArray[isa].split(":");
                           if (SemEl.length==2 && SemEl[0].replace(" ","").equals("\"rownum\"")){lrc.Add_Element("ri",SemEl[1].replace(" ",""));break;}
                         }
                         }
                         LeafRows.add(lrc);
                     } else {
                         LuiCollection lcc=new LuiCollection("");
                         lcc.Add_Element("cc",TupleEl.Code);
                         lcc.Add_Element("cn",String.valueOf(ltn));
                         String its=TupleEl.getProperty("ITEM_SEMANTICS");
                         if (is_not_null(its)){
                         String[] itemSemArray=its.split(",");
                         for (int isa=0;isa<itemSemArray.length;isa++){
                           String[] SemEl= itemSemArray[isa].split(":");
                           if (SemEl.length==2 && SemEl[0].replace(" ","").equals("\"colnum\"")){lcc.Add_Element("ci",SemEl[1].replace(" ",""));break;}
                         }
                         }
                         LeafColumns.add(lcc);
                     }
                   }  
                  if (TupleType.equals("ROW"))XMLcmd=XMLcg.addCommand("Row");
                          else XMLcmd=XMLcg.addCommand("Column");
                  XMLcmd.addScaPar("Code",TupleCode);
                  if (leaf){
                  sizeNameUsr="ColWidth";
                  sizeNameVS="Width";  
                  if (TupleType.equals("ROW")){sizeNameUsr="RowHeight";sizeNameVS="Height";}
                  String width=TupleEl.getProperty("WIDTH");
                  width=this.getVisualPref(TupleCode, sizeNameUsr,width);
                  XMLcmd.addScaParameter(sizeNameVS,width,def(""));
                  }
                  XMLcmd.addScaParameter("Parent",NVL(TupleParent,Axisel.getCode()),Axisel.getCode());
                  XMLcmd.addScaParameter("Caption",LTrim(TupleEl.getProperty("CAPTION")),def(""));//-- Заголовок
                  switch (NVL(getProperty(TupleType+"_INFO"),"N").charAt(0)){ //-- Доп информация
                      case 'N':
                      break;
                      case 'Y':
                      XMLcmd.addScaParameter("Info",TupleEl.getProperty("TUPLE_INFO"),def(""));
                      break;
                      case 'O':
                      XMLcmd.addScaParameter("Info",TupleEl.getProperty("ltn"),def(""));
                      break;
                  }
                  XMLcmd.addScaParameter("Visible",TupleEl.getProperty("VISIBLE"),def("Y"));
                  
                  String dir="H";
            switch (TupleEl.getProperty("DIRECTION")){
                case "TOP2BOTTOM":dir="TB";break;
                case "BOTTOM2TOP":dir="BT";break;
                case "VERTICAL":dir="V";break;
            }
             
                  XMLcmd.addScaParameter("Direction",dir,def(""));//-- Направление текста в заголовке
                  XMLcmd.addScaParameter("H-align",LuiString.withCapitalLetter(TupleEl.getProperty("ALIGNMENT")),def(""));
                  XMLcmd.addScaParameter("V-align",LuiString.withCapitalLetter(TupleEl.getProperty("VALIGN")),def(""));
                  XMLcmd.addScaParameter("Overflow",LuiString.withCapitalLetter(TupleEl.getProperty("OVERFLOW")),def(""));
                  XMLcmd.addScaParameter("Class",TupleEl.getProperty("VISUALIZATION"),def(""));
                  XMLcmd.addScaParameter("SelfClass",TupleEl.getProperty("TUPLE_VISUAL"),def(""));
                  XMLcmd.addScaParameter("MakeUp",TupleEl.getProperty("MAKEUP"),def(""));
                  XMLcmd.addScaParameter("GetUp",TupleEl.getProperty("GETUP"),def(""));
                  XMLcmd.addScaParameter("Hint",TupleEl.getProperty("HINT"),def(""));
              }
              
              if (TupleType.equals("ROW")) addProperty("LastROW",String.valueOf(ltn));
              if (TupleType.equals("COL")) addProperty("LastCOL",String.valueOf(ltn));
             } // конец кручения размерностей
           addProperty("CurrentTuple","");  
             
           if (MatrixMap!=null)MatrixMap.clear();// очищаем карту колоды карт
           
           else MatrixMap=new HashMap<>(); // а если её нет - создаём
           ArrayList<String> cellArray; // массив колоды в ячейке
           String MK; //адрес колоды в карте колод
           // -- Обходим прототипы ячеек
           
           
           /*for (String ck:elements.keySet()){
            if (!elements.get(ck).getType().equals("CELL"))continue; 
            clist.add(elements.get(ck));
           }*/
           for (int ci=0;ci<clist.size();ci++){ 
            LuiAppElement cellEl=clist.get(ci);
            String ck=cellEl.getCode();
             muSQLText=cellEl.getProperty("CELL_MULTIPLIER");
       //      muSQLText="[]";
             
             if (muSQLText.length()>5 && muSQLText.charAt(0)=='J' && muSQLText.charAt(1)=='S' && muSQLText.charAt(2)==':'){//-- JS- Размножитель есть. Значит, это ячейка-репликатор
                 cellEl.properties.remove("CELL_MULTIPLIER");
                try { 
                  JsonElement JSONRoot;  
                  if (muSQLText.charAt(3)=='<'&&muSQLText.charAt(4)=='=') {
                  JSONRoot = application.JSONparser.parse(getVal(muSQLText.substring(5)));    
                  } else
                  JSONRoot = application.JSONparser.parse(muSQLText.substring(3));
                  JsonArray JSONArray=JSONRoot.getAsJsonArray();
                  for (int mr=0;mr<JSONArray.size();mr++){
                      
                      //Первым делом нужно выяснить код строки и код колонки и их номера как личтовых
                      ColCode=null;RowCode=null;
                      for (Map.Entry<String, JsonElement> pEntry:JSONArray.get(mr).getAsJsonObject().entrySet()){
                      String pk=pEntry.getKey();
                      if (pk.equals("ROW")) {//-- Ищем  с кодом ROW
                      RowCode=pEntry.getValue().getAsString();//-- Нашли. Записем в RowCode   
                        }
                     if (pk.equals("COL")) {//-- Ищем  с кодом ROW
                       ColCode=pEntry.getValue().getAsString();;//-- Нашли. Записем в RowCode     
                       }
                     if (ColCode!=null && RowCode!=null)break;
                        }
                     if (is_not_null(RowCode)&& is_not_null(ColCode)){//-- есть куда поместить ячейку
                        newCellCode="c"+String.valueOf(CellNum); //-- 25.03.2021  
                        if (!elements.containsKey(RowCode)){//-- Нет ещё такой строки
                        
                        LuiAppElement rowEl=addElement(RowCode,"ROW",null,RowCode,NVL(firstRowsGroup,"Form")); // -- Создадим элемент типа строка у себя
                        int ltn=Integer.valueOf(getProperty("LastROW"));
                        ltn++;
                        rowEl.addProperty("ltn",String.valueOf(ltn));
                        setProperty("LastROW",String.valueOf(ltn));
                        XMLcmd=XMLcg.addCommand("Row"); //-- Отправим команду создания строки Жене
                        XMLcmd.addScaPar("Code",RowCode);
                       }
                       if (!elements.containsKey(ColCode)){//-- Нет ещё такого столбца
                        LuiAppElement colEl=addElement(ColCode,"COL",null,ColCode,NVL(firstColsGroup,"Form")); // -- Создадим элемент типа столбей у себя
                        int ltn=Integer.valueOf(getProperty("LastCOL"));
                        ltn++;
                        colEl.addProperty("ltn",String.valueOf(ltn));
                        setProperty("LastCOL",String.valueOf(ltn));
                        XMLcmd=XMLcg.addCommand("Column"); //-- Отправим команду создания колонки Жене
                        XMLcmd.addScaPar("Code",ColCode);   
                       }
                       MatrixCopyElement(ck,newCellCode,"CELL");//-- копируем шаблон ячейки в новый элемент
                       elements.get(newCellCode).addProperty("ROW",RowCode);
                       elements.get(newCellCode).addProperty("COL",ColCode);
                       elements.get(newCellCode).addProperty("Metadata","N");
                       elements.get(newCellCode).addProperty("MetadataElement",ck);
                       for (Map.Entry<String, JsonElement> pEntry:JSONArray.get(mr).getAsJsonObject().entrySet()){
                           String pk=pEntry.getKey();
                           if (pk.equals("ROW")||pk.equals("COL"))continue;
                           String pv=pEntry.getValue().getAsString();
                           PropName=pk.toUpperCase()/*.substring(0,31)*/; //-- каждое выражение - некое свойство ячейки
                           ItemText=pv;//-- значение свойства
                           if (PropName.toUpperCase().equals("PARENT")) {
                                if (ItemText==null) continue;
                                if (ItemText.equals(""))ItemText="Form";          
                                PropName=LuiString.withCapitalLetter(PropName);
                            }
                            if (ItemText!=null){
                            elements.get(newCellCode).addProperty(PropName,ItemText);//-- заменяем свойство в новом элементе типа CELL
                            //elements.get(newCellCode).properties.get(PropName).is_meta=false; // и говорим что оно не мета
                            elements.get(newCellCode).properties.get(PropName).genesis='A'; // и говорим что оно не мета
                            }
                       }
                       // теперь кладём ячейку в колоду карт
                       RowNum=elements.get(RowCode).getProperty("ltn");
                       elements.get(newCellCode).addProperty("RowNo",RowNum);
                       ColNum=elements.get(ColCode).getProperty("ltn");
                       elements.get(newCellCode).addProperty("ColNo",ColNum);
                       MK=RowNum+":"+ColNum;
                       if (!MatrixMap.containsKey(MK)){
                       cellArray=new ArrayList<>();
                       cellArray.add(newCellCode);
                       MatrixMap.put(MK, cellArray);
                       } else {
                       MatrixMap.get(MK).add(newCellCode);
                       }
                       CellNum++; //-- 25.03.2021
                     } //-- есть куда поместить ячейку
                  }
             } catch (Exception ex) {
                 String here="here2";
             }
            
             }//JS конец
                 if (like(muSQLText,"SQL:%") && is_not_null(muSQLText.substring(4))){//-- SQL- Размножитель есть. Значит, это ячейка-репликатор
                 cellEl.properties.remove("CELL_MULTIPLIER");
              try{
                   //берём соединение
                   ArrayList<String> acstmt=application.getAppConnAndStmt(muSQLText.substring(4),getProperty("Connection"),getFormID());
                   String acc=acstmt.get(0);
                   String stmt=acstmt.get(1);
                 if (lcComConn.containsKey(acc)){
                     lc=lcComConn.get(acc);
                 } else {
                   ac=application.userDbConnections.get(acc);
                  if (ac.getSessionMode()=='C') { // если соединение коммунальное
                   lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
                   lcComConn.put(acc, lc);
                   } else{                  //в противном случае возьмём текущее cоединение
                       lc=ac.currentConnection;
                         }
                    if (lc==null) // если соединение так и не нашлось, то
                     application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
                 }
                    // соединение есть
                 stmt=stmt.replace("$LeafRows$", application.toJSON(LeafRows));
                 stmt=stmt.replace("$LeafColumns$", application.toJSON(LeafColumns));
            
                    cursor=lc.CursorOpen(stmt); // открываем курсор
                    cols=lc.CursorGetColsInfo(cursor);//получаем список столбцов
                    L=cols.size();
                    while(lc.CursorNextRow(cursor)){ //крутим строки  -- Каждая строка - новая ячейка
                      cursRow=lc.CursorGetRow(cursor);
                      RowCode=null;
                      ColCode=null;
                      RowNum=null;
                      RowNum=null;
                       for (int F=0;F<L;F++){//-- Первым делом нужно выяснить код строки и код колонки
                        if (cols.get(F).alias.toUpperCase().equals("ROW")) {//-- Ищем столбец с кодом ROW
                         RowCode=cursRow.get(F);//-- Нашли. Записем в RowCode
                        }
                        if (cols.get(F).alias.toUpperCase().equals("COL")) {//-- Ищем столбец с кодом ROW
                         ColCode=cursRow.get(F);//-- Нашли. Записем в RowCode
                        }
                      }
                     if (is_not_null(RowCode)&& is_not_null(ColCode)){//-- есть куда поместить ячейку
                        //newCellCode=RowCode+rcs+ColCode;   
                        newCellCode="c"+String.valueOf(CellNum); //-- 25.03.2021  
                       if (!elements.containsKey(RowCode)){//-- Нет ещё такой строки
                        LuiAppElement rowEl=addElement(RowCode,"ROW",null,RowCode,NVL(firstRowsGroup,"Form")); // -- Создадим элемент типа строка у себя
                        int ltn=Integer.valueOf(NVL(getProperty("LastROW"),"0"));
                        ltn++;
                        rowEl.addProperty("ltn",String.valueOf(ltn));
                        setProperty("LastROW",String.valueOf(ltn));
                        XMLcmd=XMLcg.addCommand("Row"); //-- Отправим команду создания строки Жене
                        XMLcmd.addScaPar("Code",RowCode);
                       }
                       if (!elements.containsKey(ColCode)){//-- Нет ещё такого столбца
                        LuiAppElement colEl=addElement(ColCode,"COL",null,ColCode,NVL(firstColsGroup,"Form")); // -- Создадим элемент типа строка у себя
                        int ltn=Integer.valueOf(NVL(getProperty("LastCOL"),"0"));
                        ltn++;
                        colEl.addProperty("ltn",String.valueOf(ltn));
                        setProperty("LastCOL",String.valueOf(ltn));
                        
                        XMLcmd=XMLcg.addCommand("Column"); //-- Отправим команду создания колонки Жене
                        XMLcmd.addScaPar("Code",ColCode);   
                       }
                       MatrixCopyElement(ck,newCellCode,"CELL");//-- копируем шаблон ячейки в новый элемент
                       elements.get(newCellCode).addProperty("ROW",RowCode);
                       elements.get(newCellCode).addProperty("COL",ColCode);
                       elements.get(newCellCode).addProperty("Metadata","N");
                       elements.get(newCellCode).addProperty("MetadataElement",ck);
                       
                       for (int F=0;F<L;F++){//-- Смотрим все выражения запроса
                            PropName=cols.get(F).alias.toUpperCase()/*.substring(0,31)*/; //-- каждое выражение - некое свойство ячейки
                            ItemText=cursRow.get(F);//-- значение свойства
                            
                            if (PropName.toUpperCase().equals("PARENT")) {
                                if (ItemText==null) continue;
                                if (ItemText.equals(""))ItemText="Form";          
                                PropName=LuiString.withCapitalLetter(PropName);
                            }
                         /* 25.03.2021if (!elements.get(RowCode+rcs+ColCode).properties.containsKey(PropName)
                             || !elements.get(RowCode+rcs+ColCode).properties.get(PropName).is_meta){   
                            elements.get(RowCode+rcs+ColCode).addProperty(PropName,ItemText);//-- заменяем свойство в новом элементе типа CELL
                            elements.get(RowCode+rcs+ColCode).properties.get(PropName).is_meta=false;
                         } */
                         if (ItemText!=null){
                         elements.get(newCellCode).addProperty(PropName,ItemText);//-- заменяем свойство в новом элементе типа CELL 25.03.2021
                         //elements.get(newCellCode).properties.get(PropName).is_meta=false; // и говорим что оно не мета 25.03.2021
                          elements.get(newCellCode).properties.get(PropName).genesis='A'; // и говорим что оно не мета
                         }
                       }
                      // теперь кладём ячейку в колоду карт
                       RowNum=elements.get(RowCode).getProperty("ltn");
                       elements.get(newCellCode).addProperty("RowNo",RowNum);
                       ColNum=elements.get(ColCode).getProperty("ltn");
                       elements.get(newCellCode).addProperty("ColNo",ColNum);
                       MK=RowNum+":"+ColNum;
                       if (!MatrixMap.containsKey(MK)){
                       cellArray=new ArrayList<>();
                       cellArray.add(newCellCode);
                       MatrixMap.put(MK, cellArray);
                       } else {
                       MatrixMap.get(MK).add(newCellCode);
                       }
                     CellNum++; // 25.03.2021
                     } //-- есть куда поместить ячейку
                    } //крутим строки курсора
              } finally{
                  if (cursor!=null)
                    lc.CursorClose(cursor);
                    cursor=null;
                  /*  if (lc!=null && ac!=null) {
                       if (ac.getSessionMode()=='C') { // если соединение коммунальное
                        ac.connPool.SetFree(lc); // освободим его
                      }
                       ac=null;
                       lc=null;
                     } */
              }
             }  
             if (!like(muSQLText,"SQL:%") && !like(muSQLText,"JS:%")){//-- это не шаблон ячеек, а статическая ячейка
               RowCode=cellEl.getProperty("ROW");
               ColCode=cellEl.getProperty("COL");
               if (is_not_null(RowCode)&& is_not_null(ColCode)){//-- найдена координата ячейки
                  if (!elements.containsKey(RowCode)){//-- Нет ещё такой строки
                        LuiAppElement rowEl=addElement(RowCode,"ROW",null,RowCode,NVL(firstRowsGroup,"Form")); // -- Создадим элемент типа строка у себя
                        int ltn=Integer.valueOf(NVL(getProperty("LastROW"),"0"));
                        ltn++;
                        rowEl.addProperty("ltn",String.valueOf(ltn));
                        setProperty("LastROW",String.valueOf(ltn));
                        XMLcmd=XMLcg.addCommand("Row"); //-- Отправим команду создания строки Жене
                        XMLcmd.addScaPar("Code",RowCode);
                       } 
                  if (!elements.containsKey(ColCode)){//-- Нет ещё такого столбца
                        LuiAppElement colEl=addElement(ColCode,"COL",null,ColCode,NVL(firstColsGroup,"Form")); // -- Создадим элемент типа строка у себя
                        int ltn=Integer.valueOf(NVL(getProperty("LastCOL"),"0"));
                        ltn++;
                        colEl.addProperty("ltn",String.valueOf(ltn));
                        setProperty("LastCOL",String.valueOf(ltn));
                        
                        XMLcmd=XMLcg.addCommand("Column"); //-- Отправим команду создания колонки Жене
                        XMLcmd.addScaPar("Code",ColCode);   
                       }
                  //newCellCode=RowCode+rcs+ColCode;   
                        newCellCode="c"+String.valueOf(CellNum); //-- 25.03.2021  
                  MatrixCopyElement(ck,newCellCode,"CELL");//
                  LuiAppElement newel=elements.get(newCellCode);
                  for (String pk:newel.properties.keySet()){
                     if (cellEl.properties.containsKey(pk))
                      newel.properties.get(pk).genesis=cellEl.properties.get(pk).genesis;
                  }
                  newel.addProperty("MetadataElement",ck);
                  newel.addProperty("Metadata","Y");
                   // теперь кладём ячейку в колоду карт
                       
                       RowNum=elements.get(RowCode).getProperty("ltn");
                       elements.get(newCellCode).addProperty("RowNo",RowNum);
                       ColNum=elements.get(ColCode).getProperty("ltn");
                       elements.get(newCellCode).addProperty("ColNo",ColNum);
                       MK=RowNum+":"+ColNum;
                       if (!MatrixMap.containsKey(MK)){
                       cellArray=new ArrayList<>();
                       cellArray.add(newCellCode);
                       MatrixMap.put(MK, cellArray);
                       } else {
                       MatrixMap.get(MK).add(newCellCode);
                       }
                  CellNum++; // 25.03.2021
               }
             }
            elements.remove(ck);
           }// -- Обходим прототипы ячеек
           //MatrixMakeMap();  
         for (String acc:lcComConn.keySet()){ //Освобождаем занятые комуналки
             ac=application.userDbConnections.get(acc);
             ac.connPool.SetFree(lcComConn.get(acc));
         }
         
         
         //передаём ячейки на клиента
         for (String mk:MatrixMap.keySet()){
             RowCode=MatrixGetCellProperty(mk,"ROW");
             if (is_null(RowCode))continue;
             ColCode=MatrixGetCellProperty(mk,"COL");
             if (is_null(ColCode))continue;
              XMLcmd=XMLcg.addCommand("Cell");
              XMLcmd.addScaPar("R-code",RowCode);
              XMLcmd.addScaPar("C-code",ColCode);
              XMLcmd.addScaParameter("Show",MatrixGetCellProperty(mk,"CELL_SHOW"), def());
              //XMLcmd.addScaParameter("Show",MatrixGetCellProperty(mk,"CELL_TEXT"), def());
              XMLcmd.addScaParameter("Hint",MatrixGetCellProperty(mk,"CELL_HINT"), def());
                  XMLcmd.addScaParameter("H-align",LuiString.withCapitalLetter(MatrixGetCellProperty(mk,"CELL_HALIGN")), def());
                  XMLcmd.addScaParameter("V-align",LuiString.withCapitalLetter(MatrixGetCellProperty(mk,"CELL_VALIGN")), def());
                  XMLcmd.addScaParameter("Overflow",LuiString.withCapitalLetter(MatrixGetCellProperty(mk,"CELL_OVERFLOW")), def());
                  XMLcmd.addScaParameter("MakeUp",MatrixGetCellProperty(mk,"CELL_MAKEUP"), def());
                  XMLcmd.addScaParameter("GetUp",MatrixGetCellProperty(mk,"CELL_GETUP"), def());
                  XMLcmd.addScaParameter("Class",MatrixGetCellVisual(mk), def());
         }
//} //   WaitingForFormActivate
     this.MatrixPrepareToolBar(XMLcg);
     //this.prepareSysMenu();
     XMLcmd=XMLcg.addCommand("Form");
     properties.get("HEADER").clearCompVal();
     XMLcmd.addScaPar("Title",this.getFormTtle());
     XMLcmd=XMLcg.addCommand("Matrix");
     XMLcmd.addScaPar("AngularText",LTrim(getProperty("ANGULAR")));
     XMLcmd.addScaPar("AngularHint",getProperty("ANGULAR_HINT"));
     XMLcmd.addScaParameter("AngularWidth",this.getVisualPref("", "AngularWidth",getProperty("ANGULAR_WIDTH")),def(""));
     if (getProperty("COL_INFO").equals("O"))setProperty("COL_INFO","OB");
     if (getProperty("COL_INFO").equals("Y"))setProperty("COL_INFO","YB");
     if (NVL(getProperty("COL_INFO"),"N").equals("N")) XMLcmd.addScaPar("ColumnInfo","None");
     else { 
         switch (getProperty("COL_INFO").charAt(1)){
             case 'B':XMLcmd.addScaPar("ColumnInfo","Bottom");break;
             case 'T':XMLcmd.addScaPar("ColumnInfo","Top");break;
             case 'O':XMLcmd.addScaPar("ColumnInfo","Only");break;
        }
         
     }
     if (getProperty("ROW_INFO").equals("O"))setProperty("ROW_INFO","OR");
     if (getProperty("ROW_INFO").equals("Y"))setProperty("ROW_INFO","YR");
     if (NVL(getProperty("ROW_INFO"),"N").equals("N")) XMLcmd.addScaPar("RowInfo","None");
     else {
         switch (getProperty("ROW_INFO").charAt(1)){
             case 'R':XMLcmd.addScaPar("RowInfo","Right");break;
             case 'L':XMLcmd.addScaPar("RowInfo","Left");break;
             case 'O':XMLcmd.addScaPar("RowInfo","Only");break;
        }
     }
     XMLcmd.addScaPar("SelectionMode",withCapitalLetter(NVL(getProperty("SELECTION_MODE"),"STANDARD")));
     XMLcmd.addScaPar("Title", getProperty("GRID_TITLE"));
     XMLcmd.addScaPar("TitleBorder", getProperty("GRID_TITLE_BORDER"));
     XMLcmd.addScaPar("FormID",this.getFormID());
 } //   WaitingForFormActivate
     XMLcmd=XMLcg.addCommand("Form");
     properties.get("HEADER").clearCompVal();
     XMLcmd.addScaPar("Title",this.getFormTtle());
     XMLcmd=XMLcg.addCommand("MatrixFinish");
     XMLcmd.addScaPar("FormID",getFormID());

     break;
     
     case "LIST": // если список
        boolean totalstatus=false;
        XMLcmd.addScaParameter("ConfigList",this.getUsrConfigList(),def()); // передаём список конфигураций
        XMLcmd.addScaParameter("FetchedRows","",def());//-- количество выфетченных строк - не определено
        clearElementsValue();// Ни одной строки ещё не выбрано. Нет контекстной строки
        XMLcmdG=XMLcg.addCommand("Grid");
        switch (getProperty("START_MODE")) {
            case "MULTI_ROW":
             setProperty("SelectionMode","Multi");
             XMLcmdG.addScaPar("SelectionMode","Multi");    
             break;
            case "SINGLE_ROW":
             setProperty("SelectionMode","Single");
             XMLcmdG.addScaPar("SelectionMode","Single");    
             break;
            case "INVERSION":
             setProperty("SelectionMode","Multi");
             setVal("Inversion","Y");
             XMLcmdG.addScaPar("SelectionMode","Multi");                  
             break;
            default:
             setProperty("SelectionMode","Single");
             XMLcmdG.addScaPar("SelectionMode","None");    
             break;
        }
        
        XMLcmdG.addScaParameter("OverflowWidth",this.getVisualPref("","Overflow",""), def());
        
        //prepareSysMenu();                                  
        LuiCollection defGTypes = new LuiCollection(""); // коллекция типов групп, она нужна, чтобы понять нужна ли будет суррогатная группа для перемещений в неё стобцов
        
        // сортируем элементы в elList
//        makeElList("Form");
        //наполним тулбар иконками системного меню
        toolBarButtons=this.makeSysToolBarBattonCollection();
        // обходим элементы и применяем к ним сохранённое пользователем св-во Parent
        for (String ec:elements.keySet()){
            LuiAppElement el=elements.get(ec);
            if (el.getType().equals("ITEM")){
                     String elParent=NVL(getVisualPref(el.getCode(), "Parent", el.getParent()),"Form");
                     el.setProperty("Parent", elParent);
            }
        }
        this.elSortedList.clear(); // чистим сортированный массив чтобы заново построился
        for (int i=0;i<elements.size();i++){ // обходим последовательно элементы
            LuiAppElement el=elSort(i);
            switch (el.getType()) {
                case "ITEM":
                   if (el.getProperty("Read").equals("Y")) { // если поле показывается
                     //String elParent=NVL(getVisualPref(el.getCode(), "Parent", el.getParent()),"Form");
                     //el.setProperty("Parent", elParent);
                     String elParent=el.getParent();
                     if (!elParent.equals("Form")) {// если родитель не форма
                        if (!elements.containsKey(elParent)) { // и такого элемента нет
                          if (!defGTypes.containsKey(elParent)) { // и даже в коллекции типов нет 
                          XMLcmd=XMLcg.addCommand("Group");
                          XMLcmd.addScaParameter("Code", elParent,def());
                          //XMLcmd.addScaParameter("Type", "Overflow",def());
                          XMLcmd.addScaParameter("Type", LuiString.withCapitalLetter(elParent),def());
                          XMLcmd.addScaParameter("State", "Visible",def());
                          XMLcmd.addScaParameter("FieldsPromptPosition", "Up",def());
                          //addGroup(elParent,"","Y",null,"","OVERFLOW","Form","UP"); // создаём группу
                          addGroup(elParent,"","Y",null,"",elParent,"Form","UP"); // создаём группу
                          } else
                          {
                           elParent=defGTypes.get(elParent);
                          }    
                        }
                     }
                    XMLcmd=XMLcg.addCommand("Column"); 
                    XMLcmd.addScaParameter("Code", el.getCode(), def());
                    XMLcmd.addScaParameter("Group", elParent.replace("Form", ""), def());
                    XMLcmd.addScaParameter("Title",LuiString.LTrim(el.getProperty("CAPTION")), def());
                    XMLcmd.addScaParameter("Alignment",LuiString.withCapitalLetter(el.getProperty("ALIGNMENT")), def("Left"));
                    if (el.getProperty("VISIBLE").equals("Y")){
                    XMLcmd.addScaPar("Enable","Y");
                    
                    String tstr=this.getVisualPref(el.getCode(),"ColWidth", null);
                    if (tstr==null){
                        if (el.getProperty("WIDTH").equals("0")) tstr="0";
                        if (el.getProperty("WIDTH").equals("6")) tstr="6";
                    } 
                    if (el.getProperty("WIDTH").equals("Y")) tstr="";   // плевать на предпочтения если Y или 
                    if (el.getProperty("WIDTH").equals("N")) tstr="0";  // N
                    XMLcmd.addScaPar("Width",tstr);
                    el.setRTproperty();
                    if (is_not_null(el.getNonCompValue("TOTAL")))totalstatus=true;
                    if (is_null(getVal("CurrentColumn"))) { // если текущий столбей ещё неизвестен
                        if (NVL(el.getProperty("VISIBLE"),"N").equals("Y")) {// и если изуаемый - видим
                            if (Integer.valueOf(NVL(tstr,"1"))>0){
                                setVal("CurrentColumn",el.getCode());
                            }
                        }
                    }
                    }
                    else {
                    XMLcmd.addScaPar("Enable","N");
                    XMLcmd.addScaPar("Width","0");
                    }
                    XMLcmd.addScaParameter("Sort",el.getProperty("SORT"), def());
                    XMLcmd.addScaParameter("Hint",el.getProperty("HINT"), def());
                    //if (!like(el.properties.get("VISUALIZATION").startedValue,"%{%:%}%")) 19/05/2020
                    if (el.properties.containsKey("VISUALIZATION") && !el.properties.get("VISUALIZATION").isDynamic())
                    XMLcmd.addScaParameter("Style", el.getProperty("VISUALIZATION"), def()); // 29.04.2020
                   // if (!like(el.properties.get("MAKEUP").startedValue,"%{%:%}%"))
                    XMLcmd.addScaParameter("MakeUp", LuiString.RTrim(el.getProperty("MAKEUP"),"\n"), def()); // 29.04.2020
                    XMLcmd.addScaParameter("GetUp", el.getProperty("GETUP"), def()); // 05.11.2020
                   } // если поле показывается по правам
                break;
                case "ACTION":
                if (!like(el.getCode(),"SYSMENU$%") && !like(el.getCode(),"COLMENU$%") && !like(el.getCode(),"CELLMENU$%")){ // исключаем действия системного меню и меню столбца и QBE меню ячейки
                if (el.getProperty("Run").equals("Y")) { // если действие разрешено
                 if (!el.getNonCompValue("ACTIVE").equals("N")) { // Действие не вечно заблокировано 
                   // Это не понятно зачем и было решено закоментировать 17.02.2020
                   //if ((!"N".equals(el.getNonCompValue("ACTIVE"))) && (!"N".equals(el.getNonCompValue("APPLY_TO")))) {
                   //    if (!sm.equals("Multi")) sm="Single";
                   //}  
                   if (el.getParent().equals("Form")){ //если действие уровня формы
                     if (is_not_null(el.getProperty("ICON"))){ // если есть иконка 
                        if (toolBarButtons.containsKey(el.getProperty("ICON"))){ // если эта иконка уже в тулбаре 
                          //toolBarButtons.Add_Element(el.getProperty("ICON"),""); // гасим промпт иконки
                          this.addIcon2ToolBar(el.getProperty("ICON"),"","", toolBarButtons);
                     } else{ // если её ещё в тулбаре нет - выводим её хинт
                            // но перед выводом сравним декады
                            
                            if ((int)el.sourceSeq/10!=toolBarDecade){ // если декады разные ставим сепаратор
                             String keySep="";
                             for(int j=0;j<toolBarDecade;j++) keySep=keySep+" ";
                             //toolBarButtons.Add_Element(keySep,"");  
                             this.addIcon2ToolBar(keySep,"","", toolBarButtons);
                             toolBarDecade=(int)el.sourceSeq/10;
                            }
                             //toolBarButtons.Add_Element(el.getProperty("ICON"),el.getProperty("EXPLANATION"));  
                             this.addIcon2ToolBar(el.getProperty("ICON"),el.getProperty("PROMPT"),el.getProperty("EXPLANATION"), toolBarButtons);
                           }
                     }
                   } else { // если действие относится не к форме
                       // то готовим перечень иконок в свойстве "LocBut" поля для которого это действие предназначено
                       if (is_not_null(el.getNonCompValue("ICON"))){
                       String aPa=el.getParent();
                       String LocBut =elements.get(aPa).getProperty("LocBut");
                       LocBut=LocBut+el.getCode()+":;";
                       elements.get(aPa).setProperty("LocBut", LocBut);
                       }
                   }
                   if (is_not_null(el.getNonCompValue("BUTTON"))){ // -- Действие может вызываться кнопкой в подвале
                      XMLcmd=XMLcg.addCommand("Button");  
                      XMLcmd.addScaParameter("Code", el.getCode(), def());
                      XMLcmd.addScaParameter("Group",el.getParent(),"Form");
                      XMLcmd.addScaParameter("Hint",el.getProperty("EXPLANATION"),def());
                      XMLcmd.addScaParameter("Label",el.getProperty("BUTTON"),def());
                      if (is_null(el.getProperty("BUTTON"))){ // если не видна
                      XMLcmd.addScaParameter("State","Hidden",def());    
                      } 
                      else if(el.isActionActive()) { // и активна
                           XMLcmd.addScaParameter("State","Active",def());
                      }
                      else {
                          if (is_not_null(defIfSecond)) {
                           XMLcmd.addScaParameter("State","Disable",def());   
                          } 
                      }
                   } // 
                   if (like(el.getNonCompValue("ROLE"),"%TIME%")) {
                       setVal("Timer","Y"); // Признак необходимости "тикания таймером"
                   }
                 } // Действие не вечно заблокировано 
                }// если действие разрешено
                }
                break;
                case "GROUP":
                String groupType=el.getProperty("GROUP_TYPE");
                XMLcmd=XMLcg.addCommand("Group"); 
                XMLcmd.addScaParameter("Code", el.getCode(), def());
                XMLcmd.addScaParameter("Type",getGroupNameForXML(groupType),def("Tab"));
                XMLcmd.addScaParameter("Prompt",el.getProperty("TITLE"),def());
                String pp=getPromptPositionForGroup(el.getCode());
                pp=pp.replace("AUTO", "UP");
                pp=NVL(pp,"LEFT");
                pp=LuiString.withCapitalLetter(pp);
                XMLcmd.addScaParameter("FieldsPromptPosition",pp,def("Left"));
                XMLcmd.addScaParameter("Hint",el.getProperty("GROUP_HINT"),"");
                XMLcmd.addScaParameter("State","Visible",def(""));
                if (!defGTypes.containsKey(groupType)){
                    defGTypes.put(groupType,el.getCode());
                }
                break;
            }
        } // обходим последовательно элементы
    // 17,02.2020 if (is_null(getProperty("SelectionMode"))) { // если режима SelectionMode ещё нет
    //     setProperty("SelectionMode",sm);
    //     XMLcmdG.addScaPar("SelectionMode",sm);
    //   }
  //   setVal("ToolBarButtons",toolBarButtons.toString()); // установим значение св-ва ToolBarButtons
        setVal("ToolBarButtons",this.getToolBarString(toolBarButtons));
     XMLcmd=XMLcg.addCommand("Form");
     if (!getProperty("ToolBar").equals("N"))
     XMLcmd.addScaPar("ToolBarButtons",getVal("ToolBarButtons"));
     if (is_not_null(getNonCompValue("ADD_INFO"))){//Есть подвал -- LIST_ADD_INFO
       if (is_null(getVal("Second"))) { // Первый проход
           String tstr=this.getVisualPref("", "Underside", "");
           if (is_not_null(tstr)) {// Устанавливалась высота подвала
               XMLcmd=XMLcg.addCommand("Underside");
               XMLcmd.addScaParameter("Height", tstr, def());
           }
       }
     }
     if (totalstatus) { //Имеет смысл запрос итоговой строки
         XMLcmd=XMLcg.addCommand("Grid");
         XMLcmd.addScaParameter("Totals","Y",def());
         setVal("TotalStatus","Enable");
     }
     if (getVal("MasterForm").equals("Y")) { // Если это мастер
         XMLcmd=XMLcg.addCommand("Event");
         XMLcmd.addScaParameter("EventCode", "Requery", def());
         XMLcmd.addScaParameter("MS", "1", def());
         XMLcmd.addScaParameter("Params", "FormID:"+getFormID()+";", def());
     } else
     { // если не мастер то делаем перезапрос
         if (like(this.code,"O2P_L_OBJECT%")){
          System.out.println("before requery:"+System.currentTimeMillis()); // perfomance

      }
         requery();
         if (like(this.code,"O2P_L_OBJECT%")){
          System.out.println("after requery:"+System.currentTimeMillis()); // perfomance

      }
     }
     break;
     case "BLANK": // если бланк
      HashMap<String,Integer> grpFieldsCount=new HashMap<>(); //заводим перечень групп с указанием количества полей в них
      for (String elCode:elements.keySet()){
          if (elements.get(elCode).getType().equals("FIELD")){
              elements.get(elCode).addProperty("Correct","");
              if (!elements.get(elCode).getParent().equals("Form")) {
                String gCode=elements.get(elCode).getParent();
                Integer fldCnt=1;
                if  (grpFieldsCount.containsKey(gCode)){
                     fldCnt=grpFieldsCount.get(gCode);
                     fldCnt++;
                     grpFieldsCount.replace(gCode, fldCnt);
                } else grpFieldsCount.put(gCode, fldCnt);
                    
              }
          }
      }
     
     for (int i=0;i<elements.size();i++) { //циклим последовательно элементы
         LuiAppElement el=elSort(i);
         switch (el.getType()){
             case "GROUP": // если элемент это группа
             XMLcmd=XMLcg.addCommand("Group"); // Группа бланка
             XMLcmd.addScaParameter("Group",el.getParent() ,"Form"); //Вложенность
             XMLcmd.addScaParameter("Code",el.getCode() ,def()); //Вложенность
             if (NVL(el.getProperty("Read"),"Y").equals("Y")) { //если есть право читать
                 XMLcmd.addScaParameter("Type",getGroupNameForXML(el.getProperty("GROUP_TYPE")),def("Tab")); //Тип
                 XMLcmd.addScaParameter("Prompt",el.getProperty("TITLE"),def()); //Заголовок
                // if (NVL(el.getProperty("SEEABLE"),"N").equals("Y")) { // -- Состояние -- GROUP_VISIBLE
                if (el.isVisible(false)){
                     XMLcmd.addScaParameter("State","Visible",def());
                 } else {
                      if (is_not_null(defIfSecond)) {
                           XMLcmd.addScaParameter("State","Hidden",def());   
                          } 
                 }
                String pp=getPromptPositionForGroup(el.getCode()); //Положение подписей элементов группы
                if (pp.equals("AUTO")) { // если авто
                    if (el.getProperty("GROUP_TYPE").equals("SIDE_BY_SIDE")) { // и группа всё поперёк
                      if  (grpFieldsCount.containsKey(el.getCode())){
                       if (grpFieldsCount.get(el.getCode())>1) { // и полей на ней более одного
                           pp="UP";
                       }
                      }
                    }
                }
                pp=LuiString.withCapitalLetter(pp);
                XMLcmd.addScaParameter("FieldsPromptPosition", pp ,def("Left"));
             }//если есть право читать
             break;
             case "FIELD": // если поле
             if (NVL(el.getProperty("Read"),"Y").equals("Y")) { //если есть право читать
             XMLcmd=XMLcg.addCommand("Field"); // Поле бланка  
             if (el.getCode().equals("NODE_TEXT")){
                String s=el.getProperty("Value");
             }
             XMLcmd.addScaParameter("Code",el.getCode() ,def()); //код
             XMLcmd.addScaParameter("Group",el.getParent() ,"Form"); //группа в которой поле
             XMLcmd.addScaParameter("Prompt",el.getProperty("CAPTION") ,def()); // подпись
             XMLcmd.addScaParameter("Hint",el.getProperty("HINT") ,def()); // подпись
             XMLcmd.addScaParameter("HelperText",el.getProperty("HELPER_TEXT") ,def()); // подпись
             //XMLcmd.addScaParameter("Show",el.getFieldShow(),def());
             XMLcmd.addScaPar("Show",el.getFieldShow());
             XMLcmd.addScaParameter("Pattern",el.getProperty("PATTERN"),def());
                    if (el.properties.containsKey("StartCursor") && el.properties.containsKey("EndCursor")){
                      if (el.getProperty("StartCursor").equals(el.getProperty("EndCursor"))) {
                        XMLcmd.addScaPar("Cursor",el.getProperty("StartCursor"));  
                      }  
                    }
             String Lines=el.getLines4view(); // количество строк
             if (Lines.equals("0")) { // если ноль - не показываем
               XMLcmd.addScaParameter("Type","Hidden",def());
               el.setVal("Type","Hidden");
             }else{ // если поле надо показать
               XMLcmd.addScaParameter("Lines",Lines,def("1"));     
              if (is_not_null(el.getProperty("VERIFIER"))){//если есть верификатор
                   String vrf=el.getProperty("VERIFIER");
                   if (application.verifiers!=null){// и верификаторы есть 
                   if (application.verifiers.containsKey(vrf)) { // и он есть в массиве верификаторов
                       XMLcmd.addScaPar("Verifier", "fieldVerify."+vrf);
                       // крутим параметры верификатора
                       LuiCollection vrfParams=new LuiCollection("");
                       boolean vpp=false;
                       for (String vpC:el.properties.keySet()){
                        if (like(vpC,"VERIFIER_PARAM.%")){
                            vrfParams.Add_Element(vpC.replace("VERIFIER_PARAM.",""), el.getProperty(vpC));
                            vpp=true;
                        }
                       }
                       if (!vpp) { // если параметров верификатора не найдено
                          LuiAppForm vrfFrm=application.verifiers.get(vrf); 
                          if (vrfFrm.elements!=null){
                              if (vrfFrm.elements.size()>0){ // а параметры, то есть!
                                  for(int pind=0;pind<vrfFrm.elements.size();pind++){ // крутим его элементы (там бывают только параметры
                                    LuiAppElement pvrf=vrfFrm.elSort(pind); 
                                    if (pvrf.getType().equals("PARAM")){ 
                                    if (pvrf.properties.containsKey("Value")){
                                     el.addProperty("VERIFIER_PARAM."+pvrf.getCode(),pvrf.properties.get("Value").startedValue); // под каждый параметр создаём проперть подя
                                    }
                                    }
                                }
                                // и вновь вычисляем значения  
                                for (String vpC:el.properties.keySet()){
                                if (like(vpC,"VERIFIER_PARAM.%")){
                                    vrfParams.Add_Element(vpC.replace("VERIFIER_PARAM.",""), el.getProperty(vpC));
                                 }
                                }
                              }
                          }
                       }
                       if (vrfParams.size()>0){
                         XMLcmd.addScaPar("VerifierParams", vrfParams.toString());   
                        }
                 }
                }
               }//если есть верификатор
              if (el.getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR")) {//если цвет
                 XMLcmd.addScaPar("Type","Color"); 
                 XMLcmd.addScaPar("Mode","Readonly");              
              } else { // не цвет
               LuiCollection sliderCol=null;
               if (like(el.getProperty("VARIANTS"),"LOVR:%")){ // слайдер
                 sliderCol=new LuiCollection(el.getProperty("VARIANTS").substring(5));
                 if (!(sliderCol.containsKey("Min") && sliderCol.containsKey("Max") && sliderCol.containsKey("Step") && sliderCol.containsKey("ValueFrom"))){
                     sliderCol=null;
                 } 
               }
               if (sliderCol!=null) { //если слайдер LOVR:Min:;Max:;Step:;
                 // создадим у поля свойства для контроля изменений грациц и положения
                 el.setVal("OldMin",sliderCol.Get_Element("Min"));
                 el.setVal("OldMax",sliderCol.Get_Element("Max"));
                 el.setVal("OldValueFrom",sliderCol.Get_Element("ValueFrom"));
                 el.setVal("OldStep",sliderCol.Get_Element("Step"));
                 XMLcmd.addScaPar("Type","Range");   
                 if (el.getProperty("CHANGING").equals("N")){
                   XMLcmd.addScaPar("Mode","Readonly");                
                 }
                 XMLcmd.addScaPar("ValueMin", sliderCol.Get_Element("Min"));
                 XMLcmd.addScaPar("ValueMax", sliderCol.Get_Element("Max"));
                 XMLcmd.addScaPar("ValueStep", sliderCol.Get_Element("Step"));
                 XMLcmd.addScaPar("ValueFrom", sliderCol.Get_Element("ValueFrom"));
               } else { // не слайдер 
               String Type="Text";
               String Mode="Editable";
               String KeyNavigable="Y";
               if (el.getProperty("DATA_TYPE").equals("RICHTEXT")) {//если это форматированный текст
                 Type="RichText";
                 XMLcmd.addScaParameter("Type","RichText",def());          
                 if (!el.getProperty("CHANGING").equals("Y")){ //а руками вводить нельзя
                 Mode="ReadOnly";
                 XMLcmd.addScaPar("Mode","Readonly");     
                 }
               } else { //если не RICHTEXT
                 String tstr="ReadOnly";
                 switch (el.getProperty("CHANGING")){
                     case "Y": tstr="Text";break;
                     case "U": tstr="Upper";break;
                     case "H": tstr="Passwd";break;
                     case "B": tstr="CheckBox";break;
                     case "R": tstr="Radio";break;
                 }
                 if (tstr.equals("ReadOnly")){
                   Type="Text";
                   XMLcmd.addScaParameter("Type","Text",def("Text"));
                   if (!el.getProperty("DATA_TYPE").equals("WIKITEXT")){Mode="Readonly";XMLcmd.addScaPar("Mode","Readonly");}
                   else {Mode="Wiki";XMLcmd.addScaPar("Mode","Wiki");}
                   if (!el.getProperty("CHANGING").equals("C")){
                       KeyNavigable="N";
                       XMLcmd.addScaParameter("KeyNavigable","N",def("Y"));
                   } 
                 } else {
                   Type=tstr;
                   XMLcmd.addScaParameter("Type",tstr,def("Text"));
                   if (!el.getProperty("DATA_TYPE").equals("WIKITEXT")){Mode="Editable";XMLcmd.addScaPar("Mode","Editable");}
                   else {
                       if (el.Code.equals(getVal("CurrentField"))){Mode="Editable";XMLcmd.addScaPar("Mode","Editable");}
                       else {Mode="Wiki";XMLcmd.addScaPar("Mode","Wiki");}
                   }
                   KeyNavigable="Y";
                   XMLcmd.addScaParameter("KeyNavigable","Y",def("Y"));
                 }
                     
               }
               el.setVal("Type",Type);
               el.setVal("Mode",Mode);
               el.setVal("KeyNavigable",KeyNavigable);
               } // не слайдер
              }// если не цвет
             } // если поле надо показать 
             XMLcmd.addScaParameter("Style",el.getProperty("VISUALIZATION"),def());
//05.11.2020             if (is_not_null(el.getProperty("MAKEUP"))){
//05.11.2020                 XMLcmd.addScaPar("MakeUp",LuiString.RTrim(el.getProperty("MAKEUP"),"\n"));
//05.11.2020             }
             XMLcmd.addScaParameter("MakeUp",el.getProperty("MAKEUP"),def()); //05.11.2020       
             XMLcmd.addScaParameter("GetUp",el.getProperty("GETUP"),def());  //05.11.2020          
             if (el.getProperty("OPTIONAL").equals("N")){
                   XMLcmd.addScaParameter("Mandatory","Y",def());
             } else {
               if (is_not_null(defIfSecond)){
                   XMLcmd.addScaParameter("Mandatory","N",def());
               }
             }
             if (el.getProperty("CHANGING").equals("B")){
                 if (is_null(el.getProperty("Value")) && is_not_null(application.getProperty("CHECK_BOX_NULL_ICON")))
                 XMLcmd.addScaParameter("Icon",application.getProperty("CHECK_BOX_NULL_ICON"),def());    
                 else
                 XMLcmd.addScaParameter("Icon",el.getProperty("Value"),def());
             }
             if (el.getProperty("CHANGING").equals("R")){
                 XMLcmd.addScaParameter("Values",el.getProperty("VARIANTS"),def());
             }    
             }//если есть право читать    
             break;
             case "ACTION":
             if (!like(el.getCode(),"SYSMENU$%")){//если не SYSMENU
             if (NVL(el.getProperty("Run"),"Y").equals("Y")) { //если есть право выполнять
               String elParent=el.getParent();
               String FormOrGroupOrField=elParent;
               if (!"Form".equals(elParent)){ // не уровень формы
                   if (elements.containsKey(elParent)) { // и есть такой элемент
                       FormOrGroupOrField=elements.get(elParent).getType();
                   }
               }
               if (FormOrGroupOrField.equals("Form") || FormOrGroupOrField.equals("GROUP")) { //если действие уровня бланка или группы
                XMLcmd=XMLcg.addCommand("Button");
                XMLcmd.addScaParameter("Code", el.getCode(), def());
                XMLcmd.addScaParameter("Group", elParent, def("Form"));
                String prompt=el.getProperty("PROMPT");
                
                 String ico=el.getProperty("ICON");
                 if (is_not_null(ico)){
                     prompt="[ico:"+ico+"]"+prompt;
                 }
                
                XMLcmd.addScaParameter("Label",prompt, def());
                XMLcmd.addScaParameter("Hint", el.getProperty("EXPLANATION"), def());
                if (el.getProperty("DISPLAY").equals("Y")){ // если кнопка видима
                    if (el.getProperty("ACTIVE").equals("Y")){ // если кнопка активна
                      XMLcmd.addScaParameter("State","Active", def());    
                    } else {// если не активна
                        if (is_not_null(defIfSecond)){
                            XMLcmd.addScaParameter("State","Disable", def());    
                        }
                    }
                } else {// если кнопка невидима
                  XMLcmd.addScaParameter("State","Hidden", def());  
                }
               }
               else if (FormOrGroupOrField.equals("FIELD")) {//если действие уровня поля
                String icon=el.getProperty("ICON");
               if (is_not_null(icon)) {//если есть иконка
                 el.setProperty("OldIcon",icon); //Сохраним текущее значение для следующего события
                 String fieldButtons=elements.get(elParent).getProperty("Buttons"); // читаем кнопки поля
                 XMLcmd=XMLcg.addCommand("FieldButton");
                 if (elements.get(elParent).getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR")){// если цвет
                 XMLcmd.addScaPar("Code",icon);
                 // добавим в действие свойство BgColor если его там нет
                 if (!el.properties.containsKey("BgColor")){
                    el.addProperty("BgColor","{JS:if (\"{Property:Value}\")\"{Property:Value}\";else \"rgba(255,255,255,0)\";}");
                 }
                 XMLcmd.addScaPar("BgColor", el.getProperty("BgColor"));
                 XMLcmd.addScaParameter("Field",elParent, def());
                 if (el.getProperty("ACTIVE").equals("Y")){
                 XMLcmd.addScaParameter("State","Active", def());    
                 } else {
                 XMLcmd.addScaParameter("State","Disable", def());    
                 }
                 } else {// если не цвет
                 XMLcmd.addScaParameter("Code", icon, def());
                 
                 XMLcmd.addScaParameter("Field",elParent, def());
                 if (el.getProperty("ACTIVE").equals("Y")){
                 XMLcmd.addScaParameter("State","Active", def());    
                 } else {
                 XMLcmd.addScaParameter("State","Hidden", def());    
                 }
                 }
                 LuiCollection cFieldButtons=new LuiCollection(fieldButtons);
                 if (cFieldButtons.Exists_Element(icon)) {//если в поле такая кнопка уже есть
                    if (is_not_null(cFieldButtons.Get_Element(icon))){ // и у кнопки есть код действия
                        cFieldButtons.replace(icon,""); // удаояем описание
                    } 
                 } else{//а если в поле такой кнопки ещё нет
                   XMLcmd.addScaParameter("Hint",el.getProperty("EXPLANATION"), def());// рисуем на ней хинт
                  cFieldButtons.Add_Element(icon, el.getCode());//добавляем кнопку в поле
                  fieldButtons=cFieldButtons.toString();
                  elements.get(elParent).setProperty("Buttons", fieldButtons);
                 }
               }//если есть иконка
               }//если действие уровня поля
              }//если есть право выполнять
             if (!"N".equals(el.getNonCompValue("ACTIVE"))) { //если действие незакрыто наглухо
             if (like(el.getNonCompValue("ROLE"),"%TIME%")) {  // а роль содержит TIME
                        setVal("Timer","Y");                                   //Признак необходимости "тикания таймером"
             }
             }
             } //если не SYSMENU
             break;
         } // switch по типам элементов
     } // цикл по элементам бланка
//06/06/2020 ещё цикл по полям
     for (String elCode:elements.keySet()){
       if (elements.get(elCode).getType().equals("FIELD")){
           LuiAppElement el=elements.get(elCode);
           LuiAppProp p=el.properties.get("Value");
           for (LuiAppProp fp: p.dependsFrom.keySet()){
             fp.depends.remove(p);
           }
           if (p.dependsFrom!=null)p.dependsFrom.clear();
           
           application.parser.findDependency(el.getNonCompValue("RENOVATION"), el.properties.get("RENOVATION"));
       }  
     }
//06/06/2020 конец     
     setField(getVal("CurrentField"));
     String cf=getVal("CurrentField");
     if (is_not_null(cf)){
         XMLcmd=XMLcg.addCommand("CurrentField");
         XMLcmd.addScaParameter("FormID",this.getFormID(), def());
         XMLcmd.addScaParameter("Field",cf, def());
     }
    
    checkFrmDetails();
    //19.03.2020    checkRunTimer();
    application.pingRunTimer();
     break;
     case "TREE":
     setVal("CurrentRow","");//-- текущая строка - неопределена
     setVal("FetchedRows","");// -- количество выфетченных строк - не определено
     setProperty("SelectionMode","Single");
     // для того чтобы отработал построитель тулбара присвоим всем столюцам пустые строки
    /* for (String ik:elements.keySet()){
         if (elements.get(ik).getType().equals("ITEM")){
             if (!elements.get(ik).properties.containsKey("Value")){
                 addProperty("Value","");
             }
         }
     }
     this.prepareTB4TREE(XMLcg);*/
     requery();
     this.prepareTB4TREE(XMLcg);
     break;
     case "MENU":
       showMenuItems("",null);
     break;
     case "HTML":
     XMLcmd=XMLcg.addCommand("Html");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Data",this.getProperty("HTML_GENERATOR"));
     for (String ec:elements.keySet()){
         if (!elements.get(ec).getType().equals("ITEM"))continue;
         LuiAppElement el=elements.get(ec);
         try{
         JsonElement JSONRoot = application.JSONparser.parse(el.getProperty("DEFINITION"));
         JsonElement JSONEl=null;
         for (Map.Entry<String, JsonElement> pEntry:JSONRoot.getAsJsonObject().entrySet()){
         String pk=pEntry.getKey();
         String pv=pEntry.getValue().getAsString();
         el.addProperty(pk, pv);
         el.properties.get(pk).gen4html=true;
         }
         } catch(Exception ex){}
         XMLcmd=XMLcg.addCommand("SetProperties");
         XMLcmd.addScaPar("FormID",this.getFormID());
         XMLcmd.addScaPar("Code",el.getProperty("HTML_ELEMENT"));
         XMLcmd.addScaPar("Properties",el.getProperty("DEFINITION"));
     }
     checkFrmDetails();
     //19.03.2020    checkRunTimer();
     application.pingRunTimer();
     break;    
     case "CHART":
       //chartDestroy(XMLcg);
       
       if (is_not_null(getProperty("ChartStatus"))){
     if (XMLcg==null) XMLcg=application.commandWriter.addGroup("ChartDestroy");
     XMLcmd=XMLcg.addCommand("Chart");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Command", "clear");
     XMLcmd=XMLcg.addCommand("Chart");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Command", "clear");
     XMLcmd=XMLcg.addCommand("Chart");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Command", "dispose");
     properties.remove("ChartStatus");
       }
       setVal("Scale","1");
       prepareTB4Chart(XMLcg);   
       chart(XMLcg);
       
       
       checkFrmDetails();
     //19.03.2020    checkRunTimer();
     application.pingRunTimer();
     break;
 }
 setVal("Status","Run");
 setAllChangeFlag(false);
 setVal("Second","Y");
}
/** Подготовка тулбара для древовидных списоков
 * 
 **/
 public void prepareTB4TREE(LuiClCommandGroup pXMLcg) throws LuiException, SQLException{
    String root = getVal("RootFormID");
    if (!application.forms.containsKey(root))return;
    String context=application.forms.get(root).getVal("ContextForkID");
    
    if (!getFormID().equals(NVL(context,getFormID()))) return;
     LuiClCommand XMLcmd=null; 
    LuiCollection toolBarButtons=new LuiCollection("");
     //обрабатываем действия
     int toolBarDecade=0;
        //for (String elcode:elements.keySet())  {
        for (int ei=0;ei<elements.size();ei++){
         LuiAppElement el=elSort(ei);
         if (!el.getType().equals("ACTION"))continue;
         if (!el.getProperty("Run").equals("Y"))continue;
         if (like(el.getCode(),"SYSMENU$%") )continue; // исключаем действия системного меню 
         if (!el.getParent().equals("Form"))continue; //если действие уровня формы
         el.properties.get("BUTTON").recalculate=true;
         el.properties.get("ICON").recalculate=true;
         el.properties.get("ACTIVE").recalculate=true;
         if (!like(el.getProperty("BUTTON"),"[ico:_%]%")) continue; // если есть описание иконочки в button
         // вычленяем иконочеу
         String ico=LuiString.getIco(el.getProperty("BUTTON"));
         String prompt=el.getProperty("BUTTON").replace("["+ico+"]","");
         ico=ico.replace("ico:","");
         if (is_null(ico))continue; // если нет иконки пропускаем
         if (toolBarButtons.containsKey(ico)){ // если эта иконка уже в тулбаре 
            this.addIcon2ToolBar(ico,"","", toolBarButtons); // гасим промпт иконки 
         } else {
             // если её ещё в тулбаре нет - выводим её хинт
                            // но перед выводом сравним декады
         if (((int)el.sourceSeq/10)!=toolBarDecade){ // если декады разные ставим сепаратор
                             String keySep="";
                             for(int j=0;j<toolBarDecade;j++) keySep=keySep+" ";
                             addIcon2ToolBar(keySep,"","", toolBarButtons); 
                             toolBarDecade=(int)el.sourceSeq/10;
                            }
                            this.addIcon2ToolBar(ico,prompt,el.getProperty("EXPLANATION"), toolBarButtons);
                                                          
         }
        }
        String ntb=this.getToolBarString(toolBarButtons);
        if (!getVal("ToolBarButtons").equals(ntb)){
        setVal("ToolBarButtons",this.getToolBarString(toolBarButtons)); // установим значение св-ва ToolBarButtons
        XMLcmd=pXMLcg.addCommand("Form");   
        XMLcmd.addScaPar("FormID",getVal("RootFormID"));
        if (!getProperty("ToolBar").equals("N"))    
         XMLcmd.addScaPar("ToolBarButtons",getVal("ToolBarButtons"));
        }
 }
    /**
     * Подготовка тулбара для диаграмм м матриц
     * @param pXMLcg неизвестно что
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
    public void prepareTB4Chart(LuiClCommandGroup pXMLcg) throws LuiException, SQLException{
        LuiClCommand XMLcmd=null;
         
        LuiCollection toolBarButtons=this.makeSysToolBarBattonCollection();
     //обрабатываем действия
     int toolBarDecade=0;
        //for (String elcode:elements.keySet())  {
        for (int ei=0;ei<elements.size();ei++){
            
         
         LuiAppElement el=elSort(ei);
        
         if (el.getType().equals("ACTION")){
         if (!like(el.getCode(),"SYSMENU$%") ){ // исключаем действия системного меню 
                if (el.getProperty("Run").equals("Y")) { // если действие разрешено
                 if (!el.getNonCompValue("ACTIVE").equals("N")) { // Действие не вечно заблокировано 
                   
                   if (el.getParent().equals("Form")){ //если действие уровня формы
                     if (is_not_null(el.getProperty("ICON"))){ // если есть иконка 
                         
                        if (toolBarButtons.containsKey(el.getProperty("ICON"))){ // если эта иконка уже в тулбаре 
                            
                            this.addIcon2ToolBar(el.getProperty("ICON"),"","", toolBarButtons); // гасим промпт иконки
                     } else{ // если её ещё в тулбаре нет - выводим её хинт
                            // но перед выводом сравним декады
                            if (((int)el.sourceSeq/10)!=toolBarDecade){ // если декады разные ставим сепаратор
                             String keySep="";
                             for(int j=0;j<toolBarDecade;j++) keySep=keySep+" ";
                             addIcon2ToolBar(keySep,"","", toolBarButtons); 
                             toolBarDecade=(int)el.sourceSeq/10;
                            }
                            this.addIcon2ToolBar(el.getProperty("ICON"),el.getProperty("PROMPT"),el.getProperty("EXPLANATION"), toolBarButtons);
                              
                           }
                     }
                   } else { // если действие относится не к форме
                       // то готовим перечень иконок в свойстве "LocBut" поля для которого это действие предназначено
                       if (is_not_null(el.getNonCompValue("ICON"))){
                       String aPa=el.getParent();
                       String LocBut =elements.get(aPa).getProperty("LocBut");
                       LocBut=LocBut+el.getCode()+":;";
                       elements.get(aPa).setProperty("LocBut", LocBut);
                       }
                   }
                   if (is_not_null(el.getNonCompValue("BUTTON"))){ // -- Действие может вызываться кнопкой в подвале
                      XMLcmd=pXMLcg.addCommand("Button");  
                      XMLcmd.addScaParameter("Code", el.getCode(), def());
                      XMLcmd.addScaParameter("Group",el.getParent(),"Form");
                      XMLcmd.addScaParameter("Hint",el.getProperty("EXPLANATION"),def());
                      XMLcmd.addScaParameter("Label",el.getProperty("BUTTON"),def());
                      if (is_null(el.getProperty("BUTTON"))){ // если не видна
                      XMLcmd.addScaParameter("State","Hidden",def());    
                      } 
                      else if(el.isActionActive()) { // и активна
                           XMLcmd.addScaParameter("State","Active",def());
                      }
                      else {
                          if (is_not_null(defIfSecond)) {
                           XMLcmd.addScaParameter("State","Disable",def());   
                          } 
                      }
                   } // 
                   if (like(el.getNonCompValue("ROLE"),"%TIME%")) {
                      if (is_null(getVal("Timer"))) 
                       setVal("Timer","Y"); // Признак необходимости "тикания таймером"
                   }
                   if (like(el.getNonCompValue("ROLE"),"%TIME500%")) {
                       setVal("Timer","500"); // Признак необходимости "тикания таймером"
                   }
                 } // Действие не вечно заблокировано 
                }// если действие разрешено
                }    
         
       /*  if (like(el.getNonCompValue("ROLE"),"%TIME%")) {
                 setVal("Timer","Y");//Признак необходимости "тикания таймером"
             }
         if (!like(el.getCode(),"SYSMENU$%") && !like(el.getCode(),"COLMENU$%") && !like(el.getCode(),"CELLMENU$%")){ // исключаем действия системного меню и меню столбца и QBE меню ячейки
         if (el.getProperty("Run").equals("Y")) { // если действие разрешено
             if (!el.getNonCompValue("ACTIVE").equals("N")) { // Действие не вечно заблокировано 
             if (el.getParent().equals("Form")){ //если действие уровня формы
                if (is_not_null(el.getNonCompValue("BUTTON"))){ // -- Действие может вызываться кнопкой в подвале
                    if (pXMLcg==null) pXMLcg=application.commandWriter.addGroup("Chart");
                      XMLcmd=pXMLcg.addCommand("Button");  
                      XMLcmd.addScaParameter("Code", el.getCode(), def());
                      XMLcmd.addScaParameter("Group",el.getParent(),"Form");
                      XMLcmd.addScaParameter("Hint",el.getProperty("EXPLANATION"),def());
                      XMLcmd.addScaParameter("Label",el.getProperty("BUTTON"),def());
                      if (is_null(el.getProperty("BUTTON"))){ // если не видна
                      XMLcmd.addScaParameter("State","Hidden",def());    
                      } 
                      else if(el.isActionActive()) { // и активна
                           XMLcmd.addScaParameter("State","Active",def());
                      }
                      else {
                          if (is_not_null(defIfSecond)) {
                           XMLcmd.addScaParameter("State","Disable",def());   
                          } 
                      }
                   }
             }
          }
         }    
         }
         }*/
       }   
    
    
    }
        setVal("ToolBarButtons",this.getToolBarString(toolBarButtons)); // установим значение св-ва ToolBarButtons
        XMLcmd=pXMLcg.addCommand("Form");    
    if (!getProperty("ToolBar").equals("N"))    
    XMLcmd.addScaPar("ToolBarButtons",getVal("ToolBarButtons"));
    }
    /**
     * 
     * Графопостроитель
     * @param pXMLcg неизвестно что 
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
 public void chart(LuiClCommandGroup pXMLcg) throws LuiException, SQLException{
     if (!application.luiSeance.getServer().getChartLicense()){
       /*  String url=application.parser.plugins.get("ENCR").compute("lui.fors.ru", properties.get("CHART_GENERATOR"));
         String text=application.parser.plugins.get("ENCR").compute(
                 "Этот функционал доступен только в лицензионной версии LUI, приобрести которую можно на сайте ", 
                 properties.get("CHART_GENERATOR"));
         String appologies=application.parser.plugins.get("ENCR").compute("Приносим свои извинения", 
                 properties.get("CHART_GENERATOR"));
         String urle=application.parser.plugins.get("ENCR").compute("lui.fors.com", properties.get("CHART_GENERATOR"));
         String texte=application.parser.plugins.get("ENCR").compute(
                 "This feature is only available in the licensed version which can be purchased on the site ", 
                 properties.get("CHART_GENERATOR"));
         String appologiese=application.parser.plugins.get("ENCR").compute(
                 "Please accept our apologies", 
                 properties.get("CHART_GENERATOR"));*/
        String text="";
        String url="";
        String apologies="";
         if (application.NLS.currentLang.code.equals("RUS")){
           url=application.parser.plugins.get("DECR").compute("24PrypEfa5mlXHG57iL1zg==",properties.get("CHART_GENERATOR"));
           text=application.parser.plugins.get("DECR").compute("ojfTxt5ajcu8OMBrkQKaRI1zExxnzO9tCJFtUROIkw9zY8i8JHLOfLAIXjdRE+yK/McI3ANH6TN4KDUpxp5Br4e0q7UNhf4wfC3ivvofwEcq3d1yPxQF5nwtSqYQ2Z8JC9UKCjsGMCnrkiaPcKeCd0LCp0cXOtCvB6qe0US9zG/6LE4WsRjXo/uRsg5bdMtrI8INKcVeCprpMMCvnxRHl5mAOe1oxXxs3xZ2/rmwXFs=",properties.get("CHART_GENERATOR"));
           apologies=application.parser.plugins.get("DECR").compute("BWRD44+WF2tFbn0vUEdIl8KZs8wt0eyKPxIM01d8TDIi2eGhozy4D7H7hy8768qc",properties.get("CHART_GENERATOR")); 
         } else {
           url=application.parser.plugins.get("DECR").compute("24PrypEfa5l/Tsp1XGMn4A==",properties.get("CHART_GENERATOR"));
           text=application.parser.plugins.get("DECR").compute("yNYQ3CYeCD2R++uq24E/10s9IKP1XN4KLOPgfmdkIZe/ed6c7IeOKpE36NO7sFlvthaWwklH4B+RgRAKJMxwT5mbTfSB9YhAvf/FOvySZ84g/v+T1jHh25ex3otAmQ+z",properties.get("CHART_GENERATOR"));
           apologies=application.parser.plugins.get("DECR").compute("RUqSDl6R7FjsyxxVzqO2Q0Rm7zayB2W2nbMhfRJLTZI=",properties.get("CHART_GENERATOR"));
         }
         
         application.parser.plugins.get("HTML").execute("(Title:;Target:M;)"+
                 "<!DOCTYPE html>\n" +
"<HTML><HEAD>\n" +
"<meta http-equiv=\"X-UA-Compatible\" content=\"IE=Edge\">\n" +
"<META http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">\n" +
"</HEAD>\n" +
"<body style=\"padding:10pt;margin:0;background-color: #172548;color:white;text-align:center;font: normal 200% Arial, Verdana, Tahoma;\">\n" +
"<h3 style=\"font-size:200%;color: #FFB;text-shadow: 1px 2px 3px red;\">"+text+"<a href=\"http://"+url+"/#distribution\" target=\"_blank\" style=\"color:#BFF\">"+url+"</a></h3>\n" +
"<p>"+apologies+"</p>\n" +
"</body>\n" +
"</html>"
                 ,application.properties.get("AppCode"));
         
         return;
     }
     /* перенесено перед ON-LOAD
     for(String pc:properties.keySet()){
     
         if (like(pc,"Click%")){
             properties.get(pc).currentValue="";
             properties.get(pc).startedValue="";
         }
         
     }*/
 
     LuiClCommand XMLcmd=null;
     if (pXMLcg==null) pXMLcg=application.commandWriter.addGroup("Chart");
       XMLcmd=pXMLcg.addCommand("Chart");
       XMLcmd.addScaPar("FormID",this.getFormID());
   if (is_not_null(getProperty("ChartStatus"))){
       XMLcmd.addScaPar("Command","setOption");
   }   
   if (elements.containsKey("LEGEND_MODE")){
     if  (elements.get("LEGEND_MODE").getProperty("Value").equals("EVENT")){
         XMLcmd.addScaPar("LegendMode", "Event");
     }
   }
/*   if (elements.containsKey("MAP_POSITION")){
     if (is_not_null(elements.get("MAP_POSITION").getProperty("Value"))){
         String pos=elements.get("MAP_POSITION").getProperty("Value");
         XMLcmd.addScaPar("Position", pos);
         pos=pos.replace("[","").replace("]","");
         addProperty("MapPositionX",pos.split(",")[0]);
         addProperty("MapPositionY",pos.split(",")[1]);
     }  
   }*/
   //XMLcmd.addScaPar("Data",getProperty("CHART_GENERATOR"));    // это без отложенного выполнения
   XMLcmd.addJPar("Data", getProperty("CHART_GENERATOR"));
   if (properties.containsKey("CHART_GEN_ADDON") && is_not_null(properties.get("CHART_GEN_ADDON").startedValue))
   XMLcmd.addJPar("PieData", getProperty("CHART_GEN_ADDON"));
   String orientation="";
    if (NVL(getProperty("CHART_SIZE_TYPE"),"W").equals("WL"))orientation="Landscape";
    if (NVL(getProperty("CHART_SIZE_TYPE"),"W").equals("WP"))orientation="Portrait";
    XMLcmd.addScaPar("Orientation",orientation);
    setVal("LastChartOrientation",orientation);
    String chartWidth="";
    String chartHeight="";
    if (is_null(orientation) && like(NVL(getProperty("CHART_SIZE_TYPE"),"W"),"F%")){
      chartWidth=getProperty("CHART_WIDTH");
      chartHeight=getProperty("CHART_HEIGHT");
    }
    XMLcmd.addScaPar("Width",chartWidth);
    setVal("LastChartWidth",chartWidth);
    XMLcmd.addScaPar("Height",chartHeight);
    setVal("LastChartHeight",chartHeight);
       addProperty("ChartStatus","RUN");
}
 /**
  * Реакция на клик в легенду
  */
 public void chartLegendClick(String legendElementName,String pDC) throws LuiException, SQLException{
     addProperty("ClickLegendElementName",legendElementName);
     // ищем действие с ролью CHART_CLICK пока только уровня формы
      for (int i=0;i<elements.size();i++){ // крутим элементы
           LuiAppElement el=elSort(i);
       if (el.getParent().equals("Form")){ // относящиеся к заданному отцу
         if (el.getType().equals("ACTION")) { // только действия 
           if (like(el.getProperty("ROLE"),"%CHART_LEGEND_"+pDC+"%")){ // если действие имееет роль CHART_LEGEND...
             if (el.getProperty("ACTIVE").equals("Y")){
               if (el.getProperty("Run").equals("Y")){
                       runAction(el.getCode());
               }  
             }
           }
         }
       }
      }
 }
 
public void htmlAction(String pActionCode,String pParams) throws LuiException, SQLException {
    if (!elements.containsKey(pActionCode)) return;
    LuiAppElement action=elements.get(pActionCode);
    if (!action.getType().equals("ACTION")) return;
    JsonElement JSONRoot = application.JSONparser.parse(pParams);
    if (JSONRoot.getAsJsonObject().isJsonObject()){
    JsonObject jo=JSONRoot.getAsJsonObject();
    for (int i=0;i<elements.size();i++){
         LuiAppElement el=elSort(i);
         if (!el.getParent().equals(pActionCode))continue;
         if (!el.getType().equals("PARAM"))continue;
         JsonElement je=null;
         try{
         je=jo.get(el.getCode());
         } catch (Exception ex){je=null;}
         if (je!=null && !je.isJsonNull() && je.isJsonPrimitive() ){
           if (el.properties.containsKey("Value")){  
           //el.properties.get("Value").setCurrentValue(je.getAsString());
           //el.properties.get("Value").setChangeFlag();
           el.addProperty("Value",je.getAsString());
           }
           if (el.properties.containsKey("VALUE")){  
           //el.properties.get("VALUE").setCurrentValue(je.getAsString());
           //el.properties.get("VALUE").setChangeFlag();
           el.addProperty("VALUE",je.getAsString());
           }
         }
         
    }
    }
   this.runAction(pActionCode);
}
/*public void chartMapPosition(String pParams){
    addProperty("MapPosition",pParams);
    JsonElement JSONRoot = application.JSONparser.parse(pParams);
    if (JSONRoot.isJsonArray()){
           JsonArray JSONdataArray=JSONRoot.getAsJsonArray();
           addProperty("MapPositionX",JSONdataArray.get(0).getAsString());
           addProperty("MapPositionY",JSONdataArray.get(1).getAsString());
    }
}*/
    public void chartResize(String pWidth,String pHeight) throws LuiException, SQLException{
   addProperty("ChartWidth",pWidth);  
   addProperty("ChartHeight",pHeight);  
     String ROLE="%CHART_RESIZE%";
  // ищем действие с ролью ROLE пока только уровня формы
      for (int i=0;i<elements.size();i++){ // крутим элементы
           LuiAppElement el=elSort(i);
       if (el.getParent().equals("Form")){ // относящиеся к заданному отцу
         if (el.getType().equals("ACTION")) { // только действия 
           if (like(el.getProperty("ROLE"),ROLE)){ // если действие имееет роль CHART_CLICK
             if (el.getProperty("ACTIVE").equals("Y")){
               if (el.getProperty("Run").equals("Y")){
                       runAction(el.getCode());
               }  
             }
           }
         }
       }
      }

}
public void chartZoom(String pScale) throws LuiException, SQLException{
  addProperty("MapScale",pScale); 
  String ROLE="%CHART_ZOOM%";
  // ищем действие с ролью ROLE пока только уровня формы
      for (int i=0;i<elements.size();i++){ // крутим элементы
           LuiAppElement el=elSort(i);
       if (el.getParent().equals("Form")){ // относящиеся к заданному отцу
         if (el.getType().equals("ACTION")) { // только действия 
           if (like(el.getProperty("ROLE"),ROLE)){ // если действие имееет роль CHART_CLICK
             if (el.getProperty("ACTIVE").equals("Y")){
               if (el.getProperty("Run").equals("Y")){
                       runAction(el.getCode());
               }  
             }
           }
         }
       }
      }
  
}
 /**
  * Реакция на клик в данные на диаграмме
  * @param pParams параметры
  * @throws LuiException логическая ошибка
  * @throws SQLException ошибка выполнения SQL-запроса
  */
 public void chartClick(String pMouseButton,String pParams,String defComponetType) throws LuiException, SQLException{
     //boolean b =true;
     //while(b){
     //b=false;    
     String click="Click";
     
     if (pMouseButton.equals("Over"))click="MouseOver";
     if (pMouseButton.equals("Out"))click="MouseOut";
     for(String pc:properties.keySet()){
         if (like(pc,click+"%")){
      //       properties.get(pc).currentValue="";
      //       properties.get(pc).startedValue="";
      //properties.remove(pc);b=true;break;
         addProperty(pc,"");
         }
     }
     //}
     String clct="";
     if (is_not_null(defComponetType)){addProperty(click+"ComponentType",defComponetType);
     clct=defComponetType;
             }
     addProperty(click+"JSONparams",pParams);
     JsonElement JSONRoot = application.JSONparser.parse(pParams);
     JsonElement JSONEl=null;
     try{
     JSONEl=JSONRoot.getAsJsonObject().get("componentType");
     }catch (Exception ex){} 
     if (JSONEl!=null) {
     clct=JSONEl.getAsString();    
     addProperty(click+"ComponentType",JSONEl.getAsString());
     }
     clct=LuiString.withCapitalLetter(clct);
      JSONEl=null;
     try{
     JSONEl=JSONRoot.getAsJsonObject().get("componentIndex"); 
     }catch (Exception ex){} 
     if (JSONEl!=null){// есть ID компонента
         String compind=JSONEl.getAsString();
         if (DgV2!=null && clct.equals("Series") && (DgV2.chartType.equals("BOX") || DgV2.chartType.equals("BOXG"))){
             compind=String.valueOf(Integer.valueOf(compind)/2);
         }
         addProperty(click+clct+"СomponentIndex",compind);
         addProperty(click+"СomponentIndex",compind);
         
     }
     JSONEl=JSONRoot.getAsJsonObject().get("targetType");
     if (JSONEl!=null){// есть адресность тыка axisName или axisLabel
              addProperty(click+"TargetType",JSONEl.getAsString());
              addProperty(click+clct+"TargetType",JSONEl.getAsString());
     }
     JSONEl=JSONRoot.getAsJsonObject().get("name");
     if (JSONEl!=null){// есть имя оси или данных
              addProperty(click+"Name",JSONEl.getAsString());
              addProperty(click+clct+"Name",JSONEl.getAsString());
              addProperty(click+"DataName",JSONEl.getAsString());
              addProperty(click+clct+"DataName",JSONEl.getAsString());
     }
     JSONEl=JSONRoot.getAsJsonObject().get("value");
     
     if (JSONEl!=null){ 
         if (!JSONEl.isJsonArray()){// есть значение метки оси или значение серии
             if (JSONEl.isJsonNull()) {
              addProperty(click+"Value","");
              addProperty(click+clct+"Value","");
              addProperty(click+"DataValue","");   
              addProperty(click+clct+"DataValue","");   
             }else {
              addProperty(click+"Value",NVL(JSONEl.getAsString(),""));
              addProperty(click+clct+"Value",NVL(JSONEl.getAsString(),""));
              addProperty(click+"DataValue",NVL(JSONEl.getAsString(),""));
              addProperty(click+clct+"DataValue",NVL(JSONEl.getAsString(),""));
             }
         }
         else {addProperty(click+"DataValue",JSONEl.toString());
               addProperty(click+clct+"DataValue",JSONEl.toString());
         // данные массив - раскроем элементы массива
           JsonArray JSONdataArray=JSONEl.getAsJsonArray();
         for (int i=0;i<JSONdataArray.size();i++){
            if (!JSONdataArray.get(i).isJsonNull()){ 
             addProperty(click+"DataValueDim"+String.valueOf(i),JSONdataArray.get(i).getAsString());
             addProperty(click+clct+"DataValueDim"+String.valueOf(i),JSONdataArray.get(i).getAsString());
         } else {
             addProperty(click+"DataValueDim"+String.valueOf(i),"nUlL");    
             addProperty(click+clct+"DataValueDim"+String.valueOf(i),"nUlL");    
                 } 
         }
         }
     } else { // нет значения
         addProperty(click+"DataValue","nUlL");  
         addProperty(click+clct+"DataValue","nUlL");  
         addProperty(click+"DataValueDim0","nUlL");  
         addProperty(click+clct+"DataValueDim0","nUlL");  
         addProperty(click+clct+"DataValueDim1","nUlL");  
         addProperty(click+clct+"DataValueDim1","nUlL");  
         addProperty(click+clct+"DataName","nUlL");  
         addProperty(click+clct+"DataName","nUlL");  
     }
     JSONEl=JSONRoot.getAsJsonObject().get("seriesID"); 
     if (JSONEl!=null){// есть ID серии
         addProperty(click+"SeriesID",JSONEl.getAsString());
     }
     JSONEl=JSONRoot.getAsJsonObject().get("seriesIndex"); 
     if (JSONEl!=null){// есть ID серии
         addProperty(click+"SeriesID",JSONEl.getAsString());
     }
     JSONEl=JSONRoot.getAsJsonObject().get("seriesName");
     if (JSONEl!=null){// есть имя серии
     addProperty(click+"SeriesName",JSONEl.getAsString());
     JSONEl=JSONRoot.getAsJsonObject().get("dataIndex"); // берём индекс данных
     if (JSONEl!=null){// есть индекс
         addProperty(click+"DataIndex",JSONEl.getAsString());
         addProperty(click+clct+"DataIndex",JSONEl.getAsString());
     }
     JSONEl=JSONRoot.getAsJsonObject().get("data"); // берём данные
          if (JSONEl!=null){// есть данные
         JsonElement JSONdataValue=null;
     JsonElement JSONdataName=null;
         addProperty(click+"Data",JSONEl.toString());
         addProperty(click+clct+"Data",JSONEl.toString());
         try{
         JSONdataName=JSONEl.getAsJsonObject().get("name");
         } catch (Exception ex){}
       if (JSONdataName!=null){// есть имя у точки данных
         if (!JSONdataName.isJsonNull()){
         addProperty(click+"DataName",JSONdataName.getAsString());
         addProperty(click+clct+"DataName",JSONdataName.getAsString());
         JSONdataValue=JSONEl.getAsJsonObject().get("value");
         if (JSONdataValue!=null){// есть value у точки данных
           addProperty(click+"DataValue",JSONdataValue.toString());  
           addProperty(click+clct+"kDataValue",JSONdataValue.toString());  
         }} else {
             JSONdataValue=JSONEl;  
         addProperty(click+"DataValue",JSONdataValue.toString());  
         addProperty(click+clct+"DataValue",JSONdataValue.toString());  
         addProperty(click+"DataValueDim0","nUlL");  
         addProperty(click+clct+"DataValueDim0","nUlL");  
         addProperty(click+"DataName","nUlL");  
         addProperty(click+clct+"DataName","nUlL");  
         }
        } else {// у точки данных имени нет
         JSONdataValue=JSONEl;  
         addProperty(click+"DataValue",JSONdataValue.toString());  
         addProperty(click+clct+"DataValue",JSONdataValue.toString());  
       }
      if (JSONdataValue!=null && JSONdataValue.isJsonArray()){ // если точка данных это массив
         JsonArray JSONdataArray=JSONdataValue.getAsJsonArray();
         for (int i=0;i<JSONdataArray.size();i++){
            if (!JSONdataArray.get(i).isJsonNull()){ 
             addProperty(click+"DataValueDim"+String.valueOf(i),JSONdataArray.get(i).getAsString());
             addProperty(click+clct+"DataValueDim"+String.valueOf(i),JSONdataArray.get(i).getAsString());
         } else {
             addProperty(click+"DataValueDim"+String.valueOf(i),"nUlL");    
             addProperty(click+clct+"DataValueDim"+String.valueOf(i),"nUlL");    
                 } 
         }
      } 
      
     }// есть данные
         }
     String ROLE="CHART";
     String MB="";
     switch (pMouseButton){
         case "Left":MB="_CLICK";break;
         case "Right":MB="_RMB";break;
         case "LeftDouble":MB="_DC";break;
         case "Over":MB="_OVER";break;
         case "Out":MB="_OUT";break;
     }
     String COMPONENT="";
     boolean dmc=true;
     if (elements.containsKey("DECOMPOSE_MOUSE_CLICK")){
        LuiAppElement eldmc =elements.get("DECOMPOSE_MOUSE_CLICK");
        if (eldmc.getType().equals("PARAM")&& eldmc.getParent().equals("Form")){
          if (eldmc.getValue().equals("N")) dmc=false; 
        }
     }
     if (dmc){
        switch (getProperty(click+"ComponentType")){
            case "xAxis":
            case "yAxis":
            COMPONENT="_AXIS";break;
            case "series":
            COMPONENT="";break;
            case "legend":
            COMPONENT="_LEGEND";break;
            case "title":
            COMPONENT="_TITLE";break;  
            case "graphic":
            COMPONENT="_GRAPHIC";break;  
            case "colormap":
            COMPONENT="_COLORMAP";break;    
        } 
     }
     ROLE=ROLE+COMPONENT+MB;
     ROLE="%"+ROLE+"%";
     if (ROLE.equals("%CHART_RMB%")){ // если ПКМ на графике прям
          LuiClCommandGroup XMLcg=null;
          LuiClCommand XMLcmd=null;
                  
          
          for (int i=0;i<elements.size();i++){ // крутим элементы
           LuiAppElement el=elSort(i);
       if (!el.getParent().equals("Form")) continue;
         if (!el.getType().equals("ACTION")) continue;
           if (!like(el.getProperty("ROLE"),ROLE)) continue;
              if (el.getNonCompValue("ACTIVE").equals("N")) continue; // Действие не вечно заблокировано 
               if (!el.getProperty("Run").equals("Y")) continue;
                  if (XMLcg==null){
                      XMLcg=application.commandWriter.addGroup("ToolBarButton"); 
                      XMLcg.addParameter("FormID",getFormID());
                      XMLcg.addParameter("Type","Action");
                      /*XMLcmd=*/XMLcg.addCommand("LocalMenu");
                  }
                  XMLcmd=XMLcg.addCommand("AddMenuItem");
                  XMLcmd.addScaPar("Code",el.getCode());
        String Prompt=el.getProperty("PROMPT");  
        if (!el.getProperty("ACTIVE").equals("Y")){
          XMLcmd.addScaPar("Enable","N");
          Prompt=RemoveIco(Prompt);
        }
        XMLcmd.addScaPar("Label",Prompt);
                 
         }
      
     } else { // если не ПКМ на графике
     // ищем действие с ролью ROLE пока только уровня формы
      for (int i=0;i<elements.size();i++){ // крутим элементы
           LuiAppElement el=elSort(i);
       if (el.getParent().equals("Form")){ // относящиеся к заданному отцу
         if (el.getType().equals("ACTION")) { // только действия 
           if (like(el.getProperty("ROLE"),ROLE)){ // если действие имееет роль CHART_CLICK
             if (el.getProperty("ACTIVE").equals("Y")){
               if (el.getProperty("Run").equals("Y")){
                       runAction(el.getCode());
               }  
             }
           }
         }
       }
      }
     }
 }
 public void chartClearClickData(){
     for(String pc:properties.keySet()){
         if (like(pc,"Click%") || like(pc,"MouseOver%")|| like(pc,"MouseOut%")){
      //       properties.get(pc).currentValue="";
      //       properties.get(pc).startedValue="";
      //properties.remove(pc);b=true;break;
         addProperty(pc,"");
         properties.get(pc).clearCompVal();
         }
     }

 }
 /**
  * Графочиститель
  * @param pXMLcg неизвестно что
  * @throws LuiException логическая ошибка
  * @throws SQLException ошибка выполнения SQL-запроса
  */
 public void chartClear(LuiClCommandGroup pXMLcg) throws LuiException, SQLException{
    LuiClCommand XMLcmd=null;
    if (is_not_null(getProperty("ChartStatus"))){
     if (pXMLcg==null) pXMLcg=application.commandWriter.addGroup("ChartClear");
     XMLcmd=pXMLcg.addCommand("Chart");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Command", "clear");}
     for(String pc:properties.keySet()){
         properties.get(pc).clearCompVal();
         if (like(pc,"Click%")){
             properties.get(pc).currentValue="";
             properties.get(pc).startedValue="";
         }
         
     }
     if (elements!=null){
     for(String ec:elements.keySet()){
      if (elements.get(ec).properties!=null){
         for(String pc:elements.get(ec).properties.keySet()){
         elements.get(ec).properties.get(pc).clearCompVal();
     }   }
     }
     }
     setProperty("ChartStatus","CLEAR");
  
 }
 
 /**
  * Графоразрушитель
  * @param pXMLcg неизвестно что
  * @throws LuiException логическая ошибка
  * @throws SQLException ошибка выполнения SQL-запроса
  */
 public void chartDestroy(LuiClCommandGroup pXMLcg) throws LuiException, SQLException{
     LuiClCommand XMLcmd=null;
     if (is_not_null(getProperty("ChartStatus"))){
     if (pXMLcg==null) pXMLcg=application.commandWriter.addGroup("ChartDestroy");
     XMLcmd=pXMLcg.addCommand("Chart");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Command", "clear");
     XMLcmd=pXMLcg.addCommand("Chart");
     XMLcmd.addScaPar("FormID",this.getFormID());
     XMLcmd.addScaPar("Command", "dispose");
     }
     properties.remove("ChartStatus");
     for(String pc:properties.keySet()){
         properties.get(pc).clearCompVal();
     }
     if (elements!=null){
     for(String ec:elements.keySet()){
      if (elements.get(ec).properties!=null){
         for(String pc:elements.get(ec).properties.keySet()){
         elements.get(ec).properties.get(pc).clearCompVal();
     }   }
     }
     }
 }
public void saveChart() throws LuiException, SQLException{
    LuiClCommandGroup XMLcg=application.commandWriter.addGroup("SaveChart");
    LuiClCommand XMLcmd=null;
    XMLcmd=XMLcg.addCommand("SaveChart");
    XMLcmd.addScaPar("FormID",this.getFormID());
    XMLcmd.addScaPar("FileName",NVL(NVL(Jsoup.parse(getProperty("HEADER")).text(),getProperty("ConfigName")),this.code));
}
public void showMenuItems(String pParent,String pContext) throws LuiException, SQLException{
 LuiClCommandGroup XMLcg=null;
 LuiClCommand XMLcmd=null;
 String initialItem=null;
 if (is_null(pParent)){
  setVal("info4chageNode","");
 }
 XMLcg=application.commandWriter.addGroup("DeleteChilds "+pParent);
       XMLcmd=XMLcg.addCommand("DeleteChilds");
       XMLcmd.addScaPar("FormID", this.getFormID());
       XMLcmd.addScaPar("Node",pParent);
 XMLcg=application.commandWriter.addGroup("Menu "+pParent);
       XMLcg.addParameter("FormID", this.getFormID());
       XMLcg.addParameter("ParentNode",pParent);
       XMLcg.addParameter("Type","ROW");
       XMLcg.addParameter("ContextID",this.getFormID());
       XMLcg.addParameter("ChildsContextID",this.getFormID());
       // родительский пункт мог быть сгенерирован запросом поэтому вычислим елемент который стал источником
       String trueParent=pParent;
       if (is_null(trueParent)) trueParent="Form";
       else { 
           //LuiAppElement pmel=this.findAndMute(trueParent);
           LuiAppElement pmel=null;
           LuiCollection concol=new LuiCollection(pContext);
           if (elements.containsKey(concol.Get_Element("ACTION"))){
              pmel= elements.get(concol.Get_Element("ACTION"));
           }
             if (pmel!=null)  trueParent=pmel.getCode();
               
       
       }
      String mute=pParent.replace(trueParent,"");
      
      for (int i=0;i<elements.size();i++){ // крутим элементы
       LuiAppElement el=elSort(i);
       if (el.getParent().equals(trueParent)){ // относящиеся к заданному отцу
         if (el.getType().equals("ACTION")) { // только действия не системного меню и не CLOSE и не RMB
          if (!like(el.getCode(),"SYSMENU$%") && !like(el.getProperty("ROLE"),"%CLOSE%") && !like(el.getProperty("ROLE"),"%RMB%")){ 
             ShowMenuItem(el.getCode(),XMLcg,mute,pContext);
             } 
         }
      
       }
      }
   if (is_not_null(pParent)){
            XMLcmd=XMLcg.addCommand("Node");
            XMLcmd.addScaPar("Node",pParent);
            XMLcmd.addScaPar("State","Expanded");
   }
   if (is_null(pParent)){
      if (is_not_null(getVal("info4chageNode"))){
          LuiCollection cncol=new LuiCollection(getVal("info4chageNode"));
          this.changeNode(cncol.Get_Element("Node"),cncol.Get_Element("State"),cncol.Get_Element("Type"),cncol.Get_Element("Context"),cncol.Get_Element("ContextID"),cncol.Get_Element("ChildsContextID"));
         /* XMLcmd=XMLcg.addCommand("Event");
          XMLcmd.addScaPar("EventCode","ChangeNode");
          XMLcmd.addScaPar("MS","0");
          XMLcmd.addScaPar("Params",getVal("info4chageNode"));*/
          //setVal("info4chageNode","");
      }
   }
}
/**
 * Вычисляет  видимость хотябы одного подузла (включая всех потомков) заданного узла
 * @param pCode код узла
 * @return признак видимости 
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean isVisibleSubNode(String pCode) throws LuiException, SQLException{
    for (String subel: elements.keySet()){
       if (elements.get(subel).getParent().equals(pCode)){ 
         if (elements.get(subel).getType().equals("ACTION")){
          if ( !like(elements.get(subel).getProperty("ROLE"),"%CLOSE%") 
                  && !like(elements.get(subel).getProperty("ROLE"),"%RMB%")){   
             if (elements.get(subel).getProperty("DISPLAY").equals("Y") || like(elements.get(subel).getProperty("DISPLAY"),"SQL:%")){
               return true;                    
             } else {
               if (isVisibleSubNode(subel))  return true;
             }
          }   
         }  
       }
    }
 return false;  
}

public void setDynamicMenuItemProperty(LuiAppElement el,LuiCollection concol) throws LuiException, SQLException{
    boolean setted=false;
    for (String pc:concol.keySet()){
      if (pc.equals("ACTION"))continue;  
       if (el.properties.containsKey(pc)){
        LuiAppProp p=el.properties.get(pc);
        if (!p.gen4menu) continue; //  из контекста забираем только сгенерированные новые
        
        p=el.addProperty(pc, concol.Get_Element(pc));
        setted=true;
        p.gen4menu=true;
       } else {
        LuiAppProp p=el.addProperty(pc, concol.Get_Element(pc));
        setted=true;
        p.gen4menu=true;
       }
    }
    if (setted) { // если что-то поменялось установим признак необходимости вычислений у всех св-в элемента и у его подчинённых элементов
        for (String pc:el.properties.keySet()){
            el.properties.get(pc).clearCompVal();
        }
        for (String ec:elements.keySet()){
            if (elements.get(ec).getParent().equals(el.getCode())){
         for (String pc:elements.get(ec).properties.keySet()){
            elements.get(ec).properties.get(pc).clearCompVal();
        }       
            }
        }
    }
}
/**
 * Отображение пункта меню
 * @param elCode -- код пункта (действия)
 * @param coGr -- группа команд
 * @param pMute -- муть
 * @param pContext -- контекст
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void ShowMenuItem(String elCode,LuiClCommandGroup coGr,String pMute,String pContext) throws LuiException, SQLException{
     LuiClCommandGroup XMLcg=null;
     LuiClCommand XMLcmd=null;
      LuiCollection concol=new LuiCollection(NVL(pContext,""));
      concol.Add_Element("ACTION", elCode);
    LuiAppElement el=elements.get(elCode);
    // устанавливаем у элемента проперти из контекста
    setDynamicMenuItemProperty(el, concol);
    boolean vsn=this.isVisibleSubNode(elCode);
     if (coGr!=null) {
        XMLcg=coGr; 
     } else     {
       XMLcg=application.commandWriter.addGroup("Menu Item");
       XMLcg.addParameter("FormID", this.getFormID());
       String parent=el.getParent();
       if (!parent.equals("Form")) {
       XMLcg.addParameter("ParentNode",el.getParent());
       }
       XMLcg.addParameter("Type","ROW");
       XMLcg.addParameter("ContextID",this.getFormID());
       XMLcg.addParameter("ChildsContextID",this.getFormID());
     } 
       // необходимость показа пункта определяется видимостью самого пункта или наличием видимых подпунктов
       // кроме того пункты могут размножаться, но без переноса вложенных веток
       String AV; 
       AV=el.getProperty("DISPLAY");
       if (like(AV,"SQL:%")) { // в DISPLAY лежит SELECT
          ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4),getProperty("Connection"),this.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          // крутим извлечённые строки
          if (propsArray!=null){  // если строки извлечены
            for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                boolean gen4menu=true;
                LuiAppProp p;
                if (el.properties.containsKey(propkey)) { // если свойство уже есть
                  p=el.properties.get(propkey); 
                  gen4menu=p.gen4menu; //берём его признак генерированности
                p=el.addProperty(propkey, props.Get_Element(propkey));
                } else {
                p=el.addProperty(propkey, props.Get_Element(propkey)); //  пополняем проперти действия
                }
                p.gen4menu=gen4menu;
                if (gen4menu){ //  в контекст закидывали только сгенерированные новые
                concol.Add_Element(propkey, props.Get_Element(propkey));//и добавляем их в контекст если это сгенерировано для меню
                }
             } // конец кручения пропертей строки
            el.properties.get("PROMPT").clearCompVal();
            el.properties.get("EXPLANATION").clearCompVal();
            el.properties.get("ACTIVE").clearCompVal();
            el.properties.get("ICON").clearCompVal();
            el.properties.get("QUESTION").clearCompVal();
            String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) { // если PROMPT не пуст
               // делаем узел для этой строки
            XMLcmd=XMLcg.addCommand("Node");
            XMLcmd.addScaPar("Node",el.getCode()+pMute+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            //сохраним в контексте данные об настоящем элементе
           
            concol.Add_Element("ACTION",el.getCode());
            XMLcmd.addScaPar("Context",concol.toString());
            XMLcmd.addScaPar("Label",Prompt);
            XMLcmd.addScaPar("Icon",el.getProperty("ICON"));
            XMLcmd.addScaPar("Hint",el.getProperty("EXPLANATION"));
            if (!vsn) {
                XMLcmd.addScaPar("State","Leaf");
            }
            else {
               if (el.getProperty("AUTOEXPAND").equals("Y")) XMLcmd.addScaPar("State","AutoExpand");
               else XMLcmd.addScaPar("State","Collapsed");
            }
            if (is_null(getVal("info4chageNode"))) {
                XMLcmd.addScaPar("Selected","Y");
                LuiCollection i4cn=new LuiCollection("");
                i4cn.Add_Element("Node",XMLcmd.params.get("Node").get_value());
                //i4cn.Add_Element("State",XMLcmd.params.get("State").get_value());
                i4cn.Add_Element("Type","ROW");
                i4cn.Add_Element("Context",XMLcmd.params.get("Context").get_value());
                i4cn.Add_Element("ContextID",this.getFormID());
                i4cn.Add_Element("ChildsContextID",this.getFormID());
                i4cn.Add_Element("FormID",this.getFormID());
                setVal("info4chageNode",i4cn.toString());
            }
            } // если PROMPT не пуст
            } // конец кручения строк
           } // если строки извлечены
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }
       } else { // в DISPLAY НЕ лежит SELECT
           if (AV.equals("Y") || vsn){       
            XMLcmd=XMLcg.addCommand("Node");
            XMLcmd.addScaPar("Node",el.getCode()+pMute);
            XMLcmd.addScaPar("Context",concol.toString());
            XMLcmd.addScaPar("Label",el.getProperty("PROMPT"));
            XMLcmd.addScaPar("Icon",el.getProperty("ICON"));
            XMLcmd.addScaPar("Hint",el.getProperty("EXPLANATION"));
            if (!vsn) {
                XMLcmd.addScaPar("State","Leaf");
            }
            else {
               if (el.getProperty("AUTOEXPAND").equals("Y")) XMLcmd.addScaPar("State","AutoExpand");
               else XMLcmd.addScaPar("State","Collapsed");
            }
            if (is_null(getVal("info4chageNode"))) {
                XMLcmd.addScaPar("Selected","Y");
                LuiCollection i4cn=new LuiCollection("");
                i4cn.Add_Element("Node",XMLcmd.params.get("Node").get_value());
             //   i4cn.Add_Element("State",XMLcmd.params.get("State").get_value());
                i4cn.Add_Element("Type","ROW");
                i4cn.Add_Element("Context",XMLcmd.params.get("Context").get_value());
                i4cn.Add_Element("ContextID",this.getFormID());
                i4cn.Add_Element("ChildsContextID",this.getFormID());
                i4cn.Add_Element("FormID",this.getFormID());
                setVal("info4chageNode",i4cn.toString());
            }
      } 
       }
}
/**
 *  Проверка необходимости вызова/обновления/закрытия детальных форм
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void checkFrmDetails() throws LuiException, SQLException{
    if (getVal("Is_QBE_Mode").equals("Y"))return; // если QBE - уходим
    if (is_not_null(getVal("CurrentAction"))) return; // если идёт выполнение действия то уходим
    if (getVal("Status").equals("Closing"))return; // страховка если форма закрывается
    boolean isCalledDetail=false;
    String menuItemcode="";
    if (getType().equals("MENU")){
     //LuiAppElement mel=this.findAndMute(getProperty("CurrentMenuItem"));
     LuiAppElement mel=null;
     LuiCollection concol=new LuiCollection(getProperty("CurrentMenuItemContext"));
     if (elements.containsKey(concol.Get_Element("ACTION"))){
       mel=elements.get(concol.Get_Element("ACTION"));
       this.setDynamicMenuItemProperty(mel, concol);
     }
     if (mel==null) return;
     menuItemcode=mel.getCode();
    }
    if (!getType().equals("GRID"))elSort(0); // сортируем если не матрица в матрице до хрена элементов и они уже сортированы
    for (int i=0;i<elSortedList.size();i++) {
     LuiAppElement el=elSortedList.get(i);
     if (el.getType().equals("ACTION") && !like(el.getCode(),"SYSMENU$%") && !like(el.getCode(),"COLMENU$%")&& !like(el.getCode(),"CELLMENU$%")){   // если это действие
      if ((el.getParent().equals("Form") && !getType().equals("MENU")) ||
              (getType().equals("MENU")&& el.getCode().equals(menuItemcode))){ // если действие уровня формы или формой является меню и действие = текущий пункт
        String calledFormID=el.getProperty("CalledFormID");
        String runFormStatus="";
        if (is_not_null(calledFormID)){
          if (application.forms.containsKey(calledFormID))  // такая форма есть?
            runFormStatus=application.forms.get(calledFormID).getVal("Status");// работает ли эта форма, которая вызвана этим действием?
        }
        String lastParamsHash=NVL(el.getProperty("ParamsHash"),"");// читаем у действия последнюю сигнатуру выходных параметров
              if (el.isActionActive() && like(el.getProperty("ROLE"),"%AUTO%")){ // если действие активно и её надо как деталь
         
          String currentParamsHash=String.valueOf(el.getProperty("COMMAND").hashCode()); 
        // начинаем вычислять новую сигнатуру параметров и формы
        for(int pi=0;pi<elSortedList.size();pi++){ // циклим все элементы формы
            LuiAppElement par=elSortedList.get(pi);
            if (par.getParent().equals(el.getCode()) && par.getType().equals("PARAM")){// если это параметр действия el
                //String v=par.getProperty("VALUE");
                currentParamsHash=String.valueOf(Integer.valueOf(currentParamsHash)+(par.getCode()+":"+par.getProperty("VALUE")+";").hashCode());
            }
        }
        if (is_not_null(calledFormID) //-- Когда-то открывалась детальная форма по этому действию
                && is_null(runFormStatus)//но сейчас формы нет
                ) {
           lastParamsHash="-1"; //Сбросим старую сигнатуру параметров чтобы форма открылась снова.
        }
        if (!lastParamsHash.equals(currentParamsHash)){//Необходимо обновить или заново открыть детальную форму
         if (runFormStatus.equals("Run")) {//Работает форма, запущенная в предыдущий раз. Обновить!
           //проверим а должна ли будет эта форма закрываться?
           String detailAction=application.forms.get(calledFormID).checkHotKey("DETAIL","Form"); // если у формы кчть действие с ролью DETAIL то не будем такую форму закрывать
                                                                                                // это предотвратит моргание графиков
           if (is_null(detailAction)) { // да должна
           String closeAction=application.forms.get(calledFormID).checkHotKey("CLOSE","Form");//Вдруг есть действие при закрытии?
           if (is_not_null(closeAction)){//Если есть действие при закрытии
             el.setProperty("ParamsHash",currentParamsHash);// запоминаем сигнатуру
             application.forms.get(calledFormID).runAction(closeAction);
             //            Run_Action(CalledFormID, CloseAction); -- Выполним его.
             // 27/10/2019 место загадочного удаления группы команд
             //application.commandWriter.removeGroup(calledFormID);//-- Если действие вдруг таки закрыло форму - отменим отправку этой команды клиенту.
           }
         }
         }
            el.setProperty("ParamsHash",currentParamsHash);// запоминаем сигнатуру
            runAction(el.getCode());
            //Run_Action(MasterWorkID, A.Code); -- Выполним действие
            if (is_not_null(calledFormID)){
            if(!el.getProperty("CalledFormID").equals(calledFormID)){
              if (application.forms.containsKey(calledFormID)) {  
                   
                application.forms.get(calledFormID).close(null,null);
              }
            }
            }
            isCalledDetail=true;
        } else {
          if (runFormStatus.equals("Closing") ){//Возможно, что форма уже закрылась сама. Но мы же её снова будем открывать
          el.setProperty("ParamsHash",currentParamsHash);// запоминаем сигнатуру
            runAction(el.getCode());
            isCalledDetail=true;
          // Run_Action(MasterWorkID, A.Code); -- Выполним действие
          if (el.getProperty("CalledFormID").equals(calledFormID)){
              application.commandWriter.removeGroup(calledFormID);//-- Удалим команду клиенту на закрытие.
          }
          }  
        }
      } else { // нет это не деталь (т.е. либо не активно либо не AUTO
                  
          if (runFormStatus.equals("Run") 
                  //&&   !getType().equals("CHART") //13.03.2020 коронавирус real-time dashboard // не из chart'а !!!
                  ){//Есть работающая форма, вызванная этим действием ранее 
              if (!application.forms.get(calledFormID).getProperty("NeverClosing").equals("Y")){ 
              //-- Раскрытые ветки дерева - не детали! Нужно тут учитывать?
              String closeAction=application.forms.get(calledFormID).checkHotKey("CLOSE","Form");//Вдруг есть действие при закрытии?
              if (is_not_null(closeAction)){//Если есть действие при закрытии
                  application.forms.get(calledFormID).runAction(closeAction);
                  //          Run_Action(CalledFormID, CloseAction); -- Выполним его.
              } else {
               
                application.forms.get(calledFormID).close(null,null);  
              }
          } else isCalledDetail=true;
          }
      }// конец не деталь
      } // конец уровня формы 
     } 
    }
    if (isCalledDetail    
            //&& application.windows.get(getVal("WindowID")).isActivable() не очень понятно зачем тут проверка на ктивируемость окна тыком мыши
            // поэтому исправил пока с учётом того что это прверяем только для форм с графикой
            //&& (application.windows.get(getVal("WindowID")).isActivable() || this.DgV2==null) //29 07 2021
            && application.windows.get(getVal("WindowID")).getProperty("AbleStatus").equals("Enable") // 29 07 2021 ваще сделал так, что только если активно окно
            ){
       
        application.commandWriter.removeCommadFromAllGroups("Focus");
        LuiClCommandGroup XMLcg=application.commandWriter.addGroup("");
        LuiClCommand XMLcmd=XMLcg.addCommand("Focus");
        XMLcmd.addScaParameter("FormID",NVL(getVal("RootFormID"),getFormID()), null);
    }
}
/**
 * Выполнение действия в форме (1-я фаза)
 * @param pActionCode код действия
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void runAction(String pActionCode) throws LuiException, SQLException{
 if (!elements.containsKey(pActionCode)) return;
 String lastAction=getVal("CurrentAction");//Последнее выполнявшееся действие
 //если щас выполняется какое-то действие уже и
  //это не таймер в чарте
 if (is_not_null(lastAction) 
         //&& (!"CHART".equals(getType())|| !like(elements.get(pActionCode).getProperty("ROLE"),"%TIME%"))
         && this.RTdashboard==null
         ) return;
 application.backList.clear(); // очистка бэк листа
     //    Set_LOV_ReturnPoint('Run_Action', FormID, UI_P_Collection.Add_Element(null, 'ActionCode', ActionCode));
     lastAction=pActionCode;
 LuiAppElement el=elements.get(lastAction);
   if (!el.getProperty("LastRunSCN").equals(getCurrentSCN())){
   el.makeRunSCN("LastRunSCN",String.valueOf(getCurrentSCN())); 
  //el.setProperty("LastRunSCN",String.valueOf(getCurrentSCN())); 
   if (NVL(el.getProperty("Run"),"Y").equals("Y")) {//-- А есть ли у пользователя право?
      if (!like(el.getProperty("ROLE"),"%TIME%")){
       el.properties.get("ACTIVE").clearComputedValue(false);
       if (!el.isActionActive()) return;//Последняя проверка активности действия
           }
        application.setVal("LastButtonPressed", lastAction);//Установим код текущего действия
        if (getType().equals("BLANK")) { // если бланк
        application.setVal("FieldsValues",getFieldsValues().toString()); //Заполним коллекцию значений полей (бланка) 
        application.setVal("JSONFieldsValues",getFieldsValues().ToJSON());
        if (el.getProperty("COMMAND").indexOf("(LUI$LOG:N;")<0){
        application.log.addRecord("ACTION", 
                application.NLS.NLS("RUS:Выполнение действия;ENG:Run action;")+" \""+
                lastAction+"-"+el.getProperty("PROMPT")+"\"", application.NLS.NLS("RUS:Значения полей;ENG:Field values;")+"=\n"+
                getFieldsValues().toMultiLineText(),this.getFormID());}
        if(!like(el.getProperty("ROLE"),"%CLOSE%")){//Если форма не закрывается -- ACTION_HOTKEY
         if (el.getProperty("VERIFICATION").equals("Y")){//Если перед действием должна проводится проверка полей -- ACTION_IS_CHECK
         checkFields();   
         }
        }
       el.setProperty("Item",getVal("CurrentField"));//зачемто запихиваем в св-во действия Item cв-во формы CurrentField        
     } 
     else if (getType().equals("LIST")){// если список
         setVal("Mult","N");
         el.setProperty("Item",getVal("CurrentColumn"));//-- столбец на начало действия
         //el.setProperty("UK", getUK().toString());//Коллекция ключевых столбцов на начало действия
         el.setProperty("Row", getVal("CurrentRow"));//Номер строки на начало действия
     }
     else if (getType().equals("TREE")){// если дерево
         //el.setProperty("UK", getUK().toString());//Коллекция ключевых столбцов на начало действия
         el.setProperty("Row", getVal("CurrentRow"));//Номер строки на начало действия
     }
     savepoint(); //-- Если будем после действия закрывать транзакцию - нужно установить Save Point
     setVal("CurrentAction",lastAction);
     runActionLockAndAsk();
    
    //}
   }
  }
 
}   
/**
 * Выполнение действия в форме (2-я фаза)
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void runActionLockAndAsk() throws LuiException, SQLException{
 String lastAction=getVal("CurrentAction");
 if (is_not_null(lastAction)){
  LuiAppElement el=elements.get(lastAction);
  //if (el.getProperty("ACTIVE").equals("EXIST")){//-- Если действие для строки, -- ACTION_IS_ACTIVE
  if (el.properties.get("ACTIVE").currentValue.equals("EXIST")){//-- Если действие для строки, -- ACTION_IS_ACTIVE
   if (el.getProperty("APPLY_TO").equals("N") || //Если действие только одной строки
           (getProperty("SelectionMode").equals("Single")&& like(el.getProperty("APPLY_TO"),"PROGRESS%"))){//или для всех выделенных, но выделена только одна
    String lock=el.getProperty("LOCK");
    if (!lock.equals("N")){  //Если  требуется блокировка -- ACTION_LOCK
      setVal("IsNeedLock","Y");//установим признак необходимости наложения блокировки
       if (is_not_null(el.getNonCompValue("PROC_INITIALLY")) || is_null(el.getNonCompValue("COMMAND"))){//-- ACTION_SQL
         locking(true,lock);  
       } else {
         locking(false,lock);
         if (getVal("IsNeedLock").equals("Y")){//не заблокировалось
//              GL_P_Param.Set_Value('COMMIT', 'Insert into SY_T_DUMMY_COMMIT values (null)'); -- запрет commit             
            application.setProperty("COMMIT","'Insert into SY_T_DUMMY_COMMIT values (null)')");
            
         }
       }
    }
   }
  }
  //-- Вычисляем текст вопроса и задаём его, если есть что спросить
  if (NVL(application.ask(el.getProperty("QUESTION"),null,null),"Y").equals("Y")){
     runActionExecutePre(); 
  } else {
      addProperty("ActionBreak",el.Code);
      application.actionBreak();
  }
      
 }
}
/**
 * Выполнение действия часть 3-я
 * выполнение программы (перед вызовом вызываемой формы)
 * ui_p_service.Run_Action_Action_SQL
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void runActionExecutePre() throws LuiException, SQLException{
 String lastAction=getVal("CurrentAction");
 if (is_not_null(lastAction)){
  
  String tcm=getVal("CommitMode");
  if (is_not_null(tcm)){
      application.setProperty("COMMIT",tcm);
  }
  
  exeProc(lastAction,"PROC_INITIALLY");
  
  elements.get(lastAction).setProperty("LastRunSCN",String.valueOf(application.getCurrentSCN()));
      //UI_P_Virtual.Set_Element_Property(FormID, LastAction, 'LastRunSCN', GL_P_PARAM.Get_Current_Value('SCN'));
    runActionRunCommand();
 }   
}
/**
 * Выполнение действия часть 4-я
 * выполнение команды по вызову (например формы)
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void runActionRunCommand() throws LuiException, SQLException{
    
String lastAction=getVal("CurrentAction");
String actionType;
String actionCommand;
String safeActionCommand;
String params="";
LuiClCommandGroup XMLcg;
LuiClCommand XMLcmd;
LuiAppElement el;
 if (is_not_null(lastAction)){
    el=elements.get(lastAction);
    actionType=el.getProperty("COMMAND_TYPE");
    actionCommand=el.getProperty("COMMAND");
    XMLcg=application.commandWriter.addGroup("RunCommand");    
    if (like(actionCommand,"%(%)")){ //За командой может идти коллекция параметров в круглых скобках
      int i=actionCommand.indexOf("(");
      actionCommand=actionCommand.substring(0, i);
    }
    safeActionCommand=actionCommand;
    params=getParamString(lastAction,params);
    
    
    
    setVal("NewAfter","NotDefined");
    setVal("NewRole", null); //-- Старый способ
    /*
        switch (NVL(el.getProperty("CHILD_COMMIT"),"")){
        case "N":
        application.setProperty("COMMIT","Insert into SY_T_DUMMY_COMMIT values (null)");
        break;
        case "Y":
        application.setProperty("COMMIT","Commit");
    
    }*/
  if (is_not_null(actionType)){
    if (like(el.getProperty("ROLE"),"%WIZARD%") 
            && (!like(el.getProperty("ROLE"),"%AUTO%") // это добавлено 03 06 2021 для эксперимента с действиями двойного назначения
            || like(el.getProperty("ROLE"),"%CLOSE%")) // а это добавлено 11 06 2021 потомучто перестало закрываться окно входа после внешней авторизации
            ){ //-- Действие - шаг визарда. Нужно скрыть текущее окно
      XMLcmd=XMLcg.addCommand("HideWindow");
      XMLcmd.addScaParameter("WindowID", getVal("WindowID"),"");
    }
  }
 if (is_not_null(actionCommand)){ 
     
  switch (actionType){
      case "TASK":
      case "DG":

int point=actionCommand.indexOf(".");
String mdApp=application.code;
    if (point>1) {
      mdApp=actionCommand.substring(0, point);
      actionCommand=actionCommand.substring(point+1);
    }  
      LuiCollection paramCol=new LuiCollection(params);
      LuiCollection newParamCol=new LuiCollection("");
      String config=NVL(paramCol.Get_Element("CONFIG"),"DEFAULT"); // конфигурация
      String stmt="";
      
      // для поиска формы нам потребуется соединение с метаданными
      LuiDbConn lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
      try {
          stmt="select app_code from lui_t_form where app_code='"+mdApp+"' and code='"+actionCommand+"' and formtype_code='"+actionType+"'"; 
      String app_sign=lc.getSelectResult(stmt);
      if (is_null(app_sign)) {
          stmt="select app_code from lui_t_form where app_code='LUI' and code='"+actionCommand+"' and formtype_code='"+actionType+"' and formgroup_code='LUI_COMMON'";
      app_sign=lc.getSelectResult(stmt);    
      if (is_null(app_sign)) {
       application.msgWriter.write("LUI-00015", actionCommand+" ("+actionType+")");
      }
      }
      if (actionType.equals("TASK")){
      // крутим параметры с целью изменить их коды добавив $ cзади иначе метод app.load
      // может создать там такие элементы
      for (String parcode:paramCol.keySet()){
          stmt="select code from lui_t_element where app_code='"+app_sign+"' and form_code='"+actionCommand+"' and code='"+parcode+"'";
          String elcode=lc.getSelectResult(stmt);
          if (is_not_null(elcode)){
              newParamCol.Add_Element(parcode+"$",paramCol.Get_Element(parcode));
          } else {
              newParamCol.Add_Element(parcode,paramCol.Get_Element(parcode));
          }
      }
      }
      if (actionType.equals("DG")){
          String pDgId=getProperty("do_not_read_dg_from_meta");
          if (is_null(pDgId)){
          pDgId=lc.getSelectResult("select lui_f_dg_read_form(null,'"+actionCommand+"','"+app_sign+"',"+application.luiSession.id+")");
          if (is_null(pDgId)){
              application.msgWriter.write("LUI-00015",actionCommand);
          }
          if (application.getProperty("FORM_STAT").equals("Y")){
          lc.getSelectResult("select lui_r_register_form_event('"+application.code+"','"+actionCommand+"','RUN')");
          }
          this.prepareDataGrouping(pDgId);
      } else this.properties.remove("do_not_read_dg_from_meta");
          paramCol.Add_Element("P_DG_ID",pDgId);
      }
      
      if (actionType.equals("TASK")){
          application.metaConnPool.SetFree(lc);lc=null;
      newParamCol.Add_Element("P_TASKCODE",actionCommand);
      newParamCol.Add_Element("P_CONFIGCODE",config);
      newParamCol.Add_Element("CONFIG","DEFAULT");
      actionCommand=application.getProperty("FORM2RUN_SYNC_TASK");
      if (is_null(actionCommand)){
       application.msgWriter.write("LUI-00107");
      }    
      actionType="BLANK";  
      application.executeForm(actionCommand, newParamCol.toString(), getFormID(), actionType/*,"N"*/);
      }
      if (actionType.equals("DG")){
      actionType="BLANK";  
      actionCommand="LUI_L_DATA_GROUPING_SHOW";
      paramCol.Add_Element("P_DGROOT","Y");
      String whisOutList="N";
      if (lc!=null)whisOutList=lc.getSelectResult("select run_without_list from lui_T_data_grouping where id="+paramCol.Get_Element("P_DG_ID"));
      if (whisOutList.equals("Y")){
      actionType=lc.getSelectResult("select lui_f_dg_get_clause4chart('command_type',"+paramCol.Get_Element("P_DG_ID")+",1)");
      actionCommand=lc.getSelectResult("select lui_f_dg_get_clause4chart('chart_code"+application.getProperty("DG_VERSION")+"',"+paramCol.Get_Element("P_DG_ID")+",1)");
    //  application.metaConnPool.SetFree(lc);lc=null;
      paramCol.Add_Element("P_LEVEL","1");
      }
      // добавление параметров ранее взятых с собой при сохоранении группировки в метаданных
      String ss="select code, lui_f_get_value(app_code,form_code,'DEFAULT','VALUE',code) \"value\" from lui_t_element where  app_code='"+app_sign+"' and form_code='"+safeActionCommand+"' and etype='PARAM'";
      ArrayList<LuiCollection> pDgPList=lc.getSelectAllRowsValMap(ss);
      if (pDgPList!=null){
      for (int dgpi=0;dgpi<pDgPList.size();dgpi++){
          LuiCollection DgPCol=pDgPList.get(dgpi);
            
          paramCol.Add_Element(DgPCol.get("code"),DgPCol.get("value"));
      }
      }    
      // конец добавления параметров
      application.executeForm(actionCommand, paramCol.toString(), getFormID(), actionType/*,"N"*/);
      }
      } catch (SQLException ex){
          application.msgWriter.sqlError(lc, ex, el.properties.get("COMMAND"), stmt);
      }
      finally {
          if (lc!=null)
          application.metaConnPool.SetFree(lc);
      }
      break;
      case "HOST":
      application.HOST(actionCommand, params);
      break;
      case "BLANK":
      case "CHART":
      case "LIST":
      case "TREE":
      case "MENU":
      case "HTML":
      case "GRID":
      if (params.indexOf("LUI$LOG:N;")<0){
      application.log.addRecord("ACTION",
      application.NLS.NLS("RUS:Вызов формы;ENG:Call form;")+" "+actionType+" "+actionCommand,
      application.NLS.NLS("RUS:Выходные параметры;ENG:Output paremeters;")+":\n"+params.replace(";","\n"),
      this.getFormID());
      }
      if (like(actionCommand,"O2P_L_OBJECT%")){
          System.out.println("before execute:"+System.currentTimeMillis()); // perfomance

      }
      application.executeForm(actionCommand, params, getFormID(), actionType/*,"Y"*/);
      if (like(actionCommand,"O2P_L_OBJECT%")){
          System.out.println("after execute:"+System.currentTimeMillis()); // perfomance
          
      }
      break;
      case "LINK":
      case "URL":
      XMLcmd=XMLcg.addCommand("URL");
      XMLcmd.addScaParameter("Address",actionCommand,"");
      if (is_not_null(params)){
      LuiCollection parcol=new LuiCollection(params);
      if (parcol.containsKey("Target")){
      XMLcmd.addScaPar("Target",parcol.Get_Element("Target"));
      parcol.Erase_Element("Target");
      params=parcol.toString();
      }
      if (is_not_null(params)){
      XMLcmd.addScaPar("Params",params);    
      }
      }
      break;
      case "APPLICATION":
      XMLcmd=XMLcg.addCommand("Event");
      XMLcmd.addScaPar("EventCode","Open" );
      XMLcmd.addScaPar("Params","Code:"+actionCommand+";"+params);
       XMLcmd.addScaPar("MS","1");
       break;
/*      case "REPGROUP":
      {
      LuiCollection pc=new LuiCollection(params);
      pc.Add_Element("REP_GROUP",actionCommand);
      pc.Add_Element("APP_CODE",application.code);
      application.executeForm("LUI_L_REPORT",pc.toString(),getFormID(),"LIST");
      }
      break;
      case "REPORT":
      break;
      case "UPLOAD":
      XMLcmd=XMLcg.addCommand("UploadFile");
      XMLcmd.addScaParameter("FormID",getFormID(),"");
      XMLcmd.addScaParameter("Title",el.getProperty("EXPLANATION"),"");
      XMLcmd.addScaParameter("Path",actionCommand,"");
//    UI_P_XML.AddParam('Params', GL_P_FILE_STORAGE.CREATE_INTERFACE_ACCESS_TICKET(ActionComand));
      break;
      case "DOWNLOAD":
      XMLcmd=XMLcg.addCommand("DownloadFile");
      XMLcmd.addScaParameter("FormID",getFormID(),"");
      XMLcmd.addScaParameter("Path",actionCommand,"");
      //UI_P_XML.AddParam('Params', GL_P_FILE_STORAGE.CREATE_INTERFACE_ACCESS_TICKET(ActionComand));
      LuiCollection pc=new LuiCollection(params);
      XMLcmd.addScaParameter("FileName",pc.Get_Element("DEF_FILE_NAME"),"");
      break;*/
      default:
      if (is_not_null(actionType)){
        application.msgWriter.write("LUI-00030", actionType);
      }
  }
 }
  if (is_null(actionType) || application.getVal("FormID").equals(getFormID())){
      runActionExecutePost("");
     if (!getVal("Is_QBE_Mode").equals("Y")) {  
      if (!getVal("Status").equals("Closing")){
      if(application.eventReader.is_non_interval || !this.getFormID().equals(application.getVal("rtChart"))) {
          // только если не интервальное событие или не realTime - chart форма
      checkFrmDetails();
      }
      }
     }
  } else { //-- Открылась новая форма
      if (getType().equals("LIST")||getType().equals("TREE")){ //Открылась из списка или дерева
         if (like(elements.get(lastAction).getProperty("ROLE"),"%AUTO%")){//-- действием с ролью AUTO
            elements.get(lastAction).setVal("Context",this.getFieldsValues().toString()); // == сохраняем контекст
         } 
      }
  }
 }
 
}
/**
 * Обновление ячейки списка после действия
 * @param pElCode
 * @param pOldValue
 * @param pIsCurrentCell
 * @throws LuiException
 * @throws SQLException 
 */
private void sendCellInfo (LuiClCommandGroup XMLcg,String pElCode,String pOldValue,boolean pIsCurrentCell) throws LuiException, SQLException{
String NewValue;
String LocBut;

LuiClCommand XMLcmd;
LuiAppElement el;

XMLcmd=XMLcg.addCommand("Cell");
//XMLcmd.really=false;
XMLcmd.addScaParameter("Code", pElCode, "");
el=elements.get(pElCode);
NewValue=el.getProperty("Value");
//if (NVL(NewValue,"!").equals(NVL(pOldValue,"!"))){
if (!compareString(NewValue,pOldValue) || pOldValue==null){
  XMLcmd.addScaPar("Show",NewValue);
  XMLcmd.really=true;
  if (is_not_null(el.properties.get("TOTAL").getNonCompValue())){
      changeTotal=true;
  }
}
if (el.properties.get("VISUALIZATION").recalculate!=null) {
    XMLcmd.addScaPar("Style",el.getProperty("VISUALIZATION"));
}
if (el.properties.get("MAKEUP").recalculate!=null) {
    XMLcmd.addScaPar("MakeUp",LuiString.RTrim(el.getProperty("MAKEUP"),"\n"));
}
if (el.properties.get("GETUP").recalculate!=null) {
    XMLcmd.addScaPar("GetUp",el.getProperty("GETUP"));
}
if (el.properties.get("HINT").recalculate!=null) {
    XMLcmd.addScaPar("Hint",el.getProperty("HINT"));
}
if (pIsCurrentCell){
  if (el.isEditCell()) {
      XMLcmd.addScaPar("Edit","Y");
 
   } else {
      XMLcmd.addScaPar("Edit","N");
 
  } 
}
LocBut=el.getProperty("LocBut");
if (is_not_null(LocBut)){
    LuiCollection Buttons=new LuiCollection("");
    LuiCollection LB=new LuiCollection(LocBut);
    for (String butc:LB.keySet()){
      LuiAppElement be=elements.get(butc);
      /*Здесь наверно надо сделать не вычесленными все свойства кнопочек внутри ячейки*/
      for (String pc:be.properties.keySet()){
         be.properties.get(pc).clearCompVal();
      } /*как-то так*/
      String icon=be.getProperty("ICON");
      if (is_not_null(icon)){
          Buttons.Add_Element(icon, be.getProperty("EXPLANATION"));
      }
    }
  XMLcmd.addScaPar("Buttons",Buttons.toString());
  
}
}
/** 
 *  Выполнение действия часть 5-я
 *  выполнение последействий
 * @param pActionCode код действия
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void runActionExecutePost(String pActionCode) throws LuiException, SQLException{
//  application.changeSCN("After Action");
  String lastAction=NVL(pActionCode,getVal("CurrentAction"));
  if (!elements.containsKey(lastAction))return;
  LuiClCommandGroup XMLcg = null;
  LuiClCommand XMLcmd;
  String lastColCode;
  Integer lastRowNum;
  String currentColCode;
  Integer currentRowNum;
  LuiCollection CurrentContextStr = null;
  LuiCollection LastContextStr;
  //String OldValue;
  //LuiCollection OldValues;
  LuiAppElement el=elements.get(lastAction);
  boolean B=true;
  changeTotal=false;
  if (is_null(lastAction)) return;
  String tcm=getVal("CommitMode");
  if (is_not_null(tcm)){
      application.setProperty("COMMIT",tcm);
  }
  // восстановление контекста
  currentRowNum=Integer.valueOf(NVL(getVal("CurrentRow"),"0"));//-- Текущая строка (если не применимо - null) 
  lastRowNum=Integer.valueOf(NVL(el.getProperty("Row"),"0"));//-- Строка при начале действия (если не применимо - null)
  if (currentRowNum!=lastRowNum){ //-- требуется восстановление контекста
      LastContextStr=new LuiCollection(el.getVal("Context"));//-- Контекст при начале действия
      if (LastContextStr.size()>0) {//-- есть что восстанавливать
          CurrentContextStr=this.getFieldsValues(); //-- Тeкущий контекст. Похраним его до конца действия
          this.setElementsValues(LastContextStr);
      }
      setVal("CurrentRow",String.valueOf(lastRowNum));//-- Временно восстанавливаем номер той строки
  }
  lastColCode=el.getProperty("Item");//-- Столбец при начале действия
  currentColCode=getVal("CurrentColumn");//-- Текущий столбец
  if (!lastColCode.equals(currentColCode)){//-- требуется восстановление текущего столбца
      setVal("CurrentColumn",lastColCode);//-- Временно восстанавливаем столбец в той строке
  }
  String ARQ=el.getProperty("AFTER"); // Установленное значение AFTER (ACTION_REQUERY)
  String AHK=el.getProperty("ROLE"); // Установленное значение ROLE (ACTION_HOTKEY)

    String newAfter=getVal("NewAfter"); //Возврат из формы (при закрытии она устанавливает NewAfter и NewRole)
    if (!"NotDefined".equals(newAfter)){
    //if (!newAfter.equals("NotDefined!")){ // 
      if (like(AHK,"%WIZARD%") && (!like(AHK,"%AUTO%") // -- Форма вызывалась как шаг визарда (но без AUTO добавлено 03 06 2021 для действий двойного назначения
     || like(AHK,"%CLOSE%") // а это добавлено 11 06 2021 потомучто перестало закрываться окно входа после внешней авторизации
              )  
              ) {// -- Форма вызывалась как шаг визарда (но без AUTO добавлено 03 06 2021 для действий двойного назначения
        if (!like(NVL(getVal("NewRole"),"!"),"%WIZARD%")) { //Действие возврата не имеет роль WIZARD
           AHK=""; //Текущее действие уже не будет WIZARD - форма всё равно закрывается
           ARQ=NVL(newAfter,"EXIT"); //текущее действие закроет форму также, как закрылась дочерняя форма (EXIT или CANCEL)
           B=false;
        } 
      }
    }
    if (   ("CANCEL".equals(newAfter) //-- дочерняя форма закрылась с отказом
            && !(lastAction.equals("RUN")&&(code.equals("LUI_L_CONFIGS")||code.equals("LUI_L_FORMS"))) // и при этом это не запуск формы из редактора форм!!!!
            )
            || getVal("isNeedLock").equals("Y") //или дочерняя форма отображала заблокированную запись
            ) { 
       if (rollback()){ //Откатываем всё назад
         if (B){ //если удалось откатить и если это не взаимодействие двух визардных шагов
           ARQ="NONE"; // текущее действие дорабатывать уже не будет
         }
       } 
    } else { //дочерняя форма не отображала заблокированную запись и не завершилась отказом
     this.exeProc(lastAction,"PROC_FINALLY");
    }
  if (ARQ.equals("EXIT") || ARQ.equals("CANCEL")) {
    close(AHK,ARQ);
    setVal("CurrentAction","");
    return;
  }
  if (ARQ.equals("NONE") || ARQ.equals("CLOSE")) { // если действие предназначено для закрытия но закрывать не хочет (передумала)
     addProperty("ActionBreak",el.Code);  // это проверит программа закрытия окна
  } 
  if (ARQ.equals("CLOSE_WINDOW")){
    application.windows.get(getVal("WindowID")).close();
    return;
  }
  if (ARQ.equalsIgnoreCase("RERUN")){// замена содержимого формы т.е. новая пойдёт с темже FormID
      
      application.reRun(this.getFormID());
      
  }
  if (ARQ.equals("RELOAD")){
      XMLcg=application.commandWriter.addGroup("ReLOAD");
      XMLcg.addParameter("FormID", getFormID());
     if (getType().equals("GRID")){
        XMLcmd=XMLcg.addCommand("ClearGrid");
        XMLcmd.addScaPar("FormID",getFormID()); 
       for (String ek:elements.keySet()){
        if (elements.get(ek).getParent().equals("ROWS")||elements.get(ek).getParent().equals("COLS")){ 
        XMLcmd=XMLcg.addCommand("Kill");
        XMLcmd.addScaPar("FormID",getFormID()); 
        XMLcmd.addScaPar("Code",elements.get(ek).Code);  
        }
       }
     } else {
     for (String ek:elements.keySet()){
      XMLcmd=XMLcg.addCommand("Kill");
      XMLcmd.addScaPar("FormID",getFormID()); 
      XMLcmd.addScaPar("Code",elements.get(ek).Code);  
     } 
     }
     elements.clear();
     elSortedList.clear();
     el4order.clear();
     resetRecalc();
     setVal("Second","");
     load(getProperty("MetadataApp"),false);
     loadParams(getVal("TransitParam"));
     reloadForm();
     elSort(0);
     
  }
  if (ARQ.equals("REOPEN")){ // переоткрытие всего окна
    String win=getVal("WindowID"); // берём окно
    if (application.windows.containsKey(win)){ //если оно есть
     // находим корневую форму этого окна
     for (String frmCode:application.forms.keySet()){
       LuiAppForm wfrm=application.forms.get(frmCode);
       if (wfrm.getVal("WindowID").equals(win)){
         if (wfrm.getVal("MasterForm").equals("Y")){
             String parentFormID=wfrm.getVal("ParentFormID");
             if (is_not_null(parentFormID)){
                 if  (application.forms.containsKey(parentFormID)){
                     LuiAppForm pFrm=application.forms.get(parentFormID); // берём мастер-форму
                     String callAction=wfrm.getVal("CallAction");    
                     if (is_not_null(callAction)){
                         if (pFrm.elements.containsKey(callAction)){
                             // запоминаем статусы всех форм кроме форм окна win
                            /* LuiCollection safeStatus = new LuiCollection("");
                             for(String fk:application.forms.keySet()){
                                 if (application.forms.get(fk).getVal("WindowID").equals(win))continue;
                                 safeStatus.Add_Element(fk,application.forms.get(fk).getVal("Status"));
                             }*/
                             application.windows.get(win).Close(false);
                             // восстанавливаем статусы
                          /*   for (String fk:safeStatus.keySet()){
                                 if (!application.forms.get(fk).getVal("Status").equals(safeStatus.get(fk))){
                                 application.forms.get(fk).setVal("Status",safeStatus.get(fk));
                                 }
                             }*/
                             pFrm.runAction(callAction);
                             return;
                         }
                     }
                 }
             }
         }
       }
     }
    
  }
  }

  if (application.getVal("Close Detail Mode").equals("Y")){ //Если режим закрытия деталей - ничего в обречённой форме делать уже не надо
    setVal("CurrentAction","");
    return;  
  }
  endAction();
  if (ARQ.equals("NEXT_ACTION")){
       //цепочка
     if (!like(AHK,"%AUTO%") && !like(AHK,"%CLOSE%") && !like(AHK,"%TIME%") && !like(AHK,"%WIZARD%") && !like(AHK,"%MATCH%") && !like(AHK,"%POSTEDIT%") && !like(AHK,"%ALLEXPAND%")){
     if (is_not_null(el.getProperty("NEXT_ACTION"))){
           setVal("CurrentAction","");
           el.getForm().runAction(el.getProperty("NEXT_ACTION"));
           return;
       }  
     }
     // конец цепочка
  }
  switch (getType()){
   case "LIST":
     if (ARQ.equals("CURRENT")){
         if (application.getProperty("SelectionMode").equals("Multi")){
           if (!el.getNonCompValue("APPLY_TO").equals("N")){
               ARQ="ALL";
           }
         }
     }
     switch (ARQ){ //Что делать после Действия
      case "FIELD": //Обновить только текущее полe
          if (is_not_null(lastColCode) && lastRowNum>0){//-- Если действие было вне ячейки, то и перечитывать нечего
            LastContextStr=requeryRow(getUK());//-- Перечитываем контекст
            if (LastContextStr.size()>0){//- Если та строка в списке ещё существует...
              elements.get(lastColCode).setValue(NVL(LastContextStr.Get_Element(lastColCode),""), false, "N", false);
                XMLcg=application.commandWriter.addGroup("ReQuery-Cell");
                XMLcg.addParameter("FormID", getFormID());
                XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
                if (compareString(lastColCode,currentColCode)&& currentRowNum==lastRowNum)
                sendCellInfo(XMLcg,currentColCode, null, true);    
                else
                sendCellInfo(XMLcg,currentColCode, null, false);
            }
          }
        /*lastColCode=el.getProperty("Item");//Столбец при начале действия
        lastRowNum=Integer.valueOf(NVL(el.getProperty("Row"),"0"));//Строка при начале действия
        if (is_not_null(lastColCode) && lastRowNum>0){//-- Если действие было вне ячейки, то и перечитывать нечего
            CurrentContextStr=this.getFieldsValues();// -- Тeкущий контекст
            LastContextStr=this.requeryRow(new LuiCollection(el.getProperty("UK")));//-- Контекст при начале действия
            currentColCode=getVal("CurrentColumn");   //Текущий столбец
            currentRowNum=Integer.valueOf(getVal("CurrentRow"));//Текущая строка
            if (LastContextStr.size()>0){ //Если та строка в списке ещё существует...
              this.setElementsValues(LastContextStr);//-- Временно восстанавливаем контекст той строки
              setVal("CurrentRow",String.valueOf(lastRowNum));//-- Временно восстанавливаем номер строки той строки
              setVal("CurrentColumn",lastColCode);//-- Временно восстанавливаем столбец в той строке
              XMLcg=application.commandWriter.addGroup("ReQuery-Cell");
              OldValue=elements.get(lastColCode).getProperty("Value");
                XMLcg.addParameter("FormID", getFormID());
                XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
                sendCellInfo(XMLcg,lastColCode, OldValue, true);
                setElementsValues(CurrentContextStr);//-- Восстанавливаем текущий контекст
                setVal("CurrentRow",String.valueOf(currentRowNum));//-- Восстанавливаем текущий номер строки
                setVal("CurrentColumn",currentColCode); //-- и текущий столбей
              
            }
        }*/
        /*  
        lastColCode=el.getProperty("Item");//Столбец при начале действия
        lastRowNum=Integer.valueOf(el.getProperty("Row"));//Строка при начале действия
        if (is_not_null(lastColCode) && lastRowNum>0){//-- Если действие было вне ячейки, то и перечитывать нечего
            currentColCode=getVal("CurrentColumn");   //Текущий столбец
            currentRowNum=Integer.valueOf(getVal("CurrentRow"));//Текущая строка
            if (currentColCode.equals(lastColCode) && currentRowNum.intValue()==lastRowNum.intValue()){//После действия ячейка осталась та же
                CurrentContextStr=requeryRow(null);
                if (CurrentContextStr.size()>0){//Если текущая строка в списке ещё существует...
                   OldValue=elements.get(currentColCode).getProperty("Value");
                   elements.get(currentColCode).setValue(CurrentContextStr.Get_Element(currentColCode),false,"N",false);
                XMLcg=application.commandWriter.addGroup("ReQuery-Cell");
                XMLcg.addParameter("FormID", getFormID());
                XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
                   sendCellInfo(XMLcg,currentColCode, OldValue, true);
                }
            } else { //Ячейка уже не та, что была при начале действия
               LastContextStr=requeryRow(new LuiCollection(el.getProperty("UK"))); //Тeкущий контекст строки, которая была при начале действия
               if (LastContextStr.size()>0){ //Если та строка в списке ещё существует...
                 CurrentContextStr=getFieldsValues();//Тeкущий контекст
                 if (currentRowNum.intValue()==lastRowNum.intValue()) {//Поменялась только ячейка
                  elements.get(lastColCode).setValue(LastContextStr.Get_Element(lastColCode),false,null,false);
                 } else {//Поменялась строка
                     setElementsValues(LastContextStr);//Восстанавливаем контекст той строки
                 }
                XMLcg=application.commandWriter.addGroup("ReQuery-Cell");
                XMLcg.addParameter("FormID", getFormID());
                XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
                 sendCellInfo(XMLcg,currentColCode, null, false);
                 if (CurrentContextStr.size()>0){
                     setElementsValues(CurrentContextStr);//Восстанавливаем текущий контекст
                 } else {
                     clearElementsValue();
                 }
               }
            }
        }*/
        break;
      case "CURRENT": //Обновить текущую строку
          if (lastRowNum>0){ //-- Если действие было вне строки, то и перечитывать нечего
            LastContextStr=requeryRow(getUK());//-- Перечитываем контекст
            if (LastContextStr.size()>0){//- Если та строка в списке ещё существует...
              this.setElementsValues(LastContextStr);//-- Временно восстанавливаем контекст той строки                
              XMLcg=application.commandWriter.addGroup("ReQuery-Current");
              XMLcg.addParameter("FormID", getFormID());
              XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
              for (String ccode:elements.keySet()){
                   LuiAppElement cell=elements.get(ccode);
                  if (cell.getType().equals("ITEM")){
                   if (compareString(lastColCode,currentColCode)&& currentRowNum==lastRowNum)   
                   sendCellInfo(XMLcg,ccode, null, true);   
                   else
                   sendCellInfo(XMLcg,ccode, null, false);       
                  } 
                }
            } else { //-- Той строки уже нет в этом списке
                setVal("PrefferedRow",String.valueOf(currentRowNum));
                setVal("UK",null);
                requery();
                CurrentContextStr=new LuiCollection("");
                lastRowNum=currentRowNum;
            }
          }
          /*
          lastRowNum=Integer.valueOf(NVL(el.getProperty("Row"),"0"));//Строка при начале действия
          if (lastRowNum>0) { //-- Если действие было вне строки, то и перечитывать нечего
            CurrentContextStr=this.getFieldsValues();// -- Тeкущий контекст
            LastContextStr=this.requeryRow(new LuiCollection(el.getProperty("UK")));//-- Контекст при начале действия
            currentRowNum=Integer.valueOf(getVal("CurrentRow"));//Текущая строка
            if (LastContextStr.size()>0){ //Если та строка в списке ещё существует...
              this.setElementsValues(LastContextStr);//-- Временно восстанавливаем контекст той строки                
              setVal("CurrentRow",String.valueOf(lastRowNum));//-- Временно восстанавливаем номер строки той строки
              XMLcg=application.commandWriter.addGroup("ReQuery-Current");
              XMLcg.addParameter("FormID", getFormID());
              XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
                for (String ccode:elements.keySet()){
                   LuiAppElement cell=elements.get(ccode);
                  if (cell.getType().equals("ITEM")){
                   sendCellInfo(XMLcg,ccode, null, false);   
                  } 
                }
              setElementsValues(CurrentContextStr);//-- Восстанавливаем текущий контекст
             setVal("CurrentRow",String.valueOf(currentRowNum));//-- Восстанавливаем текущий номер строки  
            } else { //-- Той строки уже нет в этом списке
                setVal("PrefferedRow",String.valueOf(currentRowNum));
                setVal("UK",null);
                requery();
            }
          } */
        /*if (is_null(el.getProperty("Row"))){
               lastRowNum=0;
           } else
        lastRowNum=Integer.valueOf(el.getProperty("Row"));//Строка при начале действия
        if (lastRowNum>0) {
           if (is_null(getVal("CurrentRow"))){
               currentRowNum=0;
           }
           else {
           currentRowNum =Integer.valueOf(getVal("CurrentRow"));//Текущая строка
           }
            if (currentRowNum.intValue()==lastRowNum.intValue()) { // После действия СТРОКА осталась та же
              CurrentContextStr=requeryRow(null);
             if (CurrentContextStr.size()>0) {//Если текущая строка в списке ещё существует...
                currentColCode=getVal("CurrentColumn");   //Текущий столбец   
                OldValues=getFieldsValues();
                setElementsValues(CurrentContextStr);
                XMLcg=application.commandWriter.addGroup("ReQuery-Current");
                XMLcg.addParameter("FormID", getFormID());
                XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
                for (String ccode:elements.keySet()){
                  LuiAppElement cell=elements.get(ccode);
                  if (cell.getType().equals("ITEM")){
                   sendCellInfo(XMLcg,ccode, OldValues.Get_Element(ccode), ccode.equals(currentColCode));   
                  }
                }
             } else {//Если текущая строка в списке уже не существует...
                 setVal("PrefferedRow",getVal("CurrentRow"));
                 setVal("UK","");
                 requeryRow(null);
             }
            } else { //Строка уже не та, что была при начале действия
              LastContextStr=requeryRow(new LuiCollection(el.getProperty("UK"))); // -- Тeкущий контекст строки, которая была при начале действия
              if (LastContextStr.size()>0){ //-- Если та строка в списке ещё существует...
               CurrentContextStr=getFieldsValues();// Тeкущий контекст 
               setElementsValues(LastContextStr);
               XMLcg=application.commandWriter.addGroup("ReQuery-Current");
               XMLcg.addParameter("FormID", getFormID());
               XMLcg.addParameter("RowNum", String.valueOf(lastRowNum));
               for (String ccode:elements.keySet()){
                  LuiAppElement cell=elements.get(ccode);
                  if (cell.getType().equals("ITEM")){
                   sendCellInfo(XMLcg,ccode, null, false);   
                  }
                }
                setElementsValues(CurrentContextStr); //Восстанавливаем текущий контекст
              }
            }//КОНЕЦ Строка уже не та, что была при начале действия
        }*/
      break;    
      case "ALL": //Перечитать все строки
          setVal("PrefferedRow",getVal("CurrentRow"));
          requery();
      break;
      case "REFRESH"://Переинициализировать весь список
       reloadForm();
      break;
      case "QBE":
       QBE();
     }
   if (changeTotal){
       if (getVal("TotalStatus").equals("Executed")){
         XMLcg=application.commandWriter.addGroup("ReQuery-Total");  
         XMLcmd=XMLcg.addCommand("Event");
         XMLcmd.addScaParameter("EventCode","GetTotal",null);
         XMLcmd.addScaParameter("MS","1",null);
         XMLcmd.addScaParameter("Params","FormID:"+getFormID()+";",null);
       }
   }  
   break;
   case "BLANK": //Если действие было из бланка
     switch (ARQ){ //Что делать после Действия
      case "ALL": // Если все поля
       for (String elfc:elements.keySet()){
         if (elements.get(elfc).getType().equals("FIELD")){
          LuiAppElement elf=elements.get(elfc);
          
          elf.clearComputedValues();
          elf.setProperty("UserChanged","");
          //elf.newValue(elf.getProperty("DEFINITION"), null, null, "Y", false, false);  // правка рустама 24.03.2020
          elf.setValue(elf.getProperty("DEFINITION"), false, "N",false); //правка рустама 24.03.2020
          //elf.setProperty("Value",elf.getProperty("DEFINITION"));                       правка рустама 24.03.2020
         }  
       }
      case "CURRENT": //только в текущей группе
        String GT;
        String GS;
        String parentelcode;
        for (String elfc:elements.keySet()){
          if (elements.get(elfc).getType().equals("FIELD")){
          LuiAppElement elf=elements.get(elfc);
          parentelcode=elf.getParent();
          if (parentelcode.equals("Form")) GS="Visible";
          else {
          LuiAppElement parentel=elements.get(parentelcode);
          if (parentel.getType().equals("GROUP")){
              GS=parentel.getProperty("CurrentState");
              GT=parentel.getProperty("GROUP_TYPE");
              if (is_null(GS)){
              if (GT.equals("BUNCH")) GS="Collapsed"; else GS="Visible";
                  }
          } else GS="Visible";
          }
          if (GS.equals("Visible")){
              elf.clearComputedValues();
              elf.setProperty("UserChanged","");
              //elf.newValue(elf.getProperty("DEFINITION"), null, null, "Y", false, false); //правка рустама 24.03.2020
              elf.setValue(elf.getProperty("DEFINITION"), false, "N",false); //правка рустама 24.03.2020
              //elf.setProperty("Value",elf.getProperty("DEFINITION"));                     правка рустама 24.03.2020
          }
          }
        }
        for (String elfc:elements.keySet()){
          if (elements.get(elfc).getType().equals("FIELD")){
          LuiAppElement elf=elements.get(elfc);
          parentelcode=elf.getParent();
          if (parentelcode.equals("Form")) GS="Visible";
          else {
          LuiAppElement parentel=elements.get(parentelcode);
          if (parentel.getType().equals("GROUP")){
              GS=parentel.getProperty("CurrentState");
              GT=parentel.getProperty("GROUP_TYPE");
              if (is_null(GS)){
              if (GT.equals("BUNCH")) GS="Collapsed"; else GS="Visible";
                  }
          } else GS="Visible";
          }
          if (GS.equals("Visible")){
          elf.setProperty("OldValue",elf.getProperty("Value"));
          elf.properties.get("Value").setChangeFlag();
          }
          }
        }
      case "FIELD":
        String CF=el.getProperty("Item");
        if (elements.containsKey(CF)){
        LuiAppElement elf=elements.get(CF);
        elf.clearComputedValues();
        elf.setProperty("UserChanged","");
        //elf.properties.get("DEFINITION").clearComputedValue(false);
        elf.properties.get("DEFINITION").clearCompVal(); 
        //elf.properties.get("DEFINITION").recalculate=true;
        // elf.newValue(elf.getProperty("DEFINITION"), null, null, "Y", false, false); // правка рустама 24.03.2020
        elf.setValue(elf.getProperty("DEFINITION"), false, "N",false); //правка рустама 24.03.2020
        elf.setProperty("OldValue", elf.getProperty("Value"));
        elf.properties.get("Value").setChangeFlag();                     
        }
      break;
      case "REFRESH":
        reloadForm();
      case "NONE": //22 05 2021 для случая когда вызываемая форма чтщ-то изменила в нашей форме по действию c ролью AUTO и последействием EXIT
        if (this.changeFlag) {
            this.changes();
        }
   }    
   break;
   case "HTML":
      switch (ARQ){
       case "ALL":
       this.properties.get("HTML_GENERATOR").clearCompVal();
       case "REFRESH":   
       setVal("Timer","");
       closeDetails("Y");          
       break;
      }  
   
   break;    
   case "CHART":
       switch (ARQ){
        case "ALL":
        chart(XMLcg);
        break;
        case "REFRESH":
        setVal("Timer","");
        closeDetails("Y");  
        if (XMLcg==null)XMLcg=application.commandWriter.addGroup("Refresh chart");
        chartClear(XMLcg);
        reloadForm();
        break;    
       }
   break;
   case "MENU":
    switch (ARQ){
        case "ALL":
        case "REFRESH":
        setProperty("CurrentMenuItem","");
        setProperty("CurrentMenuItemContext","");
        closeDetails("Y");
        clearChecking();
        reloadForm();
        break;
        case "CURRENT":
        //LuiAppElement mel=this.findAndMute(getProperty("CurrentMenuItem"));
            LuiAppElement mel=null;
            LuiCollection concol=new LuiCollection(getProperty("CurrentMenuItemContext"));
            if (elements.containsKey(concol.Get_Element("ACTION"))){
                mel=elements.get(concol.Get_Element("ACTION"));
            }
        if (mel!=null){
        this.showMenuItems(mel.getParent(),getProperty("CurrentMenuItemContext"));
        }
        break;
    }
    break;    
   case "TREE":
    switch (ARQ){
        case "REFRESH":
         closeDetails("Y"); 
         if (getVal("RootFormID").equals(this.getFormID())) {
             reloadForm();
         } else {
                    //Выполняем действие "при загрузке"
                     if (properties.containsKey("ON_LOAD")){
                       LuiAppProp prop=properties.get("ON_LOAD");
                       String prg=prop.getComputedProperty();
                       if (is_not_null(prg)) {
                    application.log.addRecord("ON_LOAD", application.NLS.NLS("RUS:Выполнение процедуры при загрузке формы ;ENG:Programm on load ;"), prg,this.getFormID());       
                           application.parser.execute(prg, prop);

                       }
                     }
                    //сброс вычисленных значений входных параметров (после действия REFRESH) при этом если в параметре записана динамика,
                     // то должен восстановится признак необходимости вычислений
                     for (String elKey:elements.keySet()){
                         LuiAppElement el2=elements.get(elKey);
                         if (el2.getType().equals("PARAM") && el2.getProperty("Parent").equals("Form")){
                             el2.properties.get("Value").clearComputedValue(true);
                         }
                     }
                     //вычисляем значения параметров и снимаем признак необходимости вычислений, т.е. делаем параметры статическими
                     for (String elKey:elements.keySet()){
                         LuiAppElement el2=elements.get(elKey);
                         if (el2.getType().equals("PARAM") && el2.getProperty("Parent").equals("Form")){
                             el2.properties.get("Value").getComputedProperty();
                         }
                     }
                             setVal("CurrentRow",null);
                        if (XMLcg==null)XMLcg=application.commandWriter.addGroup("Tree List refresh");
                        XMLcmd=XMLcg.addCommand("Node");
                        XMLcmd.addScaPar("FormID", getVal("RootFormID"));
                        XMLcmd.addScaPar("Node", getVal("ParentNode"));
                        XMLcmd.addScaPar("Label",Jsoup.parse(getProperty("HEADER")).text());
                        requery();
                }
         break;
        case "ALL":
        setVal("CurrentRow",null);
        closeDetails("Y");
        //clearChecking();
        requery();
        break;
        case "CURRENT":
        if (lastRowNum>0) {// -- Если действие было вне строки, то и перечитывать нечего
             LastContextStr=requeryRow(getUK()); //-- Перечитываем контекст
             if (LastContextStr.size()>0){ //-- Если та нода в списке ещё существует...
                this.setElementsValues(LastContextStr);
                XMLcg=application.commandWriter.addGroup("Refresh Current Tree Node");
              XMLcmd=XMLcg.addCommand("Node"); //-- Команда обновления ноды
              XMLcmd.really=false;  
              XMLcmd.addScaPar("FormID",getVal("RootFormID"));
              XMLcmd.addScaPar("Node",getFormID()+"#"+lastRowNum); 
              addCommandParam(XMLcmd,null,"NODE_TEXT","Label",true);
              addCommandParam(XMLcmd,null,"NODE_ICON","Icon",true);
              addCommandParam(XMLcmd,null,"NODE_VISUALIZATION","Style",true);
              addCommandParam(XMLcmd,null,"ADD_INFO","Hint",true);
              if (XMLcmd.really){ //-- Если в условный вывод попала ходь одна команда
              XMLcmd.addScaPar("Context",LastContextStr.toString());//-- отправим контекст
              } else { //-- иначе аннудируем всю команду (за ненадобностью)
                  application.commandWriter.removeGroup("Refresh Current Tree Node");
              }

             } else { //-- Той ноды уже нет в этом списке
              setVal("PrefferedRow",String.valueOf(currentRowNum));//-- Предпочтительный номер строки
              setVal("UK",null);
              requery();//-- Перезапрос
                CurrentContextStr=new LuiCollection("");//-- Не нужно восстанавливать контекст в конце действия
                lastRowNum=currentRowNum;//-- для избежания восстановления в конце действия
             }
        }
        /*
        lastRowNum=Integer.valueOf(NVL(el.getProperty("Row"),"0")); //Строка при начале действия
        if (lastRowNum>0) { //-- Если действие было вне строки, то и перечитывать нечего
          CurrentContextStr=this.getFieldsValues(); //-- Тeкущий контекст
          LuiCollection lstUK=new LuiCollection(el.getProperty("UK"));
          LastContextStr=requeryRow(lstUK); //-- Контекст при начале действия
          currentRowNum=Integer.valueOf(getVal("CurrentRow"));//-- Текущая строка
          if (LastContextStr.size()>0){//-- Если та нода в списке ещё существует...
           this.setElementsValues(LastContextStr);//-- Временно восстанавливаем контекст той ноды
           this.setVal("CurrentRow", String.valueOf(lastRowNum));//Временно восстанавливаем номер строки той ноды
              XMLcg=application.commandWriter.addGroup("Refresh Current Tree Node");
              XMLcmd=XMLcg.addCommand("Node"); //-- Команда обновления ноды
              XMLcmd.really=false;  
              XMLcmd.addScaPar("FormID",getVal("RootFormID"));
              XMLcmd.addScaPar("Node",getFormID()+"#"+currentRowNum); 
              addCommandParam(XMLcmd,null,"NODE_TEXT","Label",true);
              addCommandParam(XMLcmd,null,"NODE_ICON","Icon",true);
              addCommandParam(XMLcmd,null,"NODE_VISUALIZATION","Style",true);
              addCommandParam(XMLcmd,null,"ADD_INFO","Hint",true);
              if (XMLcmd.really){ //-- Если в условный вывод попала ходь одна команда
              XMLcmd.addScaPar("Context",LastContextStr.toString());//-- отправим контекст
              } else { //-- иначе аннудируем всю команду (за ненадобностью)
                  application.commandWriter.removeGroup("Refresh Current Tree Node");
              }
              setElementsValues(CurrentContextStr);//-- Восстанавливаем текущий контекст
              setVal("CurrentRow",String.valueOf(currentRowNum));//-- Восстанавливаем текущий номер строки
          }
          else{ //-- Той ноды уже нет в этом списке
              setVal("PrefferedRow",String.valueOf(currentRowNum));
              setVal("UK",null);
              requery();
                          
          } //-- -- Той ноды уже нет в этом списке
        }*/
        /*lastRowNum=Integer.valueOf(el.getProperty("Row"));//Строка при начале действия
        String curUK=getUK().toString();
        String lstUK=el.getProperty("UK");
        if (lastRowNum>0) {
            currentRowNum=Integer.valueOf(getVal("CurrentRow"));
          if (curUK.equals(lstUK) && lastRowNum.intValue()==currentRowNum.intValue()){//После действия нода осталась та же
            CurrentContextStr=requeryRow(getUK());
            if (CurrentContextStr.size()>0){//Нода ещё есть в ветке дерева
              setElementsValues(CurrentContextStr);
              XMLcg=application.commandWriter.addGroup("Refresh Current Tree Node");
              XMLcmd=XMLcg.addCommand("Node");
              XMLcmd.addScaParameter("FormID",getVal("RootFormID") ,"");
              XMLcmd.addScaParameter("Node",getFormID()+"#"+currentRowNum,"");
              addCommandParam(XMLcmd,null,"NODE_TEXT","Label",true);
              addCommandParam(XMLcmd,null,"NODE_ICON","Icon",true);
              addCommandParam(XMLcmd,null,"NODE_VISUALIZATION","Style",true);
              addCommandParam(XMLcmd,null,"ADD_INFO","Hint",true);
              XMLcmd.addScaParameter("Context",CurrentContextStr.toString(),"");
            }else {//Текущей ноды уже нет в этом списке
                setVal("PrefferedRow",currentRowNum.toString());
                setVal("UK",null);
                requery();
            }
          } else { //Нода уже не та, что была при начале действия
              LastContextStr=requeryRow(new LuiCollection(el.getProperty("UK")));//Тeкущий контекст строки, которая была при начале действия
              if (LastContextStr.size()>0){ //Если та нода в списке ещё существует...
                CurrentContextStr=getFieldsValues();//Тeкущий контекст
                setElementsValues(LastContextStr);// Восстанавливаем контекст той ноды
                XMLcg=application.commandWriter.addGroup("Refresh Current Tree Node");
              XMLcmd=XMLcg.addCommand("Node");
              XMLcmd.addScaParameter("FormID",getVal("RootFormID") ,"");
              XMLcmd.addScaParameter("Node",getFormID()+"#"+lastRowNum,"");
              addCommandParam(XMLcmd,null,"NODE_TEXT","Label",true);
              addCommandParam(XMLcmd,null,"NODE_ICON","Icon",true);
              addCommandParam(XMLcmd,null,"NODE_VISUALIZATION","Style",true);
              addCommandParam(XMLcmd,null,"ADD_INFO","Hint",true);
              XMLcmd.addScaParameter("Context",LastContextStr.toString(),"");
              setElementsValues(CurrentContextStr); //Восстанавливаем текущий контекст
              }
          }
          } */
    }    
   break;
  }
  setVal("CurrentAction","");
  //-- Восстановление контекста
  if (lastRowNum!=currentRowNum){
      if (CurrentContextStr!=null){
      if (CurrentContextStr.size()>0){
          this.setElementsValues(CurrentContextStr);
      }
      }
      setVal("CurrentRow",String.valueOf(currentRowNum));
  }
  if (!compareString(lastColCode,currentColCode)){
      setVal("CurrentColumn",currentColCode);
  }
 
  ////19.03.2020    checkRunTimer();
    application.pingRunTimer();

  
}
/**
 * Добавление параметра к команде на основе элемента и свойств элемента
 * @param XMLcmd - коменда
 * @param pElCode - код элемента
 * @param pPropCode - код св-ва
 * @param pParamName - имя параметра команды
 * @param pMakeCmdReally - сделать команду реальной если параметр добавился
*/
private void addCommandParam(LuiClCommand XMLcmd,String pElCode,String pPropCode,String pParamName,boolean pMakeCmdReally) throws LuiException, SQLException{
LuiAppProp p=null;
    if (is_not_null(pElCode)){
    if (elements.containsKey(pElCode)){
        p=elements.get(pElCode).properties.get(pPropCode);
    }
    }
    if (p==null){
       if (properties.containsKey(pPropCode)) 
       {p=properties.get(pPropCode);}
       else {
           return;
       }
    }
    if (p.changed()){
     XMLcmd.addScaPar(NVL(pParamName,pPropCode), p.getComputedProperty());
     if (pMakeCmdReally) XMLcmd.really=true;
    }
}

/** 
 * Выполнение запроса в списке
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void requery() throws LuiException, SQLException{
   if (is_null(getProperty("QUERY_TEXT"))) return;
    if (getVal("WaitingForFormActivate").equals("Y")) {
        return;// если форма ждёт активации уходим
    } 
   application.changeSCN("Requery");
   makeEl4order();
   LuiAppConn ac=null;
   LuiDbConn lc=null;
   if (getType().equals("BLANK")) return;
   
   for (String ec:elements.keySet()){ // расставляем или модернизируем автоловы
       if (!elements.get(ec).getType().equals("ITEM"))continue;
       if (elements.get(ec).getProperty("VARIANTS").equals("LOVQ:AUTO$")
               || like(elements.get(ec).getProperty("VARIANTS"),"%/*LOVQ:AUTO$*/%")){
          String lovcol=" /*LOVQ:AUTO$*/ distinct \""+ec+"\" \""+elements.get(ec).getProperty("CAPTION")+"\" ";
          String lovdef=elements.get(ec).getProperty("DEFINITION")+" \""+ec+"\" ";
          String qt=getProperty("QUERY_TEXT");
          try {
 Matcher M=Pattern.compile("select([\\n ]+\\*(\\n| |$)+)", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt);
 if (M.find()){
   lovcol=qt.substring(0, M.start(1))+lovcol+" from (select "+lovdef+qt.substring(M.end(1))+")lov$";  
 }
 //query=Pattern.compile("(select)[\n ]+\\*[\   n ]+", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt).replaceFirst("$1"+query);
} catch (Exception ex){
qt=qt.replace(" *\n", " * ");
lovcol=qt.replace(" * ", lovcol+" from (select "+lovdef+" ")+")lov$ ";
//query=qt.replaceFirst(" \\* ", query);
}
//lovcol="select "+lovcol+" from ("+qt+")lov$ ";
          elements.get(ec).addProperty("VARIANTS","LOVQ:"+ lovcol+" order by 1");
      }
       
   }
   LuiClCommandGroup XMLcg=application.commandWriter.addGroup("Requery");
   LuiClCommand XMLcmd;
   String rootFormID=NVL(getVal("RootFormID"),getFormID());
   if (getType().equals("LIST")){
       XMLcmd=XMLcg.addCommand("ClearGrid");
       XMLcmd.addScaParameter("FormID", getFormID(),"");
       XMLcmd=XMLcg.addCommand("Grid");
       XMLcmd.addScaParameter("FormID", getFormID(),"");
       switch (getProperty("START_MODE")){
           case "MULTI_ROW":
           case "INVERSION":
           case "SINGLE_ROW":
            XMLcmd.addScaPar("SelectionMode", getProperty("SelectionMode"));
            break;
           default:
            XMLcmd.addScaPar("SelectionMode", "None");
       }
       
   }
   if (getType().equals("TREE")){
    application.collapseNodeForms2(this.getFormID(),false);
       XMLcmd=XMLcg.addCommand("DeleteChilds");
       XMLcmd.addScaPar("FormID", rootFormID);
       XMLcmd.addScaPar("Node",getVal("ParentNode"));
   }
   
int mainCursor=0; 
//LuiAppElement rownoEl=null;
/** Костыль МВД сделаем столбец ROWNO последним */
/*if (elements.containsKey("ROWNO")){
  if (like(elements.get("ROWNO").properties.get("DEFINITION").startedValue,"%over(%)%")){
    rownoEl=elements.get("ROWNO");
    setVal("RowNoEl",rownoEl.Code);
    if (rownoEl.colinfo==null) {
                   rownoEl.colinfo=new ColInfo();
               }
                   rownoEl.colinfo.LuiDATATYPE=rownoEl.getProperty("DATA_TYPE");
                   rownoEl.colinfo.LuiFORMAT=rownoEl.getProperty("FORMAT");
                   rownoEl.colinfo.alias=rownoEl.Code;
               
    if (rownoEl.sourceSeq<9999999.0){
    rownoEl.sourceSeq=9999999.0;
    this.elSortedList.clear();
    this.elSort(0);
    }
}  
}*/
String mq=makeMainQuery();   
String  cc=getProperty("Connection");
if (!application.userDbConnections.containsKey(cc))return;
ac=application.userDbConnections.get(cc);
String queryHint=getProperty("QUERY_HINT");
try {
if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
   String queryText="Select ";
   if (is_null(lc.getOffsetColumn(OFF$SET))){
   
   if (is_not_null(queryHint)){
       queryText=queryText+" /*+"+queryHint+" */";
   }
   }
      
   setColsInfo(lc,mq,"QUERY_TEXT");
   queryText=queryText+" /* Main Query of "+getCode()+" */\n";
   /* вычисление сортировки перенесено выше вычисления фразы SELECT т.к. выражение в столбце может зависеть от сортировок*/   
   String orderby=addOrderBy(lc,el4order);
 // 16.05.2021 вычисление номера строки в МВД перенесено во обёрточный запрос 
 // if (queryHint.indexOf("/*sorting defines the select clause*/")>-1) {
 //  mq=makeMainQuery();// повтороное изготовление основной фразы SELECT c учётом сформированной ORDER BY если QUERY_HINT указывает на такую необходимость
 //  }
   String mainExp=makeMainExpression(lc,false,false,true);
   if (ac.getSessionMode()=='C'){//если соединение коммунальное
       // добавим столбец off$set если надо
     String offset=lc.getOffsetColumn(OFF$SET);
     if (is_not_null(offset))
     mainExp=offset+","+mainExp;
   }
   queryText=queryText+mainExp+"\n";
   //queryText=queryText+"From (Select * from (\n"; //13.05.2021 Вынос order by наружу
   queryText=queryText+"From (\n";//13.05.2021 Вынос order by наружу
   queryText=queryText+mq+")q$1";
   String qbewhere= addWhere(lc,getProperty("inner_qbe"),false);
   if (is_not_null(qbewhere)){
      if (getProperty("inner_qbe").equals("Y")){
         if (queryText.indexOf("/*ADD_LUI$_INNER_QBE*/")>0) queryText=queryText.replace("/*ADD_LUI$_INNER_QBE*/"," and ("+qbewhere.substring(7)+")");
         else {
             if (queryText.indexOf("/*LUI$_INNER_QBE*/")>0) queryText=queryText.replace("/*LUI$_INNER_QBE*/",qbewhere);
         }
      } 
      else queryText=queryText+qbewhere+"\n";
   }
//queryText=queryText+")q$2"+this.getWhere4DgCollapsedLevels(lc);//13.05.2021 Вынос order by наружу
   queryText=queryText+this.getWhere4DgCollapsedLevels(lc);//13.05.2021 Вынос order by наружу
   
//перенесено выше вычисления фразы SELECT т.к. выражение в столбце может зависеть от сортировок   String orderby=addOrderBy(lc,el4order);
   if (is_not_null(orderby)){
       queryText=queryText+orderby+"\n";
   }
   String where4winFunc=this.addWhere(lc, getProperty("inner_qbe"),true);
   if (is_not_null(where4winFunc) ){
       if (getProperty("inner_qbe").equals("Y")){
         if (queryText.indexOf("/*ADD_LUI$_INNER_QBE4WINFUNC*/")>0) queryText=queryText.replace("/*ADD_LUI$_INNER_QBE4WINFUNC*/"," and ("+where4winFunc.substring(7)+")");
         else {
             if (queryText.indexOf("/*LUI$_INNER_QBE4WINFUNC*/")>0) queryText=queryText.replace("/*LUI$_INNER_QBE4WINFUNC*/",where4winFunc);
         }
      }
       else{
   String OuterSelect="select o$1.*";
   queryText=OuterSelect+" from ("+queryText+") o$1 "+where4winFunc;// доп обёртка для выноса QBE над оконными 15.05.2021
           }
   }
   if (getVal("CheckQuality").equals("Y")) { // если это всего лишь проверка качества
       LuiAppProp chkProp=setVal("MainQuery",queryText); // устанавливаем свойство для текста запроса
       String ERROR=application.parser.plugins.get("ParseSQL").compute(queryText, chkProp);
       if (is_not_null(ERROR)){
       setVal("CheckQualityError",ERROR);    
       }
       if (ac.getSessionMode()=='C'){ // если коммунально
       //if (lc!=null) {
        ac.connPool.SetFree(lc); //закрываем
       //}
    }
       return;
   }
         if (ac.getSessionMode()!='C'){ //если соединение не коммунальное-закроем старый курсор
            if (is_not_null(getVal("MainCursor"))){// если есть курсор
            mainCursor=Integer.valueOf(getVal("MainCursor"));
            if (lc.CursorisOpen(mainCursor)){ // и он открыт
            lc.CursorClose(mainCursor); // закроем его
            setVal("MainCursor","");
                                            }
            }
 }
 if (is_not_null(queryText)){
     setVal("MainQuery",queryText);
     if (ac.getSessionMode()!='C'){ // открываем курсор для сессии с длинными транзакциями
                                    // для сесси с короткими транзакциями это происходит в fetchRows()
     
     application.log.addRecord("MAIN_QUERY",application.NLS.NLS("RUS:Выполнение основного запроса;ENG:Main query;")
             ,queryText,this.getFormID());
     
     try{
     mainCursor=lc.CursorOpen(queryText); // открываем курсор
     } catch (SQLException ex){
         application.msgWriter.sqlError(lc, ex, this.properties.get("QUERY_TEXT"), queryText);
         //application.msgWriter.write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+queryText+") ",properties.get("QUERY_TEXT").getGenesis4msg());
     }
     setVal("MainCursor",String.valueOf(mainCursor));
     }
 }
 if (is_null(getVal("PrefferedRow"))){
 setVal("PrefferedRow",getVal("CurrentRow"));
 }
 String safeCurrentRow=getVal("CurrentRow"); 
 setVal("CurrentRow","0");
 clearChecking();
  if (getProperty("START_MODE").equals("INVERSION")) {
      setVal("Inversion", "Y");
      setProperty("START_MODE","MULTI_ROW");
  }
  else setVal("Inversion", "N");
  setVal("IsLastPage", "N");
  setVal("FetchedRows", "0");
  setVal("RowRank", "0");
  setVal("LastValue4RowRank","");
  if (is_not_null(getVal("TotalStatus"))){
      setVal("TotalStatus","Enable");
   if (like(queryText.toUpperCase(),"%CALCULATE_TOTAL_IMMEDIATE%")){
      setVal("TotalStatus","Executed");
       XMLcmd=XMLcg.addCommand("Event");
       XMLcmd.addScaParameter("EventCode", "GetTotal", "");
       XMLcmd.addScaParameter("MS", "1", "");
       XMLcmd.addScaParameter("Params", "FormID:"+getFormID()+";", "");
   }
  }
//if (ac!=null){ // если есть прикладной коннект
    if (ac.getSessionMode()=='C'){ // но он коммунальный
    //  if (lc!=null){ // и есть коннект живой
       if (lc.CursorisOpen(mainCursor)){// закроем курсор если открыт
            lc.CursorClose(mainCursor);
        }
        ac.connPool.SetFree(lc); // и освободим коннект - fetchRows он опять будет захвачен
        lc=null;
    //  }  
    }
//}  
String safeQBEmode=getVal("Is_QBE_Mode"); // запоминаем QBE mode

setVal("Is_QBE_Mode","N"); // говорим что QBE уже нет
try{ // пытаемся извлеч строки
if (like(this.code,"O2P_L_OBJECT%")){
          System.out.println("before fetch:"+System.currentTimeMillis()); // perfomance

      }    
fetchRows();
if (like(this.code,"O2P_L_OBJECT%")){
          System.out.println("after fetch:"+System.currentTimeMillis()); // perfomance

      }    

} catch (Exception QBEex){ // если случилась ошибка
    //setVal("Is_QBE_Mode","N"); // восстанавливаем запомненный режим QBE
    setVal("Is_QBE_Mode",safeQBEmode); // восстанавливаем запомненный режим QBE
    setVal("CurrentRow",safeCurrentRow); // восстанавливаем запомненный режим QBE
    throw QBEex;
}
setVal("Is_QBE_Mode","N");
application.pingRunTimer();

XMLcg=application.commandWriter.addGroup("");
if (Integer.valueOf(NVL(getVal("CurrentRow"),"0"))>0){
  if (getType().equals("LIST")) {
    gotoCell(null,null,null,true);  
    XMLcmd=XMLcg.addCommand("CurrentRow");
    XMLcmd.addScaParameter("FormID",getFormID(),null);
    XMLcmd.addScaParameter("RowNum", getVal("CurrentRow"),null);
    String curColumn=getVal("CurrentColumn");
    if (is_not_null(curColumn)){
        if (elements.containsKey(curColumn)){
        XMLcmd.addScaParameter("Cell",curColumn,null);
        if (elements.get(curColumn).isEditCell()){
            XMLcmd=XMLcg.addCommand("Cell");
            XMLcmd.addScaParameter("FormID", getFormID(),null);
            XMLcmd.addScaParameter("Code", curColumn,null);
            XMLcmd.addScaParameter("Edit","Y",null);
        }
        }else {
                setVal("CurrentColumn","");
                }
    }
  }
  if (getType().equals("TREE")){
   if (!application.forms.get(rootFormID).getVal("InsideChangeNode").equals("Y")) {    
    application.forms.get(rootFormID).setVal("CurrentNodeHash","");
    application.forms.get(rootFormID).changeNode(getFormID()+"#"+getVal("CurrentRow"),null,"ROW",
            getFieldsValues().toString(),
            getFormID(),null);
    XMLcmd=XMLcg.addCommand("Node");
    XMLcmd.addScaParameter("FormID", rootFormID, null);
    XMLcmd.addScaParameter("Node",getFormID()+"#"+getVal("CurrentRow"), null);
    XMLcmd.addScaParameter("Selected","Y",null);
   }
  }
}else{
    String parentID=getVal("ParentFormID");
    if (is_not_null(parentID)){
        String parentAction=getVal("CallAction");
        if (is_not_null(parentAction)){
            if ("AUTO".equals(application.forms.get(parentID).elements.get(parentAction).getProperty("ROLE"))){
                if (is_not_null(getNonCompValue("ADD_INFO"))) {
            XMLcmd=XMLcg.addCommand("Underside");
            XMLcmd.addScaPar("FormID", getFormID());
            XMLcmd.addScaPar("Content","");   
                }
            }
        }
    }
 clearElementsValue();//Стираем контекст
}
} finally {
    if (ac!=null && ac.getSessionMode()=='C'){
      if (lc!=null) {
        if (lc.CursorisOpen(mainCursor)){
            lc.CursorClose(mainCursor);
        }
        ac.connPool.SetFree(lc);
      }
    }
}
if (is_null(getVal("CurrentAction"))){
    checkFrmDetails();
}
setVal("Status","Run");
}

/**
 * установка информации о типах данных CУБД для элементов диаграммы
 * @param pParent -- злемент хранящий запрос
 * @param plc -- соединение
 * @param pQuery -- полусочинённый запрос
 * @param pProp -- код св-ва в элементе DIMENSION где лежит текст запроса, это чтобы ругаться нормально
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setColsInfo4Chart(String pParent,LuiDbConn plc,String pQuery,LuiAppProp pProp) throws LuiException, SQLException{
    Integer c=null;
    try{  
    c=plc.CursorOpen(pQuery); // открываем курсор
    ArrayList<ColInfo> cols=plc.CursorGetColsInfo(c); // получаем количество столбцов
    LuiAppElement el;
    for (int i=0;i<cols.size();i++){ // крутим столбцы
      ColInfo ci=cols.get(i); // берём инфу столбца
      el=null; // делаем что пока элемент соответствующий не найден
      if (elements.containsKey(ci.alias.toUpperCase())){ // если есть элемент с таким кодом столбца
      el=elements.get(ci.alias.toUpperCase()); // берём его на досмотр
      if (pParent.equals("Form") || el.getParent().equals(pParent)) {
          if (!(el.getType().equals("ITEM"))){
// если это НЕ ITEM, то елемент не годен
           el=null;
          }   
      }   }
       
         if (el!=null){ // если элемент выявлен , то делаем его членом запроса
          ci.LuiDATATYPE=NVL(el.getProperty("DATA_TYPE"),"CHAR");
          ci.LuiFORMAT=NVL(el.getProperty("FORMAT"),"");
          el.colinfo=ci;
          el.addProperty("JDBCDataType",ci.columnTypeName);
        }
      }
    
   
  } catch (SQLException ex){
    
          application.msgWriter.sqlError(plc, ex,pProp, pQuery);
    
  } finally {
      if (c!=null){
          if (plc.CursorisOpen(c)){
              try {
                  plc.CursorClose(c);
              } catch (SQLException ex) {
                  
              }
          }
      }
  }
}
/**
 * Установка информации о столбце запроса в элементы формы
 * @param plc соединение с базой данныз
 * @param pQuery текст запроса
 * @param pPropCode код свойства
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setColsInfo(LuiDbConn plc,String pQuery,String pPropCode) throws LuiException, SQLException{
Integer c=null;
  try{  
    c=plc.CursorOpen(pQuery); // открываем курсор
    ArrayList<ColInfo> cols=plc.CursorGetColsInfo(c); // получаем количество столбцов
    LuiAppElement el;
    for (int i=0;i<cols.size();i++){ // крутим столбцы
      ColInfo ci=cols.get(i); // берём инфу столбца
      el=null; // делаем что пока элемент соответствующий не найден
      //11.05.2021
      // узнаем какой элемент есть в списке в апперкейсе или точное совпадение
      if (elements.containsKey(ci.alias)) el=elements.get(ci.alias);
      else {if (elements.containsKey(ci.alias.toUpperCase())){
          el=elements.get(ci.alias.toUpperCase());
      }
      }
      //11.05.2021
//11.05.2021      if (elements.containsKey(ci.alias.toUpperCase())){ // если есть элемент с таким кодом столбца
//11.05.2021      el=elements.get(ci.alias.toUpperCase()); // берём его на досмотр
      //    if (!(el.getType().equals("ITEM")||el.getType().equals("FIELD")||(el.getType().equals("PARAM")&& getType().equals("BLANK") && el.getParent().equals("Form")))){
// если это НЕ столбец или поле или входной параметр бланка, то елемент не годен
        //   el=null;
        //  }   
        // этот фрагмент с присваиванием null я закоментил, т.к. если имя поля запроса совпадёт например с кодом группы бланка - начнётся колбаса
//11.05.2021 }   
       if (el==null)   { // если подходячий элемент не найден
          if (getType().equals("BLANK")){ // и это бланк
          el=addElement(ci.alias.toUpperCase(), "PARAM", null, "Virtual parameter","Form");  // добавим такой входной параметр
          el.addProperty("Value","");
          }
      }
         if (el!=null){ // если элемент выявлен или создан, то делаем его членом запроса
          ci.LuiDATATYPE=NVL(el.getProperty("DATA_TYPE"),"CHAR");
          ci.LuiFORMAT=NVL(el.getProperty("FORMAT"),"");
          el.colinfo=ci;
          el.addProperty("JDBCDataType",ci.columnTypeName);
          if (application.dataConverter.LuiDataTypes.get(ci.LuiDATATYPE).emptyStringFromLuiMode.equals("NULL")) {
          if (is_not_null(application.dataConverter.LuiDataTypes.get(ci.LuiDATATYPE).emptyStringSpecialSymbol))    
          el.addProperty("EmptyStringSpecialSymbol",application.dataConverter.LuiDataTypes.get(ci.LuiDATATYPE).emptyStringSpecialSymbol);
          } else {
          if (is_not_null(application.dataConverter.LuiDataTypes.get(ci.LuiDATATYPE).nullSpecialSymbol))        
          el.addProperty("NullSpecialSymbol",application.dataConverter.LuiDataTypes.get(ci.LuiDATATYPE).nullSpecialSymbol);
          }
        }
      }
    
   
  } catch (SQLException ex){
    
          application.msgWriter.sqlError(plc, ex,this.properties.get(pPropCode), pQuery);
    
  } finally {
      if (c!=null){
          if (plc.CursorisOpen(c)){
              try {
                  plc.CursorClose(c);
              } catch (SQLException ex) {
                  
              }
          }
      }
  }
}
/**
 * обработка нажатия кнопки
 * @param pCode код кнопки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void buttonPressed(String pCode) throws LuiException, SQLException{
    application.changeSCN("ButtonPressed");
    runAction(pCode);
    application.checkDetails("");
}
    /**
     * Выделение узла дерева FormID по инициативе пользователя ((клик мышью на
     * тексте узла дерева или навигация клавишами вверх и вниз)) или чьеё-то ещё
     *
     * @param pNode - ID нового текущего узла
     * @param pState - текущее состояние узла
     *
     * @param pContext - строка-коллекция контекстных значений
     * @param pContextID - ID рабочей области, к которой относится контекст
     * @param pIsCheck - выделить/не выделить
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void checkNode(String pNode,String pState,String pContext,String pContextID,String pIsCheck) throws LuiException, SQLException{
 application.changeSCN("CheckNode");   
 
 setVal("CurrentNodeHash","");
 application.setVal("FormID", pContextID);
 if (application.forms.containsKey(pContextID)){
     LuiAppForm frm=application.forms.get(pContextID);
     if (getType().equals("MENU")){
     frm.setProperty("CurrentMenuItem","");
     LuiCollection mic=new LuiCollection("");
     mic.Add_Element(pNode,frm.elements.get(pNode).getProperty("ACTIVE"));
     frm.checkRow(pIsCheck,mic);
     frm.checkFrmDetails();
     return;
     }
     frm.setVal("CurrentRow","");
     frm.setContext4Node(new LuiCollection(pContext), "");
     frm.checkRow(pIsCheck, frm.getUK());
     frm.setContext4Node(new LuiCollection(""), pContextID);
     frm.setVal("CurrentRow","");
     frm.setProperty("SelectionMode", "Multi");
 }
}

    /**
     * Изменение текущего узла дерева FormID по инициативе пользователя ((клик
     * мышью на тексте узла дерева или навигация клавишами вверх и вниз)) или
     * чьеё-то ещё
     *
     * @param pNode - ID нового текущего узла
     * @param pState - текущее состояние узла
     * @param pNType - тип нового текущего узла
     * @param pContext - строка-коллекция контекстных значений
     * @param pContextID - ID рабочей области, к которой относится контекст
     * @param pChildsContextID - Id 2-ой рабочей области
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void changeNode(String pNode,String pState,String pNType,String pContext,String pContextID,String pChildsContextID) throws LuiException, SQLException{
  application.changeSCN("ChangeNode");
  String NewRowNum;
  String OldWorkID=getVal("ContextForkID");
  String OldNodeHash=getVal("CurrentNodeHash");
  String NewNodeHash=String.valueOf((pContext.toString()+" "+pNode).hashCode());
  String oldParentNode="*";
  String contextParentNode="+";
  LuiAppForm oldfrm=null;
  LuiAppForm ctxfrm=null;
  LuiCollection ContextColl=new LuiCollection(pContext);
  if (!OldWorkID.equals(pContextID)) setVal("InsideChangeNode","Y");
  application.setVal("FormID", pContextID);
  if (is_not_null(pContextID)){
            if (application.forms.containsKey(pContextID)){
               ctxfrm=application.forms.get(pContextID);
               //ctxfrm.setVal("CurrentNode", pNode);
               if (ctxfrm.getType().equals("MENU")){
                   ctxfrm.setProperty("CurrentMenuItem", pNode);
                   ctxfrm.setProperty("CurrentMenuItemContext", pContext);
                   ctxfrm.clearChecking();
                   LuiCollection mic=new LuiCollection("");
                   if (ctxfrm.elements.containsKey(pNode)){
                   mic.Add_Element(pNode,ctxfrm.elements.get(pNode).getProperty("ACTIVE"));
                   ctxfrm.checkRow("Y", mic);
                   }
                   //ctxfrm.checkFrmDetails();
                   //return;
               }
               
            }
            }
  
  if (!NVL(OldNodeHash,"!").equals(NVL(NewNodeHash,"!"))){
      setVal("CurrentNodeHash",NewNodeHash);
      if (is_not_null(OldWorkID)){
        if (application.forms.containsKey(OldWorkID)){
            oldfrm=application.forms.get(OldWorkID);
            String opn=oldfrm.getVal("ParentNode");
            String cpn="";
            if (ctxfrm!=null){
               cpn=ctxfrm.getVal("ParentNode");
            }
            
            if (!(NVL(opn,"*").equals(NVL(cpn,"+")))){
                        oldfrm.closeDetails("N");
            }
                    if (!(oldfrm.getType().equals("MENU"))){  
            oldfrm.setVal("CurrentRow","");
            oldfrm.clearChecking();
            oldfrm.setProperty("SelectionMode","Single");
                    }
        }
      }
      if (ctxfrm!=null){
          if (!(ctxfrm.getType().equals("MENU"))){
               ctxfrm.clearChecking();
               ctxfrm.setProperty("SelectionMode","Single");
          }
               ctxfrm.setVal("CurrentNodeType",pNType);
      }
      setVal("ContextForkID","");
      if (is_not_null(pNode)){
        NewRowNum=Extract_Word(pNode, 1);
        NewRowNum=Extract_Word(NewRowNum,-1,"#");
        NewRowNum=Filter(NewRowNum,"0123456789");
        if (ctxfrm!=null){
         ctxfrm.setContext4Node(ContextColl, pChildsContextID);
           if (is_not_null(NewRowNum)){
            ctxfrm.setVal("CurrentRow",NewRowNum);
            ctxfrm.checkRow("Y",ctxfrm.getUK());//-- Выделяем текущую строку
            ctxfrm.setVal("UK",ctxfrm.getUK().toString());//-- Вычиляем новый UK
        }
        if (compareString(pState,"Continue")){
            LuiClCommandGroup XMLcg=application.commandWriter.addGroup("");
            LuiClCommand XMLcmd=XMLcg.addCommand("DeleteNode");
            XMLcmd.addScaPar("FormID", getFormID());
            XMLcmd.addScaPar("Node", pNode);
            ctxfrm.fetchRows();
        } else {
            if (NVL(pNType,"").equals("ROW") && (NVL(pState,"Collapsed").equals("Collapsed") ||
                    NVL(pState,"Collapsed").equals("Leaf")||
                    NVL(pState,"Collapsed").equals("Expanded")
                    )){
                ctxfrm.checkFrmDetails();
                setVal("ContextForkID",pContextID);
            } else {
                ctxfrm.checkFrmDetails();
            }
        }
        }
      }
  }
  if (ctxfrm!=null){
     String rootFrmID=ctxfrm.getVal("RootFormID");
     if (application.forms.containsKey(rootFrmID)){
        String forms4close[]= application.forms.get(rootFrmID).getProperty("CloseWhenChangeNode").split(",");
        for (int i=0;i<forms4close.length;i++){
            if (application.forms.containsKey(forms4close[i])){
                application.forms.get(forms4close[i]).close(null,null);
            }
        }
        //if (!rootFrmID.equals(ctxfrm.getFormID()))application.forms.get(rootFrmID).closeDetails("N");
        /*if (application.forms.get(rootFrmID).getProperty("closeDetailsIfRootInTree").equals("Y")){
            application.forms.get(rootFrmID).closeDetails("N");
        } */
     }
  }
  setVal("InsideChangeNode","");
}

    /**
     *
     * @param pContextStr - контекст
     * @param pOpenFormID - id формы которая делается текущей
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void setContext4Node(LuiCollection pContextStr,String pOpenFormID) throws LuiException, SQLException{
    LuiCollection EmptyContext=new LuiCollection("");
    setElementsValues(pContextStr);
    if (is_not_null(pOpenFormID)){
     if (application.forms.containsKey(pOpenFormID)){
         LuiAppForm ofrm=application.forms.get(pOpenFormID);
         ofrm.setVal("CurrentRow", "0");
         for (String ofrmelcode:ofrm.elements.keySet()){
           LuiAppElement oel=ofrm.elements.get(ofrmelcode);
           if (oel.getType().equals("ITEM")){
               EmptyContext.Add_Element(ofrmelcode,"");
           }
         }
     application.setVal("FormID", pOpenFormID);
     ofrm.setElementsValues(EmptyContext);
     }
    }
    
}

    /**
     * Изменение текущей ячейки Списка типовой вызов:
     * gotoCell(null,null,null,true);
     *
     * @param pRowNum -новый номер строки Значение null - текущая строка не
     * менялась. Значение "0" - нет текущей строки,
     * @param pColCode - код нового столбца Значение null - текущий столбец не
     * менялся.
     * @param pContext -коллекция значений всех ячеек новой строки.Игнорируется,
     * если RowNum=null (тоже желательно устанавливать в null)
     * @param pXpen- true если программа вызвана не по событию клиентского тыка,
     * а по другому -т.е. в рамках другого события
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void gotoCell(String pRowNum,String pColCode,LuiCollection pContext,boolean pXpen) throws LuiException, SQLException{
    application.changeSCN("GotoCell");
    String curColumn=getVal("CurrentColumn");
    boolean rowchanged=false;
    int oldRowNum=Integer.valueOf(NVL(getVal("CurrentRow"),"0"));
    if (pContext!=null){
    if (is_not_null(pContext.toString())){//-- Пришёл контекст
       setElementsValues(pContext);//-- Устанавливаем контекст
       setVal("UK",getUK().toString());//Вычиляем новый UK
    }}
    if (is_not_null(pRowNum)){
      if (!(Integer.valueOf(pRowNum).intValue()==oldRowNum)) rowchanged=true;
    } 
    if (rowchanged || pXpen){ //Поменялась строка грида или это после Requery
     if (is_not_null(pRowNum)){
         setVal("CurrentRow",pRowNum);
     }
     if (is_not_null(pRowNum)){
       if ((Integer.valueOf(pRowNum)==0)) {// Эта строка никакая
           clearElementsValue(); //-- Стираем контекст
       }           
     }
     if (!(NVL(getVal("Is_QBE_Mode"),"N").equals("Y"))){ // если не QBE
        setUnderside(); //Вычиляем подвал
        String sm=NVL(getProperty("SelectionMode"),"Single");
        
        if (sm.equals("Single") || sm.equals("None")){ //если -- Однострочный режим
            setVal("Inversion","N");
            clearChecking();
          int rownum;
          if (is_not_null(pRowNum)){
              rownum=Integer.valueOf(pRowNum);
          } else rownum=oldRowNum;
          if (rownum>0){ //если есть текущая строка
              checkRow("Y",getUK());
          }
        }
        
     } else { // если QBE
         //int maxQBERowNum=Integer.valueOf(NVL(getVal("MaxQBErows"),"1"));
         //if (maxQBERowNum<Integer.valueOf(pRowNum)) setVal("MaxQBErows",pRowNum);
     }
    }
  if (is_not_null(pColCode)) {
    if (!compareString(pColCode,curColumn)){//Поменялась колонка грида (ячейка строки)
        setVal("CurrentColumn",pColCode);
        curColumn=pColCode;
    }
  }
  if (!pXpen){
    if (is_not_null(curColumn)){
        if (!NVL(getVal("Is_QBE_Mode"),"N").equals("Y")){
          if (elements.get(curColumn).isEditCell()){
              LuiClCommandGroup XMLcg=application.commandWriter.addGroup("");
              LuiClCommand XMLcmd=XMLcg.addCommand("Cell");
              XMLcmd.addScaParameter("FormID",getFormID(),null);
              XMLcmd.addScaParameter("Code", curColumn, null);
              XMLcmd.addScaParameter("Edit","Y", null);
          }
        }
    }
    if (!NVL(getVal("Is_QBE_Mode"),"N").equals("Y")){
      checkFrmDetails();
    }
  }
}

    /**
     * Выделяет строку
     *
     * @param pNewCheck 'Y' - выделить, 'N'-развыделить
     * @param pUK коллекция уникального ключества?
     */
public void checkRow(String pNewCheck,LuiCollection pUK){
    boolean Inv=NVL(getVal("Inversion"),"N").equals("Y");
    if (pUK!=null){
     if (!pUK.isEmpty()){
        if ((Inv && pNewCheck.equals("Y")) || (!Inv && pNewCheck.equals("N"))){ //развыделить
            checkRows.remove(pUK);
        } else {
            checkRows.put(pUK, Inv);
        }
     
    
     } 
     if (properties.containsKey("CountSelected")){
     properties.get("CountSelected").clearComputedValue(false);
     }
     if (properties.containsKey("ItemsSelectedJSON")){
     properties.get("ItemsSelectedJSON").clearComputedValue(false);    
     }
     if (properties.containsKey("RowsSelectedJSON")){
     properties.get("RowsSelectedJSON").clearComputedValue(false);    
     }
    }
  //setVal("CountSelected",String.valueOf(checkRows.size())); 
}

/**
 * Очищает отмеченные строки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void clearChecking() throws LuiException, SQLException{
    checkRows.clear();
/*    if (elements.containsKey("QUERY4PRE-MARKED$")){
     LuiAppElement qel=elements.get("QUERY4PRE-MARKED$");
     if (qel.getParent().equals("Form")){
       if (qel.getType().equals("PARAM")) {
           ArrayList<String> acstmt=application.getAppConnAndStmt(qel.getProperty("Value"), "",this.getFormID());
           LuiAppConn ac=application.userDbConnections.get(acstmt.get(0));
           String stmt=acstmt.get(1);
           if (is_not_null(stmt)){
           LuiDbConn lc=null;
           if (ac.getSessionMode()=='C'){
             lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              }
           else lc=ac.currentConnection;
           if (lc!=null){
             try {
              ArrayList<LuiCollection> ArrKeyCol=lc.getSelectAllRowsValMap(stmt);
              if (ArrKeyCol!=null) {
                for (int i=0;i<ArrKeyCol.size();i++) {
                   LuiCollection keyCol=ArrKeyCol.get(i);
                   boolean is_key=true; 
                   for (String keyColElCode:keyCol.keySet()){
                      if (elements.containsKey(keyColElCode)){
                        if (!elements.get(keyColElCode).getProperty("FOR_UNIQUE_KEY").equals("Y")) {
                          is_key=false;  
                        } 
                      } else is_key=false; 
                   }
                  if (is_key){
                    checkRows.put(keyCol,true);
                  } 
                }
              }
             } finally {
                 if (ac.getSessionMode()=='C'){
                  ac.connPool.SetFree(lc);
                    }
             }  
           }
       }}
     }
    }*/
    if (properties.containsKey("CountSelected")){
    properties.get("CountSelected").clearComputedValue(false);
    }
    if (properties.containsKey("ItemsSelectedJSON")){
        properties.get("ItemsSelectedJSON").clearComputedValue(false);
    }
    if (properties.containsKey("RowsSelectedJSON")){
        properties.get("RowsSelectedJSON").clearComputedValue(false);
    }
}
/**
 * вычисляет подвал
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setUnderside() throws LuiException, SQLException{
  if (is_not_null(getNonCompValue("ADD_INFO"))) {
        if ((getVal("Is_QBE_Mode").equals("Y")||NVL(getVal("CurrentRow"),"0").equals("0"))
           && (properties.get("ADD_INFO").recalculate!=null)){
             properties.get("ADD_INFO").setCurrentValue("");
        } 
        if (properties.get("ADD_INFO").changed()){
            String NewValue=properties.get("ADD_INFO").getComputedProperty();
            if (getVal("Uderside").equals(NewValue)) return;
            setVal("Uderside",NewValue);
            LuiClCommandGroup XMLcg=application.commandWriter.addGroup("");
            LuiClCommand XMLcmd=XMLcg.addCommand("Underside");
            XMLcmd.addScaPar("FormID", getFormID());
            NewValue=LTrim(NewValue,"\n");
            NewValue=RTrim(NewValue,"\n");
            NewValue=NewValue.replace("\n\n","\n");
            NewValue=NewValue.replace("\n\n","\n");
            XMLcmd.addScaPar("Content",NewValue);
         //   SetVal("Underside")
        }
  }
}
/**
 * Извлечение строк в матрицу -- старый вариант
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
/*
public void fetchCells() throws LuiException, SQLException{
   LuiClCommandGroup XMLcg; //группа команд
   LuiClCommand XMLcmd;// команда
   XMLcg=application.commandWriter.addGroup("FetchCells");
   XMLcg.addParameter("FormID",getFormID());
   XMLcmd=XMLcg.addCommand("ClearGrid");
   boolean b=true;
   while(b){
       b=false;
   for(String cc:elements.keySet()){
       if (elements.get(cc).getType().equals("CELL")){
           b=true;
           elements.remove(cc);
           break;
       }
   }
   }
   String Query=getProperty("QUERY_TEXT");
   if (is_not_null(Query)){// если запрос не пуст
       if (Query.indexOf(" * ")>0) {// если есть звёздочка
         String SelectClause=getProperty("COORDINATE")+" COORDINATE";
         if (!getProperty("CELL_TEXT").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_TEXT"),"null")+" CELL_TEXT";
         if (!getProperty("CELL_HALIGN").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_HALIGN"),"null")+" CELL_HALIGN";
         if (!getProperty("CELL_VALIGN").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_VALIGN"),"null")+" CELL_VALIGN";
         if (!getProperty("CELL_OVERFLOW").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_OVERFLOW"),"null")+" CELL_OVERFLOW";
         if (!getProperty("CELL_VISUAL").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_VISUAL"),"null")+" CELL_VISUAL";
         if (!getProperty("CELL_MAKEUP").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_MAKEUP"),"null")+" CELL_MAKEUP";
         if (!getProperty("CELL_HINT").equals("Null")) SelectClause=SelectClause+", "+NVL(getProperty("CELL_HINT"),"null")+" CELL_HINT";
         Query=Query.replace(" * ", " "+SelectClause+" ");
       }
           application.log.addRecord("MAIN_QUERY",application.NLS.NLS("RUS:Выполнение основного запроса;ENG:Main query;"),Query,this.getFormID());
           LuiDbConn lc=null;
           Integer Cursor=null;
           LuiAppConn ac=null;
           try{
           ArrayList<String> acstmt=application.getAppConnAndStmt(Query,getProperty("Connection"),getFormID());
           String acc=acstmt.get(0);
           String stmt=acstmt.get(1);
           ac=application.userDbConnections.get(acc);
           if (ac.getSessionMode()=='C') { // если соединение коммунальное
            lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
            }
             else 
              {//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
              }
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
            Cursor=lc.CursorOpen(stmt);
            ArrayList<ColInfo> ColsInfo=lc.CursorGetColsInfo(Cursor);
            while (lc.CursorNextRow(Cursor)){// крутим строки
               ArrayList<String> row=lc.CursorGetRow(Cursor);
               LuiAppElement cel=null;
               for (int i=0;i<row.size();i++){
                   if (ColsInfo.get(i).alias.toUpperCase().equals("COORDINATE")){
                     LuiCollection col=new LuiCollection(row.get(i));
                     cel=this.addElement(col.toString(),"CELL",  getNewSeq("CELL"),col.toString(),"");
                     for (String colkey:col.keySet()) {
                         if (colkey.equals(getVal("RowDimension"))){
                           if (!elements.containsKey(col.Get_Element(colkey))){
                               this.addElement(col.Get_Element(colkey),"ITEM",  getNewSeq("ITEM"),col.Get_Element(colkey), colkey);
                               XMLcmd=XMLcg.addCommand("Row");
                               XMLcmd.addScaPar("Code",col.Get_Element(colkey));
                           } 
                           cel.addProperty("R-code",col.Get_Element(colkey));
                         }
                         if (colkey.equals(getVal("ColumnDimension"))){
                           if (!elements.containsKey(col.Get_Element(colkey))){
                               this.addElement(col.Get_Element(colkey),"ITEM",  getNewSeq("ITEM"),col.Get_Element(colkey), colkey);
                               XMLcmd=XMLcg.addCommand("Column");
                               XMLcmd.addScaPar("Code",col.Get_Element(colkey));
                           }  
                           cel.addProperty("C-code",col.Get_Element(colkey));
                         }
                     }
                     break;
                   }
               }
               for (int i=0;i<row.size();i++){ // второй круг кручения данных строки
                switch(ColsInfo.get(i).alias.toUpperCase()){
                  case "CELL_TEXT": cel.addProperty("Value",row.get(i)); break;
                  case "CELL_HINT": cel.addProperty("Hint",row.get(i)); break;
                  case "CELL_HALIGN": cel.addProperty("H-align",row.get(i)); break;
                  case "CELL_VALIGN": cel.addProperty("V-align",row.get(i)); break;
                  case "CELL_OVERFLOW": cel.addProperty("Overflow",row.get(i)); break;
                  case "CELL_MAKEUP": cel.addProperty("MakeUp",row.get(i)); break;
                  case "CELL_GETUP": cel.addProperty("GetUp",row.get(i)); break;
                  case "CELL_VISUAL": cel.addProperty("Class",row.get(i)); break;     
                }   
               }
            }//крутим строки
            lc.CursorClose(Cursor);
            Cursor=null;
   } catch (SQLException sqlex){
    LuiAppProp p=this.properties.get("QUERY_TEXT");
   application.msgWriter.sqlError(lc, sqlex, p, Query);
         }
           finally{
          if (Cursor!=null) lc.CursorClose(Cursor);
          if (lc!=null && ac!=null) {
           if (ac.getSessionMode()=='C') { // если соединение коммунальное
               ac.connPool.SetFree(lc); // освободим его
           }   
          }  
           }
   }// запрос не пуст
   for(String cc:elements.keySet()){
       LuiAppElement elcel=elements.get(cc);
       if (elcel.getType().equals("CELL")){
           XMLcmd=XMLcg.addCommand("Cell");
           XMLcmd.addScaPar("R-code",elcel.getProperty("R-code"));
           XMLcmd.addScaPar("C-code",elcel.getProperty("C-code"));
           XMLcmd.addScaParameter("Show",elcel.getProperty("Value"), def());
           XMLcmd.addScaParameter("Hint",elcel.getProperty("Hint"), def());
           XMLcmd.addScaParameter("H-align",elcel.getProperty("H-align"), def());
           XMLcmd.addScaParameter("V-align",elcel.getProperty("V-align"), def());
           XMLcmd.addScaParameter("Overflow",elcel.getProperty("Overflow"), def());
           XMLcmd.addScaParameter("MakeUp",elcel.getProperty("MakeUp"), def());
           XMLcmd.addScaParameter("GetUp",elcel.getProperty("GetUp"), def());
           XMLcmd.addScaParameter("Class",elcel.getProperty("Class"), def());
           
       }
   }
}
*/
/**
 * Извлечение строк в список
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void fetchRows() throws LuiException, SQLException{
  if (getVal("IsLastPage").equals("Y"))return;
  String queryText=LuiParser.reMask(getVal("MainQuery"));
  if (is_null(queryText)) return;  // уходим если нет запроса
  LuiClCommandGroup XMLcg; //группа команд
  LuiClCommand XMLcmd;// команда
  LuiCollection CTX; // какой-то контекст для деревьев
  String ChangingNode=null;
  String PrevNode=null;
  Integer pageSize=String2Integer(getProperty("PORTION"),32); // размер извлекаемой порции
  Integer mainCursor=null; // курсор запроса
  int fetchedRows=Integer.valueOf(NVL(getVal("FetchedRows"),"0")); // cколько строк извлечено
  int rowRank=Integer.valueOf(NVL(getVal("RowRank"),"0")); // cколько строк извлечено
  int prefRow=Integer.valueOf(NVL(getVal("PrefferedRow"),"0")); // желательный номер строки    
  String prefUK=getVal("UK"); // UK желательной строки
  LuiCollection finalRowContext=getFieldsValues(); // Контекст после Fetch
  int finalRowNum=Integer.valueOf(NVL(getVal("CurrentRow"),"0")); //Номер строки после Fetch
  boolean Inv=NVL(getVal("Inversion"),"N").equals("Y");
  String rootFormID=NVL(getVal("RootFormID"),getFormID());
  if (getType().equals("LIST")) {
      if (is_null(prefUK)){
       setVal("UK",getUK().toString());
      }
  }
  String cc=getProperty("Connection");
  LuiAppConn ac=application.userDbConnections.get(cc);
  LuiDbConn lc=null;
  try {
  if (ac.getSessionMode()=='C'){ //соединение коммунальное
      lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
      queryText=lc.setOffsetClause(queryText,OFF$SET,getProperty("QUERY_HINT"),fetchedRows); // добавим к условию, что извлекать надо rownum>fetchedrows
      queryText=queryText+" limit "+pageSize.toString();
      if (application.metaConnPool==ac.connPool){
      ac.connPool.SetFree(lc); //освобождаем для записи лога если лог и прикладуха из одного пула
      lc=null;
      }
      if (fetchedRows>0) {
          application.log.addRecord("MAIN_QUERY",application.NLS.NLS("RUS:Выполнение основного запроса;ENG:Main query;")+
             "\n ("+application.NLS.NLS("RUS:Очередная порция строк;ENG:Next rows set;")+")",queryText,this.getFormID());
      } else {
      application.log.addRecord("MAIN_QUERY",application.NLS.NLS("RUS:Выполнение основного запроса;ENG:Main query;")+
              "\n ("+application.NLS.NLS("RUS:Первая порция строк;ENG:First rows set;")+")",queryText,this.getFormID());
      }
      if (lc==null){
      lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());//вновь захватываем
      }
      mainCursor=lc.CursorOpen(queryText); //открываем курсор для коммунального 
  } else
  { //соединение частное
      lc=ac.currentConnection;
      mainCursor=Integer.valueOf(getVal("MainCursor")); // курсор
  }    
  if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!");
  ArrayList<ColInfo> colList=lc.CursorGetColsInfo(mainCursor); // список описаний столбцов
  int colsN=lc.CursorGetCols(mainCursor); //количество столбцов в запросе
  //String colCode="";//13.05.2021 Вынос order by наружу
  
  
  XMLcg=application.commandWriter.addGroup("FechRows");
  XMLcg.addParameter("FormID",rootFormID );
  if (getType().equals("TREE")){
  XMLcg.addParameter("ParentNode", getVal("ParentNode"));
  XMLcg.addParameter("Type", "ROW");
  XMLcg.addParameter("ContextID",getFormID());
  }
  LuiAppElement el;
  for(int i=1;i<=pageSize.intValue();i++){ // цикл выборки
        if (!lc.CursorNextRow(mainCursor)) { //больше строк нет
        setVal("IsLastPage","Y");//Устанавливаем признак окончания выборки
        lc.CursorClose(mainCursor); //Закрываем курсор
        mainCursor=null;
        setVal("MainCursor","");
        if (getType().equals("LIST")) {
           XMLcmd=XMLcg.addCommand("Grid");
           XMLcmd.addScaParameter("FormID",getFormID() ,"");
           XMLcmd.addScaParameter("IsLast","Y","");
           if (getVal("TotalStatus").equals("Enable")){
             XMLcmd=XMLcg.addCommand("Event");  
             XMLcmd.addScaParameter("EventCode","GetTotal", "");
             XMLcmd.addScaParameter("MS","1", "");
             XMLcmd.addScaParameter("Params","FormID:"+getFormID()+";", "");
           }
        }
        break;
    }  //конец //больше строк нет
    //-- Попали сюда - значит получили очередную строку
    fetchedRows=fetchedRows+1; //увеличим счётчик строк
    setVal("FetchedRows",String.valueOf(fetchedRows));
    setVal("CurrentRow",String.valueOf(fetchedRows));
    
    CTX=new LuiCollection("");
    String valBuffer;
    int qcc=0; //13.05.2021 Вынос order by наружу
     for(int n=0;n<elements.size();n++){
         el=elSort(n);
       if (!el.getType().equals("ITEM"))continue;
       if (el.colinfo==null) continue;
       if (NVL(el.getProperty("Read"),"Y").equals("Y")){ // и его можно показывать 
         String ldt=el.getProperty("DATA_TYPE");   
         if (ldt.equals("RICHTEXT")) valBuffer=null;
         else valBuffer=lc.CursorGetRow(mainCursor).get(qcc); // берём его значение из jdbc
         if (valBuffer==null){
              if ("EMPTY_STRING".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode) 
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
              }
              }
              if ("".equals(valBuffer)){
              if ("NULL".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
              }
              }
       el.setValue(valBuffer,false,"",false); // и запихиваем в значение столбца
      if (getType().equals("TREE")){ //Если это дерево, то контекст собираем в переменной СТХ
          CTX.Add_Element(el.Code, valBuffer);
      } else {
          CTX.Add_Element(el.Code,null);
      }      
       }
       qcc++;
     }
     // выполняем подсчёт ранка
    if (el4order!=null && el4order.size()>0){
       if (fetchedRows==1){
           rowRank=1;
           setVal("RowRank",String.valueOf(rowRank));
           setVal("LastValue4RowRank",NVL(el4order.get(0).getValue(),""));
       } else {
           if (!getVal("LastValue4RowRank").equals(NVL(el4order.get(0).getValue(),""))){
           rowRank++;
           setVal("RowRank",String.valueOf(rowRank)); 
           setVal("LastValue4RowRank",NVL(el4order.get(0).getValue(),""));
           }
       }
       if (elements.containsKey("SORT_SEPARATE_BY") && elements.containsKey("ROWNO")) {
           if (elements.get("SORT_SEPARATE_BY").getValue().equals("ORDER")) {
              elements.get("ROWNO").setValue(String.valueOf(fetchedRows), false, "",false);
           } else {
              elements.get("ROWNO").setValue(String.valueOf(rowRank), false, "",false);
           }
       }
    }
      /*//13.05.2021 Вынос order by наружу
    for(int n=0;n<colsN;n++){ //-- По всем столбцам
      //colCode=colList.get(n).name; //код столбца
      colCode=colList.get(n).alias; //код столбца
      if (elements.containsKey(colCode)) { // если есть такой элемент
      el=elements.get(colCode); // берём его
      if (NVL(el.getProperty("Read"),"Y").equals("Y")){ // и его можно показывать 
      String ldt=el.getProperty("DATA_TYPE");
      if (ldt.equals("RICHTEXT")) valBuffer=null;
      else valBuffer=lc.CursorGetRow(mainCursor).get(n); // берём его значение из jdbc
              if (valBuffer==null){
              if ("EMPTY_STRING".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode) 
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
              }
              }
              if ("".equals(valBuffer)){
              if ("NULL".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
              }
              }      
                
      el.setValue(valBuffer,false,"",false); // и запихиваем в значение столбца
      if (getType().equals("TREE")){ //Если это дерево, то контекст собираем в переменной СТХ
          CTX.Add_Element(colCode, valBuffer);
      }
      }
    }
    } *///13.05.2021 Вынос order by наружу
    
    
    if (getType().equals("LIST")) { //-- Обрабатываем FetchRows в Списке
      if (fetchedRows==1){ // Это первая строка Списка. Устанивим итоговый контекст пока на неё
          finalRowNum=1; //Возможно, до желаемой строки мы так и не дойдём,...
          finalRowContext=this.getFieldsValues(); //...тогда текущей будет первая строка
      }
      XMLcmd=XMLcg.addCommand("AddRow"); //-- Команда клиенту добавления строки
     
     /*  if (Inv) {// Инверсия включена
          
             XMLcmd.addScaPar("Check","Y");
          
      } */
      boolean rowChecked=false;
       if (Inv) {// Инверсия включена
       rowChecked=true;   
      }
      if (elements.containsKey("LUI$CHECK_ROW")){
          if (elements.get("LUI$CHECK_ROW").getValue().equals("Y")){
              checkRow("Y",getUK());
              rowChecked=true;
              if (Inv) {// Инверсия включена
              rowChecked=false;   
                }
          }
      }
      if (rowChecked)
       XMLcmd.addScaPar("Check","Y");
      //13.05.2021 Вынос order by наружу
      for (String ec:CTX.keySet()){
       el=elements.get(ec);
       
        XMLcmd=XMLcg.addCommand("Cell"); //-- Команда клиенту добавления ячейки  
        XMLcmd.addScaPar("Code",ec);//Код столбца, которому принадлежит ячейка
        //-- Значение (текст) в ячейке. Если пусто - не будем отправлять параметр. Клиент поймёт, что текста в ячейке нет.
        valBuffer=el.getProperty("Value");
        if (is_not_null(valBuffer)){
          XMLcmd.addScaParameter("Show",valBuffer,null);
        }
/*29.04.20      
        if (is_not_null(el.getNonCompValue("VISUALIZATION"))){ //Может быть расскраска ячейки
          XMLcmd.addScaParameter("Style",el.getProperty("VISUALIZATION"),null);  //Расскраска
        }
        if (is_not_null(el.getNonCompValue("MAKEUP"))){ //Может быть расскраска ячейки
          XMLcmd.addScaParameter("MakeUp",LuiString.RTrim(el.getProperty("MAKEUP"),"\n"),null);  //Расскраска
        }
29.04.20*/
 
        //if (el.properties.get("VISUALIZATION").recalculate!=null){ //Раскраска динамичная (может меняться)
        if (el.properties.get("VISUALIZATION").recalculate!=null || el.properties.get("VISUALIZATION").isDynamic()){ //Раскраска динамичная (может меняться)
            XMLcmd.addScaParameter("Style", el.getProperty("VISUALIZATION"),""); //  Безусловная отправка раскраски (пусть, даже пустой).
        } //В противном случае во всех ячейках будет отображаться раскраска, заданная при создании столбца на клиенте (команда Column)
        if (el.properties.get("MAKEUP").recalculate!=null || el.properties.get("MAKEUP").isDynamic() || el.properties.get("VISIBLE").isDynamic()|| el.properties.get("WIDTH").isDynamic()){//Раскраска динамичная (может меняться)
           XMLcmd.addScaPar("MakeUp", LuiString.RTrim(el.getProperty("MAKEUP"), "\n"));  //  Безусловная отправка раскраски (пусть, даже пустой).
        } //В противном случае во всех ячейках будет отображаться раскраска, заданная при создании столбца на клиенте (команда Column)
        if (el.properties.get("GETUP").recalculate!=null || el.properties.get("GETUP").isDynamic() || el.properties.get("VISIBLE").isDynamic()|| el.properties.get("WIDTH").isDynamic()){//05.11.2020
           XMLcmd.addScaPar("GetUp", el.getProperty("GETUP"));  //  05.11.2020
        } 
        if (el.properties.get("HINT").recalculate!=null){//Подсказка динамичная (может меняться)
            XMLcmd.addScaParameter("Hint",el.getProperty("HINT"),null); //            -- Безусловная отправка хинта (пусть, даже пустого).
        } //В противногм случае во всех ячейках юудет отображаться хинт, заданный при создании столбца на клиенте (команда Column)
        if (is_not_null(el.getProperty("LocBut"))){ //Есть иконки
          LuiCollection LocBut=new LuiCollection(el.getProperty("LocBut")); //Коллекция действий с иконками для отображения в ячейке
          String Buttons="";  
          String LocIcon;
          for (String key:LocBut.keySet()){
              if (key==null)continue;
              if (!elements.containsKey(key))continue;
              LocIcon=elements.get(key).getProperty("ICON");
              if (is_not_null(LocIcon)){
                  Buttons=Buttons+LocIcon+":"+elements.get(key).getProperty("EXPLANATION")+";";//Добавляем к коллекцию иконок с хинтом действия
              }
          }
          XMLcmd.addScaParameter("Buttons",Buttons,null);// Отправляем на клиента набор иконок с хинтами - для отображения в ячейке
        }
      
      }
      /*//13.05.2021 Вынос order by наружу
      //-- Цикл по всем доступным пользователю столбцам
      for(int n=0;n<colsN;n++){ //-- По всем столбцам 
         
      colCode=colList.get(n).alias; //код столбца
      
      if (elements.containsKey(colCode)) { // если есть такой элемент
      el=elements.get(colCode);
      if (NVL(el.getProperty("Read"),"Y").equals("Y")){ // и его можно показывать 
        XMLcmd=XMLcg.addCommand("Cell"); //-- Команда клиенту добавления ячейки  
        XMLcmd.addScaPar("Code",colCode);//Код столбца, которому принадлежит ячейка
        //-- Значение (текст) в ячейке. Если пусто - не будем отправлять параметр. Клиент поймёт, что текста в ячейке нет.
        valBuffer=el.getProperty("Value");
        if (is_not_null(valBuffer)){
          XMLcmd.addScaParameter("Show",valBuffer,null);
        }
*29.04.20      
        if (is_not_null(el.getNonCompValue("VISUALIZATION"))){ //Может быть расскраска ячейки
          XMLcmd.addScaParameter("Style",el.getProperty("VISUALIZATION"),null);  //Расскраска
        }
        if (is_not_null(el.getNonCompValue("MAKEUP"))){ //Может быть расскраска ячейки
          XMLcmd.addScaParameter("MakeUp",LuiString.RTrim(el.getProperty("MAKEUP"),"\n"),null);  //Расскраска
        }
29.04.20*
        //if (el.properties.get("VISUALIZATION").recalculate!=null){ //Раскраска динамичная (может меняться)
        if (el.properties.get("VISUALIZATION").recalculate!=null || el.properties.get("VISUALIZATION").isDynamic()){ //Раскраска динамичная (может меняться)
            XMLcmd.addScaParameter("Style", el.getProperty("VISUALIZATION"),""); //  Безусловная отправка раскраски (пусть, даже пустой).
        } //В противном случае во всех ячейках будет отображаться раскраска, заданная при создании столбца на клиенте (команда Column)
        if (el.properties.get("MAKEUP").recalculate!=null || el.properties.get("MAKEUP").isDynamic() || el.properties.get("VISIBLE").isDynamic()|| el.properties.get("WIDTH").isDynamic()){//Раскраска динамичная (может меняться)
           XMLcmd.addScaPar("MakeUp", LuiString.RTrim(el.getProperty("MAKEUP"), "\n"));  //  Безусловная отправка раскраски (пусть, даже пустой).
        } //В противном случае во всех ячейках будет отображаться раскраска, заданная при создании столбца на клиенте (команда Column)
        if (el.properties.get("GETUP").recalculate!=null || el.properties.get("GETUP").isDynamic() || el.properties.get("VISIBLE").isDynamic()|| el.properties.get("WIDTH").isDynamic()){//05.11.2020
           XMLcmd.addScaPar("GetUp", el.getProperty("GETUP"));  //  05.11.2020
        } 
        if (el.properties.get("HINT").recalculate!=null){//Подсказка динамичная (может меняться)
            XMLcmd.addScaParameter("Hint",el.getProperty("HINT"),null); //            -- Безусловная отправка хинта (пусть, даже пустого).
        } //В противногм случае во всех ячейках юудет отображаться хинт, заданный при создании столбца на клиенте (команда Column)
        if (is_not_null(el.getProperty("LocBut"))){ //Есть иконки
          LuiCollection LocBut=new LuiCollection(el.getProperty("LocBut")); //Коллекция действий с иконками для отображения в ячейке
          String Buttons="";  
          String LocIcon;
          for (String key:LocBut.keySet()){
              if (key==null)continue;
              if (!elements.containsKey(key))continue;
              LocIcon=elements.get(key).getProperty("ICON");
              if (is_not_null(LocIcon)){
                  Buttons=Buttons+LocIcon+":"+elements.get(key).getProperty("EXPLANATION")+";";//Добавляем к коллекцию иконок с хинтом действия
              }
          }
          XMLcmd.addScaParameter("Buttons",Buttons,null);// Отправляем на клиента набор иконок с хинтами - для отображения в ячейке
        }
      }// конец и его можно показывать 
      }// конец если есть такой элемент
      } //конец Цикл по всем доступным пользователю столбцам
  */
    } //конец -- Обрабатываем FetchRows в Списке
   if (getType().equals("TREE")) { //-- Обрабатываем FetchRows в дереве
      XMLcmd=XMLcg.addCommand("Node"); //Команда клиенту создания ноды
      XMLcmd.addScaParameter("Node",getFormID()+"#"+String.valueOf(fetchedRows),null);//Уникальный идентификатор ноды
      String State=NVL(getProperty("NODE_STATE"),"Leaf");//По умолчанию тип ноды - листовая
      XMLcmd.addScaParameter("Label",getProperty("NODE_TEXT"),null);//Подпись к ноде
      XMLcmd.addScaParameter("Icon",getProperty("NODE_ICON"),null);//Иконка ноды (если нет, то и параметра не будет)
      XMLcmd.addScaParameter("Style",getProperty("NODE_VISUALIZATION"),null);//Раскраска узла
      
      XMLcmd.addScaPar("Hint",getProperty("ADD_INFO"));//Хинт ноды отправляем всегда (У Жени если так не сделать, то хинт возьмётся от предка узла
      if (is_null(getProperty("NODE_STATE"))) {
      //Теперь посмотрим - можеть быть будет возмоджность раскрывать это ноду. Пройдёмся по всем глобальным действиям
      for(String elcode:elements.keySet()){
          el=elements.get(elcode);
          if (el.getType().equals("ACTION")){
            if (like(elcode,"SYSMENU$%") || like(elcode,"COLMENU$%") || like(elcode,"CELLMENU$%")) continue;  
            if (el.getProperty("ACTIVE").equals("EXIST")) { //если действие для строки
              if(el.getProperty("COMMAND_TYPE").equals("LIST")){//Еслидействие пытается открыть другой список
                  State="Collapsed"; // То тип ноды будет "нераскрытый плюс"
                if (like(el.getProperty("ROLE"),"%ALLEXPAND%")){//А если действие ещё и автораскрываемое
                  State="AutoExpand"; // То попросим клиента немедленно как бы тыркнуть за пользователя в ноду для её раскрытия  
                }
                break;//В принципе, одного такого действия достаточно - можно выходить из цикла.
              }  
            }
          }
      }
   }
     XMLcmd.addScaParameter("State",State,null); //Отправим на клиента тип ноды
     XMLcmd.addScaParameter("Context",CTX.toString(),null); //Отправим на клиента контекст (он его возвратит "как есть" при событиях на этой ноде)
     String CurrentLevel=null;
     if (CTX.containsKey("LEVEL")){
     CurrentLevel=CTX.Get_Element("LEVEL"); // Поищем в контексте столбец с кодом LEVEL.
     }
    // Тут реализован механизм, позволяющий "плоскую" структуру нод одного уровня сделать иерархической (одним основным запросом Списка)
    ChangingNode=null;
    if (is_not_null(CurrentLevel)){ //Столбец с кодом LEVEL определяет уровень ноду в иерархии (относительно базовой, плоской)
      LuiCollection Levels2Parents=new LuiCollection(getVal("NodeLevels"));//Это переменная - коллекция кодов родительских нод для каждого уровня иерархии
      LuiCollection NewLevels2Parents=new LuiCollection("");//По ходу строим новую подобнуж коллекцию
      String CurrentParent=getVal("ParentNode");
      for (String Level:Levels2Parents.keySet()){
        if (Integer.valueOf(Level)>Integer.valueOf(CurrentLevel)) break; //Пробежимся по всем уровням до текущей ноды
        CurrentParent=Levels2Parents.get(Level);
        NewLevels2Parents.Add_Element(Level, CurrentParent);//Строим новую коллекцию родителей
      }
      if (PrevNode!=null){
      if (PrevNode.equals(CurrentParent)){ //Если предыдущая нода стала родительской для текущей
          ChangingNode=PrevNode; //то эту предыдущую ноду нужно будет немного изменить
      }}
      //Добавляем родителя текущей ноды в коллекцию родительских нод (для каждого уровня)
      NewLevels2Parents.Add_Element(CurrentLevel,CurrentParent);
      PrevNode=getFormID()+"#"+fetchedRows; //Для следующей итерации
      //Добавим текущую ноду в коллекцию родительских нод (вдруг следующая нода будет ей подчиняться?)
      NewLevels2Parents.Add_Element(String.valueOf(Integer.valueOf(CurrentLevel)+1),PrevNode);
      if (State.equals("Leaf")){//Если тип текущей ноды - лист, то бессмысленно его будет менять в будущем
          PrevNode=null;
      }
      setVal("NodeLevels",NewLevels2Parents.toString()); //Сохраняем коллекцию родительских нод
      XMLcmd.addScaParameter("ParentNode",CurrentParent,"\n");//Отправляем на клиента информацию о родителе текущей ноды
      
    }
   } //конец -- Обрабатываем FetchRows в дереве

   if (getUK().toString().equals(prefUK)){//Новая строка была текущей до перезапроса (по UK)
       prefRow=fetchedRows;
   }
   if (fetchedRows==prefRow){//если дошли до желательной строки - сделаем её текущей
       finalRowNum=prefRow; 
       finalRowContext=getFieldsValues();
   }
   if (is_not_null(ChangingNode)){//Предыдущая нода была родительской для текущей и не листовой
       LuiClCommand XMLcmd2=XMLcg.addCommand("Node");//Сделаем её листовой!
       XMLcmd2.addScaParameter("Node", ChangingNode,null);
       XMLcmd2.addScaParameter("State","Leaf",null);
   }
  } // конец цикла фетчей
  //Всё, тут Fetch закончился
  if (getType().equals("TREE")){//Для дерева
     if (!NVL(getVal("IsLastPage"),"Y").equals("Y")) {//Если курсор всё ещу открыт (недофетчено)
         XMLcmd=XMLcg.addCommand("Node");//То создаём специальную технологическую ноду
         XMLcmd.addScaParameter("Node",getFormID()+"#Continue",null);
         XMLcmd.addScaParameter("ParentNode",getVal("ParentNode"),null);
         XMLcmd.addScaParameter("Type","CONTINUE",null);
         XMLcmd.addScaParameter("Label","...",null);
         XMLcmd.addScaParameter("State","Continue",null);
         XMLcmd.addScaParameter("Context",String.valueOf(fetchedRows),null);
         XMLcmd.addScaParameter("ContextID",getFormID(),null);
     }
  }
  setVal("CurrentRow",String.valueOf(finalRowNum));//(Во/У)станавливаем - что у нас там получилось текущей строкой/нодой (номер)
  setVal("PrefferedRow",null);//Своё дело оже сделало. Следующие фетчи уже точно не будут пытаться установить текущую строку
  setElementsValues(finalRowContext);
  setVal("UK",getUK().toString());
  //19.03.2020    checkRunTimer();
} catch (SQLException sqlex){
    LuiAppProp p=this.properties.get("QUERY_TEXT");
   application.msgWriter.sqlError(lc, sqlex, p, queryText);
}
  finally { // чтобы ни случилось
      if (lc!=null){   // если есть соединение
        if (ac.getSessionMode()=='C'){// если соединение коммунальное,
            if (mainCursor!=null) {
                lc.CursorClose(mainCursor); // закроем курсорчик
                mainCursor=null;
                setVal("MainCursor","");
            }
            ac.connPool.SetFree(lc); //вернём соединение в пул свободных
        }  
      }
  } 
}

/**
 * Установка значений полей из коллекции
 * @param pCollection коллекция полей
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса 
 */
public void setElementsValues(LuiCollection pCollection) throws LuiException, SQLException{
    for (String key:pCollection.keySet()){
      if (elements.containsKey(key)){
          elements.get(key).setValue(pCollection.get(key),false,"N",false);
      }
    }
}
/**
 * формирование фразы order by для текста запроса
 * @param lc соединение с базой данных
 * @param pEl4order коллекция элементов для сортировки
 * @return фраза order by 
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String addOrderBy(LuiDbConn lc,ArrayList<LuiAppElement> pEl4order ) throws LuiException, SQLException{
    String retVal="";
    String sort="";
    LuiAppElement el;
    String dt;
    String conv;
    String v;
    addProperty("SortedBy","");
    for (int i=0;i<pEl4order.size();i++){
     if (i==0) {
         addProperty("SortedBy",pEl4order.get(i).getCode());
     }
      if (is_not_null(retVal)) retVal=retVal+",";
      el=pEl4order.get(i);
      dt=NVL(el.getProperty("DATA_TYPE"));
      if (el.getProperty("Sintetic").equals("Y") 
              //&& !el.isWinFunc() // 15.05.2021 это для того чтобы правильно работал ROW$NO
                                 // 02.06.2021 но если над столбцом будет применять формат, то сортировка ролучится плохая поэтому isWinFunc закоментирован
              ) v="("+el.getProperty("DEFINITION")+")";
      else
      v="\""+pEl4order.get(i).getCode()+"\"";
      if ("Y".equals(application.dataConverter.LuiDataTypes.get(dt).tl_order_by_clause)){
          conv=application.dataConverter.getConvString(el.colinfo, lc, true,true);
          if (is_not_null(conv)){
           el.addProperty("v", v);
           el.addProperty("f", el.getProperty("FORMAT"));
           LuiAppProp p=el.addProperty("ColumnSQLClause",conv);
           conv=p.getComputedProperty();
           v=conv;
          }
      }
      retVal=retVal+v;
      sort=pEl4order.get(i).getProperty("SORT").trim();
      if (like(sort,"-%")) {
          retVal=retVal+" desc";
          if (like(sort,"-0%")) { // если указан лидирующий нолик а сортирвка по убыванию,
              retVal=retVal+" nulls last "; //то изменим умалчиваемый режим nulls first на nulls last
          }
      } else {
          retVal=retVal+" asc";
          if (like(sort,"0%")) { // если указан лидирующий нолик а сортирвка по возрастанию,
              retVal=retVal+" nulls first "; //то изменим умалчиваемый режим nulls last на nulls first
          }
      }
      
    }
    if (is_not_null(retVal)) retVal=" order by "+retVal;
    setVal("OrderByClause",retVal);
    if (this.DgV2!=null){
     if (!getProperty("DEFSORT").equals(getVal("OrderBy"))) {
      PreparedStatement pstmt=null;
      LuiDbConn mlc=null;
      try{ 
        mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());// возспользуемся соединением к метаданным  
        mlc.executeDML("delete from lui_t_data_group_chart_pref_val where prop_code in ('ORDER_BY','ORDER_BY_LIST') and grouping_id="+DgV2.DgId+"and chart_seq="+DgV2.Level);
        
         pstmt=mlc.getConnection().prepareStatement("insert into lui_t_data_group_chart_pref_val(grouping_id,chart_seq,prop_code,value)values(?::lui_d_id,?::lui_d_id,?,?)");
         pstmt.setString(1,DgV2.DgId);
         pstmt.setString(2,DgV2.Level);
         if (!DgV2.chartType.equals("MAP")){
         pstmt.setString(3,"ORDER_BY");
         pstmt.setString(4,NVL(retVal,""));
         pstmt.execute(); }
         pstmt.setString(3,"ORDER_BY_LIST");
         pstmt.setString(4,NVL(getVal("OrderBy"),""));
         pstmt.execute();
         setProperty("DEFSORT",getVal("OrderBy"));
        
      }catch(SQLException sex){
          String here="here";
      }
       finally{
          if (pstmt!=null)pstmt.close();
          application.metaConnPool.SetFree(mlc);
      }  
    }
    }
    return retVal;
}
public String getGroupingWhere4chartWhere(String pWhere,String pDgID) throws LuiException, SQLException{
    String retval="";
    LuiDbConn lc=null;
    try{
    lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); 
    retval =pWhere;
    
    if (is_not_null(retval)){
    retval=retval.substring(6);    
    ArrayList<LuiCollection> ListCol=
            lc.getSelectAllRowsValMap("select seq,src_column from lui_t_data_group_column where grouping_id="+pDgID+" order by seq");
    for (int i=0;i<ListCol.size();i++){
        LuiCollection col=ListCol.get(i);
        retval=retval.replace("\"GC"+col.Get_Element("seq")+"\"","\""+col.Get_Element("src_column")+"\"");
    }
    ListCol=
            lc.getSelectAllRowsValMap("select seq,src_column from lui_t_data_group_function where grouping_id="+pDgID+" order by seq");
    for (int i=0;i<ListCol.size();i++){
        LuiCollection col=ListCol.get(i);
        retval=retval.replace("\"SF"+col.Get_Element("seq")+"\"","\""+col.Get_Element("src_column")+"\"");
    }
    retval="("+retval+")";
    } else return "";
    } finally {
        if (lc!=null)
     application.metaConnPool.SetFree(lc);
} 
    return retval;
}
public String getWhere4DgCollapsedLevels(LuiDbConn plc) throws LuiException, SQLException{
    if (this.collapsedDGLevels==null || this.collapsedDGLevels.size()==0) return "";
    String retVal="";
    String condition="";
    String lcond="";
    LuiAppConn ac=null;
    LuiDbConn lc=null;
    if (plc==null) {
        String  cc=getProperty("Connection");
        ac=application.userDbConnections.get(cc);
        if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
 
    } else {
        lc=plc;
    }
    LuiCollection col;
    for(int i=0;i<this.collapsedDGLevels.size();i++){
      col=this.collapsedDGLevels.get(i);
      lcond="";
      for (String ec:col.keySet()){
          condition=elements.get(ec).getSQLCondition(lc,col.get(ec),"N");
          if (!lcond.equals("")) lcond=lcond+" and ";
          lcond=lcond+condition;
      }
      if (!lcond.equals("")){
          lcond="not ("+lcond+")";
         if (!retVal.equals(""))retVal=retVal+"\n and ";
         retVal=retVal+lcond;
      }
    }
if (!retVal.equals("")) retVal=" where "+retVal;
return retVal;
}
/**
 * Строит текст предикатов дополнительных условий отбора в формах-списках (QBE)
 * побочный эффект - добавление текста-индикатора в заголовок формы, если действуют условия QBE
 * @param lc соединение с базой данных
 * @return - условие на основе QBE
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String addWhere(LuiDbConn plc,String pInnerQBE,boolean p4winFunc) throws LuiException, SQLException{
    String retVal="";
    String ColVal;
    String condition="";
    LuiAppConn ac=null;
    LuiDbConn lc=null;
    if (plc==null) {
        String  cc=getProperty("Connection");
        ac=application.userDbConnections.get(cc);
        if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
 
    } else {
        lc=plc;
    }
    try {
    HashMap<String,String> rowQBE;
    if (getProperty("QBE_COL").equals("Y")) {
    HashMap<String,ArrayList<String>> colQBE=new HashMap<>();
    for (String rnqbe:listQBE.keySet()){ // цикл по сторкам QBE
     rowQBE=listQBE.get(rnqbe); 
     for (String ColCode:rowQBE.keySet()){ // цикл по столбцам QBE
      if (!p4winFunc && elements.get(ColCode).isWinFunc()) continue;
      if (p4winFunc && !elements.get(ColCode).isWinFunc()) continue;
       if (!colQBE.containsKey(ColCode)){   
        ArrayList<String> colcond=new ArrayList<>();
        colQBE.put(ColCode, colcond);
       }
       colQBE.get(ColCode).add(rowQBE.get(ColCode));
     }
    }
    int qbecols=0;
    for (String ColCode:colQBE.keySet()){// цикл по столбцам QBE
        ArrayList<String> colconds=colQBE.get(ColCode);
        int predicats_in_col=0;
        String qbecol="";
        for (int i=0;i<colconds.size();i++){
            ColVal=colconds.get(i);
            if (is_not_null(ColVal)){
               condition=elements.get(ColCode).getSQLCondition(lc, ColVal,pInnerQBE); 
               if (predicats_in_col>0)  qbecol=qbecol+"\n     or ";
               else qbecol=qbecol+"\n  ";
               qbecol=qbecol+condition;
               predicats_in_col++;  
            }
        }
        if (predicats_in_col>0){
           if (predicats_in_col>1){
               qbecol="("+qbecol+")";
           }
             if(qbecols>0) retVal=retVal + "\nand ";
             retVal=retVal+qbecol;
             qbecols++;
         }
    }
    } else {
    int qberows=0;
    String qberow="";
        for (String rnqbe:listQBE.keySet()){ // цикл по сторкам QBE
            rowQBE=listQBE.get(rnqbe);
        int predicats_in_row=0;
        qberow="";
         for (String ColCode:rowQBE.keySet()){ // цикл по столбцам QBE
                if (!p4winFunc && elements.get(ColCode).isWinFunc()) continue;
                if (p4winFunc && !elements.get(ColCode).isWinFunc()) continue;
  
             ColVal=rowQBE.get(ColCode);
            if (is_not_null(ColVal)){//только не пустые
         //condition=new LuiSQLCondition(lc,elements.get(ColCode),ColVal).condition;
           condition=elements.get(ColCode).getSQLCondition(lc, ColVal,pInnerQBE);
              if (predicats_in_row>0)  qberow=qberow+"\n     and ";
              else qberow=qberow+"\n  ";
              qberow=qberow+condition;
              predicats_in_row++;  
            }
         }
         if (predicats_in_row>0){
           if (predicats_in_row>1){
               qberow="("+qberow+")";
           }
             if(qberows>0) retVal=retVal + "\nor ";
             retVal=retVal+qberow;
             qberows++;
         }
        }
    }
 if (is_not_null(retVal)) {
   if (getProperty("QBE_NOT").equals("Y")) {
      retVal=" where not("+retVal+")";
     setVal("QBE_Title",NVL(application.getProperty("QBE_NOT_TITLE")," - NOT(QBE)"));  
   }  else {
     retVal=" where "+retVal;
     setVal("QBE_Title",NVL(application.getProperty("QBE_TITLE")," - QBE"));
   }
   if (getProperty("QBE_COL").equals("Y")) {
     setVal("QBE_Title",getVal("QBE_Title")+"[]");
   }
 }else{
     setVal("QBE_Title","");
 }        
    } finally {
        if (plc==null) {
            if (lc!=null) {
                if (ac!=null){
                    if (ac.getSessionMode()=='C'){
                        ac.connPool.SetFree(lc);
                    }
                }
            }
        }
    }
   if (!p4winFunc) 
    this.setVal("where",retVal);
   else
    this.setVal("where4winFunc",retVal);
    if (this.DgV2!=null && !p4winFunc){
     LuiDbConn mlc=null;
     String jsListQBE=application.toJSON(listQBE);
     if (!NVL(getProperty("DEFQBE"),"{}").equals(jsListQBE) ||
         !getProperty("DEFQBE_NOT").equals(getProperty("QBE_NOT")) ||
         !getProperty("DEFQBE_COL").equals(getProperty("QBE_COL"))
         ){
      PreparedStatement pstmt=null;
      try{  
        mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());// возспользуемся соединением к метаданным  
        mlc.executeDML("delete from lui_t_data_group_chart_pref_val where prop_code in ('QBE','QBE_LIST','QBE_NOT_LIST','QBE_COL_LIST') and grouping_id="+DgV2.DgId+"and chart_seq="+DgV2.Level);
        
         pstmt=mlc.getConnection().prepareStatement("insert into lui_t_data_group_chart_pref_val(grouping_id,chart_seq,prop_code,value)values(?::lui_d_id,?::lui_d_id,?,?)");
         
         pstmt.setString(1,DgV2.DgId);
         pstmt.setString(2,DgV2.Level);
         if (is_not_null(retVal)){
         pstmt.setString(3,"QBE");
         pstmt.setString(4,retVal);
         pstmt.execute(); 
         pstmt.setString(3,"QBE_LIST");
         pstmt.setString(4,jsListQBE);
         pstmt.execute();}
         if (!getProperty("QBE_NOT").equals("")) {
         pstmt.setString(3,"QBE_NOT_LIST");
         pstmt.setString(4,getProperty("QBE_NOT"));
         pstmt.execute(); }
         if (!getProperty("QBE_COL").equals("")) {
         pstmt.setString(3,"QBE_COL_LIST");
         pstmt.setString(4,getProperty("QBE_COL"));
         pstmt.execute();}

         setProperty("DEFQBE",jsListQBE);
         setProperty("DEFQBE_NOT",getProperty("QBE_NOT"));
         setProperty("DEFQBE_COL",getProperty("QBE_COL"));
        
      }catch(SQLException sex){
          String here="here";
      }
       finally{
          if (pstmt!=null)pstmt.close();
          application.metaConnPool.SetFree(mlc);
      }  
    }
    
    }
    return retVal;
}

/**
 * Перезапрос строки списка из БД
 * @param pUK -- String-коллекция уникального ключа "Код столбца:Значение;"
 * @return -- String-коллекция строки списка "Код столбца:Значение;"
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiCollection requeryRow(LuiCollection pUK) throws LuiException, SQLException{
     LuiAppConn ac=application.userDbConnections.get(getProperty("Connection"));
 LuiDbConn lc = null;
 Integer cursor = null;
 String ldt="";
 String valBuffer;
 int cols;
 ArrayList<ColInfo> colsInfo;
 if (pUK==null){
     pUK=getUK();
 }    
 if (is_null(pUK.toString())) {
     pUK=getUK();
 }
 
 LuiCollection row=new LuiCollection("");
 String selectText="";
 String stmt="";
 try{  
 if (ac.getSessionMode()=='C'){
     lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
 } else {
     lc=ac.currentConnection;
 }
 if (lc==null)application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!");
 
   // обходим элементы, входящие в UK
 for(String ukcolcode:pUK.keySet()){
   LuiAppElement el=elements.get(ukcolcode);
   String condition=el.getSQLCondition(lc, '='+pUK.Get_Element(ukcolcode),"N");
   if  (is_not_null(condition)){
    if (is_not_null(selectText)){
   selectText=selectText+" and ";
           }        
   selectText=selectText+condition; 
   }
 }
 if (is_not_null(selectText)){
     
 stmt="Select /* Requery Current Row of "+getCode()+"*/\n"+
         makeMainExpression(lc,false,false,true)+"\n From (\n"+
         makeMainQuery()+")q1 where "+LTrim(selectText," and");
 application.log.addRecord("ROW_REQUERY",application.NLS.NLS("RUS:Перезапрос строки;ENG:Requery for row;"),stmt,this.getFormID());     
 
  cursor=lc.CursorOpen(stmt);
  cols=lc.CursorGetCols(cursor);
  colsInfo=lc.CursorGetColsInfo(cursor);
  if (lc.CursorNextRow(cursor)){
      ArrayList<String> r=lc.CursorGetRow(cursor);
    int qcc=0; //13.05.2021 Вынос order by наружу
     for(int n=0;n<elements.size();n++){
      LuiAppElement   el=elSort(n);
      if (!el.getType().equals("ITEM"))continue;
       if (el.colinfo==null) continue;
       valBuffer=r.get(qcc);
       ldt=el.getProperty("DATA_TYPE");
      if (valBuffer==null){
              if ("EMPTY_STRING".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode) 
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
              }
              }
      if ("".equals(valBuffer)){
              if ("NULL".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
              }
              }      
      row.Add_Element(el.Code,valBuffer); 
      qcc++;
       
     }
 /*//13.05.2021 Вынос order by наружу     
    for (int i=0;i<cols;i++){
    valBuffer=r.get(i);
    if (elements.containsKey(colsInfo.get(i).alias)){
      ldt=elements.get(colsInfo.get(i).alias).getProperty("DATA_TYPE");
      if (valBuffer==null){
              if ("EMPTY_STRING".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode) 
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
              }
              }
      if ("".equals(valBuffer)){
              if ("NULL".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
              }
              }      
    }
      row.Add_Element(colsInfo.get(i).alias,valBuffer);
    } *////13.05.2021 Вынос order by наружу
  }
  lc.CursorClose(cursor);

}
} catch(SQLException ex){
     application.msgWriter.write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+stmt+") ",properties.get("QUERY_TEXT").getGenesis4msg());
 }
 finally{
     if (lc!=null){
     if (cursor!=null) {    
     if (lc.CursorisOpen(cursor)){
         lc.CursorClose(cursor);
     }
     }
     if (ac.getSessionMode()=='C'){
         ac.connPool.SetFree(lc);
     }
     }
 } 
 return row;
}

    /**
     * Формирование коллекции параметров для вызова вызываемой формы
     *
     * @param pActionCode -- код действия
     * @param pParams -- ранее приговленная строковая коллекция параметров
     * @return -- строковая коллекция параметров
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public String getParamString(String pActionCode,String pParams) throws LuiException, SQLException{
    LuiCollection parCollection=new LuiCollection(pParams);
    LuiAppElement el=elements.get(pActionCode);
    String command=el.getProperty("COMMAND");
    if (like(command,"%(%:%;%)")){ //19/05/2020
    //if (el.properties.containsKey("COMMAND") && el.properties.get("COMMAND").isDynamic()){
      int i=command.indexOf("(");
      //int j=command.indexOf(")");
      int j=command.length()-1;
      String cp=command.substring(i+1, j);
      LuiCollection cpc=new LuiCollection(cp);
      for (String pk:cpc.keySet()){
         parCollection.Add_Element(pk,cpc.Get_Element(pk) );
      }
    }
    if (!getType().equals("GRID"))elSort(0);
    for (int i=0;i<elSortedList.size();i++){
       LuiAppElement par=elSortedList.get(i);
       if (par.getType().equals("PARAM") && par.getParent().equals(pActionCode)){
         String cpar=par.getProperty("NAME");
         if (is_not_null(cpar)){
             //if ("MENU".equals(getType())) {
             //  par.properties.get("VALUE").clearCompVal();
             //}
             parCollection.Add_Element(cpar,par.getProperty("VALUE"));
         }
       }
    }
    String retVal=parCollection.toString();
    /*if (like(retVal,"%{%}%")){ 05.03.2020 решили финальный этот парс убрать, т.к. иногда в параметре передаётся текст к-либо программы и он содержит фигурки и если парисить то может получиться drop schema cascade например
    но для чего-то это было сделано ещё в Fastcom'е и если что-то перестанет работать при вызове списко-бланков то обрати сюда внимание
      LuiAppProp p= el.properties.get("COMMAND"); 
      retVal=application.parser.compute(retVal, p);
    }*/
    return retVal;
}

/**
 * Выполнение программы записанной в свойстве pPropCode действия pActionCode
 * @param pActionCode код действия
 * @param pPropCode код свойства
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void exeProc(String pActionCode,String pPropCode) throws LuiException, SQLException{
  LuiAppElement el=elements.get(pActionCode);
    String actionType=el.getProperty("APPLY_TO");
    String tmpValues;
    String curValues="";
    String prg="";
    boolean b=false; // нет зависимости от ключа
    String selectionMode=getProperty("SelectionMode");
    
    if (is_not_null(el.getProperty("DATA_TYPE"))){
            selectionMode="Single";
    }
    if (selectionMode.equals("Single")){
        actionType="N";
    }
    if (like(actionType,"PROGRESS%")){ // проверяем наличие зависимости от столбцов ключа
        for (int i=0;i<elements.size();i++){
            LuiAppElement eli=elSort(i);
            if (eli.getType().equals("ITEM")){
            if (!eli.getProperty("FOR_UNIQUE_KEY").equals("N")){ // берём ключевые столбцы
                
               prg=el.getNonCompValue(pPropCode); // берём не парсенную программу
               if (like(prg,"%{F:"+eli.getCode()+"}%")){ // если в ней есть упоминание ключевого столбца то зависимость есть
                   b=true;
               }
               if (like(prg,"%{F2SQL:"+eli.getCode()+"}%")){ // если в ней есть упоминание ключевого столбца то зависимость есть
                   b=true;
               }
               if (like(prg,"%{Field:"+eli.getCode()+"}%")){ // если в ней есть упоминание ключевого столбца то зависимость есть
                   b=true;
               }
               if (like(prg,"%{Column:"+eli.getCode()+"}%")){ // если в ней есть упоминание ключевого столбца то зависимость есть
                   b=true;
               }
               if (like(prg,"%{Param:"+eli.getCode()+"}%")){ // если в ней есть упоминание ключевого столбца то зависимость есть
                   b=true;
               }
            curValues=curValues+NVL(eli.getProperty("CAPTION"),eli.getCode())+"={F:"+eli.getCode()+"};"; // сочиняем текст для progress
            }
            }    
        }
        if (!b){
          actionType="N"; // если зависимость не выявлена то выполняем обычное действие
        }
        
    }
    if (like(actionType,("PROGRESS%"))){
        curValues=RTrim(curValues,"; ");
        addProperty("ProgressWorkText",curValues);
        LuiCollection tmpVals=getFieldsValues();
        application.log.addRecord("ACTION",
        application.NLS.NLS("RUS:Выполнение программы в;ENG:Execute programm in;")+" "+pActionCode+"."+pPropCode,
        prg,this.getFormID());
        exeProgress(pActionCode, pPropCode,actionType);
        setElementsValues(tmpVals);
        setVal("Mult","Y");
        //application.changeSCN("");//Возможно, в результате действия были изменены свойства элементов
    }else{
        LuiAppProp p=el.properties.get(pPropCode);
        tmpValues=p.getComputedProperty();
       if (is_not_null(tmpValues)){ 
        application.log.addRecord("ACTION",
        application.NLS.NLS("RUS:Выполнение программы в;ENG:Execute programm in;")+" "+pActionCode+"."+pPropCode, tmpValues,this.getFormID());
        application.parser.execute(tmpValues, p);
        //application.changeSCN("");//Возможно, в результате действия были изменены свойства элементов
       }
    }
    
}
/**
 * Выделить все строки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void checkAll() throws LuiException, SQLException{
    application.changeSCN("CheckAll");
    /*if (elements.containsKey("QUERY4PRE-MARKED$")){
      LuiAppElement el=elements.get("QUERY4PRE-MARKED$");
      if (el.getParent().equals("Form")){
          if (el.getType().equals("PARAM")){
              el.setProperty("Value","");
          }
      }
    }*/
    setVal("Inversion","Y");
    clearChecking();
    checkFrmDetails();
    
}
/**
 * РазВыделить все строки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void unCheckAll() throws LuiException, SQLException{
    application.changeSCN("UnCheckAll");
  /*  if (elements.containsKey("QUERY4PRE-MARKED$")){
      LuiAppElement el=elements.get("QUERY4PRE-MARKED$");
      if (el.getParent().equals("Form")){
          if (el.getType().equals("PARAM")){
              el.setProperty("Value","");
          }
      }
    }*/
    setVal("Inversion","N");
    clearChecking();
    
    checkFrmDetails();
}
/**
 * Подсчёт количества выделенных строк
 * @return количество выделенныз строк
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public int getCountMarkedRows() throws LuiException, SQLException{
 if (getType().equals("MENU")) {
     return checkRows.size();
 }
      if ("N".equals(NVL(getVal("Inversion"),"N"))){ //если инверсии нет
       return checkRows.size(); 
    }
if (is_null(getVal("MainQuery"))) return 0;      
String stmt="select count(*) from("+LuiParser.reMask(getVal("MainQuery"))+")q$count4progress";
application.log.addRecord("COUNT_QUERY",application.NLS.NLS("RUS:Подсчёт выделенных строк;ENG:Count marked rows;")
             ,stmt,this.getFormID());
  
 ArrayList<String> acstmt=application.getAppConnAndStmt(getProperty("QUERY_TEXT"), "",this.getFormID());
 LuiAppConn ac=application.userDbConnections.get(acstmt.get(0));
 LuiDbConn lc=null;
 Integer cursor=null;
 
 int total;
 if (ac.getSessionMode()=='C'){
 lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
  }
 else lc=ac.currentConnection;
 if (lc==null)application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
 try {
  total=Integer.valueOf(lc.getSelectResult(stmt));
  total=total-checkRows.size();
  return total;
 }   finally {
     //if (lc!=null) {
       if (cursor!=null){
           if (lc.CursorisOpen(cursor)){
               lc.CursorClose(cursor);
           }
       }
      if (ac.getSessionMode()=='C'){
         ac.connPool.SetFree(lc);
      }
     //}
 }
}

/**
 * Выполнение действия PROGRESS
 * @param pActioncode код действия
 * @param pPropCode код свойства
 * @param pActionType -- PROGRESS или PROGRESSHIDE
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
*/
public void exeProgress(String pActioncode,String pPropCode,String pActionType) throws LuiException, SQLException{
 LuiAppElement el=elements.get(pActioncode);
 LuiClCommandGroup XMLcg;
 LuiClCommand XMLcmd;
 String tmpValues="";
 String cmd="";
 boolean hide=false;
 
 boolean responce=true;
 if (like(pActionType,"PROGRESS_HIDE"))hide=true;
 String title=NVL(el.getProperty("EXPLANATION"),el.getProperty("PROMPT"));;
 title=LuiString.RemoveIco(title);
 if ("N".equals(NVL(getVal("Inversion"),"N"))){ //если инверсии нет
  Integer curcnt=0;
  Integer total=checkRows.size();
  if (!hide){
  responce=application.progressBar.startProgress(title, "Y",curcnt.toString()+"/"+total.toString(), "", total, curcnt); //открываем прогрессбар
  if (!responce) {
      application.progressBar.stopProgress();
      application.actionBreak();
      return; // уходим если пользователь его закрыл
    }
  }
    for (LuiCollection r:checkRows.keySet()){ //крутим выделенные строки
curcnt++;         
addProperty("ProcessingRowNumber",String.valueOf(curcnt));
      setElementsValues(r);
      LuiAppProp p=el.properties.get(pPropCode);
      cmd=p.getComputedProperty();
       if (is_not_null(cmd)){    
        application.parser.execute(cmd, p);   
       }
 if (!hide){
 responce=application.progressBar.continueProgress(title, "Y",curcnt.toString()+"/"+total.toString(),getProperty("ProgressWorkText"),total, curcnt); //продолжаем прогрессбар
 if (!responce){ // уходим если пользователь его закрыл
     application.progressBar.stopProgress();
    application.actionBreak();
    return;
 }
 }
    } //конец кручения выделенныъ строк
    if (!hide){
    application.progressBar.stopProgress();
    }
  } else { //инверсия
  Integer curcnt=0;
  Integer total=checkRows.size();
  if (!hide){
  //открываем прогрессбар
  responce=application.progressBar.startProgress(title, "Y",curcnt.toString()+"/?????", application.NLS.NLS("RUS:Подсчёт количества строк;ENG:Count the number of rows;"), total, curcnt); 
 
  if (!responce) {
      application.progressBar.stopProgress();
      application.actionBreak();
      return; // уходим если пользователь его закрыл
    }
  }
 // вычисляем количество строк select'а   
 total=Integer.valueOf(getProperty("CountSelected"));
 total=total-checkRows.size();
 
 ArrayList<String> acstmt=application.getAppConnAndStmt(getProperty("QUERY_TEXT"), "",this.getFormID());
 LuiAppConn ac=application.userDbConnections.get(acstmt.get(0));
 LuiDbConn lc=null;
 Integer cursor=null;
 if (ac.getSessionMode()=='C'){
 lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
  }
 else lc=ac.currentConnection;
 if (lc==null)application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
 try {
 String stmt=LuiParser.reMask(getVal("MainQuery"));
 cursor=lc.CursorOpen(stmt); // откоываем курсор
 ArrayList<ColInfo> ColsInfo=lc.CursorGetColsInfo(cursor);// иформация о столбцах
 while (lc.CursorNextRow(cursor)){//извлекаем строку
     LuiCollection valcol=new LuiCollection("");
     ArrayList<String> row=lc.CursorGetRow(cursor);
     int qcc=0; //13.05.2021 Вынос order by наружу
     String valBuffer;
     for(int n=0;n<elements.size();n++){
         LuiAppElement colel=elSort(n);
       if (!colel.getType().equals("ITEM"))continue;
       if (colel.colinfo==null) continue;
       String ldt=colel.getProperty("DATA_TYPE"); 
       valBuffer=row.get(qcc); // берём его значение из jdbc
         if (valBuffer==null){
              if ("EMPTY_STRING".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode) 
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
              }
              }
              if ("".equals(valBuffer)){
              if ("NULL".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol)){ //если есть особый режим работы с NULL
                  valBuffer=application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
              }
              }
       valcol.Add_Element(colel.Code,valBuffer);
       qcc++;
       }
    /* for(int i=0;i<ColsInfo.size();i++){
         String ec=ColsInfo.get(i).alias;
         valcol.Add_Element(ec,row.get(i));
     }*/
     setElementsValues(valcol);
     LuiCollection uk=getUK();
     if (!checkRows.containsKey(uk)){
      curcnt++; 
      addProperty("ProcessingRowNumber",String.valueOf(curcnt));
      LuiAppProp p=el.properties.get(pPropCode);
      
      cmd=p.getComputedProperty();
      
       if (is_not_null(cmd)){ 
        application.parser.execute(cmd, p);   
       }
      
   if (!hide){     
        responce=application.progressBar.continueProgress(title, "Y",curcnt.toString()+"/"+total.toString(), getProperty("ProgressWorkText"), total, curcnt);   
 if (!responce) {
     application.progressBar.stopProgress();
     application.actionBreak();
      return; // уходим если пользователь его закрыл
    }
   }     
     }
 }
 if (!hide){
 application.progressBar.stopProgress();
 }
 //конец строк больше нет
 } finally {
     //if (lc!=null) {
       if (cursor!=null){
           if (lc.CursorisOpen(cursor)){
               lc.CursorClose(cursor);
           }
       }
      if (ac.getSessionMode()=='C'){
         ac.connPool.SetFree(lc);
      }
     //}
 }
}
}

/** 
 * Изменение режима выделения строк в списке
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void changeMode() throws LuiException, SQLException{
  application.changeSCN("ChangeMode");
 /* if (elements.containsKey("QUERY4PRE-MARKED$")){
      LuiAppElement el=elements.get("QUERY4PRE-MARKED$");
      if (el.getParent().equals("Form")){
          if (el.getType().equals("PARAM")){
              el.setProperty("Value","");
          }
      }
    }*/
  String currentMode= getProperty("SelectionMode");  
  LuiClCommandGroup XMLcg=application.commandWriter.addGroup("ChangeMode");
  LuiClCommand XMLcmd=XMLcg.addCommand("Grid");
  XMLcmd.addScaPar("FormID",getFormID());
  if (compareString(currentMode,"Multi")){
      setProperty("SelectionMode","Single");
      setVal("Inversion","N");
      clearChecking();
      if (Integer.valueOf(NVL(getVal("CurrentRow"),"0"))>0){
          checkRow("Y",getUK());
      }
  }
  if (compareString(currentMode,"Single")){
      setProperty("SelectionMode","Multi");
      clearChecking();
  }
  addCommandParam(XMLcmd, null,"SelectionMode", "SelectionMode",true);
  checkFrmDetails();
}
/**
 * Получение итоговой строки списка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void getTotal() throws LuiException, SQLException{
application.changeSCN("GetTotal");
String queryText="";
    setVal("TotalStatus","Executed");
    for (int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.getType().equals("ITEM")){
          String totalExpr=el.getProperty("TOTAL");
          if (is_not_null(totalExpr)){
           String format=el.getProperty("FORMAT");
           if (is_not_null(format)){
               totalExpr="to_char("+totalExpr+",'"+format+"')";
           }
           //queryText=queryText+", "+totalExpr+" \""+el.getCode()+"\" \n"; //13.05.2021 Вынос order by наружу
           queryText=queryText+", "+totalExpr+" \n"; //13.05.2021 Вынос order by наружу
          }
        }
    }
    if (is_not_null(queryText)){
        queryText=LTrim(queryText,",");
        queryText="select /*Total Row*/\n"+queryText+" from (\n"+makeMainQuery()+"\n)q$total ";
        String accode=getProperty("Connection");
        if (is_null(accode))return;
        if (!application.userDbConnections.containsKey(accode))return;
        LuiDbConn lc=null;
        LuiAppConn ac=application.userDbConnections.get(accode);
        try{
        if (ac.getSessionMode()=='C') {
            lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
        } else {
            lc=ac.currentConnection;
        }
        queryText=queryText+addWhere(lc,getProperty("inner_qbe"),false);
        application.log.addRecord("TOTALS_QUERY", 
        application.NLS.NLS("RUS:Запрос итогов;ENG:Totals query;"), queryText,this.getFormID());
        //13.05.2021 Вынос order by наружу
        ArrayList<String> totArray=lc.getSelectFirstRowValList(queryText);
        if (totArray!=null) {
        LuiClCommandGroup XMLcg=application.commandWriter.addGroup("ListTotal");
        LuiClCommand XMLcmd;
        XMLcg.addParameter("FormID",getFormID());    
        int qcc=0;
        for (int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.getType().equals("ITEM")){
          String totalExpr=el.getProperty("TOTAL");
          if (is_not_null(totalExpr)){
          XMLcmd=XMLcg.addCommand("TotalCell");    
          XMLcmd.addScaParameter("Code",el.Code,"");
          XMLcmd.addScaParameter("Show",totArray.get(qcc),"");
          qcc++;    
          }
        }
        }
        } 
        /*//13.05.2021 Вынос order by наружу
        HashMap<String,String> tot=lc.getSelectFirstRowValMap(queryText);
        if (tot!=null){
        LuiClCommandGroup XMLcg=application.commandWriter.addGroup("ListTotal");
        LuiClCommand XMLcmd;
        XMLcg.addParameter("FormID",getFormID());
        for (String key:tot.keySet()){
        XMLcmd=XMLcg.addCommand("TotalCell");    
        XMLcmd.addScaParameter("Code",key,"");
        XMLcmd.addScaParameter("Show",tot.get(key),"");
        }}
        */
        } 
        catch (SQLException ex){
         application.msgWriter.sqlError(lc, ex, this.properties.get("QUERY_TEXT"), queryText);
        }
        
        finally {
            if (ac.getSessionMode()=='C'){
              ac.connPool.SetFree(lc);
            } 
        }
    }
}
/**
 * Попытка блокировки строки списка при выполнении действия
 * @param isGenEx -- следует ли генерировать сообщение о занятости ресурса
 * @param pLock -- значение св-ва LOCK действия
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void locking(boolean isGenEx,String pLock) throws LuiException, SQLException{
    ArrayList<ColInfo> ukcf=new ArrayList<>();
    String stmt="";
    if (!getVal("IsNeedLock").equals("Y")){
      setVal("IsNeedLock",""); // снимаем признак необходимости блокировки   
      return;
    }
    LuiAppConn appcon=null;
      String appcon_code=getProperty("Connection");
      if (application.userDbConnections.containsKey(appcon_code)){
          appcon=application.userDbConnections.get(appcon_code);
      }
      if (appcon==null) {
          for (String acc:application.userDbConnections.keySet()){
             if (!("LUI$".equals(acc))) { 
              appcon=application.userDbConnections.get(acc);
              break;
             }
          }
      }
     if (appcon==null) { // если нет соединения
         setVal("IsNeedLock",""); // снимаем признак необходимости блокировки   
      return; // и уходим
     }
      LuiDbConn c=null;
     try{
     if (appcon.getSessionMode()=='C'){
         c=appcon.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
     } else {
         c=appcon.currentConnection;
     }

      for (int i=0;i<elements.size();i++){
          LuiAppElement el;
          el=elSort(i);
          if (el.getType().equals("ITEM")){
            String fuk=el.getProperty("FOR_UNIQUE_KEY");
            if (!fuk.equals("N")){
               if (!like(fuk,"%_._%")) {
                fuk=el.getProperty("DEFINITION");
               }
              if (is_not_null(fuk))  {
                el.colinfo.LuiDATATYPE=el.getProperty("DATA_TYPE");
                el.colinfo.LuiFORMAT=el.getProperty("FORMAT");
                String wstr=application.dataConverter.getFromLuiConversion(el.getProperty("Value"), el, c);
                if (wstr.equals("null")) {
                    wstr="is "+wstr;
                } else {
                    wstr="="+wstr;
                }
                //el.colinfo.value=el.getProperty("Value");
                el.colinfo.value=wstr;
                ukcf.add(el.colinfo);
                
                  if (like(fuk,"%_._%")) {
                    if (is_null(el.colinfo.tableAlias)){  
                      el.colinfo.tableAlias=LuiString.Extract_Word(fuk, 1, ".");
                    }
                    if (is_null(el.colinfo.name)){
                      el.colinfo.name=LuiString.Extract_Word(fuk, 2, ".");
                    }
                  }  
                
              }
            }
          }    
      }
    
   if (ukcf.size()==0) {
    setVal("IsNeedLock",""); // снимаем признак необходимости блокировки  
    return;
   }
   
     ArrayList<String> s4u=new ArrayList<>();
     if (like(pLock.toUpperCase(),"%SELECT%")||
             like(pLock.toUpperCase(),"%UPDATE%")||
             like(pLock.toUpperCase(),"%BEGIN%")){
      s4u.add(pLock);
      }  else {
      s4u=c.select4update(makeMainQuery(), ukcf);
     }
     ArrayList<String> ac=new ArrayList<String>();
     ac.add("success");
     for (int sn=0;sn<s4u.size();sn++){
      stmt=s4u.get(sn);   
      application.log.addRecord("ROW_LOCK",
              application.NLS.NLS("RUS:Блокировка строки;ENG:Row lock;")+"\n(" +getUK()+")"
                    ,stmt,this.getFormID());
      ac=c.executeIfNoLocks(stmt);
      if (!ac.get(0).equals("success")) {
          break;
      }
     }
     if (ac.get(0).equals("success")){
       setVal("IsNeedLock",""); // снимаем признак необходимости блокировки
     } else {
         if (isGenEx){ // 
             application.msgWriter.write("LUI-00028",ac.get(1)); // это вызовет exception
         } else {
             application.msgWriter.write("LUI-00029",ac.get(1));// а это - нет
         }
     }
     } 
     catch (SQLException sqlex){
         application.msgWriter.write("LUI-00038",this.code,sqlex.getMessage(),stmt);
     }
     finally {
       if (c!=null){
          if (appcon!=null){
              if (appcon.getSessionMode()=='C'){
                  appcon.connPool.SetFree(c);
              }
          }
       }
   }
         
    
   
}
/**
 * 
 * @param pDimCode -- код измерения
 * @param lc -- cоединение
 * @return  нечто
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
/* //13.05.2021 Вынос order by наружу
public String makeMainExpression4ChartDimension(String pDimCode,LuiDbConn lc) throws LuiException, SQLException{
 String clause="";  
 int cnt=0;
 for(int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.getParent().equals(pDimCode)){
        if (el.colinfo!=null) {
        if (el.colinfo.alias!=null) {    
        //if (el.getType().equals("ITEM")||el.getType().equals("FIELD")||(el.getType().equals("PARAM")&& getType().equals("BLANK"))){
            
            if (cnt>0){
            clause=clause+",";
            }   
            cnt++;
            clause=clause+application.dataConverter.getColumnSelectClause(el, lc);
        }
    }
    }
 }
return clause+" "; 
}
*/
/* //13.05.2021 Вынос order by наружу
public String makeMainExpression4Chart1(LuiDbConn lc,String pParent,LuiAppProp pProp) throws LuiException, SQLException{
   String clause="";
   int cnt=0;
       for(int i=0;i<elements.size();i++){
           LuiAppElement el=elSort(i);
           if (pParent.equals("Form") || el.getParent().equals(pParent)){
           if (el.colinfo!=null) {
             if (el.colinfo.alias!=null) {    
                 if (cnt>0){
            clause=clause+",";
            }   
            cnt++;
            clause=clause+application.dataConverter.getColumnSelectClause(el, lc);
             }  
           }
       }
       }     
   return clause;
}*/
/**
 * Изготовление Select-фразы основного зароса списка
 * @param lc соединение с базой данных
     * @param forRT -- если надо отобрать столбцы пригоднве для real-time диаграмм
     * @param alias -- снабжать ли выражения последующими псевдонимами
     * @param format -- учиьываьб ли форматы 
 * @return Select-фраза
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String makeMainExpression(LuiDbConn lc,boolean forRT,boolean alias,boolean format) throws LuiException, SQLException{
String clause="";
int cnt=0;

    for(int i=0;i<elements.size();i++){
        
        LuiAppElement el=elSort(i);
        if (el.Code.equals(getVal("RowNoEl")))continue; // костыль МВД нумерации и отсечки строк
        if (el.getProperty("Sintetic").equals("Y")) {
            
               if (el.colinfo==null) {
                   el.colinfo=new ColInfo();
               }
                   el.colinfo.LuiDATATYPE=el.getProperty("DATA_TYPE");
                   el.colinfo.LuiFORMAT=el.getProperty("FORMAT");
                   el.colinfo.alias=el.Code;
               
               if (cnt>0){clause=clause+",";}    
               cnt++;  
               if ((alias || el.isWinFunc()) && !el.getProperty("BUT_NOT_MAKE_ALIAS").equals("Y"))
               clause=clause+application.dataConverter.getToLuiConversion(el,lc,"Y",format)+" \""+el.Code+"\""; //13.05.2021 Вынос order by наружу
               else
               clause=clause+application.dataConverter.getToLuiConversion(el,lc,"Y",format);
               //clause=clause+el.getProperty("DEFINITION") +" \""+el.Code+"\"";    
        } else {
        if (el.colinfo!=null) {
        if (el.colinfo.alias!=null) { 
        if (!forRT 
                || "Y".equals(el.getVal("rtEnable")) 
                || "Y".equals(el.getProperty("FOR_UNIQUE_KEY"))) {    
            if (cnt>0){
            clause=clause+",";
            }   
            cnt++;
            el.colinfo.LuiFORMAT=el.getProperty("FORMAT");
            clause=clause+application.dataConverter.getColumnSelectClause(el, lc,format);
            if ((alias || el.isWinFunc()) && !el.getProperty("BUT_NOT_MAKE_ALIAS").equals("Y")) clause=clause+" \""+el.Code+"\"";
        }
        }
        } 
        }
        
    }
    
return clause+" ";
}
/**
 * Построение запроса для диаграммы
     * @param pParent -- код элемента в котором лежит запрос
     * @param pProp-- свойство элемента в котором лежит запрос
 * @return текст запроса
     * @throws LuiExceptionMsg.LuiException
     * @throws java.sql.SQLException
 */
/*public String makeQuery4Chart(String pParent,LuiAppProp pProp) throws LuiException, SQLException{
    String query=" ";  
  ArrayList<String> acStmt;
  int cnt=0;
  // готовим стобцы фразы select
  for (int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.getType().equals("ITEM")) {
         if (pParent.equals("Form") || el.getParent().equals(pParent)){
          if (cnt>0){
           query=query+"  ,";
           }  
        query=query+NVL(el.getProperty("DEFINITION"),"null")+" \""+el.getCode()+"\" \n";
        cnt++;  
        }}
  }
  acStmt=application.getAppConnAndStmt(pProp.getComputedProperty(), "",this.getFormID());
  String qt=acStmt.get(1);
  
  if (pParent.equals("Form")) {
      addProperty("Connection",acStmt.get(0));
  } else {
  elements.get(pParent).addProperty("Connection",acStmt.get(0));
  }
  try {
 Matcher M=Pattern.compile("select([\\n ]+\\*[\\n ]+)", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt);
 if (M.find()){
   query=qt.substring(0, M.start(1))+query+qt.substring(M.end(1));  
 }
 //query=Pattern.compile("(select)[\n ]+\\*[\   n ]+", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt).replaceFirst("$1"+query);
} catch (Exception ex){
qt=qt.replace(" *\n", " * ");
query=qt.replace(" * ", query);
//query=qt.replaceFirst(" \\* ", query);
}
  return query;
}*/


/**
 * Изготовление основного запроса списка (без учёта QBE)
 * @return  основной запрос списка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String makeMainQuery() throws LuiException, SQLException{
  String query=" ";  
  ArrayList<String> acStmt;
  int cnt=0;
  // готовим стобцы фразы select
    for (int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.Code.equals(getVal("RowNoEl")))continue;// костыль МВД нумерации и отсечки строк
        if (el.getType().equals("ITEM")) {
        
        String defini=NVL(el.getProperty("DEFINITION"),"null");
        // синтетические столбцы
        for (String ec:elements.keySet()){
          if (ec.equals(el.Code)||!elements.get(ec).getType().equals("ITEM"))continue;
          if (defini.indexOf("\""+ec+"\"")>-1){
           el.addProperty("Sintetic", "Y");
           break;
          }
        }
        if (!el.getProperty("Sintetic").equals("Y")){ // если не синтетический
        if (cnt>0){
        query=query+"  ,";
        }
        query=query+defini+" \""+el.getCode()+"\" \n";
        
        cnt++;
        } 
        }
    }
      acStmt=application.getAppConnAndStmt(getProperty("QUERY_TEXT"), "",this.getFormID());
      String qt=acStmt.get(1);
      setProperty("Connection",acStmt.get(0));
      //query=qt.replaceFirst(" \\* ", query);
try {
 //Matcher M=Pattern.compile("select([\\n ]+\\*[\\n ]+)", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt);
 Matcher M=Pattern.compile("select([\\n ]+\\*(\\n| |$)+)", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt);
 if (M.find()){
   query=qt.substring(0, M.start(1))+query+qt.substring(M.end(1));  
 }
 //query=Pattern.compile("(select)[\n ]+\\*[\   n ]+", Pattern.CASE_INSENSITIVE+Pattern.MULTILINE).matcher(qt).replaceFirst("$1"+query);
} catch (Exception ex){
qt=qt.replace(" *\n", " * ");
query=qt.replace(" * ", query);
//query=qt.replaceFirst(" \\* ", query);
}
      
return query;
}
/** 
 * Выполняет savepoit в начале действия вызываемого в форме
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void savepoint() throws LuiException, SQLException{
 String ac_code=getProperty("Connection");
 LuiAppConn ac=null;
 if (is_not_null(ac_code)){
     if (application.userDbConnections.containsKey(ac_code)){
      ac=application.userDbConnections.get(ac_code);
     }
 }
 if (ac==null){
     for(String acc:application.userDbConnections.keySet()){
            if (!("LUI$".equals(acc))) {
         ac=application.userDbConnections.get(acc);
         break;
            }
     }
 }
 if (ac!=null){
     if (ac.getSessionMode()!='C'){
       if (is_not_null(application.getVal("Global Transaction"))){
         ac.currentConnection.savepoint("\""+this.getFormID()+"\"");
         setVal("Transaction","savepoint");
       } else{
         setVal("Transaction","rollback");  
       }
       application.setVal("Global Transaction", "Y");
     } else {
         
     }
 }
 
}
/**
 * Откат действия к его началу
 * @return произошёл откат или нет - в холостую
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean rollback() throws LuiException, SQLException{
    
String ac_code=getProperty("Connection");
 LuiAppConn ac=null;
 if (is_not_null(ac_code)){
     if (application.userDbConnections.containsKey(ac_code)){
      ac=application.userDbConnections.get(ac_code);
     }
 }
 if (ac==null){
     for(String acc:application.userDbConnections.keySet()){
            if (!("LUI$".equals(acc))) {
         ac=application.userDbConnections.get(acc);
         break;}
     }
 }
 if (ac!=null){
     if (ac.getSessionMode()!='C'){    
      LuiDbConn c=ac.currentConnection;
    switch (getVal("Transaction")){
        case "savepoint":
         if (!c.rollbackToSavepoint("\""+getFormID()+"\"")){ // точка сохранения не найденв
          c.rollback();
          endAction();
          if (!c.getDBvendor().equals("PG")) // если это не PG!!! PG если случилась ошибка теряет все savepoint's
          return false;
         }
        break;
        case "rollback":
          c.rollback();
        break;
    }
     }
 } 
    endAction();
    String tcm=getVal("CommitMode");
  if (is_not_null(tcm)){
      application.setProperty("COMMIT",tcm);
  }
return true;  
}

/**
 * Не знаю зачем
 */
private void endAction(){
    if (getVal("Transaction").equals("rollback")){
        application.setVal("Global Transaction", "");
    }
    setVal("Transaction","");
    setVal("IsNeedLock","");
}
/**
 * Вычисляет String-коллекцию уникального ключа списка (столбец1:значение1;столбец2:значение2;.:..;
 * @return коллекция уникального ключа списка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiCollection getUK() throws LuiException, SQLException{
String uk=getVal("UK");
LuiCollection ukc=null;
 if (is_not_null(uk)){
     ukc=new LuiCollection(uk);
     for (String key:ukc.keySet()){
       ukc.Add_Element(key,elements.get(key).getProperty("Value"));
     }
 }else {
     ukc=new LuiCollection("");
   for(int i=0;i<elements.size();i++){
       LuiAppElement el=elSort(i);
       if (el.getType().equals("ITEM")){
        if (!el.getProperty("FOR_UNIQUE_KEY").equals("N")) {
         ukc.Add_Element(el.getCode(),el.getProperty("Value"));
         //ukc.Add_Element(el.getCode(),NVL(el.getProperty("Value"),""));
        }
       }
   }  
 }
 if (ukc==null)return new LuiCollection("");
 return ukc;
}
/**
 * Выполняет проверку качества введённых полей бланка
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void checkFields() throws LuiException, SQLException{
    LuiDbConn c=null;
    LuiAppConn ac=null;
    String ac_code="";
    String fieldValue;
    String fieldFormat;
    String fieldDatatype;
    String AorF;
    try{
    for (int i=0;i<elements.size();i++){
       LuiAppElement el=elSort(i);
       if (el.getType().equals("ITEM") || el.getType().equals("FIELD")){// поле или столбец
          
         if (el.isVisible(true)){// видимое
          if(el.isFieldChangeable()){//изменяемое
             fieldValue=el.getProperty("Value"); // берём значение
           if (el.getProperty("OPTIONAL").equals("N")) { // если не может быть пусто
//             if(is_null(el.fieldCurrentShow())) {// ... и отображаемое значение пусто, то
             if(is_null(el.getFieldShow())) {// ... и отображаемое значение пусто, то
              application.msgWriter.write("LUI-00026",el.getName4show(),el.getCode() );
             }  
           }
           if (!el.getProperty("CHANGING").equals("C")){//-- Возможен свободный ввод пользователем -- ITEM_UPD
           fieldDatatype=el.getProperty("DATA_TYPE");    
           AorF=application.dataConverter.LuiDataTypes.get(fieldDatatype).AlphabetOrFormat;
           
           fieldFormat=el.getProperty("FORMAT");       
               if (AorF.equals("A")) {
                fieldValue=el.checkAlpha(fieldFormat, fieldValue, el.getName4show());
               }
           if (is_not_null(fieldValue) ){ // если не пусто 
               if (!AorF.equals("N")){
               // причёсываем под формат и конвертируем данные
               if (c==null){//если соединение в этом цикле ещё не вычислено вычислим его
                    ac_code=getProperty("Connection");//берём код коннекта юзера из формы
                    if (is_not_null(ac_code)){ // если он есть
                      if (application.userDbConnections.containsKey(ac_code)){// таки на самом деле есть
                          ac=application.userDbConnections.get(ac_code); //берём его
                       }
                    }
                    if (ac==null){// если не нашли у формы возьмём любой
                      for (String acc:application.userDbConnections.keySet()){
                          ac_code=acc;
                          ac=application.userDbConnections.get(acc);
                         }
                    }
                    if (ac!=null){ // если что-то нашли
                        if (ac.getSessionMode()=='C'){// и оно коммунальное
                            c=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // берём из его пула свободное
                        } else {
                            c=ac.currentConnection; // если не коммунальное - берём текущее
                        }

                    } else { // если ни чего пользовательского не нашли
                        c=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());// возспользуемся соединением к метаданным
                    }
                 }
              try{
                 //fieldValue=c.StringToStringByFormat(fieldValue, fieldFormat, fieldDatatype);
                 /*fieldValue=*/application.dataConverter.convAndCheck(fieldValue, el, c);
                 } catch(Exception ex){
                    application.msgWriter.write("LUI-00027",el.getName4show(),fieldDatatype,el.getProperty("FORMAT"));
                 } 
               } //AorF!=N 
             // конец причёсывания под формат и конвертирования
             // если есть серверный верификатор 
      if (el.properties.containsKey("VERIFIER_SRV_PRG")){
         LuiAppProp p=el.properties.get("VERIFIER_SRV_PRG");
         p.recalculate=true;
         String mescode=p.getComputedProperty();
         if (is_not_null(mescode)) {
             application.msgWriter.write(mescode,NVL(el.getProperty("CAPTION"),el.getProperty("Name")));
         }
      }
           }// если значение не пусто
           
           } 
          }
         }  
       }
    }
    } finally {
    if (c!=null){
        if (ac!=null){
            if (ac.getSessionMode()=='C'){
                ac.connPool.SetFree(c);
            }
        }else{
            application.metaConnPool.SetFree(c);
        }
    }
}
}
/**
 * Вычисляет коллекцию полей и их значений
 * @return коллекция полей и их значений
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiCollection getFieldsValues() throws LuiException, SQLException{
    LuiCollection fv=new LuiCollection("");
    for (int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.getType().equals("FIELD") || el.getType().equals("ITEM"))
            fv.Add_Element(el.getCode(),el.getValue());
    }
    return fv;
}
/**
 * Вызывается при изменении текущего поля бланка FormID на FieldCode
 * @param pFieldCode -- новое текущее поле
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setField(String pFieldCode/*,String pCursor*/) throws LuiException, SQLException{
application.changeSCN("GotoField");
if (elements.containsKey(pFieldCode)){
    /*if (is_not_null(pCursor)){
        elements.get(pFieldCode).setProperty("StartCursor",pCursor);
        elements.get(pFieldCode).setProperty("EndCursor",pCursor);
    }*/
    lastCurrentField=getVal("CurrentField");
    
    if (elements.get(pFieldCode).isVisible(true)) {
        setVal("CurrentField",pFieldCode);
    }
    else {
        setVal("CurrentField","");
    }
}
}
/**
 * Возвращает true если нет отмеченных строк
 * @return признак отсутсвия отмеченных строк
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean isNoSelectedRows() throws LuiException, SQLException{
      if (getVal("Inversion").equals("Y")){
        if (getVal("IsLastPage").equals("N")){
            return false;
        } else {
            Integer cnt=Integer.valueOf(NVL(getVal("FetchedRows"),"0"))-this.checkRows.size();
                    //Integer.valueOf(getProperty("CountSelected"));
            return (cnt==0);
        }
            
    } else {
        String p=getProperty("CountSelected");
        Integer cnt=Integer.valueOf(NVL(getProperty("CountSelected"),"0"));
        return (cnt==0);
    }
}

/**
 * Установка всех свойств всех элементов в состояние изменённости
 */ 
private void setAllChangeFlag(boolean pFlag)
{
    for (String prcode:properties.keySet()){
        properties.get(prcode).changeFlag=pFlag;
    }
    for (String elcode:elements.keySet()){
       for (String prcode:elements.get(elcode).properties.keySet()){
        elements.get(elcode).properties.get(prcode).changeFlag=pFlag;
                                              }
       elements.get(elcode).changeFlag=pFlag;
    }
}
        
/**
 * заполнение дефолтных полей данными из запроса
 */
private void queryToBlank() throws LuiException, SQLException{
  String qt=getProperty("QUERY_TEXT");
  String stmt;
  String cc;
  LuiDbConn lc=null;
  LuiAppConn ac;
  Integer cursor=null;
  if (qt==null) return;
  if (qt.equals(""))return;
  ArrayList<String> al=application.getAppConnAndStmt(qt,"",this.getFormID());
  cc=al.get(0);
  stmt=al.get(1);
  
  ac=application.userDbConnections.get(cc);
  try{
  if (ac.getSessionMode()=='C') { // если соединение коммунальное, то получим из пула свободное
            lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
        }
        else {//в противном случае возьмём текущее 
            lc=ac.currentConnection;
        }
  if (lc==null){ // если соединение так и не нашлось, то
        application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!");
  }
  try{      
  setColsInfo(lc,stmt,"QUERY_TEXT");  
  stmt="select "+makeMainExpression(lc,false,true,true)+"from ("+stmt+")q$1";
  application.log.addRecord("MAIN_QUERY",application.NLS.NLS("RUS:Выполнение запроса;ENG:Execute query;")
             ,stmt,this.getFormID());
  cursor=lc.CursorOpen(stmt);
  } catch (SQLException ex){
      application.msgWriter.write("LUI-00034","State:"+ex.getSQLState()+"\n"+ex.getLocalizedMessage()+" ("+stmt+") ",properties.get("QUERY_TEXT").getGenesis4msg());
  }
  if (cursor==null)return;
  ArrayList<ColInfo> colsinfo;
  colsinfo = lc.CursorGetColsInfo(cursor); 
  int rowcount=0;
  boolean b=false;
  try{
  b=lc.CursorNextRow(cursor);// извлекаем первую строку
  }  catch (SQLException sex){
      application.msgWriter.sqlError(lc, sex, properties.get("QUERY_TEXT"), stmt);
  }
  if (b){ //если извлеклась
   for (int i=0;i<lc.CursorGetCols(cursor);i++){ //цикл по столбцам
     String elCode=colsinfo.get(i).alias.toUpperCase();//код элемента
         String elVal=lc.CursorGetRow(cursor).get(i); // значение
      if (elements.containsKey(elCode)) { // если есть элемент с таким кодом
          if (elements.get(elCode).getType().equals("FIELD")) { // если это поле
              String ldt=elements.get(elCode).getProperty("DATA_TYPE");
              if (elVal==null){
              if ("EMPTY_STRING".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode) 
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol)){ //если есть особый режим работы с NULL
                  elVal=application.dataConverter.LuiDataTypes.get(ldt).nullSpecialSymbol;
              }
              }
              if ("".equals(elVal)){
              if ("NULL".equals(application.dataConverter.LuiDataTypes.get(ldt).emptyStringFromLuiMode)
                      && is_not_null(application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol)){ //если есть особый режим работы с NULL
                  elVal=application.dataConverter.LuiDataTypes.get(ldt).emptyStringSpecialSymbol;
              }
              }
              elements.get(elCode).addProperty("DEFINITION",LuiParser.mask(elVal)); // записываем значение столбца запроса в свойство DEFINITION
              elements.get(elCode).setProperty("Diff","N"); // пометим элемент как имеющий одно значение для всех строк запроса
          }
          if (elements.get(elCode).getType().equals("PARAM")) { // если это параметр
              elements.get(elCode).addProperty("Value",elVal); // записываем значение столбца запроса в свойство Value
          }
      } else { // если элемента с таким кодом нет - создаём параметр и записываем туда value
     //     LuiAppElement newel=addElement(elCode, "PARAM", null, "Virtual parameter","Form");
     //     newel.addProperty("Value", elVal);
       // создание параметра перенесено в setColsInfo
      }
   }
   boolean B=true;
   while (lc.CursorNextRow(cursor)){ // извлекаем остальные строки
       B=true;
       for (int i=0;i<lc.CursorGetCols(cursor);i++){ //цикл по столбцам
       String elCode=colsinfo.get(i).alias.toUpperCase();//код элемента
       String elVal=lc.CursorGetRow(cursor).get(i); // значение    
       if (elements.containsKey(elCode)) { // если есть элемент с таким кодом
        if (elements.get(elCode).getType().equals("FIELD")) { // если это поле  
         /*String fieldFormat=elements.get(elCode).getProperty("FORMAT");
              if (is_not_null(fieldFormat)) { // если формат не пуст пытаемся преобразовать данные по этому формату
                try{  
                 elVal=lc.fromRStoStringByFormat(elVal,colsinfo.get(i), fieldFormat);
                  } catch(SQLException ex){
                      //Ошибка применения формата "<p1>" к значеню "<p2>" поля "<p3>" бланка "<p4>" конфигурации "<p5>" при выполнении основного запроса
                   application.msgWriter.write("LUI-00024",fieldFormat,elVal,elCode,getCode(),getConfigCode());
                  }
              }   */
         if (NVL(elements.get(elCode).getProperty("Diff"),"N").equals("N")){ // если поле ещё не помечено как поле имеющее различие по строкам
             B=false;
             if (!LuiString.compareString(elVal,elements.get(elCode).getProperty("DEFINITION"))) { // и если хранимое значение и новое извлечённое не совпадают
                 elements.get(elCode).addProperty("DEFINITION",""); // записываем значение "" в свойство DEFINITION
                 elements.get(elCode).setProperty("Diff","Y"); // пометим элемент как имеющий разные значения для разных строк запроса
             }
         } 
        } 
       }
       }
    if (B) break; //если уже все поля помечены diff - уходим из просмотра строк       
   } // конец цикла извлечения остальных строк
  } else { // запрос не извлёк ни одной строки
             //а вдруг это перезагрузка формы и во второй раз ничего не извлеклось а раньше извлеклось
      for (int i=0;i<lc.CursorGetCols(cursor);i++){ //цикл по столбцам
          String elCode=colsinfo.get(i).alias.toUpperCase();//код элемента
          if (elements.containsKey(elCode)) { // если есть элемент с таким кодом
              if (elements.get(elCode).getType().equals("PARAM")) { // если это параметр
                  if (elements.get(elCode).properties.containsKey("Value")) {
                     elements.get(elCode).properties.get("Value").clearCompVal(); // выполним сброс вычисленного значения параметра
                  }
                     // при этом у виртуальных параметров current_value свойства Value должен стать ""
              }
          }
      }
  }
  
  
  } finally {
   if (lc!=null){   // если есть соединение
    if (cursor!=null) lc.CursorClose(cursor); //закрываем курсор если он был открыт
    if (ac.getSessionMode()=='C')ac.connPool.SetFree(lc);  // если соединение коммунальное, то вернём соединение в пул свободных
   }
  }  
}
/**
 * Предобработка, специфическая для бланков (недописано)
 */
private void prepareBlank() throws LuiException, SQLException{

HashMap<String, LuiAppElement> fields=new HashMap<>();
boolean b=true;

while (b){
 b=false;
  for(String elCode:elements.keySet()){ // цикл элементов для инфекций полей действиями
   if (fields.containsKey(elCode)) continue;   
    LuiAppElement el=elements.get(elCode);
      if (el.getType().equals("FIELD")) { //если поле
      el.properties.remove("Buttons");
      LuiAppProp pSc=el.addProperty("StartCursor","0");
      LuiAppProp pEc=el.addProperty("EndCursor","0");
      el.addProperty("ValueWithCursor","{JS:/*NHJS*/\"{Property:Value}\".replace(/([\\s\\S]{{Property:EndCursor}})/, \"$1¦\")}");
      // занесём его в массив уже обработанных
      fields.put(elCode, el); 
          // Перенос выражения по умолчанию в Value с целью последующего вычисления
          //String v=el.getNonCompValue("DEFINITION"); 02/11/2018
          //elements.get(elCode).addProperty("Value",v);02/11/2018
          el.properties.get("DEFINITION").clearCompVal(); //02/11/2018
          
  //добавление шаблонных действий из соответствующего шаблона формы
  b=prepareTmplAct4Item(elCode);
  if (b) { // если добавление выполнено прервём цикл осмотра полей и начнём его заново
      break; // это нужно для правильной последовательности расположения действий в массивах элементов
  } 
  }
}
}  
/*02/11/2018
  for(String elCode:fields.keySet()){ // второй цикл элементов для вычисления Value
          String v=LuiParser.mask(fields.get(elCode).getProperty("Value"));
          fields.get(elCode).properties.get("Value").currentValue=v;
  }

  //чистим зависимости 
  for(String elCode:fields.keySet()){ 
          elements.get(elCode).properties.get("Value").depends.clear();
  }
*///02/11/2018  
 // Ещё один цикл для делания некоторых полей не динамическими и для многоязычных полей
 for(String elCode:fields.keySet()){ 
     LuiAppElement el=fields.get(elCode);
     //playRenovation(el);*/02/11/2018
       // далее идёт обработка многоязычия
       //if (is_null(el.getProperty("NLS"))){ // если ещё нет никаких NLS-св-в у этого поля
       
         if (el.getProperty("DATA_TYPE").equals("NLSTEXT")){// Только для многоязычных полей  
           //String Val=NVL(el.getProperty("Value"),"");*/02/11/2018
           String Val=NVL(el.getProperty("DEFINITION"),"");//02/11/2018

            LuiCollection lcol= new LuiCollection(Val);
            if (lcol.size()==0){
                lcol=new LuiCollection("");
                lcol.Add_Element(application.NLS.currentLang.code, application.NLS.NLS(Val));
                el.setValue(lcol.toString(), false,"N",false);
            }
          if (is_null(Val)){ //Если значение многоязычного поля - пусто
               lcol= new LuiCollection("");
               lcol.Add_Element(application.NLS.currentLang.code,"");
               //el.setValue(lcol.toString(), false,"N");02/11/2018
               Val=lcol.toString(); //02/11/2018
               el.addProperty("DEFINITION", Val);
           }//Если значение многоязычного поля - пусто
          
          if (!elements.containsKey(application.NLS.currentLang.code+"-"+el.getCode())){ // если языкового поля ещё нет
          //if (!elements.containsKey("G"+"-"+el.getCode())){ // если группы нет
            // Создаём группу если языков >1
            String gvi=el.getNonCompValue("VISIBLE");
            if (application.NLS.langList.size()>1){
                 this.addGroup("G"+"-"+el.getCode(), 
                   el.getNonCompValue("CAPTION"),
                   gvi,
                   el.sourceSeq,
                   el.getNonCompValue("NOTES"),
                   "SIDE_BY_SIDE",
                   el.getNonCompValue("Parent"),
                   "UP");
            }     
           String ItemValues="";
           String UserChanged="";
           double lseq=0;
           for(String langCode:application.NLS.langList.keySet()){ // крутим языки
              LuiLang lang=application.NLS.langList.get(langCode);
               lseq++;
            //Для каждого языка порождаем новое поле - КОД-ЯЗЫК   
            //pCode - -код элемента 
            //pType - -тип элемента (!!!если в бланк будут добавляться элементы типа ITEM, то они превратятся в FIELD) 
            //pSeq - - порядковый номер 
            //pName - - название
           
            LuiAppElement lel;
            if (application.NLS.langList.size()>1){
               lel = addElement(el.getCode()+"-"+langCode,"FIELD",lseq,application.NLS.NLS(lang.name),"G"+"-"+el.getCode());
               lel.addProperty("CAPTION", application.NLS.NLS(lang.name));
               lel.addProperty("DEFINITION", lcol.Get_Element(langCode));
               
            } else {
               lel = addElement(el.getCode()+"-"+langCode,"FIELD",el.sourceSeq,el.getNonCompValue("CAPTION"),el.getNonCompValue("Parent")); 
               lel.sourceSeq=el.sourceSeq;
               lel.mutSeq=el.mutSeq;
               lel.addProperty("CAPTION",el.getNonCompValue("CAPTION"));
               lel.addProperty("Seq", el.getNonCompValue("Seq"));
               lel.addProperty("DEFINITION", lcol.Get_Element(langCode));
            }
               
            lel.addProperty("ForNLSSupport",el.getCode());
            lel.setValue(application.NLS.NLS(Val,langCode),false, "N",false);
               lel.addProperty("CHANGING", el.getNonCompValue("CHANGING"));
               lel.addProperty("VISIBLE","{JS:/;"+langCode+":/.test(';{F:"+el.getCode()+"}')?'"+gvi+"':'N'}");
               lel.addProperty("OPTIONAL", el.getNonCompValue("OPTIONAL"));
               lel.addProperty("VISUALIZATION", el.getNonCompValue("VISUALIZATION"));
               String ren=el.getNonCompValue("RENOVATION");
               if (is_not_null(ren)) {//если задано св-во RENOVATION
                //if (like(ren,"%{%}%")) {   //19/05/2020
                if (el.properties.get("RENOVATION").isDynamic()){
                lel.addProperty("RENOVATION","{NLS:("+langCode+")"+el.getNonCompValue("RENOVATION")+"}");
//*06/06/2020*/      application.parser.findDependency(lel.getNonCompValue("RENOVATION"), lel.properties.get("RENOVATION"));//02/11/2018
                //playRenovation(lel);//02/11/2018
                }
               }
               lel.addProperty("NLS", langCode);// св-ву NLS зададим код языка
               // добавляе действие по удалеию поля
               addAction(lel.getCode()+"-D",this.getNewSeq("ACTION"),
                application.NLS.NLS("RUS:Удалить текст \";ENG:Remove text \";")+application.NLS.NLS(lang.name)+"\"",//name
                lel.getCode(),//parent
                application.NLS.NLS("RUS:Удалить текст \";ENG:Remove text \";")+application.NLS.NLS(lang.name)+"\"",//prompt
                null, // hint
                //"{JS:if ('{F:" + el.getCode() + "}'.match(/.+?:.*?;/g).length-1) 'Y';else 'N';}", // active
                //"{JS:if ('{F:" + el.getCode() + "}'.match(/.+?:[\\s\\S]*?;/g).length-1) 'Y';else 'N';}", // active
                 "{JS:if ('{F:" + el.getCode() + "}'.match(/.+?:[^;]*;/g).length-1) 'Y';else 'N';}", // active
                null,//visible
                //"JS:" + el.getCode() + "='{F:" + el.getCode() + "}'.replace(/(;|^)" + langCode + ":.*?;/, '$1')", //action pre
                // "JS:" + el.getCode() + "='{F:" + el.getCode() + "}'.replace(/(;|^)" + langCode + ":[\\s\\S]*?;/, '$1')", //action pre
                 "JS:" + el.getCode() + "='{F:" + el.getCode() + "}'.replace(/(;|^)" + langCode + ":[^;]*;/, '$1')", //action pre
                null,//action type
                null,//action command
                "N", //is check
                null,null,null,null,null,null,null,null,null
               ); 
               
               // добавляе действие по добавлению
               // сочиняем JS-размножитель
               String JSm="";
         /*      LuiCollection mlc=new LuiCollection(el.getValue());
               for (String llc:application.NLS.langList.keySet()){
                   if (!mlc.containsKey(llc)){
                       JSm=JSm+"{PROMPT:'"+application.NLS.NLS("RUS:Добавить вариант для;ENG:;Add entry for")+" \\\""+application.NLS.NLS(application.NLS.langList.get(llc).name)+"\\\"',NLS:'"+llc+"'}";
                   }
               }
               JSm="JS:["+JSm+"]";*/
        // JSm="JS:var A=[];'RUS:Русский;ENG:Английский;'.replace(/(.+?):(.+?);/g, function(E, K, V){A.push({PROMPT:'Добавить язык для \"' + V + '\"', NLS:K})});A";
        //JSm="JS:var A=[];'{ApplicationProperty:Var NLSlanguages}'.replace(/(.+?):(.+?);/g, function(E, K, V){if (!/(;|^)\" + K + \":.*?;/.test('{F:" + el.getCode() + "}'))     A.push({PROMPT:'"+application.NLS.NLS("RUS:Добавить язык для;ENG:Add field for;") +" \"' + V + '\"', NLS:K})});A";
        JSm="JS:var A=[];'{ApplicationProperty:Var NLSlanguages}'.replace(/(.+?):(.+?);/g, function(E, K, V){if (!RegExp('(;|^)' + K + ':.*?;').test('{F:" + el.getCode() + "}')) A.push({PROMPT:'"+application.NLS.NLS("RUS:Добавить текст;ENG:Add text;")+" \"' + V + '\"', NLS:K})});A";
               addAction(lel.getCode()+"-A",this.getNewSeq("ACTION"),
                application.NLS.NLS("RUS:Добавить текст ...;ENG:Add text ...;"),//name
                lel.getCode(),//parent
                application.NLS.NLS("RUS:Добавить текст ...;ENG:Add text ...;"),//name
                null, // hint
                null, // active
                JSm,//visible
                //"JS:" + el.getCode()+"='{F:"+el.getCode()+"}{Property:NLS}:;'", //action pre
                "JS:" + el.getCode()+"='{F:"+el.getCode()+"}{Property:NLS}:'+'{F:" + el.getCode() + "-{Property:NLS}}'.replace(';', '#$')+';'", //action pre
                null,//action type
                null,//action command
                "N", //is check
                null,null,null,null,null,null,null,null,null
               ); 
             //ItemValues=ItemValues+"{F:"+lel.getCode()+"}"; //По ходу строится текстовая цепочка вида {ПолеЯзыка}{ПолеЯзыка}...
             ItemValues=ItemValues+langCode+":'+'{F:"+lel.getCode()+"}'.replace(/;/g,'#$')+';"; //По ходу строится текстовая цепочка вида {ПолеЯзыка}{ПолеЯзыка}...
             if (!"".equals(UserChanged)) {UserChanged+="||";}
             UserChanged+="\"{ElementProperty:"+lel.getCode()+".UserChanged}\"";

           } //крутим языки
           //Меняем многоязычное поле
           el.addProperty("VISIBLE","N"); //Делаем его невидимым 
           el.addProperty("UserChanged", "{JS:"+UserChanged+"}");
           el.properties.get("UserChanged").getComputedProperty();
           
//           el.addProperty("RENOVATION","{JS:('"+ItemValues+"').replace(/(.*?):.*?;/g, function(E, K){if (RegExp('(;|^)' + K + ':.*?;').test('{F:"+el.getCode()+"}')) return E; else return ''})}");
//             el.addProperty("RENOVATION","{JS:('"+ItemValues+"').replace(/(.*?):[\\s\\S]*?;/g, function(E, K){if (RegExp('(;|^)' + K + ':[\\s\\S]*?;').test('{F:"+el.getCode()+"}')) return E; else return ''})}");
           el.addProperty("RENOVATION","{JS:('"+ItemValues+"').replace(/(.*?):[^;]*;/g, function(E, K){if (RegExp('(;|^)' + K + ':[^;]*;').test('{F:"+el.getCode()+"}')) return E; else return ''})}");
// Заново строим зависимости
           //playRenovation(el);} /02/11/2018
           } // если группы ещё нет
//06/06/2020*/          el.setValue(Val, false,"N",false);//02/11/2018
//06/06/2020*/          application.parser.findDependency(el.getNonCompValue("RENOVATION"), el.properties.get("RENOVATION"));//02/11/2018
         }// Только для многоязычных полей
         else { // если обычное поле //02/11/2018
//06/06/2020             el.setValue(el.getProperty("DEFINITION"), false,"N",false);//02/11/2018
//06/06/2020             application.parser.findDependency(el.getNonCompValue("RENOVATION"), el.properties.get("RENOVATION"));//02/11/2018
         }
      // }// если ещё нет никаких NLS-св-в у этого поля
      
 }
  
}
/**
 * Построение зависимостей для автообновление динамических значений полей
 */
/*// 02/11/2018public void playRenovation(LuiAppElement pEl) throws LuiException, SQLException{
    String v=pEl.getNonCompValue("RENOVATION");
    if (is_not_null(v)) {//если задано св-во RENOVATION
      if (!v.equals("{Property:DEFINITION}")){ // и его значение не равно {Property:DEFINITION}
         pEl.properties.get("Value").startedValue=v; 
         pEl.properties.get("Value").SCN=1;
      }  
    } else {
         pEl.properties.get("Value").SCN=0;
    }
    // вычисляем зависимости для свойства Value
       application.parser.findDependency(pEl.getNonCompValue("Value"), pEl.properties.get("Value"));
}*/
/**
 * Добавление к полю или ячейке столбца шаблонных действий
 * @param pElcode - код элемента (столбца или поля)
 * @return - true если добавление выполнено
 */
private boolean prepareTmplAct4Item(String pElCode) throws LuiException, SQLException{
  LuiAppElement el=elements.get(pElCode);
  boolean b=false;
  boolean lov_added=false;
  String variants=el.properties.get("VARIANTS").getNonCompValue();
  double seq=el.getMaxChildSeq();
 // String frmTemplCode=getType()+"_TEMPLATE"; // код шаблонной формы
  String frmTemplCode=getVal("FormTemplate");
  LuiAppForm frmTempl=application.formTemplates.get(frmTemplCode); 
  if (!application.formTemplates.containsKey(frmTemplCode)) return b;

  
  
  String elTemplCode=el.getProperty("ITEM_TEMPLATE"); //код шаблонного поля
  
   
    if (frmTempl.elements.containsKey(elTemplCode)){ // если есть такое шаблонное поле в шаблонной форме
        for (String actTemplCode:frmTempl.elements.keySet()){ // циклим элементы шаблонной формы   
          if (like(actTemplCode,"LOV$%") && (is_null(variants))) continue; //если имя похоже на LOV$ а в нашем поле вариантов нет уходим на продолжение
          if (is_not_null(variants)) { // если VARIANTS не пуст
            if (!like(variants,"LOV_:%")){ // и не начинается с LOV?:
            if (like(variants.toUpperCase(),"SELECT%")){ // но начинается с SELECT, то это LOVQ
                el.addProperty("VARIANTS", "LOVQ:"+variants);
              } else
            if (like(variants,"%:%;") && !el.getProperty("CHANGING").equals("R")){ // а если похож на коллекцию, то это LOVD
               el.addProperty("VARIANTS", "LOVD:"+variants);
            }
            }
          }
            LuiAppElement eTempl=frmTempl.elements.get(actTemplCode); // берём шаблонный элемент
            if (eTempl.getType().equals("ACTION")){ // если это действие   
             if (eTempl.getParent().equals(elTemplCode)){ // уровня шаблонного поля
              String ncode=eTempl.getCode().replace(elTemplCode,pElCode);// подменяем в коде шаблонного действии код шаблонного поля(столбца) на код элемента настоящей формы
              if (!elements.containsKey(ncode)){ //и если такого элемента нет   
               boolean bc=true;
                if (like(actTemplCode,"EDITOR$%")) { // если шаблонное действие начинается с EDITOR$
                  for (String edtactex:elements.keySet()){ // ищем действия EDITOR$%
                    if (elements.get(edtactex).getType().equals("ACTION")){
                      if (like(elements.get(edtactex).getCode(),"EDITOR$%")  ){ // если такое есть
                       
                        String edtactexParent=elements.get(edtactex).properties.get("Parent").startedValue; //берём отца
                        if (edtactexParent.equals(pElCode)) { 
                            // если отец равен нашему элементу ЛОКАЛЬНОЕ ПЕРЕКРЫЛО ШАБЛОННОЕ
                            bc=false;
                            break;
                        } else  {
                        if (like(edtactexParent,"%CurrentField}%")|| like(edtactexParent,"%CurrentColumn}%")) { 
                         // если отец  зависит от Var CurrentField или CurrentColumn
                         // но уходим на след оборот если код местного элемента начинается с FIELD но не совпадает с шаблонным элементом
                          if (like(edtactex,"EDITOR$FIELD%") && !elTemplCode.equals(edtactex.replace("EDITOR$",""))) continue;
                            bc=false;
                            break;
                        }}
                      }  
                    }  
                  } 
               }
              if (bc){
              LuiAppElement nel=addElement(ncode,"ACTION",seq+1000.0+eTempl.sourceSeq,eTempl.getProperty("Name"),pElCode);// создаём действие 
              nel.addProperty("FromTemplate", frmTempl.getCode()+"."+actTemplCode); // указываем, что оно пришло из шаблона
              for (String pk:eTempl.properties.keySet()){ // крутим свойства шаблонного действия
               if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
               nel.addProperty(pk,eTempl.properties.get(pk).getNonCompValue()); // добавляем их
                }
              // а теперь добавляем параметры из шаблонного действия 
               for (String PactTemplCode:frmTempl.elements.keySet()){ // циклим элементы шаблонной формы   
                LuiAppElement PeTempl=frmTempl.elements.get(PactTemplCode); // берём шаблонный 
                if (PeTempl.getParent().equals(actTemplCode) && PeTempl.getType().equals("PARAM")){ // параметр шаблонного действия
                   if (elements.containsKey(PactTemplCode)){ // если такой элемент у нас уже есть
                     PactTemplCode=PactTemplCode+String.valueOf(PeTempl.sourceSeq); // мутируем код 
                   }
                  LuiAppElement Pnel=addElement(PactTemplCode,"PARAM",PeTempl.sourceSeq,PeTempl.getProperty("Name"),ncode);// создаём параметр  
                   Pnel.addProperty("FromTemplate", frmTempl.code+"."+PactTemplCode);
                  // и его св-ва
                  for (String pk:PeTempl.properties.keySet()){ // крутим свойства шаблонного параметра
               if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
               Pnel.addProperty(pk,PeTempl.properties.get(pk).getNonCompValue()); // добавляем их
                }
                }
               }
              b=true;
              if (like(actTemplCode,"LOV$%")) lov_added=true;
              }
             }
             }
            }
        }
    }

//////////////////////////
if (!lov_added){
variants=el.properties.get("VARIANTS").getNonCompValue();
  if (is_not_null(variants) && !like(variants,"LOVR:%")) { // если есть список значений и это не LOVR:%
      if (!like(variants,"LOV_:%")){ // и не начинается с LOV?:
            if (like(variants.toUpperCase(),"SELECT%")){ // но начинается с SELECT, то это LOVQ
                el.addProperty("VARIANTS", "LOVQ:"+variants);
              } else
            if (like(variants,"%:%;")&& !el.getProperty("CHANGING").equals("R")){ // а если похож на коллекцию, то это LOVD
               el.addProperty("VARIANTS", "LOVD:"+variants);
            }
            }
    String actTempl4lov="LOV$"+el.getType();
    if (frmTempl.elements.containsKey(actTempl4lov)){
      LuiAppElement eTempl=frmTempl.elements.get(actTempl4lov); // берём шаблонный элемент
      if (eTempl.getType().equals("ACTION")){ // если это действие   
        String ncode=eTempl.getCode().replace(el.getType(),pElCode);// подменяем в коде шаблонного действии код типа поля(столбца) на код элемента настоящей формы  
        if (!elements.containsKey(ncode)){ //и если такого элемента нет   
          LuiAppElement nel=addElement(ncode,"ACTION",seq+1000.0+eTempl.sourceSeq,eTempl.getProperty("Name"),pElCode);// создаём действие   
          nel.addProperty("FromTemplate", frmTempl.code+"."+actTempl4lov); // указываем, что оно пришло из шаблона
              for (String pk:eTempl.properties.keySet()){ // крутим свойства шаблонного действия
               if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
               nel.addProperty(pk,eTempl.properties.get(pk).getNonCompValue()); // добавляем их
                }
             // а теперь добавляем параметры из шаблонного действия 
               for (String PactTemplCode:frmTempl.elements.keySet()){ // циклим элементы шаблонной формы   
                LuiAppElement PeTempl=frmTempl.elements.get(PactTemplCode); // берём шаблонный 
                if (PeTempl.getParent().equals(actTempl4lov) && PeTempl.getType().equals("PARAM")){ // параметр шаблонного действия
                   if (elements.containsKey(PactTemplCode)){ // если такой элемент у нас уже есть
                     PactTemplCode=PactTemplCode+String.valueOf(PeTempl.sourceSeq); // мутируем код 
                   }
                  LuiAppElement Pnel=addElement(PactTemplCode,"PARAM",PeTempl.sourceSeq,PeTempl.getProperty("Name"),ncode);// создаём параметр  
                   Pnel.addProperty("FromTemplate", frmTempl.code+"."+PactTemplCode);
                  // и его св-ва
                  for (String pk:PeTempl.properties.keySet()){ // крутим свойства шаблонного параметра
               if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
               Pnel.addProperty(pk,PeTempl.properties.get(pk).getNonCompValue()); // добавляем их
                }
                }
               }
              b=true; 
        }
      }
      
    }  
  }
}
return b;
}
/**
 * Предобработка, специфическая для списков 
 * задачи - добавка действия типа "покажитесь числа точнее"
 *        - подготовка сортировки
 */
private void prepareList() throws LuiException, SQLException{
HashMap<String,Boolean> done=new HashMap<>();
boolean b=true;

while (b){
 b=false;
  for(String elCode:elements.keySet()){ // цикл элементов для инфекций полей действиями
   if (done.containsKey(elCode)) continue;   
    LuiAppElement el=elements.get(elCode);
      if (el.getType().equals("ITEM")) { //если столбец
      done.put(elCode, b); // занесём его в массив уже обработанных
          // Перенос выражения по умолчанию в Value с целью последующего вычисления
          String v=el.getNonCompValue("DEFINITION");
          elements.get(elCode).setProperty("Value",v);
  //добавление шаблонных действий из соответствующего шаблона формы
  b=prepareTmplAct4Item(elCode);
      }
  if (b) { // если добавление выполнено прервём цикл осмотра полей и начнём его заново
      break; // это нужно для правильной последовательности расположения действий в массивах элементов
  } 
  }
}    
    
    makeEl4order(); // готовим перечень столбцов сортировки
    // нормализуем св-во SORT
    
    for (int i=0;i<el4order.size();i++){
        
        String newS="";
        if (el4order.get(i).getProperty("SORT").contains("-")){
            newS=newS+"-";
        }
        if (el4order.get(i).getProperty("SORT").contains("0")){
            newS=newS+"0";
        }
        newS=newS+String.valueOf(i+1);
        el4order.get(i).setProperty("SORT",newS);
        
    }
    //setVal("Elements","");// Сброс кэш элементов
    // подготовим список пользовательских конфигураций
    setUsrConfigListProperty();
}
private void prepareAddons()throws LuiException, SQLException{
  
  LuiDbConn lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
  try {    
  ArrayList<LuiCollection> AddonsProps=lc.getSelectAllRowsValMap(
"select an.form_type,af.keys,af.addon_code,"+ 
"af.formtype_code,af.title,af.parent,af.is_auto,af.list_is_toolbar,af.list_is_button,"+ 
"an.form_code,an.config_code,an.icon," +
" nls(an.name) \"name\",nls(an.description) \"description\"," +
" is_active"+
" from lui_t_addon2form af,lui_t_addon an" +
" where af.app_code='"+application.code+"'" +
" and af.config_code='"+getProperty("ConfigCode")+"'" +
" and af.form_code='"+getProperty("MetadataCode")+"'" +
" and af.app_code=an.app_code" +
" and an.form_code is not null"+
" and af.addon_code=an.code");
  if (AddonsProps==null) return;
  if (AddonsProps.size()<1) return;
  LuiAppElement el;
  LuiAppElement pel;
String parent=null;
String icon=null;
String prompt=null;
String visible=null;
String role=null;
String button=null;
String icon_in_toolbar=null;
double seq=0;
String requery=null;
String active=null;
String command_type=null;
String is_active=null;
String keys=null;
  // крутим дополнения
    for(int i=0;i<AddonsProps.size();i++){
      LuiCollection AddonProps=AddonsProps.get(i);
      if (AddonProps.Get_Element("form_code").equals("LUI_L_LINKED_FILE")) {//если дополнение обслуживается формой файлов
          //то проверим наличие к-либо прав на к-либо файлы этого дополнения
       String tmp=lc.getSelectResult("select 'Y' from lui_t_addon_filetype where app_code='"+ 
               application.code+"' and addon_code='"+
               AddonProps.Get_Element("addon_code")+
          "' and lui_f_get_effective_right('FILES',filetype_code,'%','"+application.getProperty("AppUser")+
          "','"+application.code+"')='Y' limit 1");
      if (NVL(tmp,"N").equals("N")) continue;
      }
      parent=AddonProps.Get_Element("parent");if (!elements.containsKey(parent))parent="Form";
      icon=AddonProps.Get_Element("icon");
      role=null;
      button=null;
      visible="Y";
      prompt=AddonProps.Get_Element("name");
      button=null;
      active=null;
      requery=null;
      command_type=null;
      is_active=null;
      keys="";
      //icon_in_toolbar=icon;
      icon_in_toolbar="{JS:if(\"{Property:ACTIVE}\"==\"Y\")\""+icon+"\";else \"\";}";
      seq=0;
      ArrayList<LuiCollection> keyElements=lc.getSelectAllRowsValMap("select replace(value::text,'\"','') \"CODE\" from json_array_elements('"+AddonProps.Get_Element("keys")+"') order by 1");
      if (keyElements==null) continue;
      int ji=0;
      for (int k=0;k<keyElements.size();k++){
          String code=keyElements.get(k).Get_Element("CODE");
          if (elements.containsKey(code)) {
           if (ji>0)keys=keys+",";   
           keys=keys+"\"{F:"+code+"}\"";
           ji++;
          }
      }
      keys="["+keys+"]";
      is_active=AddonProps.Get_Element("is_active");
      is_active=is_active.replace("{AppCode}",application.code);
      is_active=is_active.replace("{AddonCode}",AddonProps.Get_Element("addon_code"));
      is_active=is_active.replace("{Keys}",keys);
      switch (getType()){
          case "BLANK":
       visible="{JS: if (";
       ji=0;
       for (int k=0;k<keyElements.size();k++){
               String code=keyElements.get(k).Get_Element("CODE");
       if (elements.containsKey(code)) {
           if (ji>0)visible=visible+" && ";
           visible=visible+"\"{F:"+code+"}\"";
           ji++;
       }
           }
       //visible=visible+")\"Y\";else \"N\";}";
       visible=visible+")\""+is_active+"\";else \"N\";}";
       active=visible;
       if (!AddonProps.Get_Element("is_auto").equals("N")){
             button=null;
             icon=null;
             visible="N";
             prompt=null;
             icon_in_toolbar=null;
             role="AUTO";
             parent="Form";
             if (AddonProps.Get_Element("is_auto").equals("R")){
                 command_type="BLANK";
             }
             if (AddonProps.Get_Element("is_auto").equals("B")){
                 command_type="LIST";
             }
            }
          break;
          case "LIST":
                 icon_in_toolbar="{JS:if(\"{Property:ACTIVE}\"==\"EXIST\")\""+icon+"\";else \"\";}";   
           if (AddonProps.Get_Element("list_is_toolbar").equals("N")
                && parent.equals("Form")){
                icon_in_toolbar=null;
            }
            prompt="[ico:"+icon+"]"+prompt;
            if (AddonProps.Get_Element("list_is_button").equals("Y")) button="[ico:"+icon+"]"+AddonProps.Get_Element("name");
            if (!AddonProps.Get_Element("is_auto").equals("N")){
             button=null;
             icon=null;
             visible="N";
             prompt=null;
             icon_in_toolbar=null;
             role="AUTO";
             parent="Form";
             if (AddonProps.Get_Element("is_auto").equals("R")){
                 command_type="BLANK";
             }
             if (AddonProps.Get_Element("is_auto").equals("B")){
                 command_type="LIST";
             }
            } 
            if (!parent.equals("Form")){
             button=null; 
            }
            seq=this.getNewSeq("ACTION")+10;
            //active="EXIST";
            active="{JS:if(\""+is_active+"\"==\"Y\")\"EXIST\";else \"N\";}";
            requery="CURRENT";
          break;
      }
               el=this.addAction("ADDON$"+AddonProps.Get_Element("addon_code"), // код действия 
                              seq, // номер
                              AddonProps.Get_Element("name"), // название
                              parent, // parent
                              prompt, // prompt
                              AddonProps.Get_Element("description"), // hint
                              active, // Active
                              visible, // Visible
                              null, // actionPRE
                              NVL(command_type,AddonProps.Get_Element("formtype_code")),// actionType
                              AddonProps.Get_Element("form_code")+"(CONFIG:"+AddonProps.Get_Element("config_code")+")",// COMMAND
                              "N",//pIsCheck
                              null,//pActionQuestion
                              null,//pNotes 
                              role,//pRole VALPREFIX
                              requery,//pRequery 
                              "",//pCloseTransaction
                              "",//pActionPost code
                              icon_in_toolbar,// icon
                              button,"N"
            );
            // добавляем параметры
            
           pel=addElement("ADDON$"+String.valueOf(i)+"APPCODE","PARAM",this.getNewSeq("PARAM"),"AppCode", el.getCode()); // код приложения
           pel.addProperty("Value",application.code);pel.addProperty("VALUE",application.code);
           pel.addProperty("NAME", "P_APPCODE");
           
           pel=addElement("ADDON$"+String.valueOf(i)+"ADDCODE","PARAM",this.getNewSeq("PARAM"),"AddonCode", el.getCode()); // код дополнения
           pel.addProperty("Value",AddonProps.Get_Element("addon_code"));pel.addProperty("VALUE",AddonProps.Get_Element("addon_code"));
           pel.addProperty("NAME", "P_ADDONCODE");
           
           pel=addElement("ADDON$"+String.valueOf(i)+"ADDNAME","PARAM",this.getNewSeq("PARAM"),"AddonName", el.getCode()); // название дополнения
           pel.addProperty("Value",AddonProps.Get_Element("name"));
           pel.addProperty("VALUE",AddonProps.Get_Element("name"));
           pel.addProperty("NAME", "P_ADDONNAME");
           
           pel=addElement("ADDON$"+String.valueOf(i)+"KEYS","PARAM",this.getNewSeq("PARAM"),"Keys", el.getCode()); // ключ
           // сочиним выражение для вычисления ключа
           String kv="[";
           for (int k=0;k<keyElements.size();k++){
               String code=keyElements.get(k).Get_Element("CODE");
               kv=kv+"\"{F:"+code+"}\",";
           }
           kv=LuiString.RTrim(kv,",")+"]";
           pel.addProperty("Value",kv);pel.addProperty("VALUE",kv);
           pel.addProperty("NAME", "P_KEYS");
           
           pel=addElement("ADDON$"+String.valueOf(i)+"TITLE","PARAM",this.getNewSeq("PARAM"),"Title", el.getCode()); // заголовок
           pel.addProperty("Value",AddonProps.Get_Element("title"));
           pel.addProperty("VALUE",AddonProps.Get_Element("title"));
           pel.addProperty("NAME", "P_TITLE");
    /* теперь посмотрим наличие файлов в этом дополнении*/
    /*
   if (getType().equals("LIST")){ // если список
      String files=lc.getSelectResult("select min(filetype_code) from lui_t_addon_filetype where app_code='"+application.code+"' and addon_code='"+
              AddonProps.Get_Element("addon_code")+"'");
      if (is_not_null(files)){ // /если файлы есть, то
          for (String elcode:elements.keySet()){
              if (elements.get(elcode).getType().equals("ACTION")) {
                 if (like(elements.get(elcode).getProperty("ROLE"),"%DELETE%" )){ // находим действие с ролью удаления
                   String action=elements.get(elcode).properties.get("PROC_INITIALLY").startedValue;
                   if (is_not_null(action)) action=action+"\n\n";
                   action=action+"SQL:select lui_r_remove_addon_files('{ApplicationProperty:AppCode}','"+
                           AddonProps.Get_Element("addon_code")+"','"+kv+"')";
                   elements.get(elcode).addProperty("PROC_INITIALLY",action);
                 }
              }
          }
      }
    }*/
    }
  // 
  } 
  catch (SQLException ex){
      String s=ex.getMessage();
      s="here";
  }
  
  finally {
      application.metaConnPool.SetFree(lc);
  }
    
}
/**
 * Шаблонные элементы матрицы
 */
public void prepareMatrixTemplEl(String eType)throws LuiException, SQLException{
    
    //String tfrmcode=getType()+"_TEMPLATE";
    String tfrmcode=getVal("FormTemplate");
    double aseq=0;
    if (!eType.equals("CELL"))aseq=getNewSeq(eType);
    LuiAppProp p;
    if (application.formTemplates.containsKey(tfrmcode)){
    LuiAppForm tfrm=application.formTemplates.get(tfrmcode);
    for (String telcode:tfrm.elements.keySet()){
        LuiAppElement tel=tfrm.elements.get(telcode);
        if (!tel.getType().equals(eType))continue;
        if (!tel.getParent().equals("Form"))continue;
        LuiAppElement el;
         if (!elements.containsKey(telcode)) {
          el=addElement(telcode,eType,aseq+tel.sourceSeq,tel.getProperty("Name"),"Form");  
          el.addProperty("FromTemplate", tfrm.code+"."+telcode);
         } else el=elements.get(telcode);
        
        for (String pk:tel.properties.keySet()){ // копируем свойства действия
            if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
            if (el.properties.containsKey(pk)){
              //  if (el.properties.get(pk).is_meta==false)
                  if (el.properties.get(pk).genesis!='M')
                    el.addProperty(pk,tel.properties.get(pk).getNonCompValue());
                    
            } else {
          p=el.addProperty(pk,tel.properties.get(pk).getNonCompValue());
          p.genesis=tel.properties.get(pk).genesis;
            }
        }
    }
    }
}
/**
 * Подготовка системного меню
*/
private void prepareSysMenu() throws LuiException, SQLException{
   //String tfrmcode=getType()+"_TEMPLATE";
   String tfrmcode=getVal("FormTemplate");
 String visible="N";
 double aseq=getNewSeq("ACTION");
 double pseq=getNewSeq("PARAM");
 if (application.formTemplates.containsKey(tfrmcode)){
    LuiAppForm tfrm=application.formTemplates.get(tfrmcode);
    for (String telcode:tfrm.elements.keySet()){
       if (tfrm.elements.get(telcode).getType().equals("ACTION")){ 
        if (tfrm.elements.get(telcode).getParent().equals("Form")){   
        if (tfrm.getType().equals("GRID")|| tfrm.getType().equals("MENU") || like(telcode,"SYSMENU$%") || like(telcode,"SYSACTION$%") || like(telcode,"COLMENU$%") || like(telcode,"CELLMENU$%") || like(telcode,"QBE$ITEM%")
                ){
         if (!elements.containsKey(telcode)){ // если такого элемента нет 
        LuiAppElement tel=tfrm.elements.get(telcode);
        //addElement(String pCode,String pType,Integer pSeq,String pName,String pParent) throws LuiException, SQLException{
        
        LuiAppElement el=addElement(telcode,"ACTION",aseq+tel.sourceSeq,tel.getProperty("Name"),"Form");
        el.addProperty("FromTemplate", tfrm.code+"."+telcode);
        //new LuiAppElement(this,tel.getType(),tel.getCode(),tel.seq);
        for (String pk:tel.properties.keySet()){ // копируем свойства действия
            if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
          el.addProperty(pk,tel.properties.get(pk).getNonCompValue());
        }
        
        // теперь осмотрим параметры шаблонного действия
         for (String tpelcode:tfrm.elements.keySet()){
             if (tfrm.elements.get(tpelcode).getType().equals("PARAM")){ 
                if (tfrm.elements.get(tpelcode).getParent().equals(telcode)){
                 LuiAppElement tpel=tfrm.elements.get(tpelcode);
                 if (elements.containsKey(tpelcode)){ // если вдруг такой элемент уже есть подправим его код
                   tpelcode=tpelcode+String.valueOf(tpel.sourceSeq);
                 }
                 //создаём параметр
                 LuiAppElement pel=addElement(tpelcode,"PARAM",pseq+tpel.sourceSeq,tpel.getProperty("Name"),el.getCode());
                 pel.addProperty("FromTemplate", tfrm.code+"."+tpelcode);
                    for (String pk:tpel.properties.keySet()){ // копируем свойства параметра
                        if (pk.equals("Parent") || pk.equals("Code")) continue;  // если это не код и не родитель
                    pel.addProperty(pk,tpel.properties.get(pk).getNonCompValue());
                    }
        
                }         
             }
         }
        // если это меню столбца, то подменим Parent на динамику вычисления текущего столбца
        if (like(telcode,"COLMENU$%") && el.getProperty("Parent").equals("Form")) {
          el.addProperty("Parent","{FormProperty:Var ColumnMenu}");
        }
        // если это QBE меню столбца, то подменим Parent на динамику вычисления текущего столбца
        if (like(telcode,"CELLMENU$%") && el.getProperty("Parent").equals("Form")) {
          el.addProperty("Parent","{FormProperty:Var CurrentColumn}");
        }
        } else { // такой элемент есть - меняем ему seq
            LuiAppElement el=elements.get(telcode);
            LuiAppElement tel=tfrm.elements.get(telcode);
            el.sourceSeq=aseq+tel.sourceSeq;
            el.mutSeq=aseq+tel.sourceSeq;
         }  
        }
    }}}}
 for (String elcode:elements.keySet()){
     if (like(elcode,"SYSMENU$%")){
         if (elements.get(elcode).getType().equals("ACTION")){ 
        if (elements.get(elcode).getParent().equals("Form")){   
        LuiAppElement el=elements.get(elcode);
        if (el.properties.get("ACTIVE").recalculate!=null) {
            visible="Y";
        }
        if (!el.properties.get("ACTIVE").currentValue.equals("N")){
            visible="Y";
        }
  
     }
 }}}
setVal("SysMenu",visible);
        
}

/** 
 * Вычисление очередного seq для добавляемого элемента действи
 * @param pElType - тип элемента
 * @return значение очередного seq
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public double getNewSeq(String pElType) throws LuiException, SQLException{
    Double newseq = null;
    for(String elCode:elements.keySet()){
      if (!elements.get(elCode).properties.containsKey("FromTemplate")) {  // просматриваем только не шаблоны
       if (!like(elCode,"SYSMENU$%")){   
        if (elements.get(elCode).getType().equals(pElType)){
/*          if (is_not_null(pParent)) { 
          String parent=elements.get(elCode).getParent();
          if (!parent.equals(pParent))continue;
          }*/
            if (newseq==null)newseq=elements.get(elCode).sourceSeq;
            if (newseq<elements.get(elCode).sourceSeq) newseq=elements.get(elCode).sourceSeq;
            
        }
    }
    }
    }
    if (newseq==null) return 1;
    return ++newseq;
}

    /**
     * Добавление группы в список или в бланк
     *
     * @param pCode - код группы
     * @param pPrompt - надпись на корешке
     * @param pVisible - признак видимости
     * @param pSeq- номер элемента
     * @param pNotes- Текст для генерации Help
     * @param pGtype- Тип группы (если не передать сам определится) для бланков:
     * TAB,BUNCH,SIDE_BY_SIDE и "". Для списков:FIXED, OVERFLOW
     * @param pParentGroup -Родительская группа или "Form"
     * @param pPromptPos - Расположение подписей к полям внутри группы
     * относительно самих полей:AUTO - решает интефейс,LEFT - слева от всех
     * полей, UP - сверху всех полей. -- Любое другое значение эквивалентно
     * значению LEFT
     * @return LuiAppElement
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public LuiAppElement addGroup(String pCode,String pPrompt,String pVisible,Double pSeq,String pNotes,String pGtype,String pParentGroup,String pPromptPos)throws LuiException, SQLException{
    LuiAppElement el=null;
    el=this.addElement(pCode,"GROUP",pSeq,pPrompt,pParentGroup);
    if (is_null(pGtype)){
    if (getType().equals("BLANK")) pGtype="TAB";   
    else pGtype="OVERFLOW";
    }
    el.addProperty("GROUP_TYPE", pGtype);
    el.addProperty("TITLE", pPrompt);
    el.addProperty("SEEABLE",pVisible);
    el.addProperty("POSITION",pPromptPos);
    el.addProperty("GROUP_NOTES", pNotes);
    return el;
}
/**
 * Добавление действия
     * @param pCode - код действия
     * @param pSeq - номер элемента
     * @param pName - название
     * @param pParent - код старшего элемента если null или "", то будет "Form"
     * @param pPrompt - надпись на кнопке
     * @param pHint - пояснение к действию
     * @param pActive - признак активности 
     * @param pVisible - признак видимости
     * @param pActionPre - Программа выполняемая перед выполнением команды
     * @param pActionType - тип команды LIST/BLANK/TREE/URL
     * @param pActionCommand - команда
     * @param pIsCheck -Y (по умолчанию) - делать проверку форматов и обязательных полей, любое другое значение - нет
     * @param pActionQuestion -Вопрос перед выполнением
     * @param pNotes - Текст для генерации Help
     * @param pRole - Роль действия:DBLCLICK - выполняется при нажати ENTER в любом поле бланка
                                    CLOSE - Выполняется при попытке закрытия формы
                                    WIZARD - Если действие вызывает новую форму, то текущий бланк скрывается (Next, вперёд).
                                              Если действие закрывает текущую форму, то попадаем на предыдущий шаг (Prev, назад)
     * @param pRequery -Что делать после выполнения действия
                        FIELD   - Обновить текущее поле
                        CURRENT - Обновить поля текущей группы
                        ALL     - Обновить все поля
                        REFRESH - Переинициировать форму (как бы выйти и войти)
                        EXIT    - Закрыть бланк и выйти
                        CANCEL  - Закрыть бланк и отменить действие, вызвавшее бланк
                        NONE (и любое другое значение) - Ничего не делать
     * @param pCloseTransaction - Разрешение транзакции в вызываемой форме:
                                - Пусто (по умолчанию) - режим транзакций не меняется (наследуется из вызываемой формы или берётся значение по умолчанию)
                                - N - в вызываемой форме не будет завершаться транзакция (все commit не выполняются)
                                - Y - в вызываемой форме все commit всегда безусловно выполняются (устанавливать не рекомендуется)
                                - Любое другое значение эквивалентно "Пусто"
     * @param pActionPost - Этот код выполняется уже после команды и аналогичен pActionPre
     * @param pActionIcon - Имя иконки, активирующей действие
     * @param pButton - Имя кнопки в подвале
     * @param pLock - накладывать блокировку
 * @return LuiAppElement
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiAppElement addAction(String pCode,double pSeq, String pName,String pParent,String pPrompt,String pHint,
        String pActive,String pVisible,String pActionPre,String pActionType,String pActionCommand,
        String pIsCheck,String pActionQuestion, String pNotes,String pRole, String pRequery,String pCloseTransaction,String pActionPost,
        String pActionIcon,String pButton,String pLock) throws LuiException, SQLException{
    LuiAppElement el=null;
    el=this.addElement(pCode,"ACTION",pSeq,pName,pParent);
   if (pPrompt!=null) el.addProperty("PROMPT", pPrompt);
   if (pHint!=null) el.addProperty("EXPLANATION", pHint);
   if (pActive!=null) el.addProperty("ACTIVE", pActive);
   if (pVisible!=null) el.addProperty("DISPLAY", pVisible);
   if (pRole!=null) el.addProperty("ROLE", pRole);
   if (pActionQuestion!=null) el.addProperty("QUESTION", pActionQuestion);
   if (pIsCheck!=null) el.addProperty("VERIFICATION", pIsCheck);
   if (pActionPre!=null) el.addProperty("PROC_INITIALLY", pActionPre);
   if (pActionType!=null) el.addProperty("COMMAND_TYPE", pActionType);
   if (pActionCommand!=null) el.addProperty("COMMAND", pActionCommand);
   if (pActionPost!=null) el.addProperty("PROC_FINALLY", pActionPost);
   if (pCloseTransaction!=null) el.addProperty("CHILD_COMMIT", pCloseTransaction);
   if (pRequery!=null) el.addProperty("AFTER", pRequery);
   if (pActionIcon!=null) el.addProperty("ICON", pActionIcon);
    if (pCode.equals("OK") && is_null(el.properties.get("ROLE").getCurrentValue())){
        el.addProperty("ROLE", "DBLCLICK");
    }
   if (pButton!=null)el.addProperty("BUTTON", pButton);
   if (pLock!=null)el.addProperty("LOCK", pLock);
    return el;
}
/**
 * Загрузка пользовательских предпочтений
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void loadUserPrefs() throws LuiException, SQLException{
    LuiDbConn c=null;
    if (is_null(application.getAppUser())) return; // если пользователь не известен уходим
    if (is_null(getVal("UserConfiguration"))){
      try{
       c=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
       String defCid=c.getSelectResult("Select p.value from lui_t_user_pref p where p.user_name='"+application.getAppUser()+"' and p.form_code='"+getProperty("FEATURE")+"' and p.config_code='"+getConfigCode()+"' and p.app_code='"+application.code+"' and p.property_code='DEF_USER_CFG' and p.element_code is null");   
       if (is_not_null(defCid)){
           loadUserConfig(defCid);
       }
      } finally{
             if (c!=null){
            try {
                application.metaConnPool.SetFree(c);
            } catch (Exception ex){
                
            }
      }  
    }
} else {
           loadUserConfig(getVal("UserConfiguration"));
    }
/*
try{
   String defCid=getVal("UserConfiguration");
   if (is_null(defCid)) {
       c=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
       defCid=c.getSelectResult("Select p.value from lui_t_user_pref p where p.user_name='"+application.getAppUser()+"' and p.form_code='"+getCode()+"' and p.config_code='"+getConfigCode()+"' and p.app_code='"+application.code+"' and p.property_code='DEF_USER_CFG' and p.element_code is null");    
   }
   if (is_not_null(defCid)){
           loadUserConfig(defCid);
   }
} finally{
             if (c!=null){
            try {
                application.metaConnPool.SetFree(c);
            } catch (Exception ex){
                
            }
      }
}
*/
}
    /**
     * Загрузка пользовательской конфигурации
     * @param pUserConfId идентификатор пользовательской конфигурации
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
    public void loadUserConfig(String pUserConfId) throws LuiException, SQLException
{
    LuiDbConn c=null;
    if (is_null(application.getAppUser())) return; // если пользователь не известен уходим
    try {
    c=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
    if (is_null(pUserConfId)){// Параметр пуст выявим "выбираемую по умолчанию"
       String defCid=c.getSelectResult("Select p.value from lui_t_user_pref p where p.user_name='"+application.getAppUser()+"' and p.form_code='"+getProperty("FEATURE")+"' and p.config_code='"+getConfigCode()+"' and p.app_code='"+application.code+"' and p.property_code='DEF_USER_CFG' and p.element_code is null");
       if (is_not_null(defCid)) pUserConfId=defCid;   
       
    }   
    if (is_not_null(pUserConfId)){ // если конфигурация определена
     // извлекаем её
     ArrayList<String> confal=c.getSelectFirstRowValList("Select order_by,name,parent_Form,parent_Config from lui_t_user_config where id="+pUserConfId);
     if (confal!=null){
     if (confal.get(3).equals(getConfigCode()) && confal.get(2).equals(getProperty("FEATURE"))){//если код формы и конфигурации соответствуют данной
      setVal("UserConfiguration",pUserConfId);
      setVal("UserConfigurationName",confal.get(1));
           listQBE.clear(); // очищаем QBE
          //и заполняем его
      ArrayList<ArrayList<String>> qbe=c.getSelectAllRowsValList("select column_code,condition,row_number from lui_t_user_condition where usrcfg_id="+pUserConfId+" order by row_number");
      if (qbe!=null){
      for (int ri=0;ri<qbe.size();ri++){
          ArrayList<String> qberow=qbe.get(ri);
        if (!elements.containsKey(qberow.get(0))) continue;  // если такой элемент есть
        if (!elements.get(qberow.get(0)).getType().equals("ITEM")) continue; // и это столбец
          HashMap<String,String> hmc=new HashMap<>();  
          if (listQBE.containsKey(qberow.get(2))) {
              hmc=listQBE.get(qberow.get(2));
          }            
          hmc.put(qberow.get(0),qberow.get(1));
          listQBE.put(qberow.get(2), hmc);
        }
      
      
      }
      String QBE_NOT=c.getSelectResult("select value from lui_t_user_pref prf where prf.property_code='QBE_NOT' and prf.element_code is null and prf.usrcfg_id="+pUserConfId);
      if (is_not_null(QBE_NOT)) {
          addProperty("QBE_NOT",QBE_NOT);
      }
      String QBE_COL=c.getSelectResult("select value from lui_t_user_pref prf where prf.property_code='QBE_COL' and prf.element_code is null and prf.usrcfg_id="+pUserConfId);
      if (is_not_null(QBE_COL)) {
          addProperty("QBE_COL",QBE_COL);
      }
      // применяем order by
      LuiCollection lc=new LuiCollection(confal.get(0));
      int n=0;
      for (String ec:lc.keySet()) {
        if (!elements.containsKey(ec)) continue; // если такой элемент есть
        if (!elements.get(ec).getType().equals("ITEM")) continue; // и это столбец
           /* для чего это надо????? 
            n++;  
            int sign=1;
            if (Integer.valueOf(lc.get(ec))<0) sign=-1;
            Integer so=(sign*n);
            elements.get(ec).setProperty("SORT", so.toString()); */
                elements.get(ec).addProperty("SORT", lc.Get_Element(ec));

        
      
      }
      }
    }
    }
    } catch (SQLException ex){
      application.msgWriter.write("LUI-00024",pUserConfId,getProperty("FEATURE")+"."+this.getConfigCode(),ex.getLocalizedMessage());
    } finally {
        if (c!=null){
            try {
                application.metaConnPool.SetFree(c);
            } catch (Exception ex){
                
            }
        }
    }

}
        
    /**
     * Загрузка списка пользовательских конфигураций списка
     *
     * @return список пользовательских конфигураций
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public String getUsrConfigList() throws LuiException, SQLException{
    //LuiDbConn c=null;
    String retVal = null;
    retVal="0:"+application.NLS.NLS("RUS:Базовая конфигурация;ENG:Base configuration;")+";";
    retVal=retVal+getVal("UsrConfList");
    String curconf=NVL(getVal("UserConfiguration"),"0");
    LuiCollection cc = new LuiCollection(retVal);
    retVal="";
    if (cc.size()>1 ||getVal("UserConfigurationShow").equals("Y")){
    for (String key:cc.keySet()){
       if (key.equals(curconf)){
           retVal=retVal+":"+cc.Get_Element(key)+";";
       } else
       {
           retVal=retVal+key+":"+cc.Get_Element(key)+";";
       }
           
    }
    }
    
    return retVal;
}

/**
 * Чтение визуального предпочтения пользователя
 * @param pElemCode - код элемента (если нет,то ставь ""
 * @param pProperty - код свойства
 * @param pDef - значение по умолчанию т.е. что вернуть если строк нет в БД (если нет, то ставь null
 * @return - предпочитаемое пользователем значение свойства
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getVisualPref(String pElemCode,String pProperty,String pDef) throws LuiException, SQLException{
   LuiDbConn c = null;
   String retVal=null;
   String usrCfgId=getVal("UserConfiguration");
   if (usrCfgId.equals("0")) usrCfgId="";
   if (is_null(application.getAppUser())) return pDef;
   String stmt="select value from lui_t_user_pref where pk='"+
   application.getAppUser()+"."+application.code+"."+getProperty("FEATURE")+"."+getConfigCode()+"."+
           usrCfgId+"."+pElemCode+"."+application.getWorkPlace()+"."+pProperty+"'";
   try{
   c=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
   retVal=c.getSelectResult(stmt,1);
   if (c.getRowCount()==0){
   if (("LUI".equals(application.code)) && (!"lui".equals(application.getAppUser()))) {
    stmt=stmt.replace("pk='"+application.getAppUser()+".","pk='lui.");
     retVal=c.getSelectResult(stmt,1);  
    if (c.getRowCount()==0) {   
       retVal=pDef;
    }  
   }
   else
   retVal=pDef;
   }
   
   } finally {
        if (c!=null){
            try {
                application.metaConnPool.SetFree(c);
            } catch (Exception ex){
                
            }
        }
    }
//   if (is_null(retVal)) return pDef;
   return retVal;
}
/**
 * Вычиляет тип позиции промптов для полей или столбцов группы списка или бланка
 * @param pGroupCode - код группы
 * @return "UP" или "LEFT"
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String getPromptPositionForGroup(String pGroupCode) throws LuiException, SQLException{
if (!elements.containsKey(pGroupCode))return "";
    if (!elements.get(pGroupCode).properties.containsKey("POSITION")){
    elements.get(pGroupCode).addProperty("POSITION", "AUTO");
}
    String pp=elements.get(pGroupCode).getProperty("POSITION");
 switch (pp){
     case "AUTO":
     String gParent=elements.get(pGroupCode).getParent();
     if (!gParent.equals("Form")) {//если группа вложена в другую группу
         return getPromptPositionForGroup(gParent);
     }
     return pp;
     case "UP":
     return pp;    
 }
return "LEFT";
}
public String getName4Report() throws LuiException, SQLException{
    String retVal=this.getFormID()+" - ";
    String meta=this.getProperty("MetadataCode");
    switch (this.getType()){
        case "LIST":
         retVal=retVal+application.NLS.NLS("RUS:Список;ENG:List;")+" ";
         break;
        case "HOST":
         retVal=retVal+application.NLS.NLS("RUS:Команда ОС;ENG:OS command;")+" ";
         break; 
        case "BLANK":
         retVal=retVal+application.NLS.NLS("RUS:Бланк;ENG:Blank;")+" ";
         break;
        case "MENU":
         retVal=retVal+application.NLS.NLS("RUS:Меню;ENG:Menu;")+" ";
         break;
        case "TREE":
         if (NVL(getVal("RootFormID"),getFormID()).equals(getFormID())) {
            retVal=retVal+application.NLS.NLS("RUS:Дерево;ENG:Tree;")+" "; 
         } else {
            retVal=retVal+application.NLS.NLS("RUS:Ветвь;ENG:Tree branch;")+" ";   
         }
         break;
        default:
         retVal=retVal+application.NLS.NLS("RUS:Форма;ENG:Form;")+" ";
    }
    retVal=retVal+this.getCode();
    if (!getCode().equals(meta)){
        retVal=retVal+" ["+meta+"]";
    }
    if (getVal("MasterForm").equals("Y")){
        retVal=retVal+" "+application.NLS.NLS("RUS:Окно;ENG:Window;")+" "+getVal("WindowID")/*+" - "+application.windows.get(getVal("WindowID").)*/;
    }
    return retVal;
}

/**
 * Вычисляет тип группы для передачи этого типа в командах клиенту
 * @param pString какая-то строка
 * @return тип группы
 */
public String getGroupNameForXML(String pString){
    switch (pString) {
        case "TAB":
        return "Tab";
        case "BUNCH":
        return "Bunch";
        case "SIDE_BY_SIDE":
        return "SideBySide";
        case "FIXED":
        return "Fixed";
        case "OVERFLOW":
        return "Overflow";
}
    return "Virtual";
}


/**
 * Закрытие детальных форм для данной формы
 * @param pCloseDetailMode -- этот параметр позволяет предотвратить лишние вызыовы Post-программ при закрытии деталей деталей
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void closeDetails(String pCloseDetailMode) throws LuiException, SQLException{
String actionCode;    
String oldCloseDetailMode=application.getVal("Close Detail Mode");
String oldcurrentFormId=application.getVal("FormID");
application.setVal("Close Detail Mode",NVL(pCloseDetailMode,"N")); //Возможно, включить режим "Закрытие деталей"
String thisFormID="";
String pFid="";
thisFormID=getFormID(); // берём ID данной формы
boolean b=true;
int excCount=0;
while (b){
 b=false;
 try{
 for (String frmId:application.forms.keySet()){ // крутим формы
   if (!(application.forms.get(frmId).getVal("Status").equals("Closing"))){ // если форма не находится в состоянии закрытия
       pFid=application.forms.get(frmId).getVal("ParentFormID");
       if (is_not_null(pFid)){
        if (!application.forms.containsKey(pFid)){ // родительской формы уже нет
           application.forms.get(frmId).close(null,null); //закрываем тогда и дочернюю
        } else { // родитель есть
          if (application.forms.get(pFid).getVal("Status").equals("Closing")){ // но она закрывается
           if (!thisFormID.equals(frmId))   {
           application.forms.get(frmId).close(null,null); //закрываем тогда и дочернюю  если она не та которую закрываем щас
           }
          }  
        }
        
       }
    if (pFid.equals(thisFormID)){ // для каждой формы являющейся деталью данной
       application.forms.get(frmId).closeDetails(pCloseDetailMode); // выполняем закрытие ёё деталей 
      if (is_null(application.forms.get(frmId).getVal("ParentNode"))) { // и не являющейся Раскрытой веткой дерева 
     // а потом её самой, при этом
     application.setVal("FormID",frmId); // временно установим её текущей
     //actionCode=this.checkHotKey(frmId, "CLOSE"); // поищем в ней действие c ролью CLOSE
     actionCode=application.forms.get(frmId).checkHotKey("CLOSE", "Form");
     if (is_not_null(actionCode)){ // если такое действие есть 
        application.forms.get(frmId).runAction(actionCode); // выполняем его
     } else { // если нет - закроем насильно
        application.forms.get(frmId).close(null,null);
     }
     } 
    }
   }
  }
} catch (Exception ex) {
   String m=ex.getMessage();
   if (m!=null) {
       throw ex;
   } else {
       b=true;
       excCount++;
       if (excCount>100)break;
   }
}
}
 application.setVal("Close Detail Mode",oldCloseDetailMode);
 application.setVal("FormID",oldcurrentFormId);
}
/**
 * Подготовка команд закрытия формы и её деталей причём от листьев
 * @return последняя сформированная группа команд
 */
private LuiClCommandGroup makeClosingCommands() throws LuiException, SQLException{
// сначала подчинённые  
    for(String cfrmid:application.forms.keySet()){
        if (getFormID().equals(application.forms.get(cfrmid).getVal("ParentFormID"))){
          if (!application.forms.get(cfrmid).getVal("Status").equals("Closing")){ // если не помечена как закрываемая
           application.forms.get(cfrmid).makeClosingCommands();
          }  
        }  
      }  
    
    setVal("Status","Closing"); // форма закрывается такой процесс 
    LuiClCommandGroup XMLcg=application.commandWriter.addGroup(getFormID()); // Группа команд имя=ID формы
    LuiClCommand XMLcmd=XMLcg.addCommand("CloseForm"); // команда закрытия
    XMLcmd.addScaParameter("FormID", getFormID(),""); // параметр команды    
    return XMLcg;
}
/**
 * Закрытие формы 
     * @param pRole -- роль на которую надо поменять роль действия, вызвавшего данную форму 
     * @param pPostAction -- после действие которое тоже подменяется
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void close(String pRole,String pPostAction) throws LuiException, SQLException{
 if (getProperty("NeverClosing").equals("Y"))return;
//String oldStatus=getVal("Status");
    if (getVal("Status").equals("Closing")){
  // повторное закрытие
 return;  
}
    /*
if (getCode().equals("LUI_B_RT_DASHBOARD")){
    int i=1/0; -- отладка
}*/
if (is_not_null(pPostAction)){
if (getType().equals("TREE")){ //если это дерево
    if (is_not_null(getVal("RootFormID"))){
        if (!getVal("RootFormID").equals(getFormID())){
            application.forms.get(getVal("RootFormID")).close(pRole, pPostAction);
        }
    }
}}
boolean rt=false;    
if (RTdashboard!=null) { // если есть дашборд
  if ("run".equals(RTdashboard.getStatus())) {
  RTdashboard.setStatus("stop");
  RTrunThread.interrupt();
  while (RTdashboard.getStatus().equals("stop")){
      try {
          Thread.sleep(1000);
      } catch (InterruptedException ex) {
          
      }
      
  }
  }
  // гасим все pStmt в дашбоарде
  if (RTdashboard.BdataSources!=null)
  for (String ds:RTdashboard.BdataSources.keySet()){
    if (RTdashboard.BdataSources.get(ds).pStmt!=null && !RTdashboard.BdataSources.get(ds).pStmt.isClosed())
    {
        try{
            RTdashboard.BdataSources.get(ds).pStmt.close();
        } catch (Exception ex){}
    }
  }
  // гасим все соединения в дашбоарде
  if (RTdashboard.Bconnections!=null)
  for(String concode:RTdashboard.Bconnections.keySet()){
    if (application.userDbConnections.containsKey(concode)){
        try {
        application.userDbConnections.get(concode).connPool.SetFree(RTdashboard.Bconnections.get(concode));
        }catch(Exception ex){}
    }  else {
    if (RTdashboard.Bconnections.get(concode)!=null) {
       try{
       RTdashboard.Bconnections.get(concode).disconnect();
       }catch(Exception ex){}
     }
    }
    RTdashboard.Bconnections.put(concode,null);
  }
  RTdashboard.frm=null;
  RTdashboard=null;
  application.setVal("rtChart","");
  if (application.forms.containsKey(getVal("rtParentFormID"))){
  setVal("ParentFormID",getVal("rtParentFormID"));
  } 
  rt=true;
}
if (getType().equals("HOST")){ // если HOST
 return;   
}
int i=0;    
    if (getType().equals("BLANK")){ // если бланк
        for (String elcode:elements.keySet()){
            LuiAppElement el=elements.get(elcode);
            if (el.getType().equals("FIELD")) {
                if (i<1){
                if (application.lastFieldValues==null){
                    application.lastFieldValues=new LuiCollection("");
                } else {
                    application.lastFieldValues.clear();
                }
                }
                i++;
                application.lastFieldValues.Add_Element(elcode, el.properties.get("Value").currentValue);
                application.addProperty("LastFieldValues",application.lastFieldValues.toString());
            }
        }
    }
    LuiClCommand XMLcmd;
    
    setVal("Status","Closing");
    closeDetails("Y");
    LuiClCommandGroup XMLcg=makeClosingCommands();
    if (getVal("MasterForm").equals("Y")){//Если форма была последней в окне - закрыть окно и активировать родительское окно
      XMLcmd=XMLcg.addCommand("CloseWindow");
      XMLcmd.addScaParameter("WindowID",getVal("WindowID"),"");
      if (application.windows.containsKey(getVal("WindowID"))) application.windows.get(getVal("WindowID")).disable(); // почему-то этого сдесь небыло
    }
    String parentFormID=getVal("ParentFormID");
    
    if (is_not_null(parentFormID) && application.forms.containsKey(parentFormID)){//Есть вызывающая форма. Решим что делать с ней
     LuiAppForm pFrm=application.forms.get(parentFormID); // берём мастер-форму
     String callAction=getVal("CallAction");// узнаём в результате какого действия нас позвали
     application.setVal("FormID", parentFormID);// укажем, что текущаяя форма теперь - вызывающая
     String parentWindowID=pFrm.getVal("WindowID"); // код окна вызывающей формы
     if (!parentWindowID.equals(getVal("WindowID"))){// если окна разные
       application.windows.get(parentWindowID).enable();   // почему-то этого сдесь небыло
       XMLcmd=XMLcg.addCommand("EnableWindow");
       XMLcmd.addScaParameter("WindowID", parentWindowID,"");
       XMLcmd=XMLcg.addCommand("ShowWindow");
       XMLcmd.addScaParameter("WindowID", parentWindowID,"");
       //XMLcmd=XMLcg.addCommand("Focus");
       //XMLcmd.addScaPar("FormID",parentFormID);
       application.pingRunTimer();//19.03.2020
     }
     application.commandWriter.removeCommadFromAllGroups("Focus");
       XMLcmd=XMLcg.addCommand("Focus");
       
       XMLcmd.addScaParameter("FormID",NVL(pFrm.getVal("RootFormID"), parentFormID), "");
     
     if (is_not_null(callAction) ) { // если нас вызвали по действию
       application.setVal("FormID", parentFormID);// укажем, что текущаяя форма теперь - вызывающая  
      if ( !code.equals(application.getProperty("ERROR_MESSAGE_FORM"))){
       pFrm.setVal("NewRole", pRole);
       pFrm.setVal("NewAfter", pPostAction);
       pFrm.setVal("CurrentAction", callAction);
       //if (!getVal("DoNotRunPostAction").equals("Y")){
       if (NVL(getProperty("TASK_SHOW_PARAM_FORM"),"Y").equals("Y")) { //если это синхронна задача без видимой формы параметров
       pFrm.runActionExecutePost(callAction); // то это последейсттвие здесь не выполняем, оно будет выполнено в вызывающем действии само иначе будет задуливание действий
       } 
     }
       //}
     } else {
        //19.03.2020 pFrm.checkRunTimer();
     }
     
     if (!parentWindowID.equals(getVal("WindowID"))){// если окна разные
               //DBMS_APPLICATION_INFO.set_action(GetWindowTitle(ParentWindowID));
     }
     
   }else { // нет вызывающей формы
 if (!rt) {
      XMLcmd=XMLcg.addCommand("LogOff");      
      if (is_not_null(application.getProperty("RetValues"))){
          XMLcmd.addScaParameter("Params",application.getProperty("RetValues"),"");
      }
     } 
    }
    
}
/**
 * Действие по умолчанию для узла дерева
 * (клик мышкой в области отображения состояний, клавиши влево-вправо для текущего узла и т.п.)
     * @param pNode -ID узла, с которым производится манипуляция
     * @param pState -текущее состояние узла
     * @param pNType -тип этого узла
     * @param pContext -коллекция контекстных значений
     * @param pContextID -ID рабочей области, к которой относится контекст
     * @param pChildsContextID -???
* @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса 
 */
public void nodeDC(String pNode,String pState,String pNType,String pContext,String pContextID,String pChildsContextID) throws LuiException, SQLException{
  application.changeSCN("NodeDC"); 
  LuiAppForm contextFrm=application.forms.get(pContextID);  
  if (contextFrm==null)return;
  if (contextFrm.getType().equals("MENU")){
      //LuiAppElement mel=this.findAndMute(pNode);
      LuiAppElement mel=null;
      LuiCollection concol=new LuiCollection(pContext);
      if (elements.containsKey(concol.Get_Element("ACTION"))){
          mel=elements.get(concol.Get_Element("ACTION"));
          this.setDynamicMenuItemProperty(mel, concol);
      }
     if (mel!=null){
       if (NVL(mel.getProperty("ROLE"),"DBLCLICK").equals("DBLCLICK")){    
       contextFrm.runAction(mel.getCode());
       application.checkDetails("");
       }
     }
   return;
  }
  if (pNType.equals("ROW")){
   application.setContext4Node(pContextID, pContext, pChildsContextID);   
   String actionCode=contextFrm.checkHotKey("DBLCLICK", "Form");
   if (is_not_null(actionCode)){
       contextFrm.runAction(actionCode);
       application.checkDetails("");
   } else {
       changeState(pNode, pState, pNType, pContext, pContextID,pChildsContextID);
   }
  } else {
      if (pNType.equals("ACTION")){
          contextFrm.setContext4Node(new LuiCollection(pContext), pChildsContextID);
          changeState(pNode, pState, pNType, pContext, pContextID,pChildsContextID);
      }
  }
}

    /**
     * Изменение состояния (раскрыт, схлопнут и т.п.) узла дерева FormID тем или
     * иным способом (клик мышкой в области отображения состояний, клавиши
     * влево-вправо для текущего узла и т.п.)
     *
     * @param pNode -ID узла, с которым производится манипуляция
     * @param pState -текущее состояние узла
     * @param pNType -тип этого узла
     * @param pContext -коллекция контекстных значений
     * @param pContextID -ID рабочей области, к которой относится контекст
     * @param pChildsContextID -???
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void changeState(String pNode,String pState,String pNType,String pContext,String pContextID,String pChildsContextID) throws LuiException, SQLException{
 if (!application.forms.containsKey(pContextID))return;
    application.changeSCN("ChangeState");
    
    LuiAppForm contextFrm=application.forms.get(pContextID);  
    
    LuiCollection SaveOldContext=contextFrm.getFieldsValues(); //-- Тeкущий контекст ContextID. Похраним его до конца действия. Добавлено 07.04.2020
    
    
    application.setContext4Node(pContextID, pContext, pChildsContextID);
    String ActionCode;
    LuiClCommandGroup XMLcg;
    LuiClCommand XMLcmd;
    
    
    switch (pState){
        case "Collapsed": //Раскрытие закрытой ноды
        case "AutoExpand": // И ЭТО ТОЖЕ РАСКРЫТИЕ ЗАКРЫТОЙ НОДЫ  
          if (pNType.equals("ACTION")){ //Раскрытие действия
            ActionCode=Extract_Word(pNode,-1);
            contextFrm.expandAction(ActionCode, pNode);
          } else {//-- Раскрытие строки списка
            contextFrm.expandRow(pNode,pContext);
           // contextFrm.checkFrmDetails();
          } 
        break;
        case "Expanded": //Сворачивание раскрытой ноды
            application.collapseNodeForms(pContextID /*getFormID()*/, pNode);
            
        XMLcg=application.commandWriter.addGroup("ChangeState");
        XMLcg.addParameter("FormID", getFormID());
        XMLcg.addParameter("Node", pNode);
        XMLcg.addCommand("DeleteChilds");
        if (pNType.equals("ACTION")){//Закрытие действия
         ActionCode=Extract_Word(pNode,-1);   
         LuiAppElement ael=contextFrm.findAndMute(ActionCode);
            if (ael!=null){
                   if (ael.getProperty("ACTIVE").equals("EXIST")){//Не потеряло ли действие активность?
                       XMLcmd=XMLcg.addCommand("Node");
                       XMLcmd.addScaPar("Label", LuiString.RemoveIco(ael.getProperty("PROMPT")));
                       XMLcmd.addScaPar("State", "Collapsed");
                       XMLcmd.addScaPar("ChildsContextID","");
                   } else {
                       XMLcg.addCommand("DeleteNode");
                   }
           } else {
                       XMLcg.addCommand("DeleteNode");
            }
        } else{//Закрытие строки
            XMLcmd=XMLcg.addCommand("Node");
            XMLcmd.addScaPar("State", "Collapsed");
            String after="";
            String currentAction=contextFrm.getVal("CurrentAction");
            if (contextFrm.elements.containsKey(currentAction)){
                after=contextFrm.elements.get(currentAction).getProperty("AFTER");
            }
            if (after.equals("CURRENT")){
             String label=contextFrm.getProperty("NODE_TEXT");
             LuiCollection contextUK=contextFrm.getUK();
             LuiCollection contextColl=contextFrm.requeryRow(contextUK);
             if (contextColl.size()>0){//Нода ещё есть в ветке дерева
                 contextFrm.setElementsValues(contextColl);
                 XMLcmd.addScaPar("Label", LuiString.RemoveIco(contextFrm.getProperty("NODE_TEXT")));
                 XMLcmd.addScaPar("Icon", contextFrm.getProperty("NODE_ICON"));
                 XMLcmd.addScaPar("Hint", contextFrm.getProperty("ADD_INFO"));
                 XMLcmd.addScaPar("Context", contextColl.toString());
                 contextFrm.checkFrmDetails();
             } else {
                 XMLcmd.addScaPar("Label", label);
             }
            }else{
              if (contextFrm.getType().equals("MENU")) {
                  LuiAppElement mel=null;
                  LuiCollection concol=new LuiCollection(pContext);
                  if (contextFrm.elements.containsKey(concol.Get_Element("ACTION"))){
                      mel=contextFrm.elements.get(concol.Get_Element("ACTION"));
                      this.setDynamicMenuItemProperty(mel, concol);
                      
                  }
                //LuiAppElement mel=contextFrm.findAndMute(pNode);
                if (mel!=null) {
                    XMLcmd.addScaPar("Label",mel.getProperty("PROMPT"));
                }
              } else
              XMLcmd.addScaPar("Label", LuiString.RemoveIco(contextFrm.getProperty("NODE_TEXT")));  
            }
             XMLcmd.addScaPar("ChildsContextID","");
        }    
        break;
        case "Continue":
        this.changeNode(pNode, pState, pNType, pContext, pContextID, pChildsContextID);
        break;    
    }
  
  contextFrm.setElementsValues(SaveOldContext);//-- Восстановление контекста ContextID. Добавлено 07.04.2020
}
/**
 * Поиск распропагандированного элемента и мутация его selecto-m
 */
LuiAppElement findAndMute(String pActionCode) throws LuiException, SQLException{
  LuiAppElement el=null;
  if (elements.containsKey(pActionCode)) { //если есть такое действие то берём его
  el=elements.get(pActionCode);
  } else {
             for (String elcode:elements.keySet()){
       LuiAppElement bel=elements.get(elcode);
       if (bel.getType().equals("ACTION")){
          if (like(pActionCode,elcode+"-%")) { 
            String AV=bel.getProperty("DISPLAY");
            String hashprompt=LuiString.Extract_Word(pActionCode,-1,"-");  
            if (is_not_null(hashprompt)) {
            if (like(AV,"JS:%")){  // DISPALY содержит JS
          try { 
            jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) application.JSengine.eval(AV.substring(3));
            jdk.nashorn.api.scripting.ScriptObjectMirror props;
            for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
              props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind); // берём нижний массив
              for (String propkey:props.keySet()){ // крутим элементы вложенного массива
                bel.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
              }
              String newhashprompt=bel.getProperty("PROMPT"); // берём prompt
              if (is_not_null(newhashprompt)) { // если не пуст
            newhashprompt=String.valueOf(Math.abs(newhashprompt.hashCode())); // берём его hash
             if (hashprompt.equals(newhashprompt)) {  // сравниваем hash prompta
             el=bel;
             pActionCode=el.getCode();
             break;
             }
              }
            }                    

          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          }

            }
               if (like(AV,"SQL:%")) { // DISPLAY содержит select
                ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4), getProperty("Connection"),this.getFormID());
                LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
                LuiDbConn lc=null;
            try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
                // извлекаем данные из select'а
              ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
              if (propsArray!=null){
              for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                bel.setProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
           
            String newhashprompt=bel.getProperty("PROMPT"); // берём prompt
            if (is_not_null(newhashprompt)) { // если не пуст
            newhashprompt=String.valueOf(Math.abs(newhashprompt.hashCode())); // берём его hash
             if (hashprompt.equals(newhashprompt)) {  // сравниваем hash prompta
             el=bel;
             pActionCode=el.getCode();
             break;
             }
            }
              }
            }
            }finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }
               } 
            }
          }
       }  
      }
  }
  return el;
}

/**
 * Раскрытие узла дерева типа "действие" как бы выполняя действие ActionCode
 * @param pActionCode -- код действия
 * @param pNode --узел
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void expandAction(String pActionCode,String pNode) throws LuiException, SQLException{
    String RootFormID =getVal("RootFormID");
    LuiAppElement el=null;
    el=this.findAndMute(pActionCode);
if (el==null) return;    
pActionCode=el.getCode();
    if (el.getProperty("ACTIVE").equals("EXIST")){//Не потеряло ли действие активность?
//    if (application.getEffectiveRight("UI_ELEMENT",getCode()+"."+pActionCode,"RUN").equals("Y")) { // если доступ есть
    if (application.getEffectiveRight("UI_ELEMENT",getProperty("FEATURE")+"."+pActionCode,"RUN").equals("Y")) { // если доступ есть
      String ActionCommand=el.getProperty("COMMAND");
      if (like(ActionCommand,"%(%:%;)")){ //- параметры могут передаваться коллекцией в скобочках
      int i=ActionCommand.indexOf("(");
      ActionCommand=ActionCommand.substring(0, i);//очистим команду от скобочек
      }
      String params="";
      params=getParamString(pActionCode,params);
      LuiCollection paramsCollection=new LuiCollection(params);
      String ActionConfig=paramsCollection.Get_Element("CONFIG");
      if (is_null(ActionConfig)){
          ActionConfig=getProperty("ConfigCode");
      }
      if (!NVL(ActionConfig,"DEFAULT").equals("DEFAULT")){ //Конфигурацию попробуем  открыть ту же, что и у родителя
         LuiDbConn lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
         String cnt="1";
          try { 
          cnt=lc.getSelectResult("select count(*) from lui_t_config where app_code='"+application.code+
                  "' and form_code='"
                  +ActionCommand+"' and code in('"
                  +ActionConfig+"','DEFAULT')");
         } finally {
             application.metaConnPool.SetFree(lc);
         } 
          if (cnt.equals("1")){
              ActionConfig="DEFAULT";
          } 
          if (cnt.equals("0")){
           application.msgWriter.write("LUI-00015",ActionCommand);
          }
      }
      String NewFormID="FRM"+application.getNextUniqueID();
     // setVal("CurrentAction",pActionCode); //надо наверно удалить
      LuiAppForm newfrm=application.addForm(NewFormID,ActionCommand,"TREE"/*,"Y"*/);
      
      newfrm.setProperty("ConfigCode", ActionConfig);
      application.loadFormFromMetadata( NewFormID, params, "REQUERY");
      
//            PLSQL4ShowLOV(UI_P_Virtual.Get_Element_Property(NewWorkID, 'Form', 'ON_LOAD'));
      newfrm.setVal("ParentFormID",getFormID());
      newfrm.setConnectionProperty(params);
      newfrm.setVal("ParentNode",pNode);
      newfrm.setVal("RootFormID",RootFormID);
      newfrm.setVal("WindowID",getVal("WindowID"));
      newfrm.setVal("CommitMode",application.getProperty("COMMIT"));
      newfrm.setProperty("FType","TREE");
      newfrm.setProperty("SelectionMode","Single");
      //шаблонные элементы!!!
//      newfrm.makeElList("Form");
//Выполняем действие "при загрузке"
 if (newfrm.properties.containsKey("ON_LOAD")){
   LuiAppProp prop=newfrm.properties.get("ON_LOAD");
   String prg=prop.getComputedProperty();
   if (is_not_null(prg)) {
application.log.addRecord("ON_LOAD", application.NLS.NLS("RUS:Выполнение процедуры при загрузке формы ;ENG:Programm on load ;"), prg,newfrm.getFormID());       
       application.parser.execute(prg, prop);
   
   }
 }
 newfrm.addProperty("NeverClosing","N"); //-- это чтобы удалялись поднинённые формы всё таки при схлопывании узлов
 //вычисляем значения параметров и снимаем признак необходимости вычислений, т.е. делаем параметры статическими
                     for (String elKey:newfrm.elements.keySet()){
                         LuiAppElement el2=newfrm.elements.get(elKey);
                         if (el2.getType().equals("PARAM") && el2.getProperty("Parent").equals("Form")){
                             el2.properties.get("Value").getComputedProperty();
                         }
                     }
//-- установка Run для действий списка подветки
 for(String elCode:newfrm.elements.keySet()){
     if (newfrm.elements.get(elCode).getType().equals("ACTION")){ // просматриваем только действия
        if (like(elCode,"SYSMENU$%") || like(elCode,"COLMENU$%")  || like(elCode,"COLMENU$%")|| like(elCode,"CELLMENU$%")){ // если это действие системного меню или меню столбца или QBE меню ячейки
               newfrm.elements.get(elCode).setProperty("Run",application.getEffectiveRight("UI_ELEMENT",newfrm.getProperty("FEATURE")+"."+elCode ,"RUN"));     
             } else {
             if ((getVal("ReadOnly").equals("Y"))) {
              newfrm.elements.get(elCode).setProperty("Run","N");
             } else{
                  newfrm.elements.get(elCode).setProperty("Run",application.getEffectiveRight("UI_ELEMENT",newfrm.getProperty("FEATURE")+"."+elCode ,"RUN"));   
             } }
     }
 }                           
      newfrm.requery();
      //      -- Модификация родительской ноды
      LuiClCommandGroup XMLcg=application.commandWriter.addGroup("EA1");
      XMLcg.addParameter("FormID", RootFormID);
      LuiClCommand XMLcmd=XMLcg.addCommand("Node");
      XMLcmd.addScaPar("Node", pNode);
      XMLcmd.addScaPar("State", "Expanded");
      XMLcmd.addScaPar("ChildsContextID", NewFormID);
      XMLcmd.addScaPar("Label",Jsoup.parse(newfrm.getProperty("HEADER")).text());
    }
}
    
    
    }


    
public void expandRow(String pNode,String pContext) throws LuiException, SQLException{
    if (getType().equals("MENU")){
//     LuiAppElement mel=this.findAndMute(pNode);
     LuiCollection concol=new LuiCollection(pContext);
     LuiAppElement mel = this.elements.get(concol.Get_Element("ACTION"));
     
     if (mel!=null){
        showMenuItems(pNode,pContext);
     }  
    return;    
    }
  String RootFormID=getVal("RootFormID");
  String OnlySelfCommand=null;
  LuiClCommandGroup XMLcg=application.commandWriter.addGroup("Expand_Row");
  LuiClCommand XMLcmd;
  XMLcg.addParameter("FormID", RootFormID);
  XMLcg.addParameter("Type", "ACTION");
  XMLcg.addParameter("Icon", "Catalog");
  XMLcg.addParameter("State", "Collapsed");
  for (int i=0;i<elements.size();i++){
    LuiAppElement el=elSort(i);
    if ((el.getType().equals("ACTION")) && (el.getParent().equals("Form"))){
      if (like(el.getCode(),"SYSMENU$%") ||like(el.getCode(),"COLMENU$%") ||like(el.getCode(),"CELLMENU$%")) continue;  
      if (NVL(el.getProperty("Run"),"Y").equals("Y")) {
         if (el.getProperty("ACTIVE").equals("EXIST")) {
            if (el.getProperty("APPLY_TO").equals("N")) { 
              if (el.getProperty("COMMAND_TYPE").equals("LIST")) {   
                if (is_null(el.properties.get("PROC_INITIALLY").getComputedProperty())){
                  if (is_null(RTrim(LTrim(el.properties.get("ROLE").getComputedProperty().replace("DBLCLICK","")))) ||
                          like(el.properties.get("ROLE").getComputedProperty(),"%ALLEXPAND%")
                          ){
                    if (is_null(el.getProperty("PROC_FINALLY"))){
                      String ActionCommand=el.getProperty("COMMAND");  
                      if (is_not_null(ActionCommand)){
                          String AV; 
                          AV=el.getProperty("DISPLAY");
                        if (like(AV,"SQL:%")) { // в DISPLAY лежит SELECT
          ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4),getProperty("Connection"),this.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          // крутим извлечённые строки
          if (propsArray!=null){
          OnlySelfCommand=" ";     
          for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                el.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("Node");
            XMLcmd.addScaPar("Node",pNode+" "+el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("ParentNode", pNode);
            String ico=LuiString.GetIcoFromSquareBracket(Prompt);
            Prompt=LuiString.RemoveIco(Prompt);
            XMLcmd.addScaPar("Label",Prompt);
            if (is_not_null(ico))
            XMLcmd.addScaPar("Icon", ico);
            if (like(el.getProperty("ROLE"),"%ALLEXPAND%")){
                            XMLcmd.addScaPar("State","AutoExpand");
            }
            }
          }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }
                               } // SELECT
                        else {
                        XMLcmd=XMLcg.addCommand("Node");
                        XMLcmd.addScaPar("Node", pNode+" "+el.getCode());
                        XMLcmd.addScaPar("ParentNode", pNode);
                        String Prompt=el.getProperty("PROMPT");
                        String ico=LuiString.GetIcoFromSquareBracket(Prompt);
                        Prompt=LuiString.RemoveIco(Prompt);
                        XMLcmd.addScaPar("Label",Prompt);
                        if (is_not_null(ico))
                            XMLcmd.addScaPar("Icon", ico);
                        if (like(el.getProperty("ROLE"),"%ALLEXPAND%")){
                            XMLcmd.addScaPar("State","AutoExpand");
                        }
                        if (is_null(OnlySelfCommand)){
                           OnlySelfCommand=el.getCode();
                        } else {
                           OnlySelfCommand=" "; 
                        }
                      }
                      }
                    }  
                  }
                }
              }
            }
         }
      }
    }
  }
  //Доработка для "Короткой" иерархии
  if (is_not_null(RTrim(OnlySelfCommand))){
      application.commandWriter.removeGroup("Expand_Row");
      expandAction(OnlySelfCommand,pNode);
  } else {
      XMLcg=application.commandWriter.addGroup("ER1");
      XMLcmd=XMLcg.addCommand("Node");
      XMLcmd.addScaPar("FormID", RootFormID);
      XMLcmd.addScaPar("Node", pNode);
      if (is_null(OnlySelfCommand)){
          XMLcmd.addScaPar("State", "Leaf");
      } else {
          XMLcmd.addScaPar("State", "Expanded");
      }
          
  }
}
/**
 * Проверка необходимости взведения таймера на клиенте для обработки возможных действий формы WorkID с ACTION_ROLE = TIME
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
/* 19.03.2020 теперь это pingRunTimer в LuiApplication
public void checkRunTimer() throws LuiException, SQLException{
    if (is_null(application.getProperty("LUI Timer"))){
     String Timer=getVal("Timer");
     if (is_null(Timer))Timer="";
     if (Timer.equals("Y"))Timer="1000";
     if (Timer.equals("N"))Timer="";
     if (is_not_null(Timer)) {
       application.setProperty("LUI Timer","Y");
     LuiClCommandGroup XMLcg=application.commandWriter.addGroup("");
     LuiClCommand XMLcmd=XMLcg.addCommand("Event");
     XMLcmd.addScaParameter("EventCode", "Interval","");
     XMLcmd.addScaParameter("MS", Timer,"");
     }   
    }
}*/
/**
 * Очичстка (присвоить пусто) всем полям или столбцам
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void clearElementsValue() throws LuiException, SQLException{
    for (String elCode:elements.keySet()){
        if (elements.get(elCode).getType().equals("FIELD") || elements.get(elCode).getType().equals("ITEM")){
            elements.get(elCode).setValue("", false, "N",false);
            //elements.get(elCode).properties.get("Value").clearComputedValue();
        }
    }
}
/**
 * Нажатие удаления строки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void deleteBar() throws LuiException, SQLException{
    deleteRow();
}
/**
 * Нажатие удаления строки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void deleteRow() throws LuiException, SQLException{
    if (!compareString(getVal("Is_QBE_Mode"),"Y")){
        application.changeSCN("DeleteRow");
        String action=checkHotKey("DELETE", NVL(getVal("CurrentColumn"),getVal("CurrentField")));
        if (action!=null) {
            runAction(action);
            application.checkDetails("");
        }
    }
}

/**
 * Определяет первое активное действие в форме  с ролью pRole (при текущем столбце CurrItem)
 * @param pRole -- роль
 * @param pCurItem -- код текущего элемента или "Form"
 * @return -- код элемента
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public String checkHotKey(String pRole,String pCurItem) throws LuiException, SQLException{
  String action_code=null;  
  String parent_type="";
  if (!getType().equals("GRID"))elSort(0);
  for (int i=0;i<this.elSortedList.size();i++){
    LuiAppElement el=elSortedList.get(i);
    if (el.getType().equals("ACTION")){ // если это действие
     parent_type=el.getParent();
     if (elements.containsKey(parent_type))parent_type=elements.get(parent_type).getType();
     else parent_type="";
     if (el.getParent().equals(pCurItem) || (pRole.equals("CLOSE") && parent_type.equals("GROUP")) ) { // и если оно относится к заданному элементу или имеет роль CLOSE и запрятано в группу ради экранного выравнивания
      if (NVL(el.getProperty("Run"),"Y").equals("Y")) { // и если право выполнения есть  
       if (like(el.getProperty("ROLE"),"%"+pRole+"%")) { // и если роль действия совпадает с заданной
         String active=el.getProperty("ACTIVE");
         if (active.equals("Y")||active.equals("EXIST")){ // и действие активно  
          action_code=el.getCode();   
          break;
         }
       }   
      }
     }
    }
  }
  
  if (action_code!=null) return action_code;
  if (pCurItem.equals("Form")) return action_code;
  if (!elements.containsKey(pCurItem))return action_code;           
  return checkHotKey(pRole,elements.get(pCurItem).getParent());
}
/**
 * Выполняет действие по умолчанию (кнопка Enter)
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void defaultAction() throws LuiException, SQLException {
 LuiAppForm contextForm=this;   
 String role="DBLCLICK";
    if ((NVL(getVal("Is_QBE_Mode"),"N").equals("Y"))){
        role="DBLCLCK_IN_QUERY_MODE";
    }
        application.changeSCN("DefaultAction");
        if (getType().equals("MENU")){
          if (is_not_null(getProperty("CurrentMenuItem"))) {
                  //LuiAppElement mel=this.findAndMute(getProperty("CurrentMenuItem"));
                  LuiCollection concol=new LuiCollection(getProperty("CurrentMenuItemContext"));
                  LuiAppElement mel=null;
                  if (elements.containsKey((concol.Get_Element("ACTION")))){
                  mel=elements.get(concol.Get_Element("ACTION"));
                  }
               if (mel!=null){
                  if (NVL(mel.getProperty("ROLE"),role).equals(role)){
                      
                      this.setDynamicMenuItemProperty(mel, concol);
                      this.runAction(mel.getCode());
                      application.checkDetails("");
                  }
              }
          }  
         return; 
        }
        if (getType().equals("TREE")){
           String ContextFormID=getVal("ContextForkID");
           if (application.forms.containsKey(ContextFormID)){
           contextForm=application.forms.get(ContextFormID);
           } else contextForm=null;
        }
        if (contextForm!=null) {
            String parent=NVL(contextForm.getVal("CurrentColumn"),contextForm.getVal("CurrentField"));
            String elCode=contextForm.checkHotKey(role, NVL(parent,"Form"));
            if (is_not_null(elCode)) {
                contextForm.runAction(elCode);
                application.checkDetails("");
            }
            
        }
       
    //}
}
/**
 * Устанавливает индивидуальное значение pValue свойства pProperty элемента pElement формы,
 * @param pElement -- код элемента
 * @param pProperty -- код св-ва
 * @param pValue --значение св-ва
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void changeVisualParam(String pElement,String pProperty,String pValue) throws LuiException, SQLException{
 
 String Conf=NVL(getProperty("ConfigCode"),"Virtual");
 String UserCfg=getVal("UserConfiguration");
/* это не удачная попытка отработать изменение сортировок
  -- результат: зацикливание т.к. addProperty тоже вызывает changeVisualParam
   if (elements.containsKey(pElement)){
     elements.get(pElement).addProperty(pProperty, pValue);
     if (pProperty.equals("LeftNeighbor")){
         application.setVal("NeedSort",getFormID());
     }
 }*/
 char c=0;
 //String VassalChangedValue=String.valueOf(c);
 String VassalChangedValue=null;
 String VassalElementCode=null;
 String AfterElementCode =null;
 LuiAppElement el;
 boolean b=true;
 if (pProperty.equals("Group")) {
   pProperty="Parent";
 }
 if (pProperty.equals("Parent")) {
   if (is_null(pValue)){
       pValue="Form";
   }  
 }
 //  if GL_P_PARAM.Get_Current_Value('CURRENT_MODE') in ('SYSTEM', 'ADMIN', 'USER') then -- Не работает для режима CLIENT и др.
 /*
 if (pProperty.equals("LeftNeighbor")) {//Если меняем порядок столбцов
   for (int i=0;i<this.elements.size();i++){ // крутим элементы
        el=elSort(i); // берём очередной по -порядку
        if (el.getType().equals("ITEM")){ // если это столбец
         String leftNei=null; // установим что св-во не определено
         if (el.properties.containsKey(pProperty)){ // если у него есть такое
          leftNei=el.getProperty(pProperty); // читаем
         }
            if (compareString(leftNei,pElement)){//Кто-то ранее следовал за перемещаемым элементом el 
                // т.е. код элемента которму меняют соседа совпал со значением свойства "левый сосед" елемента el
                VassalElementCode=el.getCode();//-- Старый хвост перемещаемого элемента Запоминаем код элемента el
            } else{    
             if (compareString(leftNei,pValue)){
                VassalChangedValue=el.getCode();//-- Старый хвост нового босса (B) 
             }
            }
         if (el.getCode().equals(pElement)){
             b=false;
         }
         if (b){
           AfterElementCode=el.getCode();//Старый босс перемещаемого элемента (С)
         }
        }
    }
   if (is_not_null(VassalElementCode)){
      if (elements.containsKey(VassalElementCode)) {
      LuiAppElement vel=elements.get(VassalElementCode);
      vel.setProperty(pProperty, AfterElementCode);
      //setVisualParams(String pFormCode,String pConfigCode,String pElementCode,String pPropertyCode,String pUserCfgID,String pValue,boolean pIns)
      application.setVisualParams(getCode(),Conf,VassalElementCode, pProperty, UserCfg, AfterElementCode, false);
   }
   }
   if (is_not_null(VassalChangedValue)){//Есть старый хвост нового босса
      if (elements.containsKey(VassalChangedValue)){
       LuiAppElement vel=elements.get(VassalChangedValue);   
       vel.setProperty(pProperty, pElement);
       application.setVisualParams(getCode(),Conf,VassalChangedValue, pProperty, UserCfg, pElement, false);
      }
   }
   if (elements.containsKey(pElement)){
     el=elements.get(pElement);
     el.setProperty(pProperty, pValue);
   }
 }//конец Если меняем порядок столбцов
 */
 application.setVisualParams(getProperty("FEATURE"),Conf,pElement, pProperty, UserCfg, pValue, true);
 
}
/** удаление значения QBE
 * 
 */
public void delQBErow() throws LuiException, SQLException {
    if (!getVal("Is_QBE_Mode").equals("Y")) return;
    String cr=getVal("CurrentRow");
    if (listQBE.containsKey(cr)) {
    listQBE.remove(cr);
    }
    HashMap<String,HashMap<String,String>> NewlistQBE=new HashMap<>();
    int crn=Integer.valueOf(cr);
    int rsn=1;
    for (String rs:listQBE.keySet()){
         rsn=Integer.valueOf(rs);
         if (rsn>crn)rsn--;
         NewlistQBE.put(String.valueOf(rsn),listQBE.get(rs));
    }
    listQBE=NewlistQBE;
    if (crn>1 && crn>rsn)cr=String.valueOf(rsn);
    QBEr(cr);
}
public void QBEr(String cr)throws LuiException, SQLException{
    HashMap<String,String> rowQBE;
        
    LuiClCommandGroup XMLcg=application.commandWriter.addGroup("QBE");
    XMLcg.addParameter("FormID",getFormID());
    LuiClCommand XMLcmd=XMLcg.addCommand("Grid");
    //XMLcmd.addScaPar("SelectionMode","Single");
    XMLcmd.addScaPar("SelectionMode","None");
    XMLcmd=XMLcg.addCommand("ClearGrid");
    XMLcmd=XMLcg.addCommand("QBEMode");
    XMLcmd=XMLcg.addCommand("Form");
    setVal("Is_QBE_Mode","Y");
    setVal("CurrentRow",cr);
    // циклим строки для выявления максимума
    int maxRowN=0;
    for (String rnqbe:listQBE.keySet()){
        if (Integer.valueOf(rnqbe)>maxRowN) maxRowN=Integer.valueOf(rnqbe);
    }
    // берём имеющийся максимум
    
    int maxQBERows=Integer.valueOf(NVL(getVal("MaxQBErows"),"1"));
    if (maxQBERows<=maxRowN) { // если старый макимум меньше нового
    for (int i=maxQBERows;i<=maxRowN;i++){
        /*XMLcmd=*/XMLcg.addCommand("AddRow"); // то добавляем строки в грид QBE
    }
    setVal("MaxQBErows",String.valueOf(maxRowN));
    }
    
        // циклим строки
      for (String rnqbe:listQBE.keySet()){
       rowQBE=listQBE.get(rnqbe);
           // циклим столбцы
       for (String ec:rowQBE.keySet()){
           XMLcmd=XMLcg.addCommand("Cell");
           XMLcmd.addScaParameter("RowNum",rnqbe,"");
           XMLcmd.addScaParameter("Code",ec,"");
           XMLcmd.addScaParameter("Show",rowQBE.get(ec),"");
       }
      }
     XMLcmd=XMLcg.addCommand("CurrentRow");
     XMLcmd.addScaParameter("FormID", getFormID(),"");
     XMLcmd.addScaParameter("RowNum",cr,"");
     XMLcmd.addScaParameter("Cell",getVal("CurrentColumn"),"");
     clearElementsValue();
     setUnderside();
}
/**
 * Перевод списка в QBE-режим
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void QBE() throws LuiException, SQLException{
    if (getVal("Is_QBE_Mode").equals("Y")){
       listQBE.clear();
    }
    QBEr("1");
    // String qbe=application.toJSON(XMLcg);
}
/**
 * Обработка события InvertChecking
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void invertChecking() throws LuiException, SQLException{
    application.changeSCN("InvertChecking");
 /*   if (elements.containsKey("QUERY4PRE-MARKED$")){
      LuiAppElement el=elements.get("QUERY4PRE-MARKED$");
      if (el.getParent().equals("Form")){
          if (el.getType().equals("PARAM")){
              el.setProperty("Value","");
          }
      }
    }*/
    String inv=NVL(getVal("Inversion"),"N");
    if (inv.equals("Y")) {
    setVal("Inversion","N");
    } else {
    setVal("Inversion","Y");    
    }
    for (LuiCollection keycol:this.checkRows.keySet()){
      boolean b=checkRows.get(keycol);
      checkRows.replace(keycol, !b);
    }
    
     if (properties.containsKey("CountSelected")){
     properties.get("CountSelected").clearComputedValue(false);
     }
     if (properties.containsKey("ItemsSelectedJSON")){
     properties.get("ItemsSelectedJSON").clearComputedValue(false);    
     }
     if (properties.containsKey("RowsSelectedJSON")){
     properties.get("RowsSelectedJSON").clearComputedValue(false);    
     }
    checkFrmDetails();
}

/**
 * Обработка нажатия разных функциональных клавиш клавиатуры
 * @param pEv клиентское событие
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void fBar(LuiClEvent pEv) throws LuiException, SQLException{
 String role=null;
 String actionCode;
 String pKey=pEv.name;
 application.changeSCN(pKey);
    switch (pKey){
        case "F4Bar":role="DUPLICATE";break;
        case "InsertBar": case "F6Bar":role="ADD";break;
        case "QBE":case "F7Bar":role="QBE";break;
        case "F8Bar":role="EXECUTEQUERY";break;
        case "F9Bar":role="F9";break;
        case "F1Bar":role="F1";break;
        case "F5Bar":role="F5";break;
        case "ShiftF6Bar":role="DELETE";break;
}
 if (role==null) return;
String curItem="Form";
    switch (getType()){
        case "LIST":
        if (getProperty("Var Is_QBE_Mode").equals("Y") && 
                       !role.equals("QBE") && 
                       !role.equals("EXECUTEQUERY")&&
                       !role.equals("DELETE") &&
                       !role.equals("F5")&& 
                       !role.equals("F1") && 
                       !role.equals("F1") && 
                       !role.equals("F9") && !role.equals("DUPLICATE")
                ) return;
        curItem=getVal("CurrentColumn");
        break;
        case "BLANK":curItem=getVal("CurrentField");break;
        case "TREE":curItem="Form";break;
    }
    actionCode=checkHotKey(role, NVL(curItem,"Form"));
    if (is_not_null(actionCode)){
    if (role.equals("F5")) {
        if (application.getProperty("DUMP_KEY_DISABLE").equals("Y")){
          if (elements.containsKey(actionCode)){
             if (like(elements.get(actionCode).properties.get("PROC_INITIALLY").getNonCompValue(),"Statistic:%") ||
                 like(elements.get(actionCode).properties.get("PROC_FINALLY").getNonCompValue(),"Statistic:%")    ){
                 return;
             } 
          }
        }
    }
    runAction(actionCode);
    } else {
    if (role.equals("F5") && !application.getProperty("DUMP_KEY_DISABLE").equals("Y")) {
        application.parser.plugins.get("Statistic").execute(application.NLS.NLS("RUS:Дамп ;ENG:Memory Dump ;")+application.code, null);
    }
    if (role.equals("F1")) {
        application.parser.plugins.get("Help").execute(getProperty("HEADER"), properties.get("HEADER"));
    }
    }
}

    /**
     * Обработка нажатия иконки в поле
     *
     * @param pIconName - имя иконки
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void fieldButton(String pIconName) throws LuiException, SQLException{
application.changeSCN("FieldButton");
String cf=getVal("CurrentField");
String action=null;
if (elements.containsKey(cf)){
  /* if (elements.get(cf).getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR")) {
       if (elements.containsKey(pIconName)){
         if (elements.get(pIconName).getParent().equals(cf)){
             action=pIconName;
         }  
       }
   } else {*/
   LuiCollection buttons=new LuiCollection(elements.get(cf).getProperty("Buttons"));
   if (!buttons.containsKey(pIconName)){
      for (String ac:elements.keySet()){
        if (elements.get(ac).getType().equals("ACTION")){
          if (elements.get(ac).getParent().equals(cf)){
            String icon=elements.get(ac).getProperty("ICON");
            buttons.Add_Element(icon, ac);
          }
        }  
      }
   }
   if (buttons.containsKey(pIconName)){
     for (int i=0;i<elements.size();i++){
       LuiAppElement el=elSort(i);
       if (el.getType().equals("ACTION")){
         if (el.getParent().equals(cf)){
           if (el.getProperty("ICON").equals(pIconName)){
             action=el.getCode();
             break;
           }  
         }
       }
     }
   }
//}
}
if (is_not_null(action)){
    runAction(action);
}
}
/**
 * Обработка нажатия иконки тулбара
 * @param pIconName имя иконки
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void toolBarButton(String pIconName) throws LuiException, SQLException{
   application.changeSCN("ToolBarButton");
   if (is_null(pIconName))return;
   //if (getVal("Is_QBE_Mode").equals("Y"))return;
   if (!getType().equals("GRID"))elSort(0);
   for (int i=0;i<elSortedList.size();i++) {
     if (elSortedList.get(i).getType().equals("ACTION")){
       LuiAppElement el=elSortedList.get(i);
       if (getVal("Is_QBE_Mode").equals("Y") && (!like(el.getCode(),"SYSMENU$%"))&& (!like(el.getCode(),"COLMENU$%")) && (!like(el.getCode(),"CELLMENU$%"))) continue;
       if (el.getParent().equals("Form")){
        String Icon="";   
        if (getType().equals("GRID") ||getType().equals("TREE")) {
         Icon=LuiString.GetIcoFromSquareBracket(el.getProperty("BUTTON"));
        }
        else Icon=el.getProperty("ICON");
        if (Icon.equals(pIconName)){
           String AV=el.getProperty("DISPLAY");
                if (AV.indexOf("SQL:", 0)>-1) { // в DISPLAY размножитель SELECT
          ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4),getProperty("Connection"),this.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(this.application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          if (propsArray!=null){
          if (propsArray.size()>0) {
          LuiClCommandGroup XMLcg=application.commandWriter.addGroup("ToolBarButton"); 
          XMLcg.addParameter("FormID",getFormID());
          XMLcg.addParameter("Type","Action");
          LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
              // крутим извлечённые строки
              for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                el.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
            if (!compareString(el.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N"); 
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          }
          
           } finally {
                  if (lc!=null) {
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }         
                } else if (AV.indexOf("JS:", 0)>-1) { // в DISPLAY размножитель JS
                         try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.application.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          LuiClCommandGroup XMLcg=application.commandWriter.addGroup("ToolBarButton"); 
          XMLcg.addParameter("FormID",getFormID());
          XMLcg.addParameter("Type","Action");
          LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");

          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного массива
           el.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
            if (!compareString(el.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          } 
                    
                }  else { // нет размножителя
             if (el.isActionActive()){
                 runAction(el.getCode());
                 if (!getVal("Is_QBE_Mode").equals("Y")) {  
                 //application.checkDetails(""); -- 03/09/2020 из-за странного глюка в диаграммвх
                 el.getForm().checkFrmDetails();
                 }
                 break;
             }
           }
        }   
       }
     }
   }
}
/**
 * Меняет текущую конфигурацию списка на заданную пользовательскую или на базовую (0)
 * Кроме того, если текуща форма - это chart группировки данных, то выполняется смена
 * отображаемого агрегатного столбца
 * @param pUsrConfig пользовательская конфигурация
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void changeConfig(String pUsrConfig) throws SQLException, LuiException{
    
    if (getType().equals("CHART")){
        if (this.DgV2!=null && is_not_null(this.DgV2.CurCfg)) pUsrConfig=NVL(pUsrConfig,this.DgV2.CurCfg);
        else pUsrConfig=NVL(pUsrConfig,"0");
      if (elements.containsKey("P_DG_ID")) {
       String dg_id=elements.get("P_DG_ID").getValue();
       if (elements.containsKey("P_LEVEL")){
        String level=elements.get("P_LEVEL").getValue(); 
       // setVal("UserChartDGConfiguration",pUsrConfig);
        LuiDbConn lc=null;
         try{
         lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
         
         if (this.DgV2!=null){
          String cfg="CFG3D";
          if (DgV2.chartType.equals("MAP") ||DgV2.chartType.equals("PIE")  ||DgV2.chartType.equals("WHR")){
              setVal("UserChartDGConfiguration",pUsrConfig);
              cfg="CFG";
              if (elements.containsKey("P_SERIES_ID")){
              LuiAppElement el=elements.get("P_SERIES_ID");
              if (el.getType().equals("PARAM") && el.getParent().equals("Form")){
                  if (is_not_null(el.getProperty("Value"))){
                     el.addProperty("Value", pUsrConfig);
                  }}}
          //this.DgV2.makeFormParameters4Titles();
          } else { // сохранение состояния делаем для матрёшка/стопка но для карты и пирога с выбором кривой
          lc.getSelectResult("select lui_r_dg_store_state("+dg_id+","+level+")");
          }
          lc.executeDML("delete from lui_t_data_group_chart_pref_val where prop_code='"+cfg+"' and grouping_id="+dg_id+" and chart_seq="+level);
          lc.executeDML("insert into lui_t_data_group_chart_pref_val(grouping_id,chart_seq,prop_code,value) values("+dg_id+","+level+",'"+cfg+"','"+pUsrConfig+"')");        
          
          
          //this.DgV2.get("Y");ssdfsd
          
          this.chart(null);
         } else {
         lc.executeDML("update lui_t_data_group_chart_pref set group_function_seq="+pUsrConfig+
                 " where grouping_id="+dg_id+" and seq="+level);
         this.reloadForm();
         }
                  
                  
         } finally{
             if (lc!=null){
               application.metaConnPool.SetFree(lc);
             }
         }
       }
       return;  
      }  
    }
    String ParentFormID = getVal("ParentFormID");
    application.changeSCN("ChangeConfig");
    if (NVL(pUsrConfig,"0").equals("0")){
        application.loadFormFromMetadata(getFormID(),getVal("TransitParam"),"REFRESH");
        setVal("ParentFormID",ParentFormID);
        setVal("UserConfiguration",pUsrConfig);
    } else { //-- Что-то непонятное тут
        loadUserConfig(pUsrConfig);
        
    }
    reloadForm();
}
/**
 * Передача изменений на клиент
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void changes() throws LuiException, SQLException{
 String FType=getType();
 
 LuiClCommandGroup XMLcg;
 LuiClCommand XMLcmd = null;
 XMLcg=application.commandWriter.addGroup("Changes "+getFormID());
 XMLcg.addParameter("FormID",getFormID());
 if (!(FType.equals("TREE"))){

    String title=null;
     if (properties.containsKey("HEADER")) {
         if (properties.get("HEADER").changed()) {
            title=Jsoup.parse(getProperty("HEADER")).text(); 
         }
     }    
     if (properties.containsKey("Var QBE_Title")){
        if (properties.get("Var QBE_Title").changed()){
            title=title+" "+getProperty("Var QBE_Title");
        }
      }
     
     if (title!=null){
       XMLcmd=XMLcg.addCommand("Form"); 
       if (!getVal("Title").equals(this.getFormTtle()+getProperty("Var QBE_Title"))) {
           
       XMLcmd.addScaPar("Title",this.getFormTtle()+getProperty("Var QBE_Title"));
       setVal("Title",this.getFormTtle()+getProperty("Var QBE_Title"));
     }
     }
 } else {
     //if (application.getVal("FormID").equals(getFormID())) // обновим тулбар но только для текущего списка дерева
      prepareTB4TREE(XMLcg);  
 }
 LuiCollection ToolBarButtons;
 int toolBarDecade=0;
 switch (FType){
     case "LIST":
        ToolBarButtons=makeSysToolBarBattonCollection();
        properties.remove("Var rtEnable");//удалим признак доступности RT-curve
        properties.remove("Var chartEnable");//удалим признак доступности простых диаграмм
      if (!compareString(getVal("Is_QBE_Mode"),"Y")){ // если щас не QBE
       int columnsReadable=0;  
       int columnsReadableNumber=0;  
        for (int i=0;i<elements.size();i++) { //крутим элементы по порядку
          LuiAppElement el=elSort(i);
          switch(el.getType()){
              case "ACTION": // если это действие
                  //???????????? тут сто-то менялось в Fastcom после 23.04.2018
      if (!like(el.getCode(),"SYSMENU$%") && !like(el.getCode(),"COLMENU$%") && !like(el.getCode(),"CELLMENU$%")) { // не системного меню 
          if (NVL(el.getProperty("Run"),"Y").equals("Y")){ // ЕСЛИ ЕГО МОЖНО ВЫПОЛНЯТЬ                    
                if (el.getParent().equals("Form")){ // уровня формы
                    String icon=el.getProperty("ICON");
                    if (is_not_null(icon)){ // c иконкой
                      if (ToolBarButtons.containsKey(icon)) { // и оно есть уже в коллекции
                          //ToolBarButtons.Add_Element(icon,"");// стираем ей хинт
                          this.addIcon2ToolBar(icon, "", "", ToolBarButtons);
                      } else { // если же ещё нет
                          // если её ещё в тулбаре нет - выводим её хинт
                            // но перед выводом сравним декады
                            if ((el.sourceSeq/10)!=toolBarDecade){ // если декады разные ставим сепаратор
                             String keySep="";
                             for(int j=0;j<toolBarDecade;j++) keySep=keySep+" ";
                             //ToolBarButtons.Add_Element(keySep,"");  
                             this.addIcon2ToolBar(keySep, "", "", ToolBarButtons);
                             toolBarDecade=(int)el.sourceSeq/10;
                            }
                             //ToolBarButtons.Add_Element(el.getProperty("ICON"),el.getProperty("EXPLANATION"));  
                             this.addIcon2ToolBar(el.getProperty("ICON"),el.getProperty("PROMPT"),el.getProperty("EXPLANATION"), ToolBarButtons);
                      } 
                    }
                    if (is_not_null(el.properties.get("BUTTON").getNonCompValue())){//Действие может вызываться кнопкой в подвале
                       XMLcmd=XMLcg.addCommand("Button");XMLcmd.really=false; //готовим возможную команду для отображения кнопки
                       XMLcmd.addScaPar("Code",el.getCode());
                       addCommandParam(XMLcmd, el.getCode(),"BUTTON", "Label", true);//Отправим надпись на кнопке, если она изменилась
                       addCommandParam(XMLcmd, el.getCode(),"EXPLANATION", "Hint", true);
                       if (el.properties.get("BUTTON").changed() || //Если изменилась надпись на кнопке
                           el.properties.get("ACTIVE").changed() || //или активность кнопки
                           properties.get("Var CurrentRow").changed() || //или текущая строка
                           properties.get("CountSelected").changed() || //или количество выделенных строк
                           properties.get("SelectionMode").changed()){ //или режим выделения строк
                         if (is_null(el.getProperty("BUTTON"))) { // если кнопка скрыты
                           XMLcmd.addScaPar("State", "Hidden");XMLcmd.really=true; // скрыть
                         } else { // если не скрыта
                             if (el.isActionActive()) { // и активна
                                 XMLcmd.addScaPar("State", "Active");XMLcmd.really=true; // Активировать, если не скрыта, но активна
                             } else {
                                 XMLcmd.addScaPar("State", "Disable");XMLcmd.really=true; // Задизейблить, если не скрыта, но не активна
                             }
                         }
                       }
                    }
                }
                }
          }
              break;  
              case "GROUP":// если группа
              if (el.properties.get("TITLE").changed() || el.properties.get("GROUP_HINT").changed()){
                XMLcmd=XMLcg.addCommand("Group"); 
                XMLcmd.addScaPar("Code", el.getCode());
                XMLcmd.addScaPar("Prompt",el.getProperty("TITLE"));
                XMLcmd.addScaPar("Hint",el.getProperty("GROUP_HINT"));
              }
              break;
              case "ITEM": // если столбец
              el.setRTproperty();
              if (el.getProperty("Read").equals("Y")){ //только для доступных пользователю столбцов
               columnsReadable++;
               if (el.getProperty("DATA_TYPE").equals("NUMBER")){
               columnsReadableNumber++;    
               }
               XMLcmd=null;
               if (el.properties.get("SORT").changed()){ //Если изменилось свойство SORT...
                 XMLcmd=XMLcg.addCommand("Column"); //Отправим клиенту новый порядок в сортировке столбца
                 XMLcmd.addScaPar("Code",el.getCode());
                 XMLcmd.addScaPar("Sort",el.getProperty("SORT"));
               }
               //if (el.properties.get("WIDTH").changed()){ // Если изменилась видимость
               //гггг    
               //}
               if (el.properties.get("VISIBLE").changed() || el.properties.get("WIDTH").changed()){ //Если изменилась доступность столбца...
                 if (XMLcmd==null){
                 XMLcmd=XMLcg.addCommand("Column"); 
                 XMLcmd.addScaPar("Code",el.getCode());
                 }  
                 //XMLcmd.addScaPar("Enable","N"); //... Скроем столбец???
                 if (el.getProperty("VISIBLE").equals("Y"))
                 XMLcmd.addScaPar("Enable","Y");
                 else
                 XMLcmd.addScaPar("Enable","N");
                 String tstr=this.getVisualPref(el.getCode(),"ColWidth", null);
                    if (tstr==null){
                        if (el.getProperty("WIDTH").equals("0")) tstr="0";
                        if (el.getProperty("WIDTH").equals("6")) tstr="6";
                    } 
                 if (el.getProperty("WIDTH").equals("Y")) tstr="";// плевать на предпочтения если Y или 
                 if (el.getProperty("WIDTH").equals("N")) tstr="0";// N
                    XMLcmd.addScaPar("Width",tstr);
               }
               if (el.properties.get("CAPTION").changed()){ // Если изменился заголосок столбца...
                 if (XMLcmd==null){
                 XMLcmd=XMLcg.addCommand("Column"); 
                 XMLcmd.addScaPar("Code",el.getCode());
                 }  
                 XMLcmd.addScaPar("Title",LTrim(el.getProperty("CAPTION"))); 
               }
              }
              break;
          }
        }
       if (columnsReadable>1 && columnsReadableNumber>0)setVal("chartEnable","Y"); 
      }  // конец если щас не QBE
      else { // если QBE 
          for (int i=0;i<elements.size();i++) { //крутим элементы по порядку
          LuiAppElement el=elSort(i);
          if (el.getType().equals("ITEM")){
              el.properties.get("VISIBLE").recalculate=true;
              String oldv=el.properties.get("VISIBLE").currentValue;
              String newv=el.getProperty("VISIBLE");
           if (!compareString(oldv,newv) || el.properties.get("WIDTH").changed()){ //Если изменилась доступность столбца...
                 
                 XMLcmd=XMLcg.addCommand("Column"); 
                 XMLcmd.addScaPar("Code",el.getCode());
                 if (newv.equals("Y"))
                 XMLcmd.addScaPar("Enable","Y");
                 else
                 XMLcmd.addScaPar("Enable","N"); 
                 String tstr=this.getVisualPref(el.getCode(),"ColWidth", null);
                    if (tstr==null){
                        if (el.getProperty("WIDTH").equals("0")) tstr="0";
                        if (el.getProperty("WIDTH").equals("6")) tstr="6";
                    } 
                 if (el.getProperty("WIDTH").equals("Y")) tstr="";// плевать на предпочтения если Y или 
                 if (el.getProperty("WIDTH").equals("N")) tstr="0";// N
                    XMLcmd.addScaPar("Width",tstr);
               }
      }
 }
          String rown=getVal("CurrentRow"); //берём код строки
          if (listQBE.containsKey(rown)){ // если в QBE есть такая строка
           HashMap<String,String> rowQBE;
           rowQBE=listQBE.get(rown);
           String elcode=getVal("CurrentColumn"); //берём код столбца
           if (rowQBE.containsKey(elcode)) { // и если в QBE такой есть посылаем его содержимое
             XMLcmd=XMLcg.addCommand("Cell");
             XMLcmd.addScaPar("Code", elcode);
             XMLcmd.addScaPar("RowNum", rown);
             XMLcmd.addScaPar("Show",rowQBE.get(elcode));
           }
          }
      }
    // if (!compareString(ToolBarButtons.toString(),getVal("ToolBarButtons"))) {//Если построенная коллекция кнопок тулбара отличается от того, что было
       if (!compareString(this.getToolBarString(ToolBarButtons),getVal("ToolBarButtons"))) {//Если построенная коллекция кнопок тулбара отличается от того, что было
     
     XMLcmd=XMLcg.addCommand("Form");
     if (!getProperty("ToolBar").equals("N"))
     XMLcmd.addScaPar("ToolBarButtons", this.getToolBarString(ToolBarButtons));
     setVal("ToolBarButtons",this.getToolBarString(ToolBarButtons));
     }
     setUnderside();
     break;        
     case "BLANK":
        for (String elcode:elements.keySet()) { // крутим элементы
        LuiAppElement el=elements.get(elcode);
       switch (el.getType()){
           case "FIELD": // если это поле
            /*//06.11.2018
               if (el.properties.get("RENOVATION").changed()){ // 02/11/2018
               el.setValue(el.getProperty("RENOVATION"),false,"N");// 02/11/2018
             }// 02/11/2018
            *///06.11.2018   
             if (el.getProperty("Read").equals("Y")){ //только для доступных пользователю полей
               XMLcmd=XMLcg.addCommand("Field");
               XMLcmd.really=false;
               XMLcmd.addScaPar("Code", elcode);
               String s1=application.getVal("SCN name")+el.getCode();
               String s2="EditValue"+getVal("CurrentField");
               if (!compareString(application.getVal("SCN name")+el.getCode(),"EditValue"+getVal("CurrentField"))){
                   //сюда не попадём, если с клиента поступило промежуточное значение ввода пользователем в текущее поле
                /*if ("PROMPT".equals(el.getCode())) {
                    //    String here="here";
                }*/
              if (el.properties.containsKey("StartCursor") && el.properties.containsKey("EndCursor")){
                if (el.properties.get("StartCursor").changed() || el.properties.get("EndCursor").changed()){
                        if (el.getProperty("StartCursor").equals(el.getProperty("EndCursor")))                           
                        XMLcmd.addScaPar("Cursor",el.getProperty("StartCursor"));XMLcmd.really=true;
                }
              } 
              /*String oldShow="";
              if (el.properties.containsKey("Show")){
              oldShow=el.properties.get("Show").currentValue;
              } */
                if (el.properties.get("CHANGING").changed() || //Если изменился способ ввода данных
                    el.properties.get("SHOW").changed() || //или отображаемое значение
                    el.properties.get("VARIANTS").changed() || //или набор возможных значений
                    el.properties.get("Value").changed() || //или просто текущее значение поля
                    properties.get("Var CurrentField").changed()  // или просто переставили курсор в ...
                && (compareString(el.getCode(),lastCurrentField) || //... это или из этого поля
                        compareString(getVal("CurrentField"),elcode) )        
                        ){
                 /*String newShow= el.getFieldShow();
                 if (!newShow.equals(oldShow)) {
                    XMLcmd.addScaPar("Show", el.getFieldShow());XMLcmd.really=true;
                 }*/
                 //вот этот фрагмент был добавлен 23.04.2020 коронавирус
                 // по какм-то причинам иногда у поля бланка к этому моменту в свойстве Show
                 // уже changeFlag=true и если Show вычислить опять, то флаг станет false
                 // поэтому мы сначала этот флаг запомним в переменной sh
                 // а потом уж будем вычисляь
                 // и при принятии решения о выдаче на клиента будем анализировать и sh тоже
                 boolean sh=false;
                 if (el.properties.containsKey("Show")){
                     sh=el.properties.get("Show").changeFlag;
                 } 
                 // конец добавления
                  String newShow= el.getFieldShow();
                  if (el.properties.get("Show").changed() || sh ||
                      (el.properties.get("Value").changeFlag && 
                          el.getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR"))
                          ) {
                      // Это слайдер?
                           LuiCollection sliderCol=null;
               if (like(el.getProperty("VARIANTS"),"LOVR:%")){
                 sliderCol=new LuiCollection(el.getProperty("VARIANTS").substring(5));
                 if (!(sliderCol.containsKey("Min") && sliderCol.containsKey("Max") && sliderCol.containsKey("Step") && sliderCol.containsKey("ValueFrom"))){
                     sliderCol=null;
                 } 
               }
               if (sliderCol== null)
                      XMLcmd.addScaPar("Show",newShow);
               else  XMLcmd.addScaPar("ValueFrom", newShow);
                XMLcmd.really=true;
                  }
                    
                }   
               }
              
             if (el.properties.get("VERIFIER").changed()){
                 if (is_null(el.getProperty("VERIFIER")))
                     XMLcmd.addScaPar("Verifier","");
                 else XMLcmd.addScaPar("Verifier", "fieldVerify."+el.getProperty("VERIFIER"));
                 XMLcmd.really=true;
             }
             if (el.properties.get("VARIANTS").changed() || el.properties.get("VISIBLE").changed() || el.properties.get("CHANGING").changed() 
                     || el.properties.get("DATA_TYPE").changed()
                     || el.getProperty("DATA_TYPE").equals("WIKITEXT")
                     //|| el.properties.get("VISUALIZATION").changed() // добалено для обеспечения подсветки
                     ){ //если изменилась видимость или способ ввода        
                 if (!el.isVisible(false)){ //видимость стала нулевая
                    XMLcmd.addScaPar("Type","Hidden");XMLcmd.really=true; //отправим команду скрыть поле
                    el.setVal("Type","Hidden");
                } else { //поле видимо
                  if (el.getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR")) {//если цвет
                 XMLcmd.addScaPar("Type","Color"); //скажем клиенту об этом
                  }
                 else {// если не цвет
                   LuiCollection sliderCol=null;
               if (like(el.getProperty("VARIANTS"),"LOVR:%")){
                 sliderCol=new LuiCollection(el.getProperty("VARIANTS").substring(5));
                 if (!(sliderCol.containsKey("Min") && sliderCol.containsKey("Max") && sliderCol.containsKey("Step") && sliderCol.containsKey("ValueFrom"))){
                     sliderCol=null;
                 } 
               }
               if (sliderCol!=null) { //если слайдер LOVR:Min:;Max:;Step:;
                 XMLcmd.addScaPar("Type","Range");   
                 if (el.getProperty("CHANGING").equals("N")){
                   XMLcmd.addScaPar("Mode","Readonly");                
                 }
                 if (!sliderCol.Get_Element("Min").equals(el.getVal("OldMin")))
                 XMLcmd.addScaPar("ValueMin", sliderCol.Get_Element("Min"));
                 if (!sliderCol.Get_Element("Max").equals(el.getVal("OldMax")))
                 XMLcmd.addScaPar("ValueMax", sliderCol.Get_Element("Max"));
                 if (!sliderCol.Get_Element("Step").equals(el.getVal("OldStep")))
                 XMLcmd.addScaPar("ValueStep", sliderCol.Get_Element("Step"));
                 if (!sliderCol.Get_Element("ValueFrom").equals(el.getVal("OldValueFrom")))
                 XMLcmd.addScaPar("ValueFrom", sliderCol.Get_Element("ValueFrom"));
                 setVal("OldMin",sliderCol.Get_Element("Min"));
                 setVal("OldMax",sliderCol.Get_Element("Max"));
                 setVal("OldStep",sliderCol.Get_Element("Step"));
                 setVal("OldValueFrom",sliderCol.Get_Element("ValueFrom"));
               } else { // не слайдер     
                   String Type="Text";
                   String Mode="Editable";
                   String KeyNavigable="Y";
                   if (el.getProperty("DATA_TYPE").equals("RICHTEXT")){ //это поле с картинками
                    Type="RichText";
                   // XMLcmd.addScaPar("Type","RichText");XMLcmd.really=true; //скажем клиенту об этом
                  } else { //это обычное поле
                     String chan=el.getProperty("CHANGING");
                     if (chan.equals("Y")||chan.equals("C")||chan.equals("U")||chan.equals("H")||chan.equals("B")||chan.equals("R")){ // оно не ReadOnly
                    KeyNavigable="Y";
                    //XMLcmd.addScaPar("KeyNavigable","Y");XMLcmd.really=true; //туда можно прийти табом
                     } else { //это поле ReadOnly
                    KeyNavigable="N";     
                    //XMLcmd.addScaPar("KeyNavigable","N");XMLcmd.really=true; //туда запрет прийти табом 
                    
                     }
                      String tstr="ReadOnly";
                      switch (el.getProperty("CHANGING")){
                     case "Y": tstr="Text";break;
                     case "U": tstr="Upper";break;
                     case "H": tstr="Passwd";break;
                     case "B": tstr="CheckBox";break;
                     case "R": tstr="Radio";break;
                                                        }
                     if (tstr.equals("ReadOnly")){ //Если поле ReadOnly или выбор только из LOV
                      Type="Text";
                      /*XMLcmd.addScaPar("Type","Text");XMLcmd.really=true;*/   //тип - текстовое значение
                      if (!el.getProperty("DATA_TYPE").equals("WIKITEXT")){Mode="Readonly";/*XMLcmd.addScaPar("Mode","Readonly");*/}
                      else {Mode="Wiki";/*XMLcmd.addScaPar("Mode", "Wiki");*/}
                      /*XMLcmd.really=true;*/   //Скажем клиенту, чтобы не разрешал портить поле руками
                     } else { //Поле типа Свободный ввод просто или в верхнем регистре, пароль, чекбокс
                      Type=tstr;
                      //XMLcmd.addScaPar("Type",tstr);XMLcmd.really=true; //тип - Text, Upper, Passwd или CheckBox
                      if (!el.getProperty("DATA_TYPE").equals("WIKITEXT")){Mode="Editable";/*XMLcmd.addScaPar("Mode","Editable");*/}//Скажем клиенту, чтобы разрешал портить поле руками
                        else {
                       if (el.Code.equals(getVal("CurrentField"))){Mode="Editable";/*XMLcmd.addScaPar("Mode","Editable");*/}
                       else {Mode="Wiki";/*XMLcmd.addScaPar("Mode","Wiki");*/}
                      }
                      
                      //XMLcmd.really=true;   
                     }
                  } 
               if (!el.getVal("Type").equals(Type)){
                   XMLcmd.addScaPar("Type",Type);XMLcmd.really=true;el.setVal("Type",Type);
               }
               if (!el.getVal("Mode").equals(Mode)){
                   XMLcmd.addScaPar("Mode",Mode);XMLcmd.really=true; el.setVal("Mode",Mode);
               }
               if (!el.getVal("KeyNavigable").equals(KeyNavigable)){
                   XMLcmd.addScaPar("KeyNavigable",KeyNavigable);XMLcmd.really=true;el.setVal("KeyNavigable",KeyNavigable);
               }
               
               }//не слайдер
               
                         } // если не цвет
                }
             } // конец если изменилась видимость или способ ввода  
             addCommandParam(XMLcmd, elcode, "CAPTION", "Prompt",true);//Отправим Подпись к полю, если она изменилась
             addCommandParam(XMLcmd, elcode, "HINT", "Hint",true);//Отправим Подсказку к полю, если она изменилась
             addCommandParam(XMLcmd, elcode, "HELPER_TEXT", "HelperText",true);//Отправим Подсказку к полю, если она изменилась
             addCommandParam(XMLcmd, elcode, "PATTERN", "Pattern",true);//Отправим шаблон Автоперехода, если он изменился
//05.11.2020 if (el.properties.get("VISUALIZATION").changed() ||el.properties.get("MAKEUP").changed() ||el.properties.get("CHANGING").changed()){//Если изменась раскраска или способ ввода
//05.11.2020 XMLcmd.addScaPar("Style",el.getProperty("VISUALIZATION"));
//05.11.2020 XMLcmd.addScaPar("MakeUp",LuiString.RTrim(el.getProperty("MAKEUP"),"\n"));
//05.11.2020 XMLcmd.really=true;
//05.11.2020 }
             addCommandParam(XMLcmd, elcode, "VISUALIZATION", "Style",true);
             if (!elements.get(elcode).getVal("MakeUp").equals(elements.get(elcode).getProperty("MAKEUP"))){
             elements.get(elcode).setVal("MakeUp","");
             elements.get(elcode).properties.get("Var MakeUp").currentValue=elements.get(elcode).getProperty("MAKEUP");
             XMLcmd.addScaPar("MakeUp", elements.get(elcode).getProperty("MAKEUP"));
             addCommandParam(XMLcmd, elcode, "MAKEUP", "MakeUp",true);
             }
             addCommandParam(XMLcmd, elcode, "GETUP", "GetUp",true);
             if (el.properties.get("VISIBLE").changed()){ //Если изменилось Количество строк в поле
              String lines=el.getLines4view();
              if (lines.equals("0")){
              XMLcmd.addScaPar("Type","Hidden");
              el.setVal("Type","Hidden");
              } else {
              XMLcmd.addScaPar("Lines",lines);
              }
              XMLcmd.really=true;
             }
             if (el.properties.get("OPTIONAL").changed()){ //Если изменилась Обязательность данных в поле
               if (el.getProperty("OPTIONAL").equals("N")){
                   XMLcmd.addScaPar("Mandatory","Y");XMLcmd.really=true;
               } else {
                   XMLcmd.addScaPar("Mandatory","N");XMLcmd.really=true;
               }
             }
             if (el.getProperty("CHANGING").equals("B")){//Если изменилась Иконка
              if (is_null(el.getProperty("Value")) && is_not_null(application.getProperty("CHECK_BOX_NULL_ICON")))
                 XMLcmd.addScaParameter("Icon",application.getProperty("CHECK_BOX_NULL_ICON"),def());    
                 else
               addCommandParam(XMLcmd, elcode, "Value","Icon",true);
             }
             if (el.getProperty("CHANGING").equals("R")){//Если изменилась Иконка
               addCommandParam(XMLcmd, elcode, "VARIANTS","Values",true);
             }
               addCommandParam(XMLcmd,elcode,"Parent","Group",true);
             }
           break;
           case "ACTION": // если это действие в бланке
            if (!like(el.getCode(),"SYSMENU$%") /*&& !like(el.getCode(),"COLMENU$%")*/){
            if (NVL(el.getProperty("Run"),"Y").equals("Y")){ // ЕСЛИ ЕГО МОЖНО ВЫПОЛНЯТЬ
             String parType=el.getParent();
             if (!parType.equals("Form")){
                 if (elements.containsKey(parType)) {
                 LuiAppElement pel=elements.get(parType);
                 parType=pel.getType(); }
                 else {
                   parType="";  
                 }
             }
             if (parType.equals("Form") || parType.equals("GROUP")){ // если это действие уровня формы или группы
                      XMLcmd=XMLcg.addCommand("Button");XMLcmd.really=false;
              XMLcmd.addScaPar("Code", elcode);
              if (el.properties.get("ICON").changed() || el.properties.get("PROMPT").changed()) {//Надпись и/или иконка
                  XMLcmd.really=true;
                 String prompt=el.getProperty("PROMPT");
                
                 String ico=el.getProperty("ICON");
                 if (is_not_null(ico)){
                     prompt="[ico:"+ico+"]"+prompt;
                 }
                
                XMLcmd.addScaPar("Label",prompt); 
              }
              //addCommandParam(XMLcmd, elcode, "PROMPT","Label",true);
              addCommandParam(XMLcmd, elcode, "EXPLANATION","Hint",true);//Пояснения
              addCommandParam(XMLcmd, elcode, "Parent","Group",true);//Отправим Родителя, если он изменился
              if (el.properties.get("DISPLAY").changed() || el.properties.get("ACTIVE").changed()){//Если изменилось висимость или активность
                if (el.getProperty("DISPLAY").equals("Y")){//Если видимо
                  if (el.getProperty("ACTIVE").equals("Y")){  // Если активно
                     XMLcmd.addScaPar("State","Active");XMLcmd.really=true;
                  } else {//Если не активно
                     XMLcmd.addScaPar("State","Disable");XMLcmd.really=true; 
                  }
                } else {//Если не видимо
                    XMLcmd.addScaPar("State","Hidden");XMLcmd.really=true; 
                }
              }
             } else { // это действие не уровня формы или группы
               if (parType.equals("FIELD")){ // Это действие уровня поля (локальное)
                 LuiAppElement pel=elements.get(el.getParent());
                 /*
                 if (pel.getProperty("ITEM_TEMPLATE").equals("FIELD_CHAR_COLOR")){// если цвет
                   if (like(el.Code,"EDITOR$%")){  
                   String newIcon=el.getProperty("ICON");//новая иконка (цвет) действия 
                   String oldIcon=el.getProperty("OldIcon"); //Старая иконка действия
                   if (!oldIcon.equals(newIcon) || el.properties.get("ACTIVE").changed()){
                       //Иконка (цвет) сменилась или активность действия сменилась
                       el.setProperty("OldIcon",newIcon);//Сохраним текущее значение для следующего события
                       XMLcmd=XMLcg.addCommand("FieldButton");
                       XMLcmd.addScaPar("Code",el.Code);
                       XMLcmd.addScaPar("Field",pel.getCode());
                       XMLcmd.addScaPar("BgColor",newIcon);
                       if (el.getProperty("ACTIVE").equals("Y")){//Сейчас активна
                          XMLcmd.addScaPar("State","Active");
                      } else {//Сейчас не активна
                          XMLcmd.addScaPar("State","Disable");//Скрывавем
                      }
                   }
                 }*/
               //  } else {// если не цвет
                 String newIcon=el.getProperty("ICON");//новая иконка действия
                 if (el.properties.get("ICON").changed()){ //Иконка сменилась
                  String oldIcon=el.getProperty("OldIcon"); //Старая иконка действия
                  el.setProperty("OldIcon",newIcon);//Сохраним текущее значение для следующего события
                  ToolBarButtons=new LuiCollection(pel.getProperty("Buttons"));//коллекция иконок поля
                  if (is_not_null(oldIcon)){ //Ранее иконка была
                    if (ToolBarButtons.Get_Element(oldIcon).equals(elcode)) {// Иконка соответствовала текущему действию
                       XMLcmd=XMLcg.addCommand("FieldButton");
                       XMLcmd.addScaPar("Code",oldIcon);
                       XMLcmd.addScaPar("Field",pel.getCode());
                       XMLcmd.addScaPar("State","Hidden");
                       ToolBarButtons.Erase_Element(oldIcon);
                    }  
                  }
                  if (is_not_null(newIcon)){//Сейчас иконка есть
                      XMLcmd=XMLcg.addCommand("FieldButton");
                      XMLcmd.addScaPar("Code",newIcon);
                      XMLcmd.addScaPar("Field",pel.getCode());
                      if (el.getProperty("ACTIVE").equals("Y")){//Сейчас активна
                          XMLcmd.addScaPar("State","Active");
                      } else {//Сейчас не активна
                          XMLcmd.addScaPar("State","Hidden");//Скрывавем
                      }
                      if (ToolBarButtons.containsKey(newIcon)) {//Иконка есть в перечне иконок поля
                        if (!ToolBarButtons.Get_Element(newIcon).equals(elcode)){//-- Иконка соответствует не текущему действию
                            ToolBarButtons.Add_Element(newIcon,"");//Иконка не связана с каким-либо действием
                        }
                      } else { //Иконки нет в перечне иконок поля
                          XMLcmd.addScaPar("Hint", el.getProperty("EXPLANATION"));//Добавляем хинт 
                          ToolBarButtons.Add_Element(newIcon,elcode);//Привязываем иконку к действию
                      }
                  }
                  pel.setProperty("Buttons", ToolBarButtons.toString());
                 } else { //иконка не сменилась
                   if (is_not_null(newIcon)){// но есть
                    if (el.properties.get("ACTIVE").changed() ||
                        (el.properties.containsKey("BgColor") &&
                         el.properties.get("BgColor").changed())   
                            ){//Изменилась активность или цвет
                        XMLcmd=XMLcg.addCommand("FieldButton");
                        XMLcmd.addScaPar("Code",newIcon);
                        XMLcmd.addScaPar("Field",pel.getCode());
                        if (el.properties.containsKey("BgColor")){
                         XMLcmd.addScaPar("BgColor",el.getProperty("BgColor"));   
                        }       
                        if (el.getProperty("ACTIVE").equals("Y")){//Сейчас активна
                          XMLcmd.addScaPar("State","Active");
                      } else {//Сейчас не активна
                         if (!el.properties.containsKey("BgColor"))   
                          XMLcmd.addScaPar("State","Hidden");//Скрывавем для обычных
                         else
                          XMLcmd.addScaPar("State","Disable");//Дизаблим для цвета
                      }
                    }  
                   }  
                 }
               //}  //если не цвет
             }
             }
            }
            }
           break;
           case "GROUP": //Это граппа элементов бланка
           XMLcmd=XMLcg.addCommand("Group");XMLcmd.really=false;
           XMLcmd.addScaPar("Code", elcode);
           addCommandParam(XMLcmd, elcode, "TITLE", "Prompt", true);
           if (el.properties.get("GROUP_TYPE").changed()){//Если изменился тип группы
               XMLcmd.addScaPar("Type", getGroupNameForXML(el.getProperty("GROUP_TYPE")));XMLcmd.really=true;
           }
           if (el.properties.get("SEEABLE").changed()){//Если изменался признак видимости группы
             //if (NVL(el.getProperty("SEEABLE"),"N").equals("Y"))  {//Если видима
               if (el.isVisible(false))  {//Если видима    
                 XMLcmd.addScaPar("State", "Visible");XMLcmd.really=true;
             } else {//Если не видима
                 XMLcmd.addScaPar("State", "Hidden");XMLcmd.really=true;
             }
           }
           if (el.properties.get("POSITION").changed()){//Если изменилось положение подписей к полям в группе
            String pp=getPromptPositionForGroup(el.getCode());
                pp=pp.replace("AUTO", "UP");
                pp=NVL(pp,"LEFT");
                pp=LuiString.withCapitalLetter(pp);
                XMLcmd.addScaPar("FieldsPromptPosition",pp);XMLcmd.really=true;
           }
           break;    
       }
      }
     lastCurrentField=null; 
     setField(getVal("CurrentField"));
     break;
     case "MENU":
      for (int i=0;i<elements.size();i++){
       LuiAppElement el=elSort(i);
       if (el.getType().equals("ACTION")){
        if (like(el.getCode(),"SYSMENU%")) continue;  
        if (like(el.getProperty("ROLE"),"%RMB%")) continue;
        if (like(el.getProperty("ROLE"),"%CLOSE%")) continue;
        if (el.properties.get("PROMPT").changed() ||
            el.properties.get("ICON").changed() ||
            el.properties.get("EXPLANATION").changed() ||
            el.properties.get("DISPLAY").changed()  ) {
             
        }
       }
      }
     break;    
     case "GRID":
     if (application.windows.containsKey(getVal("WindowID")) && application.windows.get(getVal("WindowID")).getProperty("AbleStatus").equals("Enable") 
             && !application.windows.get(getVal("WindowID")).getProperty("Discover").equals("Y")
             ){ // если окно есть и оно активно
     
     boolean b=false;
     /*for (String wk:application.windows.keySet()){
         if (application.windows.get(wk).getProperty("ParentID").equals(getVal("WindowID"))) {
             b=true;break;
         }
     }*/
     if (!b){// и нет окон порождённых этим окном
     if (!getVal("WaitingForFormActivate").equals("Y")) {
     // свойства самой матрицы
     if (properties.get("ANGULAR").changed() ||
         properties.get("ANGULAR_HINT").changed() ||
         properties.get("COL_INFO").changed() ||
         properties.get("ROW_INFO").changed() ||
         properties.get("SELECTION_MODE").changed() ||
         properties.get("GRID_TITLE").changed() ||
         properties.get("GRID_TITLE_BORDER").changed()
             ){
      XMLcmd=XMLcg.addCommand("Matrix");
      if (properties.get("GRID_TITLE").changed())XMLcmd.addScaPar("Title",LTrim(getProperty("GRID_TITLE")));
      if (properties.get("ANGULAR").changed())XMLcmd.addScaPar("AngularText",LTrim(getProperty("ANGULAR")));
      if (properties.get("ANGULAR_HINT").changed())XMLcmd.addScaPar("AngularHint",getProperty("ANGULAR_HINT"));
      if (properties.get("COL_INFO").changed()){
        if (NVL(getProperty("COL_INFO"),"N").equals("N")) XMLcmd.addScaPar("ColumnInfo","None");
        else { 
         switch (getProperty("COL_INFO").charAt(1)){
             case 'B':XMLcmd.addScaPar("ColumnInfo","Bottom");break;
             case 'T':XMLcmd.addScaPar("ColumnInfo","Top");break;
             case 'O':XMLcmd.addScaPar("ColumnInfo","Only");break;
           }      
        }  
      }
      if (properties.get("ROW_INFO").changed()){
        if (NVL(getProperty("ROW_INFO"),"N").equals("N")) XMLcmd.addScaPar("RowInfo","None");
         else {
         switch (getProperty("ROW_INFO").charAt(1)){
             case 'R':XMLcmd.addScaPar("RowInfo","Right");break;
             case 'L':XMLcmd.addScaPar("RowInfo","Left");break;
             case 'O':XMLcmd.addScaPar("RowInfo","Only");break;
        }
          }  
      }
     }
     
     this.MatrixPrepareToolBar(XMLcg);
     // столбы и строки
// обновление строк и столбцов
String safeCurrentTuple=getProperty("CurrentTuple"); // сохраняем значение CurrentTuple
     for (String TupleCode:elements.keySet()){
       if (!elements.get(TupleCode).getType().equals("ROW") && !elements.get(TupleCode).getType().equals("COL"))continue;
       if (!elements.get(TupleCode).changeFlag)continue;
       LuiAppElement TupleEl=elements.get(TupleCode);
       String TupleType=TupleEl.getType();
       addProperty("CurrentTuple",TupleCode);
       if (TupleType.equals("ROW"))XMLcmd=XMLcg.addCommand("Row");
                              else XMLcmd=XMLcg.addCommand("Column");
       XMLcmd.addScaPar("Code",TupleCode);
       if (TupleEl.properties.containsKey("CAPTION") && TupleEl.properties.get("CAPTION").changed()) {
        XMLcmd.addScaPar("Caption",LTrim(TupleEl.getProperty("CAPTION")));//-- Заголовок   
       }
       switch (NVL(getProperty(TupleType+"_INFO"),"N").charAt(0)){ //-- Доп информация
                      case 'N':
                      break;
                      case 'Y':
                      XMLcmd.addScaPar("Info",TupleEl.getProperty("TUPLE_INFO"));
                      break;
                      case 'O':
                      XMLcmd.addScaPar("Info",TupleEl.getProperty("ltn"));
                      break;
                  }
       if (TupleEl.properties.containsKey("VISIBLE") && TupleEl.properties.get("VISIBLE").changed()) {
        XMLcmd.addScaPar("Visible",TupleEl.getProperty("VISIBLE"));
       }
       if (TupleEl.properties.containsKey("DIRECTION") && TupleEl.properties.get("DIRECTION").changed()) {
        String dir="H";
           switch (TupleEl.getProperty("DIRECTION")){
                case "TOP2BOTTOM":dir="TB";break;
                case "BOTTOM2TOP":dir="BT";break;
                case "VERTICAL":dir="V";break;
            }
             
                  XMLcmd.addScaPar("Direction",dir);//-- Направление текста
       }
       if (TupleEl.properties.containsKey("HINT") && TupleEl.properties.get("HINT").changed()) {
        XMLcmd.addScaPar("Hint",TupleEl.getProperty("HINT"));   
       }
       if (TupleEl.properties.containsKey("ALIGNMENT") && TupleEl.properties.get("ALIGNMENT").changed()) {
         XMLcmd.addScaPar("H-align",LuiString.withCapitalLetter(TupleEl.getProperty("ALIGNMENT")));  
       }
       if (TupleEl.properties.containsKey("VALIGN") && TupleEl.properties.get("VALIGN").changed()) {
         XMLcmd.addScaPar("V-align",LuiString.withCapitalLetter(TupleEl.getProperty("VALIGN")));  
       }
       if (TupleEl.properties.containsKey("OVERFLOW") && TupleEl.properties.get("OVERFLOW").changed()) {
         XMLcmd.addScaPar("Overflow",LuiString.withCapitalLetter(TupleEl.getProperty("OVERFLOW")));
       }
       if (TupleEl.properties.containsKey("VISUALIZATION") && TupleEl.properties.get("VISUALIZATION").changed()) {
         XMLcmd.addScaPar("Class",TupleEl.getProperty("VISUALIZATION"));
       }
       if (TupleEl.properties.containsKey("TUPLE_VISUAL") && TupleEl.properties.get("TUPLE_VISUAL").changed()) {
         XMLcmd.addScaPar("SelfClass",TupleEl.getProperty("TUPLE_VISUAL"));
       }
       if (TupleEl.properties.containsKey("MAKEUP") && TupleEl.properties.get("MAKEUP").changed()) {
        XMLcmd.addScaPar("MakeUp",TupleEl.getProperty("MAKEUP"));   
       }
       if (TupleEl.properties.containsKey("GETUP") && TupleEl.properties.get("GETUP").changed()) {
        XMLcmd.addScaPar("GetUp",TupleEl.getProperty("GETUP"));   
       }
       if (properties.get(TupleType+"_INFO").changed()|| 
           (TupleEl.properties.containsKey("TUPLE_INFO") && TupleEl.properties.get("TUPLE_INFO").changed()) ||
           (TupleEl.properties.containsKey("ltn") && TupleEl.properties.get("ltn").changed())) {
         switch (NVL(getProperty(TupleType+"_INFO"),"N").charAt(0)){ //-- Доп информация
                      case 'N':
                      break;
                      case 'Y':
                      XMLcmd.addScaPar("Info",TupleEl.getProperty("TUPLE_INFO"));
                      break;
                      case 'O':
                      XMLcmd.addScaPar("Info",TupleEl.getProperty("ltn"));
                      break;
                  }  
       }
               
     }     
     // восстанавливаем CurrentTuple
     addProperty("CurrentTuple",safeCurrentTuple);
     // крутим карту матрицы
     String RowCode;
     String ColCode;
     ArrayList<String> acell;
     LuiAppElement ecell;
     
     boolean stackCh=false;
     
     for (String mk:MatrixMap.keySet()){
      stackCh=false;
      acell=MatrixMap.get(mk);
      if (acell!=null){
        for (int ci=0;ci<acell.size();ci++){
         if (is_null(acell.get(ci)))continue;
         ecell=elements.get(acell.get(ci));
         if (ecell==null)continue;
         if (ecell.properties!=null){
         for (String cpc:ecell.properties.keySet()){
            if (cpc.equals("CELL_SHOW") 
                    || cpc.equals("CELL_HINT") 
                    || cpc.equals("CELL_HALIGN")
                    || cpc.equals("CELL_VALIGN")
                    || cpc.equals("CELL_OVERFLOW")
                    || cpc.equals("CELL_MAKEUP")
                    || cpc.equals("CELL_GETUP")
                    || cpc.equals("CELL_VISUAL")){
            if (ecell.properties.get(cpc).changed()){
              stackCh=true;break;
            } 
            }
         }}
         if (stackCh)break;
        }
      }
      if (!stackCh)continue;
             RowCode=MatrixGetCellProperty(mk,"ROW");
             if (is_null(RowCode))continue;
             ColCode=MatrixGetCellProperty(mk,"COL");
             if (is_null(ColCode))continue;
              XMLcmd=XMLcg.addCommand("Cell");
              XMLcmd.addScaPar("R-code",RowCode);
              XMLcmd.addScaPar("C-code",ColCode);
                XMLcmd.addScaParameter("Show",MatrixGetCellProperty(mk,"CELL_SHOW"), def());
              XMLcmd.addScaParameter("Hint",MatrixGetCellProperty(mk,"CELL_HINT"), def());
                  XMLcmd.addScaParameter("H-align",LuiString.withCapitalLetter(MatrixGetCellProperty(mk,"CELL_HALIGN")), def());
                  XMLcmd.addScaParameter("V-align",LuiString.withCapitalLetter(MatrixGetCellProperty(mk,"CELL_VALIGN")), def());
                  XMLcmd.addScaParameter("Overflow",LuiString.withCapitalLetter(MatrixGetCellProperty(mk,"CELL_OVERFLOW")), def());
                  XMLcmd.addScaParameter("MakeUp",MatrixGetCellProperty(mk,"CELL_MAKEUP"), def());
                  XMLcmd.addScaParameter("GetUp",MatrixGetCellProperty(mk,"CELL_GETUP"), def());
                  XMLcmd.addScaParameter("Class",MatrixGetCellVisual(mk), def()); 
                  
     }
     XMLcmd=XMLcg.addCommand("MatrixFinish");
     XMLcmd.addScaPar("FormID",getFormID());
     setAllChangeFlag(false);
     }
 }
     }
     break;    
     case "CHART":
     case "HTML":
     
         
     //  во-первых тулбар и кнопки подвала и таймер
     if (application.eventReader.is_non_interval){
     String Timer="";    
     ToolBarButtons=makeSysToolBarBattonCollection();
     for (int i=0;i<elements.size();i++) { //крутим элементы по порядку
          LuiAppElement el=elSort(i);
          switch(el.getType()){
           case "ACTION": // если это действие              
             if (!like(el.getCode(),"SYSMENU$%")) { // не системного меню    
              if (NVL(el.getProperty("Run"),"Y").equals("Y")){ // ЕСЛИ ЕГО МОЖНО ВЫПОЛНЯТЬ
                  el.properties.get("ACTIVE").clearCompVal();//пересчитываем активность
                  el.properties.get("ICON").clearCompVal();//пересчитываем иконку
               if (el.getParent().equals("Form")){ // уровня формы                  
                    String icon=el.getProperty("ICON");
                    if (is_not_null(icon)){ // c иконкой
                      if (ToolBarButtons.containsKey(icon)) { // и оно есть уже в коллекции
                          this.addIcon2ToolBar(icon, "","",ToolBarButtons);
                         // ToolBarButtons.Add_Element(icon,"");// стираем ей хинт
                      } else { // если же ещё нет
                          // если её ещё в тулбаре нет - выводим её хинт
                            // но перед выводом сравним декады
                            if ((el.sourceSeq/10)!=toolBarDecade){ // если декады разные ставим сепаратор
                             String keySep="";
                             for(int j=0;j<toolBarDecade;j++) keySep=keySep+" ";
                             //ToolBarButtons.Add_Element(keySep,"");  
                             this.addIcon2ToolBar(keySep, "","",ToolBarButtons);
                             toolBarDecade=(int)el.sourceSeq/10;
                            }
                             //ToolBarButtons.Add_Element(el.getProperty("ICON"),el.getProperty("EXPLANATION"));  
                             this.addIcon2ToolBar(el.getProperty("ICON"), el.getProperty("PROMPT"),el.getProperty("EXPLANATION"),ToolBarButtons);
                      }                        
                    }//c иконкой
                  if (is_not_null(el.properties.get("BUTTON").getNonCompValue())){//Действие может вызываться кнопкой в подвале
                       XMLcmd=XMLcg.addCommand("Button");XMLcmd.really=false; //готовим возможную команду для отображения кнопки
                       XMLcmd.addScaPar("Code",el.getCode());
                       addCommandParam(XMLcmd, el.getCode(),"BUTTON", "Label", true);//Отправим надпись на кнопке, если она изменилась
                       addCommandParam(XMLcmd, el.getCode(),"EXPLANATION", "Hint", true);

                       if (el.properties.get("BUTTON").changed() || //Если изменилась надпись на кнопке
                           el.properties.get("ACTIVE").changed()  //или активность кнопки
                           ){ 
                         if (is_null(el.getProperty("BUTTON"))) { // если кнопка скрыты
                           XMLcmd.addScaPar("State", "Hidden");XMLcmd.really=true; // скрыть
                         } else { // если не скрыта
                             if (el.isActionActive()) { // и активна
                                 XMLcmd.addScaPar("State", "Active");XMLcmd.really=true; // Активировать, если не скрыта, но активна
                             } else {
                                 XMLcmd.addScaPar("State", "Disable");XMLcmd.really=true; // Задизейблить, если не скрыта, но не активна
                             }
                         } // если не скрыта                          
                            }//или активность кнопки
                  }//Действие может вызываться кнопкой в подвале
                  if (like(el.getProperty("ROLE"),"%TIME%")) {
                      if (is_null(Timer)||Timer.equals("N")) 
                       Timer="Y"; // Признак необходимости "тикания таймером"
                   }
                   if (like(el.getProperty("ROLE"),"%TIME500%")) {
                       Timer="500"; // Признак необходимости "тикания таймером"
                   }
               }
              }
             }
           break;
          }//switch
     }//цикл по элеиентам
     setVal("Timer",Timer);
   //  if (!compareString(ToolBarButtons.toString(),getVal("ToolBarButtons"))) {//Если построенная коллекция кнопок тулбара отличается от того, что было
       if (!compareString(this.getToolBarString(ToolBarButtons),getVal("ToolBarButtons"))) {//Если построенная коллекция кнопок тулбара отличается от того, что было
     XMLcmd=XMLcg.addCommand("Form");
     if (!getProperty("ToolBar").equals("N"))
     XMLcmd.addScaPar("ToolBarButtons", this.getToolBarString(ToolBarButtons));
     setVal("ToolBarButtons",this.getToolBarString(ToolBarButtons));
     }
     //  во-вторых свойства генерации
     if (FType.equals("CHART") && (properties.get("CHART_GENERATOR").changed()||properties.get("CHART_GEN_ADDON").changed())){
         chart(XMLcg);
     }
     if (FType.equals("HTML")){
         if (properties.get("HTML_GENERATOR").changed()){
         XMLcmd=XMLcg.addCommand("Html");
         XMLcmd.addScaPar("FormID",this.getFormID());
         XMLcmd.addScaPar("Data",this.getProperty("HTML_GENERATOR"));
         }
       for (String ec:elements.keySet()){
         if (!elements.get(ec).getType().equals("ITEM"))continue;
         //if (!elements.get(ec).properties.get("DEFINITION").changed())continue;
         LuiCollection col=new LuiCollection("");
         for (String pk:elements.get(ec).properties.keySet()){
             if (!elements.get(ec).properties.get(pk).gen4html)continue;
             col.Add_Element(pk,elements.get(ec).getProperty(pk));
         } 
         if (!col.ToJSON().equals(elements.get(ec).getProperty("DEFINITION"))){
         elements.get(ec).addProperty("DEFINITION",col.ToJSON());
         XMLcmd=XMLcg.addCommand("SetProperties");
         XMLcmd.addScaPar("FormID",this.getFormID());
         XMLcmd.addScaPar("Code",elements.get(ec).getProperty("HTML_ELEMENT"));
         XMLcmd.addScaPar("Properties",elements.get(ec).getProperty("DEFINITION"));
         }
        }
     }
     }
     break;  
     case "TREE":
     if (properties.get("HEADER").changed()){
       if (!getVal("OldHeader").equals(getProperty("HEADER"))){
       XMLcmd=XMLcg.addCommand("Node");
       XMLcmd.addScaPar("FormID", getVal("RootFormID"));
       XMLcmd.addScaPar("Node", getVal("ParentNode"));
       XMLcmd.addScaPar("Label", LuiString.RemoveIco(getProperty("HEADER")));
       if (is_not_null(LuiString.GetIcoFromSquareBracket(getProperty("HEADER")))) {
       XMLcmd.addScaPar("Icon",LuiString.GetIcoFromSquareBracket(getProperty("HEADER")));  
       }
       setVal("OldHeader",getProperty("HEADER"));
     }
     }
    /* if (properties.get("NODE_ICON").changed() && is_not_null(getVal("CurrentNode"))){
          XMLcmd=XMLcg.addCommand("Node");
       XMLcmd.addScaPar("FormID", getVal("RootFormID"));
       XMLcmd.addScaPar("Node", getVal("CurrentNode"));
     XMLcmd.addScaPar("Icon", getProperty("NODE_ICON"));
     }*/
 }
}

    /**
     * обработка нажатия ПКМ в узле дерева
     *
     * @param pNode -ID узла, с которым производится манипуляция. Если клик
     * производится вне какого-то узла (ниже последнего), то параметр Node
     * должен быть "" или null
     * @param pState -текущее состояние узла
     * @param pNType -тип этого узла, может быть "" или null, если Node не
     * определён
     * @param pContext -Контекст,может быть "" или null, если Node не определён
     * @param pContextID -D рабочей области, к которой относится контекст, может
     * быть "" или null, если Node не определён
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void nodeRMB(String pNode, String pState, String pNType, String pContext, String pContextID ) throws LuiException, SQLException{

String FormID;
    if (is_null(pNode)){
        this.clearElementsValue();
        this.changeNode(pNode, pState, pNType, pContext, pContextID, "");
     FormID=this.getFormID();
    } else {
     FormID=pContextID;   
    }
 if (!application.forms.containsKey(FormID)) {
   LuiClCommandGroup XMLcg=application.commandWriter.addGroup("NodeRMB");
   LuiClCommand XMLcmd=XMLcg.addCommand("DeleteNode");
   XMLcmd.addScaPar("FormID", this.getFormID());
   XMLcmd.addScaPar("Node", pNode);
    
   return;  
 }
 application.setVal("FormID",FormID);
 LuiAppForm frm=application.forms.get(FormID);
 String menuitemcode="";
 if (frm.getType().equals("MENU")) {
     LuiCollection concol=new LuiCollection(pContext);
     if(frm.elements.containsKey(concol.Get_Element("ACTION"))){
         LuiAppElement mel=frm.elements.get(concol.Get_Element("ACTION"));
         menuitemcode=mel.getCode();
         this.setDynamicMenuItemProperty(mel, concol);
     }
 }
 LuiClCommandGroup XMLcg=application.commandWriter.addGroup("NodeRMB");
 XMLcg.addParameter("FormID", FormID);
 XMLcg.addParameter("Type","Action");
 LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
 if (NVL(pNType,"").equals("ACTION") && NVL(pState,"").equals("Collapsed")){
   XMLcmd=XMLcg.addCommand("AddMenuItem");
   XMLcmd.addScaPar("Code", Extract_Word(pNode,-1));
   XMLcmd.addScaPar("Label",application.NLS.NLS("RUS:Открыть как список;ENG:Open as list;"));
 } else {
     for (int i=0;i<frm.elements.size();i++) { //крутим элементы
       LuiAppElement el=frm.elSort(i);  
       
       //if (el.getType().equals("ACTION") && (el.getProperty("Parent").equals("Form") || (frm.getType().equals("MENU") && like(el.getProperty("ROLE"),"%RMB%")))) { // если это действие уровня формы или формой является меню а действие имеет роль RMB
       if (!el.getType().equals("ACTION")) continue;
       if (frm.getType().equals("MENU")){
           if (!like(el.getProperty("ROLE"),"%RMB%")) continue;
           if (!el.getProperty("Parent").equals("Form") && !el.getCode().equals(menuitemcode)) continue;          
       } else {
           if (!el.getProperty("Parent").equals("Form")) continue;
       }
           
       if (like(el.getCode(),"SYSMENU$%") || like(el.getCode(),"COLMENU$%") || like(el.getCode(),"CELLMENU$%")) continue;
         if (NVL(el.getProperty("Run"),"Y").equals("Y")){ // если действие разрешено
          // вычисляем видимость
          String AV=el.getProperty("DISPLAY");
          if (AV.indexOf("SQL:", 0)>-1) { // в DISPLAY размножитель SELECT
          ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4), frm.getProperty("Connection"),this.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(this.application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          if (propsArray!=null){
          // крутим извлечённые строки
          for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                el.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt) 
                    && (!el.getProperty("ACTIVE").equals("EXIST")|| frm.getType().equals("MENU"))//вот это доп условие предотвращает генерацию засеренных пунктов
                                                                                                  // в списковом дереве пунктов меню, которые на самом деле должны становиться
                                                                                                  // действиями-подузлами
                    ) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
                if (!compareString(el.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N"); 
            //Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }    
          } 
          else if (AV.indexOf("JS:", 0)>-1) { // в DISPLAY размножитель JS
                         try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.application.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного массива
           el.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          String Prompt=el.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",el.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
            if (!compareString(el.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");
            //Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          } 
          }
          else { //// в DISPLAY нет размножителя SELECT и JS
              if (AV.equals("Y") || (AV.equals("EXIST") && pNType.equals("ROW"))){ // если видимо = Y или когда EXIST
                String Prompt= el.getProperty("PROMPT");
                if (is_not_null(Prompt)){
                    if (!(el.getProperty("ACTIVE").equals("EXIST") 
                      && el.getProperty("COMMAND_TYPE").equals("LIST")
                      && is_null(el.getNonCompValue("PROC_INITIALLY"))
                      && is_null(el.getNonCompValue("PROC_FINALLY"))
                      && el.getProperty("APPLY_TO").equals("N")
                      && is_not_null(el.getProperty("COMMAND"))
                      && !like(NVL(el.getProperty("ROLE"),""),"%AUTO%"))) {
                     XMLcmd=XMLcg.addCommand("AddMenuItem");
                     XMLcmd.addScaPar("Code",el.getCode());
                     XMLcmd.addScaPar("Group",el.getParent()+String.valueOf(el.sourceSeq/10));
                     if (!el.isActionActive()){
                         XMLcmd.addScaPar("Enable","N");
//                         Prompt=RemoveIco(Prompt);
                     }
                     XMLcmd.addScaPar("Label",Prompt);
                    }
                }
              }
          }
         }  // если действие разрешено
           
       
     }//крутим элементы
 }
   if (XMLcg.commandList.size()<2) {
       application.commandWriter.removeGroup("NodeRMB");
   }
}
/* Обработка события совпадения введённого в поле значения с заданной маской
*/
public void patternMatch() throws LuiException, SQLException{
    String oldCurrentField=getVal("CurrentField");
    String actionCode=this.checkHotKey("MATCH", oldCurrentField);
    if (is_not_null(actionCode)){
        application.changeSCN("PatternMatch");
        this.runAction(actionCode);
        String newCurrentField=getVal("CurrentField");
        if (!compareString(oldCurrentField,newCurrentField)){
         LuiClCommandGroup XMLcg=application.commandWriter.addGroup("SetCurrentField");
         LuiClCommand XMLcmd=XMLcg.addCommand("CurrentField");
         XMLcmd.addScaPar("FormID", getFormID());
         XMLcmd.addScaPar("Field", newCurrentField);
        }
        checkFrmDetails();
    } else {
         LuiClCommandGroup XMLcg=application.commandWriter.addGroup("GotoNextField");
         LuiClCommand XMLcmd=XMLcg.addCommand("NextField");
         XMLcmd.addScaPar("FormID", getFormID());
    }
}
/**
 * Дублирование в бланк значений предыдущего похожего
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void duplicate() throws LuiException, SQLException{
    application.changeSCN("Duplicate");
    if (application.lastFieldValues!=null) {
      for (String ec:elements.keySet()){
          LuiAppElement el=elements.get(ec);
          if (el.getType().equals("FIELD")){
             switch (NVL(el.getProperty("CHANGING"),"N")) {
                 case "C":
                 case "Y":
                 case "B":
                 case "U":
                 case "P":
                 case "R":
                 if (application.lastFieldValues.containsKey(ec)){
                     el.setValue(application.lastFieldValues.get(ec), false,"Y",false);
                 }    
                 
                 
             }
          }
      }}
      checkFrmDetails();
}

/**
 * Добавление иконки в тулбар
 * @param el -- элемент
 * @param tb -- тулбар-коллекция
 * @throws LuiException
 * @throws SQLException 
 */
public void addIcon2ToolBar(String icon,String Prompt,String Hint,LuiCollection tb) throws LuiException, SQLException{
    if (application.getProperty("TOOLBAR_MODE").equals("PH")){
     if (is_not_null(icon.replace(" ","").replace("_",""))){
      if (application.getProperty("HIDE_TBL").equals("Y"))
       tb.Add_Element(icon,"†"+Hint);
      else
       tb.Add_Element(icon,LuiString.RemoveIco(Prompt)+"†"+Hint);
     }
    } else {
      tb.Add_Element(icon,Hint);  
    }
}
public String getToolBarString(LuiCollection tb) throws LuiException, SQLException{
    if (application.getProperty("TOOLBAR_MODE").equals("PH")){
      String s="";
      for (String k:tb.keySet()){
          String[] sa=tb.Get_Element(k).split("†");
       //   s=s+"[ico:"+k+"]"+tb.Get_Element(k).replace("†",":")+";";
       s=s+k+"\n"+tb.Get_Element(k).replace("†",":")+";";
      }
      return s;
    } else {
      return tb.toString();
    }
}
/** 
 * Создаёт коллеуцию кнопок системного меню
 * @return коллеуция кнопок системного меню
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public LuiCollection makeSysToolBarBattonCollection() throws LuiException, SQLException{
  LuiCollection toolBarButtons=new LuiCollection("");
if (getType().equals("LIST") ||getType().equals("CHART")||getType().equals("HTML")
        /*||getType().equals("GRID")*/ //закоментировал т.к. для матриц почему-то иная логика
        ){
        for (int i=0;i<elements.size();i++){
               LuiAppElement el=elSort(i);
               if (el.getType().equals("ACTION")){
                 if (el.getProperty("Run").equals("Y")) { // если действие разрешено
                    if (!el.getNonCompValue("ACTIVE").equals("N")) { // Действие не вечно заблокировано 
                       if (el.getParent().equals("Form")){ //если действие уровня формы 
                           if (like(el.getCode(),"SYSMENU$%")) {//и оно из системного меню
                             if (is_not_null(el.getProperty("ICON")))  {
                              addIcon2ToolBar(el.getProperty("ICON"),el.getProperty("PROMPT"),el.getProperty("EXPLANATION"),toolBarButtons);
                             }
                           }
                       }
                    } 
                 }  
               }
        }
        }
  return toolBarButtons;
}
/** 
 *  устанавливает свойство Connection для формы
 * @param pParams коллекцция параметров
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void setConnectionProperty(String pParams) throws LuiException, SQLException{
if (is_not_null(pParams)){
LuiCollection pc=new LuiCollection(pParams);
if (pc.containsKey("CONNECTION$")){
 String c=pc.Get_Element("CONNECTION$");
 if (is_not_null(c)){
     setProperty("Connection",c);
     return;
 }
} }
if (application.commonForms!=null){
if (application.commonForms.containsKey(code) && !application.code.equals("LUI")){
     setProperty("Connection","LUI$");
   } else
     setProperty("Connection",application.getProperty("Connection"));
} else 
    setProperty("Connection",application.getProperty("Connection"));
}
/**
 * сбрасывает флаг изменённости
 */
public void resetChangeFlag(){
    if (this.changeFlag){
     for (String pkey:properties.keySet()){
       properties.get(pkey).changeFlag=false;
     }   
     for (String elc:elements.keySet()){
       elements.get(elc).resetChangeFlag();
     }
     this.changeFlag=false;    
    }
}
@Override
public String getGenesis(){
    try {    
        return ownerObj.getGenesis()+" "+getType()+":"+getCode()+";FormID:"+getFormID()+";";
    } catch (LuiException ex) {
       return "";
    } catch (SQLException ex) {
       return "";
    }
}
@Override
public String getGenesis4dmp(){
    try {    
        return getType()+":"+getCode();
    } catch (LuiException ex) {
       return "";
    } catch (SQLException ex) {
       return "";
    }
}

@Override
public LuiApplication getApplication(){
    return application;
}
@Override
public LuiAppForm getForm(){
    return this;
}
/**
 * Сравнение элементов для сортировки
 * @param el1 первый элемент
 * @param el2 второй элемент
 * @return true если первый el1 меньше второго
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public boolean compareElements(LuiAppElement el1, LuiAppElement el2) throws LuiException, SQLException { // true, если el1 идёт раньше el2
  int S1=2, S2=2;
  if (!el1.getParent().equals(el2.getParent())) { // Разные родители
    if ("Form".equals(el1.getParent())) return compareElements(el1, elements.get(el2.getParent())); // Форма и Родитель
    if ("Form".equals(el2.getParent())) return compareElements(elements.get(el1.getParent()), el2); // Родитель и Форма
    return compareElements(elements.get(el1.getParent()), elements.get(el2.getParent())); // Родитель1 и Родитель2
  }
  if ("PARAM".equals(el1.getType())) return true; // Параметры идут всегда в начале (в любом порядке)
  switch (getType()){
    case "LIST":
      if ("ACTION".equals(el1.getType())) S1=4; // Действия всегда в конце
      if ("ACTION".equals(el2.getType())) S2=4; // Действия всегда в конце
      if ("GROUP".equals(el1.getType())) { 
        if ("OVERFLOW".equals(el1.getProperty("GROUP_TYPE"))) S1=3; // Группа OVERFLOW в конце грида
        if ("FIXED".equals(el1.getProperty("GROUP_TYPE"))) S1=1; // Группа FIXED в начале
      }
      if ("GROUP".equals(el2.getType())) {
        if ("OVERFLOW".equals(el2.getProperty("GROUP_TYPE"))) S2=3; // Группа OVERFLOW в конце грида
        if ("FIXED".equals(el2.getProperty("GROUP_TYPE"))) S2=1; // Группа FIXED в начале
      }
      if (S1!=S2) return (S1<S2); else return (el1.mutSeq < el2.mutSeq); // если уровни одитнаковы - играет Seq
    case "BLANK":
           if ("ACTION".equals(el1.getType()) && 
              !"Form".equals(el1.getParent()) && 
              !"SIDE_BY_SIDE".equals(elements.get(el1.getParent()).getProperty("GROUP_TYPE")
              )) S1=3; // Действия в группе не - в конце
      if ("ACTION".equals(el2.getType()) && !"Form".equals(el2.getParent()) && !"SIDE_BY_SIDE".equals(elements.get(el2.getParent()).getProperty("GROUP_TYPE"))) S2=3; // Действия в группе не - в конце

      if (S1!=S2) return (S1<S2); else return (el1.mutSeq < el2.mutSeq); // если уровни одитнаковы - играет Seq
    default:
      return (el1.mutSeq < el2.mutSeq);
  }
}

public LuiAppElement findLeastThen(LuiAppElement pEl) throws LuiException, SQLException{
  String d;  
    for (String ec:elements.keySet()){
        if (!pEl.getCode().equals(ec)){ // если коды сопадают не сравниваем чтобы предотвратить сравнение с собой
         
          LuiAppElement el=elements.get(ec);
          if (!this.elSortedList.contains(el)){ // если элемент уже есть в отсортированном списке не сравниваем
             if (this.compareElements(el, pEl)) { // если el меньше sEl
                 d=el.getCode()+"seq="+String.valueOf(el.mutSeq)+"<"+pEl.getCode()+"seq="+String.valueOf(pEl.mutSeq);
              return findLeastThen(el);
             }
          }  
        }
        
    }
        return pEl;
}

public String getFullPath() throws LuiException, SQLException{
    String retVal="";
    String fType=this.getType();
    String parentID="";
    String parentAction="";
    String rootFormID="";
    parentID=getVal("ParentFormID");
    switch (fType){
        case "MENU":
            retVal=";\n"+application.NLS.NLS("RUS:Меню;ENG:Меню;")+": "+getProperty("HEADER");
            String items="";
            String currentItemContext=getProperty("CurrentMenuItemContext");
            if (is_not_null(currentItemContext)){
              LuiCollection concol=new LuiCollection(currentItemContext);
              String elcode=concol.Get_Element("ACTION");
              while (true)  {
                LuiAppElement el=elements.get(elcode);
                if (is_null(items)) items=el.getProperty("PROMPT");
                else    
                items=el.getProperty("PROMPT")+"->"+items;
                elcode=el.getParent();
                if (elcode.equals("Form")) break;
              }
            }
            retVal=retVal+application.NLS.NLS("RUS: Пункт: ;ENG: Item: ;")+items;
         break;
        case "LIST":
          LuiCollection uk=this.getUK();
          // крутим элементы ключа
          for (String field:uk.keySet()){
            retVal=retVal+", "+elements.get(field).getProperty("CAPTION")+" = "+uk.Get_Element(field);
          }
          if (is_not_null(retVal)){//-- в начале приляпывается фраза Строка, где ...
              retVal=retVal+";\n"+application.NLS.NLS("RUS:Строка где ;ENG:At row where ;")+LTrim(retVal," ,");
          } // В итоге будет текст Список: [Заголовок списка] Строка, где [Столбец] = [знач.], [Столбец] = [знач.]...
          retVal=";\n"+application.NLS.NLS("RUS:Список;ENG:List;")+": "+getProperty("HEADER")+retVal;
        break;
        case "BLANK":
          retVal=";\n"+application.NLS.NLS("RUS:Бланк;ENG:Blank;")+": "+getProperty("HEADER");
        break;
        case "TREE":
          if (properties.containsKey("Var CurrentRow")) { 
          if (is_not_null(getVal("CurrentRow"))) {
          if (Integer.valueOf(getVal("CurrentRow"))>0) {
          retVal="/"+getProperty("NODE_TEXT");
          } 
          }
          }
          rootFormID=getVal("RootFormID");
          if (compareString(rootFormID,this.getFormID())){
           retVal=";\n"+application.NLS.NLS("RUS:Дерево;ENG:Tree;")+": "+getProperty("HEADER")+";\n"+application.NLS.NLS("RUS:Узел;ENG:Node;")+": "+retVal;
          } else {
             return application.forms.get(parentID).getFullPath()+"/"+getProperty("HEADER")+retVal;
          }
        break;
        default:
            
    }
    if (is_not_null(parentID)){
        parentAction=getVal("CallAction");
        if (is_not_null(parentAction)){
            if ("AUTO".equals(application.forms.get(parentID).elements.get(parentAction).getProperty("ROLE"))){
                retVal=";\n"+application.NLS.NLS("RUS:Деталь;ENG:Detail;")+": "+retVal;
            } else {
                String prompt=application.forms.get(parentID).elements.get(parentAction).getProperty("PROMPT");
                String role=application.forms.get(parentID).elements.get(parentAction).getProperty("ROLE");
                if (is_not_null(role) && is_null(prompt)){
                  prompt=prompt+"("+role+")";
                }
                prompt=LuiString.RemoveIco(prompt);
                String aparent=application.forms.get(parentID).elements.get(parentAction).getParent();
                if (!aparent.equals("Form")){
                    aparent=" "+application.NLS.NLS("RUS:элемента;ENG: at level of element;")+" "+NVL(application.forms.get(parentID).elements.get(aparent).getProperty("CAPTION"),application.forms.get(parentID).elements.get(aparent).getProperty("Name"));
                } else
                    aparent="";
                retVal=";\n"+application.NLS.NLS("RUS:Действие;ENG:Action;")+": \""+prompt+"\""+aparent+retVal;
            }
        }
        retVal=application.forms.get(parentID).getFullPath()+retVal;
    }   
    
    return LTrim(retVal,";");
}
/**
 * Возвращает коллекцию отмеченных строк в виде массива JSON
 * @return JSON-массив отмеченных строк
 */
public String getSelectedRowsJSON(){
  ArrayList<LuiCollection> alc = new ArrayList<LuiCollection>();
  for (LuiCollection keycol:checkRows.keySet()){
    alc.add(keycol);
  }  
  String s=application.toJSON(alc);
  return s;
}
/**
 * Добавление ссылки в знечение поля бланка (используется в редакторе)
 * @param pElCode код элемента
 * @param pActionCode код действия
 * @param pPropertyCode код свойства
 * @throws LuiException логическая ошибка
 * @throws SQLException ошибка выполнения SQL-запроса
 */
public void addRef(String pElCode,String pActionCode,String pPropertyCode) throws LuiException, SQLException{
    if (elements.containsKey(pElCode)){
      LuiAppElement el=elements.get(pElCode);
      if (el.properties.containsKey("Value")){
        if (elements.containsKey(pActionCode)){
            LuiAppElement ael=elements.get(pActionCode);
            if (ael.properties.containsKey(pPropertyCode)){
                String vwc=el.getProperty("Value");
                String ec=NVL(el.getProperty("EndCursor"),"0");
                String ref=ael.getProperty(pPropertyCode);
                int curs=Integer.valueOf(ec);
                if (curs>vwc.length()) curs=vwc.length();
                vwc=vwc.substring(0, curs)+ref+CURSORSIGN+vwc.substring(curs);
                
                el.newValue(vwc,null,null,"Y",false,false); 
            }
        }  
      }
    }
}
/**
 * Возвращает коллекцию отмеченных пунктов в виде JSON
 * @return JSON-коллекция отмеченных пунктов
 */
public String getSelectedMenuItemsJSON(){
    LuiCollection c=new LuiCollection("");
    for (LuiCollection mic:this.checkRows.keySet()){
      for (String mitem:mic.keySet()){  
        c.Add_Element(mitem,"Y");
      }
    }
    return c.ToJSON();
}
/* Возвращает объект LuiSQLCondition - разобранное QBE условие задананное в указанном столбце
* 

public LuiSQLCondition getSQLCondition(String pColCode,String QBEtext) throws LuiException, SQLException{
    SQLCondition=new LuiSQLCondition();
    if (pColCode==null){
        pColCode=getProperty("Var CurrentColumn");
    }
    if (!elements.containsKey(pColCode)) return SQLCondition;
    
    if (QBEtext==null){
      String currentRow=getProperty("Var CurrentRow");
      if (listQBE.containsKey(currentRow)){
          if (listQBE.get(currentRow).containsKey(pColCode)) {
           QBEtext=listQBE.get(currentRow).get(pColCode);
          } else 
           QBEtext="";    
      } else 
           QBEtext="";
      
    }
    if (is_null(QBEtext)) return SQLCondition;
    String connCode=getProperty("Connection");
    LuiDbConn lc=null;
    if (is_null(connCode)) connCode="LUI$";
    LuiAppConn ac=application.userDbConnections.get(connCode);
if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
   lc=ac.currentConnection;
}
   if (lc==null) { // если так и не найдено
       lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
   }
   LuiAppElement el=elements.get(pColCode);
   SQLCondition= new LuiSQLCondition(lc,el,QBEtext);
   return SQLCondition;
}*/
public String getQBEfromCell(Integer pRow,String pCol){
    if (listQBE.containsKey(pRow.toString())) {
       if (listQBE.get(pRow.toString()).containsKey(pCol)){
           return listQBE.get(pRow.toString()).get(pCol);
       }
    }
return "";
}
/**
 * Применение сортировок установленных в LUI_L_ORDER_BY
 * @param p_json -- массив элементов json
 */
public void JSON2sort(String p_json) throws LuiException, SQLException {
   LuiDbConn lc=null;
   try{
   lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());    
   // чистим свойства сортировок
   for (String ec:elements.keySet()){
       if (elements.get(ec).getType().equals("ITEM")){
         if (elements.get(ec).properties.containsKey("SORT")){
             elements.get(ec).setProperty("SORT","");
             elements.get(ec).properties.get("SORT").clearCompVal();
             
         }
       }
   }
   // устанавливаем новые
    ArrayList<LuiCollection> newSorts=lc.getSelectAllRowsValMap("select code,sort from json_to_recordset('"+p_json+"')as rec(code text,sort text,ad text,sort_nulls text)order by abs(sort::numeric)");
    if (newSorts!=null){
    for (int i=0;i<newSorts.size();i++){
      LuiCollection sortCol=newSorts.get(i);
      String sortN=String.valueOf(i+1);
      String sortSign="";
      String sortNull="";
      if (sortCol.Get_Element("sort").charAt(0)=='-'){sortSign="-";
          if (sortCol.Get_Element("sort").charAt(1)=='0') sortNull="0";
      } else {
          if (sortCol.Get_Element("sort").charAt(0)=='0') sortNull="0";
      }
      elements.get(sortCol.Get_Element("code")).setProperty("SORT",sortSign+sortNull+sortN);
    }
    }
   }finally{
    if (lc!=null) application.metaConnPool.SetFree(lc);
   }
}
/**
 * Сочинение запроса без order by
 * @param p_where
 * @return
 * @throws LuiException
 * @throws SQLException 
 */
public String getQueryWithoutOrder(String p_where) throws LuiException, SQLException{
LuiDbConn lc=null;
LuiAppConn ac=null;
String p_whe=p_where;
if (p_where.equals("do$nt$w"))p_whe="";
String  cc=getProperty("Connection");
if (!application.userDbConnections.containsKey(cc))return "";
ac=application.userDbConnections.get(cc);
String retval="";
try {
if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
    String whe="";
    if (!p_where.equals("do$nt$w"))whe=this.addWhere(lc,getProperty("inner_qbe"),false);
    if (is_not_null(p_whe) && is_null(whe)){
     whe="where "+p_whe;
    } else {
    if (is_not_null(p_whe) && is_not_null(whe)){
     whe=whe.replace("where ","where (")+") and ("+p_whe+")";
    }
    }
    retval="select * from ("+this.makeMainQuery()+")q$0 "+whe;//this.addWhere(lc);
    
} finally{
    //if (ac!=null){
      //if (lc!=null){
          if (ac.getSessionMode()=='C')ac.connPool.SetFree(lc);
      //}  
    //}
}
return retval;
}

    /**
     * Сочинение запроса для диаграммы группировки данных
     * @param p_dg_id - id группировки
     * @param p_level - номер диаграммы по порядку
     * @return сочинённый запро для диаграммы группировки данных
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void sendDataGroupConfigList(String p_dg_id,String p_level) throws LuiException, SQLException{
   LuiDbConn lc=null;
   String curGfSeq="";
   
   try{
   lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
   curGfSeq=lc.getSelectResult("select group_function_seq from lui_t_data_group_chart_pref where grouping_id="+
           p_dg_id+" and seq="+p_level);
    ArrayList<LuiCollection> gflist=lc.getSelectAllRowsValMap("select seq,name from lui_t_data_group_function where grouping_id="+p_dg_id+ " order by seq");
    if (gflist==null || gflist.size()<2) return;
    String gfcol="";
    for (int i=0;i<gflist.size();i++){
     if (!curGfSeq.equals(gflist.get(i).Get_Element("seq")))
      gfcol=gfcol+gflist.get(i).Get_Element("seq");
      gfcol=gfcol+":"+gflist.get(i).Get_Element("seq")+"-"+gflist.get(i).Get_Element("name").replace(";",",")+";";
    }
    
    LuiClCommandGroup XMLcg=application.commandWriter.addGroup("DgChartConfList");
    LuiClCommand XMLcmd=XMLcg.addCommand("Form");
    XMLcmd.addScaPar("FormID",getFormID());
    XMLcmd.addScaPar("ConfigList",gfcol);
   } finally {
     if (lc!=null){
         application.metaConnPool.SetFree(lc);
     }  
}
}
public String getDataGroupSelectClause4Chart(String p_dg_id,String p_level,LuiDbConn lc,LuiDbConn mlc,boolean p_native) throws LuiException, SQLException{
    String stmt="";
    Integer cursor=null;
    Connection c;
    c=mlc.getConnection();
    PreparedStatement ps=null;
    // для сочинения фразы select составим список описаний столбцов запроса 
// узнаем количество осей в диаграмме
// String tmp=mlc.getSelectResult("select chart_type from lui_t_data_group_chart_pref where grouping_id="+p_dg_id+" and seq="+p_level);

String tmp=mlc.getSelectResult("select lui_f_dg_get_clause4chart('chart_code',"+p_dg_id+","+p_level+")");
int  axescnt=1;
if (like(tmp,"%3D")) axescnt=2;
// узнаем с какого seq начинается отсчёт осей
tmp=mlc.getSelectResult("select key from lui_f_collection2table(lui_f_dg_get_clause4chart('XYCOL',"+p_dg_id+","+p_level+")) order by key limit 1");
int axesfrom=Integer.valueOf(tmp);
// для этого сначала сочиним простенький запрос
//if (p_native)
//stmt=getQueryWithoutOrder("").substring(8);
//else {
    stmt=mlc.getSelectResult("select last_select from lui_t_data_grouping where id="+p_dg_id).substring(8);
    //stmt=stmt.replace("{ApplicationProperty:SessionID}","0");
//}
String group_by=mlc.getSelectResult("select lui_f_dg_get_clause4chart('group by',"+p_dg_id+","+p_level+")");
stmt=mlc.getSelectResult("select lui_f_dg_get_clause4chart('select',"+p_dg_id+","+p_level+")")+stmt+" "+group_by;
cursor=lc.CursorOpen(stmt); // откроем курсор и 
ArrayList<ColInfo> colsInfo=lc.CursorGetColsInfo(cursor); // получим его столбцы
// теперь составим коллекцию осей

// крутим список столбцов
stmt="select ";
 for(int i=0;i<colsInfo.size();i++){   
    ColInfo cf=colsInfo.get(i);
    String columnfunc=mlc.getSelectResult("select COALESCE(min(sql_expression),'{COLUMN}') from lui_f_dg_get_function_list("+p_dg_id+") where gora='TO_CHART' and dbms_type='"+lc.getDBvendor()+"' and src_datatype='"+cf.columnTypeName+"'");
    String colexp="";
    if (i<axescnt){ // если этот столбец - ось
       colexp=mlc.getSelectResult("select lui_f_dg_get_expression4group("+p_dg_id+","+String.valueOf(i+axesfrom)+")");    
       colexp=columnfunc.replace("{COLUMN}", colexp);
       colexp=colexp+" \"GC"+String.valueOf(i+1)+"\"";
       String oldColExp=mlc.getSelectResult("select colexp from lui_t_data_group_column where grouping_id="+p_dg_id+
               " and seq="+String.valueOf(i+axesfrom));
       if (!NVL(oldColExp,"").equals(colexp)){
       
       ps=c.prepareStatement("update lui_t_data_group_column set colexp=? where grouping_id="+p_dg_id+
               " and seq="+String.valueOf(i+axesfrom));
       ps.setString(1, colexp);
       ps.execute();
       ps.close();
       ps=null;}
    } else { // если не ось
       
       //colexp=mlc.getSelectResult("select lui_f_dg_get_expression4agg("+p_dg_id+","+String.valueOf(i-axescnt+1)+")");     
    /* управление функциями агрегирования из диаграмм
       colexp=columnfunc.replace("{COLUMN}", colexp);
       colexp=colexp+" \"SF"+String.valueOf(i-axescnt+1)+"\"";
       String oldColExp=mlc.getSelectResult("select colexp from lui_t_data_group_function where grouping_id="+p_dg_id+
               " and seq="+String.valueOf(i-axescnt+1));
       if (!NVL(oldColExp,"").equals(colexp)){
       
       ps=c.prepareStatement("update lui_t_data_group_function set colexp=? where grouping_id="+p_dg_id+
               " and seq="+String.valueOf(i-axescnt+1));
       ps.setString(1, colexp);
       ps.execute();
       ps.close();
       ps=null;}
        управление функциями агрегирования из диаграмм */
       continue;
    }
    
    if (i>0) stmt=stmt+",";
    stmt=stmt+colexp;
}
if (cursor!=null) {
    if (lc.CursorisOpen(cursor)){
        lc.CursorClose(cursor);
    }
}
    return stmt;
}
/*public String getDataGroupSelectClause4Chart(String p_dg_id,String p_level) throws LuiException, SQLException{
LuiDbConn lc=null;
LuiDbConn mlc=null;
LuiAppConn ac=null;

String  cc=getProperty("Connection");
ac=application.userDbConnections.get(cc);

try {
mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());    
if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 

return getDataGroupSelectClause4Chart(p_dg_id,p_level,lc,mlc);


} finally{
    if (mlc!=null) application.metaConnPool.SetFree(mlc);
    if (ac!=null){
      if (lc!=null){
          lc.CloseAllCursors();
              
          
          if (ac.getSessionMode()=='C')ac.connPool.SetFree(lc);
      }  
    }
}


       
}*/
/**
 * Формирование условия для data_group-диаграммы
 * @param p_dg_id
 * @param p_level
 * @param p_dim0
 * @param p_dim1
 * @param p_where
 * @return
 * @throws LuiException
 * @throws SQLException 
 */
public String getDataGroupCondition(String p_dg_id,String p_level,String p_dim0,String p_dim1,LuiDbConn lc,LuiDbConn mlc,String p_where,boolean p_native) throws LuiException, SQLException{
if (is_null(p_dim0))return p_where;
Integer cursor=null;
String retVal=p_where;
String stmt="select"+mlc.getSelectResult("select lui_f_dg_get_clause4chart(' ',"+p_dg_id+","+p_level+")");
if (p_native)
  stmt=stmt+getQueryWithoutOrder("").substring(8);
else
  stmt=stmt+mlc.getSelectResult("select last_select from lui_t_data_grouping where id="+p_dg_id).substring(8);
  //stmt=stmt.replace("{ApplicationProperty:SessionID}", "0");
  cursor=lc.CursorOpen(stmt);
  ArrayList<ColInfo> colsInfo=lc.CursorGetColsInfo(cursor);
  LuiCollection xycol=mlc.getSelectAllRowsTwoColsValMap("select key,value from lui_f_collection2table(lui_f_dg_get_clause4chart('xycol',"+p_dg_id+","+p_level+")) order by key");
  int cnt=0;
  String dimCond="";
  ColInfo colinfo=null;
  LuiAppElement nel;
  for (String colNo:xycol.keySet()){
  if (cnt==0) dimCond=p_dim0; 
  else dimCond=p_dim1;
  if (is_not_null(dimCond)) {
    String expr=mlc.getSelectResult("select lui_f_dg_get_expression4group("+p_dg_id+","+colNo+")");  
    if (dimCond.equals("nUlL")){
     dimCond=expr+" is null";
    }  else {
  
  colinfo=colsInfo.get(cnt);
  colinfo.LuiDATATYPE=xycol.get(colNo);
  // попробуем найти конвертер среди функций группировок
  String convexp=mlc.getSelectResult("select sql_expression from lui_f_dg_get_function_list("+p_dg_id+") where gora='FROM_CHART' and dbms_type='"+lc.getDBvendor()+"' and res_datatype='"+colinfo.columnTypeName+"' limit 1");
  if (is_not_null(convexp)){
   dimCond=convexp.replace("{COLUMN}",dimCond);
   dimCond=expr+"="+dimCond;
  } else { // если не нашли конвертируем обычным конвертером
  nel=new LuiAppElement(this,"ITEM","ppp$qqq",0);
  nel.addProperty("DATA_TYPE",xycol.get(colNo));
  nel.colinfo=colinfo;
  dimCond=application.dataConverter.getFromLuiConversion(dimCond, nel, lc);
  dimCond=expr+"="+dimCond;
    } 
  }
  if (is_not_null(retVal))retVal=retVal+" and ";
  retVal=retVal+dimCond;
  
  }
  cnt++;
  }
  if (cursor!=null) {
    if (lc.CursorisOpen(cursor)){
        lc.CursorClose(cursor);
    }
}
return retVal;    
}
/*public String getDataGroupCondition(String p_dg_id,String p_level,String p_dim0,String p_dim1,String p_where) throws LuiException, SQLException{
if (is_null(p_dim0))return p_where;

LuiDbConn lc=null;
LuiDbConn mlc=null;
LuiAppConn ac=null;
int cursor=0;
String  cc=getProperty("Connection");
ac=application.userDbConnections.get(cc);

try {
mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());    
if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
  
  
  return getDataGroupCondition(p_dg_id,p_level,p_dim0,p_dim1,lc,mlc,p_where);
} finally{
    if (mlc!=null) application.metaConnPool.SetFree(mlc);
    if (ac!=null){
      if (lc!=null){
          
              if (lc.CursorisOpen(cursor)){
              lc.CursorClose(cursor);
              }
          
          if (ac.getSessionMode()=='C')ac.connPool.SetFree(lc);
      }  
    }

 
}
}*/
/**
 * Подготовка к демонстрации группировки данных и диаграмм
 */
public void prepareDataGrouping(String p_dg_id) throws LuiException, SQLException{
   LuiDbConn lc=null;    
LuiAppConn ac=null;
LuiDbConn mlc=null;

PreparedStatement pst=null;
String  cc=getProperty("Connection");
ac=application.userDbConnections.get(cc);
try{
mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());    
if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) // если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 

// выясним группировка создана именно для данной формы?
Boolean isnative=false;
String gfc=mlc.getSelectResult("select form_code from lui_t_data_grouping where id="+p_dg_id);
String gff=mlc.getSelectResult("select config_code from lui_t_data_grouping where id="+p_dg_id);
if (getCode().equals(gfc) && getProperty("FEATURE").equals(gff)){
    isnative=true;
}
Connection mcc=mlc.getConnection();
if (isnative){ // если форма текущая совпадает с формой для которой сделана группировка
// читаем старые свойства формирования запроса
LuiCollection oldCol=mlc.getSelectFirstRowValMap("select last_select,last_qbe,db_connect,last_qbe4show,last_order_by,last_qbe4winfunc from lui_t_data_grouping where id="+p_dg_id);
// обновляем last_select и last_qbe и last qbe4winfunc
String mq=this.makeMainQuery();
String me=this.makeMainExpression(mlc, false, true, false);
String last_select="select * from (select "+me+" from ("+mq+")gr$1 )gr$2";
String last_qbe=this.addWhere(lc,getProperty("inner_qbe"),false);
if (is_null(last_qbe)) {
     last_qbe=" /*LUI$_INNER_QBE*/";
 } else {
     last_qbe=last_qbe+" /*ADD_LUI$_INNER_QBE*/";
 }
String last_qbe4winfunc=this.addWhere(lc,getProperty("inner_qbe"),true);
if (is_null(last_qbe4winfunc)) {
     last_qbe4winfunc=" /*LUI$_INNER_QBE4WINFUNC*/";
 } else {
     last_qbe4winfunc=last_qbe4winfunc+" /*ADD_LUI$_INNER_QBE4WINFUNC*/";
 }
String last_qbe4show="";
 for (String qbeRowN:this.listQBE.keySet()){
  HashMap<String,String> QBErow=listQBE.get(qbeRowN);
  for (String ColCode:QBErow.keySet()){
      if (elements.containsKey(ColCode)){
          if (!last_qbe4show.equals(""))last_qbe4show=last_qbe4show+", ";
          last_qbe4show=last_qbe4show+"\""+elements.get(ColCode).getName4show()+"\""+QBErow.get(ColCode);
          //last_qbe4show=last_qbe4show.replace("\""+ColCode+"\"", "\""+elements.get(ColCode).getName4show()+"\"");
      }
  }
 }
 String db_conn=application.parser.plugins.get("DBConnect").compute(cc, properties.get("Connection"));
 LuiCollection oldConnect=new LuiCollection(oldCol.Get_Element("db_connect"));
 oldConnect.Erase_Element("USERNAME");
 oldConnect.Erase_Element("PASSWORDC");
 LuiCollection newConnect=new LuiCollection(db_conn);
 newConnect.Erase_Element("USERNAME");
 newConnect.Erase_Element("PASSWORDC");
 if (!NVL(last_qbe,"").equals(NVL(oldCol.Get_Element("last_qbe"),"")) ||
     !NVL(last_qbe4show,"").equals(NVL(oldCol.Get_Element("last_qbe4show"),"")) ||
     !NVL(last_qbe,"").equals(NVL(oldCol.Get_Element("last_order_by"),"")) ||
     !NVL(last_select,"").equals(NVL(oldCol.Get_Element("last_select"),""))||
     !NVL(last_qbe4winfunc,"").equals(NVL(oldCol.Get_Element("last_qbe4winfunc"),""))||
     !oldConnect.equals(newConnect)
         )
 {
 pst=mcc.prepareStatement("update lui_t_data_grouping set last_select=?, last_qbe=?, db_connect=?, last_qbe4show=?,last_order_by=?,last_qbe4winfunc=?,last_select_safe=null where id="+p_dg_id);
 pst.setString(1, last_select.replace("right outer join Regl_Bok_Sostav ","join Regl_Bok_Sostav ")); ///Ради быстродействия ВПС
 pst.setString(2, last_qbe);
 pst.setString(3,db_conn);
 pst.setString(4,last_qbe4show);
 pst.setString(5,"");//last_order_by
 pst.setString(6,last_qbe4winfunc);
 //try{
 pst.execute();
 pst.close();
 pst=null;
 //} catch (SQLException sex){
 //    String here="here";
 //}
}
 //pst.close();
}// isnsmtive
else { // если форма не соответствует той для которй делалась группировка то надо искать прикладное соединение в по свойствам группировки
       
       LuiCollection co=new LuiCollection(mlc.getSelectResult("select db_connect from lui_t_data_grouping where id="+p_dg_id));
        cc=co.Get_Element("CODE");
        if (application.userDbConnections.containsKey(cc)){
        ac=application.userDbConnections.get(cc);
        if (ac.getSessionMode()=='C') { // если соединение коммунальное
    lc=ac.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); // то берём свободное
        }
        else 
{//в противном случае возьмём текущее cоединение
            lc=ac.currentConnection;
}
            if (lc==null) {// если соединение так и не нашлось, то
            application.msgWriter.write("LUI-00020",". Lui:Null pointer detected!"); 
            return;
            }
        }
        else {
          String name=mlc.getSelectResult("select name from lui_t_data_grouping where id="+p_dg_id);
          
          application.msgWriter.write("LUI-00141",cc,name);
          return;
        }
 String lss=mlc.getSelectResult("select last_select from lui_t_data_grouping where id="+p_dg_id);
 String ls=lss.replace("{ApplicationProperty:SessionID}",application.luiSession.id);
 ls=ls.replace("'{F:P_DG_ID}'", "'"+p_dg_id+"'");
 pst=mcc.prepareStatement("update lui_t_data_grouping set last_select=? , last_select_safe=? where id="+p_dg_id);
 pst.setString(1, ls);
 pst.setString(2, lss);
 pst.execute();
 pst.close();        
}
 
// генерируем для диаграмм фразы select и where и прочее
 // узнаем их количество - т.е. количество диаграмм
 String grByCols=mlc.getSelectResult("select count(*) from lui_t_data_group_column where grouping_id="+p_dg_id);
 int NrgByCols=Integer.valueOf(grByCols); // количество столбцов группировки
 ArrayList<LuiCollection> diagList= mlc.getSelectAllRowsValMap("select seq,lui_f_dg_get_chart_type("+p_dg_id+",seq) chart_type "
         + "from lui_t_data_group_chart_pref where grouping_id="+p_dg_id+" order by seq");
 pst=mcc.prepareStatement("update lui_t_data_group_chart_pref set select_clause=?,where_clause=?,where_clause_0_null=?,where_clause_1_null=?,where_clause_01_null=? where grouping_id="+p_dg_id+" and seq=?");
 for (int i=0;i<diagList.size();i++){
     
   if (NrgByCols<1)break;
   int axes=1;
   if (like(diagList.get(i).Get_Element("chart_type"),"%3D%"))axes=2;
   if (NrgByCols<axes)axes=1;
   NrgByCols=NrgByCols-axes;
   // читаем старые свойства формирования запроса
   LuiCollection oldCol=mlc.getSelectFirstRowValMap("select select_clause,where_clause,where_clause_0_null,where_clause_1_null,where_clause_01_null "
           + "from lui_t_data_group_chart_pref where grouping_id="+p_dg_id+" and seq="+diagList.get(i).Get_Element("seq"));
   String select_clause=getDataGroupSelectClause4Chart(p_dg_id,diagList.get(i).Get_Element("seq"),lc,mlc,isnative);
   String where_clause=getDataGroupCondition(p_dg_id,diagList.get(i).Get_Element("seq"),":p_dim0",":p_dim1",lc,mlc,"",isnative);
   String where_clause_dim0_null=getDataGroupCondition(p_dg_id,diagList.get(i).Get_Element("seq"),"nUlL",":p_dim1",lc,mlc,"",isnative);
   String where_clause_dim1_null=getDataGroupCondition(p_dg_id,diagList.get(i).Get_Element("seq"),":p_dim0","nUlL",lc,mlc,"",isnative);
   String where_clause_dim01_null=getDataGroupCondition(p_dg_id,diagList.get(i).Get_Element("seq"),"nUlL","nUlL",lc,mlc,"",isnative);
   if (!select_clause.equals(oldCol.Get_Element("select_clause")) ||
       !where_clause.equals(oldCol.Get_Element("where_clause"))    ||
       !where_clause_dim0_null.equals(oldCol.Get_Element("where_clause_0_null"))  ||
       !where_clause_dim1_null.equals(oldCol.Get_Element("where_clause_1_null"))  ||              
       !where_clause_dim01_null.equals(oldCol.Get_Element("where_clause_01_null"))                  
           ) {
   pst.setString(1,select_clause);
   pst.setString(2,where_clause);
   pst.setString(3,where_clause_dim0_null);
   pst.setString(4,where_clause_dim1_null);
   pst.setString(5,where_clause_dim01_null);
   pst.setInt(6,Integer.valueOf(diagList.get(i).Get_Element("seq")));
   pst.execute();
   } 
 }
 //if (pst!=null && !pst.isClosed()) pst.close();
 // удаляем undo-информацию и статус изменённости
mlc.executeDML("delete from lui_t_data_group_chart_pv_undo where grouping_id="+p_dg_id);
mlc.executeDML("update lui_t_data_group_chart_pref set current_state_no=null,max_state_no=null where grouping_id="+p_dg_id);
mlc.executeDML("update lui_t_data_grouping set changed_after_read_from_meta=null where id="+p_dg_id);
 } finally {
    if (mlc!=null) {
       if (pst!=null){
         if (!pst.isClosed()) try{pst.close();}catch(Exception ex){}
       } 
        try{application.metaConnPool.SetFree(mlc);}catch(Exception ex){}
    }
    if (ac!=null){
      if (lc!=null){
   //       lc.CloseAllCursors();
         if (ac.getSessionMode()=='C')ac.connPool.SetFree(lc);
          
      }  
    }
}
}

/**
 * Поиск дочерней формы с кодом=pFormCode и выполнение в ней действия pActionCode
 * и выполнение перезапроса списка элементов в дереве редактора
 
public void findChildFormAndRunAction(String pFormCode,String pActionCode) throws LuiException, SQLException{
  for (String fc:application.forms.keySet()){
      if (application.forms.get(fc).code.equals(pFormCode) && 
              application.forms.get(fc).getParentID().equals(this.getFormID())){
          LuiAppForm frm=application.forms.get(fc);
          if (frm.elements.containsKey(pActionCode)){
              frm.runAction(pActionCode);
          }
      }
  }

}*/
/**
 * Поиск работающей формы LUI_L_ELEMENTS по пути родитель-родитель-LUI_L_CONFIGS-ребёнок LUI_L_ELEMENTS
 * и выполнение перезапроса списка элементов в дереве редактора
 */
public void requeryLuiElementsFormID() throws LuiException, SQLException{
  String fc;
  LuiAppForm frm;
  frm=this;
  while(true){
   fc=frm.getParentID();
   if (application.forms.containsKey(fc)) {
       frm=application.forms.get(fc);
       if (frm.code.equals("LUI_L_CONFIGS")) break;
   } else {
       break;
   }
  }
  if (!frm.code.equals("LUI_L_CONFIGS"))   return;
  for (String fcc:application.forms.keySet()){
    if (application.forms.get(fcc).getParentID().equals(frm.getFormID()) && application.forms.get(fcc).code.equals("LUI_L_ELEMENTS")) {
        frm=application.forms.get(fcc);
        break;
    }  
  }
  if (!frm.code.equals("LUI_L_ELEMENTS"))   return;
  frm.requery();
}
/*
 Делает все динамические св-ва формы и элементов формвы не вычисленными и они будут вычисляться при первом к ним обращении
Это надо для повторного вызова REFRESH в действиях
*/
public void resetRecalc() throws LuiException, SQLException{
    for (String pc:properties.keySet()){
        if (properties.get(pc).recalculate!=null)properties.get(pc).recalculate=true;
    }
    for (String ec:elements.keySet()){
      if (elements.get(ec).getType().equals("PARAM") && elements.get(ec).getParent().equals("Form")) continue; 
        for (String pc:elements.get(ec).properties.keySet()){
        if (elements.get(ec).properties.get(pc).recalculate!=null)elements.get(ec).properties.get(pc).recalculate=true;
    }
    }
}
public void showCellLocMenu(String pColumn,Integer Row,LuiClCommand XMLcmd,LuiClCommandGroup XMLcg) throws LuiException, SQLException{
   //Готовим список действий 
    ArrayList<LuiAppElement> actionListCol=new ArrayList<>();
    ArrayList<LuiAppElement> actionListRow=new ArrayList<>();
    String AV;
    String parent;
    LuiAppElement ael;
    for (int i=0;i<this.elements.size();i++) {
        ael=this.elSort(i);
        //if (compareString(frm.getVal("Is_QBE_Mode"),"Y") && !like(ael.getCode(),"CELLMENU$%")) continue;
        if (compareString(this.getVal("Is_QBE_Mode"),"Y") && !like(ael.getCode(),"QBE$ITEM%")) continue;
            if (ael.getType().equals("ACTION") && !like(ael.getCode(),"SYSMENU$%")&& !like(ael.getCode(),"GRPMENU$%") && !like(ael.getCode(),"COLMENU$%")){  
          parent=ael.getParent();  
         if ((parent.equals("Form") && (Row==null || Row>0 ))
                 ||parent.equals(pColumn)){ 
         if (ael.getProperty("Run").equals("Y")) {
          AV=ael.getProperty("DISPLAY");
          if (AV.equals("Y") || (AV.equals("EXIST") && (Row!=null &&Row>0)) || like(AV,"SQL:%") || like (AV,"JS:%")){
            if (is_not_null(ael.getProperty("PROMPT"))){  
             if (parent.equals("Form")){    
               actionListRow.add(ael);
             } else {
               actionListCol.add(ael);  
             } 
            }
          }  
         }
        }
    }
    }
    // слияние
    for (int i=0;i<actionListRow.size();i++) actionListCol.add(actionListRow.get(i));
    // изготовление команд
    for (int i=0;i<actionListCol.size();i++){
       ael=actionListCol.get(i);
       AV=ael.getProperty("DISPLAY");
       if (like(AV,"SQL:%")) { // в DISPLAY лежит SELECT
          ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4), ownerObj.getProperty("Connection"),this.getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          // крутим извлечённые строки
          if (propsArray!=null){
          for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                ael.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            String Prompt=ael.getProperty("PROMPT");
            
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
            if (!ael.isActionActive()){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }
       } else if (like(AV,"JS:%")) { // JSON-размножитель
          try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.application.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного массива
           ael.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          String Prompt=ael.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
                if (!ael.isActionActive()){
            XMLcmd.addScaPar("Enable","N");   
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          }
        } // JSON
         else{ // не Select и не JS
       String Prompt=ael.getProperty("PROMPT");           
       if (is_not_null(Prompt)) {
       XMLcmd=XMLcg.addCommand("AddMenuItem");
       XMLcmd.addScaPar("Code",ael.getCode());
       XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
       if (!ael.isActionActive()){
       XMLcmd.addScaPar("Enable","N");   
       Prompt=RemoveIco(Prompt);
       }
       XMLcmd.addScaPar("Label",Prompt);
       }
       }
    } 
}
/**
 * обработка мышиных кликов в HTML-форме
 * @param pEvent
 * @param pParams
 * @param pElement 
 */
public void htmlMouseEvent(String pEvent,String pParams,String pElement) throws LuiException, SQLException{
    if (pEvent.equals("DBLCLK")) pEvent="DBLCLICK";
    addProperty("HTMLelement",pElement);
    addProperty("HTMLevent",pEvent);
    LuiAppElement el=null;
    for (String ec:elements.keySet()){
     if (!elements.get(ec).getType().equals("ITEM"))continue;
      if (elements.get(ec).getProperty("HTML_ELEMENT").equals(pElement)) {
          el=elements.get(ec);
          pElement=el.Code;
          break;
      }
    }
   
     
    //раскручиваем параметры в свойства элемента или формы
     if (is_not_null(pParams)){
     JsonElement JSONRoot = application.JSONparser.parse(pParams);
     JsonElement JSONEl=null;
     for (Map.Entry<String, JsonElement> pEntry:JSONRoot.getAsJsonObject().entrySet()){
         String pk=pEntry.getKey();
         String pv=pEntry.getValue().getAsString();
         if (el!=null) {
             el.addProperty(pk, pv);
             el.properties.get(pk).gen4html=true;
         }else {
             addProperty(pk,pv);
             properties.get(pk).gen4html=true;
         }
     }
     }
      String action=this.checkHotKey(pEvent,pElement);
      if (is_not_null(action))
          this.runAction(action);
      else {
          if (pEvent.equals("RMB")){
              
              if (el!=null){
                  el.fieldRMB();
              }
              
          }
      }
}

    /**
     * Обработка нажатия ПКМ в ячейке сипска 
     *
     * @param pRow - номер строки
     * @param pContext -контекст коллекция
     * @throws LuiException логическая ошибка
     * @throws SQLException ошибка выполнения SQL-запроса
     */
public void cellRMB(String pColumn,String pRow,String pContext) throws LuiException, SQLException{
  LuiCollection Context=new LuiCollection(pContext);
  LuiCollection CurrentContext = null;
  Integer Row=null;
  int RN=Integer.valueOf(NVL(getVal("CurrentRow"),"0"));
  setVal("ActionColumn",pColumn);
    LuiClCommandGroup XMLcg=application.commandWriter.addGroup("CellRMB");
    XMLcg.addParameter("FormID",this.getFormID());
    XMLcg.addParameter("Type", "Action");
    if (is_not_null(pRow))
    Row=Integer.valueOf(pRow);
    
    if (NVL(pRow,"0").equals("0")){//клик в области грида, где нет строк
        CurrentContext=this.getFieldsValues();
        this.setVal("CurrentRow","");
        this.clearElementsValue();
        this.clearChecking();
    } else {
        this.setElementsValues(Context);
    }
    LuiClCommand XMLcmd=XMLcg.addCommand("LocalMenu");
    showCellLocMenu(pColumn,Row,XMLcmd,XMLcg);
    //Чисто клиентские пункты
    if (application.getProperty("LISTCELL_RMB_DEFAULT_MENU").equals("N"))return;
      XMLcmd=XMLcg.addCommand("AddLocalItems");
      XMLcmd.addScaPar("Group","Local");
      this.setVal("CurrentRow",String.valueOf(RN));
      if (Row==null || Row==0){
          if (CurrentContext!=null) setElementsValues(CurrentContext);
        if (RN>0){
            this.checkRow("Y",this.getUK());
        }  
      }
  //} 
}

/**
 * Инициализация формы с rtChart
 * эта программа вызывается из LUI_C_REALTIME на действии ON_LOAD
 * 
 * @param pDBid
 * @throws LuiException
 * @throws SQLException 
 */
public void RTinit(String pDBid) throws LuiException, SQLException{
 if (RTdashboard!=null) {
if ("run".equals(RTdashboard.getStatus())) {
  RTdashboard.setStatus("stop");
  while (RTdashboard.getStatus().equals("Stop")){
      try {
          Thread.sleep(1000);
      } catch (InterruptedException ex) {
          
      }
      
  }
  }
  RTdashboard.frm=null;
  RTdashboard=null;
  application.setVal("rtChart","");
    if (application.forms.containsKey(getVal("rtParentFormID"))){
  setVal("ParentFormID",getVal("rtParentFormID"));
  } 
 }
 String parentFormID=getVal("ParentFormID");
 if (application.forms.containsKey(parentFormID)){
     application.forms.get(parentFormID).setVal("CurrentAction","");//в родительской форме скажем, что ничего не делается
                                                 // это для того, чтобы в ней все действия работали как ни в чём не бывало
 }
 setVal("rtParentFormID",parentFormID); // запоминаем откуда нас вызвали
 setVal("ParentFormID",""); // отвязываемся от вызывающей формы для предотвращения каскадного закрытия нас
 application.setVal("rtChart",getFormID()); // запомним в приложении ID формы с графиками
 String curWin=getVal("WindowID"); // текущее окно
 if (application.windows.containsKey(curWin)){
     application.windows.get(curWin).setProperty("ParentID",""); // тоже отвязываем от родительского окна
 }

//try { 
 RTdashboard=new LuiRTdashboard(Integer.valueOf(pDBid),application,this.getFormID()); // создаём структуры dashboard
 RTdashboard.start();// запускае тред постоянного тока графиков
 application.setVal("rtDashBoardID",pDBid);
/*} finally {
  if (RTdashboard!=null) {
if ("run".equals(RTdashboard.getStatus())) {
  RTdashboard.setStatus("stop");
  while (RTdashboard.getStatus().equals("Stop")){
      try {
          Thread.sleep(1000);
      } catch (InterruptedException ex) {
          
      }
      
  }
  }
  RTdashboard.frm=null;
  RTdashboard=null;
  application.setVal("rtChart","");
  this.close("", "");
  application.actionBreak();
 }  
} */
} 
/**
 * Переход в форму графиков реального времени
 * может выполняется из любой прикладной формы
 * @return String -- пусто если переход выполнен в живую форму или код формы LUI_C_REALTIME для классического вызова
 * @throws LuiException
 * @throws SQLException 
 */
public void goToRTchart(String pDbId) throws LuiException, SQLException{
  String chartFormID=null;
  if (is_not_null(application.getVal("rtChart"))) { // а имеется ли форма c такими графиками
   if (application.forms.containsKey(application.getVal("rtChart"))){
      chartFormID= application.getVal("rtChart"); // берём её ID
   } 
  }
  if (is_not_null(chartFormID)){ // если форма существует
   LuiAppForm cfrm=application.forms.get(chartFormID); // берём её
   cfrm.RTdashboard.tryReconnect();// попытка восстановлений утраченных соединений
   cfrm.setVal("rtParentFormID",this.getFormID()); // и запоминаем в ней ID прикладной формы
   String chartWinID=cfrm.getVal("WindowID"); // берём окно формы с графиками
   String curWinID=getVal("WindowID"); //берём окно текущей прикладной формы
   LuiClCommandGroup XMLcg=application.commandWriter.addGroup("Go to rtChart");
      LuiClCommand  XMLcmd;
     //XMLcmd=XMLcg.addCommand("DisableWindow");  // деактивируем окно прикладной формы
     //XMLcmd.addScaPar("WindowID",curWinID);
     application.windows.get(curWinID).disCover(XMLcg,null); // это новый способ гашения
     XMLcmd=XMLcg.addCommand("EnableWindow"); // активируем окно с графиками
     XMLcmd.addScaPar("WindowID",chartWinID);
     XMLcmd=XMLcg.addCommand("Focus"); // фокус на форму с графиками
     XMLcmd.addScaPar("FormID",chartFormID);
   if (is_not_null(pDbId)){  
    if (cfrm.elements.containsKey("P_DB_ID")){
     if (cfrm.elements.get("P_DB_ID").properties.containsKey("Value")){
        if (!cfrm.elements.get("P_DB_ID").properties.get("Value").currentValue.equals(pDbId)){
            cfrm.elements.get("P_DB_ID").addProperty("Value", pDbId);
            cfrm.chartClear(XMLcg);
            cfrm.reloadForm();
            return;
        } 
     }
    } 
   }
   //cfrm.RTdashboard.refreshDetails("");
  }
}
/**
 * Возврат из формы c графиками реального времени в прикладную форму
 * вызывается в контексте формы с графиками реального времени
 * возвращает true если такой форзврат выполнен
 * Возврат выполняется нажатием Escape или спец действия в форме с графиками реального времени
     * @return boolean
 **/
public boolean backFromRTchart() throws LuiException, SQLException{
   if (application.forms.containsKey(getVal("rtParentFormID"))){ // если есть форма вызвашая графики
       String parentFormID=getVal("rtParentFormID");
       String parentWindowID=application.forms.get(parentFormID).getVal("WindowID");
       String chartWindowID=getVal("WindowID");
       LuiClCommandGroup XMLcg=application.commandWriter.addGroup("Back from rtChart");
       LuiClCommand  XMLcmd;
     //LuiClCommand  XMLcmd=XMLcg.addCommand("DisableWindow");  // деактивируем окно графиков
     //XMLcmd.addScaPar("WindowID",chartWindowID);
               application.windows.get(chartWindowID).disCover(XMLcg,null); // это новый способ деактивации
     XMLcmd=XMLcg.addCommand("EnableWindow"); // активируем окно прикладной формы
     XMLcmd.addScaPar("WindowID",parentWindowID);
     XMLcmd=XMLcg.addCommand("Focus"); // фокус на форму с графиками
     XMLcmd.addScaPar("FormID",parentFormID);
     application.forms.get(parentFormID).setVal("CurrentAction", "");
     application.setVal("FormID",parentFormID);
     return true;
   }
   return false; 
}
public String getNames4rtCurve() throws LuiException, SQLException{
    ArrayList<String> names=new ArrayList<>();
LuiCollection uk=getUK();
String ukn="";
for (String ue:uk.keySet()){
  ukn=ukn+uk.Get_Element(ue)+"/";  
}
ukn=ukn+elements.get(getVal("CurrentColumn")).getName4show();
names.add(ukn);
names.add(elements.get(getVal("CurrentColumn")).getName4show());
for(int i=0;i<elements.size();i++){ // крутим его элементы (там бывают только параметры
LuiAppElement el=elSort(i); 
if (!el.getType().equals("ITEM")) continue;
if (el.getCode().equals(getVal("CurrentColumn")))continue;
if (!el.isVisible(true)) continue;
names.add(el.getValue());
}
return application.toJSON(names);
}
/**
 * Инициализация диаграмм группировок данных
 * @param pDgId id группировки
 * @param pLevel уровень
 * @param pWhere условие сгенерированное вызывающей диаграммой
 * @param pChartTitle заголовок диаграммы сгенерированный в вызывающей диаграмме
 * @throws LuiException
 * @throws SQLException 
 */
public void initDgV2(String pDgId,String pLevel,String pWhere,String pChartTitle) throws LuiException, SQLException{
  DgV2=new LuiDgV2(this,pDgId,pLevel,pWhere,pChartTitle);
  //this.chartClear(null);
  //this.chartDestroy(null);
  if (is_not_null(DgV2.mapCode)){
      application.gMap.registerMap(DgV2.mapCode, false,null);
          addProperty("MAP_FORM_CODE",DgV2.mapCode);
          addProperty("MAP_APP_CODE",application.gMap.maps.get(DgV2.mapCode).appCode);
          addProperty("MAP_NAME",application.gMap.maps.get(DgV2.mapCode).name);
          //DgV2.sendConfigList();
          //addProperty("OldLabelSize","12");
          //addProperty("CurrentLabelSize","");
  }
}
/* добавление к группировке сохранённой в метаданных входных параметров*/
public void DgV2GenInpParam2Meta(String pDgFormCode) throws LuiException, SQLException{
    if (!elements.containsKey("FOR_DG2SAVE_AS_FORM"))return ;
    if (!elements.get("FOR_DG2SAVE_AS_FORM").getType().equals("PARAM"))return;
    LuiCollection pCol=new LuiCollection(elements.get("FOR_DG2SAVE_AS_FORM").getProperty("Value"));
    if (pCol.size()<1) return;
   LuiDbConn mlc=null;
   PreparedStatement ps1=null;
   PreparedStatement ps2=null; 
   String safe;
   try {
       mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); 
       safe=mlc.getSelectResult("select current_setting('application_name')");
       mlc.executeDML("set session application_name='psql'");
       mlc.executeDML("delete from lui_t_element where etype='PARAM' and element_code is null and form_code='"+pDgFormCode+"' and app_code='"+application.code+"'");
       ps1=mlc.getConnection().prepareStatement("insert into lui_t_element(etype,app_code,form_code,code,name,seq)values ('PARAM',?,?,?,?,?)");
       ps2=mlc.getConnection().prepareStatement("select lui_r_save_value(?,?,'DEFAULT','VALUE',?,?,null)");
       int n=1;
       for (String pk:pCol.keySet()){
       LuiAppElement pel=null;
       if (elements.containsKey(pk)){
           pel=elements.get(pk);}
       else {
          for (String ok:elements.keySet()){
            if (!elements.get(ok).getParent().equals("Form")){
              if (elements.get(ok).getType().equals("PARAM")){
                if (elements.get(ok).getProperty("NAME").equals(pk)) {pel=elements.get(ok);break;}
              }  
            }  
          } 
       }
       if (pel==null)continue;
       ps1.setString(1,application.code);
       ps1.setString(2,pDgFormCode);
       ps1.setString(3,pk);
       ps1.setString(4,pel.getName4show());
       ps1.setInt(5,n);
       ps1.execute();
       ps2.setString(1,application.code);
       ps2.setString(2,pDgFormCode);
       ps2.setString(3,pk);
       String pv="Value";
       if (!pel.getParent().equals("Form")) pv="VALUE";
       if (pCol.get(pk).equals("C")) ps2.setString(4,pel.getProperty(pv));
       else ps2.setString(4,pel.properties.get(pv).startedValue);
       ps2.execute();
       n++;
       }
       mlc.executeDML("set session application_name='"+safe+"'");
   } catch(SQLException sex){
       String here="here";
   }
      finally {
       try{
       if (ps1!=null && !ps1.isClosed())ps1.close();
       }catch (Exception ex){}
       try{
       if (ps2!=null && !ps2.isClosed())ps2.close();
       }catch (Exception ex){}
       if (mlc!=null)application.metaConnPool.SetFree(mlc);
       
   }
    /*ArrayList<LuiCollection>aList = new ArrayList<>();
    for (String pk:pCol.keySet()){
       if (!elements.containsKey(pk))continue;
       LuiCollection aListElement = new LuiCollection("");
       aListElement.Add_Element("code",pk);
       aListElement.Add_Element("name",elements.get(pk).getName4show());
       if (pCol.Get_Element(pk).equals("C"))
       aListElement.Add_Element("value",elements.get(pk).getProperty("Value"));
       else
       aListElement.Add_Element("value",elements.get(pk).properties.get("Value").startedValue);
       aList.add(aListElement);
    }
 return   application.toJSON(aList);
 */
}

/*
 создание выходных параметров графика
*/
public void DgV2output(String pDgId,String pLevel) throws SQLException, LuiException{
   LuiDbConn mlc=null;
   LuiDbConn lc=null;    
   PreparedStatement ps=null;
   LuiAppConn appConn=null;
   LuiAppElement nel;
   String value;
   String expr;
   String columnTypeName;
   boolean nonPredicatable=false;
   try {
       mlc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser()); 
       String usrConnCode=getProperty("Connection");
       if (application.userDbConnections.containsKey(usrConnCode)){
       appConn=application.userDbConnections.get(usrConnCode);
       if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
        } else {
            String name=mlc.getSelectResult("select name from lui_t_data_grouping where id="+pDgId);
            application.msgWriter.write("LUI-00141",usrConnCode,name);
            return;
        }
       // удаляем ранее сгенерированное этим и более низкими уровнями
       mlc.executeDML("delete from lui_t_data_grouping_output where grouping_id="+pDgId+" and chart_seq>="+pLevel);
       // количество осей
       int axisN=1;
       
       String chartT=mlc.getSelectResult("select lui_f_dg_get_chart_type("+pDgId+","+pLevel+")");
       if (like(chartT,"%3D"))axisN=2;
       // начало осей
       String startFrom=mlc.getSelectResult("select sum((case lui_f_dg_get_chart_type("+pDgId+",seq) "
               + "when 'BAR3D' then '2'\n" +
"                       when 'SCT3d' then '2'\n" +
"                       else '1' END)::NUMERIC) \n" +
"from lui_t_data_group_chart_pref where grouping_id="+pDgId+" and seq<"+pLevel);
       if (is_null(startFrom)) startFrom="0";
       // извлекаем описание группирующих столбцов
       ArrayList<LuiCollection> GClist=mlc.getSelectAllRowsValMap("select src_column,res_datatype,coalesce(lui_f_dg_get_chart_pref_val("+pDgId+",1,'DIMENSION_NAME'||seq::text,null),name) \"name\" ,hint,colexp"
               + " from lui_t_data_group_column "
               + "where grouping_id="+pDgId+" and seq>"+startFrom+" order by seq limit "+String.valueOf(axisN));
       if (GClist==null) return;
       ps=mlc.getConnection().prepareStatement("insert into lui_t_data_grouping_output(grouping_id,chart_seq,code,name,src_column,datatype,hint,value,predicat,predicat4show,predicat_template)"
               + "values("+pDgId+","+pLevel+",?,?,?,?,?,?,?,?,?)");
       for(int i=0;i<GClist.size();i++){
           LuiCollection GCcol=GClist.get(i);
           ps.setString(1,"P$"+pLevel+"GC"+String.valueOf(i+1));
           ps.setString(2, GCcol.Get_Element("name"));
           ps.setString(3, GCcol.Get_Element("src_column"));
           ps.setString(4, GCcol.Get_Element("res_datatype"));
           ps.setString(5, GCcol.Get_Element("hint"));
           expr=GCcol.Get_Element("colexp");
           expr=expr.replace(" \"GC"+String.valueOf(i+1)+"\"","");
           value=null;
           if (DgV2!=null && getType().equals("CHART")){ // если есть график
           switch (DgV2.chartType) {
                   case "WHR":
                   value=getProperty("ClickSeriesName");
                   break;
                   case "MAP":   
                   case "PIE":
                   value=getProperty("ClickSeriesDataName");    
                   break;
                   case "BAR":
                   case "BARG":
                   case "SCT":
                   case "LINE":  
                   case "BAR3D":
                   case "SCT3D":
                   value=getProperty("ClickSeriesDataValueDim"+String.valueOf(i));
                   break;   
            }
           if (is_null(value)) nonPredicatable=true;
           } else { // если нет, то график отобпажён списком
               value=elements.get("GC"+String.valueOf(i+1)).getProperty("Value");
           } 
        if (nonPredicatable) {
            value=null;
            ps.setString(6,value);
            ps.setString(7,value);
            ps.setString(8,value);
            ps.setString(9,value);
        }else {
        if (value==null || value.equals("nUlL")){
            value=null;
            ps.setString(6,value);
            ps.setString(7, expr+" is null");
            ps.setString(8, GCcol.Get_Element("name")+":");
            ps.setString(9, expr.replace("\""+GCcol.Get_Element("src_column")+"\"","{COLUMN}")+" is null");
        }else {
            ps.setString(6,value);
            ps.setString(8, GCcol.Get_Element("name")+":"+value);
            // конвертируем значение к нужному типу данных
            // попробуем найти конвертер среди функций группировок
            if (DgV2!=null)
            columnTypeName=DgV2.colsinfo.get(i).columnTypeName;
            else
            columnTypeName=elements.get("GC"+String.valueOf(i+1)).colinfo.columnTypeName;
           String convexp=mlc.getSelectResult("select sql_expression from lui_f_dg_get_function_list("+pDgId+") where gora='FROM_CHART' and dbms_type='"+lc.getDBvendor()+"' and res_datatype='"+columnTypeName+"' limit 1");
           if (is_not_null(convexp)){
               value=convexp.replace("{COLUMN}",value);
           } else {// если не нашли конвертируем обычным конвертером 
               if (elements.containsKey("GC"+String.valueOf(i+1))) { // если уже есть нужный элемент
               nel=elements.get("GC"+String.valueOf(i+1)); // берём
               } else { // нет - создаём
                  nel=new LuiAppElement(this,"ITEM","GC"+String.valueOf(i+1),i+1); 
                  nel.colinfo=DgV2.colsinfo.get(i);
               }
               value=application.dataConverter.getFromLuiConversion(value, nel, lc);
           }
           ps.setString(7, expr+"="+value);
           ps.setString(9, expr.replace("\""+GCcol.Get_Element("src_column")+"\"","{COLUMN}")+"="+value);
        } 
       }
         ps.execute();
       }
       // теперь кликнутая кривая
       // вычислим номер кликнутой кривой и значение
       value=null;
       int SFid=0;
       String SFs="SF1";
       
       if (DgV2==null){
           if (like(getVal("CurrentColumn"),"SF%")){
               SFid=Integer.valueOf(getVal("CurrentColumn").replace("SF", ""))-1;
               SFs=getVal("CurrentColumn");
           }
           value=elements.get(SFs).getProperty("Value");
       } else {
       if (DgV2.chartType.equals("MAP")||DgV2.chartType.equals("PIE")||DgV2.chartType.equals("WHR")){
           if (is_null(getVal("UserChartDGConfiguration"))){
               setVal("UserChartDGConfiguration","0");
           } else {
              try{
              SFid=Integer.valueOf(getVal("UserChartDGConfiguration"));
              }catch(Exception ex){SFid=0;}
              SFs="SF"+String.valueOf(SFid+1);
           }
       value=getProperty("ClickSeriesDataValue");
       } else {
          if (is_null(getProperty("ClickSeriesСomponentIndex"))){
              
           } else {
              SFid=Integer.valueOf(getProperty("ClickSeriesСomponentIndex"));
              SFs="SF"+String.valueOf(SFid+1);
              value=getProperty("ClickSeriesDataValueDim"+String.valueOf(SFid+GClist.size()));
           } 
       }
       }
       // управление функциями агрегирования LuiCollection SFcol=mlc.getSelectFirstRowValMap("select src_column,name,res_datatype,hint,colexp from lui_t_data_group_function where grouping_id="+pDgId+" and seq="+String.valueOf(SFid+1));
       LuiCollection SFcol=mlc.getSelectFirstRowValMap("select src_column,name,res_datatype,hint,lui_f_dg_get_expression4agg(grouping_id,seq) \"colexp\" from lui_t_data_group_function where grouping_id="+pDgId+" and seq="+String.valueOf(SFid+1));
       ps.setString(1,"P$"+pLevel+"SFNAME");
       ps.setString(2, application.NLS.NLS("RUS:Имя выбранного показателя;ENG:Selected indicator name;"));
       ps.setString(3, SFs);
       ps.setString(4, "CHAR");
       ps.setString(5, "");
       if (nonPredicatable){
           value=null;
            ps.setString(6,value);
            ps.setString(7,value);
            ps.setString(8,value);
            ps.setString(9,value);
       } else {
           ps.setString(6, SFcol.Get_Element("name"));
           ps.setString(7, String.valueOf(SFid));
           ps.setString(8, application.NLS.NLS("RUS:Показатель;ENG:indicator;")+":"+SFcol.Get_Element("name"));
           ps.setString(9, String.valueOf(SFid));
       }
       ps.execute();
       ps.setString(1,"P$"+pLevel+"SFVALUE");
       ps.setString(2, application.NLS.NLS("RUS:Значение выбранного показателя;ENG:Indicator value;"));
       if (nonPredicatable){
         value=null;  
         ps.setString(3,value);
         ps.setString(4,value);
         ps.setString(5,value);
         ps.setString(6,value);
         ps.setString(7,value);
         ps.setString(8,value);
         ps.setString(9,value);
       } else {
       ps.setString(3, SFcol.Get_Element("src_column"));
       ps.setString(4, SFcol.Get_Element("res_datatype"));
       ps.setString(5, SFcol.Get_Element("hint"));
          expr=SFcol.Get_Element("colexp");
           expr=expr.replace(" \"SF"+String.valueOf(SFid+1)+"\"","");
           if (value==null || value.equals("nUlL")){
            value=null;
            ps.setString(6,value);
            ps.setString(7, expr+" is null");
            ps.setString(8, SFcol.Get_Element("name")+":");
            ps.setString(9, expr.replace("\""+SFcol.Get_Element("src_column")+"\"","{COLUMN}")+" is null");
            }else {
            ps.setString(6,value);
            ps.setString(8, SFcol.Get_Element("name")+":"+value);
            // конвертируем значение к нужному типу данных
            // попробуем найти конвертер среди функций группировок
            if (DgV2!=null)
            columnTypeName=DgV2.colsinfo.get(SFid+GClist.size()).columnTypeName;
            else
            columnTypeName=elements.get(SFs).colinfo.columnTypeName;
           String convexp=mlc.getSelectResult("select sql_expression from lui_f_dg_get_function_list("+pDgId+") where gora='FROM_CHART' and dbms_type='"+lc.getDBvendor()+"' and res_datatype='"+columnTypeName+"' limit 1");
           if (is_not_null(convexp)){
               value=convexp.replace("{COLUMN}",value);
           } else {// если не нашли конвертируем обычным конвертером 
               if (elements.containsKey(SFs)) { // если уже есть нужный элемент
               nel=elements.get(SFs); // берём
               } else { // нет - создаём
                  nel=new LuiAppElement(this,"ITEM",SFs,GClist.size()+SFid); 
                  nel.colinfo=DgV2.colsinfo.get(GClist.size()+SFid);
               }
               value=application.dataConverter.getFromLuiConversion(value, nel, lc);
           }
           ps.setString(7, expr+"="+value);
           ps.setString(9, expr.replace("\""+SFcol.Get_Element("src_column")+"\"","{COLUMN}")+"="+value);
        }
       }
          
         ps.execute();
   } finally {
       if (ps!=null)ps.close();
       if (mlc!=null)application.metaConnPool.SetFree(mlc);
       if (lc!=null){
        if (appConn.getSessionMode()=='C')  appConn.connPool.SetFree(lc);    
       }
   }
}
// генерация команд с кортежами матрицы старый вариант
/*public void MatrixGenTupleCommands(String pParent,String pTupleType,LuiClCommandGroup XMLcg) throws LuiException, SQLException {
    LuiClCommand XMLcmd;
    for(int i=0;i<elements.size();i++){
        LuiAppElement el=elSort(i);
        if (el.getType().equals("ITEM")&&el.getParent().equals(pParent)){
            XMLcmd=XMLcg.addCommand(pTupleType);
            XMLcmd.addScaPar("Code",el.getCode());
            if (!pParent.equals("Form")){
                if (!elements.get(pParent).getType().equals("DIMENSION"))
                XMLcmd.addScaPar("Parent",pParent);
            }
            XMLcmd.addScaParameter("Caption",el.getProperty("CAPTION"),def(""));
            XMLcmd.addScaParameter("H-align",LuiString.withCapitalLetter(el.getProperty("ALIGNMENT")),def("Center"));
            XMLcmd.addScaParameter("V-align",LuiString.withCapitalLetter(el.getProperty("VALIGN")),def("Center"));
            XMLcmd.addScaParameter("Overflow",LuiString.withCapitalLetter(el.getProperty("OVERFLOW")),def("Expand"));
            XMLcmd.addScaParameter("Class",el.getProperty("VISUALIZATION"),def(""));
            XMLcmd.addScaParameter("MakeUp",el.getProperty("MAKEUP"),def(""));
            XMLcmd.addScaParameter("GetUp",el.getProperty("GETUP"),def(""));
            String dir="H";
            switch (el.getProperty("DIRECTION")){
                case "TOP2BOTTOM":dir="TB";break;
                case "BOTTOM2TOP":dir="BT";break;
            }
            XMLcmd.addScaParameter("Direction",dir,def("H"));
            XMLcmd.addScaParameter("Hint",el.getProperty("HINT"),def(""));
            MatrixGenTupleCommands(el.getCode(),pTupleType,XMLcg);
        }
    }
} */
/**
 * Изготовление иерархии кортежей матриц упрощённое т.е. считаем что кортежи уже сортированы
 * 
 */
public void MatrixMakeItemIerarSimpleRec(ArrayList<LuiCollection> pItemIerar,ArrayList<LuiAppElement> ellist,String parent)throws LuiException, SQLException{
    for (int i=0;i<ellist.size();i++){
        LuiAppElement el=ellist.get(i);
        if (!el.getParent().equals(parent))continue;
        LuiCollection lci=new LuiCollection("");
        lci.Add_Element("PARENT",parent);
        lci.Add_Element("CODE",el.Code);
        pItemIerar.add(lci);
        if (!el.getProperty("LEAF").equals("Y")) {
        MatrixMakeItemIerarSimpleRec(pItemIerar,ellist,el.Code);
        } else {
            String here="here";
        }
    }
}
 /*public static Comparator<Double> MatrixMakeItemIerarSimpleComparator = new Comparator<Double>() {
 
        @Override
        public int compare(Double d1, Double d2) {
            if (d1<d2) return -1;
    if (d1>d2) return 1;
    return 0;
        }
    };
 */
public void MatrixMakeItemIerarSimple2(String parent,LinkedHashMap<String,ArrayList<String>> cortageTree, ArrayList<LuiCollection> pCortIerar)throws LuiException, SQLException{
    if (cortageTree.containsKey(parent)){
        for (int i=0;i<cortageTree.get(parent).size();i++){
            LuiCollection lci=new LuiCollection("");
            lci.Add_Element("PARENT", parent);
            lci.Add_Element("CODE", cortageTree.get(parent).get(i));
            pCortIerar.add(lci);
            MatrixMakeItemIerarSimple2(cortageTree.get(parent).get(i),cortageTree,pCortIerar);
        }
    }
}
public ArrayList<LuiCollection> MatrixMakeItemIerarSimple(String fParent,String pType) throws LuiException, SQLException{
    ArrayList<Double> seqlist=new ArrayList<>();
    for (String ec:elements.keySet()){
        LuiAppElement el=elements.get(ec);
        if (!el.getType().equals(pType)) continue;
        seqlist.add(el.sourceSeq);
    }
    Collections.sort(seqlist);
    //,this.MatrixMakeItemIerarSimpleComparator);
    ArrayList<LuiAppElement> ellist=new ArrayList<>();
    for (int i=0;i<seqlist.size();i++){
        for (String ec:elements.keySet()){
        LuiAppElement el=elements.get(ec);
        if (!el.getType().equals(pType)) continue;
        if (el.sourceSeq!=seqlist.get(i)) continue;
        ellist.add(el);
        }
    }
     ArrayList<LuiCollection> pItemIerar=new ArrayList<>();
    MatrixMakeItemIerarSimpleRec(pItemIerar,ellist,fParent);
        return pItemIerar;
}
/**
 * Изготовление иерархии кортежей матриц новых 05.02.21
 * 
 */
public void MatrixMakeItemIerar(ArrayList<LuiCollection>pItemIerar,String pParent,String pType) throws LuiException, SQLException{
    for (int i=0;i<elements.size();i++){
      LuiAppElement el=elSort(i); 
      if (!el.getType().equals(pType)) continue;
      if (!el.getParent().equals(pParent))continue;
      LuiCollection lci=new LuiCollection("");
      lci.Add_Element("PARENT",pParent);
      lci.Add_Element("CODE",el.Code);
      pItemIerar.add(lci);
      MatrixMakeItemIerar(pItemIerar,el.Code,pType);
    }
}
public void MatrixCopyElement(String fromCode,String toCode,String newType) throws LuiException, SQLException{
    LuiAppElement oldEl=elements.get(fromCode);
    LuiAppElement newEl;
    if (elements.containsKey(toCode)){
        newEl=elements.get(toCode);
     for (String pk:oldEl.properties.keySet()){
         if (pk.equals("Parent")||pk.equals("Code")||pk.equals("Name")||pk.equals("EType")||pk.equals("Seq")||pk.equals("Metadata")||pk.equals("MetadataElement"))continue;
         if (newEl.properties.containsKey(pk)) {
             //if (!newEl.properties.get(pk).is_meta && oldEl.properties.get(pk).is_meta) {
             if (newEl.properties.get(pk).genesis=='A' && oldEl.properties.get(pk).genesis!='A') {
              newEl.addProperty(pk,oldEl.properties.get(pk).startedValue);   
             }
         }else {
             newEl.addProperty(pk,oldEl.properties.get(pk).startedValue);
         }
     }   
    } else {
    newEl=addElement(toCode,newType,oldEl.sourceSeq+0.001,oldEl.getProperty("Name"),oldEl.getParent());
    oldEl.sourceSeq=oldEl.sourceSeq+0.001;
    for (String pk:oldEl.properties.keySet()){
        if (pk.equals("Parent")||pk.equals("Code")||pk.equals("Name")||pk.equals("EType")||pk.equals("Seq"))continue;
           newEl.addProperty(pk,oldEl.properties.get(pk).startedValue);
    }
    }
}
public void MatrixChangeCell(String cCode,String rCode) throws LuiException, SQLException{
    String r="";
    String c="";
    addProperty("CurrentCOL",cCode);
        if (is_not_null(cCode))   { addProperty("CurrentCOLN",elements.get(cCode).getProperty("ltn"));c=elements.get(cCode).getProperty("ltn");}
    else addProperty("CurrentCOLN","");
    addProperty("CurrentROW",rCode);
    if (is_not_null(rCode))  {addProperty("CurrentROWN",elements.get(rCode).getProperty("ltn"));r=elements.get(rCode).getProperty("ltn");}
    else addProperty("CurrentROWN","");
    addProperty("CurrentCELL","");
    if (is_not_null(r)&& is_not_null(c)) {
        if (MatrixMap.containsKey(r+":"+c)){
            ArrayList<String> acell=MatrixMap.get(r+":"+c);
            if (acell!=null && acell.size()>0) addProperty("CurrentCELL",acell.get(acell.size()-1));
        }
    }
    addProperty("CurrentTuple","");
    if (is_not_null(getProperty("CurrentCOL")) && is_null(getProperty("CurrentROW"))){
    addProperty("CurrentTuple",getProperty("CurrentCOL"));    
    }
    if (is_null(getProperty("CurrentCOL")) && is_not_null(getProperty("CurrentROW"))){
    addProperty("CurrentTuple",getProperty("CurrentROW"));    
    }
    /*if (is_not_null(cCode) && is_not_null(rCode)){      
    for (String ek:elements.keySet()){
        if (!elements.get(ek).getType().equals("CELL"))continue;
        if (elements.get(ek).getProperty("ROW").equals(rCode) && elements.get(ek).getProperty("COL").equals(cCode))
        {addProperty("CurrentCELL",ek);break;}
    }
    } */
    this.changes();
}
public void MatrixPrepareToolBar(LuiClCommandGroup pXMLcg) throws LuiException, SQLException{
    if (getProperty("ToolBar").equals("N"))return;  
    LuiClCommand XMLcmd=null;     
    LuiCollection toolBarButtons=new LuiCollection("");
    for (int isys=0;isys<2;isys++){
         int toolBarDecade=0;
        //for (int i=0;i<elements.size();i++){
        //LuiAppElement ael=elSort(i);
        for (int i=0;i<elSortedList.size();i++){
        LuiAppElement ael=elSortedList.get(i);
        if (!ael.getType().equals("ACTION"))continue; 
        if (isys>0 && like(ael.getCode(),"SYSMENU$%"))continue;
        if (isys==0 && !(like(ael.getCode(),"SYSMENU$%")))continue;
        if (!ael.getNonCompValue("ACTIVE").equals("N")) { // Действие не вечно заблокировано 
        if (like(ael.getNonCompValue("ROLE"),"%TIME%")) {
                      if (is_null(getVal("Timer"))) 
                       setVal("Timer","Y"); // Признак необходимости "тикания таймером"
                   }
                   if (like(ael.getNonCompValue("ROLE"),"%TIME500%")) {
                       setVal("Timer","500"); // Признак необходимости "тикания таймером"
                   }
        }
        if (is_null(ael.getProperty("BUTTON")))continue;
        if (!ael.getProperty("Run").equals("Y")) continue; // если действие разрешено
        if (!ael.getProperty("ACTIVE").equals("Y")) continue; // и активно
        String button=ael.getProperty("BUTTON");
        String icon=LuiString.GetIcoFromSquareBracket(button);
        String prompt=LuiString.RemoveIco(button);
        if (toolBarButtons.containsKey(icon)){ // если эта иконка уже в тулбаре 
         this.addIcon2ToolBar(icon,"","", toolBarButtons); // гасим промпт иконки
         } else {
           // если её ещё в тулбаре нет - выводим её хинт
                            // но перед выводом сравним декады
                            if (isys>0){
                            if ((ael.sourceSeq/10)!=toolBarDecade){ // если декады разные ставим сепаратор
                             String keySep="";
                             for(int j=0;j<toolBarDecade;j++) keySep=keySep+" ";
                             addIcon2ToolBar(keySep,"","", toolBarButtons); 
                             toolBarDecade=(int)ael.sourceSeq/10;
                            }
                            this.addIcon2ToolBar(icon,prompt,ael.getProperty("EXPLANATION"), toolBarButtons);
                            }   
        }
    }
}
    String ntb=this.getToolBarString(toolBarButtons);
    if (!getVal("ToolBarButtons").equals(ntb)){
    setVal("ToolBarButtons",ntb); // установим значение св-ва ToolBarButtons
        XMLcmd=pXMLcg.addCommand("Form");    
      
    XMLcmd.addScaPar("ToolBarButtons",getVal("ToolBarButtons"));
    }
}

public void MatrixRMB(String cCode,String rCode) throws LuiException, SQLException{
    String r="";
    String c="";
    addProperty("CurrentCOL",cCode);
    if (is_not_null(cCode))   { addProperty("CurrentCOLN",elements.get(cCode).getProperty("ltn"));c=elements.get(cCode).getProperty("ltn");}
    else addProperty("CurrentCOLN","");
    addProperty("CurrentROW",rCode);
    if (is_not_null(rCode)) {addProperty("CurrentROWN",elements.get(rCode).getProperty("ltn"));r=elements.get(rCode).getProperty("ltn");}
    else addProperty("CurrentROWN","");
    addProperty("CurrentCELL","");
    if (is_not_null(r)&& is_not_null(c)) {
        if (MatrixMap.containsKey(r+":"+c)){
            ArrayList<String> acell=MatrixMap.get(r+":"+c);
            if (acell!=null && acell.size()>0) addProperty("CurrentCELL",acell.get(acell.size()-1));
        }
    
    }
    addProperty("CurrentTuple","");
    if (is_not_null(getProperty("CurrentCOL")) && is_null(getProperty("CurrentROW"))){
    addProperty("CurrentTuple",getProperty("CurrentCOL"));    
    }
    if (is_null(getProperty("CurrentCOL")) && is_not_null(getProperty("CurrentROW"))){
    addProperty("CurrentTuple",getProperty("CurrentROW"));    
    }
    /*if (is_not_null(cCode) && is_not_null(rCode)){
    for (String ek:elements.keySet()){
        if (!elements.get(ek).getType().equals("CELL"))continue;
        if (elements.get(ek).getProperty("ROW").equals(rCode) && elements.get(ek).getProperty("COL").equals(cCode))
        {addProperty("CurrentCELL",ek);
        break;}
    }
    } */
   LuiClCommandGroup XMLcg=null;
   LuiClCommand XMLcmd=null;
    for (int i=0;i<elSortedList.size();i++){
        LuiAppElement ael=elSortedList.get(i);
        if (!ael.getType().equals("ACTION"))continue;
        if (like(ael.getCode(),"SYSMENU$%")) continue;// исключаем действия системного меню 
        if (is_null(ael.getProperty("PROMPT")))continue;
        if (!ael.getProperty("Run").equals("Y")) continue; // если действие разрешено
        if (ael.getNonCompValue("ACTIVE").equals("N")) continue; // Действие не вечно заблокировано 
        String AV=ael.getProperty("DISPLAY");
        String Prompt=ael.getProperty("PROMPT");
        if (AV.equals("")||(!like(AV,"SQL:%") && !like(AV,"JS:%"))){ //не SQL и не JS
        if (XMLcg==null){
            XMLcg=application.commandWriter.addGroup("LocalMenu");
            XMLcmd=XMLcg.addCommand("LocalMenu");
            XMLcmd.addScaPar("FormID",getFormID());
            XMLcmd.addScaPar("Type", "Action");
        }
        XMLcmd=XMLcg.addCommand("AddMenuItem");
        XMLcmd.addScaPar("Code",ael.getCode());
          
        if (!ael.getProperty("ACTIVE").equals("Y")){
          XMLcmd.addScaPar("Enable","N");
          Prompt=RemoveIco(Prompt);
        }
        XMLcmd.addScaPar("Label",Prompt);
        XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
        }
        if (like(AV,"SQL:%")){ //SQL размножитель
           ArrayList<String> ac=application.getAppConnAndStmt(AV.substring(4),getProperty("Connection"),getFormID());
          LuiAppConn appConn=getApplication().userDbConnections.get(ac.get(0));
          LuiDbConn lc=null;
          try {
              if(appConn.getSessionMode()=='C') { // если работаем через пул коммунальных
              lc=appConn.connPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
              } else lc=appConn.currentConnection;
          // извлекаем данные из select'а
          ArrayList<LuiCollection> propsArray=lc.getSelectAllRowsValMap(ac.get(1));
          if (propsArray!=null){
          // крутим извлечённые строки
          for (int j=0;j<propsArray.size();j++){
            LuiCollection props=propsArray.get(j); // читаем проперти каждой строки
            for (String propkey:props.keySet()){ // крутим их
                ael.addProperty(propkey, props.Get_Element(propkey)); // изменяем или пополняем проперти действия
            }
            Prompt=ael.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
             if (XMLcg==null){
              XMLcg=application.commandWriter.addGroup("LocalMenu");
              XMLcmd=XMLcg.addCommand("LocalMenu");
              XMLcmd.addScaPar("FormID",getFormID());
              XMLcmd.addScaPar("Type", "Action");
                            }
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
            if (!compareString(ael.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
          }
          }
          } finally {
                  if (lc!=null) {
                     
                      if (appConn.getSessionMode()=='C') {
                          appConn.connPool.SetFree(lc);
                      }
                  }
              }  
        }
        if (like(AV,"JS:%")){ //JS размножитель
           try { 
          jdk.nashorn.api.scripting.ScriptObjectMirror Arr=(jdk.nashorn.api.scripting.ScriptObjectMirror) this.application.JSengine.eval(AV.substring(3));
          jdk.nashorn.api.scripting.ScriptObjectMirror props;
          for (String Ind:Arr.keySet()) { // крутим элементы верхего массива
          props = (jdk.nashorn.api.scripting.ScriptObjectMirror) Arr.get(Ind);
          for (String propkey:props.keySet()){ // крутим элементы вложенного 
           ael.addProperty(propkey.toString(), props.get(propkey).toString()); // изменяем или пополняем проперти действия
           }
          Prompt=ael.getProperty("PROMPT");
            if (is_not_null(Prompt)) {
             if (XMLcg==null){
            XMLcg=application.commandWriter.addGroup("LocalMenu");
            XMLcmd=XMLcg.addCommand("LocalMenu");
            XMLcmd.addScaPar("FormID",getFormID());
            XMLcmd.addScaPar("Type", "Action");
        }
            XMLcmd=XMLcg.addCommand("AddMenuItem");
            XMLcmd.addScaPar("Code",ael.getCode()+"-"+String.valueOf(Math.abs(Prompt.hashCode())));
            XMLcmd.addScaPar("Group",ael.getParent()+String.valueOf(ael.sourceSeq/10));
            if (!compareString(ael.getProperty("ACTIVE"),"Y")){
            XMLcmd.addScaPar("Enable","N");    
            Prompt=RemoveIco(Prompt);
            } 
            XMLcmd.addScaPar("Label",Prompt);
            }
           }
          } catch (ScriptException sex) {
              
          } catch (Exception ex){
              
          }  
             
        }
    }
    //this.changes();
}
/** реакция на выделение ячеек матрицы
 * 
 */
public void MatrixSelect(String pColumns,String pRows,String pCells) throws LuiException, SQLException{
    addProperty("SelectedColumns",pColumns);
    addProperty("SelectedRows",pRows);
    addProperty("SelectedCells",pCells);
    ArrayList<LuiCollection> selCelN=new ArrayList<>();
    ArrayList<LuiCollection> selCelC=new ArrayList<>();
    String mapKey;
    JsonElement JSONRoot = application.JSONparser.parse(pCells);
    JsonArray JSONArray=JSONRoot.getAsJsonArray();
    for (int mr=0;mr<JSONArray.size();mr++){
      LuiCollection rcn=new LuiCollection("");
      LuiCollection rcc=new LuiCollection("");
        for (Map.Entry<String, JsonElement> pEntry:JSONArray.get(mr).getAsJsonObject().entrySet()){
                      String pk=pEntry.getKey();
                      String pv=pEntry.getValue().getAsString();
                      if (!elements.containsKey(pv))continue;
                      rcc.Add_Element(pk, pv);
                      if (!elements.get(pv).properties.containsKey("ltn"))continue;
                      rcn.Add_Element(pk,elements.get(pv).getProperty("ltn"));
       }
    selCelN.add(rcn);
    mapKey=rcn.Get_Element("row")+":"+rcn.Get_Element("col");
    if (MatrixMap!=null && MatrixMap.containsKey(mapKey) && MatrixMap.get(mapKey)!=null && MatrixMap.get(mapKey).size()>0){
        rcc.Add_Element("code", MatrixMap.get(mapKey).get(0));
        selCelC.add(rcc);
    }
    }  
    addProperty("SelectedCellsC",application.toJSON(selCelC));
    addProperty("SelectedCellsN",application.toJSON(selCelN));
    if (is_not_null(getVal("ParentFormID"))){
        if (application.forms.containsKey(getVal("ParentFormID"))){
            LuiAppForm waitedCellsParent=application.forms.get(getVal("ParentFormID"));
            if (waitedCellsParent.properties.containsKey("SelectedCellsForm")){
              if (pCells.equals("[]"))
                waitedCellsParent.addProperty("SelectedCellsForm","");
              else waitedCellsParent.addProperty("SelectedCellsForm",this.getFormID());
            }
        }
    }
    addProperty("SelectedRCdomains",this.MatrixCreateRCdomains());
}
public String MatrixGetCellVisual(String pMk)throws LuiException, SQLException{
    if (!MatrixMap.containsKey(pMk)){
        String words[]=pMk.split(":",2);
        String r=words[0];
        String c=words[1];
        if (!elements.containsKey(r))return"";
        if (!elements.containsKey(c))return"";
        r=elements.get(r).getProperty("ltn");
        c=elements.get(c).getProperty("ltn");
        pMk=r+":"+c;
        if (!MatrixMap.containsKey(pMk))return "";
    }
    String rc="";
    ArrayList<String> acell=MatrixMap.get(pMk);
    for (int i=0;i<acell.size();i++){
        if (!elements.containsKey(acell.get(i)))continue;
        LuiAppElement cell=elements.get(acell.get(i));
        if (!cell.properties.containsKey("CELL_VISUAL"))continue;
        rc=rc+" "+cell.getProperty("CELL_VISUAL");
    }
    return LuiString.LTrim(rc);
}
public LuiAppProp MatrixGetCellP(String pMk,String pPropCode) throws LuiException, SQLException{
    if (MatrixMap==null)return null;
    if (!MatrixMap.containsKey(pMk))return null;
    ArrayList<String> acell=MatrixMap.get(pMk);
     LuiAppProp p=null;
    for (int i=acell.size()-1;i>=0;i--){
        if (!elements.containsKey(acell.get(i)))continue;
        LuiAppElement cell=elements.get(acell.get(i));
        if (!cell.properties.containsKey(pPropCode))continue;
        p=cell.properties.get(pPropCode);
        if (p.genesis!='D')return p;
    }
    if (p!=null)return p;
    return null;
}
public String MatrixGetCellProperty(String pMk,String pPropCode) throws LuiException, SQLException{
    
    if (!MatrixMap.containsKey(pMk))return "";
    ArrayList<String> acell=MatrixMap.get(pMk);
     LuiAppProp p=null;
    for (int i=acell.size()-1;i>=0;i--){
        if (!elements.containsKey(acell.get(i)))continue;
        LuiAppElement cell=elements.get(acell.get(i));
        if (!cell.properties.containsKey(pPropCode))continue;
        p=cell.properties.get(pPropCode);
        if (p.genesis!='D')return p.getComputedProperty();
    }
    if (p!=null)return p.getComputedProperty();
    return "";
}

/*public void MatrixMakeMap() throws LuiException, SQLException{
    if (MatrixMap!=null)MatrixMap.clear();
    else MatrixMap=new HashMap<>();
    String r;
    String c;
    String rn;
    String cn;
    String mk;
    ArrayList<String> cellArray;
     for (int i=0;i<elements.size();i++){//-- перебираем все ячейки по порядку
            LuiAppElement elcel=elSort(i);
            if (!elcel.getType().equals("CELL"))continue;
            r=elcel.getProperty("ROW");
            c=elcel.getProperty("COL");
            if (elements.containsKey(r)&& elements.containsKey(c)){
                rn=elements.get(r).getProperty("ltn");
                cn=elements.get(c).getProperty("ltn");
                elcel.addProperty("RowNo",rn);
                elcel.addProperty("ColNo",cn);
                mk=rn+":"+cn;
                if (!MatrixMap.containsKey(mk)){
                    cellArray=new ArrayList<>();
                    cellArray.add(elcel.Code);
                    MatrixMap.put(mk, cellArray);
                } else {
                    MatrixMap.get(mk).add(elcel.Code);
                }
            }
         }
}*/
public String MatrixIsCellSeleced(String sc,int rn,int cn) throws LuiException, SQLException{
     // узнаём код элемента ROW 
                 String ROW="";
                 for (String rowCode:elements.keySet()){
                 if (!elements.get(rowCode).getType().equals("ROW"))continue;
                 if (!elements.get(rowCode).getProperty("ltn").equals(String.valueOf(rn)))continue;
                 ROW=rowCode;break;
                 }
                 // узнаём код элемента COL 
                 String COL="";
                 for (String colCode:elements.keySet()){
                 if (!elements.get(colCode).getType().equals("COL"))continue;
                 if (!elements.get(colCode).getProperty("ltn").equals(String.valueOf(cn)))continue;
                 COL=colCode;break;
                 }
                 String search="{\"row\":\""+ROW+"\",\"col\":\""+COL+"\"}";
                 
                 if (sc.indexOf(search)<0 ) return null;
                 return sc.replace(search,"");
                 
}
/**
 * Создание перечня RC-доменов на основе выбранных ячеек
 * @return 
 */
public String MatrixCreateRCdomains() throws LuiException, SQLException{
    String sc="";
    // составляем массиы отмеченных ячеек
Integer r1=null;
Integer r2=null;
Integer c1=null;
Integer c2=null;
if (!getProperty("SelectedCells").equals("[]")){
   String sel=getProperty("SelectedCells").replace("},{","}{").replace("[", "").replace("]","");
   String newsel=null;
   int rows=Integer.valueOf(getProperty("LastROW"));
   int cols=Integer.valueOf(getProperty("LastCOL")); 
   while (is_not_null(sel)){ //пока массив выделенных не пуст
        r1=null;
        r2=null;
        c1=null;
        c2=null;
       // найдём левую верхнюю
       for (int r=1;r<=rows;r++){
       for (int c=1;c<=cols;c++){
        newsel=MatrixIsCellSeleced(sel,r,c);
        if (newsel!=null){
          sel=newsel;  r1=r;r2=r;c1=c;c2=c; break;
        }
       }
        if (r1!=null)break;
       }
       if (r1==null)break;
       //левый верхний найден двигаемся по столбцам тойже строки
       if (c2!=null && r2!=null)
       for (int c=c2+1;c<=cols;c++){
         newsel=MatrixIsCellSeleced(sel,r2,c);
        if (newsel!=null){ // если соседка слева выделена
          sel=newsel; c2=c;
        } else { // соседка слева не выделена надо посмотреть на следующую строку и прсмреть выделенность в ней всех ячеек
           break; 
        }  
       }
       //теперь двигаемся вниз и смотрим выделенность в строке всех ячеек от с1 до с2
       if (r2!=null && c2!=null)
       for (int r=r2+1;r<=rows;r++){
                  String safecell=sel;
                  boolean yes=true;
              for (int c=c1;c<=c2;c++) {
                  newsel=MatrixIsCellSeleced(sel,r,c);
                  if (newsel!=null){sel=newsel;}
                  else {sel=safecell;yes=false;break;}
              }
              if(!yes) { // оказалось не все выделены
                  break;
              } else r2=r;
            }
       if (r1!=null) {
          if (r1==r2 && c1==c2)
          sc=sc+"{"+String.valueOf(r1)+","+String.valueOf(c1)+"}";    
          else
          sc=sc+"{"+String.valueOf(r1)+","+String.valueOf(c1)+":"+String.valueOf(r2)+","+String.valueOf(c2)+"}";
       }
   }
}
   
    return sc.replace("}{","},{");
    
}
/**
 * Создание и обновление диаграммы матриц
 * @param pRdepth
 * @param pCdepth
 * @param pGCNames
 * @param pGCTypes
 * @return
 * @throws LuiException
 * @throws SQLException 
 */
public String MatrixCreateNewDgChart(int pRdepth,int pCdepth, String pGCNames,String pGCTypes) throws LuiException, SQLException{
String DgId=null;
LuiDbConn lc=null;
PreparedStatement pstmt=null;
PreparedStatement pstmt2=null;
Connection con=null;
String Select;
String stmt;
LuiCollection Cols=null;
LuiCollection Hints=null;
LuiCollection GCNames=new LuiCollection(pGCNames);
int GCShift =0;
LuiCollection GCChartTypes=new LuiCollection(pGCTypes);
ArrayList<LuiCollection> O$array=new ArrayList<>();
String matrix_data_show_mode=null;  
String matrix_data_fixed_cells=null;
String matrix_category_axes=null;
String mca;
String sc=getProperty("SelectedRCdomains");
int rdepth=0;
int cdepth=0;

try{
   lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
   // поищем похожую по сигнатуре параметров
   int Axes=1+pRdepth;
   String pSignature=String.valueOf(pRdepth)+";"+String.valueOf(pCdepth)+";"+NVL(pGCNames,"")+";"+NVL(pGCTypes,"")+";";
   pSignature=pSignature.replace("'","`");
   ArrayList<String> oldDG=lc.getSelectFirstRowValList("select id,matrix_data_show_mode,matrix_data_fixed_cells,matrix_category_axes "
           + " from lui_t_data_grouping dg where dg.app_code='"+application.code+"' and "
           + "dg.meta_app_code='"+application.code+"' and dg.form_code='"+this.code+"' and dg.config_code='"+getProperty("FEATURE")+"' and matrix_signature='"
           +pSignature+"'");
   if (oldDG==null){ // есди не нашли - создаём
   DgId=lc.getSelectResult("select nextval('lui_q_data_grouping')");
       
   } else {
 //      newCh=false;
       DgId=oldDG.get(0);
       matrix_data_show_mode=NVL(oldDG.get(1),"AS");
       matrix_data_fixed_cells=oldDG.get(2);
       if (matrix_data_show_mode.equals("F") || matrix_data_show_mode.equals("FI")) {
           sc=matrix_data_fixed_cells;
       }
       
       matrix_category_axes=oldDG.get(3);
   }
   LuiCollection GroupFunctions=MatrixCopy2TempData(NVL(matrix_data_show_mode,"AS"),pRdepth,pCdepth,lc,sc);
   mca=NVL(matrix_category_axes,"A");
   if (mca.equals("A")){
       if (Integer.valueOf(GroupFunctions.Get_Element("COLS"))>Integer.valueOf(GroupFunctions.Get_Element("ROWS"))){
          mca="C";
       } else  mca="R";
   }
           
           
   if (mca.equals("R")){
        LuiCollection O$col=new LuiCollection("");
        O$col.Add_Element("CODE","O$1");
        O$col.Add_Element("NAME",application.NLS.NLS("RUS:№ строки;ENG:Row#;"));
       if (GroupFunctions.Get_Element("RI2NUM").equals("::numeric"))
        O$col.Add_Element("DATATYPE","NUMBER");
       else
        O$col.Add_Element("DATATYPE","CHAR");
       O$array.add(O$col);
   Cols=   
   lc.getSelectAllRowsTwoColsValMap("select * FROM("
           + "select distinct 'C'||c C,'' v from lui_t_tem_data where c<=0 and r>0 and form_id='"+getFormID()+"' and session_id="+application.luiSession.id+
   "union all select 'C'||c C,value v from lui_t_tem_data where r=0  and form_id='"+getFormID()+"' and session_id="+application.luiSession.id+
   ")q order by replace(C,'C','')::lui_d_id");
   Hints=lc.getSelectAllRowsTwoColsValMap("select 'C'||c,hint from lui_t_tem_data where r=0  and form_id='"+getFormID()+"' and session_id="+application.luiSession.id);
   }
   if (mca.equals("C")){ 
       LuiCollection O$col=new LuiCollection("");
        O$col.Add_Element("CODE","O$1");
        O$col.Add_Element("NAME",application.NLS.NLS("RUS:№ столбца;ENG:Column#;"));
       if (GroupFunctions.Get_Element("CI2NUM").equals("::numeric"))
        O$col.Add_Element("DATATYPE","NUMBER");
       else
        O$col.Add_Element("DATATYPE","CHAR");
       O$array.add(O$col);
   Cols=   
   lc.getSelectAllRowsTwoColsValMap("select * FROM("
           + "select distinct 'C'||r C,'' v from lui_t_tem_data where r<=0 and c>0 and form_id='"+getFormID()+"' and session_id="+application.luiSession.id+
   "union all select 'C'||r C,value v from lui_t_tem_data where c=0  and form_id='"+getFormID()+"' and session_id="+application.luiSession.id+
   ")q order by replace(C,'C','')::lui_d_id");
   Hints=lc.getSelectAllRowsTwoColsValMap("select 'C'||r,hint from lui_t_tem_data where c=0  and form_id='"+getFormID()+"' and session_id="+application.luiSession.id);
   }
   if (mca.equals("RC")){ 
    LuiCollection O$col=new LuiCollection("");
        O$col.Add_Element("CODE","O$1");
        O$col.Add_Element("NAME",application.NLS.NLS("RUS:№ строки;ENG:Row#;"));
       if (GroupFunctions.Get_Element("RI2NUM").equals("::numeric"))
        O$col.Add_Element("DATATYPE","NUMBER");
       else
        O$col.Add_Element("DATATYPE","CHAR");
       O$array.add(O$col);
       O$col.clear();
        O$col.Add_Element("CODE","O$2");
        O$col.Add_Element("NAME",application.NLS.NLS("RUS:№ столбца;ENG:Column#;"));
       if (GroupFunctions.Get_Element("CI2NUM").equals("::numeric"))
        O$col.Add_Element("DATATYPE","NUMBER");
       else
        O$col.Add_Element("DATATYPE","CHAR");
       O$array.add(O$col);
       
   Cols=new LuiCollection("");    
   // глубина по строкам и cтолбцам
   rdepth=Integer.valueOf(lc.getSelectResult("select count(distinct c) from lui_t_tem_data c where c<=0 and form_id='"+getFormID()+"' and session_id="+application.luiSession.id));
   cdepth=Integer.valueOf(lc.getSelectResult("select count(distinct r) from lui_t_tem_data c where r<=0 and form_id='"+getFormID()+"' and session_id="+application.luiSession.id));
   for (int i=0;i<rdepth+cdepth;i++){
      String key="C"+String.valueOf(0-(rdepth+cdepth-1-i));
      Cols.Add_Element(key,"");
      
   }
      Cols.Add_Element("C1",application.NLS.NLS("RUS:Показатель;ENG:Value;"));
      GCShift=cdepth;
      Hints=new LuiCollection("");
   }
   if (Cols==null || Cols.size()==0)return null;
   Axes=0;
   for(String ck:Cols.keySet()){
      if (Integer.valueOf(ck.replace("C",""))<=0){
        if (is_null(Cols.Get_Element(ck))){
           String kGC=String.valueOf(Integer.valueOf(ck.replace("C",""))+GCShift);
           String name=GCNames.Get_Element(kGC);
           
           if (is_null(name))name=LuiString.LPad("",' ', Axes+1);
                   Cols.replace(ck, name);
        }  
        Axes++;  
      }
   }
   con=lc.getConnection();
   // сочиняем запрос для группировки данных
   Select="";
   
   if (mca.equals("RC")){
        Select="select * from (select * from (select ri"+GroupFunctions.Get_Element("RI2NUM")+" \"O$1\" \n";
        Select=Select+",ci"+GroupFunctions.Get_Element("СI2NUM")+" \"O$2\" \n";
        
   for (String ck:Cols.keySet()){ // крутим столбцы
       Select=Select+",\""+ck+"\"";
       if (Integer.valueOf(ck.replace("C",""))>0)Select=Select+"::lui_d_number";
       Select=Select+" \""+ck+"\"\n";
   }
       Select=Select+" from (select ri,ci";
       int cn=(0-(rdepth+cdepth));
       for(int i=0;i<rdepth;i++){
           cn++;
           Select=Select+"\n,lui_f_get_tem_data(r,"+String.valueOf(1-rdepth+i)+",form_id,session_id) \"C"+String.valueOf(cn)+"\" ";
       }
       for(int i=0;i<cdepth;i++){
           cn++;
           Select=Select+"\n,lui_f_get_tem_data("+String.valueOf(1-cdepth+i)+",c,form_id,session_id) \"C"+String.valueOf(cn)+"\" ";
       }
       Select=Select+"\n,value \"C1\" from lui_t_tem_data ";
       Select=Select+"\n where form_id='"+getFormID()+"' and session_id="+application.luiSession.id+" and r>0 and c>0)vps$)q$0 )q$0";
   }
   if (mca.equals("R")){
   Select="select * from (select * from (select ri"+GroupFunctions.Get_Element("RI2NUM")+" \"O$1\" \n";
   for (String ck:Cols.keySet()){ // крутим столбцы
       Select=Select+",\""+ck+"\"";
       if (Integer.valueOf(ck.replace("C",""))>0)Select=Select+"::lui_d_number";
       Select=Select+" \""+ck+"\"\n";
   }
   Select=Select+" from (select ri ";
   for (String ck:Cols.keySet()){ // крутим столбцы
    Select=Select+"\n,lui_f_get_tem_data(row.r,"+ck.replace("C","")+",row.form_id,row.session_id) \""+ck+"\"";
   }
   Select=Select+"\n from lui_t_tem_data row ";
   Select=Select+"\n where form_id='"+getFormID()+"' and c=0 and session_id="+application.luiSession.id+" and r>0)vps$)q$0 )q$0";
   }
   if (mca.equals("C")){ 
   Select="select * from (select * from (select ci"+GroupFunctions.Get_Element("CI2NUM")+" \"O$1\" \n";
   for (String ck:Cols.keySet()){ // крутим столбцы
       Select=Select+",\""+ck+"\"";
       if (Integer.valueOf(ck.replace("C",""))>0)Select=Select+"::lui_d_number";
       Select=Select+" \""+ck+"\"\n";
   }
   Select=Select+" from (select ci ";
   for (String ck:Cols.keySet()){ // крутим столбцы
    Select=Select+"\n,lui_f_get_tem_data("+ck.replace("C","")+",row.c,row.form_id,row.session_id) \""+ck+"\"";
   }
   Select=Select+"\n from lui_t_tem_data row ";
   Select=Select+"\n where form_id='"+getFormID()+"' and r=0 and session_id="+application.luiSession.id+" and c>0)vps$)q$0 )q$0";
   }
   stmt="insert into lui_t_data_grouping"+
"(id,user_name,name, meta_app_code, app_code, form_code, config_code, last_select, \n" +
"db_connect, is_stored, cube_func,last_qbe,last_qbe4show,run_from_main_menu,\n" +
"run_without_list, last_order_by, is_not_grouping, narrow, dg_form_code, session_id,matrix_data_show_mode,matrix_signature,matrix_category_axes,matrix_data_fixed_cells)\n" +
"VALUES (?::lui_d_id,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?::lui_d_id,?,?,?,?) on conflict(id) do update set "
           + "last_select=EXCLUDED.last_select,matrix_data_fixed_cells=EXCLUDED.matrix_data_fixed_cells";
   pstmt=con.prepareStatement(stmt);
   pstmt.setString(1,DgId);
   pstmt.setString(2,application.luiSession.userName);
   // название группировки
   boolean b=true;
   String name=Jsoup.parse(getProperty("HEADER")).text();
 /*  int n=1;
   while (b) {
       String oldname=lc.getSelectResult("select 1 from lui_t_data_grouping where app_code='"+application.code+"' and form_code='"
               +this.code // это надо будет поправить если формы будут копироваться в VIRTUAL
               +"' and name='"+name+"'");
       if (is_null(oldname)) b=false;
       else {
           n++;
           String words[]=name.split("\\(", 2);
           name=words[0];
           name=name+"("+String.valueOf(n)+")";
       }
   }*/
   pstmt.setString(3, name);
   pstmt.setString(4, application.code);
   pstmt.setString(5, application.code);
   pstmt.setString(6, this.code);// это надо будет поправить если формы будут копироваться в VIRTUAL
   pstmt.setString(7, getProperty("FEATURE"));// это надо будет поправить если формы будут копироваться в VIRTUAL
   pstmt.setString(8, Select);
   // соединение
   LuiAppProp p=addProperty("dbConn","{DBConnect:LUI$}");
   pstmt.setString(9, p.getComputedProperty());
   pstmt.setString(10,"Y");
   pstmt.setString(11,"rollup");
   pstmt.setString(12,"/*LUI$_INNER_QBE*/");
   pstmt.setString(13,null);
   pstmt.setString(14,"N");
   pstmt.setString(15,"N");
   pstmt.setString(16,null);
   //pstmt.setString(16,"\"O$1\"");
   pstmt.setString(17,"N");
   pstmt.setString(18,"N");
   pstmt.setString(19,null);
   pstmt.setString(20,application.luiSession.id);
   pstmt.setString(21,matrix_data_show_mode);
   pstmt.setString(22,pSignature);
   pstmt.setString(23,matrix_category_axes);
   pstmt.setString(24,sc);
   
   pstmt.execute();
   pstmt.close();
   //lui_t_data_group_src_column
   int seq=0;
   stmt="INSERT INTO lui_t_data_group_src_column"
     + "(grouping_id,src_column,seq,        name,        src_datatype,src_format,pk,parent_code,parent_name,parent_seq, hint,width,visible,variants,sintetic,definition)\n" +
"VALUES (?::lui_d_id,?,         ?::lui_d_id,?,           ?,           ?,         ?, ?,          ?,          ?::lui_d_id,?,   ?,    ?,      ?,       ?,       ?)";
//       1           2          3           4            5            6          7  8           9           10          11   12    13      14       15       16
   stmt=stmt+" on conflict(grouping_id,src_column) do update set src_column=EXCLUDED.src_column "
           + ",name=EXCLUDED.name "
           + ",src_datatype=EXCLUDED.src_datatype "
           + ",src_format=EXCLUDED.src_format "
           + ",pk=EXCLUDED.pk "
           + ",parent_code=EXCLUDED.parent_code "
           + ",parent_name=EXCLUDED.parent_name "
           + ",parent_seq=EXCLUDED.parent_seq "
           + ",hint=EXCLUDED.hint "
           + ",width=EXCLUDED.width "
           + ",visible=EXCLUDED.visible "
           + ",sintetic=EXCLUDED.sintetic "
           + ",definition=EXCLUDED.definition ";
   pstmt=con.prepareStatement(stmt);
   for (int i=0;i<O$array.size();i++){
   LuiCollection o$col=O$array.get(i);
   pstmt.setString(1,DgId);
   pstmt.setString(2,o$col.Get_Element("CODE"));  
   pstmt.setString(3,String.valueOf(++seq));
   pstmt.setString(4,o$col.Get_Element("NAME"));
   pstmt.setString(5,o$col.Get_Element("DATATYPE"));
   pstmt.setString(6,null); // format
   pstmt.setString(7,"Y"); //pk
   pstmt.setString(8,null);
   pstmt.setString(9,null);
   pstmt.setString(10,"-9999999");
   pstmt.setString(11,o$col.Get_Element("NAME"));
   pstmt.setString(12,null);
   pstmt.setString(13,"Y");
   pstmt.setString(14,"LOVQ:AUTO$");
   pstmt.setString(15,null);
   pstmt.setString(16,o$col.Get_Element("CODE"));
   pstmt.execute();}
   
   for (String ck:Cols.keySet()){ // крутим столбцы для src_col
    pstmt.setString(2,ck);  
    pstmt.setString(3,String.valueOf(++seq));
    pstmt.setString(4,Cols.Get_Element(ck));
    pstmt.setString(5,"CHAR");
    pstmt.setString(6,null); // format
    pstmt.setString(7,"N"); //pk
    pstmt.setString(11,NVL(Hints.Get_Element(ck),Cols.Get_Element(ck)));//hint
    pstmt.setString(16,ck);
    if (Integer.valueOf(ck.replace("C",""))>0){
    pstmt.setString(5,"NUMBER"); 
    String format="";
    for(String ckey:elements.keySet()){
        if (!elements.get(ckey).getType().equals("COL"))continue;  // подозрительно
        if (!elements.get(ckey).getProperty("ltn").equals(ck.replace("C","")))continue;
        LuiAppElement colel=elements.get(ckey);
        p=colel.addProperty("fractional", "{MatrixSemantics:fractional}");
        int f=Integer.valueOf(NVL(p.getComputedProperty(),"2"));
        format="FM999999999990."+RPad("0",'0', f);
        break;
    }
    pstmt.setString(6,format); // format
    }
    pstmt.execute();
   }
   pstmt.close();
   // групповые функции lui_t_data_group_function
   stmt="INSERT INTO lui_t_data_group_function"
      + "( grouping_id, seq,          func, src_column, src_datatype, res_datatype, format, name, hint)\n" +
" VALUES (?::lui_d_id ,?::lui_d_id,   ?,    ?,          ?,            ?            ,?      ,?    ,?   )";
//        1            2              3     4           5             6             7       8     9      
   stmt=stmt+" on conflict(grouping_id,seq) do update set "
           +  "func=EXCLUDED.func "
           + ",src_column=EXCLUDED.src_column "
           + ",src_datatype=EXCLUDED.src_datatype "
           + ",res_datatype=EXCLUDED.res_datatype "
           + ",format=EXCLUDED.format "
           + ",name=EXCLUDED.name "
           + ",hint=EXCLUDED.hint";
   pstmt=con.prepareStatement(stmt);
   pstmt.setString(1,DgId);
   
   pstmt.setString(5,"NUMBER");
   pstmt.setString(6,"NUMBER");
   int sfCount=0;
   
   for (String ck:Cols.keySet()){ // крутим столбцы для функций группировки данных
     if (Integer.valueOf(ck.replace("C",""))>0){
         pstmt.setString(2,String.valueOf(++sfCount));
         if (mca.equals("R")||mca.equals("C"))
         pstmt.setString(3,NVL(GroupFunctions.Get_Element(ck),"SUM"));
         else
         pstmt.setString(3,"SUM");
         pstmt.setString(4,ck);
         String format="";
    for(String ckey:elements.keySet()){
        if (!elements.get(ckey).getType().equals("COL"))continue; // подозрительно
        if (!elements.get(ckey).getProperty("ltn").equals(ck.replace("C","")))continue;
        LuiAppElement colel=elements.get(ckey);
        p=colel.addProperty("fractional", "{MatrixSemantics:fractional}");
        int f=Integer.valueOf(NVL(p.getComputedProperty(),"2"));
        format="FM999999999990."+RPad("0",'0', f);
        break;
    }
    pstmt.setString(7,format); // format
    pstmt.setString(8,Cols.Get_Element(ck));
    pstmt.setString(9,NVL(Hints.Get_Element(ck),Cols.Get_Element(ck)));//hint);
    pstmt.execute();
     }
   }
    pstmt.close();
    // группирующие столбцы   
    stmt="INSERT INTO lui_t_data_group_column"
         + "(src_column,src_datatype,res_datatype,name,lov,hint,colexp,order_by,order_by_mode,order_by_datatype,grouping_id,seq)\n" +
" VALUES (   ?,         ?,           ?,           ?,   ?,  ?,   ?,     ?,       ?,            ?,                ?::lui_d_id,?::lui_d_id)";
//           1          2            3            4    5   6    7      8        9             10                11          12
  
   pstmt=con.prepareStatement(stmt);
   pstmt.setString(11,DgId);
   pstmt.setString(2,"CHAR");
   pstmt.setString(3,"CHAR");
   
   stmt="update lui_t_data_group_column set "
            + " src_column=? "
           + ",src_datatype=? "
           + ",res_datatype=? "
            + ",name=? "
           + ",lov=? "
           + ",hint=? "
              + ",colexp=? "
           + ",order_by=? "
           + ",order_by_mode=? "
           + ",order_by_datatype=? "
           + "where grouping_id=?::lui_d_id and seq=?::lui_d_id";
   pstmt2=con.prepareStatement(stmt);
   pstmt2.setString(11,DgId);
   pstmt2.setString(2,"CHAR");
   pstmt2.setString(3,"CHAR");
   seq=0;
   for (String ck:Cols.keySet()){ // крутим столбцы для группирующих столбцов
     if (Integer.valueOf(ck.replace("C",""))<=0){
      //определим номер сортирующего элемента
      int nse=O$array.size()-Axes+seq;
      
      pstmt2.setString(12,String.valueOf(++seq));
      pstmt2.setString(1,ck);
      pstmt2.setString(4,Cols.Get_Element(ck));
      pstmt2.setString(5,"LOVQ:AUTO$");
      pstmt2.setString(6,NVL(Hints.Get_Element(ck),Cols.Get_Element(ck)));
     // pstmt.setString(9,"\""+ck+"\" \"GC"+String.valueOf(seq)+"\"");
      pstmt2.setString(7,"\""+ck+"\" \"GC1\"");
      if (nse>=0){
      pstmt2.setString(8,"\""+O$array.get(nse).Get_Element("CODE")+"\"");
      pstmt2.setString(9,"a nl");
      pstmt2.setString(10,O$array.get(nse).Get_Element("DATATYPE"));
      } else {
      pstmt2.setString(8,null);
      pstmt2.setString(9,null);
      pstmt2.setString(10,null);
      }
      pstmt2.execute();
      if (pstmt2.getUpdateCount()==0){
      pstmt.setString(12,String.valueOf(seq));
      pstmt.setString(1,ck);
      pstmt.setString(4,Cols.Get_Element(ck));
      pstmt.setString(5,"LOVQ:AUTO$");
      pstmt.setString(6,Cols.Get_Element(ck));
     // pstmt.setString(9,"\""+ck+"\" \"GC"+String.valueOf(seq)+"\"");
      pstmt.setString(7,"\""+ck+"\" \"GC1\"");
      if (nse>=0){
      pstmt.setString(8,"\""+O$array.get(nse).Get_Element("CODE")+"\"");
      pstmt.setString(9,"a nl");
      pstmt.setString(10,O$array.get(nse).Get_Element("DATATYPE"));
      } else {
      pstmt.setString(8,null);
      pstmt.setString(9,null);
      pstmt.setString(10,null);
      }
      pstmt.execute();
      }
     }
   }
   pstmt.close();
   // диаграммы
   stmt="INSERT INTO lui_t_data_group_chart_pref"
           + "(grouping_id,seq,        chart_type,call_mode,select_clause,where_clause,\n" +
"                where_clause_0_null,where_clause_1_null,where_clause_01_null,map_form_code,show_as_list)\n" +
"VALUES (      ?::lui_d_id,?::lui_d_id,?,             ?,        ?,            ?,\n" +
//             1           2           3              4         5             6 
"                ?,                  ?,                  ?,                   ?,            ?)";
//               7                   8                   9                   10             11
   stmt=stmt+" on conflict(grouping_id,seq) do update set "
           + " chart_type=EXCLUDED.chart_type "
           + ",call_mode=EXCLUDED.call_mode "
           + ",select_clause=EXCLUDED.select_clause "
           + ",where_clause=EXCLUDED.where_clause "
           +",where_clause_0_null=EXCLUDED.where_clause_0_null "
           +",where_clause_1_null=EXCLUDED.where_clause_1_null "
           +",where_clause_01_null=EXCLUDED.where_clause_01_null "
           +",map_form_code=EXCLUDED.map_form_code "
           //+ ",show_as_list=EXCLUDED.show_as_list"
           ;
           
   pstmt=con.prepareStatement(stmt);
   pstmt.setString(1,DgId);
   seq=0;
   ArrayList<String> AxCol=new ArrayList<>();
   for (String ck:Cols.keySet()){ // крутим столбцы
    if (Integer.valueOf(ck.replace("C",""))<=0){
    AxCol.add(ck);
    pstmt.setString(2,String.valueOf(++seq));   
    
    pstmt.setString(3,"BAR");
    pstmt.setString(10,null);
    
      String kGC=String.valueOf(Integer.valueOf(ck.replace("C",""))+GCShift);
      if (GCChartTypes.containsKey(kGC)){
      String ct=GCChartTypes.get(kGC);  
      if (like(ct,"MAP,%")){
          pstmt.setString(10,ct.split(",",2)[1]);
          ct=ct.split(",",2)[0];
      }
      pstmt.setString(3,ct);
    }
    if (seq==1)pstmt.setString(4,"MODAL");
    else pstmt.setString(4,"AUTO-R");
   // pstmt.setString(5,"select \""+ck+"\" \"GC"+String.valueOf(seq)+"\"");
    pstmt.setString(5,"select \""+ck+"\" \"GC1\"");
    pstmt.setString(6,"\""+ck+"\"=':p_dim0'");
    pstmt.setString(7,"\""+ck+"\" is null");
    pstmt.setString(8,"\""+ck+"\"=':p_dim0'");
    pstmt.setString(9,"\""+ck+"\" is null");
    pstmt.setString(11,null);
    pstmt.execute();
     }
   }
   pstmt.close();
   // выравнивание 3D/2D
   int AxColpos=0;
   //                                               1           2               3                    4                     5                    
   stmt="update lui_t_data_group_chart_pref set select_clause=?,where_clause=?,where_clause_0_null=?,where_clause_1_null=?,where_clause_01_null=?"
           + " where grouping_id=?::lui_d_id and seq=?::lui_d_id";
   pstmt=con.prepareStatement(stmt);
   pstmt.setString(6,DgId);
   ArrayList<LuiCollection> charts=lc.getSelectAllRowsValMap("select lui_f_dg_get_chart_type(grouping_id,seq) \"chart_type\" from lui_t_data_group_chart_pref where grouping_id="+DgId+" order by seq");
   for (int i=0;i<charts.size();i++){
      
       if (AxColpos>=AxCol.size()){
            //lc.executeDML("delete from lui_t_data_group_chart_pref where grouping_id="+DgId+" and seq="+String.valueOf(AxColpos+1));
            lc.executeDML("delete from lui_t_data_group_column where grouping_id="+DgId+" and seq="+String.valueOf(AxColpos+1));
            continue;
       }
       LuiCollection col=charts.get(i);
       String ct=col.Get_Element("chart_type");
       
       int localAxes=1;
       if  (like(ct,"%3D")){localAxes=2;
       if ((AxColpos+localAxes)>AxCol.size()) {
        //ct=ct.replace("3D","");
        localAxes=1;
       }
       }
       String selectClause="";
       String whereClause="";
       String whereClause0n="";
       String whereClause1n="";
       String whereClause01n="";
       int dim=0;
       for (int j=AxColpos;j<AxColpos+localAxes;j++){
           if (j<AxCol.size()){
           if (is_not_null(selectClause))selectClause=selectClause+" ,";
           if (is_not_null(whereClause))whereClause=whereClause+" and ";
           selectClause=selectClause+"\""+AxCol.get(j)+"\" \"GC"+String.valueOf(dim+1)+"\"";
           whereClause=whereClause+"\""+AxCol.get(j)+"\"=':p_dim"+String.valueOf(dim)+"'";
                   }
       dim++;
       }
       whereClause0n=whereClause.replace("=':p_dim0'"," is null");
       whereClause1n=whereClause.replace("=':p_dim1'"," is null");
       whereClause01n=whereClause0n.replace("=':p_dim1'"," is null");
       if (is_not_null(selectClause)) {
       
       pstmt.setString(1,"select "+selectClause);
       pstmt.setString(2,whereClause);
       pstmt.setString(3,whereClause0n);
       pstmt.setString(4,whereClause1n);
       pstmt.setString(5,whereClause01n);
       pstmt.setString(7, String.valueOf(i+1));
       pstmt.execute();
       }
       AxColpos=AxColpos+localAxes;
   }
   lc.executeDML("delete from lui_t_data_group_function where grouping_id="+DgId+" and seq>"+String.valueOf(sfCount));
   }
  catch (SQLException sex){
      throw sex;
  }
  finally {
       if (lc!=null){
           if (pstmt!=null)pstmt.close();
           if (pstmt2!=null)pstmt2.close();
           application.metaConnPool.SetFree(lc);
       }
   }
   
return DgId;

}
public boolean MatrixIsCellInDomains(int r,int c,String domains){
    domains=domains.replace("},{"," ").replace("}{"," ").replace("{","").replace("}","");
    String[] Domains=domains.split(" ");
    if (Domains==null)return false;
    if (Domains.length==0)return false;
    for(int i=0;i<Domains.length;i++){
        String domain=Domains[i];
        String r1c1=null;
        String r2c2=null;
        if (domain.indexOf(":")>0){
            r1c1=domain.split(":")[0];
            r2c2=domain.split(":")[1];
        } else {
            r1c1=domain;
            r2c2=domain;
        }
        if (is_null(r1c1)||is_null(r2c2))return false;
        if (r1c1.indexOf(",")<0 ||r2c2.indexOf(",")<0)return false;
        int r1=Integer.valueOf(r1c1.split(",")[0]);
        int c1=Integer.valueOf(r1c1.split(",")[1]);
        int r2=Integer.valueOf(r2c2.split(",")[0]);
        int c2=Integer.valueOf(r2c2.split(",")[1]);
        if (r>=r1 && r<=r2 && c>=c1 && c<=c2)return true;
    }
    return false;
}
/**
 * копирование матрицы в таблицу для построения графиков
     * @param pRdepth - глубина вложенности по строкам
     * @param pCdepth - глубина вложенности по столбцам
     * @param pLc - cокдинение
     * @return - коллекция кодов  строк и столбцов c рекомендуемой функцией группировки
 * @throws LuiException
 * @throws SQLException 
 */

public LuiCollection MatrixCopy2TempData(String pShowMode,int pRdepth,int pCdepth, LuiDbConn pLc,String sc) throws LuiException, SQLException{
 /*  if (!getProperty("SelectedCells").equals("[]")){
       if (pShowMode.equals("AS"))pShowMode="S";
       if (pShowMode.equals("AI"))pShowMode="I";
   } else {
       if (pShowMode.charAt(0)=='A')pShowMode="A";
   }
   */
   if (is_not_null(sc)){
       if (pShowMode.equals("AS"))pShowMode="S";
       if (pShowMode.equals("AI"))pShowMode="I";
       if (pShowMode.equals("F"))pShowMode="S";
       if (pShowMode.equals("FI"))pShowMode="I";
   } else {
       if (pShowMode.charAt(0)=='A')pShowMode="A";
       if (pShowMode.charAt(0)=='F')pShowMode="A";
   }
   LuiDbConn lc=null;
   LuiCollection GroupFunctions=new LuiCollection("");
   String stmt="insert into lui_t_tem_data(form_id,session_id,r,c,value,hint)values(?,?::lui_d_id,?::lui_d_id,?::lui_d_id,?,?)";
   //LuiCollection tvc;
   PreparedStatement pstmt=null;
   Connection c=null;
   String v;
   String pc;
   String colcode;
   String rowcode;
   Integer cursor=null;
   try{
   if (pLc==null){
   lc=application.metaConnPool.GetFree(application.NLS.currentLang.locale,application.NLS.currentLang.code,application.code,application.getLuiSessionID(),application.getAppUser());
   } else {
       lc=pLc;
   }
   lc.executeDML("delete from lui_t_tem_data where session_id="+application.luiSession.id+" and form_id='"+getFormID()+"'");
   c=lc.getConnection();
   pstmt=c.prepareStatement(stmt);
   
   pstmt.setString(1,getFormID());
   pstmt.setString(2,application.luiSession.id);
   int rows=Integer.valueOf(getProperty("LastROW"));
   int cols=Integer.valueOf(getProperty("LastCOL"));
   boolean isRowMember=false;
   for (int rn=0;rn<=rows;rn++){
       isRowMember=false;
       for (int cn=0;cn<=cols;cn++){
           
           if (rn==0 && cn==0)continue;
           if (rn==0){
               for (String colCode:elements.keySet()){
                 if (!elements.get(colCode).getType().equals("COL"))continue;
                 if (!elements.get(colCode).getProperty("ltn").equals(String.valueOf(cn)))continue;
                 if (elements.get(colCode).getProperty("VISIBLE").equals("N"))continue;
                 addProperty("CurrentTuple",colCode);
                 //tvc=new LuiCollection("");
                 v=elements.get(colCode).getProperty("TUPLE_SIGN");
                 if (is_null(v))v=elements.get(colCode).getProperty("CAPTION");
                 v=application.clearWiki(v).replace("\n"," ");
                 pstmt.setString(3, String.valueOf(rn)); 
                 pstmt.setString(4, String.valueOf(cn));
                 pstmt.setString(5, v);
                 v=elements.get(colCode).getProperty("HINT");
                 pstmt.setString(6, v);
                 pstmt.execute();
                 pc=colCode;
                 for (int pi=0;pi<pCdepth;pi++){
                   pc=elements.get(pc).getProperty("Parent");  
                   if (is_null(pc) || pc.equals("Form")) break;
                   if (!elements.containsKey(pc))break;
                   if (elements.get(pc).getType().equals("GROUP"))break;
                   addProperty("CurrentTuple",pc);
                   v=elements.get(pc).getProperty("TUPLE_SIGN");
                   if (is_null(v))v=elements.get(pc).getProperty("CAPTION");
                   v=application.clearWiki(v).replace("\n"," ");
                   pstmt.setString(3, String.valueOf(rn-(1+pi)));
                   pstmt.setString(4, String.valueOf(cn));  
                   pstmt.setString(5, v);
                    v=elements.get(pc).getProperty("HINT");
                   pstmt.setString(6, v);
                   pstmt.execute();
                 }
                 break;
               }
           }
           if (cn==0){
               for (String rowCode:elements.keySet()){
                 if (!elements.get(rowCode).getType().equals("ROW"))continue;
                 if (!elements.get(rowCode).getProperty("ltn").equals(String.valueOf(rn)))continue;
                 if (elements.get(rowCode).getProperty("VISIBLE").equals("N"))continue;
                 addProperty("CurrentTuple",rowCode);
                 v=elements.get(rowCode).getProperty("TUPLE_SIGN");
                 if (is_null(v))v=elements.get(rowCode).getProperty("CAPTION");
                 if (is_not_null(v)) v=application.clearWiki(v).replace("\n"," ");
                 pstmt.setString(3, String.valueOf(rn));  
                 pstmt.setString(4, String.valueOf(cn));
                 pstmt.setString(5, v);
                  v=elements.get(rowCode).getProperty("HINT");
                 pstmt.setString(6, v);
                 pstmt.execute();
                 pc=rowCode;
                 for (int pi=0;pi<pRdepth;pi++){
                   pc=elements.get(rowCode).getProperty("Parent");  
                   if (is_null(pc) || pc.equals("Form")) break;
                   if (!elements.containsKey(pc))break;
                   if (elements.get(pc).getType().equals("GROUP"))break;
                   addProperty("CurrentTuple",pc);
                   v=elements.get(pc).getProperty("TUPLE_SIGN");
                   if (is_null(v))v=elements.get(pc).getProperty("CAPTION");
                   v=application.clearWiki(v).replace("\n"," ");
                   pstmt.setString(3, String.valueOf(rn));
                   pstmt.setString(4, String.valueOf(cn-(1+pi)));  
                   pstmt.setString(5, v);
                    v=elements.get(pc).getProperty("HINT");
                   pstmt.setString(6, v);
                   pstmt.execute();
                 }
                 break;
               }
           }
           if (cn>0 && rn>0){
               colcode=MatrixGetCellProperty(String.valueOf(rn)+":"+String.valueOf(cn),"COL");
               if (!elements.containsKey(colcode))continue;
               if (elements.get(colcode).getProperty("VISIBLE").equals("N"))continue;
               rowcode=MatrixGetCellProperty(String.valueOf(rn)+":"+String.valueOf(cn),"ROW");
               if (!elements.containsKey(rowcode))continue;
               if (elements.get(rowcode).getProperty("VISIBLE").equals("N"))continue;
               if (!pShowMode.equals("A")){ // если надо выводить не всё
               if (is_not_null(sc)){ //  и есть отмеченные
                  boolean b=MatrixIsCellInDomains(rn,cn,sc);
                  if (pShowMode.equals("S")) {
                     if (!b) continue;
                  }
                  if(pShowMode.equals("I")){
                     if (b) continue; 
                  }
                 
               } else { // нет отмеченных
                  if (pShowMode.equals("S")) continue; // уходим если надо выводить только отмеченные
               }
               }    
               v=this.MatrixGetCellProperty(String.valueOf(rn)+":"+String.valueOf(cn),"CELL_SHOW");
               if (is_null(v))continue;
               v=v.replace(" ","");
               v=v.replace(String.valueOf((char)160),"");
               v=v.replace(",",".");
               Double dv=null;
               pstmt.setString(3, String.valueOf(rn));  
               pstmt.setString(4, String.valueOf(cn));
               try{
               dv=Double.valueOf(v);
               if (dv!=Math.ceil(dv)){
               
                   GroupFunctions.Add_Element("C"+String.valueOf(cn),"AVG");
                   GroupFunctions.Add_Element("R"+String.valueOf(rn),"AVG");
               }
               pstmt.setString(5, dv.toString());
               pstmt.setString(6, null);
               pstmt.execute();
               isRowMember=true;
               } catch (Exception ex){
              //  pstmt.setString(5, null);
               }      
               
               
           }
          
       }
       if (!isRowMember && rn>0){
            lc.executeDML("delete from lui_t_tem_data where  session_id="+application.luiSession.id+" and form_id='"+getFormID()+"' and r="+String.valueOf(rn));
        } 
       
   }
   for (int cn=1;cn<=cols;cn++){
       String s=lc.getSelectResult("select 1 from lui_t_tem_data where value is not null and r>0 and session_id="+application.luiSession.id+" and form_id='"+getFormID()+"' and c="+String.valueOf(cn)+" limit 1");
           if (is_null(s)){
            lc.executeDML("delete from lui_t_tem_data where session_id="+application.luiSession.id+" and form_id='"+getFormID()+"' and c="+String.valueOf(cn));               
           }
       }
   // среди оставшихся строк выполним заполнение доп. информации
   cursor=lc.CursorOpen("select distinct r from lui_t_tem_data where r>0 and session_id="+application.luiSession.id+" and form_id='"+getFormID()+"'");
   ArrayList<String> arow;
   pstmt.close();
   pstmt=c.prepareStatement("update lui_t_tem_data set ri=? where session_id=?::lui_d_id and form_id=? and r=?::lui_d_id");
   pstmt.setString(2,application.luiSession.id );
   pstmt.setString(3,getFormID());
   String RI2NUM="::numeric";
   rows=0;
   while (lc.CursorNextRow(cursor)){
       rows++;
       arow=lc.CursorGetRow(cursor);
       String r=arow.get(0);
        // узнаём код элемента ROW 
                 String ROW="";
                 for (String rowCode:elements.keySet()){
                 if (!elements.get(rowCode).getType().equals("ROW"))continue;
                 if (!elements.get(rowCode).getProperty("ltn").equals(r))continue;
                 ROW=rowCode;break;
                 }
       String ri=r;
       if (like(getProperty("ROW_INFO"),"Y%")){
          if (is_not_null(elements.get(ROW).getProperty("TUPLE_INFO"))){
              ri=elements.get(ROW).getProperty("TUPLE_INFO");
          } 
       try {
           Double.valueOf(ri);
       } catch(Exception ex){RI2NUM="";}
       }
       pstmt.setString(1, ri);
       pstmt.setString(4, r);
       pstmt.execute();
   }
   lc.CursorClose(cursor);cursor=null;
   
   // среди оставшихся столбцов выполним заполнение доп. информации
   cursor=lc.CursorOpen("select distinct c from lui_t_tem_data where c>0 and session_id="+application.luiSession.id+" and form_id='"+getFormID()+"'");
   ArrayList<String> acol;
   pstmt.close();
   pstmt=c.prepareStatement("update lui_t_tem_data set ci=? where session_id=?::lui_d_id and form_id=? and c=?::lui_d_id");
   pstmt.setString(2,application.luiSession.id );
   pstmt.setString(3,getFormID());
   String CI2NUM="::numeric";
   cols=0;
   while (lc.CursorNextRow(cursor)){
       cols++;
       acol=lc.CursorGetRow(cursor);
       String co=acol.get(0);
        // узнаём код элемента ROW 
                 String COL="";
                 for (String rowCode:elements.keySet()){
                 if (!elements.get(rowCode).getType().equals("COL"))continue;
                 if (!elements.get(rowCode).getProperty("ltn").equals(co))continue;
                 COL=rowCode;break;
                 }
       String ci=co;
       if (like(getProperty("COL_INFO"),"Y%")){
          if (is_not_null(elements.get(COL).getProperty("TUPLE_INFO"))){
              ci=elements.get(COL).getProperty("TUPLE_INFO");
          } 
       }
       try {
           Double.valueOf(ci);
       } catch(Exception ex){CI2NUM="";}
       pstmt.setString(1, ci);
       pstmt.setString(4, co);
       pstmt.execute();
   }
   lc.CursorClose(cursor);
   cursor=null;
   GroupFunctions.Add_Element("RI2NUM",RI2NUM);
   GroupFunctions.Add_Element("CI2NUM",CI2NUM);
   GroupFunctions.Add_Element("ROWS",String.valueOf(rows));
   GroupFunctions.Add_Element("COLS",String.valueOf(cols));
   // исключение дублирования в именах столбцов
   lc.executeDML("update lui_t_tem_data set value=value||' ('||COALESCE(NULLIF(ci,''),c::text)||')' "
           + "where session_id="+application.luiSession.id+" and form_id='"+getFormID()+"' "
                   + "and r=0 and value in "
                   + "(select td.value from lui_t_tem_data td where td.session_id="+application.luiSession.id+" and td.form_id='"+getFormID()+"' "
                           + " and td.r=0 group by td.value having count(*)>1)");
   // исключение дублирования в именах строк
   lc.executeDML("update lui_t_tem_data set value=value||' ('||COALESCE(NULLIF(ri,''),r::text)||')' "
           + "where session_id="+application.luiSession.id+" and form_id='"+getFormID()+"' "
                   + "and c=0 and value in "
                   + "(select td.value from lui_t_tem_data td where td.session_id="+application.luiSession.id+" and td.form_id='"+getFormID()+"' "
                           + " and td.c=0 group by td.value having count(*)>1)");
   
   } finally {
       if (lc!=null){
           if (cursor!=null)lc.CursorClose(cursor);
           if (pstmt!=null)pstmt.close();
           if (pLc==null)
           application.metaConnPool.SetFree(lc);
       }
   }
   return GroupFunctions;
}
// свёртывание уровня списка группировки
public void dgLcollapse() throws LuiException, SQLException{
    
    //-- осматриваем индикаторы группировки
    String GI;
    String GC;
    LuiCollection col=new LuiCollection("");
    for (int i=1;i<1000;i++){
        GI="GI"+String.valueOf(i);
        GC="GC"+String.valueOf(i);
        if (!elements.containsKey(GI))break;
        if (!elements.containsKey(GC))break;
        if (elements.get(GI).getProperty("Value").equals("1")) {
            col.Add_Element(GI,"0");break;
        }
        col.Add_Element(GC,elements.get(GC).getProperty("Value"));    
    }
    this.collapsedDGLevels.add(col);
}
// развёртывание уровня
public void dgLexpand() throws LuiException, SQLException{
    
    //-- осматриваем индикаторы группировки
    String GI;
    String GC;
    LuiCollection col=new LuiCollection("");
    for (int i=1;i<1000;i++){
        GI="GI"+String.valueOf(i);
        GC="GC"+String.valueOf(i);
        if (!elements.containsKey(GI))break;
        if (!elements.containsKey(GC))break;
        if (elements.get(GI).getProperty("Value").equals("1")) {col.Add_Element(GI,"0");break;}
        col.Add_Element(GC,elements.get(GC).getProperty("Value"));    
    }
    this.collapsedDGLevels.remove(col);
}
// свёртывание уровня
public String dgIsCollapse() throws LuiException, SQLException{
    if (this.collapsedDGLevels.size()==0) return "N";
    //-- осматриваем индикаторы группировки
    String GI;
    String GC;
    LuiCollection col=new LuiCollection("");
    for (int i=1;i<1000;i++){
        GI="GI"+String.valueOf(i);
        GC="GC"+String.valueOf(i);
        if (!elements.containsKey(GI))break;
        if (!elements.containsKey(GC))break;
        if (elements.get(GI).getProperty("Value").equals("1")) {col.Add_Element(GI,"0");break;}
        col.Add_Element(GC,elements.get(GC).getProperty("Value"));    
    }
   if  (this.collapsedDGLevels.contains(col)) return "Y";
   return "N";
}
public String getFormTtle()throws LuiException, SQLException{
    if (getVal("MasterForm").equals("Y"))
    return Jsoup.parse(getProperty("HEADER")).text();
    if (NVL(application.getProperty("SHOW_SUBWTITLE_IFEQ_WTITLE"),"Y").equals("Y"))
    return Jsoup.parse(getProperty("HEADER")).text();
    String wid=getVal("WindowID");
    for (String fc:application.forms.keySet()){
        if (!application.forms.get(fc).getVal("WindowID").equals(wid)) continue;
        if (!application.forms.get(fc).getVal("MasterForm").equals("Y"))continue;
        if (getProperty("HEADER").equals(application.forms.get(fc).getProperty("HEADER")))return "";
        else return Jsoup.parse(getProperty("HEADER")).text();
    }
    return Jsoup.parse(getProperty("HEADER")).text();
}
}
