/*
 * Copyright (c) 2021 «FORS Telecom» LLC.
 * All Rights Reserved.
 *
 * The code is published under the terms of the «FORS Telecom» LLC Licens,
 * published in file LICENSE of this package or at
 *
 * http://lui.fors.ru/license	- in Russian
 * http://lui.fors.com/license	- in English.
 *
 */

package ru.fors.telecom.webui.core;

import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.io.Writer;
import java.net.Socket;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.Array;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;
import java.util.Random;
import java.util.StringJoiner;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.lang3.StringUtils;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.jdom2.output.Format;
import org.jdom2.output.Format.TextMode;
import org.jdom2.output.XMLOutputter;
import ru.fors.fastcom.container.client.fastcom.FastcomConnectionFabric;
import ru.fors.fastcom.container.client.fastcom.FastcomFileStorage;
import ru.fors.fastcom.container.client.fastcom.FastcomFileStorage.FileStat;
import ru.fors.fastcom.container.client.tcp.MultipleExceptions;
import ru.fors.fastcom.container.client.tcp.ProtocolError;
import ru.fors.fastcom.container.client.tcp.RemoteError;
import ru.fors.telecom.webui.common.Defaults;
import ru.fors.telecom.webui.common.Defaults.FormPosition;
import ru.fors.telecom.webui.common.Defaults.FormType;
import ru.fors.telecom.webui.common.Defaults.Instructions;
import ru.fors.telecom.webui.common.Defaults.NodeState;
import ru.fors.telecom.webui.common.Defaults.NodeType;
import ru.fors.telecom.webui.common.Defaults.ButtonState;
import ru.fors.telecom.webui.common.Defaults.ColumnAlign;
import ru.fors.telecom.webui.common.Defaults.DateDetail;
import ru.fors.telecom.webui.common.Defaults.FieldType;
import ru.fors.telecom.webui.common.Defaults.GroupState;
import ru.fors.telecom.webui.common.Defaults.GroupType;
import ru.fors.telecom.webui.common.Defaults.InfoTarget;
import ru.fors.telecom.webui.common.Defaults.SelectionMode;
import ru.fors.telecom.webui.common.Defaults.WindowPosition;
import ru.fors.telecom.webui.common.UiCollection;
import ru.fors.telecom.webui.common.Utils;
import ru.fors.telecom.webui.error.UiException;
import ru.fors.telecom.webui.event.CollectionParameter;
import ru.fors.telecom.webui.event.SingleEvent;
import ru.fors.telecom.webui.event.SingleEventParameters;
import ru.fors.telecom.webui.event.StringParameter;
import ru.fors.telecom.webui.event.TotalEvent;
import ru.fors.telecom.webui.iface.UiFormContent;
import ru.fors.telecom.webui.iface.UiHasHeader;
import ru.fors.telecom.webui.menu.UiLocalMenuItem;
import ru.fors.telecom.webui.menu.UiSubMenu;
import ru.fors.telecom.webui.common.Defaults.LOVRowType;
import ru.fors.telecom.webui.common.Defaults.MenuItemRole;
import ru.fors.telecom.webui.common.Defaults.MenuItemType;
import ru.fors.telecom.webui.common.Defaults.SortDirection;
import ru.fors.telecom.webui.common.Defaults.Sounds;
import ru.fors.telecom.webui.common.Defaults.TimeDetail;
import ru.fors.telecom.webui.common.Defaults.URLTarget;
import ru.fors.telecom.webui.common.PollResponse;
import ru.fors.telecom.webui.core.WebUI.Dispatcher;
import ru.fors.telecom.webui.error.UiError;
import ru.fors.telecom.webui.iface.UiFormCurrent;
import ru.fors.telecom.webui.menu.UiAddonMenuItem;
import ru.fors.telecom.webui.menu.UiGroupMenuItem;
import ru.fors.telecom.webui.menu.UiLocalMenuSeparator;
import ru.fors.telecom.webui.menu.UiMenuObject;
import java.util.logging.Logger;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.FileUtils;
import static org.apache.commons.lang3.StringUtils.substringAfter;
import static org.apache.commons.lang3.StringUtils.substringBefore;
import ru.fors.telecom.webui.common.Defaults.CaptionOrientation;
import ru.fors.telecom.webui.common.Defaults.ChartOrientation;
import ru.fors.telecom.webui.common.Defaults.FieldMode;
import ru.fors.telecom.webui.common.Defaults.Overflow;
import ru.fors.telecom.webui.common.Defaults.HorizontalAlign;
import ru.fors.telecom.webui.common.Defaults.MatrixColumnInfo;
import ru.fors.telecom.webui.common.Defaults.MatrixRowInfo;
import ru.fors.telecom.webui.common.Defaults.MatrixSelectionMode;
import ru.fors.telecom.webui.common.Defaults.VerticalAlign;
import ru.fors.telecom.webui.common.HtmlProperties;
import ru.fors.telecom.webui.common.MatrixCellRecord;



/**
 *
 * @author john
 */



/*******************************************************************************
*
* Рабочая область, связанная с подключённым пользователем
*
********************************************************************************/
public class WorkArea {

    private boolean mobile;
    private String uid = "";
    private String token;
    private Connection connection;
    private CallableStatement eventStatement;
    private String param;
    private String param2;
    private XMLOutputter prettyOutputter;
    private XMLOutputter rawOutputter;
    private UiRootUnit uiRoot; 
    private ArrayList<HashMap<String, String>> globalParams;
    private HashMap<String, String> localParams;
    private HashMap<String, String> effectiveParams;
    private ArrayList<JsonObject> instructions;
    private SAXBuilder xmlBuilder;
    private String commandTag;
    private int commandLevel;
    private String commandToProcess;
    private UiWindow contextWindow = null;
    private UiForm contextForm = null;
    private UiUnit contextFormContent = null;
    private UiTreeNode contextNode = null;
    private UiToolBar contextToolBar = null;
    private UiBaseButton contextButton = null;
    private UiBasement contextBasement = null;
    private UiGridColumn contextColumn = null;
    private UiGridRow contextRow = null;
    private UiGridCell contextCell = null;
    private UiGrid contextGrid = null;
    private UiChart contextChart = null;
    private UiHtml contextHtml = null;
    private UiGroup contextGroup = null;
    private UiField contextField = null;
    private UiMatrix contextMatrix = null;
    private UiMatrixColumn contextMatrixColumn = null;
    private UiMatrixRow contextMatrixRow = null;
    private UiMatrixCell contextMatrixCell = null;
    private ArrayList<UiSubMenu> localMenu = null;
    private HashMap<String, UiLocalMenuItem> localMenuItems = null;
    private boolean showMenu = false;
    private StringBuilder menuHTML = null;
    private HashSet<UiUnit> unitsToCheck;
    private TotalEvent totalEvent;
    private HashMap<String, UiUnit> units;
    private UiQuestion question;
    private UiLov lov = null;
    //private UiProgressWindow progressWindow = null;
    private UiCalendar calendar = null;
    private UiHeader contextHeader = null;
    private UiButtonBar contextButtonBar = null;
    private UiConfigs contextConfigs = null;
    private final static FastcomConnectionFabric ST_AGENT_FABRIC = new FastcomConnectionFabric();
    private int aliveCounter = 0;
    private Timer monitor;
    private Dispatcher dispatcher;
    private MonitorTask monitorTask;
    private boolean stopMonitoring = false;
    private Class clForm, clNode, clField, clGroup, clButton, clToolButton, clLovRow, clTabs, clMatrixRow, clMatrixColumn, clMatrixCell;
    private Class clButtonBar, clToolBar, clHeader, clFormContent, clBasement, clConfigs, clFieldButton, clMatrixButton, clWindow;
    private UiFileUpload fileUpload = null;
    private UiFileDownload fileDownload = null;
    private SharedArea sharedArea;
    private String userName;
    private int protocol;
    private boolean customMenu = false;
    private BufferedWriter sockWriter;
    private BufferedReader sockReader;
    private Socket socket;
    private String remoteHost;
    private MessageHandler msgHandler = null;
    private boolean running = false;
    private String msgPort = "";
    private PollResponse defaultPollResponse;
    private static Logger LOGGER = Logger.getLogger(WorkArea.class.getName());
    private String tempDir = "";
    private Path eventsLogFile;
    private Writer eventsLogWriter;
    private boolean eventsLogging = false;
    private boolean testing = false;
    private ArrayList<String> eventsLog;
    private String formToFocus = null;
    private String unitToFocus = null;
    

    public WorkArea(String user, Connection conn, Dispatcher disp, SharedArea shar, String ip, boolean mob, boolean fst) throws UiException {
        
        try {
            protocol = 1;
            userName = user.toUpperCase();
            connection = conn;
            mobile = mob;
            dispatcher = disp;
            sharedArea = shar;
            init();
            if (fst) {
                try (Statement stmtRole = connection.createStatement()) {
                    stmtRole.execute("set role fastcom_main identified by abrakadabra");
                    try (
                        CallableStatement stmtAddress = connection.prepareCall("begin ui_p_common.set_client_ip(?); end;")) {
                        stmtAddress.setString(1, ip);
                        stmtAddress.execute();
                    }
                }
            }
        } catch (Exception ex) {
            ex.printStackTrace();
            try {
                connection.close();
            } catch (SQLException e) {}
            UiError.raise("Failed to create application data");
        }
        
    }
    
    public WorkArea (Socket socket, Dispatcher disp, String mport, String host, boolean mob) throws UiException {
        
        try {
            protocol = 2;
            mobile = mob;
            this.socket = socket;
            dispatcher = disp;
            msgPort = mport;
            remoteHost = host;
            init();
            //sockWriter = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), "Windows-1251"));
            //sockReader = new BufferedReader(new InputStreamReader(socket.getInputStream(), "Windows-1251"));
            sockWriter = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream(), "UTF8"));
            sockReader = new BufferedReader(new InputStreamReader(socket.getInputStream(), "UTF8"));
        } catch (Exception ex) {
            ex.printStackTrace();
            try {
                socket.close();
            } catch (IOException e) {}
            UiError.raise("Failed to create application data");
        }
        
    }
    
    /**
     * Конструктор через функцию-геренатор локальных сеансов, для работы в Bundle
     * @param pGenerator функция-генератор сеансов
     * @param pDispatcher дитспетчер
     * @param host какой-то хост
     * @param mob какой-то моб
     * @throws UiException что-то пошло не так
     */
    public WorkArea(Function<?extends Object, ?extends Object> pGenerator, Dispatcher pDispatcher, String host, boolean mob) throws UiException  {
        dispatcher = pDispatcher;
        try {
            protocol = 2;
            mobile = mob;
            this.socket = null;
            remoteHost = host;
            
            sockWriter = null;
            sockReader = null;
            
            Object aAnswer = null;
            synchronized(pGenerator) {
                aAnswer = pGenerator.apply(null);
            }
            
            if(aAnswer == null || ! (aAnswer instanceof Object[])) {
                UiError.raise("Corrupted local session generator");
                return;
            }
            
            Object[] aArray = (Object[]) aAnswer;
            for(Object aElement: aArray) {
                if(aElement instanceof BufferedReader) { sockReader = (BufferedReader) aElement; }
                else if(aElement instanceof BufferedWriter) { sockWriter = (BufferedWriter) aElement; }
            }
            
            if(sockWriter == null || sockReader == null) {
                UiError.raise("Unexpected answer from session generator");
                return;
            }
            
            init();
            
        } catch (Exception ex) {
            if(sockWriter != null) try { sockWriter.close(); } catch(Exception ignore) {}
            if(sockReader != null) try { sockReader.close(); } catch(Exception ignore) {}
            UiError.raise("Failed to create application data");
        }
    }
    
    private void init() throws ClassNotFoundException {
        
        instructions = new ArrayList<>();
        token = UUID.randomUUID().toString();
        
        prettyOutputter = new XMLOutputter();
        prettyOutputter.setFormat(Format.getPrettyFormat());
        rawOutputter = new XMLOutputter();
        rawOutputter.setFormat(Format.getRawFormat().setLineSeparator("").setTextMode(TextMode.PRESERVE));
        
        commandToProcess = null;
        commandLevel = -1;
        commandTag = null;
        units = new HashMap<>();
        xmlBuilder = new SAXBuilder();
        localParams = new HashMap<>();
        effectiveParams = new HashMap<>();
        globalParams = new ArrayList<>();
        unitsToCheck = new HashSet<>();
        
        uiRoot = new UiRootUnit("/");
        units.put(uiRoot.getId(), uiRoot);
        
        monitorTask = new MonitorTask();
        monitor = new Timer();
        monitor.schedule(monitorTask, 5000, 30000);
        
        clForm = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiForm");
        clNode = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiTreeNode");
        clField = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiField");
        clGroup = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiGroup");
        clButton = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiButton");
        clToolButton = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiToolButton");
        clLovRow = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiLov$UiLovRow");
        clButtonBar = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiButtonBar");
        clHeader = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiHeader");
        clToolBar = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiToolBar");
        clFormContent = Class.forName("ru.fors.telecom.webui.iface.UiFormContent");
        clBasement = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiBasement");
        clConfigs = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiConfigs");
        clTabs = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiTabs");
        clFieldButton = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiFieldButton");
        clMatrixButton = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiMatrixButton");
        clWindow = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiWindow");
        clMatrixRow = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiMatrixRow");
        clMatrixColumn = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiMatrixColumn");
        clMatrixCell = Class.forName("ru.fors.telecom.webui.core.WorkArea$UiMatrixCell");
        
        defaultPollResponse = new PollResponse();
        defaultPollResponse.setRetry(Defaults.NORMAL_POLL_INTERVAL);
        
        instructions.add(Utils.instruction(Instructions.root, uiRoot.getId()));
        instructions.add(Utils.instruction(Instructions.init, token));
        
    }
    
    public ArrayList<JsonObject> getInstructions() {
        return instructions;
    }
    
    public final String getToken() {
        return token;
    }
    
    public void close() {
        monitor.cancel();
        monitor.purge();
        if (protocol == 1) {
            try {
                connection.rollback();
            } catch (SQLException ex) {}
            try {
                connection.close();
            } catch (SQLException ex) {}
        } else if (protocol == 2 || protocol == 3) {
            try {
                msgHandler.finish();
            } catch (Exception ex) {}
                //++BUNDLE
            try {
                sockWriter.close();
            } catch (Exception ex) {}
            try {
                sockReader.close();
            } catch (Exception ex) {}
            try {
                socket.close();
            } catch (Exception ex) {}
                // --BUNDLE
        }
        dispatcher.rm(token);
    }
    
    public synchronized void processCommands(String rawCommands) throws JDOMException, IOException {
        Document XMLCommands = xmlBuilder.build(new StringReader(rawCommands));
        Utils.reorderXML(XMLCommands.getRootElement());
        //System.out.println(prettyOutputter.outputString(XMLCommands));
        if (!mobile) instructions.add(Utils.instruction(Instructions.lc, prettyOutputter.outputString(XMLCommands)));
        List<Element> commandsList = XMLCommands.getRootElement().getChildren("W");
        if (commandsList.isEmpty()) {
            buildInstructions(XMLCommands.getRootElement());
        } else {
            Iterator<Element> iter = commandsList.iterator();
            while (iter.hasNext()) {
                buildInstructions(iter.next());
            }
        }
    }
    
    private void buildInstructions(Element elem) {
        globalParams.clear();
        localParams.clear();
        commandTag = "";
        processXML(elem, 0);
        processCommand();
    }
        
    private void processXML(Element e, int level) {
        if (e.getName().equalsIgnoreCase("Command") || (e.getName().equalsIgnoreCase("W"))) {
            commandTag = "W";
            processCommand();
        } else if (e.getName().equalsIgnoreCase("C")) {
            processCommand();
            commandTag = "C";
            commandToProcess = e.getAttributeValue("N");
            commandLevel = level;
        }
        if (e.getName().equalsIgnoreCase("P")) {
            if (commandTag.equalsIgnoreCase("W")) {
                if (globalParams.size() < level) {
                    for (int i = globalParams.size() + 1; i <= level; i++) {
                        globalParams.add(new HashMap<String, String>());
                    }
                }
                globalParams.get(level - 1).put(e.getAttributeValue("N"), e.getText());
            } else if (commandTag.equalsIgnoreCase("C")) {
                localParams.put(e.getAttributeValue("N"), e.getText());
            }
        }
        List<Element> children = e.getChildren();
        Iterator<Element> iter = children.iterator();
        while (iter.hasNext()) {
            Element child = iter.next();
            processXML(child, level + 1);
        }
    }
        
    private String checkParameter(String param) {
        if (effectiveParams.containsKey(param)) {
            return effectiveParams.get(param);
        }
        return null;
    }
    
    private synchronized void setUid(String uid) {
        if (!WorkArea.this.uid.equals(uid)) {
            WorkArea.this.uid = uid;
            instructions.add(Utils.instruction(Instructions.b, uid));
        }
    }
    
    public void setTempDir(String d) {
        tempDir = d;
    }
    
    private boolean checkWindow(String code) {
        contextWindow = null;
        UiUnit u = uiRoot.findChild(code, clWindow);
        if (u != null) {
            contextWindow = (UiWindow)u;
            return true;
        }
        return false;
    }
    
    private boolean checkForm(UiUnit start, String code) {
        contextForm = null;
        contextFormContent = null;
        UiUnit u = start.findChild(code, clForm);
        if (u != null) {
            contextForm = (UiForm)u;
            contextFormContent = contextForm.getContent();
            return true;
        }
        return false;
    }
    
    private boolean checkNode(UiUnit start, String code) {
        contextNode = null;
        UiUnit u = start.findChild(code, clNode);
        if (u != null) {
            contextNode = (UiTreeNode)u;
            return true;
        }
        return false;
    }
    
    private boolean checkField(UiUnit start, String code) {
        contextField = null;
        UiUnit u = start.findChild(code, clField);
        if (u != null) {
            contextField = (UiField)u;
            return true;
        }
        return false;
    }
    
    private boolean checkGroup(UiUnit start, String code) {
        contextGroup = null;
        UiUnit u = start.findChild(code, clGroup);
        if (u != null) {
            contextGroup = (UiGroup) u;
            return true;
        }
        return false;
    }
    
    private boolean checkButton(UiUnit start, String code) {
        contextButton = null;
        UiUnit u = start.findChild(code, clButton);
        if (u != null) {
            contextButton = (UiBaseButton)u;
            return true;
        }
        return false;
    }
    
    private boolean checkFieldButton(UiUnit start, String code) {
        contextButton = null;
        UiUnit u = start.findChild(code, clFieldButton);
        if (u != null) {
            contextButton = (UiBaseButton)u;
            return true;
        }
        return false;
    }
    
    private boolean checkMatrixButton(UiUnit start, String code) {
        contextButton = null;
        UiUnit u = start.findChild(code, clMatrixButton);
        if (u != null) {
            contextButton = (UiBaseButton)u;
            return true;
        }
        return false;
    }
    
    private boolean checkToolButton(UiUnit start, String code) {
        contextButton = null;
        UiUnit u = start.findChild(code, clToolButton);
        if (u != null) {
            contextButton = (UiBaseButton)u;
            return true;
        }
        return false;
    }
    
    private UiSubMenu findSubMenu(String id) {
        for (UiSubMenu sm : localMenu) {
            if (sm.getId().equals(id)) {
                return sm;
            }
        }
        return null;
    }
        
    private void buildMenu(String itemId) {
        UiSubMenu subMenu = findSubMenu(itemId);
        menuHTML.append(subMenu.HTMLstart());
        for (UiMenuObject menuObj : subMenu) {
            if (menuObj instanceof UiLocalMenuItem) {
                UiLocalMenuItem item = (UiLocalMenuItem) menuObj;
                localMenuItems.put(menuObj.getUid(), (UiLocalMenuItem) menuObj);
                menuHTML.append(item.HTML());
            } else if (menuObj instanceof UiGroupMenuItem) {
                UiGroupMenuItem item = (UiGroupMenuItem) menuObj;
                menuHTML.append(item.HTMLstart());
                buildMenu(item.getId());
                menuHTML.append(item.HTMLend());
            } else if (menuObj instanceof UiLocalMenuSeparator) {
                menuHTML.append(((UiLocalMenuSeparator) menuObj).HTML());
            } else if (menuObj instanceof UiAddonMenuItem) {
                menuHTML.append(((UiAddonMenuItem) menuObj).HTML());
            }
        }
        menuHTML.append(subMenu.HTMLend());
    }
    
    private void processCommand() {
        if (commandToProcess != null) {
            effectiveParams.clear();
            try {
                int i;
                for (i = 0; i < commandLevel; i++) {
                    if ((!globalParams.isEmpty()) && (!globalParams.get(i).isEmpty())) {
                        effectiveParams.putAll(globalParams.get(i));
                    }
                }
            } catch (Exception ex) {
            }
            effectiveParams.putAll(localParams);
            
            //System.out.println("---");
            //System.out.println("command to process> "+commandToProcess);
            
            switch (commandToProcess) {
                
                case "Cell": 
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                param = checkParameter("Code");
                                if (param != null) {
                                    contextGrid = (UiGrid) contextFormContent;
                                    contextColumn = contextGrid.checkColumn(param);
                                    if (contextColumn != null) {
                                        contextRow = contextGrid.getContextRow();
                                        param = checkParameter("RowNum");
                                        if (param != null) {
                                            contextRow = contextGrid.getRowByNum(param);
                                        }
                                        if (contextRow != null) {
                                            contextCell = contextColumn.getCell(contextRow);
                                            if (contextCell != null) {
                                                contextCell.setValue(checkParameter("Show"));
                                                contextCell.setButtons(checkParameter("Buttons"));
                                                contextCell.setEditable(checkParameter("Edit"));
                                                contextCell.setStyle(checkParameter("Style"));
                                                contextCell.setStyle(checkParameter("Class"));
                                                contextCell.setColoringStyle(checkParameter("GetUp"));
                                                contextCell.setColoring(checkParameter("MakeUp"));
                                                contextCell.setHint(checkParameter("Hint"));
                                                contextColumn.setAlign(checkParameter("Alignment"));
                                            }    
                                        }
                                    }
                                }
                            } else if (contextForm.type == FormType.Grid) {
                                param = checkParameter("R-code");
                                if (param != null) {
                                    contextMatrix = (UiMatrix) contextFormContent;
                                    contextMatrixRow = contextMatrix.checkRow(param);
                                    if (contextMatrixRow != null) {
                                        param = checkParameter("C-code");
                                        if (param != null) {
                                            contextMatrixColumn = contextMatrix.checkColumn(param);
                                            if (contextMatrixColumn != null) {
                                                contextMatrixCell = contextMatrix.getCell(contextMatrixRow, contextMatrixColumn);
                                                if (contextMatrixCell == null) {
                                                    contextMatrixCell = new UiMatrixCell(contextMatrixRow, contextMatrixColumn);
                                                    contextMatrix.addChild(contextMatrixCell);
                                                }
                                                contextMatrixCell.setValue(checkParameter("Show"));
                                                contextMatrixCell.setHint(checkParameter("Hint"));
                                                contextMatrixCell.setHAlign(checkParameter("H-align"));
                                                contextMatrixCell.setHAlign(checkParameter("Alignment"));
                                                contextMatrixCell.setVAlign(checkParameter("V-align"));
                                                contextMatrixCell.setOverflow(checkParameter("Overflow"));
                                                contextMatrixCell.setClasses(checkParameter("Class"));
                                                contextMatrixCell.setClasses(checkParameter("Style"));
                                                contextMatrixCell.setMakeup(checkParameter("MakeUp"));
                                                contextMatrixCell.setMUClasses(checkParameter("GetUp"));
                                                contextMatrixCell.setEditable(checkParameter("Edit"));
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                   
                    break;
                
                case "Column": 
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                param = checkParameter("Code");
                                if (param != null) {
                                    contextColumn = ((UiGrid)contextFormContent).checkColumn(param);
                                    if (contextColumn == null) {
                                        contextColumn = new UiGridColumn(param);
                                        contextFormContent.addChild(contextColumn);
                                    } else {
                                        
                                    }
                                    param = checkParameter("Group");
                                    if (param != null) {
                                        if (Utils.testEmptyString(param)) {
                                            contextFormContent.addChild(contextColumn);
                                        } else if (checkGroup(contextFormContent, param)) {
                                            contextGroup.addChild(contextColumn);
                                        }    
                                    }
                                    contextColumn.setTitle(checkParameter("Title"));
                                    contextColumn.setEnabled(checkParameter("Enable"));
                                    contextColumn.setWidth(checkParameter("Width"));
                                    contextColumn.setAlign(checkParameter("Alignment"));
                                    contextColumn.setHint(checkParameter("Hint"));
                                    contextColumn.setStyle(checkParameter("Style"));
                                    contextColumn.setStyle(checkParameter("Class"));
                                    contextColumn.setColoringStyle(checkParameter("GetUp"));
                                    contextColumn.setColoring(checkParameter("MakeUp"));
                                    contextColumn.setSort(checkParameter("Sort"));
                                }
                            } else if (contextForm.type == FormType.Grid) {
                                param = checkParameter("Code");
                                if (param != null) {
                                    contextMatrixColumn = ((UiMatrix)contextFormContent).checkColumn(param);
                                    if (contextMatrixColumn == null) {
                                        contextMatrixColumn = new UiMatrixColumn(param);
                                        param2 = checkParameter("Parent");
                                        if (param2 == null) {
                                            contextFormContent.addChild(contextMatrixColumn);
                                        } else {
                                            UiMatrixColumn parentMatrixColumn = ((UiMatrix)contextFormContent).checkColumn(param2);
                                            if (parentMatrixColumn != null) {
                                                parentMatrixColumn.addChild(contextMatrixColumn);
                                            }
                                        }
                                    } else {
                                        
                                    }
                                    contextMatrixColumn.setVisible(checkParameter("Visible"));
                                    contextMatrixColumn.setInfo(checkParameter("Info"));
                                    contextMatrixColumn.setOrientation(checkParameter("Orientation"));
                                    contextMatrixColumn.setOrientation(checkParameter("Direction"));
                                    contextMatrixColumn.setCaption(checkParameter("Caption"));
                                    contextMatrixColumn.setCaption(checkParameter("Title"));
                                    contextMatrixColumn.setHAlign(checkParameter("H-align"));
                                    contextMatrixColumn.setVAlign(checkParameter("V-align"));
                                    contextMatrixColumn.setOverflow(checkParameter("Overflow"));
                                    contextMatrixColumn.setClasses(checkParameter("Class"));
                                    contextMatrixColumn.setSelfClasses(checkParameter("SelfClass"));
                                    contextMatrixColumn.setClasses(checkParameter("Style"));
                                    contextMatrixColumn.setMakeup(checkParameter("MakeUp"));
                                    contextMatrixColumn.setMUClasses(checkParameter("GetUp"));
                                    contextMatrixColumn.setHint(checkParameter("Hint"));
                                    contextMatrixColumn.setWidth(checkParameter("Width"));
                                }
                            }
                        }
                    }
                    
                    break;
                
                case "AddRow":
                case "Row": 
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                UiGridRow row = new UiGridRow();
                                contextFormContent.addChild(row);
                                row.select(checkParameter("Check"));
                            } else if (contextForm.type == FormType.Grid) {
                                param = checkParameter("Code");
                                if (param != null) {
                                    contextMatrixRow = ((UiMatrix)contextFormContent).checkRow(param);
                                    if (contextMatrixRow == null) {
                                        contextMatrixRow = new UiMatrixRow(param);
                                        param2 = checkParameter("Parent");
                                        if (param2 == null) {
                                            contextFormContent.addChild(contextMatrixRow);
                                        } else {
                                            UiMatrixRow parentMatrixRow = ((UiMatrix)contextFormContent).checkRow(param2);
                                            if (parentMatrixRow != null) {
                                                parentMatrixRow.addChild(contextMatrixRow);
                                            }
                                        }
                                    } else {
                                        
                                    }
                                    contextMatrixRow.setVisible(checkParameter("Visible"));
                                    contextMatrixRow.setInfo(checkParameter("Info"));
                                    contextMatrixRow.setOrientation(checkParameter("Orientation"));
                                    contextMatrixRow.setOrientation(checkParameter("Direction"));
                                    contextMatrixRow.setCaption(checkParameter("Caption"));
                                    contextMatrixRow.setCaption(checkParameter("Title"));
                                    contextMatrixRow.setHAlign(checkParameter("H-align"));
                                    contextMatrixRow.setVAlign(checkParameter("V-align"));
                                    contextMatrixRow.setOverflow(checkParameter("Overflow"));
                                    contextMatrixRow.setClasses(checkParameter("Class"));
                                    contextMatrixRow.setSelfClasses(checkParameter("SelfClass"));
                                    contextMatrixRow.setClasses(checkParameter("Style"));
                                    contextMatrixRow.setMakeup(checkParameter("MakeUp"));
                                    contextMatrixRow.setMUClasses(checkParameter("GetUp"));
                                    contextMatrixRow.setHint(checkParameter("Hint"));
                                    contextMatrixRow.setHeight(checkParameter("Height"));
                                }
                            }
                        }
                    }
                    
                    break;    
                    
                case "Grid":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                contextGrid = (UiGrid) contextFormContent;
                                param = checkParameter("HeadLines");
                                if (param != null) {
                                    contextGrid.setHeader(param);
                                }   
                                param = checkParameter("Totals");
                                if (param != null) {
                                    contextGrid.setFooter(param);
                                } 
                                param = checkParameter("IsLast");
                                if (param != null) {
                                    contextGrid.setAllRows(param);
                                } 
                                contextGrid.setSelectionMode(checkParameter("SelectionMode"));
                                contextGrid.setOverflow(checkParameter("OverflowWidth"));
                            }
                        }
                    }    
                    
                    break;
                    
                case "QBEMode":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                ((UiGrid) contextFormContent).setQbeMode(true);
                            }
                        }
                    }    
                    
                    break;    
                    
                case "CurrentRow":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                UiGrid grid = ((UiGrid) contextFormContent);
                                grid.setCell(checkParameter("RowNum"), checkParameter("Cell"));
                            }
                        }
                    }    
                    
                    break;     
                    
                case "ClearGrid":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                ((UiGrid) contextFormContent).clear();
                            } else if (contextForm.type == FormType.Grid) {
                                ((UiMatrix) contextFormContent).clear();
                            }
                        }
                    }    
                    
                    break;      
                    
                case "Window":
                    param = checkParameter("WindowID");
                    if (param != null) {
                        WindowPosition pos = WindowPosition.Cascade;
                        if (!checkWindow(param)) {
                            param2 = checkParameter("WindowType");
                            setUid(uiRoot.getId());
                            if ("Logon".equals(param2)) {
                                contextWindow = new UiLogonWindow(param);
                            } else if ("MainMenu".equals(param2)) {
                                contextWindow = new UiMenuWindow(param);
                            } else if ("Service".equals(param2)) {
                                contextWindow = new UiServiceWindow(param);    
                            } else {
                                String p = checkParameter("Position");
                                try {
                                    pos = WindowPosition.valueOf(p);
                                } catch (Exception ex) {
                                    pos = WindowPosition.Cascade;
                                }
                                if (pos == WindowPosition.LastClick) {
                                    contextWindow = new UiEmbeddedWindow(param); 
                                } else {
                                    contextWindow = new UiWindow(param);
                                }    
                            }
                            uiRoot.addChild(contextWindow);
                            if (!(contextWindow instanceof UiEmbeddedWindow)) {
                                UiWindow win = contextWindow;
                                param = checkParameter("Position");
                                try {
                                    pos = WindowPosition.valueOf(param);
                                } catch (Exception ex) {
                                    pos = WindowPosition.Cascade;
                                }
                                param = checkParameter("ParentID");
                                if (Utils.testEmptyString(param)) {
                                    if (pos != WindowPosition.WorkCenter) {
                                        pos = WindowPosition.Zero;
                                    }
                                } else {
                                    if (checkWindow(param)) {
                                        win.setParentWindow(contextWindow);
                                    } else {
                                        if (pos != WindowPosition.WorkCenter) {
                                            pos = WindowPosition.Zero;
                                        }
                                    }
                                }
                                win.setPosition(pos);
                                contextWindow = win;
                                contextWindow.applyPosition();
                                contextWindow.unlock();
                            }    
                        }
                        contextWindow.setWidth(checkParameter("Width"));
                        contextWindow.setHeight(checkParameter("Height"));
                        contextWindow.setTitle(checkParameter("Title"));
                        contextWindow.setSysMenu(checkParameter("SysMenu"));
                        contextWindow.setCloseButton(checkParameter("CloseButton"));
                        contextWindow.setMaximizeButton(checkParameter("MaxButton"));
                        contextWindow.setMaximized(checkParameter("Max"));
                        contextWindow.setLeft(checkParameter("Left"));
                        contextWindow.setTop(checkParameter("Top"));    
                    }
                    
                    break;
                    
                case "CloseWindow":
                    param = checkParameter("WindowID");
                    if (checkWindow(param)) {
                        contextWindow.getParent().rmChild(contextWindow);
                    }
                    
                    break;
                    
                case "EnableWindow":
                    param = checkParameter("WindowID");
                    if (checkWindow(param)) {
                        contextWindow.enable();
                    }
                    
                    break;
                    
                case "DisableWindow":
                    param = checkParameter("WindowID");
                    if (checkWindow(param)) {
                        contextWindow.disable(checkParameter("Cover"));
                    }
                    
                    break;    
                    
                case "ShowWindow":
                    param = checkParameter("WindowID");
                    if (checkWindow(param)) {
                        contextWindow.show();
                    }
                    
                    break;
                    
                case "HideWindow":
                    param = checkParameter("WindowID");
                    if (checkWindow(param)) {
                        contextWindow.hide();
                    }
                    
                    break;       
                    
                case "Form":
                    param = checkParameter("FormID");
                    if (checkForm(uiRoot, param)) {

                    } else {
                        param2 = checkParameter("WindowID");
                        if (param2 != null) {
                            if (checkWindow(param2)) {
                                String formCode = param;
                                UiForm form = null;
                                FormPosition fp = FormPosition.Main;
                                FormType ft = null;
                                contextForm = null;
                                param2 = checkParameter("Type");
                                if (param2 != null) {
                                    try {
                                        ft = FormType.valueOf(param2);
                                    } catch (Exception ex) {
                                        ft = null;
                                    }
                                    if (ft != null) {
                                        param2 = checkParameter("ParentID");
                                        if ((!Utils.testEmptyString(param2)) && checkForm(contextWindow, param2)) {
                                            param2 = checkParameter("Position");
                                            try {
                                                fp = FormPosition.valueOf(param2);
                                            } catch (Exception ex) {
                                                fp = FormPosition.Bottom;
                                            }
                                            form = contextForm;
                                            contextForm = new UiForm(formCode);
                                            contextForm.setPosition(fp);
                                            form.addChild(contextForm);
                                            contextForm.setType(ft);
                                        } else {
                                            if (customMenu) {
                                                contextForm = contextWindow instanceof UiMenuWindow ? new UiMenuForm(formCode) : new UiForm(formCode);
                                            } else {
                                                contextForm = new UiForm(formCode);
                                            }    
                                            contextWindow.addChild(contextForm);
                                            contextForm.setPosition(FormPosition.Main);
                                            contextForm.setType(ft);
                                        }
                                        contextForm.unlock();
                                    }
                                }
                            }
                        }
                    }
                    if (contextForm != null) {
                        param = checkParameter("Title");
                        if (param != null || protocol == 3) {
                            contextHeader = (UiHeader)contextForm.getHeader();
                            if (contextHeader == null) {
                                contextHeader = new UiFormHeader();
                                contextForm.addChild(contextHeader);
                            }
                        }
                        if (param != null) {
                            contextHeader.setText(param);
                        }
                        param = checkParameter("ToolBarButtons");
                        if (param != null) {
                            contextToolBar = (UiToolBar)contextForm.getToolBar();
                            //if (!Utils.testEmptyString(param)) {
                                if (contextToolBar == null) {
                                    contextToolBar = new UiToolBar();
                                    contextForm.addChild(contextToolBar);
                                }
                                contextButtonBar = (UiButtonBar) contextToolBar.getButtonBar();
                                if (contextButtonBar == null) {
                                    contextButtonBar = new UiButtonBar();
                                    contextToolBar.addChild(contextButtonBar);
                                }
                                contextButtonBar.set(param);
                            //}
                        }
                        param = checkParameter("ConfigList");
                        if (param != null) {
                            contextToolBar = (UiToolBar)contextForm.getToolBar();
                            if (!Utils.testEmptyString(param)) {
                                if (contextToolBar == null) {
                                    contextToolBar = new UiToolBar();
                                    contextForm.addChild(contextToolBar);
                                }
                                contextConfigs = (UiConfigs) contextToolBar.getConfigs();
                                if (contextConfigs == null) {
                                    contextConfigs = new UiConfigs();
                                    contextToolBar.addChild(contextConfigs);
                                }
                                contextConfigs.set(param);
                            }
                        }
                        contextForm.setEastSize(checkParameter("SplitterRight"));
                        contextForm.setSouthSize(checkParameter("SplitterBottom"));
                        contextForm.setWidth(checkParameter("Width"));
                        contextForm.setHeight(checkParameter("Height"));
                        contextForm.setCover(checkParameter("Cover"));
                    }
            
                    break;   
                    
                case "CloseForm":    
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            setUid(contextForm.getId());
                            instructions.add(Utils.instruction(Instructions.clos));
                            contextForm.remove();
                        }
                    }
                    
                    break;
                    
                case "Matrix":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Grid) {
                                contextMatrix = (UiMatrix) contextFormContent;
                                contextMatrix.setAngularText(checkParameter("AngularText"));
                                contextMatrix.setAngularHint(checkParameter("AngularHint"));
                                contextMatrix.setTitle(checkParameter("Title"));
                                contextMatrix.setTitleBorder(checkParameter("TitleBorder"));
                                contextMatrix.setColumnInfo(checkParameter("ColumnInfo"));
                                contextMatrix.setRowInfo(checkParameter("RowInfo"));
                                contextMatrix.setAngularWidth(checkParameter("AngularWidth"));
                                contextMatrix.setSelectionMode(checkParameter("SelectionMode"));
                            }
                        }
                    }
                    
                    break;
                    
                case "MatrixFinish":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Grid) {
                                contextMatrix = (UiMatrix) contextFormContent; 
                                contextMatrix.finish();
                            }
                        }
                    }
                    
                    break;
                    
                case "Focus":
                    if (!mobile) {
                        param = checkParameter("FormID");
                        if (checkForm(uiRoot, param)) {
                            formToFocus = contextForm.getId();
                            setUid(formToFocus);
                            instructions.add(Utils.instruction(Instructions.cfcs));
                        }
                    }
                    
                    break;
                    
                case "Node":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Tree) {
                                param = checkParameter("Node");
                                if (param != null) {
                                    if (checkNode(contextFormContent, param)) {

                                    } else {
                                        if (customMenu) {
                                            contextNode = new UiMenuNode(param);
                                        } else {
                                            contextNode = new UiTreeNode(param);
                                        }    
                                        param = checkParameter("ParentNode");
                                        if (param == null) {
                                            contextFormContent.addChild(contextNode);
                                        } else {
                                            UiTreeNode n = contextNode;
                                            if (checkNode(contextFormContent, param)) {
                                                contextNode.addChild(n);
                                            } else {
                                                contextFormContent.addChild(n);
                                            }
                                            contextNode = n;
                                        }
                                    }
                                    contextNode.setLabel(checkParameter("Label"));
                                    contextNode.setHint(checkParameter("Hint"));
                                    contextNode.setIcon(checkParameter("Icon"));
                                    contextNode.setType(checkParameter("Type"));
                                    contextNode.setStyle(checkParameter("Style"));
                                    contextNode.setContext(checkParameter("Context"));
                                    contextNode.setContextId(checkParameter("ContextID"));
                                    contextNode.setChildsContextId(checkParameter("ChildsContextID"));
                                    contextNode.setState(checkParameter("State"));
                                    contextNode.setSelected(checkParameter("Selected"));
                                }
                            }
                        }
                    }
                    
                    break;
                    
                case "DeleteNode":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Node");
                            if (param != null) {
                                if (checkNode(contextFormContent, param)) {
                                    contextNode.remove();
                                }
                            }
                        }
                    }
                    
                    break;
                
                case "DeleteChilds":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Node");
                            if (Utils.testEmptyString(param)) {
                                if (contextForm.type == FormType.Tree) {
                                    contextFormContent.rmChildren();
                                }
                            } else if (checkNode(contextFormContent, param)) {
                                contextNode.rmChildren();
                            }
                        }
                    }
                    
                    break;
                
                case "ToolBarButton":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Name");
                            if (param != null) {
                                contextToolBar = (UiToolBar)contextForm.getToolBar();
                                if (contextToolBar == null) {
                                    contextToolBar = new UiToolBar();
                                    contextForm.addChild(contextToolBar);
                                }
                                if (param.equals("")) {
                                    UiSeparator s = new UiSeparator();
                                    contextToolBar.addChild(s);
                                } else {
                                    if (checkToolButton(contextToolBar, param)) {
                                        
                                    } else {
                                        contextButton = new UiToolButton(param);
                                        contextToolBar.addChild(contextButton);
                                    }
                                    contextButton.setHint(checkParameter("Hint"));
                                    contextButton.setState(checkParameter("State"));
                                }
                            }
                        }
                    }
                    
                    break;
                    
                case "Button":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Code");
                            if (param != null) {
                                if (checkButton(contextForm, param)) {
                                    
                                } else {
                                    contextButton = new UiButton(param);
                                    param = checkParameter("Group");
                                    if ((param != null) && checkGroup(contextFormContent, param)) {
                                        contextGroup.addChild(contextButton);
                                    } else {
                                        contextForm.addChild(contextButton);
                                    }
                                    contextButton.setState("Disable");
                                }
                                contextButton.setHint(checkParameter("Hint"));
                                contextButton.setState(checkParameter("State"));
                                ((UiButton)contextButton).setLabel(checkParameter("Label"));
                            }
                        }
                    }
                    
                    break;    
                    
                case "Underside":
                    param = checkParameter("FormID");
                    if ((param != null) && (!mobile)) {
                        if (checkForm(uiRoot, param)) {
                            UiUnit u = contextForm.getBasement();
                            contextBasement = null;
                            if (u != null) {
                                contextBasement = (UiBasement) u;
                            }
                            param = checkParameter("Content");
                            if (contextBasement == null) {
                                contextBasement = new UiBasement();
                                contextForm.addChild(contextBasement);
                            }
                            contextBasement.setContent(param);
                            param = checkParameter("Height");
                            if (param != null) {
                                if (param.equals("")) {
                                    if (contextBasement != null) {
                                        contextBasement.remove();
                                    }
                                } else {
                                    if (contextBasement == null) {
                                        contextBasement = new UiBasement();
                                        contextForm.addChild(contextBasement);
                                    }
                                    contextBasement.setHeight(param);
                                }
                            }
                        }
                    }
                    
                    break;
                    
                case "LocalMenu":
                    showMenu = false;
                    localMenu = new ArrayList<>();
                    localMenu.add(new UiSubMenu("@"));
                    localMenuItems = new HashMap<>();
                    
                    break;    
                
                case "AddMenuItem":
                    param = checkParameter("Label");
                    if (param != null) {
                        showMenu = true;
                        String[] splittedLabel = param.split("\\|");
                        UiSubMenu subMenu;
                        UiGroupMenuItem grpMenuItem;
                        UiLocalMenuItem menuItem;
                        UiMenuObject menuObj;
                        String subMenuId = "@";
                        String menuItemId = "@";
                        subMenu = findSubMenu(subMenuId);
                        if (splittedLabel.length > 1) {
                            for (int i = 0; i < splittedLabel.length - 1; i++) {
                                menuItemId = menuItemId + splittedLabel[i] + "@";
                                menuObj = subMenu.findObject(menuItemId);
                                if ((menuObj != null) && (!(menuObj instanceof UiGroupMenuItem))) {
                                    subMenu.remove(menuObj);
                                }
                                if (subMenu.findObject(menuItemId) == null) {
                                    grpMenuItem = new UiGroupMenuItem(menuItemId, splittedLabel[i]);
                                    grpMenuItem.setGroup(menuItemId);
                                    if (subMenu.size() > 0) {
                                        subMenu.add(new UiLocalMenuSeparator());
                                    }
                                    subMenu.add(grpMenuItem);
                                } else {
                                    grpMenuItem = (UiGroupMenuItem) subMenu.findObject(menuItemId);
                                }
                                subMenuId = menuItemId;
                                if (grpMenuItem.getSubMenu() == null) {
                                    subMenu = new UiSubMenu(subMenuId);
                                    grpMenuItem.setSubMenu(subMenu);
                                    localMenu.add(subMenu);
                                } else {
                                    subMenu = grpMenuItem.getSubMenu();
                                }
                            }
                        }
                        menuItemId = menuItemId + splittedLabel[splittedLabel.length - 1] + "@";
                        menuItem = new UiLocalMenuItem(menuItemId);
                        menuItem.setLabel(splittedLabel[splittedLabel.length - 1]);
                        param = checkParameter("Group");
                        String grpCurrent = (param == null) ? "menugroup$" : param;
                        menuItem.setGroup(grpCurrent);
                        if (subMenu.size() > 0) {
                            String grpPrev = subMenu.get(subMenu.size() - 1).getGroup();
                            if (grpPrev != null && (!grpPrev.equals(grpCurrent))) {
                                subMenu.add(new UiLocalMenuSeparator());
                            }
                        }
                        subMenu.add(menuItem);
                        param = checkParameter("FormID");
                        if (param != null) {
                            menuItem.setFormId(param);
                        }
                        param = checkParameter("Type");
                        if (param != null) {
                            try {
                                MenuItemType mit = MenuItemType.valueOf(param);
                                menuItem.setType(param);
                            } catch (Exception e) {
                            }
                        }
                        param = checkParameter("Code");
                        if (param != null) {
                            menuItem.setCode(param);
                        }
                        param = checkParameter("Params");
                        if (param != null) {
                            menuItem.setParams(param);
                        }
                        param = checkParameter("Role");
                        if (param != null) {
                            try {
                                MenuItemRole mir = MenuItemRole.valueOf(param);
                                menuItem.setRole(param);
                            } catch (Exception e) {
                            }
                        }
                        param = checkParameter("Enable");
                        if (param != null) {
                            menuItem.setEnabled(Utils.testYString(param));
                        }
                        param = checkParameter("Style");
                        if (param != null) {
                            menuItem.setStyle(param);
                        }
                    }
                    
                    break;

                case "AddLocalItems":
                    UiSubMenu subMenu = findSubMenu("@");
                    if (subMenu != null) {
                        showMenu = true;
                        param = checkParameter("Group");
                        if (param != null) {
                            if (subMenu.size() > 0) {
                                String grpPrev = subMenu.get(subMenu.size() - 1).getGroup();
                                if (grpPrev != null && (!grpPrev.equals(param))) {
                                    subMenu.add(new UiLocalMenuSeparator());
                                }
                            }
                            UiAddonMenuItem addonItem = new UiAddonMenuItem();
                            subMenu.add(addonItem);
                        }
                    }
                    
                    break;

                case "TotalCell":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.List) {
                                param = checkParameter("Code");
                                if (param != null) {
                                    contextColumn = ((UiGrid)contextFormContent).checkColumn(param);
                                    if (contextColumn != null) {
                                        contextColumn.setFooter(checkParameter("Show"));
                                    } 
                                }
                            }
                        }
                    }
                    
                    break;
                    
                case "Field":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Code");
                            if (param != null) {
                                if (checkField(contextForm.getContent(), param)) {
                                    
                                } else {
                                    contextField = new UiField(param);
                                    param = checkParameter("Group");
                                    if (param == null) {
                                        contextFormContent.addChild(contextField);
                                    } else {
                                        if (checkGroup(contextForm, param)) {
                                            contextGroup.addChild(contextField);
                                        } else {
                                            contextFormContent.addChild(contextField);
                                        }
                                    }
                                }
                                contextField.setType(checkParameter("Type"));
                                contextField.setPrompt(checkParameter("Prompt"));
                                contextField.setLines(checkParameter("Lines"));
                                contextField.setValues(checkParameter("Values"));
                                contextField.setValue(checkParameter("Show"));
                                contextField.setReadOnly(checkParameter("ReadOnly"));
                                contextField.setMandatory(checkParameter("Mandatory"));
                                contextField.setMode(checkParameter("Mode"));
                                contextField.setHint(checkParameter("Hint"));
                                contextField.setIcon(checkParameter("Icon"));
                                contextField.setStyle(checkParameter("Style"));
                                contextField.setNavigable(checkParameter("KeyNavigable"));
                                contextField.setCursor(checkParameter("Cursor"));
                                contextField.setPattern(checkParameter("Pattern"));
                                param = checkParameter("Verifier");
                                if (param != null) {
                                    contextField.setVerifier(param);
                                }
                                //contextField.setVerifierParams(checkParameter("VerifierParams"));
                                param = checkParameter("VerifierParams");
                                if (param != null) {
                                    contextField.setVerifierParams(param);
                                }
                                contextField.setColoringStyle(checkParameter("GetUp"));
                                contextField.setColoring(checkParameter("MakeUp"));
                                contextField.setHelperText(checkParameter("HelperText"));
                                contextField.setMin(checkParameter("ValueMin"));
                                contextField.setMax(checkParameter("ValueMax"));
                                contextField.setStep(checkParameter("ValueStep"));
                                contextField.setFrom(checkParameter("ValueFrom"));
                            }    
                        }
                    }
                    
                    break;
                    
                case "CurrentField":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Field");
                            if (Utils.testEmptyString(param)) {
                                setUid(contextForm.getContent().getId());
                            } else {    
                                if (checkField(contextForm.getContent(), param)) {
                                    unitToFocus = contextField.getId();
                                    setUid(unitToFocus);
                                } 
                            }
                            instructions.add(Utils.instruction(Instructions.curr));
                        }
                    }
                    
                    break; 
                    
                case "NextField":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {    
                            //if (contextForm.type == FormType.Blanc) {
                                unitToFocus = contextForm.getId();
                                setUid(contextForm.getId());
                                instructions.add(Utils.instruction(Instructions.nxt));
                            //}
                        }
                    }
                    
                    break;
                    
                case "FieldButton":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Field");
                            param2=param;
                            if (param != null) {
                                if (checkField(contextForm.getContent(), param)) {
                                    UiUnit u = contextField.getChildByClass(clButtonBar);
                                    if (u == null) {
                                        contextButtonBar = new UiButtonBar();
                                        contextField.addChild(contextButtonBar);
                                        if (protocol == 3) {
                                            UiFieldButton b = new UiFieldButton("more");
                                            contextButtonBar.addChild(b);
                                            b.setState("More");
                                        }
                                    } else {
                                        contextButtonBar = (UiButtonBar)u;
                                    }
                                    param = checkParameter("Code");
                                    if (param != null) {
                                        checkFieldButton(contextButtonBar, param);
                                        if (contextButton == null) {
                                            contextButton = new UiFieldButton(param);
                                            contextButtonBar.addChild(contextButton);
                                        } else {
                                            
                                        } 
                                        contextButton.setHint(checkParameter("Hint"));
                                        contextButton.setState(checkParameter("State"));
                                        contextButton.setBgColor(checkParameter("BgColor"));
                                    }
                                } 
                            }    
                        }
                    }
                    
                    break;    
                    
                case "DropDown":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Field");
                            if (param != null) {
                                if (checkField(contextForm.getContent(), param)) {
                                    contextField.dropDownSelect(checkParameter("Selected"));
                                    contextField.dropDown(checkParameter("Contents"));
                                } 
                            }    
                        }
                    } 
                    
                    break;   
                    
                case "ColorPicker":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Field");
                            if (param != null) {
                                if (checkField(contextForm.getContent(), param)) {
                                    contextField.colorPalette(checkParameter("Palette"));
                                    contextField.colorPicker(checkParameter("Color"));
                                } 
                            }    
                        }
                    } 
                    
                    break;     
                    
                case "Calendar":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (calendar != null) {
                                calendar.remove();
                            }
                            calendar = new UiCalendar();
                            uiRoot.addChild(calendar);
                            calendar.setForm(param);
                            calendar.setSpecialDates(checkParameter("SpecialDates"));
                            calendar.setDefault(checkParameter("Default"));
                            calendar.setDateDetail(checkParameter("DateDetail"));
                            //calendar.setDateDetail("M");//
                            calendar.setTimeDetail(checkParameter("TimeDetail"));
                            //calendar.setTimeDetail("S");
                        }
                    }
                    
                    break;            
                    
                case "Group":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Code");
                            if (param != null) {
                                if (checkGroup(contextForm, param)) {
                                    
                                } else {
                                    param2 = checkParameter("Type");
                                    switch (Utils.string2String(param2)) {
                                        case "Virtual" :    switch (contextForm.type) {
                                                                case List:  contextGroup = new UiGridGroup(param);
                                                                            break;
                                                                case Blanc: contextGroup = new UiFieldsGroup(param);
                                                                            break;
                                                                default:    contextGroup = null;         
                                                            }
                                                            break;
                                                            
                                        case "SideBySide" : switch (contextForm.type) {
                                                                case List:  contextGroup = new UiGridGroup(param);
                                                                            break;
                                                                case Blanc: contextGroup = new UiHorizontalGroup(param);
                                                                            break;
                                                                default:    contextGroup = null;         
                                                            }
                                                            break;
                                                                  
                                        case "Bunch" :      switch (contextForm.type) {
                                                                case List:  contextGroup = new UiGridGroup(param);
                                                                            break;
                                                                case Blanc: contextGroup = new UiBunchGroup(param);
                                                                            break;
                                                                default:    contextGroup = null;         
                                                            }
                                                            break;                     
                                                             
                                        case "Fixed" :      switch (contextForm.type) {
                                                                case List:  contextGroup = new UiFixedGroup(param);
                                                                            break;
                                                                case Blanc: contextGroup = new UiFieldsGroup(param);
                                                                            break;
                                                                default:    contextGroup = null;         
                                                            }
                                                            break;
                                                            
                                        case "Overflow" :   switch (contextForm.type) {
                                                                case List:  contextGroup = new UiOverflowGroup(param);
                                                                            break;
                                                                case Blanc: contextGroup = new UiFieldsGroup(param);
                                                                            break;
                                                                default:    contextGroup = null;         
                                                            }
                                                            break;
                                        
                                        case "Tab":
                                        case "":            switch (contextForm.type) {
                                                                case List:  contextGroup = new UiOverflowGroup(param);
                                                                            break;
                                                                case Blanc: contextGroup = new UiTabGroup(param);
                                                                            break;
                                                                default:    contextGroup = null;         
                                                            }
                                                            break;
                                                            
                                        default :           contextGroup = null; 
                                      
                                    }
                                    if (contextGroup != null) {
                                        param = checkParameter("Group");
                                        if (param == null) {
                                            contextFormContent.addChild(contextGroup);
                                        } else {
                                            UiGroup saveGroup = contextGroup;
                                            if (checkGroup(contextForm, param)) {
                                                contextGroup.addChild(saveGroup);
                                                contextGroup = saveGroup;
                                            } else {
                                                contextGroup = saveGroup;
                                                contextFormContent.addChild(contextGroup);
                                            }
                                        }
                                    }
                                }
                                if (contextGroup != null) {
                                    contextGroup.setState(checkParameter("State"));
                                    contextGroup.setTitle(checkParameter("Prompt"));
                                    contextGroup.setPromptsPos(checkParameter("FieldsPromptPosition"));
                                    contextGroup.setHint(checkParameter("Hint"));
                                }   
                            }
                        }
                    }
                    
                    break;
                    
                case "CellButton":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Grid) {
                                param = checkParameter("R-code");
                                if (param != null) {
                                    contextMatrix = (UiMatrix) contextFormContent;
                                    contextMatrixRow = contextMatrix.checkRow(param);
                                    if (contextMatrixRow != null) {
                                        param = checkParameter("C-code");
                                        if (param != null) {
                                            contextMatrixColumn = contextMatrix.checkColumn(param);
                                            if (contextMatrixColumn != null) {
                                                contextMatrixCell = contextMatrix.getCell(contextMatrixRow, contextMatrixColumn);
                                                if (contextMatrixCell != null) {
                                                    param = checkParameter("Code");
                                                    if (param != null) {
                                                        checkMatrixButton(contextMatrixCell, param);
                                                        if (contextButton == null) {
                                                            contextButton = new UiMatrixButton(param);
                                                            contextMatrixCell.addChild(contextButton);
                                                        } else {

                                                        }
                                                    }
                                                }
                                                contextButton.setHint(checkParameter("Hint"));
                                                contextButton.setState(checkParameter("State"));
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    break;        
                
                case "Kill":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Code");
                            if (param != null) {
                                UiUnit u = contextForm.findByCode(param);
                                if (u != null) {
                                    u.remove();
                                }
                            }    
                        }
                    } 
                    
                    break;      
                   
                case "Undo":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Code");
                            if (param != null) {
                                UiUnit u = contextForm.findByCode(param);
                                if (u != null) {
                                    u.undo();
                                }
                            }    
                        }
                    } 
                    
                    break;
                    
                case "Redo":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Code");
                            if (param != null) {
                                UiUnit u = contextForm.findByCode(param);
                                if (u != null) {
                                    u.redo();
                                }
                            }    
                        }
                    } 
                    
                    break;    
                    
                case "Chart":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Chart) {
                                contextChart = (UiChart) contextFormContent;
                                contextChart.setOrientation(checkParameter("Orientation"));
                                contextChart.setWidth(checkParameter("Width"));
                                contextChart.setHeight(checkParameter("Height"));
                                param = checkParameter("Command");
                                if (Utils.testEmptyString(param) || param.equalsIgnoreCase("generate")) {
                                    if (!contextChart.isGenerated()) {
                                        contextChart.generate(checkParameter("PieData"), checkParameter("Data"));  
                                    }
                                } else {
                                    if (contextChart.isGenerated()) {
                                        contextChart.setCommand(param);
                                        contextChart.setData(checkParameter("PieData"), checkParameter("Data"));
                                        contextChart.execCommand();
                                    }
                                }
                                if (contextChart.isGenerated()) {
                                    contextChart.setLegendMode(checkParameter("LegendMode"));
                                    contextChart.setPosition(checkParameter("Position"));
                                }
                            }
                        }
                    }    
                    
                    break;
                    
                case "SaveChart":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Chart) {
                                contextChart = (UiChart) contextFormContent;
                                contextChart.save(checkParameter("FileName"));
                            }
                        }
                    }    
                    
                    break; 
                    
                case "GetChart":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Chart) {
                                contextChart = (UiChart) contextFormContent;
                                contextChart.get();
                            }
                        }
                    }    
                    
                    break;     
                    
                case "Html":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            if (contextForm.type == FormType.Html) {
                                contextHtml = (UiHtml) contextFormContent;
                                contextHtml.setHtml(checkParameter("Data"));
                            }
                        }
                    }    
                    
                    break;     
                    
                case "SetProperties":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            contextForm.setProperties(checkParameter("Code"), checkParameter("Properties"));
                        }
                    }    
                    
                    break;      
                
                case "LOV":
                    if (lov != null) {
                        lov.remove();
                    }
                    lov = new UiLov();
                    uiRoot.addChild(lov);
                    lov.setTitle(checkParameter("Title"));
                    lov.setColumns(checkParameter("HeadLines"));
                    lov.setSensitivity(checkParameter("CaseSensitive"));
                    
                    break;     
                    
                case "RowLOV":
                    if (lov != null) {
                        param = checkParameter("ID");
                        //if (param != null) {
                            lov.addRow(Utils.string2String(param), checkParameter("Type"), checkParameter("Parent"));
                            lov.setContext(checkParameter("Context"));
                            lov.setAllRows(checkParameter("IsLast"));
                        //}
                        
                    }
                    
                    break;
                    
                case "ClearLOV":
                    if (lov != null) {
                        lov.clear();
                    }
                    
                    break;
                    
                case "ShowQuestion":
                    param = checkParameter("Question");
                    if (param != null) {
                        question = new UiQuestion();
                        uiRoot.addChild(question);
                        question.setText(param);
                        question.setTitle(checkParameter("Title"));
                        question.setButtons(checkParameter("Buttons"));
                        question.initQuestion();
                    }
                    
                    break;    
                    
                case "Message":
                    UiMessage mw = new UiMessage();
                    uiRoot.addChild(mw);
                    mw.setTitle(checkParameter("Title")); 
                    mw.setType(checkParameter("Type"));   
                    mw.setText(checkParameter("Text")); 
                    mw.setNotes(checkParameter("Notes"));
                    mw.show();
                    
                    break;        
                    
                case "Event":
                    param = checkParameter("EventCode");
                    if (param != null) {
                        instructions.add(Utils.instruction(Instructions.evtc, param));
                        param = checkParameter("MS");
                        Integer ms;
                        try {
                            ms = Integer.parseInt(param);
                            if (ms < 1) {
                                ms = 1;
                            }
                        } catch (Exception ex) {
                            ms = 1;
                        }
                        instructions.add(Utils.instruction(Instructions.evtd, ms.toString()));
                        param = checkParameter("Params");
                        if (param != null) {
                            UiCollection c = new UiCollection(param);
                            if (!c.isEmpty()) {
                                while (!c.isLast()) {
                                    instructions.add(Utils.instruction(Instructions.par, c.getCode()));
                                    instructions.add(Utils.instruction(Instructions.parv, c.getValue()));
                                    c.getNext();
                                }
                            }
                        }
                        instructions.add(Utils.instruction(Instructions.evt));
                    }
                    
                    break;
                    
                case "Info":
                    param = checkParameter("Target");
                    InfoTarget infoTarget;
                    try {
                        infoTarget = InfoTarget.valueOf(param);
                    } catch (Exception e) {
                        infoTarget = InfoTarget.Float;
                    }
                    instructions.add(Utils.instruction(Instructions.inft, Utils.string2String(checkParameter("Title"))));
                    instructions.add(Utils.instruction(Instructions.infc, Utils.string2String(checkParameter("Content"))));
                    instructions.add(Utils.instruction(Instructions.infa, ""));
                    instructions.add(Utils.instruction(Instructions.infp, ""));
                    instructions.add(Utils.instruction(Instructions.inf, infoTarget.toString()));
                    
                    break;
                    
                case "RegisterMap":
                    param = checkParameter("Code");
                    if (!Utils.testEmptyString(param)) {
                        param2 = checkParameter("Data");
                        if (!Utils.testEmptyString(param2)) {
                            instructions.add(Utils.instruction(Instructions.mapc, param));
                            instructions.add(Utils.instruction(Instructions.map, param2));
                        }
                    }
                    break;
                    
                case "URL":
                    param = checkParameter("Address");
                    if (!Utils.testEmptyString(param)) {
                        instructions.add(Utils.instruction(Instructions.infc, ""));
                        instructions.add(Utils.instruction(Instructions.inft, ""));
                        instructions.add(Utils.instruction(Instructions.infa, param));
                        param = checkParameter("Params");
                        instructions.add(Utils.instruction(Instructions.infp, Utils.string2String(param)));
                        URLTarget target;
                        param = checkParameter("Target");
                        try {
                            target = URLTarget.valueOf(param);
                        } catch (Exception e) {
                            target = URLTarget.Tab;
                        }
                        instructions.add(Utils.instruction(Instructions.inf, target.toString()));
                    }
                    
                    break;    
                    
                case "Dict":
                    param = checkParameter("Key");
                    if (param != null) {
                        instructions.add(Utils.instruction(Instructions.dk, param));
                        instructions.add(Utils.instruction(Instructions.dv, checkParameter("Value")));
                    }
                    break;
                    
                case "Init":
                    if (!mobile) {
                        if (!running) {
                            running = true;
                            param = checkParameter("Appearance");
                            if (param != null) {
                                if (param.equals("Trendy")) {
                                    protocol = 3;
                                    instructions.add(Utils.instruction(Instructions.prot, String.valueOf(protocol)));
                                }
                                instructions.add(Utils.instruction(Instructions.appr, param));
                            }
                            /*  VPS TESTING  */
                            //param = "vps";
                            //instructions.add(Utils.instruction(Instructions.appr, param.));
                            /* VPS */
                            instructions.add(Utils.instruction(Instructions.run));
                        }
                        param = checkParameter("Application");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.app, param));
                        }
                        param = checkParameter("Favicon");
                        if (!mobile && param != null) {
                            instructions.add(Utils.instruction(Instructions.fav, param));
                        }
                        param = checkParameter("Skin");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.skin, param));
                        }
                        param = checkParameter("Locale");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.loc, param.toLowerCase()));
                        }
                        param = checkParameter("Cursor");
                        if (param != null) {
                            if (Utils.testEmptyString(param)) {
                                instructions.add(Utils.instruction(Instructions._cur));
                            } else {
                                instructions.add(Utils.instruction(Instructions.cur, param.toLowerCase()));
                            }    
                        }
                        /* VPS TESTING */
                        //instructions.add(Utils.instruction(Instructions.skin, "vps"));
                    }
                    break;    
                   
                case "Sound":
                    param = checkParameter("Type");
                    Defaults.Sounds sound = Sounds.OK;
                    if (param != null) {
                        try {
                            sound = Sounds.valueOf(param);
                        } catch (Exception e) {
                            sound = Sounds.OK;
                        }
                    }
                    instructions.add(Utils.instruction(Instructions.snd, sound.toString().toLowerCase()));
                    
                    break;
                    
                case "Trace":
                    if (!mobile) {
                        instructions.add(Utils.instruction(Instructions.trb));
                        instructions.add(Utils.instruction(Instructions.trti, checkParameter("Time")));
                        instructions.add(Utils.instruction(Instructions.trtl, checkParameter("Title")));
                        instructions.add(Utils.instruction(Instructions.trty, checkParameter("Type")));
                        instructions.add(Utils.instruction(Instructions.trms, checkParameter("Message")));
                        instructions.add(Utils.instruction(Instructions.tre));
                    }
                    
                    break;    
                    
                case "Tracer":
                    if (!mobile) {
                        instructions.add(Utils.instruction(Instructions.trc));
                    }
                    
                    break;     
                    
                case "Editor":
                    param = checkParameter("FormID");
                    if (param != null) {
                        if (checkForm(uiRoot, param)) {
                            param = checkParameter("Field");
                            if (param != null) {
                                if (checkField(contextForm.getContent(), param)) {
                                    contextField.rtfEditor();
                                }
                            }
                        }
                    }

                    break;
                
                case "Capture":
                    param = checkParameter("Mask");
                    if (param == null) {
                        param = "Y";
                    } else {
                        param = Utils.testYString(param) ? "Y" : "N";
                    }
                    setUid(uiRoot.getId());
                    instructions.add(Utils.instruction(Instructions.capt, param));
                        
                    break;    
                    
                case "DownloadFile":
                    param = checkParameter("Base64");
                    if (Utils.testEmptyString(param)) {
                        param = checkParameter("Path");
                        if (param != null && fileDownload == null) {
                            fileDownload = new UiFileDownload();
                            fileDownload.setPath(Utils.string2String(param));
                            fileDownload.setFileName(Utils.string2String(checkParameter("FileName")));
                        }
                    } else if (fileDownload == null) {
                        fileDownload = new UiFileDownload();
                        fileDownload.setBase64(param);
                        fileDownload.setFileName(Utils.string2String(checkParameter("FileName")));
                    }
                    
                    break;	
                    
                case "UploadFile":
                    param = checkParameter("Path");
                    if (param != null && fileUpload == null) {
                        fileUpload = new UiFileUpload();
                        uiRoot.addChild(fileUpload);
                        fileUpload.setPath(Utils.string2String(param));
                        param = checkParameter("FormID");
                        fileUpload.setForm(Utils.string2String(param));
                        fileUpload.setTitle(checkParameter("Title"));
                        fileUpload.setMultiple(checkParameter("Multiple"));
                        if (!fileUpload.isMultiple()) {
                            fileUpload.setStoreName(checkParameter("StoreFileName"));
                        }
                    }
                    
                    break;
                
                case "Progress":
                    param = checkParameter("ProcessID");
                    if (!Utils.testEmptyString(param)) {
                        setUid(uiRoot.getId());
                        instructions.add(Utils.instruction(Instructions.prog, param));
                        param = checkParameter("Title");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.prtl, param));
                        }    
                        param = checkParameter("PercentText");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.prpt, param));
                        }
                        param = checkParameter("WorkText");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.prwt, param));
                        }
                        param = checkParameter("Percent");
                        if (param != null) {
                            instructions.add(Utils.instruction(Instructions.prct, param));
                        }
                        param = checkParameter("Cancelable");
                        if ("Y".equals(param)) {
                            instructions.add(Utils.instruction(Instructions.prca));
                        }    
                    }
                    break;
                    
                case "ProgressClose":
                    param = checkParameter("ProcessID");
                    if (!Utils.testEmptyString(param)) {
                        setUid(uiRoot.getId());
                        instructions.add(Utils.instruction(Instructions.prcl, param));
                    }
                    break;
                    
                case "Script":
                    param = checkParameter("Name");
                    if (!Utils.testEmptyString(param)) {
                        setUid(uiRoot.getId());
                        instructions.add(Utils.instruction(Instructions.js, param));
                    }
                    break;     
                    
            }
         
            localParams.clear();
            commandToProcess = null; 
        }
    }
    
    
    // Приблизительное вычисление ширины строки
    public int proportional(String s) {
        int ret = 0;
        String symbols = "!\"'(),.:;I[\\/]`fijlt|‚„‘’“”¦                                          "
                + "*-J^ry{}­гзктухь                                                      "
                + "#$+0123456789<=>?FLTZ_abcdeghknopqsuvxz~§«ёє»ГЗКТУХабвдеийлнопрсцчъэя "
                + "QERYUOPASGHKXVBNwЁЄАБВДЕИЙЛНОПРСФЦЧЬЭЯжмыю                            "
                + "%CDMm©®МЪфшщ                                                          "
                + "@W…ЖШЩЮЫ№";
        if (!Utils.testEmptyString(s)) {
            for (int i = 0; i < s.length(); i++) {
                ret = ret + (int) ((symbols.indexOf(s.charAt(i)) + 1) / 70.0) + 2;
            }
        }
        return (int) Math.ceil(ret / 4.0);
    }
    
    // Обработка события от клиента
    public final void processEvent(JsonElement msg) throws UiException {
        
        SingleEventParameters sep;
        
        aliveCounter = 0;
        instructions = new ArrayList<>();
        totalEvent = new TotalEvent();
        instructions.add(Utils.instruction(Instructions.s));
        
        JsonElement paramElement = msg.getAsJsonObject().get("par");
        JsonElement eventsElement = paramElement.getAsJsonObject().get("es");
        
        for (JsonElement eventElement : eventsElement.getAsJsonArray()) {
            HashMap<String, String> event = new HashMap<>();
            Iterator<Entry<String, JsonElement>> paramsIter = eventElement.getAsJsonObject().entrySet().iterator();
            while (paramsIter.hasNext()) {
                Entry<String, JsonElement> paramsEntry = paramsIter.next();
                JsonElement eventParamElement = paramsEntry.getValue();
                event.put(paramsEntry.getKey(), eventParamElement.getAsString());
            }
            
            //String evnt = event.get("e");
            UiUnit unit = null;
            if (event.containsKey("uid")) {
                String x = event.get("uid");
                if (units.containsKey(x)) {
                    unit = units.get(x);
                }
            }   
            //try {
                switch (event.get("e")) {
                    // Событие GotoCell
                    case "gtc":
                        if (units.containsKey(event.get("gid"))) {
                            UiGridCell gtc = ((UiGrid) units.get(event.get("gid"))).reqCell(event.get("uid"), event.get("cid"), event.get("row"));
                            if (gtc != null) {
                                gtc.reqCurrent();
                            }
                        }
                        break;
                    // Событие NewValue
                    case "nvl":
                        if (unit != null) {
                            ((UiGridCell) unit).reqNewValue(event.get("val"));
                        }
                        break;
                    // Событие InterValue
                    case "civ":
                        if (unit != null) {
                            ((UiGridCell) unit).reqInterValue(event.get("val"));
                        }
                        break;    
                    // Событие Enter2Fastcom
                    case "e2f":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("Enter2Fastcom"));
                        sep.put("Need2Convert", new StringParameter("Не надо"));
                        if (!Utils.testEmptyString(event.get("params"))) {
                            UiCollection c = new UiCollection(event.get("params"));
                            if (!c.isEmpty()) {
                                CollectionParameter cp = new CollectionParameter();
                                while (!c.isLast()) {
                                    cp.put(c.getCode(), c.getValue());
                                    c.getNext();
                                }
                                sep.put("Parameters", cp);
                            }
                        }
                        totalEvent.add(new SingleEvent(sep));

                        break;
                    // Событие Open
                    case "opn":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("Open"));
                        sep.put("Params", new StringParameter(event.get("params")));
                        sep.put("URL", new StringParameter(event.get("url")));
                        sep.put("UserLang", new StringParameter(event.get("userlang")));
                        sep.put("HostName", new StringParameter(remoteHost));
                        sep.put("ClientID", new StringParameter(token));
                        sep.put("MsgPort", new StringParameter(msgPort));
                        totalEvent.add(new SingleEvent(sep));

                        break;
                    // Событие LMB
                    case "lmb":
                        if (unit != null) {
                            ((UiForm) unit).reqLMB(event.get("code"), event.get("par"));
                        }
                        break;   
                    // Событие LMB
                    case "rmb":
                        if (unit != null) {
                            ((UiForm) unit).reqRMB(event.get("code"), event.get("par"));
                        }
                        break;    
                    // Событие DBLCLK
                    case "dcl":
                        if (unit != null) {
                            ((UiForm) unit).reqDBLCLK(event.get("code"), event.get("par"));
                        }
                        break;    
                    // Событие CloseWindow
                    case "clw":
                        if (unit != null) {
                            ((UiWindow) unit).reqClose();
                        }
                        break;
                    // Событие SystemMenu
                    case "sys":
                        if (unit != null) {
                            ((UiWindow) unit).reqSystemMenu();
                        }
                        break;
                    // Событие ChangeNode
                    case "chn":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqChangeNode();
                        }
                        break;
                    // Событие CheckNode
                    case "ckn":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqCheckNode(event.get("check"));
                        }
                        break;
                    // Событие ChangeState
                    case "chs":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqChangeState();
                        }
                        break;
                    // Событие NodeDC
                    case "ndc":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqDoubleClick();
                        }
                        break;
                    // Событие NodeRMB
                    case "nrb":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqRMB();
                        }
                        break;
                    case "cnb":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqRMBChecked();
                        }
                        break;
                    case "trb":
                        if (unit != null) {
                            ((UiTree) unit).reqRMB();
                        }
                        break;
                    // Событие FetchRows
                    case "fch":
                        if (unit != null) {
                            ((UiForm) unit).reqFetchRows();
                        }
                        break;
                    // Событие CellRMB
                    case "crb":
                        if (units.containsKey(event.get("gid"))) {
                            UiGridCell crb = ((UiGrid) units.get(event.get("gid"))).reqCell(event.get("uid"), event.get("cid"), event.get("row"));
                            if (crb != null) {
                                crb.reqRMB();
                            }
                        }
                        break;
                    case "grb":
                        if (unit != null) {
                            ((UiGrid) unit).reqRMB();
                        }
                        break;
                    // Событие CellDC
                    case "cdc":
                        if (unit != null) {
                            ((UiGridCell) unit).reqDC();
                        }
                        break;
                    // Событие ColumnLMB
                    case "clb":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqLMB();
                        }
                        break;
                    // Событие ColumnDC
                    case "clu":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqDC();
                        }
                        break;    
                    // Событие ColumnRMB
                    case "crc":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqRMB();
                        }
                        break;    
                    // Событие GroupLMB
                    case "glb":
                        if (unit != null) {
                            ((UiGroup) unit).reqLMB();
                        }
                        break;
                    // Событие GroupDC
                    case "gdc":
                        if (unit != null) {
                            ((UiGroup) unit).reqDC();
                        }
                        break;    
                    // Событие GroupRMB
                    case "grc":
                        if (unit != null) {
                            ((UiGroup) unit).reqRMB();
                        }
                        break;      
                    case "hlb":
                        if (unit != null) {
                            ((UiGrid) unit).reqLMB();
                        }
                        break;
                    // Событие CheckRow
                    case "chk":
                        if (unit != null) {
                            ((UiGridRow) unit).reqCheck(event.get("check"));
                        }
                        break;
                    // Событие ChangeRow
                    case "chr":
                        if (unit != null) {
                            ((UiGridRow) unit).reqChange();
                        }
                        break;
                    // Событие ChangeMode
                    case "mod":
                        if (unit != null) {
                            ((UiForm) unit).reqChangeMode();
                        }
                        break;
                    // Событие CheckAll
                    case "cal":
                        if (unit != null) {
                            ((UiForm) unit).reqCheckAll();
                        }
                        break;
                    // Событие UnCheckAll
                    case "ual":
                        if (unit != null) {
                            ((UiForm) unit).reqUnCheckAll();
                        }
                        break;
                    // Событие InvertChecking
                    case "inv":
                        if (unit != null) {
                            ((UiForm) unit).reqInvertChecking();
                        }
                        break;
                    // Событие ToolBarButton
                    case "tbb":
                        if (unit != null) {
                            ((UiToolButton) unit).reqPress();
                        }
                        break;
                    // Событие CellButton
                    case "cbp":
                        if (unit != null) {
                            ((UiCellButton) unit).reqPress();
                        }
                        break;
                    // Событие ChangeConfig
                    case "cfg":
                        if (unit != null) {
                            ((UiForm) unit).reqChangeConfig(event.get("cfgid"));
                        }
                        break;
                    // Событие GotoField
                    case "gtf":
                        if (unit != null) {
                            ((UiField) unit).reqCurrent(event.get("car"));
                        }
                        break;
                    // Событие GotoButton
                    case "gtb":
                        if (unit != null) {
                            ((UiButton) unit).reqCurrent();
                        }
                        break;
                    // Событие NewValue
                    case "nwv":
                        if (unit != null) {
                            ((UiField) unit).reqNewValue(event.get("val"), event.get("start"), event.get("end"));
                        }
                        break;
                    // Событие NewValue - промежуточное
                    case "nwi":
                        if (unit != null) {
                            ((UiField) unit).reqNewInterValue(event.get("val"), event.get("start"), event.get("end"));
                        }
                        break;
                    // Событие NewValue - для радиогруппы
                    case "rvl":
                        if (unit != null) {
                            ((UiField) unit).reqRadioValue(event.get("val"));
                        }
                        break;    
                    // Событие NewValue при изменении только позиции курсора
                    case "car":
                        if (unit != null) {
                            ((UiField) unit).reqCaretPos(event.get("start"), event.get("end"));
                        }
                        break;
                    // Событие NewValue для слайдера
                    case "rfr":
                        if (unit != null) {
                            ((UiField) unit).reqRangeFrom(event.get("par"));
                        }
                        break;    
                    // Событие PatternMatch
                    case "mch":
                        if (unit != null) {
                            ((UiField) unit).reqPatternMatch(event.get("val"), event.get("start"), event.get("end"));
                        }
                        break;
                    // Событие SetUserProperty - установка ширины столбца
                    case "cpw":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqPropertyWidth(event.get("width"));
                        }
                        break;
                    // Событие SetUserProperty - перемещение столбца
                    case "cmv":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqMove(event.get("left"));
                        }
                        break;
                    // Событие SetUserProperty - перемещение столбца в другую группу
                    case "cmg":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqGroup(event.get("grp"));
                        }
                        break;
                    // Событие GetTotal - запрос итоговой строки
                    case "tot":
                        if (unit != null) {
                            ((UiGrid) unit).reqSummary();
                        }
                        break;
                    // Событие ButtonPressed
                    case "bpr":
                        if (unit != null) {
                            ((UiButton) unit).reqPress();
                        }
                        break;
                    // Событие FieldButton
                    case "fib":
                        if (unit != null) {
                            ((UiFieldButton) unit).reqPress();
                        }
                        break;
                    // Событие FieldRMB
                    case "frb":
                        if (unit != null) {
                            ((UiField) unit).reqRMB();
                        }
                        break;
                    // Событие DropDown
                    case "drd":
                        if (unit != null) {
                            ((UiField) unit).reqDropDown(event.get("key"));
                        }
                        break;
                    // Событие ColorPick
                    case "cpk":
                        if (unit != null) {
                            ((UiField) unit).reqColorPick(event.get("par"));
                        }
                        break;    
                    // Событие SetUserProperty - изменение размеров окна
                    case "wsz":
                        if (unit != null) {
                            ((UiWindow) unit).reqResize(event.get("w"), event.get("h"));
                        }
                        break;
                    // Событие SetUserProperty - изменение размеров подвала
                    case "bsz":
                        if (unit != null) {
                            ((UiBasement) unit).reqResize(event.get("h"));
                        }
                        break;
                    // Событие SetUserProperty - изменение ширины столбца
                    case "csz":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqPropertyWidth(event.get("w"));
                        }
                        break;
                    // Событие SetUserProperty - изменение размеров overflow
                    case "osz":
                        if (unit != null) {
                            ((UiForm) unit).reqResizeOverflow(event.get("w"));
                        }
                        break;
                    // Событие SetUserProperty - изменение высоты нижних дочерних форм
                    case "scs":
                        if (unit != null) {
                            ((UiSouthContainer) unit).reqResize(event.get("h"));
                        }
                        break;
                    // Событие SetUserProperty - изменение ширины правых дочерних форм
                    case "ecs":
                        if (unit != null) {
                            ((UiEastContainer) unit).reqResize(event.get("w"));
                        }
                        break;
                    // Событие SetUserProperty - изменение ширины формы, являющейся деталью
                    case "fws":
                        if (unit != null) {
                            ((UiForm) unit).reqResizeWidth(event.get("w"));
                        }
                        break;
                    // Событие SetUserProperty - изменение высоты формы, являющейся деталью
                    case "fhs":
                        if (unit != null) {
                            ((UiForm) unit).reqResizeHeight(event.get("h"));
                        }
                        break;
                    // Событие CancelLOV - закрытие списка значений
                    case "lcl":
                        if (lov != null) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("CancelLOV"));
                            totalEvent.add(new SingleEvent(sep));
                            lov.remove();
                            lov = null;
                        }
                        break;
                    // Событие ChoiceLOV - выбор строки в LOV
                    case "lch":
                        if (!(lov == null || unit == null)) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("ChoiceLOV"));
                            sep.put("ID", new StringParameter(unit.getCode()));
                            totalEvent.add(new SingleEvent(sep));
                            lov.remove();
                            lov = null;
                        }
                        break;

                    // Событие NextPartLOV - запрос очередной порции строк для LOV
                    case "npl":
                        if (lov != null) {
                            lov.reqFetch();
                        }
                        break;
                    // Событие SetSortLOV - установка сортировки столбца LOV
                    case "ssl":
                        if (lov != null) {
                            lov.setSortColumn(event.get("uid"));
                        }
                        break;
                    // Событие ExpandLOV - раскрытие строки LOV
                    case "exp":
                        if (lov != null) {
                            lov.reqExpand(event.get("uid"));
                        }
                        break;
                    // Схлопывание (collapse) строки LOV
                    case "clp":
                        if (lov != null) {
                            lov.reqCollapse(event.get("uid"));
                        }
                        break;
                    // Событие LivePickLOV - фильтрация строк в LOV
                    case "lpl":
                        if (lov != null) {
                            for (String s : event.get("cols").split(";")) {
                                lov.setFilter(s, event.get(s));
                            }
                            lov.applyFilter();
                        }
                        break;
                    // Событие CancelLOV - закрытие окна с вопросом
                    case "qcl":
                        if (unit != null) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("CancelLOV"));
                            totalEvent.add(new SingleEvent(sep));
                            unit.remove();
                        }
                        break;
                    // Событие ChoiceLOV - в окне с вопросом
                    case "qch":
                        if (unit != null) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("ChoiceLOV"));
                            sep.put("ID", new StringParameter(unit.getCode()));
                            totalEvent.add(new SingleEvent(sep));
                            unit.getParent().remove();
                        }
                        break;
                    // Image event
                    case "img":
                        if (unit != null) {
                            ((UiChart) unit).reqImage(event.get("data"));
                        }
                        break;
                    // Событие закрытие с системным сообщением
                    case "mcl":
                        if (unit != null) {
                            unit.remove();
                        }
                        break;
                    // Запрос на добавление строки в список
                    case "rwa":
                        if (unit != null) {
                            ((UiGrid) unit).reqAddRow();
                        }
                        break;
                    case "rwn":
                        if (unit != null) {
                            ((UiGrid) unit).reqNewRow();
                        }
                        break;    
                    // Событие ChoiceMenuItem
                    case "mnu":
                        UiLocalMenuItem mi = localMenuItems.get(event.get("uid"));
                        if (mi != null) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("ChoiceMenuItem"));
                            sep.put("FormID", new StringParameter(mi.getFormId()));
                            sep.put("Code", new StringParameter(mi.getCode()));
                            sep.put("Type", new StringParameter(mi.getType()));
                            sep.put("Params", new StringParameter(mi.getParams()));
                            totalEvent.add(new SingleEvent(sep));
                        }
                        break;
                    // Событие EscapeBar
                    case "esc":
                        if (unit != null) {
                            ((UiForm) unit).reqEscape();
                        }
                        break;
                    // Событие F1Bar
                    case "f1":
                        if (unit != null) {
                            ((UiForm) unit).reqF1();
                        }
                        break;
                    // Событие F4Bar
                    case "f4":
                        if (unit != null) {
                            ((UiForm) unit).reqF4();
                        }
                        break;
                    // Событие F6Bar
                    case "f6":
                        if (unit != null) {
                            ((UiForm) unit).reqF6();
                        }
                        break;
                    // Событие ShiftF6Bar
                    case "sf6":
                        if (unit != null) {
                            ((UiForm) unit).reqShiftF6();
                        }
                        break;
                    // Событие F7Bar
                    case "f7":
                        if (unit != null) {
                            ((UiForm) unit).reqF7();
                        }
                        break;
                    // Событие F8Bar
                    case "f8":
                        if (unit != null) {
                            ((UiForm) unit).reqF8();
                        }
                        break;
                    // Событие F5Bar
                    case "f5":
                        if (unit != null) {
                            ((UiForm) unit).reqF5();
                        }
                        break;
                    // Событие F10Bar
                    case "f10":
                        if (unit != null) {
                            ((UiForm) unit).reqF10();
                        }
                        break;
                    // Событие SpaceBar в дереве
                    case "tsb":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqSpace();
                        }
                        break;
                    // Событие SpaceBar в списке
                    case "gsb":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqSpace();
                        }
                        break;
                    // Событие SpaceBar в бланке
                    case "bsb":
                        if (unit != null) {
                            ((UiField) unit).reqSpace();
                        }
                        break;
                    // Событие EnterBar в дереве
                    case "teb":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqEnter();
                        }
                        break;
                    // Событие EnterBar в списке
                    case "geb":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqEnter();
                        }
                        break;
                    // Событие EnterBar в бланке
                    case "beb":
                        if (unit != null) {
                            ((UiField) unit).reqEnter();
                        }
                        break;
                    // Событие F9Bar в бланке
                    case "bf9":
                        if (unit != null) {
                            ((UiField) unit).reqF9();
                        }
                        break;
                    // Событие F9Bar в списке
                    case "gf9":
                        if (unit != null) {
                            ((UiGridColumn) unit).reqF9();
                        }
                        break;
                    // Событие InsertBar в дереве
                    case "tib":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqInsert();
                        }
                        break;
                    // Событие DeleteBar в дереве
                    case "tdb":
                        if (unit != null) {
                            ((UiTreeNode) unit).reqDelete();
                        }
                        break;
                    // Событие InsertBar в списке
                    case "gib":
                        if (unit != null) {
                            ((UiForm) unit).reqInsert();
                        }
                        break;
                    // Событие DeleteBar в списке
                    case "gdb":
                        if (unit != null) {
                            ((UiForm) unit).reqDelete();
                        }
                        break;
                    // Событие CheckCell в матрице
                    case "mas":
                        if (unit != null) {
                            ((UiMatrix) unit).reqSelect(event.get("rows"), event.get("cols"), event.get("cells"));
                        }
                        break;    
                    // Событие ChangeCell в матрице
                    case "mac":
                        if (unit != null) {
                            ((UiMatrixCell) unit).reqChange();
                        }
                        break;    
                    // Событие CellMenu в матрице
                    case "mcm":
                        if (unit != null) {
                            ((UiMatrixCell) unit).reqMenu();
                        }
                        break;  
                    // Событие MatrixRMB
                    case "mfm":
                        if (unit != null) {
                            ((UiMatrix) unit).reqBlankMenu(event.get("row"), event.get("col"));
                        }
                        break; 
                    case "mfc":
                        if (unit != null) {
                            ((UiMatrix) unit).reqBlankClick(event.get("row"), event.get("col"));
                        }
                        break; 
                    case "mfd":
                        if (unit != null) {
                            ((UiMatrix) unit).reqBlankDC(event.get("row"), event.get("col"));
                        }
                        break;    
                    // Событие CellDC в матрице
                    case "mad":
                        if (unit != null) {
                            ((UiMatrixCell) unit).reqDC();
                        }
                        break;   
                    // Событие ColMenu в матрице
                    case "msm":
                        if (unit != null) {
                            ((UiMatrixCellset) unit).reqMenu();
                        }
                        break; 
                    // Событие Cellset CLick в матрице
                    case "msc":
                        if (unit != null) {
                            ((UiMatrixCellset) unit).reqClick();
                        }
                        break;   
                    // Событие Cellset DblCLick в матрице
                    case "msd":
                        if (unit != null) {
                            ((UiMatrixCellset) unit).reqDblClick();
                        }
                        break;    
                    // Событие MatrixRMB в матрице
                    case "mrm":
                        if (unit != null) {
                            ((UiMatrix) unit).reqMenu();
                        }
                        break;   
                    case "mri":
                        if (unit != null) {
                            ((UiMatrix) unit).reqClick();
                        }
                        break;    
                    case "mrd":
                        if (unit != null) {
                            ((UiMatrix) unit).reqDblClick();
                        }
                        break;    
                    case "mrc":
                        if (unit != null) {
                            ((UiMatrix) unit).reqMenuRC();
                        }
                        break;   
                    case "rcc":
                        if (unit != null) {
                            ((UiMatrix) unit).reqClickRC();
                        }
                        break;
                    case "rcd":
                        if (unit != null) {
                            ((UiMatrix) unit).reqDCRC();
                        }
                        break;    
                    // Событие CellButton в матрице
                    case "mcb":
                        if (unit != null) {
                            ((UiMatrixButton) unit).reqPress();
                        }
                        break;
                    // Событие NewValue в матрице
                    case "mvl":
                        if (unit != null) {
                            ((UiMatrixCell) unit).reqNewValue(event.get("val"));
                        }
                        break;
                    // Событие InterValue в матрице
                    case "miv":
                        if (unit != null) {
                            ((UiMatrixCell) unit).reqInterValue(event.get("val"));
                        }
                        break;
                    // Событие SetUserProperty - установка ширины столбца матрицы
                    case "mcw":
                        if (unit != null) {
                            ((UiMatrixColumn) unit).reqPropertyWidth(event.get("width"));
                        }
                        break; 
                    // Событие SetUserProperty - установка высоты строки матрицы
                    case "mrh":
                        if (unit != null) {
                            ((UiMatrixRow) unit).reqPropertyHeight(event.get("height"));
                        }
                        break;   
                    // Событие SetUserProperty - установка ширины угловой ячейки матрицы
                    case "maw":
                        if (unit != null) {
                            ((UiMatrix) unit).reqAngularWidth(event.get("width"));
                        }
                        break;     
                    // Событие ChartOver
                    case "cho":
                        if (unit != null) {
                            ((UiChart) unit).reqOver(event.get("par"));
                        }
                        break;  
                    // Событие ChartOut
                    case "chu":
                        if (unit != null) {
                            ((UiChart) unit).reqOut(event.get("par"));
                        }
                        break;    
                    // Событие ChartClick
                    case "chc":
                        if (unit != null) {
                            ((UiChart) unit).reqClick(event.get("par"));
                        }
                        break; 
                    // Событие ChartRMB
                    case "chm":
                        if (unit != null) {
                            ((UiChart) unit).reqRMB(event.get("par"));
                        }
                        break;    
                    // Событие ChartDC
                    case "chd":
                        if (unit != null) {
                            ((UiChart) unit).reqDC(event.get("par"));
                        }
                        break;
                    // Событие ChartGraphicClick
                    case "cgc":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("ChartGraphicClick"));
                        sep.put("FormID", new StringParameter(event.get("fid")));
                        sep.put("ElementID", new StringParameter(event.get("eid")));
                        totalEvent.add(new SingleEvent(sep)); 
                        break;
                    // Событие ChartLegendClick
                    case "clc":
                        if (unit != null) {
                            ((UiChart) unit).reqLegendClick(event.get("par"));
                        } 
                        break;
                    // Событие ChartLegendDC
                    case "cll":
                        if (unit != null) {
                            ((UiChart) unit).reqLegendDC(event.get("par"));
                        } 
                        break;
                    // Событие ChartLegendRMB
                    case "clr":
                        if (unit != null) {
                            ((UiChart) unit).reqLegendRMB(event.get("par"));
                        } 
                        break; 
                    // Событие ChartMapPosition
                    case "mps":
                        if (unit != null) {
                            ((UiChart) unit).reqMapPosition(event.get("par"));
                        } 
                        break;    
                    // Событие ChartResize
                    case "crz":
                        if (unit != null) {
                            ((UiChart) unit).reqResize(event.get("w"), event.get("h"));
                        } 
                        break;     
                    // Событие ColormapClick
                    case "mck":
                        if (unit != null) {
                            ((UiChart) unit).reqColormapClick(event.get("par"));
                        } 
                        break;     
                    // Событие ColormapDC
                    case "mdb":
                        if (unit != null) {
                            ((UiChart) unit).reqColormapDC(event.get("par"));
                        } 
                        break;
                    // Событие ColormapRMB
                    case "mrb":
                        if (unit != null) {
                            ((UiChart) unit).reqColormapRMB(event.get("par"));
                        } 
                        break;   
                    // Событие PiecewiseClick
                    case "pwc":
                        if (unit != null) {
                            ((UiChart) unit).reqPiecewiseClick(event.get("par"));
                        } 
                        break;     
                    // Событие PiecewiseDC
                    case "pwd":
                        if (unit != null) {
                            ((UiChart) unit).reqPiecewiseDC(event.get("par"));
                        } 
                        break;
                    // Событие ColormapRMB
                    case "pwr":
                        if (unit != null) {
                            ((UiChart) unit).reqPiecewiseRMB(event.get("par"));
                        } 
                        break;     
                    // Событие ChartZoomClick
                    case "czc":
                        if (unit != null) {
                            ((UiChart) unit).reqZoomClick(event.get("par"));
                        } 
                        break;   
                    // Событие ChartZoomDC
                    case "czd":
                        if (unit != null) {
                            ((UiChart) unit).reqZoomDC(event.get("par"));
                        } 
                        break;    
                    // Событие ChartAxisClick
                    case "cac":
                        if (unit != null) {
                            ((UiChart) unit).reqAxisClick(event.get("par"));
                        } 
                        break; 
                    // Событие ChartZoom
                    case "czm":
                        if (unit != null) {
                            ((UiChart) unit).reqZoom(event.get("par"));
                        } 
                        break; 
                    // Событие Undo
                    case "und":
                        if (unit != null) {
                            ((UiForm) unit).reqUndo();
                        }
                        break;    
                    // Событие Redo
                    case "red":
                        if (unit != null) {
                            ((UiForm) unit).reqRedo();
                        }
                        break;    
                    // Событие Activate
                    case "act":
                        if (unit != null) {
                            ((UiWindow) unit).reqActivate();
                        }
                        break;  
                    // Событие "Максимизация окна"
                    case "wmx":
                        if (unit != null) {
                            ((UiWindow) unit).reqMaximize();
                        }
                        break; 
                    // Событие "Положение окна"
                    case "wpo":
                        if (unit != null) {
                            ((UiWindow) unit).reqPosition(event.get("left"), event.get("top"));
                        }
                        break;    
                    // Событие "Восстановление размеров окна"
                    case "wrs":
                        if (unit != null) {
                            ((UiWindow) unit).reqRestore();
                        }
                        break;      
                    // Закрытие календаря
                    case "crm":
                        if (unit != null) {
                            ((UiCalendar) unit).reqClose();
                        }
                        break;
                    // Событие Calendar
                    case "cld":
                        if (unit != null) {
                            ((UiCalendar) unit).reqChoice(event.get("day"),
                                    event.get("mon"),
                                    event.get("year"),
                                    event.get("hour"),
                                    event.get("min"),
                                    event.get("sec"));
                        }
                        break;
                    // Событие FormActivate
                    case "afr":
                        if (unit != null) {
                            ((UiForm) unit).reqActivate();
                        }
                        break;    
                    // Событие Event
                    case "evt":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter(event.get("uid")));
                        event.remove("e");
                        event.remove("uid");
                        for (Map.Entry<String, String> entry : event.entrySet()) {
                            sep.put(entry.getKey(), new StringParameter(Utils.string2String(entry.getValue())));
                        }
                        totalEvent.add(new SingleEvent(sep));
                        uiRoot.emulatedEvent();
                        break;
                    // Событие Action    
                    case "acn":
                        String formCode = event.get("frm");
                        if (checkForm(uiRoot, formCode)) {
                            contextForm.reqAction(event.get("act"), event.get("par"));
                        }
                    // Событие UploadOk
                    case "uok":
                        if (fileUpload != null) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("UploadOk"));
                            sep.put("FormID", new StringParameter(fileUpload.getForm()));
                            sep.put("Files", new StringParameter(fileUpload.getFiles()));
                            totalEvent.add(new SingleEvent(sep));
                            fileUpload.remove();
                            fileUpload = null;
                        }
                        break;
                    // Событие UploadCancel
                    case "cup":
                        if (fileUpload != null) {
                            String evt = event.get("event");
                            if (!Utils.testEmptyString(evt)) {
                                sep = new SingleEventParameters();
                                sep.put("e", new StringParameter("UploadCancel"));
                                sep.put("FormID", new StringParameter(fileUpload.getForm()));
                                totalEvent.add(new SingleEvent(sep));
                            }
                            fileUpload.remove();
                            fileUpload = null;
                        }

                        break;
                    // Событие GetTraceInfo
                    case "tra":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("GetTraceInfo"));
                        sep.put("State", new StringParameter("All"));
                        totalEvent.add(new SingleEvent(sep));

                        break;

                    // Событие ProgressCancel
                    case "prc":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("ProgressCancel"));
                        sep.put("ProcessID", new StringParameter(event.get("uid")));
                        totalEvent.add(new SingleEvent(sep));

                        break;

                    // Событие Verify
                    case "vrf":
                        if (unit != null) {
                            sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("Verify"));
                            sep.put("Field", new StringParameter(((UiField) unit).getCode()));
                            sep.put("FormID", new StringParameter(((UiField) unit).getForm().getCode()));
                            sep.put("ExitCode", new StringParameter(event.get("exit")));
                            sep.put("ExitMessage", new StringParameter(event.get("msg")));
                            sep.put("Value", new StringParameter(event.get("val")));
                            totalEvent.add(new SingleEvent(sep));
                        }
                        
                        break;
                        
                    // Событие screenshot    
                    case "cpt":
                        sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("Capture"));
                        sep.put("Data", new StringParameter(event.get("data")));
                        sep.put("Height", new StringParameter(event.get("h")));
                        sep.put("Width", new StringParameter(event.get("w")));
                        totalEvent.add(new SingleEvent(sep));
                        
                        break;    
                        
                    // Начать запись событий
                    case "rec":
                        eventsLogging = true;
                        eventsLog = new ArrayList<>();

                        break;

                    // Завершить запись событий
                    case "_rec":                        
                        try {
                            eventsLogFile = Paths.get(tempDir, "luievents.log");
                            eventsLogWriter = new OutputStreamWriter(new FileOutputStream(eventsLogFile.toFile()), StandardCharsets.UTF_8);
                            for (int i = 0; i < eventsLog.size(); i++) {
                                eventsLogWriter.write(eventsLog.get(i) + "\n");
                            }
                            eventsLogWriter.flush();
                            eventsLogWriter.close();
                        } catch (Exception ex) {}
                        eventsLogging = false;
                        
                        break;    
                    
                    // Установить преобразование для главного меню
                    case "csm":                        
                        customMenu = true;
                        
                        break;    
                    
                    default:
                }
            //} catch (Exception ex) {
            //    LOGGER.log(Level.SEVERE, "--> Error while processing {0} event> \n{1}", new Object[]{evnt, ex.getMessage()});
            //    UiError.raise("Server error> \n" + ex.getMessage() + "\n See log for details");
            //}
        }

        sendEvent();
        
        unitToFocus = formToFocus == null ? unitToFocus : formToFocus;
        if (unitToFocus != null && units.containsKey(unitToFocus)) {
            setUid(unitToFocus);
            instructions.add(Utils.instruction(Instructions.fcs));
        }
        formToFocus = null;
        unitToFocus = null;
        
        if (lov != null) {
            if (lov.getState() == 1) {
                lov.initialize();
            }
            lov.reset();
            if ((lov.getState() == 1 || lov.getState() == 2) && (!mobile)) {
                setUid(lov.getId());
                instructions.add(Utils.instruction(Instructions.curr));
            }
            lov.setState(0);
        }
        
        if (!unitsToCheck.isEmpty()) {
            Iterator<UiUnit> iter = unitsToCheck.iterator();
            while (iter.hasNext()) {
                UiUnit u = iter.next();
                iter.remove();
                if (units.containsValue(u)) {
                    u.preCheck();
                    instructions.addAll(u.getInstructions());
                    u.postCheck();
                }
            }
        }
        
        setUid("dummy");
        if (showMenu) {
            showMenu = false;
            menuHTML = new StringBuilder();
            buildMenu("@");
            instructions.add(Utils.instruction(Instructions.menu, menuHTML.toString()));
        }
        instructions.add(Utils.instruction(Instructions.f));
    }

    // Посылка события в движок
    public final void sendEvent() throws UiException {
        if (!totalEvent.isEmpty()) {
            Document event = totalEvent.xml();
            StringBuilder sb = new StringBuilder();
            if (!mobile) instructions.add(Utils.instruction(Instructions.le, prettyOutputter.outputString(event)));
            try {
                //Date dateStart = new Date();
                //Date dateFinish = new Date();;
                //stopMonitoring = true;
                if (protocol == 1) {
                    //dateStart = new Date();
                    eventStatement = connection.prepareCall("begin ? := ui_p_xml.events(?); end;");
                    eventStatement.registerOutParameter(1, Types.ARRAY, "XMLVARRAY");
                    String e = rawOutputter.outputString(event);
                    //System.out.println(">> EVENT > "+prettyOutputter.outputString(event));
                    eventStatement.setCharacterStream(2, new StringReader(e), e.length());
                    eventStatement.execute();
                    Array a = eventStatement.getArray(1);
                    String[] s = (String[]) a.getArray();
                    for (String item : s) {
                        sb.append(item);
                    }
                    //dateFinish = new Date();
                } else if (protocol == 2 || protocol == 3) {
                    //System.out.println("EVENT > "+prettyOutputter.outputString(event));
                    String e = rawOutputter.outputString(event).replaceAll("\n", "&#10;").replaceAll("\r", "&#13;");
                    //Date dateStart = new Date();
                    sockWriter.write(e);
                    if (eventsLogging) {
                        eventsLog.add(e);
                    }
                    sockWriter.write("\n");
                    sockWriter.flush();
                    String line;
                    while ((line = sockReader.readLine()) != null) {
			if (line.equals("")) {
                            break;
                        } else {
                            sb.append(line.replaceAll(Defaults.LF_REPLACE, "\n"));
                        } 
                    }
                    //Date dateFinish = new Date();
                    //System.out.println(">>> socket write - read time> "+(dateFinish.getTime() - dateStart.getTime()));
                }
                //System.out.println(">> COMMAND > "+sb.toString());
                //Date dateStart = new Date();
                processCommands(sb.toString());
                //Date dateFinish = new Date();
                //System.out.println(">>> commands processing time> "+(dateFinish.getTime() - dateStart.getTime()));
            } catch (SQLException | JDOMException | IOException ex) {
                //UiError.raise(sb.toString() + "\n" + ex.getMessage());
                //System.out.println("ERROR\n "+sb.toString());
                //ex.printStackTrace();
                UiError.raise(ex.getMessage());
            } finally {
                stopMonitoring = false;
                if (protocol == 1 & eventStatement != null) {
                    try {
                        eventStatement.close();
                    } catch (SQLException ex) {
                    }
                    eventStatement = null;
                }
            }
        }
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiGridChild - "дочерний элемент списка"                    *
 *                                                                      *
 ************************************************************************/
    
    public class UiGridChild extends UiFormChild {
        
        public UiGridChild() {
            super();
        }
        
        public UiGridChild(String code) {
            super(code);
        }
                
        // Получить родительский список
        public UiGrid getGrid() {
            return new Object() {
                UiGrid find(UiUnit u) {
                    if (u == null) {
                        return null;
                    } else if (u instanceof UiGrid) {
                        return (UiGrid) u;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this.getParent());
        }
        
    }        

    
/************************************************************************
 *                                                                      *
 *     Класс UiBlankChild - "дочерний элемент бланка"                   *
 *                                                                      *
 ************************************************************************/
    
    public class UiBlankChild extends UiFormChild {
        
        public UiBlankChild() {
            super();
        }
        
        public UiBlankChild(String code) {
            super(code);
        }
        
        // Получить родительский бланк
        public UiBlank getBlank() {
            return new Object() {
                UiBlank find(UiUnit u) {
                    if (u == null) {
                        return null;
                    } else if (u instanceof UiBlank) {
                        return (UiBlank) u;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this.getParent());
        }
        
    }        
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBasement - "подвал"                                      *
 *                                                                      *
 ************************************************************************/    
    
    public class UiBasement extends UiFormChild {
        
        String content = "";
        String height = "Auto";
        
        public UiBasement() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.bsmt;
        }
        
        public void setContent(String content) {
            if (!(mobile || protocol == 3)) {
                setUid(getId());
                this.content = content;
                instructions.add(Utils.instruction(Instructions.cont, content));
            }
        }
        
        public void setHeight(String height) {
            if (!(mobile || protocol == 3)) {
                setUid(getId());
                if (height.equals("Auto")) {
                    this.height = height;
                    instructions.add(Utils.instruction(Instructions.ha));
                } else {
                    this.height = height;
                    if ("%".equals(height.substring(0, 1))) {
                        instructions.add(Utils.instruction(Instructions.h$, height.substring(1)));
                    } else {
                        instructions.add(Utils.instruction(Instructions.h, height));
                    }
                }
            }
        }
        
        public void reqResize(String size) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Property", new StringParameter("Underside"));
            sep.put("Value", new StringParameter(size));
            totalEvent.add(new SingleEvent(sep));
        };
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiUnit - базовая единица интерфейса                        *
 *                                                                      *
 ************************************************************************/
    public class UiUnit {
        
        private final String id;
        private String code = "";
        private UiUnit parent = null;
        final ArrayList<UiUnit> children;
        private Object data = null;
        private boolean hidden = false;
        private boolean render = true;
        
        public UiUnit() {
            id = Utils.generateId();
            children = new ArrayList<>();
        }
        
        public UiUnit(String code) {
            id = Utils.generateId();
            this.code = code;
            children = new ArrayList<>();
        }
        
        public UiUnit(boolean render) {
            id = Utils.generateId();
            children = new ArrayList<>();
            this.render = render;
        }
        
        public UiUnit(String code, boolean render) {
            id = Utils.generateId();
            this.code = code;
            children = new ArrayList<>();
            this.render = render;
        }
        
        public final String getId() {
            return id;
        }
        
        public final String getCode() {
            return code;
        }
        
        public void setData(Object data) {
            this.data = data;
        }
        
        public Object getData() {
            return data;
        }
        
        public final void setParent(UiUnit parent) {
            this.parent = parent;
        }
        
        public final UiUnit getParent() {
            return parent;
        }
        
        public Instructions getCreateInstruction() {
            return Instructions.c;
        }
               
        public void addChild(UiUnit child) {
            if (child.getParent() == null) {
                if (!render) child.render = false;
                if (child.render) {
                    setUid(this.getId());
                    instructions.add(Utils.instruction(child.getCreateInstruction(), child.getId()));
                    if (!Utils.testEmptyString(child.getCode())) {
                        setUid(child.getId());
                        instructions.add(Utils.instruction(Instructions.code, child.getCode()));
                    }
                }    
                child.setParent(this);
                children.add(child);
                units.put(child.getId(), child);
            } else if (this != child.getParent()) {
                if (render == child.render) {
                    setUid(this.getId());
                    instructions.add(Utils.instruction(child.getCreateInstruction(), child.getId()));
                    child.getParent().children.remove(child);
                    child.setParent(this);
                    children.add(child);
                } 
            }
        }
        
        public void changeParent(UiUnit parent) {
            if (parent != null) {
                if (getParent() == null) {
                    parent.addChild(this);
                } else if (parent != getParent()) {
                    getParent().children.remove(this);
                    setParent(parent);
                    parent.children.add(this);
                }
            }
        }
        
        public void remove() {
            this.getParent().rmChild(this);
        }
        
        public void rmChild(UiUnit child) {
            if (children.contains(child)) {
                child.preRemove();
                unitsToCheck.remove(child);
                child.rmChildren();
                //setUid(uiRoot.getId());
                //if (child.render) instructions.add(Utils.instruction(Instructions.rm, child.getId()));
                instructions.add(Utils.instruction(Instructions.rm, child.getId()));
                children.remove(child);
                units.remove(child.getId());
            }
        }
        
        public void rmChildren() {
            //setUid(uiRoot.getId());
            for (UiUnit u : getAllChildren()) {
                u.preRemove();
                unitsToCheck.remove(u);
                //if (u.render) instructions.add(Utils.instruction(Instructions.rm, u.getId()));
                instructions.add(Utils.instruction(Instructions.rm, u.getId()));
                units.remove(u.getId());
            }
            children.clear();
        }
        
        public void show() {
            hidden = false;
        }
        
        public void hide() {
            hidden = true;
        }
        
        public boolean isHidden() {
            return hidden;
        }
        
        public void undo() {
            
        }
        
        public void redo() {
            
        }
        
        public void preRemove() {
            
        }
        
        public void preCheck() {
            
        }
        
        public void postCheck() {
            
        }
        
        public ArrayList<JsonObject> getInstructions() {
            return new ArrayList<>();
        }
        
        public ArrayList<UiUnit> getAllChildren() {
            ArrayList<UiUnit> allChildren = new ArrayList<>();
            getAllChildren_(allChildren);
            return allChildren;
        }
        
        private void getAllChildren_(ArrayList<UiUnit> a) {
            for (UiUnit u : children) {
                a.add(u);
                u.getAllChildren_(a);
            }
        }

        public ArrayList<UiUnit> getChildrenByClass(Class childClass) {
            ArrayList<UiUnit> ret = new ArrayList<>();
            for (UiUnit u : children) {
                if (childClass.isInstance(u)) {
                    ret.add(u);
                }
            }
            return ret;
        }
        
        public ArrayList<UiUnit> getVisibleChildrenByClass(Class childClass) {
            ArrayList<UiUnit> ret = new ArrayList<>();
            for (UiUnit u : children) {
                if ((!u.isHidden()) && childClass.isInstance(u)) {
                    ret.add(u);
                }
            }
            return ret;
        }

        public ArrayList<UiUnit> getAllChildrenByClass(Class childClass) {
            ArrayList<UiUnit> allChildren = new ArrayList<>();
            getAllChildrenByClass_(allChildren, childClass);
            return allChildren;
        }
        
        private void getAllChildrenByClass_(ArrayList<UiUnit> a, Class childClass) {
            for (UiUnit u : children) {
                if (childClass.isInstance(u)) a.add(u);
                u.getAllChildrenByClass_(a, childClass);
            }
        }
        
        public ArrayList<UiUnit> getAllVisibleChildrenByClass(Class childClass) {
            ArrayList<UiUnit> allChildren = new ArrayList<>();
            getAllVisibleChildrenByClass_(allChildren, childClass);
            return allChildren;
        }
        
        private void getAllVisibleChildrenByClass_(ArrayList<UiUnit> a, Class childClass) {
            for (UiUnit u : children) {
                if ((!u.isHidden()) && childClass.isInstance(u)) a.add(u);
                u.getAllVisibleChildrenByClass_(a, childClass);
            }
        }
        
        public UiUnit getChildByClass(Class childClass) {
            for (UiUnit u : children) {
                if (childClass.isInstance(u)) {
                    return u;
                }
            }
            return null;
        }
        
        public UiUnit findChild(String code, Class cl) {
            if (cl.isInstance(this) && this.getCode().equals(code)) {
                return this;
            }
            UiUnit u = null;
            for (int i = 0; u == null && i < this.children.size(); i++) {
                u = this.children.get(i).findChild(code, cl);
            }
            return u;
        }
        
        public UiUnit findByCode(String code) {
            if (this.getCode().equals(code)) {
                return this;
            }
            UiUnit u = null;
            for (int i = 0; u == null && i < this.children.size(); i++) {
                u = this.children.get(i).findByCode(code);
            }
            return u;
        }
        
    }
    
    public UiUnit findUnitByCode(String code, String className) {
        Class c;
        try {
            c = Class.forName(className);
            for (Map.Entry<String, UiUnit> entry : units.entrySet()) {
                if (entry.getValue().getCode().equals(code) && c.isInstance(entry.getValue())) {
                    return entry.getValue();
                }
            }
        } catch (ClassNotFoundException ex) {}
        return null;
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiRootUnit - корневой узел                                 *
 *                                                                      *
 ************************************************************************/
    
    public class UiRootUnit extends UiUnit {
        
        private final ArrayList<JsonObject> addonInstructions;
        
        public UiRootUnit(String code) {
            super(code);
            addonInstructions = new ArrayList<>();
        }
        
        public void emulatedEvent() {
            addonInstructions.add(Utils.instruction(Instructions.b, getId()));
            addonInstructions.add(Utils.instruction(Instructions.ee));
            unitsToCheck.add(this);
        }
        
        @Override
        public void postCheck() {
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
    }
    
/************************************************************************
 *                                                                      *
 *     Класс UiWindow - "окно"                                          *
 *                                                                      *
 ************************************************************************/
    
    public class UiWindow extends UiUnit {
        
        String height = null;
        String width = null;
        String title = "";
        WindowPosition position;
        UiUnit parentWindow = null;
        private final ArrayList<JsonObject> addonInstructions;
        
        public UiWindow() {
            super();
            addonInstructions = new ArrayList<>();
            winit();
        }
        
        public UiWindow(String code) {
            super(code);
            addonInstructions = new ArrayList<>();
            winit();
        }
        
        private void winit() {
            setData("autosize");
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.win;
        }
        
        public void setWidth(String w) {
            if ((protocol != 3) && (!mobile) && (!Utils.testEmptyString(w))) {
                setUid(getId());
                width = w;
                instructions.add(Utils.instruction(Instructions.w, w));
                fixSize();
                /*if (position == WindowPosition.WorkCenter) {
                    setPosition(position, null);
                }*/
            } 
        }
        
        public void setHeight(String h) {
            if ((protocol != 3) && (!mobile) && (!Utils.testEmptyString(h))) {
                setUid(getId());
                height = h;
                instructions.add(Utils.instruction(Instructions.h, h));
                fixSize();
                /*if (position == WindowPosition.WorkCenter) {
                    setPosition(position, null);
                }*/
            } 
        }
        
        public void setMaximized(String p) {
            if ((protocol != 3) && (!mobile) && (!Utils.testEmptyString(p))) {
                setUid(getId());
                if (Utils.testYString(p)) {
                    instructions.add(Utils.instruction(Instructions.max));
                } else {
                    instructions.add(Utils.instruction(Instructions.rest));
                }
            }
        }
        
        public void setLeft(String p) {
            if ((protocol != 3) && (!mobile) && (!Utils.testEmptyString(p))) {
                setUid(getId());
                if ("%".equals(p.substring(0, 1))) {
                    instructions.add(Utils.instruction(Instructions.lft$, p.substring(1)));
                } else {
                    instructions.add(Utils.instruction(Instructions.lft, p));
                }
            }
        }
        
        public void setTop(String p) {
            if ((protocol != 3) && (!mobile) && (!Utils.testEmptyString(p))) {
                setUid(getId());
                if ("%".equals(p.substring(0, 1))) {
                    instructions.add(Utils.instruction(Instructions.top$, p.substring(1)));
                } else {
                    instructions.add(Utils.instruction(Instructions.top, p));
                }
            }
        }
        
        public void fixSize() {
            setData(null);
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.fsz));
            lock();
        }
        
        public void setTitle(String t) {
            if ((protocol != 3) && (t != null)) {
                title = t;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.ttl, t));
            } 
        }
        
        public void setSysMenu(String s) {
            if (s != null) {
                Defaults.Instructions cmd = (Utils.testYString(s)) ? Defaults.Instructions.mnu : Defaults.Instructions._mnu; 
                setUid(getId());
                instructions.add(Utils.instruction(cmd));
            } 
        }
        
        public void setCloseButton(String s) {
            if (s != null) {
                Defaults.Instructions cmd = (Utils.testYString(s)) ? Defaults.Instructions.cls : Defaults.Instructions._cls; 
                setUid(getId());
                instructions.add(Utils.instruction(cmd));
            } 
        }
        
        public void setMaximizeButton(String s) {
            if (s != null) {
                Defaults.Instructions cmd = (Utils.testYString(s)) ? Defaults.Instructions.mxm : Defaults.Instructions._mxm; 
                setUid(getId());
                instructions.add(Utils.instruction(cmd));
            } 
        }
        
        public void setParentWindow(UiUnit u) {
            parentWindow = u;
        }
        
        public void setPosition(WindowPosition pos) {
            position = pos;
        }
        
        public void applyPosition() {
            if ((protocol != 3) && (!mobile)) {
                setUid(this.getId());
                switch (position) {
                    case Zero:
                        instructions.add(Utils.instruction(Instructions.posz));
                        break;
                        
                    case Parent:
                        instructions.add(Utils.instruction(Instructions.posp, parentWindow.getId()));
                        break;
                        
                    case ParentRight:
                        instructions.add(Utils.instruction(Instructions.pspr, parentWindow.getId()));
                        break;
                        
                    case ParentBottom:
                        instructions.add(Utils.instruction(Instructions.pspb, parentWindow.getId()));
                        break;

                    case WorkCenter:
                        addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                        addonInstructions.add(Utils.instruction(Instructions.post));
                        unitsToCheck.add(contextWindow);
                        break; 

                }
            }
        }
        
        public void enable() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.en));
        }
        
        public void disable(String cover) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.dis, Utils.string2String(cover)));
        }
        
        @Override
        public void show() {
            super.show();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.show));
            unlock();
        }
        
        @Override
        public void hide() {
            super.hide();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.hide));
        }
        
        public void lock() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.lck));
        }
        
        public void unlock() {
            if ("autosize".equals(getData())) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._lck));
            }
        }
        
        @Override
        public void postCheck() {
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
        public void reqClose() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CloseWindow"));
            sep.put("WindowID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqSystemMenu() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SystemMenu"));
            sep.put("WindowID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqResize(String w, String h) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("WinWidth"));
            sep.put("Value", new StringParameter(w));
            totalEvent.add(new SingleEvent(sep));
            sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("WinHeight"));
            sep.put("Value", new StringParameter(h));
            totalEvent.add(new SingleEvent(sep));
            fixSize();
        }
        
        public void reqMaximize() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("WinMax"));
            sep.put("Value", new StringParameter("Y"));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqPosition(String left, String top) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("WinLeft"));
            sep.put("Value", new StringParameter(left));
            totalEvent.add(new SingleEvent(sep));
            sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("WinTop"));
            sep.put("Value", new StringParameter(top));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRestore() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("WinMax"));
            sep.put("Value", new StringParameter("N"));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqActivate() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("Activate"));
            sep.put("WindowID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
    }    
    

/************************************************************************
 *                                                                      *
 *     Класс UiMenuWindow - "главное меню системы" для современного     *
 *     приложения                                                       *
 *                                                                      *
 ************************************************************************/   
    public class UiMenuWindow extends UiWindow {
        
        public UiMenuWindow() {
            super();
        }
        
        public UiMenuWindow(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.mm;
        }
        
    }    

    
/************************************************************************
 *                                                                      *
 *     Класс UiLogonWindow - "форма входа в систему" для современного   *
 *     приложения                                                       *
 *                                                                      *
 ************************************************************************/   
    public class UiLogonWindow extends UiWindow {
        
        public UiLogonWindow() {
            super();
        }
        
        public UiLogonWindow(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.lgn;
        }
        
    } 
    
/************************************************************************
 *                                                                      *
 *     Класс UiServiceWindow -"Сервисное окно"                          *
 *                                                                      *
 ************************************************************************/   
    public class UiServiceWindow extends UiWindow {
        
        public UiServiceWindow() {
            super();
        }
        
        public UiServiceWindow(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.swin;
        }
        
    }     
    
 /***********************************************************************
 *                                                                      *
 *     Класс UiEmbeddedWindow - "встраиваемое окно"                     *
 *                                                                      *
 ************************************************************************/
    
    public class UiEmbeddedWindow extends UiWindow {
        
        public UiEmbeddedWindow() {
            super();
        }
        
        public UiEmbeddedWindow(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.ewin;
        }
        
        @Override
        public void setWidth(String w) {
            
        }
        
        @Override
        public void setHeight(String h) {
            
        }
        
        @Override
        public void setMaximized(String p) {
            
        }
        
        @Override
        public void setLeft(String p) {
            
        }
        
        @Override
        public void setTop(String p) {
            
        }
        
        @Override
        public void fixSize() {
            
        }
        
        @Override
        public void setTitle(String t) {
            
        }
        
        @Override
        public void setSysMenu(String s) {
             
        }
        
        @Override
        public void setCloseButton(String s) {
            
        }
        
        @Override
        public void setMaximizeButton(String s) {
            
        }
        
        @Override
        public void applyPosition() {

        }
        
        @Override
        public void enable() {
            
        }
        
        @Override
        public void disable(String cover) {
            
        }
        
        @Override
        public void show() {

        }
        
        @Override
        public void hide() {
            
        }
        
        @Override
        public void lock() {
            
        }
        
        @Override
        public void unlock() {
            
        }
        
        @Override
        public void postCheck() {

        }
        
    }    
       
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiMenuForm                                                 *
 *                                                                      *
 ************************************************************************/   
    public class UiMenuForm extends UiForm {
        
        public UiMenuForm(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.mnf;
        }
        
        @Override
        public void setType(FormType type) {
            this.type = type;
            if (type == FormType.Tree) {
                this.addChild(new UiMenuTree());
            }
        }
        
        
    }     
    
/************************************************************************
 *                                                                      *
 *     Класс UiMenuTree                                                 *
 *                                                                      *
 ************************************************************************/   
    public class UiMenuTree extends UiTree {
        
        public UiMenuTree() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.mtr;
        }
        
        @Override
        public void addChild(UiUnit child) {
            if ((child instanceof UiMenuNode) && (child.getParent() != UiMenuTree.this)) { 
                if (child.getParent() == null) {
                    setUid(this.getId());
                    instructions.add(Utils.instruction(child.getCreateInstruction(), child.getId()));
                    if (!Utils.testEmptyString(child.getCode())) {
                        setUid(child.getId());
                        instructions.add(Utils.instruction(Instructions.code, child.getCode()));
                    }
                    child.setParent(this);
                    children.add(child);
                    units.put(child.getId(), child);
                    unitsToCheck.add(this);
                }
            }
        }
        
        @Override
        public void preCheck() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.ini));
        }
        
    }         
    
    
 /************************************************************************
 *                                                                      *
 *     Класс UiMenuNode                                                 *
 *                                                                      *
 ************************************************************************/ 
    public class UiMenuNode extends UiTreeNode {
        
        public UiMenuNode(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.mnd;
        }
        
        @Override
        public void setState(String state) {
            try {
                NodeState ns = NodeState.valueOf(state);
                this.state = ns;
                if (ns == NodeState.Collapsed || ns == NodeState.AutoExpand) {
                    totalEvent = new TotalEvent();
                    reqChangeState();
                    sendEvent();
                }
            } catch (Exception ex) {}
        }
        
        @Override
        public void addChild(UiUnit child) {
            if ((child instanceof UiMenuNode)) { 
                if (child.getParent() == null) {
                    setUid(this.getId());
                    instructions.add(Utils.instruction(child.getCreateInstruction(), child.getId()));
                    if (!Utils.testEmptyString(child.getCode())) {
                        setUid(child.getId());
                        instructions.add(Utils.instruction(Instructions.code, child.getCode()));
                    }
                    child.setParent(this);
                    children.add(child);
                    units.put(child.getId(), child);
                    unitsToCheck.add(getTree());
                }
            }
        }
        
        @Override
        public void setIcon(String icon) {
            
        }
        
    }    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiForm - "форма"                                           *
 *                                                                      *
 ************************************************************************/   
    public class UiForm extends UiWindowChild implements UiHasHeader {
        
        private FormPosition position;
        public FormType type;
        private UiEastContainer eastContainer = null;
        private UiSouthContainer southContainer = null;
        private UiTabForm eastTab = null;
        private UiTabForm southTab = null;
        private int eastSize = -1;
        private int southSize = -1;
        private UiUnit currentUnit = null;
        private HashMap<String, String> autoCss;
        

        public UiForm(String code) {
            super(code);
            finit();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.form;
        }
        
        private void finit() {
            setData("autosize");
            autoCss = new HashMap<>();
        }
        
        public void lock() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.lck));
        }
        
        public void unlock() {
            getWindow().unlock();
            if ("autosize".equals(getData())) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._lck));
            }
        }
        
        public void fixSize() {
            setData(null);
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.fsz));
            lock();
        }
        
        public void setPosition(FormPosition position) {
            this.position = position;
            //instructions.add(Utils.instruction(Instructions.position$, position.toString()));
        }
        
        public FormPosition getPosition() {
            return position;
        }

        public void setType(FormType type) {
            this.type = type;
            switch (type) {
                case Tree:
                    this.addChild(new UiTree());
                    break;
                case List:
                    this.addChild(new UiGrid());
                    break;
                case Blanc:
                    this.addChild(new UiBlank());
                    break;
                case Chart:
                    this.addChild(new UiChart());
                    break;
                case Html:
                    this.addChild(new UiHtml());
                    break;    
                case Grid:
                    this.addChild(new UiMatrix(false));
                    break;
            }
        }
        
        public void setHeight(String height) {
            if (!mobile) {
                if (!Utils.testEmptyString(height)) {
                    setUid(getId());
                    if ("%".equals(height.substring(0, 1))) {
                        instructions.add(Utils.instruction(Instructions.h$, height.substring(1)));
                    } else {
                        instructions.add(Utils.instruction(Instructions.h, height));
                    }
                    fixSize();
                }
            }
        }
        
        public void setWidth(String width) {
            if (!mobile) {
                if (!Utils.testEmptyString(width)) {
                    setUid(getId());
                    if ("%".equals(width.substring(0, 1))) {
                        instructions.add(Utils.instruction(Instructions.w$, width.substring(1)));
                    } else {
                        instructions.add(Utils.instruction(Instructions.w, width));
                    }
                    fixSize();
                }
            }
        }
        
        public void setCover(String c) {
            if (!Utils.testEmptyString(c)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.cvr, c));
            }
        }
        
        public String parseCss(String val, String delim, boolean join) {
            String p, p1, p2;
            String css = "";
            String n = "class-" + Utils.numberSeq();
            String ret = join ? n : "";
            String[] classes = StringUtils.splitByWholeSeparator(val, delim);
            for (String s : classes) {
                String t = "";
                String m = "";
                String tmp = ".grid-cell.%n .cell-content%m{%t} .ocolumn.%n .ocolumn-content%m{%t} .ui-field.%n .field-value%m{%t} .ui-matrix-cell.%n{%t} td.ui-matrix-row.%n{%t} td.ui-matrix-column.%n{%t}";
                if (join || ((!join) && Utils.testEmptyString(autoCss.get(s)))) {
                    if (StringUtils.startsWith(s, "MARKED(")) {
                        tmp = ".grid-cell.%n .cell-content%m{%t} .ocolumn.%n .ocolumn-content%m{%t} .ui-field.%n .field-value%m{%t} .ui-matrix-cell.%n .ui-matrix-cell-content%m{%t}";
                        p = StringUtils.substringAfter(StringUtils.substringBeforeLast(s, ")"), "(");
                        if (!Utils.testEmptyString(p)) {
                            p1 = Utils.string2String(StringUtils.substringBefore(p, ","));
                            p2 = Utils.string2String(StringUtils.substringAfter(p, ","));
                            t = "align-self:center;padding-right:5px;content:'" + p1 + "';color:" + p2 + ";";
                            m = ":before";
                        }
                    } else if (StringUtils.startsWith(s, "UNDERLINE(")) {
                        tmp = tmp + " span.%n{%t}";
                        t = "text-decoration:underline;text-decoration-color:" + StringUtils.substringAfter(StringUtils.substringBeforeLast(s, ")"), "(") + ";";
                    } else if (StringUtils.startsWith(s, "BOUNDARY(")) {
                        p = StringUtils.substringAfter(StringUtils.substringBeforeLast(s, ")"), "(");
                        if (!Utils.testEmptyString(p)) {
                            p1 = Utils.string2String(StringUtils.substringBefore(p, ","));
                            p2 = Utils.string2String(StringUtils.substringAfter(p, ","));
                            if (p2.toLowerCase().equals("none")) {
                                t = "border-" + p1 + ":none;";
                            } else {
                                t = "border-" + p1 + ":2px solid " + p2 + ";";
                            }    
                        }
                    } else if (StringUtils.startsWith(s, "RECTANGLE(")) {
                        p1 = StringUtils.substringAfter(StringUtils.substringBeforeLast(s, ")"), "(");
                        if (p1.toLowerCase().equals("none")) {
                            t = "border:none;";
                        } else {
                            t = "border:2px solid " + p1 + ";";
                        }    
                    } else if (StringUtils.startsWith(s, "BACKGROUND(")) {
                        tmp = "div.grid-cell.%n{%t} .ocolumn.%n .ocolumn-content%m{%t} .ui-field.%n .field-value%m{%t} .ui-matrix-cell.%n{%t} td.ui-matrix-row.%n{%t} td.ui-matrix-column.%n{%t}";
                        t = "background-color:" + StringUtils.substringAfter(StringUtils.substringBeforeLast(s, ")"), "(") + ";";
                    } else if (StringUtils.startsWith(s, "TEXTCOLOR(")) {
                        tmp = tmp + " span.%n{%t}";
                        t = "color:" + StringUtils.substringAfter(StringUtils.substringBeforeLast(s, ")"), "(") + ";";
                    } else if (StringUtils.startsWith(s, "HIDDEN")) {
                        tmp = "span.%n{%t}";
                        t = Utils.string2String(Defaults.CLASS_TEMPLATES.get(s));
                    } else {
                        tmp = tmp + " span.%n{%t}";
                        t = Utils.string2String(Defaults.CLASS_TEMPLATES.get(s));
                    }
                    if (!Utils.testEmptyString(t)) {
                        autoCss.put(s, n);
                        css = css + StringUtils.replace(StringUtils.replace(StringUtils.replace(tmp, "%n", n), "%t", t), "%m", m);
                    }
                } else {
                    n = autoCss.get(s);
                }
                if (!join) {
                    ret = Utils.testEmptyString(ret) ? n : ret + delim + n;
                    n = "class-" + Utils.numberSeq();
                }
            }
            if (!Utils.testEmptyString(css)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.css, css));
            }
            return StringUtils.strip(ret);
        }
        
        public String clearCss(String val) {
            String ret = val.replaceAll("\\s+", " ");
            ret = StringUtils.replace(StringUtils.replace(ret, " ,", ","), ", ", ",");
            ret = StringUtils.replace(StringUtils.replace(ret, " (", "("), "( ", "(");
            ret = StringUtils.replace(StringUtils.replace(ret, "; ", ";"), " ;", ";");
            ret = StringUtils.replace(ret, " )", ")");
            return ret;
        }
        
        public String parseMultiCss(String val){
            String ret = "";
            String vl = clearCss(val);
            if (StringUtils.contains(vl, ";")) {
                String[] a = StringUtils.split(vl);
                for (String css : a) {
                    ret = Utils.testEmptyString(ret) ? parseCss(css, ";", true) : ret + " " + parseCss(css, ";", true);
                }
            } else {
                return parseCss(vl, " ", false);
            }
            return ret;
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (child.getParent() != UiForm.this) {
                if (child instanceof UiFormContent) {
                    if (getContent() == null) {
                        super.addChild(child);
                    }
                } else if (child instanceof UiHeader) {
                    if (getHeader() == null) {
                        super.addChild(child);
                    }
                } else if (child instanceof UiToolBar) {
                    if (getToolBar() == null) {
                        super.addChild(child);
                    }
                } else if (child instanceof UiButton) {
                    UiButtonBar buttonBar = (UiButtonBar) getButtonBar();
                    if (buttonBar == null) {
                        buttonBar = new UiButtonBar();
                        super.addChild(buttonBar);
                    }
                    buttonBar.addChild(child);
                } else if (child instanceof UiForm) {
                    UiForm f = (UiForm) child;
                    if (f.getPosition() == FormPosition.Right || f.getPosition() == FormPosition.TabRight) {
                        if (eastContainer == null) {
                            eastContainer = new UiEastContainer();
                            this.addChild(eastContainer);
                        }
                        if (f.getPosition() == FormPosition.Right) {
                            eastContainer.addChild(f);
                        } else {
                            if (eastTab == null) {
                                eastTab = new UiTabForm();
                                eastContainer.addChild(eastTab);
                            }
                            eastTab.addChild(f);
                        }    
                    } else if (f.getPosition() == FormPosition.Bottom || f.getPosition() == FormPosition.TabBottom) {
                        if (southContainer == null) {
                            southContainer = new UiSouthContainer();
                            this.addChild(southContainer);
                        }
                        if (f.getPosition() == FormPosition.Bottom) {
                            southContainer.addChild(f);
                        } else {
                            if (southTab == null) {
                                southTab = new UiTabForm();
                                southContainer.addChild(southTab);
                            }
                            southTab.addChild(f);
                        }    
                    }
                } else {
                    super.addChild(child);
                }
            }
        }
        
        @Override
        public UiUnit getHeader() {
            return getChildByClass(clHeader);
        }
        
        public UiUnit getToolBar() {
            return getChildByClass(clToolBar);
        }
        
        public UiUnit getButtonBar() {
            return getChildByClass(clButtonBar);
        }
        
        public UiUnit getContent() {
            return getChildByClass(clFormContent);
        }
        
        public UiUnit getBasement() {
            return getChildByClass(clBasement);
        }
        
        public void setEastSize(String size) {
            if (!mobile) {
                if (!Utils.testEmptyString(size)) {
                    setUid(getId());
                    if ("%".equals(size.substring(0, 1))) {
                        instructions.add(Utils.instruction(Instructions.ec$, size.substring(1)));
                    } else {
                        instructions.add(Utils.instruction(Instructions.ec, size));
                    }
                    fixSize();
                }
            }
        }
        
        public void setSouthSize(String size) {
            if (!mobile) {
                if (!Utils.testEmptyString(size)) {
                    setUid(getId());
                    if ("%".equals(size.substring(0, 1))) {
                        instructions.add(Utils.instruction(Instructions.sc$, size.substring(1)));
                    } else {
                        instructions.add(Utils.instruction(Instructions.sc, size));
                    }
                    fixSize();
                }    
            }
        }
        
        public void setCurrentUnit(UiUnit u) {
            currentUnit = u;
        }
        
        public UiUnit getCurrentUnit() {
            return currentUnit;
        }
        
        public void setProperties(String c, String p) {
            if (!Utils.testEmptyString(p)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.ecod, c));
                instructions.add(Utils.instruction(Instructions.prop, p));
            }
        }
        
        public void reqFetchRows() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("FetchRows"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));   
        }
        
        public void reqChangeMode() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChangeMode"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));   
        }
        
        public void reqCheckAll() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CheckAll"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));   
        }
        
        public void reqUnCheckAll() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("UnCheckAll"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));   
        }
        
        public void reqInvertChecking() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("InvertChecking"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));   
        }
        
        public void reqChangeConfig(String id) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChangeConfig"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("ConfigID", new StringParameter(id));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqAction(String act, String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("Action"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Code", new StringParameter(act));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqActivate() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("FormActivate"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqInsert() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("InsertBar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqDelete() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("DeleteBar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqEscape() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("EscapeBar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF1() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F1Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF4() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F4Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF5() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F5Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF6() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F6Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqShiftF6() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ShiftF6Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF7() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F7Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF8() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F8Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqF10() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F10Bar"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqUndo() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("Undo"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRedo() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("Redo"));
            sep.put("FormID", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqLMB(String c, String p) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("LMB"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Code", new StringParameter(c));
            if (!Utils.testEmptyString(p)) {
                sep.put("Parameters", new StringParameter(p));
            }
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRMB(String c, String p) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("RMB"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Code", new StringParameter(c));
            if (!Utils.testEmptyString(p)) {
                sep.put("Parameters", new StringParameter(p));
            }
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqDBLCLK(String c, String p) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("DBLCLK"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Code", new StringParameter(c));
            if (!Utils.testEmptyString(p)) {
                sep.put("Parameters", new StringParameter(p));
            }
            totalEvent.add(new SingleEvent(sep));
        }
                
        public void reqResizeWidth(String size) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("ChildWidth"));
            sep.put("Value", new StringParameter(size));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqResizeHeight(String size) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("ChildHeight"));
            sep.put("Value", new StringParameter(size));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqResizeOverflow(String size) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("Overflow"));
            sep.put("Value", new StringParameter(size));
            totalEvent.add(new SingleEvent(sep));
        }
                
        @Override
        public void rmChild(UiUnit child) {
            if (child == eastContainer) {
                eastContainer = null;
            } else if (child == southContainer) {
                southContainer = null;
            }
            super.rmChild(child);
        }

        public void rmTab(UiUnit tab) {
            if (tab == eastTab) {
                eastTab = null;
            } else if (tab == southTab) {
                southTab = null;
            }
        }
                
    }
   

    
/************************************************************************
 *                                                                      *
 *     Класс UiWindowChild - "дочерний элемент окна"                    *
 *                                                                      *
 ************************************************************************/
    
    public class UiWindowChild extends UiUnit {
        
        public UiWindowChild() {
            super();
        }
        
        public UiWindowChild(boolean render) {
            super(render);
        }
        
        public UiWindowChild(String code) {
            super(code);
        }
        
        // Получить родительское окно
        public UiWindow getWindow() {
            return new Object() {
                UiWindow find(UiUnit u) {
                    if (u == null) {
                        return null;
                    } else if (u instanceof UiWindow) {
                        return (UiWindow) u;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this.getParent());
        }
        
    }

    
    
/************************************************************************
 *                                                                      *
 *     Класс UiFormChild - "дочерний элемент формы"                     *
 *                                                                      *
 ************************************************************************/
    
    public class UiFormChild extends UiWindowChild {
        
        public UiFormChild() {
            super();
        }
        
        public UiFormChild(boolean render) {
            super(render);
        }
        
        public UiFormChild(String code) {
            super(code);
        }
        
        // Получить родительскую форму
        public UiForm getForm() {
            return new Object() {
                UiForm find(UiUnit u) {
                    if (u == null) {
                        return null;
                    } else if (u instanceof UiForm) {
                        return (UiForm) u;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this.getParent());
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiHeader - "заголовок" чего-либо                           *
 *                                                                      *
 ************************************************************************/ 
    
    public class UiHeader extends UiFormChild {
        
        String text;
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.head;
        }
        
        public void setText(String text) {
            this.text = Utils.string2String(text);
            setUid(getId());
            if (Utils.testEmptyString(this.text)) {
                this.remove();
            } else {
                instructions.add(Utils.instruction(Instructions.txt, this.text));
            } 
        }
        
        public String getText() {
            return text;
        }
        
    }

    
/************************************************************************
 *                                                                      *
 *     Класс UiFormHeader - "заголовок" формы                          *
 *                                                                      *
 ************************************************************************/ 
    
    public class UiFormHeader extends UiHeader {
        
        @Override
        public void setText(String text) {
            this.text = Utils.string2String(text);
            setUid(getId());
            if (Utils.testEmptyString(this.text)) {
                if (protocol == 3) {
                    instructions.add(Utils.instruction(Instructions._txt));
                } else {
                    this.remove();
                }
            } else {
                instructions.add(Utils.instruction(Instructions.txt, this.text));
            } 
        }
        
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiTabForm - "закладочная мастер-форма"                     *
 *                                                                      *
 ************************************************************************/     
    
    public class UiTabForm extends UiFormChild {
        
        public UiTabForm() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.tabf;
        }
    
        @Override
        public void rmChild(UiUnit child) {
            super.rmChild(child);
            if (children.isEmpty()) {
                remove();
            }
        }
        
    }   
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiEastContainer - "контейнер для дочерних объектов"        *
 *                                                                      *
 ************************************************************************/    
    
    public class UiEastContainer extends UiFormChild {
        
        public UiEastContainer() {
            super();
        }
        
        public void reqResize(String size) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Property", new StringParameter("SplitterRight"));
            sep.put("Value", new StringParameter(size));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.econ;
        }
        
        @Override
        public void rmChild(UiUnit child) {
            super.rmChild(child);
            if (child instanceof UiTabForm) {
                getForm().rmTab(child);
            }
            if (children.isEmpty()) {
                remove();
            }
        }
        
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiSouthContainer - "контейнер для дочерних объектов"       *
 *                                                                      *
 ************************************************************************/    
    
    public class UiSouthContainer extends UiFormChild {
        
        public UiSouthContainer() {
            super();
        }
        
        public void reqResize(String size) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Property", new StringParameter("SplitterBottom"));
            sep.put("Value", new StringParameter(size));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.scon;
        }
        
        @Override
        public void rmChild(UiUnit child) {
            super.rmChild(child);
            if (child instanceof UiTabForm) {
                getForm().rmTab(child);
            }
            if (children.isEmpty()) {
                remove();
            }
        }
        
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiToolBar                                                  *
 *                                                                      *
 ************************************************************************/
    
    public class UiToolBar extends UiFormChild {
        
        public UiToolBar() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.tbar;
        }
        
        public UiUnit getButtonBar() {
            return getChildByClass(clButtonBar);
        }
        
        public UiUnit getConfigs() {
            return getChildByClass(clConfigs);
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (child instanceof UiToolButton || child instanceof UiSeparator) {
                UiUnit bb = getButtonBar();
                if (bb == null) {
                    bb = new UiButtonBar();
                    super.addChild(bb);
                }
                bb.addChild(child);
            } else {
                super.addChild(child);
            }
        }
        
        @Override
        public void preCheck() {
            if (getButtonBar() == null && getConfigs() == null) {
                remove();
            }
        }
        
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiConfigs - "выбиралка конфигурации"                       *
 *                                                                      *
 ************************************************************************/
    
    public class UiConfigs extends UiFormChild {
        
        public UiConfigs() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.cfg;
        }
        
        public void set(String cfg) {
            if (cfg != null) {
                clear();
                unitsToCheck.add(getParent());
                if (Utils.testEmptyString(cfg)) {
                    remove();
                } else {    
                    UiCollection c = new UiCollection(cfg);
                    if (c.isEmpty()) {
                        remove();
                    } else {    
                        setUid(getId());
                        while (!c.isLast()) {
                            addItem(c.getCode(), c.getValue());
                            c.getNext();
                        }
                        if (protocol == 3) {
                            setUid(getId());
                            instructions.add(Utils.instruction(Instructions._cfg));
                        }
                    }
                }
            }
        }
        
        public void addItem(String key, String value) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.itm));
            instructions.add(Utils.instruction(Instructions.k, key));
            instructions.add(Utils.instruction(Instructions.v, value));
            if (Utils.testEmptyString(key)) {
                instructions.add(Utils.instruction(Instructions.curr));
            }
            instructions.add(Utils.instruction(Instructions._itm));
        }
        
        public void clear() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.clr));
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiButtonBar - "панель с кнопками"                          *
 *                                                                      *
 ************************************************************************/    
    
    public class UiButtonBar extends UiFormChild {
        
        private final ArrayList<JsonObject> addonInstructions;
        
        public UiButtonBar() {
            super();
            addonInstructions = new ArrayList<>();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.bbar;
        }
        
        public void clear() {
            this.rmChildren();
        }
        
        public void set(String buttons) {
            if (buttons != null) {
                clear();
                unitsToCheck.add(getParent());
                if (Utils.testEmptyString(buttons)) {
                    remove();
                } else {
                    UiCollection bc = new UiCollection(buttons);
                    if (bc.isEmpty()) {
                        remove();
                    } else {    
                        while (!bc.isLast()) {
                            String code = bc.getCode();
                            if (Utils.testEmptyString(code)) {
                                addButton(new UiSeparator());
                            } else {
                                ArrayList<JsonObject> el = new ArrayList<>();
                                String ico;
                                String lbl = "";
                                if (StringUtils.contains(code, "\n")) {
                                    ico = substringBefore(code, "\n");
                                    lbl = substringAfter(code, "\n");
                                } else {
                                    ico = code;
                                }
                                JsonObject o;
                                if (!Utils.testEmptyString(ico)) {
                                    o = new JsonObject();
                                    o.addProperty("type", "icon");
                                    o.addProperty("val", ico);
                                    el.add(o);
                                    if (!Utils.testEmptyString(lbl)) {
                                        o = new JsonObject();
                                        o.addProperty("type", "text");
                                        o.addProperty("val", lbl);
                                        el.add(o);
                                    }
                                    UiToolButton tb = new UiToolButton(ico);
                                    addButton(tb);
                                    tb.setExtLabel(el);
                                    tb.setHint(bc.getValue());
                                }
                            }
                            bc.getNext();
                        } 
                        UiToolButton tb = new UiToolButton("");
                        addButton(tb);
                        tb.setState("More");
                    }
                }
            }
        }
        
        public void addButton(UiButton btn) {
            this.addChild(btn);
        }
        
        public void addButton(UiToolButton btn) {
            this.addChild(btn);
        }
        
        public void addButton(UiSeparator sep) {
            if (!(mobile)) {
                this.addChild(sep);
            }
        }
        
        @Override
        public void preCheck() {
            if (!(getParent() instanceof UiToolBar)) {
                Instructions instr = Instructions.hide;
                for (int i = 0; i < this.children.size(); i++) {
                    if (this.children.get(i) instanceof UiBaseButton) {
                        ButtonState state = ((UiBaseButton)this.children.get(i)).getState();
                        if (state != ButtonState.Hidden) {
                            instr = Instructions.show;
                            break;
                        }
                    }
                }
                addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                addonInstructions.add(Utils.instruction(instr));
            }
        }
        
        @Override
        public void postCheck() {
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiSeparator - некий "разделитель"                          *
 *                                                                      *
 ************************************************************************/     
    
    public class UiSeparator extends UiUnit {
        
        public UiSeparator() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.sep;
        }
        
    }

    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBaseButton - "базовая кнопка"                            *
 *                                                                      *
 ************************************************************************/    
    
    public class UiBaseButton extends UiFormChild {
        
        String hint = null;
        ButtonState state = ButtonState.Active;
        
        public UiBaseButton(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.btn;
        }
        
        public void setHint(String hint) {
            if (hint != null) {
                setUid(getId());
                if (Utils.testEmptyString(hint)) {
                    this.hint = null;
                    instructions.add(Utils.instruction(Instructions._hnt));
                } else {
                    this.hint = hint;
                    instructions.add(Utils.instruction(Instructions.hnt, hint));
                }
            }
        }
        
        public void setBgColor(String c) {
            
        }
        
        public void setState(String state) {
            if (state != null) {
                setUid(getId());
                try {
                    this.state = ButtonState.valueOf(state);
                    switch (this.state) {
                        case Active:    instructions.add(Utils.instruction(Instructions.show));
                                        instructions.add(Utils.instruction(Instructions.en));
                                        break;

                        case Disable:   instructions.add(Utils.instruction(Instructions.show));
                                        instructions.add(Utils.instruction(Instructions.dis));
                                        break;    

                        case Hidden:    instructions.add(Utils.instruction(Instructions.hide));
                                        break;
                                        
                        case More:      instructions.add(Utils.instruction(Instructions.more));
                                        break;           
                        
                        default:        instructions.add(Utils.instruction(Instructions.stat, state.toLowerCase()));
                    }
                    unitsToCheck.add(this.getParent());
                } catch (Exception e) {}
            }
        }
        
        public ButtonState getState() {
            return state;
        }
        
        public void reqPress() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ButtonPressed"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Code", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep)); 
        }
        
    }
        
    

/************************************************************************
 *                                                                      *
 *     Класс UiButton - "кнопка"                                        *
 *                                                                      *
 ************************************************************************/    
    
    public class UiButton extends UiBaseButton {
        
        public UiButton(String code) {
            super(code);
        }
                
        public void setLabel(String label) {
            if (label != null) {
                setUid(getId());
                String tmp = label;
                StringBuilder l = new StringBuilder();
                l.append("<div class='ui-button-content lui2-unit'>");
                while (StringUtils.contains(tmp, "[ico:")) {
                    String txt = StringUtils.substringBefore(tmp, "[ico:");
                    if (!Utils.testEmptyString(txt)) {
                        l.append("<div class='ui-button-label lui2-unit'>");
                        l.append(txt);
                        l.append("</div>");
                    }    
                    tmp = StringUtils.removeStart(tmp, txt);
                    l.append("<div class='ui-button-icon ico-");
                    l.append(StringUtils.substringAfter(StringUtils.substringBefore(tmp, "]"), "[ico:"));
                    l.append(" lui2-unit'></div>");
                    tmp = StringUtils.substringAfter(tmp, "]");
                }
                if (!Utils.testEmptyString(tmp)) {
                    l.append("<div class='ui-button-label lui2-unit'>");
                    l.append(tmp);
                    l.append("</div>");
                }
                l.append("</div>");
                instructions.add(Utils.instruction(Instructions.lbl, l.toString()));
            }
        }
        
        public void reqCurrent() {
            UiForm form = getForm();
            UiUnit curr = form.getCurrentUnit();
            if ((curr != null) && (curr instanceof UiFormCurrent)) {
                ((UiFormCurrent)curr).unsetCurrent();
            }    
            getForm().setCurrentUnit(this);
        }
        
        @Override
        public void reqPress() {
            //reqCurrent();
            super.reqPress();
        }
        
        @Override
        public void setHint(String hint) {
            if (!mobile) {
                super.setHint(hint);
            }
        }
        
    }    
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiToolButton - "кнопка на toolbar"                         *
 *                                                                      *
 ************************************************************************/    
    
    public class UiToolButton extends UiBaseButton {
        
        private String iconCode = null;
        
        public UiToolButton(String code) {
            super(code);
        }
        
        public void setIcon(String ico) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.ico, ico));  
        }
        
        public void setExtLabel(ArrayList<JsonObject> label) {
            Gson gson = new Gson();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.xlbl, gson.toJson(label)));
        }
        
        public void setIconCode(String c) {
            iconCode = c;
        }
        
        @Override
        public void reqPress() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ToolBarButton"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            if (Utils.testEmptyString(iconCode)) {
                sep.put("Name", new StringParameter(getCode()));
            } else {
                sep.put("Name", new StringParameter(iconCode));
            }
            totalEvent.add(new SingleEvent(sep)); 
        }
        
    }        
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiTree - "контейнер для узлов дерева"                      *
 *                                                                      *
 ************************************************************************/     
    
    public class UiTree extends UiFormChild implements UiFormContent {
        
        public UiTree() {
            super();
        }  
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.tree;
        }
        
        @Override
        public void addChild(UiUnit child) {
            if ((child instanceof UiTreeNode) && (child.getParent() != UiTree.this)) {
                super.addChild(child);
                unlock();
            }
        }
        
        public void unlock() {
            getForm().unlock();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions._lck));
        }
        
        public void reqRMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NodeRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            totalEvent.add(new SingleEvent(sep));
        }  
        
    }

    
/************************************************************************
 *                                                                      *
 *     Класс UiTreeChild - "дочерний элемент дерева"                    *
 *                                                                      *
 ************************************************************************/
    
    public class UiTreeChild extends UiFormChild {
        
        public UiTreeChild() {
            super();
        }
        
        public UiTreeChild(String code) {
            super(code);
        }
        
        // Получить родительское дерево
        public UiTree getTree() {
            return new Object() {
                UiTree find(UiUnit u) {
                    if (u == null) {
                        return null;
                    } else if (u instanceof UiTree) {
                        return (UiTree) u;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this.getParent());
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiTreeNode - "узел дерева"                                 *
 *                                                                      *
 ************************************************************************/    
    
    public class UiTreeNode extends UiTreeChild implements UiFormCurrent {
        
        String label = null;
        NodeState state = NodeState.Empty;
        NodeType type = NodeType.ROW;
        String context = "";
        String contextId = "";
        String childsContextId = "";
        
        public UiTreeNode(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.nd;
        }
        
        public void setLabel(String label) {
            if (label != null) {
                this.label = label;
                setUid(getId());
                if (Utils.hasIcons(label)) {
                    Gson gson = new Gson();
                    instructions.add(Utils.instruction(Instructions.xlbl, gson.toJson(Utils.parseCaption(label))));
                } else {
                    instructions.add(Utils.instruction(Instructions.lbl, label));
                }    
            }
        }
        
        public void setHint(String hint) {
            if ((!mobile) && (hint != null)) {
                setUid(getId());
                if (Utils.testEmptyString(hint)) {
                    instructions.add(Utils.instruction(Instructions._hnt));
                } else {
                    instructions.add(Utils.instruction(Instructions.hnt, hint));
                }
            }
        }
        
        public void setIcon(String icon) {
            if (icon != null) {
                setUid(getId());
                if (Utils.testEmptyString(icon)) {
                    instructions.add(Utils.instruction(Instructions._ico));
                } else {
                    instructions.add(Utils.instruction(Instructions.ico, icon));
                }
            }
        }
        
        public void setSelected(String selected) {
            if (!Utils.testEmptyString(selected)) {
                if (selected.equals("Y")) {
                    getForm().setCurrentUnit(this);
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.sel));
                }
            }
        }
        
        public void setState(String state) {
            try {
                NodeState ns = NodeState.valueOf(state);
                if ((mobile || protocol == 3) && ns == NodeState.AutoExpand) {
                    ns = NodeState.Collapsed;
                }
                this.state = ns;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.stat, this.state.toString()));
                if (ns == NodeState.Expanded) {
                    getTree().unlock();
                }
            } catch (Exception ex) {}
        }
        
        public void setType(String type) {
            try {
                NodeType nt = NodeType.valueOf(type);
                this.type = nt;
            } catch (Exception e) {}
        }
        
        public void setStyle(String style) {
            if (style != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._sty));
                if (!Utils.testEmptyString(style)) {
                    instructions.add(Utils.instruction(Instructions.sty, style));
                }  
            }
        }
        
        @Override
        public void unsetCurrent() {
            getForm().setCurrentUnit(null);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChangeNode"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Node", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void setContext(String context) {
            if (context != null) this.context = context;
        }
        
        public void setContextId(String contId) {
            if (contId != null) this.contextId = contId;
        }
        
        public void setChildsContextId(String chContId) {
            if (chContId != null) this.childsContextId = chContId;
        }
        
        public String findContext() {
            return new Object() {
                String find(UiUnit u) {
                    if (u == null) {
                        return "";
                    } else if ((u instanceof UiTreeNode) && (!Utils.testEmptyString(((UiTreeNode) u).contextId))) {
                        return ((UiTreeNode) u).context;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this);
        }
        
        public String findContextId() {
            return new Object() {
                String find(UiUnit u) {
                    if (u == null) {
                        return "";
                    } else if ((u instanceof UiTreeNode) && (!Utils.testEmptyString(((UiTreeNode) u).contextId))) {
                        return ((UiTreeNode) u).contextId;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this);
        }
        
        public void reqDoubleClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NodeDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Node", new StringParameter(getCode()));
            sep.put("NType", new StringParameter(type.toString()));
            sep.put("State", new StringParameter(state.toString()));
            sep.put("Context", new StringParameter(findContext()));
            sep.put("ContextID", new StringParameter(findContextId()));
            sep.put("ChildsContextID", new StringParameter(childsContextId));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqChangeNode() {
            UiTree tree = getTree();
            if (this == getForm().getCurrentUnit()) {
                if (mobile) {
                    reqDoubleClick();
                }
            } else {
                getForm().setCurrentUnit(this);
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("ChangeNode"));
                sep.put("FormID", new StringParameter(tree.getForm().getCode()));
                sep.put("Node", new StringParameter(getCode()));
                sep.put("NType", new StringParameter(type.toString()));
                sep.put("State", new StringParameter(state.toString()));
                sep.put("Context", new StringParameter(findContext()));
                sep.put("ContextID", new StringParameter(findContextId()));
                sep.put("ChildsContextID", new StringParameter(childsContextId));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public void reqCheckNode(String c) {
            UiTree tree = getTree();
            getForm().setCurrentUnit(null);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CheckNode"));
            sep.put("FormID", new StringParameter(tree.getForm().getCode()));
            sep.put("Node", new StringParameter(getCode()));
            sep.put("State", new StringParameter(state.toString()));
            sep.put("Context", new StringParameter(findContext()));
            sep.put("ContextID", new StringParameter(findContextId()));
            sep.put("IsCheck", new StringParameter(c));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqChangeState() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChangeState"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Node", new StringParameter(getCode()));
            sep.put("NType", new StringParameter(type.toString()));
            NodeState ns = state == NodeState.AutoExpand ? NodeState.Collapsed : state;
            sep.put("State", new StringParameter(ns.toString()));
            sep.put("Context", new StringParameter(findContext()));
            sep.put("ContextID", new StringParameter(findContextId()));
            sep.put("ChildsContextID", new StringParameter(childsContextId));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRMB() {
            if (!mobile) {
                reqChangeNode();
            }
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NodeRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Node", new StringParameter(getCode()));
            sep.put("NType", new StringParameter(type.toString()));
            sep.put("State", new StringParameter(state.toString()));
            if (Utils.testEmptyString(childsContextId)) {
                sep.put("ContextID", new StringParameter(findContextId()));
            } else {
                sep.put("ContextID", new StringParameter(childsContextId));
            }
            /*if (this.state == NodeState.Expanded) {
                sep.put("ContextID", new StringParameter(childsContextId));
            } else {
                sep.put("ContextID", new StringParameter(findContextId()));
            }*/
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRMBChecked() {
            UiTree tree = getTree();
            getForm().setCurrentUnit(null);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NodeRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Node", new StringParameter(getCode()));
            sep.put("NType", new StringParameter(type.toString()));
            sep.put("State", new StringParameter(state.toString()));
            if (Utils.testEmptyString(childsContextId)) {
                sep.put("ContextID", new StringParameter(findContextId()));
            } else {
                sep.put("ContextID", new StringParameter(childsContextId));
            }
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqSpace() {
            try {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("SpaceBar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Element", new StringParameter(getCode()));
                totalEvent.add(new SingleEvent(sep));
            } catch (Exception ex) {
            }
        }
        
        public void reqEnter() {
            try {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("EnterBar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Element", new StringParameter(getCode()));
                totalEvent.add(new SingleEvent(sep));
            } catch (Exception ex) {
            }
        }
        
        public void reqInsert() {
            if ((!Utils.testEmptyString(contextId)) || (!Utils.testEmptyString(childsContextId))) {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("InsertBar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public void reqDelete() {
            if (!Utils.testEmptyString(contextId)) {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("DeleteBar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiGrid - "список"                                          *
 *                                                                      *
 ************************************************************************/     
    
    public class UiGrid extends UiFormChild implements UiFormContent {
        
        UiUnit currentRow = null;
        UiUnit contextRow = null;
        UiUnit currentColumn = null;
        boolean empty = true;
        boolean header = true;
        boolean footer = false;
        boolean allRows = false;
        boolean qbeMode = false;
        boolean hadData = false;
        private Defaults.SelectionMode selMode = SelectionMode.Single;
        private final ArrayList<UiGridColumn> columns;
        private final ArrayList<UiUnit> dataRows;
        private final ArrayList<UiUnit> qbeRows;
        private ArrayList<UiUnit> rows;
        boolean checkFetch = true;
        private final ArrayList<JsonObject> addonInstructions;
        
        public UiGrid() {
            super();
            dataRows = new ArrayList<>();
            qbeRows = new ArrayList<>();
            rows = dataRows;
            columns = new ArrayList<>();
            addonInstructions = new ArrayList<>();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.grid;
        }
        
        public UiGridRow getContextRow() {
            return (UiGridRow)contextRow;
        }
        
        public UiGridRow getCurrentRow() {
            return (UiGridRow)currentRow;
        }
        
        public UiGridColumn getCurrentColumn() {
            return currentColumn == null ? null : (UiGridColumn)currentColumn;
        }
        
        public void setCurrentRow(UiUnit row) {
            this.currentRow = row;
            this.contextRow = row;
        }    
        
        public void setCurrentColumn(UiUnit column) {
            this.currentColumn = column;
        } 
        
        public UiGridCell reqCell(String uid, String cid, String row) {
            if (qbeMode) {
                return ((UiGridCell) units.get(uid));
            } else {
                Integer idx;
                try {
                    idx = Integer.valueOf(row);
                    return ((UiGridColumn) units.get(cid)).getCell(getRowByIndex(idx));
                } catch (Exception ex) {

                }
            }
            return null;
        }
        
        public void setCell(String row, String col) {
            if (col != null) {
                UiGridColumn c = checkColumn(col);
                if (c == null) {
                    c = getCurrentColumn();
                }
                if ((c != null) && (row != null)) {
                    UiGridRow r = getRowByNum(row);
                    if (r == null) {
                        r = getCurrentRow();
                    }
                    if (r != null) {
                        setCurrentRow(r);
                        setCurrentColumn(c);
                        c.setCell(r);
                    }
                }
            }
        }
        
        public UiGridRow getRowByIndex(int idx) {
            return (UiGridRow)rows.get(idx);
        }
        
        public UiGridRow getRowByNum(String n) {
            Integer num;
            try {
                num = Integer.parseInt(n);
                return getRowByIndex(num - 1);
            } catch (Exception ex) {
            }
            return null;
        }
        
        public int getRowNumber(UiGridRow row) {
            return rows.indexOf(row) + 1;
        }
        
        public UiGridColumn checkColumn(String code) {
            for (UiGridColumn c : columns) {
                if (c.getCode().equals(code)) {
                    return c;
                }
            }
            return null;
        }
        
        public void rmColumn(UiGridColumn column) {
            columns.remove(column);
            if (columns.isEmpty()) {
                for (UiUnit r : dataRows) {
                    r.remove();
                }
                dataRows.clear();
                for (UiUnit r : qbeRows) {
                    r.remove();
                }
                qbeRows.clear();        
            }
        }
        
        public void setHeader(String h) {
            setUid(getId());
            header = Utils.testYString(h);
            if (header) {
                instructions.add(Utils.instruction(Instructions.hdr));
            } else {
                instructions.add(Utils.instruction(Instructions._hdr));
            }
        }
        
        public void setFooter(String f) {
            setUid(getId());
            footer = Utils.testYString(f);
            if (!mobile) {
                if (footer) {
                    instructions.add(Utils.instruction(Instructions.ftr));
                } else {
                    instructions.add(Utils.instruction(Instructions._ftr));
                }
            }
        }
        
        public void setAllRows(String ar) {
            if (Utils.testYString(ar)) {
                allRows = true;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rall));
                //setNotEmpty();
            }
        }
        
        public void setEmpty() {
            empty = true;
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.em));
        }
        
        public void setNotEmpty() {
            if (empty) {
                empty = false;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._em));
            }
        }
        
        public void setOverflow(String w) {
            if (!mobile) {
                if (!Utils.testEmptyString(w)) {
                    setUid(getId());
                    if ("%".equals(w.substring(0, 1))) {
                        instructions.add(Utils.instruction(Instructions.ow$, w.substring(1)));
                    } else {
                        instructions.add(Utils.instruction(Instructions.ow, w));
                    }
                }
            }
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (child.getParent() != UiGrid.this) {
                if (child instanceof UiGridRow) {
                    if (checkFetch) {
                        checkFetch = false;
                        unitsToCheck.add(this);
                        //addonInstructions.add(Utils.instruction(Instructions.e));
                        addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                        addonInstructions.add(Utils.instruction(Instructions.ftch));
                    }
                    contextRow = child;
                    super.addChild(contextRow);
                    if (rows == qbeRows) {
                        ((UiGridRow)contextRow).setQbe(true);
                    } else if (!hadData) {
                        hadData = true;
                        for (UiGridColumn c : columns) {
                            if (Utils.testEmptyString(c.width)) {
                                setUid(c.getId());
                                instructions.add(Utils.instruction(Instructions.wa));
                            }
                        }
                    }
                    rows.add(contextRow);
                    for (UiGridColumn c : columns) {
                        setUid(getId());
                        instructions.add(Utils.instruction(Instructions.scol, c.getId()));
                        UiGridCell cell = new UiGridCell();
                        contextRow.addChild(cell);
                        c.addChild(cell);
                    }
                    setNotEmpty();
                } else if (child instanceof UiGridColumn) {
                    super.addChild(child);
                    if (!columns.contains((UiGridColumn) child)) {
                        columns.add((UiGridColumn) child);
                    }
                } else if (child instanceof UiGroup) {
                    super.addChild(child);
                    ((UiGroup)child).setState("Hidden");
                }
            }
        }

        public void setSelectionMode(String mode) {
            Defaults.SelectionMode m;
            try {
                m = Defaults.SelectionMode.valueOf(mode);
                selMode = m;
                setUid(getId());
                if (m == Defaults.SelectionMode.None) {
                    instructions.add(Utils.instruction(Instructions._sel));
                } else {
                    instructions.add(Utils.instruction(Instructions.sel, mode));
                }
            } catch (Exception ex) {

            }
        }
        
        public boolean getQbeMode() {
            return qbeMode;
        }
        
        public void setQbeMode(boolean mode) {
            setUid(getId());
            if (!qbeMode && mode) {
                qbeMode = true;
                rows = qbeRows;
                instructions.add(Utils.instruction(Instructions.qbe));
            } else if (qbeMode && !mode) {
                qbeMode = false;
                rows = dataRows;
                instructions.add(Utils.instruction(Instructions._qbe));
            }
            for (UiGridColumn c : columns) {
                c.setQbeMode(qbeMode);
            }
            if (qbeMode && rows.isEmpty()) this.addChild(new UiGridRow());
        }
        
        public void clear() {
            allRows = false;
            this.currentRow = null;
            this.contextRow = null;
            this.currentColumn = null;
            for (UiUnit r : dataRows) {
                r.remove();
            }
            dataRows.clear();
            for (UiGridColumn c : columns) {
                c.clear();
            }
            setQbeMode(false);
            setEmpty();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.clr));
        }
        
        public void reqAddRow() {
            UiUnit saveRow = contextRow;
            addChild(new UiGridRow());
            contextRow = saveRow;
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.srow, contextRow.getId()));
        }
        
        public void reqNewRow() {
            UiGridRow r = new UiGridRow();
            addChild(r);
            if (!mobile) {
                setUid(r.getId());
                instructions.add(Utils.instruction(Instructions.curf));
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.acel));
            }
        }
        
        public void reqSummary() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("GetTotal"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            totalEvent.add(new SingleEvent(sep)); 
        }
        
        public void reqLMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColumnLMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqRMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CellRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        @Override
        public void postCheck() {
            checkFetch = true;
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiGridColumn - "столбец списка"                            *
 *                                                                      *
 ************************************************************************/     
    
    public class UiGridColumn extends UiGridChild implements UiFormCurrent {
        
        String title = "";
        String width = null;
        ColumnAlign align = ColumnAlign.Left;
        String hint = null;
        String style = null;
        String coloring = null;
        String colorStyle = null;
        String sort = null;
        String group = null;
        boolean calcWidth = true;
        int calcStatus = 0;
        int autoWidth = 0;
        String maxValue = "";
        private HashMap<UiGridRow, UiGridCell> cells;
        private final HashMap<UiGridRow, UiGridCell> dataCells;
        private final HashMap<UiGridRow, UiGridCell> qbeCells;
        private final ArrayList<JsonObject> addonInstructions;
        
        public UiGridColumn(String code) {
            super(code);
            dataCells = new HashMap<>();
            qbeCells = new HashMap<>();
            cells = dataCells;
            addonInstructions = new ArrayList<>();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.col;
        }
        
        @Override
        public void preRemove() {
            UiGrid grid = this.getGrid();
            grid.rmColumn(this);
        }
        
        public UiGridCell getCell(UiGridRow row) {
            return cells.get(row);
        }
        
        public void setQbeMode(boolean mode) {
            cells = mode ? qbeCells : dataCells;
            if (!mode) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._qbe));
            }
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (UiGridColumn.this != child.getParent()) {
                if (child instanceof UiGridCell) {
                    UiGridCell cell = (UiGridCell) child;
                    children.add(cell);
                    units.put(cell.getId(), cell);
                    cell.setParent(this);
                    cells.put(cell.getRow(), cell);
                    if (!mobile) {
                        //cell.setHint(hint);
                        cell.setStyle(style);
                        cell.setColoringStyle(colorStyle);
                        cell.setColoring(coloring);
                    }
                }
            }
        }    
        
        public void setTitle(String title) {
            if (title != null) {
                setUid(getId());
                this.title = title;
                instructions.add(Utils.instruction(Instructions.ttl, title));
                //autoWidth = proportional(title);
                if (calcWidth && (calcStatus == 0)) {
                    unitsToCheck.add(this);
                }
            } 
        }
        
        public void setAlign(String align) {
            if (align != null) {
                try {
                    this.align = ColumnAlign.valueOf(align);
                } catch (Exception ex) {
                    this.align = ColumnAlign.Left;
                }
                String s = "left";
                switch (this.align) {
                    case Right:
                        s = "right";
                        break;
                    case Center:
                        s = "center";
                        break;
                }
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.alig, s));
            }
        }
        
        public void setHint(String hint) {
            if (hint != null) {
                setUid(getId());
                if (Utils.testEmptyString(hint)) {
                    instructions.add(Utils.instruction(Instructions._hnt));
                } else {
                    instructions.add(Utils.instruction(Instructions.hnt, hint));
                }
            }
        }
        
        public void setStyle(String style) {
            if (style != null) {
                this.style = style;
            }
        }
        
        public void setColoring(String coloring) {
            if (coloring != null) {
                this.coloring = coloring;
            }
        }
        
        public void setColoringStyle(String style) {
            if (style != null) {
                this.colorStyle = style;
            }
        }
        
        public void setSort(String sort) {
            if (sort != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._srt));
                try {
                    Integer s = Integer.parseInt(sort);
                    this.sort = sort;
                    String direction = s < 0 ? "desc" : "asc";
                    int abs = Math.abs(s);
                    if ((abs > 0) && (abs <= 9)) {
                        instructions.add(Utils.instruction(Instructions.srt, direction + String.valueOf(abs)));
                    } else if (abs > 9) {
                        instructions.add(Utils.instruction(Instructions.srt, direction));
                    }
                } catch (Exception ex) {
                }
            }
        }
        
        public void setWidth(String width) {
            Integer w;
            if (width != null) {
                this.width = width;
                setUid(getId());
                if (Utils.testEmptyString(width)) {
                    if (protocol != 3) {
                        instructions.add(Utils.instruction(Instructions.w, "auto"));
                    }
                    instructions.add(Utils.instruction(Instructions.show));
                } else {
                    try {
                        w = Integer.parseInt(width);
                        if (w <= 0) {
                            if (protocol != 3) {
                                instructions.add(Utils.instruction(Instructions.w, "0"));
                            }
                            instructions.add(Utils.instruction(Instructions.hide));
                        } else if (w > 0) {
                            if (protocol != 3) {
                                instructions.add(Utils.instruction(Instructions.w, width));
                            }
                            instructions.add(Utils.instruction(Instructions.show));
                        }
                    } catch (Exception ex) {
                        if (protocol != 3) {
                            instructions.add(Utils.instruction(Instructions.w, "0"));
                        }
                        instructions.add(Utils.instruction(Instructions.hide));
                    }
                }
            }
        }
        
        public void setEnabled(String enable) {
            if (enable != null) {
                setUid(getId());
                if (Utils.testYString(enable)) {
                    instructions.add(Utils.instruction(Instructions.en));
                } else {
                    instructions.add(Utils.instruction(Instructions.dis));
                }
            }
        }
        
        public void setFooter(String footer) {
            if (footer != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.foot, footer));
            }
        }   
        
        public void clear() {
            if (cells == dataCells) {
                for (UiGridCell c : cells.values()) {
                    c.remove();
                }
                cells.clear();
            } else if (cells == qbeCells) {
                for (UiGridCell c : cells.values()) {
                    c.clearValue();
                    c.rmButtons();
                }
            }
            addonInstructions.clear();
            maxValue = "";
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.clr));
        }
        
        public void updateWidth(String val) {
            if (mobile) {
                int w = proportional(val);
                if (w > autoWidth) {
                    autoWidth = w;
                    if ((calcWidth) && (calcStatus == 1)) {
                        unitsToCheck.add(this);
                    }
                }
            } else if (protocol == 3) {
                if (val.length() > maxValue.length()) {
                    maxValue = val;
                    unitsToCheck.add(this);
                }
            }
        }
        
        @Override
        public void unsetCurrent() {
            getGrid().setCurrentColumn(null);
            getForm().setCurrentUnit(null);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("GotoCell"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Column", new StringParameter(""));
            UiGridRow r = getGrid().getCurrentRow();
            if (r == null) {
                sep.put("Context", new StringParameter(""));
            } else {
                sep.put("Context", r.getContext());
            }
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void preCheck() {
            if (mobile) {
                if (calcWidth && (calcStatus == 0 || calcStatus == 1)) {
                    addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                    addonInstructions.add(Utils.instruction(Instructions.w, String.valueOf(autoWidth)));
                }
            } else if (protocol == 3) {
                addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                addonInstructions.add(Utils.instruction(Instructions.maxv, maxValue));
            }
        }
        
        @Override
        public void postCheck() {
            calcStatus++;
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }

        public void setCell(UiGridRow row) {
            try {
                getCell(row).asCurrent();
            } catch (Exception ex) {}    
        }
        
        public void reqLMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColumnLMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Column", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqDC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColumnDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Column", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqRMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColumnRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Column", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqMove(String left) {
            String code = (Utils.testEmptyString(left)) ? "" : ((UiGridColumn) units.get(left)).getCode();
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(this.getWindow().getCode()));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("LeftNeighbor"));
            sep.put("Value", new StringParameter(code));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqGroup(String grp) {
            UiUnit u = units.get(grp);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("WindowID", new StringParameter(this.getWindow().getCode()));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("Group"));
            if (u instanceof UiGrid) {
                sep.put("Value", new StringParameter(""));
            } else {
                sep.put("Value", new StringParameter(u.getCode()));
            }
            totalEvent.add(new SingleEvent(sep));
            changeParent(u);
        }
        
        public void reqPropertyWidth(String width) {
            this.width = width;
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("Element", new StringParameter(getCode()));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Property", new StringParameter("ColWidth"));
            sep.put("Value", new StringParameter(width));
            totalEvent.add(new SingleEvent(sep));
        };
        
        public void reqSpace() {
            try {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("SpaceBar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Element", new StringParameter(getCode()));
                totalEvent.add(new SingleEvent(sep));
            } catch (Exception ex) {
            }
        }
        
        public void reqEnter() {
            try {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("EnterBar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Element", new StringParameter(getCode()));
                totalEvent.add(new SingleEvent(sep));
            } catch (Exception ex) {
            }
        }
        
        public void reqF9() {
            try {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("F9Bar"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Element", new StringParameter(getCode()));
                totalEvent.add(new SingleEvent(sep));
            } catch (Exception ex) {
            }
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiGridRow - "строка списка"                                *
 *                                                                      *
 ************************************************************************/     
    
    public class UiGridRow extends UiGridChild {
        
        final ArrayList<UiUnit> cells;
        private boolean qbeRow = false;
        
        public UiGridRow() {
            super();
            cells = new ArrayList<>();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.row;
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (child instanceof UiGridCell) {
                ((UiGridCell)child).setRow(this);
                cells.add(child);
                instructions.add(Utils.instruction(Instructions.cell, child.getId()));
            }
        }
        
        public void setQbe(boolean qbe) {
            qbeRow = qbe;
        }
        
        public boolean isQbeRow() {
            return qbeRow;
        }
        
        public void select(String sel) {
            if (Utils.testYString(sel)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.sel));
            }
        }
        
        public CollectionParameter getContext() {
            CollectionParameter context = new CollectionParameter();
            for (UiUnit u : cells) {
                UiGridCell cell = (UiGridCell) u;
                context.put(cell.getParent().getCode(), cell.getValue());
            }
            return context;
        }    
        
        public void reqCheck(String c) {
            SingleEventParameters sep;
            UiGrid grid = getGrid();
            if (this != grid.getCurrentRow()) {
                sep = new SingleEventParameters();
                sep.put("e", new StringParameter("GotoCell"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("RowNum", new StringParameter(String.valueOf(grid.getRowNumber(this))));
                grid.setCurrentRow(this);
                sep.put("Context", getContext());
                totalEvent.add(new SingleEvent(sep));    
            }
            sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CheckRow"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            //sep.put("Row", new StringParameter(String.valueOf(getGrid().getRowNumber(this))));
            sep.put("IsCheck", new StringParameter(c));
            //sep.put("Context", getContext());
            totalEvent.add(new SingleEvent(sep)); 
        }
        
        public void reqChange() {
            SingleEventParameters sep;
            UiGrid grid = getGrid();
            if (this != grid.getCurrentRow()) {
                sep = new SingleEventParameters();
                sep.put("e", new StringParameter("GotoCell"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("RowNum", new StringParameter(String.valueOf(grid.getRowNumber(this))));
                grid.setCurrentRow(this);
                sep.put("Context", getContext());
                totalEvent.add(new SingleEvent(sep));    
            }
        }
        
    }       
    

    
/************************************************************************
 *                                                                      *
 *     Класс UiGridCell - "ячейка списка"                               *
 *                                                                      *
 ************************************************************************/     
    
    public class UiGridCell extends UiGridChild {
        
        String value = "";
        UiGridRow row;
        private boolean qbeCell = false;
        ColumnAlign align = ColumnAlign.Left;
        
        public UiGridCell() {
            super();
        }
            
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.cell;
        }
        
        public boolean isQbeCell() {
            return qbeCell;
        }
        
        public void setRow(UiGridRow row) {
            qbeCell = row.isQbeRow();
            this.row = row;
        }
        
        public UiGridRow getRow() {
            return row;
        }
        
        public void clearValue() {
            assignValue("");
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.v, this.value));
        }

        public void setValue(String value) {
            if (value != null) {
                assignValue(value);
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.v, this.value));
                if (qbeCell) {
                    UiGrid grid = (UiGrid)row.getParent();
                    if (grid.qbeRows.size() > 0 && grid.qbeRows.indexOf(row) == grid.qbeRows.size() - 1) {
                        grid.reqAddRow();
                    }
                }
                if (mobile || protocol == 3) {
                    ((UiGridColumn) this.getParent()).updateWidth(this.value);
                }
            }
        }
        
        public void assignValue(String value) {
            this.value = value;
        }
        
        public String getValue() {
            return value;
        }
        
        public void asCurrent() {
            unitToFocus = getId();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.curr));
            if (getGrid().getQbeMode()) {
                instructions.add(Utils.instruction(Instructions.edit, "Y"));
            }
        }
        
        private void rmButtons() {
            UiUnit b = null;
            for (UiUnit u : children) {
                if (u instanceof UiButtonBar) {
                    b = u;
                    break;
                }
            }
            if (!(b == null)) {
                b.remove();
            }
        }
        
        public void setButtons(String buttons) {
            if (buttons != null) {
                rmButtons();
                UiCollection c = new UiCollection(buttons);
                if (!c.isEmpty()) {
                    UiButtonBar bbar = new UiButtonBar();
                    this.addChild(bbar);
                    String s = value;
                    if (protocol == 3) {
                        if (c.getSize() > 1) {
                            UiCellButton b = new UiCellButton("more");
                            bbar.addChild(b);
                            b.setState("More");
                        }
                    }
                    while (!c.isLast()) {
                        UiCellButton b = new UiCellButton(c.getCode());
                        bbar.addChild(b);
                        if (mobile) {
                            s = s + "XXXx";
                        } else {
                            b.setHint(c.getValue());
                        }
                        c.getNext();
                    }
                    if (mobile) {
                        ((UiGridColumn) this.getParent()).updateWidth(s);
                    }  
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.cb));
                }
            }
        }
        
        public void setEditable(String e) {
            if (e != null) {
                unitToFocus = getId();
                setUid(getId());
                String s = Utils.testYString(e) ? e : "";
                instructions.add(Utils.instruction(Instructions.edit, s));
            }
        }
        
        public void setStyle(String style) {
            if (style != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._sty));
                if (!Utils.testEmptyString(style)) {
                    String s = getForm().parseMultiCss(style);
                    if (!Utils.testEmptyString(s)) {
                        setUid(getId());
                        instructions.add(Utils.instruction(Instructions.sty, s));
                    }
                }
            }
        }
        
        public void setHint(String h) {
            if (h != null) {
                setUid(getId());
                if (Utils.testEmptyString(h)) {
                    instructions.add(Utils.instruction(Instructions._hnt));
                } else {
                    instructions.add(Utils.instruction(Instructions.hnt, h));
                }
            }
        }
        
        public void setColoring(String c) {
            if (c != null) {
                setUid(getId());
                if (Utils.testEmptyString(c)) {
                    instructions.add(Utils.instruction(Instructions._mkp));
                } else {
                    instructions.add(Utils.instruction(Instructions.mkup, c));
                }
            }
        }
        
        public void setColoringStyle(String p) {
            if (p != null) {
                String s = getForm().parseMultiCss(p);
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.mkps, s));
            }
        }
        
        public void setAlign(String align) {
            if (align != null) {
                try {
                    this.align = ColumnAlign.valueOf(align);
                } catch (Exception ex) {
                    this.align = ColumnAlign.Left;
                }
                String s = "left";
                switch (this.align) {
                    case Right:
                        s = "right";
                        break;
                    case Center:
                        s = "center";
                        break;
                }
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.alig, s));
            }
        }
        
        private void makeCurrent(boolean dbl) {
            UiGrid grid = getGrid();
            if ((getRow() == grid.getCurrentRow()) && (getParent() == grid.getCurrentColumn())) {
                if (mobile && dbl)  {
                    reqDC();
                }    
            } else {
                if (getGrid().getQbeMode()) {
                    unitToFocus = getId();
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.edit, "Y"));
                }
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("GotoCell"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                if (getRow() != grid.getCurrentRow()) {
                    sep.put("RowNum", new StringParameter(String.valueOf(grid.getRowNumber(getRow()))));
                    grid.setCurrentRow(getRow());
                } 
                if (getParent() != grid.getCurrentColumn()) {
                    sep.put("Column", new StringParameter(getParent().getCode()));
                    grid.setCurrentColumn(getParent());
                    getForm().setCurrentUnit(getParent());
                }    
                sep.put("Context", getRow().getContext());
                totalEvent.add(new SingleEvent(sep));    
            }
        }
        
        public void reqCurrent() {
            makeCurrent(true);
        }
        
        public void reqRMB() {
            if (!mobile) {
                reqCurrent();
            }
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CellRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Row", new StringParameter(String.valueOf(getGrid().getRowNumber(getRow()))));
            sep.put("Column", new StringParameter(getParent().getCode()));
            sep.put("Context", getRow().getContext());
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqDC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("DefaultAction"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqNewValue(String val) {
            assignValue(val);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NewValue"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getParent().getCode()));
            sep.put("Value", new StringParameter(getValue()));
            sep.put("IsCheck", new StringParameter("Y"));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqInterValue(String val) {
            assignValue(val);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NewValue"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getParent().getCode()));
            sep.put("Value", new StringParameter(getValue()));
            sep.put("IsCheck", new StringParameter("N"));
            totalEvent.add(new SingleEvent(sep));  
        }
    }      
    
    

/************************************************************************
 *                                                                      *
 *     Класс UiCellButton - "кнопка в ячейке"                           *
 *                                                                      *
 ************************************************************************/    
    
    public class UiCellButton extends UiBaseButton {
        
        public UiCellButton(String code) {
            super(code);
        }
        
        @Override
        public void reqPress() {
            UiUnit cell = this.getParent().getParent();
            UiUnit column = cell.getParent();
            ((UiGridCell) cell).makeCurrent(false);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CellButton"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Column", new StringParameter(column.getCode()));
            sep.put("Name", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
    }            
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiGroup - "группирующий элемент"                           *
 *                                                                      *
 ************************************************************************/     
    
    public class UiGroup extends UiFormChild {
        
        GroupType type = GroupType.Tab;
        GroupState state = GroupState.Hidden;
        String title = "";
        
        public UiGroup() {
            super();
        }
        
        public UiGroup(String code) {
            super(code);
        }
                      
        public void setTitle(String title) {
            if (title != null) {
                setUid(getId());
                this.title = title;
                if (title.equals("")) {
                    instructions.add(Utils.instruction(Instructions._ttl));
                } else {
                    instructions.add(Utils.instruction(Instructions.ttl, title));
                }
            }
        }
        
        public void setState(GroupState gs) {
            this.state = gs;
            setUid(getId());
            if (this.state == GroupState.Hidden) {
                instructions.add(Utils.instruction(Instructions.hide));
            } else {
                instructions.add(Utils.instruction(Instructions.show));
            }
        }
        
        public void setState(String state) {
            if (state != null) {
                try {
                    setState(GroupState.valueOf(state));
                } catch (Exception ex) {
                    
                }
            }
        }
        
        public void setPromptsPos(String pos) {
            if (!(mobile || protocol == 3 || pos == null)) {
                setUid(getId());
                if (pos.equals("Left")) {
                    instructions.add(Utils.instruction(Instructions.prml));
                } else if (pos.equals("Up")) {
                    instructions.add(Utils.instruction(Instructions.prmt));
                }
            }
        }
        
        public void setHint(String hint) {
            
        }
        
        public void reqLMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("GroupLMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Group", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqRMB() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("GroupRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Group", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqDC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("GroupDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Group", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
    }     
    

    
/************************************************************************
 *                                                                      *
 *     Класс UiGridGroup - "группа столбцов списка"                     *
 *                                                                      *
 ************************************************************************/     
    
    public class UiGridGroup extends UiGroup {
        
        public UiGridGroup(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.gcol;
        }
        
        @Override
        public void setTitle(String title) {
            if (!(mobile || protocol == 3)) {
                super.setTitle(title);
            }
        }
        
        @Override
        public void setPromptsPos(String pos) {
            
        }
        
        @Override
        public void setHint(String hint) {
            if (hint != null) {
                setUid(getId());
                if (Utils.testEmptyString(hint)) {
                    instructions.add(Utils.instruction(Instructions._hnt));
                } else {
                    instructions.add(Utils.instruction(Instructions.hnt, hint));
                }
            }
        }
        
        public void groupPromptsPos(String pos) {
            super.setPromptsPos(pos);
        }
        
    }    
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiFixedGroup - "группа непрокручиваемых столбцов списка"   *
 *                                                                      *
 ************************************************************************/     
    
    public class UiFixedGroup extends UiGridGroup {
        
        public UiFixedGroup(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.gfix;
        }
        
    }            
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiOverflowGroup - "группа столбцов типа Overflow"          *
 *                                                                      *
 ************************************************************************/     
    
    public class UiOverflowGroup extends UiGridGroup {
        
        public UiOverflowGroup(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.govr;
        }
        
        @Override
        public void setPromptsPos(String pos) {
            if (protocol != 3) {
                groupPromptsPos(pos);
            }
        }
        
    }            

    
    
/************************************************************************
 *                                                                      *
 *     Класс UiQuestion - Окно с вопросами и кнопками                   *
 *                                                                      *
 ************************************************************************/
    public class UiQuestion extends UiUnit {
        
        public UiQuestion() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.qstn;
        }
        
        public void setText(String text) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.txt, text));
        }
        
        public void setTitle(String title) {
            if (title != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.ttl, title));
            }
        }
        
        public void setButtons(String buttons) {
            if (!Utils.testEmptyString(buttons)) {
                UiCollection c = new UiCollection(buttons);
                if (!c.isEmpty()) {
                    while (!c.isLast()) {
                        UiButton b = new UiButton(c.getCode());
                        addChild(b);
                        b.setLabel(c.getValue());
                        c.getNext();
                    }
                }
            }
        }
        
        public void initQuestion() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.ini));
        }
        
    }

    

/************************************************************************
 *                                                                      *
 *     Класс UiMessage - Системное сообщение                            *
 *                                                                      *
 ************************************************************************/
    public class UiMessage extends UiUnit {
        
        public UiMessage() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.msg;
        }
        
        public void setTitle(String t) {
            if (t != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.ttl, t));
            }
        }
        
        public void setType(String t) {
            String s = "fail";
            if (!Utils.testEmptyString(t)) {
                s = t.toLowerCase();
            }
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.type, s));
        }
        
        public void setText(String t) {
            if (t != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.txt, t));
            }
        }
        
        public void setNotes(String n) {
            if (n != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.note, n));
            }
        }
        
        public void show() {
            if (protocol == 3) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.show));
            }
        }
        
    }
    

/************************************************************************
 *                                                                      *
 *     Класс UiCalendar - Календарь                                     *
 *                                                                      *
 ************************************************************************/
    public class UiCalendar extends UiUnit {
        
        private String form = null;
        
        public UiCalendar() {
            super();
        }    
             
        public void setForm(String form) {
            this.form = form;
        }
        
        public String getForm() {
            return form;
        }
        
        public void setSpecialDates(String dates) {
            if (!Utils.testEmptyString(dates)) {
                UiCollection sd = new UiCollection(dates);
                if (!sd.isEmpty()) {
                    setUid(getId());
                    while (!sd.isLast()) {
                        instructions.add(Utils.instruction(Instructions.csd, sd.getCode()));
                        String t;
                        switch (sd.getValue().toUpperCase()) {
                            case "ПР":
                                t = "p";
                                break;
                            case "ВЫХ":
                                t = "v";
                                break;
                            case "РАБ":
                                t = "r";
                                break;
                            default:
                                t = "r";
                        }
                        instructions.add(Utils.instruction(Instructions._csd, t));
                        sd.getNext();
                    }
                }
            }
        }
        
        public void setDefault(String dt) {
            String def = dt;
            Integer currentYear;
            Integer currentMonth;
            Integer currentDay;
            Integer currentHour;
            Integer currentMin;
            Integer currentSec;
            if (Utils.testEmptyString(dt)) {
                def = "20180518112517";
            }
            setUid(getId());
            String d = Utils.string2String(def);
            Calendar cal = Calendar.getInstance();
            cal.setTime(new Date());
            try {
                currentYear = Integer.parseInt(d.substring(0, 4));
            } catch (Exception ex) {
                currentYear = cal.get(Calendar.YEAR);
            }
            instructions.add(Utils.instruction(Instructions.cyr, String.valueOf(currentYear)));
            try {
                currentMonth = Integer.parseInt(d.substring(4, 6));
            } catch (Exception ex) {
                currentMonth = cal.get(Calendar.MONTH) + 1;
            }
            instructions.add(Utils.instruction(Instructions.cmon, String.valueOf(currentMonth)));
            try {
                currentDay = Integer.parseInt(d.substring(6, 8));
            } catch (Exception ex) {
                currentDay = cal.get(Calendar.DAY_OF_MONTH);
            }
            instructions.add(Utils.instruction(Instructions.cday, String.valueOf(currentDay)));
            try {
                currentHour = Integer.parseInt(d.substring(8, 10));
            } catch (Exception ex) {
                currentHour = cal.get(Calendar.HOUR_OF_DAY);
            }
            instructions.add(Utils.instruction(Instructions.chr, String.valueOf(currentHour)));
            try {
                currentMin = Integer.parseInt(d.substring(10, 12));
            } catch (Exception ex) {
                currentMin = cal.get(Calendar.MINUTE);
            }
            instructions.add(Utils.instruction(Instructions.cmin, String.valueOf(currentMin)));
            try {
                currentSec = Integer.parseInt(d.substring(12, 14));
            } catch (Exception ex) {
                currentSec = cal.get(Calendar.SECOND);
            }
            instructions.add(Utils.instruction(Instructions.csec, String.valueOf(currentSec)));
            instructions.add(Utils.instruction(Instructions.crfr));
        }
        
        public void setDateDetail(String dd) {
            setUid(getId());
            if (dd == null) {
                instructions.add(Utils.instruction(Instructions.cdd, DateDetail.D.toString()));
            } else {
                try {
                    DateDetail d = DateDetail.valueOf(dd);
                    instructions.add(Utils.instruction(Instructions.cdd, d.toString()));
                } catch (Exception ex) {
                    instructions.add(Utils.instruction(Instructions.cdd, "N"));
                }
            }
        }
        
        public void setTimeDetail(String td) {
            setUid(getId());
            //instructions.add(Utils.instruction(Instructions.ctd, "S"));
            if (td == null) {
                instructions.add(Utils.instruction(Instructions.ctd, "N"));
            } else {
                try {
                    TimeDetail t = TimeDetail.valueOf(td);
                    instructions.add(Utils.instruction(Instructions.ctd, t.toString()));
                } catch (Exception ex) {
                    instructions.add(Utils.instruction(Instructions.ctd, "N"));
                }
            }
        }

        @Override
        public Instructions getCreateInstruction() {
            return Instructions.cal;
        }
        
        public void reqClose() {
            remove();
        }
        
        public void reqChoice  (String day,
                                String month,
                                String year,
                                String hour,
                                String minute,
                                String second) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("Calendar"));
            sep.put("FormID", new StringParameter(getForm()));
            Integer m;
            m = Integer.parseInt(month);
            m += 1;
            String dt = String.format("%04d", Integer.parseInt(year)) +
                        String.format("%02d", m) +
                        String.format("%02d", Integer.parseInt(day)) +
                        String.format("%02d", Integer.parseInt(hour)) +
                        String.format("%02d", Integer.parseInt(minute)) +
                        String.format("%02d", Integer.parseInt(second));
            sep.put("DateTime", new StringParameter(dt));
            totalEvent.add(new SingleEvent(sep)); 
            remove();
        }
        
    }    
        

    
/************************************************************************
 *                                                                      *
 *     Класс UiLov - Список возможных значений                          *
 *                                                                      *
 ************************************************************************/
    public class UiLov extends UiUnit {
        
        private final HashMap<String, UiLovColumn> columns;
        private final ArrayList<UiLovRow> rows;
        private int state = 1;
        private UiLovRow contextRow = null;
        private boolean allRows = false;
        private boolean extFilter = false;
        private UiLovColumn lastSortColumn = null;
        private UiLovRow rowToExpand = null; 
        private boolean caseSensitive = true;
        private final ArrayList<JsonObject> addonInstructions;
        
        public UiLov() {
            super();
            columns = new HashMap<>();
            rows = new ArrayList<>();
            addonInstructions = new ArrayList<>();
        }    
                
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.lov;
        }
        
        public void setColumns(String columns) {
            if (!Utils.testEmptyString(columns)) {
                UiCollection c = new UiCollection(columns);
                if (!c.isEmpty()) {
                    while (!c.isLast()) {
                        UiLovColumn col = new UiLovColumn(c.getCode());
                        this.columns.put(c.getCode(), col);
                        setUid(getId());
                        instructions.add(Utils.instruction(Instructions.col, col.getId()));
                        col.setHeader(c.getValue());
                        c.getNext();
                    }    
                }
            }
        }
        
        public void setTitle(String title) {
            if (title != null) {
                setUid(getId());
                if (Utils.testEmptyString(title)) {
                    instructions.add(Utils.instruction(Instructions._ttl));
                } else {
                    instructions.add(Utils.instruction(Instructions.ttl, title));
                }
            }
        }
        
        public int getState() {
            return state;
        }
        
        public void setState(int state) {
            this.state = state;
        }
        
        public void initialize() {
            if (protocol != 3) {
                for (UiLovColumn c : columns.values()) {
                    setUid(c.getId());
                    if (mobile) {
                        instructions.add(Utils.instruction(Instructions.w, c.getWidth()));
                    } else {
                        instructions.add(Utils.instruction(Instructions.w));
                    }
                }
            }
            if (!mobile) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.ini));
            }
        }
        
        public void setSensitivity(String s) {
            if (s != null) {
                caseSensitive = Utils.testYString(s);
            }
        }
        
        public void addRow(String code, String type, String parent) {
            contextRow = new UiLovRow(code);
            //instructions.add(Utils.instruction(Instructions.row, contextRow.getId()));
            if (Utils.testEmptyString(parent)) {
                addChild(contextRow);
                rows.add(contextRow);
                setUid(contextRow.getId());
                instructions.add(Utils.instruction(Instructions.lvl, "0"));
            } else {
                UiUnit par = findChild(parent, clLovRow);
                if (par != null) {
                    par.addChild(contextRow);
                }
            }
            contextRow.setType(type);
        }
        
        public void setContext(String context) {
            if (!Utils.testEmptyString(context)) {
                setUid(getId());
                UiCollection c = new UiCollection(context);
                if (!c.isEmpty()) {
                    while (!c.isLast()) {
                        UiLovColumn col = columns.get(c.getCode());
                        ArrayList<UiLovContent> cont = new ArrayList<>();
                        String tmp = Utils.string2String(c.getValue());
                        while (StringUtils.contains(tmp, "[ico:")) {
                            String txt = StringUtils.substringBefore(tmp, "[ico:");
                            if (!Utils.testEmptyString(txt)) {
                                cont.add(new UiLovContentText(txt));
                            }    
                            tmp = StringUtils.removeStart(tmp, txt);
                            cont.add(new UiLovContentIcon(StringUtils.substringAfter(StringUtils.substringBefore(tmp, "]"), "[ico:")));
                            tmp = StringUtils.substringAfter(tmp, "]");
                        }
                        if (!Utils.testEmptyString(tmp)) {
                            cont.add(new UiLovContentText(tmp));
                        }
                        String cellText = "";
                        String colText = "";
                        String html = "";
                        for (int i = 0; i < cont.size(); i++) {
                            UiLovContent lc = cont.get(i);
                            if (lc instanceof UiLovContentIcon) {
                                html = html + "<div class='lov-cell-icon ico-noicon " + Utils.encodeBrackets(lc.getValue()) + " lui2-unit'></div>";
                                colText = colText + "XXXX";
                            } else if (lc instanceof UiLovContentText) {
                                html = html + "<div class='lov-cell-text lui2-unit'>" + Utils.encodeBrackets(lc.getValue()) + "</div>";
                                cellText = cellText + lc.getValue();
                                colText = colText + lc.getValue();
                            }
                        }
                        col.setText(colText);
                        UiLovCell cell = new UiLovCell();
                        col.addCell(cell);
                        contextRow.addCell(cell);
                        cell.setText(cellText);
                        cell.setHtml(html);
                        instructions.add(Utils.instruction(Instructions.k, col.getId()));
                        if (protocol == 3) {
                            instructions.add(Utils.instruction(Instructions.v, cellText));
                        } else {
                            instructions.add(Utils.instruction(Instructions.v, html));
                        }
                        c.getNext();
                    }
                }
            }
        }
             
        public void setAllRows(String isLast) {
            if (rowToExpand == null) {
                if (Utils.testYString(isLast)) {
                    allRows = true;
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.rall));
                }
            } else {
                rowToExpand.setAllRows();
            }
        }
        
        private UiLovColumn getColumnById(String id) {
            for (UiLovColumn c : columns.values()) {
                if (c.getId().equals(id)) {
                    return c;
                }
            }
            return null;
        }
        
        public void setFilter(String col, String filter) {
            UiLovColumn c = getColumnById(col);
            if (c != null) {
                c.setFilter(filter, caseSensitive);
            }
        }
        
        public void applyFilter() {
            if (allRows & (!extFilter)) {
                refreshView();
            } else {
                extFilter = false;
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("LivePickLOV"));
                CollectionParameter cp = new CollectionParameter();
                for (UiLovColumn c : columns.values()) {
                    cp.put(c.getCode(), c.getFilter());
                    if (!Utils.testEmptyString(c.getFilter())) {
                        extFilter = true;
                    }    
                }
                sep.put("Conditions", cp);
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public UiLovColumn getLastSortColumn() {
            return lastSortColumn;
        }
        
        public void setSortColumn(String col) {
            UiLovColumn c = getColumnById(col);
            if (c != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._srt));
                c.changeSort();
                lastSortColumn = c;
                if (allRows) {
                    refreshView();
                } else {
                    SingleEventParameters sep = new SingleEventParameters();
                    sep.put("e", new StringParameter("SetSortLOV"));
                    String s = "";
                    if (c.getSort() == SortDirection.Desc) {
                        s = "-";
                    }
                    sep.put("OrderBy", new StringParameter(s + c.getCode()));
                    totalEvent.add(new SingleEvent(sep));
                }
            }
        }
        
        public void clear() {
            setState(2);
            this.rmChildren();
            for (UiLovColumn c : columns.values()) {
                c.clear();
            }
            allRows = false;
            contextRow = null;
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.clr));
        }
        
        @Override
        public ArrayList<UiUnit> getAllChildren() {
            ArrayList<UiUnit> allChildren = new ArrayList<>();
            getAllChildren_(allChildren);
            return allChildren;
        }
        
        private void getAllChildren_(ArrayList<UiUnit> a) {
            for (UiUnit u : children) {
                a.add(u);
                u.getAllChildren_(a);
            }
        }
                
        public void refreshView() {
            setState(2);
            ArrayList<UiLovRow> view = null;
            if (lastSortColumn != null) {
                view = new ArrayList<>();
                for (UiLovCell c : lastSortColumn.getSortedCells()) {
                    view.add(c.getRow());
                }
            }
            for (UiLovColumn col : columns.values()) {
                if (view == null) {
                    view = new ArrayList<>();
                    for (UiLovRow r : col.getFilteredRows(caseSensitive)) {
                        view.add(r);
                    }
                } else {
                    view.retainAll(col.getFilteredRows(caseSensitive));
                }
            }
            if (view != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.clr));
                ArrayList<UiLovRow> level0 = new ArrayList<>();
                for (UiLovRow row : view) {
                    if (row.getLevel() == 0) {
                        level0.add(row);
                    }
                }
                if (level0.size() > 0) {
                    for (UiLovRow row : level0) {
                        contextRow = row;
                        instructions.add(Utils.instruction(Instructions.row, contextRow.getId()));
                        setUid(contextRow.getId());
                        instructions.add(Utils.instruction(Instructions.type, contextRow.getType()));
                        instructions.add(Utils.instruction(Instructions.lvl, String.valueOf(contextRow.getLevel())));
                        setUid(getId());
                        for (UiLovCell c2 : contextRow.getCells()) {
                            instructions.add(Utils.instruction(Instructions.k, c2.getColumn().getId()));
                            if (protocol == 3) {
                                instructions.add(Utils.instruction(Instructions.v, c2.getText()));
                            } else {
                                instructions.add(Utils.instruction(Instructions.v, c2.getHtml()));
                            }    
                        }
                    }
                }
            }
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.rall));
        }
        
        public void reset() {
            rowToExpand = null;
        }

        public void reqFetch() {
            if (!allRows) {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("NextPartLOV"));
                totalEvent.add(new SingleEvent(sep));
            }    
        }
        
        public void reqExpand(String id) {
            rowToExpand = (UiLovRow)units.get(id);
            if (rowToExpand.isAllRows()) {
                rowToExpand.expand();
            } else {
                rowToExpand.setAllRows();
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("ExpandLOV"));
                sep.put("ID", new StringParameter(rowToExpand.getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
            addonInstructions.add(Utils.instruction(Instructions.b, getId()));
            addonInstructions.add(Utils.instruction(Instructions.rfr));
            unitsToCheck.add(this);
        }
        
        public void reqCollapse(String id) {
            ((UiLovRow)units.get(id)).collapse();
        }

        @Override
        public void postCheck() {
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
        // Столбец LOV
        public class UiLovColumn {
            
            private String id;
            private String code;
            private ArrayList<UiLovCell> cells;
            private String maxText = "";
            private SortDirection sort = SortDirection.Asc;
            private String filter = "";
            private Pattern filterPattern = null;
            private ArrayList<UiLovRow> filteredRows;
            private ArrayList<UiLovCell> sortedCells;
            
            public UiLovColumn(String code) {
                this.code = code;
                id = Utils.generateId();
                cells = new ArrayList<>();
            }
            
            public String getId() {
                return id;
            }
            
            public String getCode() {
                return code;
            }
            
            public void clear() {
                cells.clear();
            }
            
            public void addCell(UiLovCell cell) {
                cells.add(cell);
                cell.setColumn(this);
            }
            
            public void setHeader(String h) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.hdr, h));
                setText(h + "XXXX");
            }
            
            public void setFilter(String filter, boolean sensitive) {
                this.filter = filter;
                String x = filter;
                StringBuilder p = new StringBuilder();
                String sp;
                while (x.length() > 0) {
                    int i = StringUtils.indexOfAny(x, '%', '_');
                    if (i >= 0) {
                        String s = x.substring(0, i);
                        String d = x.substring(i, i + 1);
                        if (!s.equals("")) {
                            sp = sensitive ? s : s.toLowerCase();
                            p.append(Utils.patternSafe(sp));
                        }
                        String delim = d.equals("%") ? ".*" : ".";
                        p.append(delim);
                        x = x.substring(i + 1);
                    } else {
                        if (!x.equals("")) {
                            sp = sensitive ? x : x.toLowerCase();
                            p.append(Utils.patternSafe(sp));
                        }
                        x = "";
                    }
                }
                p.append(".*");
                filterPattern = Pattern.compile(p.toString());
            }
            
            public String getFilter() {
                return filter;
            }
            
            public ArrayList<UiLovRow> getFilteredRows(boolean sensitive) {
                filteredRows = new ArrayList<>();
                if (Utils.testEmptyString(filter)) {
                    for (UiLovCell c : cells) {
                        filteredRows.add(c.getRow());
                    }
                } else {
                    for (UiLovCell c : cells) {
                        String s = sensitive ? c.getText() : c.getText().toLowerCase();
                        Matcher m = filterPattern.matcher(s);
                        if (m.lookingAt()) {
                            filteredRows.add(c.getRow());
                        } 
                    }
                }
                return filteredRows;
            }    
            
            public void setText(String text) {
                if (text.length() > maxText.length()) {
                    maxText = text;
                }
            }
            
            public String getWidth() {
                return String.valueOf(proportional(maxText));
            }
            
            public void changeSort() {
                if (this == UiLov.this.getLastSortColumn()) {
                    if (sort == SortDirection.Asc) {
                        sort = SortDirection.Desc;
                    } else {
                        sort = SortDirection.Asc;
                    }
                }
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.srt, sort.toString().toLowerCase()));
            }    
                
            public SortDirection getSort() {
                return sort;
            }    
            
            public ArrayList<UiLovCell> getSortedCells() {
                sortedCells = new ArrayList<>();
                sortedCells.addAll(cells);
                Collections.sort(sortedCells, new Comparator<UiLovCell>() {
                    @Override
                    public int compare(UiLovCell c1, UiLovCell c2) {
                        if (sort == SortDirection.Asc) {
                            return c1.getText().compareTo(c2.getText());
                        } else {
                            return c2.getText().compareTo(c1.getText());
                        }
                    }
                });
                return sortedCells;
            }
        }

        
        // Ячейка LOV
        public class UiLovCell {

            private UiLovColumn column;
            private UiLovRow row;
            private String text;
            private String html;
            
            public UiLovCell() {
                
            }

            public void setRow(UiLovRow row) {
                this.row = row;
            }
            
            public UiLovRow getRow() {
                return row;
            }
            
            public void setColumn(UiLovColumn column) {
                this.column = column;
            }
            
            public UiLovColumn getColumn() {
                return column;
            }
            
            public void setText(String text) {
                this.text = text;
            }
            
            public String getText() {
                return text;
            }
            
            public void setHtml(String html) {
                this.html = html;
            }
            
            public String getHtml() {
                return html;
            }

        }
        
        // Строка LOV
	public class UiLovRow extends UiUnit {

            private ArrayList<UiLovCell> cells;
            private ArrayList<UiLovRow> rows;
            private boolean allRows = false;
            private String type = "leaf";
            private int level = 0;
            
            public UiLovRow(String code) {
                super(code);
                cells = new ArrayList<>();
                rows = new ArrayList<>();
            }
            
            @Override
            public Instructions getCreateInstruction() {
                return Instructions.row;
            }
            
            @Override
            public void addChild(UiUnit child) {
                if (child instanceof UiLovRow) {
                    super.addChild(child);
                    UiLovRow r = (UiLovRow)child;
                    rows.add(r);
                    UiUnit p = r.getParent();
                    if ((p != null) && (p instanceof UiLovRow)) {
                        r.setLevel(((UiLovRow)p).getLevel() + 1);
                        setUid(r.getId());
                        instructions.add(Utils.instruction(Instructions.lvl, String.valueOf(r.getLevel())));
                    }
                }
            }
            
            public void addCell(UiLovCell cell) {
                cell.setRow(this);
                cells.add(cell);
            }
            
            public void asCurrent() {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.curr));
            }
            
            public ArrayList<UiLovCell> getCells() {
                return cells;
            }
            
            public void setAllRows() {
                allRows = true;
            }
            
            public boolean isAllRows() {
                return allRows;
            }
            
            public void setLevel(int level) {
                this.level = level;
            }
            
            public int getLevel() {
                return level;
            }
            
            public String getType() {
                return type;
            }
            
            public void setType(String type) {
                if (!Utils.testEmptyString(type)) {
                    try {
                        LOVRowType rt = LOVRowType.valueOf(type);
                        this.type = type.toLowerCase();
                        setUid(getId());
                        instructions.add(Utils.instruction(Instructions.type, this.type));
                    } catch (Exception ex) {}
                }
            }
            
            public void expand() {
                for (UiUnit u : children) {
                    if (u instanceof UiLovRow) {
                        UiLovRow r = (UiLovRow)u;
                        UiLov.this.contextRow = r;
                        setUid(r.getParent().getId());
                        instructions.add(Utils.instruction(Instructions.row, r.getId()));
                        setUid(r.getId());
                        instructions.add(Utils.instruction(Instructions.type, r.getType()));
                        instructions.add(Utils.instruction(Instructions.lvl, String.valueOf(r.getLevel())));
                        setUid(UiLov.this.getId());
                        for (UiLovCell c : r.getCells()) {
                            instructions.add(Utils.instruction(Instructions.k, c.getColumn().getId()));
                            instructions.add(Utils.instruction(Instructions.v, c.getText()));
                        }
                    }
                }
            }
            
            public void collapse() {
                setUid(uiRoot.getId());
                for (UiUnit u : getAllChildren()) {
                    if (u instanceof UiLovRow) {
                        instructions.add(Utils.instruction(Instructions.rm, u.getId()));
                    }
                }
            }
            
        }
        
    }    


/************************************************************************
 *                                                                      *
 *     Класс UiProgressWindow - Окно с индикаторами прогресса           *
 *                                                                      *
 ************************************************************************/
    public class UiProgressWindow extends UiUnit {
        
        public UiProgressWindow() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.prw;
        }
        
    }
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiProgress - индикатор прогресса                           *
 *                                                                      *
 ************************************************************************/
    public class UiProgress extends UiUnit {
        
        public UiProgress() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.prog;
        }
        
    }    
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBlankGroup - "группа принадлежащая бланку"               *
 *                                                                      *
 ************************************************************************/     
    
    public class UiBlankGroup extends UiGroup {
        
        public UiBlankGroup() {
            super();
        }
        
        public UiBlankGroup(String code) {
            super(code);
        }
        
        public void lock() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.lck));
        }
        
        public void unlock() {
            getForm().unlock();
            setUid(getId());
            instructions.add(Utils.instruction(Instructions._lck));
        }
        
    }     
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBlankExtraGroup - "специальная группа в бланке"          *
 *                                                                      *
 ************************************************************************/     
    
    public class UiBlankExtraGroup extends UiBlankGroup {
        
        public UiBlankExtraGroup() {
            super();
        }
        
    }             
    
 
    
/************************************************************************
 *                                                                      *
 *     Класс UiTabs - "группа закладок"                                 *
 *                                                                      *
 ************************************************************************/     
    
    public class UiTabs extends UiBlankExtraGroup {
        
        public UiTabs() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.tabs;
        }
        
    }
        
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBlankRegularGroup - "группа в бланке содержащая поля     *
 *     и другие группы"                                                 *
 *                                                                      *
 ************************************************************************/     
    
    public class UiBlankRegularGroup extends UiBlankGroup {
        
        public UiBlankRegularGroup() {
            super();
        }
        
        public UiBlankRegularGroup(String code) {
            super(code);
        }
        
        @Override
        public void setState(GroupState gs) {
            super.setState(gs);
            if (!mobile && this.state == GroupState.Visible) {
                unlock();
            }
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (!mobile && children.isEmpty()) {
                unlock();
            }
            if (child instanceof UiTabGroup) {
                UiUnit tabs = this.getChildByClass(clTabs);
                if (tabs != null) {
                    tabs.addChild(child);
                } else {
                    tabs = new UiTabs();
                    this.addChild(tabs);
                    tabs.addChild(child);
                }
                ((UiTabGroup)child).setTitle("-");
                ((UiGroup)child).setState("Hidden");
            } else if (child instanceof UiTabs) {
                superChild(child);
            } else if (child instanceof UiGroup) {
                superChild(child);
                ((UiGroup)child).setState("Hidden");
            } else if (child instanceof UiButton) {
                UiUnit buttonBar = this.getChildByClass(clButtonBar);
                if (buttonBar != null) {
                    buttonBar.addChild(child);
                } else {
                    buttonBar = new UiButtonBar();
                    this.addChild(buttonBar);
                    buttonBar.addChild(child);
                }
            } else {
                superChild(child);
            }
        }
        
        public void superChild(UiUnit child) {
            super.addChild(child);
        }
        
        @Override
        public void setTitle(String title) {
            if (title != null) {
                UiUnit header = this.getChildByClass(clHeader);
                if (header == null) {
                    header = new UiHeader();
                    this.addChild(header);
                }
                ((UiHeader)header).setText(title);
            }
        }
        
    }             
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiFieldsGroup - "группа полей в бланке"                    *
 *                                                                      *
 ************************************************************************/     
    
    public class UiFieldsGroup extends UiBlankRegularGroup {
        
        public UiFieldsGroup(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.gfld;
        }
        
    }  
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiHorizontalGroup - "горизонтальная группа полей в бланке" *
 *                                                                      *
 ************************************************************************/     
    
    public class UiHorizontalGroup extends UiBlankRegularGroup {
        
        public UiHorizontalGroup(String code) {
            super(code);
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (child instanceof UiButton) {
                super.superChild(child);
            } else {
                super.addChild(child);
            }
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.ghrz;
        }
        
    }    
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBunchGroup - "схлопывающаяся группа полей в бланке"      *
 *                                                                      *
 ************************************************************************/     
    
    public class UiBunchGroup extends UiBlankRegularGroup {
        
        public UiBunchGroup(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.gbch;
        }
        
    }      
    
    
    
 /************************************************************************
 *                                                                      *
 *     Класс UiTabGroup - "закладка в группе закладок"                  *
 *                                                                      *
 ************************************************************************/     
    
    public class UiTabGroup extends UiBlankRegularGroup {
        
        public UiTabGroup(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.gtab;
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiBlank - "бланк"                                          *
 *                                                                      *
 ************************************************************************/     
    
    public class UiBlank extends UiBlankRegularGroup implements UiFormContent {
        
        private boolean sendPromptWidth = true;
        private int promptWidth = 0;
        private final ArrayList<JsonObject> addonInstructions;
        private boolean initBlank;
        
        public UiBlank() {
            super();
            addonInstructions = new ArrayList<>();
            init();
        }
        
        private void init() {
            initBlank = true;
            unitsToCheck.add(this);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.blnk;
        }
        
        public void calcPromptWidth (String prompt) {
            if (prompt != null) {
                int i = proportional(prompt);
                if (i > promptWidth) {
                    promptWidth = i;
                    if (sendPromptWidth) {
                        unitsToCheck.add(this);
                    }
                }
            } 
        }
        
        @Override
        public void preCheck() {
            if (initBlank) {
                addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                addonInstructions.add(Utils.instruction(Instructions.ini));
            }
        }
        
        @Override
        public void postCheck() {
            initBlank = false;
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
    }
    
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiField - "поле в бланке"                                  *
 *                                                                      *
 ************************************************************************/     
    
    public class UiField extends UiBlankChild implements UiFormCurrent {
        
        private FieldType baseType = FieldType.text;
        private FieldType type = baseType;
        private String value = "";
        private String cursorStart = "";
        private String cursorEnd = "";
        boolean checkFormat = false;
        boolean readOnly = false;
        private String dropDownSelected = "";
        
        public UiField(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.fld;
        }
        
        @Override
        public void undo() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.undo));
        }
        
        @Override
        public void redo() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.redo));
        }
        
        public void setPrompt(String prompt) {
            if (prompt != null) {
                setUid(getId());
                if (Utils.testEmptyString(prompt)) {
                    instructions.add(Utils.instruction(Instructions._prm));
                } else {
                    instructions.add(Utils.instruction(Instructions.prm, prompt));
                }
                getBlank().calcPromptWidth(prompt);
            }
        }
        
        public void setType(String type) {
            if (type != null) {
                if (!Utils.testEmptyString(type)) {
                    try {
                        FieldType ft = FieldType.valueOf(type.toLowerCase());
                        if (ft != baseType) {
                            changeType(type);
                            baseType = ft;
                        }
                    } catch (Exception ex) {
                    }
                }
            }
        }
        
        public void changeType(String type) {
            try {
                FieldType ft = FieldType.valueOf(type.toLowerCase());
                if (ft != this.type) {
                    setUid(getId());
                    if (ft == FieldType.hidden) {
                        instructions.add(Utils.instruction(Instructions.hide));
                    } else {
                        instructions.add(Utils.instruction(Instructions.show));
                        if (ft == FieldType.code) {
                            instructions.add(Utils.instruction(Instructions.type, "area"));
                        }
                        instructions.add(Utils.instruction(Instructions.type, type.toLowerCase()));
                    }
                    this.type = ft;
                }
            } catch (Exception ex) {
            }
        }
        
        public void setLines(String lines) {
            if (!(mobile || lines == null)) {
                int m = StringUtils.countMatches(lines, "/");
                if (m == 1) {
                    if (protocol != 3) {
                        String[] a = lines.split("/");
                        double dr = 1.0;
                        double dd = 1.0;
                        Integer x;
                        if (a.length > 0) {
                            try {
                                x = Integer.parseInt(a[0]);
                                dr = x;
                            } catch (Exception ex) {
                            }
                            if (a.length == 2) {
                                try {
                                    x = Integer.parseInt(a[1]);
                                    dd = x;
                                } catch (Exception ex) {
                                }
                            }
                        }
                        if (dd == 0.0) dd = 1.0;
                        if (!(this.type == FieldType.hidden || this.type == FieldType.code || this.type == FieldType.area || this.type == FieldType.richtext)) {
                            changeType("area");
                        }
                        setUid(getId());
                        instructions.add(Utils.instruction(Instructions.fh, String.valueOf(Math.round((dr / dd) * 100.0) / 100.0)));
                    }
                } else {
                    Integer i;
                    try {
                        i = Integer.parseInt(lines);
                    } catch (Exception ex) {
                        i = 1;
                    }
                    if (i > 1) {
                        if (this.type == FieldType.text) {
                            changeType("area");
                        }
                        if (this.type == FieldType.area || this.type == FieldType.code || this.type == FieldType.richtext) {
                            setUid(getId());
                            instructions.add(Utils.instruction(Instructions.h, lines));
                        }
                    } else if (this.type == FieldType.area || this.type == FieldType.code || this.type == FieldType.richtext) {
                        changeType("text");
                        setUid(getId());
                        instructions.add(Utils.instruction(Instructions.h, "1"));
                    }
                }
            }
        }
        
        public void setValues(String values) {
            if (values != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rd));
                if (!Utils.testEmptyString(values)) {
                    UiCollection c = new UiCollection(values);
                    if (!c.isEmpty()) {
                        while (!c.isLast()) {
                            instructions.add(Utils.instruction(Instructions.rdp));
                            instructions.add(Utils.instruction(Instructions.rdpn, c.getCode()));
                            String val = c.getValue();
                            if (val.startsWith("--")) {
                                instructions.add(Utils.instruction(Instructions.rdpd));
                                instructions.add(Utils.instruction(Instructions.rdpv, val.substring(2)));
                            } else {
                                instructions.add(Utils.instruction(Instructions.rdpv, val));
                            }
                            instructions.add(Utils.instruction(Instructions._rdp));
                            c.getNext();
                        }
                    }
                }
                instructions.add(Utils.instruction(Instructions._rd));
            }
        }
        
        public void setReadOnly(String ro) {
            if (ro != null) {
                setUid(getId());
                if (Utils.testYString(ro)) {
                    instructions.add(Utils.instruction(Instructions.ro));
                    readOnly = true;
                } else {
                    instructions.add(Utils.instruction(Instructions.rw));
                    readOnly = false;
                }
            }
        }
        
        public void setMode(String mode) {
            if (mode != null) {
                try {
                    FieldMode fm = FieldMode.valueOf(mode);
                    if (fm == FieldMode.Editable) {
                        setReadOnly("N");
                    } else if (fm == FieldMode.Readonly) {
                        setReadOnly("Y");
                    }
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.swtc, fm.toString()));
                } catch (Exception ex) {}
            }
        }
        
        public void setHelperText(String hlp) {
            if (hlp != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.fhlp, Utils.string2String(hlp)));
            }    
        }
        
        public void setMin(String min) {
            if (!Utils.testEmptyString(min)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rmin, min));
            }
        }
        
        public void setMax(String max) {
            if (!Utils.testEmptyString(max)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rmax, max));
            }
        }
        
        public void setStep(String step) {
            if (!Utils.testEmptyString(step)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rstp, step));
            }
        }
        
        public void setFrom(String frm) {
            if (!Utils.testEmptyString(frm)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rfrm, frm));
            }
        }
        
        public void setHint(String hint) {
            if ((!mobile) && (hint != null)) {
                setUid(getId());
                if (Utils.testEmptyString(hint)) {
                    instructions.add(Utils.instruction(Instructions._hnt));
                } else {
                    instructions.add(Utils.instruction(Instructions.hnt, hint));
                }
            }
        }
        
        public void setMandatory(String mandatory) {
            if (mandatory != null) {
                setUid(getId());
                if (Utils.testYString(mandatory)) {
                    instructions.add(Utils.instruction(Instructions.mndt));
                } else {
                    instructions.add(Utils.instruction(Instructions.ord));
                }
            }
        }
        
        public void setIcon(String icon) {
            if (icon != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._ico));
                if (!Utils.testEmptyString(icon)) {
                    String ico = icon;
                    switch (icon) {
                        case "ЮЛ":
                            ico = "UL";
                            break;
                        case "ФЛ":
                            ico = "FL";
                            break;
                        case "ПР":
                            ico = "PR";
                            break;
                        default:
                            break;
                    }
                    instructions.add(Utils.instruction(Instructions.ico, ico));
                }
            }
        }
        
        public void setValue(String val) {
            if (val != null) {
                this.value = val;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.vl, val));
            }
        }
        
        public void setCursor(String c) {
            try {
                Integer i;
                i = Integer.parseInt(c);
                if (!(mobile || i < 0)) {
                    cursorStart = c;
                    cursorEnd = c;
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.car, c));
                }
            } catch (Exception ex) {
            }
        }
        
        private void changeStyle(String style) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions._sty));
            if (!Utils.testEmptyString(style)) {
                String s = getForm().parseMultiCss(style);
                if (!Utils.testEmptyString(s)) {
                    setUid(getId());
                    instructions.add(Utils.instruction(Instructions.sty, s));
                }
            }
        }
        
        public void setStyle(String style) {
            if (style != null) {
                boolean b = true;
                int i = StringUtils.lastIndexOf(style, ".");
                if (i > -1) {
                    b = StringUtils.lastIndexOf(style, ")") > i;
                }
                setUid(getId());
                if (b) {
                    changeStyle(style);
                    if (this.type == FieldType.code) {
                        changeType("area");
                    }
                } else {
                    changeStyle(StringUtils.substringBeforeLast(style, "."));
                    String md = StringUtils.substringAfterLast(style, ".");
                    if (Utils.testEmptyString(md)) {
                        if (this.type == FieldType.code) {
                            changeType("area");
                        }
                    } else {
                        if (!(mobile || protocol == 3) && this.type == FieldType.area) {
                            changeType("code");
                        }
                        if (this.type == FieldType.code && md != null) {
                            instructions.add(Utils.instruction(Instructions.mode, md.toLowerCase()));
                        }
                    }
                }
            }
        }
        
        public void setNavigable(String nav) {
            if (nav != null) {
                setUid(getId());
                if (Utils.testYString(nav)) {
                    instructions.add(Utils.instruction(Instructions.nav));
                } else {
                    instructions.add(Utils.instruction(Instructions._nav));
                }
            }
        }
        
        public void setPattern(String p) {
            if (!Utils.testEmptyString(p)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.pat, p));
            }
        }
        
        public void setVerifierParams(String vp) {
            if (!Utils.testEmptyString(vp)) {
                setUid(getId());
                UiCollection c = new UiCollection(vp);
                if (!c.isEmpty()) {
                    while (!c.isLast()) {
                        instructions.add(Utils.instruction(Instructions.vrp));
                        instructions.add(Utils.instruction(Instructions.vrpn, c.getCode()));
                        instructions.add(Utils.instruction(Instructions.vrpv, c.getValue()));
                        instructions.add(Utils.instruction(Instructions._vrp));
                        c.getNext();
                    }
                }
            }
        }
                
        public void setVerifier(String v) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.vr, v));
        }
        
        public void setColoring(String c) {
            if (c != null) {
                setUid(getId());
                if (Utils.testEmptyString(c)) {
                    instructions.add(Utils.instruction(Instructions._mkp));
                } else {
                    instructions.add(Utils.instruction(Instructions.mkup, c));
                }
            }
        }
        
        public void setColoringStyle(String p) {
            if (p != null) {
                String s = getForm().parseMultiCss(p);
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.mkps, s));
            }
        }
        
        public void dropDown(String items) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.drp));
            if (!Utils.testEmptyString(items)) {
                UiCollection c = new UiCollection(items);
                if (!c.isEmpty()) {
                    while (!c.isLast()) {
                        instructions.add(Utils.instruction(Instructions.dit));
                        String code = c.getCode();
                        instructions.add(Utils.instruction(Instructions.dkey, code));
                        instructions.add(Utils.instruction(Instructions.dval, c.getValue()));
                        if (code.equals(dropDownSelected)) {
                            instructions.add(Utils.instruction(Instructions.drs));
                        }
                        instructions.add(Utils.instruction(Instructions._dit));
                        c.getNext();
                    }
                }
            }
            instructions.add(Utils.instruction(Instructions._drp));
        }
        
        public void dropDownSelect(String s) {
            dropDownSelected = s;
        }
        
        public void colorPicker(String color) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.cpk, color));
        }
        
        public void colorPalette(String palette) {
            setUid(getId());
            if (Utils.testEmptyString(palette)) {
                instructions.add(Utils.instruction(Instructions._cpa));
            } else {
                instructions.add(Utils.instruction(Instructions.cpa, palette));
            }
        }
        
        public void rtfEditor() {
            if (!mobile) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.rtf));
            }
        }

        @Override
        public void unsetCurrent() {
            getForm().setCurrentUnit(null);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("GotoField"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        private void makeCurrent() {
            if (this != getForm().getCurrentUnit()) {
                getForm().setCurrentUnit(this);
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("GotoField"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Field", new StringParameter(getCode()));
                //sep.put("Cursor", new StringParameter(cursorEnd));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public void reqCurrent(String car) {
            /*cursorStart = car;
            cursorEnd = car;*/
            if (!(mobile || this == getForm().getCurrentUnit())) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._gtf));
            }
            makeCurrent();
        }
        
        public void reqNewValue(String val, String sc, String ec) {
            makeCurrent();
            if (!this.readOnly) {
                int i = 1;
                String s = val;
                if (this.type == FieldType.upper) {
                    s = s.toUpperCase();
                }
                boolean changeValue;
                boolean changeCursorStart = false;
                boolean changeCursorEnd = false;
                if (this.type == FieldType.checkbox) {
                    changeValue = true;
                    checkFormat = true;
                } else {
                    changeValue = false;
                    if (!value.equals(s)) {
                        changeValue = true;
                        checkFormat = true;
                        value = s;
                    }
                    if (!cursorStart.equals(sc)) {
                        changeCursorStart = true;
                        cursorStart = sc;
                    }
                    if (!cursorEnd.equals(ec)) {
                        //changeCursorStart = true;
                        changeCursorEnd = true;
                        cursorEnd = ec;
                    }
                }
                if (checkFormat || changeCursorStart || changeCursorEnd) {
                    checkFormat = false;
                    String p;
                    if (protocol == 1) {
                        do {
                            p = s.length() > 3999 ? s.substring(0, 3999) : s;
                            SingleEventParameters sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("NewValue"));
                            sep.put("FormID", new StringParameter(getForm().getCode()));
                            sep.put("Field", new StringParameter(getCode()));
                            if (changeValue) {
                                sep.put("Value", new StringParameter(p));
                            }
                            sep.put("Part", new StringParameter(String.valueOf(i)));
                            sep.put("IsCheckFormat", new StringParameter("Y"));
                            if (changeCursorStart) {
                                sep.put("StartCursor", new StringParameter(sc));
                            }
                            if (changeCursorEnd) {
                                sep.put("EndCursor", new StringParameter(ec));
                            }
                            totalEvent.add(new SingleEvent(sep));
                            s = s.length() > 3999 ? s.substring(3999) : "";
                            i++;
                        } while (s.length() != 0);
                    } else if (protocol == 2 || protocol == 3) {
                        SingleEventParameters sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("NewValue"));
                        sep.put("FormID", new StringParameter(getForm().getCode()));
                        sep.put("Field", new StringParameter(getCode()));
                        if (changeValue) {
                            sep.put("Value", new StringParameter(s));
                        }
                        //sep.put("Part", new StringParameter("1"));
                        sep.put("IsCheckFormat", new StringParameter("Y"));
                        if (changeCursorStart) {
                            sep.put("StartCursor", new StringParameter(sc));
                        }
                        if (changeCursorEnd) {
                            sep.put("EndCursor", new StringParameter(ec));
                        }
                        totalEvent.add(new SingleEvent(sep));
                    }
                }
            }
        }
        
        public void reqRadioValue(String val) {
            makeCurrent();
            if (!this.readOnly) {
                value = val;
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("NewValue"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Field", new StringParameter(getCode()));
                sep.put("Value", new StringParameter(val));
                sep.put("IsCheckFormat", new StringParameter("Y"));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public void reqNewInterValue(String val, String sc, String ec) {
            makeCurrent();
            if (!(this.readOnly || this.type == FieldType.checkbox)) {
                int i = 1;
                String s = val;
                if (this.type == FieldType.upper) {
                    s = s.toUpperCase();
                }
                boolean changeValue = false;
                if (!value.equals(s)) {
                    changeValue = true;
                    checkFormat = true;
                    value = s;
                }
                boolean changeCursorStart = false;
                if (!cursorStart.equals(sc)) {
                    changeCursorStart = true;
                    cursorStart = sc;
                }
                boolean changeCursorEnd = false;
                if (!cursorEnd.equals(ec)) {
                    changeCursorEnd = true;
                    cursorEnd = ec;
                }
                if (changeValue || changeCursorStart || changeCursorEnd) {
                    String p;
                    if (protocol == 1) {
                        if (changeValue) {
                            do {
                                p = s.length() > 3999 ? s.substring(0, 3999) : s;
                                SingleEventParameters sep = new SingleEventParameters();
                                sep.put("e", new StringParameter("NewValue"));
                                sep.put("FormID", new StringParameter(getForm().getCode()));
                                sep.put("Field", new StringParameter(getCode()));
                                sep.put("Value", new StringParameter(p));
                                sep.put("Part", new StringParameter(String.valueOf(i)));
                                sep.put("IsCheckFormat", new StringParameter("N"));
                                if (changeCursorStart) {
                                    sep.put("StartCursor", new StringParameter(sc));
                                }
                                if (changeCursorEnd) {
                                    sep.put("EndCursor", new StringParameter(ec));
                                }
                                totalEvent.add(new SingleEvent(sep));
                                s = s.length() > 3999 ? s.substring(3999) : "";
                                i++;
                            } while (s.length() != 0);
                        } else {
                            SingleEventParameters sep = new SingleEventParameters();
                            sep.put("e", new StringParameter("NewValue"));
                            sep.put("FormID", new StringParameter(getForm().getCode()));
                            sep.put("Field", new StringParameter(getCode()));
                            sep.put("Part", new StringParameter("1"));
                            sep.put("IsCheckFormat", new StringParameter("N"));
                            if (changeCursorStart) {
                                sep.put("StartCursor", new StringParameter(sc));
                            }
                            if (changeCursorEnd) {
                                sep.put("EndCursor", new StringParameter(ec));
                            }
                            totalEvent.add(new SingleEvent(sep));
                        }
                    } else if (protocol == 2 || protocol == 3) {
                        SingleEventParameters sep = new SingleEventParameters();
                        sep.put("e", new StringParameter("NewValue"));
                        sep.put("FormID", new StringParameter(getForm().getCode()));
                        sep.put("Field", new StringParameter(getCode()));
                        if (changeValue) {
                            sep.put("Value", new StringParameter(s));
                        }
                        sep.put("Part", new StringParameter("1"));
                        sep.put("IsCheckFormat", new StringParameter("N"));
                        if (changeCursorStart) {
                            sep.put("StartCursor", new StringParameter(sc));
                        }
                        if (changeCursorEnd) {
                            sep.put("EndCursor", new StringParameter(ec));
                        }
                        totalEvent.add(new SingleEvent(sep));
                    }
                }
            }
        }
        
        public void reqCaretPos(String sc, String ec) {
            makeCurrent();
            if (!(this.readOnly || this.type == FieldType.checkbox)) {
                boolean changeCursorStart = false;
                if (!cursorStart.equals(sc)) {
                    changeCursorStart = true;
                    cursorStart = sc;
                }
                boolean changeCursorEnd = false;
                if (!cursorEnd.equals(ec)) {
                    changeCursorEnd = true;
                    cursorEnd = ec;
                }
                if (changeCursorStart || changeCursorEnd) {
                    SingleEventParameters sep = new SingleEventParameters();
                    sep.put("e", new StringParameter("NewValue"));
                    sep.put("FormID", new StringParameter(getForm().getCode()));
                    sep.put("Field", new StringParameter(getCode()));
                    sep.put("Part", new StringParameter("1"));
                    sep.put("IsCheckFormat", new StringParameter("N"));
                    if (changeCursorStart) {
                        sep.put("StartCursor", new StringParameter(sc));
                    }
                    if (changeCursorEnd) {
                        sep.put("EndCursor", new StringParameter(ec));
                    }
                    totalEvent.add(new SingleEvent(sep));
                }
            }
        }
        
        public void reqPatternMatch(String val, String sc, String ec) {
            reqNewInterValue(val, sc, ec);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("PatternMatch"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            totalEvent.add(new SingleEvent(sep));
        }

        public void reqRMB() {
            makeCurrent();
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("FieldRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqDropDown(String value) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("DropDown"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getCode()));
            sep.put("Value", new StringParameter(value));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqColorPick(String color) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NewValue"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getCode()));
            sep.put("Value", new StringParameter(color));
            sep.put("IsCheckFormat", new StringParameter("Y"));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRangeFrom(String v) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NewValue"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getCode()));
            sep.put("Value", new StringParameter(v));
            sep.put("IsCheckFormat", new StringParameter("Y"));
            totalEvent.add(new SingleEvent(sep));
        }
               
        public void reqSpace() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SpaceBar"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqEnter() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("EnterBar"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqF9() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("F9Bar"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));  
        }

        
    }
    

/************************************************************************
 *                                                                      *
 *     Кнопка в поле FieldButton                                        *
 *                                                                      *
 ************************************************************************/
    public class UiFieldButton extends UiBaseButton {
        
        public UiFieldButton(String code) {
            super(code);
        }
        
        @Override
        public void reqPress() {
            UiUnit u = this;
            while ((u != null) && (!(u instanceof UiField))) {
                u = u.getParent();
            }
            if (u != null) {
                ((UiField) u).makeCurrent();
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("FieldButton"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Name", new StringParameter(getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        @Override
        public void setBgColor(String c) {
            //setUid(getId());
            if (Utils.testEmptyString(c)) {
                //instructions.add(Utils.instruction(Instructions._bg));
            } else {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.bg, c));
            }
        }
       
    }            
    
    
/************************************************************************
 *                                                                      *
 *     Операции с файлами (загрузка, выгрузка)                          *
 *                                                                      *
 ************************************************************************/    
    
    /**
     * 
     * @return поток загружаемых байт
     * @throws SQLException что-то пошло не так
     */
    public InputStream getDownloadStream() throws SQLException {
        try {
            if (Utils.testEmptyString(fileDownload.getBase64())) {
                String[] filesPath = fileDownload.getPath().split(":");
                if (filesPath.length == 2) {
                    String storage = filesPath[0];
                    String filename = filesPath[1];
                    FastcomFileStorage aStorage = ST_AGENT_FABRIC.connectToStorage(connection, storage);
                    FileStat fs = aStorage.stat(filename);
                    if (fs.isFile && (aStorage != null)) {
                        return aStorage.openForRead(filename);
                    }
                }
            } else {
                Base64 decoder = new Base64();
                return new ByteArrayInputStream(decoder.decode(fileDownload.getBase64()));
            }    
        } catch (IOException | MultipleExceptions | ProtocolError | RemoteError ex) {}
        
        return null;
    }
    
    public String getDownloadFilename() {
        
        String filename = fileDownload.getFileName();
        if (Utils.testEmptyString(filename)) {
            String[] filePathArray = fileDownload.getPath().split(":");
            if (filePathArray.length == 2) {
                filename = filePathArray[1];
                int lastSlash = filename.lastIndexOf("/");
                if (lastSlash > -1) {
                    filename = filename.substring(lastSlash + 1);
                }
            }
        }
        
        return filename;
    }
    
    public class UiFileUpload extends UiUnit {
        
        private String path = "";
        private String form = "";
        private String files = "";
        private boolean multiple = false;
        private String storeName = null;
        
        public UiFileUpload () {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.upl;
        }
        
        public void setTitle(String t) {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.ttl, t));
        }
        
        public void setMultiple(String m) {
            multiple = Utils.testYString(m);
            if (!multiple) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions._mlt));
            }    
        }
        
        public boolean isMultiple() {
            return multiple;
        }
        
        public void setStoreName(String name) {
            if (!Utils.testEmptyString(name)) {
                storeName = name;
            }
        }
        
        public String getStoreName() {
            return storeName;
        }
        
        public void setPath(String p) {
            path = p;
        }
        
        public void setForm(String f) {
            form = f;
        }
        
        public String getForm() {
            return form;
        }
        
        public String getPath() {
            return path;
        }
        
        public void addFiles(String f) {
            files += f + ";";
        }
        
        public String getFiles() {
            return files;
        }
    
    }
    
    public class UiFileDownload {
        
        private String path = "";
        private String file = "";
        private String base64 = "";
        
        public UiFileDownload () {
            setUid(uiRoot.getId());
            instructions.add(Utils.instruction(Instructions.dl));
        }
        
        public void setPath(String p) {
            path = p;
        }
        
        public String getPath() {
            return path;
        }
        
        public void setFileName(String f) {
            file = f;
        }
        
        public String getFileName() {
            return file;
        }
        
        public void setBase64(String s) {
            base64 = s;
        }
        
        public String getBase64() {
            return base64;
        }
        
    }    
    
    public synchronized void uploadFile(String fileName, InputStream fileStream) throws UiException {
        
        String fname;
        String origName;
        
        int pos = fileName.lastIndexOf("\\");
        if (pos > -1) {
            fname = fileName.substring(pos + 1);
        } else {
            fname = fileName;
        }
        if (protocol == 1) {
            origName = fname;
            if (!fileUpload.isMultiple() && fileUpload.getStoreName() != null) {
                fname = fileUpload.getStoreName();
            }
            String[] filesPath = fileUpload.getPath().split(":");
            if (filesPath.length == 2) {
                String storage = filesPath[0];
                String directory = filesPath[1];
                if ((storage != null) && (directory != null)) {
                    if (!String.valueOf(directory.charAt(directory.length() - 1)).equals("/")) {
                        directory = directory + "/";
                    }
                    FastcomFileStorage aStorage;
                    FileStat fs;
                    OutputStream aWriteTo;
                    try {
                        aStorage = ST_AGENT_FABRIC.connectToStorage(connection, storage);
                        if (aStorage != null) {
                            try {
                                fs = aStorage.stat(directory + fname);
                                while (fs.isFile) {
                                    String fn = fname;
                                    String fext = "";
                                    String fadd = "-1";
                                    int dotPos = fname.lastIndexOf(".");
                                    if (dotPos > -1) {
                                        fn = fname.substring(0, dotPos);
                                        fext = fname.substring(dotPos);
                                    }
                                    if (fn.length() > 2) {
                                        int minusPos = fn.lastIndexOf("-");
                                        if (minusPos > -1) {
                                            String s = fn.substring(minusPos);
                                            try {
                                                Integer i = Integer.parseInt(s);
                                                i--;
                                                fadd = String.valueOf(i.intValue());
                                                fn = fn.substring(0, minusPos);
                                            } catch (NumberFormatException ex) {
                                            }
                                        }
                                    }
                                    fname = fn + fadd + fext;
                                    fs = aStorage.stat(directory + fname);
                                }
                            } catch (MultipleExceptions | ProtocolError | RemoteError | IOException ex) {
                            }

                            try {
                                aWriteTo = aStorage.openForWrite(directory + fname, FastcomFileStorage.OpenMode.REWRITE);
                                byte buf[] = new byte[65536];  // больше не ставить - заболеет!
                                int len;
                                while ((len = fileStream.read(buf)) > 0) {
                                    aWriteTo.write(buf, 0, len);
                                }
                                if (!fname.equals(origName)) {
                                    fname = fname + ":" + origName;
                                }
                                aWriteTo.close();
                                fileUpload.addFiles(fname);
                            } catch (IOException | IllegalArgumentException | ProtocolError | RemoteError ex) {
                                UiError.raise(ex.getMessage());
                            }
                        } else {
                            UiError.raise("Error connecting to file storage");
                        }
                        
                    } catch (SQLException e) {
                        UiError.raise(e.getMessage());
                    }
                } else {
                    UiError.raise("Invalid file storage or path specified");
                }
            } else {
                UiError.raise("Invalid file path specified");
            }
        } else {
            File fl = org.apache.commons.io.FileUtils.getFile(fileUpload.getPath(), fname);
            try {
                org.apache.commons.io.FileUtils.copyInputStreamToFile(fileStream, fl);
                fileUpload.addFiles(fname);
            } catch (IOException ex) {
                UiError.raise(ex.getMessage());
            }
            
        }
    }

    // Окончание загрузки файлов
    public void uploadFinish() {
        instructions = new ArrayList<>();
        instructions.add(Utils.instruction(Instructions.s));
        setUid(fileUpload.getId());
        instructions.add(Utils.instruction(Instructions.ok));
        instructions.add(Utils.instruction(Instructions.f));
    }
    
    public void downloadFinish() {
        fileDownload = null;
    }
    
    
/************************************************************************
 *                                                                      *
 *     Keep alive от клиента                                            *
 *                                                                      *
 ************************************************************************/ 
    
    /**
     * @return команда?
     */
    public Object procPolling() {
        Object ret;
        aliveCounter = 0;
        try {
            if (protocol == 1) {
                ret = sharedArea.getCommands(userName);
            } else {
                ret = msgHandler.getResponse();
            }
        } catch (Exception ex) {
            ret = defaultPollResponse;
        }
        return ret;
    }
    
    public class MonitorTask extends TimerTask {
        
        private final WorkArea wa = WorkArea.this;
        
        @Override
        public void run() {
            if (!stopMonitoring) {
                aliveCounter += 1;
            }
            if (aliveCounter > 7) {
                wa.close();
            }
        }
    }
    
    
/************************************************************************
 *                                                                      *
 *     Прослушивание сообщений сервера                                  *
 *                                                                      *
 ************************************************************************/     
    
    /**
     * Установить обработчик сообщений
     * @param handler обработчик сообщений
     */
    public void setMessageHandler (MessageHandler handler) {
        
        if (msgHandler != null) {
            msgHandler.finish();
        }
        msgHandler = handler;
        
    }
    
    
/************************************************************************
 *                                                                      *
 *     Контент для LOV                                                  *
 *                                                                      *
 ************************************************************************/     
    
    public class UiLovContent {

        String value;

        public UiLovContent(String value) {
            this.value = value;
        }

        public String getValue() {
            return Utils.string2String(value);
        }

    }
    
    public class UiLovContentIcon extends UiLovContent {

        public UiLovContentIcon(String value) {
            super(value);
        }

        @Override
        public String getValue() {
            if (Utils.testEmptyString(Utils.string2String(value))) {
                return "";
            } else {
                return "ico-" + value;
            }
        }

    }
    
    public class UiLovContentText extends UiLovContent {

        public UiLovContentText(String value) {
            super(value);
        }

    }
    
 /***********************************************************************
 *                                                                      *
 *     Класс UiChart - "диаграмма"                                      *
 *                                                                      *
 ************************************************************************/     
    
    public class UiChart extends UiFormChild implements UiFormContent {
        
        boolean generated = false;
        boolean command = false;
                
        public UiChart() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.chrt;
        }
        
        public boolean isGenerated() {
            return generated;
        }
        
        public void generate(String pie, String data) {
            setUid(getId());
            if (Utils.testEmptyString(pie)) {
                if (!Utils.testEmptyString(data)) {
                    generated = true;
                    instructions.add(Utils.instruction(Instructions.pie, ""));
                    instructions.add(Utils.instruction(Instructions.gen, data));
                }
            } else {
                generated = true;
                instructions.add(Utils.instruction(Instructions.pie, pie));
                String d = Utils.testEmptyString(data) ? "" : data;
                instructions.add(Utils.instruction(Instructions.gen, d));
            }
        }
        
        public void setCommand(String cmd) {
            if (!Utils.testEmptyString(cmd)) {
                command = true;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.cmd, cmd));
                if (cmd.toLowerCase().equals("clear")) {
                    instructions.add(Utils.instruction(Instructions.clr, cmd));
                }
                if (cmd.toLowerCase().equals("dispose")) {
                    generated = false;
                    instructions.add(Utils.instruction(Instructions.pie, ""));
                    instructions.add(Utils.instruction(Instructions.data, ""));
                }
            }
        }
        
        public void setData(String pieData, String data) {
            if (command) {
                String p = Utils.string2String(pieData);
                String d = Utils.string2String(data);
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.pie, p));
                instructions.add(Utils.instruction(Instructions.data, d));
            }
        }
        
        public void execCommand() {
            if (command) {
                command = false;
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.exec));
            }
        }
        
        public void save(String fn) {
            String fileName = Utils.testEmptyString(fn) ? "luichart" : fn;
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.save, fileName + ".png"));
        }
        
        public void setLegendMode(String mode) {
            if (mode != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.lm, mode));
            }    
        }
        
        public void setPosition(String pos) {
            if (!Utils.testEmptyString(pos)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.pos, pos));
            }    
        }
        
        public void get() {
            setUid(getId());
            instructions.add(Utils.instruction(Instructions.get));
        }
        
        public void setOrientation(String orientation) {
            if (orientation != null) {
                ChartOrientation co;
                setUid(getId());
                if (Utils.testEmptyString(orientation)) {
                    instructions.add(Utils.instruction(Instructions._orn));
                } else {
                    try {
                        co = ChartOrientation.valueOf(orientation);
                        instructions.add(Utils.instruction(Instructions.ornt, orientation));
                    } catch (Exception ex) {}
                    
                }
            }
        }
        
        public void setWidth(String width) {
            if (width != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.w, width));
            }
        }
        
        public void setHeight(String height) {
            if (height != null) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.h, height));
            }
        }
        
        public void reqOver(String e) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartOver"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(e));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqOut(String e) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartOut"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(e));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqClick(String e) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(e));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqRMB(String e) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(e));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqDC(String e) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(e));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqLegendClick(String name) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartLegendClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Name", new StringParameter(name));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqLegendDC(String name) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartLegendDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Name", new StringParameter(name));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqLegendRMB(String name) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartLegendRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Name", new StringParameter(name));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqMapPosition(String p) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartMapPosition"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(p));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqColormapClick(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColormapClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqColormapDC(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColormapDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqColormapRMB(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ColormapRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqPiecewiseClick(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("PiecewiseClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqPiecewiseDC(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("PiecewiseDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqPiecewiseRMB(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("PiecewiseRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqZoomClick(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartZoomClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqZoomDC(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartZoomDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqResize(String width, String height) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartResize"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Width", new StringParameter(width));
            sep.put("Height", new StringParameter(height));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqAxisClick(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartAxisClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Parameters", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqZoom(String par) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("ChartZoom"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Scale", new StringParameter(par));
            totalEvent.add(new SingleEvent(sep));
        } 
        
        public void reqImage(String img) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("Image"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Data", new StringParameter(img));
            totalEvent.add(new SingleEvent(sep));
        }
        
    }
    
 /***********************************************************************
 *                                                                      *
 *     Класс UiHtml - "форма содержащая html"                           *
 *                                                                      *
 ************************************************************************/     
    
    public class UiHtml extends UiFormChild implements UiFormContent {
       
        public UiHtml() {
            super();
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.html;
        }
        
        public void setHtml(String data) {
            if (!Utils.testEmptyString(data)) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.data, data));
            }
        }
    }
    
 
/************************************************************************
 *                                                                      *
 *     Класс UiMatrix - "матрица"                                          *
 *                                                                      *
 ************************************************************************/     
    
    public class UiMatrix extends UiFormChild implements UiFormContent {
        
        private final LinkedHashMap<String, UiMatrixRow> rows;
        private final LinkedHashMap<String, UiMatrixColumn> columns;
        private final HashMap<String, UiMatrixCell> cells;
        private final ArrayList<JsonObject> addonInstructions;
        private final ArrayList<JsonObject> asyncInstructions;
        private UiMatrixCell currentCell = null;
        private String selectedRows = "";
        private String selectedColumns = "";
        private String selectedCells = "";
        private boolean initialized = false;
        private UiMatrixTitleCell htmlTitle;
        private ArrayList<ArrayList<UiUnit>> htmlHeader;
        private ArrayList<ArrayList<UiUnit>> htmlSidewall;
        private ArrayList<ArrayList<UiUnit>> htmlData;
        private MatrixColumnInfo columnInfo = MatrixColumnInfo.None;
        private MatrixRowInfo rowInfo = MatrixRowInfo.None;
        private String prevId = "";

        
        public UiMatrix() {
            super();
            rows = new LinkedHashMap<>();
            columns = new LinkedHashMap<>();
            cells = new HashMap<>();
            addonInstructions = new ArrayList<>();
            asyncInstructions = new ArrayList<>();
        }
        
        public UiMatrix(boolean render) {
            super(render);
            rows = new LinkedHashMap<>();
            columns = new LinkedHashMap<>();
            cells = new HashMap<>();
            addonInstructions = new ArrayList<>();
            asyncInstructions = new ArrayList<>();
        }
        
        private void refresh() {
            if (isInitialized()) {
                setUid(getId());
                instructions.add(Utils.instruction(Instructions.wipe));
                asyncInstructions.clear();
                prevId = "";
                initialized = false;
                unitsToCheck.add(this);
            }    
        }
        
        public void plain() {
            ArrayList<ArrayList<String>> plainMatrix = new ArrayList<>();
            ArrayList<ArrayList<String>> plainHeader = new ArrayList<>();
            ArrayList<ArrayList<String>> plainSidewall = new ArrayList<>();
            ArrayList<ArrayList<String>> plainCells = new ArrayList<>();
            int cSize = 0;
            int rSize = 0;
            ArrayList<String> row;
            if (htmlHeader.size() > 3) {
                cSize = htmlHeader.get(1).size();
                if (columnInfo != MatrixColumnInfo.Only) {
                    for (int i = 0; i < htmlHeader.size() - 3; i++) {
                        row = new ArrayList<>();
                        for (int j = 0; j < cSize; j++) {
                            row.add(null);
                        }
                        plainHeader.add(row);
                    }
                    for (int i = 0; i < htmlHeader.size() - 3; i++) {
                        int m = htmlHeader.get(i + 2).size();
                        for (int j = 0; j < m; j++) {
                            UiMatrixColumn c = (UiMatrixColumn)htmlHeader.get(i + 2).get(j);
                            ArrayList<String> ca = plainHeader.get(i);
                            int n = 0;
                            while (!(ca.get(n) == null)) {
                                n += 1;
                            }
                            int cs = c.getHtmlProperties().getColSpan() > 0 ? c.getHtmlProperties().getColSpan() - 1 : 0;
                            int rs = c.getHtmlProperties().getRowSpan() > 0 ? c.getHtmlProperties().getRowSpan() - 1 : 0;
                            for (int l = n; l <= n + cs; l++) {
                                for (int k = i; k <= i + rs; k++) {
                                    plainHeader.get(k).set(l, "x");
                                }
                            }
                            ca.set(n, c.getId());
                        }
                    }
                }
                if (columnInfo == MatrixColumnInfo.Only || columnInfo == MatrixColumnInfo.Top) {
                    row = new ArrayList<>();
                    for (int i = 0; i < cSize; i++) {
                        UiUnit u = htmlHeader.get(1).get(i);
                        row.add("info" + ((UiMatrixInfoCell)u).getCellset().getId());
                    }
                    if (row.size() > 0) {
                        if (plainHeader.isEmpty()) {
                            plainHeader.add(row);
                        } else {
                            plainHeader.set(0, row);
                        }
                    }
                } else if (columnInfo == MatrixColumnInfo.Bottom) {
                    row = new ArrayList<>();
                    for (int i = 0; i < cSize; i++) {
                        UiUnit u = htmlHeader.get(1).get(i);
                        row.add("info" + ((UiMatrixInfoCell)u).getCellset().getId());
                    }
                    if (row.size() > 0) plainHeader.add(row);
                }
            }
            if (htmlSidewall.size() > 0) {
                ArrayList<UiUnit> a = htmlSidewall.get(0);
                for (int i = 1; i < a.size() - 1; i++) {
                    int cs = ((UiMatrixRow)a.get(i)).getHtmlProperties().getColSpan();
                    if (cs == 0) cs = 1;
                    rSize = rSize + cs;
                }
                for (int i = 0; i < htmlSidewall.size(); i++) {
                    row = new ArrayList<>();
                    for (int j = 0; j < rSize; j++) {
                        row.add(null);
                    }
                    plainSidewall.add(row);
                }
                if (rowInfo != MatrixRowInfo.Only) {    
                    for (int i = 0; i < htmlSidewall.size(); i++) {
                        a = htmlSidewall.get(i);
                        ArrayList<String> ra = plainSidewall.get(i);
                        for (int j = 1; j < a.size() - 1; j++) {
                            UiMatrixRow r = (UiMatrixRow)a.get(j);
                            int n = 0;
                            while (!(ra.get(n) == null)) {
                                n += 1;
                            }
                            int cs = r.getHtmlProperties().getColSpan() > 0 ? r.getHtmlProperties().getColSpan() - 1 : 0;
                            int rs = r.getHtmlProperties().getRowSpan() > 0 ? r.getHtmlProperties().getRowSpan() - 1 : 0;
                            for (int l = n; l <= n + cs; l++) {
                                for (int k = i; k <= i + rs; k++) {
                                    plainSidewall.get(k).set(l, "x");
                                }
                            }
                            ra.set(n, r.getId());
                        }
                    }
                }    
                if (rowInfo == MatrixRowInfo.Only || rowInfo == MatrixRowInfo.Left) {
                    for (int i = 0; i < htmlSidewall.size(); i++) {
                        UiMatrixInfoCell ic = (UiMatrixInfoCell)htmlSidewall.get(i).get(0);
                        plainSidewall.get(i).add(0, "info" + ic.getCellset().getId());
                    }
                } else if (rowInfo == MatrixRowInfo.Right) {
                    for (int i = 0; i < htmlSidewall.size(); i++) {
                        UiMatrixInfoCell ic = (UiMatrixInfoCell)htmlSidewall.get(i).get(0);
                        plainSidewall.get(i).add("info" + ic.getCellset().getId());
                    }
                }
                rSize = plainSidewall.get(0).size();
            }
            for (int i = 0; i < plainSidewall.size(); i++) {
                row = new ArrayList<>();
                for (int j = 0; j < cSize; j++) {
                    row.add(null);
                }
                plainCells.add(row);
            }
            for (int i = 0; i < plainSidewall.size(); i++) {
                ArrayList<UiUnit> a = htmlData.get(i);
                ArrayList<String> xa = plainCells.get(i);
                for (int j = 0; j < a.size(); j++) {
                    UiMatrixCellWrap w = (UiMatrixCellWrap)a.get(j);
                    if (w.getCell() != null) {
                        int n = 0;
                        while (!(xa.get(n) == null)) {
                            n += 1;
                        }
                        int cs = w.getHtmlProperties().getColSpan() > 0 ? w.getHtmlProperties().getColSpan() - 1 : 0;
                        int rs = w.getHtmlProperties().getRowSpan() > 0 ? w.getHtmlProperties().getRowSpan() - 1 : 0;
                        for (int l = n; l <= n + cs; l++) {
                            for (int k = i; k <= i + rs; k++) {
                                plainCells.get(k).set(l, "x");
                            }
                        }
                        xa.set(n, w.getCell().getId());
                    }    
                }
            }
            if (plainSidewall.size() > 0 || plainHeader.size() > 0) {
                int m = plainSidewall.size() > 0 ? rSize : 1;   
                if (plainHeader.size() > 0) {
                    for (int i = 0; i < plainHeader.size(); i++) {
                        row = new ArrayList<>();
                        for (int j = 0; j < m; j++) {
                            row.add("x");
                        }
                        row.addAll(plainHeader.get(i));
                        plainMatrix.add(row);
                    }
                    plainMatrix.get(0).set(0, "r");
                } else {
                    row = new ArrayList<>();
                    row.add("r");
                    plainMatrix.add(row);
                }
                if (plainSidewall.size() > 0) {
                    for (int i = 0; i < plainSidewall.size(); i++) {
                        row = new ArrayList<>();
                        row.addAll(plainSidewall.get(i));
                        row.addAll(plainCells.get(i));
                        plainMatrix.add(row);
                    }
                }
            }
            setUid(getId());
            Gson gson = new Gson();
            instructions.add(Utils.instruction(Instructions.pln, gson.toJson(plainMatrix)));
        }
        
        public void finish() {
            if (!isInitialized()) {
                updateProperties();
                htmlHeader = new ArrayList<>();
                StringJoiner columnsIds = new StringJoiner(" ");
                ArrayList<UiUnit> title = new ArrayList<>();
                ArrayList<UiUnit> topInfos = new ArrayList<>();
                ArrayList<UiUnit> bottomInfos = new ArrayList<>();
                htmlTitle = new UiMatrixTitleCell();
                title.add(htmlTitle);
                htmlHeader.add(title);
                htmlHeader.add(topInfos);
                ArrayList<UiUnit> rowLeafs = new ArrayList<>();
                ArrayList<UiUnit> columnLeafs = new ArrayList<>();
                ArrayList<UiUnit> cols = this.getVisibleChildrenByClass(clMatrixColumn);
                while (!cols.isEmpty()) {
                    ArrayList<UiUnit> row = new ArrayList<>();
                    ArrayList<UiUnit> chld = new ArrayList<>();
                    for (UiUnit c : cols) {
                        UiMatrixColumn col = (UiMatrixColumn)c;
                        row.add(col);
                        col.getHtmlProperties().clear();
                        col.clearCellsIds();
                        columnsIds.add(col.getId());
                        ArrayList<UiUnit> temp = c.getVisibleChildrenByClass(clMatrixColumn);
                        if (temp.isEmpty()) {
                            col.getHtmlProperties().setLeaf();
                            //topInfos.add(col.getTopInfo());
                            //bottomInfos.add(col.getBottomInfo());
                        }
                        chld.addAll(temp);
                    }
                    htmlHeader.add(row);
                    cols = chld;
                }
                htmlHeader.add(bottomInfos);
                for (int i = 2; i < htmlHeader.size() - 1; i++) {
                    for (int j = 0; j < htmlHeader.get(i).size(); j++) {
                        if (htmlHeader.get(i).get(j) instanceof UiMatrixColumn) {
                            UiMatrixColumn col = (UiMatrixColumn)htmlHeader.get(i).get(j);
                            UiUnit parent = col.getParent();
                            StringJoiner parents = new StringJoiner(" ");
                            while (!(parent instanceof UiMatrix)) {
                                parents.add(parent.getId());
                                parent = parent.getParent();
                            }
                            col.getHtmlProperties().setParents(parents.toString());
                            if (col.getHtmlProperties().isLeaf()) {
                                int rs = htmlHeader.size() - i - 1;
                                if (rs > 1) col.getHtmlProperties().setRowSpan(rs);
                            } else {
                                ArrayList<UiUnit> temp = col.getAllVisibleChildrenByClass(clMatrixColumn);
                                ArrayList<UiMatrixCellset> leafs = new ArrayList<>();
                                StringJoiner allChildren = new StringJoiner(" ");
                                for (int k = 0; k < temp.size(); k++) {
                                    allChildren.add(temp.get(k).getId());
                                    if (((UiMatrixColumn)temp.get(k)).getHtmlProperties().isLeaf()) {
                                        leafs.add((UiMatrixColumn)temp.get(k));
                                    }
                                }
                                col.setLeafs(leafs);
                                col.getHtmlProperties().setAllChildren(allChildren.toString());
                                if (leafs.size() > 1) {
                                    col.getHtmlProperties().setColSpan(leafs.size());
                                }
                            }
                        }
                    }
                }
                htmlSidewall = new ArrayList<>();
                StringJoiner rowsIds = new StringJoiner(" ");
                ArrayList<UiUnit> rws = this.getVisibleChildrenByClass(clMatrixRow);
                ArrayList<UiUnit> mr;
                while (!rws.isEmpty()) {
                    ArrayList<UiUnit> chld = new ArrayList<>();
                    UiUnit parent = null;
                    int idx = 0;
                    int delta = 0;
                    for (UiUnit r : rws) {
                        UiMatrixRow row = (UiMatrixRow)r;
                        row.getHtmlProperties().clear();
                        row.clearCellsIds();
                        rowsIds.add(row.getId());
                        ArrayList<UiUnit> temp = r.getVisibleChildrenByClass(clMatrixRow);
                        if (temp.isEmpty()) {
                            row.getHtmlProperties().setLeaf();
                        }
                        if (r.getParent() instanceof UiMatrixRow) {
                            if (r.getParent() != parent) {
                                parent = r.getParent();
                                delta = 0;
                                for (int i = 0; i < htmlSidewall.size(); i++) {
                                    if (htmlSidewall.get(i).contains(parent)) {
                                        idx = i;
                                        break;
                                    }
                                }
                            }
                            if (delta > 0) {
                                mr = new ArrayList<>();
                                mr.add(r);
                                htmlSidewall.add(idx + delta, mr);
                            } else {
                                htmlSidewall.get(idx).add(r);
                            }
                            delta += 1;
                        } else {
                            mr = new ArrayList<>();
                            mr.add(r);
                            htmlSidewall.add(mr);
                        }
                        chld.addAll(temp);
                    }
                    rws = chld;
                }
                int maxRowParents = 0;
                for (int i = 0; i < htmlSidewall.size(); i++) {
                    UiMatrixRow lr = null;
                    for (int j = 0; j < htmlSidewall.get(i).size(); j++) {
                        if (htmlSidewall.get(i).get(j) instanceof UiMatrixRow) {
                            UiMatrixRow row = (UiMatrixRow)htmlSidewall.get(i).get(j);
                            UiUnit parent = row.getParent();
                            StringJoiner parents = new StringJoiner(" ");
                            int ps = 0;
                            while (!(parent instanceof UiMatrix)) {
                                parents.add(parent.getId());
                                parent = parent.getParent();
                                ps += 1;
                            }
                            row.getHtmlProperties().setParents(parents.toString());
                            row.getHtmlProperties().setParentsCount(ps);
                            maxRowParents = Math.max(ps, maxRowParents);
                            if (row.getHtmlProperties().isLeaf()) {
                                lr = row;
                                rowLeafs.add(row);
                                //break;
                            }
                        }
                    }
                    if (lr != null) {
                        htmlSidewall.get(i).add(0, lr.getLeftInfo());
                        htmlSidewall.get(i).add(lr.getRightInfo());
                    }
                }
                for (int i = 0; i < htmlSidewall.size(); i++) {
                    for (int j = 0; j < htmlSidewall.get(i).size(); j++) {
                        if (htmlSidewall.get(i).get(j) instanceof UiMatrixRow) {
                            UiMatrixRow row = (UiMatrixRow)htmlSidewall.get(i).get(j);
                            if (row.getHtmlProperties().isLeaf()) {
                                int cs = maxRowParents - row.getHtmlProperties().getParentsCount();
                                if (cs > 0) row.getHtmlProperties().setColSpan(cs + 1);
                            } else {
                                ArrayList<UiUnit> temp = row.getAllVisibleChildrenByClass(clMatrixRow);
                                ArrayList<UiMatrixCellset> leafs = new ArrayList<>();
                                StringJoiner allChildren = new StringJoiner(" ");
                                for (int k = 0; k < temp.size(); k++) {
                                    allChildren.add(temp.get(k).getId());
                                    if (((UiMatrixRow)temp.get(k)).getHtmlProperties().isLeaf()) {
                                        leafs.add((UiMatrixRow)temp.get(k));
                                    }
                                }
                                row.setLeafs(leafs);
                                row.getHtmlProperties().setAllChildren(allChildren.toString());
                                if (leafs.size() > 1) {
                                    row.getHtmlProperties().setRowSpan(leafs.size());
                                }
                            }
                        }
                    }
                }
                htmlData = new ArrayList<>();
                for (UiMatrixColumn c : columns.values()) {
                    if (!(c.isHidden()) && c.getHtmlProperties().isLeaf()) {
                        columnLeafs.add(c);
                        topInfos.add(c.getTopInfo());
                        bottomInfos.add(c.getBottomInfo());
                    }
                }
                for (int i = 0; i < rowLeafs.size(); i++) {
                    ArrayList<UiUnit> dataRow = new ArrayList<>();
                    UiMatrixRow r = (UiMatrixRow)rowLeafs.get(i);
                    for (int j = 0; j < columnLeafs.size(); j++) {
                        UiMatrixCellWrap cw = new UiMatrixCellWrap(r, (UiMatrixColumn)columnLeafs.get(j));
                        dataRow.add(cw);
                    }
                    htmlData.add(dataRow);
                }
                StringJoiner cellsIds = new StringJoiner(" ");
                for (UiMatrixRow row : rows.values()) {
                    if (!row.isHidden()) {
                        ArrayList<UiMatrixCell> rc = row.getCells();
                        for (int i = 0; i < rc.size(); i++) {
                            UiMatrixCell c = rc.get(i);
                            UiMatrixCellset cl = (UiMatrixCellset)units.get(c.getColumnId());
                            if (!cl.isHidden()) {
                                int ridx = rowLeafs.indexOf(row);
                                int cidx = columnLeafs.indexOf(cl);
                                ArrayList<UiMatrixCellset> sr = new ArrayList<>();
                                ArrayList<UiMatrixCellset> sc = new ArrayList<>();
                                if (ridx < 0 || cidx < 0) {
                                    if (ridx < 0) {
                                        for (int n = 0; n < row.getLeafs().size(); n++) {
                                            sr.add(row.getLeafs().get(n));
                                        }
                                        ridx = rowLeafs.indexOf(sr.get(0));
                                    } else {
                                        sr.add(row);
                                    }
                                    if (cidx < 0) {
                                        for (int n = 0; n < cl.getLeafs().size(); n++) {
                                            sc.add(cl.getLeafs().get(n));
                                        }
                                        cidx = columnLeafs.indexOf(sc.get(0));
                                    } else {
                                        sc.add(cl);
                                    }
                                }
                                UiMatrixCellWrap cw = (UiMatrixCellWrap)htmlData.get(ridx).get(cidx);
                                if (!cw.getHtmlProperties().isOccupied()) {
                                    for (int n = 0; n < sr.size(); n++) {
                                        ridx = rowLeafs.indexOf(sr.get(n));
                                        for (int m = 0; m < sc.size(); m++) {
                                            cidx = columnLeafs.indexOf(sc.get(m));
                                            UiMatrixCellWrap wr = (UiMatrixCellWrap)htmlData.get(ridx).get(cidx);
                                            wr.setCell(null);
                                            wr.getHtmlProperties().setOccupied();
                                        }
                                    }
                                    cw.setCell(c);
                                    cellsIds.add(c.getId());
                                    row.addCellId(c.getId());
                                    cl.addCellId(c.getId());
                                    if (sr.size() > 1) cw.getHtmlProperties().setRowSpan(sr.size());
                                    if (sc.size() > 1) cw.getHtmlProperties().setColSpan(sc.size());
                                }
                            }
                        }
                    }    
                }
                //  Подготовка HTML
                ArrayList<JsonObject> instr = new ArrayList<>();
                ArrayList<JsonObject> propInstr = new ArrayList<>();
                StringBuilder sb = new StringBuilder();
                sb.append("<div class='ui-matrix ui-selectable context-menu lui2-unit' id='");
                sb.append(getId());
                //sb.append("' rowinfo='none' colinfo='none'>");
                sb.append("' rowinfo='");
                sb.append(rowInfo.toString().toLowerCase());
                sb.append("' colinfo='");
                sb.append(columnInfo.toString().toLowerCase());
                sb.append("'>");
                sb.append("<div class='ui-matrix-preloader preloader-ballsline lui2-unit'></div>");
                sb.append("<table class='ui-matrix-layout lui2-unit'><tbody>");
                sb.append("<tr class='ui-matrix-title hidden lui2-unit'>");
                sb.append("<td class='ui-title-cell ui-select-handle context-menu lui2-unit' colspan='");
                sb.append(String.valueOf(maxRowParents + 3 + columnLeafs.size()));
                sb.append("'>");
                sb.append("<div class='ui-title-content lui2-unit'></div>");
                sb.append("</td></tr>");
                sb.append("<tr class='ui-matrix-infoline ui-matrix-infoline-top lui2-unit' id='");
                sb.append(Utils.generateId());
                sb.append("'>");
                sb.append("<td class='ui-matrix-fixed-cell ui-root-cell ui-select-handle ui-resizable context-menu lui2-unit' rowspan='");
                sb.append(String.valueOf(htmlHeader.size() - 1));
                sb.append("' colspan='");
                sb.append(String.valueOf(maxRowParents + 3));
                sb.append("'>");
                sb.append("<div class='ui-root-header lui2-unit'><div class='ui-root-text lui2-unit'></div></div>");
                sb.append("<div class='ui-root-resize ui-resizable-handle lui2-unit'></div></td>");
                for (int i = 0; i < htmlHeader.get(1).size(); i++) {
                    UiMatrixInfoCell ic = (UiMatrixInfoCell)htmlHeader.get(1).get(i);
                    sb.append("<td class='ui-info-cell ui-select-handle ui-resizable ui-info-cell-top leaf lui2-unit' column='");
                    sb.append(ic.getCellset().getId());
                    sb.append("'>");
                    sb.append("<div class='ui-column-info lui2-unit'><div class='ui-info-text lui2-unit'></div></div>");
                    sb.append("<div class='ui-column-resize ui-column-info-resize ui-resizable-handle lui2-unit'></div>");
                    sb.append("</td>");
                }
                sb.append("</tr>");
                for (int i = 2; i < htmlHeader.size() - 1; i++) {
                    if (i == 2) {
                        sb.append("<tr class='ui-matrix-columns ui-root-columns lui2-unit' id='");
                    } else {
                        sb.append("<tr class='ui-matrix-columns lui2-unit' id='");
                    }
                    sb.append(Utils.generateId());
                    sb.append("'>");
                    for (UiUnit u : htmlHeader.get(i)) {
                        UiMatrixColumn col = (UiMatrixColumn)u;
                        sb.append("<td class='ui-matrix-column ui-resizable context-menu ui-select-handle halign-Center valign-Center lui2-unit ");
                        if (col.getHtmlProperties().isLeaf()) {
                            sb.append("leaf ");
                        }
                        sb.append(col.getId());
                        sb.append(" orientation-");
                        sb.append(col.getOrientation().toString());
                        propInstr.add(Utils.instruction(Instructions.b, col.getId()));
                        propInstr.add(Utils.instruction(Instructions.ornt_, col.getOrientation().toString()));
                        String val = col.getCurrentProperties().get("scls");
                        if (val != null) {
                            String v = getForm().parseMultiCss(val);
                            if (!Utils.testEmptyString(v)) {
                                sb.append(" ");
                                sb.append(v);
                                propInstr.add(Utils.instruction(Instructions.b, col.getId()));
                                propInstr.add(Utils.instruction(Instructions.scls_, v));
                            }
                        }   
                        sb.append("' id='");
                        sb.append(col.getId());
                        sb.append("' ");
                        sb.append("column='");
                        sb.append(col.getId());
                        sb.append("' ");
                        sb.append("code='");
                        sb.append(col.getCode());
                        sb.append("'");
                        if (col.getHtmlProperties().getColSpan() > 0) {
                            sb.append(" colspan='");
                            sb.append(String.valueOf(col.getHtmlProperties().getColSpan()));
                            sb.append("'");
                        }
                        if (col.getHtmlProperties().getRowSpan() > 0) {
                            sb.append(" rowspan='");
                            sb.append(String.valueOf(col.getHtmlProperties().getRowSpan()));
                            sb.append("'");
                        }
                        if (!Utils.testEmptyString(col.getHtmlProperties().getAllChildren())) {
                            sb.append(" children='");
                            sb.append(col.getHtmlProperties().getAllChildren());
                            sb.append("'");
                        }   
                        if (!Utils.testEmptyString(col.getHtmlProperties().getParents())) {
                            sb.append(" parents='");
                            sb.append(col.getHtmlProperties().getParents());
                            sb.append("'");
                        } 
                        sb.append(">");
                        sb.append("<div class='ui-matrix-column-header lui2-unit'>");
                        sb.append("<div class='ui-matrix-column-title lui2-unit'>");
                        sb.append("<div class='ui-matrix-column-title-text lui2-unit'></div>");
                        sb.append("</div></div><div class='ui-column-resize ui-matrix-column-resize ui-resizable-handle lui2-unit'></div>");
                        sb.append("</td>");
                        instr.add(Utils.instruction(Instructions.b, getId()));
                        instr.add(Utils.instruction(Instructions.id, col.getId()));
                        instr.add(Utils.instruction(Instructions.ln, "column"));
                        instr.add(Utils.instruction(Instructions.b, col.getId()));
                        instr.add(Utils.instruction(Instructions.ids, col.getCellsIds()));
                    }
                    sb.append("</tr>");
                }
                sb.append("<tr class='ui-matrix-infoline ui-matrix-infoline-bottom lui2-unit' id='");
                sb.append(Utils.generateId());
                sb.append("'>");
                for (int i = 0; i < htmlHeader.get(1).size(); i++) {
                    UiMatrixInfoCell ic = (UiMatrixInfoCell)htmlHeader.get(1).get(i);
                    sb.append("<td class='ui-info-cell ui-select-handle ui-resizable ui-info-cell-bottom leaf lui2-unit' column='");
                    sb.append(ic.getCellset().getId());
                    sb.append("'>");
                    sb.append("<div class='ui-column-info lui2-unit'><div class='ui-info-text lui2-unit'></div></div>");
                    sb.append("<div class='ui-column-resize ui-column-info-resize ui-resizable-handle lui2-unit'></div>");
                    sb.append("</td>");
                }
                sb.append("</tr>");
                for (int i = 0; i < htmlSidewall.size(); i++) {
                    sb.append("<tr class='ui-matrix-data lui2-unit' id='");
                    sb.append(Utils.generateId());
                    sb.append("'>");
                    for (UiUnit u : htmlSidewall.get(i)) {
                        if (u instanceof UiMatrixInfoCell) {
                            UiMatrixInfoCell ic = (UiMatrixInfoCell)u;
                            sb.append("<td class='ui-row-info ui-select-handle ui-resizable lui2-unit ");
                            if (ic.getPosition() == Defaults.InfoCellPosition.Left) {
                                sb.append("ui-row-info-left");
                            } else {
                                sb.append("ui-row-info-right");
                            }
                            sb.append("' row='");
                            sb.append(ic.getCellset().getId());
                            sb.append("'>");
                            sb.append("<div class='ui-info-text lui2-unit'></div>");
                            sb.append("<div class='ui-row-resize ui-row-info-resize ui-resizable-handle lui2-unit'></div></td>");
                        } else if (u instanceof UiMatrixRow) {
                            UiMatrixRow rw = (UiMatrixRow)u;
                            sb.append("<td class='ui-matrix-row ui-resizable context-menu ui-select-handle halign-Left valign-Center lui2-unit");
                            if (rw.getHtmlProperties().isLeaf()) {
                                sb.append(" leaf");
                            }
                            sb.append(" orientation-");
                            sb.append(rw.getOrientation().toString());
                            propInstr.add(Utils.instruction(Instructions.b, rw.getId()));
                            propInstr.add(Utils.instruction(Instructions.ornt_, rw.getOrientation().toString()));
                            String val = rw.getCurrentProperties().get("scls");
                            if (val != null) {
                                String v = getForm().parseMultiCss(val);
                                if (!Utils.testEmptyString(v)) {
                                    sb.append(" ");
                                    sb.append(v);
                                    propInstr.add(Utils.instruction(Instructions.b, rw.getId()));
                                    propInstr.add(Utils.instruction(Instructions.scls_, v));
                                }
                            } 
                            sb.append("' id='");
                            sb.append(rw.getId());
                            sb.append("' ");
                            sb.append("row='");
                            sb.append(rw.getId());
                            sb.append("' ");
                            sb.append("code='");
                            sb.append(rw.getCode());
                            sb.append("'");
                            if (rw.getHtmlProperties().getColSpan() > 0) {
                                sb.append(" colspan='");
                                sb.append(String.valueOf(rw.getHtmlProperties().getColSpan()));
                                sb.append("'");
                            }
                            if (rw.getHtmlProperties().getRowSpan() > 0) {
                                sb.append(" rowspan='");
                                sb.append(String.valueOf(rw.getHtmlProperties().getRowSpan()));
                                sb.append("'");
                            }
                            if (!Utils.testEmptyString(rw.getHtmlProperties().getAllChildren())) {
                                sb.append(" children='");
                                sb.append(rw.getHtmlProperties().getAllChildren());
                                sb.append("'");
                            }  
                            if (!Utils.testEmptyString(rw.getHtmlProperties().getParents())) {
                                sb.append(" parents='");
                                sb.append(rw.getHtmlProperties().getParents());
                                sb.append("'");
                            } 
                            sb.append(">");                    
                            sb.append("<div class='ui-matrix-row-header lui2-unit'>");
                            sb.append("<div class='ui-matrix-row-title lui2-unit'></div></div>");
                            sb.append("<div class='ui-row-resize ui-matrix-row-resize ui-resizable-handle lui2-unit'></div>");
                            sb.append("</td>");
                            instr.add(Utils.instruction(Instructions.b, getId()));
                            instr.add(Utils.instruction(Instructions.id, rw.getId()));
                            instr.add(Utils.instruction(Instructions.ln, "row"));
                            instr.add(Utils.instruction(Instructions.b, rw.getId()));
                            instr.add(Utils.instruction(Instructions.ids, rw.getCellsIds()));
                        }
                    }
                    for (UiUnit u : htmlData.get(i)) {
                        UiMatrixCellWrap cw = (UiMatrixCellWrap)u; 
                        UiMatrixCell cl = cw.getCell();
                        boolean closeTag = false;
                        //propInstr.clear();
                        if (cl == null) {
                            if (!cw.getHtmlProperties().isOccupied()) {
                                sb.append("<td class='ui-matrix-cell context-menu ui-select-handle lui2-unit'");
                                sb.append(" row='");
                                sb.append(cw.getRow().getId());
                                sb.append("' column='");
                                sb.append(cw.getColumn().getId());
                                sb.append("'");
                                instr.add(Utils.instruction(Instructions.b, getId()));
                                instr.add(Utils.instruction(Instructions.id, cw.getId()));
                                instr.add(Utils.instruction(Instructions.ln, "wrap"));
                                closeTag = true;
                            }    
                        } else {
                            sb.append("<td class='ui-matrix-cell context-menu ui-select-handle lui2-unit");
                            String val;
                            val = cl.getCurrentProperties().get("halg");
                            if (val == null) val = "Center";
                            sb.append(" halign-");
                            sb.append(val);
                            propInstr.add(Utils.instruction(Instructions.b, cl.getId()));
                            propInstr.add(Utils.instruction(Instructions.halg_, val));
                            val = cl.getCurrentProperties().get("valg");
                            if (val == null) val = "Center";
                            sb.append(" valign-");
                            sb.append(val);
                            propInstr.add(Utils.instruction(Instructions.valg_, val));
                            val = cl.getCurrentProperties().get("ovrf");
                            if (val == null) val = "Expand";
                            sb.append(" overflow-");
                            sb.append(val);
                            propInstr.add(Utils.instruction(Instructions.ovrf_, val));
                            val = cl.getCurrentProperties().get("clss");
                            if (val != null) {
                                String v = getForm().parseMultiCss(val);
                                if (!Utils.testEmptyString(v)) {
                                    sb.append(" ");
                                    sb.append(v);
                                    propInstr.add(Utils.instruction(Instructions.clss_, v));
                                }
                            }    
                            sb.append("' id='");
                            sb.append(cl.getId());
                            sb.append("' row='");
                            sb.append(cl.getRowId());
                            sb.append("' column='");
                            sb.append(cl.getColumnId());
                            sb.append("' cell='");
                            sb.append(cl.getId());
                            sb.append("' notext='true'");
                            closeTag = true;
                        }
                        if (closeTag) {
                            if (cw.getHtmlProperties().getColSpan() > 0) {
                                sb.append(" colspan='");
                                sb.append(String.valueOf(cw.getHtmlProperties().getColSpan()));
                                sb.append("'");
                            }
                            if (cw.getHtmlProperties().getRowSpan() > 0) {
                                sb.append(" rowspan='");
                                sb.append(String.valueOf(cw.getHtmlProperties().getRowSpan()));
                                sb.append("'");
                            }
                            sb.append(">");
                            if (cl != null) {
                                sb.append("<div class='ui-matrix-cell-content lui2-unit'>");
                                sb.append("<div class='ui-matrix-cell-data lui2-unit'></div>");
                                sb.append("<div class='ui-matrix-cell-buttons lui2-unit'></div></div>");
                                instr.add(Utils.instruction(Instructions.b, getId()));
                                instr.add(Utils.instruction(Instructions.id, cl.getId()));
                                instr.add(Utils.instruction(Instructions.ln, "cell"));
                            }
                            sb.append("</td>");
                        }    
                    }
                    sb.append("</tr>");
                }
                sb.append("</tbody></table>");
                sb.append("</div>");

                instructions.add(Utils.instruction(Instructions.b, getForm().getId()));
                instructions.add(Utils.instruction(Instructions.cont, sb.toString()));
                instructions.add(Utils.instruction(Instructions.id, getId()));
                instructions.add(Utils.instruction(Instructions.ln, "matrix"));
                instructions.add(Utils.instruction(Instructions.b, getId()));
                instructions.add(Utils.instruction(Instructions.rids, rowsIds.toString()));
                instructions.add(Utils.instruction(Instructions.cids, columnsIds.toString()));
                instructions.add(Utils.instruction(Instructions.ids, cellsIds.toString()));
                for (int i = 0; i < instr.size(); i++) {
                    instructions.add(instr.get(i));
                }
                for (int i = 0; i < propInstr.size(); i++) {
                    instructions.add(propInstr.get(i));
                }
                for (int i = 0; i < asyncInstructions.size(); i++) {
                    instructions.add(asyncInstructions.get(i));
                }
                instructions.add(Utils.instruction(Instructions.b, getId()));
                instructions.add(Utils.instruction(Instructions.ini));
                initialized = true;
                setUid("");
                plain();
            }
        }
        
        public boolean isInitialized() {
            return initialized;
        }
        
        public void addInstruction(String id) {
            if (!id.equals(prevId)) {
                asyncInstructions.add(Utils.instruction(Instructions.b, id));
            }
            if (isInitialized()) {
                setUid(id);
            }
        }
        
        public void addInstruction(JsonObject instr) {
            asyncInstructions.add(instr);
            if (isInitialized()) {
                instructions.add(instr);
            }
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.mtrx;
        }
        
        @Override
        public void addChild(UiUnit child) {
            if (child instanceof UiMatrixCellset) {
                UiMatrixCellset cs = (UiMatrixCellset)child;
                super.addChild(cs);
                addCellset(cs);
                cs.setMatrix(this);
                refresh();
            } else if (child instanceof UiMatrixCell) {
                UiMatrixCell cell = (UiMatrixCell)child;
                if (rows.containsKey(cell.getRowCode()) && columns.containsKey(cell.getColumnCode())) {
                    cells.put(cellKey(cell), cell);
                    super.addChild(cell);
                    cell.setMatrix(this);
                    cell.refresh();
                    refresh();
                }
            }
        }
        
        public ArrayList<UiMatrixCell> getCells(UiMatrixCellset cs) {
            ArrayList<String> keys = new ArrayList<>();
            if (cs instanceof UiMatrixRow) {
                String row = cs.getCode();
                columns.keySet().forEach((key) -> {
                    keys.add(cellKey(row, key));
                });
            } else if (cs instanceof UiMatrixColumn) {
                String column = cs.getCode();
                rows.keySet().forEach((key) -> {
                    keys.add(cellKey(key, column));
                });
            }
            ArrayList<UiMatrixCell> ret = new ArrayList<>();
            for (int i = 0; i < keys.size(); i++) {
                UiMatrixCell cell = cells.get(keys.get(i));
                if (cell != null) {
                    ret.add(cell);
                }    
            }
            return ret;
        }
        
        public void addCellset(UiMatrixCellset cs) {
            if (cs instanceof UiMatrixRow) {
                rows.put(cs.getCode(), (UiMatrixRow)cs);
            } else if (cs instanceof UiMatrixColumn) {
                columns.put(cs.getCode(), (UiMatrixColumn)cs);
            }
            if (isInitialized()) {
                refresh();
            }
        }
        
        public void rmCell(UiMatrixCell cell) {
            cells.values().remove(cell);
        }
        
        public void rmCellset(UiMatrixCellset cs) {
            if (cs instanceof UiMatrixRow) {
                rows.remove(cs.getCode());
            } else if (cs instanceof UiMatrixColumn) {
                columns.remove(cs.getCode());
            }     
            /*if (rows.isEmpty() && columns.isEmpty()) {
                initialized = false;
                addonInstructions.add(Utils.instruction(Instructions.b, getId()));
                addonInstructions.add(Utils.instruction(Instructions._ini));
                unitsToCheck.add(this);
            }*/
        }
        
        public void setColumnInfo(String info) {
            if (!Utils.testEmptyString(info)) {
                try {
                    MatrixColumnInfo cinf = MatrixColumnInfo.valueOf(info);
                    if (cinf == MatrixColumnInfo.Y) cinf = MatrixColumnInfo.Bottom;
                    if (cinf != columnInfo) {
                        columnInfo = cinf;
                        addInstruction(getId());
                        addInstruction(Utils.instruction(Instructions.cnf, cinf.toString().toLowerCase()));
                        if (isInitialized()) {
                            plain();
                        }
                    }    
                    
                } catch (Exception ex) {}
            }
        }
        
        public void setRowInfo(String info) {
            if (!Utils.testEmptyString(info)) {
                try {
                    MatrixRowInfo rinf = MatrixRowInfo.valueOf(info);
                    if (rinf == MatrixRowInfo.Y) rinf = MatrixRowInfo.Right;
                    if (rinf != rowInfo) {
                        rowInfo = rinf;
                        addInstruction(getId());
                        addInstruction(Utils.instruction(Instructions.rnf, rinf.toString().toLowerCase()));
                        if (isInitialized()) {
                            plain();
                        }
                    }    
                } catch (Exception ex) {}
            }
        }
        
        public UiMatrixColumn getColumn(String code) {
            return columns.get(code);
        }
        
        public UiMatrixRow getRow(String code) {
            return rows.get(code);
        }
        
        public UiMatrixColumn checkColumn(String code) {
            return getColumn(code);
        }
        
        public UiMatrixRow checkRow(String code) {
            return getRow(code);
        }
        
        public UiMatrixCell getCell(String key) {
            return cells.get(key);
        }
        
        public UiMatrixCell getCell(String row, String column) {
            return getCell(cellKey(row, column));
        }
        
        public UiMatrixCell getCell(UiMatrixRow row, UiMatrixColumn column) {
            return getCell(row.getCode(), column.getCode());
        }
        
        private String cellKey(String row, String column) {
            return row + "." + column;
        }
        
        private String cellKey(UiMatrixCell cell) {
            return cellKey(cell.getRowCode(), cell.getColumnCode());
        }
        
        public void clear() {
            HashMap<String, UiMatrixCell> tmp = new HashMap<>();
            tmp.putAll(cells);
            for (Map.Entry<String, UiMatrixCell> entry : tmp.entrySet()) {
                entry.getValue().remove();
            }
            cells.clear();
            currentCell = null;
            selectedRows = "";
            selectedColumns = "";
            selectedCells = "";
        }
        
        public void setAngularText(String text) {
            if (text != null) {
                addInstruction(getId());
                addInstruction(Utils.instruction(Instructions.lbl, Utils.string2String(text)));
            }    
        }
        
        public void setAngularHint(String hint) {
            if (hint != null) {
                addInstruction(getId());
                if (Utils.testEmptyString(hint)) {
                    addInstruction(Utils.instruction(Instructions._hnt));
                } else {
                    addInstruction(Utils.instruction(Instructions.hnt, hint));
                }
            }    
        }
        
        public void setTitle(String t) {
            if (t != null) {
                addInstruction(getId());
                if (Utils.testEmptyString(t)) {
                    addInstruction(Utils.instruction(Instructions._ttl));
                } else {
                    addInstruction(Utils.instruction(Instructions.ttl, t));
                }    
            }    
        }
        
        public void setTitleBorder(String b) {
            if (b != null) {
                addInstruction(getId());
                if (Utils.testYString(b)) {
                    addInstruction(Utils.instruction(Instructions.brd));
                } else {
                    addInstruction(Utils.instruction(Instructions._brd));
                }
            }
        }
        
        public void setAngularWidth(String width) {
            if (!Utils.testEmptyString(width)) {
                addInstruction(getId());
                addInstruction(Utils.instruction(Instructions.aw, width));   
            } 
        }
        
        public void setSelectionMode(String sm) {
            if (!Utils.testEmptyString(sm)) {
                try {
                    MatrixSelectionMode mode = MatrixSelectionMode.valueOf(sm);
                    addInstruction(getId());
                    if (mode == MatrixSelectionMode.Standard) {
                        addInstruction(Utils.instruction(Instructions._sm)); 
                    } else {
                        addInstruction(Utils.instruction(Instructions.sm, mode.toString()));
                    }
                } catch (Exception ex) {}
            }
        }
        
        public void reqAngularWidth(String width) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Property", new StringParameter("AngularWidth"));
            sep.put("Value", new StringParameter(width));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqMenu() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("R-code", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("R-code", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        public void reqDblClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("R-code", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqMenuRC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("R-code", new StringParameter(""));
            sep.put("C", new StringParameter("0"));
            sep.put("R", new StringParameter("0"));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqClickRC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("R-code", new StringParameter(""));
            sep.put("C", new StringParameter("0"));
            sep.put("R", new StringParameter("0"));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqDCRC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("R-code", new StringParameter(""));
            sep.put("C", new StringParameter("0"));
            sep.put("R", new StringParameter("0"));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqBlankMenu(String row, String col) {
            UiUnit r = units.get(row);
            UiUnit c = units.get(col);
            if (!(r == null || c == null)) {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("MatrixRMB"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("C-code", new StringParameter(""));
                sep.put("R-code", new StringParameter(""));
                sep.put("C", new StringParameter(c.getCode()));
                sep.put("R", new StringParameter(r.getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public void reqBlankClick(String row, String col) {
            UiUnit r = units.get(row);
            UiUnit c = units.get(col);
            if (!(r == null || c == null)) {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("MatrixClick"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("C-code", new StringParameter(""));
                sep.put("R-code", new StringParameter(""));
                sep.put("C", new StringParameter(c.getCode()));
                sep.put("R", new StringParameter(r.getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        public void reqBlankDC(String row, String col) {
            UiUnit r = units.get(row);
            UiUnit c = units.get(col);
            if (!(r == null || c == null)) {
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("MatrixDC"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("C-code", new StringParameter(""));
                sep.put("R-code", new StringParameter(""));
                sep.put("C", new StringParameter(c.getCode()));
                sep.put("R", new StringParameter(r.getCode()));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
        private void updateProperties() {
            for (Map.Entry<String, UiMatrixRow> entry : rows.entrySet()) {
                UiMatrixRow row = entry.getValue();
                if (row.needUpdate()) {
                    row.updateProperties();
                }
            }
            for (Map.Entry<String, UiMatrixColumn> entry : columns.entrySet()) {
                UiMatrixColumn col = entry.getValue();
                if (col.needUpdate()) {
                    col.updateProperties();
                }
            }
            for (Map.Entry<String, UiMatrixCell> entry : cells.entrySet()) {
                UiMatrixCell cell = entry.getValue();
                if (cell.needUpdate()) {
                    cell.updateProperties();
                }
            }
        }
         
        @Override
        public void preCheck() {
            finish();
            updateProperties();
        }

        @Override
        public void postCheck() {
            addonInstructions.clear();
        }
        
        @Override
        public ArrayList<JsonObject> getInstructions() {
            return addonInstructions;
        }
        
        public void changeCell(UiMatrixCell cell) {
            //if (cell != currentCell) {
                currentCell = cell;
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("ChangeCell"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("R-code", new StringParameter(cell.getRowCode()));
                sep.put("C-code", new StringParameter(cell.getColumnCode()));
                totalEvent.add(new SingleEvent(sep));
            //}
        }
        
        public void reqSelect(String rows, String cols, String cells) {
            String pRows, pColumns, pCells;
            String[] a = StringUtils.split(rows, ",");
            ArrayList<String> r = new ArrayList<>();
            for (String s : a) {
                r.add(units.get(s).getCode());
            }
            Gson gson = new Gson();
            pRows = gson.toJson(r);
            a = StringUtils.split(cols, ",");
            r = new ArrayList<>();
            for (String s : a) {
                r.add(units.get(s).getCode());
            }
            gson = new Gson();
            pColumns = gson.toJson(r);
            a = StringUtils.split(cells, ",");
            ArrayList<MatrixCellRecord> rc = new ArrayList<>();
            for (String s : a) {
                UiMatrixCell cl = (UiMatrixCell)units.get(s);
                MatrixCellRecord mcr = new MatrixCellRecord(cl.getRowCode(), cl.getColumnCode());
                rc.add(mcr);
            }
            gson = new Gson();
            pCells = gson.toJson(rc);
            if (!(pRows.equals(selectedRows) && pColumns.equals(selectedColumns) && pCells.equals(selectedCells))) {
                selectedRows = pRows;
                selectedColumns = pColumns;
                selectedCells = pCells;
                SingleEventParameters sep = new SingleEventParameters();
                sep.put("e", new StringParameter("MatrixSelect"));
                sep.put("FormID", new StringParameter(getForm().getCode()));
                sep.put("Rows", new StringParameter(selectedRows));
                sep.put("Columns", new StringParameter(selectedColumns));
                sep.put("Cells", new StringParameter(selectedCells));
                totalEvent.add(new SingleEvent(sep));
            }
        }
        
    }

/************************************************************************
 *                                                                      *
 *     Класс UiMatrixChild - "дочерний элемент матрицы"                 *
 *                                                                      *
 ************************************************************************/
    
    public class UiMatrixChild extends UiFormChild {
        
        private boolean needUpdate = false;
        public final HashMap<String, String> properties;
        private String hint = "";
        private final HtmlProperties htmlProperties;
        private UiMatrix matrix = null;
        
        public UiMatrixChild() {
            super();
            properties = new HashMap<>();
            htmlProperties = new HtmlProperties();
        }
        
        public UiMatrixChild(String code) {
            super(code);
            properties = new HashMap<>();
            htmlProperties = new HtmlProperties();
        }
                
        public HashMap<String, String> getProperties() {
            return properties;
        }
        
        // Получить матрицу
        public UiMatrix getMatrixParent() {
            return new Object() {
                UiMatrix find(UiUnit u) {
                    if (u == null) {
                        return null;
                    } else if (u instanceof UiMatrix) {
                        return (UiMatrix) u;
                    } else {
                        return find(u.getParent());
                    }
                }
            }.find(this.getParent());
        }
        
        public void setMatrix(UiMatrix matrix) {
            this.matrix = matrix;
        }
        
        public UiMatrix getMatrix() {
            return matrix;
        }
        
        public HtmlProperties getHtmlProperties() {
            return htmlProperties;
        } 
        
        public void forUpdate() {
            needUpdate = true;
        }
        
        public boolean needUpdate() {
            return needUpdate;
        }
        
        public void updated() {
            needUpdate = false;
        }
        
        public void refresh() {
            forUpdate();
            if (getMatrix().isInitialized()) unitsToCheck.add(getMatrix());
        }
        
        public void setHint(String hint) {
            if (hint != null) {
                if (!this.hint.equals(hint)) {
                    this.hint = hint;
                    getMatrix().addInstruction(getId());
                    if (Utils.testEmptyString(hint)) {
                        getMatrix().addInstruction(Utils.instruction(Instructions._hnt));
                    } else {
                        getMatrix().addInstruction(Utils.instruction(Instructions.hnt, hint));
                    }
                }
            } 
        }
       
        public void setHAlign(String p) {
            if (p != null) {
                try {
                    HorizontalAlign h = HorizontalAlign.valueOf(p);
                    properties.put("halg", p);
                    refresh();
                } catch (Exception ex) {
                    
                }
            }
        }

        public void setVAlign(String p) {
            if (p != null) {
                try {
                    VerticalAlign v = VerticalAlign.valueOf(p);
                    properties.put("valg", p);
                    refresh();
                } catch (Exception ex) {
                    
                }
            }
        }
        
        public void setOverflow(String p) {
            if (p != null) {
                try {
                    Overflow o = Overflow.valueOf(p);
                    properties.put("ovrf", p);
                    refresh();
                } catch (Exception ex) {
                    
                }
            }
        }
        
        public void setClasses(String p) {
            if (p != null) {
                //String s = getForm().parseMultiCss(p, false);
                //properties.put("clss", s);
                properties.put("clss", p);
                refresh();
            }
        }
        
        public void setMakeup(String p) {
            if (p != null) {
                properties.put("mkup", p);
                refresh();
            }
        }
        
        public void setMUClasses(String p) {
            if (p != null) {
                String s = getForm().parseMultiCss(p);
                properties.put("mucs", s);
                refresh();
            }
        }

        public void updateProperties() {
            
        }
        
    }
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixRootCell - "угловая ячейка"                        *
 *                                                                      *
 ************************************************************************/      
    
    public class UiMatrixRootCell extends UiMatrixChild {    
        
    }
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixTitleCell - "заглавная ячейка"                     *
 *                                                                      *
 ************************************************************************/      
    
    public class UiMatrixTitleCell extends UiMatrixChild {    
        
    }      
    
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixInfoCell - "информационная ячейка"                 *
 *                                                                      *
 ************************************************************************/      
    
    public class UiMatrixInfoCell extends UiMatrixChild {    
        
        private final Defaults.InfoCellPosition position;
        private final UiMatrixCellset cellset;
        
        public UiMatrixInfoCell(UiMatrixCellset set, Defaults.InfoCellPosition position) {
            super();
            this.position = position;
            this.cellset = set;
        }
        
        public Defaults.InfoCellPosition getPosition() {
            return position;
        }
        
        public UiMatrixCellset getCellset() {
            return cellset;
        }
        
    }    
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixCellWrap - "контейнер для ячейки матрицы"          *
 *                                                                      *
 ************************************************************************/      
    
    public class UiMatrixCellWrap extends UiMatrixChild {
        
        private UiMatrixRow row;
        private UiMatrixColumn column;
        private UiMatrixCell cell;
        
        public UiMatrixCellWrap(UiMatrixRow row, UiMatrixColumn column) {
            super();
            this.row = row;
            this.column = column;
            cell = null;
        }
        
        public UiMatrixRow getRow() {
            return row;
        }
        
        public UiMatrixColumn getColumn() {
            return column;
        }
        
        public void setCell(UiMatrixCell cell) {
            this.cell = cell;
        }
        
        public UiMatrixCell getCell() {
            return cell;
        }
        
    }       
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixCellset - "набор ячеек матрицы"                    *
 *                                                                      *
 ************************************************************************/      
    
    public class UiMatrixCellset extends UiMatrixChild {
        
        private String caption = "";
        private String visible = "Y";
        private CaptionOrientation orientation = CaptionOrientation.H;
        private final HashMap<String, String> currentProperties;
        private final HashMap<String, String> effectiveProperties;
        private final ArrayList<UiUnit> parents;
        private UiMatrixCellset root = this;
        private ArrayList<UiMatrixCellset> leafs;
        private StringJoiner cellsIds;
        
        
        public UiMatrixCellset(String code) {
            super(code);
            parents = new ArrayList<>();
            currentProperties = new HashMap<>();
            effectiveProperties = new HashMap<>();
            init();
        }
        
        private void init() {
            leafs = new ArrayList<>();
            cellsIds = new StringJoiner(" ");
        }
        
        public void clearCellsIds() {
            cellsIds = new StringJoiner(" ");
        }
        
        public void addCellId(String id) {
            cellsIds.add(id);
        }
        
        public String getCellsIds() {
            return cellsIds.toString();
        }
        
        public UiMatrixCellset getRoot() {
            return root;
        }
        
        public void setRoot(UiMatrixCellset r) {
            root = r;
        }
        
        public void setLeafs(ArrayList<UiMatrixCellset> leafs) {
            this.leafs = leafs;
        }

        public ArrayList<UiMatrixCellset> getLeafs() {
            return leafs;
        }
        
        public HashMap<String, String> getCurrentProperties() {
            return currentProperties;
        }
        
        @Override
        public void preRemove() {
            ArrayList<UiMatrixCell> cells = getCells();
            for (int i = 0; i < cells.size(); i++) {
                cells.get(i).remove();
            }
            getMatrix().rmCellset(this);
        }
        
        @Override
        public void refresh() {
            getRoot().forUpdate();
            if (getMatrix().isInitialized()) unitsToCheck.add(getMatrix());
        }
        
        @Override
        public void addChild(UiUnit child) {
            super.addChild(child);
            if (child instanceof UiMatrixCellset) {
                UiMatrixCellset c = (UiMatrixCellset)child;
                c.setMatrix(getMatrix());
                getMatrix().addCellset(c);
                c.setRoot(this.getRoot());
                /*c.addParents(this.getParents());
                c.addParent(this);
                String p = "";
                for (UiUnit u : c.getParents()) {
                    p = p + u.getId() + " ";
                    ArrayList<UiUnit> ch = u.getAllChildren();
                    String cs = "";
                    for (UiUnit cu : ch) {
                        if (u instanceof UiMatrixCellset) {
                            cs = cs + cu.getId() + " ";
                        }
                    }
                    if (!Utils.testEmptyString(cs)) {
                        setUid(u.getId());
                        instructions.add(Utils.instruction(Instructions.chld, StringUtils.strip(cs)));
                    }
                }
                if (!Utils.testEmptyString(p)) {
                    setUid(c.getId());
                    instructions.add(Utils.instruction(Instructions.prnt, StringUtils.strip(p)));
                }*/
                refresh();
            }
        }
        
        public void setOrientation(String orient) {
            if (orient != null) {
                try {
                    CaptionOrientation o = CaptionOrientation.valueOf(orient);
                    if (o != orientation) {
                        orientation = o;
                        if (getMatrix().isInitialized()) {
                            setUid(getId());
                            instructions.add(Utils.instruction(Instructions.ornt, orient));
                        }  
                        //getMatrix().addInstruction(getId());
                        //getMatrix().addInstruction(Utils.instruction(Instructions.ornt, orient));
                    }
                } catch (Exception ex) {
                    
                }
            } 
        }
        
        public CaptionOrientation getOrientation() {
            return orientation;
        }
        
        public void setCaption(String caption) {
            if (caption != null) {
                if (!this.caption.equals(caption)) {
                    this.caption = caption;
                    getMatrix().addInstruction(getId());
                    if (Utils.testEmptyString(caption)) {
                        getMatrix().addInstruction(Utils.instruction(Instructions._ttl));
                    } else {
                        getMatrix().addInstruction(Utils.instruction(Instructions.ttl, caption));
                    }
                }
            } 
        }
         
        public void setVisible(String visible) {
            if (visible != null) {
                if (!this.visible.equals(visible)) {
                    this.visible = visible;
                    //setUid(getId());
                    if (Utils.testYString(visible)) {
                        show();
                        //instructions.add(Utils.instruction(Instructions.show));
                    } else {
                        hide();
                        //instructions.add(Utils.instruction(Instructions.hide));
                    }
                    getMatrix().refresh();
                }
            } 
        }
        
        public void setInfo(String info) {
            if (info != null) {
                getMatrix().addInstruction(getId());
                getMatrix().addInstruction(Utils.instruction(Instructions.minf, info));
            }
        }
        
        public void setSelfClasses(String p) {
            if (p != null) {
                properties.put("scls", p);
                refresh();
            }
        }
        
        public void addParent(UiUnit p) {
            parents.add(p);
        }
        
        public void addParents(ArrayList<UiUnit> p) {
            parents.addAll(p);
        }
        
        public ArrayList<UiUnit> getParents() {
            return parents;
        }
        
        public ArrayList<UiMatrixCell> getCells() {
            return getMatrix().getCells(this);
        }

        public HashMap<String, String> getEffectiveProperties() {
            return effectiveProperties;
        }

        private void calcProperties() {
            effectiveProperties.clear();
            UiUnit prnt = this.getParent();
            List<String> jpl = Arrays.asList(Defaults.MATRIX_JOIN_PROPERTIES);
            List<String> spl = Arrays.asList(Defaults.MATRIX_CELLSET_PROPERTIES);
            if (prnt != null && prnt instanceof UiMatrixCellset) {
                effectiveProperties.putAll(((UiMatrixCellset) prnt).getEffectiveProperties());
            }
            for (String p : Defaults.MATRIX_PROPERTIES) {
                if (getProperties().containsKey(p)) {
                    if (jpl.contains(p)) {
                        if (effectiveProperties.containsKey(p)) {
                            String s = effectiveProperties.get(p) + " " + getProperties().get(p);
                            if (StringUtils.contains(s, Defaults.MATRIX_RESET_CLASS)) {
                                s = StringUtils.substringAfterLast(s, Defaults.MATRIX_RESET_CLASS);
                            }
                            effectiveProperties.put(p, s);
                        } else {
                            effectiveProperties.put(p, getProperties().get(p));
                        }
                    } else {
                        effectiveProperties.put(p, getProperties().get(p));
                    }
                }
                String val = null;
                if (effectiveProperties.containsKey(p)) {
                    if (currentProperties.containsKey(p)) {
                        if (!effectiveProperties.get(p).equals(currentProperties.get(p))) {
                            val = effectiveProperties.get(p);
                        }
                    } else {
                        val = effectiveProperties.get(p);
                    }
                }
                if (val != null) {
                    currentProperties.put(p, val);
                    if (getMatrix().isInitialized()) {
                        if (spl.contains(p)) {
                            if (jpl.contains(p)) {
                                String v = getForm().parseMultiCss(val);
                                //if (!Utils.testEmptyString(v)) {
                                    setUid(getId());
                                    instructions.add(Utils.instruction(Instructions.valueOf(p), v));
                                //}
                            } else {
                                setUid(getId());
                                instructions.add(Utils.instruction(Instructions.valueOf(p), val));
                            }    
                        }    
                    }    
                }
            }
        }
        
        @Override
        public void updateProperties() {
            calcProperties();
            for (int i = 0; i < children.size(); i++) {
                if (children.get(i) instanceof UiMatrixCellset) {
                    ((UiMatrixCellset) children.get(i)).updateProperties();
                }
            }
            ArrayList<UiMatrixCell> cells = getCells();
            for (int i = 0; i < cells.size(); i++) {
                cells.get(i).forUpdate();
            }
            updated();
        }

        public void reqMenu() {
            
        }
        
        public void reqClick() {
            
        }
        
        public void reqDblClick() {
            
        }
                
    } 
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixColumn - "столбец матрицы"                         *
 *                                                                      *
 ************************************************************************/     
    
    public class UiMatrixColumn extends UiMatrixCellset {
        
        private final UiMatrixInfoCell topInfo = new UiMatrixInfoCell(this, Defaults.InfoCellPosition.Top);
        private final UiMatrixInfoCell bottomInfo = new UiMatrixInfoCell(this, Defaults.InfoCellPosition.Bottom);
        
        public UiMatrixColumn(String code) {
            super(code);
        }    
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.col;
        }
        
        public void setWidth(String width) {
            if (!Utils.testEmptyString(width)) {
                getMatrix().addInstruction(getId());
                getMatrix().addInstruction(Utils.instruction(Instructions.w, width));
            }
        }

        public UiUnit getTopInfo() {
            return topInfo;
        }

        public UiUnit getBottomInfo() {
            return bottomInfo;
        }
    
        public void reqPropertyWidth(String width) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("ColWidth"));
            sep.put("Value", new StringParameter(width));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void reqMenu() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(getCode()));
            sep.put("R-code", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void reqClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(getCode()));
            sep.put("R-code", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void reqDblClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(getCode()));
            sep.put("R-code", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
    }    
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixRow - "строка матрицы"                             *
 *                                                                      *
 ************************************************************************/     
    
    public class UiMatrixRow extends UiMatrixCellset {
        
        private UiMatrixInfoCell leftInfo = new UiMatrixInfoCell(this, Defaults.InfoCellPosition.Left);
        private UiMatrixInfoCell rightInfo = new UiMatrixInfoCell(this, Defaults.InfoCellPosition.Right);
        
        public UiMatrixRow(String code) {
            super(code);
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.row;
        }
        
        public void setHeight(String height) {
            if (!Utils.testEmptyString(height)) {
                getMatrix().addInstruction(getId());
                getMatrix().addInstruction(Utils.instruction(Instructions.h, height));
            }
        }
        
        public void setLeftInfo(UiMatrixInfoCell info) {
            leftInfo = info;
        }

        public void setRightInfo(UiMatrixInfoCell info) {
            rightInfo = info;
        }

        public UiUnit getLeftInfo() {
            return leftInfo;
        }

        public UiUnit getRightInfo() {
            return rightInfo;
        }
        
        public void reqPropertyHeight(String height) {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("SetUserProperty"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Element", new StringParameter(getCode()));
            sep.put("Property", new StringParameter("RowHeight"));
            sep.put("Value", new StringParameter(height));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void reqMenu() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("R-code", new StringParameter(getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void reqClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixClick"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("R-code", new StringParameter(getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        @Override
        public void reqDblClick() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("R-code", new StringParameter(getCode()));
            sep.put("C-code", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            sep.put("R", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
       
    }
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixCell - "ячейка матрицы"                            *
 *                                                                      *
 ************************************************************************/     
    
    public class UiMatrixCell extends UiMatrixChild {
        
        private String editable = "N";
        private String value = "";
        private final UiMatrixRow row;
        private final UiMatrixColumn column;
        private final HashMap<String, String> effectiveProperties;
        private final HashMap<String, String> currentProperties;
        
        public UiMatrixCell(UiMatrixRow row, UiMatrixColumn column) {
            super();
            this.row = row;
            this.column = column;
            effectiveProperties = new HashMap<>();
            currentProperties = new HashMap<>();
            init();   
        }
        
        private void init() {
            /*setUid(row.getId());
            instructions.add(Utils.instruction(Instructions.col, column.getId()));
            instructions.add(Utils.instruction(Instructions.cell, getId()));*/
        }
        
        @Override
        public Instructions getCreateInstruction() {
            return Instructions.cell;
        }
        
        @Override
        public void preRemove() {
            getMatrix().rmCell(this);
        }
        
        public String getRowCode() {
            return row.getCode();
        }
        
        public String getColumnCode() {
            return column.getCode();
        }
        
        public String getRowId() {
            return row.getId();
        }
        
        public String getColumnId() {
            return column.getId();
        }
        
        public HashMap<String, String> getCurrentProperties() {
            return currentProperties;
        }
        
        public void setEditable(String editable) {
            if (editable != null) {
                if (!this.editable.equals(editable)) {
                    this.editable = editable;
                    getMatrix().addInstruction(getId());
                    if (Utils.testYString(editable)) {
                        getMatrix().addInstruction(Utils.instruction(Instructions.edit));
                    } else {
                        getMatrix().addInstruction(Utils.instruction(Instructions._edt));
                    }
                }
            } 
        }
        
        public void setValue(String value) {
            if (value != null) {
                assignValue(value);
                getMatrix().addInstruction(getId());
                getMatrix().addInstruction(Utils.instruction(Instructions.v, value));
            }
        }
        
        @Override
        public void updateProperties() {
            effectiveProperties.clear();
            effectiveProperties.putAll(row.getEffectiveProperties());
            List<String> jpl = Arrays.asList(Defaults.MATRIX_JOIN_PROPERTIES);
            List<String> cpl = Arrays.asList(Defaults.MATRIX_CELL_PROPERTIES);
            for (String p : Defaults.MATRIX_PROPERTIES) {
                if (column.getEffectiveProperties().containsKey(p)) {
                    if (jpl.contains(p)) {
                        if (effectiveProperties.containsKey(p)) {
                            String s = effectiveProperties.get(p) + " " + column.getEffectiveProperties().get(p);
                            if (StringUtils.contains(s, Defaults.MATRIX_RESET_CLASS)) {
                                s = StringUtils.substringAfterLast(s, Defaults.MATRIX_RESET_CLASS);
                            }
                            effectiveProperties.put(p, s);
                        } else {
                            effectiveProperties.put(p, column.getEffectiveProperties().get(p));
                        }
                    } else {
                        effectiveProperties.put(p, column.getEffectiveProperties().get(p));
                    }
                }
            }    
            for (String p : Defaults.MATRIX_PROPERTIES) {    
                if (getProperties().containsKey(p)) {
                    if (jpl.contains(p)) {
                        if (effectiveProperties.containsKey(p)) {
                            String s = effectiveProperties.get(p) + " " + getProperties().get(p);
                            if (StringUtils.contains(s, Defaults.MATRIX_RESET_CLASS)) {
                                s = StringUtils.substringAfterLast(s, Defaults.MATRIX_RESET_CLASS);
                            }
                            effectiveProperties.put(p, s);
                        } else {
                            effectiveProperties.put(p, getProperties().get(p));
                        }
                    } else {
                        effectiveProperties.put(p, getProperties().get(p));
                    }
                }
                String val = null;
                if (effectiveProperties.containsKey(p)) {
                    if (currentProperties.containsKey(p)) {
                        if (!effectiveProperties.get(p).equals(currentProperties.get(p))) {
                            val = effectiveProperties.get(p);
                        }
                    } else {
                        val = effectiveProperties.get(p);
                    }
                }
                if (val != null) {
                    currentProperties.put(p, val);
                    if (getMatrix().isInitialized()) {
                        if (cpl.contains(p)) {
                            if (jpl.contains(p)) {
                                String v = getForm().parseMultiCss(val);
                                //if (!Utils.testEmptyString(v)) {
                                    setUid(getId());
                                    instructions.add(Utils.instruction(Instructions.valueOf(p), v));
                                //}
                            } else {
                                setUid(getId());
                                instructions.add(Utils.instruction(Instructions.valueOf(p), val));
                            }    
                        }    
                    }    
                }
            }
            updated();
        }
        
        public void assignValue(String value) {
            this.value = value;
        }
        
        public String getValue() {
            return value;
        }
        
        public void reqChange() {
            getMatrix().changeCell(this);
        }
        
        public void reqMenu() {
            reqChange();
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixRMB"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("R-code", new StringParameter(row.getCode()));
            sep.put("C-code", new StringParameter(column.getCode()));
            sep.put("R", new StringParameter(""));
            sep.put("C", new StringParameter(""));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqDC() {
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("MatrixDC"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("R-code", new StringParameter(row.getCode()));
            sep.put("C-code", new StringParameter(column.getCode()));
            sep.put("R", new StringParameter(row.getCode()));
            sep.put("C", new StringParameter(column.getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
        public void reqNewValue(String val) {
            assignValue(val);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NewValue"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(column.getCode()));
            sep.put("R-code", new StringParameter(row.getCode()));
            sep.put("Value", new StringParameter(getValue()));
            sep.put("IsCheck", new StringParameter("Y"));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        public void reqInterValue(String val) {
            assignValue(val);
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("NewValue"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("Field", new StringParameter(getParent().getCode()));
            sep.put("Value", new StringParameter(getValue()));
            sep.put("IsCheck", new StringParameter("N"));
            totalEvent.add(new SingleEvent(sep));  
        }
        
        /*
        @Override
        public void preCheck() {
            row.getRoot().updateProperties();
            column.getRoot().updateProperties();
        }
        */
        
    }
    
/************************************************************************
 *                                                                      *
 *     Класс UiMatrixButton - "кнопка в ячейке матрицы"                 *
 *                                                                      *
 ************************************************************************/    
    
    public class UiMatrixButton extends UiBaseButton {
        
        public UiMatrixButton(String code) {
            super(code);
        }
        
        @Override
        public void reqPress() {
            UiMatrixCell cell = (UiMatrixCell)this.getParent();
            cell.reqChange();
            SingleEventParameters sep = new SingleEventParameters();
            sep.put("e", new StringParameter("CellButton"));
            sep.put("FormID", new StringParameter(getForm().getCode()));
            sep.put("C-code", new StringParameter(cell.getColumnCode()));
            sep.put("R-code", new StringParameter(cell.getRowCode()));
            sep.put("Name", new StringParameter(getCode()));
            totalEvent.add(new SingleEvent(sep));
        }
        
    }       
      
    
    
/*******************************************************************/    
    public void startTest(String app) {
        testing = true;
        //stopMonitoring = true;
        aliveCounter = 0;
        String appCode = "";
        try {
            Random rnd = new Random();
            eventsLog = new ArrayList<>();
            if (!Utils.testEmptyString(app)) {
                appCode = "AppCode:" + app + ";";
            }
            eventsLog.add("<?xml version=\"1.0\" encoding=\"UTF-8\"?><Events><E N=\"Open\"><P N=\"Params\">" + appCode + "</P><P N=\"ClientID\">" + token + "</P><P N=\"MsgPort\">" + msgPort + "</P><P N=\"UserLang\">ru-RU</P><P N=\"URL\">stresstest.lui</P><P N=\"HostName\">" + remoteHost + "</P></E></Events>");
            eventsLog.addAll(Files.readAllLines(Paths.get(tempDir, "luievents.log"), StandardCharsets.UTF_8));
            int i = 0;
            while (testing && i < eventsLog.size()) {
                instructions = new ArrayList<>();
                totalEvent = new TotalEvent();
                StringBuilder sb = new StringBuilder();
                sockWriter.write(eventsLog.get(i));
                sockWriter.write("\n");
                sockWriter.flush();
                String line;
                while (testing && ((line = sockReader.readLine()) != null)) {
                    if (line.equals("")) {
                        break;
                    } else {
                        sb.append(line.replaceAll(Defaults.LF_REPLACE, "\n"));
                    }
                }
                if (testing) {
                    processCommands(sb.toString());
                    Thread.currentThread().sleep((rnd.nextInt(9) + 1) * 200);
                }
                //System.out.println("Event > " + eventsLog.get(i));
                //System.out.println("Command > " + sb.toString());
                i++;
            }
        } catch (Exception ex) {
            System.out.println("Test > " + token + " failed with message:");
            ex.printStackTrace();
        } finally {
            stopTest();
        }
    }
    
    public void stopTest() {
        System.out.println("Stop > " + this.getToken());
        testing = false;
        aliveCounter = 2;
        stopMonitoring = false;
    }
    
    
}
